module adcheck
  use constants, only : zero, one
  implicit none
  private

  interface ADcheckIce
    module procedure check_ice
    module procedure print_ice
  end interface
  private :: check_ice, print_ice

  public  :: ADcheck2, ADcheckIce, STlimit

contains

  !=============================================================================

subroutine ADcheck2 (mm, n2d, kh, dt, u, v, w, dx, dy,                         &
                     hs, hx, hy, ndtp, lmn, ind, area, ompidx)
!-------------------------------------------------------------------------------
!  Refined stability check for tracer advection scheme based on the principle:
!  "You cannot cut a bald-head".
!
!  Initiated Jan. 28, 2006, by Per Berg DMI/COI
!-------------------------------------------------------------------------------

  !- modules:
  use dmi_mpi_global, only : iu06
  use dmi_mpi,        only : mpi_rank
  use dmi_omp,        only : domp_get_thread_no, domp_get_domain
  use io_subs,        only : flush_unit

  !- directives:
  implicit none

  !- arguments:
  integer(4), intent(in)    :: n2d, lmn, area
  integer(4), intent(in)    :: mm(:,0:,0:), kh(0:), ind(:,:)
  real(8), intent(in)       :: u(0:), v(0:), w(0:)
  real(8), intent(in)       :: dx(:), dy, dt
  real(8), intent(in)       :: hs(0:), hx(0:), hy(0:)
  integer(4), intent(inout) :: ndtp(:)
  integer(4), intent(inout) :: ompidx(:,:)

  !- locals:
  integer(4) :: idxsrf, idx, m, n, k, kb, isrfl, isrfu, tnum, ndt
  real(8)    :: maxu, maxv, maxw, maxcfl, tout, ddx, ddy

  !- initialise:
  call domp_get_domain(kh, 1, n2d, isrfl, isrfu, ompidx)
  call domp_get_thread_no (tnum)
  ndt = 1
  ddy = one/dy

  !- do the stability check (k-loop unrolled):
  maxcfl = zero
  do idxsrf=isrfl,isrfu
    m = ind(1,idxsrf)
    n = ind(2,idxsrf)

    ddx = one/dx(m)
    idx = idxsrf
    kb  = kh(idxsrf)

    !- do the surface layer:
    maxu = max(zero,hx(idx)*u(idx)) - min(zero,hx(mm(1,m,n-1))*u(mm(1,m,n-1)))
    maxv =-min(zero,hy(idx)*v(idx)) + max(zero,hy(mm(1,m-1,n))*v(mm(1,m-1,n)))
    tout = maxu*ddx + maxv*ddy   
    if (kb > 1) tout = tout - min(zero,w(mm(2,m,n)))
    maxcfl = max( dt*tout/hs(idx), maxcfl )

    !- do all layers, except for top and bottom:
    do k=2,kb-1
      idx  = mm(k,m,n)
      maxu = max(zero,hx(idx)*u(idx))-min(zero,hx(mm(k,m,n-1))*u(mm(k,m,n-1)))
      maxv =-min(zero,hy(idx)*v(idx))+max(zero,hy(mm(k,m-1,n))*v(mm(k,m-1,n)))
      maxw = max(zero,w(idx))        -min(zero,w(mm(k+1,m,n)))
      tout = maxu*ddx + maxv*ddy + maxw
      maxcfl = max( dt*tout/hs(idx), maxcfl )
    enddo

    !- do the bottom layer:
    if (kb < 2) cycle
    idx  = mm(kb,m,n)
    maxu = max(zero,hx(idx)*u(idx))-min(zero,hx(mm(kb,m,n-1))*u(mm(kb,m,n-1)))
    maxv =-min(zero,hy(idx)*v(idx))+max(zero,hy(mm(kb,m-1,n))*v(mm(kb,m-1,n)))
    maxw = max(zero,w(idx))
    tout = maxu*ddx + maxv*ddy + maxw
    maxcfl = max( dt*tout/hs(idx), maxcfl )

  enddo


  !- Do we need to issue a warning?
  ndt = int(maxcfl) + 1
  if (ndt > 1) then
    write(iu06,*) 'Warning: stability criteria exceeded', ndt
    write(iu06,*) ' in area:', area
    write(iu06,*) ' thread No:', tnum,', task No:',mpi_rank+1 
    call flush_unit(iu06)
  endif


  !- very critical alert:
  if (ndt > lmn) then
    write(iu06,*) 'RED ALERT: Time step is too large!!!'
    call flush_unit(iu06)
  endif


  !- transfer to output array:
  ndtp(tnum) = ndt

end subroutine ADcheck2

  !=============================================================================

  subroutine check_ice (me, n2d, dt, u, v, dx, dy, ndtp, ind)
  !-----------------------------------------------------------------------------
  !  Stability check for ice advection scheme based on a simple cfl criterion.
  !-----------------------------------------------------------------------------

    !- modules:
    use dmi_omp, only : domp_get_thread_no, domp_get_domain

    !- directives:
    implicit none

    !- arguments:
    integer(4), intent(in)  :: n2d
    integer(4), intent(in)  :: me(0:,0:), ind(:,:)
    real(8),    intent(in)  :: u(0:), v(0:)
    real(8),    intent(in)  :: dx(:), dy, dt
    integer(4), intent(out) :: ndtp(:)

    !- locals:
    integer(4) :: n, nl, nu, tnum, i, j
    real(8)    :: maxu, maxv, maxcfl, tout, ddx, ddy

    !- initialize:
    call domp_get_domain(1, n2d, nl, nu)
    call domp_get_thread_no (tnum)
    ddy = one/dy

    !- do the stability check:
    maxcfl = zero
    do n=nl,nu
      i = ind(1,n)
      j = ind(2,n)

      ddx = one/dx(i)

      maxu = max(zero,u(n)) - min(zero,u(me(i,j-1)))
      maxv =-min(zero,v(n)) + max(zero,v(me(i-1,j)))
      tout = dt*(maxu*ddx + maxv*ddy)
      maxcfl = max( tout, maxcfl )
    enddo

    !- transfer to output array:
    ndtp(tnum) = int(maxcfl) + 1

  end subroutine check_ice

  !=============================================================================

  subroutine print_ice (ndtp, lmn, area)
  !-----------------------------------------------------------------------------
  !  Print stability violations
  !-----------------------------------------------------------------------------

    !- modules:
    use dmi_mpi_global, only : iu06
    use dmi_mpi,        only : mpi_rank
    use io_subs,        only : flush_unit
    use cmod_params,    only : nthreads

    !- directives:
    implicit none

    !- arguments:
    integer(4), intent(in) :: lmn, area
    integer(4), intent(in) :: ndtp(:)

    !- locals:
    integer(4) :: tnum, ndt

    !- do the warning:
    !    requires a previous call of check_ice to fill array ndtp(:)
    do tnum=1,nthreads
      ndt = ndtp(tnum)

      !- Do we need to issue a warning?
      if (ndt > 1) then
        write(iu06,*) 'Warning: stability criteria exceeded, ice', ndt
        write(iu06,*) ' in area:', area
        write(iu06,*) ' thread No:', tnum,', task No:',mpi_rank+1 
        call flush_unit(iu06)
      endif

      !- very critical alert:
      if (ndt > lmn) then
        write(iu06,*) 'RED ALERT: Time step is too large in ice advection!!!'
        call flush_unit(iu06)
      endif
    enddo

  end subroutine print_ice

  !=============================================================================

subroutine STlimit( n2d, kh, ind, mk, cmp, ompidx )

  !- modules -------------------------------------------------------------------
  use dmi_omp,   only : domp_get_domain
  use constants, only : cfreeze, saltmax, wtmpmax

  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in)    :: n2d, kh(0:), ind(:,:), mk(:,0:,0:)
  real(8),    intent(inout) :: cmp(1:,0:)
  integer(4), intent(inout) :: ompidx(:,:)

  !- local vars ----------------------------------------------------------------
  integer(4) :: n, nl, nu, i, j, mi, kb
  real(8)    :: tfreeze

  call domp_get_domain(kh, 1, n2d, nl, nu, ompidx)

  do n = nl, nu
    cmp(2,n)  = min( max(zero, cmp(2,n)), saltmax)
    tfreeze  = -cfreeze*cmp(2,n)
    cmp(1,n) = min( max(tfreeze, cmp(1,n)), wtmpmax)
  enddo
  do n = nl, nu
    kb = kh(n)
    if (kb <= 1) cycle
    i = ind(1,n)
    j = ind(2,n)
    do mi=mk(2,i,j),mk(2,i,j)+kb-2
      cmp(2,mi) = min( max(zero, cmp(2,mi)), saltmax)
      tfreeze   = -cfreeze*cmp(2,mi)
      cmp(1,mi) = min( max(tfreeze, cmp(1,mi)), wtmpmax)
    enddo
  enddo

end subroutine STlimit

end module adcheck
