module assimilation

  implicit none
  private

  public  :: assimilate
  private :: kalmanfilter, readsat, modify_freeze

  real(8), parameter, private :: undefined = 99999.0_8, eps = 0.001_8

contains

  !=============================================================================

  subroutine kalmanfilter(tim,m,n,k,iw2_l,mmk_l,mksrf,z,hz,avt,ts,             &
                          asst,obsvar,lastasstime,errvar,t_var_sat,iu06,nc)

  !-----------------------------------------------------------------------------
  !**
  !** Local variables:
  !** ----------------
  !**
  !** Adaptation of James Annan routine for data assimilation.
  !** Modifications:
  !** --------------
  !**    - update of the code to the new f90 version (IAB, march 2003)
  !**    - saturation of the model error variance
  !**    - adaption to cmod (jl@dmi.dk).
  !**    - tons of optimizations (per@dmi.dk).
  !
  !     Common declarations for the assimilation routines.
  !
  ! Variables declared:
  ! -------------------
  ! rbias_(day/night) : biases for data to be assimilated (day/night) time 
  ! data (negative if data has too low values)
  ! avt_thresh: prescribed value associated to upper limit for thermocline
  !         (criterion based on eddy diffusivity) [m^2 s^-1]
  ! secshour: seconds in an hour
  ! hoursday: hours in a day
  !
  ! Authors: I. Andreu-Burillo, March 2003.
  !          Jesper Larsen, December 2003.
  !          Jesper Larsen, November 2005.
  !-----------------------------------------------------------------------------
    !
    use constants, only : ldebug
    !
    implicit none
    !
    ! 1.a.  Argument declarations
    integer(4), intent(in) :: m, n, k, iu06, nc
    real(8), intent(in)    :: tim, t_var_sat

    !- global arrays:
    real(8), intent(inout) :: lastasstime(0:), errvar(0:)
    real(8), intent(in)    :: asst(:), obsvar(:)
    integer(4), intent(in) :: mksrf(:,0:,0:)

    !- task-local arrays:
    real(8), intent(inout) :: ts(1:,0:)
    real(8), intent(in)    :: avt(1:,0:), hz(0:), z(0:)
    integer(4), intent(in) :: iw2_l, mmk_l(:,0:,0:)
  

    ! 1.b.  Local declarations
    real(8), parameter :: rbias_night=0.0
    !real(8), parameter :: rbias_day=0.0, rbias_night=0.0
    ! real(8), parameter :: rbias_day=0.0, rbias_night=-0.12
    ! real(8), parameter :: rbias_day=0.0, rbias_night=-0.27
    ! real(8), parameter :: bvf2_thresh=1.e-2
  
    real(8), parameter :: avt_thresh=1.e-3_8, zero = 0.0_8, one = 1.0_8
    real(8), parameter :: magic = 1.0_8*0.0625_8*20.0_8
    real(8)      :: dep, dd, rbias, innovation !, bvf2(0:iw3)
    integer(4)   :: i, j, kk, ilayer, mi, mis, mil, ml
    !character(5) :: cpass
  
    !     Variables used for diagnostic output statistics 
    !     -----------------------------------------------
    real(8)    :: tmodave, tobsave, dtmodave, dzave, told, asstsfc,            &
                  sst, dtmodobsave, rmsbefore, rmsafter, errvarbefore,         &
                  errvarafter, obsvsfc, maxdiffold, maxdiff, sstatmaxdiffold,  &
                  sstatmaxdiff, dasstold, dasst
    integer(4) :: cnt, imxold, jmxold, imx, jmx
    logical    :: asstdef
  
    !     Don't do assimilation if there are no surface wet points
    if (iw2_l == 0) return

    !     Initialise diagnostic variables
    !     -------------------------------
    tmodave      = zero
    tobsave      = zero
    dtmodave     = zero
    dzave        = zero
    dtmodobsave  = zero
    cnt          = 0
    rmsbefore    = zero
    rmsafter     = zero
    errvarbefore = zero
    errvarafter  = zero
    if (ldebug) then
      maxdiffold      = -undefined
      maxdiff         = -undefined
      imxold          = -1
      jmxold          = -1
      imx             = -1
      jmx             = -1
      sstatmaxdiffold = -undefined
      sstatmaxdiff    = -undefined
      dasstold        = -undefined
      dasst           = -undefined
    endif
    !
    ! 2. Check if it is a day or night pass (currently not used)
    ! ----------------
    !      if (ihour < 10.or.ihour >= 22) then
    !cpass(1:5) = 'night'
    !      else
    !        cpass(1:3) = 'day'
    !      endif
    !
    !if (cpass(1:1) == 'n') then
      rbias = rbias_night
    !elseif (cpass(1:1) == 'd') then
    !  rbias = rbias_day
    !endif
    !
    ! 3. Analysis step
    ! ----------------
    !      call bruntvaisala(m,n,k,iw2,iw3,z,hz,rho,mksrf,bvf2)
    do j = 1,n
      do i = 1,m
        ! surface index:
        mil = mmk_l(1,i,j)

        if (0 < mil .and. mil <= iw2_l) then
          ! some surface vars:
          mis     = mksrf(1,i,j)
          obsvsfc = obsvar(mis)
          asstsfc = asst(mis)
          asstdef = (abs(asstsfc-undefined) > eps)
          told    = ts(nc,mil)
          !
          ! 3.1. calculate the assimilation depth ...
          ! -----------------------------------------
          dep    = z(mil) + hz(mil)
          ilayer = 1
          do kk = 2,k
            ml = mmk_l(kk,i,j)
            if (ml /= 0 .and. avt(min(3,nc),ml) > avt_thresh) then
            ! if (mi /= 0 .and. bvf2(mi) < bvf2_thresh) then
              dep    = dep + hz(ml)
              ilayer = kk
            else
              exit
            endif
          enddo
          !
          ! 3.2. Compute forecast error variance prior to assimilation
          !      (introducing a saturation forecast error-variance)
          ! ----------------------------------------------------------
          dd = magic*(tim-lastasstime(mis))/dep
          do kk=1,k
            mi = mksrf(kk,i,j)
            if (mi == 0) exit
            errvar(mi) = min( errvar(mi) + dd, t_var_sat )
          enddo
          errvarbefore = errvarbefore + errvar(mis)
          !
          if (asstdef) then
            !
            ! 3.3. Compute innovation using uppermost model layer
            ! ---------------------------------------------------
            innovation = (asstsfc - rbias) - ts(nc,mil)
            !
            ! 3.4. Compute Kalman gain and produce analysis down the ML
            ! ---------------------------------------------------------
            do kk = 1,ilayer
              ml = mmk_l(kk,i,j)
              mi = mksrf(kk,i,j)
              ts(nc,ml) = ts(nc,ml) + innovation*errvar(mi)/(obsvsfc+errvar(mi))
            enddo
            !
            ! 3.5. Update errvar to new improved value
            ! ----------------------------------------
            do kk = 1,ilayer
              mi = mksrf(kk,i,j)
              errvar(mi) = one/(one/errvar(mi) + one/obsvsfc)
            enddo
            errvarafter = errvarafter + errvar(mis)
            !
            ! 3.6. Updating lastasstime
            ! -------------------------
            lastasstime(mis) = tim
            !
            ! 3.7. Calculate statististics
            ! ----------------------------
            sst         = ts(nc,mil)
            dzave       = dzave       + dep
            tmodave     = tmodave     + sst
            tobsave     = tobsave     + asstsfc - rbias
            dtmodave    = dtmodave    + sst - told
            dtmodobsave = dtmodobsave + innovation

            cnt       = cnt + 1
            rmsbefore = rmsbefore + (told - asstsfc)**2
            rmsafter  = rmsafter  + (sst  - asstsfc)**2
            if (ldebug) then
              if (abs(told - asstsfc) > maxdiffold) then
                maxdiffold      = abs(told - asstsfc)
                sstatmaxdiffold = told
                dasstold        = asstsfc
                imxold          = i
                jmxold          = j
              endif
              if (abs(sst - asstsfc) > maxdiff) then
                maxdiff      = abs(sst - asstsfc)
                sstatmaxdiff = sst
                dasst        = asstsfc
                imx          = i
                jmx          = j
              endif
            endif
          endif
  
        endif
      enddo
    enddo
  
    !
    ! 4. Finish statistics and write it out
    ! ------------------------------------
    write(iu06,'(a49,f14.8)')                                                  &
                      'Assimilation time:                               ',tim
    write(iu06,'(a49,i5)')                                                     &
                      'Number of observations:                          ',cnt
    if (cnt > 1) then
      dzave        = dzave / cnt
      tmodave      = tmodave / cnt
      tobsave      = tobsave / cnt
      dtmodobsave  = dtmodobsave / cnt
      errvarbefore = errvarbefore / (cnt-1)
      errvarafter  = errvarafter / (cnt-1)
      rmsbefore    = sqrt(rmsbefore / cnt)
      rmsafter     = sqrt(rmsafter / cnt)
      write(iu06,'(a49,f7.4)')                                                 &
        'Average t_mod at assimilation points:            ', tmodave
      write(iu06,'(a49,f7.4)')                                                 &
        'Average t_obs at assimilation points:            ', tobsave
      write(iu06,'(a49,f7.4)')                                                 &
        'Average t_mod change at assimilation points:     ', dtmodave/cnt
      write(iu06,'(a49,f7.4)')                                                 &
        'Average t_mod change overall surface:            ', dtmodave/iw2_l
      write(iu06,'(a49,f7.4)')                                                 &
        'Average innovation vector at assimilation points:', dtmodobsave
      write(iu06,'(a49,f7.4)')                                                 &
        'Average error variance before assimilation:      ', errvarbefore
      write(iu06,'(a49,f7.4)')                                                 &
        'Average error variance after assimilation:       ', errvarafter
      write(iu06,'(a49,f7.4)')                                                 &
        'RMS Error before assimilation:                   ', rmsbefore
      write(iu06,'(a49,f7.4)')                                                 &
        'RMS Error after assimilation:                    ', rmsafter
      if (ldebug) then
        write(iu06,'(a49,f7.4)')                                               &
          'Max deviation before assimilation, value:        ', maxdiffold
        write(iu06,'(a49,2i5)')                                                &
          'Max deviation before assimilation, position:     ', imxold, jmxold
        write(iu06,'(a49,f7.4)')                                               &
          'SST before assimilation at max deviation:        ', sstatmaxdiffold
        write(iu06,'(a49,f7.4)')                                               &
          'SST obs at max deviation position, before:       ', dasstold
        write(iu06,'(a49,f7.4)')                                               &
          'Max deviation after assimilation, value:         ', maxdiff
        write(iu06,'(a49,2i5)')                                                &
          'Max deviation after assimilation, position:      ', imx, jmx
        write(iu06,'(a49,f7.4)')                                               &
          'SST after assimilation at max deviation:         ', sstatmaxdiff
        write(iu06,'(a49,f7.4)')                                               &
          'SST obs at max deviation position, after:        ', dasst
      endif
    endif

  end subroutine kalmanfilter
  
  !=============================================================================

  subroutine assimilate(minut,nsecs,tim,m,n,k,iw2_l,mmk_l,mksrf,z,hz,avt,ts,   &
                        asst,obserr,lastasstime,errvar,xkoor,ykoor,tmin,tmax,  &
                        t_var_sat,iu06,nc)
  !*****************************************************************************
  !
  ! Performs Annan-Hargreaves data assimilation
  !
  ! Jesper Larsen, September 2006.
  !
  !*****************************************************************************

    use io_subs, only : flush_unit
  
    implicit none
  
    integer(4), intent(in) :: minut, nsecs, m, n, k, iu06, nc
    real(8), intent(inout) :: lastasstime(0:), errvar(0:)
    integer(4), intent(in) :: mksrf(:,0:,0:)
    real(8), intent(out)   :: asst(:), obserr(:)
    real(8), intent(in)    :: xkoor(:), ykoor(:)
    real(8), intent(in)    :: tmin, tmax, t_var_sat, tim

    !- task-local arrays:
    integer(4), intent(in) :: iw2_l, mmk_l(:,0:,0:)
    real(8), intent(inout) :: ts(1:,0:)
    real(8), intent(in)    :: avt(1:,0:), hz(0:), z(0:)

    logical :: lfile
  
    if (minut /= 0 .or. nsecs /= 0) return

    ! Read in optimally interpolated satellite data field and error
    call readsat(tim,m,n,mksrf,asst,obserr,lfile,xkoor,ykoor,tmin,tmax,iu06)
  
    ! Assimilate data
    if (lfile) then
      if (nc == 1) call modify_freeze( m,n,iw2_l,mmk_l,mksrf,ts,asst )
      call kalmanfilter(tim,m,n,k,iw2_l,mmk_l,mksrf,z,hz,avt,ts,               &
                        asst,obserr,lastasstime,errvar,t_var_sat,iu06,nc)
    endif

    ! Flush the printouts:
    call flush_unit(iu06)
  
  end subroutine assimilate
  
  !=============================================================================

  subroutine readsat(tim,m,n,m1,asst,obserr,lfile,xkoor,ykoor,tmin,tmax,iu06)

  !-----------------------------------------------------------------------------
  ! Reads in optimally interpolated data data and error variances.
  !
  ! Arguments:
  ! ----------
  ! m,n:      model grid dimensions
  ! iday:     day  
  ! imon:     month
  ! iyyy:     year
  ! ihour:    hour
  ! asst:     satellite values interpolated to model grid
  ! obserr:   estimated satellite sst error interpolated to model grid
  ! lfile:    true if file is found
  !
  ! Jesper Larsen, 29/11-2005
  !-----------------------------------------------------------------------------

    use constants, only : grarad
    use io_subs,   only : io_new_unit
    use cmod_date, only : caltime
  
    implicit none
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: m, n, iu06
    integer(4), intent(in) :: m1(0:,0:,:)
    real(8), intent(in)    :: tim
    real(8), intent(out)   :: asst(:)
    real(8), intent(out)   :: obserr(:)
    logical, intent(out)   :: lfile
    real(8), intent(in)    :: xkoor(:)
    real(8), intent(in)    :: ykoor(:)
    real(8), intent(in)    :: tmin, tmax
  
    !- local vars --------------------------------------------------------------
    real(8), parameter :: inundefined=99.
  
    real(8), dimension(:,:), allocatable :: rawdata, rawerr
  
    integer(4)        :: iday, imon, iyyy, ihour, minut, nsecs
    integer(4)        :: i,j,lun,is,js, ios, mi
    real(8)           :: trest
    real(8)           :: xmin, xmax, ymin, ymax, dx, dy, dummy
    real(8)           :: x0c, y0c, dxc, dyc
    integer(4)        :: nx, ny
    character(60)     :: sdate
    character(len=19) :: ctim
    character(len=4)  :: cy
    character(len=2)  :: cm,cd,ch
    character(len=60) :: fnam
    character(len=10) :: cfmt
    !NOTE: Intenitionally, dummy,sdate,xmax,ymin are only read in but never used
  
    ! Read data from file ------------------------------------------------------

    ! Initialize on every entry:
    lfile     = .false.
    asst(:)   = undefined
    obserr(:) = undefined
  
    call caltime(tim,iday,imon,iyyy,ihour,minut,nsecs,trest,ctim)
    write(cy,'(i4.4)') iyyy
    write(cm,'(i2.2)') imon
    write(cd,'(i2.2)') iday
    write(ch,'(i2.2)') ihour
  
    ! Check if files exist
    fnam = 'err_'//cy//cm//cd//ch//'.asc'
    inquire(file=trim(fnam),exist=lfile)
    if (.not. lfile) then
      write(iu06,'(a)') 'Warning: Cannot find '//trim(fnam)
      write(iu06,'(a)') '         ... continuing without data assimilation'
      return
    endif
    fnam = 'oi_'//cy//cm//cd//ch//'.asc'
    inquire(file=trim(fnam),exist=lfile)
    if (.not. lfile) then
      write(iu06,'(a)') 'Warning: Cannot find '//trim(fnam)
      write(iu06,'(a)') '         ... continuing without data assimilation'
      return
    else
      lfile = .true.
    endif
  
    ! Read in data estimates
    lun = io_new_unit()
    open(lun,file=fnam,iostat=ios)
    if (ios /= 0) then
      write(iu06,'(a)') 'Warning: Cannot open '//trim(fnam)
      write(iu06,'(a)') '         ... continuing without data assimilation'
      lfile = .false.
      return
    endif
    read(lun,*) xmin, xmax, ymin, ymax, dummy, dummy, dx, dy
    read(lun,*) ny, nx, dummy
    read(lun,*) sdate
    write(cfmt(2:5),'(i4.4)') nx
    cfmt = "(" // cfmt(2:5) // "f6.2)"
        
    allocate(rawdata(ny, nx))
    do i = ny,1,-1
      read(lun,trim(cfmt),iostat=ios) (rawdata(i,j),j = 1,nx)
      if (ios /= 0) exit
    enddo
    close(lun)
    if (ios /= 0) then
      write(iu06,'(a,i5)') ' Warning: Cannot read '//trim(fnam)//' at ',i
      write(iu06,'(a)')    '         ... continuing without data assimilation'
      lfile = .false.
      return
    endif
  
    ! Read in error variance estimates
    fnam = 'err_'//cy//cm//cd//ch//'.asc'
    lun = io_new_unit()
    open(lun,file=fnam,iostat=ios)
    if (ios /= 0) then
      write(iu06,'(a)') 'Warning: Cannot open '//trim(fnam)
      write(iu06,'(a)') '         ... continuing without data assimilation'
      lfile = .false.
      return
    endif
    read(lun,*) xmin, xmax, ymin, ymax, dummy, dummy, dx, dy
    read(lun,*) ny, nx, dummy
    read(lun,*) sdate
    write(cfmt(2:5),'(i4.4)') nx
    cfmt = "(" // cfmt(2:5) // "f6.2)"
        
    allocate(rawerr(ny, nx))
    do i = ny,1,-1
      read(lun,trim(cfmt),iostat=ios) (rawerr(i,j),j = 1,nx)
      if (ios /= 0) exit
    enddo
    close(lun)
    if (ios /= 0) then
      write(iu06,'(a,i5)') ' Warning: Cannot read '//trim(fnam)//' at ',i
      write(iu06,'(a)')    '         ... continuing without data assimilation'
      lfile = .false.
      return
    endif
  
    ! Interpolate to model grid (nearest neighbour)
    !  - fields were initialized as "undefined" at entry to this sub
    ! FIXME: nearest neighbour interpolation makes assumptions about indexing
    ! order in surface layer.
    x0c = xkoor(1)/grarad
    y0c = ykoor(1)/grarad
    dxc = (xkoor(n-1)-xkoor(1))/grarad/(n-2)
    dyc = (ykoor(m-1)-ykoor(1))/grarad/(m-2)
    do j = 1,n-1
      do i = 1,m-1
        mi = m1(i,j,1)
        if (mi /= 0) then
          is = int((ymax - y0c - dyc*(i-1)) / dy + 1)
          js = int((x0c + dxc*(j-1) - xmin) / dx + 1)
          if (abs(rawdata(is,js)-inundefined) > eps) then
            asst(mi)   = rawdata(is,js)
            obserr(mi) = rawerr(is,js)
          endif
        endif
      enddo
    enddo
  
    where (asst <= tmin .or. asst > tmax) asst = undefined
  
    if (lfile) then
      write(iu06,'(a5,a,a2)') 'Read:',trim(fnam),':T'
    else
      write(iu06,'(a5,a,a2)') 'Read:',trim(fnam),':F'
    endif
  
  end subroutine readsat

  !=============================================================================

  subroutine modify_freeze( m,n,iw2_l,mmk_l,mksrf,ts,asst )
  !-----------------------------------------------------------------------------
  !
  ! modify the assimilation temperature field so that we do not do assimilation
  ! in points which corresponds to:
  !
  !  a:  freezing assimilation temperature
  !
  !  b:  freezing model ocean surface temperature
  ! 
  !-----------------------------------------------------------------------------
    use constants, only : cfreeze
    implicit none

    !- global stuff:
    integer(4), intent(in) :: m, n
    real(8), intent(inout) :: asst(:)
    integer(4), intent(in) :: mksrf(:,0:,0:)

    !- task-local things:
    real(8), intent(in)    :: ts(1:,0:)
    integer(4), intent(in) :: iw2_l, mmk_l(:,0:,0:)

    integer(4) :: i, j, mil, mis
    real(8)    :: tfreeze, asstsfc, wsfctmp

    do j = 1,n
      do i = 1,m
        ! surface index:
        mil = mmk_l(1,i,j)

        if (0 < mil .and. mil <= iw2_l) then
          ! some surface vars:
          mis     = mksrf(1,i,j)
          asstsfc = asst(mis)
          wsfctmp = ts(1,mil)
          tfreeze = -cfreeze*ts(2,mil)
          if (abs(asstsfc-undefined) > eps) then
            if ( (asstsfc <= tfreeze) .or. (wsfctmp <= tfreeze) ) then
              asst(mis) = undefined
            endif            
          endif
        endif
      enddo
    enddo

  end subroutine modify_freeze

end module assimilation
