module coupler
  implicit none
  private

  ! private scalars:
  integer(4), private, save :: coupler_tracers = 2
  integer(4), private, save :: coupler_NDT     = 2
  real(8),    private, save :: coupler_dt      = 60.0_8
  real(8),    private, save :: coupler_outdt   = 3600.0_8
  logical,    private, save :: coupler_init_sp = .false.

  integer(4), private, save :: coupler_offset  = 2
  integer(4), private, save :: olun            = 0
  integer(4), private, save :: itime           = 1
  integer(4), private, save :: otime           = 1
  integer(4), private, save :: out_nt          = 60

  logical,       parameter, private :: coupler_mpi     = .true.
  character(13), parameter, private :: coupler_output = 'tracers_stats'
  real(8),       parameter, private :: zero = 0.0_8, one = 1.0_8, two = 2.0_8
  real(8),       parameter, private :: three = 3.0_8
  integer(4),    parameter, private :: io_min=7, io_max=130

  logical, private :: file_is_open

  ! public methods:
  public :: coupler_usage, coupler_init, coupler_dynamics, coupler_finish

  ! private methods:
  ! none

contains

  !-----------------------------------------------------------------------------

  subroutine coupler_usage(iu06,nt,nt2d, irc)
    implicit none
    integer(4), intent(in)  :: iu06
    integer(4), intent(out) :: nt   ! number of tracers to be inc. in cmp array 
    integer(4), intent(out) :: nt2d ! number of 2d fields
    integer(4), intent(out) :: irc  ! return code; 0 means 'no errors'

    integer(4) :: lun, ios
    character(5), parameter :: w = '(a57)'

    namelist /coupleroptions/ coupler_tracers, coupler_NDT, coupler_dt,        &
                              coupler_outdt, coupler_init_sp

    ! read specs from a namlelist:
    file_is_open = .true.
    lun = io_min-1
    do while (file_is_open)
      lun = lun+1
      if (lun > io_max) then
        write(iu06,*) 'ERROR: Could not find new I/O unit number'
        irc = 1 
        return 
      endif
      inquire(unit=lun,opened=file_is_open)
    enddo
    open (unit=lun, file='coupler.nml', status='old', iostat=ios)
    if (ios /= 0) then
      write(iu06,w)  'Warning: Cannot open file coupler.nml ...                '
      write(iu06,w)  '         ... continue with default options.              '
      write(iu06,'(a25,i5)') 'I/O problem - iostat is: ',ios
    else
      read (unit=lun, nml=coupleroptions, iostat=ios)
      if (ios /= 0) then
        write(iu06,w)'Warning: Cannot read coupleroptions in coupler.nml file  '
        write(iu06,w)'         ... continue with default coupler settings.     '
        write(iu06,'(a25,i5)') 'I/O problem - iostat is: ',ios
      endif
      close (unit=lun)
    endif

    ! transfer characteristics of this sub model to cmod:
    nt            = coupler_tracers
    nt2d          = 0

    ! FIXME consider sub. the * below with some proper format string
    write(iu06,*) 'coupler_usage succesfully called ...'
    write(iu06,*) '  coupler_tracers:', coupler_tracers
    write(iu06,*) '  coupler_NDT:    ', coupler_NDT
    write(iu06,*) '  coupler_dt:     ', coupler_dt
    write(iu06,*) '  coupler_outdt:  ', coupler_outdt
    write(iu06,*) '  coupler_output: ', coupler_output
    write(iu06,*) '  coupler_init_sp:', coupler_init_sp

    ! some private vars:
    out_nt = coupler_NDT*coupler_outdt/coupler_dt

    ! return code:
    if (coupler_init_sp) then
      irc = 0
    else
      irc = 1
    endif
    if (coupler_tracers <= 0) irc = -1

  end subroutine coupler_usage

  !-----------------------------------------------------------------------------

  subroutine coupler_init(ia, narea, iu06, mainarea, ncoff,                    &
                          mmx, nmx, kmx, nq, nz, mmk, kh, hz,                  &
                          components, rho, bndq, bndzk, on_IO_task, irc )
    !
    !Called from main as:
    !
    !if (mpi_io_serial) then
    !  do ia=1,narea
    !    call coupler_init(ia, narea, iu06, (ia==mainarea), ncoff,             &
    !                      mmx(ia),nmx(ia),kmx(ia),nq(ia),nz(ia),mm1k(ia)%p,   &
    !                      cmp(ia)%p, bndq(ia)%p, bndzk, irc )
    !    if (irc /= 0) call exitme(irc, 'coupler_init failed')
    !  enddo
    !endif
    !do ia=1,narea
    !  ! scatter only the tracers to task-local arrays (don't touch S and T):
    !  call dmpi_scatter(ia,iw3(ia),mm1k_l(ia)%p,cmp_l(ia)%p,kh_l(ia)%p,       &
    !                    mm1k(ia)%p,cmp(ia)%p,(ntracers-2),(2),mpi_io_rank)
    !  ! a haloswap is generally needed here:
    !  call dmpi_halo(kmx(ia),mm1k_l(ia)%p,ia,cmp_l(ia)%p,ntracers)
    !enddo
    !
    implicit none

    ! arguments ----------------------------------------------------------------
    integer(4), intent(in)    :: ia, narea, iu06, mmx, nmx, kmx, nq, nz, ncoff
    logical,    intent(in)    :: mainarea
    integer(4), intent(in)    :: mmk(:,0:,0:), kh(0:)
    real(8),    intent(in)    :: hz(0:), rho(0:)
    real(8),    intent(inout) :: components(:,0:)
    real(8),    intent(inout) :: bndq(:,:,0:), bndzk(:,:,0:)
    logical,    intent(in)    :: on_IO_task
    integer(4), intent(out)   :: irc

    ! local vars ---------------------------------------------------------------
    integer(4)                :: nct, i, ios, nlow, nup
    logical, save             :: FirstEntry = .true.


    ! return code (assume no errors): ------------------------------------------
    irc = 0

    ! some short-hand notations ------------------------------------------------
    nct  = coupler_tracers
    nlow = ncoff + 1
    nup  = ncoff + nct

    ! set river loadings -------------------------------------------------------
    do i=1,nct/2
      bndq(i+ncoff,1:kmx,0:nq) = real(i,8) 
    enddo
    bndq(nct/2+1+ncoff:nct+ncoff,1:kmx,0:nq) = three

    if (.not.on_IO_task) then
      write(iu06,'(a49,i4)')'... done coupler module initialisations for area '&
                            ,ia
      return
    endif

    ! on first entry, open a unit for output -----------------------------------
    if (FirstEntry) then
      FirstEntry = .false.

      write(iu06,'(a33,i4,a8)') 'Initialising coupler module with ',           &
                                nct,' tracers'

      coupler_offset = ncoff
      write(iu06,'(a33,i4)')    '                          offset ', ncoff

      file_is_open = .true.
      i = io_min-1
      do while (file_is_open)
        i = i+1
        if (i > io_max) then
          write(iu06,*) 'COUPLER ERROR: Could not find new I/O unit number'
          irc = 1
          return
        endif
        inquire (unit=i,opened=file_is_open)
      enddo
      olun = i

      open(olun,file=coupler_output,status='replace',iostat=ios)
      if (ios /= 0) then
        irc = ios
        write(iu06,*) 'COUPLER ERROR: Could not open '//coupler_output
        return
      endif

      if (nct < 1) then
        irc = 1
        write(iu06,*) 'COUPLER ERROR: Invalid No. of tracers:', nct
        return
      endif

      if (ncoff < 0) then  ! allowing for S/T, but be careful ...
        if (ncoff < 2) then
          write(iu06,*) 'COUPLER WARNING: This submodel accesses S/T.'
        endif
        irc = 1
        write(iu06,*) 'COUPLER ERROR: Invalid tracer offset No.:', ncoff
        return
      endif
    endif    

    ! print log logfile --------------------------------------------------------
    write(iu06,'(a20,i4)') '  Initialising area ',ia


    ! allocate local module arrays (if any) ------------------------------------
    if (ia == 1) then
      ! none
    endif


    ! initialize the tracers ---------------------------------------------------
    do i=1,nct/2
      components(i+ncoff,0:) = real(i,8)
    enddo
    components(nct/2+1+ncoff:nct+ncoff,0:) = one


    ! set boundary values ------------------------------------------------------
    if (mainarea) then 
      do i=1,nct/2
        bndzk(i+ncoff,1:kmx,0:nz) = real(i,8)
      enddo
      bndzk(nct/2+1+ncoff:nct+ncoff,1:kmx,0:nz) = two
    endif


    ! That's it! ---------------------------------------------------------------
    write(iu06,'(a49,i4)')'... done coupler module initialisations for area ',ia
 
  end subroutine coupler_init

  !-----------------------------------------------------------------------------

  subroutine coupler_dynamics(ia, iu06, mmx, nmx, kmx, mmk, kh, hz,            &
                              components, dispv, rho, u, v, z, casus,          &
                              atemp, humid, cloud, wu, wv, pl, precip,         &
                              update_counters, idx, irc)
  !
  !Called from main something like this:
  !
  !
  ! ! mpi coupler:
  ! do ia=1,narea
  !   call coupler_dynamics(ia, iu06, ..., cmp_l(ia)%p, ..., (ia==narea), irc)
  !   if (irc /= 0) call exitme(irc, 'coupler_dynamics failed')
  ! enddo
  ! halo_swap( cmp )
  !

    implicit none

    ! arguments ----------------------------------------------------------------
    integer(4), intent(in)    :: ia, iu06, mmx, nmx, kmx, mmk(:,0:,0:), kh(0:)
    real(8),    intent(in)    :: hz(0:)
    real(8),    intent(in)    :: dispv(:,0:)
    real(8),    intent(inout) :: components(:,0:)
    real(8),    intent(in)    :: u(0:), v(0:), z(0:), rho(0:)
    logical,    intent(in)    :: casus(0:)
    real(8),    intent(in)    :: atemp(0:), humid(0:), cloud(0:)
    real(8),    intent(in)    :: wu(0:), wv(0:), pl(0:), precip(0:)
    logical,    intent(in)    :: update_counters
    integer(4), intent(inout) :: idx(1:,1:)
    integer(4), intent(out)   :: irc


    ! local vars ---------------------------------------------------------------
    ! none
    

    ! return code (assume no errors) -------------------------------------------
    irc = 0


    ! do the sub model dynamics ------------------------------------------------
    if (mod(itime,coupler_NDT) == 0) then
      write(iu06,'(a28)') '... doing sub model dynamics'
    endif


    ! do the sub model outout --------------------------------------------------
    if (mod(otime,out_nt) == 0) then
      write(iu06,'(a26)') '... doing sub model output'
    endif


    ! update time step counters ------------------------------------------------
    if (update_counters) then
      if (mod(itime,coupler_NDT) == 0) then
        itime = 1
      else
        itime = itime + 1 
      endif  
      if (mod(otime,out_nt) == 0) then
        otime = 1
      else
        otime = otime + 1 
      endif  
    endif

  end subroutine coupler_dynamics

  !-----------------------------------------------------------------------------

  subroutine coupler_finish(iu06, irc)
    implicit none

    integer(4), intent(in)   :: iu06
    integer(4), intent(out)  :: irc

    ! return code (assume no errors):
    irc = 0

    ! close the units
    close (olun)

  end subroutine coupler_finish

  !-----------------------------------------------------------------------------
  ! Below here goes all the sub-model private subroutines:
  ! These guys could also be placed in seperate files in case this is desired, 
  ! cf. EXTRA_SOURCES in Makefile.am
  !-----------------------------------------------------------------------------

end module coupler
