module biobenthic
  implicit none

  !---------------------------------------------------------
  ! This module contains the Benthic biogeochemical submodel
  ! part of the ecological model used in CMOD at DMI
  !---------------------------------------------------------

  private

  public :: bio_benthic  

  real(8), parameter, private :: zero = 0.0_8, one = 1.0_8

contains

  subroutine bio_benthic(eco,bottemp,ben,deriv,benderiv,rfr,hbot,velkb,kb,irc)
    use bioparam, only : dn_sed,dn_sed_anox,q10_recs, NOR, idx_oxy, idx_phos,  &
                         idx_amm, idx_nit, ONamup, ONnitr, ksdenit, NNdenit,   &
                         wdz, fnitr, pfrac, Kspburial, bidx_nitr, bidx_sil,    &
                         ntben 
    implicit none

    real(8),    intent(in)    :: eco(:,0:), ben(0:), rfr, hbot
    real(8),    intent(in)    :: bottemp, velkb
    real(8),    intent(inout) :: deriv(0:), benderiv(0:)
    integer,    intent(in)    :: kb
    integer,    intent(inout) :: irc

    !-----------------
    ! local variables
    !-----------------
    real(8), parameter :: max_exp     = 0.4_8
    real(8), parameter :: burialcoeff = 5.0e-6_8
    real(8), parameter :: half        = 0.5_8

    integer(4) :: iben
    real(8)    :: ldn_N,ldn_O,denitscal

    ! kinetic parameters
    real(8) :: recs, Kspburial2

    ! process rates
    real(8) :: nitrification, pburial, burial(0:ntben), mineralization(0:ntben)

    ! other quantities
    real(8) :: bot_nit, bot_oxy, bot_pho, bot_oxy2
    real(8) :: dt_amm, dt_nit, dt_oxy, dt_pho
    real(8) :: exparg

    !- init some arrays where dummy index 0 can be used
    mineralization(0) = zero
    
    ! avoid unnecessary array access:
    bot_nit  = eco(kb,idx_nit)
    bot_oxy  = eco(kb,idx_oxy)
    bot_pho  = eco(kb,idx_phos)

    dt_amm  = deriv(idx_amm)
    dt_nit  = deriv(idx_nit)
    dt_pho  = deriv(idx_phos)
    dt_oxy  = deriv(idx_oxy)

    ! set up some numbers
    Kspburial2 = Kspburial*Kspburial
    bot_oxy2   = bot_oxy*bot_oxy

    !- permanent burial of benthos when the concentration is very large
    do iben=1,ntben
      exparg = ben(iben)*ben(iben)*burialcoeff
      if (exparg > max_exp) then
        burial(iben) = half*ben(iben)
      else
        burial(iben) = (exp(exparg)-one)*ben(iben)
      endif
    enddo

    ! Mineralized detritus becomes ammonium and nitrified under oxic conditions
    ! The produced nitrate is assumed to be used in coupled denitrification
    ! Thus no nitrate is produced but N2 which is lost from the system
    ! 
    ! Under oxic conditions benthic OM is oxidated by oxygen.
    ! Under anoxic conditions benthic OM is oxidated by nitrate, if available,
    ! otherwise by sulfate as negative o2
    ldn_N = zero
    ldn_O = ONamup

    !fixme: denitrification in oxic conditions not same as Neumann
    if (bot_oxy <= zero) then
      !-----------------
      ! Anoxic conditions
      !-----------------

      ! using nitrate (and suphalte) to oxidize benthic organic material
      denitscal = bot_nit/(ksdenit+bot_nit) 
      ldn_N     = NNdenit * denitscal
      ldn_O     = ONamup * (one - denitscal)

      recs           = dn_sed_anox*exp(q10_recs*bottemp) ! d^-1
      mineralization(bidx_nitr) = recs*ben(bidx_nitr) ! mmol N *m^-2*d^-1
      mineralization(bidx_sil)  = recs*ben(bidx_sil) ! mmol N *m^-2*d^-1
      
      ! No nitrification or P burial under anoxic conditions
      nitrification  = zero
      pburial        = zero
    else
      !----------------
      ! Oxic conditions
      !----------------
      
      recs                      = dn_sed*exp(q10_recs*bottemp) ! d^-1
      mineralization(bidx_nitr) = recs*ben(bidx_nitr) ! mmol N *m^-2*d^-1
      mineralization(bidx_sil)  = recs*ben(bidx_sil)  ! mmol N *m^-2*d^-1
      nitrification             = fnitr * mineralization(bidx_nitr)

      ! Phosphate is buried under oxic conditions
      pburial        = pfrac*bot_oxy2*(bot_oxy2 + Kspburial2)
    endif

    ! Net fluxes
    deriv(idx_amm)  = dt_amm + ( mineralization(bidx_nitr) - nitrification)/hbot
    deriv(idx_nit)  = dt_nit + (-ldn_N*mineralization(bidx_nitr))/hbot
    deriv(idx_phos) = dt_pho + (rfr*mineralization(bidx_nitr)*(one-pburial))   &
                               /hbot

    deriv(idx_oxy)  = dt_oxy - (ldn_O*mineralization(bidx_nitr) +              &
                               ONnitr*nitrification)*NOR/hbot

    do iben=1,ntben
      benderiv(iben) = -mineralization(iben)-burial(iben)
    enddo
  end subroutine bio_benthic

end module biobenthic
