module biobenthic
  implicit none

  !---------------------------------------------------------
  ! This module contains the Benthic biogeochemical submodel
  ! part of the ecological model used in CMOD at DMU
  !---------------------------------------------------------

  private

  public :: bio_benthic  

  real(8), parameter, private :: zero = 0.0_8, one = 1.0_8

contains

  subroutine bio_benthic(bwconc,bwtemp,seddetritus,dt,hbot)

    use bioparam, only : dn_sed,dn_sed_anox,q10_recs, NOR, idx_oxy,            &
                         idx_phos, idx_amm, idx_nit, idx_det, idx_sil, ONamup, &
                         ONnitr,Ksdenit, NNdenit, wdz, fnitr, rfr, pfrac,      &
                         Kspburial, rfr_s
    implicit none

    real(8),    intent(inout) :: bwconc(:)  ! (1:ntbio)
    real(8),    intent(in)    :: bwtemp
    real(8),    intent(inout) :: seddetritus
    real(8),    intent(in)    :: dt,hbot

    !-----------------
    ! local variables
    !-----------------

    real(8) :: ldn_N,ldn_O,nit_mod

    ! kinetic parameters
    real(8) :: recs, Ksdenit2, Kspburial2

    ! process rates
    real(8) :: mineralization,deposition,nitrification,pburial

    ! rates of change
    real(8) :: dammonium,dnitrate,dphosphate,doxygen,ddet,dseddet,dsilicate

    ! other quantities
    real(8) :: dth, bwc_nit, bwc_oxy, bwc_det, bwc_pho,bwc_sil

    !-----------------
    ! local params
    !-----------------
    real(8), parameter :: silscale = 0.4_8 

    ! avoid unnecessary array access:
    bwc_nit = bwconc(idx_nit)
    bwc_oxy = bwconc(idx_oxy)
    bwc_det = bwconc(idx_det)
    bwc_pho = bwconc(idx_phos)
    bwc_sil = bwconc(idx_sil)


    ! set up some numbers
    Ksdenit2   = Ksdenit*Ksdenit
    Kspburial2 = Kspburial*Kspburial

    ! Deposition of sinking organic matter 
    deposition = -wdz*bwc_det  ! mmolN * m^-2 * d^-1

    ! Recycling of benthic organic matter produces ammonium and phosphate,
    !   and consumes oxygen or nitrate

    if (bwc_oxy <= zero .and. bwc_nit > zero ) then
      nit_mod = one/(one + Ksdenit2/bwc_nit**2)
      ldn_N   = NNdenit * nit_mod
      ldn_O   = ONamup * (one - nit_mod)
    else
      ldn_N   = zero
      ldn_O   = ONamup
    endif
    ! Mineralized detritus becomes ammonium and nitrified under oxic conditions
    ! The produced nitrate is assumed to be used in coupled denitrification
    ! Thus no nitrate is produced but N2 which is lost from the system
    if (bwc_oxy <= zero) then
      ! Anoxic conditions
      recs = dn_sed_anox*exp(q10_recs*bwtemp) ! d^-1
      mineralization = recs*seddetritus ! mmol N *m^-2*d^-1
      ! No nitrification or P burial under anoxic conditions
      nitrification = zero
      pburial = zero
    else
      ! Oxic conditions
      recs = dn_sed*exp(q10_recs*bwtemp) ! d^-1
      mineralization = recs*seddetritus ! mmol N *m^-2*d^-1
      nitrification  = fnitr * mineralization
      ! Phosphate is buried under oxic conditions
      pburial = pfrac/(one + Kspburial2/bwc_oxy**2)
    endif


    ! Mineralized detritus becomes ammonium and nitrified under oxic conditions
    ! The produced nitrate is assumed to be used in coupled denitrification
    ! Thus no nitrate is produced but N2 which is lost from the system

    ! Net fluxes
    dammonium  =  mineralization - nitrification
    dnitrate   = -ldn_N*mineralization
    dphosphate = rfr*mineralization*(one-pburial)
    ddet       = -deposition
    doxygen    = NOR*(-ldn_O*mineralization -ONnitr*nitrification)
    dseddet    = deposition - mineralization
    dsilicate  = rfr_s*mineralization*(one- pburial)*silscale

    ! Integration
    dth = dt/hbot
    bwconc(idx_amm)  = bwconc(idx_amm) + dth*dammonium
    bwconc(idx_nit)  = bwc_nit         + dth*dnitrate
    bwconc(idx_phos) = bwc_pho         + dth*dphosphate
    bwconc(idx_det)  = bwc_det         + dth*ddet
    bwconc(idx_oxy)  = bwc_oxy         + dth*doxygen
    bwconc(idx_sil)  = bwc_sil         + dth*dsilicate
    seddetritus      = seddetritus     + dt *dseddet

  end subroutine bio_benthic
  
end module biobenthic
