module biofish

  implicit none
  private

  integer(4), private, save :: lunfish

  public :: fishinit, fishread, fishnull, fishclose, fishfeeding

contains

  !=============================================================================

  subroutine fishinit(narea, fishn, fisho)
    !--------------------------------------------------------------------------!
    ! Open the fish-forcings file
    !--------------------------------------------------------------------------!

    use exits,          only : exitme
    use io_subs,        only : io_new_unit
    use dmi_mpi_global, only : iu06
    use cmod_miscvars,  only : mdatei,tima
    use cmod_mem,       only : cmr1
   
    implicit none

    ! arguments
    integer(4),  intent(in)    :: narea
    type (cmr1), intent(inout) :: fishn(:),fisho(:)

    ! local vars
    integer(4)    :: ios
    character(11) :: fdatei
    real(8)       :: tfish

    fdatei  = 'F'//trim(mdatei(2:11))
    lunfish = io_new_unit()
    open(lunfish,file=fdatei,form='unformatted',status='old',iostat=ios)
    if (ios /= 0) call exitme(1,'Could not open '//fdatei)

    read(lunfish,iostat=ios) fdatei
    if (ios /= 0) call exitme(1,'Could not read or read past eof, fish file')
    write(iu06,*) 'fdatei=',fdatei

    ! read the first time
    tfish = tima
    
    do while (tfish <= tima)
      call fishread(tfish,narea,fishn)
      if (tfish == tima) then
        !- save old fish fields:
        call fishsave(narea, fishn, fisho )
      endif
    enddo

  end subroutine fishinit

  !=============================================================================

  subroutine fishread(tfish,narea,fishn)
    !---------------------------------------------------------------------------
    ! read the fish-forcings
    !---------------------------------------------------------------------------

    use exits,     only : exitme
    use cmod_mem,  only : cmr1
    use cmod_date, only : jultimea
    use constants, only : ldebug
    use dmi_mpi_global, only : iu06

    implicit none

    ! arguments
    real(8),      intent(out) :: tfish
    integer(4),   intent(in)  :: narea
    type (cmr1),  intent(inout) :: fishn(:)

    ! local vars
    integer(4)    :: ia, ios
    character(19) :: cfish

    read(lunfish,iostat=ios) cfish
    if (ios /= 0) call exitme(1,'Fish file truncated')

    call jultimea (cfish,tfish)

    write(iu06,*) 'Fish pressure read, valid at: ',cfish

    read(lunfish,iostat=ios) (fishn(ia)%p, ia=1,narea)
    if (ios /= 0) call exitme(1,'Could not read or read past eof, fish file')
        
    !- Optionally write out debug information ----------------------------------
    if (ldebug) then
      write(iu06,*) 'Fish forcing min and max values (area, min, max)'
      do ia = 1,narea
        write(iu06,*) 'fish ',ia,minval(fishn(ia)%p(1:)),maxval(fishn(ia)%p(1:))
      enddo
      write(iu06,*) '------------------------------'
    endif
  end subroutine fishread
  
  !=============================================================================

  subroutine fishsave(narea, fishn, fish )
    !---------------------------------------------------------------------------
    ! save old fish-forcings 
    !---------------------------------------------------------------------------

    use cmod_mem,  only : cmr1

    implicit none

    integer(4),  intent(in)    :: narea
    type (cmr1), intent(inout) :: fishn(:)
    type (cmr1), intent(inout) :: fish(:)

    integer :: ia

    do ia=1,narea
      fish(ia)%p = fishn(ia)%p
    enddo

  end subroutine fishsave

  !=============================================================================

  subroutine fishnull(fish)
    !---------------------------------------------------------------------------
    ! set up "neutral" fish-forcings
    !---------------------------------------------------------------------------

    implicit none

    ! argument
    real(8), intent(out), dimension(:) :: fish

    fish(:) =  0.d0

  end subroutine fishnull

  !=============================================================================

  subroutine fishclose()
    !---------------------------------------------------------------------------
    ! close the fish-forcings file
    !---------------------------------------------------------------------------
    implicit none

    close(lunfish)
  end subroutine fishclose

  !=============================================================================

  subroutine fishfeeding( fishn,fisho,temp,fish)
    !---------------------------------------------------------------------------
    ! feeding pressure from fish on zooplakton
    ! index (values between zero and one)
    !---------------------------------------------------------------------------

    use constants,  only : rzeite, zzeits
    implicit none

    ! arguments
    real(8), intent(in) :: fishn, fisho   ! fish feeding pressure at surf. point
    real(8), intent(in) :: temp           ! temperature at current location
    real(8), intent(out) :: fish          ! fish feeding pressure at surf. point
    
    ! local vars
    real(8) :: tfrac
    tfrac = zzeits/rzeite
    fish =  (1.0-tfrac)*fisho + tfrac*fishn

    ! this is the place to add temperature dependence

  end subroutine fishfeeding

end module biofish

 


