module bioinput

  implicit none
  private

  public  :: bioread, biobndcl, bioread_npr

  real(8), parameter, private :: zero = 0.0_8, half = 0.5_8, hpm = 720.0_8
  real(8), parameter, private :: onemill = 1000000.0_8
contains

!===============================================================================

  subroutine bioread( iact, n3d, narea, coldst, iu06, benthos, cmp, nl, nu )

    use io_subs,   only : io_new_unit
    use exits,     only : exitme
    use constants, only : aufdat
    use bioparam,  only : Nnorm, Onorm, ntbio, idx_oxy, ibiofields
    use biogas,    only : oxygensat

    implicit none

    integer(4), intent(in)    :: iact, narea, iu06, n3d, nl, nu
    logical,    intent(in)    :: coldst
    real(8),    intent(out)   :: benthos(:,0:) 
    real(8),    intent(inout) :: cmp(1:,0:)  

    integer(4)       :: ios, i
    integer(4), save :: lun
    character(11)    :: starta, fnam
    character(3)     :: c3


    if (.not.coldst) then
      ! read from a bio_start file (i.e. like a hotstarted simulation)

      if (iact == 1) then
        lun = io_new_unit()
        open(lun,file='bio_start',form='unformatted',iostat=ios,               &
             action='read',status='old')
        if (ios /= 0) call exitme (1, 'Could not open bio_start' )
        read(lun,iostat=ios) starta
        if (ios /= 0) then
          write(iu06,*) 'I/O problem - iostat is:',ios
          call exitme (2, 'Could not read starta from bio_start' )
        endif
        if (starta(2:11) /= aufdat(2:11)) then
          call exitme(3,'Time stamp in bio_start does not match starting time')
        endif
      endif

      read(lun,iostat=ios) cmp(nl:nu,:), benthos(:,:)
      if (ios /= 0) then
        write(iu06,*) 'I/O problem - iostat is:',ios
        write(iu06,*) '                   area:',iact
        call exitme (4, 'Could not read bio_start' )
      endif
      cmp(nl:nl-2+ntbio,:) = cmp(nl:nl-2+ntbio,:)/Nnorm
      cmp(nl-1+idx_oxy,:)    = cmp(nl-1+idx_oxy,:)/Onorm
      benthos(:,:) = benthos(:,:)/Nnorm

      if (iact == narea) close(lun)

    else
      ! assign bio vars as constants or read from a file (like coldstarted)

      cmp(nl:nu,:) = zero
      benthos(:,:) = zero

      ! Initialize oxygen at 100% saturation
      do i=1,n3d
        cmp(nl-1+idx_oxy,i) = oxygensat(cmp(1,i),cmp(2,i)) / Onorm
      enddo

      if (ibiofields == 1) then
        !- read from a dummy restart file containing only biological tracers:
        if (iact == 1) then
          lun = io_new_unit()
          open(lun,file='restart.bio',form='unformatted',iostat=ios,           &
               action='read',status='old')
          if (ios /= 0) call exitme (5, 'Could not open restart.bio' )
        endif

        read(lun,iostat=ios) cmp(nl:nu,:), benthos(:,:)
        if (ios /= 0) then
          write(iu06,*) 'I/O problem - iostat is:',ios
          write(iu06,*) '                   area:',iact
          call exitme (6, 'Could not read restart.bio' )
        endif
        if (iact == narea) close(lun)

      elseif (ibiofields == 2) then
        !- read biological tracers (excl. benthos) from an ascii data file:

        if (iact < 10) then
          write(c3,'(i1)') iact
        elseif (iact < 100) then
          write(c3,'(i2)') iact
        elseif (iact < 1000) then
          write(c3,'(i3)') iact
        else
          call exitme(7,'Cannot handle more than 999 areas.')
        endif

        fnam = 'bio.init'//trim(c3)
        lun = io_new_unit()
        write(iu06,*) 'Reading biological fields from ',fnam
        open(lun,file=fnam,form="formatted",status="old",iostat=ios)
        if (ios /= 0) call exitme(8,'Could not open '//trim(fnam))

        cmp(nl:nu,0) = onemill
        do i=1,n3d
          read(lun,*) cmp(nl:nu,i)
        enddo
        cmp(nl:nl-2+ntbio,:) = cmp(nl:nl-2+ntbio,:)/Nnorm
        cmp(nl-1+idx_oxy,:)    = cmp(nl-1+idx_oxy,:)/Onorm

        if (iact == narea) close(lun)
      endif

    endif

  end subroutine bioread

!===============================================================================

  subroutine biobndcl(kmx, kh, mmk, nz, nlow, nup, bndzk, iu06)

    use cmod_miscvars, only : jstuna
    use cmod_params,   only : krz, mainarea
    use exits,         only : exitme
    use io_subs,       only : io_new_unit
    implicit none

    integer(4), intent(in)    :: kmx, kh(0:), mmk(:,0:,0:), nz, nlow, nup, iu06
    real(8),    intent(inout) :: bndzk(:,:,0:)        ! (ntracers,kmx,0:nz)

    character(20)  :: fnam
    character(256) :: cformat
    integer(4)     :: lune, nbp, k, t, mon, monat, kb, n, ios, ibnd, jbnd, ntr
    integer(4)     :: khbnd(nz)
    real(8)        :: pqr, w1, w2
    real(8), dimension(nup-nlow+1,kmx,0:nz,12) :: rwzke_a

    !- initialise --------------------------------------------------------------
    ntr = nup-nlow+1
    rwzke_a(:,:,:,:) = zero
    khbnd(:) = 0
    do n=1,nz
      ibnd = krz(mainarea)%p(1,n)
      jbnd = krz(mainarea)%p(2,n)
      khbnd(n) = kh(mmk(1,ibnd,jbnd)) 
    enddo

    !- read climatology ecological or passive tracers --------------------------
    if (nz > 0) then
      fnam = 'rwzkebnd'
      lune = io_new_unit()
      open (lune,file=fnam,status='old',iostat=ios,pad='NO')
      if (ios /= 0) then
        write(iu06,*) 'Cannot open rwzkebnd ... continue without'
      else
        if     (kmx < 10) then
          write(cformat,'(i1)') kmx
        elseif (kmx < 100) then
          write(cformat,'(i2)') kmx
        elseif (kmx < 1000) then
          write(cformat,'(i3)') kmx
        else
          call exitme(2,'Cannot handle > 999 layers for tracer climatology.')
        endif

        cformat = '(i4,'//trim(cformat)//'f8.3)'

        do t=1,ntr
          do n=1,nz
            read (lune,*,iostat=ios)
            if (ios /= 0) call exitme (2,'Cannot read rwzkebnd')
            read (lune,'(i6)',iostat=ios) nbp
            if (ios /= 0) call exitme (3,'Cannot read rwzkebnd')
            kb = khbnd(n)
            do mon=1,12
              read (lune,cformat,iostat=ios) monat,(rwzke_a(t,k,n,mon),k=1,kb)
              if (ios /= 0) call exitme (4,'Cannot read rwzkebnd')
            enddo
            read (lune,*,iostat=ios)
            if (ios /= 0) call exitme (5,'Cannot read rwzkebnd')
          enddo
        enddo
        close(lune)
      endif

      !- interpolation to actual day:
      mon = (jstuna+720/2)/720       !       mon=0,1,2,...,11,12
      pqr = real(jstuna-mon*720,8)/hpm
      w1  = half - pqr
      w2  = half + pqr

      do n=1,nz
        kb = khbnd(n)
        do k=1,kb
          bndzk(nlow:nup,k,n) = w1*rwzke_a(1:ntr,k,n,mod(mon+11,12)+1)         &
                              + w2*rwzke_a(1:ntr,k,n,mod(mon+12,12)+1)
        enddo
        do k=kb+1,kmx
          bndzk(nlow:nup,k,n) = bndzk(nlow:nup,kb,n)
        enddo
      enddo
      bndzk(nlow:nup,:,0) = bndzk(nlow:nup,:,1)


    !- Dummy values if no bnd points -------------------------------------------
    else
      bndzk(:,:,:) = zero
    endif

  end subroutine biobndcl

  !=============================================================================
  
  subroutine bioread_npr( iact, narea, iu06, npr )
    use io_subs,   only : io_new_unit
    use exits,     only : exitme
    use bioparam,  only : rfr, npread

    implicit none

    integer(4), intent(in)    :: iact, narea, iu06
    real(8),    intent(out)   :: npr(0:)

    integer(4)       :: ios
    logical, save    :: fopened = .false.
    integer(4), save :: lun

    if (iact == 1 .and. npread) then
      fopened = .true.
      lun = io_new_unit()
      open(lun,file='NPr.dat',form='unformatted',iostat=ios,action='read',     &
           status='old')
      if (ios /= 0) then
        write(iu06,*) 'WARNING: Cannot open NPr.dat, using a fixed NP-ratio'
        fopened = .false.
      endif
    endif

    if(fopened) then
      read(lun,iostat=ios) npr(0:)
      if (ios /= 0) then
        write(iu06,*) 'WARNING: cannot read NPr.dat, using a fixed N/P-ratio'
        npr(:) = rfr
      endif
    else
      npr(:)=rfr
    endif

    if (iact == narea .and. fopened) close(lun)

  end subroutine bioread_npr

end module bioinput
