module bioinput

  implicit none
  private

  public  :: bioread, biobndcl

  real(8), parameter, private :: zero = 0.0_8, half = 0.5_8, hpm = 720.0_8
  real(8), parameter, private :: onemill = 1000000.0_8
contains

!===============================================================================

  subroutine bioread(iact,n3d,narea,coldst,iu06,benthos,cmp,nl,nu)

    use io_subs,    only : io_new_unit
    use exits,      only : exitme
    use constants,  only : aufdat
    use bioparam,   only : Nnorm, Onorm, idx_det, idx_oxy, ibiofields
    use biogas,     only : oxygensat

    implicit none

    integer(4), intent(in)    :: iact, narea, iu06, n3d, nl, nu
    logical,    intent(in)    :: coldst
    real(8),    intent(out)   :: benthos(0:) 
    real(8),    intent(inout) :: cmp(1:,0:)  

    integer(4)       :: ios, i
    integer(4), save :: lun
    character(11)    :: starta, fnam
    character(3)     :: c3

    if (.not.coldst) then
      ! read from a bio_start file (i.e. like a hotstarted simulation)

      if (iact == 1) then
        lun = io_new_unit()
        open(lun,file='bio_start',form='unformatted',iostat=ios,               &
             action='read',status='old')
        if (ios /= 0) call exitme (1, 'Could not open bio_start' )
        read(lun,iostat=ios) starta
        if (ios /= 0) then
          write(iu06,*) 'I/O problem - iostat is:',ios
          call exitme (2, 'Could not read starta from bio_start' )
        endif
        if (starta(2:11) /= aufdat(2:11)) then
          call exitme(3,'Time stamp in bio_start does not match starting time')
        endif
      endif

      read(lun,iostat=ios) cmp(nl:nu,:), benthos(:)
      if (ios /= 0) then
        write(iu06,*) 'I/O problem - iostat is:',ios
        write(iu06,*) '                   area:',iact
        call exitme (4, 'Could not read bio_start' )
      else
        call scaleRestart(cmp,benthos,nl,nu)
      endif

      if (iact == narea) close(lun)

    else
      ! assign bio vars as constants or read from a file (like coldstarted)

      cmp(nl:nu,:) = zero
      benthos(:)   = zero

      ! Initialize oxygen at 100% saturation
      do i=1,n3d
        cmp(nl-1+idx_oxy,i) = oxygensat(cmp(1,i),cmp(2,i)) / Onorm
      enddo

      if (ibiofields == 1) then
        !- read from a dummy restart file containing only biological tracers:
        if (iact == 1) then
          lun = io_new_unit()
          open(lun,file='restart.bio',form='unformatted',iostat=ios,           &
               action='read',status='old')
          if (ios /= 0) call exitme (5, 'Could not open restart.bio' )
        endif
        
        ! FIXME: should we read starta here?
        read(lun,iostat=ios) starta
        if (ios /= 0) then
          write(iu06,*) 'I/O problem - iostat is:',ios
          call exitme (2, 'Could not read starta from bio_start' )
        endif

        read(lun,iostat=ios) cmp(nl:nu,:), benthos(:)
        if (ios /= 0) then
            write(iu06,*) 'I/O problem - iostat is:',ios
            write(iu06,*) '                   area:',iact
            call exitme (4, 'Could not read restart_bio' )
        else
          call scaleRestart(cmp,benthos,nl,nu)
        endif

        if (iact == narea) close(lun)

      elseif (ibiofields == 2) then
        !- read biological tracers (excl. benthos) from an ascii data file:

        if (iact < 10) then
          write(c3,'(i1)') iact
        elseif (iact < 100) then
          write(c3,'(i2)') iact
        elseif (iact < 1000) then
          write(c3,'(i3)') iact
        else
          call exitme(7,'Cannot handle more than 999 areas.')
        endif

        fnam = 'bio.init'//trim(c3)
        lun = io_new_unit()
        write(iu06,*) 'Reading biological fields from ',fnam
        open(lun,file=fnam,form="formatted",status="old",iostat=ios)
        if (ios /= 0) call exitme(8,'Could not open '//trim(fnam))

        cmp(nl:nu,0) = onemill
        do i=1,n3d
          read(lun,*) cmp(nl:nu,i)
        enddo
        call scaleRestart(cmp,benthos,nl,nu)

        if (iact == narea) close(lun)
      endif

    endif

  end subroutine bioread
!===============================================================================

  subroutine biobndcl(kmx, kh, mmk, nz, nlow, nup, bndzk)

    use cmod_miscvars, only : jstuna
    use cmod_params,   only : krz, mainarea
    use exits,         only : exitme
    use io_subs,       only : io_new_unit
    
    implicit none

    integer(4), intent(in)    :: kmx, kh(0:), mmk(:,0:,0:), nz, nlow, nup 
    real(8),    intent(inout) :: bndzk(:,:,0:)        ! (ntracers,kmx,0:nz)

    character(20)  :: fnam
    character(256) :: cformat
    integer(4)     :: lune, nbp, k, t, mon, monat, kb, n, ios, ibnd, jbnd, ntr
    integer(4)     :: khbnd(nz)
    real(8)        :: pqr, w1, w2
    real(8), dimension(nup-nlow+1,kmx,0:nz,12) :: rwzke_a

    !- initialise --------------------------------------------------------------
    ntr = nup-nlow+1
    rwzke_a(:,:,:,:) = zero
    khbnd(:) = 0
    do n=1,nz
      ibnd = krz(mainarea)%p(1,n)
      jbnd = krz(mainarea)%p(2,n)
      khbnd(n) = kh(mmk(1,ibnd,jbnd)) 
    enddo

    !- read climatology ecological or passive tracers --------------------------
    if (nz > 0) then
      fnam = 'rwzkebnd'
      lune = io_new_unit()
      open (lune,file=fnam,status='old',iostat=ios,pad='NO')
      if (ios /= 0) then
        write(*,*) 'Cannot open rwzkebnd ... continue without'
      else
        if     (kmx < 10) then
          write(cformat,'(i1)') kmx
        elseif (kmx < 100) then
          write(cformat,'(i2)') kmx
        elseif (kmx < 1000) then
          write(cformat,'(i3)') kmx
        else
          call exitme(2,'Cannot handle > 999 layers for tracer climatology.')
        endif

        cformat = '(i4,'//trim(cformat)//'f9.4)'

        do t=1,ntr
          do n=1,nz
            read (lune,*,iostat=ios)
            if (ios /= 0) call exitme (2,'Cannot read rwzkebnd')
            read (lune,'(i6)',iostat=ios) nbp
            if (ios /= 0) call exitme (3,'Cannot read rwzkebnd')
            kb = khbnd(n)
            do mon=1,12
              read (lune,cformat,iostat=ios) monat,(rwzke_a(t,k,n,mon),k=1,kb)
              if (ios /= 0) call exitme (4,'Cannot read rwzkebnd')
            enddo
            read (lune,*,iostat=ios)
            if (ios /= 0) call exitme (5,'Cannot read rwzkebnd')
          enddo
        enddo
        close(lune)
      endif

      !- interpolation to actual day:
      mon = (jstuna+720/2)/720       !       mon=0,1,2,...,11,12
      pqr = real(jstuna-mon*720,8)/hpm
      w1  = half - pqr
      w2  = half + pqr

      do n=1,nz
        kb = khbnd(n)
        do k=1,kb
          bndzk(nlow:nup,k,n) = w1*rwzke_a(1:ntr,k,n,mod(mon+11,12)+1)         &
                              + w2*rwzke_a(1:ntr,k,n,mod(mon+12,12)+1)
        enddo
        do k=kb+1,kmx
          bndzk(nlow:nup,k,n) = bndzk(nlow:nup,kb,n)
        enddo
      enddo
      bndzk(nlow:nup,:,0) = bndzk(nlow:nup,:,1)


    !- Dummy values if no bnd points -------------------------------------------
    else
      bndzk(:,:,:) = zero
    endif

  end subroutine biobndcl

!===============================================================================
  
  subroutine scaleRestart(cmp,benthos,nl,nu)
    use bioparam, only : stage_model, zpmodel,Nnorm,Onorm, idx_det, idx_oxy,   &
                         idx_sil,idx_proto,idx_eggs
    
    implicit none

    ! args
    real(8), intent(inout) :: cmp(1:,0:), benthos(0:)
    integer(4), intent(in) :: nl,nu

    ! parameters
    real(8) :: dia_correct    = 1.0_8/3.0_8
    real(8) :: flag_correct   = 1.0_8/3.0_8
    real(8) :: zoo_correct    = 0.03_8/Nnorm
    real(8) :: det_coorect    = 0.05_8/Nnorm
    real(8) :: ben_correct    = 50.0_8/Nnorm 
    
    real(8) :: proto_correct  = 0.02_8/Nnorm
     
    real(8) :: finm_correct   = zero
    real(8) :: helgo_correct1 = zero 
    real(8) :: helgo_correct2 = 0.015_8/2.0_8/ Nnorm
    real(8) :: helgo_correct3 = 11.5_8/2.0_8

    ! scale everything with Nnorm execpt for oxygen
    cmp(nl:nl-1+idx_det,:) = cmp(nl:nl-1+idx_det,:)/Nnorm
    cmp(nl-1+idx_oxy,:)    = cmp(nl-1+idx_oxy,:)/Onorm
    cmp(nl-1+idx_sil,:)    = cmp(nl-1+idx_sil,:)/Nnorm
    benthos(:)             = benthos(:)/Nnorm

    if (zpmodel) then
      cmp(nl-1+idx_proto,:)    = cmp(nl-1+idx_proto,:)/Nnorm
    endif
    if (stage_model) then
      cmp(nl-1+idx_eggs:,:) = cmp(nl-1+idx_eggs:,:)/Nnorm
    endif
    
    
    !cmp(ia)%p(idx_dia,1:) = cmp(ia)%p(idx_dia,1:)*dia_correct
    !cmp(ia)%p(idx_flag,1:)= cmp(ia)%p(idx_flag,1:)*flag_correct
    !cmp(ia)%p(idx_zoo,1:) = zoo_correct
    !cmp(ia)%p(idx_det,1:) = det_correct
    !benthos(ia)%p(1:)     = ben_correct

    !if (zpmodel) then
    !  cmp(ia)%p(idx_proto,i) = proto_correct    
    !endif

    !if (stage_model) then
    !  ! C. finmarchicus
    !  cmp(ia)%p(idx_eggs,:)     = finm_correct         
    !  cmp(ia)%p(idx_naup_m,:)   = finm_correct
    !  cmp(ia)%p(idx_naup_a,:)   = finm_correct           
    !  cmp(ia)%p(idx_cop1_m,:)   = finm_correct
    !  cmp(ia)%p(idx_cop1_a,:)   = finm_correct           
    !  cmp(ia)%p(idx_cop2_m,:)   = finm_correct
    !  cmp(ia)%p(idx_cop2_a,:)   = finm_correct
    !  cmp(ia)%p(idx_adul_m,:)   = finm_correct
    !  cmp(ia)%p(idx_adul_a,:)   = finm_correct

      ! C. helgolandicus
    !  cmp(ia)%p(idx_eggs_ch,:)   = helgo_correct1           
    !  cmp(ia)%p(idx_naup_m_ch,:) = helgo_correct1
    !  cmp(ia)%p(idx_naup_a_ch,:) = helgo_correct1
    !  cmp(ia)%p(idx_cop1_m_ch,:) = helgo_correct1
    !  cmp(ia)%p(idx_cop1_a_ch,:) = helgo_correct1
    !  cmp(ia)%p(idx_cop2_m_ch,:) = helgo_correct1
    !  cmp(ia)%p(idx_cop2_a_ch,:) = helgo_correct1
    !  cmp(ia)%p(idx_adul_m_ch,:) = helgo_correct2
    !  cmp(ia)%p(idx_adul_a_ch,:) = helgo_correct3
    ! endif


  end subroutine scaleRestart

end module bioinput
