module biolight 

  implicit none
  private
  
  public :: bio_light
  
contains
  
  !----------------------------------------------------------------------------!
  !     calculate the light considering self shadowing
  !----------------------------------------------------------------------------!
  subroutine bio_light(kb,j,hi,hsrf,swrad,salt,ecoa,PPI,Idep)

    use bioparam,    only : Nnorm, idx_dia, idx_flag, idx_cyano, idx_det,      &
                            att_par_1, att_par_2, Imin
    use constants,   only : zeitstu
    
    !--------------------------------------------------------------------------!
    !     cmod variables
    !--------------------------------------------------------------------------!
    integer(4), intent(in) :: kb,j
    real(8), intent(in)    :: hi(:)             ! (ke)
    real(8), intent(in)    :: hsrf
    real(8), intent(in)    :: swrad
    real(8), intent(in)    :: salt(:)

    !--------------------------------------------------------------------------!
    !     biomod variables
    !--------------------------------------------------------------------------!
    real(8), intent(in)    :: ecoa(:,:)         ! (ke,ntbio)  old values
    real(8), intent(out)   :: PPI(:)            ! (1:ke)
    real(8), intent(out)   :: Idep(:)            ! (1:ke)

    !--------------------------------------------------------------------------!
    !     local variables
    !--------------------------------------------------------------------------!
    real(8) :: I0norm, IIopt   ! light
    real(8) :: ss
    real(8) :: ssz             ! light path
    real(8) :: shade
    real(8) :: hzk, exparg
    real(8) :: att_Nsea, att_BS, doy, att

    !--------------------------------------------------------------------------!
    ! constants
    !--------------------------------------------------------------------------!
    real(8), parameter    :: max_exp = 20.0_8, half = 0.5_8, zero = 0.0_8
    real(8), parameter    :: twentyfour = 24.0_8, quarter = 0.25_8
    real(8), parameter    :: one = 1.0_8
    real(8), parameter    :: att_Nsea_val = 0.2_8
    
    real(8), parameter    :: bs_a  = 0.26_8
    real(8), parameter    :: bs_b  = 0.003_8
    real(8), parameter    :: bs_c  = 0.23_8

    real(8), parameter    :: ns_a  = 0.206_8
    real(8), parameter    :: ns_b  = 0.05714_8
    real(8), parameter    :: ns_c  = 0.14_8
    
    integer(4), parameter :: ldaylim = 74, udaylim= 300
    integer(4), parameter :: deplim  = 21
    integer(4), parameter :: jlim  = 88
    integer(4), parameter :: saltlim = 25

    !--------------------------------------------------------------------------!
    ! loop counters
    !--------------------------------------------------------------------------!
    integer(4) :: k

    !--------------------------------------------------------------------------!
    ! level depths
    !--------------------------------------------------------------------------!
    real(8) :: zw              ! from surface to bottom of k'th layer
    real(8) :: zt              ! from surface to center of k'th layer

    ! day of year
    doy = zeitstu/twentyfour

    !--------------------------------------------------------------------------!
    !     calculate light attenuation due to transparency
    !--------------------------------------------------------------------------!
    if (kb.gt.deplim) then  !>45 m depth
      att_Nsea = zero
    else
      if  ((doy.gt.ldaylim).AND.(doy.lt.udaylim)) then
        att_Nsea = zero
      else
        att_Nsea = att_Nsea_val
      endif
    endif

    att_BS = zero

    !--------------------------------------------------------------------------!
    !     calculate the light considering self shadowing
    !--------------------------------------------------------------------------!
    I0norm = half*swrad/max(quarter*swrad,Imin)

    !--------------------------------------------------------------------------!
    ! surface layer
    !--------------------------------------------------------------------------!
    k  = 1

    !--------------------------------------------------------------------------!
    ! calculate depth of (center,bottom) of this layer
    !--------------------------------------------------------------------------!
    zw = hsrf
    zt = half*zw
    
    !--------------------------------------------------------------------------!
    !           biology self shading only in this loop
    !--------------------------------------------------------------------------!
    if ((j.lt.jlim).and.(salt(k).gt.saltlim)) then
      att = att_Nsea + max(ns_a - salt(k)*ns_b,ns_c)      !0.14 COHERENS
    else
      att = att_BS + max(bs_a - salt(k)*bs_b, bs_c)    !Colin- BS
    endif

    shade = ecoa(k,idx_dia  )                                                  &
          + ecoa(k,idx_flag )                                                  &
          + ecoa(k,idx_cyano)                                                  &
          + ecoa(k,idx_det  )

    ss    = shade*zt
    ssz   = ss/zw*Nnorm*att_par_2

    exparg = (ssz+att)*zt
    if (exparg > max_exp) then
      PPI(k)  = zero
      Idep(k) = zero
    else
       IIopt  = I0norm*exp(-exparg)
       PPI(k) = I0norm*exp(one - IIopt)
       Idep(k) = half*swrad*exp(-exparg)
    endif
    
    !--------------------------------------------------------------------------!
    ! layers below surface
    !--------------------------------------------------------------------------!
    do k=2,kb
      hzk = hi(k)

      !------------------------------------------------------------------------!
      ! calculate depth of (center,bottom) of this layer
      !------------------------------------------------------------------------!
      zt = zw + half*hzk
      zw = zw + hzk

      !------------------------------------------------------------------------!
      !           biology self shading only in this loop
      !------------------------------------------------------------------------!
      if ((j.lt.jlim).and.(salt(k).gt.saltlim)) then
        att = att_Nsea + max(ns_a - salt(k)*ns_b,ns_c)      ! COHERENS
      else
        att = att_BS + max(bs_a - salt(k)*bs_b, bs_c)    !Colin- BS
      endif

      shade = ecoa(k,idx_dia  )                                                &
            + ecoa(k,idx_flag )                                                &
            + ecoa(k,idx_cyano)                                                &
            + ecoa(k,idx_det  )

      ss  = ss + shade*hzk
      ssz = ss/zw*Nnorm*att_par_2
      
      exparg = (ssz+att)*zt
      if (exparg > max_exp) then
        PPI(k)  = zero
        Idep(k) = zero
      else
        IIopt   = I0norm*exp(-exparg)
        PPI(k)  = I0norm*exp(one - IIopt)
        Idep(k) = half*swrad*exp(-exparg)
      endif

    enddo ! k


  end subroutine bio_light
  !============================================================================!

end module biolight
