module biomod

  !============================================================================!
  ! A module containing core subroutines for the eco model. The model is based 
  ! on Neumann 2000, Neumann 2002 and DMI developments.
  !============================================================================!

  implicit none
  private
  
  public :: bio_dynamics

contains

  subroutine bio_dynamics(iw2,ke,mmk,ind,kh,h,u,v,wu,wv,ts,benthos,swrad,      &
                          dispvc,casus,npr, dts, idx, irc)
    !==========================================================================!
    !
    ! Biogeochemical dynamics
    !
    ! Biogeochemical-active tracers (physically passive) are solved with an 
    ! operator splitting scheme. tflow solves the tracer advection while the 
    ! change in concentration due to biogeochemical intercations and vertical 
    ! movement are calculated in the subroutine bio_dynamics. 
    !
    ! bio_dynamics:
    !
    ! The subroutine is called from the coupler at every #NTBIOMOD timestep 
    !
    !   i m                   zonal dimension (n-->s)
    !   i n                   merid dimension (w-->e)
    !
    ! input:
    !
    !   i iw2                 2d (surface) wet points
    !   i ke                  number of layers
    !   i mmk(ke,0:m+1,0:n+1) padded land/sea mask (0=land,1..iw3=sea)
    !   i ind(2,iw2)          index i,j of each surface point
    !   i kh(0:iw2)           No. of active layers at each wet surface point
    !   r h(0:iw3)            layer thicknesses
    !   r wu(0:iw2)           meridional wind speed 
    !   r wv(0:iw2)           zonal wind speed 
    !   r ts(1:2+nrt,0:iw3)   temperature, salinity and ecotracers
    !   r benthos(1:ntben,0:iw2)      benthos nitrogen
    !   r swrad(0:iw2)        shortwave radiation under the air-sea interface
    !   r dts                 time step  (seconds)
    !
    !   parameter:
    !
    !   i ntbio=9             number of ecological variables
    !
    !   returns:
    !
    !   r ts                  with updated ecological tracer concentrations
    !   r benthos             with uptaded concentrations
    !
    !   ecological variables, optional tracers in ():
    !
    !   0     dummy values
    !   1     ammonium
    !   2     nitrate
    !   3     phosphate
    !   4     diatoms
    !   5     flagellates
    !   6     cyano bacteria
    !   7     mesozooplankton (or bulk zooplankton)
    !   (8)   Optional microzooplankton
    !   8(9)  detritus
    !   9(10) oxygen
    !
    ! TODO:
    !   * Sediment erosion
    !   * Implicit scheme
    !   * Should nutrients be accumulated on the ice?
    !   * Should we really use the NP array in the surface loop if it is const.?
    !
    !==========================================================================!

    !--------------------------------------------------------------------------!
    ! Biomod params: constants, water type specific
    !--------------------------------------------------------------------------!
    use dmi_omp,    only : domp_get_domain
    use bioparam,   only : idx_dia, idx_dia, idx_flag, idx_det, idx_cyano,     &
                           idx_phos, idx_nit, idx_amm, idx_oxy,idx_miz,idx_sil,&
                           idx_mez, alphab, alphaf, alphap, ntbio, iv, deltao, &
                           cyanosul, b0, anit, cyanotll, dn,ds, f0, flagtll,   &
                           ksdenit, ksnit, nb,  nor, onnitr, ONamup, ntben,    &
                           onniup, onorm, p0, q10_rec, rb0, rf0,rp0, rnit,     &
                           wbz, wdz, wpz, NNdenit, bidx_nitr, bidx_sil,        &
                           srfamflx,srfniflx, srfpflx, Nnorm, Imin1, Imin2,    &
                           Imin3, ppispmmod, zpmodel, version, rfs, idx_dets,  &
                           flagtsc
    use biogas,     only : oxygensat, pistonvel
    use biolight,   only : bio_light
    use biovmove,   only : vmove, sedimentation_resuspension, drying_case
    use biobenthic, only : bio_benthic
    use biosolver,  only : solvebio
    use biozoo,     only : zoodynamics, zoorates

    !--------------------------------------------------------------------------!
    ! Cmod params: dimension, depth, lsm
    !--------------------------------------------------------------------------!
    integer(4), intent(in)    :: iw2, ke
    integer(4), intent(in)    :: mmk(:,0:,0:)     ! (ke,0:m+1,0:n+1)
    integer(4), intent(in)    :: ind(:,:)         ! (2,iw2)
    integer(4), intent(inout) :: idx(1:,1:)       ! (2,iw2) thread decompo
    integer(4), intent(in)    :: kh(0:)           ! (0:iw2)
    integer(4), intent(inout) :: irc              ! error status
    real(8),    intent(in)    :: h(0:)            ! (0:iw3)
    real(8),    intent(in)    :: u(0:), v(0:)     ! (0:iw3)
    real(8),    intent(in)    :: wu(0:), wv(0:)   ! (0:iw2)
    real(8),    intent(in)    :: dispvc(:,0:)     ! (1:3,0:iw3) eddyvdiff
    real(8),    intent(inout) :: ts(1:,0:)        ! (1:ntbio+2,0:iw3) T&S&eco
    logical,    intent(in)    :: casus(0:)        ! (0:iw2) ice cover
    real(8),    intent(inout) :: benthos(1:,0:)   ! sediment
    real(8),    intent(inout) :: npr(0:)          ! N/P ratio

    !--------------------------------------------------------------------------!
    ! Solar forcing: short wave radiation and solar elevation 
    !--------------------------------------------------------------------------!
    real(8), intent(in) :: swrad(0:)              ! (0:iw2)    sw radiation

    !--------------------------------------------------------------------------!
    ! Model time step (in seconds)
    !--------------------------------------------------------------------------!
    real(8), intent(in) :: dts

    !--------------------------------------------------------------------------!
    ! local arrays for the whole water column 
    !--------------------------------------------------------------------------!
    real(8) :: ecoa(1:ke,0:ntbio)   ! eco variables (inactive vars at 0)
    real(8) :: ben(0:ntben)         ! benthos variable
    real(8) :: deriv(0:ntbio)       ! Derivatves of eco variables
    real(8) :: benderiv(0:ntben)    ! Deriviate of benthos
    real(8) :: light(1:ke)          ! PAR in water column
    real(8) :: mmi(1:ke), hi(1:ke)  ! grid index and grid depth
    real(8) :: t_k(1:ke), s_k(1:ke) ! salt and temp
    real(8) :: eddyvdiff(1:ke)      ! vertical diffusivity

    !--------------------------------------------------------------------------!
    ! local variables
    !--------------------------------------------------------------------------!
    real(8) :: dt             ! time step in days
    real(8) :: rfr            ! N/P ratio
    real(8) :: amon,nitr,phos,dia,sil,flag,cyan,mezoo,mizoo,detn,dets,oxyg 
    real(8) :: phyto          ! total phyto
    real(8) :: diaeps         ! total diatoms including p0
    real(8) :: flageps        ! total flagellates including f0
    real(8) :: cyaneps        ! total cyanos including b0
    real(8) :: phos2,din2,sil2! squared nutrients
    real(8) :: ldn_N          ! loss rates times
    real(8) :: ldn_O          ! for N, O eq.
    real(8) :: invdin_eps     ! total inorganic N inversed and eps
    real(8) :: temp,temp2     ! temperature, temperature**2
    real(8) :: ppik           ! light limitaion
    real(8) :: lightk         ! light
    real(8) :: sali           ! salinity
    real(8) :: otemp          ! makes things go more easy
    real(8) :: osat           ! oxygen saturation 
    real(8) :: hsrf, hbot     ! surfacei/bottom layer height
    real(8) :: ksdenit2       ! half saturation oxygen dependence of nitrif.
    real(8) :: wui, wvi, wind ! wind components and windspeed
                              ! dia and flag nutrent half saturation squared
    real(8) :: alphap2, alphaf2 
                              ! nutrent half saturation times redfield squared
    real(8) :: alfrfr2, albrfr2, alprfr2, alprfs2, alphab2
    real(8) :: nutlimd, nutlimf, nutlimc
    real(8) :: ust,vst,velkb
    logical :: icefree        ! ice cover 

    !--------------------------------------------------------------------------!
    ! Rates
    !--------------------------------------------------------------------------!
    real(8) :: lpn, lpd, lp   ! loss rates
    real(8) :: ldn            ! loss rates D => A
    real(8) :: lds            ! loss rates Ds => S
    real(8) :: rp,rf,rb       ! uptake rates
    real(8) :: nf             ! nitrification rate
    real(8) :: o2flx          ! O2 surface flux
    real(8) :: uptake_df      ! uptake diatoms and flagellates
    real(8) :: uptake_dia     ! uptake diatoms
    real(8) :: uptake_flag    ! uptake flagellates
    real(8) :: pistonvel0     ! piston velocity
    real(8) :: denitnscal     ! denitrification
    type(zoorates) :: zrate   ! zooplankton rates
 
    !--------------------------------------------------------------------------!
    ! Loop counters
    !--------------------------------------------------------------------------!
    integer(4) :: i,k,j,nsrf,mi,mww,mnn,n2dl,n2du,kb

    !--------------------------------------------------------------------------!
    ! Local parameters
    !--------------------------------------------------------------------------!
    real(8), parameter :: zero  = 0.0_8, one = 1.0_8, ten = 10.0_8
    real(8), parameter :: four  = 4.0_8, half = 0.5_8
    real(8), parameter :: eps   = 0.00000000001_8            ! no division by 0
    real(8), parameter :: pd2ps = 1.0_8/(24.0_8*60.0_8*60.0_8)
    real(8), parameter :: pi    = 3.14159265358979323846_8 
    real(8), parameter :: hnull = 0.1_8

    !--------------------------------------------------------------------------!
    ! Bio model params are in units of days, so we scale the coefficients by
    ! a simple scale of the time step size and hope that time-params only
    ! enter the model as coefficients to dt:
    !--------------------------------------------------------------------------!
    dt = dts*pd2ps

    !--------------------------------------------------------------------------!
    ! some local parameters to ease computation
    !--------------------------------------------------------------------------!

    ksdenit2 = ksdenit*ksdenit
    alphap2  = alphap*alphap
    alphaf2  = alphaf*alphaf
    alphab2  = alphab*alphab
    alprfs2  = alphap2*rfs*rfs !silicate redfield ratio

    call domp_get_domain(kh, 1, iw2, n2dl, n2du, idx)

    !--------------------------------------------------------------------------!
    ! Storing input tracers in old array ecoa
    !--------------------------------------------------------------------------!
    ! init dummy vars for unused parameters
    ben(0)       = zero 
    benderiv(0)  = zero 
    deriv(0)     = zero
    ecoa(:,0)    = zero
 
    do nsrf=n2dl,n2du
      kb = kh(nsrf)
      !- Do we have a drying situation
      if (kb <= 1 .and. h(nsrf)<hnull) then
        call drying_case(hnull, ts(3:2+ntbio,nsrf),benthos(1:ntben,nsrf))
        cycle
      endif
 
      i = ind(1,nsrf)
      j = ind(2,nsrf)

      icefree      = casus(nsrf)
      ben(1:ntben) = benthos(1:ntben,nsrf)
      rfr          = npr(nsrf)
      alprfr2      = alphap2*rfr*rfr
      alfrfr2      = alphaf2*rfr*rfr
      albrfr2      = alphab2*rfr*rfr

      do k=1,kb
        mi = mmk(k,i,j)

        mmi(k) = mi
        hi(k)  = h(mi)

        t_k(k)          = max(ts(1,mi),zero) ! all bioeq's are based on temp>0
        s_k(k)          = ts(2,mi)
        ecoa(k,1:ntbio) = ts(3:2+ntbio,mi)
        eddyvdiff(k)    = dispvc(3,mi)
      enddo

      hsrf = max(hi(1),hnull)
      if (kb == 1) then
        hbot = hsrf
      else
        hbot = hi(kb)
      endif

      ! bottom velocity
      mi    = mmk(kb,i,j)         
      mww   = mmk(kb,i,j-1)      
      mnn   = mmk(kb,i-1,j)      
      ust   = half*(u(mi)+u(mww))
      vst   = half*(v(mi)+v(mnn))
      velkb = sqrt(ust*ust+vst*vst)

      !------------------------------------------------------------------------!
      ! Sinking and rising of particles and algae
      !------------------------------------------------------------------------!
      if ( kb > 1 ) then
        ! biological vertical movement neglected for single layer column
        call vmove(kb,hi,ecoa(:,idx_dia  ),wpz,dt,irc)
        call vmove(kb,hi,ecoa(:,idx_cyano),wbz,dt,irc)
        call vmove(kb,hi,ecoa(:,idx_det  ),wdz,dt,irc)
        call vmove(kb,hi,ecoa(:,idx_dets ),wdz,dt,irc)
        if (irc /= 0) return
      endif
      call sedimentation_resuspension(hbot, dt, velkb, ecoa(kb,:), ben(:))

      !------------------------------------------------------------------------!
      !  Light limitation considering self shading
      !------------------------------------------------------------------------!
      if (ppispmmod) then
        call bio_light(kb,hi,hsrf,swrad(nsrf),ecoa,eddyvdiff(:),light)
      else
        call bio_light(kb,hi,hsrf,swrad(nsrf),ecoa,light)
      endif

      !------------------------------------------------------------------------!
      !
      !            surface fluxes
      !
      !------------------------------------------------------------------------!
      !fixme: srf flux should probably be a function of the ice concentration
      !       nutrient should also be accumulated on the snow
      if (icefree) then
        osat = oxygensat(t_k(1),s_k(1)) / Onorm  !using biogas
               
        !----------------------------------------------------------------------!
        !    Gas transfer at surface using a fixed piston-velocity 
        !----------------------------------------------------------------------!
        oxyg = ecoa(1,idx_oxy)
        wui  = wu(nsrf)
        wvi  = wv(nsrf)
        wind = sqrt(wui*wui + wvi*wvi)      ! wind velocity
        pistonvel0 = pistonvel(wind,t_k(1)) ! using a variable piston-velocity 
        o2flx      = pistonvel0*(osat - oxyg) ! o2_rel * m/day
        ecoa(1,idx_oxy) = oxyg + dt*o2flx/hsrf

        !----------------------------------------------------------------------!
        !   Atmospheric deposistion of nitr, amon and phos
        !----------------------------------------------------------------------!
        ecoa(1,idx_nit)  = ecoa(1,idx_nit)  + srfniflx/hsrf/Nnorm*dt
        ecoa(1,idx_amm)  = ecoa(1,idx_amm)  + srfamflx/hsrf/Nnorm*dt
        ecoa(1,idx_phos) = ecoa(1,idx_phos) + srfpflx/hsrf/Nnorm*dt
      endif
      
      !- done with surface fluxes, now loop through the water column
      do k=1,kb
        mi = mmi(k)

        temp   = t_k(k)
        temp2  = temp*temp
        sali   = s_k(k)
        lightk = light(k)

        amon =  ecoa(k,idx_amm  ) 
        nitr =  ecoa(k,idx_nit  ) 
        phos =  ecoa(k,idx_phos ) 
        sil  =  ecoa(k,idx_sil ) 
        dia  =  ecoa(k,idx_dia  )
        flag =  ecoa(k,idx_flag )
        cyan =  ecoa(k,idx_cyano)
        mezoo=  ecoa(k,idx_mez  )  
        mizoo=  ecoa(k,idx_miz  )  
        detn =  ecoa(k,idx_det  )
        dets =  ecoa(k,idx_dets )
        oxyg =  ecoa(k,idx_oxy  )

        !----------------------------------------------------------------------!
        ! combined variables
        !----------------------------------------------------------------------!
        invdin_eps = one/(nitr+amon + eps)
        din2       = (nitr+amon)*(nitr+amon)
        phos2      = phos * phos
        sil2       = sil  * sil
        phyto      = dia + flag + cyan
        diaeps     = dia + p0
        flageps    = flag + f0
        cyaneps    = cyan + b0

        if (oxyg <= zero) then
          !--------------------------------------------------------------------!
          !
          !            Hypoxic conditions
          !
          !--------------------------------------------------------------------!
 
          !-----------------
          ! no phyto uptake under anoxic conditions
          !-----------------
          rp = zero
          rf = zero
          rb = zero
          uptake_dia  = zero
          uptake_flag = zero
          uptake_df   = zero

          !-----------------
          ! phyto mortality increases and no respiration
          !-----------------
          lpd = ten*deltao
          lpn = zero
          lp  = lpd

          !-----------------
          ! zooplankton dynamics 
          !-----------------
          call zoodynamics(temp,dia,flag,cyan,mezoo,mizoo,.true.,zpmodel,zrate,&
                           irc)
          if (irc /= 0) return
 
          !--------------------------------------------------------------------!
          !            detritus/recycling rates
          !--------------------------------------------------------------------!
          ! Pelagic recycling, temperature dependent, based on Savchuk
          !--------------------------------------------------------------------!
          !
          ! 1. O2 non-present + NO3 present ->  denitrification + 
          !    anoxic recycling depending on NO3 conc
          !
          ! 2. No oxygen + no nitrate -> anoxic recycling as negative oxygen 
          !    conc
          !             
          !--------------------------------------------------------------------!
          ldn = dn/ten*exp(q10_rec*temp)
          lds = ds/ten*exp(q10_rec*temp)

          denitnscal = nitr*nitr/(ksdenit2+nitr*nitr)
          ldn_N = NNdenit * denitnscal       ! (respiration of nitrate if avail
          ldn_O = ONamup*(one - denitnscal)  ! (less hydrogen sulphite formation
                                             ! if nitrate is avaialable)
 
          !- No Nitrification
          nf    = zero
        else  
          !--------------------------------------------------------------------!
          !
          !            Oxic conditions
          !
          !--------------------------------------------------------------------!

          !--------------------------------------------------------------------!
          !Phyoplankton rates
          !--------------------------------------------------------------------!
         
          nutlimf = min(din2/(alphaf2+din2),phos2/(alfrfr2+phos2))
          nutlimc = phos2/(albrfr2+phos2)
          if (version == 3) then 
            ! include silicate
            nutlimd = min(din2/(alphap2+din2),phos2/(alprfr2+phos2),           &
                          sil2/(alprfs2+sil2))
          else
            nutlimd = min(din2/(alphap2+din2),phos2/(alprfr2+phos2))
          endif
         
          ! NOTE!! The light description below is problematic since we get very
          ! low ppik for cyano and dia at the surface...consider change the 
          ! formulation to the one outcomment for the flag.

          !-----------------
          ! Diatom uptake
          !-----------------
          ppik = min(lightk/Imin1,one) ! ratio of optimal light for diatoms 
          ppik = ppik*exp(one-ppik)
          rp   = rp0*min(nutlimd, ppik)

          !-----------------
          ! Flagellate Uptake
          !-----------------
          !Ifac = light/Imin1
          !ppik = Ifac/sqrt(1+Ifac*Ifac)
          ppik = min(lightk/Imin2,one) ! ratio of optimal light for flagellate 
          ppik = ppik*exp(one-ppik)
          rf   = min(nutlimf, ppik)

          !-----------------
          ! T dependend Flagellates
          !-----------------
          !rf   = rf * rf0* (one + temp2/(flagtll+temp2))
          rf = rf * rf0 * flagtsc*exp(flagtll*temp)

          !-----------------
          ! Bluegreen Alge Uptake
          !-----------------
          ppik = min(lightk/Imin3,one) ! ratio of optimal light for bluegreen
          ppik = ppik*exp(one-ppik)
          rb = min(nutlimc, ppik)

          !-----------------
          ! T dependent Bluegreen Algae
          !-----------------
          ! fixme: check this formulation 
          rb = rb*rb0*four/(four + exp(cyanotll - temp))
          
          !-----------------
          ! S dependent Bluegreen Algae
          !-----------------
          rb = rb * (atan(-sali+cyanosul)/pi+half)
          
          !-----------------
          ! loss rates of phytoplankton
          !-----------------
          lpn = nb
          lpd = deltao
          lp  = lpn + lpd

          uptake_dia  = rp*diaeps
          uptake_flag = rf*flageps
          uptake_df   = uptake_dia + uptake_flag

          !--------------------------------------------------------------------!
          !            Nitrification rate
          !--------------------------------------------------------------------!
          otemp = max(zero, oxyg)
          nf    = rnit*otemp/(ksnit + otemp)*exp(anit*temp)

          !--------------------------------------------------------------------!
          ! Zooplankton dynamics
          !--------------------------------------------------------------------!
          call zoodynamics(temp,dia,flag,cyan,mezoo,mizoo,.false.,zpmodel,     &
                           zrate,irc) 
          if (irc /= 0) return
          
          !--------------------------------------------------------------------!
          ! oxic detritus/recycling rates
          !--------------------------------------------------------------------!
          ldn = dn*exp(q10_rec*temp)
          lds = ds*exp(q10_rec*temp)
          ldn_N = zero    ! oxic recycling (no respiration of nitrate)
          ldn_O = ONamup  ! oxic recycling (no respiration of nitrate)
 
        endif

        !----------------------------------------------------------------------!
        !
        !            biochemical tracer dynamics
        !
        !----------------------------------------------------------------------!

        !-----------------
        ! Ammonium Tracer
        !-----------------
        deriv(idx_amm) =                                                       &
                          ! recycling
                          ldn*detn                                             &
                          ! uptake and nitrification
                          - (uptake_df*invdin_eps + nf)*amon                   &
                          ! phytoplankton respiration losses
                          + lpn*phyto                                          &
                          ! zooplankton respiration losses
                          + zrate%lnzoo

        !-----------------
        ! Nitrate Tracer
        !-----------------
        deriv(idx_nit) =                                                       &
                          ! nitrification
                          nf*amon                                              &
                          ! uptake
                          - uptake_df*invdin_eps*nitr                          &
                          ! denitrification
                          - ldn_N*ldn*detn

        !-----------------
        ! Phosphate Tracer
        !-----------------
        deriv(idx_phos) = rfr*(                                                &
                                ! recycling
                                ldn*detn                                       &
                                ! uptake
                                - (uptake_df + rb*cyan)                        &
                                ! phytoplankton respiration losses
                                + lpn*phyto                                    &
                                ! zooplankton respiration losses
                                + zrate%lnzoo                                  &
                               )
        !-----------------
        ! Silicate Tracer
        !-----------------
        deriv(idx_sil) =                                                       &
                         ! recycling
                         lds*dets                                              &
                         ! uptake
                         - rfs*uptake_dia                                      &
                         ! diatom respiration
                         + rfs*lpn*dia
 
        !-----------------
        ! Diatoms Tracer
        !-----------------
        deriv(idx_dia) =                                                       &
                         ! uptake
                         uptake_dia                                            &
                         ! grazing, respiration, mortality
                         - (zrate%grazdia + lp)*dia

        !-----------------
        ! Flagellates Tracer
        !-----------------
        deriv(idx_flag) =                                                      &
                          ! uptake
                          uptake_flag                                          &
                          ! grazing, respiration, mortality
                          - (zrate%grazflag + lp)*flag
                           
        !-----------------
        ! Bluegreen Algae Tracer
        !-----------------
        deriv(idx_cyano) =                                                     &
                           ! uptake
                           rb*cyaneps                                          &
                           ! grazing, respiration, mortality
                           - (zrate%grazcyan + lp)*cyan
                            
        !-----------------
        ! Mesozooplankton Tracer
        !-----------------
        deriv(idx_mez) =                                                       &
                         ! grazing
                           zrate%mezgrazdia*dia   + zrate%mezgrazflag*flag     &
                         + zrate%mezgrazcyan*cyan + zrate%mezgrazmiz*mizoo     &
                         ! respiration
                         - zrate%lmez

        !-----------------
        ! Microzooplankton Tracer
        !-----------------
        deriv(idx_miz) =                                                       &
                         ! grazing
                           zrate%mizgrazdia*dia + zrate%mizgrazflag*flag       &
                         + zrate%mizgrazcyan*cyan                              &
                         ! eaten
                         - zrate%mezgrazmiz*mizoo                              &
                         ! respiration
                         - zrate%lmiz
                         
        !-----------------
        ! N/P Detritus Tracer
        !-----------------
        deriv(idx_det) =                                                       &
                         ! respiration phytoplankton
                         lpd*phyto                                             &
                         ! respiration zooplankton
                         + zrate%ldzoo                                         &
                         ! recycling of nitrogen detritus
                         - ldn*detn

        !-----------------
        ! Silicate Detritus Tracer
        !-----------------
        deriv(idx_dets) =                                                      &
                         ! respiration
                         rfs*lpd*dia                                           &
                         ! all silicate in dia turn to dets during zoo grazing 
                         + rfs*zrate%grazdia*dia                               &
                         ! recycling
                         - lds*dets

        !-----------------
        ! Oxygen Tracer
        !-----------------
        deriv(idx_oxy) = NOR*(                                                 &
                               ! Primary production
                               uptake_df*invdin_eps*                           &
                               (                                               &
                                 nitr*ONniup + amon*ONamup                     &
                               )                                               &
                               ! Primary production
                               + rb*cyan*ONamup                                &
                               ! Remineralization of detritus
                               - ldn_O*ldn*detn                                &
                               ! Phyto and zoo respiration
                               - ONamup*( lpn*phyto + zrate%lmiz + zrate%lmez) &
                               ! Nitrification
                               - ONnitr*nf*amon                                &
                              )
        
        !----------------------------------------------------------------------!
        ! Integration 
        !----------------------------------------------------------------------!
        if (k==kb) then
          !-----------------
          ! Add benthic dynamics
          !-----------------
          call bio_benthic(ecoa,temp,ben(:),deriv,benderiv(:),rfr,hbot,velkb,  &
                           kb,irc)
          
          call solvebio(ecoa,deriv,dt,irc,kb,ben(:),benderiv(:))
        else
          ! pelagic 
          call solvebio(ecoa,deriv,dt,irc,k)
        endif
        
        if (irc /= 0) return
      enddo  ! k 

      !-  transfer ecoarray back to the tracer array
      do k=1,kb
        mi = mmi(k)
        do i = 1,ntbio-1
          ts(2+i,mi) = max(ecoa(k,i),zero)
        enddo
        ! oxygen can be negative since it represent hydrogen sulphite
        ts(2+idx_oxy,mi) = ecoa(k,idx_oxy)
      enddo
      benthos(1:ntben,nsrf) = max(ben(1:ntben),zero)

    enddo    ! nsrf

  end subroutine bio_dynamics

end module biomod
