module biomod

!==============================================================================!
!
! biomod:
!
! biogeochemical dynamics
!
! parameters defined in module bioparam
!
! biogeochemical-active tracers are solved with a time splitting scheme, 
! tflow.f90 solves hydrodynamics assuming the tracers are 
! non-biogeochemical-active, while this subroutine solves 
! biogeochemical dynamics assuming the seawater is static. 
! an euler forward schema is applied here.
!
!------------------------------------------------------------------------------!
!
! version 0.91 beta
! zooplankton as in ERGOM4
!
!------------------------------------------------------------------------------!
!
! dependencies:
!
! module biodynamics
! contains:
!   subroutine bio_dynamics
! uses:
!   module bioparam
!    various parameters
!   module biogas
!    function oxygensat
!   module biovmove
!    subroutine vmove
!   module biolight
!     subroutine bio_light
!   module biobenthic
!     subroutine bio_benthic 
!   module constants
!     grarad
!   module dmi_omp
!     subroutine domp_get_domain
!
!==============================================================================!

  implicit none
  private
  
  public :: bio_dynamics

contains

  subroutine bio_dynamics(iw2,ke,mmk,ind,kh,h,wu,wv,ts,benthos,primprod,fisho, &
                          fishn,swrad,light,dts)

    !--------------------------------------------------------------------------!
    !
    ! bio_dynamics:
    !
    ! called from Cmod every time biogeochemical params are updated
    !
    !   i m                   zonal dimension (n-->s)
    !   i n                   merid dimension (w-->e)
    !
    ! input:
    !
    !   i iw2                 2d (surface) wet points
    !   i ke                  number of layers
    !   i mmk(ke,0:m+1,0:n+1) padded land/sea mask (0=land,1..iw3=sea)
    !   i ind(2,iw2)          index i,j of each surface point
    !   i kh(0:iw2)           No. of active layers at each wet surface point
    !   r hz(0:iw3)           layer thicknesses
    !   r z(0:iw2)            elevation of free surface layer
    !   r ts(1:,0:iw3)        temperature and salinity
    !   r swrad(0:iw2)        shortwave radiation under the air-sea interface
    !   r dts                 time step  (seconds)
    !
    !   parameter:
    !
    !   i ntbio=9             number of ecological variables
    !
    !   returns:
    !
    !   r eco(ntbio,0:iw3)    ecological variables, new timestep
    !
    !   ecological variables:
    !
    !   1     ammonium
    !   2     nitrate
    !   3     phosphate
    !   4     diatoms
    !   5     flagellates
    !   6     cyano bacteria
    !   7     zooplankton
    !   8     detritus
    !   9     oxygen
    !   10    silicate
    !   11    protozooplankton (if swithed on)
    !   12-29 zooplankton stagemodel (if stage_model)
    !
    ! NOTES:
    !   NTBIO SHOULD BE A CALL VARIABLE?
    !   INDEX 0 IS A DUMMY VALUE
    !   STILL NEDD TO SPECIFY BIOTF
    !   STILL NEED TO DEFINE WHAT TO DO ABOUT SEDIMENT
    !   STILL NEED TO CONSIDER WAVE EROSION
    !   STILL NEED TO CHECK SOLAR ANGLE/ SWRAD
    !   STILL NEED TO CHECK UNITS AND SCALINGS
    !
    !--------------------------------------------------------------------------!

    !--------------------------------------------------------------------------!
    ! Biomod params: constants, water type specific
    !--------------------------------------------------------------------------!
    use dmi_omp,    only : domp_get_domain
    use bioparam,   only : idx_dia, idx_dia, idx_flag, idx_det, idx_cyano,     &
                           idx_phos, idx_nit, idx_amm, idx_oxy, idx_zoo,       &
                           alphab, alphaf, alphap, ntbio, iv, deltao, cyanosul,&
                           b0, anit, cyanosll, cyanotll, dn, graz, f0, flagtll,&
                           ksdenit, ksnit, nb,  nor, onnitr, ONamup, nue,      &
                           onniup, onorm, p0, q10_rec, rb0, rf0,               &
                           rp0, rfr, rnit, sigma_b, toptz, wbz, wdz, wpz,      &
                           zooprefflag, zcl1, z0, zooprefdia, zooprefcyano,    &
                           NNdenit, srfamflx, srfniflx, srfpflx, Nnorm, Imin,  &
                           ecofish,fish, zpmodel, stage_model,zooprefpz,       &
                           pzprefdia, pzprefflag, pzprefcya, idx_proto,idx_sil,&
                           pistonvel, ivi, lzd_a, lzd_b, lzn_a, pz_kpz,        &
                           pz_lpzd, pz_lpzn, lzproto_a, gg_pz_a,pz_imax,rfr_s
                           
    use biogas,     only : oxygensat
    use biolight,   only : bio_light
    use biovmove,   only : vmove
    use biobenthic, only : bio_benthic
    use biofish,    only : fishfeeding
    use biozoo,     only : stage_dynamics
    ! fixme: should be passed as an argument
    use constants,   only : zeitstu

    !--------------------------------------------------------------------------!
    ! Cmod params: dimension, depth, lsm
    !--------------------------------------------------------------------------!
    integer(4), intent(in)    :: iw2, ke
    integer(4), intent(in)    :: mmk(:,0:,0:)        ! (ke,0:m+1,0:n+1)
    integer(4), intent(in)    :: ind(:,:)            ! (2,iw2)
    integer(4), intent(in)    :: kh(0:)              ! (0:iw2)
    real(8), intent(in)       :: h(0:)              ! (0:iw3)
    real(8), intent(in)       :: wu(0:)              ! (0:iw2)
    real(8), intent(in)       :: wv(0:)              ! (0:iw2)
    real(8), intent(inout)    :: ts(1:,0:)           ! (1:ntbio+2,0:iw3) T&S&eco
    real(8), intent(inout)    :: light(0:)           ! (0:iw3) Idep vars
    real(8), intent(inout)    :: primprod(0:)        ! (0:iw3) primary prod

    !--------------------------------------------------------------------------!
    ! data structures for biogeochemical model replace sbcocn vars.
    !--------------------------------------------------------------------------!
    real(8) :: ecoa(1:ke,ntbio)                   ! biomod variables, old values
    real(8) :: econ(1:ke,4)                       ! biovmove variables + oxy
    real(8), intent(inout) :: benthos(0:)         ! biomod variables, new values
                                                  ! (0:iw2)
    real(8), intent(in) :: fisho(0:), fishn(0:)   ! (0:iw2) feeding pres. (fish)

    !--------------------------------------------------------------------------!
    ! solar forcing: short wave radiation and solar elevation 
    !--------------------------------------------------------------------------!
    real(8), intent(in) :: swrad(0:)              ! (0:iw2)    sw radiation

    !--------------------------------------------------------------------------!
    ! model time step
    !--------------------------------------------------------------------------!
    real(8), intent(in) :: dts

    !--------------------------------------------------------------------------!
    ! local variables
    !--------------------------------------------------------------------------!
    real(8) :: dt, doy        ! time step in days, day of year
    real(8) :: foodtau(1:ke)  ! food at t=tau n - c5
    real(8) :: foodpz(1:ke)   ! food for protozoo
    real(8) :: aa,nn,po,pp,ff,bb,zz,dd,oo,oop1,nnp1,si,pzz ! local values of 
                                                         ! variables
    real(8) :: nnp1_mod
    real(8) :: phyto          ! total phyto
    real(8) :: ppg            ! total diatoms including p0
    real(8) :: ffg            ! total flagellates including f0
    real(8) :: bbg            ! total cyanos including b0
    real(8) :: part_uptake    ! uptake specific N (river or baltic) for A/N eq.
    real(8) :: uptake_p       ! uptake diatoms for P equation
    real(8) :: uptake_f       ! uptake flagellates for P equation
    real(8) :: graz_z         ! grazing rate for z equation
    real(8) :: o2flx          ! O2 surface flux
    real(8) :: sini           ! Si:N ratio-uptake function

    !--------------------------------------------------------------------------!
    ! local variables
    !--------------------------------------------------------------------------!
    real(8), parameter :: zero  = 0.0_8, one = 1.0_8, two = 2.0_8, ten = 10.0_8
    real(8), parameter :: four  = 4.0_8, twentyfour = 24.0_8
    real(8), parameter :: eps   = 0.00000000001_8            ! no division by 0
    real(8), parameter :: pd2ps = 1.0_8/(24.0_8*60.0_8*60.0_8)
    real(8), parameter :: cnst  = 2.7183_8

    real(8) :: dtoptz, ttoptz, toptz2, Ksdenit2, alphap2, alphaf2, alprfr2  
    real(8) :: alfrfr2, albrfr2, pz_kpz2

    real(8) :: food, food_eps
    real(8) :: ptemp,stemp
    real(8) :: plim,slim,nlim,sinifrac

    !--------------------------------------------------------------------------!
    !   rates
    !--------------------------------------------------------------------------!
    real(8) :: gg,ivs          ! grazing rate zoo, half saturation-salinity
    real(8) :: gg_pz           ! grazing rate proto
    real(8) :: graz_pz         !      - || -
    real(8) :: lzproto         ! mortality proto
    real(8) :: lpn, lpd, lp    ! loss rates
    real(8) :: lzn,lzd,lz      ! loss rates
    real(8) :: ldn             ! loss rates D => A
    real(8) :: ldn_N           ! loss rates times
    real(8) :: ldn_O           ! for N, O eq.
    real(8) :: rp,rf,rb        ! uptake rates
    real(8) :: nf              ! nitrification rate
    real(8) :: ntemp           ! limitation function
    real(8) :: ntemp_eps       ! total N > eps
    real(8) :: temp,tempq      ! temperature, temperature**2
    real(8) :: tempk           ! temperature in kelvin
    real(8) :: sali            ! salinity
    real(8) :: otemp           ! makes things go more easy
    real(8) :: osat            ! oxygen saturation 
    real(8) :: ppik            ! temp. storage of PPI(k)
    real(8) :: zsrf            ! temp. storage of z(nsrf)
    real(8) :: hsrf, hbot      ! surfacei/bottom layer height

    real(8) :: PPI(1:ke), Idep(1:ke)
    real(8) :: mmi(1:ke), hi(1:ke), t_k(1:ke), s_k(1:ke)

    !--------------------------------------------------------------------------!
    ! Stage model interaction rates (lossrate to amm, det & grazing)
    !--------------------------------------------------------------------------!
    real(8) :: gpress, lzd_hz, lzn_hz

    !--------------------------------------------------------------------------!
    ! loop counters
    !--------------------------------------------------------------------------!
    integer(4)         :: i,k,j,nsrf,mi,n2dl,n2du,kb
    real(8), parameter :: hnull   = 0.1_8
    real(8), parameter :: cel2kel = 273.15_8

    !--------------------------------------------------------------------------!
    ! some local parameters to ease computation
    !--------------------------------------------------------------------------!
    dtoptz   = one/toptz
    ttoptz   = two*dtoptz
    toptz2   = cnst/(toptz*toptz)
    Ksdenit2 = Ksdenit*Ksdenit
    alphap2  = alphap*alphap
    alphaf2  = alphaf*alphaf
    alprfr2  = alphap2*rfr*rfr
    alfrfr2  = alphaf2*rfr*rfr
    albrfr2  = alphab*alphab*rfr*rfr
    pz_kpz2  = pz_kpz*pz_kpz
    doy      = zeitstu/twentyfour

    call domp_get_domain(1, iw2, n2dl, n2du)
    
    !--------------------------------------------------------------------------!
    ! Storing input tracers in old array ecoa
    !
    ! Available food for zooplankton
    !     note the preference against the cyanos
    !--------------------------------------------------------------------------!
    do nsrf=n2dl,n2du
      i = ind(1,nsrf)
      j = ind(2,nsrf)

      kb = kh(nsrf)

      !------------------------------------------------------------------------!
      ! Feeding pressure from fish interpolated to current time step.
      ! (for now, not temp dep. and default values set in bioparam)
      !------------------------------------------------------------------------!
      if (ecofish) then
           call fishfeeding( fishn(nsrf),fisho(nsrf),temp,fish)
      endif

      do k=1,kb
        mi = mmk(k,i,j)

        mmi(k) = mi
        hi(k)  = h(mi)

        ecoa(k,1:ntbio)  = ts(3:2+ntbio,mi)
        t_k(k) = ts(1,mi)
        s_k(k) = ts(2,mi)
        foodtau(k) = zooprefdia   * ecoa(k,idx_dia  )                          &
                   + zooprefflag  * ecoa(k,idx_flag )                          &
                   + zooprefcyano * ecoa(k,idx_cyano)
        
        if (zpmodel) then
          foodtau(k) = foodtau(k) + zooprefpz*max(ecoa(k,idx_proto),zero)
          foodpz(k)  = pzprefdia  * max(ecoa(k,idx_dia),zero)                  &
                     + pzprefflag * max(ecoa(k,idx_flag),zero)                 &
                     + pzprefcya  * max(ecoa(k,idx_cyano),zero)
        else
          foodtau(k) = zooprefdia   * ecoa(k,idx_dia  )                        &
                     + zooprefflag  * ecoa(k,idx_flag )                        &
                     + zooprefcyano * ecoa(k,idx_cyano)
        endif
      enddo

      hsrf = max(hi(1),hnull)

      !------------------------------------------------------------------------!
      ! Sinking and rising of particles and algae
      !------------------------------------------------------------------------!
      !-  use time step size in secs:
      dt = dts
      if ( kb > 1 ) then
        call vmove(kb,hi,ecoa(:,idx_dia  ),econ(:,1),wpz,dt)
        call vmove(kb,hi,ecoa(:,idx_cyano),econ(:,2),wbz,dt)
        call vmove(kb,hi,ecoa(:,idx_det  ),econ(:,3),wdz,dt)

      ! biological vertical movement neglected for single layer column
      else
        econ(1,1) = ecoa(1,idx_dia)  
        econ(1,2) = ecoa(1,idx_cyano)
        econ(1,3) = ecoa(1,idx_det)  
      endif

      !------------------------------------------------------------------------!
      !  Light limitation considering self shading
      !------------------------------------------------------------------------!
      call bio_light(kb,j,hi,hsrf,swrad(nsrf),s_k,ecoa,PPI,Idep)

      !------------------------------------------------------------------------!
      !     the biological model
      !------------------------------------------------------------------------!
      !-  bio model params are in units of days, so we scale the coefficients by
      !   a simple scale of the time step size and hope that time-params only
      !   enter the model as coefficients to dt:
      dt = dts*pd2ps

      !------------------------------------------------------------------------!
      !
      !            surface fluxes
      !
      !------------------------------------------------------------------------!
      osat = oxygensat(t_k(1),s_k(1)) / Onorm  !using biogas

      !------------------------------------------------------------------------!
      !    Gas transfer at surface using a fixed piston-velocity 
      !------------------------------------------------------------------------!
      o2flx     = pistonvel*(osat - ecoa(1,idx_oxy)) ! o2_rel * m/day
      econ(1,4) = ecoa(1,idx_oxy) + dt*o2flx/hsrf
      if (kb > 1) econ(2:kb,4) = ecoa(2:kb,idx_oxy)
      
      !------------------------------------------------------------------------!
      !   Atmospheric deposistion of nitr, amon and phos
      !------------------------------------------------------------------------!
      !ecoa(1,idx_nit)  = ecoa(1,idx_nit)  + srfniflx/hsrf/Nnorm*dt
      !ecoa(1,idx_amm)  = ecoa(1,idx_amm)  + srfamflx/hsrf/Nnorm*dt
      !ecoa(1,idx_phos) = ecoa(1,idx_phos) + srfpflx/hsrf/Nnorm*dt
      do k=1,kb
        mi = mmi(k)

        temp = t_k(k)
        sali = s_k(k)

        !----------------------------------------------------------------------!
        !  tempq is used to scale Temp. dependent rates only
        !  hence, if T is less than zero T is set to zero instead
        !----------------------------------------------------------------------!
        tempq = max(temp,zero) ! biology hiberate at freezing temp
        tempq = tempq*tempq
        tempk = temp + cel2kel  ! temperature in Kelvin

        aa   = ecoa(k,idx_amm  ) 
        nn   = ecoa(k,idx_nit  ) 
        nnp1 = nn
        po   = ecoa(k,idx_phos ) 
        pp   = ecoa(k,idx_dia  )
        ff   = ecoa(k,idx_flag )
        bb   = ecoa(k,idx_cyano)
        zz   = ecoa(k,idx_zoo  )  
        dd   = ecoa(k,idx_det  )
        oo   = ecoa(k,idx_oxy  )
        oop1 = econ(k,4)
        si   = ecoa(k,idx_sil)
        if (zpmodel) then
          pzz = ecoa(k,idx_proto)
        endif
        
        !----------------------------------------------------------------------!
        ! combined variables
        !----------------------------------------------------------------------!
        ntemp     = nn + aa
        ntemp_eps = one/(ntemp + eps)
        ntemp     = ntemp*ntemp
        ptemp     = po * po
        phyto     = pp + ff + bb
        stemp     = si*si
        ppg       = pp + p0
        ffg       = ff + f0
        bbg       = bb + b0
        sinifrac  = si/(nn+eps)*si/(nn+eps)
        sini      = two*sinifrac/(sinifrac + one)

        !----------------------------------------------------------------------!
        !
        !            phytoplankton rates
        !
        !----------------------------------------------------------------------!
       
        !-----------------
        !   no uptake under anoxic conditions
        !-----------------
        if (oo <= zero) then
          rp = zero
          rf = zero
          rb = zero
          !-----------------
          !   mortality increases and no respiration
          !-----------------
          lpd = ten*deltao
          lpn = zero
          lp  = lpd

          uptake_p    = zero
          uptake_f    = zero
          part_uptake = zero
        else
          ppik = PPI(k)

          !-----------------
          ! Diatom uptake
          !-----------------
          slim = stemp/ (alphap2*(rfr_s*rfr_s) + stemp)
          nlim = ntemp / (alphap2 + ntemp)
          plim = ptemp / (alprfr2 + ptemp)
          rp   = rp0*min(slim,nlim,plim,PPI(k))
   
         !-----------------
         ! Flagellate Uptake
         !-----------------
         nlim = ntemp / (alphaf2 + ntemp)
         plim = ptemp / (alfrfr2 + ptemp)
         rf   = min(nlim,plim,PPI(k))

        !-----------------
        ! T dependend Flagellates
        !-----------------
        rf = rf * rf0*(one + tempq/(flagtll+tempq))

        !-----------------
        ! Bluegreen Alge Uptake
        !-----------------
        plim = ptemp / (albrfr2 + ptemp)
        rb   = min(plim,PPI(k))

        !-----------------
        ! T dependent Bluegreen Algae
        !-----------------
        rb = rb*rb0 / (one + exp(cyanotll - temp))

        !-----------------
        ! S dependent Bluegreen Algae
        !-----------------
        rb = rb/(one + exp(sali - cyanosul))/(one + exp(cyanosll - sali))
  
        !-----------------
        ! loss rates 
        !-----------------
        lpn = nb
        lpd = deltao
        lp  = lpn + lpd

        uptake_p    = rp*ppg*ntemp_eps
        uptake_f    = rf*ffg*ntemp_eps
        part_uptake = uptake_p + uptake_f
      endif ! oxic

        !----------------------------------------------------------------------!
        !
        !            Nitrification rate
        !
        !----------------------------------------------------------------------!
        otemp = max(zero, oop1)
        nf    = rnit*otemp/(Ksnit + otemp)*exp(anit*temp)
      
        !----------------------------------------------------------------------!
        !
        !            zooplankton rates
        !
        !----------------------------------------------------------------------!

        !-----------------
        ! under anoxic conditions no growth,
        ! increased mortality and no respiration
        !-----------------
        if (oo <= zero) then
          gg  = zero
          lzd = ten * sigma_b
          lzn = zero
          ! fixme (laj) should graz_z also be zero
          graz_z = zero
        else
        
          food     = foodtau(k)
          food_eps = max(food,eps)
        
          !-------------------------------------------
          ! the grazing pressure is spread on 3 species 
          ! Temperature dependent grazing, at optimal Temp. double grazing
          !--------------------------------------------
          ivs      = 0.189_8
          gg       = graz*(food*food/(food*food+ivs*ivs))*                     &
                   (one + toptz2*tempq*exp(one - temp*ttoptz))
          graz_z   = gg*(zz+z0)/food_eps
 
          ! egestion and mortality
          lzd = zcl1*zz/(zz+lzd_a)+lzd_b*gg

          ! Basal respiration, active respiration, egestion
          lzn = nue + lzn_a*gg

          lz = lzn + lzd

          !--------------------------------------------------------------------!
          !
          !            protozooplankton rates
          !
          !--------------------------------------------------------------------!
          if (zpmodel) then
            food     = foodpz(k)
            food_eps = max(food,eps)

            gg_pz = pz_imax/(food*food + pz_kpz2)                              &
                  * food*food*(one+tempq/(gg_pz_a+tempq))

            graz_pz = gg_pz*pzz/food_eps

            lzproto = (pz_lpzd + pz_lpzn)*gg_pz + lzproto_a

            ! fixme (laj) What about anoxic conditions?
          endif
        endif ! oxic

        !----------------------------------------------------------------------!
        !
        !            detritus/recycling rates
        !
        !----------------------------------------------------------------------!
        !
        ! Pelagic recycling, temperature dependant
        !     based on Savchuk
        !----------------------------------------------------------------------!
        ! 1. O2 present -> oxic recycling
        !
        ! 2. O2 non-present + NO3 present ->  denitrification + 
        !    anoxic recycling depending on NO3 conc
        !
        ! 3. No oxygen + no nitrate -> anoxic recycling as negative oxygen 
        !    conc
        !             
        !----------------------------------------------------------------------!
        ldn = dn*exp(q10_rec*temp)

        !-----------------
        ! stoechiometric rates
        !-----------------
        if (oop1 <= zero .and. nnp1 > zero ) then
          nnp1_mod = nnp1*nnp1/(Ksdenit2+nnp1*nnp1)
          ldn_N = NNdenit * nnp1_mod    !NO3 is respired
          ldn_O = ONamup *(one - nnp1_mod)
        else
          ldn_N = zero
          ldn_O = ONamup
        endif

        !----------------------------------------------------------------------!
        !
        !  Stage structured zooplankton model
        !
        !----------------------------------------------------------------------!
        if (stage_model) then
          call stage_dynamics(i,k,kb,dt,doy,ecoa,foodtau,temp,sali,oo,gpress,&
                              lzd_hz,lzn_hz)
        endif

        !----------------------------------------------------------------------!
        !
        !            biochemical tracer dynamics
        !
        !----------------------------------------------------------------------!

        !-----------------
        ! Oxygen Tracer
        !-----------------
        ecoa(k,idx_oxy) = econ(k,4)                                            &
                        + dt*NOR*(                                             &
                                    ! Primary production
                                    part_uptake*(nn*ONniup + aa*ONamup)        &
                                    ! Primary production
                                  + rb*bbg*ONamup                              &
                                    ! Remineralization of detritus
                                  - ldn_O*ldn*dd                               &
                                    ! Phyto respiration
                                  - ONamup*(  lpn*phyto                        &
                                    ! Zoo   respiration
                                            + lzn*zz   )                       &
                                    ! Nitrification
                                  - ONnitr*nf*aa                               &
                                 )

        !-----------------
        ! Ammonium Tracer
        !------------------
        ecoa(k,idx_amm) = ecoa(k,idx_amm)                                      &
                        + dt*(                                                 &
                                ! recycling
                                ldn*               dd                          &
                                ! uptake and nitrification
                              - (part_uptake + nf)*aa                          &
                                ! phytoplankton respiration losses
                              + lpn*phyto                                      &
                                ! zooplankton respiration losses
                              + lzn*zz                                         &
                             )

        !-----------------
        ! Nitrate Tracer
        !-----------------
        ecoa(k,idx_nit) = ecoa(k,idx_nit)                                      &
                        + dt*(                                                 &
                                ! nitrification
                                nf*           aa                               &
                                ! uptake
                              - part_uptake*  nn                               &
                                ! denitrification
                              - ldn_N * ldn*  dd                               &
                             )   

        !-----------------
        ! Phosphate Tracer
        !-----------------
        ecoa(k,idx_phos) = ecoa(k,idx_phos)                                    &
                        + dt*rfr*(                                             &
                                    ! recycling
                                    ldn*dd                                     &
                                    ! uptake
                                  - (part_uptake*(aa+nn) + rb*bbg)             &
                                    ! phytoplankton respiration losses
                                  + lpn*phyto                                  &
                                    ! zooplankton respiration losses
                                  + lzn*zz                                     &
                                 )
        !-----------------
        ! Silicate Tracer
        !-----------------
        ecoa(k,idx_sil) = ecoa(k,idx_sil)                                      &
                        + dt*rfr_s*(                                           &
                                    ! recycling
                                      ldn*dd*sini*0.4_8                        &
                                    ! uptage by diatoms
                                    - (rp*ppg)*sini                            &
                                    ! diatom respiration loss
                                    + lpn*ppg*sini                             &
                                    )

        !-----------------
        ! Diatoms Tracer
        !-----------------
        ecoa(k,idx_dia) = econ(k,1)                                            &
                        + dt*(                                                 &
                                ! uptake
                                uptake_p*(aa+nn)                               &
                                ! grazing, respiration, mortality
                              - (zooprefdia*graz_z + lp)*pp                    &
                             )

        !---------------------
        ! Flagellates Tracer
        !---------------------
        ecoa(k,idx_flag) = ecoa(k,idx_flag)                                    &
                        + dt*(                                                 &
                                ! uptake
                                uptake_f*(aa+nn)                               &
                                ! grazing, respiration, mortality
                              - (zooprefflag*graz_z + lp)*ff                   &
                             )

        !-------------------------
        ! Bluegreen Algae Tracer
        !    remember the preference 0.5 of zooplankton grazing
        !-------------------------
        ecoa(k,idx_cyano) = econ(k,2)                                          &
                        + dt*(                                                 &
                                ! uptake
                                rb*bbg                                         &
                                ! grazing, respiration, mortality
                              - (zooprefcyano*graz_z + lp)*bb                  &
                             )

        !---------------------
        ! Zooplankton Tracer
        !---------------------
        ecoa(k,idx_zoo) = ecoa(k,idx_zoo)                                      &
                        + dt*(                                                 &
                                ! grazing
                                graz_z* (zooprefdia*pp                         &
                                ! more grazing
                              + zooprefflag*ff + zooprefcyano*bb)              &
                                ! respiration, mortality
                              - lz*zz                                          &
                             )

        !-------------------
        ! Protozooplankton Tracer
        !-------------------
        if (zpmodel) then
          ecoa(k,idx_proto) = ecoa(k,idx_proto)                                &
                            + dt*(                                             &
                                   ! grazing
                                   graz_pz * (pzprefdia*pp                     &
                                   + pzprefflag*ff + pzprefcya*bb)             &
                                   ! respiration and mortality
                                 - (lzproto+zooprefpz+graz_z)*pzz              &
                                 )
          ! interaction with other variables

          !nutrients
          ecoa(k,idx_amm)   = ecoa(k,idx_amm)  + dt*pz_lpzn*pzz*gg_pz
          ecoa(k,idx_phos)  = ecoa(k,idx_phos) + dt*pz_lpzn*rfr*pzz*gg_pz
          ecoa(k,idx_sil)   = ecoa(k,idx_sil)  + dt*pz_lpzn*rfr_s*pzz*gg_pz
          ecoa(k,idx_det)   = ecoa(k,idx_det)  + dt*pz_lpzd*gg_pz*pzz
          ecoa(k,idx_oxy)   = ecoa(k,idx_oxy)  - dt*NOR*ONamup*lzproto*pzz
 
          ! more grazing 
          ecoa(k,idx_zoo)   = ecoa(k,idx_zoo)   + dt*graz_z*zooprefpz*pzz
          ecoa(k,idx_cyano) = ecoa(k,idx_cyano) - dt*pzprefcya*graz_pz*bb
          ecoa(k,idx_flag)  = ecoa(k,idx_flag)  - dt*pzprefflag*graz_pz*ff
          ecoa(k,idx_dia)   = ecoa(k,idx_dia)   - dt*pzprefdia*graz_pz*pp
        endif       

        !-------------------
        ! Detritus Tracer
        !-------------------
        ecoa(k,idx_det) = econ(k,3)                                            &
                        + dt*(                                                 &
                                ! respiration
                                lpd*phyto                                      &
                                ! respiration
                              + lzd*zz                                         &
                                ! recycling
                              - ldn*dd                                         &
                             )
        !------------------
        ! Stage model
        !------------------
        if (stage_model) then
          ! interaction between stagemodel and remaining ergom
           ecoa(k,idx_oxy)   = ecoa(k,idx_oxy)   - dt*NOR*ONamup*lzn_hz
           ecoa(k,idx_amm)   = ecoa(k,idx_amm)   + dt*lzn_hz
           ecoa(k,idx_phos)  = ecoa(k,idx_phos)  + dt*lzn_hz*rfr
           ecoa(k,idx_sil)   = ecoa(k,idx_sil)   + dt*lzn_hz*rfr_s
           ecoa(k,idx_dia)   = ecoa(k,idx_dia)   - dt*gpress*pp*zooprefdia
           ecoa(k,idx_flag)  = ecoa(k,idx_flag)  - dt*gpress*ff*zooprefflag
           ecoa(k,idx_cyano) = ecoa(k,idx_cyano) - dt*gpress*bb*zooprefcyano
           ecoa(k,idx_det)   = ecoa(k,idx_det)   + dt*lzd_hz
        endif

        !------------------
        ! Calculate primary production
        !------------------
        primprod(mi) = 1.0_8
      enddo !k loop
      
      !-  transfer back to eco(:,:) array:
      do k=1,kb
        mi = mmi(k)
        ts(2+1:2+idx_det,mi) = ecoa(k,1:idx_det)
        ts(2+idx_oxy,mi)     = ecoa(k,idx_oxy)
        ts(2+idx_sil,mi)     = ecoa(k,idx_sil)
        if (zpmodel) then
          ts(2+idx_proto,mi) = ecoa(k,idx_proto)
        endif
        if (stage_model) then
          ts(2+idx_proto+1:,mi) = ecoa(k,idx_proto+1:)
        endif

        light(mi) = Idep(k)

      enddo
      !-----------------
      ! Benthic dynamics
      !-----------------
      k  = kb
      mi = mmi(k)
      if (k == 1) then 
        hbot = hsrf
      else
        hbot = hi(kb)
      endif
      call bio_benthic(ts(2+1:2+ntbio,mi),t_k(k),benthos(nsrf),dt,hbot)

    enddo !surf loop
  end subroutine bio_dynamics

end module biomod

