module biooutput
  !-----------------------------------------------------------------------------
  ! stuff for output of ecomodel vars
  !-----------------------------------------------------------------------------

  use cmod_mem, only : cmi1, cmi2

  implicit none 
  private

  integer(4),                 private, save :: lunbiodat, itempd
  type(cmi2),    allocatable, private       :: eco_i4(:)
  type(cmi2),    allocatable, private       :: ben_i4(:)
  character(40),              private, save :: tdnam

  private :: wcompr_bio

  public  :: biofile_new, biofile_close, biofile_write, biofile_next
  public  :: write_bio_restart, val_benthos

contains

  !=============================================================================

  subroutine biofile_new ( )

    use constants,     only : archive, aufdat
    use cmod_params,   only : narea, iw2, iw3
    use bioparam,      only : ntbio, ntben
    use io_subs,       only : io_new_unit
    use exits,         only : exitme
    use cmod_miscvars, only : ihour,ctim
  
    implicit none

    logical,       save :: FirstTime = .true.
    character(11), save :: aufdat_bio
    character(2)        :: ctmp
    integer(4)          :: ia, iben, ios

    !- allocate on first entry -------------------------------------------------
    if (FirstTime) then
      FirstTime = .false.

      allocate( eco_i4(narea), ben_i4(narea) )
      do ia=1,narea
        iben = iw2(ia)
        allocate( eco_i4(ia)%p(1:ntbio,0:iw3(ia)),                             &
                  ben_i4(ia)%p(1:ntben,0:iben) )
      enddo

      itempd = 0
      if (archive) then
        tdnam = 'biodat'
      else
        write(ctmp,'(i2.2)') itempd
        tdnam = 'biodat.'//ctmp
      endif

      aufdat_bio(1:11) = aufdat(1:11)
    else
      itempd = itempd + 1
      write(ctmp,'(i2.2)') itempd
      tdnam = 'biodat.'//ctmp
    endif

    lunbiodat = io_new_unit()
    open(lunbiodat,file=tdnam,form='unformatted',status='replace',iostat=ios)
    if (ios /= 0) call exitme(1, 'Cannot open '//trim(tdnam))
    aufdat_bio(2:5)  = ctim(1:4)
    aufdat_bio(6:10) = ctim(6:10)
    write(aufdat_bio(11:11),'(I1.1)') ihour/3
    write(lunbiodat,iostat=ios) aufdat_bio
    if (ios /= 0) call exitme(1,'Could not write header in '//trim(tdnam))

  end subroutine biofile_new

  !=============================================================================

  subroutine biofile_close ()
    implicit none
    close (lunbiodat)
  end subroutine biofile_close

  !=============================================================================

  subroutine biofile_write (ia, iu06, mi, ni, ki, n2d, n3d, mmk, nl,           &
                            cmp, benthos)

    use exits,         only : exitme
    use cmod_miscvars, only : ctim
    use constants,     only : tim

    implicit none

    real(8),    intent(in) :: cmp(1:,0:), benthos(1:,0:)
    integer(4), intent(in) :: ia, iu06, mi, ni, ki, n2d, n3d, mmk(:,0:,0:), nl

    !- local vars --------------------------------------------------------------
    integer(4) :: ios

    if (ia == 1) then
      write(iu06,'(a9,a,a10,a19)') 'write on ',trim(tdnam),' at time: ',ctim

      write(lunbiodat,iostat=ios) ctim,tim
      if (ios /= 0) call exitme(1,'Could not write time stamp in '//trim(tdnam))
    endif

    call wcompr_bio(n2d, n3d, nl, cmp, benthos, mi, ni, ki, mmk,               &
                   eco_i4(ia)%p, ben_i4(ia)%p)

    write(lunbiodat) eco_i4(ia)%p
    write(lunbiodat) ben_i4(ia)%p

  end subroutine biofile_write

  !=============================================================================

  subroutine biofile_next ( iu06 )
    implicit none

    integer(4), intent(in) :: iu06

    call biofile_close ()

    write(iu06,'(a22,a)') 'bio output stored on: ',trim(tdnam)

    call biofile_new ()

  end subroutine biofile_next

  !=============================================================================

  subroutine wcompr_bio(iw2, iw3, n, cmp, benthos, mmx, nmx, kmx, mmk, e_i, b_i)
  !----------------------------------------------------------------------------+
  ! Integer compression of biological fields
  !----------------------------------------------------------------------------+
    use bioparam,   only : Nnorm, Onorm, ntbio, ntben
    use biopermute, only : bio_depermute, eco_o, ben_o

    implicit none
    
    integer(4), intent(in)  :: iw2, iw3, mmx, nmx, kmx, n
    integer(4), intent(in)  :: mmk(:,0:,0:)   ! (1:kmx,0:mmx,0:nmx)
    real(8),    intent(in)  :: cmp(1:,0:)     ! (1:2+ntbio,0:iw3)
    real(8),    intent(in)  :: benthos(1:,0:) ! (1:ntben,0:iw2)
    integer(4), intent(out) :: e_i(1:,0:)     ! (ntbio,0:iw3)
    integer(4), intent(out) :: b_i(1:,0:)     ! (ntben,0:iw2)

    real(8), parameter :: onland    = 99.99e-2_8
    real(8), parameter :: scalebioN = 1.e4_8*Nnorm
    real(8), parameter :: scalebioO = 1.e4_8*Onorm

    integer(4) :: iw, isc

    !  de-permute eco and benthos, and scale before output:
    call bio_depermute( mmx,nmx,kmx,iw2,iw3,ntbio,ntben,n,mmk,cmp,benthos )

    isc = nint(onland*scalebioN)
    e_i(1:ntbio-1,0) = isc
    e_i(ntbio,0)     = nint(onland*scalebioO)
    do iw=1,iw3
      e_i(1:ntbio-1,iw) = nint(eco_o(1:ntbio-1,iw)*scalebioN)
      e_i(ntbio,iw)     = nint(eco_o(ntbio,iw)*scalebioO)
    enddo

    b_i(1:ntben,0) = isc
    do iw=1,iw2
      b_i(1:ntben,iw) = nint(ben_o(1:ntben,iw)*scalebioN)
    enddo

  end subroutine wcompr_bio

  !=============================================================================

  subroutine write_bio_restart(ia, iu06, doclose, mmx, nmx, kmx, iw2, iw3, mmk,&
                               nl, cmp, benthos)

    use bioparam,      only : Nnorm, Onorm, idx_oxy, ntbio, ntben
    use biopermute,    only : bio_depermute, eco_o, ben_o
    use io_subs,       only : io_new_unit
    use exits,         only : exitme
    use constants,     only : aufdat
    use cmod_miscvars, only : ihour,ctim
 
    implicit none

    integer(4), intent(in) :: ia, iu06, iw2, iw3, mmx, nmx, kmx, mmk(:,0:,0:)
    integer(4), intent(in) :: nl
    logical,    intent(in) :: doclose
    real(8),    intent(in) :: cmp(1:,0:)  
    real(8),    intent(in) :: benthos(1:,0:)

    integer(4), save :: restartunit = 0
    integer(4)       :: ios
    character(11)    :: aufdat_bio

    !  on first entry, open the restart data file:
    if (ia == 1) then
      restartunit = io_new_unit()
      open(restartunit,file='bio_restart',form='unformatted',status='replace', &
           iostat=ios)
      if (ios /= 0) call exitme (1, 'Could not open bio_restart' )
      aufdat_bio(1:11) = aufdat(1:11)
      aufdat_bio(2:5)  = ctim(1:4)
      aufdat_bio(6:10) = ctim(6:10)
      write(aufdat_bio(11:11),'(I1.1)') ihour/3
      write(iu06,'(a37,a11)') 'writing restart file bio_restart at: ',aufdat_bio
      write(restartunit) aufdat_bio
    endif

    !  de-permute eco and benthos before output:
    call bio_depermute( mmx,nmx,kmx,iw2,iw3,ntbio,ntben,nl,mmk,cmp,benthos )

    !  scale data before output:
    eco_o(1:ntbio-1,0:iw3) = eco_o(1:ntbio-1,0:iw3)*Nnorm
    eco_o(idx_oxy,  0:iw3) = eco_o(idx_oxy,  0:iw3)*Onorm
    ben_o(1:ntben,0:iw2)   = ben_o(1:ntben,0:iw2)*Nnorm

    !  write data to restart file:
    write(restartunit,iostat=ios) eco_o(1:ntbio,0:iw3), ben_o(1:ntben,0:iw2)
    if (ios /= 0) then
      write(iu06,*) 'I/O problem - iostat is:',ios
      write(iu06,*) '                   area:',ia
      call exitme (2, 'Could not write bio_restart' )
    endif

    !  on last entry, close restart data file:
    if (doclose) then
      close(restartunit)
    endif

  end subroutine write_bio_restart

  !=============================================================================
  subroutine val_benthos(ia, iu06, n2d, benthos, ntben, header)

    use io_subs, only : flush_unit

    implicit none

    integer(4), intent(in) :: ia, iu06, n2d, ntben
    real(8),    intent(in) :: benthos(:,0:)
    logical,    intent(in) :: header

    integer(4) :: ib, iben
    real(8)    :: bmean, fac1, fac2

    real(8), parameter :: zero = 0.0_8, one = 1.0_8

    if (header) write(iu06,'(a26)') 'Benthos validation prints:'
    write(iu06,'(a72)')                                                        &
      '------------------------------------------------------------------------'
    write(iu06,'(a33,i5)') 'Statistics for domain:           ', ia
    
    do iben=1,ntben
      write(iu06,'(a33,i5)') '  Benthos variable:              ', iben
      bmean = zero
      do ib=0,n2d-1
        fac2  = real(ib,8)
        fac1  = one/(fac2 + one)
        bmean = (bmean*fac2 + benthos(iben,ib+1))*fac1
      enddo
      write(iu06,'(a33,f16.10)')           '    Average value:               ',&
                                        bmean
      write(iu06,'(a33,f16.10,5x,f16.10)') '    Min and max:                 ',&
                  minval(benthos(iben,1:)), maxval(benthos(iben,1:))
    enddo
    write(iu06,'(a72)')                                                      &
    '------------------------------------------------------------------------'
    call flush_unit(iu06)

  end subroutine val_benthos
  !=============================================================================

end module biooutput
