module biopermute

  implicit none
  private

  public  :: bio_permute, bio_depermute
  private :: bio_mpi_task_perm, bio_mpi_task_deperm

  real(8), allocatable, save, public :: eco_o(:,:), ben_o(:,:)

contains

!===============================================================================

  subroutine bio_permute ( mi, ni, ki, n2d, n3d, mmk, cmp, ben, nl, nu,nben )
  !  wrapper for permuting eco and benthos arrays as needed

    implicit none

    integer(4), intent(in)    :: mi, ni, ki, n2d, n3d, mmk(:,0:,0:), nl, nu,nben
    real(8),    intent(inout) :: cmp(1:,0:)
    real(8),    intent(inout) :: ben(:,0:)

    integer(4) :: neco

    neco = (nu-nl+1)
    call bio_mpi_task_perm (mi,ni,ki,n2d,n3d,neco,nben,mmk,cmp,ben,nl)

  end subroutine bio_permute

!===============================================================================

  subroutine bio_depermute( mi, ni, ki, n2d, n3d, neco,nben, nl, mmk, cmp, ben )
  !  wrapper for depermuting eco and benthos arrays as needed

    implicit none

    integer(4), intent(in) :: mi, ni, ki, n2d, n3d, neco, nben, mmk(:,0:,0:), nl
    real(8),    intent(in) :: cmp(1:,0:)
    real(8),    intent(in) :: ben(1:,0:)

    logical :: do_alloc

    do_alloc = .true.
    if (allocated(eco_o)) then
      do_alloc = .false.
      if (n3d > size(eco_o(1,:)) - 1) then
        deallocate(eco_o)
        do_alloc = .true.
      endif
    endif
    if (do_alloc) allocate( eco_o(neco,0:n3d) )
    do_alloc = .true.
    if (allocated(ben_o)) then
      do_alloc = .false.
      if (n2d > size(ben_o(1,:)) - 1) then
        deallocate(ben_o)
        do_alloc = .true.
      endif
    endif
    if (do_alloc) allocate( ben_o(nben,0:n2d) )

    call bio_mpi_task_deperm( mi, ni, ki, neco, nben, nl, mmk, cmp, ben)

  end subroutine bio_depermute

!===============================================================================

  subroutine bio_mpi_task_perm(mi,ni,ki,n2d,n3d,neco,nben,mmk,cmp,ben,nl)
  !  mpi task permute eco and benthos arrays to full domain

    implicit none

    integer(4), intent(in)    :: mi,ni,ki,n2d,n3d,neco,nben,mmk(:,0:,0:),nl
    real(8),    intent(inout) :: cmp(1:,0:)
    real(8),    intent(inout) :: ben(1:,0:)

    integer(4) :: i, j, k, mo, mc
    real(8)    :: eco_n(1:neco,1:n3d), ben_n(1:nben,1:n2d)

    !  mo: index in old permutation
    !  mc: index in new permutation (mmk already permuted from cmod_main)
    k  = 1
    mo = 0
    do j=1,ni
      do i=1,mi
        mc = mmk(k,i,j)
        if (mc <= 0) cycle
        mo = mo + 1
        eco_n(1:neco,mc) = cmp(nl:nl-1+neco,mo)
        ben_n(1:nben,mc) = ben(1:nben,mo)
      enddo
    enddo
    do k=2,ki
      do j=1,ni
        do i=1,mi
          mc = mmk(k,i,j)
          if (mc <= 0) cycle
          mo = mo + 1
          eco_n(1:neco,mc) = cmp(nl:nl-1+neco,mo)
        enddo
      enddo
    enddo

    ben(1:nben,1:n2d)       = ben_n(1:nben,1:n2d)
    cmp(nl:nl-1+neco,1:n3d) = eco_n(1:neco,1:n3d)

  end subroutine bio_mpi_task_perm

!===============================================================================

  subroutine bio_mpi_task_deperm(mi, ni, ki, neco, nben, nl, mmk, cmp, ben)
  !  mpi task de-permute eco and benthos arrays to full domain

    use exits, only : exitme

    implicit none

    integer(4), intent(in) :: mi, ni, ki, neco, nben, mmk(:,0:,0:), nl
    real(8),    intent(in) :: cmp(1:,0:)
    real(8),    intent(in) :: ben(1:,0:)

    integer(4) :: i, j, k, mc, mn

    if (.not.allocated(eco_o)) then
      call exitme('Error in bio_mpi_task_deperm: eco_o not allocated')
    endif
    if (.not.allocated(ben_o)) then
      call exitme('Error in bio_mpi_task_deperm: ben_o not allocated')
    endif

    ! now, let's permute data arrays:
    eco_o(1:neco,0) = cmp(nl:nl-1+neco,0)
    mc = 0
    do k=1,ki
      do j=1,ni
        do i=1,mi
          mn = mmk(k,i,j)
          if (mn == 0) cycle
          mc = mc + 1
          eco_o(1:neco,mc) = cmp(nl:nl-1+neco,mn)
        enddo
      enddo
    enddo
    ben_o(1:nben,0) = ben(1:nben,0)
    mc = 0
    k  = 1
    do j=1,ni
      do i=1,mi
        mn = mmk(k,i,j)
        if (mn == 0) cycle
        mc = mc + 1
        ben_o(1:nben,mc) = ben(1:nben,mn)
      enddo
    enddo

  end subroutine bio_mpi_task_deperm

!===============================================================================

end module biopermute
