module biopermute

  implicit none
  private

  public  :: bio_depermute
  private :: bio_mpi_task_perm, bio_mpi_task_deperm

  real(8), allocatable, public :: eco_o(:,:), ben_o(:), pp_o(:)

contains

!===============================================================================

  subroutine bio_depermute ( mi, ni, ki, n2d, n3d, neco, nl, mmk, cmp, ben, pp )
  !  wrapper for depermuting eco and benthos arrays as needed

    implicit none

    integer(4), intent(in) :: mi, ni, ki, n2d, n3d, neco, mmk(:,0:,0:), nl
    real(8),    intent(in) :: cmp(1:,0:)
    real(8),    intent(in) :: ben(0:)
    real(8),    intent(in) :: pp(0:)

    logical :: do_alloc

    do_alloc = .true.
    if (allocated(eco_o)) then
      do_alloc = .false.
      if (n3d > size(eco_o(1,:)) - 1) then
        deallocate(eco_o)
        do_alloc = .true.
      endif
    endif
    if (do_alloc) allocate( eco_o(neco,0:n3d) )
    do_alloc = .true.
    if (allocated(ben_o)) then
      do_alloc = .false.
      if (n2d > size(ben_o(:)) - 1) then
        deallocate(ben_o)
        do_alloc = .true.
      endif
    endif
    if (do_alloc) allocate( ben_o(0:n2d) )

    do_alloc = .true.
    if (allocated(pp_o)) then
      do_alloc = .false.
      if (n3d > size(pp_o(:)) - 1) then
        deallocate(pp_o)
        do_alloc = .true.
      endif
    endif
    if (do_alloc) allocate( pp_o(0:n3d) )
    do_alloc = .true.

    call bio_mpi_task_deperm( mi, ni, ki, neco, nl, mmk, cmp, ben, pp)

  end subroutine bio_depermute

!===============================================================================

  subroutine bio_mpi_task_perm ( mi, ni, ki, n2d, n3d, neco, mmk, cmp, ben, nl )
  !  mpi task permute eco and benthos arrays to full domain

    implicit none

    integer(4), intent(in)    :: mi, ni, ki, n2d, n3d, neco, mmk(:,0:,0:), nl
    real(8),    intent(inout) :: cmp(1:,0:)
    real(8),    intent(inout) :: ben(0:)

    integer(4) :: i, j, k, mo, mc
    real(8)    :: eco_n(1:neco,1:n3d), ben_n(1:n2d)

    k  = 1
    mo = 0
    do j=1,ni
      do i=1,mi
        mc = mmk(k,i,j)
        if (mc <= 0) cycle
        mo = mo + 1
        eco_n(1:neco,mc) = cmp(nl:nl-1+neco,mo)
        ben_n(mc)        = ben(mo)
      enddo
    enddo
    do k=2,ki
      do j=1,ni
        do i=1,mi
          mc = mmk(k,i,j)
          if (mc <= 0) cycle
          mo = mo + 1
          eco_n(1:neco,mc) = cmp(nl:nl-1+neco,mo)
        enddo
      enddo
    enddo

    ben(1:n2d)              = ben_n(1:n2d)
    cmp(nl:nl-1+neco,1:n3d) = eco_n(1:neco,1:n3d)
  end subroutine bio_mpi_task_perm

!===============================================================================

  subroutine bio_mpi_task_deperm(mi, ni, ki, neco, nl, mmk, cmp, ben, pp)
  !  mpi task de-permute eco and benthos arrays to full domain

    use exits, only : exitme

    implicit none

    integer(4), intent(in) :: mi, ni, ki, neco, mmk(:,0:,0:), nl
    real(8),    intent(in) :: cmp(1:,0:)
    real(8),    intent(in) :: pp(0:)
    real(8),    intent(in) :: ben(0:)

    integer(4) :: i, j, k, mc, mn

    ! now, let's permute data arrays:
    eco_o(1:neco,0) = cmp(nl:nl-1+neco,0)
    pp_o(0)  = pp(0)
    mc = 0
    do k=1,ki
      do j=1,ni
        do i=1,mi
          mn = mmk(k,i,j)
          if (mn == 0) cycle
          mc = mc + 1
          eco_o(1:neco,mc) = cmp(nl:nl-1+neco,mn)
          pp_o(mc)  = pp(mn)
        enddo
      enddo
    enddo
    ben_o(0) = ben(0)
    mc = 0
    k  = 1
    do j=1,ni
      do i=1,mi
        mn = mmk(k,i,j)
        if (mn == 0) cycle
        mc = mc + 1
        ben_o(mc) = ben(mn)
      enddo
    enddo

  end subroutine bio_mpi_task_deperm

!===============================================================================

end module biopermute


