module bioriver

  use exits,   only : exitme
  use io_subs, only : io_new_unit

  implicit none
  private
  
  public :: bioloadings

  private :: ReadLoadings

contains

!===============================================================================

  subroutine bioloadings ( ia, nlow, nup, bndq )

    use dmi_mpi,        only : mpi_rank
    use dmi_mpi_global, only : iu06
    use cmod_params,    only : nq, narea
    use cmod_arrays,    only : nqlist, nqfile

    implicit none

    integer(4), intent(in)  :: ia, nlow, nup
    real(8),    intent(out) :: bndq(:,:,0:)  

    integer(4),    save :: lun, ntc
    integer(4)          :: ios, iq, iqa
    character(1)        :: c1
    character(2)        :: c2
    character(4),  save :: ifmt
    character(11), save :: cfmt
    character(16)       :: qlogfile
    logical,       save :: firsttime = .true.

    real(8), parameter :: zero = 0.0_8


    if (firsttime) then
      firsttime = .false.
      lun = io_new_unit()
      write (qlogfile,'(a,i3.3)') 'loadings.out.',mpi_rank
      open(lun,file=qlogfile,status='unknown',iostat=ios)
      if (ios /= 0) call exitme (1,'input error - could open '//qlogfile)

      ntc = nup - nlow + 1
      if (ntc < 10) then
        ifmt = '(i1)'
        write(c1,ifmt) ntc
        cfmt = '(i3, '//c1//'f7.1)'
      elseif (ntc < 100) then
        ifmt = '(i2)'
        write(c2,ifmt) ntc
        cfmt = '(i3,'//c2//'f7.1)'
      else 
        call exitme (1,'dmi-ergom not prepared for so many tracers')
      endif
    endif

    ! FIXME: pls, come up with a better climatology, especially for 
    !        fall-back during operational runs, but until then:
    ! better safe than sorry:
    bndq(nlow:nup,:,0:) = zero

    ! read the data:
    if (nq(ia) > 0) then
      write(iu06,'(a32,i3)') 'Reading river loadings for area ',ia
      call ReadLoadings( ia, nqfile(ia), nlow, nup, bndq )

      ! shuffle the read source data:
      do iq=1,nqfile(ia)
        iqa = nqlist(ia)%p(iq)
        if (iqa > 0) bndq(nlow:nup,1,iqa) = bndq(nlow:nup,1,iq)
      enddo

      ! print loadings:
      write(lun,'(/,'' Loadings in area '',i3)') ia
      do iq=1,nqfile(ia)
        iqa = nqlist(ia)%p(iq)
        if (iqa > 0) write(lun,cfmt) iqa,bndq(nlow:nup,1,iqa)
      enddo
    else
      write(lun,'(/,'' No rivers in area '',i3)') ia
    endif

    ! clean up:
    if (ia == narea) close (lun)

  end subroutine bioloadings

!===============================================================================

  subroutine ReadLoadings( ia, nr, ncl, ncu, load )
    ! straight forward read of nutrient loadings from an ascii file

    use cmod_date,      only : jultimea
    use constants ,     only : TIM
    use bioparam,       only : Nnorm, Onorm, idx_oxy,ntbio
    use dmi_mpi_global, only : iu06

    implicit none

    integer(4), intent(in) :: ia, nr, ncl, ncu
    real(8), intent(inout) :: load(:,:,0:)    ! (1:ntracers,1:kmax,0:nr)

    character(1)  :: c1
    character(2)  :: c2
    character(3)  :: c3
    character(4)  :: iformat
    character(13) :: cformat
    character(11) :: fname
    character(19) :: cdat
    integer(4)    :: lun, iq, ic, ios, nrc, rclbioloads,ii
    real(8)       :: timfl

    real(8), parameter :: twenty = 20.0_8

    !- set up file name, format, etc ... ---------------------------------------
    nrc = nr*(ncu-ncl+1)
    if (nrc < 1) then
      return
    elseif (nrc < 10) then
      iformat = '(i1)'
      write(c1,iformat) nrc
      cformat = '(a19,  '//c1//'f8.1)'
    elseif (nrc < 100) then
      iformat = '(i2)'
      write(c2,iformat) nrc
      cformat = '(a19, '//c2//'f8.1)'
    elseif (nrc < 1000) then
      iformat = '(i3)'
      write(c3,iformat) nrc
      cformat = '(a19,'//c3//'f8.1)'
    else
      ! trap
      call exitme(1,'ERROR: Number of bio sources too large.')
    endif
    if (ia < 10) then
      iformat = '(i1)'
      write(c1,iformat) ia
      fname   = 'bioloads_0'//c1
    elseif (ia < 100) then
      iformat = '(i2)'
      write(c2,iformat) ia
      fname   = 'bioloads_'//c2
    else
      ! trap
      call exitme(1,'ERROR: Number of areas too large.')
    endif

    !- open the file and read it -----------------------------------------------
    lun = io_new_unit()
    rclbioloads = 1024
    ii = 19+nrc*8+1 
    do while (ii > rclbioloads)
      ! adjust record length of bioloads file
      rclbioloads = rclbioloads + 1024
    enddo

    open (lun,file=fname,action='read',status='old',iostat=ios,recl=rclbioloads)
    if (ios /= 0) then
      write(iu06,*) 'Unable to open '//fname//', continuing without'
    else
      read (lun,*)
      read (lun,'(a)') cdat
      call jultimea(cdat,timfl)
      if ( abs(TIM-timfl) <= twenty ) then
        backspace (lun)
        read (lun,cformat) cdat, ((load(ic,1,iq), ic=ncl,ncu), iq=1,nr)
        load(ncl:ncl-2+ntbio,1,:) = load(ncl:ncl-2+ntbio,1,:) / Nnorm
        load(    ncl-1+idx_oxy,1,:) = load(    ncl-1+idx_oxy,1,:) / Onorm
        close (lun)
      else 
        write(iu06,*) "Warning, no river loadings found for this day in ",fname
      endif
    endif
    

  end subroutine ReadLoadings

!===============================================================================

end module bioriver
