module biosolver
  !=============================================================================
  ! Module containg timestepping schemes for the biogeochemical model
  !=============================================================================
  implicit none

  ! public routines
  public :: solvebio

  ! private routines
  private :: empatankar, bisec, f

contains
  subroutine solvebio(eco, deriv, dt,irc,k, ben, benderiv)
    use bioparam, only : ntbio,ntben, idx_oxy, impl
    implicit none

    ! arguments
    integer(4), intent(inout)           :: irc
    integer(4), intent(in)              :: k
    real(8),    intent(in)              :: deriv(0:), dt
    real(8),    intent(inout)           :: eco(:,0:)
    real(8),    intent(inout), optional :: ben(0:)
    real(8),    intent(in),    optional :: benderiv(0:)
       
    ! locals
    integer(4) :: itr, numimpltr, instablei
    real(8)    :: ecotmp(ntbio+ntben-1), derivtmp(ntbio+ntben-1) 
    real(8)    :: instable,ecot,deco
    logical    :: bottom

    ! parameters
    real(8), parameter :: zero = 0.0_8
    real(8), parameter :: eps  = 1.0e-14_8

    if (present(benderiv)) then
      bottom = .true.
    else
      bottom = .false.
    endif

    if (impl) then
      !-------------------------------------------------------------------------
      ! Implicit scheme
      !-------------------------------------------------------------------------
      if (bottom) then
        ! solution fot bottom point
        ecotmp(1:ntbio-1)   = eco(k,1:idx_oxy-1)
        ecotmp(ntbio:)      = ben(1:ntben) ! change oxyg with benthos
        derivtmp(1:ntbio-1) = deriv(1:idx_oxy-1)
        derivtmp(ntbio:)    = benderiv(1:ntben) ! change oxyg with benthos
        numimpltr = ntbio+ntben-1 ! eco + benthos - oxygen   
     else 
        ! solution of pelagic
        ecotmp(1:ntbio-1)   = eco(k,1:idx_oxy-1)
        ecotmp(ntbio)       = zero 
        derivtmp(1:ntbio-1) = deriv(1:idx_oxy-1)
        derivtmp(ntbio)     = zero
        numimpltr = ntbio-1 ! eco - oxygen 
     endif

      call empatankar(ecotmp, derivtmp, numimpltr, dt,irc) ! update tracers
      if (irc /= 0) return

      ! copy values back to eco and ben array
      eco(k,1:ntbio-1) = ecotmp(1:idx_oxy-1)
      if (bottom) then
        do itr=1,ntben
          ben(itr) = ecotmp(ntbio+itr-1)
        enddo
      endif

      ! udate oxyg with an euler step since it can be negative
      eco(k,idx_oxy) = eco(k,idx_oxy) + deriv(idx_oxy)*dt
      
      ! should we have a check here as well?
      instablei = 0

    else 

      !-------------------------------------------------------------------------
      ! Explicit scheme
      !-------------------------------------------------------------------------
      instable = zero
      do itr=1,ntbio-1
        ecot = eco(k,itr)
        deco = deriv(itr)*dt
        eco(k,itr) = ecot + deco
        instable = max(instable,-deco/(ecot+eps))
      enddo
      
      ! oxygen can be negative so no stability check  
      eco(k,idx_oxy) = eco(k,idx_oxy) + deriv(idx_oxy)*dt
      
      !update benthos
      if (bottom) then
        do itr=1,ntben
          ben(itr) = ben(itr) + benderiv(itr)*dt
          instable = max(instable,-benderiv(itr)*dt/(ben(itr)+eps))
        enddo
      endif

      instablei = int(instable) + 1
   endif
 
   if (instablei > 1) then
     ! we will get negative values
     irc = 3
   endif
    
  end subroutine solvebio

  !=============================================================================

  subroutine empatankar(tracer, deriv, numtr, dt, irc)
    implicit none

    ! args
    integer(4), intent(inout) :: irc
    integer(4), intent(in)    :: numtr
    real(8), intent(in)       :: dt,deriv(numtr)
    real(8), intent(inout)    :: tracer(numtr)

    ! local
    real(8)    :: pmin,pmax,pdum,a(numtr),p
    integer(4) :: numnegdt,indnegdt(numtr), itr
    logical    :: pmaxfirst

    integer(4), parameter :: maxit = 50
    real(8), parameter    :: tol   = 0.00001_8
    real(8), parameter    :: zero  = 0.0_8
    real(8), parameter    :: one   = 1.0_8
    real(8), parameter    :: eps   = 1.0e-14_8
    real(8), parameter    :: tinya = 1.0e-15_8

    pmax        = zero
    pmaxfirst   = .true.
    numnegdt    = 0
    indnegdt(:) = 0
    a(:)        = zero

    do itr=1,numtr
      if (deriv(itr) < zero) then
        numnegdt = numnegdt + 1
        indnegdt(numnegdt) = itr

        ! find pmax
        a(itr) = deriv(itr)*dt/(tracer(itr)+eps)

        if (abs(a(itr)) < tinya) then
          pdum = -one/(a(itr)-eps)
        else
          pdum = -one/(a(itr))
        endif

        ! make sure we always go in here at least once
        if ( pdum < pmax .or. pmaxfirst ) then
          pmax = pdum
          pmaxfirst = .false.
        endif
      endif
    enddo

    pmax = min(one,pmax)
    pmin = zero

    if (numnegdt > 0) then
      call bisec(a, numnegdt,indnegdt, pmin, pmax,maxit,tol, p,irc )
      if (irc /= 0) return
    else
      p = one
    endif

    ! update tracers
    do itr=1,numtr
      tracer(itr) = tracer(itr) + deriv(itr)*p*dt
    enddo
  
  end subroutine empatankar

  !============================================================================
  subroutine bisec(a,numnegdt,indnegdt,x1,x2,maxit,eps_step,c,irc )
    !--------
    ! simple bisection routine for finding the root of function f in the 
    ! low-high p-space
    ! remember dg/dp is always negative, thus we can simplify the bisection
    !--------

    use dmi_mpi_global, only : iu06
    
    implicit none

    !args
    real(8), intent(in)       :: a(:), eps_step
    real(8), intent(inout)    :: c, x1, x2
    integer(4), intent(in)    :: numnegdt, indnegdt(:), maxit
    integer(4), intent(inout) :: irc

    ! local
    real(8)    :: rtb,fc, ff,dx
    integer(4) :: k

    real(8), parameter :: zero = 0.0_8
    real(8), parameter :: half = 0.5_8
    real(8), parameter :: ten = 10.0_8
 

    ! Usually the root is located close to the upper bound, make a quick look
    ! around this point
    fc = f(x2,a,numnegdt,indnegdt)
    ff = f(x2-ten*eps_step,a,numnegdt,indnegdt)
    if ( ff > 0 ) then
      x1 = x2-ten*eps_step
    else
      ff = f(x2-ten*ten*eps_step,a,numnegdt,indnegdt)
      if (ff > 0) then
        x1 = x2-ten*ten*eps_step
      else
        ff = f(x1,a,numnegdt,indnegdt)
      endif
    endif
    
    ! Does the bracketing contains the root?
    if (ff*fc >= 0) then
      ! Root is not bracketed'
      irc = 4
      return
    endif

    ! we now the oriantation already since the dg/dp is negative
    dx  = x1-x2
    rtb = x2
    do k=1,maxit
      dx   = dx*half
      c = rtb+dx
      fc = f(c,a,numnegdt,indnegdt)

      if (fc <0 ) rtb = c
      if (abs(dx) < eps_step .or. fc == zero) return
    enddo

    !Implicit bioscheme did not converge'
    irc = 5
  end subroutine bisec

  !============================================================================

  real(8) function f(p,a,numnegdt,indnegdt)
    implicit none

    ! args
    real(8), intent(in)    :: p, a(:)
    integer(4), intent(in) :: numnegdt, indnegdt(:)

    ! local
    integer(4)        :: ii, i
    real(8)           :: y
    real(8),parameter :: one = 1.0_8

    ii = indnegdt(1);
    y = (one+a(ii)*p)
    do i=2,numnegdt
      ii = indnegdt(i)
      y = y * (one+a(ii)*p)
    enddo
    y = y - p

    f = y
  end function f

  !============================================================================

end module biosolver
