module biosun
  implicit none
  private
  
  public :: shortwaverad
contains

subroutine shortwaverad (n2d,ind,xkoor,ykoor,cloudi,hum,casus,swrad)

  !-----------------------------------------------------------------------------
  ! Short wave radiation subroutine for biomod. Based on subroutine q_i
  ! 
  ! Outputs incoming radiation at the surface
  !-----------------------------------------------------------------------------

  !- modules  ------------------------------------------------------------------
  use dmi_omp,    only : domp_get_domain
  use constants,  only : zeitstu, pair, pi

  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in)              :: n2d
  logical, intent(in)                 :: casus(0:)                      !(0:n2d)
  integer(4), intent(in)              :: ind(:,:)                       !(2,n2d)
  real(8), intent(in)                 :: xkoor(:), ykoor(:)             !(n) (m)
  real(8), intent(in),  dimension(0:) :: hum, cloudi                    !(0:n2d)
  real(8), intent(inout), dimension(0:) :: swrad                        !(0:n2d)
  
  !
  !  Formel von Zillman, siehe Maykut,
  !       Introduction to Ice in the Polar Oceans, 1985, S.80,81
  !
  real(8), parameter :: twopi     = 2.0_8*pi, zero = 0.0_8, half = 0.5_8
  real(8), parameter :: twentyfour = 24.0_8, eighty = 80.0_8, milli = 0.001_8
  real(8), parameter :: one = 1.0_8, a1 = 0.622_8, a2 = 0.378_8, a3 = 1.2_8
  real(8), parameter :: a4 = 0.045_8, a5 = 0.6_8
  real(8), parameter :: q0        = 1350.0_8
  real(8), parameter :: sindek    = 0.39870_8    ! sindek = sin (23.5 grad)
  real(8), parameter :: albedo0   = 0.070_8
  real(8), parameter :: albedoi   = 0.600_8
  real(8), parameter :: ddayspryr = 1.0_8/365.2422_8

  real(8)    :: phi, rlam, albedo, hd24, dekl, ea, sa, sw, sd, cd, ph
  real(8)    :: humid, cloud
  integer(4) :: nsrf, n2dl, n2du

!$OMP MASTER
  swrad(0) = zero
!$OMP END MASTER

  call domp_get_domain(1, n2d, n2dl, n2du)

  hd24 = zeitstu/twentyfour
  dekl = asin( sindek*sin((hd24-eighty)*ddayspryr*twopi) )

  sd = sin(dekl)
  cd = cos(dekl)
 
  ph = (hd24-half)*twopi
  
  do nsrf=n2dl, n2du

    !- albedo:
    if (casus(nsrf)) then
      albedo = albedo0
    else
      albedo = albedoi
    endif

    !- time-center humidity and cloud cover:
    humid = hum(nsrf)*milli
    cloud = cloudi(nsrf)

    !- lat/lon coords:
    phi  = ykoor(ind(1,nsrf))
    rlam = xkoor(ind(2,nsrf))

    !- sun angle:
    sa = sin(phi)*sd + cos(phi)*cd*cos(ph + rlam)
    sa = max(sa,zero)
    
    !- short wave radiation:
    ea = pair*humid/(a1 + a2*humid)
    sw = q0*sa*sa/(a3*sa + (one+sa)*ea*milli + a4) 
    swrad(nsrf) = sw*(one - a5*cloud)*(one - albedo)

  enddo

end subroutine shortwaverad

end module biosun
