module biovmove 

  implicit none
  private
  
  public :: vmove, sedimentation_resuspension, drying_case

contains
  
  subroutine vmove(kb,hi,zooa,wzoo,dt,irc)
    
    !--------------------------------------------------------------------------!
    !
    !   vmove:
    !
    !   calculate vertical movement of plankton
    !
    !   i ke                   number of layers
    !
    !   input:
    !
    !   i kb                   No of active layers at this wet surface point
    !   r hi(ke)               layer thicknesses
    !   r zooa(1:ke)           plankton, old timestep
    !   r wzoo                 plankton vertical eigenvelocity
    !   r dt                   time step
    !
    !   returns:
    !
    !
    !--------------------------------------------------------------------------!
    !
    !  Method: A conservative and stable scheme for determining the biological 
    !          redistribution which, assuming a constant vertical velocity w, 
    !          goes like this:
    !    
    !            zoo_new(k) = zoo_old(k) + (zoo_in - zoo_out)/h(k)
    !
    !          where, according to simple upwinding:
    !
    !            zoo_out = dt*abs(w)*zoo_old(k)    for all w
    !
    !          and 
    !
    !            zoo_in  = dt*w*zoo_old(k+1)       if  w >= 0
    !            zoo_in  = dt*abs(w)*zoo_old(k-1)  if  w <  0
    !
    !          Another formulation, in terms of transports through the upper
    !          and the lower cell faces, goes like this:
    !
    !            zoo_new(k) = zoo_old(k) + (zt_up + zt_lo)/h(k)
    !
    !          where
    !
    !            zt_up = -dt*w*zoo_old(k)   for w >= 0
    !            zt_lo =  dt*w*zoo_old(k+1) for w >= 0
    !
    !            zt_up =  dt*abs(w)*zoo_old(k-1) for w < 0
    !            zt_lo = -dt*abs(w)*zoo_old(k)   for w < 0
    !
    !          or:
    !
    !            zt_up = -zflux(k-1)  
    !            zt_lo =  zflux(k)
    !            zflux(k) = dt*w*zoo_old(k+1) for w >= 0
    !            zflux(k) = dt*w*zoo_old(k)   for w <  0
    !
    !--------------------------------------------------------------------------!

    implicit none

    real(8),    intent(in)    :: dt
    integer(4), intent(in)    :: kb
    integer(4), intent(inout) :: irc
    real(8),    intent(in)    :: hi(:)              ! (1:ke)
    real(8),    intent(inout) :: zooa(1:)           ! (1:ke)
    real(8),    intent(in)    :: wzoo

    real(8)    :: wdt, zoo_old, zoo_new, maxcfl
    integer(4) :: k,ndt

    !  convert m/d to m/s
    real(8), parameter :: hnull = 0.1_8
    real(8), parameter :: zero  = 0.0_8

    ! skip trivial case:
    if (kb <= 1) return

    !--- compute the mass flow through the bottom of layer k -------------------
    wdt = wzoo*dt
    maxcfl = zero

    if (wzoo >= 0) then
      zooa(1) = zooa(1) + wdt*zooa(2)/max(hi(1),hnull)
      do k=2,kb-1
        maxcfl  = max(wdt/hi(k),maxcfl)
        zooa(k) = zooa(k) + wdt*(zooa(k+1) - zooa(k))/hi(k) 
      enddo
      maxcfl  = max(wdt/hi(kb),maxcfl)
      zooa(kb) = zooa(kb) - wdt*zooa(kb)/hi(kb)
    else ! sinking
      maxcfl  = max(-wdt/max(hi(1),hnull),maxcfl)
      zoo_old = zooa(1)
      zooa(1) = zooa(1) + wdt*zooa(1)/max(hi(1),hnull)    
      do k=2,kb-1
        maxcfl  = max(-wdt/hi(k),maxcfl)
        zoo_new = zooa(k) + wdt*(zooa(k) - zoo_old)/hi(k)
        zoo_old = zooa(k)
        zooa(k) = zoo_new
      enddo
      zooa(kb) = zooa(kb) - wdt*zoo_old/hi(kb)    
    endif

    ! did we have stability issues?
    ndt = int(maxcfl) + 1
    if (ndt > 1) then
      ! Yes, write an error flag'
      irc = 1
    endif
 
  end subroutine vmove

  !============================================================================!

  subroutine sedimentation_resuspension(hbot,dt, velkb,eco,ben)
    !- modules
    use bioparam, only : reslim, wdz, resuspension, resusprate, ntben,         &
                         idx_dets, idx_det, ntben
    
    implicit none

    !- args
    real(8),    intent(in)    :: velkb, hbot, dt
    real(8),    intent(inout) :: eco(0:), ben(0:)

    !- local paramters
    real(8),    parameter :: one     = 1.0_8
    real(8),    parameter :: sqrtcd0 = 0.4_8
    real(8),    parameter :: z0      = 0.0002_8

    !- local variables
    integer(4) :: i, detidx(2)
    real(8)    :: z0hbot, vstar


    detidx(1) = idx_det
    detidx(2) = idx_dets

    z0hbot  = abs( one+log(z0/hbot) )
    vstar   = sqrtcd0/z0hbot*velkb

    ! Use implicit euler
    if (vstar > reslim .and. resuspension) then
      ! resuspension
      do i=1,ntben
        ben(i)         = ben(i) / (one + dt*resusprate)
        eco(detidx(i)) = eco(detidx(i)) + dt*resusprate*ben(i)/hbot
      enddo
    else
      ! deposition
      do i=1,ntben
        eco(detidx(i)) = hbot*eco(detidx(i)) / (hbot - dt*wdz)
        ben(i)         = ben(i) - dt*wdz*eco(detidx(i))
      enddo
    endif
  end subroutine sedimentation_resuspension

  !============================================================================!
  subroutine drying_case(hbot, eco, ben)
    !- modules
    use bioparam, only : idx_dets, idx_det, ntben

    implicit none

    !- args
    real(8), intent(in)    :: hbot
    real(8), intent(inout) :: eco(1:), ben(1:)

    !- local variables
    integer(4) :: i, detidx(ntben)
    integer(4), parameter :: zero = 0.0_8
    integer(4), parameter :: half = 0.5_8

    detidx(1) = idx_det
    detidx(2) = idx_dets
    
    !- half the detrtus becomes benthos (remainder goes to terrestial ecosystem)
    do i=1,ntben
      ben(i)         = ben(i) + eco(detidx(i))*hbot*half
      eco(detidx(i)) = zero
    enddo
  end subroutine drying_case
end module biovmove
