module biovmove 

  implicit none
  private
  
  public :: vmove

contains
  
  subroutine vmove(kb,hi,zooa,zoon,wzoo,dt)    
    !--------------------------------------------------------------------------!
    !
    !   vmove:
    !
    !   calculate vertical movement of plankton
    !
    !   i ke                   number of layers
    !
    !   input:
    !
    !   i kb                   No of active layers at this wet surface point
    !   r hi(ke)               layer thicknesses
    !   r zooa(1:ke)           plankton, old timestep
    !   r wzoo                 plankton vertical eigenvelocity
    !   r dt                   time step
    !
    !   returns:
    !
    !   r zoon(1:ke)           plankton, new timestep
    !
    !
    !--------------------------------------------------------------------------!
    !
    !  Method: A conservative and stable scheme for determining the biological 
    !          redistribution which, assuming a constant vertical velocity w, 
    !          goes like this:
    !    
    !            zoo_new(k) = zoo_old(k) + (zoo_in - zoo_out)/hz(k)
    !
    !          where, according to simple upwinding:
    !
    !            zoo_out = dt*abs(w)*zoo_old(k)    for all w
    !
    !          and 
    !
    !            zoo_in  = dt*w*zoo_old(k+1)       if  w >= 0
    !            zoo_in  = dt*abs(w)*zoo_old(k-1)  if  w <  0
    !
    !          Another formulation, in terms of transports through the upper
    !          and the lower cell faces, goes like this:
    !
    !            zoo_new(k) = zoo_old(k) + (zt_up + zt_lo)/hz(k)
    !
    !          where
    !
    !            zt_up = -dt*w*zoo_old(k)   for w >= 0
    !            zt_lo =  dt*w*zoo_old(k+1) for w >= 0
    !
    !            zt_up =  dt*abs(w)*zoo_old(k-1) for w < 0
    !            zt_lo = -dt*abs(w)*zoo_old(k)   for w < 0
    !
    !          or:
    !
    !            zt_up = -zflux(k-1)  
    !            zt_lo =  zflux(k)
    !            zflux(k) = dt*w*zoo_old(k+1) for w >= 0
    !            zflux(k) = dt*w*zoo_old(k)   for w <  0
    !
    !--------------------------------------------------------------------------!

    implicit none

    real(8),    intent(in)  :: dt
    integer(4), intent(in)  :: kb
    real(8),    intent(in)  :: hi(:)              ! (1:ke)
    real(8),    intent(in)  :: zooa(1:)           ! (1:ke)
    real(8),    intent(out) :: zoon(1:)           ! (1:ke)
    real(8),    intent(in)  :: wzoo

    real(8)    :: wdt
    integer(4) :: k

    !  convert m/d to m/s
    real(8), parameter :: mpd2mps = 1.0_8/(24.0_8*60.0_8*60.0_8)
    real(8), parameter :: hnull = 0.1_8

    ! skip trivial case:
    if (kb <= 1) return

    !--- compute the mass flow through the bottom of layer k -------------------
    wdt = mpd2mps*wzoo*dt

    if (wzoo >= 0) then
      zoon(1) = zooa(1) + wdt*zooa(2)/max(hi(1),hnull)
      do k=2,kb-1
        zoon(k) = zooa(k) + wdt*(zooa(k+1) - zooa(k))/hi(k) 
      enddo
      zoon(kb) = zooa(kb) - wdt*zooa(kb)/hi(kb)
    else
      zoon(1) = zooa(1) + wdt*zooa(1)/max(hi(1),hnull)    
      do k=2,kb-1
        zoon(k) = zooa(k) + wdt*(zooa(k) - zooa(k-1))/hi(k)
      enddo
      zoon(kb) = zooa(kb) - wdt*zooa(kb-1)/hi(kb)    
    endif

  end subroutine vmove

  !============================================================================!

end module biovmove
