module biozoo
  ! ----------------------------------------------------------------------------
  ! A module containing the zooplankton dynamics 
  !
  ! The process rates are collected on the zoorates tape and return to the 
  ! biomod
  !
  ! Options:
  !
  ! if zpmodel == true:  then using two stage zooplankton model (DMU version)
  ! if zpmodel == false: using the original zooplankton model in ergom
  !
  ! output: zrates
  ! 
  ! Author: Lars Jonasson
  !-----------------------------------------------------------------------------
  implicit none
  private

  type zoorates
    real(8) :: ldzoo
    real(8) :: lnzoo
    real(8) :: lmez
    real(8) :: lmiz
    real(8) :: grazdia
    real(8) :: grazflag
    real(8) :: grazcyan
    real(8) :: mezgrazdia 
    real(8) :: mezgrazflag
    real(8) :: mezgrazcyan
    real(8) :: mezgrazmiz
    real(8) :: mizgrazdia 
    real(8) :: mizgrazflag
    real(8) :: mizgrazcyan
  end type zoorates
  public :: zoorates
  
  ! public routines
  public :: zoodynamics  

contains

!===============================================================================

  subroutine zoodynamics(temp,dia,flag,cyano,mezoo,mizoo, hypoxia, zpmodel,    &
                         zrate, ierr)
    !---------------------------------------------------------------------------
    ! A routine for calculating the zooplankton rates
    ! The zpmodel is based on DMU-ergom
    !
    ! Author: Lars Jonasson, DMI
    !---------------------------------------------------------------------------

    !- Modules 
    use bioparam, only : toptz, miztk, mizgraz,mizgrazf,mezgraz, mezgrazf, nue,&
                         mezprefmiz,zcl1,mezcl, ldmez, lnmez,ldmiz, lnmiz, iv, &
                         sigma_b, mizprefdia, mizpreffla, mizprefcya, graz,    &
                         mezprefdia, mezpreffla, mezprefcya, mez_tscale, meztk
    !- Directives
    implicit none

    !- Arguments
    logical, intent(in)         :: hypoxia, zpmodel
    integer(4), intent(inout)   :: ierr
    real(8), intent(in)         :: temp, mezoo, mizoo, dia, flag, cyano
    type(zoorates), intent(out) :: zrate

    ! Parameters
    real(8), parameter :: zero=0.0_8, half = 0.5_8, one=1.0_8, two=2.0_8
    real(8), parameter :: cnst = 2.7183_8, eps = 0.0000000001_8, ten=10.0_8

    ! Locals
    real(8) :: tldmez, tldmiz, mezgt, mezgscal, zz0, foodmiz, foodmiz2, foodmez
    real(8) :: foodmez2, mizgrazf2, mezgrazf2, mizgt,mizgscal, tlnmiz, tlnmez
    real(8) :: meztotgraz, miztotgraz
    real(8) :: temp2, toptz2

    ! Initialize zrates to zero
    zrate%mezgrazdia  = zero
    zrate%mezgrazflag = zero
    zrate%mezgrazcyan = zero
    zrate%mezgrazmiz  = zero
    zrate%mizgrazdia  = zero
    zrate%mizgrazflag = zero
    zrate%mizgrazcyan = zero
    zrate%grazdia     = zero 
    zrate%grazflag    = zero
    zrate%grazcyan    = zero
    zrate%ldzoo       = zero
    zrate%lnzoo       = zero
    zrate%lmez        = zero
    zrate%lmiz        = zero
 
    if (hypoxia .and. zpmodel) then
      ! High mortality
      tldmez = ten*sigma_b*mezoo
      tldmiz = ten*sigma_b*mizoo
      zrate%ldzoo = tldmiz + tldmez
      
      ! total zoo loss
      zrate%lmez = tldmez
      zrate%lmiz = tldmiz
      
      ! No respiration or grazing so these are kept zero from above
      return
    elseif (hypoxia .and. (.not. zpmodel)) then ! orig zoo model and hypoxia
      ! High mortality
      zrate%ldzoo = ten*sigma_b*mezoo
      tldmez = ten*sigma_b*mezoo
      
      zrate%ldzoo = tldmez
      ! total zoo loss
      zrate%lmez  = tldmez

      ! No respiration or grazing so these are kept zero from above as well
      ! as all miz rates since they are not used
      return
    endif

    ! Help variables
    toptz2    = cnst/(toptz*toptz)
    temp2     = temp*temp
 
    if (zpmodel) then 
      !-------------------------------------------------------------------------
      ! DMU zooplankton model
      !-------------------------------------------------------------------------

      ! Note that the fraction (ldmez_2 + lnmez_1) of the grazing is lost due to
      ! active respiration, egestion and excretion...same for miz

      ! More help variables
      mezgrazf2 = mezgrazf*mezgrazf
      mizgrazf2 = mizgrazf*mizgrazf
      foodmez   = mezprefdia*dia + mezpreffla*flag + mezprefcya*cyano +        &
                  mezprefmiz*mizoo
      foodmez2  = foodmez*foodmez
      foodmiz   = mizprefdia*dia + mizpreffla*flag + mizprefcya*cyano
      foodmiz2  = foodmiz*foodmiz

      ! Grazing temperature dependece
      mezgt = half + mez_tscale*temp2/(meztk+temp2)
      mizgt = one  + temp2/(miztk+temp2)     

      ! food dependent grazing
      ! these have been divided by food ! mezfgrazf2 low compared to paper
      mezgscal = mezgraz*foodmez/(foodmez2+mezgrazf2)
      mizgscal = mizgraz*foodmiz/(foodmiz2+mizgrazf2)

      ! total grazing
      meztotgraz = mezgscal*mezgt
      miztotgraz = mizgscal*mizgt

      ! grazing rates
      zrate%mezgrazdia  = mezprefdia*meztotgraz*mezoo
      zrate%mezgrazflag = mezpreffla*meztotgraz*mezoo
      zrate%mezgrazcyan = mezprefcya*meztotgraz*mezoo
      zrate%mezgrazmiz  = mezprefmiz*meztotgraz*mezoo

      zrate%mizgrazdia  = mizprefdia*miztotgraz*mizoo
      zrate%mizgrazflag = mizpreffla*miztotgraz*mizoo
      zrate%mizgrazcyan = mizprefcya*miztotgraz*mizoo

      ! Total grazing on phytoplankton
      zrate%grazdia  = zrate%mezgrazdia + zrate%mizgrazdia 
      zrate%grazflag = zrate%mezgrazflag + zrate%mizgrazflag
      zrate%grazcyan = zrate%mezgrazcyan + zrate%mizgrazcyan

      ! zoo plankton loss rates (mortality)
      tldmez = ldmez*mezoo + mezcl*mezoo*mezoo
      tldmiz = ldmiz*mizoo
      zrate%ldzoo = tldmez + tldmiz

      ! zoo plankton loss rates (Basal- and active respiration for oxygen;
      ! excretion for nutrients)
      tlnmez = lnmez*meztotgraz*foodmez*mezoo
      tlnmiz = lnmiz*miztotgraz*foodmiz*mizoo
      zrate%lnzoo = tlnmez + tlnmiz

      ! total zoo loss
      zrate%lmez = tlnmez + tldmez
      zrate%lmiz = tlnmiz + tldmiz
    else
      !-------------------------------------------------------------------------
      ! Original zooplankton model (1 tracer)
      !-------------------------------------------------------------------------
      foodmez = dia+flag+cyano

      ! food dependent grazing
      if (foodmez > eps) then
        mezgscal = graz*(one - exp(-iv*foodmez*foodmez))/foodmez
      else
        ! expanding (one-exp(-iv*phyto*phyto))/phyto using L'hopital's rule
        mezgscal = iv*foodmez
      endif

      zz0 = mezoo*mezoo*zcl1
      ! Grazing temperature dependece
      mezgt = one + toptz2*temp2*exp(one-two*temp/toptz)

      ! mesozooplankton grazing rates
      zrate%mezgrazdia   = mezgscal*mezgt*mezoo
      zrate%mezgrazflag  = zrate%mezgrazdia
      zrate%mezgrazcyan  = half*zrate%mezgrazdia
      
      ! Total grazing on phytoplankton
      zrate%grazdia   = zrate%mezgrazdia
      zrate%grazflag  = zrate%mezgrazflag
      zrate%grazcyan  = zrate%mezgrazcyan

      ! mesozooplankton egestion and mortality
      zrate%ldzoo = sigma_b*zz0

      ! mesozooplankton loss rates (egestion and respiration)
      zrate%lnzoo = nue*zz0

      ! total zoo loss
      zrate%lmez = zrate%lnzoo+zrate%ldzoo

      ! All miz rates are zero from abobe
    endif

  end subroutine zoodynamics


end module biozoo
