module botfricmaps
  ! Container for various bottom friction maps.
  !   ... so far, only a very simple map of relative magnitudes is implemented.
  !
  !   How to:
  !
  !   In options.nml:
  !     &optionlist
  !       simplebfmap = .true.
  !     /
  !
  !   In botfricmaps.nml e.g.:
  !     &nsegnml
  !       nsegments = 0, 2, 0, 0
  !     /
  !     &ijsegnml
  !       ijsegment = 192, 208, 230, 396
  !       bfricfac  = 3.0
  !     /
  !     &ijsegnml
  !       ijsegment = 209, 240, 233, 396
  !       bfricfac  = 3.0
  !     /


  use cmod_params, only : maxareas

  implicit none
  private

  interface get_simple_bf_map
    module procedure read_nsegments ! starting point
    module procedure return_flag
    module procedure assign_map
    module procedure get_value
  end interface

  public  :: get_simple_bf_map
  private :: read_nsegments, return_flag, assign_map, get_value

  integer(4), save, private :: nsegments(maxareas), lun, ialast
  integer(4),       private :: ios

contains

  !-----------------------------------------------------------------------------

  subroutine read_nsegments (narea)
    use exits,          only : exitme
    use io_subs,        only : io_new_unit
    use dmi_mpi_global, only : iu06

    implicit none

    integer(4), intent(in) :: narea

    integer(4) :: ia

    namelist /nsegnml/ nsegments

    ialast = narea
    nsegments(:) = 0

    lun = io_new_unit()
    open (unit=lun, file='botfricmaps.nml', status='old', iostat=ios)
    if (ios /= 0) then
      !  apply default settings:
      write(iu06,'(a71)')                                                      &
      'Warning: File botfricmaps.nml not found. Continue with default settings'
    else
      read (unit=lun, nml=nsegnml, iostat=ios)
      if (ios /= 0) then
        write(iu06,*) 'Error: Cannot read basic numbers from botfricmaps.nml.'
        write(iu06,*) 'Error: The iostat failure was: ', ios
        call exitme(1)
      endif
      do ia=1,narea
        if (nsegments(ia) < 0) call exitme(1,'Invalid No. of segments')
      enddo
    endif
  end subroutine read_nsegments

  !-----------------------------------------------------------------------------

  subroutine return_flag (ia,flag)
    implicit none
    integer(4), intent(in)    :: ia
    logical,    intent(inout) :: flag
    flag = (nsegments(ia) > 0)
    !-   That's all, folks!
  end subroutine return_flag

  !-----------------------------------------------------------------------------

  subroutine assign_map (ia,iw2,mmx,nmx,mmk,sbfmap)
    use constants,      only : zero, one
    use exits,          only : exitme
    use dmi_mpi_global, only : iu06

    implicit none

    integer(4), intent(in)    :: ia, iw2, mmx, nmx, mmk(1:,0:,0:)
    real(8),    intent(inout) :: sbfmap(0:)

    integer(4) :: is, i, j, n, ijsegment(4)
    real(8)    :: bfricfac = one

    namelist /ijsegnml/ ijsegment, bfricfac

    ! relative factor 1 everywhere per default:
    sbfmap(0:) = one

    do is=1,nsegments(ia)
      read (unit=lun, nml=ijsegnml, iostat=ios)
      if (ios /= 0) then
        write(iu06,*) 'Error: Cannot read ij-segment from botfricmaps.nml.'
        write(iu06,*) 'Error: The iostat failure was: ', ios
        call exitme(1)
      endif    
      if (bfricfac <= zero) call exitme(1,'Invalid bottom friction factor')
      do j=max(1,ijsegment(3)),min(ijsegment(4),nmx)
        do i=max(1,ijsegment(1)),min(ijsegment(2),mmx)
          n = mmk(1,i,j)
          if (n > 0 .and. n <= iw2) sbfmap(n) = bfricfac
        enddo
      enddo
    enddo

    if (ia == ialast) close(lun)
  end subroutine assign_map

  !-----------------------------------------------------------------------------

  subroutine get_value(ia,n,bfmap,rt,fricfac)
    use constants, only : half
    implicit none

    integer(4), intent(in)    :: ia, n
    real(8),    intent(in)    :: bfmap(0:), rt
    real(8),    intent(inout) :: fricfac

    if (nsegments(ia) > 0) fricfac = half*rt*bfmap(n)
  end subroutine get_value

  !-----------------------------------------------------------------------------

end module botfricmaps
