module boundary
  use constants, only : zero, one, two, half
  implicit none
  private

  public  :: bndbclz, bndiceno, bndicez, bndstcl, bndstz, bnduvice, bnduvst,   &
             bndzice, bndzst, rand_st, bndzcorr, daybnd, vobcalc, vabcalc,     &
             gradiz, zrano, densbno, zetar, ice_vel, checkopenbnd, compbno,    &
             masseqs_bnd_dyn, zbnd_rbc

   interface gradiz
     module procedure gradiz_dyn
     module procedure gradiz_w
     module procedure gradiz_q
   end interface

   private :: gradiz_dyn, gradiz_w, gradiz_q

  real(8), parameter, private :: hpm  = 720.0_8, cm2m = 0.01_8, mm2m = 0.001_8
contains

!===============================================================================

  subroutine bndbclz (nzno, zetac2, jstuna, zetacno, krz, NBND1, NBND2, task)
  
    use exits,   only : exitme
    use io_subs, only : io_new_unit
  
    implicit none
  
    integer(4), intent(in)    :: nzno,jstuna,task
    integer(4), intent(in)    :: krz(:,0:)
    real(8), intent(inout)    :: zetac2(0:)
    real(8), intent(inout)    :: zetacno(0:)
    integer(4), intent(inout) :: NBND1, NBND2
  
    character(70) :: mes
    character(20) :: fnam
    integer(4)    :: lun, mon, n, m1, m2, ios
    real(8)       :: pqr, w1, w2
    real(8), allocatable :: rwbclzcor_a(:,:)
  
    !############################################
    ! HARD CODED weights:
    real(8), parameter :: A =  zero, B = one
    !############################################
    integer(4)         :: nz

    if (nzno <= 0) return
  
    if (task == 1) then
      ! relaxation:
      zetacno(NBND1:NBND2) = A*zetacno(NBND1:NBND2) + B*zetac2(NBND1:NBND2)
  
    else
  
      !- get North boundary coordinates ----------------------------------------
      NBND1 = nzno + 1
      NBND2 = 0
      do nz=1,nzno
        if (krz(3,nz) == 2) then
          NBND1 = min(NBND1, nz)
          NBND2 = max(NBND2, nz)
        endif
      enddo
  
      !- allocate and initialise -----------------------------------------------
      allocate( rwbclzcor_a(1:nzno,12) )
      rwbclzcor_a(:,:) = zero
      zetac2(:)        = zero
  
      !- read climatology baroclinic elevation ---------------------------------
      fnam = 'rwzkbcl'
      lun  = io_new_unit()
      open (lun,file=fnam,iostat=ios,pad='NO')
      if (ios /= 0) then
        write (mes, '(1x,"input error - could not open, ",a20)') trim(fnam)
        call exitme(mes)
      endif
  
      do mon=1,12
        read(lun,*,iostat=ios) (rwbclzcor_a(n,mon),n=NBND1,NBND2)
        if (ios /= 0) then
          write (mes,                                                          &
                 '(1x,"input error - could not read or read past eof, ",a20)') &
                 trim(fnam)
          call exitme(mes)
        endif
      enddo
  
      close(lun)
  
      !- interpolation to actual day -------------------------------------------
      mon = (jstuna+720/2)/720       !       mon=0,1,2,...,11,12
      pqr = real(jstuna-mon*720,8)/hpm
      w1  = (half - pqr)*cm2m
      w2  = (half + pqr)*cm2m
      m1  = mod(mon+11,12)+1
      m2  = mod(mon+12,12)+1
  
      zetac2(NBND1:NBND2) = w1*rwbclzcor_a(NBND1:NBND2,m1)                     &
                          + w2*rwbclzcor_a(NBND1:NBND2,m2)
  
      deallocate( rwbclzcor_a )
  
    endif
  
  
  end subroutine bndbclz

!===============================================================================

  subroutine bndiceno(nzno,iceobc,jstuna)
   
     use constants, only : nicepar
     use exits,     only : exitme
     use io_subs ,  only : io_new_unit
     implicit none
  
     integer(4), intent(in)  :: jstuna, nzno
     real(8),    intent(out) :: iceobc(1:,0:)
  
     character(7) :: fnamed, fnamec, fnamsd
     integer(4)   :: lunec, luned, lunsd, mon, monat, n, nbp, ios
                     ! NOTE: monat and nbp are only read in but never used
     real(8)      :: pqr, hp, hm
     real(8), dimension(0:nzno,12) :: rwzecno_a, rwzedno_a, rwzsdno_a
   
     !- intrinsic  -------------------------------------------------------------
     intrinsic :: real
  
     iceobc(1:nicepar,0) = zero

     if (nzno > 0) then
       fnamed = 'rwzedno'
       luned = io_new_unit()
       open (luned,FILE=fnamed,iostat=ios)
       if (ios /= 0) call exitme(1,'input error - cannot open'//trim(fnamed))
       fnamec = 'rwzecno'
       lunec = io_new_unit()
       open (lunec,FILE=fnamec,iostat=ios)
       if (ios /= 0) call exitme(1,'input error - cannot open'//trim(fnamec))
       fnamsd = 'rwzsdno'
       lunsd = io_new_unit()
       open (lunsd,FILE=fnamsd,iostat=ios)
       if (ios /= 0) call exitme(1,'input error - cannot open'//trim(fnamsd))
   
       do n = 1 , nzno
         read (luned,'(1x)',iostat=ios)
         if (ios /= 0) call exitme(1,'input error - cannot read'//trim(fnamed))
         read (lunec,'(1x)',iostat=ios)
         if (ios /= 0) call exitme(1,'input error - cannot read'//trim(fnamec))
         read (lunsd,'(1x)',iostat=ios)
         if (ios /= 0) call exitme(1,'input error - cannot read'//trim(fnamsd))
         read (luned,'(i6)',iostat=ios) nbp
         if (ios /= 0) call exitme(1,'input error - cannot read'//trim(fnamed))
         read (lunec,'(i6)',iostat=ios) nbp
         if (ios /= 0) call exitme(1,'input error - cannot read'//trim(fnamec))
         read (lunsd,'(i6)',iostat=ios) nbp
         if (ios /= 0) call exitme(1,'input error - cannot read'//trim(fnamsd))
         do mon = 1 , 12
           read (luned,'(i4,f8.1)',iostat=ios) monat, rwzedno_a(n,mon)
           if (ios /=0) call exitme(1,'input error - cannot read'//trim(fnamed))
           read (lunec,'(i4,f8.1)',iostat=ios) monat, rwzecno_a(n,mon)
           if (ios /=0) call exitme(1,'input error - cannot read'//trim(fnamec))
           read (lunsd,'(i4,f8.1)',iostat=ios) monat, rwzsdno_a(n,mon)
           if (ios /=0) call exitme(1,'input error - cannot read'//trim(fnamsd))
         enddo
         read (luned,'(1x)',iostat=ios)
         if (ios /= 0) call exitme(1,'input error - cannot read'//trim(fnamed))
         read (lunec,'(1x)',iostat=ios)
         if (ios /= 0) call exitme(1,'input error - cannot read'//trim(fnamec))
         read (lunsd,'(1x)',iostat=ios)
         if (ios /= 0) call exitme(1,'input error - cannot read'//trim(fnamsd))
       enddo
       close (luned)
       close (lunec)
       close (lunsd)
   
       mon = (jstuna+720/2)/720
       pqr = real(jstuna-mon*720,8)/hpm
       hp  = half+pqr
       hm  = half-pqr
   
       do n = 1,nzno
         iceobc(1,n) = hm*rwzedno_a(n,mod(mon+11,12)+1)                        &
                     + hp*rwzedno_a(n,mod(mon+12,12)+1)
         iceobc(2,n) = hm*rwzecno_a(n,mod(mon+11,12)+1)                        &
                     + hp*rwzecno_a(n,mod(mon+12,12)+1)
         iceobc(3,n) = hm*rwzsdno_a(n,mod(mon+11,12)+1)                        &
                     + hp*rwzsdno_a(n,mod(mon+12,12)+1)
       enddo
  
     else
       iceobc(1:nicepar,1:) = zero
     endif
   
  end subroutine bndiceno

!===============================================================================

  subroutine bndicez(mmk,iga,jga,ice,mmrk,krz,bndice,nz1,nz2)
    !---------------------------------------------------------------------------
    ! boundary values for ice-parameters for nesting from coarse to fine grid
    !---------------------------------------------------------------------------

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: mmk(:,0:,0:)
    integer(4), intent(in)    :: iga(:), jga(:)
    real(8),    intent(in)    :: ice(:,0:)
    integer(4), intent(in)    :: mmrk(:,0:,0:)
    integer(4), intent(in)    :: nz1, nz2
    integer(4), intent(in)    :: krz(:,0:)
    real(8),    intent(inout) :: bndice(:,0:)

    !- local vars --------------------------------------------------------------
    integer(4) :: iw, i, j, ig, jg, ic, jc, iww
    integer(4), parameter :: k=1, kg=1


    do iw=nz1,nz2
      i  = krz(1,iw)
      j  = krz(2,iw)
      if (mmrk(k,i,j) <= 0) cycle

      ic = iga(1)+int((i-1)/iga(2))
      jc = jga(1)+int((j-1)/jga(2))
      if (mmk(kg,ic,jc) <= 0) cycle

      ig = ic
      jg = jc
      if     (krz(3,iw) == 1 .and. mmk(kg,ic,jc-1) > 0) then
        jg = jc-1
      elseif (krz(3,iw) == 2 .and. mmk(kg,ic-1,jc) > 0) then
        ig = ic-1
      elseif (krz(3,iw) == 3 .and. mmk(kg,ic,jc+1) > 0) then
        jg = jc+1
      elseif (krz(3,iw) == 4 .and. mmk(kg,ic+1,jc) > 0) then
        ig = ic+1
      endif

      iww = mmk(kg,ig,jg)

      bndice (:,iw) = ice (:,iww)
    enddo

  end subroutine bndicez

!===============================================================================

  subroutine bndstcl (kmx,khbnd,nz,bndzk,jstuna)
  
    use exits,   only : exitme
    use io_subs, only : io_new_unit
  
    implicit none
  
    integer(4), intent(in) :: kmx, nz, jstuna, khbnd(0:)
    real(8), intent(out)   :: bndzk(1:,1:,0:)
  
    character(20)  :: fnam
    character(256) :: cformat
    integer(4)     :: lunt, luns, nbp, k, mon, monat, kb, n, ios
                     ! NOTE: monat and nbp are only read in but never used
    real(8)        :: pqr, w1, w2
  
    real(8), dimension(kmx,0:nz,12) :: rwzks_a, rwzkt_a
  
    !- initialise
    rwzks_a(:,:,:) = zero
    rwzkt_a(:,:,:) = zero
  
    !- read climatology S and T ------------------------------------------------
    if (nz > 0) then
      fnam = 'rwzktbnd'
      lunt = io_new_unit()
      open (lunt,file=fnam,iostat=ios,pad='NO')
      if (ios /= 0) call exitme (2,'Cannot open rwzktbnd')
      fnam = 'rwzksbnd'
      luns = io_new_unit()
      open (luns,file=fnam,iostat=ios,pad='NO')
      if (ios /= 0) call exitme (2,'Cannot open rwzksbnd')
  
      if     (kmx < 10) then
        write(cformat,'(i1)') kmx
      elseif (kmx < 100) then
        write(cformat,'(i2)') kmx
      elseif (kmx < 1000) then
        write(cformat,'(i3)') kmx
      else
        call exitme(2,'Cannot handle more than 999 layers for S/T climatilogy.')
      endif
    
      cformat = '(i4,'//trim(cformat)//'f8.3)'
  
      do n=1,nz
        read (lunt,*,iostat=ios)
        if (ios /= 0) call exitme (2,'Cannot read rwzktbnd')
        read (luns,*,iostat=ios)
        if (ios /= 0) call exitme (3,'Cannot read rwzksbnd')
        read (lunt,'(i6)',iostat=ios) nbp
        if (ios /= 0) call exitme (4,'Cannot read rwzktbnd')
        read (luns,'(i6)',iostat=ios) nbp
        if (ios /= 0) call exitme (5,'Cannot read rwzksbnd')
        kb = khbnd(n)
        do mon=1,12
          read (lunt,cformat,iostat=ios) monat,(rwzkt_a(k,n,mon),k=1,kb)
          if (ios /= 0) call exitme (6,'Cannot read rwzktbnd')
          read (luns,cformat,iostat=ios) monat,(rwzks_a(k,n,mon),k=1,kb)
          if (ios /= 0) call exitme (7,'Cannot read rwzksbnd')
        enddo
        read (lunt,*,iostat=ios)
        if (ios /= 0) call exitme (8,'Cannot read rwzktbnd')
        read (luns,*,iostat=ios)
        if (ios /= 0) call exitme (9,'Cannot read rwzksbnd')
      enddo
      close(lunt)
      close(luns)
  
      !- interpolation to actual day:
      mon = (jstuna+720/2)/720       !       mon=0,1,2,...,11,12
      pqr = real(jstuna-mon*720,8)/hpm
      w1  = half - pqr
      w2  = half + pqr
  
      do n=1,nz
        kb = khbnd(n)
        do k=1,kb
          bndzk(1,k,n) = w1*rwzkt_a(k,n,mod(mon+11,12)+1)                      &
                       + w2*rwzkt_a(k,n,mod(mon+12,12)+1)
          bndzk(2,k,n) = w1*rwzks_a(k,n,mod(mon+11,12)+1)                      &
                       + w2*rwzks_a(k,n,mod(mon+12,12)+1)
        enddo
        if (kb < kmx) then
          bndzk(1,kb+1:kmx,n) = bndzk(1,kb,n)
          bndzk(2,kb+1:kmx,n) = bndzk(2,kb,n)
        endif
      enddo
      bndzk(1:2,:,0) = bndzk(1:2,:,1)
  
    !- Dummy values if no bnd points -------------------------------------------
    else
      bndzk(1:2,:,:) = zero
    endif
  
  end subroutine bndstcl

!===============================================================================

  subroutine bndstz(mmk,iga,jga,t,ker,mmrk,krz,bndz,kfg,nz1,nz2)
    !---------------------------------------------------------------------------
    !
    ! Tracer boundary values:    coarse grid-->fine grid
    !
    !---------------------------------------------------------------------------

    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain
  
    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: mmk(:,0:,0:)
    integer(4), intent(in)    :: iga(:), jga(:)
    integer(4), intent(in)    :: kfg(:)
    real(8),    intent(in)    :: t(:,0:)
    integer(4), intent(in)    :: ker
    integer(4), intent(in)    :: mmrk(:,0:,0:)
    integer(4), intent(in)    :: nz1, nz2
    integer(4), intent(in)    :: krz(:,0:)
    real(8),    intent(inout) :: bndz(:,:,0:)
    
  
    !- local vars --------------------------------------------------------------
    integer(4) :: iw, i, j, ig, jg, k, kg, kgg, ic, jc, iww, nzl, nzu
  
    call domp_get_domain(nz1, nz2, nzl, nzu)
    do iw=nzl,nzu
      i  = krz(1,iw)
      j  = krz(2,iw)
      if (mmrk(1,i,j) <= 0) cycle

      ic = iga(1)+int((i-1)/iga(2))
      jc = jga(1)+int((j-1)/jga(2))
      if (mmk(1,ic,jc) <= 0) cycle
  
      kg  = 1
      kgg = 1
      do k=1,ker
        if (mmrk(k,i,j) <= 0) exit

        ig = ic
        jg = jc
        if     (krz(3,iw) == 1 .and. mmk(kg,ic,jc-1) > 0) then
          jg = jc-1
        elseif (krz(3,iw) == 2 .and. mmk(kg,ic-1,jc) > 0) then
          ig = ic-1
        elseif (krz(3,iw) == 3 .and. mmk(kg,ic,jc+1) > 0) then
          jg = jc+1
        elseif (krz(3,iw) == 4 .and. mmk(kg,ic+1,jc) > 0) then
          ig = ic+1
        endif
  
        if (mmk(kg,ig,jg) > 0) then
          iww = mmk(kg,ig,jg)
        else
          iww = mmk(kg-1,ig,jg)  
        endif
        bndz(:,k,iw) = t(:,iww)
  
        !- find next k in coarse grid kg:
        kgg = kgg + 1
        if (kgg > kfg(kg)) then
          kg  = kg + 1
          kgg = 1
        endif
      enddo
    enddo
  
  end subroutine bndstz

!===============================================================================
  subroutine bnduvice(me, kru, krv, ice, u, v, casus, dt, dx, dy, cosphi,      &
                      nu1, nu2, nv1, nv2)
  
    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    real(8),    intent(in)    :: dt
    integer(4), intent(in)    :: nu1, nu2, nv1, nv2
    logical,    intent(inout) :: casus(0:)                  ! (0:n2d) 
    real(8),    intent(inout) :: ice(:,0:)
    real(8),    intent(in)    :: cosphi(:,0:)               ! (2,0:imax) 
    real(8),    intent(in)    :: dx(0:), dy(0:)             ! (0:jmax), (0:imax)
    real(8),    intent(in)    :: u(0:), v(0:)               ! (0:n2d)
    integer(4), intent(in)    :: kru(:,0:), krv(:,0:)       ! (3,0:nu), (3,0:nv)
    integer(4), intent(in)    :: me(0:,0:)                  ! (0:imax,0:jmax)
  
    !- local vars --------------------------------------------------------------
    integer(4) :: i , j , n , nn, mi
    real(8)    :: fac
  
    do n = max(1,nu1) , nu2
      i = kru(1,n)
      j = kru(2,n)
      nn = kru(3,n)
      if ( nn == 1 ) then
        j  = j-1
        mi = me(i,j+1)
        if ( mi > 0 .and. me(i,j) > 0 ) then
          if ( u(me(i,j)) > zero ) then
            fac = dt*u(me(i,j))/(dx(j+1)*cosphi(1,i))
            ice(:,mi) = ice(:,mi) + fac*(ice(:,me(i,j))-ice(:,mi))
            casus(mi) = ( ice(1,mi) <= zero )
          endif
        endif
      elseif ( nn == 3 ) then
        j  = j+1
        mi = me(i,j)
        if ( mi > 0 .and. me(i,j+1) > 0 ) then
          if ( u(mi) < zero ) then
            fac = dt*u(mi)/(dx(j)*cosphi(1,i))
            ice(:,mi) = ice(:,mi) - fac*(ice(:,me(i,j+1))-ice(:,mi))
            casus(mi) = ( ice(1,mi) <= zero )
          endif
        endif
      endif
    enddo
  
    do n = max(1,nv1) , nv2
      i = krv(1,n)
      j = krv(2,n)
      nn = krv(3,n)
      if ( nn == 2 ) then
        i = i-1
        mi = me(i+1,j)
        if ( mi > 0 .and. me(i,j) > 0 ) then
          if ( v(me(i,j)) < zero ) then
            fac = dt*v(me(i,j))/dy(i+1)
            ice(:,mi) = ice(:,mi) - fac*(ice(:,me(i,j))-ice(:,mi))
            casus(mi) = ( ice(1,mi) <= zero )
          endif
        endif
      elseif ( nn == 4 ) then
        i = i+1
        mi = me(i,j)
        if ( mi > 0 .and. me(i+1,j) > 0 ) then
          if ( v(mi) > zero ) then
            fac = dt*v(mi)/dy(i)
            ice(:,mi) = ice(:,mi) + fac*(ice(:,me(i+1,j))-ice(:,mi))
            casus(mi) = ( ice(1,mi) <= zero )
          endif
        endif
      endif
    enddo
  
  end subroutine bnduvice

!===============================================================================

  subroutine bnduvst(mk,kh,kru,krv,t,u,v,dt,dx,dy,cosphi,nu1,nu2,nv1,nv2)
  
    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: nu1, nu2, nv1, nv2
    integer(4), intent(in) :: mk(:,0:,0:)
    integer(4), intent(in) :: kh(0:)
    integer(4), intent(in) :: kru(:,0:), krv(:,0:)
    real(8), intent(inout) :: t(:,0:)
    real(8), intent(in)    :: u(0:), v(0:)
    real(8), intent(in)    :: dt, dx(0:), dy(0:)
    real(8), intent(in)    :: cosphi(:,0:)
  
    !- local vars --------------------------------------------------------------
    integer(4) :: n, i, j, nn, kb, k, nul, nuu, nvl, nvu
    real(8)    :: fac, fac0


    call domp_get_domain(max(1,nu1), nu2, nul, nuu)
    do n=nul,nuu
      i=kru(1,n)
      j=kru(2,n)
      nn=kru(3,n)
      if (nn == 1) then
        j = j-1
        kb=kh(mk(1,i,j))
        fac0 = dt/(dx(j+1)*cosphi(1,i))
        do k=1,kb
          if ( mk(k,i,j+1) > 0 ) then
            if ( u(mk(k,i,j)) > zero ) then
              fac = fac0*u(mk(k,i,j))
              t(:,mk(k,i,j+1)) = t(:,mk(k,i,j+1))                              &
                               + fac*(t(:,mk(k,i,j)) - t(:,mk(k,i,j+1)))
            endif
          endif
        enddo
      elseif (nn == 3) then
        j = j+1
        kb=kh(mk(1,i,j))
        fac0 = dt/(dx(j)*cosphi(1,i))
        do k=1,kb
          if ( mk(k,i,j+1) > 0 ) then
            if ( u(mk(k,i,j)) < zero ) then
              fac = fac0*u(mk(k,i,j))
              t(:,mk(k,i,j)) = t(:,mk(k,i,j))                                  &
                             - fac*(t(:,mk(k,i,j+1)) - t(:,mk(k,i,j)))
            endif
          endif
        enddo
      endif
    enddo

!  barrier needed here 'coz t(:,mm_corner), where mm_corner is a corner point,
!  is assigned both above and below.
!$OMP BARRIER

    call domp_get_domain(max(1,nv1), nv2, nvl, nvu)
    do n=nvl,nvu
      i=krv(1,n)
      j=krv(2,n)
      nn=krv(3,n)
      if (nn == 2) then
        i = i-1  
        kb=kh(mk(1,i,j))
        fac0 = dt/dy(i+1)
        do k=1,kb
          if ( mk(k,i+1,j) > 0 ) then
            if ( v(mk(k,i,j)) < zero ) then
              fac = fac0*v(mk(k,i,j))
              t(:,mk(k,i+1,j)) = t(:,mk(k,i+1,j))                              &
                               - fac*(t(:,mk(k,i,j)) - t(:,mk(k,i+1,j)))
            endif
          endif
        enddo
      elseif (nn == 4) then
        i = i+1  
        kb=kh(mk(1,i,j))
        fac0 = dt/dy(i)
        do k=1,kb
          if ( mk(k,i+1,j) > 0 ) then
            if ( v(mk(k,i,j)) > zero ) then
              fac = fac0*v(mk(k,i,j))
              t(:,mk(k,i,j)) = t(:,mk(k,i,j))                                  &
                             + fac*(t(:,mk(k,i+1,j)) - t(:,mk(k,i,j)))
            endif
          endif
        enddo
      endif
    enddo
  
  end subroutine bnduvst

!===============================================================================

  subroutine bndzice(me,krz,bndice,ice,u,v,casus,dt,dx,dy,cosphi,nz1,nz2)
  
     implicit none
     integer(4), intent(in)    :: nz1, nz2
     integer(4), intent(in)    :: me(0:,0:)                 ! (0:imax,0:jmax) 
     integer(4), intent(in)    :: krz(:,0:)                 ! (3,0:nz) 
     real(8),    intent(in)    :: dt, dx(0:), dy(0:)        ! (0:jmax), (0:imax)
     real(8),    intent(in)    :: u(0:), v(0:)
     real(8),    intent(in)    :: cosphi(:,0:)              ! (2,0:imax) 
     real(8),    intent(in)    :: bndice(1:,0:)
     real(8),    intent(inout) :: ice(1:,0:)
     logical,    intent(inout) :: casus(0:)                 ! (0:n2d) 
  
     integer(4) :: i, j, n, nn, mm
     real(8)    :: fac

  !-----------------------------------------------------------------------------

     do n = nz1, nz2
       i = krz(1,n)
       j = krz(2,n)
       nn = abs(krz(3,n))
       mm = me(i,j)
       if (mm > 0) then
         if ( nn == 1 .and. u(mm) > zero ) then
           fac = dt*u(mm)/(dx(j)*cosphi(1,i))
           ice(:,mm) = ice(:,mm) + fac*(bndice(:,n)-ice(:,mm))
         elseif ( nn == 2 .and. v(mm) < zero ) then
           fac = dt*v(mm)/dy(i)
           ice(:,mm) = ice(:,mm) - fac*(bndice(:,n)-ice(:,mm))
         elseif ( nn == 3 .and. u(me(i,j-1)) < zero ) then
           fac = dt*u(me(i,j-1))/(dx(j)*cosphi(1,i))
           ice(:,mm) = ice(:,mm) - fac*(bndice(:,n)-ice(:,mm))
         elseif ( nn == 4 .and. v(me(i-1,j)) > zero ) then
           fac = dt*v(me(i-1,j))/dy(i)
           ice(:,mm) = ice(:,mm) + fac*(bndice(:,n)-ice(:,mm))
         endif
         casus(mm) = ( ice(1,mm) <= zero )
       else
         ice(:,mm) = zero
         casus(mm) = .true.
       endif
     enddo

  end subroutine bndzice

!===============================================================================

  subroutine bndzst(mk,kh,krz,bndz,t,u,v,dt,dx,dy,cosphi, nz1, nz2)
  
    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: nz1, nz2
    integer(4), intent(in) :: mk(:,0:,0:), kh(0:), krz(:,0:)
    real(8), intent(in)    :: bndz(:,:,0:)
    real(8), intent(in)    :: u(0:), v(0:)
    real(8), intent(in)    :: dt, dx(0:), dy(0:), cosphi(:,0:)
    real(8), intent(inout) :: t(:,0:)
  
    !- local vars --------------------------------------------------------------
    integer(4) :: n, i, j, nn, kb, k, indx, nzl, nzu
    real(8)    :: fac, dtdx, dtdy
  
    call domp_get_domain(nz1, nz2, nzl, nzu)

    do n=nzl,nzu
      i=krz(1,n)
      j=krz(2,n)
      nn=abs(krz(3,n))
      kb=kh(mk(1,i,j))
      dtdx = dt/(dx(j)*cosphi(1,i))
      dtdy = dt/dy(i)
      do k=1,kb
        indx = mk(k,i,j)
  
        if ( nn == 1 .and. u(indx) > zero) then
          fac = dtdx*u(indx)
          t(:,indx) = t(:,indx) + fac*(bndz(:,k,n)-t(:,indx))
  
        elseif ( nn == 2 .and. v(indx) < zero) then
          fac = dtdy*v(indx)
          t(:,indx) = t(:,indx) - fac*(bndz(:,k,n)-t(:,indx))
  
        elseif ( nn == 3 .and. u(mk(k,i,j-1)) < zero) then
          fac = dtdx*u(mk(k,i,j-1))
          t(:,indx) = t(:,indx) - fac*(bndz(:,k,n)-t(:,indx))
  
        elseif ( nn == 4 .and. v(mk(k,i-1,j)) > zero) then
          fac = dtdy*v(mk(k,i-1,j))
          t(:,indx) = t(:,indx) + fac*(bndz(:,k,n)-t(:,indx))
  
        endif
      enddo
    enddo
  
  end subroutine bndzst

!===============================================================================

  subroutine rand_st (mmk, kh, ts, nbpq, krq, rwqk, bndq)

  ! FIXME this subroutine doesn't really belong in this module

    implicit none

    !- args:
    integer(4), intent(in)    :: nbpq
    integer(4), intent(in)    :: mmk(:,0:,0:)
    integer(4), intent(in)    :: kh(0:)
    integer(4), intent(in)    :: krq(:,0:)
    real(8),    intent(in)    :: ts(1:,0:), rwqk(0:,:)
    real(8),    intent(inout) :: bndq(1:,1:,0:)

    !- locals:
    integer(4) :: i, j, k, kb, n
    !  rather arbitrary choice: 0.01 degr.C.
    real(8), parameter :: Tfdelta = 0.01_8

    if (nbpq > 0) then
      do n=1,nbpq
        i  = krq(1,n)
        j  = krq(2,n)
        kb = kh(mmk(1,i,j))
        do k=1,kb
          if (rwqk(n,k) >= zero) then
            ! Source term -------------------------------
            !- fresh water inflow, s=0
            bndq(2,k,n) = zero

            !- avoid freezing river inflow, i.e.
            !    Tin > Tfreeze( S ) >= Tfreeze( 0 ) = 0
            bndq(1,k,n) = max(Tfdelta,ts(1,mmk(k,i,j)))
          else
            ! Sink term ---------------------------------
            bndq(1:2,k,n) = ts(1:2,mmk(k,i,j))
          endif
        enddo
      enddo
    endif

  end subroutine rand_st

!===============================================================================

  subroutine bndzcorr (nbpz,krz,rwz,zetac)

    implicit none

    integer(4), intent(in) :: nbpz
    integer(4), intent(in) :: krz(:,0:)
    real(8), intent(inout) :: rwz(0:)
    real(8), intent(inout) :: zetac(0:)
  
    integer(4) :: n
  
    do n=1,nbpz
      if (krz(3,n) > 0) then
        rwz(n) = rwz(n) + zetac(n)
      endif
    enddo

  end subroutine bndzcorr

!===============================================================================

subroutine daybnd (ctim, lmnts, kmx, khbnd, nz, bndzk, setTSbnd, bndz)

  !-----------------------------------------------------------------------------
  !
  ! Get dayly values for S/T at the open lateral boundaries.
  !
  ! Method: Will scan files 'dynbnds' and 'dynbndt' for the actual date stamp
  !         in format
  !           CCYYMMDD(hhmmss)
  !         Hour, minute and second is only read in when  lmnts > 0
  !         If the time stamp is found, we attempt to read the boundary data 
  !         column-wise for each boundary point using format '(i4,17f8.3)' with
  !         17 layers as an example. 
  !         The files should thus look like, starting in column #1:
  !
  !           20061127
  !              1  35.100  35.100  35.100
  !              2  35.200  35.200  35.200
  !           .
  !           .
  !           .
  !            128  35.200  35.200  35.200  35.200  35.200  35.200
  !           20061128
  !              1  35.100  35.100  35.100
  !              2  35.200  35.200  35.200
  !           .
  !           .
  !           .
  !         Only if successful in opening the file(s), finding correct time 
  !         stamp, and reading the data, the corresponding array argument 
  !         (i.e. rwzks and/or rwzkt) will be updated with read data; otherwise,
  !         these are not touched.
  !
  !-----------------------------------------------------------------------------

  !- Modules and directives ----------------------------------------------------
  use dmi_mpi_global, only : iu06
  use exits,          only : exitme
  use io_subs,        only : io_new_unit
  implicit none

  !- Arguments -----------------------------------------------------------------
  character(19), intent(in)    :: ctim
  integer(4),    intent(in)    :: lmnts
  integer(4),    intent(in)    :: kmx, nz, khbnd(0:)
  logical,       intent(in)    :: setTSbnd
  real(8),       intent(inout) :: bndzk(1:,1:,0:)
  real(8),       intent(inout) :: bndz(1:,1:,0:)

  !- Local vars ----------------------------------------------------------------
  logical,       save :: FirstEntry = .true., fs = .true., ft = .true.
  logical,       save :: SecondEntry = .true.
  integer(4),    save :: luns, lunt
  character(14), save :: sstamp, tstamp
  character(13), save :: cformat, stampformat, stampformat2
  integer(4)          :: ios, n, k, kb, nbnd
  character(14)       :: acttime
  real(8)             :: rwdum(kmx)

  !  hardcoded filenames:
  character(7), parameter :: OBCS = 'dynbnds', OBCT = 'dynbndt'

  !- On first entry, get logical unit Nos, open files, etc ... -----------------
  if (FirstEntry) then
    FirstEntry = .false.

    if (nz > 0) then
      ! prepare boundary files, stamps, format etc:
      luns = io_new_unit()
      open(luns,file=OBCS,form='formatted',status='old',iostat=ios,pad='NO')
      if (ios /= 0) then
        write(iu06,'(a47)') 'Warning: Cannot open '//OBCS//'. Continue without.'
        fs = .false.
      endif
      sstamp = '00000000000000'

      lunt = io_new_unit()
      open(lunt,file=OBCT,form='formatted',status='old',iostat=ios,pad='NO')
      if (ios /= 0) then
        write(iu06,'(a47)') 'Warning: Cannot open '//OBCT//'. Continue without.'
        ft = .false.
      endif
      tstamp = '00000000000000'

      if     (kmx < 10  ) then
        write(cformat,'(i1)') kmx
      elseif (kmx < 100 ) then
        write(cformat,'(i2)') kmx
      elseif (kmx < 1000) then
        write(cformat,'(i3)') kmx
      elseif (kmx < 10000) then
        write(cformat,'(i4)') kmx
      else
        call exitme('Error: Cannot handle more than 9999 open boundary points.')
      endif
      cformat = '(i4,'//trim(cformat)//'f8.3)'

      !- Set the stampformat
      stampformat='(a8)'
      if (lmnts > 0) then
        stampformat2='(a14)'
      endif
    else
      ! nothing to do here, let's escape asap:
      fs = .false.
      ft = .false.
    endif
 
  elseif (SecondEntry) then
    SecondEntry = .false.
  endif

  !- Do we have anything to do here, at all? -----------------------------------
  if (.not.(fs.or.ft) .or. nz <=0) return

  !- Set actual day to look for ------------------------------------------------
  if (lmnts == 0) then
    acttime = ctim(1:4)//ctim(6:7)//ctim(9:10)//'000000'
  elseif (lmnts > 0 .and. SecondEntry) then
    acttime = ctim(1:4)//ctim(6:7)//ctim(9:10)//ctim(12:13)//'0000'
  else
    acttime = ctim(1:4)//ctim(6:7)//ctim(9:10)//ctim(12:13)//ctim(15:16)//     &
              ctim(18:19)
  endif

  !- Read boundary S -----------------------------------------------------------
  ios = 0
  if (fs) then
    ! scan to find actual date stamp:
    do while(acttime /= sstamp .and. ios == 0)
      read(luns,stampformat,iostat=ios) sstamp(1:8)
      if (acttime(1:8) == sstamp(1:8) .and. lmnts > 0) then
        backspace(luns)
        read(luns,stampformat2,iostat=ios) sstamp
      endif
    enddo

    ! read data:
    if (ios == 0) then
      do n=1,nz
        kb = khbnd(n)
        if (kb >= 1) then
          read (luns,cformat,iostat=ios) nbnd,(rwdum(k),k=1,kb)
          if (n /= nbnd) then
            write(iu06,*) 'Failed reading boundary point No:',n
            ios = 1
          endif
          if (ios /= 0) exit
          ! transfer to boundary value array:
          if (acttime == sstamp) then
            bndzk(2,1:kb,n) = rwdum(1:kb)
            if (setTSbnd) bndz(2,1:kb,n) = rwdum(1:kb)
          endif
        endif
      enddo
    endif

    ! on failure, don't try reading file again
    if (ios /= 0) then
      write(iu06,*) 'Unable to find data for day '//acttime//' in '//OBCS//'.'
      fs = .false.
    else
      write(iu06,'(a58)') 'Boundary data for day '//acttime//                  &
                          ' read from '//OBCS//'.'
    endif
  endif

  !- Read boundary T -----------------------------------------------------------
  ios = 0
  if (ft) then
    ! scan to find actual date stamp:
    do while(acttime /= tstamp .and. ios == 0)
      read(lunt,stampformat,iostat=ios) tstamp(1:8)
      if (acttime(1:8) == tstamp(1:8) .and. lmnts > 0) then
        backspace(lunt)
        read(lunt,stampformat2,iostat=ios) tstamp
      endif
    enddo

    ! read data:
    if (ios == 0) then
      do n=1,nz
        kb = khbnd(n)
        if (kb >= 1) then
          read (lunt,cformat,iostat=ios) nbnd,(rwdum(k),k=1,kb)
          if (n /= nbnd) then
            write(iu06,*) 'Failed reading boundary point No:',n
            ios = 1
          endif
          if (ios /= 0) exit
          ! transfer to boundary value array:
          if (acttime == tstamp) then
            bndzk(1,1:kb,n) = rwdum(1:kb)
            if (setTSbnd) bndz(1,1:kb,n) = rwdum(1:kb)
          endif
        endif
      enddo
    endif

    ! on failure, don't try reading file again
    if (ios /= 0) then
      write(iu06,*) 'Unable to find data for day '//acttime//' in '//OBCT//'.'
      ft = .false.
    else
      write(iu06,'(a58)') 'Boundary data for day '//acttime//                  &
                          ' read from '//OBCT//'.'
    endif
  endif

end subroutine daybnd

!===============================================================================

!-------------------------------------------------------------------------------
!
! Calculates Velocity Orthogonal to Boundary, 
! i.e. the velocity component in the direction across the open boundary.
!
! vtype=2: Applies the "usual" mom.eqs. except that 
!          - convective terms are ignored during inflow, i.e. only convection
!            during outflow
!          - horizontal diffusion is a simple Laplace filter smoothing, 
!            i.e. Smagorinsky terms are ignored
!
!
! Per Berg, DMI, August 25 2005
!
!-------------------------------------------------------------------------------

subroutine vobcalc ( ke,kdry,mmk,me,dt,ft,rt,ty,tx,gty,gtx,z,u,v,hz,un,vn,     &
                     taux,tauy,pl,press,geopot,rho,avv,casus,ice,ueis,veis,    &
                     vtype,nz,krz,z0srf )

  !- modules -------------------------------------------------------------------
  use exits,         only : exitme     
  use cmod_params,   only : kmax, kmaxp
  use constants,     only : dyn, hgr, rhow, rhoi, cdrag
  use cmod_miscvars, only : dtdscale
  use drying_cases,  only : drying_case

  ! ----------------------------------------------------------------------------
  !    hgr: grenztiefe (0.05m)
  ! ----------------------------------------------------------------------------

  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in) :: ke              ! number of layers
  integer(4), intent(in) :: kdry
  integer(4), intent(in) :: mmk(:,0:,0:)    ! index
  integer(4), intent(in) :: me(0:,0:)       ! index
  real(8), intent(in)    :: dt              ! zeitschritt
  real(8), intent(in)    :: ft(:)           ! ft(i):f(i)*dt
  real(8), intent(in)    :: rt              ! rt: r*dt
  real(8), intent(in)    :: ty              ! ty: dt/dy
  real(8), intent(in)    :: tx(:)           ! tx(i): dt/dx(i)
  real(8), intent(in)    :: gty             ! gty:g*dt/dy
  real(8), intent(in)    :: gtx(:)          ! gtx(i):g*dt/dx(i)
  real(8), intent(in)    :: z(0:)
  real(8), intent(in)    :: u(0:)
  real(8), intent(in)    :: v(0:)
  real(8), intent(in)    :: hz(0:)
  real(8), intent(inout) :: un(0:)
  real(8), intent(inout) :: vn(0:)
  real(8), intent(in)    :: taux(0:)
  real(8), intent(in)    :: tauy(0:)
  real(8), intent(in)    :: pl(0:)
  logical, intent(in)    :: casus(0:)
  real(8), intent(in)    :: ice(1:,0:)
  real(8), intent(in)    :: ueis(0:)
  real(8), intent(in)    :: veis(0:)
  real(8), intent(in)    :: press(0:)
  real(8), intent(in)    :: geopot(0:)
  real(8), intent(in)    :: rho(0:)
  real(8), intent(in)    :: avv(0:)
  real(8), intent(in)    :: z0srf(0:)
  integer(4), intent(in) :: vtype           ! type of extrapolation scheme
  integer(4), intent(in) :: nz, krz(:,0:)   ! bnd specs

  !- local vars ----------------------------------------------------------------
  logical, save :: FirstTime = .true.
  integer(4) :: nbnd,b
  integer(4) :: i,j,kev,keu,k,i1,i2,i3,i4,j1,j2,j3,j4
  integer(4) :: mi,ee,ww,nn,ss,ne,se,sw
  real(8) :: uneu,tutop,decfu,vnn,vne,eiscol,eiscor,eiscox
  real(8) :: vneu,tvtop,eiscoo,eiscou,eiscoy,diffuo,diffuu,diffvo
  real(8) :: diffvu,diffv,decfv,vlu,vzet,vsrf,vss,vww,ust
  real(8) :: ustar,vaus,vpress,sdv
  real(8) :: sdvlum
  real(8) :: diffu,difful,diffur,diffvl,diffvr,diffuv
  real(8) :: ulu,uzet,usrf,vmi,vee,umi,uee,uww,tmp,tmp2,tmp3,fac
  real(8) :: unn,uss,usw,vst,vstar,uaus,upress,urd,vrd
  real(8) :: sdu,hrt,sdulum, zu, zr, zm
  real(8) :: dtd, ucon, vcon, aurd, avrd

  real(8), allocatable, save :: ea(:),eb(:),ec(:),alpha(:),beta(:), tm(:)
  real(8), parameter :: three = 3.0_8, quart=0.25_8
  real(8), parameter :: scaling = (6.0_8*1852.0_8)**2/1000.0_8

  !- Trap! ---------------------------------------------------------------------
  if (FirstTime) then
    FirstTime = .false.
    if (vtype /= 2) then
      call exitme('ERROR: vtype /= 2 is not implemented yet')
    endif
    allocate( ea(kmaxp),eb(kmax),ec(kmax),alpha(kmax),beta(kmax), tm(kmax) )
  endif

  !- initialise on every entry -------------------------------------------------
  dtd = scaling*dtdscale*ty**2/dt

  !- run though all boundary points --------------------------------------------
  nbndloop: do nbnd=1,nz
    i = krz(1,nbnd)
    j = krz(2,nbnd)
    b = krz(3,nbnd)

    mi = mmk(1,i,j)
    if (mi <= 0) cycle nbndloop
    
    if (b == 2) then
      !- North boundary: =======================================================

      !----------------- v-equation --------------
      kev=0
      do k=1,ke
        mi=mmk(k,i,j)
        ss=mmk(k,i+1,j)
        if (mi <= 0 .or. ss <= 0) exit
        kev    = k
        tm(k)  = half*(hz(mi)+hz(ss))
      enddo
      mi=mmk(1, i  ,j)
      if(kev == 0) then
        vn(mi)=zero
        cycle nbndloop
      endif

      ss=mmk(1,i+1,j)

      tmp = half*(z0srf(mi) + z0srf(ss))
      if (casus(me(i,j)).and.casus(ss)) then
        tvtop = tauy(mi)/rhow*dt/(tm(1)+tmp)
      else
        eiscoo=ice(2,me(i,j))
        eiscoo=max(eiscoo,zero)
        eiscoo=min(eiscoo,one)
        eiscou=ice(2,me(i+1,j))
        eiscou=max(eiscou,zero)
        eiscou=min(eiscou,one)
        eiscoy=half*(eiscoo+eiscou)
        diffuo=half*( ueis(me(i,j-1))-u(mmk(1,i,j-1))+ueis(me(i,j))-u(mi))
        diffuu=half*(  ueis(me(i+1,j-1))-u(mmk(1,i+1,j-1))                     &
                      +ueis(me(i+1,j))-u(ss))
        diffvo=zero 
        diffvu=half*( veis(me(i,j))-v(mi)+veis(me(i+1,j))-v(ss))
        diffv=veis(me(i,j))-v(mi)
        diffuv=half*(  eiscoo*sqrt(diffuo**2+diffvo**2)                        &
                      +eiscou*sqrt(diffuu**2+diffvu**2))/ eiscoy

        fac = dt/rhow/(tm(1)+tmp)

        decfv=cdrag*rhoi*diffuv*fac
        tvtop= ( eiscoy*rhoi*cdrag*diffuv*diffv +(one-eiscoy)*tauy(mi) )*fac
      endif

      ! begin ea(:), ec(:) computation
      ea=zero ! i.e. ea(1)=0
      ec=zero
      do k=1,kev-2,2
        tmp=tm(k)
        tmp2=tm(k+1)
        ea(k+1)=dt*(avv(mmk(k+1,i,j))+avv(mmk(k+1,i+1,j)))/(tmp+tmp2)
        ec(k)=one+(ea(k)+ea(k+1))/tmp
        tmp3=tm(k+2)
        ea(k+2)=dt*(avv(mmk(k+2,i,j))+avv(mmk(k+2,i+1,j)))/(tmp2+tmp3)
        ec(k+1)=one+(ea(k+1)+ea(k+2))/tmp2
      enddo
      if ((mod(kev,2) /= 0).and.(kev > 0)) then
        ec(kev)=one+(ea(kev)+ea(kev+1))/tm(kev)
      else
        ea(kev)=dt*(avv(mmk(kev,i,j))+avv(mmk(kev,i+1,j)))/(tm(kev-1)+tm(kev))
        ec(kev-1)=one+(ea(kev-1)+ea(kev))/tm(kev-1)
        ec(kev)=one+(ea(kev)+ea(kev+1))/tm(kev)
      endif
      ! end ea(:) computation

      ! begin eb(:) computation
      ! --- luftdruckterm
      vlu=(pl(mi)-pl(ss))*ty/rhow ! used in eb(:) computation

      ! zeta:
      zm = z(mi)
      zu = z(ss)

      ! pressure term:
      vzet = gty*(zm-zu)

      ! collect different surface terms:
      vsrf = vlu + vzet

      eb=zero
      do k=1,kev
        ww=mmk(k,i  ,j-1)
        sw=mmk(k,i+1,j-1)
        mi=mmk(k,i  ,j  )
        ss=mmk(k,i+1,j  )
        ee=mmk(k,i  ,j+1)
        se=mmk(k,i+1,j+1)

        uww=u(ww)
        usw=u(sw)
        uss=u(ss)
        umi=u(mi)
        vmi=v(mi)
        vss=v(ss)
        vee=v(ee)
        vww=v(ww)

        ust=quart*(uww+umi+usw+uss)

        !---  barokliner druck
        vpress=(press(mi)-press(ss))*ty/(half*(rho(mi)+rho(ss)))

        ! convective terms using vector upwind scheme:
        ! ... neglected on boundary during inflow
        if (vmi > zero) then
          urd  = ust*tx(i)
          vrd  = vmi*ty
          aurd = abs(urd)
          avrd = abs(vrd)

          if (vrd >= zero) then
            i2 = i+1
          else
            i2 = i-1
          endif
          if (urd >= zero) then
            j4 = j-1
          else
            j4 = j+1
          endif
          if (avrd >= aurd) then
            i3 = i2
            j1 = j
          else
            i3 = i
            j1 = j4
          endif

          vcon = avrd*(v(mmk(k,i,j1))-v(mmk(k,i2,j1)))                         &
               + aurd*(v(mmk(k,i3,j))-v(mmk(k,i3,j4)))
        else
          vcon = zero
        endif

        !---  austauschterm
        ! ... Smagorinsky terms ignored on boundary; 
        !     only back-ground Laplace filter diffusion
        vaus = vss - three*vmi
        if (ee > 0 .and. se > 0) then
          vaus = vaus + vee
        else
          vaus = vaus + vmi
        endif
        if (ww > 0 .and. sw > 0) then
          vaus = vaus + vww
        else
          vaus = vaus + vmi
        endif
        vaus = dtd*vaus

        !- RHS of y-mom-eq stored in eb(:) ------------------------------------
        eb(k)=vmi-ft(i)*ust-vpress+vaus-vsrf-vcon
      enddo

      if (dyn) then
        !- The geopotential term
        do k=1,kev
          mi = mmk(k,i  ,j  )
          ss = mmk(k,i+1,j  )
          eb(k) = eb(k) - (geopot(mi)-geopot(ss))*ty
        enddo
      endif

      ustar = ust

      hrt     = tm(kev)+hgr
      tmp     = half*rt/hrt*sqrt(v(mmk(kev,i,j))*v(mmk(kev,i,j))+ustar*ustar) 
      eb(kev) = eb(kev) - tmp*v(mmk(kev,i,j))
      ec(kev) = ec(kev) + tmp

      eb(1)=eb(1)+tvtop
      if (.not.(casus(me(i,j)).and.casus(me(i+1,j)))) then
        eb(1)=eb(1)+half*eiscoy*decfv*v( mmk(1,i,j))
        ec(1)=ec(1)+half*eiscoy*decfv
      endif

      ! begin alpha, beta computation
      alpha=zero
      beta=zero
      alpha(kev)=eb(kev)
      beta(kev)=ec(kev)
      do k=kev-1,1,-1
        sdv      = ea(k+1)/(tm(k)*beta(k+1))
        sdvlum   = ea(k+1)/tm(k+1)
        alpha(k) = eb(k) + sdv*alpha(k+1)
        beta(k)  = ec(k) - sdv*sdvlum
      enddo
      ! end alpha, beta computation

      ! begin vn(:) computation
      vn( mmk(1,i,j))=alpha(1)/beta(1)
      do k=2,kev
        sdv=tm(k)
        vn( mmk(k,i,j))=(alpha(k)+ea(k)/sdv*vn(mmk(k-1,i,j)))/beta(k)
      enddo

      !---  handle a drying situation
      if (kev <= kdry) then
        ss = mmk(1,i+1,j)
        mi = mmk(1,i,j)
        call drying_case(ss, mi, vn(mi), hz(ss), hz(mi), z) 
        if (kdry > 1) then
          do k = 2,min(kev,kdry)
            call drying_case(ss,mi,vn(mmk(1,i,j)),hz(mmk(1,i,j)),              &
                             hz(mmk(1,i+1,j)),z)
          enddo
        endif
      endif
      ! end vn(:) computation
    elseif (b == 1) then
      !- West boundary: ========================================================

      ! --- u-equation  -----------
      keu=0
      do k=1,ke
        mi=mmk(k,i,j)
        ee=mmk(k,i,j+1)
        if (mi <= 0 .or. ee <= 0) exit
        keu    = k
        tm(k)  = half*(hz(mi)+hz(ee))
      enddo
      mi=mmk(1, i  ,j)
      if(keu == 0) then
        un(mi)=zero
        cycle nbndloop
      endif

      ee=mmk(1,i  ,j+1)

      tmp = half*(z0srf(mi) + z0srf(ee))
      if (casus(me(i,j)).and.casus(ee)) then
        tutop = taux(mi)/rhow*dt/(tm(1)+tmp)
      else
        eiscol=ice(2,me(i,j))
        eiscol=max(eiscol,zero)
        eiscol=min(eiscol,one)
        eiscor=ice(2,me(i,j+1))
        eiscor=max(eiscor,zero)
        eiscor=min(eiscor,one)
        eiscox=half*(eiscol+eiscor)
        diffu=ueis(me(i,j))-u(mi)
        difful=half*( ueis(me(i,j-1))-u(mmk(1,i,j-1))+ueis(me(i,j))-u(mi))
        diffur=half*( ueis(me(i,  j))-u(mi)+ueis(me(i,j+1))-u(ee))
        diffvl=half*( veis(me(i-1,j))-v(mmk(1,i-1,j))+veis(me(i  ,j))-v(mi)) 
        diffvr=half*( veis(me(i-1,j+1))-v(mmk(1,i-1,j+1))                      &
                     +veis(me(i  ,j+1))-v(ee))
        diffuv=half*( eiscol*sqrt(difful**2+diffvl**2)                         &
                     +eiscor*sqrt(diffur**2+diffvr**2))/eiscox

        fac = dt/rhow/(tm(1)+tmp)

        decfu=cdrag*rhoi*diffuv*fac
        tutop= ( eiscox*cdrag*rhoi*diffuv*diffu +(one-eiscox)*taux(mi) )*fac
      endif

      ! begin ea(:), ec(:) computation -----------------------------------------
      ea=zero ! i.e. ea(1)=0
      ec=zero
      do k=1,keu-2,2
        tmp=tm(k)
        tmp2=tm(k+1)
        ea(k+1)=dt*(avv(mmk(k+1,i,j))+avv(mmk(k+1,i,j+1)))/(tmp+tmp2)
        ec(k)=one+(ea(k)+ea(k+1))/tmp
        tmp3=tm(k+2)
        ea(k+2)=dt*(avv(mmk(k+2,i,j))+avv(mmk(k+2,i,j+1)))/(tmp2+tmp3)
        ec(k+1)=one+(ea(k+1)+ea(k+2))/tmp2
      enddo
      if ((mod(keu,2) /= 0).and.(keu > 0)) then
        ec(keu)=one+(ea(keu)+ea(keu+1))/tm(keu)
      else
        ea(keu)=dt*(avv(mmk(keu,i,j))+avv(mmk(keu,i,j+1)))/(tm(keu-1)+tm(keu))
        ec(keu-1)=one+(ea(keu-1)+ea(keu))/tm(keu-1)
        ec(keu)=one+(ea(keu)+ea(keu+1))/tm(keu)
      endif
      ! end ea(:) computation 

      ! begin eb(:) computation
      ! --- luftdruckterm 
      ulu=(pl(ee)-pl(mi))*tx(i)/rhow ! not symm., used in eb(:) computation

      ! zeta:
      zm = z(mi)
      zr = z(ee)

      ! pressure term:
      uzet = gtx(i)*(zr-zm)

      ! collect different surface terms:
      usrf = ulu + uzet

      eb    = zero
      do k=1,keu
        nn=mmk(k,i-1,j  )
        mi=mmk(k,i  ,j  )
        ss=mmk(k,i+1,j  )
        ne=mmk(k,i-1,j+1)
        ee=mmk(k,i  ,j+1)
        se=mmk(k,i+1,j+1)

        vnn=v(nn)
        vne=v(ne)
        vmi=v(mi)
        vee=v(ee)
        umi=u(mi)
        uee=u(ee)
        unn=u(nn)
        uss=u(ss)

        vst=quart*(vnn+vne+vmi+vee) ! not symm.

        !---  barokliner druck 
        upress=(press(ee)-press(mi))*tx(i)/(half*(rho(ee)+rho(mi)))

        ! convective terms using vector upwind scheme:
        ! ... neglected on boundary during inflow
        if (umi < zero) then
          urd  = umi*tx(i)
          vrd  = vst*ty
          aurd = abs(urd)
          avrd = abs(vrd)

          if (urd >= zero) then
            j2 = j-1
          else
            j2 = j+1
          endif
          if (vrd >= zero) then
            i4 = i+1
          else
            i4 = i-1
          endif
          if (aurd >= avrd) then
            i1 = i
            j3 = j2
          else
            i1 = i4
            j3 = j
          endif

          ucon = aurd*(u(mmk(k,i1,j))-u(mmk(k,i1,j2)))                         &
               + avrd*(u(mmk(k,i,j3))-u(mmk(k,i4,j3)))
        else
          ucon = zero
        endif

        !---  austauschterm
        ! ... Smagorinsky terms ignored on boundary;
        !     only back-ground Laplace filter diffusion
        uaus = uee - three*umi
        if (ne > 0 .and. nn > 0) then
          uaus = uaus + unn
        else
          uaus = uaus + umi
        endif
        if (se > 0 .and. ss > 0) then
          uaus = uaus + uss
        else
          uaus = uaus + umi
        endif
        uaus = dtd*uaus

        !- RHS of x-mom-eq stored in eb(:) -------------------------------------
        eb(k)=umi+ft(i)*vst-ucon+uaus-usrf-upress
      enddo

      if (dyn) then
        !- The geopotential term
        do k=1,keu
          mi = mmk(k,i  ,j  )
          ee = mmk(k,i  ,j+1)
          eb(k) = eb(k) - (geopot(ee)-geopot(mi))*tx(i)
        enddo
      endif

      vstar = vst

      hrt     = tm(keu)+hgr
      tmp     = half*rt/hrt*sqrt(u(mmk(keu,i,j))*u(mmk(keu,i,j))+vstar*vstar)
      eb(keu) = eb(keu) - tmp*u(mmk(keu,i,j))
      ec(keu) = ec(keu) + tmp

      eb(1)=eb(1)+tutop
      if (.not.(casus(me(i,j)).and.casus(me(i,j+1)))) then
        eb(1)=eb(1)+half*eiscox*decfu*u( mmk(1,i,j))
        ec(1)=ec(1)+half*eiscox*decfu
      endif

      ! begin alpha, beta computation
      alpha=zero
      beta=zero
      alpha(keu)=eb(keu)
      beta(keu)=ec(keu)
      do k=keu-1,1,-1
        sdu      = ea(k+1)/(tm(k)*beta(k+1))
        sdulum   = ea(k+1)/tm(k+1)
        alpha(k) = eb(k) + sdu*alpha(k+1)
        beta(k)  = ec(k) - sdu*sdulum
      enddo
      ! end alpha, beta computation

      ! begin un(:) computation
      un( mmk(1,i,j))=alpha(1)/beta(1)
      do k=2,keu
        sdu=tm(k)
        un( mmk(k,i,j))=(alpha(k)+ea(k)/sdu*un(mmk(k-1,i,j)))/beta(k)
      enddo

      !---  handle a drying situation
      if (keu <= kdry) then
        ee = mmk(1,i,j+1)
        mi = mmk(1,i,j)
        call drying_case(mi, ee, un(mi), hz(mi), hz(ee), z) 
        if (kdry > 1) then
          do k = 2,min(keu,kdry)
            call drying_case(mi,ee,un(mmk(k,i,j)),hz(mmk(k,i,j)),              &
                             hz(mmk(1,i,j+1)),z)
          enddo
        endif
      endif
      ! end un(:) computation


    elseif (b == 3) then
      !- East boundary: ========================================================

      !- check if we need to do this point at all
      if (mmk(1,i,j) <= 0) cycle nbndloop

      !- temporary re-define j and mi:
      j  = j-1
      mi = mmk(1,i,j)

      !- make sure to have a nice value outside domain
      ee = mmk(1,i,j+1)
      un(ee) = zero

      ! --- u-equation  -----------
      keu=0
      do k=1,ke
        mi=mmk(k,i,j)
        ee=mmk(k,i,j+1)
        if (mi <= 0 .or. ee <= 0) exit
        keu    = k
        tm(k)  = half*(hz(mi)+hz(ee))
      enddo
      mi=mmk(1, i  ,j)
      if(keu == 0) then
        un(mi)=zero
        cycle nbndloop
      endif

      ee=mmk(1,i  ,j+1)

      tmp = half*(z0srf(mi) + z0srf(ee))
      if (casus(me(i,j)).and.casus(ee)) then
        tutop = taux(mi)/rhow*dt/(tm(1)+tmp)
      else
        eiscol=ice(2,me(i,j))
        eiscol=max(eiscol,zero)
        eiscol=min(eiscol,one)
        eiscor=ice(2,me(i,j+1))
        eiscor=max(eiscor,zero)
        eiscor=min(eiscor,one)
        eiscox=half*(eiscol+eiscor)
        diffu=ueis(me(i,j))-u(mi)
        difful=half*( ueis(me(i,j-1))-u(mmk(1,i,j-1))+ueis(me(i,j))-u(mi))
        diffur=half*( ueis(me(i,  j))-u(mi)+ueis(me(i,j+1))-u(ee))
        diffvl=half*( veis(me(i-1,j))-v(mmk(1,i-1,j))+veis(me(i  ,j))-v(mi)) 
        diffvr=half*( veis(me(i-1,j+1))-v(mmk(1,i-1,j+1))                      &
                     +veis(me(i  ,j+1))-v(ee))
        diffuv=half*( eiscol*sqrt(difful**2+diffvl**2)                         &
                     +eiscor*sqrt(diffur**2+diffvr**2))/eiscox

        fac = dt/rhow/(tm(1)+tmp)

        decfu=cdrag*rhoi*diffuv*fac
        tutop= ( eiscox*cdrag*rhoi*diffuv*diffu +(one-eiscox)*taux(mi) )*fac
      endif

      ! begin ea(:), ec(:) computation -----------------------------------------
      ea=zero ! i.e. ea(1)=0
      ec=zero
      do k=1,keu-2,2
        tmp=tm(k)
        tmp2=tm(k+1)
        ea(k+1)=dt*(avv(mmk(k+1,i,j))+avv(mmk(k+1,i,j+1)))/(tmp+tmp2)
        ec(k)=one+(ea(k)+ea(k+1))/tmp
        tmp3=tm(k+2)
        ea(k+2)=dt*(avv(mmk(k+2,i,j))+avv(mmk(k+2,i,j+1)))/(tmp2+tmp3)
        ec(k+1)=one+(ea(k+1)+ea(k+2))/tmp2
      enddo
      if ((mod(keu,2) /= 0).and.(keu > 0)) then
        ec(keu)=one+(ea(keu)+ea(keu+1))/tm(keu)
      else
        ea(keu)=dt*(avv(mmk(keu,i,j))+avv(mmk(keu,i,j+1)))/(tm(keu-1)+tm(keu))
        ec(keu-1)=one+(ea(keu-1)+ea(keu))/tm(keu-1)
        ec(keu)=one+(ea(keu)+ea(keu+1))/tm(keu)
      endif
      ! end ea(:) computation 

      ! begin eb(:) computation
      ! --- luftdruckterm 
      ulu=(pl(ee)-pl(mi))*tx(i)/rhow ! not symm., used in eb(:) computation

      ! zeta:
      zm = z(mi)
      zr = z(ee)

      ! pressure term:
      uzet = gtx(i)*(zr-zm)

      ! collect different surface terms:
      usrf = ulu + uzet

      eb    = zero
      do k=1,keu
        ww=mmk(k,i  ,j-1)
        nn=mmk(k,i-1,j  )
        mi=mmk(k,i  ,j  )
        ss=mmk(k,i+1,j  )
        ne=mmk(k,i-1,j+1)
        ee=mmk(k,i  ,j+1)
        se=mmk(k,i+1,j+1)

        vnn=v(nn)
        vne=v(ne)
        vmi=v(mi)
        vee=v(ee)
        umi=u(mi)
        uww=u(ww)
        unn=u(nn)
        uss=u(ss)

        vst=quart*(vnn+vne+vmi+vee) ! not symm.

        !---  barokliner druck 
        upress=(press(ee)-press(mi))*tx(i)/(half*(rho(ee)+rho(mi)))
  

        ! convective terms using vector upwind scheme:
        ! ... neglected on boundary during inflow
        if (umi > zero) then
          urd  = umi*tx(i)
          vrd  = vst*ty
          aurd = abs(urd)
          avrd = abs(vrd)

          if (urd >= zero) then
            j2 = j-1
          else
            j2 = j+1
          endif
          if (vrd >= zero) then
            i4 = i+1
          else
            i4 = i-1
          endif
          if (aurd >= avrd) then
            i1 = i
            j3 = j2
          else
            i1 = i4
            j3 = j
          endif

          ucon = aurd*(u(mmk(k,i1,j))-u(mmk(k,i1,j2)))                         &
               + avrd*(u(mmk(k,i,j3))-u(mmk(k,i4,j3)))
        else
          ucon = zero
        endif

        !---  austauschterm
        ! ... Smagorinsky terms ignored on boundary;
        !     only back-ground Laplace filter diffusion
        uaus = uww - three*umi
        if (ne > 0 .and. nn > 0) then
          uaus = uaus + unn
        else
          uaus = uaus + umi
        endif
        if (se > 0 .and. ss > 0) then
          uaus = uaus + uss
        else
          uaus = uaus + umi
        endif
        uaus = dtd*uaus

        !- RHS of x-mom-eq stored in eb(:) -------------------------------------
        eb(k)=umi+ft(i)*vst-usrf-upress-ucon+uaus
      enddo

      if (dyn) then
       !- The geopotential term
       do k=1,keu
          mi = mmk(k,i  ,j  )
          ee = mmk(k,i  ,j+1)
          eb(k) = eb(k) - (geopot(ee)-geopot(mi))*tx(i)
        enddo
      endif

      vstar = vst

      hrt     = tm(keu)+hgr
      tmp     = half*rt/hrt*sqrt(u(mmk(keu,i,j))*u(mmk(keu,i,j))+vstar*vstar)
      eb(keu) = eb(keu) - tmp*u(mmk(keu,i,j))
      ec(keu) = ec(keu) + tmp

      eb(1)=eb(1)+tutop
      if (.not.(casus(me(i,j)).and.casus(me(i,j+1)))) then
        eb(1)=eb(1)+half*eiscox*decfu*u( mmk(1,i,j))
        ec(1)=ec(1)+half*eiscox*decfu
      endif

      ! begin alpha, beta computation
      alpha=zero
      beta=zero
      alpha(keu)=eb(keu)
      beta(keu)=ec(keu)
      do k=keu-1,1,-1
        sdu      = ea(k+1)/(tm(k)*beta(k+1))
        sdulum   = ea(k+1)/tm(k+1)
        alpha(k) = eb(k) + sdu*alpha(k+1)
        beta(k)  = ec(k) - sdu*sdulum
      enddo
      ! end alpha, beta computation

      ! begin un(:) computation
      !    force "nice" value outside domain j+1
      uneu = alpha(1)/beta(1)
      un(mmk(1,i,j))  = uneu
      un(mmk(1,i,j+1)) = uneu
      do k=2,keu
        sdu  = tm(k)
        uneu = (alpha(k)+ea(k)/sdu*un(mmk(k-1,i,j)))/beta(k)
        un(mmk(k,i,j))   = uneu
        un(mmk(k,i,j+1)) = uneu
      enddo

      !---  trockenfallabfrage
      if (keu <= kdry) then
        ee = mmk(1,i,j+1)
        mi = mmk(1,i,j)
        call drying_case(mi, ee, un(mi), hz(mi), hz(ee), z) 
        if (kdry > 1) then
          do k = 2,min(keu,kdry)
            call drying_case(mi,ee,un(mmk(k,i,j)),hz(mmk(k,i,j)),              &
                             hz(mmk(1,i,j+1)),z)
          enddo
        endif
      endif
      ! end un(:) computation
      
      !- turn back to original values for j and mi for safety:
      j  = j+1
      mi = mmk(1,i,j)


    elseif (b == 4) then
      !- South boundary: =======================================================

      !- check if we need to do this point at all
      if (mmk(1,i,j) <= 0) cycle nbndloop

      !- temporary re-define i and mi:
      i  = i - 1
      mi = mmk(1,i,j)
 
      !- make sure to have a nice value outside domain:
      ss = mmk(1,i+1,j)
      vn(ss) = zero

      !----------------- v-equation --------------
      kev=0
      do k=1,ke
        mi=mmk(k,i,j)
        ss=mmk(k,i+1,j)
        if (mi <= 0 .or. ss <= 0) exit
        kev    = k
        tm(k)  = half*(hz(mi)+hz(ss))
      enddo
      mi=mmk(1, i  ,j)
      if(kev == 0) then
        vn(mi)=zero
        cycle nbndloop
      endif

      ss=mmk(1,i+1,j)

      tmp = half*(z0srf(mi) + z0srf(ss))
      if (casus(me(i,j)).and.casus(ss)) then
        tvtop = tauy(mi)/rhow*dt/(tm(1)+tmp)
      else
        eiscoo=ice(2,me(i,j))
        eiscoo=max(eiscoo,zero)
        eiscoo=min(eiscoo,one)
        eiscou=ice(2,me(i+1,j))
        eiscou=max(eiscou,zero)
        eiscou=min(eiscou,one)
        eiscoy=half*(eiscoo+eiscou)
        diffuo=half*( ueis(me(i,j-1))-u(mmk(1,i,j-1))+ueis(me(i,j))-u(mi))
        diffuu=half*( ueis(me(i+1,j-1))-u(mmk(1,i+1,j-1))                      &
                     +ueis(me(i+1,j))-u(ss))
        diffvo=zero
        diffvu=half*( veis(me(i,j))-v(mi)+veis(me(i+1,j))-v(ss))
        diffv=veis(me(i,j))-v(mi)
        diffuv=half*( eiscoo*sqrt(diffuo**2+diffvo**2)                         &
                     +eiscou*sqrt(diffuu**2+diffvu**2))/ eiscoy

        fac = dt/rhow/(tm(1)+tmp)

        decfv=cdrag*rhoi*diffuv*fac
        tvtop= ( eiscoy*rhoi*cdrag*diffuv*diffv +(one-eiscoy)*tauy(mi) )*fac
      endif

      ! begin ea(:), ec(:) computation
      ea=zero ! i.e. ea(1)=0
      ec=zero
      do k=1,kev-2,2
        tmp=tm(k)
        tmp2=tm(k+1)
        ea(k+1)=dt*(avv(mmk(k+1,i,j))+avv(mmk(k+1,i+1,j)))/(tmp+tmp2)
        ec(k)=one+(ea(k)+ea(k+1))/tmp
        tmp3=tm(k+2)
        ea(k+2)=dt*(avv(mmk(k+2,i,j))+avv(mmk(k+2,i+1,j)))/(tmp2+tmp3)
        ec(k+1)=one+(ea(k+1)+ea(k+2))/tmp2
      enddo
      if ((mod(kev,2) /= 0).and.(kev > 0)) then
        ec(kev)=one+(ea(kev)+ea(kev+1))/tm(kev)
      else
        ea(kev)=dt*(avv(mmk(kev,i,j))+avv(mmk(kev,i+1,j)))/(tm(kev-1)+tm(kev))
        ec(kev-1)=one+(ea(kev-1)+ea(kev))/tm(kev-1)
        ec(kev)=one+(ea(kev)+ea(kev+1))/tm(kev)
      endif
      ! end ea(:) computation

      ! begin eb(:) computation
      ! --- luftdruckterm
      vlu=(pl(mi)-pl(ss))*ty/rhow ! used in eb(:) computation

      ! zeta:
      zm = z(mi)
      zu = z(ss)

      ! pressure term:
      vzet = gty*(zm-zu)

      ! collect different surface terms:
      vsrf = vlu + vzet

      eb=zero
      do k=1,kev
        ww=mmk(k,i  ,j-1)
        sw=mmk(k,i+1,j-1)
        nn=mmk(k,i-1,j  )
        mi=mmk(k,i  ,j  )
        ss=mmk(k,i+1,j  )
        ee=mmk(k,i  ,j+1)
        se=mmk(k,i+1,j+1)

        uww=u(ww)
        usw=u(sw)
        uss=u(ss)
        umi=u(mi)
        vmi=v(mi)
        vnn=v(nn)
        vee=v(ee)
        vww=v(ww)

        ust=quart*(uww+umi+usw+uss)

        !---  barokliner druck
        vpress=(press(mi)-press(ss))*ty/(half*(rho(mi)+rho(ss)))

        ! convective terms using vector upwind scheme:
        ! ... neglected on boundary during inflow
        if (vmi < zero) then
          urd  = ust*tx(i)
          vrd  = vmi*ty
          aurd = abs(urd)
          avrd = abs(vrd)

          if (vrd >= zero) then
            i2 = i+1
          else
            i2 = i-1
          endif
          if (urd >= zero) then
            j4 = j-1
          else
            j4 = j+1
          endif
          if (avrd >= aurd) then
            i3 = i2
            j1 = j
          else
            i3 = i
            j1 = j4
          endif

          vcon = avrd*(v(mmk(k,i,j1))-v(mmk(k,i2,j1)))                         &
               + aurd*(v(mmk(k,i3,j))-v(mmk(k,i3,j4)))
        else
          vcon = zero
        endif

        !---  austauschterm
        ! ... Smagorinsky terms ignored on boundary; 
        !     only back-ground Laplace filter diffusion
        vaus = vnn - three*vmi
        if (ee > 0 .and. se > 0) then
          vaus = vaus + vee
        else
          vaus = vaus + vmi
        endif
        if (ww > 0 .and. sw > 0) then
          vaus = vaus + vww
        else
          vaus = vaus + vmi
        endif
        vaus = dtd*vaus

        !- RHS of y-mom-eq stored in eb(:) -------------------------------------
        eb(k)=vmi-ft(i)*ust-vpress-vcon-vsrf+vaus
      enddo

      if (dyn) then
        !- The geopotential term
        do k=1,kev
          mi = mmk(k,i  ,j  )
          ss = mmk(k,i+1,j  )
          eb(k) = eb(k) - (geopot(mi)-geopot(ss))*ty
        enddo
      endif

      ustar = ust

      hrt     = tm(kev)+hgr
      tmp     = half*rt/hrt*sqrt(v(mmk(kev,i,j))*v(mmk(kev,i,j))+ustar*ustar) 
      eb(kev) = eb(kev) - tmp*v(mmk(kev,i,j))
      ec(kev) = ec(kev) + tmp

      eb(1)=eb(1)+tvtop
      if (.not.(casus(me(i,j)).and.casus(me(i+1,j)))) then
        eb(1)=eb(1)+half*eiscoy*decfv*v( mmk(1,i,j))
        ec(1)=ec(1)+half*eiscoy*decfv
      endif

      ! begin alpha, beta computation
      alpha=zero
      beta=zero
      alpha(kev)=eb(kev)
      beta(kev)=ec(kev)
      do k=kev-1,1,-1
        sdv      = ea(k+1)/(tm(k)*beta(k+1))
        sdvlum   = ea(k+1)/tm(k+1)
        alpha(k) = eb(k) + sdv*alpha(k+1)
        beta(k)  = ec(k) - sdv*sdvlum
      enddo
      ! end alpha, beta computation

      ! begin vn(:) computation
      !    force "nice" value outside domain i+1
      vneu = alpha(1)/beta(1)
      vn(mmk(1,i,  j)) = vneu
      vn(mmk(1,i+1,j)) = vneu
      do k=2,kev
        sdv=tm(k)
        vneu = (alpha(k)+ea(k)/sdv*vn(mmk(k-1,i,j)))/beta(k)
        vn(mmk(k, i, j)) = vneu
        vn(mmk(k,i+1,j)) = vneu
      enddo


      !---  trockenfallabfrage
      if (kev <= kdry) then
        ss = mmk(1,i+1,j)
        mi = mmk(1,i,j)
        call drying_case(ss, mi, vn(mi), hz(ss), hz(mi), z) 
        if (kdry > 1) then
          do k = 2,min(kev,kdry)
            call drying_case(ss,mi,vn(mmk(1,i,j)),hz(mmk(1,i,j)),              &
                             hz(mmk(1,i+1,j)),z)
          enddo
        endif
      endif
      ! end vn(:) computation

      !- turn back to original values for i and mi for safety:
      i  = i+1
      mi = mmk(1,i,j)

    endif

  enddo nbndloop

  !- That's all folks! ---------------------------------------------------------

end subroutine vobcalc

!===============================================================================

!-------------------------------------------------------------------------------
!
! calculates Velocity Along Boundary, 
! i.e. the velocity component in the direction along the open boundary.
!
! VAB = 0 @ inflow
! VAB is extracpolated from inside @ outflow
!
! Per Berg, DMI, August 25 2005
!
!-------------------------------------------------------------------------------

subroutine vabcalc  (mmk,khu,khv,u,v,un,vn,nz,krz,kdry,hz,z)

  !- modules -------------------------------------------------------------------
  use drying_cases,  only : drying_case

  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in) :: mmk(:,0:,0:)    ! index
  integer(4), intent(in) :: khu(0:),khv(0:)
  real(8), intent(in)    :: u(0:)
  real(8), intent(in)    :: v(0:)
  real(8), intent(inout) :: un(0:)
  real(8), intent(inout) :: vn(0:)
  integer(4), intent(in) :: nz              ! no. of bnd points
  integer(4), intent(in) :: krz(:,0:)       ! bnd poins and type
  integer(4), intent(in) :: kdry
  real(8), intent(in)    :: hz(0:)
  real(8), intent(in)    :: z(0:)

  !- local vars ----------------------------------------------------------------
  integer(4) :: i,j,k,nbnd,n,m,b,kb
  integer(4) :: nextn, nexti, nextj
  real(8)    :: dir 


  !- run though all boundary points --------------------------------------------
  nbndloop: do nbnd=1,nz
    i = krz(1,nbnd)
    j = krz(2,nbnd)
    b = krz(3,nbnd)
    m = mmk(1,i,j)
    
    if (b == 2) then
      !- North boundary:
      kb = khu(m)
      if (kb <= 0) cycle nbndloop

      ! find relevant point for determining the directional branch 
      nextn = min(nbnd+1, nz)
      nextj = j + 1
      if ((krz(3,nextn) == 2 .and. krz(2,nextn) == nextj .and.                 &
           krz(1,nextn) == i) .or. (mmk(1,i-1,j) <= 0) ) then
        ! E-neighbour point is also a N-boundary point 
        ! or N-neighbour point is outside domain
        ! and we will proceed as usual:
        nexti = i
        do k=1,kb
          n = mmk(k,i,j)
          dir = v(n)+v(mmk(k,nexti,nextj))
          if (dir <= zero) then
            un(n) = zero
          else
            un(n) = un(mmk(k,i+1,j))
          endif
        enddo   

      else
        ! interpolate u(north) and u(south) to obtain u.
        ! Note: skew time-centering since u(north) unavail at new time step.
        nexti = i-1
        nextj = j
        do k=1,kb
          n = mmk(k,i,j)
          un(n) = 0.5*(un(mmk(k,i+1,j))+u(mmk(k,nexti,nextj)))
        enddo

      endif 

      if (kb <= kdry) then
        do k=1,kb
          n = mmk(k,i,j)
          call drying_case(mmk(1,i,j),mmk(1,i,j+1),un(n),hz(n),                &
                           hz(mmk(k,i,j+1)),z)
        enddo
      endif

    elseif (b == 1) then
      !- West boundary:
      kb = khv(m)
      do k=1,kb
        n = mmk(k,i,j)
        dir = u(n)+u(mmk(k,i+1,j))
        if (dir >= zero) then
          vn(n) = zero
        else
          vn(n) = vn(mmk(k,i,j+1))
        endif
      enddo   

      if (kb <= kdry) then
        do k=1,kb
          n = mmk(k,i,j)
          call drying_case(mmk(1,i+1,j),mmk(1,i,j),vn(n),hz(mmk(k,i+1,j)),     &
                           hz(n),z)
        enddo
      endif

    elseif (b == 4) then
      !- South boundary:
      kb = khu(m)
      do k=1,kb
        n = mmk(k,i,j)
        dir = v(mmk(k,i-1,j))+v(mmk(k,i-1,j+1))
        if (dir >= zero) then
          un(n) = zero
        else
          un(n) = un(mmk(k,i-1,j))
        endif
      enddo   

      if (kb <= kdry) then
        do k=1,kb
          n = mmk(k,i,j)
          call drying_case(mmk(1,i,j),mmk(1,i,j+1),un(n),hz(n),                &
                           hz(mmk(k,i,j+1)),z)
        enddo
      endif

    elseif (b == 3) then
      !- East boundary:
      kb = khv(m)
      do k=1,kb
        n = mmk(k,i,j)
        dir = u(mmk(k,i,j-1))+u(mmk(k,i+1,j-1))
        if (dir <= zero) then
          vn(n) = zero
        else
          vn(n) = vn(mmk(k,i,j-1))
        endif
      enddo   

      if (kb <= kdry) then
        do k=1,kb
          n = mmk(k,i,j)
          call drying_case(mmk(1,i+1,j),mmk(1,i,j),vn(n),hz(mmk(k,i+1,j)),     &
                           hz(n),z)
        enddo
      endif

    endif

  enddo nbndloop


  !- That's all folks! ---------------------------------------------------------

end subroutine vabcalc

!===============================================================================

subroutine gradiz_w ( ke,mmk,w,nz,krz )
  !-----------------------------------------------------------------------------
  !
  ! - set vertical velocities to zero along boundary, and
  !
  !-----------------------------------------------------------------------------
  use dmi_omp, only : domp_get_domain

  implicit none

  integer(4), intent(in)    :: ke, nz, mmk(:,0:,0:), krz(:,0:)
  real(8),    intent(inout) :: w(0:)

  integer(4) :: iw, i, j, k, mi, nzl, nzu

  call domp_get_domain(1, nz, nzl, nzu)
  do iw=nzl,nzu
    i = krz(1,iw)
    j = krz(2,iw)
    do k=1,ke
      mi = mmk(k,i,j)
      if (mi <= 0) exit
      w(mi) = zero
    enddo
  enddo

end subroutine gradiz_w

!===============================================================================

subroutine gradiz_q ( mmk,dt,dx,dy,zn,nq,krq,rwq )
  !-----------------------------------------------------------------------------
  !
  ! - add source-inflow to z 
  !
  !-----------------------------------------------------------------------------
  use dmi_omp, only : domp_get_domain

  implicit none

  integer(4), intent(in)    :: nq, mmk(:,0:,0:), krq(:,0:)
  real(8),    intent(in)    :: dt, dx(:), dy, rwq(0:)
  real(8),    intent(inout) :: zn(0:)

  integer(4) :: iq, i, mi, nql, nqu
  real(8)    :: dtdy

  dtdy = dt/dy

  call domp_get_domain(1, nq, nql, nqu, .true.)
  do iq=nql,nqu
    i  = krq(1,iq)
    mi = mmk(1,i,krq(2,iq))

    if (mi > 0) zn(mi) = zn(mi) + rwq(iq)*dtdy/dx(i)
  enddo

end subroutine gradiz_q

!===============================================================================

subroutine gradiz_dyn ( mmk,dt,dx,dy,zn,nq,krq,rwq,h )
  !-----------------------------------------------------------------------------
  !
  ! In the dynamic case:
  ! - add source-inflow to z
  !
  !-----------------------------------------------------------------------------

  use dmi_omp, only : domp_get_domain

  implicit none

  integer(4), intent(in) :: nq
  integer(4), intent(in) :: mmk(:,0:,0:)
  integer(4), intent(in) :: krq(:,0:)
  real(8), intent(in)    :: dt, dx(:), dy
  real(8), intent(in)    :: rwq(0:)
  real(8), intent(inout) :: zn(0:)
  real(8), intent(inout) :: h(0:)

  integer(4) :: iw, i, mi, nql, nqu
  real(8)    :: dtdy

  dtdy = dt/dy

  call domp_get_domain(1, nq, nql, nqu)

  do iw=nql,nqu
    i  = krq(1,iw)
    mi = mmk(1,i,krq(2,iw))

    if (mi > 0) then
      zn(mi) = zn(mi) + rwq(iw)*dtdy/dx(i)
      h(mi)  = h(mi) + rwq(iw)*dtdy/dx(i)  
    endif   
  enddo

end subroutine gradiz_dyn

!===============================================================================

subroutine zrano ( lun,nzno,krzno,rwzno,lmn,coldst,OPRFMT )
!-------------------------------------------------------------------------------
!
!     z-boundary values at the open boundaries of the model
!
!     NOTE:
!     switch OPRFMT chooses between integer(4) (i5) and real (f7.2)
!     format of the surge input file
!     A wrong selection will most likely cause model blow-up
!     and certainly wrong results.
!
!-------------------------------------------------------------------------------

  !- modules -------------------------------------------------------------------
  use exits,     only : exitme
  use constants, only : omg, hg, gg, aufdat, izeits, zeitstu, zref

  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in) :: lun,nzno,lmn
  integer(4), intent(in) :: krzno(:,0:)
  real(8), intent(inout) :: rwzno(0:)
  logical, intent(in)    :: coldst,OPRFMT

  !- local vars ----------------------------------------------------------------
  character(256),          save :: cformat
  character(4),            save :: iformat
  real(8), allocatable,    save :: zr(:)
  integer(4), allocatable, save :: izr(:)
  logical,                 save :: firsttime = .true.
  integer(4)                    :: klaus, l, mrno, ios
  real(8)                       :: timsurge, zet, arg
                                   ! NOTE: timsurge is read in but never used
  real(8),                 save :: sc

  !- return if we have nothing to do here --------------------------------------
  if (nzno == 0) return

  !- determine some vars dependent on surge file format ------------------------
  if (firsttime) then
    if (OPRFMT) then
      ! mm
      sc = mm2m
    else
      ! cm
      sc = cm2m
    endif
  endif

  !- read from surge model -----------------------------------------------------
  if (aufdat(1:1) /= 'G') then 
    if (mod(izeits-1,lmn) == 0) then
      if (firsttime) then
        !- initialise on first entry:
        firsttime = coldst

        mrno = nzno

        allocate( zr(mrno) )

        if (mrno < 10) then
          iformat = '(i1)'
        elseif (mrno < 100) then
          iformat = '(i2)'
        elseif (mrno < 1000) then
          iformat = '(i3)'
        elseif (mrno < 10000) then
          iformat = '(i4)'
        else
          ! trap 
          call exitme('ERROR: Dimension of open surge boundary too large.')
        endif
        write(cformat,iformat) mrno

        if (OPRFMT) then
          allocate( izr(mrno) )
          ! integer(4)s in mm:
          cformat = '(20X,f15.8,'//trim(cformat)//'i5)'
        else
          ! reals in cm:
          cformat = '(20X,f15.8,'//trim(cformat)//'f7.2)'
        endif
      endif

      !- read new bnd surge values:
      if (OPRFMT) then
        ! integer(4)s in mm:
        read(lun,cformat,iostat=ios) timsurge, izr
        if (ios /= 0) call exitme(1, 'Cannot read surge data file' )
        zr = real( izr, 8 )
      else
        ! reals in cm:
        read(lun,cformat,iostat=ios) timsurge, zr
        if (ios /= 0) call exitme(1, 'Cannot read surge data file' )
      endif

      if (firsttime .and. coldst) then
        !- initialise on first entry:
        firsttime = .false.

        zr(1:nzno) = rwzno(1:nzno)/sc
      endif
    endif ! end surge 
  elseif (firsttime) then
    firsttime = .false.
    allocate( zr(nzno) )
    zr(:) = zero
    sc    = zero
  endif 

  !- scale to m and add tidal signal -------------------------------------------
  do klaus=1,nzno 

    if (krzno(3,klaus) > 0) then

      !- convert cm or mm to m:
      zet = zr(klaus)*sc

      !- add tidal konstituents:
      do l=1,17
        arg = omg(l)*zeitstu + gg(l,klaus)
        zet = zet + hg(l,klaus)*cos(arg)
      enddo

      ! add zref and transfer:
      rwzno(klaus) = zet + zref
    endif

  enddo

  !- That's all folks! ---------------------------------------------------------

end subroutine zrano

!===============================================================================

subroutine compbno (mmk,khu,khv,h,rho,press,geopot,nbpz,krz,zetacno)

  !- modules -------------------------------------------------------------------
  use constants, only : g
  
  !- directives ----------------------------------------------------------------
  implicit none
  
  !- arguments -----------------------------------------------------------------
  integer(4), intent(in) :: nbpz
  integer(4), intent(in) :: mmk(:,0:,0:)
  integer(4), intent(in) :: khu(0:),khv(0:)
  integer(4), intent(in) :: krz(:,0:)
  real(8), intent(in)    :: h(1:,0:)
  real(8), intent(in)    :: rho(1:,0:), press(1:,0:),geopot(1:,0:)
  real(8), intent(out)   :: zetacno(0:)
 
  !- local vars ----------------------------------------------------------------
  real(8), parameter   :: half = 0.5_8
  real(8), parameter   :: dg = 1.0_8/g
  integer(4)           :: n,i,j,k,na,nn,ne,i1,j1,kbu,kbv,ii,jj,n1
  real(8)              :: pressc,dpress,dpress1,dpress2,rwzm,srwz
  real(8)              :: tiefex,tiefey,hhx,hhy
  logical              :: end_of_bnd
  
  !- nothing to do here ?
  if (nbpz <= 0) return
  
  !- initialise on every entry:
  zetacno(0:nbpz) = zero
 
  na = 1
  nn = 0
  srwz = zero
  pressc = zero
  end_of_bnd = .false.
  do n=1,nbpz
    nn = nn+1

    i  = krz(1,n)
    j  = krz(2,n)

    n1 = n+1
    i1 = krz(1,min(n1,nbpz))
    j1 = krz(2,min(n1,nbpz))

    if (n == nbpz .or. abs(i1-i) > 1 .or. abs(j1-j) > 1) then
      !- end point found
      ne = n
      end_of_bnd = .true.

    else
      !- this is not an end point, but ...

      if (i1 == i) then
        !- ... it's a point on a straight N or S bnd:
        jj     = min( j, j1 )
        kbu    = khu(mmk(1,i,jj))
        tiefex = zero
        dpress = zero
        do k=1,kbu
          hhx    = half*(h(k,n)+h(k,n1))
          tiefex = tiefex+hhx
          dpress = dpress - hhx*(((press(k,n1) -press(k,n)) /                  &
                       (half*(rho(k,n1)+rho(k,n))))+(geopot(k,n1)-geopot(k,n)))
        enddo
        dpress = dpress/tiefex

      elseif (j1 == j) then
        !- ... it's a point on a straight E or W bnd:
        ii     = min( i, i1 )
        kbv    = khv(mmk(1,ii,j))
        tiefey = zero
        dpress = zero
        do k=1,kbv
          hhy    = half*(h(k,n)+h(k,n1))
          tiefey = tiefey+hhy
          dpress = dpress - hhy*(((press(k,n1) - press(k,n)) /                 &
                       (half*(rho(k,n1)+rho(k,n))))+(geopot(k,n1)-geopot(k,n)))
        enddo
        dpress = dpress/tiefey 

!     else
!FIXME: The implementation is broken for non-grid-aligned bnd-points
!       !- ... it's a point on a non-grid-aligned bnd:
!       ii  = min( i, i1 )
!       jj  = min( j, j1 )
!       ma  = 0
!       mb  = 0

!       !- treat a NW-SE sloping N-bnd, or SW-NE sloping S-bnd:
!       kbu = khu(mmk(1,i1,jj))
!       kbv = khv(mmk(1,ii,j))
!       if (kbu*kbv > 0) then
!         ma = 1
          dpress1 = zero
!         tiefex  = zero
          dpress  = zero
!         do k=1,kbu
!           hhx    = half*(h(k,mmk(1,i1,j))+h(k,n1))
!           tiefex = tiefex+hhx
!           dpress =dpress-hhx*(((press(k,n1)-press(k,mmk(1,i1,j)))/           &
!                             (half*(rho(k,n1)+rho(k,mmk(1,i1,j))))+           &
!                             (geopot(k,n1)-geopot(k,mmk(1,i1,j)))  )
!         enddo
!         dpress1 = dpress1+dpress/tiefex

!         tiefey = zero
          dpress = zero
!         do k=1,kbv
!           hhy    = half*(h(k,n)+h(k,mmk(1,i1,j)))
!           tiefey = tiefey+hhy
!           dpress=dpress-hhy*(((press(k,mmk(1,i1,j)) - press(k,n))/           &
!                             (half*(rho(k,mmk(1,i1,j))+rho(k,n))))+           &
!                             (geopot(k,mmk(1,i1,j))-geopot(k,n))   )
!         enddo
!         dpress1 = dpress1+dpress/tiefey
!       endif

!       !- treat a SW-NE sloping N-bnd, or NW-SE sloping S-bnd:
!       kbu = khu(mmk(1,i,jj))
!       kbv = khv(mmk(1,ii,j1))
!       if (kbu*kbv > 0) then
!         mb = 1
          dpress2 = zero
!         tiefey  = zero
          dpress  = zero
!         do k=1,kbv
!           hhy    = half*(h(k,mmk(1,i,j1))+h(k,n1))
!           tiefey = tiefey+hhy
!           dpress=dpress-hhy*(((press(k,n1)-press(k,mmk(1,i,j1)))/            &
!                           (half*(rho(k,n1)+rho(k,mmk(1,i,j1)))))+            &
!                           (geopot(k,n1)-geopot(k,mmk(1,i,j1))) )
!         enddo
!         dpress2 = dpress2+dpress/tiefey

!         tiefex = zero
          dpress = zero
!         do k=1,kbu
!           hhx    = half*(h(k,n)+h(k,mmk(1,i,j1)))
!           tiefex = tiefex+hhx
!           dpress = dpress-hhx*(((press(k,mmk(1,i,j1))-press(k,n))/           &
!                             (half*(rho(k,mmk(1,i,j1))+rho(k,n))))+           &
!                             (geopot(k,mmk(1,i,j1))-geopot(k,n)) )
!         enddo
!         dpress2 = dpress2+dpress/tiefex
!       endif

!       if (ma*mb > 0) then
!         dpress = half*(dpress1+dpress2)
!       elseif (ma > 0) then
!         dpress = dpress1
!       elseif (mb > 0) then
!         dpress = dpress2
!       endif
      endif

      !- correction at neighbour point:
      pressc       = pressc + dpress
      zetacno(n+1) = pressc*dg
      srwz = srwz + zetacno(n+1)
    endif

    if (end_of_bnd) then
      !- finnish this bnd by making corrections:
      rwzm = srwz/nn
      zetacno(na:ne) = zetacno(na:ne) - rwzm

      !- prepare for next boundary:
      if (n < nbpz) then
        na = n+1
        nn = 0
        end_of_bnd = .false.
        pressc = zero
        srwz = zero
      endif
    endif
  enddo
  
end subroutine compbno

!===============================================================================

subroutine densbno (mmk,khu,khv,hx,hy,rho,press,nbpz,krz,zetacno)

  use constants, only : g

  implicit none
 
  integer(4), intent(in) :: nbpz
  integer(4), intent(in) :: mmk(:,0:,0:)
  integer(4), intent(in) :: khu(0:),khv(0:)
  integer(4), intent(in) :: krz(:,0:)
  real(8), intent(in)    :: hx(1:,0:), hy(1:,0:)
  real(8), intent(in)    :: rho(1:,0:), press(1:,0:)
  real(8), intent(out)   :: zetacno(0:)
 
 
  real(8), parameter   :: dg = 1.0_8/g
  integer(4)           :: n,i,j,k,na,nn,ne,i1,j1,kbu,kbv,ii,jj,n1,nh
  real(8)              :: pressc,dpress,rwzm
  real(8)              :: tiefex,tiefey,hhx,hhy
  logical              :: end_of_bnd
 
  !- nothing to do here ?
  if (nbpz <= 0) return
 
  !- initialise on every entry:
  zetacno(0:nbpz) = zero
 
  na = 1
  nn = 0
  pressc = zero
  end_of_bnd = .false.
  do n=1,nbpz
    nn = nn+1

    i  = krz(1,n)
    j  = krz(2,n)

    n1 = n+1
    i1 = krz(1,min(n1,nbpz))
    j1 = krz(2,min(n1,nbpz))

    if (n == nbpz .or. abs(i1-i) > 1 .or. abs(j1-j) > 1) then
      !- end point found
      ne = n
      end_of_bnd = .true.

    else
      !- this is not an end point, but ...

      if (i1 == i) then
        !- ... it's a point on a straight N or S bnd:
        if (j <= j1) then
          jj = j
          nh = n
        else
          jj = j1
          nh = min(n1,nbpz)
        endif
        kbu    = khu(mmk(1,i,jj))
        tiefex = zero
        dpress = zero
        do k=1,kbu
          hhx    = hx(k,nh)
          tiefex = tiefex+hhx
          dpress = dpress-hhx*(press(k,n1)-press(k,n))
        enddo
        dpress = dpress/(tiefex*rho(1,n1))
  
      elseif (j1 == j) then
        !- ... it's a point on a straight E or W bnd:
        if (i <= i1) then
          ii = i
          nh = n
        else
          ii = i1
          nh = min(n1,nbpz)
        endif
        kbv    = khv(mmk(1,ii,j))
        tiefey = zero
        dpress = zero
        do k=1,kbv
          hhy    = hy(k,nh)
          tiefey = tiefey+hhy
          dpress = dpress-hhy*(press(k,n1)-press(k,n))
        enddo
        dpress = dpress/(tiefey*rho(1,n1)) 

      else
        !- ... it's a point on a non-grid-aligned bnd:
!FIXME: the implementation is broken for non-grid-aligned bnds!!!!
        dpress = zero
!       ii  = min( i, i1 )
!       jj  = min( j, j1 )
!       ma  = 0
!       mb  = 0

        !- treat a NW-SE sloping N-bnd, or SW-NE sloping S-bnd:
!       kbu = khu(mmk(1,i1,jj))
!       kbv = khv(mmk(1,ii,j))
!       if (kbu*kbv > 0) then
!         ma = 1
!         dpress1 = zero
!         tiefex  = zero
!         dpress  = zero
!         do k=1,kbu
!           hhx    = hx(mmk(k,i1,jj))
!           tiefex = tiefex+hhx
!           dpress = dpress-hhx*(pressbno(k,n1)-press(mmk(k,i1,j)))
!         enddo
!         dpress1 = dpress1+dpress/(tiefex*rhobno(1,n1))

!         tiefey = zero
!         dpress = zero
!         do k=1,kbv
!           hhy    = hy(mmk(k,ii,j))
!           tiefey = tiefey+hhy
!           dpress = dpress-hhy*(press(mmk(k,i1,j))-pressbno(k,n))
!         enddo
!         dpress1 = dpress1+dpress/(tiefey*rho(mmk(1,i1,j)))
!       endif

        !- treat a SW-NE sloping N-bnd, or NW-SE sloping S-bnd:
!       kbu = khu(mmk(1,i,jj))
!       kbv = khv(mmk(1,ii,j1))
!       if (kbu*kbv > 0) then
!         mb = 1
!         dpress2 = zero
!         tiefey  = zero
!         dpress  = zero
!         do k=1,kbv
!           hhy    = hy(mmk(k,ii,j1))
!           tiefey = tiefey+hhy
!           dpress = dpress-hhy*(pressbno(k,n1)-press(mmk(k,i,j1)))
!         enddo
!         dpress2 = dpress2+dpress/(tiefey*rhobno(1,n1))

!         tiefex = zero
!         dpress = zero
!         do k=1,kbu
!           hhx    = hx(mmk(k,i,jj))
!           tiefex = tiefex+hhx
!           dpress = dpress-hhx*(press(mmk(k,i,j1))-pressbno(k,n))
!         enddo
!         dpress2 = dpress2+dpress/(tiefex*rho(mmk(1,i,j1)))
!       endif

!       if (ma*mb > 0) then
!         dpress = half*(dpress1+dpress2)
!       elseif (ma > 0) then
!         dpress = dpress1
!       elseif (mb > 0) then
!         dpress = dpress2
!       endif

      endif

      !- correction at neighbour point:
      pressc       = pressc + dpress
      zetacno(n+1) = pressc*dg
    endif

    if (end_of_bnd) then
      !- finnish this bnd by making corrections to each side of midpoint:
      rwzm = zetacno((ne+na)/2)
      zetacno(na:ne) = zetacno(na:ne) - rwzm

      !- prepare for next boundary:
      if (n < nbpz) then
        na = n+1
        nn = 0
        end_of_bnd = .false.
        pressc = zero
      endif
    endif
  enddo

end subroutine densbno

!===============================================================================

subroutine zetar ( mmk,zn,nz,krz,rwz,kh,h,dpth )
  !-----------------------------------------------------------------------------
  !
  ! prescribe the z-boundary conditions
  !
  ! in the dynamical case there is also a prescription of the boundary 
  ! layer thicknesses.
  !
  !-----------------------------------------------------------------------------

  use constants,     only : dyn, hgr
  use exits,         only : exitme

  implicit none

  integer(4), intent(in) :: nz, mmk(:,0:,0:), krz(:,0:)
  real(8), intent(in)    :: rwz(0:)
  real(8), intent(inout) :: zn(0:)

  integer(4), intent(in),    optional :: kh(0:)
  real(8),    intent(in),    optional :: dpth(0:)
  real(8),    intent(inout), optional :: h(0:)

  integer(4) :: iw, mi, i, j, k, kb
  real(8)    :: scal, zzn
  
  if (.not.dyn) then
    do iw=1,nz
      if (krz(3,iw) > 0) then
        mi = mmk(1,krz(1,iw),krz(2,iw))
        if (mi <= 0) cycle
        zn(mi) = rwz(iw)
      endif
    enddo
  else ! dyn
    do iw=1,nz
      if (krz(3,iw) > 0) then

        mi = mmk(1,krz(1,iw),krz(2,iw))
        if (mi <= 0) cycle

        !- prescribed boundary conditions:
        zzn = max(rwz(iw),-dpth(mi)+hgr)
        i   = krz(1,iw)
        j   = krz(2,iw)
        kb  = kh(mi)

        scal = (dpth(mi)+zzn)/(dpth(mi)+zn(mi))
        if (scal <= zero) then
          call exitme(1,'A watercolum with a negative depth ... program stops')
        endif
        do k=1,kb
          h(mmk(k,i,j)) = scal*h(mmk(k,i,j))
        end do
      
        zn(mi) = max(rwz(iw),-dpth(mi)+hgr)
      endif
    enddo
  endif
 
end subroutine zetar

!===============================================================================

  subroutine ice_vel(nuvs2d, induvse, mmk, il, iu, jl, ju, casus, ue, ve)
    implicit none

    integer(4), intent(in) :: nuvs2d, induvse(:,:) ! global
    integer(4), intent(in) :: mmk(1:,0:,0:), il, iu, jl, ju
    logical, intent(in)    :: casus(0:)            
    real(8), intent(inout) :: ue(0:), ve(0:)

    integer(4) :: nn, i, j, ml, mr, mo, mu, mi

    do nn=1,nuvs2d
      i = induvse(1,nn)
      j = induvse(2,nn)

      if (i < il .or. iu < i .or. j < jl .or. ju < j) cycle

      mi = mmk(1,i,j)
      if (mi <= 0) then
        ue(mi) = zero
        ve(mi) = zero
        cycle
      endif

      if (induvse(3,nn) == 1) then
        ml = mi
        mr = mmk(1,i,j+1)
        if (casus(ml).and.casus(mr)) then
          ue(ml) = zero
        endif
      elseif (induvse(3,nn) == 2) then
        mo = mi
        mu = mmk(1,i+1,j)
        if (casus(mo).and.casus(mu)) then
          ve(mo) = zero
        endif
      endif
    enddo

  end subroutine ice_vel

!===============================================================================

  subroutine checkopenbnd( kh, me, khbnd )
  !-----------------------------------------------------------------------------
  ! Make sure that we only treat open boundaries that are defined on the
  ! present MPI task.
  !-----------------------------------------------------------------------------

    use dmi_mpi,     only : dd, mpi_io_serial
    use cmod_params, only : mainarea, nz, krz
    use old_ice_dyn, only : ne
    use constants,   only : icedynactive, shortcircuit

    implicit none

    integer(4), intent(in)  :: kh(0:), me(0:,0:)
    integer(4), intent(out) :: khbnd(0:)

    integer(4) :: ia, il, iu, jl, ju, ibnd, jbnd, n
    logical    :: found

    !  handle only the main area:
    ia = mainarea

    !  always assign khbnd(0):
    khbnd(0) = 0
  
    !  skip, if we have nothing to do here:
    if (nz(ia) <= 0) return

    ! adjust the No. of open bnd points if they are not included in this task
    !   - check all boundary points, 
    !   - if none are on this task we take nz and ne equal to zero, otherwise we
    !     will do nothing; always keep I/O task as is
    !   - assign values to khbnd(:)
    il    = dd(ia)%low_i
    iu    = dd(ia)%up_i
    jl    = dd(ia)%low_j
    ju    = dd(ia)%up_j
    found = .false.
    do n=1,nz(ia)
      ibnd = krz(ia)%p(1,n)
      jbnd = krz(ia)%p(2,n)
      if (.not.found) then
        found = (ibnd >= il .and. ibnd <= iu .and. jbnd >= jl .and. jbnd <= ju) 
      endif
      khbnd(n) = kh(me(ibnd,jbnd))
    enddo
    if ((.not.found) .and. (.not.mpi_io_serial)) nz(ia) = 0

    if ((.not.found) .and. (.not.mpi_io_serial) .and.                          &
        icedynactive .and. (.not.shortcircuit))  ne(ia) = 0

  end subroutine checkopenbnd

!===============================================================================

  subroutine masseqs_bnd_dyn (ia,nz,kmax,mmk,krz,khu,khv,kh,dx,dy,dt,covo,covu,&
                              hu,hv,un,vn,z,zn,ti,h,h0,hn,ws,avv)

  !-----------------------------------------------------------------------------
  ! Use momentum equation and coarse grid velocity components to 
  ! obtain fine grid velocity components at border.
  !
  !
  ! The method works with the following border types (as seen from fine grid):
  !   unknown border type:  < 0, no action
  !   not a border:         0, no action
  !   W border:             1
  !   N border:             2
  !   E border:             3
  !   S border:             4
  ! Each border must be enumerated in a specific order in increments of 1:
  !   W: from S to N
  !   N: from W to E
  !   E: from N to S
  !   S: from E to W
  ! Each border must be fully connected in the specifications.
  ! There can be more than one border of each type.
  ! 
  ! Convex corners (as seen from fine grid) need special treatment. The corner 
  ! point (in fine grid) must be defined on BOTH of the joined borders, 
  ! enumerated by an increment of 1 clockwise, eg 
  !   for a W/N corner, where the W border ends at brd-point No. 7, the N border
  !   must start at brd-point No. 8.
  ! 
  ! Concave corners (as seen from fine grid) need no special treatment with the 
  ! present method.
  !
  !
  ! Initiated Nov. 28, 2005, by Per Berg, COI/DMI.
  !-----------------------------------------------------------------------------

      !- modules  --------------------------------------------------------------
      use constants,      only : two, hgr, wghtz, wghthar, wghteta, wghtturb,  &
                                 gammadt
      use dmi_mpi_global, only : iu06
      use exits,          only : exitme
    
    
      !- directives ------------------------------------------------------------
      implicit none
    
    
      !- arguments -------------------------------------------------------------
      integer(4), intent(in)       :: ia
      integer(4), intent(in)       :: nz,kmax
      integer(4), intent(in)       :: mmk(:,0:,0:)
      integer(4), intent(in)       :: krz(:,0:)
      integer(4), intent(in)       :: khu(0:),khv(0:)
      integer(4), intent(in)       :: kh(0:)
      real(8), intent(in)          :: dy,dt
      real(8), intent(in)          :: dx(:),covo(:),covu(:)
      real(8), intent(in)          :: hu(0:),hv(0:),un(0:),vn(0:)
      real(8), intent(in)          :: avv(0:)
      real(8), intent(in)          :: z(0:),ti(0:)
      real(8), intent(in)          :: h(0:),h0(0:)
      real(8), intent(inout)       :: hn(0:)
      real(8), intent(inout)       :: ws(0:), zn(0:)

      !- local vars ------------------------------------------------------------
      real(8),dimension(1:kmax)   :: h_tmp, h_inc
      real(8),dimension(1:kmax)   :: div,diverg,wpsd
      real(8),dimension(1:kmax)   :: a_h,b_h,c_h,e_h
      real(8),dimension(1:kmax+1) :: r_h,z_h
      integer(4)           :: nnz,nn,kb,ncorner
      integer(4)           :: i,j,k
      integer(4)           :: kbue,kbuw,kbvn,kbvs
      integer(4)           :: iterat,it_max
      integer(4)           :: kbw,kbe,kbn,kbs
      integer(4)           :: mi
      real(8)              :: divk, dxi, covoidy, covuidy
      real(8)              :: depth_n,w_pseudo
      real(8)              :: hue,huw,hvn,hvs
      real(8)              :: norm,scal
      real(8)              :: dh,fac
      real(8)              :: length_a,length_b
      real(8)              :: h_tmp_a, h_tmp_b
    
      !- local parameters ------------------------------------------------------
      real(8),parameter    ::  scal_max = 1.0e+6_8, norm_min = 1.0e-9_8
                        
     
      do nnz = 1, nz
        i = krz(1,nnz)
        j = krz(2,nnz)

        nn = mmk(1,i,j)
        if (nn <= 0) cycle
        kb = kh(nn)

        dxi   = dx(i)
        covoidy = covo(i)/dy
        covuidy = covu(i)/dy

        ! only do corner points once
        if (nnz > 1) then
          ! skip a corner point that has already been handled
          if (krz(3,nnz-1) > 0  .and.                                          &
              i == krz(1,nnz-1) .and. j == krz(2,nnz-1)) cycle
        endif

        ncorner = 0
        if (nnz < nz) then
          ! first-time identification of a convex corner point
          if (krz(3,nnz+1) > 0  .and.                                          &
              i == krz(1,nnz+1) .and. j == krz(2,nnz+1)) then
            ! identify this corner by a code:
            !   unknown corner type:  < 0
            !   not a corner:           0
            !   N/W corner:             1
            !   N/E corner:             2
            !   S/E corner:             3
            !   S/W corner:             4
            ncorner = krz(3,nnz)
          endif
        endif

        h_inc(:) = zero
        div(:)   = zero
        diverg(:)= zero
        wpsd(:)  = zero

        depth_n = zero
      
        !- Calculate v * dz/ds and Grad_s * v * dz/ds, i.e. the component of 
        !- vertical velocity which depends on the horizontal velocity resp. the 
        !- horizontal water flux. 
        do k=1,kb
          mi = mmk(k, i, j)
          !-- default four-point stencil around actual grid point
          hue=hu(mi)*un(mi)
          huw=hu(mmk(k,i,j-1))*un(mmk(k,i,j-1))
          hvn=hv(mmk(k,i-1,j))*vn(mmk(k,i-1,j))
          hvs=hv(mi)*vn(mi)

          ! -- overrule defaults according to border and corner
          !at points at western boundary
          if (krz(3,nnz) == 1) then
            huw=hu(mi)*un(mi)
            if(ncorner > 0) then
              hvn=hv(mi)*vn(mi)
            endif

          !at points at northern boundary
          elseif (krz(3,nnz) == 2) then
            hvn=hv(mi)*vn(mi)
            if(ncorner > 0) then
              hue=hu(mmk(k,i,j-1))*un(mmk(k,i,j-1))
            endif

          !at points at eastern boundary 
          elseif (krz(3,nnz) == 3) then
            hue=hu(mmk(k,i,j-1))*un(mmk(k,i,j-1))
            if(ncorner > 0) then
              hvs=hv(mmk(k,i-1,j))*vn(mmk(k,i-1,j))
            endif

          !at points at southern boundary 
          elseif (krz(3,nnz) == 4) then
            hvs=hv(mmk(k,i-1,j))*vn(mmk(k,i-1,j))
            if(ncorner > 0) then
              huw=hu(mi)*un(mi)
            endif
          endif

          divk=(hue-huw)/dxi+(hvn*covoidy-hvs*covuidy)

          depth_n = depth_n + h(mi) - dt*divk
          div(k)    = divk
          diverg(k) = divk          
          h_tmp(k) = h (mi)
        enddo
     

        !- vertical redistribution of divergence so as to smooth layer 
        !- thicknesses wghthar*dh/dx(i) stands for Grad_s 
        !- (velocity * length * Grad_s (dz/ds)), whereas the inner Grad_s is 
        !- already in the factor dh, since dh = dh+h(mm(i,j,k))-h(mm(i,j+1,k))
        !- depends of course direcly on the space between two cells, which is 
        !- described by dx or dy.

        ! Grad_s(dz/ds) conform to ds*Grad_s(dz/ds) in descrete.
        ! dh = lenght * Grad_s (dz/ds)

        kbue = khu(mmk(1,i,j))
        dh = zero
        do k=1,kbue-1
          dh=dh+h(mmk(k,i,j))-h(mmk(k,i,j+1))
          div( k )=div( k )+wghthar*dh/dxi
          div(k+1)=div(k+1)-wghthar*dh/dxi
        end do
        kbuw=khu(mmk(1,i,j-1))
        dh = zero
        do k=1,kbuw-1
          dh=dh+h(mmk(k,i,j))-h(mmk(k,i,j-1))
          div( k )=div( k )+wghthar*dh/dxi
          div(k+1)=div(k+1)-wghthar*dh/dxi
        end do
        kbvn=khv(mmk(1,i-1,j))
        dh = zero
        do k=1,kbvn-1
          dh=dh+h(mmk(k,i,j))-h(mmk(k,i-1,j))
          div( k )=div( k )+wghthar*dh*covoidy
          div(k+1)=div(k+1)-wghthar*dh*covoidy
        end do
        kbvs=khv(mmk(1,i,j))
        dh = zero
        do k=1,kbvs-1
          dh=dh+h(mmk(k,i,j))-h(mmk(k,i+1,j))
          div( k )=div( k )+wghthar*dh*covuidy
          div(k+1)=div(k+1)-wghthar*dh*covuidy
        end do


        !- levelling at bottom steps, this part smoothes the water-transfer at 
        !- bottom steps, i.e. there where a water column is deeper than the 
        !- neighbour-water-column.

        kbw = kh(mmk(1,i,j-1))
        if (kb > kbw .and. kbw > 0) then
          dh = ti(mmk(1,i,j-1))+z(mmk(1,i,j-1))
          do k=1,kbw
            dh = dh-h(mmk(k,i,j))
          end do
          dh = max(dh,zero)
          div(kbw)   = div(kbw)   - gammadt*dh
          div(kbw+1) = div(kbw+1) + gammadt*dh
        end if

        kbe = kh(mmk(1,i,j+1))
        if (kb > kbe .and. kbe > 0) then
          dh = ti(mmk(1,i,j+1))+z(mmk(1,i,j+1))
          do k=1,kbe
            dh = dh-h(mmk(k,i,j))
          end do
          dh = max(dh,zero)
          div(kbe)   = div(kbe)   - gammadt*dh
          div(kbe+1) = div(kbe+1) + gammadt*dh
        end if

        kbn = kh(mmk(1,i-1,j))
        if (kb > kbn .and. kbn > 0) then
          dh = ti(mmk(1,i-1,j))+z(mmk(1,i-1,j))
          do k=1,kbn
            dh = dh-h(mmk(k,i,j))
          end do
          dh = max(dh,zero)
          div(kbn)   = div(kbn)   - gammadt*dh
          div(kbn+1) = div(kbn+1) + gammadt*dh
        end if

        kbs = kh(mmk(1,i+1,j))
        if (kb > kbs .and. kbs > 0) then
          dh = ti(mmk(1,i+1,j))+z(mmk(1,i+1,j))
          do k=1,kbs
            dh = dh-h(mmk(k,i,j))
          end do
          dh = max(dh,zero)
          div(kbs)   = div(kbs)   - gammadt*dh
          div(kbs+1) = div(kbs+1) + gammadt*dh
        end if

        norm = one
        scal = zero
        iterat = 0
      
        a_h(1:kmax)   = zero
        b_h(1:kmax)   = zero
        c_h(1:kmax)   = zero
        e_h(1:kmax)   = zero
        r_h(1:kmax+1) = zero
        z_h(1:kmax+1) = zero
      
        it_max = 25 * kb

        !- Now the last term of the equation which is to be solved will be 
        !- integrated:
        !- dt*(wghteta+(wghtturb*avv))*(length_b/h_tmp( k )-length_a/h_tmp(k-1))
        !- is the parametrisation of the term d/ds * (d/ds ( mu / (dz/ds) ) ). 
        !- This parametrisation prevents the case that one layer disappears, 
        !- i.e. the layer thickness becomes zero, because this term trys always 
        !- to push the thicknesses in the origin state.

        !- To solve the whole equation of continuity over the vertical the 
        !- Newton iteration is used.

        if (depth_n > zero) then
          iteration: do while ( norm > norm_min .and. scal <= scal_max )

            do k=1,kb
              c_h(k) = one+wghtz
              e_h(k) = (one+wghtz)*(h(mmk(k,i,j))-h_tmp(k)) - dt*div(k)
            end do

            do k=1, kb
              e_h(1) = e_h(1) - dt*wghtz*div(k)
            enddo

            do k=2,kb
              !- the layer thicknesses h0 act like a reference thickness or
              !- resting state
              length_a = h0(mmk(k-1,i,j))
              length_b = h0(mmk(k,i,j))

              h_tmp_a  = h_tmp(k-1)
              h_tmp_b  = h_tmp(k)
 
              fac = wghteta + wghtturb*avv(mmk(k,i,j))
   
              a_h( k ) =        - dt*fac*length_a/(h_tmp_a*h_tmp_a)
              c_h( k ) = c_h(k) + dt*fac*length_b/(h_tmp_b*h_tmp_b)
              e_h( k ) = e_h(k) + dt*fac*( length_b/h_tmp( k )                 &
                                - length_a/h_tmp(k-1) )

              b_h(k-1) =         - dt*fac*length_b/(h_tmp_b*h_tmp_b)
              c_h(k-1) = c_h(k-1)+ dt*fac*length_a/(h_tmp_a*h_tmp_a)
              e_h(k-1) = e_h(k-1)+ dt*fac*( length_a/h_tmp(k-1)                &
                                  - length_b/h_tmp( k ) )
            end do


            !- line inversion (factorization) method to compute iterative 
            !- increments of cell thickness
            !- solved by using the Thomas algorithm the system of equations 
            !- c_h(k)*h_inc(k)+a_h(k)*h_inc(k-1)+b_h(k)*h_inc(k+1)=e_h(k) to 
            !- get the  increment h_inc of the layer thicknesses to calculate
            !- the new h(:) (hn(:)). 

            do k=1,kb
              r_h(k+1)=(e_h(k)-a_h(k)*r_h(k))/(c_h(k)+a_h(k)*z_h(k))
              z_h(k+1)=-b_h(k)/(c_h(k)+a_h(k)*z_h(k))
            end do

            h_inc(kb)=r_h(kb+1)
            do k=kb-1,1,-1
              h_inc(k)=z_h(k+1)*h_inc(k+1)+r_h(k+1)
            end do

            norm = zero
            scal = one
            do k=1,kb
              norm = max(norm,abs(h_inc(k)))
              scal = max(scal,-two*(h_inc(k)/h_tmp(k)))
            end do

            do k=1,kb
              h_inc(k) = h_inc(k) / scal
              h_tmp(k) = h_tmp(k) + h_inc(k)
            enddo

            iterat = iterat + 1
            if ( iterat > it_max ) call exitme(1,'Undue number of iterations')

          end do iteration
    
        elseif ( depth_n <= zero ) then
          !- In this case the water column would run originally dry - so set
          !- all layersthicknesses to 5 cm and write a WARNING to logfile.
          !- Most probably a coarse grid cell fell dry, which is also covered 
          !- by a finer grid, so that the real cell thickness (at the end of the
          !- timestep comes from fine grid anyway! 
          write(iu06,*) 'WARNING: In Grid ', ia, ' a column ran empty: line: ',&
                         i, ' column: ', j,' Number of layers: ', kb

          do k=1,kb
            h_tmp(k) = hgr
          enddo
        endif

        !- Calculate the pseudo vertical velocity by using the new layer 
        !- thicknesses.
        w_pseudo = zero
        do k=kb,1,-1
          w_pseudo = w_pseudo - (diverg(k) + (h_tmp(k)-h(mmk(k,i,j)))/dt )
          wpsd(k) = w_pseudo
        end do

        dh = zero
        do k=1,kb
          mi = mmk(k,i,j)
          hn(mi) = h_tmp(k)
          ws(mi) = wpsd(k)
          dh = dh + h_tmp(k)
        end do
        ws(mmk(1,i,j)) = zero

        !- Calculate the new sea level zn.      
        zn(mmk(1,i,j))= dh - ti(mmk(1,i,j))
            
      end do

end subroutine masseqs_bnd_dyn

!===============================================================================

subroutine zbnd_rbc ( mmk, khu, khv, un, vn, hu, hv, pl, nz, krz, rwz )
  !-----------------------------------------------------------------------------
  !     radiation boundary condition for open boundaries 
  !     with hydrostatic air pressure correction
  !-----------------------------------------------------------------------------

  use constants, only : g, rhow, pair

  implicit none

  integer(4), intent(in) :: mmk(:,0:,0:)
  integer(4), intent(in) :: khu(0:)
  integer(4), intent(in) :: khv(0:) 
  integer(4), intent(in) :: nz
  integer(4), intent(in) :: krz(:,0:)
  real(8), intent(in)    :: pl(0:)
  real(8), intent(in)    :: un(0:)
  real(8), intent(in)    :: vn(0:)
  real(8), intent(in)    :: hu(0:)
  real(8), intent(in)    :: hv(0:)
  real(8), intent(inout) :: rwz(0:)
  
  integer(4) :: klaus
  integer(4) :: i, j, k, kbu, kbv
  real(8)    :: zet, vdiff, depth


  do klaus=1,nz

    zet = zero
    i = krz(1,klaus)
    j = krz(2,klaus)

    if (mmk(1,i,j) <= 0) cycle

    !- hydrostatic air pressure correction:
    zet   = zet + (pair*100.0_8 - pl(mmk(1,i,j)))/(rhow*g)

    vdiff = zero
    depth = zero

    if (krz(3,klaus) == 1) then
      kbu = khu(mmk(1,i,j))
      do  k = 1, kbu
        vdiff = -un(mmk(k,i,j))
        depth = hu(mmk(k,i,j))
        if (depth > zero) zet = zet+vdiff*sqrt(depth/g)
      enddo
    end if

    if (krz(3,klaus) == 2) then
      kbv = khv(mmk(1,i,j))
      do  k = 1, kbv
        vdiff = vn(mmk(k,i,j))
        depth = hv(mmk(k,i,j))
        if (depth > zero) zet = zet+vdiff*sqrt(depth/g)
      enddo
    end if

    if (krz(3,klaus) == 3) then
      kbu = khu(mmk(1,i,j-1))
      do  k = 1, kbu
        vdiff = un(mmk(k,i,j-1))
        depth = hu(mmk(k,i,j-1))
        if (depth > zero) zet = zet+vdiff*sqrt(depth/g)
      enddo
    end if

    if (krz(3,klaus) == 4) then
      kbv = khv(mmk(1,i-1,j))
      do  k = 1, kbv
        vdiff = -vn(mmk(k,i-1,j))
        depth = hv(mmk(k,i-1,j))
        if (depth > zero) zet = zet+vdiff*sqrt(depth/g)
      enddo
    end if

    rwz(klaus) = zet

    ! corner points needs a special treatment
    if (klaus > 1) then
      if (krz(3,klaus-1) > 0  .and. i == krz(1,klaus-1) .and.                  &
          j == krz(2,klaus-1)) then
        rwz(klaus)   = (rwz(klaus) + rwz(klaus-1))/two
        rwz(klaus-1) = rwz(klaus)
      endif
    endif

  enddo 

end subroutine zbnd_rbc

end module boundary
