module checkborder_wrap 
  use constants, only : zero

  implicit none
  private
  public :: checkborder

contains
  !============================================================================+
  !
  !  set up vertical nesting and
  !  check for consistency along border between fine and coarse grid
  !
  !============================================================================+

  subroutine checkborder( kc, mmc, dzc, iga, jga, kf, mmf, dzf,                &
                          krz, kfg, nzfirst, nzlast, iac, iaf )

    !- modules -----------------------------------------------------------------
    use dmi_mpi_global, only : iu06
    use exits,          only : exitme

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)  :: kc
    integer(4), intent(in)  :: mmc(0:,0:,:)
    real(8), intent(in)     :: dzc(:)
    integer(4), intent(in)  :: iga(:), jga(:)
    integer(4), intent(in)  :: kf
    integer(4), intent(in)  :: mmf(0:,0:,:)
    real(8), intent(in)     :: dzf(:)
    integer(4), intent(in)  :: nzfirst, nzlast, iac, iaf
    integer(4), intent(in)  :: krz(:,0:)
    integer(4), intent(out) :: kfg(:)
  

    !- local vars --------------------------------------------------------------
    integer(4) :: iw, i, j, ic, jc, ig, jg, k, kkc, kkf, kg, kgg
    real(8)    :: df, dc
    logical    :: lerr, cerr

    real(8), parameter :: onepeps = 1.001_8

    lerr = .false.

    !- find No. of fine grid points pr. coarse grid point vertically -----------
    kfg(:) = 0
    kkf = 0
    do kkc=1,kc
      df = zero
      dc = dzc(kkc)
      do while (kkf < kf .and. onepeps*df < dc)
        kkf      = kkf + 1
        kfg(kkc) = kfg(kkc) + 1
        df       = df + dzf(kkf)
      enddo
    enddo


    !- run through all points along this border --------------------------------
    do iw=nzfirst,nzlast
      !- fine grid border coordinates:
      i = krz(1,iw)
      j = krz(2,iw)

      !- coarse grid coordinates along border:
      ic = iga(1) + int((i-1)/iga(2))
      jc = jga(1) + int((j-1)/jga(2))

      !- run through the water column and check for consistency:
      kg  = 1
      kgg = 1
      do k=1,kf
        if (mmf(i,j,k) > 0) then
          !- coarse grid coordinates (possibly) outside border:
          ig = ic
          jg = jc
          if     (krz(3,iw) == 1 .and. mmc(ig,jg-1,kg) > 0) then
            jg = jc-1
          elseif (krz(3,iw) == 2 .and. mmc(ig-1,jg,kg) > 0) then
            ig = ic-1
          elseif (krz(3,iw) == 3 .and. mmc(ig,jg+1,kg) > 0) then
            jg = jc+1
          elseif (krz(3,iw) == 4 .and. mmc(ig+1,jg,kg) > 0) then
            ig = ic+1
          endif

          !- check for consistency:
          if (mmc(ig,jg,kg) <= 0) then
            cerr = .false.
            if (kg == 1) then
              cerr = .true.
            elseif (kg > 1) then
              if (mmc(ig,jg,kg-1) <= 0) cerr = .true.
            endif
            if (cerr) then
              if (.not.lerr) write(iu06,*)'Error: inconsistent number of layers'
              write(iu06,*) '  border point no:   ',iw
              write(iu06,*) '  coarse grid no:    ',iac
              write(iu06,*) '  fine   grid no:    ',iaf
              write(iu06,*) '  fine   grid (ijk): ',i,j,k
              write(iu06,*) '  coarse grid (ijk): ',ig,jg,kg
              write(iu06,*) ' '
              lerr =.true. 
            endif
          endif

          !- find next k in coarse grid kg:
          kgg = kgg + 1
          if (kgg > kfg(kg)) then
            kg  = kg + 1
            kgg = 1
          endif
        endif
      enddo

    enddo

    if (lerr) call exitme('exit from checkborder')

  end subroutine checkborder
end module checkborder_wrap 
