module checknesting_wrap 

  use constants, only : zero, centi, hgr

  implicit none
  private
  public :: checknesting_dyn

contains
  !============================================================================+
  !
  !  check for consistency between fine and coarse grid
  !  (needed if dynamical vertical co-ordinates are used )
  !
  !============================================================================+

  subroutine checknesting_dyn ( mg,ng,khg,mmg,mf,nf,kfmx,mmf,igaf,jgaf,kfg,dxf,&
                                dyf,hzg,hzf)
  
    !- modules -----------------------------------------------------------------
    use dmi_mpi_global, only : iu06
    use exits,          only : exitme

    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: mg, ng
    integer(4), intent(in) :: mf, nf, kfmx
    integer(4), intent(in) :: khg(0:)
    integer(4), intent(in) :: mmg(0:,0:,:),mmf(0:,0:,:)
    integer(4), intent(in) :: igaf(:),jgaf(:), kfg(1:)
    !     coarse grid field:
    real(8), intent(in)    :: hzg(0:)
    !     fine grid field:
    real(8), intent(in)    :: dxf(:)
    real(8), intent(in)    :: hzf(0:)
    real(8), intent(in)    :: dyf
  
    
    !- local vars --------------------------------------------------------------
    !  coarse grid indices:
    integer(4) :: i, j, ia, ja, ie, je, k, mmc, kbg, idxg
    !  fine grid indices:
    integer(4) :: ifa, jfa, ife, jfe, iff, jf, kf, kk, idx
    !  temporary quatities:
    real(8)    :: height, area_f, cellvol, hfine
    real(8)    :: vol(1:kfmx)
    logical    :: lerr1, lerr2

    lerr1 = .false.
    lerr2 = .false.

    !- find extreme indices:
    ia = igaf(1) + (igaf(2)/2+igaf(3))/igaf(2)
    ie = igaf(1) + (igaf(2)/2+igaf(3)+mf-1)/igaf(2)
    ja = jgaf(1) + (jgaf(2)/2+jgaf(3))/jgaf(2)
    je = jgaf(1) + (jgaf(2)/2+jgaf(3)+nf-1)/jgaf(2)
  
    do j=max(1,ja),min(ng,je)
      do i=max(1,ia),min(mg,ie)
  
        idxg = mmg(i,j,1)
        if (idxg <= 0) cycle

        kbg        = khg(mmg(i,j,1))
        height     = zero
        area_f     = zero
        vol(:)     = zero

        ifa = igaf(2)*(i-igaf(1)) + 1
        ife = igaf(2)*(i-igaf(1)+1)
        jfa = jgaf(2)*(j-jgaf(1)) + 1
        jfe = jgaf(2)*(j-jgaf(1)+1)

        do jf=jfa,jfe
          do iff=ifa,ife
            kf = 1
            idx = mmf(iff,jf,kf)
            
            if (idx > 0) then
              hfine   = max(hzf(idx), hgr)
              vol(kf) = vol(kf) + dxf(iff)*dyf*hfine
              area_f  = area_f + dxf(iff)*dyf
            endif

            do kf = 2,kfmx
              idx = mmf(iff,jf,kf)
              if (idx <= 0) exit  ! kf-loop

              hfine   = max(hzf(idx), hgr)
              vol(kf) = vol(kf) + dxf(iff)*dyf*hfine
            enddo
          enddo
        enddo

        if ( area_f > zero ) then
          kk = 1
          do k = 1, kbg-1
            mmc = mmg(i,j,k)
            cellvol = zero
            do kf = kk, min(kk-1+kfg(k),kfmx)
              cellvol   = cellvol + vol(kf)
            enddo
            if (cellvol == zero) then
              write(iu06,*)'Error: inconsistent nesting!!!'
              write(iu06,*)'   Coarser grid deeper than finer grid!'
              write(iu06,*)'   coarse grid (ijk): ',i,j,k
              write(iu06,*)'    Column should have only ', k-1, ' or less cells'
              write(iu06,*)'                 instead of ', kbg
              write(iu06,*) ' '
              lerr1 =.true.
            else
              height   = cellvol/area_f
              if (abs(height-hzg(mmc)) > centi)  then
                write(iu06,*)'Error: inconsistent nesting!!!'
                write(iu06,*)'   Coarser grid cell height do not fit to '
                write(iu06,*)'                       finer cell heights!'
                write(iu06,*)'   coarse grid (ijk): ',i,j,k
                write(iu06,*)'   coarse grid height:', hzg(mmc)
                write(iu06,*)'                it should be: ', height
                write(iu06,*) ' '
                lerr2 =.true.
              endif
            endif
            kk = kk+kfg(k)
          enddo

          !- unroll k = kbg
          k = kbg
          mmc = mmg(i,j,k)
          cellvol = zero
          do kf = kk, kfmx
            cellvol   = cellvol + vol(kf)
          enddo
          if (cellvol == zero) then
            write(iu06,*)'Error: inconsistent nesting!!!'
            write(iu06,*)'   Coarser grid deeper than finer grid!'             
            write(iu06,*)'   coarse grid (ijk): ',i,j,k
            write(iu06,*)'    Column should have only ', k-1, ' or less cells'
            write(iu06,*)'                 instead of ', kbg
            write(iu06,*) ' '
            lerr1 =.true.
          else
            height   = cellvol/area_f
            if (hzg(mmc)-height > centi) then
              write(iu06,*)'Error: inconsistent nesting!!!'
              write(iu06,*)'   Coarser grid bottom cell is too height'
              write(iu06,*)'   coarse grid (ijk): ',i,j,k
              write(iu06,*)'   coarse grid height:', hzg(mmc)
              write(iu06,*)'                it should be: ', height
              write(iu06,*) ' '
              lerr2 =.true.
            endif
          endif
        endif
      enddo
    enddo
    

    if(lerr1) call exitme('Exit from checknesting_dyn: coarse grid is too deep')
    if(lerr2) call exitme('Exit from checknesting_dyn: heights does not fit')
    

  end subroutine checknesting_dyn

end module checknesting_wrap 
