module cmod_advecsubs
! module that wraps some advection related subroutines
  use constants, only : zero, one
  implicit none
  private

  interface traceroutflow
    module procedure traceroutflow3D
    module procedure traceroutflow2D
  end interface

  public :: traceroutflow, resuvwzi

  private :: traceroutflow2D, traceroutflow3D

contains

!===============================================================================

subroutine traceroutflow3D( ncmp,bndno,krz,kbh,mmk,dt,dx,dy,cosphi,uadc,vadc,c )
  !-----------------------------------------------------------------------------
  !
  ! Update the AD outflow boundaries after the 3D AD calculations.
  !
  !-----------------------------------------------------------------------------

  !- directives ----------------------------------------------------------------
  implicit none

  !- Arguments -----------------------------------------------------------------
  integer(4), intent(in) :: ncmp, bndno
  integer(4), intent(in) :: krz(:,0:),kbh(0:),mmk(:,0:,0:)
  real(8), intent(in)    :: dt, dx(0:), dy(0:)
  real(8), intent(in)    :: cosphi(1:,0:)

  real(8), intent(in),    dimension(0:)   :: uadc,vadc
  real(8), intent(inout), dimension(:,0:) :: c


  !- Locals --------------------------------------------------------------------
  integer(4) :: i, j, k, kb
  integer(4) :: nccc, npcc, ncmc, ncpc, nmcc, bno
  real(8)    :: dtdx, dtdy, x, y, fc



  !- Loop over all boundaries --------------------------------------------------
  bndloop: do bno=1,bndno

    i  = krz(1,bno)
    j  = krz(2,bno)
    kb = kbh(mmk(1,i,j))
    if (kb < 1) cycle bndloop
    dtdx = dt/(dx(j)*cosphi(1,i))
    dtdy = dt/dy(i)

    !- West boundary -----------------------------------------------------------
    if (krz(3,bno) == 1) then
      kwloop: do k=1,kb
              nccc = mmk(k,i,j)

              if (uadc(nccc) < zero) then
                ncpc = mmk(k,i,j+1)

                x  = -uadc(nccc)*dtdx
                fc = one - x
                c(1:ncmp,nccc) = fc*c(1:ncmp,nccc) + x*c(1:ncmp,ncpc)
              endif
      enddo kwloop

    !- North boundary ----------------------------------------------------------
    elseif (krz(3,bno) == 2) then
      knloop: do k=1,kb
              nccc = mmk(k,i,j)

              if (vadc(nccc) > zero) then
                npcc = mmk(k,i+1,j)

                y  = vadc(nccc)*dtdy
                fc = one - y
                c(1:ncmp,nccc) = fc*c(1:ncmp,nccc) + y*c(1:ncmp,npcc)
              endif
      enddo knloop

    !- East boundary -----------------------------------------------------------
    elseif (krz(3,bno) == 3) then
      keloop: do k=1,kb
              nccc = mmk(k,i,j)
              ncmc = mmk(k,i,j-1)

              if (uadc(ncmc) > zero) then
                x  = uadc(ncmc)*dtdx
                fc = one - x
                c(1:ncmp,nccc) = fc*c(1:ncmp,nccc) + x*c(1:ncmp,ncmc)
              endif
      enddo keloop

    !- South boundary ----------------------------------------------------------
    elseif (krz(3,bno) == 4) then
      ksloop: do k=1,kb
              nccc = mmk(k,i,j)
              nmcc = mmk(k,i-1,j)

              if (vadc(nmcc) < zero) then
                y  = -vadc(nmcc)*dtdy
                fc = one - y
                c(1:ncmp,nccc) = fc*c(1:ncmp,nccc) + y*c(1:ncmp,nmcc)
              endif
      enddo ksloop

    endif
  enddo bndloop
  !- End of bndno loop ---------------------------------------------------------

        
  !- That's all folks ! --------------------------------------------------------

end subroutine traceroutflow3D

!===============================================================================

subroutine traceroutflow2D( ncmp,bndno,krz,me,dt,dx,dy,cosphi,u,v,c )
  !-----------------------------------------------------------------------------
  !
  ! Update the AD outflow boundaries after the 2D AD calculations.
  !
  !-----------------------------------------------------------------------------

  !- directives ----------------------------------------------------------------
  implicit none

  !- Arguments -----------------------------------------------------------------
  integer(4), intent(in)    :: ncmp, bndno
  integer(4), intent(in)    :: krz(:,0:),me(0:,0:)
  real(8),    intent(in)    :: dt, dx(0:), dy(0:)
  real(8),    intent(in)    :: cosphi(1:,0:)
  real(8),    intent(in)    :: u(0:),v(0:)
  real(8),    intent(inout) :: c(:,0:)


  !- Locals --------------------------------------------------------------------
  integer(4) :: i, j
  integer(4) :: nccc, npcc, ncmc, ncpc, nmcc, bno
  real(8)    :: dtdx, dtdy, x, y, fc



  !- Loop over all boundaries --------------------------------------------------
  bndloop: do bno=1,bndno

    i = krz(1,bno)
    j = krz(2,bno)
    if (me(i,j) <= 0) cycle bndloop
    dtdx = dt/(dx(j)*cosphi(1,i))
    dtdy = dt/dy(i)

    !- West boundary -----------------------------------------------------------
    if (krz(3,bno) == 1) then
      nccc = me(i,j)

      if (u(nccc) < zero) then
        ncpc = me(i,j+1)

        x  = -u(nccc)*dtdx
        fc = one - x
        c(1:ncmp,nccc) = fc*c(1:ncmp,nccc) + x*c(1:ncmp,ncpc)
      endif

    !- North boundary ----------------------------------------------------------
    elseif (krz(3,bno) == 2) then
      nccc = me(i,j)

      if (v(nccc) > zero) then
        npcc = me(i+1,j)

        y  = v(nccc)*dtdy
        fc = one - y
        c(1:ncmp,nccc) = fc*c(1:ncmp,nccc) + y*c(1:ncmp,npcc)
      endif

    !- East boundary -----------------------------------------------------------
    elseif (krz(3,bno) == 3) then
      nccc = me(i,j)
      ncmc = me(i,j-1)

      if (u(ncmc) > zero) then
        x  = u(ncmc)*dtdx
        fc = one - x
        c(1:ncmp,nccc) = fc*c(1:ncmp,nccc) + x*c(1:ncmp,ncmc)
      endif

    !- South boundary ----------------------------------------------------------
    elseif (krz(3,bno) == 4) then
      nccc = me(i,j)
      nmcc = me(i-1,j)

      if (v(nmcc) < zero) then
        y  = -v(nmcc)*dtdy
        fc = one - y
        c(1:ncmp,nccc) = fc*c(1:ncmp,nccc) + y*c(1:ncmp,nmcc)
      endif

    endif
  enddo bndloop
  !- End of bndno loop ---------------------------------------------------------

        
  !- That's all folks ! --------------------------------------------------------

end subroutine traceroutflow2D

!===============================================================================

subroutine resuvwzi ( iwet2,mm,ind,kh,khu,khv,hx,hy,ui,vi,idx )

  !- modules -------------------------------------------------------------------
  use dmi_omp, only : domp_get_domain

  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in)    :: iwet2
  integer(4), intent(in)    :: mm(:,0:,0:), ind(:,:)
  integer(4), intent(in)    :: khu(0:),khv(0:),kh(0:)
  real(8), intent(in)    :: hx(0:),hy(0:)
  real(8), intent(inout) :: ui(0:),vi(0:)
  integer(4), intent(inout) :: idx(:,:)

  !- local vars ----------------------------------------------------------------
  integer(4) :: nn, i, j, kbu, kbv, k, mi, nnl, nnu

  call domp_get_domain(kh, 1, iwet2, nnl, nnu, idx)

  do nn=nnl,nnu
    i = ind(1,nn)
    j = ind(2,nn)

    kbu = khu(nn)
    do k=1,kbu
      mi = mm(k,i,j)
      ui(mi) = ui(mi)/hx(mi)
    enddo

    kbv = khv(nn)
    do k=1,kbv
      mi = mm(k,i,j)
      vi(mi) = vi(mi)/hy(mi)
    enddo
  enddo

end subroutine resuvwzi

!===============================================================================

end module cmod_advecsubs
