module cmod_arrays

  !- Modules -------------------------------------------------------------------
  use cmod_mem, only : cml1, cmi1, cmi2, cmi3, cmr1, cmr2, cmr3

  !- Directives ----------------------------------------------------------------
  implicit none
  private


  !- Basic arrays (indices, bathymetries, etc) ---------------------------------
  !    Some fundamental arrays are declared, allocated and set or initialised 
  !    in module cmod_params
  !
  !  All the arrays declared below are allocated through routines
  !    cmod_arrays::AllocPointers
  !    cmod_arrays::AllocArrays


  !- Index arrays --------------------------------------------------------------
  type(cmi3), allocatable, save, public :: mm1k(:)  ! Will eventually subs. mm1

  !- Index help arrays ---------------------------------------------------------
  type(cmi2), allocatable, save, public :: ind(:)
  type(cmi1), allocatable, save, public :: kh(:), khu(:), khv(:), kh_fake(:)
  integer(4), allocatable, save, public :: nqfile(:)
  type(cmi1), allocatable, save, public :: nqlist(:)
  integer(4), allocatable, save, public :: kdry(:)


  !- dx,dy,dz,dt and combinations hereof ---------------------------------------
  real(8),    allocatable, save, public :: dt(:), dy(:), ty(:), gty(:)
  real(8),    allocatable, save, public :: dxk(:), dyk(:)
  type(cmr1), allocatable, save, public :: xkoor(:), ykoor(:), dz(:)
  type(cmr1), allocatable, save, public :: dx(:)
  type(cmr1), allocatable, save, public :: tx(:), gtx(:)
  type(cmr1), allocatable, save, public :: covo(:), covu(:)
  type(cmr1), allocatable, save, public :: ft(:)
  type(cmr1), allocatable, save, public :: dxx(:)
  type(cmr1), allocatable, save, public :: dyy(:)
  type(cmr2), allocatable, save, public :: cosphi(:), fcorio(:)


  !- Indices for dams ----------------------------------------------------------
  type(cmi1), allocatable, save, public :: lvd(:), lud(:)

  !- Indices, heights and flow rates for weirs ---------------------------------
  type(cmi2), allocatable, save, public :: weir_idx(:)
  type(cmr1), allocatable, save, public :: zweir(:)
  type(cmr1), allocatable, save, public :: wflowr(:)

  !- Depths --------------------------------------------------------------------
  type(cmr1), allocatable, save, public :: ti(:)

  !- Sea level and current -----------------------------------------------------
  type(cmr1), allocatable, save, public :: h(:)
  type(cmr1), allocatable, save, public :: hn(:)
  type(cmr1), allocatable, save, public :: z(:)
  type(cmr1), allocatable, save, public :: u(:), v(:)


  !- Wind speed, -stress and air presssure -------------------------------------
  type(cmr1), allocatable, save, public :: wua(:), wun(:), wva(:), wvn(:)
  type(cmr1), allocatable, save, public :: wu(:), wv(:)
  type(cmr1), allocatable, save, public :: pla(:), pln(:), pl(:)


  !- Air temperature, -humidity and cloud cover --------------------------------
  type(cmr1), allocatable, save, public :: atempa(:), atempn(:), atemp(:)
  type(cmr1), allocatable, save, public :: humida(:), humidn(:), humid(:)
  type(cmr1), allocatable, save, public :: clouda(:), cloudn(:), cloud(:)


  !- Work arrays used here and there where gathering to global is needed -------
  type(cmr1), allocatable, save, public :: wrk(:), wrk2(:)
  real(8),    allocatable,       public :: obcwrk(:,:,:)
  ! FIXME: this is just a preliminary hack, must be coded properly when we find
  !        time and energy to do so. See code where these are used.
  real(8),    allocatable,       public :: wma1(:), wma2(:), wma3(:), wma4(:)


  !- Geopot for dyn.vert.coords.  ----------------------------------------------
  type(cmr1), allocatable, save, public :: geopot(:)


  ! - vertical divergence of shortwave radiation -------------------------------
  type(cmr1), allocatable, save, public :: divrad(:), divrad_s(:) 


  !- Sea bed temperature -------------------------------------------------------
  type(cmr2), allocatable, save, public :: t_soil(:)


  !- Precipitation and evaporation ---------------------------------------------
  type(cmr1), allocatable, save, public :: precipo(:), precipn(:), precip(:)
  type(cmr1), allocatable, save, public :: evapo(:), tpe(:)


  !- Parameters for turbulence model -------------------------------------------
  type(cmr1), allocatable, save, public :: avv(:), tke(:), diss(:)
  real(8),    allocatable, save, public :: ahc(:), avc(:), rt(:)


  !- Arrays for A/D scheme with internal component loop ------------------------
  type(cmr3), allocatable, save, public :: bndz(:), bndq(:)
  type(cmr2), allocatable, save, public :: dispv(:), cmp(:)


  !- Fields for the ice model --------------------------------------------------
  type(cml1), allocatable, save, public :: casus(:)
  type(cmr1), allocatable, save, public :: tsnei(:)
  type(cmr1), allocatable, save, public :: ueis(:), veis(:)
  type(cmr2), allocatable, save, public :: ice(:)
  type(cmr2), allocatable, save, public :: iota(:)

  !- Boundary point coordinates ------------------------------------------------
  type(cmi2), allocatable, save, public :: kru(:), krv(:)
  type(cmi2), allocatable, save, public :: krq(:)


  !- Arrays for boundary values ------------------------------------------------
  type(cmr1), allocatable, save, public :: rwz(:)
  type(cmr2), allocatable, save, public :: rwqk(:)
  type(cmr1), allocatable, save, public :: rwq(:)


  !- Boundary data arrays for ice model ----------------------------------------
  type(cmr2), allocatable, save, public :: bndice(:)
  real(8),    allocatable, save, public :: iceobc(:,:)


  !- Work arrays for boundary values -------------------------------------------
  integer(4), allocatable, save, public :: khbnd(:)
  real(8),    allocatable, save, public :: zetac(:), zetac2(:)
  real(8),    allocatable, save, public :: bndzk(:,:,:)


  !- Kalman Filter arrays ------------------------------------------------------
  type(cmr1), allocatable, save, public :: a_asstemp(:), a_obserr(:)
  type(cmr1), allocatable, save, public :: a_errvar(:), a_lastasstime(:)

  !- Arrays used in OMP --------------------------------------------------------
  type(cmi2), allocatable, save, public :: idx(:)
  integer(4), allocatable, save, public :: ndtp(:)


  !- Subroutines/functions -----------------------------------------------------
  public  :: AllocArrays
  private :: AllocPointers


contains

  !============================================================================+
  !
  !  Allocate arrays 
  !
  !============================================================================+
  subroutine AllocArrays (ia)

    !- modules -----------------------------------------------------------------
    use cmod_params,   only : nthreads, ntotal, mainarea, mmx, nmx, kmx,       &
                              mmxp, nmxp, iw2, iw3, nud, nvd, nweir,           &
                              nq, nz, nu, nv, nsoilpar
    use cmod_miscvars, only : ntracers
    use constants,     only : usezbcl, kalman, precevap, pen_type, dyn,        &
                              nicepar, printturb, zero, iotamodel, niotapar
    use exits,         only : exitme
 
    !- dircetives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: ia

    !- local vars --------------------------------------------------------------
    logical, save :: FirstEntry = .true.
    integer(4)    :: mnp, nnp, mn, nn, kn, n2d, n3d,                           &
                     nudn, nvdn, nweirn, nqn, nzn, nun, nvn, ierr


    !- Allocate all the pointers on first entry only ---------------------------
    ierr = 0
    if (FirstEntry) then
      FirstEntry = .false.
      call AllocPointers( ntotal, ierr )
      if (ierr /= 0) call exitme(-1,'AllocArrays failed, AllocPointers')
      allocate( ndtp(1:nthreads), stat=ierr )
      if (ierr /= 0) call exitme(-1,'AllocArrays failed, ndtp')
      n3d = maxval(iw3(:))
      allocate( wma3(0:n3d), wma4(0:n3d), stat=ierr )
      if (ierr /= 0) call exitme(-1,'AllocArrays failed, wma3 and wma4')
    endif

    !- Make a local copy of dim nos for simplicity -----------------------------
    mn     = mmx(ia)
    nn     = nmx(ia)
    kn     = kmx(ia)
    mnp    = mmxp(ia)
    nnp    = nmxp(ia)
    n2d    = iw2(ia)
    n3d    = iw3(ia)
    nudn   = nud(ia)
    nvdn   = nvd(ia)
    nweirn = nweir(ia)
    nqn    = nq(ia)
    nzn    = nz(ia)
    nun    = nu(ia)
    nvn    = nv(ia)

    !- Allocate all the arrays -------------------------------------------------
    allocate( idx(ia)%p(1:2,1:nthreads),                                       &
      mm1k(ia)%p(kn,0:mnp,0:nnp), ind(ia)%p(1:2,n2d),                          &
      kh(ia)%p(0:n2d), khu(ia)%p(0:n2d), khv(ia)%p(0:n2d), dz(ia)%p(kn),       &
      xkoor(ia)%p(nn), dx(ia)%p(mn), ft(ia)%p(mn), kh_fake(ia)%p(0:n2d),       &
      ykoor(ia)%p(mn), tx(ia)%p(mn), gtx(ia)%p(mn),                            &
      covo(ia)%p(mn), covu(ia)%p(mn), cosphi(ia)%p(2,0:mnp),                   &
      dxx(ia)%p(0:nnp), dyy(ia)%p(0:mnp), fcorio(ia)%p(2,0:mnp),               &
      lvd(ia)%p(nvdn), lud(ia)%p(nudn),                                        &
      weir_idx(ia)%p(3,0:nweirn),zweir(ia)%p(0:nweirn),wflowr(ia)%p(0:nweirn), &
      ti(ia)%p(0:n2d), h(ia)%p(0:n3d),                                         &
      z(ia)%p(0:n2d), u(ia)%p(0:n3d), v(ia)%p(0:n3d),                          &
      wua(ia)%p(0:n2d), wun(ia)%p(0:n2d), wva(ia)%p(0:n2d), wvn(ia)%p(0:n2d),  &
      wu(ia)%p(0:n2d), wv(ia)%p(0:n2d), pla(ia)%p(0:n2d), pln(ia)%p(0:n2d),    &
      pl(ia)%p(0:n2d),atemp(ia)%p(0:n2d),cloud(ia)%p(0:n2d),humid(ia)%p(0:n2d),&
      atempa(ia)%p(0:n2d), atempn(ia)%p(0:n2d), clouda(ia)%p(0:n2d),           &
      humida(ia)%p(0:n2d), humidn(ia)%p(0:n2d), cloudn(ia)%p(0:n2d),           &
      cmp(ia)%p(1:ntracers,0:n3d), wrk(ia)%p(0:n3d), wrk2(ia)%p(0:n3d),        &
      t_soil(ia)%p(1:nsoilpar,n2d), avv(ia)%p(0:n3d), casus(ia)%p(0:n2d),      &
      tsnei(ia)%p(0:n2d),                                                      &
      ueis(ia)%p(0:n2d), veis(ia)%p(0:n2d), ice(ia)%p(1:nicepar,0:n2d),        &
      kru(ia)%p(4,0:nun), krv(ia)%p(4,0:nvn), krq(ia)%p(2,0:nqn),              &
      rwz(ia)%p(0:nzn), nqlist(ia)%p(0:nqn),                                   &
      rwqk(ia)%p(0:nqn,1:kn), bndz(ia)%p(1:ntracers,1:kn,0:nzn),               &
      rwq(ia)%p(0:nqn), bndq(ia)%p(1:ntracers,1:kn,0:nqn),                     &
      bndice(ia)%p(1:nicepar,0:nzn), stat=ierr )
      if (ierr /= 0) call exitme(-1,'AllocArrays failed, global arrays')

      if (iotamodel) then  
        allocate(iota(ia)%p(1:niotapar,0:n2d), stat=ierr )
        if (ierr /= 0) call exitme(-1,'AllocArrays iota failed, global arrays')
      else
        allocate(iota(ia)%p(1:niotapar,0:0), stat=ierr )
        if (ierr /= 0) call exitme(-1,'AllocArrays iota failed, global arrays')
      end if

    !- make sure to have sane values at land in the work arrays ----------------
    wrk (ia)%p(0) = zero
    wrk2(ia)%p(0) = zero

    !- define two-equation turbulence model variables --------------------------
    allocate(tke(ia)%p(0:n3d), diss(ia)%p(0:n3d), stat=ierr)
    if (ierr /= 0) call exitme(-1,'AllocArrays failed, turbulence arrays')

    !- define tracer diffusivity -----------------------------------------------
    if (printturb) then
      if (ntracers == 2) then
        allocate(dispv(ia)%p(1:2,0:n3d),stat=ierr)
      else
        allocate(dispv(ia)%p(1:3,0:n3d),stat=ierr)
      endif
      if (ierr /= 0) call exitme(-1,'AllocArrays failed, diffusivity array')
    endif
 
    !- define vertical divergence of sw-penetration (only needed if pen_type==2)
    if (pen_type == 2) then
      allocate( divrad(ia)%p(0:n3d), divrad_s(ia)%p(0:n2d), stat=ierr )
    elseif (pen_type <= 1) then
      allocate( divrad(ia)%p(0:0), divrad_s(ia)%p(0:0), stat=ierr )
    endif
    if (ierr /= 0) call exitme(-1,'AllocArrays failed, sw-penetration arrays')
    
    !- define the additional variables needed for dynamical vertical co-ord. ---
    if (dyn) then
      allocate( geopot(ia)%p(0:n3d), hn(ia)%p(0:n3d), stat=ierr )
    else
      allocate( geopot(ia)%p(0:0), hn(ia)%p(0:0), stat=ierr )
    endif
    if (ierr /= 0) call exitme(-1,'AllocArrays failed, geopot arrays')

 
    !- define work arrays for main area ----------------------------------------
    if (ia == mainarea) then
      if (dyn) then
        allocate(wma1(0:n3d), wma2(0:n3d), stat=ierr)
        if (ierr /= 0) call exitme(-1,'AllocArrays failed, wma1/2 arrays')
        wma1(0) = zero
        wma2(0) = zero
      else 
        allocate(wma1(0:0), wma2(0:0), stat=ierr)
        if (ierr /= 0) call exitme(-1,'AllocArrays failed, wma1/2 arrays')
        wma1(0) = zero
        wma2(0) = zero
      endif
      allocate(obcwrk(1:kn,0:nzn,1:4), stat=ierr)
      if (ierr /= 0) call exitme(-1,'AllocArrays failed, obcwrk array')
      obcwrk(:,0,:) = zero
    endif

    !- define sponges/obc for main area only -----------------------------------
    if (ia == mainarea) then
      allocate(bndzk(1:ntracers,1:kn,0:nzn), iceobc(1:nicepar,0:nzn), stat=ierr)
      if (ierr /= 0) call exitme(-1,'AllocArrays failed, sponge/bnd arrays')
    endif

    !- define bnd zeta correction variables and khbnd, main area only ----------
    if (ia == mainarea) then
      allocate( zetac(0:nzn), khbnd(0:nzn), stat=ierr )
      if (ierr /= 0) call exitme(-1,'AllocArrays failed, zetac arrays')
      if (usezbcl) then
        allocate( zetac2(0:nzn), stat=ierr )
      else
        allocate( zetac2(0:0), stat=ierr ) ! ensure that bounds are well-def. 
      endif
      if (ierr /= 0) call exitme(-1,'AllocArrays failed, zetac2 array')
    endif

    ! Precipitation and evaporation --------------------------------------------
    if (precevap) then
      allocate( precipo(ia)%p(0:n2d), precipn(ia)%p(0:n2d),precip(ia)%p(0:n2d),&
                evapo(ia)%p(0:n2d), tpe(ia)%p(0:n2d), stat=ierr )
    else
      allocate( precipo(ia)%p(0:0), precipn(ia)%p(0:0),precip(ia)%p(0:0),      &
                evapo(ia)%p(0:0), tpe(ia)%p(0:0), stat=ierr )
    endif
    if (ierr /= 0) call exitme(-1,'AllocArrays failed, precip arrays')

    ! Kalman Filter ------------------------------------------------------------
    if (kalman) then
      allocate( a_asstemp(ia)%p(1:n2d), a_obserr(ia)%p(1:n2d),                 &
                a_lastasstime(ia)%p(0:n2d), a_errvar(ia)%p(0:n3d), stat=ierr )
    else
      allocate( a_asstemp(ia)%p(1:1), a_obserr(ia)%p(1:1),                     &
                a_lastasstime(ia)%p(0:0), a_errvar(ia)%p(0:0), stat=ierr )
    endif
    if (ierr /= 0) call exitme(-1,'AllocArrays failed, kalman arrays')

    !- No kind of sub-model selected -------------------------------------------

  end subroutine AllocArrays

  !============================================================================+
  !
  !  Allocate basic pointers and some arrays
  !
  !============================================================================+
  subroutine AllocPointers (n, ierr)

    use constants, only : printturb

    implicit none

    integer(4), intent(in)    :: n
    integer(4), intent(inout) :: ierr
   
    allocate( mm1k(n), ind(n), kh(n), khu(n), khv(n), dz(n), kh_fake(n),       &
      dt(n), dy(n), ty(n), gty(n), tx(n), gtx(n),                              &
      ft(n), xkoor(n), ykoor(n), dx(n), covo(n), covu(n),                      &
      dxk(n), dxx(n), dyk(n), dyy(n), cosphi(n), fcorio(n), ti(n),             &
      h(n), z(n), u(n), v(n), lvd(n), lud(n), weir_idx(n),                     &
      zweir(n), wflowr(n), wua(n), wun(n), wva(n), wvn(n),                     &
      pla(n), pln(n), pl(n), wu(n), wv(n), atemp(n), humid(n), cloud(n),       &
      divrad(n), divrad_s(n), atempa(n), atempn(n), humida(n), humidn(n),      &
      clouda(n), cloudn(n), wrk(n), wrk2(n), t_soil(n), avv(n),                &
      tke(n), diss(n), ahc(n), avc(n), rt(n), bndz(n), bndq(n), cmp(n),        &
      casus(n), tsnei(n), ueis(n), veis(n), ice(n),                            &
      kru(n), krv(n), krq(n), nqfile(n), nqlist(n),                            &
      rwz(n), rwqk(n), rwq(n), bndice(n), idx(n),                              &
      a_asstemp(n), a_obserr(n), a_errvar(n), a_lastasstime(n),                &
      precipo(n), precipn(n), precip(n), evapo(n), tpe(n),                     &
      geopot(n), hn(n), kdry(n), iota(n), stat=ierr )

      if (ierr == 0) then
        if (printturb) allocate(dispv(n), stat=ierr)
      endif

  end subroutine AllocPointers

end module cmod_arrays
