module cmod_auto_decompo

  !- modules: ------------------------------------------------------------------
  !  none

  !- implicit directives: ------------------------------------------------------
  implicit none
  private

  !- public stuff --------------------------------------------------------------
  public :: autodecompose, decompose_optimal_1xN, decompose_write_1xN,         &
            get_forbidden_i
  
  !- private stuff -------------------------------------------------------------
  private :: decompose_optimal_Mx1, decompose_write_MxN, read_data
  
contains

!===============================================================================

  subroutine autodecompose(ntaskmax, narea, mmx, nmx, kmx, iw3, mm1, iu06,     &
                           nforbidden, iforbiddenp, iforbiddenm)

    use constants, only : only_islices
    use cmod_mem,  only : cmi3

    implicit none

    integer(4),  intent(in) :: ntaskmax, narea, iu06
    integer(4),  intent(in) :: mmx(:), nmx(:), kmx(:), iw3(:)
    type (cmi3), intent(in) :: mm1(:)
    integer(4),  intent(in) :: nforbidden(:), iforbiddenp(:,:), iforbiddenm(:,:)


    integer(4)              :: nproci, nprocj, nact, maxit, maxtt, ia, it
    integer(4)              :: ntasksj
    integer(4), allocatable :: jtask(:,:), jlows(:,:), jhghs(:,:)
    integer(4), allocatable :: ilows(:,:), ihghs(:,:)

    !- set No. of tasks in J-direction:
    if (only_islices) then
      ntasksj = 1
    else
      ntasksj = ntaskmax
    endif

    !- do some allocations
    allocate(jtask(ntasksj,narea),jlows(ntasksj,narea),jhghs(ntasksj,narea),   &
             ilows(ntaskmax,narea),ihghs(ntaskmax,narea))

    !- run through number of i-slices
    do nproci=1,ntaskmax

      !- autodecompose with attempt to keep the tasks well-balanced ------------
      do nprocj=1,ntasksj
  
        !- find j-slices:
        nact = 0
        jtask(1:nprocj,:) = 0
        jlows(1:nprocj,:) = 0
        jhghs(1:nprocj,:) = 0
        do ia=1,narea
          ! make the sliced, optimal decompo:
          call decompose_optimal_1xN(ia, nprocj, mmx(ia), nmx(ia), kmx(ia),    &
                                     iw3(ia), mm1(ia)%p, nact,                 &
                                     jtask, jlows, jhghs)

          ! find sliced task with max No. of wet points:
          maxit = 0
          maxtt = 0
          do it=1,nprocj
            if (maxtt < jtask(it,ia)) then
              maxtt = jtask(it,ia)
              maxit = it
            endif
          enddo

          ! sub-divide the largest slice:
          ilows(1:nproci,ia) = 0
          ihghs(1:nproci,ia) = 0
          call decompose_optimal_Mx1(nproci, mmx(ia), kmx(ia), mm1(ia)%p,      &
                                     jlows(maxit,ia), jhghs(maxit,ia), maxtt,  &
                                     ilows(:,ia), ihghs(:,ia), nforbidden(ia), &
                                     iforbiddenp(:,ia), iforbiddenm(:,ia))
        enddo        

        ! sub-divide each j-slice according to the largest one, and 
        ! write this nproci by nprocj decomposition: 
        call decompose_write_MxN(nproci, nprocj, narea, mm1, kmx,              &
                                 ilows, ihghs, jlows, jhghs, ntaskmax, iu06)

      enddo

    enddo
    deallocate( jtask, jlows, jhghs, ilows, ihghs )

  end subroutine autodecompose

!===============================================================================

  subroutine decompose_optimal_1xN(ia, nj, mmx, nmx, kmx, iw3, mm1,            &
                                   nact, jtask, jlows, jhghs, jnsrf)
    implicit none

    integer(4), intent(in)    :: ia, nj, mmx, nmx, kmx, iw3, mm1(0:,0:,1:)
    integer(4), intent(inout) :: nact
    integer(4), intent(inout) :: jtask(:,:), jlows(:,:), jhghs(:,:)
    integer(4), intent(inout), optional :: jnsrf(:,:)

    integer(4) :: iweven, iw, i, j, k, jl, jsum, jsl, js, tsum, jsum2
    integer(4) :: jdist(nmx)

    !- even split:
    iweven = iw3/nj

    !- distribution of wet points in j-direction:
    do j=1,nmx
      iw = 0
      do k=1,kmx
        do i=1,mmx
          if (mm1(i,j,k) > 0) iw = iw + 1
        enddo
      enddo
      jdist(j) = iw
    enddo

    !- find j-slices:
    tsum  = 0
    jsum  = 0
    jsum2 = 0
    jl    = 1
    jsl   = 1
    do j=1,nmx
      jsum = jsum + jdist(j)
      if (j < nmx) jsum2 = jsum + jdist(j+1)
      ! either we have enough wetpoints, or we run out of 
      ! j-lines or of tasks
      if (jsum >= iweven .or. j == nmx .or. jsl == nj .or.                     &
          (j < nmx .and. jsum2 > iweven .and. jsum2-iweven > iweven-jsum) ) then
        jtask(jsl,ia) = jsum
        jlows(jsl,ia) = jl
        jhghs(jsl,ia) = j
        if (jsl == nj) then
          jtask(jsl,ia) = iw3 - tsum
          jhghs(jsl,ia) = nmx
          exit
        endif
        jl    = j + 1
        jsl   = jsl + 1
        tsum  = tsum + jsum
        jsum  = 0
        jsum2 = 0
      endif
    enddo
    nact = max(jsl, nact)

    if (present(jnsrf)) then
      !- find No of srf wetpoints in each slice:
      do js=1,jsl
        if (jtask(js,ia) <= 0) cycle
        iw = 0
        do j=jlows(js,ia),jhghs(js,ia)
          do i=1,mmx
            if (mm1(i,j,1) > 0) iw = iw + 1
          enddo
        enddo
        jnsrf(js,ia) = iw
      enddo
    endif

  end subroutine decompose_optimal_1xN

!===============================================================================

  subroutine decompose_write_1xN(narea, mmx, iw2, iw3, nact,                   &
                                 jtask, jlows, jhghs, jnsrf)

    use io_subs,        only : io_new_unit
    use exits,          only : exitme
    use dmi_mpi_global, only : mpi_decomp_file

    implicit none 

    integer(4), intent(in) :: narea, nact, mmx(:), iw2(:), iw3(:)
    integer(4), intent(in) :: jtask(:,:), jlows(:,:), jhghs(:,:), jnsrf(:,:)

    integer(4) :: lun, ios, il, iu, it, ia, n_w, n_e, tmin, tave, tmax, tnul
    integer(4) :: n, neven, jl, ju, tminp
    integer(4) :: tmp(nact,1:2)
    integer(4), parameter :: n_fake = -1

    !- write decompo:
    lun = io_new_unit()
    open (unit=lun, file=trim(mpi_decomp_file), status='replace', iostat=ios)
    if (ios /= 0) then
      call exitme(1,'Cannot create '//trim(mpi_decomp_file))
    endif
    write(lun, '(I5)') nact
    il = 1
    do it=1,nact
      do ia=1,narea
        iu = mmx(ia)

        if (jtask(it,ia) > 0) then
          ! define index range:
          jl = jlows(it,ia)
          ju = jhghs(it,ia)
          write(lun, '(6I5)') it, ia, il, iu, jl, ju

          ! define neighbours:
          n_w = -1
          if (it > 1) then
            if (jtask(it-1,ia) > 0) n_w = it-1
          endif
          n_e = -1
          if (it < nact) then
            if (jtask(it+1,ia) > 0) n_e = it+1
          endif
          write(lun, '(10x,4I5)') n_w, n_fake, n_e, n_fake

        else
          write(lun, '(6I5)') it, ia, 0, 0, 0, 0
          write(lun, '(10x,4I5)') n_fake, n_fake, n_fake, n_fake
        endif

        ! define corner-neighbours:
        write(lun, '(10x,4I5)') n_fake, n_fake, n_fake, n_fake
      enddo
    enddo
    close(lun)

    ! print stats for this decompo:
    lun = io_new_unit()
    open (unit=lun, file='stats_'//trim(mpi_decomp_file), status='replace',    &
          iostat=ios)
    if (ios /= 0) then
      call exitme(1,'Cannot create stats_'//trim(mpi_decomp_file))
    endif
    write(lun,*) 'Statistics for 1 by ',nact,' auto-decomposition'
    do ia=1,narea
      neven = iw2(ia)/nact
      write(lun,*) ia,'    surface wetpoints: ',iw2(ia)
      write(lun,*) ia,'           even split: ',neven
      tmin = iw3(ia) + 1
      tave = 0
      tmax = -tmin
      tnul = 0
      tmp(1:,1:) = -1
      il = 0
      iu = 0

      do it=1,nact
        n = jnsrf(it,ia)
        if (n == 0) tnul = tnul + 1
        tmin = min(tmin, n)
        tmax = max(tmax, n)
        tave = tave + n
        if (n < 0.5*neven) then
          il = il + 1
          tmp(il,1) = it
        endif
        if (n > 1.5*neven) then
          iu = iu + 1
          tmp(iu,2) = it
        endif
      enddo

      write(lun,*) ia,'        min, ave, max: ',tmin,tave/nact,tmax
      if (il > 0) write(lun,*) ia,'  < 50% of even split: ',(tmp(jl,1),jl=1,il)
      if (iu > 0) write(lun,*) ia,'    > 1.5x even split: ',(tmp(ju,2),ju=1,iu)

      write(lun,*) ia,'          empty tasks: ',tnul

      neven = iw3(ia)/nact
      write(lun,*) ia,'    3D, No. wetpoints: ',iw3(ia)
      write(lun,*) ia,'           even split: ',neven
      tmin = iw3(ia) + 1
      tave = 0
      tmax = -tmin
      tmp(1:max(il,iu),1:2) = -1
      il = 0
      iu = 0
      tminp = tmin

      do it=1,nact
        n = jtask(it,ia)
        tmin = min(tmin, n)
        tmax = max(tmax, n)
        tave = tave + n
        if (n < 0.75*neven) then
          il = il + 1
          tmp(il,1) = it
        endif
        if (n > 1.25*neven) then
          iu = iu + 1
          tmp(iu,2) = it
        endif
        if (n > 0) tminp = min(tminp, n)
      enddo

      write(lun,*) ia,'        min, ave, max: ',tmin,tave/nact,tmax
      write(lun,*) ia,'        skewness,   C: ',real(tmax,4)/real(tminp,4)
      if (il > 0) write(lun,*) ia,'  < 75% of even split: ',(tmp(jl,1),jl=1,il)
      if (iu > 0) write(lun,*) ia,'   > 1.25x even split: ',(tmp(ju,2),ju=1,iu)
    enddo
    close(lun)
 
  end subroutine decompose_write_1xN

!===============================================================================

  subroutine decompose_write_MxN(nproci, nprocj, narea, mm1, kmx,              &
                                 ilows, ihghs, jlows, jhghs, ntaskmax, iu06)

    use cmod_mem, only : cmi3
    use io_subs,  only : io_new_unit
    use exits,    only : exitme

    implicit none 

    integer(4), intent(in) :: nproci, nprocj, narea, kmx(:), ntaskmax
    integer(4), intent(in) :: iu06
    type(cmi3), intent(in) :: mm1(:)
    integer(4), intent(in) :: ilows(:,:), ihghs(:,:)
    integer(4), intent(in) :: jlows(:,:), jhghs(:,:)

    logical,    save :: FirstEntry = .true.
    integer(4), save :: luns
    integer(4)       :: lund, ios, i, j, k, ip, jp, ia, ntmx, nt
    integer(4)       :: il, iu, jl, ju, iw
    integer(4)       :: n_w, n_n, n_e, n_s, n_nw, n_ne, n_se, n_sw
    integer(4)       :: iwp(nproci,nprocj,narea), itt(nproci*nprocj,1:2,narea)
    integer(4)       :: ntt(nproci,nprocj,narea)
    character(1)     :: c1
    character(2)     :: c2
    character(3)     :: c3
    character(4)     :: c4
    character(256)   :: cfn, cfn2
    real(8)          :: chgh, clow, cha(narea), cla(narea), riw

    !- on first entry, open stats file:
    if (FirstEntry) then
      luns = io_new_unit()
      if     (ntaskmax < 10   ) then
        write(c1,'(i1)') ntaskmax
        cfn  = 'decompo_stats_'//c1//'.txt'
      elseif (ntaskmax < 100  ) then
        write(c2,'(i2)') ntaskmax
        cfn  = 'decompo_stats_'//c2//'.txt'
      elseif (ntaskmax < 1000 ) then
        write(c3,'(i3)') ntaskmax
        cfn  = 'decompo_stats_'//c3//'.txt'
      elseif (ntaskmax < 10000) then
        write(c4,'(i4)') ntaskmax
        cfn  = 'decompo_stats_'//c4//'.txt'
      else
        call exitme(1,'Cannot handle more than 9999 by 9999 tasks, sorry')
      endif
      open (unit=luns, file=trim(cfn), status='replace', iostat=ios)
      if (ios /= 0) call exitme(1,'Cannot create decompo_stats file')
      FirstEntry = .false.
    endif

    !- on every entry, open decompo file:
    lund = io_new_unit()
    if     (nproci < 10   ) then
      write(c1,'(i1)') nproci
      cfn  = 'mpi_decompo_'//c1//'x'
    elseif (nproci < 100  ) then
      write(c2,'(i2)') nproci
      cfn  = 'mpi_decompo_'//c2//'x'
    elseif (nproci < 1000 ) then
      write(c3,'(i3)') nproci
      cfn  = 'mpi_decompo_'//c3//'x'
    elseif (nproci < 10000) then
      write(c4,'(i4)') nproci
      cfn  = 'mpi_decompo_'//c4//'x'
    endif
    if     (nprocj < 10   ) then
      write(c1,'(i1)') nprocj
      cfn2 = trim(cfn)//c1//'.txt'
    elseif (nprocj < 100  ) then
      write(c2,'(i2)') nprocj
      cfn2 = trim(cfn)//c2//'.txt'
    elseif (nprocj < 1000 ) then
      write(c3,'(i3)') nprocj
      cfn2 = trim(cfn)//c3//'.txt'
    elseif (nprocj < 10000) then
      write(c4,'(i4)') nprocj
      cfn2 = trim(cfn)//c4//'.txt'
    endif
    open (unit=lund, file=trim(cfn2), status='replace', iostat=ios)
    if (ios /= 0) call exitme(1,'Cannot create decompo file')

    !---------------------------------------------------------------------------
    !   run through this nproci by nprocj decompo, find pruned decompo,
    !   obtain statistics, and dump the decompo.
    itt(:,:,:) = -1
    ntt(:,:,:) = -1
    ntmx = 0
    chgh = 0.0_8
    clow = 0.0_8
    do ia=1,narea
      clow = max( clow, real(maxval(mm1(ia)%p(0:,0:,1:)),8) )
    enddo
    do ia=1,narea
      nt = 0
      cha(ia) = 0.0_8
      cla(ia) = real(maxval(mm1(ia)%p(0:,0:,1:)),8)
      do jp=1,nprocj
        do ip=1,nproci
          iw = 0
          do j=jlows(jp,ia),jhghs(jp,ia)
            do i=ilows(ip,ia),ihghs(ip,ia)
              do k=1,kmx(ia)
                if (mm1(ia)%p(i,j,k) <= 0) exit
                iw = iw + 1
              enddo
            enddo
          enddo
          if (iw > 0) then
            nt = nt + 1
            iwp(ip,jp,ia) = iw
            itt(nt,1,ia)  = ip
            itt(nt,2,ia)  = jp
            ntt(ip,jp,ia) = nt
            riw     = real(iw,8)
            chgh    = max( chgh, riw )
            clow    = min( clow, riw )
            cha(ia) = max( cha(ia), riw )
            cla(ia) = min( cla(ia), riw )
          else
            iwp(ip,jp,ia) = 0
          endif
        enddo
      enddo
      ntmx = max( ntmx, nt )
    enddo
    if (ntmx > 0) then
      write(luns,'(a50)') '=================================================='
      write(luns,'(a18,2i5)')   'Auto-decompo:     ',nproci, nprocj
      write(luns,'(a18,i5)')    'non-empty tasks:  ',ntmx
!fixme: shoud have a nicer format:
      write(luns,*)             'global Cmax/Cmin: ',chgh/clow
      do ia=1,narea
        write(luns,*)           'area, Cmax/Cmin:  ',ia, cha(ia)/cla(ia)
        write(luns,'(a18,2i5)') 'area, empty tasks:',ia,ntmx-maxval(ntt(:,:,ia))
      enddo

      write(lund,'(I5)') ntmx
    else
      write(iu06,*) 'Cannot decompose to zero tasks'
      write(iu06,*) ' ... continue with next decompo'
      close(lund)
      return
    endif

    do nt=1,ntmx
        iw = 0
        do ia=1,narea
          if (itt(nt,1,ia) <= 0) cycle
          ip = itt(nt,1,ia)
          jp = itt(nt,2,ia)
          iw = iw + iwp(ip,jp,ia)
        enddo
        if (iw <= 0) cycle
        do ia=1,narea
          if (itt(nt,1,ia) > 0) then
            ip = itt(nt,1,ia)
            jp = itt(nt,2,ia)

            ! define task:
            il = ilows(ip,ia)
            iu = ihghs(ip,ia)
            jl = jlows(jp,ia)
            ju = jhghs(jp,ia)

            ! define neighbour tasks:
            write(lund,'(6i5)') nt, ia, il, iu, jl, ju
            if (jp > 1) then
              n_w = ntt(ip,jp-1,ia)
            else
              n_w = -1
            endif
            if (ip > 1) then
              n_n = ntt(ip-1,jp,ia)
            else
              n_n = -1
            endif
            if (jp < nprocj) then
              n_e = ntt(ip,jp+1,ia)
            else
              n_e = -1
            endif
            if (ip < nproci) then
              n_s = ntt(ip+1,jp,ia)
            else
              n_s = -1
            endif
            write(lund,'(10x,4i5)') n_w, n_n, n_e, n_s

            ! define neighbour-corner tasks:
            if (ip > 1 .and. jp > 1) then
              n_nw = ntt(ip-1,jp-1,ia)
            else
              n_nw = -1
            endif
            if (ip > 1 .and. jp < nprocj) then
              n_ne = ntt(ip-1,jp+1,ia)
            else
              n_ne = -1
            endif
            if (ip < nproci .and. jp < nprocj) then
              n_se = ntt(ip+1,jp+1,ia)
            else
              n_se = -1
            endif
            if (ip < nproci .and. jp > 1) then
              n_sw = ntt(ip+1,jp-1,ia)
            else
              n_sw = -1
            endif
            write(lund,'(10x,4i5)') n_nw, n_ne, n_se, n_sw

          else
            write(lund,'(6i5)') nt, ia, 0, 0, 0, 0
            write(lund,'(10x,4i5)')    -1,-1,-1,-1
            write(lund,'(10x,4i5)')    -1,-1,-1,-1
          endif  
        enddo
    enddo
    !---------------------------------------------------------------------------

    !- close files
    close(lund)
    if (nproci == ntaskmax .and. nprocj == ntaskmax) close(luns)

  end subroutine decompose_write_MxN

!===============================================================================

  subroutine decompose_optimal_Mx1(ni, mmx, kmx, mm1, jlow, jhgh, maxtt,       &
                                   ilows, ihghs,                               &
                                   nforbidden, iforbiddenp, iforbiddenm)

    implicit none 
 
    integer(4), intent(in)    :: ni, mmx, kmx, jlow, jhgh, maxtt
    integer(4), intent(in)    :: mm1(0:,0:,1:)
    integer(4), intent(in)    :: nforbidden, iforbiddenp(:), iforbiddenm(:)
    integer(4), intent(inout) :: ilows(:), ihghs(:)

    integer(4) :: iweven, iw, i, j, k, il, isum, isl, tsum, isum2, ifbn
    integer(4) :: idist(mmx)
    logical    :: notforbidden

    !- even split:
    iweven = maxtt/ni

    !- distribution of wet points in i-direction:
    do i=1,mmx
      iw = 0
      do k=1,kmx
        do j=jlow,jhgh
          if (mm1(i,j,k) > 0) iw = iw + 1
        enddo
      enddo
      idist(i) = iw
    enddo

    !- find i-slices:
    tsum  = 0
    isum  = 0
    isum2 = 0
    il    = 1
    isl   = 1
    do i=1,mmx
      isum = isum + idist(i)
      if (i < mmx) isum2 = isum + idist(i+1)
      ! raise a flag if we are on a forbidden i-line:
      notforbidden = .true.
      do ifbn=1,nforbidden
        if (i == iforbiddenp(ifbn)) then
          notforbidden = .false.
          exit
        endif
        if (i == iforbiddenm(ifbn)) then
          notforbidden = .false.
          exit
        endif
      enddo
            
      ! out of i-lines, put remainding points on the last mpi task
      if (isl == ni) then
        ilows(isl) = il
        ihghs(isl) = mmx
        exit
      endif

      ! Do we have enough wet points to split the domain here?
      if ( notforbidden .and. (isum >= iweven .or.                             &
         (i < mmx .and. isum2 > iweven .and. isum2-iweven > iweven-isum)) ) then

        ilows(isl) = il
        ihghs(isl) = i
        il    = i + 1
        isl   = isl + 1
        tsum  = tsum + isum
        isum  = 0
        isum2 = 0
      elseif (i==mmx) then
        ilows(isl) = il
        ihghs(isl) = i
      endif
    enddo
  end subroutine decompose_optimal_Mx1

!===============================================================================

  subroutine get_forbidden_i(narea, cfgfile, nz, nu, nv, kmx,                  &
                             nforbidden, iforbiddenp, iforbiddenm)

    implicit none

    character(256), intent(in)  :: cfgfile(:)
    integer(4),     intent(in)  :: narea
    integer(4),     intent(in)  :: nz(:), nu(:), nv(:), kmx(:)
    integer(4),     intent(out) :: nforbidden(:)
    integer(4),     intent(out) :: iforbiddenp(:,:), iforbiddenm(:,:)

    integer(4) :: nzn, nvs, ia

    nforbidden(1:narea) = 0
    iforbiddenp(:,1:narea) = -2
    iforbiddenm(:,1:narea) = -2

    ! fixme: This will only fix known problems near open North and South 
    !        z-boundaries and near nesting South and North v-borders.  If we see
    !        issues at other kinds of boundaries/borders in the future, then 
    !        these must be dealt with.

    ! Read bnds and brds from file, and find possible
    ! N-z-bnds (iforbiddenp) and S-v-brds (iforbiddenm):
    do ia=1,narea
      call read_data(cfgfile(ia), kmx(ia), nz(ia), nu(ia), nv(ia), nzn, nvs,   &
                     iforbiddenp(:,ia), iforbiddenm(:,ia) )
      nforbidden(ia) = max(nzn, nvs)
    enddo

  end subroutine get_forbidden_i

!===============================================================================

  subroutine read_data (cfgf, ke, nz, nu, nv, nzn, nvs, iforbp, iforbm)

    use io_subs,        only : io_new_unit
    use exits,          only : exitme

    implicit none
  
    character(256), intent(in)    :: cfgf
    integer(4),     intent(in)    :: ke, nz, nu, nv
    integer(4),     intent(out)   :: nzn, nvs
    integer(4),     intent(inout) :: iforbp(:), iforbm(:)
  
    character(3) :: ch3
    character(1) :: ch1
    integer(4)   :: idum, idums,is, knl, ios, ir, k
    integer(4)   :: krz(1:3,0:nz), kru(1:3,0:nu), krv(1:3,0:nv)
    real(8)      :: rdum
    logical      :: nsbound
    ! NOTE: Intenitionally, ch1,kru,rdum are read in but never used.

    !- Initialise some parameters with sane values -----------------------------
    krz(:,0:) = 0
    kru(:,0:) = 0
    krv(:,0:) = 0
    nzn = 0
    nvs = 0
    
    !---- Open data file -------------------------------------------------------
    knl = io_new_unit()
    open(knl,file=trim(cfgf),status='old',iostat=ios)
    if (ios /= 0) call exitme (1,'Error: Cannot open configuration file')

    !---- Skip irrelevant parameters -------------------------------------------
    do is=1,9999
      read(knl,'(2X,A3)') ch3
      if(ch3 == 'POS') exit
    enddo
    read(knl,'(1X)')
  
    read(knl,'(17X,3I3,A1)') idum,idum,idum,ch1
    read(knl,'(17X,3I3,A1)') idum,idum,idum,ch1
  
    read(knl,'(1X,/)')
    read(knl,'(17X,3I3)') idum,idum,idum
    read(knl,'(17X,3I3)') idum,idum,idum
  
    read(knl,'(1X,/)')
    do k=1,ke
      read(knl,'(16X,F10.5)') rdum
    enddo
  
    read(knl,'(1X,/)')
    read(knl,'(16X,F10.5)') rdum
  
    read(knl,'(1X,/)')
    read(knl,'(16X,F10.5)') rdum
    read(knl,'(16X,F10.5)') rdum
  
    read(knl,'(1X,/)')
    read(knl,'(16X,F10.5)') rdum
  
    do is=1,9999
      read(knl,'(A3)') ch3
      if(ch3 == '===') exit
    enddo

    !---- Read relevant parameters ---------------------------------------------
    read(knl,'(1X,//)')
    do ir=1,nz
      read(knl,'(40X,3I5)') krz(1,ir), krz(2,ir), krz(3,ir)
    enddo
  
    read(knl,'(1X,///)')
    do ir=1,nu
      read(knl,'(40X,3I5)') kru(1,ir),kru(2,ir),kru(3,ir)
    enddo
  
    read(knl,'(1X,///)')
    do ir=1,nv
      read(knl,'(40X,3I5)') krv(1,ir),krv(2,ir),krv(3,ir)
    enddo

    close(knl)
  
    !---- Look for NS-z-bnds ---------------------------------------------------
    if (nz > 0) then
      idum  = 0
      idums = 0
      do ir=1,nz
        ! Northern boundary and the i-line below
        if (krz(3,ir) == 2 .and. krz(1,ir) /= idum) then
          nzn  = nzn + 2
          idum = krz(1,ir)
          iforbp(nzn-1) = idum
          iforbp(nzn)   = idum+1 ! the line below 
        endif
        ! Southern boundary and the i-line above
        if (krz(3,ir) == 4 .and. krz(1,ir) /= idums) then
          nzn  = nzn + 2
          idums = krz(1,ir)
          iforbp(nzn-1) = idums-1 ! the line above 
          iforbp(nzn)   = idums
        endif
      enddo
    endif

    !---- Look for NS-v-brds ---------------------------------------------------
    if (nv > 0) then
      idum = 0
      do ir=1,nv
        nsbound = .false.
        if (krv(3,ir) == 2 .or. krv(3,ir) == 4) nsbound = .true. 

        if ( nsbound .and. krv(1,ir) /= idum) then
          nvs  = nvs + 1
          idum = krv(1,ir)
          iforbm(nvs) = idum
        endif
      enddo
    endif

    !- That's all, folks! 

  end subroutine read_data

!===============================================================================

end module cmod_auto_decompo

