module cmod_date

  implicit none
  private

  public  :: jultimea, caltime
  private :: caldat
  public  :: julday

contains

!===============================================================================

  integer(4) function julday(mm,id,iyyy)
  
    !- modules -----------------------------------------------------------------
    use exits, only : exitme
  
    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: mm, id, iyyy
  
    !- local vars --------------------------------------------------------------
    integer(4), parameter  :: igreg = 15 + 31*(10+12*1582)
    integer(4)             :: ja, jm, jy
    real(8),    parameter  :: dpry = 365.25_8, dprm = 30.6001_8
    real(8),    parameter  :: a = 0.01_8, b = 0.25_8

    !- intrinsic ---------------------------------------------------------------
    intrinsic :: real, int
  
    jy = iyyy
    if ( jy == 0 ) then
      call exitme("function julday: there is no year zero")
    endif
    if ( jy < 0 ) jy = jy + 1
    if ( mm > 2 ) then
      jm = mm + 1
    else
      jy = jy - 1
      jm = mm + 13
    endif
    julday = int(dpry*real(jy,8)) + int(dprm*real(jm,8)) + id + 1720995
    if ( id+31*(mm+12*iyyy) >= igreg ) then
      ja     = int(a*real(jy,8))
      julday = julday + 2 - ja + int(b*real(ja,8))
    endif
  
  end function julday

!===============================================================================

  subroutine jultimea(At,Tex)
  ! time calculation from string input
  ! calendar time -> julian time (since 1/1/1900)
  ! string format: '0000.00.00 00:00:00', 19 digits
  ! tex: julian time since 1.1.1900
  ! id: calendar day
  ! mm: calendar month
  ! iy: calendar year
  ! ih: hour
  ! minn: minute
  ! nsecs: second
  ! at: time in string format
  !
    !- directives --------------------------------------------------------------
    implicit none
  
    !- args --------------------------------------------------------------------
    character(19), intent(in)  :: At     ! watch out here
    real(8),       intent(out) :: Tex
  
    !- locals ------------------------------------------------------------------
    integer(4)         :: id , ih , iy , minn , mm , nsecs
    real(8), parameter :: hprd = 24.0_8, mprd = 1440.0_8, sprd = 86400.0_8

    !- intrinsic ---------------------------------------------------------------
    intrinsic :: real
  
    read (At( 1:4), '(i4.4)') iy
    read (At( 6:7), '(i2.2)') mm
    read (At( 9:10),'(i2.2)') id
    read (At(12:13),'(i2.2)') ih
    read (At(15:16),'(i2.2)') minn
    read (At(18:19),'(i2.2)') nsecs
    Tex = real(julday(mm,id,iy)-julday(1,1,1900),8)
    Tex = Tex + real(ih,8)/hprd + real(minn,8)/mprd+ real(nsecs,8)/sprd

  end subroutine jultimea

!===============================================================================

  subroutine caltime(tex,id,mm,iy,ih,minn,nsecs,tr,at)
  ! julian time (since 1/1/1900) -> calendar time
  ! tex: julian time since 1.1.1900
  ! id: calendar day
  ! mm: calendar month
  ! iy: calendar year
  ! ih: hour
  ! minn: minute
  ! nsecs: second
  ! tr: time residual
  ! at: time in string format

    !- directives --------------------------------------------------------------
    implicit none

    !- args --------------------------------------------------------------------
    character(19), intent(out) :: at         ! watch out here
    integer(4),    intent(out) :: ih, id, iy, minn, mm, nsecs
    real(8),       intent(in)  :: tex
    real(8),       intent(out) :: tr

    !- locals ------------------------------------------------------------------
    integer(4)         :: juld, nsec
    real(8)            :: texs, tpart
    real(8), parameter :: ONE = 1.0_8, EPS = 2.0e-7_8
    real(8), parameter :: SPD = 24.0_8*60.0_8*60.0_8  ! seconds per day

    !- intrinsic ---------------------------------------------------------------
    intrinsic :: real, mod, int, nint
   
  !     plus EPS to avoid 24:00:00 because of computional problems
  !     2**15 = 32768 -> 19.09.1989
   
    texs  = tex + EPS
    tpart = mod(texs,ONE)
    nsec  = nint(tpart*SPD)
    tr    = tpart - real(nsec,8)/SPD
    nsecs = mod(nsec,60)
    minn  = mod(nsec-nsecs,3600)/60
    ih    = (nsec-nsecs-minn*60)/3600
    juld  = int(texs) + julday(1,1,1900)
    call caldat(juld,mm,id,iy)
   
    at = '0000.00.00 00:00:00'
    write (at( 1:4), '(i4.4)') iy
    write (at( 6:7), '(i2.2)') mm
    write (at( 9:10),'(i2.2)') id
    write (at(12:13),'(i2.2)') ih
    write (at(15:16),'(i2.2)') minn
    write (at(18:19),'(i2.2)') nsecs
  
  end subroutine caltime
  
!===============================================================================

  subroutine caldat(julian,mm,id,iyyy)
    implicit none

    integer(4), intent(in)  :: julian
    integer(4), intent(out) :: id, iyyy, mm

    integer(4), parameter   :: igreg = 2299161
    integer(4)              :: ja, jalpha, jb, jc, jd, je

    real(8),    parameter   :: q = 0.25_8, c = 36524.25_8, dprm = 30.6001_8
    real(8),    parameter   :: f1 = 6680.0_8, f2 = 122.1_8, f3 = 365.25_8

    intrinsic :: real, int

    ! julian days --> calendar date
    if ( julian >= igreg ) then
      jalpha = int((real(julian-1867216,8)-q)/c)
      ja = julian + 1 + jalpha - int(q*real(jalpha,8))
    else
      ja = julian
    endif
    jb = ja + 1524
    jc = int(f1 + (real(jb-2439870,8)-f2)/f3)
    jd = 365*jc + int(q*real(jc,8))
    je = int(real(jb-jd,8)/dprm)
    id = jb - jd - int(dprm*real(je,8))
    mm = je - 1
    if ( mm > 12 ) mm = mm - 12
    iyyy = jc - 4715
    if ( mm > 2 ) iyyy = iyyy - 1
    if ( iyyy <= 0 ) iyyy = iyyy - 1
  
  end subroutine caldat
  
!===============================================================================

end module cmod_date
