module cmod_dens
  use constants, only : one
  implicit none
  private
  public :: dens
  private :: densit

contains

  real(8) function densit (s,t,p)
  !-----------------------------------------------------------------------------
  !
  ! Function that calculates the density of sea water using
  ! the UNESCO equation of state.
  !
  ! Comments
  ! Re-written from the original F77 function in cmodlib 
  ! to F90 such that it is ready for in-lining if we choose to do so
  !
  ! Then, optimised with respect to No. of floating point multiplication
  !   (from 46 to 39)
  !
  !-----------------------------------------------------------------------------
  
    implicit none
  
    ! salinity, temperature, pressure (in bar):
    real(8), intent(in) :: s, t, p 
  
  
    real(8) :: sq, rhonul, sbm
    !**********      coefficients   ********************************************
    real(8), parameter :: a0 = 999.842594_8 ,                                  &
                          a1 =   6.793952e-2_8 ,                               &
                          a2 =  -9.095290e-3_8 ,                               &
                          a3 =   1.001685e-4_8 ,                               &
                          a4 =  -1.120083e-6_8 ,                               &
                          a5 =   6.536332e-9_8 ,                               &
                          b0 =  +8.24493e-1_8  ,                               &
                          b1 =  -4.0899e-3_8  ,                                &
                          b2 =  +7.6438e-5_8  ,                                &
                          b3 =  -8.2467e-7_8  ,                                &
                          b4 =  +5.3875e-9_8  ,                                &
                          c0 =  -5.72466e-3_8  ,                               &
                          c1 =  +1.0227e-4_8  ,                                &
                          c2 =  -1.6546e-6_8  ,                                &
                          d0 =  +4.8314e-4_8  ,                                &
                          e0 = 19652.21_8     ,                                &
                          e1 =   148.4206_8   ,                                &
                          e2 =    -2.327105_8 ,                                &
                          e3 =     1.360477e-2_8 ,                             &
                          e4 =    -5.155288e-5_8                           
  
    real(8), parameter :: f0 =    54.6746_8  ,                                 &
                          f1 =    -0.603459_8 ,                                &
                          f2 =     1.09987e-2_8 ,                              &
                          f3 =    -6.1670e-5_8 ,                               &
                          g0 =     7.944e-2_8  ,                               &
                          g1 =     1.6483e-2_8  ,                              &
                          g2 =    -5.3009e-4_8  ,                              &
                          h0 =  3.239908_8    ,                                &
                          h1 =  1.43713e-3_8 ,                                 &
                          h2 =  1.16092e-4_8 ,                                 &
                          h3 = -5.77905e-7_8 ,                                 &
                          i0 =  2.2838e-3_8  ,                                 &
                          i1 = -1.0981e-5_8  ,                                 &
                          i2 = -1.6078e-6_8  ,                                 &
                          j0 =  1.91075e-4_8 ,                                 &
                          k0 =  8.50935e-5_8 ,                                 &
                          k1 = -6.12293e-6_8 ,                                 &
                          k2 =  5.2787e-8_8 ,                                  &
                          l0 = -9.9348e-7_8  ,                                 &
                          l1 =  2.0816e-8_8  ,                                 &
                          l2 =  9.1697e-10_8 
    !***************************************************************************
    sq = sqrt(s)
  
    rhonul =  a0+s*(b0+c0*sq+d0*s) +                                           &
             (a1+s*(b1+c1*sq)      +                                           &
             (a2+s*(b2+c2*sq)      +                                           &
             (a3+s*b3              +                                           &
             (a4+s*b4              +                                           &
              a5                    *t)*t)*t)*t)*t
  
    sbm    =  e0+s*(f0+g0*sq)+p*(h0+s*(i0+j0*sq)+p*(k0+l0*s)) +                &
             (e1+s*(f1+g1*sq)+p*(h1+s* i1       +p*(k1+l1*s)) +                &
             (e2+s*(f2+g2*sq)+p*(h2+s* i2       +p*(k2+l2*s)) +                &
             (e3+s* f3       +p* h3                           +                &
              e4                                               *t)*t)*t)*t
             
    densit = rhonul/(one-p/sbm)
  
  end function densit

subroutine dens (m,ind,n2d,kh,dz,cmp,rho,press,ompidx)
!subroutine dens (m,ind,n2d,kh,h,dz,cmp,rho,press,ompidx)

  use dmi_omp,    only : domp_get_domain
  use constants,  only : dyn, g

  implicit none

  integer(4), intent(in)    :: n2d
  integer(4), intent(in)    :: ind(:,:),kh(0:)
  integer(4), intent(in)    :: m(:,0:,0:)
  real(8),    intent(in)    :: dz(:)
!  real(8),    intent(in)    :: h(0:)
  real(8),    intent(inout) :: cmp(1:,0:)
  real(8),    intent(inout) :: rho(0:), press(0:)
  integer(4), intent(inout) :: ompidx(:,:)

  integer(4)         :: n,i,j,k,kb,mc, mm2
  real(8)            :: pp
  real(8)            :: hho,rhoo,presso,hhu,rhou,pressu
  real(8), parameter :: gh = 0.5_8*g, a = 0.1_8, b = 0.05_8
  integer(4)         :: n2dl, n2du

 
  call domp_get_domain(kh, 1, n2d, n2dl, n2du, ompidx)

  !- Obtain density from S, T and P --------------------------------------------
  do n=n2dl, n2du

    kb = kh(n)

    hho = dz(1)
    pp  = a + b*hho

    rho(n) = densit(cmp(2,n),cmp(1,n),pp)             !     pp  [ druck in bar ]

    if (kb < 2) cycle

    i  = ind(1,n)
    j  = ind(2,n)

    hhu = dz(2)
    pp  = pp + b*(hho+hhu)  ! 0.05 = 0.1/2.0
    hho = hhu
    mm2 = m(2,i,j) - 2

    !- treat water column down to one above bottom:
    do k=2,kb-1
      mc      = mm2 + k
      rho(mc) = densit(cmp(2,mc),cmp(1,mc),pp)       !     pp  [ druck in bar ]
      hhu     = dz(k+1)
      pp      = pp + b*(hho+hhu)  ! 0.05 = 0.1/2.0
      hho     = hhu
    enddo

    !- treat bottom:
    mc      = mm2 + kb 
    rho(mc) = densit(cmp(2,mc),cmp(1,mc),pp)

  enddo

  !- Obtain baroclinic pressure ------------------------------------------------
  if (.not. dyn) then
    do n=n2dl, n2du

      kb = kh(n)

      !- treat k=1:
      hhu      = dz(1)
      rhou     = rho(n)
      pressu   = gh*hhu*rhou
      press(n) = pressu

      if (kb < 2) cycle

      i = ind(1,n)
      j = ind(2,n)

      mm2 = m(2,i,j) - 2

      !- treat rest of water column:
      do k=2,kb
        hho       = hhu
        rhoo      = rhou
        presso    = pressu
        mc        = mm2 + k
        hhu       = dz(k)      !sd       hhu=max(h(m(i,j,k)),0.0d00)
        rhou      = rho(mc)
        pressu    = presso + gh*(hho*rhoo+hhu*rhou)
        press(mc) = pressu
      enddo
    enddo
  endif

end subroutine dens

end module cmod_dens
