module cmod_init

  use constants, only : zero

  implicit none
  private

  public  :: init_z, initrwqk, inituvwzi, init, prep, entr
  public  :: ice_vel_init, kh_numa, init_uvdams

contains

  !=============================================================================

  subroutine ice_vel_init(ue, ve)
    implicit none

    real(8), intent(out) :: ue(0:), ve(0:)

    ue(0:)   = zero
    ve(0:)   = zero

  end subroutine ice_vel_init

  !=============================================================================

  subroutine init_z (iw2, nz, me, krz, hz, h, elev_ref, z, zn, rwz, zetac )
    
    !- modules -----------------------------------------------------------------
    use dmi_omp,   only : domp_get_domain
    use constants, only : hgr, dyn, coldst, elevation_ref
  
    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: iw2, nz
    integer(4), intent(in) :: me(0:,0:), krz(1:,0:)
    real(8), intent(in)    :: hz(0:)
    real(8), intent(in)    :: elev_ref(0:)
    real(8), intent(inout) :: h(0:)
    real(8), intent(inout) :: z(0:), rwz(0:)
    real(8), intent(inout) :: zn(0:), zetac(0:)
  
    !- local vars --------------------------------------------------------------
    integer(4) :: ii, il, iu, ibnd
  
    ! --------------------------------------------------------------------------
  
    call  domp_get_domain(1, iw2, il, iu)
  
    !- Set z to reference hight when it is a colstart
    if (.not. elevation_ref) then
      if (coldst) z(il:iu) = zero
      !- treat possibly dried points:
      if (.not.dyn .or. coldst) then
        do ii=il,iu
          z(ii) = max( -hz(ii)+hgr, z(ii) )
          h(ii) = hz(ii) + z(ii)
        enddo
      endif
    else
      if (coldst) then
        do ii=il,iu
          z(ii)  = elev_ref(ii)
        enddo  
      endif
      !- treat possibly dried points:
      if (.not.dyn .or. coldst) then
        do ii=il,iu
          z(ii)  = max( -(hz(ii)-elev_ref(ii))+hgr, z(ii) )
          h(ii)  = hz(ii) + ( z(ii) - elev_ref(ii) )
        enddo
      endif
    endif

    !- initialize "new" z
    zn(il:iu) = z(il:iu)

  
!$OMP BARRIER
!$OMP MASTER
    !- initialize on-land value of "new" z:
    zn(0)  = z(0)
    rwz(0) = zero
  
    !- make boundary values consistent with interior:
    zetac(:) = zero
    do ibnd=1,nz
      ii = me(krz(1,ibnd),krz(2,ibnd))
      if (krz(3,ibnd) > 0 .and. ii > 0) then
        rwz(ibnd) = z(ii)
      else
        rwz(ibnd) = zero
      endif
    enddo

!$OMP END MASTER
  
  end subroutine init_z

  !=============================================================================

  subroutine initrwqk (rwqk)
  
    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    real(8), intent(out) :: rwqk(0:,:)
  
    rwqk(0:,:)= zero ! all points
  
  end subroutine initrwqk

  !=============================================================================

  subroutine inituvwzi (ui,vi,wi,iwet2,kh,ind,mm,idx)
  
    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain
  
    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)                :: iwet2
    integer(4), intent(in)                :: kh(0:), ind(:,:)
    integer(4), intent(in)                :: mm(:,0:,0:)
    real(8), intent(inout), dimension(0:) :: ui, vi, wi
    integer(4), intent(inout)             :: idx(1:,1:)
        
    !- local variables ---------------------------------------------------------
    integer(4)            :: nsrf, nl, nu, i, j, k, mi

!$OMP BARRIER
!$OMP MASTER
    !- initialize on-land values:
    ui(0) = zero
    vi(0) = zero
    wi(0) = zero
!$OMP END MASTER

    call domp_get_domain(kh, 1, iwet2, nl, nu, idx)
  
    do nsrf=nl,nu
      i=ind(1,nsrf)
      j=ind(2,nsrf)
      do k=1,kh(nsrf)
        mi = mm(k,i,j)
        ui(mi) = zero
        vi(mi) = zero
        wi(mi) = zero
      enddo
    enddo
  
  end subroutine inituvwzi

  !=============================================================================

  subroutine init (                                                            &
  !-----------------------------------------------------------------------------
  ! initialising some parameters from the configuration file
  !-----------------------------------------------------------------------------
         knl                                                                   &
        ,mx,nx, m,n,ke,mm,nlevs,iga,jga                                        &
        ,dxk,xkoor,dyk,ykoor,dz                                                &
        ,dt,dx,dy,ft,rt,ty,tx                                                  &
        ,gty,gtx,covo,covu                                                     &
        ,z,hz,h,elev_ref,ti                                                    &
        ,nz,nu,kru,nv,krv                                                      &
        ,nq,krq,rwq,nud,lud,nvd,lvd,nweir,weir_idx,zweir,wflowr                &
        , ahc, avc, idx,                                                       &
         nnest, znest, unest, vnest, narea, znest_s, unest_s, vnest_s, kdry )
   
    use exits,     only : exitme
    use constants, only : g,grarad,omega2,gmaequ,dyn,weirs,coriolis,           &
                          elevation_ref
  
    implicit none
  
    integer(4), intent(in)    :: knl,mx,nx,m,n,ke,nlevs,nnest, narea
    integer(4), intent(in)    :: nz,nu,nv,nq,nud,nvd,nweir
    integer(4), intent(in)    :: mm(0:,0:,:)
    integer(4), intent(out)   :: iga(:,:),jga(:,:)  !(3,1:max(1,nlevs))
    real(8), intent(in)       :: hz(0:)
    real(8), intent(in)       :: elev_ref(0:)
    real(8), intent(out)      :: xkoor(:),ykoor(:),dz(:)
    real(8), intent(out)      :: dt,rt,dy,ty,gty,dxk,dyk
    real(8), intent(out)      :: dx(:),ft(:),tx(:),gtx(:),covo(:),covu(:)
    real(8), intent(out)      :: h(0:),z(0:),ti(0:)
    real(8), intent(out)      :: ahc,avc
    real(8), intent(out)      :: rwq(0:)
    real(8), intent(out)      :: zweir(0:), wflowr(0:)
    integer(4), intent(out)   :: krq(:,0:),kru(:,0:),krv(:,0:)
    integer(4), intent(out)   :: lvd(:),lud(:)    ! (nvd),(nud)
    integer(4), intent(out)   :: weir_idx(:,0:)   ! (3,0:nweir)
    integer(4), intent(out)   :: idx(1:,1:)
    integer(4), intent(out)   :: kdry
    integer(4), intent(inout) :: znest(:,:),unest(:,:),vnest(:,:) 
    integer(4), intent(inout) :: znest_s(:,:),unest_s(:,:),vnest_s(:,:) 
                                                              ! (max(1,nlevs),2)
  
  
    character(3) :: zeile3
    character(1) :: ns,ew
    integer(4)   :: isuch,iphig,iphim,iphis,ilamg,ilamm,ilams
    integer(4)   :: idphg,idphm,idphs,idlag,idlam,idlas,i,j,k,ir,iii,jjj
    integer(4)   :: iu, iv
    real(8)      :: deltat,aush,ausv,rbot,r,dphi,dlam,phinord,phiz
    real(8)      :: sinf,cosf,cosvo,cosvu
    real(8)      :: xlamwest,xlaz,tiefe

    real(8), parameter :: hm = 60.0_8, hs = 3600.0_8
  
    !    dt:     time step size
    !    dx(i):  grid spacing in x-direction
    !    dy:     grid spacing in y-direction
    !    g:      gravity constant
    !    hgr:    limit depth (0.05m)
    !    ah:     austausch coefficient
    !    f(i):   coriolis parameter
    !    ft(i):  f(i)*dt
    !    r:      bottom roughness
    !    rt:     r*dt
    !    ty:     dt/dy
    !    tx(i):  dt/dx(i)

    !- Perform a little validation (just in case ...) --------------------------
    if (m /= mx+1) call exitme(1,'Invalid N-S dimensions')
    if (n /= nx+1) call exitme(1,'Invalid W-E dimensions')
  

    !- Initialise some parameters with sane values -----------------------------
    kru(:,0:) = 0
    krv(:,0:) = 0
    

    !---- Read in model parameters: --------------------------------------------
  
    !---------------------------------------------------------------------------
    do isuch=1,9999
      read(knl,'(2X,A3)') zeile3
      if(zeile3 == 'POS') exit
    enddo
    read(knl,'(1X)')
  
    read(knl,'(17X,3I3,A1)')iphig,iphim,iphis,ns
    read(knl,'(17X,3I3,A1)')ilamg,ilamm,ilams,ew
    if (ns /= 'N'.and.ns /= 'S') call exitme(1,'Please specify N or S')
    if (ew /= 'E'.and.ew /= 'W') call exitme(1,'Please specify E or W')
  
    read(knl,'(1X,/)')
    read(knl,'(17X,3I3)')idphg,idphm,idphs
    read(knl,'(17X,3I3)')idlag,idlam,idlas
  
    read(knl,'(1X,/)')
    do k=1,ke
      read(knl,'(16X,F10.5)')dz(k)
    enddo
  
    read(knl,'(1X,/)')
    read(knl,'(16X,F10.5)')deltat
  
    read(knl,'(1X,/)')
    read(knl,'(16X,F10.5)')aush
    read(knl,'(16X,F10.5)')ausv
  
    read(knl,'(1X,/)')
    read(knl,'(16X,F10.5)')rbot
  
    do isuch=1,9999
      read(knl,'(A3)') zeile3
      if(zeile3 == '===') exit
    enddo
    read(knl,'(1X,//)')
    do ir=1,nz
      read(knl,'(a)')
    enddo
  
    read(knl,'(1X,///)')
    do ir=1,nu
      read(knl,'(40X,3I5)')kru(1,ir),kru(2,ir),kru(3,ir)
    enddo
  
    read(knl,'(1X,///)')
    do ir=1,nv
      read(knl,'(40X,3I5)')krv(1,ir),krv(2,ir),krv(3,ir)
    enddo
  
    read(knl,'(1X,///)')
    do ir=1,nq
      read(knl,'(40X,2I5,F10.2)')krq(1,ir),krq(2,ir),rwq(ir)
    enddo
  
    read(knl,'(1X,///)')
    do ir=1,nud
      read(knl,'(40X,2I5)')iii,jjj
      lud(ir)=iii*1000+jjj
    enddo
  
    read(knl,'(1X,///)')
    do ir=1,nvd
      read(knl,'(40X,2I5)')iii,jjj
      lvd(ir)=iii*1000+jjj
    enddo

    if (weirs) then
      zweir(0)  = zero
      wflowr(0) = zero
      read(knl,'(1X,///)')
      do ir=1,nweir
        read(knl,'(40X,3I5,2F10.2)') weir_idx(1,ir), weir_idx(2,ir),           &
                                     weir_idx(3,ir), zweir(ir), wflowr(ir)
      enddo
    endif    
 
    do isuch=1,9999
      read(knl,'(A3)')zeile3
      if(zeile3 == '===') exit
    enddo
    read(knl,'(1X,/)')
    do ir=1,max(1,nlevs)
      do j=1,3
        read(knl,'(40X,2I5)')iga(j,ir),jga(j,ir)
      enddo
    enddo
  
    !- if we have more than two nested domains we must supply more nesting-info:
    if (narea > 2) then
      !- find out which z-bnd points are nesting to which enclosing areas:
      if (nlevs > 0) then
        zeile3 = 'xxx'
        do while (zeile3 /= '===')
          read(knl,'(A3)') zeile3
        enddo
        read(knl,'(1X,/)')
        do ir=1,nlevs
          read(knl,'(40X,2I5)') znest(ir,1), znest(ir,2)
        enddo
      endif
  
      !- find out which uv-bnd points are nesting to which areas:
      if (nnest > 0) then
        zeile3 = 'xxx'
        do while (zeile3 /= '===')
          read(knl,'(A3)') zeile3
        enddo
        read(knl,'(1X,/)')
        do ir=1,nnest
          read(knl,'(40X,4I5)') unest(ir,1),unest(ir,2),vnest(ir,1),vnest(ir,2)
        enddo
      endif
  
    else
      !- OK, don't worry, we can figure it out:
      if (nlevs > 0) then
        znest(1,1) = 1
        znest(1,2) = nz
      endif
      if (nnest > 0) then
        if (nu > 0) then
          unest(1,1) = 1
          unest(1,2) = nu
        endif
        if (nv > 0) then
          vnest(1,1) = 1
          vnest(1,2) = nv
        endif
      endif
    endif
  
    !- make sure to keep a copy for serial applications ------------------------
    znest_s(:,:) = znest(:,:)
    unest_s(:,:) = unest(:,:)
    vnest_s(:,:) = vnest(:,:)
  

    !- Save corner information of uv-boundary in kru(4,0:) and krv(4,:) with ---
    !-   following code:
    !-  0 - no corner
    !-  4 - corner of S and W bnd-point (corse grid) (N/E-corner in fine grid)
    !-  3 - corner of N and W bnd-point (corse grid) (S/E-corner in fine grid)
    !-  2 - corner of N and E bnd-point (corse grid) (S/W-corner in fine grid)
    !-  1 - corner of S and E bnd-point (corse grid) (N/W-corner in fine grid)
    do ir=1,nnest
      do iu=unest(ir,1),unest(ir,2)
        if (kru(3,iu) == 1) then
          do iv=vnest(ir,1),vnest(ir,2)
            if(krv(1,iv) == kru(1,iu)-1 .and. krv(2,iv) == kru(2,iu)) then
              kru(4,iu) = 4
              krv(4,iv) = 4
            elseif(krv(1,iv) == kru(1,iu) .and. krv(2,iv) == kru(2,iu)) then
              kru(4,iu) = 3
              krv(4,iv) = 3
            endif
          enddo
        elseif (kru(3,iu) == 3) then
          do iv=vnest(ir,1),vnest(ir,2)
            if(krv(1,iv) == kru(1,iu) .and. krv(2,iv) == kru(2,iu)+1) then
              kru(4,iu) = 2
              krv(4,iv) = 2
            elseif(krv(1,iv) == kru(1,iu)-1 .and. krv(2,iv) == kru(2,iu)+1) then
              kru(4,iu) = 1
              krv(4,iv) = 1
            endif
          enddo
        endif
      enddo
    enddo


    !--------- model constants -------------------------------------------------
  
    ahc = aush
    avc = ausv
    r   = rbot
  
    !---- spactial and temporal discretisations --------------------------------
  
    dphi = idphg+idphm/hm+idphs/hs
    dlam = idlag+idlam/hm+idlas/hs
    dt   = deltat
  
    !---- combined quatities ---------------------------------------------------
  
    rt  = r*dt
    dy  = dphi*gmaequ
    ty  = dt/dy
    gty = g*ty
  
    phinord = iphig + iphim/hm + iphis/hs
    if (ns == 'S') phinord = -phinord
    phiz = phinord*grarad
    dphi = dphi*grarad
  
    do i=1,mx
      sinf    = sin(phiz)
      cosf    = cos(phiz)
      cosvo   = cos(phiz + dphi*.5_8)
      cosvu   = cos(phiz - dphi*.5_8)
      covo(i) = cosvo/cosf
      covu(i) = cosvu/cosf
      dx(i)   = dlam*gmaequ*cosf
      tx(i)   = dt/dx(i)
      gtx(i)  = g*tx(i)
      ft(i)   = omega2*sinf*dt
      phiz    = phiz - dphi
    enddo

    ! if coriolis force should be turned off, set ft to zero
    if (.not.coriolis) then
      ft(:) = zero
    endif
   
    !     grid spacings in x- and y-direction
  
    xlamwest = ilamg + ilamm/hm + ilams/hs
    if (ew == 'W') xlamwest = -xlamwest
    xlaz = xlamwest*grarad
    dlam = dlam*grarad
  
    dxk = dlam*gmaequ/grarad
    do j=1,nx
      xkoor(j) = xlaz
      xlaz     = xlaz + dlam
    enddo
  
    dyk  = dphi*gmaequ/grarad
    phiz = phinord*grarad
    do i=1,mx
      ykoor(i) = phiz
      phiz     = phiz - dphi
    enddo
  
    !-------- set the deepest layer which could be fall dry --------------------
    if (dyn) then
      kdry = ke
    else
      kdry = 1
    endif

    !----------------- initialise some fields ----------------------------------
    
    !     initial values (fields set to zero):
  
    z(:)    = zero
  
    if (dyn) then
      h(:)    = zero
    else  
      h(0)    = hz(0)
    endif
  
    if (elevation_ref) then
      do i=1,m
        do j=1,n
          tiefe = -elev_ref(mm(i,j,1))
          do k=1,ke
            if (mm(i,j,k) > 0) then
              !- Initialise the h-field and calculate the depth-field
              if(.not.dyn) then
                h(mm(i,j,k)) = hz(mm(i,j,k))
              endif
              tiefe = tiefe + hz(mm(i,j,k))
            else
              exit
            endif
          enddo
          ti(mm(i,j,1)) = tiefe
        enddo
      enddo
    elseif (dyn) then
      do i=1,m
        do j=1,n
          tiefe = zero
          do k=1,ke
            if (mm(i,j,k) > 0) then
              tiefe = tiefe + hz(mm(i,j,k))
            else
              exit
            endif
          enddo
          ti(mm(i,j,1)) = tiefe
        enddo
      enddo
    else ! default, z-coords wo/elevation_ref
      do i=1,m
        do j=1,n
          tiefe = zero
          do k=1,ke
            if (mm(i,j,k) > 0) then
              h(mm(i,j,k)) = hz(mm(i,j,k))
              tiefe = tiefe + hz(mm(i,j,k))
            else
              exit
            endif
          enddo
          ti(mm(i,j,1)) = tiefe
        enddo
      enddo
    endif
  
    !- OMP arrays are intialised here before usage -----------------------------
    idx(:,:)  = -1
  
  end subroutine init
  
  !=============================================================================

  subroutine prep (imax,jmax,kmax,m,mk,ind,n2d,phi,dy,kh,khu,khv,cosphi,fcorio)
  
    use constants, only : omega2, coriolis

    implicit none
  
    integer(4), intent(in)  :: imax,jmax,kmax,n2d
    integer(4), intent(in)  :: m(0:,0:,:)
    integer(4), intent(out) :: mk(:,0:,0:)
    integer(4), intent(out) :: ind(:,:)
    integer(4), intent(out) :: kh (0:),khu(0:),khv(0:)
    real(8), intent(in)     :: dy(0:), phi(:)
    real(8), intent(out)    :: cosphi(:,0:),fcorio(:,0:)
  
  
    integer(4)         :: n,i,j,k,kk,imaxm1
    real(8)            :: phi0, phi1, ey
  
    ! automatic arrays:
    real(8), dimension(0:imax)   :: qyp, qym
  
    imaxm1=imax-1
  
  
  !   north-south-grid
    do i=0,imaxm1
      ey       = dy(i)+dy(i+1)
      qyp(i)   = dy(i+1)/ey
      qym(i+1) = dy(i)/ey
    enddo
  
    kh(:)  = 0
    khu(:) = 0
    khv(:) = 0

    ind(:,:) = 0
  
    do i=0,imax
      do j=0,jmax
        if(m(i,j,1) == 0) cycle
        ind(1,m(i,j,1))=i
        ind(2,m(i,j,1))=j
      enddo
    enddo
  
    do n=1,n2d
      i=ind(1,n)
      j=ind(2,n)
      kk=0
      do k=1,kmax
        if(m(i,j,k) > 0) kk=kk+1
      enddo
      kh(n)=kk
    enddo
  
    do n=1,n2d
      i=ind(1,n)
      j=ind(2,n)
      khu(n)=min(kh(m(i,j,1)),kh(m(i,j+1,1)))
    enddo
  
    do n=1,n2d
      i=ind(1,n)
      j=ind(2,n)
      khv(n)=min(kh(m(i,j,1)),kh(m(i+1,j,1)))
    enddo
  
  
    phi0 = phi(1)-(phi(2)-phi(1))
    cosphi(1,0) = cos(phi0)
    do i=1,imaxm1
      cosphi(1,i) = cos(phi(i))
    enddo
    phi1 = phi(imaxm1)+(phi(imaxm1)-phi(imaxm1-1))
    cosphi(1,imax) = cos(phi1)
  
    cosphi(2,0) = cos(qyp(0)*phi0 + qym(1)*phi(1))
    do i=1,imaxm1-1
      cosphi(2,i) = cos(qyp(i)*phi(i) + qym(i+1)*phi(i+1))
    enddo
    cosphi(2,imaxm1) = cos(qyp(imaxm1)*phi(imaxm1) + qym(imax)*phi1)
  
    fcorio(1,0) = omega2*sin(phi0)
    do i=1,imaxm1
      fcorio(1,i) = omega2*sin(phi(i))
    enddo
    i = imax
    fcorio(1,i) = omega2*sin(phi1)
  
    fcorio(2,0) = omega2*sin(qyp(0)*phi0+qym(1)*phi(1))
    do i=1,imaxm1-1
      fcorio(2,i) = omega2*sin(qyp(i)*phi(i)+qym(i+1)*phi(i+1))
    enddo
    i = imaxm1
    fcorio(2,i) = omega2*sin(qyp(i)*phi(i)+qym(i+1)*phi1)
  
    ! if coriolis force should be turned off, set fcorio to zero
    if (.not.coriolis) then
      fcorio(:,0:) = zero
    endif

     
    ! Create index array with depth index first
    do j=0,jmax
      do i=0,imax
        do k=1,kmax
          mk(k,i,j) = m(i,j,k)
        enddo
      enddo
    enddo

  end subroutine prep

  !=============================================================================

  subroutine kh_numa (n2d, n3d, kh, kh_fake)
    ! proper numa layout of kh
    use dmi_omp, only : domp_get_domain
    !- directives --------------------------------------------------------------
    implicit none
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)  :: n2d, n3d
    integer(4), intent(in)  :: kh_fake(0:)
    integer(4), intent(inout) :: kh(0:)
    !- local variables ---------------------------------------------------------
    integer(4) :: np, nl, nu
    call domp_get_domain(n3d, kh_fake, 0, n2d, nl, nu)
    do np=nl,nu
      kh(np) = kh_fake(np)
    enddo
  end subroutine kh_numa 

  subroutine entr (mmk,ind,n2d,dx,dy,kh,khu,khv,hx,hy,u,v,w,                   &
                   nbpz,krz,nbpq,krq,rwq,rwqk,cosphi,un,vn)
  
    implicit none

    integer(4), intent(in)    :: n2d,nbpq,nbpz
    integer(4), intent(in)    :: mmk(:,0:,0:), ind(:,:)
    integer(4), intent(in)    :: krq(:,0:),krz(:,0:),kh(0:),khu(0:),khv(0:)
    real(8), intent(inout) :: rwqk(0:,:)
    real(8), intent(in)    :: hx(0:),hy(0:),dy(0:),dx(0:),cosphi(:,0:),rwq(0:)
    real(8), intent(inout) :: u(0:),v(0:),w(0:)
    real(8), intent(out)   :: un(0:),vn(0:)

    integer(4) :: n,i,j,k,kb,kbuw,kbue,kbvs,kbvn,nuw,nue,nvn,nvs,kbm1,kw,mi
    real(8)    :: dxj, dyi, cp1, c2i, c2m



    u(0) = zero
    v(0) = zero
    w(0) = zero


    if (nbpq > 0) then
      do n=1,nbpq
        i  = krq(1,n)
        j  = krq(2,n)
        kb = kh(mmk(1,i,j))
        if (kb <= 0) cycle
        rwqk(n,1) = rwq(n)/(dy(i)*dx(j)*cosphi(1,i))
      enddo
    endif


    do n=1,n2d
      i  = ind(1,n)
      j  = ind(2,n)
      mi = mmk(1,i,j)
      kb = kh(mi)
      do k=max(1,khu(mi)+1),kb
        u(mmk(k,i,j)) = zero
      enddo
      do k=max(1,khv(mi)+1),kb
        v(mmk(k,i,j)) = zero
      enddo
    enddo
  
    !- put some sane numbers into un/vn:
    un(0:) = u(0:)
    vn(0:) = v(0:)
  
    do n=1,n2d
      i  = ind(1,n)
      j  = ind(2,n)
      mi = mmk(1,i,j)
      kb = kh(mi)
      if (kb <= 0) cycle
      cp1  = cosphi(1,i)
      dxj  = dx(j)*cp1
      dyi  = dy(i)*cp1
      c2i  = cosphi(2,i)
      c2m  = cosphi(2,i-1)
      kbuw = khu(mmk(1,i,j-1))
      kbue = khu(mi)
      kbvn = khv(mmk(1,i-1,j))
      kbvs = khv(mi)
      nuw  = mmk(kb,i,j-1)
      nue  = mmk(kb,i, j)
      nvn  = mmk(kb,i-1,j)
      nvs  = mmk(kb,i, j)
      if (kb > kbuw) nuw = 0
      if (kb > kbue) nue = 0
      if (kb > kbvn) nvn = 0
      if (kb > kbvs) nvs = 0
      w(mmk(kb,i,j)) = -(    hx(nue)*u(nue) -     hx(nuw)*u(nuw))/dxj          &
                     -(c2m*hy(nvn)*v(nvn) - c2i*hy(nvs)*v(nvs))/dyi
      if (kb>1) then
        kbm1 = kb-1
        do k=1,kbm1
          kw  = kb - k
          nuw = mmk(kw,i,j-1)
          nue = mmk(kw,i, j)
          nvn = mmk(kw,i-1,j)
          nvs = mmk(kw,i, j)
          if (kw > kbuw) nuw = 0
          if (kw > kbue) nue = 0
          if (kw > kbvn) nvn = 0
          if (kw > kbvs) nvs = 0
          w(mmk(kw,i,j)) = w(mmk(kw+1,i,j))                                    &
                        -(    hx(nue)*u(nue) -     hx(nuw)*u(nuw))/dxj         &
                        -(c2m*hy(nvn)*v(nvn) - c2i*hy(nvs)*v(nvs))/dyi
        enddo
      endif
    enddo
  
  
    if (nbpq > 0) then
      do n=1,nbpq
        i  = krq(1,n)
        j  = krq(2,n)
        mi = mmk(1,i,j)
        w(mi) = w(mi) + rwqk(n,1)
      enddo
    endif
  
  
    if (nbpz > 0) then
      do n=1,nbpz
        i = krz(1,n)
        j = krz(2,n)
        do k=1,kh(mmk(1,i,j))
          w(mmk(k,i,j)) = zero
        enddo
      enddo
    endif
   
  end subroutine entr

  !=============================================================================

  subroutine init_uvdams( uvdam, nud, nvd, lud, lvd, mm, iw2, ih2 )

    use constants, only : one
    implicit none
    integer(4), intent(in)  :: nud, nvd, lud(:), lvd(:), mm(:,0:,0:), iw2, ih2
    real(8),    intent(out) :: uvdam(:,0:)

    integer :: lu, lv, iud, ivd, jud, jvd, mi, i2max

    uvdam(:,0:) = one

    i2max = iw2+ih2

    do lu=1,nud
      iud = int(lud(lu)/1000)
      jud = lud(lu)-iud*1000
      mi  = mm(1,iud,jud)
      if (mi > 0 .and. mi <= i2max) uvdam(1,mi) = zero
    enddo
    do lv=1,nvd
      ivd = int(lvd(lv)/1000)
      jvd = lvd(lv)-ivd*1000
      mi  = mm(1,ivd,jvd)
      if (mi > 0 .and. mi <= i2max) uvdam(2,mi) = zero
    enddo

  end subroutine init_uvdams

  !=============================================================================

end module cmod_init
