module cmod_local_arrays

  !- Modules -------------------------------------------------------------------
  use cmod_mem, only : cml1, cmi1, cmi2, cmi3, cmr1, cmr2

  !- Directives ----------------------------------------------------------------
  implicit none
  private


  !- Local versions of the arrays for MPI support ------------------------------


  !- Dimensions ----------------------------------------------------------------
  integer(4), allocatable, save, public :: iw2_l(:), iw3_l(:)


  !- Index arrays --------------------------------------------------------------
  type(cmi3), allocatable, save, public :: mm1_l(:), mm1k_l(:)
  type(cmi2), allocatable, save, public :: me_l(:)


  !- Index help arrays ---------------------------------------------------------
  type(cmi2), allocatable, save, public :: ind_l(:)
  type(cmi1), allocatable, save, public :: kh_l(:), khu_l(:), khv_l(:)

  !- Depths --------------------------------------------------------------------
  type(cmr1), allocatable, save, public :: ti_l(:), h_l(:), hn_l(:)
  type(cmr1), allocatable, save, public :: hu_l(:), hv_l(:), hz_l(:)
  type(cmr1), allocatable, save, public :: h_old_l(:), h_new_l(:), h_i_l(:)


  !- Sea level and current -----------------------------------------------------
  type(cmr1), allocatable, save, public :: z_l(:), zn_l(:), elev_ref_l(:)
  type(cmr1), allocatable, save, public :: u_l(:), v_l(:), w_l(:)
  type(cmr1), allocatable, save, public :: un_l(:), vn_l(:)
  type(cmr1), allocatable, save, public :: ui_l(:), vi_l(:), wi_l(:)


  !- Wind speed, -stress and air presssure -------------------------------------
  type(cmr1), allocatable, save, public :: wua_l(:), wun_l(:), wva_l(:),wvn_l(:)
  type(cmr1), allocatable, save, public :: wu_l(:), wv_l(:)
  type(cmr1), allocatable, save, public :: taux_l(:), tauy_l(:)
  type(cmr1), allocatable, save, public :: pla_l(:), pln_l(:), pl_l(:)


  !- Air temperature, -humidity and cloud cover --------------------------------
  type(cmr1), allocatable, save, public :: atempa_l(:), atempn_l(:), atemp_l(:)
  type(cmr1), allocatable, save, public :: humida_l(:), humidn_l(:), humid_l(:)
  type(cmr1), allocatable, save, public :: clouda_l(:), cloudn_l(:), cloud_l(:)


  !- Water temperature and salinity --------------------------------------------
  !  Density and baroclinic pressure
  type(cmr1), allocatable, save, public :: rho_l(:), press_l(:), geopot_l(:)

  ! - vertical divergence of shortwave radiation -------------------------------
  type(cmr1), allocatable, save, public :: divrad_l(:), divrad_s_l(:)


  !- Sea bed temperature -------------------------------------------------------
  type(cmr2), allocatable, save, public :: t_soil_l(:)


  !- Precipitation and evaporation ---------------------------------------------
  type(cmr1), allocatable, save, public :: precipo_l(:),precipn_l(:),precip_l(:)
  type(cmr1), allocatable, save, public :: evapo_l(:), tpe_l(:)


  !- Parameters for turbulence model -------------------------------------------
  type(cmr1), allocatable, save, public :: avv_l(:)
  type(cmr1), allocatable, save, public :: tke_l(:), diss_l(:), hflux_l(:)
  type(cmr1), allocatable, save, public :: z0srf_l(:)
  !  these four to replace former five shear,stretch,div,eddyh,eddyd
  type(cmr1), allocatable, save, public :: sm1_l(:), sm2_l(:), sm3_l(:)
  type(cmr1), allocatable, save, public :: eddyh_l(:), smu_l(:), smv_l(:)


  !- Bottom friction maps ------------------------------------------------------
  type(cmr1), allocatable, save, public :: sbfmap_l(:)


  !- Tide param maps -----------------------------------------------------------
  type(cmr2), allocatable, save, public :: tidemap_l(:)


  !- Arrays for A/D scheme with internal component loop ------------------------
  type(cmr2), allocatable, save, public :: dispv_l(:), cmp_l(:), uvdam_l(:)


  !- Fields for the ice model, iota model --------------------------------------
  type(cml1), allocatable, save, public :: casus_l(:)
  type(cmr1), allocatable, save, public :: tsnei_l(:)
  type(cmr1), allocatable, save, public :: ueis_l(:), veis_l(:)
  type(cmr2), allocatable, save, public :: ice_l(:)
  type(cmr2), allocatable, save, public :: iota_l(:)
  type(cmr1), allocatable, save, public :: iotac(:)

  real(8),    allocatable,       public :: uice_o(:), vice_o(:)
  real(8),    allocatable,       public :: shear(:), stretch(:), div(:)
  real(8),    allocatable,       public :: eddyh(:), eddyd(:)
  logical,    allocatable, save, public :: fastmask(:)

  !- Subroutines/functions -----------------------------------------------------
  public  :: Alloc_Local_Arrays
  private :: Alloc_Local, icedyn_alloc

contains


  !============================================================================+
  !
  !  Allocate arrays 
  !
  !============================================================================+
  subroutine Alloc_Local_Arrays (narea, nthr)

    !- modules -----------------------------------------------------------------
    use dmi_mpi,       only : dd
    use exits,         only : exitme
    use cmod_miscvars, only : ntracers
    use constants,     only : precevap, pen_type, dyn, nicepar, simplebfmap,   &
                              icedynactive, shortcircuit, elevation_ref,       &
                              niotapar, iotamodel, niceact
    use cmod_params,   only : nsoilpar
    use tflow,         only : tflow_alloc
    use iflow,         only : iflow_alloc
    use botfricmaps,   only : get_simple_bf_map
    use tideparam,     only : get_tide_map
    use turbmodels,    only : tidetype
    use masscorrection,only : masscorr_firstentry
 
    !- dircetives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: narea, nthr

    !- local vars --------------------------------------------------------------
    integer(4) :: mnp, nnp, mn, nn, kn, n2d, n3d, ia, n2s, n3s, n2h, n3h, n2d1
    integer(4) :: maxn3s, maxn2s, ierr
    logical    :: flag = .false.


    !- initialise --------------------------------------------------------------
    ierr = 0

    !- Allocate all the pointers -----------------------------------------------
    call Alloc_Local( narea, ierr )
    if (ierr /= 0) call exitme(-1,'Alloc_Local_Arrays failed, Alloc_Local')

    !- Allocate the local arrays -----------------------------------------------
    maxn3s = -1
    maxn2s = -1
    do ia=1,narea

      !- Full-size local arrays ------------------------------------------------
      !
      !   dim nos for the global quantities:
      mn  = dd(ia)%dim1
      nn  = dd(ia)%dim2
      kn  = dd(ia)%dim3
      mnp = mn + 1
      nnp = nn + 1
      n2d = dd(ia)%iwet2
      n3d = dd(ia)%iwet3
      !
      !   allocate full-size local arrays:
      allocate( mm1_l(ia)%p(0:mnp,0:nnp,kn), mm1k_l(ia)%p(kn,0:mnp,0:nnp),     &
                me_l(ia)%p(0:mnp,0:nnp), stat=ierr )
      if (ierr /= 0) call exitme(-1,'Alloc_Local_Arrays failed, mm arrays')

      !- Task-size local arrays ------------------------------------------------
      !
      !   dim nos for the local 2D and 3D quantities:
      n2h = dd(ia)%halo2
      n3h = dd(ia)%halo3
      n3d = dd(ia)%nwet3
      if (n3d == 0) then
        if (dd(ia)%up_ws /= 0 .or. dd(ia)%low_ws /= 0 .or.                     &
            dd(ia)%halo2 /= 0 .or. dd(ia)%halo3  /= 0      )                   &
            call exitme(1,'Cannot figure out how to set up local arrays')
        n2d  = 0 
        n2d1 = 1
      else
        n2d  = dd(ia)%up_ws - dd(ia)%low_ws + 1
        n2d1 = n2d
      endif
      n2s = n2d + n2h
      n3s = n3d + n3h
      maxn3s = max(n3s,maxn3s)
      maxn2s = max(n2s,maxn2s)
      !
      !   allocate task-size local arrays:
      allocate(  z_l(ia)%p(0:n2s), zn_l(ia)%p(0:n2s),                          &
                hz_l(ia)%p(0:n3s), h_l(ia)%p(0:n3s), un_l(ia)%p(0:n3s),        &
                vn_l(ia)%p(0:n3s), ueis_l(ia)%p(0:n2s), veis_l(ia)%p(0:n2s),   &
                u_l(ia)%p( 0:n3s), v_l(ia)%p( 0:n3s), w_l(ia)%p( 0:n3s),       &
                ui_l(ia)%p(0:n3s), vi_l(ia)%p(0:n3s), wi_l(ia)%p(0:n3s),       &
                hu_l(ia)%p(0:n3s), hv_l(ia)%p(0:n3s),                          &
                h_old_l(ia)%p(0:n3s),h_new_l(ia)%p(0:n3s),h_i_l(ia)%p(0:n3s),  &
                taux_l(ia)%p(0:n2s), tauy_l(ia)%p(0:n2s), pl_l(ia)%p(0:n2s),   &
                wua_l(ia)%p(0:n2s), wva_l(ia)%p(0:n2s), pla_l(ia)%p(0:n2s),    &
                wun_l(ia)%p(0:n2s), wvn_l(ia)%p(0:n2s), pln_l(ia)%p(0:n2s),    &
                wu_l(ia)%p(0:n2s), wv_l(ia)%p(0:n2s), atemp_l(ia)%p(0:n2d),    &
                atempa_l(ia)%p(0:n2d), atempn_l(ia)%p(0:n2d),                  &
                humida_l(ia)%p(0:n2d), humidn_l(ia)%p(0:n2d),                  &
                clouda_l(ia)%p(0:n2d), cloudn_l(ia)%p(0:n2d),                  &
                humid_l(ia)%p(0:n2d), cloud_l(ia)%p(0:n2d),                    &
                kh_l(ia)%p(0:n2s), khu_l(ia)%p(0:n2s), khv_l(ia)%p(0:n2s),     &
                ind_l(ia)%p(2,1:max(1,n2s)), z0srf_l(ia)%p(0:n2s),             &
                uvdam_l(ia)%p(1:2,0:n2s), cmp_l(ia)%p(1:ntracers,0:n3s),       &
                rho_l(ia)%p(0:n3s), press_l(ia)%p(0:n3s), avv_l(ia)%p(0:n3s),  &
                ice_l(ia)%p(1:nicepar,0:n2s), tsnei_l(ia)%p(0:n2s),            &
                casus_l(ia)%p(0:n2s), t_soil_l(ia)%p(1:nsoilpar,n2d1),         &
                sm1_l(ia)%p(0:n3s), sm2_l(ia)%p(0:n3s), sm3_l(ia)%p(0:n3s),    &
                eddyh_l(ia)%p(0:n3s), smu_l(ia)%p(0:n3s), smv_l(ia)%p(0:n3s),  &
                stat=ierr )
      if (ierr /= 0) then
        call exitme(-1,'Alloc_Local_Arrays failed, task local arrays')
      endif

      if (iotamodel) then
        allocate(iota_l(ia)%p(1:niotapar,0:n2s), iotac(ia)%p(0:n2s), stat=ierr)
        if (ierr /= 0) then
          call exitme(-1,'Alloc_Local_iota_Arrays failed, task local arrays')
        endif
      else
        allocate(iota_l(ia)%p(1:niotapar,0:0), iotac(ia)%p(0:0), stat=ierr)
      endif
      if (ierr /= 0) then
        call exitme(-1,'Alloc_Local_iota_Arrays failed, task local arrays')
      endif

      !   transfer values:
      iw2_l(ia) = n2d  
      iw3_l(ia) = n3d  
      !
      !- some setup specific allocations: --------------------------------------
      !
      !   define two-equation turbulence model variables:
      allocate(tke_l(ia)%p(0:n3s),diss_l(ia)%p(0:n3s),hflux_l(ia)%p(0:n2d),    &
               stat=ierr)
      if (ierr /= 0) call exitme(-1,'Alloc_Local_Arrays failed, turb arrays')
      !
      !   define tracer diffusivity: -------------------------------------------
      if (ntracers == 2) then 
        allocate(dispv_l(ia)%p(1:2,0:n3s),stat=ierr)
      else
        allocate(dispv_l(ia)%p(1:3,0:n3s),stat=ierr)
      endif
      if (ierr /= 0) call exitme(-1,'Alloc_Local_Arrays failed, diff array')
      !
      !   define vertical divergence of shortwave radiation:
      if (pen_type == 2) then
        allocate(divrad_l(ia)%p(0:n3s), divrad_s_l(ia)%p(0:n2d), stat=ierr)
      elseif (pen_type <= 1) then
        allocate(divrad_l(ia)%p(0:0), divrad_s_l(ia)%p(0:0), stat=ierr)
      endif
      if (ierr /= 0) call exitme(-1,'Alloc_Local_Arrays failed, pen_type')
      !
      !   define arrays for dynamical vertical co-ordinates:
      if (dyn) then
        allocate( ti_l(ia)%p(0:n2s), hn_l(ia)%p(0:n3s), geopot_l(ia)%p(0:n3s), &
                  stat=ierr )
      else
        allocate( ti_l(ia)%p(0:n2d), hn_l(ia)%p(0:0), geopot_l(ia)%p(0:0),     &
                  stat=ierr )
      endif
      if (ierr /= 0) call exitme(-1,'Alloc_Local_Arrays failed, geopot')
      !
      !   Precipitation and evaporation:
      if (precevap) then
        allocate( precipo_l(ia)%p(0:n2d), precipn_l(ia)%p(0:n2d),              &
                  precip_l(ia)%p(0:n2d), evapo_l(ia)%p(0:n2d),                 &
                  tpe_l(ia)%p(0:n2d), stat=ierr )
      else
        allocate( precipo_l(ia)%p(0:0), precipn_l(ia)%p(0:0),                  &
                  precip_l(ia)%p(0:0), evapo_l(ia)%p(0:0), tpe_l(ia)%p(0:0),   &
                  stat=ierr )
      endif
      if (ierr /= 0) call exitme(-1,'Alloc_Local_Arrays failed, precip')
      !
      ! ref. elevation map:
      if (elevation_ref) then
        allocate( elev_ref_l(ia)%p(0:n2s), stat=ierr )
      else
        allocate( elev_ref_l(ia)%p(0:0), stat=ierr )
      endif
      if (ierr /= 0) call exitme(-1,'Alloc_Local_Arrays failed, elev_ref')

    enddo
    !  tflow tmp arrays:
    if (.not.dyn) then
      call masscorr_firstentry(nthr, ierr)
      if (ierr /= 0) call exitme(-1,'Alloc_Local_Arrays failed, mass corr')
    endif
    call tflow_alloc(maxn3s, ntracers, ierr)
    if (ierr /= 0) call exitme(-1,'Alloc_Local_Arrays failed, tflow_alloc')
    !
    !   ice model tmp arrays
    if (icedynactive .and. shortcircuit) then
      call iflow_alloc (maxn2s,niceact,ierr)
      if (ierr /= 0) call exitme(ierr, 'IFLOW_ALLOC: Failed to allocate')
      call icedyn_alloc(maxn2s,ierr)
      if (ierr /= 0) call exitme(ierr, 'ICEDYN_ALLOC: Failed to allocate')
    endif
    !
    ! simple bottom friction map:
    if (simplebfmap) call get_simple_bf_map(narea)
    do ia=1,narea
      if (simplebfmap) call get_simple_bf_map(ia,flag)
      if (flag) then
        n2s = iw2_l(ia)
      else
        n2s = 0
      endif
      allocate(sbfmap_l(ia)%p(0:n2s),stat=ierr)
      if (ierr /= 0) call exitme(-1,'Alloc_Local_Arrays failed, sbfmap')
    enddo
    !
    ! tide param map:
    if (tidetype > 0) call get_tide_map(narea)
    do ia=1,narea
      if (tidetype > 0) call get_tide_map(ia,flag)
      if (flag) then
        n2s = iw2_l(ia)
      else
        n2s = 0
      endif
      allocate(tidemap_l(ia)%p(0:n2s,1:2),stat=ierr)
      if (ierr /= 0) call exitme(-1,'Alloc_Local_Arrays failed, tidemap')
    enddo

  end subroutine Alloc_Local_Arrays




  !============================================================================+
  !
  !  Allocate basic pointers according No. of areas
  !
  !============================================================================+
  subroutine Alloc_Local (n, ierr)

    implicit none

    integer(4), intent(in)    :: n
    integer(4), intent(inout) :: ierr
   
    allocate( mm1_l(n), mm1k_l(n), me_l(n), iw2_l(n), iw3_l(n), hz_l(n),h_l(n),&
              hn_l(n), hu_l(n), hv_l(n), kh_l(n), ind_l(n), z_l(n),            &
              elev_ref_l(n),zn_l(n), un_l(n), vn_l(n), u_l(n), v_l(n), w_l(n), &
              khu_l(n), khv_l(n), h_old_l(n), h_new_l(n), h_i_l(n), ti_l(n),   &
              ueis_l(n), veis_l(n), rho_l(n), press_l(n), cmp_l(n), dispv_l(n),&
              ui_l(n), vi_l(n), wi_l(n), precipo_l(n),precipn_l(n),precip_l(n),&
              evapo_l(n), tpe_l(n), pla_l(n), pln_l(n), pl_l(n), t_soil_l(n),  &
              wua_l(n), wun_l(n), wva_l(n), wvn_l(n), taux_l(n), tauy_l(n),    &
              wu_l(n), wv_l(n), atemp_l(n), humid_l(n), cloud_l(n),            &
              atempa_l(n), atempn_l(n), humida_l(n), humidn_l(n),              &
              clouda_l(n), cloudn_l(n), tke_l(n), diss_l(n), hflux_l(n),       &
              avv_l(n), casus_l(n), ice_l(n), tsnei_l(n), geopot_l(n),         &
              sm1_l(n), sm2_l(n), sm3_l(n), eddyh_l(n), smu_l(n), smv_l(n),    &
              divrad_l(n), divrad_s_l(n), uvdam_l(n), z0srf_l(n), sbfmap_l(n), &
              tidemap_l(n), iota_l(n), iotac(n), stat=ierr )

  end subroutine Alloc_Local

  !-----------------------------------------------------------------------------

  subroutine icedyn_alloc(n2dm, ierr)
    implicit none
    integer(4), intent(in)    :: n2dm
    integer(4), intent(inout) :: ierr

    allocate (uice_o(0:n2dm), vice_o(0:n2dm), fastmask(0:n2dm),                &
              shear(0:n2dm), stretch(0:n2dm), div(0:n2dm),                     &
              eddyh(0:n2dm), eddyd(0:n2dm), stat=ierr )

  end subroutine icedyn_alloc

  !-----------------------------------------------------------------------------

end module cmod_local_arrays
