module cmod_mem
  !=============================================================================
  !
  ! Module for memory allocation in cmod, especially for nested variables.
  ! Assuming shape of arrays is important.
  ! Names of predefined type defs follow the notation:
  !   cmTD   
  ! where cm is short for "Cmod Memory", T is a type identifier
  !   T  = i for integer(4)
  !   T  = r for real(8) 
  !   T  = l for logical 
  ! and D identifies the dimension (shape) of the array
  !   D  = 1 for 1D arrays
  !   D  = 2 for 2D arrays
  !   D  = 3 for 3D arrays
  ! 
  ! Example of usage: 
  ! Define a 3D integer(4) index array for each of N model domains 
  ! when N is specified at run-time, ie it is not known during compilation.
  ! Use the index arrays in subsequent calls of a subroutine requiring four
  ! arguments: the index array and its three dimensions. 
  ! Free memory when done.
  !   
  !   use cmod_mem
  !   interface
  !     subroutine sub( ix, n1, n2, n3 )
  !       integer(4), intent(in)    :: n1, n2, n3
  !       integer(4), intent(inout) :: ix(:,:,:)
  !     end subroutine sub
  !   end interface
  !   type (cmi3), allocatable :: idx(:)
  !   integer(4) :: N, ia
  !   integer(4), allocatable :: dims(:,:)
  !      ... statements getting N
  !   allocate( dims(N,3) )
  !      ... statements getting dimensions dims(:,1:3) in each area
  !   allocate( idx(N) )
  !   do ia=1,N
  !     allocate( idx(ia)%p( 1:dims(ia,1), 1:dims(ia,2), 1:dims(ia,3) ) )
  !   enddo
  !   do ia=1,N
  !     call sub( idx(ia), dims(ia,1), dims(ia,2), dims(ia,3) )
  !   enddo
  !   deallocate( idx )
  !=============================================================================
  implicit none
  private

  type cmi1
    integer(4), allocatable :: p(:)  
  end type

  type cmi2
    integer(4), allocatable :: p(:,:)  
  end type

  type cmi3
    integer(4), allocatable :: p(:,:,:)  
  end type

  type cmr1
    real(8), allocatable :: p(:)  
  end type

  type cmr2
    real(8), allocatable :: p(:,:)  
  end type

  type cmr3
    real(8), allocatable :: p(:,:,:)  
  end type

  type cml1
    logical, allocatable :: p(:)  
  end type

  type cml2
    logical, allocatable :: p(:,:)  
  end type

  type cml3
    logical, allocatable :: p(:,:,:)  
  end type

  public  :: cmi1, cmi2, cmi3, cmr1, cmr2, cmr3, cml1, cml2, cml3

end module cmod_mem
