module cmod_params

  !============================================================================+
  !
  !    obtain basic CMOD parameters
  !
  !============================================================================+

  !- Modules -------------------------------------------------------------------
  use exits,     only : exitme
  use cmod_mem,  only : cmi1, cmi2, cmi3, cmr1

  !- Implicit Directives -------------------------------------------------------
  implicit none
  private

  !- Basic numbers -------------------------------------------------------------
  integer(4), save, public :: narea    ! No. of dynamically two-way nested areas
  integer(4), save, public :: nfjord   ! No. of one-way nested fjord models
  integer(4), save, public :: ntotal   ! Total No. of nested domains
  integer(4), save, public :: maxtimelevels  ! Max No of time levels

  integer(4), parameter, public :: maxareas = 99 ! Max No. of areas
  integer(4), parameter, public :: mainarea = 1  ! No. corresponding to the main
                                              ! area/domain
  logical, save, private   ::  bathy_asc2bin, bathy_bin, bathy_asc2mpibin
  logical, save, private   ::  bathy_mpibin


  !- Config and bathy file names -----------------------------------------------
  character(LEN=256), allocatable, save, public :: cfgfile(:), bathyfile(:)


  !- Time- and nesting-levels --------------------------------------------------
  logical,    allocatable, save, public :: onthislevel(:,:)
  integer(4), allocatable, save, public :: timelevel(:), nestinglevels(:)
  integer(4), allocatable, save, public :: enclosing(:,:), timelevelsize(:)
  type(cmi1), allocatable, save, public :: nestingto(:), nestingfrom(:)
  type(cmi2), allocatable, save, public :: znesting(:), unesting(:), vnesting(:)
  type(cmi2), allocatable, save, public :: znest_serial(:)
  type(cmi2), allocatable, save, public :: unest_serial(:), vnest_serial(:)
!FIXME: arrays need while doing incremental improvements to nesting w/MPI
  type(cmi2), allocatable, save, public :: un_c_mpi(:), vn_c_mpi(:)
  type(cmi2), allocatable, save, public :: un_f_mpi(:), vn_f_mpi(:)


  !- Basic arrays (indices, bathymetries, etc) ---------------------------------
  type (cmi3), allocatable, save, private :: m1(:)
  type (cmi3), allocatable, save, public  :: mm1(:)
  type (cmi2), allocatable, save, public  :: me1(:),krz(:),kfg(:),iga(:),jga(:)
  type (cmr1), allocatable, save, public  :: elev_ref(:)
  type (cmr1), allocatable, save, public  :: hz(:)


  !- Grid dimensions -----------------------------------------------------------
  integer(4), allocatable, save, public :: mmx(:), nmx(:), kmx(:), iw2(:),iw3(:)
  integer(4), allocatable, save, public :: mmxp(:), nmxp(:) ! mxp = mx + 1
  integer(4),              save, public :: kmax, kmaxp      ! kmaxp=kmax+1

  !- Soil parameters (number of layers, height and conductivity of layers) -----
  integer(4),           save, public :: nsoilpar
  real(8), allocatable, save, public :: h_soil(:)
  real(8), allocatable, save, public :: c_soil(:)


  !- Number of boundary points -------------------------------------------------
  !  outer (nz), inner (nu,nv), freshwater (nq)
  integer(4), allocatable, save, public :: nz(:), nu(:), nv(:), nq(:)


  !- Number of dams for each grid ----------------------------------------------
  integer(4), allocatable, save, public :: nud(:), nvd(:)

  !- Number of weirs for each grid ---------------------------------------------
  integer(4), allocatable, save, public :: nweir(:)

  !- Misc parameters of global usage -------------------------------------------
  
  !- OMP params
  integer(4), save, public :: nthreads    ! #threads


  !- Subroutines/functions -----------------------------------------------------
  public  :: GetBasicParams, GetParams, PrintComputationalIndex,               &
             ReadElevationReference
  private :: ReadDataFile, CalcDerivedParams1, CheckParams, PrintParams


contains

  !============================================================================+
  !
  !  Obtain the basic parameters 
  !
  !============================================================================+
  subroutine GetBasicParams()

    !- Modules -----------------------------------------------------------------
    use io_subs,        only : io_new_unit
    use constants,      only : ldebug
    use dmi_mpi_global, only : iu06

    !- local vars --------------------------------------------------------------
    integer(4)         :: lun, ios, ia, timelevel_in, nestinglevels_in
    integer(4)         :: ii, iii, iv, iia, nlevs
    integer(4)         :: nestingto_in(maxareas)
    character(LEN=256) :: cfgfile_in, bathyfile_in
    character(LEN=1)   :: c1
    character(LEN=2)   :: c2
    character(LEN=10)  :: cfmt

    !- namelists ---------------------------------------------------------------
    namelist /cfglist/ narea, nfjord, bathy_asc2bin, bathy_asc2mpibin,         &
                       bathy_bin, bathy_mpibin
    namelist /cfgfn/   cfgfile_in, bathyfile_in, timelevel_in,                 &
                       nestinglevels_in, nestingto_in
    
    !- Default initialisations -------------------------------------------------
    narea  = 2
    nfjord = 0
    bathy_asc2bin=.false.
    bathy_asc2mpibin=.false.
    bathy_bin=.false.
    bathy_mpibin=.false.
    
    !- Use configuration file if present ---------------------------------------
    ! FIXME: hard coded file name should be input at a later stage ...
    lun = io_new_unit()
    open (unit=lun, file='cfg.nml', status='old', iostat=ios)
    if (ios /= 0) then
      !  apply default settings:
      write(iu06,'(a64)')                                                      &
        'Warning: File cfg.nml not found. Continue with default settings'

      !  total No. of domains:
      ntotal = narea
      
      allocate( cfgfile(ntotal), bathyfile(ntotal), timelevel(ntotal),         &
                nestinglevels(ntotal), nestingto(ntotal), timelevelsize(ntotal))
      allocate( nestingto(1)%p(1), nestingto(2)%p(1) )
      cfgfile(1)        = 'data_coarse'
      cfgfile(2)        = 'data_fine'
      bathyfile(1)      = 'coarse_grid_depth'
      bathyfile(2)      = 'fine_grid_depth'
      timelevel(1)      = 1
      timelevel(2)      = 2
      nestinglevels(1)  = 1
      nestinglevels(2)  = 0
      nestingto(1)%p(1) = 2
      nestingto(2)%p(1) = 0
    else
      !  apply settings from cfg file:
      read (unit=lun, nml=cfglist, iostat=ios)
      if (ios /= 0) then
        write(iu06,*) 'Error: Cannot read basic numbers from cfg file ... exit'
        write(iu06,*) 'Error: The iostat failure was:: ', ios
        call exitme(1)
      endif
      if (narea <= 0 .or. nfjord < 0) then
        write(iu06,*) 'Error: Invalid number of areas ... exit'
        call exitme(1)
      endif
      ! fjords are not implemented.
      if (nfjord /= 0) then
        write(iu06,*) 'Error: Fjord module not implemented yet ... exit'
        call exitme(1)
      endif

      !  total No. of domains:
      ntotal = narea + nfjord
      if (ntotal > maxareas) then
        write(iu06,*) 'Error: Total No. of areas cannot exceed ', maxareas
        write(iu06,*) '       Increase maxareas and re-compile ... exit'
        call exitme(1)
      endif

      !  allocate and get file names, time levels, etc:
      allocate( cfgfile(ntotal), bathyfile(ntotal), timelevel(ntotal),         &
                nestinglevels(ntotal), nestingto(ntotal), timelevelsize(ntotal))
      do ia=1,ntotal
        cfgfile_in       = ' '
        bathyfile_in     = ' '
        timelevel_in     = 1
        if (ia == ntotal) then
          nestinglevels_in = 0
          nestingto_in(:)  = 0
        else
          nestinglevels_in = 1
          nestingto_in(:)  = min( ia+1, ntotal )
        endif
        read (unit=lun, nml=cfgfn, iostat=ios)
        if ( ios /= 0 ) then
          write(iu06,*) 'Error: Cannot read file names from cfg file ... exit'
          write(iu06,*) 'Error: The iostat failure was:: ', ios
          call exitme(1)
        else
          cfgfile(ia)   = trim(cfgfile_in)
          bathyfile(ia) = trim(bathyfile_in)
          timelevel(ia) = timelevel_in
          nlevs         = nestinglevels_in
          if (nlevs < 0 .or. nlevs > maxareas) then
            write(iu06,*)                                                      &
                        'Error: Invalid No. of levels for area ',ia,' ... exit'
            call exitme(1)
          endif
          nestinglevels(ia) = nlevs

          if (nlevs > 0) then
            allocate( nestingto(ia)%p(nlevs) )
            nestingto(ia)%p(1:nlevs) = nestingto_in(1:nlevs)
          endif

          if (ia == mainarea .and. timelevel(ia) /= 1) then
            write(iu06,*) 'Error: Time level of main area must be 1 ... exit'
            call exitme(1)
          elseif (timelevel(ia) < 1) then
            write(iu06,*)                                                      &
                      'Error: Time level must be a positive integer(4) ... exit'
            call exitme(1)
          endif

          if (ia == mainarea .and. nestinglevels(ia) < 1 .and. ntotal > 1) then
            write(iu06,'(a55)')                                                &
                      'Warning: No nesting into main area ... continue wihtout'
          endif

        endif 
      enddo

      ! done getting configuration params; close unit
      close(unit=lun)

      ! validate configuration:
      do ia=1,ntotal
        do nlevs=1,nestinglevels(ia)
          if (nestingto(ia)%p(nlevs) == ia    .or.                             &
              nestingto(ia)%p(nlevs) <= 0     .or.                             &
              nestingto(ia)%p(nlevs) > ntotal      ) then
            write(iu06,*) 'Error: Invalid nesting for area ',ia,' ... exit'
            call exitme(1)
          endif

          if (timelevel(ia) > timelevel(nestingto(ia)%p(nlevs))) then
            write(iu06,*) 'Error: Cannot nest to domains on lower time levels'
            call exitme(1)
          endif
        enddo
      enddo
    endif
    

    !- How many time nesting levels? -------------------------------------------
    maxtimelevels = maxval(timelevel(:))


    !- some doc on nesting arrays ----------------------------------------------
    !  
    ! The three arrays timelevel(:), nestinglevels(:) and nestingto(:)%p(:) are
    ! initialised according to user-specs. They should now have been 
    ! "sanity-checked" to avoid the most obvious user-spec errors.
    !
    ! timelevel(ia) : identifies the time step nesting level for area #ia; it 
    !                 must be 1 for the coarsed grid (mainarea) with the largest
    !                 time step and for those domains on the same time-level. A
    !                 higher value mean higher time-nesting level (i.e. smaller
    !                 time step).
    ! nestinglevels(ia) : gives the number of immediately inclosed nesting areas
    !                     to area #ia, i.e. how many domains does area #ia nest
    !                     to. It should be zero for domains on the highest 
    !                     nesting level.
    ! nestingto(ia)%p(nl) : for nl=1,nestinglevels(ia) this gives the number of
    !                       the area(s) that area #ia is nesting to. 
    !
    !---------------------------------------------------------------------------


    !- Allocate the most basic stuff -------------------------------------------
    allocate( mmx(ntotal), nmx(ntotal), kmx(ntotal), iw2(ntotal), iw3(ntotal), &
              mmxp(ntotal), nmxp(ntotal),                                      &
              nz(ntotal), nu(ntotal), nv(ntotal), nq(ntotal),                  &
              nud(ntotal), nvd(ntotal),nweir(ntotal),                          &
              m1(ntotal), mm1(ntotal), me1(ntotal),elev_ref(ntotal),           &
              hz(ntotal), krz(ntotal), kfg(ntotal), iga(ntotal), jga(ntotal),  &
              nestingfrom(ntotal), unesting(ntotal), vnesting(ntotal),         &
              znesting(ntotal), unest_serial(ntotal), vnest_serial(ntotal),    &
              znest_serial(ntotal), enclosing(ntotal,ntotal),                  &
              onthislevel(ntotal,maxtimelevels),                               &
              un_c_mpi(ntotal), vn_c_mpi(ntotal),                              &
              un_f_mpi(ntotal), vn_f_mpi(ntotal), stat=ios)
    if (ios /= 0) then
      write(iu06,*) 'Error: Failed to allocate cmod_params, status was: ', ios
      call exitme(1)
    endif


    !- Set the nesting fine-->coarse -------------------------------------------
    !    is area #ia on time level #itl:                 onthislevel(ia,itl)
    !    order of area #iia enclosing area #ia:          enclosing(ia,iia)
    !    No. of areas nesting from area #ia:             nestingfrom(ia)%p(0)
    !    areas nesting from area #ia in given order:     nestingfrom(ia)%p(1:)
    !    i.e. we have:
    !      nestingfrom(ia)%p(enclosing(ia,iia)) = iia
    !---------------------------------------------------------------------------
    enclosing(:,:) = 0
    allocate( nestingfrom(mainarea)%p(0:0) )
    nestingfrom(mainarea)%p(0) = 0
    do ia=2,ntotal 
      ii = 0
      do iia=1,ntotal
        if (iia /= ia .and. nestinglevels(iia) > 0) then
          do iii=1,nestinglevels(iia)
            if (nestingto(iia)%p(iii) == ia) ii = ii + 1
          enddo
        endif
      enddo
      allocate( nestingfrom(ia)%p(0:ii) )
      nestingfrom(ia)%p(0) = ii
      if (ii > 0) then
        iv = 0
        do iia=1,ntotal
          if (iia /= ia .and. nestinglevels(iia) > 0) then
            do iii=1,nestinglevels(iia)
              if (nestingto(iia)%p(iii) == ia) then
                iv = iv + 1
                nestingfrom(ia)%p(iv) = iia
                enclosing(ia,iia)     = iv
              endif
            enddo
          endif
        enddo
        if (iv /= ii) then
          write(iu06,*)                                                        &
                      'Error: Cannot figure out how to do the nesting ... exit'
          call exitme(1)
        endif
      endif
    enddo

    !- set up time level vs domain table ---------------------------------------
    onthislevel(:,:) = .false.
    do ia=1,ntotal
      onthislevel(ia,timelevel(ia)) = .true.
    enddo
     
    ! possibly do extented printouts -------------------------------------------
    if (ldebug) then
      write(iu06,'(a1)' ) ' '
      write(iu06,'(a1)' ) ' '
      write(iu06,'(a4)' ) 'Area'
      write(iu06,'(a32)') 'sub level, domain, on time level'
      do ia=1,ntotal
        do iv=1,nestinglevels(ia)
          write(iu06,'( i5)') ia
          write(iu06,'(3i5)') iv, nestingto(ia)%p(iv),                         &
                              timelevel(nestingto(ia)%p(iv))
          write(iu06,'( a1)') ' '
        enddo
      enddo
      do ia=1,narea
        write(iu06,'(a6,i5)') 'Area: ', ia
        if (nestinglevels(ia) > 0) then
          if (nestinglevels(ia) < 10) then
            write(c1,'(i1)') nestinglevels(ia)
            cfmt = '(a15, '//c1//'i5)'
          elseif (nestinglevels(ia) < 100) then
            write(c2,'(i2)') nestinglevels(ia)
            cfmt = '(a15,'//c2//'i5)'
          else
            call exitme(999, 'invalid number of nesting levels')
          endif
          write(iu06,cfmt)       '  nesting to:  ',                            &
                                 nestingto(ia)%p(1:nestinglevels(ia))
        else
          write(iu06,'(a32)') '  not nesting to any sub domain.'
        endif
        if (nestingfrom(ia)%p(0) > 0) then
          if (nestingfrom(ia)%p(0) < 10) then
            write(c1,'(i1)') nestingfrom(ia)%p(0)
            cfmt = '(a15, '//c1//'i5)'
          elseif (nestingfrom(ia)%p(0) < 100) then
            write(c2,'(i2)') nestingfrom(ia)%p(0)
            cfmt = '(a15,'//c2//'i5)'
          else
            call exitme(999, 'invalid number of nesting levels')
          endif
          write(iu06,cfmt)                                                     &
                    '  nesting from:',nestingfrom(ia)%p(1:nestingfrom(ia)%p(0))
        else
          write(iu06,'(a40)') '  not nesting from any enclosing domain.'
        endif
      enddo
      write(iu06,'(a1 )') ' '
      write(iu06,'(a17)') 'Loop inside-->out'
      write(iu06,'(a25)') '  fine grid, coarse grid '
      do iv=maxval(timelevel(1:)),1,-1
        do ia=narea,1,-1
          if (.not.onthislevel(ia,iv)) cycle
          do iii=1,nestingfrom(ia)%p(0)
            write(iu06,'(2i5)') ia, nestingfrom(ia)%p(iii)
          enddo
        enddo
      enddo
      write(iu06,'(a1)') ' '
      write(iu06,'(a1)') ' '
    endif

    ! Set parameters for soil-model or (optionally) read these parameters 
    ! from file data_soil
    ! FIXME: hard coded file name should be input at a later stage ...
    lun = io_new_unit()
    open (unit=lun, file='data_soil', status='old', iostat=ios)
    if(ios /= 0) then
      nsoilpar = 3
      allocate( h_soil(nsoilpar), c_soil(nsoilpar) )
      h_soil =  (/ 0.1_8, 0.5_8, 2.5_8 /)
      c_soil =  (/ 2.0_8, 1.5_8, 1.0_8 /)
    else
      read(lun,'(1X)') 
      read(lun,'(24X,i3)') nsoilpar
      read(lun,'(1X)') 
      read(lun,'(1X)') 
      read(lun,'(1X)')
      allocate( h_soil(nsoilpar), c_soil(nsoilpar) )
      do ii = 1,nsoilpar
        read(lun,'(9X,2f7.1)') h_soil(ii), c_soil(ii)
      enddo
      write(iu06,'(a50)')  'Read soil-parameters from file data_soil          '
      write(iu06,'(a19,i3,a11)') ' Soil model has now', nsoilpar, ' layers    '
    endif

  end subroutine GetBasicParams


  !============================================================================+
  !
  !  Obtain the basic parameters from bathymetry file and specs files
  !
  !============================================================================+
  subroutine GetParams(area)

    !- Modules -----------------------------------------------------------------
!    use mpi
    use constants,      only : zero, elevation_ref
    use io_subs,        only : io_new_unit
    use dmi_mpi,        only : bathy2mpibin, bathyread_mm, bathyread_hz
    use dmi_mpi_global, only : iu06
    use cmod_timer,     only : timer, timer_print

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: area


    !- local vars --------------------------------------------------------------
    character(256) :: fnam, fnambin
    integer(4)     :: lun, ia, ios !, lunbin
    logical        :: OK
!    integer(4)     :: ierr, fh
    integer(4)     :: fh

    !- short-hand notation -----------------------------------------------------
    ia = area              ! actual area No.

    
    !- Get grid parameters -----------------------------------------------------
    call ReadDataFile (cfgfile(ia), nmx(ia), mmx(ia), kmx(ia),                 &
                       nz(ia), nu(ia), nv(ia), nq(ia), nud(ia), nvd(ia),       &
                       nweir(ia), krz(ia)%p)
    

    !- Check some of the params for sanity -------------------------------------
    call CheckParams (nmx(ia), mmx(ia), kmx(ia),                               &
                      nz(ia), nu(ia), nv(ia), nq(ia),nud(ia),nvd(ia),nweir(ia),&
                      krz(ia)%p, ia, OK)
   
    
    !- Obtain basic index arrays, No. of wet points, and depths ----------------
    call timer(1,'io_bathy')
    fnam = trim(bathyfile(ia))
    allocate( mm1(ia)%p(0:mmx(ia)+1,0:nmx(ia)+1,kmx(ia)), stat=ios )
    if (ios /= 0) then
      write(iu06,*) 'Error: Allocate failure was:: ', ios
      call exitme(1)
    endif
    if (bathy_bin) then
      lun = io_new_unit()
      open(lun,file=trim(fnam),form='unformatted',status='old',iostat=ios)
      if (ios /= 0) then
        write(iu06,*) 'Error: Cannot open '//trim(fnam)//' ... exit'
        write(iu06,*) 'Error: The iostat failure was:: ', ios
        call exitme(1)
      endif
      read(lun,iostat=ios) mm1(ia)%p
      if (ios /= 0) then
        write(iu06,*) 'Error: read mm1 from binary file '//trim(fnam)//' exit'
        write(iu06,*) 'Error: The iostat failure was:: ', ios
        call exitme(1)
      endif
      iw3(ia) = maxval(mm1(ia)%p)
      iw2(ia) = maxval(mm1(ia)%p(:,:,1))
      call PrintParams (iw2(ia), iw3(ia), OK)

      !- Obtain the 3D grid's layer thicknesses --------------------------------
      allocate( hz(ia)%p(0:iw3(ia)) )
      read(lun,iostat=ios) hz(ia)%p
      if (ios /= 0) then
        write(iu06,*) 'Error: read hz from binary file '//trim(fnam)//' exit'
        write(iu06,*) 'Error: The iostat failure was:: ', ios
        call exitme(1)
      endif
      close(lun)
    else if (bathy_mpibin) then
!FIXME: no_mpi is dead, what to do here ?  Marked with "!XX"
!XX   if (no_mpi) then
!XX     write(iu06,*)'Error: Tried to read mpi binary bathy file in serial mode'
!XX     call exitme(1)
!XX   else
        call bathyread_mm(fnam,mm1(ia)%p,(mmx(ia)+2)*(nmx(ia)+2)*kmx(ia),fh)
        iw3(ia) = maxval(mm1(ia)%p)
        iw2(ia) = maxval(mm1(ia)%p(:,:,1))
        call PrintParams (iw2(ia), iw3(ia), OK)
        allocate( hz(ia)%p(0:iw3(ia)) )
        call bathyread_hz(fh,hz(ia)%p,iw3(ia))

!      call MPI_FILE_OPEN(MPI_COMM_WORLD, fnam, MPI_MODE_RDONLY,               &
!                         MPI_INFO_NULL, fh, ierr)
!      call MPI_FILE_READ_ALL(fh,mm1(ia)%p(0:,0:,:),                           &
!          (mmx(ia)+2)*(nmx(ia)+2)*kmx(ia),MPI_INTEGER4,MPI_STATUS_IGNORE, ierr)
!      iw3(ia) = maxval(mm1(ia)%p)
!      iw2(ia) = maxval(mm1(ia)%p(:,:,1))
!      call PrintParams (iw2(ia), iw3(ia), OK)
!      allocate( hz(ia)%p(0:iw3(ia)) )
!      call MPI_FILE_READ_ALL(fh,hz(ia)%p(:), iw3(ia)+1,MPI_REAL8,             &
!           MPI_STATUS_IGNORE, ierr)
!      call MPI_FILE_CLOSE(fh,ierr)
!XX   endif
    else
      ! Default is ascii input
      lun = io_new_unit()
      open(lun,file=trim(fnam),status='old',iostat=ios)
      if (ios /= 0) then
        write(iu06,*) 'Error: Cannot open '//trim(fnam)//' ... exit'
        write(iu06,*) 'Error: The iostat failure was:: ', ios
        call exitme(1)
      endif
      allocate( m1(ia)%p(mmx(ia),nmx(ia),kmx(ia)))
      if (ios /= 0) then
        write(iu06,*) 'Error: Allocate failure was:: ', ios
        call exitme(1)
      endif
      read(lun,*,iostat=ios) m1(ia)%p
      if (ios /= 0) then
        write(iu06,*) 'Error: Tried to read index array - it failed'
        write(iu06,*) 'Error: Cannot read '//trim(fnam)//' ... exit'
        write(iu06,*) 'Error: The iostat failure was:: ', ios
        call exitme(1)
      endif
      mm1(ia)%p(0:,0:,1:) = 0
      mm1(ia)%p(1:mmx(ia),1:nmx(ia),1:kmx(ia)) =                               &
       m1(ia)%p(1:mmx(ia),1:nmx(ia),1:kmx(ia))  
      !- Find and validate the Nos. of wet points in total and in surface ------
      iw3(ia) = maxval(m1(ia)%p)
      iw2(ia) = maxval(m1(ia)%p(:,:,1))
      call PrintParams (iw2(ia), iw3(ia), OK)

      !- Obtain the 3D grid's layer thicknesses --------------------------------
      allocate( hz(ia)%p(0:iw3(ia)) )
      read(lun,'(15f8.2)',iostat=ios) hz(ia)%p
      if (ios > 0) then
        write(iu06,*) 'Error: Tried to read hz array - it failed'
        write(iu06,*) 'Error: Cannot read '//trim(fnam)//' ... exit'
        write(iu06,*) 'Error: The iostat failure was:: ', ios
        call exitme(1)
      endif
      close(lun)
    endif
    call timer(2,'io_bathy')
    call timer_print('TS - IO read bathymetry took: ','io_bathy',100)

    if (bathy_asc2bin) then
!XX   if (no_mpi) then
!XX     lunbin = io_new_unit()
!XX     fnambin = trim(fnam)//'.bin'
!XX     open(lunbin,file=trim(fnambin),form='unformatted',                     &
!XX          status='new',iostat=ios)
!XX     if (ios > 0) then
!XX       write(iu06,*) 'Error: Tried to open binary bathy file - it failed'
!XX       call exitme(1)
!XX     endif
!XX     write(lunbin) mm1(ia)%p(0:mmx(ia)+1,0:nmx(ia)+1,1:kmx(ia)) 
!XX     write(lunbin) hz(ia)%p(0:iw3(ia))
!XX     close(lunbin)
!XX     write(iu06,*) 'Bathymetry files converted to binary'
!XX   else
        write(iu06,*) 'Bathymetry file conversion must be done serial'
        call exitme(1)
!XX   endif
    endif
    if (bathy_asc2mpibin) then
      fnambin = trim(fnam)//'.mpibin'
!XX   if (no_mpi) then
!XX     write(iu06,*) 'Error: Tried to convert ascii to mpi binary bathy file'
!XX     write(iu06,*) 'in serial mode - you must do it with MPI enabled'
!XX     call exitme(1)
!XX   else
        call bathy2mpibin(fnambin,mm1(ia)%p,(mmx(ia)+2)*(nmx(ia)+2)*           &
                            kmx(ia), hz(ia)%p,iw3(ia)+1)
      endif
!XX endif

    !- Allocate the 2D grid's elevation reference and set it to zero -----------
    if (elevation_ref) then
      allocate( elev_ref(ia)%p(0:iw2(ia)) )
    else
      allocate( elev_ref(ia)%p(0:0) )
    endif
    elev_ref(ia)%p(:) = zero

    !- Allocate and dummy-initialise these nesting arrays ----------------------
    allocate( kfg(ia)%p(1:kmx(ia),1:max(nestinglevels(ia),1)),                 &
              iga(ia)%p(1:3,1:max(nestingfrom(ia)%p(0),1)),                    &
              jga(ia)%p(1:3,1:max(nestingfrom(ia)%p(0),1)),                    &
              znesting(ia)%p(max(nestingfrom(ia)%p(0),1),2),                   &
              unesting(ia)%p(max(nestinglevels(ia),1),2),                      &
              vnesting(ia)%p(max(nestinglevels(ia),1),2),                      &
              un_c_mpi(ia)%p(max(nestinglevels(ia),1),2),                      &
              vn_c_mpi(ia)%p(max(nestinglevels(ia),1),2),                      &
              un_f_mpi(ia)%p(max(nestinglevels(ia),1),2),                      &
              vn_f_mpi(ia)%p(max(nestinglevels(ia),1),2),                      &
              znest_serial(ia)%p(max(nestingfrom(ia)%p(0),1),2),               &
              unest_serial(ia)%p(max(nestinglevels(ia),1),2),                  &
              vnest_serial(ia)%p(max(nestinglevels(ia),1),2) )
    kfg(ia)%p(:,:)      = 0
    iga(ia)%p(:,:)      = 0
    jga(ia)%p(:,:)      = 0
    znesting(ia)%p(:,:) = 0
    unesting(ia)%p(:,:) = 0
    vnesting(ia)%p(:,:) = 0
    un_c_mpi(ia)%p(:,:) = 0
    vn_c_mpi(ia)%p(:,:) = 0
    un_f_mpi(ia)%p(:,:) = 0
    vn_f_mpi(ia)%p(:,:) = 0
    znest_serial(ia)%p(:,:) = 0
    unest_serial(ia)%p(:,:) = 0
    vnest_serial(ia)%p(:,:) = 0


    !- Calculate derived parameters --------------------------------------------
    call CalcDerivedParams1 (ia)


    !- Obtain 2D index arrays --------------------------------------------------
    allocate( me1(ia)%p(0:mmxp(ia),0:nmxp(ia)) )
    if (bathy_bin) then
      me1(ia)%p(0:mmx(ia),0:nmx(ia)) = mm1(ia)%p(0:mmx(ia),0:nmx(ia),1)
    else if (bathy_mpibin) then
      me1(ia)%p(0:mmx(ia),0:nmx(ia)) = mm1(ia)%p(0:mmx(ia),0:nmx(ia),1)
    else
      me1(ia)%p(0:,0:) = 0
      me1(ia)%p(1:mmx(ia),1:nmx(ia)) = m1(ia)%p(1:mmx(ia),1:nmx(ia),1)
      deallocate ( m1(ia)%p ) ! FIXME - deallocate m1(ntotal)
    endif

  end subroutine GetParams


  !============================================================================+
  !
  !  Calculate derived parameters, part 1 
  !
  !============================================================================+
  subroutine CalcDerivedParams1 (ia)
    implicit none

    integer(4), intent(in) :: ia

    logical, save :: FirstEntry = .true.

    mmxp(ia) = mmx(ia) + 1
    nmxp(ia) = nmx(ia) + 1

    if (FirstEntry) then
      FirstEntry = .false.
      kmax = -1
    endif
    kmax  = max(kmax,kmx(ia)) + 1
    kmaxp = kmax + 1

  end subroutine CalcDerivedParams1



  !============================================================================+
  !
  !  Read parameters from data file
  !
  !============================================================================+
  subroutine ReadDataFile (fn, n, m, k, nz, nu, nv, nq, nud, nvd, nweir, krz)

    !- Modules -----------------------------------------------------------------
    use constants, only : weirs
    use io_subs, only   : io_new_unit

    !- Directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    character(*),               intent(in)  :: fn
    integer(4),                 intent(out) :: n, m, k
    integer(4), optional,       intent(out) :: nz, nu, nv, nq, nud, nvd, nweir
    integer(4), optional, allocatable       :: krz(:,:)


    !- local vars --------------------------------------------------------------
    character(3)   :: ch3
    integer(4)     :: i, lun, ios


    !- Read parameters from data file ------------------------------------------
    lun = io_new_unit()
    open(lun,file=trim(fn),status='old',iostat=ios)
    if (ios /= 0) call exitme(1,'Could not open '//trim(fn))
    read(lun,'(2X,A3)',iostat=ios) ch3
    if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
    do while (ch3 /= 'DIM')
      read(lun,'(2X,A3)',iostat=ios) ch3
      if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
    enddo
    read(lun,'(14X,I4)',iostat=ios) n
    if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
    read(lun,'(14X,I4)',iostat=ios) m
    if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
    do while (ch3 /= 'SCH')
      read(lun,'(2X,A3)',iostat=ios) ch3
      if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
    enddo
    k=0
    read(lun,'(5X,A3)',iostat=ios) ch3
    if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
    do while (ch3 == 'SCH')
      k=k+1
      read(lun,'(5X,A3)',iostat=ios) ch3
      if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
    enddo


    !- Skip the rest if not required -------------------------------------------
    if (present(nz)) then
      read(lun,'(2X,A3)',iostat=ios) ch3
      if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
      do while (ch3 /= '===')
        read(lun,'(2X,A3)',iostat=ios) ch3
        if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
      enddo
      read(lun,'(27X,I4)',iostat=ios) nz
      if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
      read(lun,'(a)',iostat=ios)
      if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
      read(lun,'(a)',iostat=ios)
      if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))

      !- check for sanity 
      if (nz < 0) call exitme(1,'Invalid number of bnd points. Must be >= 0')

      !- allocate and read krz
      allocate( krz(1:3,0:nz) )
      krz(:,0) = 0
      do i=1,nz
        read(lun,'(40X,3I5)',iostat=ios) krz(1,i), krz(2,i), krz(3,i)
        if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
      enddo

      read(lun,'(a)',iostat=ios)
      if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
      read(lun,'(27X,I4)',iostat=ios) nu
      if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
      do i=1,nu+3
        read(lun,'(a)',iostat=ios)
        if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
      enddo
      read(lun,'(27X,I4)',iostat=ios) nv
      if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
      do i=1,nv+3
        read(lun,'(a)',iostat=ios)
        if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
      enddo
      read(lun,'(27X,I4)',iostat=ios) nq
      if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
      do i=1,nq+3
        read(lun,'(a)',iostat=ios)
        if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
      enddo
      read(lun,'(27X,I4)',iostat=ios) nud
      if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
      do i=1,nud+3
        read(lun,'(a)',iostat=ios)
        if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
      enddo
      read(lun,'(27X,I4)',iostat=ios) nvd
      if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
      
      if (weirs) then
        do i=1,nvd+3
          read(lun,'(a)',iostat=ios)
          if (ios /= 0) call exitme(1,'Error while reading '//trim(fn))
        enddo
        read(lun,'(27X,I4)',iostat=ios) nweir
      else
        nweir = 0
      endif

    endif


    !- Close file and return ---------------------------------------------------
    close(lun)


  end subroutine ReadDataFile

  !============================================================================+
  !
  !  Check some parameters from data file
  !
  !============================================================================+
  subroutine CheckParams (n, m, k, nz, nu, nv, nq, nud, nvd, nweir,krz, ia, OK)

    use constants, only      : weirs
    use dmi_mpi_global, only : iu06

    !- Directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4),           intent(in)  :: n, m, k, ia
    integer(4), optional, intent(in)  :: nz, nu, nv, nq, nud, nvd, nweir
    integer(4), optional, intent(in)  :: krz(:,0:)
    logical,              intent(out) :: OK

    !- local vars --------------------------------------------------------------
    integer(4) :: i 

    !- Find oout if things are OK; assume not on entry -------------------------
    OK = .false.

    !- echo dimensions ---------------------------------------------------------
    write(iu06,'(a28,i5)') 'Checking dimensions of area:', ia
    write(iu06,'(a28,i5)') '  points in N/S direction:  ', m
    write(iu06,'(a28,i5)') '  points in W/E direction:  ', n
    write(iu06,'(a28,i5)') '  layers:                   ', k
    write(iu06,'(a28,i5)') '  z-boundary points:        ', nz
    write(iu06,'(a28,i5)') '  u-boundary points:        ', nu
    write(iu06,'(a28,i5)') '  v-boundary points:        ', nv
    write(iu06,'(a28,i5)') '  sources:                  ', nq
    write(iu06,'(a28,i5)') '  u-dams:                   ', nud
    write(iu06,'(a28,i5)') '  v-dams:                   ', nvd
    if (weirs) then 
      write(iu06,'(a28,i5)') '  weirs:                    ', nweir
    endif

    !- do the checking ---------------------------------------------------------
    !  (nz already checked just after read)
    if (m < 3) call exitme(1, 'Invalid 1st dimension. Must be >= 3')
    if (n < 3) call exitme(1, 'Invalid 2nd dimension. Must be >= 3')
    if (k < 1) call exitme(1, 'Invalid 3rd dimension. Must be >= 1')

    if (nu < 0) call exitme(1, 'Invalid dimension of u-bnd. Must be >= 0')
    if (nv < 0) call exitme(1, 'Invalid dimension of v-bnd. Must be >= 0')
    if (nq < 0) call exitme(1, 'Invalid number of sources. Must be >= 0')
    if (nud < 0) call exitme(1, 'Invalid number of u-dams. Must be >= 0')
    if (nvd < 0) call exitme(1, 'Invalid number of v-dams. Must be >= 0')
    if (weirs) then 
      if (nweir < 0) call exitme(1, 'Invalid number of weirs. Must be >= 0')
    endif

    do i=1,nz
      if (krz(1,i) < 1 .or. krz(1,i) > m .or.                                  &
          krz(2,i) < 1 .or. krz(2,i) > n .or.                                  &
          krz(3,i) < 1 .or. krz(3,i) > 4      ) then
         call exitme(1, 'Invalid specification of boundary')
      endif
    enddo

    !- on success, print that later --------------------------------------------
    OK = .true.

  end subroutine CheckParams


  !============================================================================+
  !
  !  Print some parameters 
  !
  !============================================================================+
  subroutine PrintParams (n2d, n3d, OK)
    use dmi_mpi_global, only : iu06

    !- Directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: n2d, n3d
    logical, intent(in)    :: OK

    !- local vars --------------------------------------------------------------
    !  none, so far

    !- echo parameters ---------------------------------------------------------
    write(iu06,*) '  surface wet points:       ', n2d
    write(iu06,*) '  wet points:               ', n3d

    !- Do some sanity checking -------------------------------------------------
    if (n3d <= 0)  call exitme(1, 'Invalid number of wet points')
    if (n2d <= 0)  call exitme(1, 'Invalid number of surface wet points')
    if (n2d > n3d) call exitme(1, 'Invalid model setup (hmm?)')

    !- on success, print that --------------------------------------------------
    if (OK) then
      write(iu06,'(a23)') '... dimensions seems OK'
    else
      call exitme(1, 'Some kind of invalid dimension was found')
    endif

  end subroutine PrintParams

  !============================================================================+
  subroutine PrintComputationalIndex (narea, iw3, dt, irmaxa)
    use constants,      only : zero
    use dmi_mpi_global, only : iu06

    !- Directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)  :: narea, iw3(:)
    real(8),    intent(in)  :: dt(:)
    integer(4), intent(out) :: irmaxa

    !- local vars --------------------------------------------------------------
    integer(4) :: ia
    real(8)    :: I, Ii, Im

    !- Reference Computational Intensity (test03) ------------------------------
    real(8), parameter :: I1 = (200119.0_8/90.0_8 + 374801.0_8/45.0_8)
    real(8), parameter :: tenth = 0.1_8

    !- Computational Intensity for this case -----------------------------------
    I  = zero
    Im = zero
    irmaxa = 0
    do ia=1,narea
      Ii = real(iw3(ia),8)/dt(ia)
      if (Ii > Im) then
        Im = Ii
        irmaxa = ia
      endif
      I  = I + Ii
    enddo
    I = I/I1
    if (I > tenth) then
      write(iu06,'(a24,f8.1)') 'Computational Intensity:', I 
    else
      write(iu06,'(a24,f8.4)') 'Computational Intensity:', I 
    endif

  end subroutine PrintComputationalIndex


  !============================================================================+
 
  subroutine ReadElevationReference(  )  
 
    !- Modules -----------------------------------------------------------------
    use io_subs,        only : io_new_unit

    !- arguments ---------------------------------------------------------------

    !- local vars --------------------------------------------------------------
    integer(4) :: lun, ios, ia

    lun = io_new_unit()
    open(lun,file='elevation_reference',status='old',iostat=ios)
    if (ios /= 0) call exitme(1,'Could not open elevation_referene file')
    
    do ia = 1,narea
      read(lun,'(15f8.2)',iostat=ios) elev_ref(ia)%p   
      if (ios /= 0) call exitme(1,'Could not read elevation_referene file')
    enddo

    close(lun)
 
  end subroutine ReadElevationReference

end module cmod_params

