module cmod_precip
!-------------------------------------------------------------------------------
! module containing methods for handling precipitation/evaporation
!
!   mass_precevap:   add (P-E) to the rhs of the mass equation
!
!   save_old_sfc:    save old concentration to be used as the concentration of
!                    the precipitation and evaporation (divided by layer
!                    thickness h)
!
!   tracer_precevap: add (C_p*P - C_e*E)/h to the rhs of the C equation
!-------------------------------------------------------------------------------
  implicit none
  private

  public :: mass_precevap, save_old_sfc, tracer_precevap
  
contains

  !-----------------------------------------------------------------------------

  subroutine mass_precevap( iw2,dt,kh,idx,zn,wn,prec,evap,h ) 
    !- modules:
    use dmi_omp,   only : domp_get_domain
 
    !- directives:
    implicit none
 
    !- arguments:
    integer(4), intent(in)          :: iw2
    real(8), intent(in)             :: dt
    integer(4), intent(in)          :: kh(0:)
    integer(4), intent(inout)       :: idx(:,:)             ! (2,nthreads)
    real(8), intent(inout)          :: zn(0:), wn(0:)
    real(8), intent(inout),optional :: h(0:)
    real(8), intent(in)             :: prec(0:), evap(0:)

    !- locals:
    integer(4) :: mi,nsrf,nl,nu
    real(8)    :: netprec

    call domp_get_domain (kh, 1, iw2, nl, nu, idx)

    do nsrf=nl,nu
      mi = nsrf

      netprec = prec(mi) - evap(mi)

      wn(mi) = wn(mi) + netprec
      zn(mi) = zn(mi) + netprec*dt
      if (present(h)) h(mi) = h(mi) + netprec*dt
    enddo

  end subroutine mass_precevap

  !-----------------------------------------------------------------------------

  subroutine save_old_sfc( iw2, c, hsn, csrf )

    use dmi_omp,   only : domp_get_domain
    use constants, only : hgr, zero

    implicit none
  
    integer(4), intent(in) :: iw2
    real(8), intent(in)    :: c(1:,0:), hsn(0:)
    real(8), intent(inout) :: csrf(0:)

    integer(4) :: nsrf, nl, nu

    call domp_get_domain(1,iw2, nl, nu)

    !- save c/h ----------------------------------------------------------------
    !  be careful not to get strange values at dried points where we will simply
    !  ignore the contribution from the net-precipitation source.
    do nsrf=nl,nu
      if (hsn(nsrf) >= hgr) then
        csrf(nsrf) = c(1,nsrf)/hsn(nsrf)
      else
        csrf(nsrf) = zero
      endif
    enddo
    
  end subroutine save_old_sfc

  !-----------------------------------------------------------------------------

  subroutine tracer_precevap( iw2, dt, cp, ce, prec, evap, c, nc )

    use dmi_omp,   only : domp_get_domain

    implicit none

    integer(4), intent(in) :: iw2, nc
    real(8), intent(in)    :: dt
    real(8), intent(in)    :: cp(0:), ce(0:)
    real(8), intent(in)    :: prec(0:), evap(0:)
    real(8), intent(inout) :: c(1:,0:)

    integer(4) :: nsrf, nl, nu
    real(8)    :: precip

    call domp_get_domain(1,iw2, nl, nu)

    do nsrf=nl,nu
      precip  = prec(nsrf)
      c(nc,nsrf) = c(nc,nsrf) + (precip*cp(nsrf) - evap(nsrf)*ce(nsrf))*dt
    enddo

  end subroutine tracer_precevap

  !-----------------------------------------------------------------------------

end module cmod_precip
