module cmod_sponge
  use constants, only : zero, one, half
  implicit none
  private
  public  :: sponge, sponge_e, sponge_ice_io, sponge_init, sponge_io

  real(8), parameter, private :: four = 4.0_8
contains

!===============================================================================

  subroutine sponge (kmax,mk,khu,khv,nz,krz,bndzk,bndz,cmp,u,v,dt,dx,dy,cosphi)

    implicit none
  
    integer(4), intent(in) :: kmax,nz
    integer(4), intent(in) :: mk(:,0:,0:)
    integer(4), intent(in) :: khu(0:), khv(0:)
    integer(4), intent(in) :: krz(:,0:)
    real(8), intent(in)    :: bndzk(:,:,0:)
    real(8), intent(in)    :: cmp(:,0:)
    real(8), intent(in)    :: u(0:), v(0:)
    real(8), intent(in)    :: dx(0:),dy(0:)
    real(8), intent(in)    :: dt, cosphi(:,0:)
    real(8), intent(inout) :: bndz(:,:,0:)
  
    integer(4) :: i, j, k, mi, n, nn
    real(8)    :: hdtddxj, hdtddyi, uu, vv, fac, fac4, hdt
  
    logical, save :: firsttime = .true.
        
    hdt = half*dt

    do n=1,nz
      i  = krz(1,n)
      j  = krz(2,n)
      nn = abs(krz(3,n))
  
      hdtddxj = hdt/(dx(j)*cosphi(1,i))
      hdtddyi = hdt/dy(i)
  
      if (nn == 1) then
        ! west boundary:
        do k=1,khu(mk(1,i,j))
          mi  = mk(k,i,j)
          uu  = u(mi)
          fac = uu*hdtddxj
          if ( uu > zero ) then
            bndz(:,k,n) = bndz(:,k,n) + fac*(bndzk(:,k,n)-bndz(:,k,n))
          else
            fac4 = fac*four
            bndz(:,k,n) = bndz(:,k,n) + fac4*(bndz(:,k,n)-cmp(:,mi))
          endif
        enddo
  
      elseif ( nn == 2 ) then
        ! north boundary:
        do k=1,khv(mk(1,i,j)) 
          mi  = mk(k,i,j)
          vv  = v(mi)
          fac = vv*hdtddyi
          if ( vv < zero ) then
            bndz(:,k,n) = bndz(:,k,n) - fac*(bndzk(:,k,n)-bndz(:,k,n))
          else
            fac4 = fac*four
            bndz(:,k,n) = bndz(:,k,n) - fac4*(bndz(:,k,n)-cmp(:,mi))
          endif
        enddo
      
      elseif ( nn == 3 ) then
        ! east boundary:
        do k=1,khu(mk(1,i,j-1))
          mi  = mk(k,i,j)
          uu  = u(mk(k,i,j-1))
          fac = uu*hdtddxj
          if ( uu < zero ) then
            bndz(:,k,n) = bndz(:,k,n) - fac*(bndzk(:,k,n)-bndz(:,k,n))
          else
            fac4 = fac*four
            bndz(:,k,n) = bndz(:,k,n) - fac4*(bndz(:,k,n)-cmp(:,mi))
          endif
        enddo
  
      elseif ( nn == 4 ) then
        ! south boundary:
        do k=1,khv(mk(1,i-1,j))
          mi  = mk(k,i,j)
          vv  = v(mk(k,i-1,j))
          fac = vv*hdtddyi
          if ( vv > zero ) then
            bndz(:,k,n) = bndz(:,k,n) + fac*(bndzk(:,k,n)-bndz(:,k,n))
          else
            fac4 = fac*four
            bndz(:,k,n) = bndz(:,k,n) + fac4*(bndz(:,k,n)-cmp(:,mi))
          endif
        enddo
  
      endif 
  
    enddo ! n loop
  
  
    !  On first entry, make sure to have nice sponge 
    !  values in all water points everywhere on the
    !  boundaries. 
    !  Having no better suggestion, "nice values" 
    !  here simply means the contents of the 
    !  boundary arrays rwzk{s/t}; it is up to the
    !  user to assure that these really are nice!
    if (firsttime) then   
      firsttime = .false.
  
      do n=1,nz
        i  = krz(1,n)
        j  = krz(2,n)
        nn = abs(krz(3,n))
  
        if (nn == 1) then
          ! west boundary:
          do k=khu(mk(1,i,j))+1,kmax
            if ( mk(k,i,j) > 0 ) then
              bndz(:,k,n) = bndzk(:,k,n)
            else
              exit
            endif
          enddo
  
        elseif ( nn == 2 ) then
          ! north boundary:
          do k=khv(mk(1,i,j))+1,kmax
            if ( mk(k,i,j) > 0 ) then
              bndz(:,k,n) = bndzk(:,k,n)
            else
              exit
            endif
          enddo
      
        elseif ( nn == 3 ) then
          ! east boundary:
          do k=khu(mk(1,i,j-1))+1,kmax
            if ( mk(k,i,j) > 0 ) then
              bndz(:,k,n) = bndzk(:,k,n)
            else
              exit
            endif
          enddo
  
        elseif ( nn == 4 ) then
          ! south boundary:
          do k=khv(mk(1,i-1,j))+1,kmax
            if ( mk(k,i,j) > 0 ) then
              bndz(:,k,n) = bndzk(:,k,n)
            else
              exit
            endif
          enddo
  
        endif 
  
      enddo
    endif   ! firsttime
  
  end subroutine sponge

!===============================================================================

  subroutine sponge_e(me,nz,krz,iceobc,bndice,ice,u,v,dt,dx,dy,cosphi,khu,khv)
  
    implicit none
  
    integer(4), intent(in)    :: nz
    integer(4), intent(in)    :: me(0:,0:)
    integer(4), intent(in)    :: krz(:,0:) 
    integer(4), intent(in)    :: khu(0:), khv(0:)
    real(8),    intent(in)    :: dt
    real(8),    intent(in)    :: cosphi(:,0:)
    real(8),    intent(in)    :: dx(0:)
    real(8),    intent(in)    :: dy(0:)
    real(8),    intent(in)    :: ice(1:,0:)
    real(8),    intent(in)    :: iceobc(1:,0:)
    real(8),    intent(in)    :: u(0:), v(0:)
    real(8),    intent(inout) :: bndice(1:,0:)
  
    ! --------------------------------------------------------------------------
  
    integer(4) :: i , j , n , nn
    real(8)    :: uu , vv, hdt, ddx, ddy
    logical, save :: firsttime = .true.

    ! --------------------------------------------------------------------------
    hdt = half*dt
  
    do n = 1 , nz
      i = krz(1,n)
      j = krz(2,n)
      nn = abs(krz(3,n))
      if ( nn == 1 .and. khu(me(i,j)) > 0) then
        uu  = u(me(i,j))
        ddx = one/(dx(j)*cosphi(1,i))
        if ( uu > zero ) then
          bndice(:,n) = bndice(:,n) + hdt*uu*ddx*(iceobc(:,n)-bndice(:,n))
        else
          bndice(:,n) = bndice(:,n) + dt*uu*ddx*(bndice(:,n)-ice(:,me(i,j)))
        endif
      elseif ( nn == 2 .and. khv(me(i,j)) > 0) then
        vv  = v(me(i,j))
        ddy = one/dy(i)
        if ( vv < zero ) then
          bndice(:,n) = bndice(:,n) - hdt*vv*ddy*(iceobc(:,n)-bndice(:,n))
        else
          bndice(:,n) = bndice(:,n) - dt*vv*ddy*(bndice(:,n)-ice(:,me(i,j)))
        endif
      elseif ( nn == 3 .and. khu(me(i,j-1)) > 0) then
        uu  = u(me(i,j-1))
        ddx = one/(dx(j)*cosphi(1,i))
        if ( uu < zero ) then
          bndice(:,n) = bndice(:,n) - hdt*uu*ddx*(iceobc(:,n)-bndice(:,n))
        else
          bndice(:,n) = bndice(:,n) - dt*uu*ddx*(bndice(:,n)-ice(:,me(i,j)))
        endif
      elseif ( nn == 4 .and. khv(me(i-1,j)) > 0) then
        vv  = v(me(i-1,j))
        ddy = one/dy(i)
        if ( vv > zero ) then
          bndice(:,n) = bndice(:,n) + hdt*vv*ddy*(iceobc(:,n)-bndice(:,n))
        else
          bndice(:,n) = bndice(:,n) + dt*vv*ddy*(bndice(:,n)-ice(:,me(i,j)))
        endif
      endif
    enddo

    !  On first entry, make sure to have nice sponge values in all water points
    !  everywhere on the boundaries.
    !  Having no better suggestion, "nice values" here simply means the 
    !  contents of the boundary arrays rwzed, rwzec, rwzsd; it is up to the
    !  user to assure that these really are nice during input!
    if (firsttime) then
      firsttime = .false.
      do n=1,nz
        i  = krz(1,n)
        j  = krz(2,n)
        nn = abs(krz(3,n))
        if ( (nn == 1 .and. khu(me(i,  j  )) == 0) .or.                        &
             (nn == 2 .and. khv(me(i,  j  )) == 0) .or.                        &
             (nn == 3 .and. khu(me(i,  j-1)) == 0) .or.                        &
             (nn == 4 .and. khv(me(i-1,j  )) == 0)      ) then
          bndice(:,n) = iceobc(:,n)
        endif
      enddo
    endif

  end subroutine sponge_e

!===============================================================================

  subroutine sponge_ice_io( nzmax,bndice,readsponge )
    !---------------------------------------------------------------------------
    !
    ! read or write ice sponge data files using hard-coded file names
    !
    !---------------------------------------------------------------------------
  
    !- modules -----------------------------------------------------------------
    use dmi_mpi_global, only : iu06
    use io_subs,        only : io_new_unit
  
    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    !- inputs:
    integer(4), intent(in) :: nzmax
    logical,    intent(in) :: readsponge
    !- in/out:
    real(8),    intent(inout) :: bndice(1:,0:)
  
    !- local vars --------------------------------------------------------------
    logical    :: found_error
    integer(4) :: n, luned, lunec, lunsd, ios
  
    !- Set hard-coded format for read/write ------------------------------------
    character(6), parameter :: cformat = '(f8.3)' 
  
    !- Open file with hard-coded names -----------------------------------------
    found_error = .false.
    if (nzmax > 0) then
      luned = io_new_unit()
      open (luned,file='sponge_ed',iostat=ios)
      if (ios /= 0) found_error = .true.
      lunec = io_new_unit()
      open (lunec,file='sponge_ec',iostat=ios)
      if (ios /= 0) found_error = .true.
      lunsd = io_new_unit()
      open (lunsd,file='sponge_sd',iostat=ios)
      if (ios /= 0) found_error = .true.
    endif
  
    !- Read sponge data file ---------------------------------------------------
    if (readsponge) then  
      bndice(1:3,:) = zero
      if (found_error) then
        write(iu06,*) 'WARNING: could not open ice sponge file(s),'
        write(iu06,*) '         initialising open bdy ice cond. with zeroes'
      else
        do n=1,nzmax
          read (luned,cformat,iostat=ios) bndice(1,n)
          if (ios /= 0) then
            found_error = .true.
            exit
          endif
          read (lunec,cformat,iostat=ios) bndice(2,n)
          if (ios /= 0) then
            found_error = .true.
            exit
          endif
          read (lunsd,cformat,iostat=ios) bndice(3,n)
          if (ios /= 0) then
            found_error = .true.
            exit
          endif
        enddo
        if (nzmax > 0) then
          close(luned)
          close(lunec)
          close(lunsd)
        endif
        if (found_error) then
          write(iu06,*) 'WARNING: could not read ice sponge file(s),'
          write(iu06,*) '         initialising open bdy ice cond. with zeroes'
          bndice(1:3,:) = zero
        endif
      endif
  
    !- Write sponge data files -------------------------------------------------
    else 
      if (found_error) then
        write(iu06,*) 'WARNING: could not open ice sponge file(s) for write,'
        write(iu06,*) '         no ice sponge files written'
      else
        do n=1,nzmax
          write(luned,cformat) bndice(1,n)
          write(lunec,cformat) bndice(2,n)
          write(lunsd,cformat) bndice(3,n)
        enddo
        if (nzmax > 0) then
          close(luned)
          close(lunec)
          close(lunsd)
        endif
      endif
    endif
  
  end subroutine sponge_ice_io

!===============================================================================

  subroutine sponge_init (kmax,mmk,khu,khv,nz,krz,bndzk,bndz,cmp,u,v,coupler,nt)
  
    implicit none
  
    logical,    intent(in)    :: coupler
    integer(4), intent(in)    :: kmax, nz, nt, mmk(:,0:,0:)
    integer(4), intent(in)    :: khu(0:), khv(0:), krz(:,0:)
    real(8),    intent(in)    :: bndzk(1:,1:,0:), cmp(:,0:), u(0:), v(0:)
    real(8),    intent(inout) :: bndz(1:,1:,0:)
  
    integer(4) :: i, j, k, n, nn, mi
  
  
    !- First, initialise as the boundary values: -------------------------------
    bndz(1,1:kmax,0:nz) = bndzk(1,1:kmax,0:nz)
    bndz(2,1:kmax,0:nz) = bndzk(2,1:kmax,0:nz)
    if (coupler)  bndz(3:nt,1:kmax,0:nz) = bndzk(3:nt,1:kmax,0:nz)
  
  
    !- Then, possibly adjust to initial S/T: -----------------------------------
    do n=1,nz
      i  = krz(1,n)
      j  = krz(2,n)
      nn = abs(krz(3,n))
  
      if (nn == 1) then
        ! west boundary:
        do k=1,khu(mmk(1,i,j))
          mi = mmk(k,i,j)
          if ( u(mi) <= zero ) then
            bndz(1,k,n) = cmp(1,mi)
            bndz(2,k,n) = cmp(2,mi)
          endif
        enddo
  
      elseif ( nn == 2 ) then
        ! north boundary:
        do k=1,khv(mmk(1,i,j)) 
          mi = mmk(k,i,j)
          if ( v(mi) >= zero ) then
            bndz(1,k,n) = cmp(1,mi)
            bndz(2,k,n) = cmp(2,mi)
          endif
        enddo
      
      elseif ( nn == 3 ) then
        ! east boundary:
        do k=1,khu(mmk(1,i,j-1))
          if ( u(mmk(k,i,j-1)) >= zero ) then
            mi = mmk(k,i,j)
            bndz(1,k,n) = cmp(1,mi)
            bndz(2,k,n) = cmp(2,mi)
          endif
        enddo
  
      elseif ( nn == 4 ) then
        ! south boundary:
        do k=1,khv(mmk(1,i-1,j))
          if ( v(mmk(k,i-1,j)) <= zero ) then
            mi = mmk(k,i,j)
            bndz(1,k,n) = cmp(1,mi)
            bndz(2,k,n) = cmp(2,mi)
          endif
        enddo
  
      endif 
  
    enddo ! n loop

    if (coupler) then
      do n=1,nz
        i  = krz(1,n)
        j  = krz(2,n)
        nn = abs(krz(3,n))
  
        if (nn == 1) then
          ! west boundary:
          do k=1,khu(mmk(1,i,j))
            mi = mmk(k,i,j)
            if ( u(mi) <= zero ) then
              bndz(3:,k,n) = cmp(3:,mi)
            endif
          enddo
  
        elseif ( nn == 2 ) then
          ! north boundary:
          do k=1,khv(mmk(1,i,j))
            mi = mmk(k,i,j)
            if ( v(mi) >= zero ) then
              bndz(3:,k,n) = cmp(3:,mi)
            endif
          enddo
  
        elseif ( nn == 3 ) then
          ! east boundary:
          do k=1,khu(mmk(1,i,j-1))
            if ( u(mmk(k,i,j-1)) >= zero ) then
              mi = mmk(k,i,j)
              bndz(3:,k,n) = cmp(3:,mi)
            endif
          enddo
  
        elseif ( nn == 4 ) then
          ! south boundary:
          do k=1,khv(mmk(1,i-1,j))
            if ( v(mmk(k,i-1,j)) <= zero ) then
              mi = mmk(k,i,j)
              bndz(3:,k,n) = cmp(3:,mi)
            endif
          enddo
  
        endif
      enddo ! n loop
    endif   ! coupler

  end subroutine sponge_init

!===============================================================================

  subroutine sponge_io(kmax, nzmax, khbnd, bndz, coupler, rds)
    !---------------------------------------------------------------------------
    !
    ! read or write S/T and passive tracer sponge data files
    ! using hard-coded file names and semi-hard-coded format
    !
    !---------------------------------------------------------------------------
  
    !- modules -----------------------------------------------------------------
    use exits,         only : exitme
    use io_subs,       only : io_new_unit
    use cmod_miscvars, only : ntracers
    use constants,     only : coupler_psponge
  
    !- directives --------------------------------------------------------------
    implicit none
  
    
    !- arguments ---------------------------------------------------------------
    !- inputs:
    integer(4), intent(in)    :: kmax, nzmax, khbnd(0:)
    logical,    intent(in)    :: coupler, rds
    !- in/out:
    real(8),    intent(inout) :: bndz(1:,1:,0:)
  
    
    !- local vars --------------------------------------------------------------
    character(256) :: cformat,cformatc
    integer(4)     :: n, k, kb, luns, lunt, ios, nn, lunc
    character(8), parameter :: fnams = 'sponge_s', fnamt = 'sponge_t',         &
                               fnamc = 'sponge_c'
  
  
    !- Check if we should do anything at all -----------------------------------
    if (nzmax == 0) return
  
    !- Open file with hard-coded names -----------------------------------------
    lunt = io_new_unit()
    open (lunt,file=fnamt,iostat=ios)
    if (ios /= 0) call exitme(1,'Could not open '//fnamt)
    luns = io_new_unit()
    open (luns,file=fnams,iostat=ios)
    if (ios /= 0) call exitme(1,'Could not open '//fnams)
    if (coupler) then
      lunc = io_new_unit()
      open (lunc,file=fnamc,iostat=ios)
      if (ios /= 0) call exitme(1,'Could not open '//fnamc)
    endif
  
    !- Set semi-hard-coded format for read/write -------------------------------
    if     (kmax < 10  ) then
      write(cformat,'(i1)') kmax
    elseif (kmax < 100 ) then
      write(cformat,'(i2)') kmax
    elseif (kmax < 1000) then
      write(cformat,'(i3)') kmax
    elseif (kmax < 10000) then
      write(cformat,'(i4)') kmax
    else
      call exitme('Error: Cannot handle more than 9999 open layers.')
    endif
    
    if (coupler_psponge) then
      cformatc = '('//trim(cformat)//'f8.3)' ! coupler uses 3 decimals
    else
      cformatc = '('//trim(cformat)//'f8.1)' 
    endif  
    cformat  = '('//trim(cformat)//'f8.1)'
    
    !- Read sponge data files --------------------------------------------------
    if (rds) then  
      do n=1,nzmax
        kb = khbnd(n)
        read (lunt,cformat,iostat=ios) bndz(1,1:kb,n) 
        if (ios /= 0) call exitme(1,'Could not read '//fnamt)
        read (luns,cformat,iostat=ios) bndz(2,1:kb,n) 
        if (ios /= 0) call exitme(1,'Could not read '//fnams)
        if (kb < kmax) then
          bndz(1,kb+1:kmax,n) = bndz(1,kb,n)
          bndz(2,kb+1:kmax,n) = bndz(2,kb,n)
        endif
        if (coupler) then
          do nn=1,ntracers-2
            read (lunc,cformatc,iostat=ios) bndz(2+nn,1:kb,n)
            if (ios /= 0) call exitme(1,'Could not read '//fnamc)
          enddo
          do k=kb+1,kmax
            bndz(3:,k,n) = bndz(3:,kb,n)
          enddo
        endif
      enddo
      bndz(1,:,0) = bndz(1,:,1)
      bndz(2,:,0) = bndz(2,:,1)
      if (coupler) bndz(3:,:,0) = bndz(3:,:,1)
  
    !- Write S/T sponge data files ---------------------------------------------
    else 
      do n=1,nzmax
        kb = khbnd(n)
        write(lunt,cformat,iostat=ios) bndz(1,1:kb,n) 
        if (ios /= 0) call exitme(1,'Could not write '//fnamt)
        write(luns,cformat,iostat=ios) bndz(2,1:kb,n) 
        if (ios /= 0) call exitme(1,'Could not write '//fnams)
        if (coupler) then
          do nn=1,ntracers-2
            write (lunc,cformatc) bndz(2+nn,1:kb,n)
          enddo
        endif
      enddo
    endif
  
    !- Clean up ----------------------------------------------------------------
    close(lunt)
    close(luns)
    if (coupler) close(lunc)
  
  end subroutine sponge_io

end module cmod_sponge
