module coare_mod
  !-----------------------------------------------------------------------------
  ! Module containing the essentials of the COARE 3.0 algorithm
  ! 
  ! We have simply "stolen" the original source from:
  !  ftp://ftp1.esrl.noaa.gov/users/cfairall/wcrp_wgsf/computer_programs/
  !                                                      cor3_0/fortran3_0/f90/
  !
  ! THANKS A LOT, guys!
  !
  ! and modified it into a modern F90 language which comform to our standards
  ! and styling rules, and which is then much more portable, reliable, and runs
  ! more efficient.
  !-----------------------------------------------------------------------------


  !- modules: ------------------------------------------------------------------
  use constants, only : g, vkarm, rhow, cpw, pi, sph,                          &
                        milli, half, zero, one, two, three, four, six, ten,    &
                        onethird, twothird, qrt

  !- implicit directives: ------------------------------------------------------
  implicit none
  private

  !- Routines: -----------------------------------------------------------------
  public  :: coare30a, coare_init

  !- Constants: ----------------------------------------------------------------
  integer(4), parameter :: nitmin = 1, nitmax = 3

  real(8), parameter :: Beta = 1.2_8, von = vkarm, fdg = 1.0_8, Rgas = 287.1_8 
  real(8), parameter :: tdk = 273.16_8, grav = g, mbar2P = 100.0_8
  real(8), parameter :: tsle = -2370.0_8, Le0 = 2501000.0_8, wetc0 = 0.622_8
  real(8), parameter :: vsa0 = 1.326e-5_8, vsa1 = 6.542e-3_8
  real(8), parameter :: vsa2 = 8.301e-6_8, vsa3 = -4.84e-9_8
  real(8), parameter :: cpa = 1004.67_8, qcpv = 0.84_8, qra = 0.61_8
  real(8), parameter :: visw = 1e-6_8, be = 0.026_8, tcw = 0.6_8
  real(8), parameter :: bigc0 = 16.0_8*grav*cpw*(rhow*visw)**3/tcw**2 
  real(8), parameter :: Al1 = 2.1e-5_8, Al2 = 3.2_8, Alpow = 0.79_8
  real(8), parameter :: sigma = 5.67e-8_8, coolf = 0.3_8, b4 = 0.004_8
  real(8), parameter :: Rn1 = 0.945_8, Rn2 = 0.97_8, twopi = 2.0_8*pi
  real(8), parameter :: charnl = 0.011_8, charnh = 0.018_8, utup = 18.0_8
  real(8), parameter :: umin = 0.1_8, webb0 = 1.61_8, zu0 = 0.0001_8
  real(8), parameter :: wvpow = 1.94_8, dw0 = 2.11e-5_8, zetuhigh = 50.0_8
  real(8), parameter :: dtd1 = 0.003309_8, dtd2 = 0.00000144_8, dtdf = 0.02411_8
  real(8), parameter :: neutraldt = -0.0098_8, dter0 = 0.3_8, usr0 = 0.035_8
  real(8), parameter :: zouu = 0.011_8, zovu = 0.11_8, Ch10 = 0.00115_8
  real(8), parameter :: xlpow1 = 0.75_8, xlpow2 = 0.333_8, tktmax = 0.01_8
  real(8), parameter :: ug0 = 0.2_8, oostfac = 50.0_8/twopi, TYfac = 1200.0_8
  real(8), parameter :: zoqmax = 0.000115_8, zoqfac = 0.000055_8, zoqpow = 0.6_8
  real(8), parameter :: dr0 = 0.065_8, dr1 = 11.0_8, dr2 = -6.6e-5_8
  real(8), parameter :: drt = 0.0008_8, von2 = von*von, v2fdg = von2*fdg
  real(8), parameter :: cmax = 50.0_8, cslope = 0.35_8
  real(8), parameter :: c0 = 14.28_8, c1 = 8.525_8
  real(8), parameter :: p1  = 10.15_8, p2  = 1.5_8, p3 = 34.15_8
  real(8), parameter :: fifteen = 15.0_8, oneph = 1.5_8

  real(8), save :: sq3, a1, a2
contains

  !-----------------------------------------------------------------------------

  subroutine coare30a (x,y)
    ! Based on the original cor30a with shortened iteration modified Rt and Rq.
    !
    !   Uses wave information wave period [s] and wave ht [m]:
    !     jwave = 0 : No wave, standard coare 2.6 charnock
    !     jwave = 1 : Oost et al.  zo=50/2/pi L (u*/c)**4.5 
    !     jwave = 2 : Taylor and Yelland  zo=1200 h*(L/h)**4.5 jwave=2

    implicit none

    !- Arguments ---------------------------------------------------------------
    real(8), intent(in)  :: x(:)  ! (19)
    real(8), intent(out) :: y(:)  ! (22)

    !- Local vars --------------------------------------------------------------
    ! from input:
    real(8)    :: u, us, ts, t, Qs, Q, Rs, Rl, rain, zi, P, zu, zt, zq, lat
    real(8)    :: jwave, jcool, twave, hwave
    ! other locals:
    real(8)    :: Le, rhoa, Al, wetc
    real(8)    :: lwave, cwave, Rns, Rnl, du, dt, dq, qout, dels, qcol, alq
    real(8)    :: bf, CC, Cd10, charn, Ct, dtmp, dwat, hl_webb, xlamx, alfac
    real(8)    :: L10, Ribcu, Ribu, rr, ta, u10, ut, zet, zetu, zo10, zot10
    real(8)    :: hsb, hlb, tau, zo, zot, zoq, L, usr, tsr, qsr, dter, dqer
    real(8)    :: tkt, RF, wbar, Cd, Ch, Ce, Cdn_10, Chn_10, Cen_10, ug 
    real(8)    :: xx, bigc, cpv, visa, psiuo, psit, psiq, z
    real(8)    :: lzu10, lzt10, lzq10, dtjcl, dqjcl, sqrtrr, psik, psic, f
    integer(4) :: i, nits

    ! prepare input ------------------------------------------------------------
    u     = x(1)       ! wind speed (m/s) at height zu (m)
    us    = x(2)       ! surface current speed in the wind direction (m/s)
    ts    = x(3)       ! bulk water temperature (C) if jcool=1, 
                       ! interface water T if jcool=0  
    t     = x(4)       ! bulk air temperature (C), height zt
    Qs    = x(5)       ! bulk water spec hum (kg/kg) if jcool=1, ...
    Q     = x(6)       ! bulk air spec hum (kg/kg), height zq
    Rs    = x(7)       ! downward solar flux (W/m**2)
    Rl    = x(8)       ! downard IR flux (W/m**2)
    rain  = x(9)       ! rain rate (mm/hr)
    zi    = x(10)      ! PBL depth (m)
    P     = x(11)      ! Atmos surface pressure (mb)
    zu    = x(12)      ! wind speed measurement height (m)
    zt    = x(13)      ! air T measurement height (m)
    zq    = x(14)      ! air q measurement height (m)
    lat   = x(15)      ! latitude (deg, N=+)
    jcool = x(16)      ! implement cool calculation skin switch, 0=no, 1=yes
    jwave = x(17)      ! implement wave dependent roughness model
    twave = x(18)      ! wave period (s)
    hwave = x(19)      ! wave height (m)

    ! air parameters -----------------------------------------------------------
    Le   = Le0 + tsle*ts
    cpv  = cpa*(one + qcpv*Q) 
    rhoa = P*mbar2P/(Rgas*(t+tdk)*(one + qra*Q)) 
    visa = vsa0*(one + (vsa1 + (vsa2 + vsa3*t)*t)*t) 

    ! cool skin parameters -----------------------------------------------------
    Al   = Al1*(ts + Al2)**Alpow
    wetc = wetc0*Le*Qs/(Rgas*(ts+tdk)**2) 
     
    ! wave parameters ----------------------------------------------------------
    cwave = grav/twopi*twave 
    lwave = cwave*twave
     
    ! compute aux stuff --------------------------------------------------------
    Rns = Rs*Rn1
    Rnl = Rn2*(sigma*(ts-coolf*jcool+tdk)**4 - Rl) 
     
    sqrtrr = sqrt(rhoa/rhow)

    !***************   Begin bulk loop *****************************************
     
    ! first guess --------------------------------------------------------------
    du    = u - us 
    dt    = ts - t + neutraldt*zt 
    dq    = Qs - Q 
    ta    = t + tdk 
    dter  = dter0
    dqer  = wetc*dter 
    ut    = sqrt(du*du + qrt)
    u10   = ut*log(ten/zu0)/log(zu/zu0) 
    usr   = usr0*u10 
    zo10  = zouu*usr*usr/grav + zovu*visa/usr 
    xx    = von/log(ten/zo10)
    zot10 = ten*exp(-(von/Ch10)*xx) 
    Cd10  = xx**2 
    lzu10 = log(zu/zo10)
    lzt10 = log(zt/zot10)
    lzq10 = log(zq/zot10)
    Cd    = (von/lzu10)**2 
    Ct    = von/lzt10 
    CC    = von*Ct/Cd 
    dtjcl = dter*jcool
    Ribcu = -zu/(zi*b4*Beta**3) 
    Ribu  = -grav*zu*((dt-dtjcl)/ta + qra*dq)/ut**2 
    if (Ribu < zero) then 
      zetu = CC*Ribu/(one + Ribu/Ribcu) 
    else 
      zetu =    Ribu*(CC  + three*Ribu)
    endif 
    L10 = zu/zetu 
    if (zetu > zetuhigh) then 
      nits = nitmin 
    else
      nits = nitmax
    endif 
    if (L10 <= zero) then
      ! wind
      z     = zu/L10
      xx    = sqrt(sqrt(one-fifteen*z))
      psik  = log((one+xx)*(one+xx)*(one+xx*xx)) - atan(xx) + a1
      xx    = (one - p1*z)**onethird
      psic  = p2*log(one+xx+xx*xx) - sq3*atan((one+two*xx)/sq3) + a2
      f     = z*z/(one+z*z) 
      psiuo = (one-f)*psik + f*psic
      ! temperature
      z     = zt/L10
      xx    = sqrt(one-fifteen*z) 
      psik  = two*log(one+xx) - log(four) 
      xx    = (one - p3*z)**onethird
      psic  = oneph*log(one+xx+xx*xx) - sq3*atan((one+two*xx)/sq3) + a2
      f     = z*z/(one+z*z) 
      psit  = (one-f)*psik + f*psic
      ! humidity
      z     = zq/L10
      xx    = sqrt(one-fifteen*z) 
      psik  = two*log(one+xx) - log(four) 
      xx    = (one - p3*z)**onethird
      psic  = oneph*log(one+xx+xx*xx) - sq3*atan((one+two*xx)/sq3) + a2
      f     = z*z/(one+z*z) 
      psiq  = (one-f)*psik + f*psic
    else
      ! wind
      z     = zu/L10
      psiuo = -((one+z) + twothird*(z-c0)*exp(-min(cmax, cslope*z)) + c1)
      ! temperature
      z     = zt/L10
      xx    = (one+twothird*z)
      psit  = -(xx*sqrt(xx) + twothird*(z-c0)*exp(-min(cmax, cslope*z)) + c1)
      ! humidity
      z     = zq/L10
      xx    = (one+twothird*z)
      psiq  = -(xx*sqrt(xx) + twothird*(z-c0)*exp(-min(cmax, cslope*z)) + c1)
    endif
    usr = ut*von/(lzu10 - psiuo)
    tsr = -(dt-dtjcl)*von*fdg/(lzt10 - psit) 
    qsr = -(dq-wetc*dtjcl)*von*fdg/(lzq10 - psiq) 

    tkt   = milli
    if (ut > utup) then
      charn = charnh 
    elseif (ut > ten) then
      charn = charnl + (ut-ten)/(utup-ten)*(charnh-charnl)
    else
      charn = charnl
    endif 
        
    ! bulk loop ----------------------------------------------------------------
    do i=1,nits 
     
      zet = von*grav*zu/ta*(tsr*(one+qra*Q)+qra*ta*qsr)/(usr*usr)/(one+qra*Q) 
      !disp(usr)
      !disp(zet) 
      if (jwave == zero) then
        zo = charn*usr*usr/grav+zovu*visa/usr  
      elseif (jwave == one) then
        xx = usr/cwave
        zo = oostfac*lwave*sqrt(xx)*xx**4 + zovu*visa/usr !Oost et al
      elseif (jwave == two) then
        xx = hwave/lwave
        zo = TYfac*hwave*sqrt(xx)*xx**4 + zovu*visa/usr   !Taylor and Yelland
      endif
      rr    = zo*usr/visa 
      L     = zu/zet 
      zoq   = min(zoqmax, zoqfac/rr**zoqpow) 
      zot   = zoq 

      if (L <= zero) then
        ! wind
        z     = zu/L
        xx    = sqrt(sqrt(one-fifteen*z))
        psik  = log((one+xx)*(one+xx)*(one+xx*xx)) - atan(xx) + a1
        xx    = (one - p1*z)**onethird
        psic  = p2*log(one+xx+xx*xx) - sq3*atan((one+two*xx)/sq3) + a2
        f     = z*z/(one+z*z) 
        psiuo = (one-f)*psik + f*psic
        ! temperature
        z     = zt/L
        xx    = sqrt(one-fifteen*z) 
        psik  = two*log(one+xx) - log(four) 
        xx    = (one - p3*z)**onethird
        psic  = oneph*log(one+xx+xx*xx) - sq3*atan((one+two*xx)/sq3) + a2
        f     = z*z/(one+z*z) 
        psit  = (one-f)*psik + f*psic
        ! humidity
        z     = zq/L
        xx    = sqrt(one-fifteen*z) 
        psik  = two*log(one+xx) - log(four) 
        xx    = (one - p3*z)**onethird
        psic  = oneph*log(one+xx+xx*xx) - sq3*atan((one+two*xx)/sq3) + a2
        f     = z*z/(one+z*z) 
        psiq  = (one-f)*psik + f*psic
      else
        ! wind
        z     = zu/L
        psiuo = -((one+z) + twothird*(z-c0)*exp(-min(cmax, cslope*z)) + c1)
        ! temperature
        z     = zt/L
        xx    = (one+twothird*z)
        psit  = -(xx*sqrt(xx)+twothird*(z-c0)*exp(-min(cmax,cslope*z)) + c1)
        ! humidity
        z     = zq/L
        xx    = (one+twothird*z)
        psiq  = -(xx*sqrt(xx)+twothird*(z-c0)*exp(-min(cmax,cslope*z)) + c1)
      endif

      usr   = ut*von/(lzu10 - psiuo) 
      tsr   = -(dt-dtjcl)*von*fdg/(lzt10 - psit) 
      qsr   = -(dq-wetc*dtjcl)*von*fdg/(lzq10 - psiq) 

      Bf    = -grav*usr*(tsr + qra*ta*qsr)/ta
      if (Bf > zero) then
        ug = Beta*(Bf*zi)**xlpow2 
      else
        ug = ug0
      endif
      ut   = sqrt(du*du+ug*ug) 
      Rnl  = Rn2*(sigma*(ts-dtjcl+tdk)**4-Rl) 
      hsb  = -rhoa*cpa*usr*tsr 
      hlb  = -rhoa*Le*usr*qsr 
      qout = Rnl+hsb+hlb 
      dels = Rns*(dr0 + dr1*tkt + dr2/tkt*(one-exp(-tkt/drt))) ! Eq.16 SW
      qcol = qout-dels 
      alq  = Al*qcol+be*hlb*cpw/Le                       ! Eq. 7 Buoy flux water

      if (alq > zero) then 
        bigc  = bigc0/rhoa**2 
        xlamx = six/(one + (bigc*alq/usr**4)**xlpow1)**xlpow2 ! Eq 13 Saunders
        tkt   = xlamx*visw/(sqrtrr*usr)                       ! Eq.11 Sub. thk
      else
        xlamx = six
        tkt   = min(tktmax, xlamx*visw/(sqrtrr*usr))          ! Eq.11 Sub. thk
      endif 
     
      dter  = qcol*tkt/tcw                                    !  Eq.12 Cool skin
      dtjcl = dtjcl*jcool
      dqer  = wetc*dter 
   
    enddo !bulk iter loop
    tau =  rhoa*usr*usr*du/ut                 !stress
    hsb = -rhoa*cpa*usr*tsr 
    hlb = -rhoa*Le*usr*qsr 
     
    ! rain heat flux -----------------------------------------------------------
    dwat  = dw0*((t+tdk)/tdk)**wvpow                  ! water vapour diffusivity
    dtmp  = (one + (dtd1-dtd2*t)*t)*dtdf/(rhoa*cpa)   ! heat diffusivity
    alfac = one/(one+(wetc*Le*dwat)/(cpa*dtmp))       ! wet bulb factor
    dqjcl = dqer*jcool
    RF    = rain*alfac*cpw*((ts-t-dtjcl)+(Qs-Q-dqjcl)*Le/cpa)/sph

    ! Webb et al. correction ---------------------------------------------------
    ! formulation in hlb already incls webb
    wbar = (webb0*hlb/Le/(one+webb0*Q) + hsb/(cpa*ta))/rhoa
    !wbar=1.61*hlb/Le/rhoa+(1+1.61*Q)*hsb/rhoa/cpa/ta 
    hl_webb = rhoa*wbar*Q*Le 

    ! compute transfer coeffs relative to ut @meas. ht -------------------------
    Cd = tau/rhoa/ut/max(umin,du) 
    Ch = -usr*tsr/ut/(dt-dtjcl) 
    Ce = -usr*qsr/(dq-dqjcl)/ut 

    ! 10-m neutral coeffs realtive to ut ---------------------------------------
    z      = v2fdg/log(ten/zo)
    Cdn_10 = von2/(log(ten/zo)**2)
    Chn_10 = z/log(ten/zot) 
    Cen_10 = z/log(ten/zoq) 

    ! the Y array going back to the caller -------------------------------------
    y = (/hsb, hlb, tau, zo, zot, zoq, L, usr, tsr, qsr, dter, dqer, tkt, RF,  &
         wbar, Cd, Ch, Ce, Cdn_10, Chn_10, Cen_10, ug /) 
  end subroutine coare30a

  !-----------------------------------------------------------------------------

  subroutine coare_init()

    implicit none

    sq3 = sqrt(three)
    a1  = pi*half - three*log(two)
    a2  = pi/sq3 - log(three*sq3)

    ! more initial stuff to be moved to here ...

  end subroutine coare_init

  !-----------------------------------------------------------------------------

end module coare_mod
