module compress

  use cmod_mem, only : cmi1, cmi2, cmi3, cml1

  implicit none
  private

  public  :: compress_i4, compress_i4_ice, allocate_compressed_arrays,         &
             allocate_compressed_ice_arrays
  private :: wcompr, wcompr_w2, wcompr_w2p, wcompr_e, set_mm_op, wcompr_cmp,   &
             wcompr_iota

  !- Scale factors for compress ------------------------------------------------
  real(8), parameter, public :: wuvscale = 10000000.0_8     ! wind scalefactor
  real(8), parameter, public :: uvscale  = 100000000.0_8    ! current scalefac.
  real(8), parameter, public :: stscale  = 10000000.0_8     ! salt/temp scfac.
  real(8), parameter, public :: wscale   = 100000000000.0_8 ! vert. vel. scfac.
  real(8), parameter, public :: hscale   = 100000.0_8       ! layer thkn. scfac.
  real(8), parameter, public :: zscale   = 100000000.0_8    ! elevation scfac.
  real(8), parameter, public :: edscale  = 1000.0_8         ! ice thickn. scfac.
  real(8), parameter, public :: tsscale  = 10.0_8           ! tsnei scalefactor
  real(8), parameter, public :: ecscale  = 100.0_8          ! ice cover scfac.
  real(8), parameter, public :: sdscale  = 1000.0_8         ! snow thickn.scfac.
  real(8), parameter, public :: uvescale = 1000.0_8         ! ice vel. scfac.
  real(8), parameter, public :: iotascale = 100.0_8         ! iota scfac.
  
  !- Integer arrays for output format ------------------------------------------
  type(cmi1), allocatable, public, save :: wu_i4(:), wv_i4(:)
  type(cmi1), allocatable, public, save :: u_i4(:), v_i4(:), w_i4(:)
  type(cmi1), allocatable, public, save :: h_i4(:), z_i4(:), s_i4(:), t_i4(:)
  type(cmi1), allocatable, public, save :: eisdi_i4(:), eisco_i4(:)
  type(cmi1), allocatable, public, save :: snowd_i4(:), tsnei_i4(:)
  type(cmi1), allocatable, public, save :: ueis_i4(:), veis_i4(:)
  type(cml1), allocatable, public, save :: casus_out(:)
  type(cmi3), allocatable, public, save :: mm_op(:)
  type(cmi2), allocatable, public, save :: idx_op(:)
  type(cmi1), allocatable, public, save :: iota1_i4(:),iota2_i4(:),iota3_i4(:)

contains

!===============================================================================

  subroutine wcompr_e( iwet2,ice,tsnei,ueis,veis,casus,                        &
                       eisdi_i4,eisco_i4,snowd_i4,tsnei_i4,ueis_i4,veis_i4,    &
                       c_out )
    ! compress ice output-arrays to I4
  
    use dmi_omp, only : domp_get_domain
    implicit none
  
    integer(4), intent(in)                   :: iwet2
    integer(4), intent(inout), dimension(0:) :: eisdi_i4, eisco_i4,            &
                                                snowd_i4, tsnei_i4,            &
                                                ueis_i4,  veis_i4
    logical, intent(inout)              :: c_out(0:)
    logical, intent(in)                 :: casus(0:)
    real(8), intent(in)                 :: ice(1:,0:)
    real(8), intent(in),  dimension(0:) :: tsnei, ueis,  veis
  
    integer(4) :: i, nl, nu

    call domp_get_domain (0, iwet2, nl, nu)
  
    do i=nl,nu
      c_out(i)    = casus(i)
      eisdi_i4(i) = nint(ice(1,i)*edscale)
      eisco_i4(i) = nint(ice(2,i)*ecscale)
      snowd_i4(i) = nint(ice(3,i)*sdscale)
      tsnei_i4(i) = nint(tsnei(i)*tsscale)
      ueis_i4(i)  = nint(ueis(i)*uvescale)
      veis_i4(i)  = nint(veis(i)*uvescale)
    enddo
  
  end subroutine wcompr_e

!===============================================================================

  subroutine wcompr_iota( iwet2,iota,iota1_i4,iota2_i4,iota3_i4 )
    ! compress iota output-arrays to I4
  
    use dmi_omp, only : domp_get_domain
    implicit none
  
    integer(4), intent(in)                   :: iwet2
    integer(4), intent(inout), dimension(0:) :: iota1_i4, iota2_i4, iota3_i4
    real(8),    intent(in)                   :: iota(1:,0:)
  
    integer(4) :: i, nl, nu

    call domp_get_domain (0, iwet2, nl, nu)
  
    do i=nl,nu
      iota1_i4(i) = nint(iota(1,i)*iotascale)
      iota2_i4(i) = nint(iota(2,i)*iotascale)
      iota3_i4(i) = nint(iota(3,i)*iotascale)
    enddo
  
  end subroutine wcompr_iota

!===============================================================================

  subroutine wcompr_w2( iwet2,wu,wv,wul_i4,wvl_i4 )
    ! compress wind to I4 before output
  
    use dmi_omp, only : domp_get_domain

    implicit none
  
    integer(4), intent(in)                   :: iwet2
    integer(4), intent(inout), dimension(0:) :: wul_i4, wvl_i4
    real(8),    intent(in),    dimension(0:) :: wu, wv
  
  
    integer(4)         :: i, nl, nu
    real(8)            :: wuwa, wvwa
    real(8), parameter :: datamax  = 99.0_8

    call domp_get_domain (1, iwet2, nl, nu)

    do i=nl,nu
      wuwa   = min(datamax, max(-datamax,wu(i)))
      wvwa   = min(datamax, max(-datamax,wv(i)))
  
      wul_i4(i) = nint(wuwa*wuvscale)
      wvl_i4(i) = nint(wvwa*wuvscale)
    enddo
  
  end subroutine wcompr_w2

!===============================================================================

  subroutine wcompr_w2p( iwet2,mm_op,ind,wu,wv,wul_i4,wvl_i4 )
    ! de-permute and compress wind to I4 before output
  
    use dmi_omp, only : domp_get_domain
    implicit none
  
    integer(4), intent(in)                   :: iwet2, mm_op(0:,0:,:), ind(:,:)
    integer(4), intent(inout), dimension(0:) :: wul_i4, wvl_i4
    real(8),    intent(in),    dimension(0:) :: wu, wv
  
  
    integer(4)         :: mn, mo, nl, nu
    real(8)            :: wuwa, wvwa
    real(8), parameter :: datamax  = 99.0_8

    call domp_get_domain (1, iwet2, nl, nu)

    do mn=nl,nu
      mo = mm_op(ind(1,mn),ind(2,mn),1)

      wuwa = min(datamax, max(-datamax,wu(mn)))
      wvwa = min(datamax, max(-datamax,wv(mn)))
  
      wul_i4(mo) = nint(wuwa*wuvscale)
      wvl_i4(mo) = nint(wvwa*wuvscale)
    enddo

  end subroutine wcompr_w2p
  
  
!===============================================================================

  subroutine wcompr(iwet2, iwet3, u, v, h, w, z, u_i4, v_i4, h_i4, w_i4, z_i4, &
                                s, t, s_i4, t_i4 )
    ! compress output quantities to I4
  
    use constants, only : dyn
    use dmi_omp,   only : domp_get_domain

    implicit none
  
    integer(4), intent(in)                   :: iwet2,iwet3
    real(8),    intent(inout), dimension(0:) :: z
    real(8),    intent(inout), dimension(0:) :: h
    real(8),    intent(inout), dimension(0:) :: u,v,w
    real(8),    intent(inout), dimension(0:) :: s,t
    integer(4), intent(inout), dimension(0:) :: z_i4
    integer(4), intent(inout), dimension(0:) :: h_i4
    integer(4), intent(inout), dimension(0:) :: u_i4,v_i4,w_i4,s_i4,t_i4
  
    integer(4) :: i, nl, nu
      
    real(8), parameter :: datamax   = 99.0_8
    real(8), parameter :: w_datamax = 0.0099_8
    real(8), parameter :: h_datamax = 99999.0_8
  
    call domp_get_domain (0, iwet3, nl, nu)

    do i=nl,nu
      u(i) = min(datamax, max(-datamax,u(i)))
      v(i) = min(datamax, max(-datamax,v(i)))
      w(i) = min(w_datamax, max(-w_datamax,w(i)))
      s(i) = min(datamax, max(-datamax,s(i)))
      t(i) = min(datamax, max(-datamax,t(i)))
  
      u_i4(i) = nint(u(i)*uvscale)
      v_i4(i) = nint(v(i)*uvscale)
      w_i4(i) = nint(w(i)* wscale)
      s_i4(i) = nint(s(i)*stscale)
      t_i4(i) = nint(t(i)*stscale)
    enddo
    if (dyn) then
      do i=nl,nu
        h(i)    = min(h_datamax, max(-h_datamax,h(i)))
        h_i4(i) = nint(h(i)*hscale)
      enddo
    endif
    
    call domp_get_domain (0, iwet2, nl, nu)

    do i=nl,nu
      z(i)    = min(datamax, max(-datamax,z(i)))
      z_i4(i) = nint(z(i)*zscale)
    enddo
  
  end subroutine wcompr

!===============================================================================

  subroutine wcompr_cmp(iwet2, iwet3, u, v, h, w, z, u_i4, v_i4, h_i4, w_i4,   &
                        z_i4, ts, s_i4, t_i4 )
    ! compress output quantities to I4
  
    use constants, only : dyn
    use dmi_omp,   only : domp_get_domain

    implicit none
  
    integer(4), intent(in)                   :: iwet2,iwet3
    real(8),    intent(inout), dimension(0:) :: z
    real(8),    intent(inout), dimension(0:) :: h
    real(8),    intent(inout), dimension(0:) :: u,v,w
    real(8),    intent(inout)                :: ts(1:,0:)
    integer(4), intent(inout), dimension(0:) :: z_i4
    integer(4), intent(inout), dimension(0:) :: h_i4
    integer(4), intent(inout), dimension(0:) :: u_i4,v_i4,w_i4,s_i4,t_i4
  
    integer(4) :: i, nl, nu
      
    real(8), parameter :: datamax   = 99.0_8
    real(8), parameter :: w_datamax = 0.0099_8
    real(8), parameter :: h_datamax = 99999.0_8
  
    call domp_get_domain (0, iwet3, nl, nu)

    do i=nl,nu
      u(i) = min(datamax, max(-datamax,u(i)))
      v(i) = min(datamax, max(-datamax,v(i)))
      w(i) = min(w_datamax, max(-w_datamax,w(i)))
      ts(2,i) = min(datamax, max(-datamax,ts(2,i)))
      ts(1,i) = min(datamax, max(-datamax,ts(1,i)))
  
      u_i4(i) = nint(u(i)*uvscale)
      v_i4(i) = nint(v(i)*uvscale)
      w_i4(i) = nint(w(i)* wscale)
      s_i4(i) = nint(ts(2,i)*stscale)
      t_i4(i) = nint(ts(1,i)*stscale)
    enddo
    if (dyn) then
      do i=nl,nu
        h(i)    = min(h_datamax, max(-h_datamax,h(i)))
        h_i4(i) = nint(h(i)*hscale)
      enddo
    endif
    
    call domp_get_domain (0, iwet2, nl, nu)

    do i=nl,nu
      z(i)    = min(datamax, max(-datamax,z(i)))
      z_i4(i) = nint(z(i)*zscale)
    enddo
  
  end subroutine wcompr_cmp

!===============================================================================

  subroutine compress_i4()

    use perm,             only : de_permute_omp, de_permute_2dvars,            &
                                 u_o, v_o, s_o, t_o, w_o, h_o, z_o
    use cmod_params,      only : narea, mmx, nmx, iw2, iw3, mm1
    use cmod_arrays,      only : u, v, cmp, h, z, wu, wv, kh, ind, mm1k, wrk
    use constants,        only : dyn

    implicit none

    integer(4) :: ia
    logical    :: do_alloc


    do ia=1,narea
      ! de-permute 3D vars:
      do_alloc = .true.
      if (allocated(u_o)) then
        do_alloc = .false.
        if (iw3(ia) > size(u_o(:)) - 1) then
          deallocate(u_o,v_o,s_o,t_o,h_o,w_o)
          do_alloc = .true.
        endif
      endif
      if (do_alloc) then
        allocate( u_o(0:iw3(ia)), v_o(0:iw3(ia)), w_o(0:iw3(ia)),              &
                  s_o(0:iw3(ia)), t_o(0:iw3(ia)) )
        if (dyn) then
          allocate( h_o(0:iw3(ia)) )
        else
          allocate( h_o(0:0) )
        endif
      endif
!$OMP PARALLEL DEFAULT (shared)
      call de_permute_omp( iw2(ia), mm1k(ia)%p, mm_op(ia)%p, ind(ia)%p,        &
                           kh(ia)%p, idx_op(ia)%p, cmp(ia)%p, t_o, s_o,        &
                           u(ia)%p, u_o, v(ia)%p, v_o, wrk(ia)%p, w_o )
      if(dyn) then
        call de_permute_omp( iw2(ia), mm1k(ia)%p, mm_op(ia)%p, ind(ia)%p,      &
                             kh(ia)%p, idx_op(ia)%p, h(ia)%p, h_o )
      endif
!$OMP END PARALLEL

      ! de-permute z:
      do_alloc = .true.
      if (allocated(z_o)) then
        do_alloc = .false.
        if (iw2(ia) > size(z_o(:)) - 1) then
          deallocate(z_o)
          do_alloc = .true.
        endif
      endif
      if (do_alloc) then
        allocate(z_o(0:iw2(ia)))
      endif
      call de_permute_2Dvars(mmx(ia),nmx(ia),mm1(ia)%p,z(ia)%p)

!$OMP PARALLEL DEFAULT (shared)
      call wcompr(iw2(ia), iw3(ia), u_o, v_o, h_o, w_o, z_o,                   &
                  u_i4(ia)%p, v_i4(ia)%p, h_i4(ia)%p, w_i4(ia)%p,              &
                  z_i4(ia)%p, s_o, t_o, s_i4(ia)%p, t_i4(ia)%p)
!$OMP END PARALLEL
    enddo

!$OMP PARALLEL DEFAULT (shared) PRIVATE (ia)
    do ia=1,narea
      call wcompr_w2p(iw2(ia), mm_op(ia)%p, ind(ia)%p,                         &
                      wu(ia)%p, wv(ia)%p, wu_i4(ia)%p, wv_i4(ia)%p)
    enddo
!$OMP END PARALLEL

  end subroutine compress_i4

!===============================================================================

  subroutine compress_i4_ice()

    use perm,             only : de_permute_2dvars,                            &
                                 ice_o, ts_o, ui_o, vi_o, casus_o
    use cmod_params,      only : narea, mmx, nmx, iw2, mm1
    use cmod_arrays,      only : ice, tsnei, ueis, veis, casus, iota
    use constants,        only : iotamodel

    implicit none

    integer(4) :: ia
    logical    :: do_alloc


    do ia=1,narea
      do_alloc = .true.
      if (allocated(ui_o)) then
        do_alloc = .false.
        if (iw2(ia) > size(ui_o(:)) - 1) then
          deallocate(ice_o,ts_o,ui_o,vi_o,casus_o)
          do_alloc = .true.
        endif
      endif
      if (do_alloc) then
        allocate(ice_o(1:3,0:iw2(ia)),ts_o(0:iw2(ia)),                         &
                 ui_o(0:iw2(ia)),vi_o(0:iw2(ia)),casus_o(0:iw2(ia))) 
      endif
      call de_permute_2Dvars(mmx(ia), nmx(ia), mm1(ia)%p, ice(ia)%p,           &
                             tsnei(ia)%p, ueis(ia)%p, veis(ia)%p, casus(ia)%p)

!$OMP PARALLEL DEFAULT (shared)
      call wcompr_e(iw2(ia), ice_o, ts_o, ui_o, vi_o, casus_o,                 &
                    eisdi_i4(ia)%p, eisco_i4(ia)%p, snowd_i4(ia)%p,            &
                    tsnei_i4(ia)%p, ueis_i4(ia)%p, veis_i4(ia)%p,              &
                    casus_out(ia)%p )
!$OMP END PARALLEL
    enddo

    if (iotamodel) then
      do ia=1,narea
        do_alloc = .true.
        ! recycle ice_o array for iota
        if (allocated(ice_o)) then
          do_alloc = .false.
          if (iw2(ia) > size(ice_o(1,:)) - 1) then
            deallocate(ice_o)
            do_alloc = .true.
          endif
        endif
        if (do_alloc) then
          allocate(ice_o(1:3,0:iw2(ia))) 
        endif
        call de_permute_2Dvars(mmx(ia),nmx(ia),mm1(ia)%p,iotamodel,iota(ia)%p)

!$OMP PARALLEL DEFAULT (shared)
        call wcompr_iota( iw2(ia),ice_o,iota1_i4(ia)%p,iota2_i4(ia)%p,         &
                    iota3_i4(ia)%p )
!$OMP END PARALLEL
      enddo
    end if

  end subroutine compress_i4_ice

!===============================================================================

  subroutine allocate_compressed_arrays ( )
  
    use cmod_params,    only : narea, iw2, iw3, mmx, nmx, kmx, mm1, nthreads
    use constants,      only : dyn

    implicit none
        
    integer(4) :: ia, n2d, n3d

    ia = narea
    allocate( u_i4(ia), v_i4(ia), w_i4(ia), h_i4(ia), z_i4(ia),                &
              s_i4(ia), t_i4(ia), wu_i4(ia), wv_i4(ia), mm_op(ia), idx_op(ia))
    do ia=1,narea
      n2d = iw2(ia)
      n3d = iw3(ia)
      allocate( u_i4(ia)%p(0:n3d), v_i4(ia)%p(0:n3d), w_i4(ia)%p(0:n3d),       &
                z_i4(ia)%p(0:n2d), s_i4(ia)%p(0:n3d), t_i4(ia)%p(0:n3d),       &
                wu_i4(ia)%p(0:n2d), wv_i4(ia)%p(0:n2d),                        &
                mm_op(ia)%p(0:mmx(ia)+1,0:nmx(ia)+1,1:kmx(ia)),                &
                idx_op(ia)%p(1:2,1:nthreads) )
      if (dyn) then
        allocate ( h_i4(ia)%p(0:n3d) )
      else
        allocate ( h_i4(ia)%p(0:0) )
      endif
      !  assign useful values for wind at land:
      wu_i4(ia)%p(0) = 0
      wv_i4(ia)%p(0) = 0
      !  save the old permutation for use with output:
      call set_mm_op( mmx(ia), nmx(ia), kmx(ia), mm1(ia)%p, mm_op(ia)%p,       &
                      idx_op(ia)%p )
    enddo
    
  end subroutine allocate_compressed_arrays

!===============================================================================

  subroutine allocate_compressed_ice_arrays ( )
  
    use cmod_params,    only : narea, iw2
    use constants,      only : iotamodel

    implicit none
        
    integer(4) :: ia, n2d

    ia = narea
     allocate( eisdi_i4(ia), eisco_i4(ia), snowd_i4(ia), tsnei_i4(ia),         &
               ueis_i4(ia), veis_i4(ia), casus_out(ia) )

     do ia=1,narea
       n2d = iw2(ia)
       allocate( eisdi_i4(ia)%p(0:n2d), eisco_i4(ia)%p(0:n2d),                 &
                 snowd_i4(ia)%p(0:n2d), tsnei_i4(ia)%p(0:n2d),                 &
                 ueis_i4(ia)%p(0:n2d), veis_i4(ia)%p(0:n2d),                   &
                 casus_out(ia)%p(0:n2d) )
     enddo

     if (iotamodel) then
      ia = narea
       allocate( iota1_i4(ia), iota2_i4(ia), iota3_i4(ia) )

       do ia=1,narea
         n2d = iw2(ia)
         allocate( iota1_i4(ia)%p(0:n2d), iota2_i4(ia)%p(0:n2d),               &
                 iota3_i4(ia)%p(0:n2d) )
       enddo
     endif  
    
  end subroutine allocate_compressed_ice_arrays

!===============================================================================

  subroutine set_mm_op( mmx, nmx, kmx, mm1, mm_op, idx_op )

    implicit none

    integer(4), intent(in)  :: mmx, nmx, kmx, mm1(0:,0:,1:)
    integer(4), intent(out) :: mm_op(0:,0:,1:), idx_op(1:,1:)

    integer(4) :: i,j,k,mo

    idx_op(1:,1:) = -1

    mo = 0
    do k=1,kmx
      do j=0,nmx+1
        do i=0,mmx+1
          if (mm1(i,j,k) > 0) then
            mo = mo + 1
            mm_op(i,j,k) = mo
          else
            mm_op(i,j,k) = 0
          endif
        enddo
      enddo
    enddo

  end subroutine set_mm_op

!===============================================================================

end module compress
