module constants


!-------------------------------------------------------------------------------
!
! This module holds physical constants, tidal contituents, correction
! terms, special water level points, as well as different time variables.
!
! Constants are set by calling subroutine konst_set
!
! This module replaces most common blocks in older versions of BSHcmod
!
!-------------------------------------------------------------------------------

 
  !- directives ----------------------------------------------------------------
  implicit none
  private

  !- constants -----------------------------------------------------------------
  real(8), parameter, public  :: g      =    9.81_8
  real(8), parameter, public  :: hgr    =    0.05_8
  real(8), parameter, public  :: hnull  =    0.1_8
  real(8), parameter, public  :: heps   = 0.00001_8
  real(8), parameter, public  :: rhoa   =    1.28_8
  real(8), parameter, public  :: rhow   = 1027.00_8
  real(8), parameter, public  :: rhoi   =  900.00_8
  real(8), parameter, public  :: rhos   =  327.50_8
  real(8), parameter, public  :: rhosoil= 2000.00_8
  real(8), parameter, public  :: saltmax=   42.00_8
  real(8), parameter, public  :: wtmpmax=   40.00_8
  real(8), parameter, public  :: cfreeze=   0.055_8
  real(8), parameter, public  :: c2k    = 273.0_8
  real(8), parameter, public  :: cpw    = 4180.00_8
  real(8), parameter, public  :: cpi    = 2200.00_8
  real(8), parameter, public  :: cpsoil = 1500.00_8
  real(8), parameter, public  :: cpa    = 1005.00_8
  real(8), parameter, public  :: Lwater = 2500.0e3_8
  real(8), parameter, public  :: Lice   =  335.0e3_8
  real(8), parameter, public  :: cdrag  =  0.0055_8
  real(8), parameter, public  :: pair   = 1013.00_8
  real(8), parameter, public  :: vkarm  =    0.40_8
  real(8), parameter, public  :: pi     = 3.14159265358979323846_8
  real(8), parameter, public  :: grarad = pi/180._8
  real(8), parameter, private :: zwopi  = 2.0_8*pi
  real(8), parameter, private :: sternt = 86164.0_8
  real(8), parameter, public  :: omega2 = 2.0_8*zwopi/sternt
  real(8), parameter, private :: rerde  = 6371040.0_8
  real(8), parameter, public  :: gmaequ = rerde*grarad
  real(8), parameter, public  :: zero = 0.0_8, one = 1.0_8, half = 0.5_8
  real(8), parameter, public  :: two = 2.0_8, three = 3.0_8, four = 4.0_8
  real(8), parameter, public  :: five = 5.0_8, six = 6.0_8, eight = 8.0_8
  real(8), parameter, public  :: ten = 10.0_8, nineteen = 19.0_8
  real(8), parameter, public  :: qrt = 0.25_8, tenth = 0.1_8
  real(8), parameter, public  :: centi = 0.01_8, milli = 0.001_8
  real(8), parameter, public  :: twenty = 20.0_8, twentyfour = 24.0_8
  real(8), parameter, public  :: onethird=1.0_8/3.0_8, twothird=2.0_8/3.0_8
  real(8), parameter, public  :: fourthird=4.0_8/3.0_8
  real(8), parameter, public  :: sph = 3600.0_8, kbyte = 1024.0_8
  real(8), parameter, public  :: iota0 = 100.0_8
  real(8), parameter, public  :: iscale = 100.0_8
  real(8), parameter, public  :: smfi = 10.0_8, smfk = 1.0_8
!FIXME: why hardcoded to 5' and 3' ?
  real(8), parameter, public  :: dx_iota=5.0_8/60.0_8*rerde*grarad
  real(8), parameter, public  :: dy_iota=3.0_8/60.0_8*rerde*grarad
  real(8), parameter, public  :: ddys=half/(iscale*dy_iota)
  real(8), parameter, public  :: ddyssmi=smfi*ddys
  real(8), parameter, public  :: ddyssmk=smfk*ddys

  !-----------------------------------------------------------------------------


  !- misc array dims -----------------------------------------------------------
  integer(4), parameter, public :: nicepar    = 3
  integer(4), parameter, public :: niotapar   = 3
  !-----------------------------------------------------------------------------


  !- tidal constituents --------------------------------------------------------
  integer(4), parameter, public      :: ntides = 17
  real(8), save, public, allocatable :: hg(:,:), gg(:,:)
  real(8), save, public              :: omg(ntides), vnpv(ntides), zj(ntides)
  logical, save, public              :: mzwei
  !-----------------------------------------------------------------------------


  !- Some model specific parameters --------------------------------------------
  real(8), parameter, public :: CSMAG = 0.2_8
  real(8), parameter, public :: CDIV  = (15.0_8)**2
  real(8), save,      public :: zref  = 0.0_8
  ! wind drag in neutral situation:
  !     Cd = ( CDA + CDB*W + CDC*W**2 )*0.001
  !   Kara et al:
  !     Cd = ( 0.862 + 0.088*W - 0.0089*W**2 )*0.001
  !   Original BSH:
  real(8), save,      public :: CDA   = 0.5_8
  real(8), save,      public :: CDB   = 0.07_8
  real(8), save,      public :: CDC   = 0.0_8
  !-----------------------------------------------------------------------------


  !- Some heat exchange and turbulence parameters ------------------------------
  logical, save,      public :: qflx = .false., ct10m = .false.
  ! intercept T and S, approximately:
  real(8), parameter, public :: Tintercept =  4.0_8
  real(8), parameter, public :: Sintercept = 24.7_8
  ! T_max_density(S) = -(cfreeze + Tintercept/Sintercept) * S  + Tintercept
  real(8), parameter, public :: cmaxdens = cfreeze + Tintercept/Sintercept
  ! exchange coeff. for latent heat in neutral situation:
  !     Cl = ( CSA + CSB*W + CSC*W**2 )*0.001
  !   Kara et al:
  !     Cl = ( 0.994 + 0.061*W - 0.001*W**2 )*0.001
  !   Original BSH:
  real(8), save,      public :: CLA = 1.2_8
  real(8), save,      public :: CLB = 0.0_8
  real(8), save,      public :: CLC = 0.0_8
  ! exchange coeff. for sensible heat:
  !   Kara et al:
  !     Cs = 0.96*Cl
  !   Original BSH:
  real(8), save,      public :: CSCL = 1.0_8
  !-----------------------------------------------------------------------------


  !- Exchange coefficients, stability dependent --------------------------------
  !   CD1 = ( CD1A + CD1B*W + CD1C*W**2 ) * 0.001
  !   CL1 = ( CL1A + CL1B/W + CL1C/W**2 ) * 0.001
  !   C[DL] = C[dl] + C[DL]1 * (Tsea -Tair)
  !  Kara et al use:
  !   CD1[A,B,C] = 0.001*[ 0.1034, -0.00678, 0.0001147]
  !   CL1[A,B,C] = 0.001*[-0.020,   0.691,  -0.817]
  real(8), save, public :: CD1A = 0.0_8
  real(8), save, public :: CD1B = 0.0_8
  real(8), save, public :: CD1C = 0.0_8
  real(8), save, public :: CL1A = 0.0_8
  real(8), save, public :: CL1B = 0.0_8
  real(8), save, public :: CL1C = 0.0_8
  !-----------------------------------------------------------------------------


  !- dynamical vertical co-ordinates parameters --------------------------------
  real(8), save, public :: wghtz    =    0.000000001_8          !  without unit
  real(8), save, public :: wghthar  =    0.25_8                 !  unit: [m/s]
  real(8), save, public :: wghteta  = 1000.0_8                  !  unit: [m/s]
  real(8), save, public :: wghtturb =    0.0_8                  !  unit: [1/m]
  real(8), save, public :: gammadt  =    0.00003_8              !  unit: [1/s]
  !-----------------------------------------------------------------------------


  !- time step information -----------------------------------------------------
  integer(4),    save, public :: izeits, jahra
  real(8),       save, public :: rzeite, zzeits, tim, zeitstu
  character(11), save, public :: aufdat, starta
  !-----------------------------------------------------------------------------


  !- optional parameter list ---------------------------------------------------
  real(8),    save, public :: toutput, filesizemax = 2.0_8, dtwlfile, Q0
  integer(4), save, public :: itsfields, iptracers, nclss
  integer(4), save, public :: zbnd_type, pen_type, itmax
  integer(4), save, public :: nproci, nprocj, niceact
  logical,    save, public :: ionml
  logical,    save, public :: only_islices, omp_loadbalance_experiment=.false.
  logical,    save, public :: dyn, elevation_ref, weirs
  logical,    save, public :: coldst, spinup, archive, ldebug
  logical,    save, public :: setTSbnd, isponge
  logical,    save, public :: updateSTI
  logical,    save, public :: usezbcl, coriolis
  logical,    save, public :: kalman, sstrs, precevap
  logical,    save, public :: printturb, tom, init_only
  logical,    save, public :: ReadNOA, icedynactive, shortcircuit
  logical,    save, public :: iotamodel, iotars, fastice_test
  logical,    save, public :: snow
  logical,    save, public :: coupler_active, coupler_sponge
  logical,    save, public :: coupler_psponge, calctrans, botfit, simplebfmap
  !-----------------------------------------------------------------------------


  !- routines/functions --------------------------------------------------------
  public :: konst_set


contains


  subroutine konst_set( mrmain )
  !-----------------------------------------------------------------------------
  !
  ! Subroutine for reading the harmonic tidal constants.
  ! Other physical constants and parameters are set in the statements above.
  !
  !-----------------------------------------------------------------------------

    !- modules -----------------------------------------------------------------
    use exits,   only : exitme
    use io_subs, only : io_new_unit


    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: mrmain


    !- local vars --------------------------------------------------------------
    character(10), parameter :: fnam  = 'tide_const'
    character(11)            :: cfrmt, cfrma, cfrm2
    character(2)             :: ctide
    integer(4)               :: i, j, l, iblock, jahrk, lun, ios


    !- don't do anything if we have no OBC points ------------------------------
    if (mrmain == 0) return

    !- allocate arrays ---------------------------------------------------------
    allocate( hg(ntides,mrmain), gg(ntides,mrmain) )


    !- read the harmonic constants ---------------------------------------------
    lun = io_new_unit()
    open(lun,file=fnam,form='formatted',status='old',iostat=ios) 
    if (ios /= 0) call exitme(1,'Cannot open '//fnam)

    read(lun,*)

    do i=1,ntides 
      read(lun,'(f15.11)',iostat=ios) omg(i)
      if (ios /= 0) call exitme(2,'Cannot read '//fnam)
    enddo

    read(lun,*)
    read(lun,*)
    read(lun,*)

    write(ctide,'(i2)') ntides       !<== hard coded, max 99 tidal constituents
                                     !    Let's hope that's sufficient ;-)
    cfrmt = '(3x,'//trim(ctide)//'f7.2)'
    cfrma = '(3x,'//trim(ctide)//'f7.3)'
    cfrm2 = '(4x,'//trim(ctide)//'f7.2)'

    do i=1,mrmain 
      read(lun,cfrma,iostat=ios) (hg(j,i),j=1,ntides) 
      if (ios /= 0) call exitme(3,'Cannot read '//fnam)
    enddo

    read(lun,*)
    read(lun,*)
    read(lun,*)

    do i=1,mrmain 
      read(lun,cfrmt,iostat=ios) (gg(j,i),j=1,ntides) 
      if (ios /= 0) call exitme(4,'Cannot read '//fnam)
    enddo

    read(lun,*)
    read(lun,*)
    read(lun,*)

    do iblock=1,2000 
      read(lun,'(1x,i4)',iostat=ios) jahrk 
      if (ios /= 0) call exitme(5,'Cannot find year for nodal correction !')
      read(lun,cfrmt,iostat=ios) vnpv 
      if (ios /= 0) call exitme(6,'Cannot read '//fnam)
      read(lun,cfrm2,iostat=ios) zj 
      if (ios /= 0) call exitme(7,'Cannot read '//fnam)
      if (jahrk == jahra) then 
        exit 
      endif 
    enddo

    close(lun) 

    do l=1,ntides 
      omg(l)=omg(l)*grarad 

      if (mzwei) then
        vnpv(l) = zero
        if (l == 1) then
          zj(l) = one
        else
          zj(l) = zero
        endif
      endif   

      do j=1,mrmain 
        gg(l,j) = (vnpv(l)-gg(l,j))*grarad 
        hg(l,j) = hg(l,j)*zj(l) 
      enddo
    enddo
    !---------------------------------------------------------------------------


    !- That's it! --------------------------------------------------------------

  end subroutine konst_set

end module constants
