module momeqs_wrapping
  use constants, only : zero, one, two, half, qrt

  implicit none
  private

  public  :: momeqs
  ! HACK TO ENSURE INLINING - IPA and craypat doesn't work, cf. trisolver.f90
  private :: trisolv

contains 

! ------------------------------------------------------------------------------
!
! momeqs: sets up and solves the momentum equations using a tridiagonal solver
!
!         a(k)*s(k-1) + b(k)*s(k) + c(k)*s(k+1) = d(k)
!
! Based on the same numerical scheme as the original bewgl, but implementation 
! of the method into the actual code is modified completely for clairity, 
! simplicity and very much indeed for performance efficiency.
! ------------------------------------------------------------------------------
subroutine momeqs (kdry, iwet2, mm, dt, ft, rt,                                &
                   ty, tx, gty, gtx, nud, nvd, uvdam, nweir, z, u, v, hz,      &
                   un, vn, hx, hy, khu, khv, taux, tauy, pl, press, geopot,    &
                   rho, avv, usmag, vsmag,                                     &
                   casus, ice, ueis, veis, ind, kh, z0srf, idx, widx, zweir,   &
                   wflowr, ia, botfricmap)
 
  !- modules -------------------------------------------------------------------
  use dmi_omp,      only : domp_get_domain
  use cmod_params,  only : kmax
  use constants,    only : dyn, weirs, hgr, rhow, rhoi, cdrag 
  use drying_cases, only : drying_case
  use weireqs,      only : weir
  use botfricmaps,  only : get_simple_bf_map

  ! ----------------------------------------------------------------------------
  !    g:     gravity
  !    hgr:   thickness limit (0.05m)
  !    rhow:  reference density of sea water
  !    rhoi:  reference density of sea ice
  !    cdrag: drag coefficient
  !    kmax:  max No of layers in all areas
  ! ----------------------------------------------------------------------------

  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in) :: kdry       ! number of layers which could fall dry
  integer(4), intent(in) :: iwet2      ! number of wet points in me
  integer(4), intent(in) :: mm(:,0:,0:)! index
  real(8), intent(in)    :: dt         ! time step size
  real(8), intent(in)    :: ft(:)      ! ft(i):f(i)*dt, f(i) = Coriolis param
  real(8), intent(in)    :: rt         ! rt: r*dt, r = bed roughness
  real(8), intent(in)    :: ty         ! ty: dt/dy
  real(8), intent(in)    :: tx(:)      ! tx(i): dt/dx(i)
  real(8), intent(in)    :: gty        ! gty:g*dt/dy
  real(8), intent(in)    :: gtx(:)     ! gtx(i):g*dt/dx(i)
  integer(4), intent(in) :: nud        ! No. of x-dams
  integer(4), intent(in) :: nvd        ! No. of y-dams
  real(8),    intent(in) :: uvdam(1:,0:) ! 1.0/0.0 mask for dams
  integer(4), intent(in) :: nweir      ! No. of weirs
  integer(4), intent(in) :: widx(:,0:) ! weir-index
  real(8), intent(in)    :: zweir(0:)  ! congestion height of the weir
  real(8), intent(in)    :: wflowr(0:) ! flow rate of the weir
  real(8), intent(in)    :: z(0:)      ! surface elevation
  real(8), intent(in)    :: u(0:)      ! u at old time level
  real(8), intent(in)    :: v(0:)      ! v at old time level
  real(8), intent(in)    :: hz(0:)     ! layer thickness
  real(8), intent(in)    :: hx(0:)     ! layer thickness @ x-mom point
  real(8), intent(in)    :: hy(0:)     ! layer thickness @ y-mom point
  integer(4), intent(in) :: kh(0:)     ! No of layers
  integer(4), intent(in) :: ind(:,:)   ! i,j indices
  integer(4), intent(in) :: khu(0:)    ! No of x-mom layers
  integer(4), intent(in) :: khv(0:)    ! No of y-mom layers
  real(8), intent(in)    :: taux(0:)   ! wind stress, x-dir
  real(8), intent(in)    :: tauy(0:)   ! wind stress, y-dir
  real(8), intent(in)    :: pl(0:)     ! air pressure
  logical, intent(in)    :: casus(0:)  ! water-ice mask
  real(8), intent(in)    :: ice(1:,0:) ! ice parameters
  real(8), intent(in)    :: ueis(0:)   ! ice velocity in x-dir
  real(8), intent(in)    :: veis(0:)   ! ice velocity in y-dir 
  real(8), intent(in)    :: press(0:)  ! baroclinic pressure
  real(8), intent(in)    :: geopot(0:)
  real(8), intent(in)    :: rho(0:)    ! denisty
  real(8), intent(in)    :: avv(0:)    ! vertical eddy viscosity
  real(8), intent(in)    :: usmag(0:)  ! Smagorinsky terms, u-eq
  real(8), intent(in)    :: vsmag(0:)  ! Smagorinsky terms, v-eq
  real(8), intent(in)    :: z0srf(0:)  ! surface roughness length scale
  integer(4), intent(in) :: ia
  real(8), intent(in)    :: botfricmap(0:)

  !- Output variables: ---------------------------------------------------------
  !  unmodified locations must have their values maintained, therefore INOUT
  real(8), intent(inout) :: un(0:)     ! u at new time level (0:iwet3)
  real(8), intent(inout) :: vn(0:)     ! v at new time level (0:iwet3)
  integer(4), intent(inout) :: idx(:,:)

  !- local vars ----------------------------------------------------------------
  integer(4) :: i, j, kev, keu, k
  integer(4) :: mi, mie, eee, sse, ee, ww, nn, ss, ne, se, sw
  integer(4) :: nl, nu, nsurf, i1, i2, i3, i4, j1, j2, j3, j4
  integer(4) :: mmi, mme, mms
  real(8)    :: eiscol, eiscor, eiscox, eiscoo, eiscou, eiscoy
  real(8)    :: diffu, difful, diffur, diffvl, diffvr, diffuv
  real(8)    :: diffv, diffvo, diffvu, diffuo, diffuu
  real(8)    :: umi, uzet, ulu, usrf, vst
  real(8)    :: vmi, vzet, vlu, vsrf, ust
  real(8)    :: ustar, vaus, vcon, vpress
  real(8)    :: vstar, uaus, ucon, upress
  real(8)    :: urd, vrd, aurd, avrd
  real(8)    :: dtd, dtdr, txi, fti, rth, hrt, ttx, tty
  real(8)    :: fac, dec, zu, zr, zm, tmp, dam

  real(8), dimension(kmax) :: a, b, c, d, s, tm

  real(8), parameter :: laplacescale = 100000.0_8
  real(8), parameter :: cdri = cdrag*rhoi, drhow  = one/rhow

  !- intrinsic  ----------------------------------------------------------------
  intrinsic  :: sqrt, min, max, abs

  !- initialise on every entry -------------------------------------------------
  dtd   = dt/laplacescale
  dtdr  = dt/rhow

  ! split the remaining outer loop among the threads ---------------------------
  call domp_get_domain(kh, 1, iwet2, nl, nu, idx)

  do nsurf=nl,nu
    i = ind(1,nsurf)
    j = ind(2,nsurf)

    ! possibly get the bottom friction from a map:
    rth = rt*half
    call get_simple_bf_map(ia,nsurf,botfricmap,rt,rth)

    !---- u-equation  ----------------------------------------------------------
    mi  = mm(1,i,j)
    keu = khu(mi)
    if (keu == 0) then
      ! land east of point, maintain zero velocity:
      un(mi) = zero

    else
      ! one or more layers
        
      ! obtain and store surface indices:
      !       don't change mie and eee below.
      mie = mi
      eee = mm(1,i,j+1)

      ! extract layer thickness:  
      tm(1) = hx(mie)
      if (keu > 1) then
        mmi = mm(2,i,j  )
        mme = mm(2,i,j+1)
        tm(2:keu) = hx(mmi:mmi+keu-2)
      endif

      ! set up the vertical eddy viscosity terms, coefficients a, b and c:
      a(1) = zero
      do k=1,keu-1
        tmp    = one/tm(k)
        a(k+1) = -dt*(avv(mmi+k-1)+avv(mme+k-1))/(tm(k)+tm(k+1))
        a(k)   = a(k)  *tmp
        c(k)   = a(k+1)*tmp
        b(k)   = one - (a(k) + c(k))
      enddo
      a(keu) = a(keu)/tm(keu)
      b(keu) = one - a(keu)
      c(keu) = zero
      
      ! air pressure term:
      ulu = (pl(eee)-pl(mie))*tx(i)*drhow

      ! zeta:
      zm = z(mie)
      zr = z(eee)

      ! pressure term:
      uzet = gtx(i)*(zr-zm)

      ! collect different surface terms:
      usrf = ulu + uzet

      ! do each layer, coefficient d:
      txi = tx(i)
      fti = ft(i)
      ttx = two*txi

      do k=1,keu
        ! extract commonly used indices:
        mi = mm(k,i  ,j  )
        nn = mm(k,i-1,j  )
        ss = mm(k,i+1,j  )
        ne = mm(k,i-1,j+1)
        ee = mm(k,i  ,j+1)
        se = mm(k,i+1,j+1)

        ! center value of u:
        umi = u(mi)

        ! v centred @ u point:
        vst  = qrt*(v(nn) + v(ne) + v(mi) + v(ee))

        ! baroclinic pressure:
        upress = (press(ee)-press(mi))*ttx/(rho(ee)+rho(mi))

        ! convective terms using vector upwind scheme:
        urd  = umi*txi
        vrd  = vst*ty
        aurd = abs(urd)
        avrd = abs(vrd)

        if (urd >= zero) then
          j2 = j-1
        else
          j2 = j+1
        endif
        if (vrd >= zero) then
          i4 = i+1
        else
          i4 = i-1
        endif
        if (aurd >= avrd) then
          i1 = i
          j3 = j2
        else
          i1 = i4
          j3 = j
        endif

        ucon = aurd*(u(mm(k,i1,j))-u(mm(k,i1,j2)))                             &
             + avrd*(u(mm(k,i,j3))-u(mm(k,i4,j3)))

        ! Laplacian smoothing term using full-slip condition for lateral terms:
        uaus = u(ee) + u(mm(k,i,j-1)) - two*umi
        if (ne > 0 .and. nn > 0) uaus = uaus + u(nn) - umi
        if (se > 0 .and. ss > 0) uaus = uaus + u(ss) - umi

        ! Horizontal eddy viscosity:
        uaus = dtd*uaus + usmag(mi)

        ! RHS of x-mom-eq stored in d:
        d(k) = umi + fti*vst - ucon + uaus - upress - usrf
      enddo
      
      if (dyn) then
        !- The geopotential term
        do k = 1,keu
          mi = mm(k,i  ,j  )
          ee = mm(k,i  ,j+1)
          d(k) = d(k) - (geopot(ee)-geopot(mi))*txi
        enddo
      endif

      ! bottom friction:
      vstar  = vst
      hrt    = tm(keu) + hgr
      ustar  = u(mm(keu,i,j))
      tmp    = rth/hrt*sqrt(ustar*ustar + vstar*vstar)
      d(keu) = d(keu) - tmp*ustar
      b(keu) = b(keu) + tmp

      ! wind and possibly ice drag applied at the surface boundary:
      tmp = half*(z0srf(mie) + z0srf(eee))
      if (casus(mie) .and. casus(eee)) then
        ! ice free; include wind drag:
        d(1) = d(1) + taux(mie)*dtdr/(tm(1)+tmp)
      else
        ! ice covered; reduce wind drag and include ice drag:
        eiscol = min(max(ice(2,mie),zero),one)
        eiscor = min(max(ice(2,eee),zero),one)
        eiscox = half*(eiscol+eiscor)
        diffu  = ueis(mie)-u(mie)
        difful = diffu + ueis(mm(1,i,j-1)) - u(mm(1,i,j-1))
        diffur = diffu + ueis(eee)         - u(eee)
        diffvl = veis(mm(1,i-1,j  )) - v(mm(1,i-1,j  )) + veis(mie) - v(mie)
        diffvr = veis(mm(1,i-1,j+1)) - v(mm(1,i-1,j+1)) + veis(eee) - v(eee)
        diffuv = qrt*( eiscol*sqrt(difful**2+diffvl**2)                        &
                      +eiscor*sqrt(diffur**2+diffvr**2))

        fac = dtdr/(tm(1)+tmp)
        dec = half*cdri*diffuv*fac

        d(1) = d(1) + (cdri*diffuv*diffu+(one-eiscox)*taux(mie))*fac           &
                    + dec*u(mie)
        b(1) = b(1) + dec
      endif

      ! solve the system:
!dir$ inline
      call TriSolv(s, a, b, c, d, keu)
      un(mie) = s(1)
      if (keu > 1) un(mmi:mmi+keu-2) = s(2:keu)

      ! handle a drying situation:
      if (keu <= kdry) then
        call drying_case(mie,eee,un(mie),hz(mie),hz(eee),z)
        if (kdry > 1) then
          do k = 2,min(keu,kdry)
            call drying_case(mie,eee,un(mmi+k-2),hz(mmi+k-2),hz(mme+k-2),z)
          enddo
        endif
      endif

      ! handle u-dams
      if (nud > 0) then
        dam = uvdam(1,mie)
        un(mie) = un(mie)*dam
        if (keu > 1) un(mmi:mmi+keu-2) = un(mmi:mmi+keu-2)*dam
      endif
    endif
    !----------------- end of u-equation ---------------------------------------

    !----------------- v-equation ----------------------------------------------
    mi  = mm(1,i,j)
    kev = khv(mi)
    if (kev == 0) then
      ! land south of point, maintain zero velocity:
      vn(mi) = zero

    else
      ! one or more layers

      ! obtain and store surface indices:
      !       don't change mie and sse below.
      mie = mi
      sse = mm(1,i+1,j)

      ! extract layer thickness:
      tm(1) = hy(mie)
      if (kev > 1) then
        mmi = mm(2,i,  j)
        mms = mm(2,i+1,j)
        tm(2:kev) = hy(mmi:mmi+kev-2)
      endif

      ! set up the vertical eddy viscosity terms, coefficients a, b and c:
      a(1) = zero
      do k=1,kev-1
        tmp    = one/tm(k)
        a(k+1) = -dt*(avv(mmi+k-1)+avv(mms+k-1))/(tm(k)+tm(k+1))
        a(k)   = a(k)  *tmp
        c(k)   = a(k+1)*tmp
        b(k)   = one - (a(k) + c(k))
      enddo
      a(kev) = a(kev)/tm(kev)
      b(kev) = one - a(kev)
      c(kev) = zero

      ! air pressure term:
      vlu = (pl(mie)-pl(sse))*ty*drhow 

      ! zeta:
      zm = z(mie)
      zu = z(sse)

      ! pressure term:
      vzet = gty*(zm-zu)

      ! collect different surface terms:
      vsrf = vlu + vzet

      ! do each layer, coefficient d:
      txi = tx(i)
      fti = ft(i)
      tty = two*ty

      do k=1,kev
        ! extract commonly used indices:
        mi = mm(k,   i,j )
        ss = mm(k,i+1,j  )
        ee = mm(k,i  ,j+1)
        se = mm(k,i+1,j+1)
        ww = mm(k,i,  j-1)
        sw = mm(k,i+1,j-1)

        ! center value of v:
        vmi = v(mi)

        ! u centred @ v point:
        ust = qrt*(u(ww) + u(mi) + u(sw) + u(ss))

        ! baroclinic presssure:
        vpress = (press(mi)-press(ss))*tty/(rho(mi)+rho(ss))

        ! convective terms using vector upwind scheme:
        urd  = ust*txi
        vrd  = vmi*ty
        aurd = abs(urd)
        avrd = abs(vrd)

        if (vrd >= zero) then
          i2 = i+1
        else
          i2 = i-1
        endif
        if (urd >= zero) then
          j4 = j-1
        else
          j4 = j+1
        endif
        if (avrd >= aurd) then
          i3 = i2
          j1 = j
        else
          i3 = i
          j1 = j4
        endif

        vcon = avrd*(v(mm(k,i,j1))-v(mm(k,i2,j1)))                             &
             + aurd*(v(mm(k,i3,j))-v(mm(k,i3,j4)))

        ! Laplacian smoothing term using full-slip condition for lateral terms:
        vaus = v(mm(k,i-1,j)) + v(ss) - two*vmi
        if (ee > 0 .and. se > 0) vaus = vaus + v(ee) - vmi
        if (ww > 0 .and. sw > 0) vaus = vaus + v(ww) - vmi

        ! Horizontal eddy viscosity:
        vaus = dtd*vaus + vsmag(mi)

        !- RHS of y-mom-eq stored in d:
        d(k) = vmi - fti*ust - vcon + vaus - vpress - vsrf
      enddo

      if (dyn) then
        !- The geopotential term
        do k=1,kev
          mi = mm(k,i  ,j  )
          ss = mm(k,i+1,j  )
          d(k) = d(k) - (geopot(mi)-geopot(ss))*ty
        enddo
      endif

      ! bottom friction:
      ustar  = ust
      hrt    = tm(kev)+hgr
      vstar  = v(mm(kev,i,j))
      tmp    = rth/hrt*sqrt(vstar*vstar + ustar*ustar) 
      d(kev) = d(kev) - tmp*vstar
      b(kev) = b(kev) + tmp

      ! wind and possibly ice drag applied at the surface boundary:
      tmp = half*(z0srf(mie) + z0srf(sse))
      if (casus(mie) .and. casus(sse)) then
        ! ice free; include wind drag:
        d(1) = d(1) + tauy(mie)*dtdr/(tm(1)+tmp)
      else
        ! ice covered; reduce wind drag and include ice drag:
        eiscoo = min(max(ice(2,mie),zero),one)
        eiscou = min(max(ice(2,sse),zero),one)
        eiscoy = half*(eiscoo+eiscou)
        diffuo = ueis(mm(1,i,  j-1)) - u(mm(1,i,  j-1)) + ueis(mie) - u(mie)
        diffuu = ueis(mm(1,i+1,j-1)) - u(mm(1,i+1,j-1)) + ueis(sse) - u(sse)
        diffv  = veis(mie)-v(mie)
        diffvo = diffv + veis(mm(1,i-1,j)) - v(mm(1,i-1,j))
        diffvu = diffv + veis(sse)         - v(sse)
        diffuv = qrt*( eiscoo*sqrt(diffuo**2+diffvo**2)                        &
                      +eiscou*sqrt(diffuu**2+diffvu**2))

        fac = dtdr/(tm(1)+tmp)
        dec = half*cdri*diffuv*fac

        d(1) = d(1) + (cdri*diffuv*diffv+(one-eiscoy)*tauy(mie))*fac           &
                    + dec*v(mie)
        b(1) = b(1) + dec
      endif

      ! solve the system -------------------------------------------------------
!dir$ inline
      call TriSolv(s, a, b, c, d, kev)
      vn(mie) = s(1)
      if (kev > 1) vn(mmi:mmi+kev-2) = s(2:kev)

      ! handle a drying situation:
      if (kev <= kdry) then
        call drying_case(sse,mie,vn(mie),hz(sse),hz(mie),z)
        if (kdry > 1) then
          do k = 2,min(kev,kdry)
            call drying_case(sse,mie,vn(mmi+k-2),hz(mms+k-2),hz(mmi+k-2),z)
          enddo
        endif
      endif

      ! handle v-dams
      if (nvd > 0) then
        dam = uvdam(2,mie)
        vn(mie) = vn(mie)*dam
        if (kev > 1) vn(mmi:mmi+kev-2) = vn(mmi:mmi+kev-2)*dam
      endif
    endif
    !----------------- end of v-equation ---------------------------------------
  enddo ! n-loop

  !------------------- include the weirs (if any) ------------------------------
  if (weirs) then
!$OMP BARRIER
!$OMP MASTER
  ! usually too small to do in parallel
    call weir (nweir, widx, mm, ty, tx, z, zweir, wflowr, un, vn)  
!$OMP END MASTER
  endif

  !- That's all folks! ---------------------------------------------------------

end subroutine momeqs

!===============================================================================

  subroutine TriSolv(s, a, b, c, d, n)
  !-----------------------------------------------------------------------------
  !  standard tridiagonal solver
  !-----------------------------------------------------------------------------

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: n
    real(8), intent(out)   :: s(1:)
    real(8), intent(in)    :: a(1:), b(1:), c(1:), d(1:)

    !-local vars ---------------------------------------------------------------
    integer(4) :: i
    real(8)    :: e
    real(8)    :: f(n)

    !- forward sweep -----------------------------------------------------------
    e    = b(1)
    s(1) = d(1)/e
    do i = 2, n
      f(i) = c(i-1)/e
      e    = b(i) - a(i)*f(i)
      s(i) = (d(i) - a(i)*s(i-1))/e
    enddo

    !- backward sweep ----------------------------------------------------------
    do i = n-1, 1, -1
      s(i) = s(i) - f(i+1)*s(i+1)
    enddo

    !- That's it! --------------------------------------------------------------

  end subroutine TriSolv

end module momeqs_wrapping

! ------------------------------------------------------------------------------
