module output
  
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !!! Everything regarding the default-output (The output of the original    !!!
  !!! DMIcmod) of the HIROMB-BOOS-Model is handled in this module. From this !!!
  !!! module the module output.f90 is generated.                             !!!
  !!! If an institute would like to have another kind of output, it has to   !!!
  !!! write another output-file (like BSH-output.f90), which should generate !!!
  !!! output.f90. This could then be realised by an --enable-[someth.]-      !!!
  !!! instruction when configuring the code.                                 !!!
  !!! To assure that the code is still running with another output, the      !!!
  !!! interfaces of the three subroutines output_init, output_write and      !!!
  !!! output_close have to be the same in every output-module.               !!!
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


  implicit none
  private
  
  public :: output_init, output_close, output_write

  logical, private, save :: wlstat_out = .false.
  
  contains
  
    subroutine output_init(aufdat,ctim,tim)
      ! Output-files will be opend in this file
    
      use cmod_miscvars,      only : itime, tidal
      use constants,          only : calctrans
      use dmi_mpi,            only : mpi_io_serial, ice_serial, dmpi_broadcast
      use io_control,         only : io_server, ioc_open
      use transport_variables,only : transdat_new
      use wlstat_wrap,        only : wlstat 
      use writetempdat,       only : tempdat_new, icefile_new 

      implicit none
    
      character(11), intent(in)           :: aufdat
      character(19), intent(in)           :: ctim  !<==  watch out here!
      real(8),       intent(in)           :: tim

      if (io_server) then
        call ioc_open() 
        return
      endif
 
      if (mpi_io_serial) then
        !- open binary HD output file w/start time as header -------------------
        call tempdat_new ()

        !- open ascii sea level time series files, read station list file ------
        call wlstat(0,ctim,tim,itime,wlstat_out)

        !- open binary file for inline transport output ------------------------
        if (calctrans) call transdat_new ()
      endif
      !- broadcast the value of wlstat_out:
      call dmpi_broadcast(wlstat_out)

      if (ice_serial) then
        !-  open ascii ice parameter output file -------------------------------
        if (.not.tidal) call icefile_new (aufdat)
      endif

    end subroutine output_init
  
  !============================================================================!
  
  subroutine output_write(izeits,aufdat,ctim,tim)
    ! Output-files will be closed in this file
    
    use cmod_arrays,        only : u, v, h, z, wu, wv, cmp, tke, diss, avv,    &
                                   dispv, casus, ice, tsnei, ueis, veis, wrk,  &
                                   iota
    use cmod_local_arrays,  only : u_l, v_l, h_l, wi_l, z_l, wu_l, wv_l, cmp_l,&
                                   tke_l, diss_l, avv_l, dispv_l, casus_l,     &
                                   ice_l, tsnei_l, ueis_l, veis_l, iota_l
    use cmod_miscvars,      only : izeite, itime, lmn1, lmn2, lmne, tidal
    use cmod_params,        only : narea, kmx
    use compress,           only : compress_i4, compress_i4_ice
    use constants,          only : printturb, dyn, calctrans, iotamodel
    use dmi_mpi,            only : mpi_io_serial, dmpi_gather
    use io_control,         only : io_server, ioc_write
    use dmi_mpi_global,     only : mpi_io_rank
    use transports,         only : transport_output
    use transport_variables,only : NDTTRANSOUT, transdat_new
    use turbout_wrap,       only : turbout
    use wlstat_wrap,        only : wlstat
    use writetempdat,       only : tempdat_next, write_icefile, tempdat_write
 
    implicit none
    
    integer(4), intent(in)       :: izeits
    character(11), intent(in)    :: aufdat
    character(19), intent(in)    :: ctim  !<==  watch out here!
    real(8),       intent(in)    :: tim
    
    integer(4)                   :: ia       

    if (io_server) then
      call ioc_write(izeits)
      return 
    endif

    !- flow field varibles to tempdat file, etc.:
    if (mod(izeits,lmn2) == 0) then

      !- Gather to I/O task:
      do ia=1,narea
        call dmpi_gather(ia,kmx(ia),  u_l(ia)%p,  u(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,kmx(ia),  v_l(ia)%p,  v(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,kmx(ia), wi_l(ia)%p,wrk(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,(1),      z_l(ia)%p,  z(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,(1),     wu_l(ia)%p, wu(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,(1),     wv_l(ia)%p, wv(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,kmx(ia),cmp_l(ia)%p,cmp(ia)%p,2,mpi_io_rank)
      enddo
      if (dyn) then
        do ia=1,narea
          call dmpi_gather(ia,kmx(ia),h_l(ia)%p,h(ia)%p,mpi_io_rank)
        enddo
      endif

      if (mpi_io_serial) then
        !- first compress to i4:
        call compress_i4 ()

        !- output to tempdat file:
        call tempdat_write (ctim,tim,aufdat,izeits,lmn2)
      endif

      !- output turbulence profiles:
      if (printturb) then 
        do ia=1,narea
          call dmpi_gather(ia,kmx(ia),  tke_l(ia)%p,  tke(ia)%p,mpi_io_rank)
          call dmpi_gather(ia,kmx(ia), diss_l(ia)%p, diss(ia)%p,mpi_io_rank)
          call dmpi_gather(ia,kmx(ia),  avv_l(ia)%p,  avv(ia)%p,mpi_io_rank)
          call dmpi_gather(ia,kmx(ia),dispv_l(ia)%p,dispv(ia)%p,2,mpi_io_rank)
        enddo
        if (mpi_io_serial)  call turbout(ctim)
      endif
    endif

    !- output of ice vars:
    if ((.not.tidal) .and. mod(izeits,lmn1) == 0) then
      do ia=1,narea
        call dmpi_gather(ia,    casus_l(ia)%p,casus(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,(1),  ice_l(ia)%p,  ice(ia)%p,3,mpi_io_rank)
        call dmpi_gather(ia,(1), ueis_l(ia)%p, ueis(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,(1), veis_l(ia)%p, veis(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,(1),tsnei_l(ia)%p,tsnei(ia)%p,mpi_io_rank)
      enddo
      if (iotamodel) then
        do ia=1,narea
          call dmpi_gather(ia,(1), iota_l(ia)%p, iota(ia)%p,3,mpi_io_rank)
        enddo
      end if

      if (mpi_io_serial) then
        !- first compress to i4:
        call compress_i4_ice ()

        !- ice data to file:
        call write_icefile (ctim,tim)
      endif
    endif

    !- ascii sea level time series files:
    if (mod(izeits,lmne) == 0 .and. wlstat_out) then
      if (mod(izeits,lmn2) /= 0) then
        do ia=1,narea
          call dmpi_gather(ia,(1),z_l(ia)%p,z(ia)%p,mpi_io_rank)
        enddo
      endif
      if (mpi_io_serial) call wlstat(1,ctim,tim,itime)
    endif

    if (calctrans) then
      if (mod(izeits, NDTTRANSOUT) == 0) then
        call transport_output( narea, ctim )
      endif
    endif

    if (izeits == izeite) then
      !- tempdat files in numbered 6 or 12 hour chunks:
      !    With the option archive=.true. the first chunk goes to the 
      !    'archive' file, and subsequent chunks will go to 'tempdat.00', 
      !    'tempdat.01', ... .   
      !    With archive=.false., the first file is 'tempdat.00', and subsequent
      !    files are 'tempdat.01', 'tempdat.02', ... .
      !    If spinup=.true. there is new, updated header 'aufdat' in each file.
      !    If spinup=.false. there is one 'aufdat'-header in the first file
      !    corresponding to the start time, and the second and all subsequent
      !    files will have the 'aufdat'-header corresponding to the start of
      !    second 6 or 12 hr chunk.
      !    
      if (mpi_io_serial) then
        call tempdat_next ()
        if (calctrans) call transdat_new () 
      endif
    endif
      
  end subroutine output_write

  !============================================================================!
  
  subroutine output_close(ctim,tim)
      ! Output-files will be closed in this file
    
    use cmod_miscvars,      only : tidal
    use io_control,         only : io_server, ioc_close
    use constants,          only : calctrans
    use dmi_mpi,            only : mpi_io_serial, ice_serial
    use transport_variables,only : transdat_close
    use wlstat_wrap,        only : wlstat 
    use writetempdat,       only : tempdat_close, icefile_close
      
    implicit none
    
    character(19), intent(in)           :: ctim  !<==  watch out here!
    real(8),       intent(in)           :: tim

    if (io_server) then
      call ioc_close() 
      return 
    endif

    if (mpi_io_serial) then
      call tempdat_close()
      if (wlstat_out)    call wlstat(2,ctim,tim)
      if (calctrans)     call transdat_close()
    endif
    if (ice_serial) then
      if (.not.tidal) call icefile_close()
    endif

  end subroutine output_close

end module output
