module coupler

  !-  modules:
  use cmod_mem, only : cmr1, cmr2

  !-  directives:
  implicit none
  private

  !-  public coupler methods:
  public :: coupler_usage, coupler_init, coupler_dynamics, coupler_finish

  !-  coupler index variables:
  integer(4), private, save :: coupler_offset = 2
  integer(4), private, save :: nlow = 3, nup = 3

  !-  coupler time variables:
  integer(4), private, save :: out_nt
  real(8),    private, save :: dtbio

  !-  global ergom variables:
  integer(4), allocatable, private, save :: n2d(:), n3d(:)
  type(cmr1), allocatable, private, save :: npr(:)
  type(cmr2), allocatable, private, save :: benthos(:)

  !-  local ergom variables:
  type(cmr1), allocatable, private, save :: npr_l(:)
  type(cmr2), allocatable, private, save :: benthos_l(:)
  real(8),    allocatable, private       :: swrad_l(:)

  !-  private methods:
  private :: bio_limit

  !NOTES:
  ! * The implementation of the variable NP ratio is a quick-and-dirty solution.
  !   It should only allocate a 2d array if it is used, and we should try to 
  !   simply the IO, scatter and allocatation calls
  !
  ! * Consider allocating all arrays in cmod_init
contains

!===============================================================================

  subroutine coupler_usage(iu06,nt,nt2d, irc)
    !- modules
    use bioparam, only : ReadBioparam, ntbio, ntben, zpmodel, version

    implicit none

    ! args
    integer(4), intent(in)  :: iu06
    integer(4), intent(out) :: nt    ! number of traces to be inc. in cmp array
    integer(4), intent(out) :: nt2d  ! number of benthos tracers
    integer(4), intent(out) :: irc   ! return code; 0 means 'no errors'
    
    ! locals
    character(len=11) :: cfmt, cfmt2

    irc = 0
    
    !- get the ergom parameters:
    call ReadBioparam(iu06)

    !- which version of ergom are we running
    if (version == 1) then
      cfmt = '(a36,'//'i2,'//'a8)'
      zpmodel = .false.
    elseif (version == 2) then
      ntbio = ntbio + 1
      cfmt = '(a36,'//'i3,'//'a8)'
      zpmodel = .true.
    elseif (version == 3) then
      ntbio = ntbio + 3
      ntben = ntben + 1
      cfmt = '(a36,'//'i3,'//'a8)'
      zpmodel = .true.
    else
      ntben = -1
      ntbio = -1
      irc = 11
    endif
    cfmt2 = '(a4,'//'i2,'//'a21)'

    write(iu06,trim(cfmt))                                                     &
    'Running with dmi_ergom_coupler using', ntbio, ' tracers'
    write(iu06,trim(cfmt2)) ' and', ntben, ' sediment variable(s)'

    nt2d = ntben
    nt   = ntbio
  end subroutine coupler_usage

!===============================================================================

  subroutine coupler_init(ia, narea, iu06, mainarea, ncoff,                    &
                          mmx, nmx, kmx, nq, nz, mmk, kh, hz,                  &
                          components, rho, bndq, bndzk, on_IO_task, irc )
    ! Container for reading in initial values into the arrays and reading
    ! in boundaries (river, open). 
    ! Also, allocate/associate arrays for the DMI-ERGOM coupler so that we
    ! avoid messing with the cmp array.

    use bioinput,       only : bioread, biobndcl, bioread_npr
    use biooutput,      only : biofile_new
    use bioparam,       only : bioInd,ntbio,ntben,biooutdt,NDTBIOMOD,zpmodel
    use constants,      only : coldst
    use io_control,     only : io_server, ioc_open_coupler
    use biopermute,     only : bio_permute
    use bioriver,       only : bioloadings
    use cmod_miscvars,  only : dtmain
    use cmod_local_arrays, only : mm1k_l 
    use dmi_mpi,        only : dd, mpi_size, dmpi_scatter
    use dmi_mpi_global, only : mpi_io_rank

    implicit none

    integer(4), intent(in)    :: ia, narea, iu06, mmx, nmx, kmx, nq, nz, ncoff
    logical,    intent(in)    :: mainarea
    integer(4), intent(in)    :: mmk(:,0:,0:), kh(0:)
    real(8),    intent(in)    :: hz(0:), rho(0:)
    real(8),    intent(inout) :: bndq(:,:,0:), bndzk(:,:,0:)
    logical,    intent(in)    :: on_IO_task
    integer(4), intent(out)   :: irc

    real(8),    intent(inout) :: components(:,0:)

    logical,    save :: firsttime = .true.
    integer(4)       :: n2d_l, iia, iben

    if (firsttime) then
      !- some private vars:
      dtbio  = NDTBIOMOD*dtmain    ! FIXME: still no stability check here
      out_nt = nint(biooutdt/dtmain)

      !- some index stuff:
      coupler_offset = ncoff 
      nlow           = ncoff + 1
      nup            = ncoff + ntbio

      !- allocate the benthos variable:
      allocate( benthos(narea), n2d(narea), n3d(narea),          &
                benthos_l(narea),npr(narea), npr_l(narea) )
     
      n2d_l = 0
      do iia=1,narea 
        n2d_l = max(dd(iia)%up_ws - dd(iia)%low_ws + 1, n2d_l)
      enddo
      allocate( swrad_l(0:n2d_l) )

      !- tracer index
      call bioInd()

      !- open output file on files entry:
      if (io_server) then
        ! open outout files with io-server
        call ioc_open_coupler()
      else
        if (on_IO_task) call biofile_new ()
      endif

      firsttime = .false.
    endif

    !- allocate/associate the biological variables:
    n2d(ia) = maxval(mmk(1,1:mmx,1:nmx))
    n3d(ia) = maxval(mmk(1:kmx,1:mmx,1:nmx))
    allocate( benthos(ia)%p(ntben,0:n2d(ia)) )
    allocate( npr(ia)%p(0:n2d(ia)) )

    if (on_IO_task) then
      !- read the initial fields:
      call bioread( ia, n3d(ia), narea, coldst, iu06, benthos(ia)%p,           &
                    components, nlow, nup )
      call bio_permute( mmx, nmx, kmx, n2d(ia), n3d(ia), mmk, components,      &
                        benthos(ia)%p, nlow, nup,ntben )

      ! read spatially varying NP ratio if required
      ! fixme: DO WE NEED TO PERMUTE THIS ARRAY?
      call bioread_npr(ia, narea, iu06, npr(ia)%p)
      
      !- boundary values:
      if (mainarea) then
        call biobndcl(kmx, kh, mmk, nz, nlow, nup, bndzk, iu06)
      endif
    endif

    !- river loadings:
    call bioloadings ( ia, nlow, nup, bndq )

    !- allocate local arrays and scatter global2local
    n2d_l = dd(ia)%up_ws - dd(ia)%low_ws + 1 + dd(ia)%halo2
    allocate( benthos_l(ia)%p(ntben,0:n2d_l) )
    allocate( npr_l(ia)%p(0:n2d_l) )
   
 
    if (mpi_size == 1) then
      ! local == global in serial mode
      benthos_l(ia)%p(1:ntben,1:n2d(ia)) = benthos(ia)%p(1:ntben,1:n2d(ia))
      npr_l(ia)%p(1:n2d(ia))     = npr(ia)%p(1:n2d(ia))
    endif
      
    do iben=1,ntben
      call dmpi_scatter(ia,(1),n2d(ia),mm1k_l(ia)%p,benthos_l(ia)%p(iben,:),   &
                        mmk,benthos(ia)%p(iben,:),mpi_io_rank,kh)
    enddo
     
    call dmpi_scatter(ia,(1),n2d(ia),mm1k_l(ia)%p,npr_l(ia)%p,                 &
                      mmk,npr(ia)%p,mpi_io_rank,kh)

    !- that's all, folks!
    irc = 0
  end subroutine coupler_init

!===============================================================================

  subroutine coupler_dynamics(ia, iu06, mmx, nmx, kmx, mmk_l, kh, h,           &
                              components, dispv, rho, u, v, z, casus,          &
                              atemp, humid, cloud, wu, wv, pl, precip,         &
                              update_counters, idx, irc)
    ! Container for the only call in the timeloop. If the coupler needs to do
    ! different things at different timesteps, e.g. dynamics and output, then
    ! it's up to the coupler_dynamics implementation to do this differentiation.

    use dmi_mpi_global,    only : mpi_io_rank
    use dmi_mpi,           only : dmpi_gather, dmpi_scatter, mpi_io_serial,dd, &
                                  mpi_size
    use cmod_local_arrays, only : ind_l
    use cmod_arrays,       only : xkoor, ykoor, mm1k, cmp
    use cmod_miscvars,     only : izeite, i2, lvalidate, dtday, schfactor, time
    use constants,         only : spinup, izeits, tim
    use io_control,        only : io_server, ioc_write_coupler,                &
                                  ioc_writerestart_coupler
    use biooutput,         only : biofile_next, biofile_write,                 &
                                  write_bio_restart, val_benthos, biofile_new
    use bioparam,          only : NDTBIOMOD, ntbio, ntben
    use biosun,            only : shortwaverad
    use biomod,            only : bio_dynamics
    implicit none

    integer(4), intent(in)    :: ia, iu06, mmx, nmx, kmx, mmk_l(:,0:,0:), kh(0:)
    real(8),    intent(in)    :: h(0:), u(0:), v(0:), z(0:), rho(0:)
    real(8),    intent(in)    :: atemp(0:), humid(0:), cloud(0:)
    real(8),    intent(in)    :: wu(0:), wv(0:), pl(0:), precip(0:)
    logical,    intent(in)    :: casus(0:), update_counters
    integer(4), intent(inout) :: idx(1:,1:)
    integer(4), intent(out)   :: irc
    real(8),    intent(in)    :: dispv(:,0:)

    real(8),    intent(inout) :: components(:,0:)

    integer(4), save :: itime = 1
    integer(4), save :: otime = 1
    integer(4), save :: ntime = 0
    integer(4), save :: n2dswr = 0, last_area = 0
    logical,    save :: initstats = .true., did_gather, write_restart

    integer(4) :: n2d_l, irc_thr

    irc     = 0

    !- allocate local bio arrays -----------------------------------------------
    n2d_l = dd(ia)%up_ws - dd(ia)%low_ws + 1
    
    !- print initial stats of the benthos variable -----------------------------
    if (initstats) then
      if (lvalidate .and. mpi_io_serial) then
        call val_benthos(ia, iu06, n2d(ia), benthos(ia)%p,ntben, (ia==1))
      endif
      if (update_counters) initstats = .false.
    endif

    ! do the ergom dynamics ----------------------------------------------------
    if (mod(itime,NDTBIOMOD) == 0) then
!$OMP PARALLEL DEFAULT (shared) PRIVATE(irc_thr)
      irc_thr = 0
      !- check eco-var limits:
      call bio_limit(n2d_l, mmk_l, ind_l(ia)%p, kh, components,idx)

      !- obtain the insolation:
      ! FIXME: consider removing the swrad array and call this routine in 
      !        biomod instead
      call shortwaverad(n2d_l, ind_l(ia)%p, xkoor(ia)%p, ykoor(ia)%p,          &
                        cloud, humid, casus, swrad_l )

!$OMP BARRIER

      !- do the bio dynamics:
      ! fixme: we should not send full npr_l array if it is not used
      !      : can we handle the error flag nicer
      call bio_dynamics(n2d_l, kmx, mmk_l, ind_l(ia)%p, kh, h,u, v, wu, wv,    &
                        components, benthos_l(ia)%p, swrad_l, dispv,           &
                        casus, npr_l(ia)%p,dtbio,idx,irc_thr)
!$OMP CRITICAL
      irc = max(irc_thr,0)
!$OMP END CRITICAL
!$OMP END PARALLEL
      if (irc /= 0) then
        write(iu06,'(a39,i3)') 'Error in bio_dynamics, error flag was, ', irc
        write(iu06,'(a39,i3)') 'Error in domain', ia
        return
      endif
    endif

    ! do the ergom outout when running without io-server------------------------
    if (io_server) then
      ! writing with IO-server
      call ioc_write_coupler(izeits, ia, components(3:,:), benthos_l(ia)%p)
    else
      did_gather = .false.
      if (mod(otime,out_nt) == 0) then
        if (mpi_size == 1) then
          ! global==local in serial mode
          benthos(ia)%p(1:ntben,1:n2d(ia)) = benthos_l(ia)%p(1:ntben,1:n2d(ia))
          ! fixme: maybe we don't need this one:
          cmp(ia)%p(3:,1:) = components(3:,1:)
        else
          call dmpi_gather(ia,kmx,components,cmp(ia)%p,coupler_offset+ntbio,   &
                           mpi_io_rank)
          !FIXME: check this!
          call dmpi_gather(ia,(1),benthos_l(ia)%p,benthos(ia)%p,ntben,         &
                             mpi_io_rank)
        endif
        did_gather = .true.
        if (mpi_io_serial) then
          call biofile_write(ia, iu06, mmx, nmx, kmx, n2d(ia), n3d(ia),        &
                             mm1k(ia)%p, nlow, cmp(ia)%p, benthos(ia)%p)
        endif
      endif
    endif

    ! update time step counters ------------------------------------------------
    if (update_counters) then
      if (mod(itime,NDTBIOMOD) == 0) then
        itime = 1
      else
        itime = itime + 1
      endif
      if (mod(otime,out_nt) == 0) then
        otime = 1
      else
        otime = otime + 1
      endif
      ntime = ntime + 1
    endif

    ! print stats of the benthos variable --------------------------------------
    if (izeits == izeite) then
      if (.not.did_gather) then
        if (mpi_size == 1) then
          ! global==local in serial mode
          benthos(ia)%p(1:ntben,1:n2d(ia)) = benthos_l(ia)%p(1:ntben,1:n2d(ia))
        else
          !FIXME: check this!
          call dmpi_gather(ia,(1),benthos_l(ia)%p,benthos(ia)%p,ntben,         &
                           mpi_io_rank)
        endif
      endif
      if (lvalidate .and. mpi_io_serial) then
        call val_benthos(ia, iu06, n2d(ia), benthos(ia)%p,ntben, (ia==1))
      endif
    endif

    !- should we write restart?
    if ((i2 == 1 .or. spinup) .and. izeits == izeite) then
      write_restart = .true.
    else
      write_restart = .false.
    endif

    if (io_server .and. write_restart) then
      !- let IO-server write the restart file
      call ioc_writerestart_coupler(tim, ia)

    elseif (.not. io_server .and. write_restart) then
      ! write bio-restart file with mpi_io_serial
      if (.not.did_gather) then
        if (mpi_size == 1) then
          ! global==local in serial mode
          benthos(ia)%p(1:ntben,1:n2d(ia)) = benthos_l(ia)%p(1:ntben,1:n2d(ia))
          ! fixme: maybe we don't need this one:
          cmp(ia)%p(3:,1:) = components(3:,1:)
        else
          call dmpi_gather(ia,kmx,components,cmp(ia)%p,coupler_offset+ntbio,   &
                           mpi_io_rank)
          !FIXME: check this!
          call dmpi_gather(ia,(1),benthos_l(ia)%p,benthos(ia)%p,ntben,         &
                           mpi_io_rank)
        endif
      endif
      if (mpi_io_serial) then
        call write_bio_restart(ia, iu06, update_counters, mmx, nmx, kmx,       &
                               n2d(ia), n3d(ia), mm1k(ia)%p, nlow, cmp(ia)%p,  &
                               benthos(ia)%p)
      endif
    endif

    if (.not. io_server) then
      ! open the next biodat file. 
      !   biodat files in numbered 6 or 12 hour chunks:  
      !       the first file is 'biodat.00', and subsequent files are 
      !       'biodat.01', 'biodat.02', ... .  .
      if (ntime == izeite .and. (tim+dtday-dtday*schfactor <= time)) then
        if (mpi_io_serial) call biofile_next ( iu06 )
        ntime = 0
      endif
    endif


  end subroutine coupler_dynamics

!===============================================================================

  subroutine coupler_finish(iu06, irc)
    ! Container for the call after the timeloop. Can be used to close files and
    ! deallocate memory and to do final outputs.
    use biooutput,  only : biofile_close
    use io_control, only : io_server, ioc_close_coupler
    implicit none
    integer(4), intent(in)   :: iu06
    integer(4), intent(out)  :: irc

    !- close output file:
    if (io_server) then
      call ioc_close_coupler()
    else
      call biofile_close()
    endif

    irc = 0
  end subroutine coupler_finish

!===============================================================================

  subroutine bio_limit( n2d, mmk, ind, kh, cmp, idx)

    use dmi_omp, only : domp_get_domain

    implicit none

    integer(4), intent(in)    :: n2d, ind(:,:), kh(0:), mmk(:,0:,0:)
    real(8),    intent(inout) :: cmp(:,0:)
    integer(4), intent(inout) :: idx(1:,1:)

    integer(4)         :: n, nl, nu, i, j, mi, nupm1, kb, ml, mu
    real(8), parameter :: zero = 0.0_8

    call domp_get_domain( kh, 1, n2d, nl, nu, idx )

    ! avoid cutting oxygene 'coz unlike the physical oxygene, the model 
    ! state variable called oxygene can indeed have negative values: 
    if (nup-1 >= nlow) then
      nupm1 = nup - 1
    else
      nupm1 = nup
    endif

    do n = nl, nu
      cmp(nlow:nupm1,n) = max( zero, cmp(nlow:nupm1,n) )
    enddo

    do n = nl, nu
      kb = kh(n)
      if (kb > 1) then
        i  = ind(1,n)
        j  = ind(2,n)
        ml = mmk(2,i,j)
        mu = mmk(2,i,j) + kb - 2
        do mi=ml,mu
          cmp(nlow:nupm1,mi) = max( zero, cmp(nlow:nupm1,mi) )
        enddo
      endif
    enddo

  end subroutine bio_limit

end module coupler

