module coupler

  !-  modules:
  use cmod_mem, only : cmr1

  !-  directives:
  implicit none
  private

  !-  public coupler methods:
  public :: coupler_usage, coupler_init, coupler_dynamics, coupler_finish

  !-  coupler index variables:
  integer(4), private, save :: coupler_offset = 2
  integer(4), private, save :: nlow = 3, nup = 3

  !-  copuler time variables:
  integer(4), private, save :: out_nt
  real(8),    private, save :: dtbio

  !-  ergom variables:
  type(cmr1), allocatable, save, private :: benthos(:)
  type(cmr1), allocatable, save, private :: fishn(:),fisho(:) ! fish feeding
  type(cmr1), allocatable, save, private :: primprod(:)       ! fish feeding
  type(cmr1), allocatable, save, private :: light(:)          ! biolight saved
  real(8),    allocatable, save, private :: swrad(:)

  !-  private methods:
  private :: bio_limit
  
contains

!===============================================================================

  subroutine coupler_usage(iu06,nt,irc)

    use bioparam, only : ReadBioparam, ntbio, stage_model, zpmodel
    use exits,    only : exitme

    implicit none

    integer(4), intent(in)  :: iu06
    integer(4), intent(out) :: nt    ! number of traces to be inc. in cmp array
    integer(4), intent(out) :: irc   ! return code; 0 means 'no errors'

    !- get the ergom parameters:
    call ReadBioparam(iu06)

    !- check some parameters
    if ( stage_model .and. (.not. zpmodel) ) then
      call exitme(1,'Must run with zpmodel and stagemodel together')
    endif

    !- return values:
    nt  = ntbio
    irc = 0

  end subroutine coupler_usage

!===============================================================================

  subroutine coupler_init(ia, narea, iu06, mainarea, ncoff,                    &
                          mmx, nmx, kmx, nq, nz, mmk, kh, hz,                  &
                          components, rho, bndq, bndzk, on_IO_task, irc )
    ! Container for reading in initial values into the arrays and reading
    ! in boundaries (river, open). 
    ! Also, allocate/associate arrays for the DMI-ERGOM coupler so that we
    ! avoid messing with the cmp array.

    use bioinput,       only : bioread, biobndcl
    use bioparam,       only : ntbio, biooutdt, NDTBIOMOD, ecofish
    use constants,      only : coldst
    use biopermute,     only : bio_permute
    use bioriver,       only : bioloadings
    use cmod_miscvars,  only : dtmain
    use biofish,        only : fishinit
    
    implicit none

    integer(4), intent(in)    :: ia, narea, iu06, mmx, nmx, kmx, nq, nz, ncoff
    logical,    intent(in)    :: mainarea
    integer(4), intent(in)    :: mmk(:,0:,0:), kh(0:)
    real(8),    intent(in)    :: hz(0:), rho(0:)
    real(8),    intent(inout) :: bndq(:,:,0:), bndzk(:,:,0:)
    logical,    intent(in)    :: on_IO_task
    integer(4), intent(out)   :: irc
    real(8),    intent(inout) :: components(:,0:)

    integer(4)       :: n2d, n3d
    integer(4), save :: n2dswr = 0
    logical,    save :: firsttime = .true.

    !- river loadings (must be defined on each local task):
    call bioloadings ( ia, kh, mmk, ncoff+1, ncoff+ntbio, bndq )

    !- return if not on IO task:
    if (.not.on_IO_task) then
      irc = 0
      return
    endif

    if (firsttime) then
      !- some private vars:
      dtbio  = NDTBIOMOD*dtmain    ! FIXME: still no stability check here
      out_nt = nint(biooutdt/dtmain)

      !- some index stuff:
      coupler_offset = ncoff 
      nlow           = ncoff + 1
      nup            = ncoff + ntbio

      !- allocate the benthos variable and dummy swrad:
      allocate( benthos(narea), swrad(0:0) )
      !- allocate arrays for fish feeding:
      allocate( fishn(narea), fisho(narea) )
      !- allocate arrays for saving primary production:
      allocate( primprod(narea) )
      !- allocate arrays for saving biolight
      allocate( light(narea) )

      firsttime = .false.
    endif

    !- allocate/associate the biological variables:
    n2d = maxval(mmk(1,1:mmx,1:nmx))
    n3d = maxval(mmk(1:kmx,1:mmx,1:nmx))
    allocate( benthos(ia)%p(0:n2d) )
    allocate( fishn(ia)%p(0:n2d),fisho(ia)%p(0:n2d) )
    allocate( primprod(ia)%p(0:n3d) )
    allocate( light(ia)%p(0:n3d) )
    if (n2d > n2dswr) then
      !  re-alloc swrad:
      n2dswr = n2d
      deallocate( swrad )
      allocate( swrad(0:n2dswr) )
    endif

    !- read the initial fields:
    call bioread( ia, n3d, narea, coldst, iu06, benthos(ia)%p, components,&
                  nlow, nup )
    call bio_permute( mmx, nmx, kmx, n2d, n3d, mmk, components, benthos(ia)%p, &
                  nlow, nup )

    !- boundary values:
    if (mainarea) then
      call biobndcl(kmx, kh, mmk, nz, nlow, nup, bndzk)
    endif

    !- fish feeding
    ! fixme (laj): should we split this file into ndomains as done with the
    !              other biofiles. For now, this will do as a quick fix.
    if ( (ia == narea) .and. ecofish ) then
      write(iu06,*) 'running with feeding pressure from fish'
      call fishinit(narea, fishn,fisho)
    endif
    
    !- that's all, folks!
    irc = 0
  end subroutine coupler_init

!===============================================================================

  subroutine coupler_dynamics(ia, iu06, mmx, nmx, kmx, mmk, kh, h,             &
                              components, dispv, rho, u, v, z, casus,          &
                              atemp, humid, cloud, wu, wv, pl, precip,         &
                              update_counters, idx, irc)
    ! Container for the only call in the timeloop. If the coupler needs to do
    ! different things at different timesteps, e.g. dynamics and output, then
    ! it's up to the coupler_dynamics implementation to do this differentiation.

    use cmod_arrays,   only : ind, xkoor, ykoor
    use cmod_miscvars, only : izeite, i2, lvalidate, dtday, schfactor, time
    use constants,     only : spinup, izeits, tim
    use biooutput,     only : biofile_next, biofile_write, write_bio_restart,  &
                              val_benthos, biofile_new
    use bioparam,      only : NDTBIOMOD
    use biosun,        only : shortwaverad
    use biomod,        only : bio_dynamics
    implicit none

    integer(4), intent(in)  :: ia, iu06, mmx, nmx, kmx, mmk(:,0:,0:), kh(0:)
    real(8),    intent(in)  :: h(0:), u(0:), v(0:), z(0:), rho(0:)
    real(8),    intent(in)  :: atemp(0:), humid(0:), cloud(0:)
    real(8),    intent(in)  :: wu(0:), wv(0:), pl(0:), precip(0:)
    logical,    intent(in)  :: casus(0:), update_counters
    integer(4), intent(out) :: irc
    integer(4), intent(inout) :: idx(1:,1:)
    real(8), intent(inout)  :: dispv(:,0:)
    real(8), intent(inout)  :: components(:,0:)

    integer(4), save :: itime = 1
    integer(4), save :: otime = 1
    integer(4), save :: ntime = 0
    logical,    save :: initstats = .true., firstentry = .true.

    integer(4) :: n2d, n3d

    irc = 0

    !- print initial stats of the benthos variable -----------------------------
    if (initstats) then
      if (lvalidate) call val_benthos(ia, iu06, maxval(mmk(1,1:mmx,1:nmx)),    &
                                      benthos(ia)%p, (ia==1))
      if (update_counters) initstats = .false.
    endif

    !- allocate/associate the biological variables -----------------------------
    if (mod(itime,NDTBIOMOD) == 0 .or. mod(otime,out_nt) == 0) then
      n2d = maxval(mmk(1,1:mmx,1:nmx))
      n3d = maxval(mmk(1:kmx,1:mmx,1:nmx))
    endif

    ! do the ergom dynamics ----------------------------------------------------
    if (mod(itime,NDTBIOMOD) == 0) then
!$OMP PARALLEL DEFAULT (shared)

      !- check eco-var limits:
      call bio_limit(n2d, mmk, ind(ia)%p, kh, components)

!$OMP BARRIER

      !- obtain the insolation:
      call shortwaverad(n2d, ind(ia)%p, xkoor(ia)%p, ykoor(ia)%p,              &
                        cloud, humid, casus, swrad )

!$OMP BARRIER

      !- do the bio dynamics:
      call bio_dynamics(n2d, kmx, mmk, ind(ia)%p, kh, h, wu, wv,components,    &
                        benthos(ia)%p, primprod(ia)%p,fisho(ia)%p, fishn(ia)%p,&
                        swrad,light(ia)%p, dtbio)
!$OMP END PARALLEL
    endif
    
    ! do the ergom outout ------------------------------------------------------
    if (firstentry) then
      !- open output file on files entry:
      call biofile_new ()
      firstentry = .false.
    endif
    if (mod(otime,out_nt) == 0) then
      call biofile_write(ia, iu06, mmx, nmx, kmx, n2d, n3d, mmk, nlow,         &
                         components, benthos(ia)%p, primprod(ia)%p)
    endif
    
    ! update time step counters ------------------------------------------------
    if (update_counters) then
      if (mod(itime,NDTBIOMOD) == 0) then
        itime = 1
      else
        itime = itime + 1
      endif
      if (mod(otime,out_nt) == 0) then
        otime = 1
      else
        otime = otime + 1
      endif
      ntime = ntime + 1
    endif

    ! print stats of the benthos variable --------------------------------------
    if (izeits == izeite) then
      if (lvalidate) call val_benthos(ia, iu06, maxval(mmk(1,1:mmx,1:nmx)),    &
                                      benthos(ia)%p, (ia==1))
    endif

    ! open the next biodat file. 
    !   biodat files in numbered 6 or 12 hour chunks:  
    !       the first file is 'biodat.00', and subsequent files are 
    !       'biodat.01', 'biodat.02', ... .  .
    if (ntime == izeite .and. (tim+dtday-dtday*schfactor <= time)) then
      call biofile_next ( iu06 )
      ntime = 0
    endif

    ! write bio-restart file.
    if ((i2 == 1 .or. spinup) .and. izeits == izeite) then
      call write_bio_restart(ia, iu06, update_counters, mmx, nmx, kmx, n2d,    &
                             n3d, mmk, nlow, components, benthos(ia)%p)
    endif

  end subroutine coupler_dynamics

!===============================================================================

  subroutine coupler_finish(iu06, irc)
    ! Container for the call after the timeloop. Can be used to close files and
    ! deallocate memory and to do final outputs.
    use biooutput,  only : biofile_close
    implicit none
    integer(4), intent(in)   :: iu06
    integer(4), intent(out)  :: irc

    !- close output file:
    call biofile_close()

    irc = 0
  end subroutine coupler_finish

!===============================================================================

  subroutine bio_limit( n2d, mmk, ind, kh, cmp )

    use dmi_omp,  only : domp_get_domain
    use bioparam, only : idx_oxy

    implicit none

    integer(4), intent(in)    :: n2d, ind(:,:), kh(0:), mmk(:,0:,0:)
    real(8),    intent(inout) :: cmp(:,0:)

    integer(4)         :: n, nl, nu, i, j, mi, nupm1, kb, ml, mu
    real(8), parameter :: zero = 0.0_8

    call domp_get_domain( 1, n2d, nl, nu )

    ! avoid cutting oxygene 'coz unlike the physical oxygene, the model 
    ! state variable called oxygene can indeed have negative values:

    nupm1 = nup

    do n = nl, nu
      cmp(nlow:2+idx_oxy-1,n)    = max( zero, cmp(nlow:2+idx_oxy-1,n) )
      cmp(2+idx_oxy+1:nupm1,n) = max( zero, cmp(2+idx_oxy+1:nupm1,n) )
    enddo

    do n = nl, nu
      kb = kh(n)
      if (kb > 1) then
        i  = ind(1,n)
        j  = ind(2,n)
        ml = mmk(2,i,j)
        mu = mmk(2,i,j) + kb - 2
        do mi=ml,mu
          cmp(nlow:2+idx_oxy-1,mi)    = max( zero, cmp(nlow:2+idx_oxy-1,mi) )
          cmp(2+idx_oxy+1:nupm1,mi) = max( zero, cmp(2+idx_oxy+1:nupm1,mi) )
        enddo
      endif
    enddo

  end subroutine bio_limit

!===============================================================================

end module coupler

