module eflow
  use constants, only : zero, one, half

  !- directives:
  implicit none
  private

  !- allocatable local arrays:
  real(8), allocatable, private, dimension(:,:) :: ee,rin,rout,eu,ev,dxe,dye,  &
                                                   deu,dev,cx,cy

  !- simple local vars:
  integer(4), save, private :: Size2D = -1
        ! =-1: above arrays not yet allocated
        ! > 1: allocated size of above arrays

  !- methods:
  private :: eflow_alloc  !< Method for allocating/re-allocating local arrays.
  public  :: eflow_int    !< Routine doing the 2D advection of ice/snow:
                          !   When we decide again to apply snow cover of 
                          !   non-zero thickness, this routine should be called 
                          !   with nc=3. At present, when we have no snow cover
                          !   on the ice, it is fully sufficient having nc=2,
                          !   signifying advection of ice thickness and ice 
                          !   concentration.

  real(8), parameter, private :: three = 3.0_8

contains

  ! ----------------------------------------------------------------------------

subroutine eflow_int (me,muvs,n2d,ind,casus,u,v,eis,dxj,dyi,dt,                &
                      cosphi,nbpz,krz,rwze,nbpu,kru,nbpv,krv,nc)

  !- modules -------------------------------------------------------------------
  use dmi_omp, only : domp_get_domain

  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in) :: nc
  integer(4), intent(in) :: n2d,nbpz,nbpu,nbpv
  integer(4), intent(in) :: me(0:,0:)                  ! (0:imax,0:jmax)
  integer(4), intent(in) :: muvs(:,0:,0:)              ! (8,0:imax,0:jmax)
  integer(4), intent(in) :: ind(:,:)                   ! (2,n2d)
  integer(4), intent(in) :: krz(:,0:)                  ! (3,0:nbpz)
  integer(4), intent(in) :: kru(:,0:)                  ! (3,0:nbpu)
  integer(4), intent(in) :: krv(:,0:)                  ! (3,0:nbpv)
  logical, intent(in)    :: casus(0:)                  ! (0:n2d)
  real(8), intent(in)    :: u(0:), v(0:)               ! (0:n2d)
  real(8), intent(in)    :: rwze(:,0:)
  real(8), intent(in)    :: dyi(0:), dxj(0:)           ! (0:imax), (0:jmax)
  real(8), intent(in)    :: cosphi(:,0:)               ! (2,0:imax)
  real(8), intent(in)    :: dt
  real(8), intent(inout) :: eis(:,0:)


  !- local vars ----------------------------------------------------------------
  integer(4) :: n,i,j,nuw,nue,nvn,nvs,xma,yma,xmb,ymb,ntyp
  real(8)    :: udye(1:nc),udyw(1:nc),vdxn(1:nc),vdxs(1:nc),aa(1:nc),bb(1:nc)
  real(8)    :: wdxdyout,wdxdyin,wdxdyein,wdxdyeout,sx,sy
  real(8)    :: wdxdye,div,ddd,emax(1:nc),emin(1:nc),pin,qin,pout,qout
  real(8)    :: hdt,fac,facx,facy, rwzin,rwzout, vol(1:nc)
  real(8)    :: dx,dy,dxdy,dtdx,dtdy,dtddx,dtddy,hdtddx,hdtddy
  integer(4) :: nl,nu,ic

  !- some parameters -----------------------------------------------------------
  real(8), parameter :: onehalf  = half     , oneforth = onehalf*onehalf
  real(8), parameter :: onethird = one/three, onesixth = onehalf*onethird
  real(8), parameter :: tm12 = 1.0e-12_8


  !- possibly allocate or re-allocate:
  if (n2d > Size2D) then
    call eflow_alloc(n2d,nc)
  endif

  dx  = dxj(1)   ! since dxj(0:jmax) is constant
  dy  = dyi(1)   ! since dyi(0:imax) is constant

  ! derived from dx, dy and dt and used throughout:
  hdt    = half*dt
  dxdy   = dx*dy
  dtddx  = dt/dx
  dtddy  = dt/dy
  hdtddx = hdt/dx
  hdtddy = hdt/dy
  dtdx   = dt*dx
  dtdy   = dt*dy

  !- some initialisations:
  eu(1:nc,0) = zero
  ev(1:nc,0) = zero
  ee(1:nc,0) = zero
  dxe(1:nc,0) = zero
  dye(1:nc,0) = zero
  deu(1:nc,0) = zero
  dev(1:nc,0) = zero
  rin (1:nc,0) = zero
  rout(1:nc,0) = zero
  cx(1:nc,0) = zero
  cy(1:nc,0) = zero

!$OMP PARALLEL DEFAULT (SHARED) private(n,nl,nu,i,j,fac,vol)
  call domp_get_domain(1,n2d, nl, nu)

  fac = hdtddy
  do n=nl, nu
    i=ind(1,n)
    j=ind(2,n)

    eu(1:nc,n)=zero

    if (muvs(1,i,j) == 0) cycle
    if (casus(n).and.casus(me(i,j+1))) cycle

    if (u(n) >= zero) then
      eu(1:nc,n)=eis(1:nc,n)
      if (v(me(i-1,j)) < zero) then
        eu(1:nc,n) = eu(1:nc,n)                                                &
                   - fac*v(me(i-1,j))*(eis(1:nc,me(i-1,j))-eis(1:nc,n))
      endif
      if (v(n) > zero) then
        eu(1:nc,n) = eu(1:nc,n) - fac*v(n)*(eis(1:nc,n)-eis(1:nc,me(i+1,j)))
      endif
    else
      eu(1:nc,n)=eis(1:nc,me(i,j+1))
      if (v(me(i-1,j+1)) < zero) then
        eu(1:nc,n) = eu(1:nc,n)                                                &
                   - fac*v(me(i-1,j+1))                                        &
                        *(eis(1:nc,me(i-1,j+1))-eis(1:nc,me(i,j+1)))
      endif
      if (v(me(i,j+1)) > zero) then
        eu(1:nc,n) = eu(1:nc,n)                                                &
                   - fac*v(me(i,j+1))                                          &
                        *(eis(1:nc,me(i,j+1))-eis(1:nc,me(i+1,j+1)))
      endif
    endif
  enddo

  do n=nl, nu
    i=ind(1,n)
    j=ind(2,n)

    ev(1:nc,n)=zero

    if (muvs(2,i,j) == 0) cycle
    if (casus(n).and.casus(me(i+1,j))) cycle

    fac = hdtddx/cosphi(1,i)

    if (v(n) <= zero) then
      ev(1:nc,n)=eis(1:nc,n)
      if (u(me(i,j-1)) > zero) then
        ev(1:nc,n) = ev(1:nc,n)                                                &
                   - fac*u(me(i,j-1))*(eis(1:nc,n)-eis(1:nc,me(i,j-1)))
      endif
      if (u(n) < zero) then
        ev(1:nc,n) = ev(1:nc,n) - fac*u(n)*(eis(1:nc,me(i,j+1))-eis(1:nc,n))
      endif
    else
      ev(1:nc,n)=eis(1:nc,me(i+1,j))
      if (u(me(i+1,j-1)) > zero) then
        ev(1:nc,n) = ev(1:nc,n)                                                &
                   - fac*u(me(i+1,j-1))                                        &
                        *(eis(1:nc,me(i+1,j))-eis(1:nc,me(i+1,j-1)))
      endif
      if (u(me(i+1,j)) < zero) then
        ev(1:nc,n) = ev(1:nc,n)                                                &
                   - fac*u(me(i+1,j))                                          &
                        *(eis(1:nc,me(i+1,j+1))-eis(1:nc,me(i+1,j)))
      endif
    endif
  enddo

!$OMP BARRIER
  ! barrier due to eu(1:nc,me(i,j-1)) and ev(1:nc,me(i-1,j))
  do n=nl, nu
    i=ind(1,n)
    j=ind(2,n)

    if (casus(n)) then
      vol(1:nc) = zero
    else
      vol(1:nc) = eis(1:nc,n)*cosphi(1,i)*dxdy
    endif

    ee(1:nc,n)= vol(1:nc)                                                      &
         - dtdy*( u(n)*eu(1:nc,n)                                              &
                 -u(me(i,j-1))*eu(1:nc,me(i,j-1)))                             &
         + dtdx*( v(n)*ev(1:nc,n)*cosphi(2,i)                                  &
                 -v(me(i-1,j))*ev(1:nc,me(i-1,j))*cosphi(2,i-1))
  enddo
!$OMP END PARALLEL


  if (nbpz > 0) then
    do n=1,nbpz
      i=krz(1,n)
      j=krz(2,n)
      if ( n > 1 .and. i == krz(1,n-1) .and. j == krz(2,n-1) ) cycle
      rwzin =zero
      rwzout=zero
      nuw=me(i,j-1)
      nue=me(i, j )
      nvn=me(i-1,j)
      nvs=me(i, j )
      if(muvs(1,i,j-1) == 0) nuw=0
      if(muvs(1,i, j ) == 0) nue=0
      if(muvs(2,i-1,j) == 0) nvn=0
      if(muvs(2, i ,j) == 0) nvs=0
      facx = one/(dx*cosphi(1,i))
      if (abs(krz(3,n)) == 3 .and. u(nue) > zero) then
        !rwzout = rwzout
        rwzin  = rwzin  + u(nue)*facx
      else
        rwzin  = rwzin  + min(u(nue),zero)*facx
        rwzout = rwzout + max(u(nue),zero)*facx
      endif
      if (abs(krz(3,n)) == 1 .and. u(nuw) < zero) then
        !rwzout = rwzout
        rwzin  = rwzin  - u(nuw)*facx
      else
        rwzin  = rwzin  - max(u(nuw),zero)*facx
        rwzout = rwzout - min(u(nuw),zero)*facx
      endif
      facy = one/(dy*cosphi(1,i))
      if (abs(krz(3,n)) == 2 .and. v(nvn) > zero) then
        !rwzout = rwzout
        rwzin  = rwzin  + cosphi(2,i-1)*v(nvn)*facy
      else
        rwzin  = rwzin  + cosphi(2,i-1)*min(v(nvn),zero)*facy
        rwzout = rwzout + cosphi(2,i-1)*max(v(nvn),zero)*facy
      endif
      if (abs(krz(3,n)) == 4 .and. v(nvs) < zero) then
        !rwzout = rwzout
        rwzin  = rwzin  - cosphi(2, i )*v(nvs)*facy
      else
        rwzin  = rwzin  - cosphi(2, i )*max(v(nvs),zero)*facy
        rwzout = rwzout - cosphi(2, i )*min(v(nvs),zero)*facy
      endif
      fac       = dxdy*cosphi(1,i)
      wdxdyin   = rwzin *fac
      wdxdyout  = rwzout*fac
      do ic=1,nc
        wdxdyein      = wdxdyin *eis(ic,me(i,j))
        wdxdyeout     = wdxdyout*rwze(ic,n)
        wdxdye        = wdxdyein + wdxdyeout
        ee(ic,me(i,j)) = ee(ic,me(i,j)) + dt*wdxdye
      enddo
    enddo
  endif

!$OMP PARALLEL DEFAULT (shared) private(n,nl,nu,i,j,xma,yma,aa,xmb,ymb,bb)
  call domp_get_domain(1,n2d, nl, nu)

  do n=nl, nu
    i=ind(1,n)

    ee(1:nc,n) = ee(1:nc,n)/(cosphi(1,i)*dxdy)

    if (casus(n)) then
      dxe(1:nc,n) = zero
      dye(1:nc,n) = zero
    else
      j=ind(2,n)

      if (me(i,j+1) > 0) then
        xma=1
        aa(1:nc)=eis(1:nc,me(i,j+1))-eis(1:nc,n)
      else
        xma=0
        aa(1:nc)=zero
      endif
      if (me(i,j-1) > 0) then
        xmb=1
        bb(1:nc)=eis(1:nc,n)-eis(1:nc,me(i,j-1))
      else
        xmb=0
        bb(1:nc)=zero
      endif
      if (xma*xmb == 1) then
        dxe(1:nc,n) = half*(aa(1:nc) + bb(1:nc))
      else
        dxe(1:nc,n) = aa(1:nc) + bb(1:nc)
      endif

      if (me(i-1,j) > 0) then
        yma=1
        aa(1:nc)=eis(1:nc,me(i-1,j))-eis(1:nc,n)
      else
        yma=0
        aa(1:nc)=zero
      endif
      if (me(i+1,j) > 0) then
        ymb=1
        bb(1:nc)=eis(1:nc,n)-eis(1:nc,me(i+1,j))
      else
        ymb=0
        bb(1:nc)=zero
      endif
      if (yma*ymb == 1) then
        dye(1:nc,n) = half*(aa(1:nc) + bb(1:nc))
      else
        dye(1:nc,n) = aa(1:nc) + bb(1:nc)
      endif
    endif
  enddo
!$OMP END PARALLEL


  if (nbpz > 0) then
    do n=1,nbpz
      i=krz(1,n)
      j=krz(2,n)
      ntyp=abs(krz(3,n))
      if     (ntyp == 3) then
        dxe(1:nc,me(i,j)) = eis(1:nc,me(i,j))-eis(1:nc,me(i,j-1))
      elseif (ntyp == 1) then
        dxe(1:nc,me(i,j)) = eis(1:nc,me(i,j+1))-eis(1:nc,me(i,j))
      elseif (ntyp == 4) then
        dye(1:nc,me(i,j)) = eis(1:nc,me(i-1,j))-eis(1:nc,me(i,j))
      elseif (ntyp == 2) then
        dye(1:nc,me(i,j)) = eis(1:nc,me(i,j))-eis(1:nc,me(i+1,j))
      endif
    enddo
  endif


  if (nbpu > 0) then
    do n=1,nbpu
      i= kru(1,n)
      j= kru(2,n)
      ntyp=kru(3,n)
      if     (ntyp == 3) then
        if (muvs(1,i,j-1) > 0) then
          dxe(1:nc,me(i,j))=eis(1:nc,me(i,j))-eis(1:nc,me(i,j-1))
        endif
      elseif (ntyp == 1) then
        if (muvs(1,i,j+1) > 0) then
          dxe(1:nc,me(i,j+1))=eis(1:nc,me(i,j+2))-eis(1:nc,me(i,j+1))
        endif
      endif
    enddo
  endif


  if (nbpv > 0) then
    do n=1,nbpv
      i= krv(1,n)
      j= krv(2,n)
      ntyp=krv(3,n)
      if     (ntyp == 4) then
        if (muvs(2,i-1,j) > 0) then
          dye(1:nc,me(i,j))=eis(1:nc,me(i-1,j))-eis(1:nc,me(i,j))
        endif
      elseif (ntyp == 2) then
        if (muvs(2,i+1,j) > 0) then
          dye(1:nc,me(i+1,j))=eis(1:nc,me(i+1,j))-eis(1:nc,me(i+2,j))
        endif
      endif
    enddo
  endif


!$OMP PARALLEL DEFAULT (SHARED)                               &
!$OMP private(n,nl,nu,i,j,udye,udyw,vdxn,vdxs,fac,facx,facy,  &
!$OMP         div,ddd,emax,emin,pin,qin,pout,qout,sx,sy,ic)

  call domp_get_domain(1,n2d, nl, nu)

  facy = dtddy
  do n=nl,nu
    i=ind(1,n)
    j=ind(2,n)

    deu(1:nc,n) = zero

    if (muvs(1,i,j) == 0) cycle
    if (casus(n).and.casus(me(i,j+1))) cycle

    sx = dtddx*u(n)/cosphi(1,i)

    if (u(n) >= zero) then

      deu(1:nc,n) = onehalf*(one-sx)*dxe(1:nc,n)

      if (v(me(i-1,j)) < zero) then
        sy = facy*v(me(i-1,j))

        deu(1:nc,n) = deu(1:nc,n)                                              &
        -sy*( (oneforth-onethird*sx)*(dxe(1:nc,me(i-1,j))-dxe(1:nc,n))         &
             -(oneforth+onesixth*sy)*(dye(1:nc,me(i-1,j))-dye(1:nc,n)))
      endif

      if (v(n) > zero) then
        sy = facy*v(n)

        deu(1:nc,n) = deu(1:nc,n)                                              &
        -sy*( (oneforth-onethird*sx)*(dxe(1:nc,n)-dxe(1:nc,me(i+1,j)))         &
             +(oneforth-onesixth*sy)*(dye(1:nc,n)-dye(1:nc,me(i+1,j))))
      endif

    else

      deu(1:nc,n) = onehalf*(one+sx)*(-dxe(1:nc,me(i,j+1)))

      if (v(me(i-1,j+1)) < zero) then
        sy = facy*v(me(i-1,j+1))

        deu(1:nc,n) = deu(1:nc,n)                                              &
        -sy*(-(oneforth+onethird*sx)                                           &
              *(dxe(1:nc,me(i-1,j+1))-dxe(1:nc,me(i,j+1)))                     &
             -(oneforth+onesixth*sy)                                           &
              *(dye(1:nc,me(i-1,j+1))-dye(1:nc,me(i,j+1))))
      endif

      if (v(me(i,j+1)) > zero) then
        sy = facy*v(me(i,j+1))

        deu(1:nc,n) = deu(1:nc,n)                                              &
        -sy*(-(oneforth+onethird*sx)                                           &
              *(dxe(1:nc,me(i,j+1))-dxe(1:nc,me(i+1,j+1)))                     &
             +(oneforth-onesixth*sy)                                           &
              *(dye(1:nc,me(i,j+1))-dye(1:nc,me(i+1,j+1))))
      endif

    endif
  enddo


  do n=nl,nu
    i=ind(1,n)
    j=ind(2,n)

    dev(1:nc,n) = zero

    if(muvs(2,i,j) == 0) cycle
    if (casus(n).and.casus(me(i+1,j))) cycle

    sy   = dtddy*v(n)
    facx = dtddx/cosphi(1,i)

    if (v(n) <= zero) then

      dev(1:nc,n) = onehalf*(one+sy)*(-dye(1:nc,n))

      if (u(me(i,j-1)) > zero) then
        sx = facx*u(me(i,j-1))

        dev(1:nc,n) = dev(1:nc,n)                                              &
        -sx*(-(oneforth+onethird*sy)*(dye(1:nc,me(i,j-1))-dye(1:nc,n))         &
             +(oneforth-onesixth*sx)*(dxe(1:nc,me(i,j-1))-dxe(1:nc,n)))
      endif

      if (u(n) < zero) then
        sx = facx*u(n)

        dev(1:nc,n) = dev(1:nc,n)                                              &
        -sx*(-(oneforth+onethird*sy)*(dye(1:nc,n)-dye(1:nc,me(i,j+1)))         &
             -(oneforth+onesixth*sx)*(dxe(1:nc,n)-dxe(1:nc,me(i,j+1))))
      endif

    else

      dev(1:nc,n) = onehalf*(one-sy)*dye(1:nc,me(i+1,j))

      if (u(me(i+1,j-1)) > zero) then
        sx = facx*u(me(i+1,j-1))

        dev(1:nc,n) = dev(1:nc,n)                                              &
        -sx*( (oneforth-onethird*sy)                                           &
              *(dye(1:nc,me(i+1,j-1))-dye(1:nc,me(i+1,j)))                     &
             +(oneforth-onesixth*sx)                                           &
              *(dxe(1:nc,me(i+1,j-1))-dxe(1:nc,me(i+1,j))))
      endif

      if (u(me(i+1,j)) < zero) then
        sx = facx*u(me(i+1,j))

        dev(1:nc,n) = dev(1:nc,n)                                              &
        -sx*( (oneforth-onethird*sy)                                           &
              *(dye(1:nc,me(i+1,j))-dye(1:nc,me(i+1,j+1)))                     &
             -(oneforth+onesixth*sx)                                           &
              *(dxe(1:nc,me(i+1,j))-dxe(1:nc,me(i+1,j+1))))
      endif

    endif
  enddo

!$OMP BARRIER

  do n=nl,nu
    i = ind(1,n)
    j = ind(2,n)

    udye(1:nc) = u(n)*deu(1:nc,n)
    udyw(1:nc) = u(me(i,j-1))*deu(1:nc,me(i,j-1))
    
    vdxs(1:nc) = v(n)*dev(1:nc,n)*cosphi(2,i)
    vdxn(1:nc) = v(me(i-1,j))*dev(1:nc,me(i-1,j))*cosphi(2,i-1)

    fac  = one/(cosphi(1,i)*dxdy)
    facx = dtdy*fac
    facy = dtdx*fac

    do ic=1,nc
      rout(ic,n) = facx*(  max(udye(ic),zero) - min(udyw(ic),zero) )           &
                 + facy*(- min(vdxs(ic),zero) + max(vdxn(ic),zero) ) 

      rin (ic,n) = facx*(- min(udye(ic),zero) + max(udyw(ic),zero) )           &
                 + facy*(  max(vdxs(ic),zero) - min(vdxn(ic),zero) )
    enddo
  enddo

! barrier here due to rin, rout
!$OMP BARRIER

  do n=nl,nu
    i=ind(1,n)
    j=ind(2,n)

    facx = dtddx/cosphi(1,i)
    facy = dtddy/cosphi(1,i)
    div = (u(n)-u(me(i,j-1)))*facx                                             &
        + (cosphi(2,i-1)*v(me(i-1,j))-cosphi(2,i)*v(n))*facy
    fac = max(one-div,zero)
    do ic=1,nc
      ddd      = fac*eis(ic,n)
      emax(ic) = max(ddd,ee(ic,n))
      emin(ic) = min(ddd,ee(ic,n))
    enddo

    if (me(i-1,j) > 0) then
      div = (u(me(i-1,j))-u(me(i-1,j-1)))*dtddx/cosphi(1,i-1)                  &
          + (cosphi(2,i-2)*v(me(i-2,j))-cosphi(2,i-1)*v(me(i-1,j)))            &
           *dtddy/cosphi(1,i-1)
      fac = max(one-div,zero)
      do ic=1,nc
        ddd      = fac*eis(ic,me(i-1,j))
        emax(ic) = max(emax(ic),ddd,ee(ic,me(i-1,j)))
        emin(ic) = min(emin(ic),ddd,ee(ic,me(i-1,j)))
      enddo
    endif
    if (me(i+1,j) > 0) then
      div = (u(me(i+1,j))-u(me(i+1,j-1)))*dtddx/cosphi(1,i+1)                  &
          + (cosphi(2,i)*v(n)-cosphi(2,i+1)*v(me(i+1,j)))                      &
           *dtddy/cosphi(1,i+1)
      fac = max(one-div,zero)
      do ic=1,nc
        ddd      = fac*eis(ic,me(i+1,j))
        emax(ic) = max(emax(ic),ddd,ee(ic,me(i+1,j)))
        emin(ic) = min(emin(ic),ddd,ee(ic,me(i+1,j)))
      enddo
    endif
    if (me(i,j-1) > 0) then
      div = (u(me(i,j-1))-u(me(i,j-2)))*facx                                   &
          + (cosphi(2,i-1)*v(me(i-1,j-1))-cosphi(2,i)*v(me(i,j-1)))*facy        
      fac = max(one-div,zero)
      do ic=1,nc
        ddd      = fac*eis(ic,me(i,j-1))
        emax(ic) = max(emax(ic),ddd,ee(ic,me(i,j-1)))
        emin(ic) = min(emin(ic),ddd,ee(ic,me(i,j-1)))
      enddo
    endif
    if (me(i,j+1) > 0) then
      div = (u(me(i,j+1))-u(n))*facx                                           &
          + (cosphi(2,i-1)*v(me(i-1,j+1))-cosphi(2,i)*v(me(i,j+1)))*facy
      fac = max(one-div,zero)
      do ic=1,nc
        ddd      = fac*eis(ic,me(i,j+1))
        emax(ic) = max(emax(ic),ddd,ee(ic,me(i,j+1)))
        emin(ic) = min(emin(ic),ddd,ee(ic,me(i,j+1)))
      enddo
    endif

    do ic=1,nc
      pin=rin(ic,n)
      !     pin>=0
      if (pin > tm12) then
        qin=emax(ic)-ee(ic,n)
        !     qin>=0
        rin(ic,n)=min(one,qin/pin)
      else
        rin(ic,n)=one
      endif

      pout=rout(ic,n)
      !     pout>=0
      if (pout > tm12) then
        qout=ee(ic,n)-emin(ic)
        !     qout>=0
        rout(ic,n)=min(one,qout/pout)
      else
        rout(ic,n)=one
      endif
    enddo

  enddo

!$OMP BARRIER

  do n=nl, nu
    i=ind(1,n)
    j=ind(2,n)
    cx(1:nc,n) = zero
    if (muvs(1,i,j) == 0) cycle
    do ic=1,nc
      if (u(n)*deu(ic,n) >= zero) then
        cx(ic,n)=min(rin(ic,me(i,j+1)),rout(ic,n))
      else
        cx(ic,n)=min(rin(ic,n),rout(ic,me(i,j+1)))
      endif
    enddo
  enddo

  do n=nl, nu
    i=ind(1,n)
    j=ind(2,n)
    cy(1:nc,n) = zero
    if(muvs(2,i,j) == 0) cycle
    do ic=1,nc
      if (v(n)*dev(ic,n) <= zero) then
        cy(ic,n)=min(rin(ic,me(i+1,j)),rout(ic,n))
      else
        cy(ic,n)=min(rin(ic,n),rout(ic,me(i+1,j)))
      endif
    enddo
  enddo

! barrier here due to cx, cy
!$OMP BARRIER

  do n=nl,nu
    i=ind(1,n)
    j=ind(2,n)

    udye(1:nc) = u(n)*cx(1:nc,n)*deu(1:nc,n)
    udyw(1:nc) = u(me(i,j-1))*cx(1:nc,me(i,j-1))*deu(1:nc,me(i,j-1))

    vdxs(1:nc) = v(n)*cosphi(2,i)*cy(1:nc,n)*dev(1:nc,n)
    vdxn(1:nc) = v(me(i-1,j))*cosphi(2,i-1)*cy(1:nc,me(i-1,j))                 &
                             *dev(1:nc,me(i-1,j))

    fac  = one/(cosphi(1,i)*dxdy)
    facx = dtdy*fac
    facy = dtdx*fac

    eis(1:nc,n) = ee(1:nc,n) + facx*(-udye(1:nc)+udyw(1:nc))                   &
                             + facy*( vdxs(1:nc)-vdxn(1:nc))

  enddo
!$OMP END PARALLEL

  !    ende advektion


end subroutine eflow_int

  ! ----------------------------------------------------------------------------

  subroutine eflow_alloc (n2d,nc)
    ! poor-mans allocater/re-allocater for eflow arrays:

    integer(4), intent(in) :: n2d,nc

    !- surface arrays:
    if (n2d > Size2D) then
      Size2D = n2d
      if (allocated(ee)) deallocate(ee,rin,rout,eu,ev,dxe,dye,deu,dev,cx,cy)
      allocate(ee(1:nc,0:n2d),eu(1:nc,0:n2d),ev(1:nc,0:n2d))
      allocate(rin(1:nc,0:n2d),rout(1:nc,0:n2d))
      allocate(dxe(1:nc,0:n2d),dye(1:nc,0:n2d),deu(1:nc,0:n2d),dev(1:nc,0:n2d))
      allocate(cx(1:nc,0:n2d),cy(1:nc,0:n2d))
    endif

  end subroutine eflow_alloc

  ! ----------------------------------------------------------------------------

end module eflow

