module cmod_hydrodynamics
  !-----------------------------------------------------------------------------
  !  This module wraps the solutions of the nested hydrodynamic equations.
  !  It uses recursion, since solution of nesting level #lev requires 
  !  solution of the next higher level #lev+1.
  !-----------------------------------------------------------------------------

  implicit none
  private

  public  :: SolveHydrodynamics
  private :: SolveMomEq, SolveMassEq, SolveMassEq_z, SolveMassEq_dyn,          &
             SolveNextLevel, sumuvwi_n2o, MassNest_z, MassNest_dyn
contains

  !=============================================================================

  recursive subroutine SolveHydrodynamics ( level )

    use cmod_params,   only : narea, onthislevel, maxtimelevels, timelevelsize
    use cmod_miscvars, only : NDTFC

    implicit none

    integer(4), intent(in) :: level

    integer(4) :: ia, itl, itn, maxitl


    !- obtain max no of iterations on this level -------------------------------
    if (maxtimelevels == 2) then
      if (level == 1) then
        maxitl = 1
      else
        maxitl = NDTFC
      endif
    else
      maxitl = min(2,level)
    endif


    !- run through the partial time levels at the present time nesting level ---
    do itl=1,maxitl

      !- solve momentum equations ----------------------------------------------
      do ia=1,narea
        if (onthislevel(ia,level)) call SolveMomEq (ia,itl)
      enddo


      !- solve the next level hydrodynamics ------------------------------------
      call SolveNextLevel ( level+1 )


      !- solve the mass equations ----------------------------------------------
      do ia=1,narea
        if (onthislevel(ia,level)) then
          itn = timelevelsize(ia)
          call SolveMassEq (ia,itn)
        endif
      enddo

    enddo

  end subroutine SolveHydrodynamics

  !=============================================================================

  recursive subroutine SolveNextLevel ( level )

    use cmod_params, only : maxtimelevels

    implicit none

    integer(4), intent(in) :: level

    if (level <= maxtimelevels) then
      call SolveHydrodynamics ( level )
    endif

  end subroutine SolveNextLevel

  !=============================================================================

  subroutine SolveMomEq (ia, itl)
  !- solve mom equs on domain #ia

    use dmi_mpi,           only : dmpi_halo, mpi_rank, dmpi_gather,            &
                                  nftable, mctable
    use cmod_params,       only : mainarea, mmx, nmx, kmx, mmxp, nz, krz,      &
                                  iga, jga, kfg, nud, nvd, nweir,              &
                                  nestingfrom, nestingto, znesting
    use cmod_arrays,       only : idx, ft, rt, ahc, dt, dx, dxx, tx,           &
                                  gtx, dy ,ty, gty, cosphi, mm1k, kh, wrk,wrk2,&
                                  weir_idx, zweir, wflowr, kdry
    use cmod_local_arrays, only : iw2_l, iw3_l, ind_l, mm1k_l, me_l,           &
                                  kh_l, khu_l, khv_l,                          &
                                  u_l, v_l, un_l, vn_l, z_l, h_l, hu_l, hv_l,  &
                                  taux_l, tauy_l, press_l, geopot_l, rho_l,    &
                                  pl_l, hflux_l, ueis_l, veis_l,               &
                                  wu_l, wv_l, casus_l, ice_l,                  &
                                  sm1_l, sm2_l, eddyh_l, sm3_l, smu_l, smv_l,  &
                                  avv_l, dispv_l, z0srf_l, uvdam_l,            &
                                  tke_l, diss_l, cmp_l, atemp_l, sbfmap_l
    use momeqs_wrapping,   only : momeqs
    use cmod_timer,        only : timer, timer_print
    use turbmodels,        only : TurbModel
    use constants,         only : dyn, izeits, CSMAG, CDIV, iptracers,         &
                                  coupler_active, half
    use cmod_miscvars,     only : NDTEDDY,NDTADVEC,NDTVDIFF,NDTTHERMO,NDTSMAG
    use smagorinsky,       only : smag, deform, calc_uvterms
    use nesting,           only : mom_c_f, mom_drying_bnd_f_c
    use mom_heights_wrap,  only : mom_heights_dyn, mom_heights_srf
    use metsubs,           only : wspang
    use boundary,          only : vobcalc, vabcalc
    use geopotential,      only : hydrostat

    implicit none

    integer(4), intent(in) :: ia, itl

    integer(4) :: iao, iia, nz1, nz2, ii
    real(8)    :: dteddy


! MPI: do halo swap here for: h in the surface layer
! MPI: do halo swap here for: u,v, z, press, rho
! MPI   Not needed for ueis,veis,ice,casus 'coz we did scatter/swaps in main 
    call timer(1,'MPI_halo')
    call dmpi_halo(1,      mm1k_l(ia)%p,ia,h_l(ia)%p,z_l(ia)%p)
    call dmpi_halo(kmx(ia),mm1k_l(ia)%p,ia,u_l(ia)%p,v_l(ia)%p)
    if ((mod(izeits-1,NDTADVEC) == 0 .or. mod(izeits-1,NDTVDIFF) == 0 .or.     &
         mod(izeits-1,NDTTHERMO) == 0) .and. itl == 1) then
      if (.not. dyn) then
        call dmpi_halo(kmx(ia),mm1k_l(ia)%p,ia,press_l(ia)%p,rho_l(ia)%p)
      else
        call dmpi_halo(kmx(ia),mm1k_l(ia)%p,ia,rho_l(ia)%p)
      endif
    endif
    call timer(2,'MPI_halo')

! MPI: Halo swap already available for: h in the surface layer
!$OMP PARALLEL DEFAULT (shared)
    if (dyn) then
      !- update layer heights at momentum points:
      call mom_heights_dyn(iw2_l(ia),mm1k_l(ia)%p, kh_l(ia)%p, khu_l(ia)%p,    &
                           khv_l(ia)%p, ind_l(ia)%p, idx(ia)%p, h_l(ia)%p,     &
                           hu_l(ia)%p, hv_l(ia)%p )
    else
      !- update top layer heights at momentum points:
      call mom_heights_srf(iw2_l(ia), mm1k_l(ia)%p, kh_l(ia)%p, ind_l(ia)%p,   &
                           h_l(ia)%p, hu_l(ia)%p, hv_l(ia)%p)
    endif

    !- windstress
    ! MPI: NO need for halo swap 
    call wspang(iw2_l(ia),me_l(ia)%p,ind_l(ia)%p,taux_l(ia)%p,tauy_l(ia)%p,    &
                wu_l(ia)%p,wv_l(ia)%p,u_l(ia)%p, v_l(ia)%p,                    &
                atemp_l(ia)%p, cmp_l(ia)%p)

    !- calculate shear and stretch:
    if (NDTSMAG==0 .or. (mod(izeits-1,NDTSMAG) == 0 .and. itl == 1)) then
      call deform (mm1k_l(ia)%p,ind_l(ia)%p,iw2_l(ia),                         &
                 half/dxx(ia)%p(1),half/dy(ia),kh_l(ia)%p,                     &
                 u_l(ia)%p,v_l(ia)%p,sm1_l(ia)%p,sm2_l(ia)%p,sm3_l(ia)%p,      &
                 cosphi(ia)%p,idx(ia)%p)
    endif
!$OMP END PARALLEL

! MPI: do halo swap here for: taux, tauy (needed in momeqs, TurbModel)
!                             sm1 (needed in smag)
    call timer(1,'MPI_halo')
    call dmpi_halo(1,mm1k_l(ia)%p,ia,taux_l(ia)%p,tauy_l(ia)%p)
    if (NDTSMAG==0 .or. (mod(izeits-1,NDTSMAG) == 0 .and. itl == 1)) then
      call dmpi_halo(kmx(ia),mm1k_l(ia)%p,ia,sm1_l(ia)%p)
    endif
    call timer(2,'MPI_halo')

    call timer(1,'turbulence')
!$OMP PARALLEL DEFAULT (shared) PRIVATE (dteddy)
    !- calculate horizontal eddy:
    if (NDTSMAG==0 .or. (mod(izeits-1,NDTSMAG) == 0 .and. itl == 1)) then
      if (NDTSMAG==0) then
        dteddy = dt(ia)
      else
        dteddy = NDTSMAG*dt(mainarea)
      endif
      call smag (mmxp(ia),kmx(ia),mm1k_l(ia)%p,ind_l(ia)%p,iw2_l(ia),          &
                 dxx(ia)%p(1),dy(ia),dteddy,CSMAG,ahc(ia),kh_l(ia)%p,          &
                 sm1_l(ia)%p,sm2_l(ia)%p,cosphi(ia)%p,                         &
                 eddyh_l(ia)%p,CDIV,idx(ia)%p,sm3_l(ia)%p)
    endif

    !- eddy viscosity:
    if (mod(izeits-1,NDTEDDY) == 0 .and. itl == 1) then
      dteddy = NDTEDDY*dt(mainarea)
      call TurbModel (kmx(ia), iw2_l(ia), mm1k_l(ia)%p, ind_l(ia)%p,           &
                      kh_l(ia)%p, dteddy, rt(ia), casus_l(ia)%p,               &
                      ice_l(ia)%p, ueis_l(ia)%p, veis_l(ia)%p,                 &
                      hflux_l(ia)%p, wu_l(ia)%p, wv_l(ia)%p,                   &
                      taux_l(ia)%p, tauy_l(ia)%p, h_l(ia)%p, rho_l(ia)%p,      &
                      u_l(ia)%p, v_l(ia)%p, avv_l(ia)%p, dispv_l(ia)%p,        &
                      tke_l(ia)%p, diss_l(ia)%p, z0srf_l(ia)%p, cmp_l(ia)%p,   &
                      idx(ia)%p, iptracers>0 .or. coupler_active)
    endif
!$OMP END PARALLEL
    call timer(2,'turbulence')
    call timer_print('TS - turbulence took: ','turbulence',50)

! MPI: Do halo swap here for: eddyh, sm2, sm3, avv, z0srf (,pl)
    call timer(1,'MPI_halo')
    if (mod(izeits-1,NDTEDDY) == 0 .and. itl == 1) then
      call dmpi_halo(kmx(ia),mm1k_l(ia)%p,ia,avv_l(ia)%p)
      call dmpi_halo(1,      mm1k_l(ia)%p,ia,z0srf_l(ia)%p) 
    endif
    if (NDTSMAG==0 .or. (mod(izeits-1,NDTSMAG) == 0 .and. itl == 1)) then
      call dmpi_halo(kmx(ia),mm1k_l(ia)%p,ia,sm2_l(ia)%p,sm3_l(ia)%p,          &
                     eddyh_l(ia)%p)
    endif
!   call dmpi_halo(1      ,mm1k_l(ia)%p,ia,pl_l(ia)%p)
    call timer(2,'MPI_halo')


    if (NDTSMAG==0 .or. (mod(izeits-1,NDTSMAG) == 0 .and. itl == 1)) then
      if (NDTSMAG==0) then
        dteddy = dt(ia)
      else
        dteddy = NDTSMAG*dt(mainarea)
      endif
!$OMP PARALLEL DEFAULT(shared)
      call calc_uvterms(mmx(ia),iw2_l(ia),mm1k_l(ia)%p,                        &
                        dteddy,dx(ia)%p,ty(ia),tx(ia)%p,ind_l(ia)%p,           &
                        kh_l(ia)%p,khu_l(ia)%p,khv_l(ia)%p,idx(ia)%p,          &
                        sm1_l(ia)%p,sm2_l(ia)%p,sm3_l(ia)%p,eddyh_l(ia)%p,     &
                        smu_l(ia)%p,smv_l(ia)%p)
!$OMP END PARALLEL
      call dmpi_halo(kmx(ia),mm1k_l(ia)%p,ia,smu_l(ia)%p,smv_l(ia)%p)
    endif

    if (dyn) then
      call timer(1,'hydrostat')
!$OMP PARALLEL DEFAULT(shared)
      call hydrostat ( mm1k_l(ia)%p, ind_l(ia)%p, iw2_l(ia), iw3_l(ia),        &
                       kh_l(ia)%p, h_l(ia)%p, rho_l(ia)%p, geopot_l(ia)%p,     &
                       press_l(ia)%p )
!$OMP END PARALLEL
! MPI: halo swap: geopot, press 
      call dmpi_halo(kmx(ia),mm1k_l(ia)%p,ia,press_l(ia)%p,geopot_l(ia)%p)
      call timer(2,'hydrostat')
      call timer_print('TS - hydrostat took: ','hydrostat',50)
    endif

    if (ia /= mainarea) then
      !  Pre-post the receives for the mom_c_f nesting:
      do iao=1,nestingfrom(ia)%p(0)
        !  ia:  fine grid,   iia: coarse grid
        iia = nestingfrom(ia)%p(iao)
        nz1 = znesting(ia)%p(iao,1)
        nz2 = znesting(ia)%p(iao,2)

        call timer(1,'MPI_nest_1')
        call dmpi_gather(ia, iia, nz1, nz2, krz(ia)%p,                         &
                         nftable(mpi_rank+1,iia,ia), .false.,                  &
                         kh_l(iia)%p, kh(iia)%p, kmx(iia),                     &
                         mm1k_l(iia)%p, mm1k(iia)%p, mm1k(ia)%p,               &
                         iga(ia)%p(:,iao), jga(ia)%p(:,iao),                   &
                         un_l(iia)%p, wrk(iia)%p, vn_l(iia)%p, wrk2(iia)%p,    &
                         iao, 1)
        call timer(2,'MPI_nest_1')
      enddo
      !  Encode buffers and pre-post the sends for the mom_c_f nesting:
      do iao=1,nestingfrom(ia)%p(0)
        !  ia:  fine grid,   iia: coarse grid
        iia = nestingfrom(ia)%p(iao)
        nz1 = znesting(ia)%p(iao,1)
        nz2 = znesting(ia)%p(iao,2)

        call timer(1,'MPI_nest_1')
        call dmpi_gather(ia, iia, nz1, nz2, krz(ia)%p,                         &
                         .false., mctable(iia,ia),                             &
                         kh_l(iia)%p, kh(iia)%p, kmx(iia),                     &
                         mm1k_l(iia)%p, mm1k(iia)%p, mm1k(ia)%p,               &
                         iga(ia)%p(:,iao), jga(ia)%p(:,iao),                   &
                         un_l(iia)%p, wrk(iia)%p, vn_l(iia)%p, wrk2(iia)%p,    &
                         iao, 2)
        call timer(2,'MPI_nest_1')
      enddo
    endif
    
    call timer(1,'momeqs')
! MPI: Halo already available: shear, stretch, div, eddyh, eddyd, avv, u, v, 
! MPI                          ueis, veis, ice, casus,
! MPI                          pl, z, press, rho, z0srf
    !- momentum equation:
    call momeqs (kdry(ia), iw2_l(ia),mm1k_l(ia)%p,dt(ia),                      &
                 ft(ia)%p,rt(ia),ty(ia),tx(ia)%p,                              &
                 gty(ia),gtx(ia)%p,nud(ia),nvd(ia),uvdam_l(ia)%p,              &
                 nweir(ia),z_l(ia)%p,u_l(ia)%p,v_l(ia)%p,                      &
                 h_l(ia)%p,un_l(ia)%p,vn_l(ia)%p,                              &
                 hu_l(ia)%p,hv_l(ia)%p,khu_l(ia)%p,khv_l(ia)%p,taux_l(ia)%p,   &
                 tauy_l(ia)%p,pl_l(ia)%p,press_l(ia)%p,geopot_l(ia)%p,         &
                 rho_l(ia)%p, avv_l(ia)%p, smu_l(ia)%p, smv_l(ia)%p,           &
                 casus_l(ia)%p,ice_l(ia)%p,ueis_l(ia)%p,veis_l(ia)%p,          &
                 ind_l(ia)%p,kh_l(ia)%p,z0srf_l(ia)%p,idx(ia)%p,               &
                 weir_idx(ia)%p,zweir(ia)%p,wflowr(ia)%p,ia,sbfmap_l(ia)%p )
    call timer(2,'momeqs')
    call timer_print('TS - momeqs took: ','momeqs',50)

    !- do the open boundaries or the nesting to enclosing domain:
    if (ia == mainarea) then
! MPI: Halo already available: u, v
! MPI  but not for un, vn which are needed if we only have the obc point on the
!      task but not its neighbour just inside the domain. Don't know why we
!      should set up such a stupid case, but you'll never know, so ...
! FIXME: consider making rules against such cases.
      !- fix velocity along boundary (VAB):
      call vabcalc (mm1k_l(ia)%p,khu_l(ia)%p,khv_l(ia)%p,u_l(ia)%p,v_l(ia)%p,  &
                    un_l(ia)%p,vn_l(ia)%p,nz(ia),krz(ia)%p,kdry(ia),           &
                    h_l(ia)%p,z_l(ia)%p)

! MPI: Halo already available: h, casus, u, v, ueis, veis, avv, ice, press, 
! MPI                          rho, z, pl
! MPI  but not for un, vn (none needed)
      !- fix for velocity outside boundary (VOB):
      call vobcalc (kmx(ia),kdry(ia),mm1k_l(ia)%p,me_l(ia)%p,dt(ia),ft(ia)%p,  &
                    rt(ia),ty(ia),tx(ia)%p,gty(ia),gtx(ia)%p,                  &
                    z_l(ia)%p,u_l(ia)%p,v_l(ia)%p,                             &
                    h_l(ia)%p,un_l(ia)%p,vn_l(ia)%p,                           &
                    taux_l(ia)%p,tauy_l(ia)%p,pl_l(ia)%p,                      &
                    press_l(ia)%p,geopot_l(ia)%p,rho_l(ia)%p,avv_l(ia)%p,      &
                    casus_l(ia)%p,ice_l(ia)%p,ueis_l(ia)%p,veis_l(ia)%p,       &
                    (2),nz(ia),krz(ia)%p,z0srf_l(ia)%p )
    else
      !- fix velocity by nesting:
      do iao=1,nestingfrom(ia)%p(0)
        !  ia:  fine grid
        !  iia: coarse grid
        iia = nestingfrom(ia)%p(iao)
        ii  = 1
        do while (ia /= nestingto(iia)%p(ii))
          ii = ii + 1
        enddo

        !  gather coarse grid velocity to global arrays:
        nz1 = znesting(ia)%p(iao,1)
        nz2 = znesting(ia)%p(iao,2)

        call timer(1,'MPI_nest_1')
        !  Decode the buffers:
        call dmpi_gather(ia, iia, nz1, nz2, krz(ia)%p,                         &
                         nftable(mpi_rank+1,iia,ia), mctable(iia,ia),          &
                         kh_l(iia)%p, kh(iia)%p, kmx(iia),                     &
                         mm1k_l(iia)%p, mm1k(iia)%p, mm1k(ia)%p,               &
                         iga(ia)%p(:,iao), jga(ia)%p(:,iao),                   &
                         un_l(iia)%p, wrk(iia)%p, vn_l(iia)%p, wrk2(iia)%p,    &
                         iao, 3)
        call timer(2,'MPI_nest_1')

        !  use coarse grid global velocity arrays:

!$OMP PARALLEL DEFAULT (shared)
        call mom_c_f(nmx(ia),kmx(ia),kdry(ia),mm1k_l(ia)%p,me_l(ia)%p,dt(ia),  &
                     ft(ia)%p,rt(ia),ty(ia),tx(ia)%p,gty(ia),gtx(ia)%p,        &
                     z_l(ia)%p,u_l(ia)%p,v_l(ia)%p,h_l(ia)%p,un_l(ia)%p,       &
                     vn_l(ia)%p,taux_l(ia)%p,tauy_l(ia)%p,pl_l(ia)%p,          &
                     press_l(ia)%p,geopot_l(ia)%p,rho_l(ia)%p,avv_l(ia)%p,     &
                     casus_l(ia)%p,ice_l(ia)%p,ueis_l(ia)%p,veis_l(ia)%p,      &
                     kmx(iia),iga(ia)%p(:,iao),jga(ia)%p(:,iao),krz(ia)%p,     &
                     mm1k(iia)%p,kfg(iia)%p(:,ii),wrk(iia)%p,wrk2(iia)%p,      &
                     nz1,nz2,z0srf_l(ia)%p )
!$OMP END PARALLEL

        if (dyn) then
          !- check coarse grid velocities at boundary in case of drying
          call mom_drying_bnd_f_c (nz1, nz2, krz(ia)%p,mm1k_l(ia)%p,kh_l(ia)%p,&
                                   kfg(iia)%p(:,ii), iga(ia)%p(:,iao),         &
                                   jga(ia)%p(:,iao), h_l(ia)%p, mm1k_l(iia)%p, &
                                   h_l(iia)%p, z_l(iia)%p,                     &
                                   un_l(iia)%p, vn_l(iia)%p)
        endif

      enddo
    endif

  end subroutine SolveMomEq

  !=============================================================================

  subroutine sumuvwi_n2o (hu,hv,ui,vi,wi,u,v,w,teil,lmn,iwet2,kh,ind,mm,idx,   &
                          uo,vo)
    !- modules -----------------------------------------------------------------
    use dmi_omp, only: domp_get_domain
  
    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)                :: iwet2
    integer(4), intent(in)                :: lmn
    integer(4), intent(in)                :: kh(0:), ind(:,:)
    integer(4), intent(in)                :: mm(:,0:,0:)
    real(8), intent(in)                   :: teil
    real(8), intent(in),    dimension(0:) :: hu, hv, u, v, w
    real(8), intent(inout), dimension(0:) :: ui, vi, wi, uo, vo
    integer(4), intent(inout)             :: idx(1:,1:)
        
    !- local variables ---------------------------------------------------------
    real(8)    :: fac0
    integer(4) :: nsrf, nl, nu, i, j, mi

    fac0 = teil/real(lmn,8)
  
    call domp_get_domain(kh, 1, iwet2, nl, nu, idx)

    ! k=1
    do nsrf=nl,nu
      uo(nsrf) = u(nsrf)
      ui(nsrf) = ui(nsrf) + hu(nsrf)*u(nsrf)*fac0
    enddo
    do nsrf=nl,nu
      vo(nsrf) = v(nsrf)
      vi(nsrf) = vi(nsrf) + hv(nsrf)*v(nsrf)*fac0
    enddo
    do nsrf=nl,nu
      wi(nsrf) = wi(nsrf) +          w(nsrf)*fac0
    enddo

    ! k>1
    do nsrf=nl,nu
      if (kh(nsrf) <= 1) cycle
      i=ind(1,nsrf)
      j=ind(2,nsrf)
!dir$ prefervector
      do mi=mm(2,i,j),mm(kh(nsrf),i,j)
        uo(mi) = u(mi)
        ui(mi) = ui(mi) + hu(mi)*u(mi)*fac0
      enddo
    enddo
    do nsrf=nl,nu
      if (kh(nsrf) <= 1) cycle
      i=ind(1,nsrf)
      j=ind(2,nsrf)
!dir$ prefervector
      do mi=mm(2,i,j),mm(kh(nsrf),i,j)
        vi(mi) = vi(mi) + hv(mi)*v(mi)*fac0
        vo(mi) = v(mi)
      enddo
    enddo
    do nsrf=nl,nu
      if (kh(nsrf) <= 1) cycle
      i=ind(1,nsrf)
      j=ind(2,nsrf)
!dir$ prefervector
      do mi=mm(2,i,j),mm(kh(nsrf),i,j)
        wi(mi) = wi(mi) +        w(mi)*fac0
      enddo
    enddo
  
  end subroutine sumuvwi_n2o

  !=============================================================================

  subroutine SolveMassEq (ia,itn)
  !- solve mass equs on domain #ia

    use dmi_mpi,           only : dmpi_halo, mpi_rank, dmpi_gather,            &
                                  uctable, uhtable, vctable, vhtable
    use cmod_params,       only : kmx, kfg, nestingto, unesting, vnesting,     &
                                  iga, jga, nestinglevels, enclosing
    use cmod_arrays,       only : kru, krv, idx, mm1k, wma3, wma4, wrk, wrk2
    use cmod_local_arrays, only : iw2_l, mm1k_l, ind_l, kh_l, hz_l, h_l, hn_l, &
                                  z_l, zn_l, hu_l, hv_l, un_l, vn_l, u_l, v_l, &
                                  w_l, ui_l, vi_l, wi_l, elev_ref_l
    use updatearrays,      only : update_h, update_z, set_h, update_h_ref
    use cmod_timer,        only : timer, timer_print
    use constants,         only : dyn, one, elevation_ref, zero
    use cmod_miscvars,     only : NDTADVEC
    use nesting,           only : mom_f_c

    implicit none

    integer(4), intent(in) :: ia, itn

    integer(4) :: ii, iia, iao, nu1, nu2, nv1, nv2

! MPI: do halo swap here for: hu, hv in surface and for un, vn
    call timer(1,'MPI_halo')
    call dmpi_halo(1,      mm1k_l(ia)%p,ia,hu_l(ia)%p,hv_l(ia)%p)
    call dmpi_halo(kmx(ia),mm1k_l(ia)%p,ia,un_l(ia)%p,vn_l(ia)%p)
    call timer(2,'MPI_halo')

    if (nestinglevels(ia) > 0) then
      !- fix velocity along border:
      do ii=1,nestinglevels(ia)
        !  iia: fine grid
        !  ia:  coarse grid
        iia = nestingto(ia)%p(ii)
        iao = enclosing(iia,ia)
        nu1 = unesting(ia)%p(ii,1)
        nu2 = unesting(ia)%p(ii,2)
        nv1 = vnesting(ia)%p(ii,1)
        nv2 = vnesting(ia)%p(ii,2)

        call timer(1,'MPI_nest_2')
        !  gather fine grid un, vn to global arrays along u and/or v brds:
        call dmpi_gather(ia,iia,nu1,nu2,kru(ia)%p,(1),                         &
                         uctable(mpi_rank+1,ia,iia), uhtable(ia,iia),          &
                         kmx(iia),mm1k_l(iia)%p,mm1k(iia)%p,mm1k(ia)%p,        &
                         iga(iia)%p(:,iao),jga(iia)%p(:,iao),                  &
                         un_l(iia)%p,wrk(iia)%p,vn_l(iia)%p,wrk2(iia)%p)
        call dmpi_gather(ia,iia,nv1,nv2,krv(ia)%p,(1),                         &
                         vctable(mpi_rank+1,ia,iia), vhtable(ia,iia),          &
                         kmx(iia),mm1k_l(iia)%p,mm1k(iia)%p,mm1k(ia)%p,        &
                         iga(iia)%p(:,iao),jga(iia)%p(:,iao),                  &
                         un_l(iia)%p,wrk(iia)%p,vn_l(iia)%p,wrk2(iia)%p)

! fixme: consider gathering ALL hu/hv during initializations and then 
!        only k=1 here during time loop.
        !  gather fine grid hu, hv to global arrays along u and/or v brds:
        call dmpi_gather(ia,iia,nu1,nu2,kru(ia)%p,(1),                         &
                         uctable(mpi_rank+1,ia,iia), uhtable(ia,iia),          &
                         kmx(iia),mm1k_l(iia)%p,mm1k(iia)%p,mm1k(ia)%p,        &
                         iga(iia)%p(:,iao),jga(iia)%p(:,iao),                  &
                         hu_l(iia)%p,wma3,hv_l(iia)%p,wma4)
        call dmpi_gather(ia,iia,nv1,nv2,krv(ia)%p,(1),                         &
                         vctable(mpi_rank+1,ia,iia), vhtable(ia,iia),          &
                         kmx(iia),mm1k_l(iia)%p,mm1k(iia)%p,mm1k(ia)%p,        &
                         iga(iia)%p(:,iao),jga(iia)%p(:,iao),                  &
                         hu_l(iia)%p,wma3,hv_l(iia)%p,wma4)
        call timer(2,'MPI_nest_2')
          

        wma3(0) = zero
        wma4(0) = zero
!$OMP PARALLEL DEFAULT (shared)
        call mom_f_c (kmx(ia), mm1k_l(ia)%p, un_l(ia)%p, vn_l(ia)%p,           &
                      hu_l(ia)%p, hv_l(ia)%p, kru(ia)%p, krv(ia)%p,            &
                      mm1k(iia)%p, iga(iia)%p(:,iao), jga(iia)%p(:,iao),       &
                      kfg(ia)%p(:,ii), wrk(iia)%p, wrk2(iia)%p,                &
                      wma3, wma4, nu1, nu2, nv1, nv2 )
!$OMP END PARALLEL

        call timer(1,'MPI_nest_3')
        call dmpi_halo(kmx(ia),mm1k_l(ia)%p,ia,hu_l(ia)%p,hv_l(ia)%p,          &
                                               un_l(ia)%p,vn_l(ia)%p)
        call timer(2,'MPI_nest_3')
      enddo
    endif

    if (dyn) then
      call SolveMassEq_dyn (ia)
    else
      call SolveMassEq_z (ia)
    endif

    call timer(1,'save_uvz')
!$OMP PARALLEL DEFAULT (shared)
! MPI: NO need for halo swap 
    if (.not.(dyn .or. elevation_ref)) then
      ! default case, z-grid wo/ref.elev.
      !- Update the surface layer thicknesses
      call update_h (hz_l(ia)%p, iw2_l(ia), zn_l(ia)%p,                        &
                     h_l(ia)%p, kh_l(ia)%p, idx(ia)%p)
    elseif ((.not.dyn) .and. elevation_ref) then
      !- Update the surface layer thicknesses
      call update_h_ref (hz_l(ia)%p, iw2_l(ia), zn_l(ia)%p, elev_ref_l(ia)%p,  &
                         h_l(ia)%p, kh_l(ia)%p, idx(ia)%p)
 
    else
      !- Save the old layer thicknesses (new -> old)
      call set_h (mm1k_l(ia)%p, ind_l(ia)%p, kh_l(ia)%p, hn_l(ia)%p, iw2_l(ia),&
                h_l(ia)%p, idx(ia)%p)
    endif
      
    !- save old zeta values:
    call update_z (zn_l(ia)%p, iw2_l(ia), z_l(ia)%p, kh_l(ia)%p, idx(ia)%p)


    !- update cell face fluxes and save old u,v values:
    call sumuvwi_n2o(hu_l(ia)%p, hv_l(ia)%p, ui_l(ia)%p, vi_l(ia)%p,wi_l(ia)%p,&
                  un_l(ia)%p, vn_l(ia)%p, w_l(ia)%p, one/real(itn,8), NDTADVEC,&
                  iw2_l(ia), kh_l(ia)%p, ind_l(ia)%p, mm1k_l(ia)%p, idx(ia)%p, &
                  u_l(ia)%p, v_l(ia)%p )
!$OMP END PARALLEL
    call timer(2,'save_uvz')
    call timer_print('TS - save_uvz took: ','save_uvz',70)

  end subroutine SolveMassEq

  !=============================================================================

  subroutine SolveMassEq_z (ia)
  !- solve mass equs on domain #ia

    use dmi_mpi_global,    only : mpi_io_rank
    use dmi_mpi,           only : dmpi_gather, mpi_io_serial, dmpi_broadcast,  &
                                  nftable, mctable, mpi_rank
    use cmod_params,       only : mainarea, kmx, nz, krz, nq, iga, jga, kfg,   &
                                  nestingfrom, nestingto, znesting
    use cmod_arrays,       only : kh, khu, khv, krq, idx, mm1k, rwz, rwq, dt,  &
                                  dx, dy, covo, covu, zetac, zetac2, pl,       &
                                  wrk, wrk2, wma3, wma4, obcwrk
    use cmod_local_arrays, only : iw2_l, mm1k_l, ind_l, kh_l,                  &
                                  zn_l, z_l, h_l, hu_l, hv_l,                  &
                                  un_l, vn_l, w_l,                             &
                                  pl_l, press_l, rho_l, precip_l, evapo_l
    use masseqs,           only : masseqs_solver_z
    use cmod_timer,        only : timer, timer_print
    use constants,         only : coldst, usezbcl, precevap, zbnd_type, zero
    use cmod_miscvars,     only : lunsurge, lmns, jstuna, OPRFMT, nb1, nb2
    use cmod_miscsubs,     only : CheckWL
    use boundary,          only : bndbclz, bndzcorr, gradiz, zrano, densbno,   &
                                  zetar, zbnd_rbc 
    use nesting,           only : rand_z, w_c_f
    use cmod_precip,       only : mass_precevap

    implicit none

    integer(4), intent(in) :: ia

    integer(4) :: ii, iia, iao, nz1, nz2


    if (nestingfrom(ia)%p(0) > 0) then
      ! Pre-post the receives for rand_z and w_c_f:
      do iao=1,nestingfrom(ia)%p(0)
        !  ia:  fine grid,   iia: coarse grid
        iia = nestingfrom(ia)%p(iao)
        nz1 = znesting(ia)%p(iao,1)
        nz2 = znesting(ia)%p(iao,2)
        
        call timer(1,'MPI_nest_4')
        call dmpi_gather(ia, iia, nz1, nz2, krz(ia)%p,                         &
                         nftable(mpi_rank+1,iia,ia), .false.,                  &
                         kh_l(iia)%p, kh(iia)%p, kmx(iia),                     &
                         mm1k_l(iia)%p, mm1k(iia)%p, mm1k(ia)%p,               &
                         iga(ia)%p(:,iao), jga(ia)%p(:,iao),                   &
                         hu_l(iia)%p, wma3, hv_l(iia)%p, wma4, iao, 1)
        call timer(2,'MPI_nest_4')
      enddo
      ! Encode and pre-post the sends for rand_z and w_c_f:
      do iao=1,nestingfrom(ia)%p(0)
        !  ia:  fine grid,   iia: coarse grid
        iia = nestingfrom(ia)%p(iao)
        nz1 = znesting(ia)%p(iao,1)
        nz2 = znesting(ia)%p(iao,2)
        
        call timer(1,'MPI_nest_4')
        call dmpi_gather(ia, iia, nz1, nz2, krz(ia)%p,                         &
                         .false., mctable(iia,ia),                             &
                         kh_l(iia)%p, kh(iia)%p, kmx(iia),                     &
                         mm1k_l(iia)%p, mm1k(iia)%p, mm1k(ia)%p,               &
                         iga(ia)%p(:,iao), jga(ia)%p(:,iao),                   &
                         hu_l(iia)%p, wma3, hv_l(iia)%p, wma4, iao, 2)
        call timer(2,'MPI_nest_4')
      enddo
    endif
    
    call timer(1,'masseqs')
!$OMP PARALLEL DEFAULT(shared)
! MPI: Halo already available: hu, hv, un, vn
    !- mass equation:
    call masseqs_solver_z ( iw2_l(ia), mm1k_l(ia)%p, ind_l(ia)%p, dt(ia),      &
                            dx(ia)%p, dy(ia), covo(ia)%p, covu(ia)%p,          &
                            z_l(ia)%p, w_l(ia)%p,                              &
                            zn_l(ia)%p, un_l(ia)%p, vn_l(ia)%p,                &
                            hu_l(ia)%p, hv_l(ia)%p, kh_l(ia)%p, idx(ia)%p)

! MPI: NO need for halo swap
    !- add precipitation-evaporation:
    if (precevap) call mass_precevap(iw2_l(ia), dt(ia), kh_l(ia)%p, idx(ia)%p, &
                                     zn_l(ia)%p, w_l(ia)%p,                    &
                                     precip_l(ia)%p, evapo_l(ia)%p )
!$OMP END PARALLEL

    !-  Do the nesting of z fine-->coarse:
    call MassNest_z (ia)

! MPI: NO need for halo swap
!$OMP PARALLEL DEFAULT(shared)
    !- check water level for possible drying:
    call CheckWL ( iw2_l(ia), zn_l(ia)%p, h_l(ia)%p, z_l(ia)%p )
!$OMP END PARALLEL
    call timer(2,'masseqs')
    call timer_print('TS - masseqs took: ','masseqs',50)

    !- boundary values:
    call gradiz (kmx(ia),mm1k_l(ia)%p,dt(ia),dx(ia)%p,dy(ia),                  &
                 w_l(ia)%p,zn_l(ia)%p,                                         &
                 nz(ia),krz(ia)%p,rwz(ia)%p,nq(ia),krq(ia)%p,rwq(ia)%p,        &
                 (ia == mainarea) )


    if (nestingfrom(ia)%p(0) > 0) then
      !- boundary conditions coarse grid --> fine grid:
      do iao=1,nestingfrom(ia)%p(0)
        !  ia:  fine grid
        !  iia: coarse grid
        iia = nestingfrom(ia)%p(iao)
        ii  = 1
        do while (ia /= nestingto(iia)%p(ii))
          ii = ii + 1
        enddo

        nz1 = znesting(ia)%p(iao,1)
        nz2 = znesting(ia)%p(iao,2)
        ! gather coarse hu/hv; did gather coarse un/vn before mom_c_f
        call timer(1,'MPI_nest_4')
        call dmpi_gather(ia, iia, nz1, nz2, krz(ia)%p,                         &
                         nftable(mpi_rank+1,iia,ia), mctable(iia,ia),          &
                         kh_l(iia)%p, kh(iia)%p, kmx(iia),                     &
                         mm1k_l(iia)%p, mm1k(iia)%p, mm1k(ia)%p,               &
                         iga(ia)%p(:,iao), jga(ia)%p(:,iao),                   &
                         hu_l(iia)%p, wma3, hv_l(iia)%p, wma4, iao, 3)
        call timer(2,'MPI_nest_4')

!$OMP PARALLEL DEFAULT(shared)
        call rand_z(kmx(iia), mm1k(iia)%p,                                     &
                    wrk(iia)%p, wrk2(iia)%p, wma3, wma4,                       &
                    kmx(ia),mm1k_l(ia)%p, iga(ia)%p(:,iao), jga(ia)%p(:,iao),  &
                    z_l(ia)%p, zn_l(ia)%p, un_l(ia)%p, vn_l(ia)%p,h_l(ia)%p,   &
                    hu_l(ia)%p, hv_l(ia)%p, krz(ia)%p, rwz(ia)%p,              &
                    dt(ia), dx(ia)%p, dy(ia), covo(ia)%p, covu(ia)%p,nz1,nz2)

        !- update w in fine grid along border:
        call w_c_f (krz(ia)%p, mm1k_l(ia)%p, kmx(iia),  mm1k(iia)%p,           &
                    kfg(iia)%p(:,ii), iga(ia)%p(:,iao), jga(ia)%p(:,iao),      &
                    z_l(ia)%p, zn_l(ia)%p, w_l(ia)%p, un_l(ia)%p, vn_l(ia)%p,  &
                    wrk(iia)%p, wrk2(iia)%p,                                   &
                    covo(ia)%p, covu(ia)%p, dx(ia)%p, dy(ia), dt(ia),          &
                    hu_l(ia)%p, hv_l(ia)%p, nz1, nz2 )
!$OMP END PARALLEL
      enddo

    elseif  (ia == mainarea) then

      !- gather press and rho to I/O task:
      call timer(1,'gs_2')
      call timer(1,'MPI_bnd')
      call dmpi_gather(ia, 1, nz(ia), krz(ia)%p, kh(ia)%p, mm1k(ia)%p,         &
                       mm1k_l(ia)%p, hu_l(ia)%p, hv_l(ia)%p, rho_l(ia)%p,      &
                       press_l(ia)%p, obcwrk )
      call timer(2,'MPI_bnd')

      if (mpi_io_serial .and. zbnd_type == 1) then
        !- zeta boundary values, surge and harmonic constants:
        call zrano (lunsurge,nz(ia),krz(ia)%p,rwz(ia)%p,lmns,coldst,OPRFMT)
      endif

      if (zbnd_type == 2) then
        call dmpi_gather(ia,kmx(ia), un_l(ia)%p, wrk(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,kmx(ia), vn_l(ia)%p,wrk2(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,1      , pl_l(ia)%p,  pl(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,kmx(ia), hu_l(ia)%p,wma3,mpi_io_rank)
        call dmpi_gather(ia,kmx(ia), hv_l(ia)%p,wma4,mpi_io_rank)
        wma3(0) = zero
        wma4(0) = zero
        if (mpi_io_serial) then
          !- zeta boundary values from radiation boundary conditions:
          call zbnd_rbc(mm1k(ia)%p, khu(ia)%p, khv(ia)%p, wrk(ia)%p,wrk2(ia)%p,&
                        wma3, wma4, pl(ia)%p, nz(ia), krz(ia)%p, rwz(ia)%p )
        endif
      endif

      if (mpi_io_serial) then
        call densbno (mm1k(ia)%p,khu(ia)%p,khv(ia)%p,                          &
                      obcwrk(:,:,1),obcwrk(:,:,2),obcwrk(:,:,3),obcwrk(:,:,4), &
                      nz(ia),krz(ia)%p,zetac )
      endif

      !- communicate the boundary values:
      call timer(1,'MPI_bnd')
      call dmpi_broadcast(nz(ia), rwz(ia)%p)
      call dmpi_broadcast(nz(ia), zetac )
      call timer(2,'MPI_bnd')
      call timer(2,'gs_2')
      call timer_print('TS - gs_2 took: ','gs_2',100)

      !- baroclinic correction from file (stored in zetac2):
      if (usezbcl) call bndbclz (nz(ia),zetac2,jstuna,zetac,krz(ia)%p,nb1,nb2,1)

      !- zeta correction at open boundary:
      call bndzcorr (nz(ia),krz(ia)%p,rwz(ia)%p,zetac)

      !- zeta boundary values to full array:
      call zetar (mm1k_l(ia)%p,zn_l(ia)%p,nz(ia),krz(ia)%p,rwz(ia)%p )
    endif

  end subroutine SolveMassEq_z

  !=============================================================================

  subroutine MassNest_z (ia)
  !- do the remaining mass eqs nesting

    use dmi_mpi,           only : dmpi_gather, dmpi_halo, dd, cctable, cftable
    use cmod_params,       only : mmx, nmx, iga, jga,                          &
                                  nestingto, nestinglevels, enclosing
    use cmod_arrays,       only : mm1k, wrk2, dx, dy
    use cmod_local_arrays, only : mm1k_l, zn_l
    use cmod_timer,        only : timer, timer_print
    use nesting,           only : copy_f_g_sfc

    implicit none

    integer(4), intent(in) :: ia

    integer(4) :: ii, iia, iao


    if (nestinglevels(ia) > 0) then
      call timer(1,'gs_3')
      !- relax z_coarse towards z_fine at border:
      do ii=1,nestinglevels(ia)
        !  iia: fine grid
        !  ia:  coarse grid
        iia = nestingto(ia)%p(ii)
        iao = enclosing(iia,ia)

        !- gather fine grid zn  (wrk2 used as global zn)
        call timer(1,'MPI_nest_5')
        call dmpi_gather (ia,iia,cctable(ia,iia),cftable(ia,iia),              &
                          mmx(iia),nmx(iia),(1),mm1k_l(iia)%p,mm1k(iia)%p,     &
                          mm1k(ia)%p, iga(iia)%p(:,iao),jga(iia)%p(:,iao),     &
                          zn_l(iia)%p,wrk2(iia)%p)
        call timer(2,'MPI_nest_5')

        !- get z_coarse from z_fine in area covered by fine grid:
!$OMP PARALLEL DEFAULT (shared)
        call copy_f_g_sfc(mmx(iia), nmx(iia), mm1k(iia)%p,                     &
                          dd(ia)%low_i, dd(ia)%up_i,                           &
                          dd(ia)%low_j, dd(ia)%up_j, mm1k_l(ia)%p,             &
                          iga(iia)%p(:,iao), jga(iia)%p(:,iao),                &
                          dx(iia)%p, dy(iia), wrk2(iia)%p, zn_l(ia)%p )
!$OMP END PARALLEL

        !- halo swap coarse grid zn'
        call timer(1,'MPI_nest_6')
        call dmpi_halo(1,mm1k_l(ia)%p,ia,zn_l(ia)%p)
        call timer(2,'MPI_nest_6')
      enddo
      call timer(2,'gs_3')
      call timer_print('TS - gs_3 took: ','gs_3',100)
    endif

  end subroutine MassNest_z

  !=============================================================================

  subroutine MassNest_dyn (ia)
  !- do the remaining mass eqs nesting

    use dmi_mpi_global,    only : mpi_io_rank
    use dmi_mpi,           only : dmpi_gather, dmpi_scatter, mpi_io_serial
    use cmod_params,       only : iw2, iw3, mmx, nmx, kmx, iga, jga, kfg, hz,  &
                                  nestingto, nestinglevels, enclosing
    use cmod_arrays,       only : mm1k, wrk2, dx, dy, kh, hn, ti
    use cmod_local_arrays, only : mm1k_l, zn_l, kh_l, hz_l, hn_l
    use cmod_timer,        only : timer, timer_print
    use nesting,           only : copy_f_g_vol

    implicit none

    integer(4), intent(in) :: ia

    integer(4) :: ii, iia, iao


    !  NOTE: wrk2 used as global zn throughout
    if (nestinglevels(ia) > 0) then
      call timer(1,'gs_3')
      !- relax z_coarse towards z_fine at border:
      do ii=1,nestinglevels(ia)
        !  iia: fine grid
        !  ia:  coarse grid
        iia = nestingto(ia)%p(ii)
        iao = enclosing(iia,ia)

        !- gather fine grid zn
        call dmpi_gather(iia,(1),zn_l(iia)%p,wrk2(iia)%p,mpi_io_rank)

        !  better safe than sorry, gather coarse grid zn'
        call dmpi_gather(ia,(1),zn_l(ia)%p,wrk2(ia)%p,mpi_io_rank)

        !- gather fine grid hn and hz
        call dmpi_gather(iia, kmx(iia), hz_l(iia)%p, hz(iia)%p, mpi_io_rank)
        call dmpi_gather(iia, kmx(iia), hn_l(iia)%p, hn(iia)%p, mpi_io_rank)

        !- FIXME: Test if this is really needed!!
        !  better safe than sorry, gather coarse grid h and hz
        call dmpi_gather(ia, kmx(ia), hz_l(ia)%p, hz(ia)%p, mpi_io_rank)
        call dmpi_gather(ia, kmx(ia), hn_l(ia)%p, hn(ia)%p, mpi_io_rank)
 
        if (mpi_io_serial) then
!$OMP PARALLEL DEFAULT (shared)
          call copy_f_g_vol ( mmx(ia),nmx(ia),kh(ia)%p, mm1k(ia)%p,            &
                              mmx(iia), nmx(iia), kmx(iia), mm1k(iia)%p,       &
                              iga(iia)%p(:,iao), jga(iia)%p(:,iao),            &
                              kfg(ia)%p(:,ii), dx(iia)%p, dy(iia), hn(ia)%p,   &
                              hz(ia)%p, hn(iia)%p, hz(iia)%p, ti(ia)%p,        &
                              wrk2(ia)%p )
!$OMP END PARALLEL
        endif

        !- scatter coarse grid hn
        call dmpi_scatter(ia,kmx(ia),iw3(ia),mm1k_l(ia)%p,hn_l(ia)%p,          &
                          mm1k(ia)%p,hn(ia)%p,mpi_io_rank,kh_l(ia)%p)

        !- scatter coarse grid zn'
        call dmpi_scatter(ia,(1),iw2(ia),mm1k_l(ia)%p,zn_l(ia)%p,              &
                          mm1k(ia)%p,wrk2(ia)%p,mpi_io_rank,kh_l(ia)%p)
      enddo
      call timer(2,'gs_3')
      call timer_print('TS - gs_3 took: ','gs_3',100)
    endif

  end subroutine MassNest_dyn

  !=============================================================================
  
  subroutine SolveMassEq_dyn (ia)
  !- solve mass equs in dynamical case on domain #ia
  
    use dmi_mpi_global,    only : mpi_io_rank
    use dmi_mpi,           only : dmpi_gather, mpi_io_serial, dmpi_broadcast
    use cmod_params,       only : mainarea, kmx, nz, krz, nq, iga, jga, kfg,   &
                                  nestingfrom, nestingto, znesting
    use cmod_arrays,       only : khu, khv, krq, idx,  mm1k, wrk, wrk2,        &
                                  rwz, rwq, dt, dx, dy, covo, covu, z,         &
                                  zetac, zetac2, pl, wma1, wma2, h, geopot,    &
                                  wma3, wma4
    use cmod_local_arrays, only : iw2_l, iw3_l, mm1k_l, ind_l, kh_l,           &
                                  khu_l, khv_l, zn_l, z_l, hz_l, hu_l, hv_l,   &
                                  h_l, hn_l, un_l, vn_l, w_l, avv_l, ti_l,     &
                                  rho_l, press_l, geopot_l, pl_l, precip_l,    &
                                  evapo_l
    use masseqs,           only : masseqs_solver_dyn
    use cmod_timer,        only : timer, timer_print
    use constants,         only : coldst, usezbcl, precevap, zbnd_type, zero
    use cmod_miscvars,     only : lunsurge, lmns, jstuna, OPRFMT, nb1, nb2
    use boundary,          only : bndbclz, bndzcorr, gradiz, zrano, zetar,     &
                                  compbno, masseqs_bnd_dyn, zbnd_rbc
    use nesting,           only : w_c_f, masseqs_c_f_dyn
    use cmod_precip,       only : mass_precevap
    
  
    implicit none

    integer(4), intent(in) :: ia
   
    integer(4) :: ii, iia, iao, nz1, nz2      
      
    call timer( 1, 'masseqs' )
!$OMP PARALLEL DEFAULT(shared)
! MPI: Halo already available: hu, hv, un, vn
    !- mass equation:

    call masseqs_solver_dyn ( ia,kmx(ia), mm1k_l(ia)%p, ind_l(ia)%p, iw2_l(ia),&
                              iw3_l(ia), khu_l(ia)%p, khv_l(ia)%p, kh_l(ia)%p, &
                              dx(ia)%p, dy(ia), dt(ia), covo(ia)%p, covu(ia)%p,&
                              hu_l(ia)%p, hv_l(ia)%p, un_l(ia)%p, vn_l(ia)%p,  &
                              z_l(ia)%p, zn_l(ia)%p, ti_l(ia)%p, h_l(ia)%p,    &
                              hz_l(ia)%p, hn_l(ia)%p, w_l(ia)%p, avv_l(ia)%p )

!$OMP END PARALLEL
    call timer( 2, 'masseqs' )
    call timer_print( 'TS - masseqs took: ', 'masseqs', 50 )


    if (nestingfrom(ia)%p(0) > 0) then
      !- boundary conditions coarse grid --> fine grid:
      do iao=1,nestingfrom(ia)%p(0)
        !  ia:  fine grid
        !  iia: coarse grid
        iia = nestingfrom(ia)%p(iao)
        nz1 = znesting(ia)%p(iao,1)
        nz2 = znesting(ia)%p(iao,2)
        ii  = 1
        do while (ia /= nestingto(iia)%p(ii))
          ii = ii + 1
        enddo

        call masseqs_c_f_dyn ( ia,nz1, nz2,kh_l(iia)%p,kmx(iia),mm1k_l(iia)%p, &
                               hu_l(iia)%p, hv_l(iia)%p,                       &
                               un_l(iia)%p, vn_l(iia)%p, kfg(iia)%p(:,ii),     &
                               iga(ia)%p(:,iao), jga(ia)%p(:,iao), kmx(ia),    &
                               mm1k_l(ia)%p, krz(ia)%p, khu_l(ia)%p,           &
                               khv_l(ia)%p, kh_l(ia)%p, dx(ia)%p, dy(ia),      &
                               dt(ia), covo(ia)%p,covu(ia)%p, hu_l(ia)%p,      &
                               hv_l(ia)%p, un_l(ia)%p, vn_l(ia)%p, z_l(ia)%p,  &
                               zn_l(ia)%p, ti_l(ia)%p, h_l(ia)%p, hz_l(ia)%p,  &
                               hn_l(ia)%p, w_l(ia)%p, avv_l(ia)%p )

        !- update w in fine grid along border:
        call w_c_f ( krz(ia)%p, mm1k_l(ia)%p, kmx(iia), mm1k_l(iia)%p,         &
                     kfg(iia)%p(:,ii), iga(ia)%p(:,iao), jga(ia)%p(:,iao),     &
                     z_l(ia)%p, zn_l(ia)%p, w_l(ia)%p, un_l(ia)%p,             &
                     vn_l(ia)%p, un_l(iia)%p, vn_l(iia)%p, covo(ia)%p,         &
                     covu(ia)%p, dx(ia)%p, dy(ia), dt(ia), hu_l(ia)%p,         &
                     hv_l(ia)%p, nz1, nz2 )
      enddo

    elseif  (ia == mainarea) then

      call timer(1,'gs_2')
      call dmpi_gather(ia,kmx(ia),press_l(ia)%p ,wma2        ,mpi_io_rank)
      call dmpi_gather(ia,kmx(ia),geopot_l(ia)%p,geopot(ia)%p,mpi_io_rank)      
      call dmpi_gather(ia,kmx(ia),  rho_l(ia)%p ,wma1        ,mpi_io_rank)
      call dmpi_gather(ia,kmx(ia),   h_l(ia)%p  ,   h(ia)%p  ,mpi_io_rank)
      call dmpi_gather(ia,kmx(ia),   hu_l(ia)%p ,wma3        ,mpi_io_rank)
      call dmpi_gather(ia,kmx(ia),   hv_l(ia)%p ,wma4        ,mpi_io_rank)
      call dmpi_gather(ia,1      ,   z_l(ia)%p  ,    z(ia)%p ,mpi_io_rank)
      wma3(0) = zero
      wma4(0) = zero

      if (mpi_io_serial .and. zbnd_type == 1) then
        !- zeta boundary values, surge and harmonic constants:
        call zrano (lunsurge,nz(ia),krz(ia)%p,rwz(ia)%p,lmns,coldst,OPRFMT)
      endif

      if (zbnd_type == 2) then
        call dmpi_gather(ia,kmx(ia), un_l(ia)%p , wrk(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,kmx(ia), vn_l(ia)%p ,wrk2(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,1      , pl_l(ia)%p ,  pl(ia)%p,mpi_io_rank)
        if (mpi_io_serial) then
          !- zeta boundary values from radiation boundary conditions:
          call zbnd_rbc(mm1k(ia)%p, khu(ia)%p, khv(ia)%p, wrk(ia)%p,wrk2(ia)%p,&
                        wma3, wma4, pl(ia)%p, nz(ia), krz(ia)%p, rwz(ia)%p )
        endif
      endif

      if (mpi_io_serial) then
        !- density boundary values:
        call compbno (mm1k(ia)%p, khu(ia)%p, khv(ia)%p, h(ia)%p, wma1, wma2,   &
                      geopot(ia)%p, nz(ia), krz(ia)%p, zetac)
      endif

      !- communicate the boundary values:
      call dmpi_broadcast( nz(ia), rwz(ia)%p )
      call dmpi_broadcast( nz(ia), zetac )

      call timer(2,'gs_2')
      call timer_print('TS - gs_2 took: ','gs_2',100)


      !- baroclinic correction from file (stored in zetac2):
      if (usezbcl) call bndbclz ( nz(ia), zetac2, jstuna, zetac, krz(ia)%p,    &
                                  nb1, nb2, 1 )

      !- zeta correction at open boundary:
      call bndzcorr ( nz(ia), krz(ia)%p, rwz(ia)%p, zetac )

      call masseqs_bnd_dyn ( ia, nz(ia), kmx(ia), mm1k_l(ia)%p, krz(ia)%p,     &
                             khu_l(ia)%p, khv_l(ia)%p, kh_l(ia)%p, dx(ia)%p,   &
                             dy(ia), dt(ia), covo(ia)%p, covu(ia)%p,hu_l(ia)%p,&
                             hv_l(ia)%p, un_l(ia)%p, vn_l(ia)%p, z_l(ia)%p,    &
                             zn_l(ia)%p, ti_l(ia)%p, h_l(ia)%p, hz_l(ia)%p,    &
                             hn_l(ia)%p, w_l(ia)%p, avv_l(ia)%p )

      !- zeta boundary values to full array:
      call zetar ( mm1k_l(ia)%p, zn_l(ia)%p, nz(ia), krz(ia)%p, rwz(ia)%p,     &
                   kh_l(ia)%p, hn_l(ia)%p, ti_l(ia)%p ) 
    endif


!$OMP PARALLEL DEFAULT(shared)
! MPI: NO need for halo swap
    !- add precipitation-evaporation:
    if (precevap) call mass_precevap(iw2_l(ia), dt(ia), kh_l(ia)%p, idx(ia)%p, &
                                     zn_l(ia)%p, w_l(ia)%p, precip_l(ia)%p,    &
                                     evapo_l(ia)%p, hn_l(ia)%p )
!$OMP END PARALLEL

    !-  Do the nesting of z fine-->coarse:
    call MassNest_dyn (ia)

    !- boundary values:
    call gradiz ( kmx(ia),mm1k_l(ia)%p,dt(ia),dx(ia)%p,dy(ia),w_l(ia)%p,       &
                  zn_l(ia)%p,nz(ia),krz(ia)%p,rwz(ia)%p,nq(ia),krq(ia)%p,      &
                  rwq(ia)%p,(ia == mainarea),hn_l(ia)%p)

  end subroutine SolveMassEq_dyn
  
end module cmod_hydrodynamics
