module ice_smagorinsky
  use constants, only : zero, one, qrt
  !-----------------------------------------------------------------------------
  !  This module wraps some of the routines for doing the Smagorinsky
  !  parameterization and horizontal deformation terms for sea ice
  !-----------------------------------------------------------------------------

  implicit none
  private

  public  :: ice_smag, ice_deform

contains

  !=============================================================================

  subroutine ice_smag (imax,mm,ind,n2d,dx,dy,dt,cs,ah,kh,casus,fastmask,       &
                       shear,stretch,cosphi,eddyh,eddyd,cdiv,idx,div)
    !---------------------------------------------------------------------------
    !  Calculate horizontal eddy viscosity for ice
    !---------------------------------------------------------------------------
  
    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain
  
    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    logical,    intent(in)    :: casus(0:), fastmask(0:)
    integer(4), intent(in)    :: imax, n2d
    integer(4), intent(in)    :: mm(:,0:,0:), kh(0:), ind(:,:)
    integer(4), intent(inout) :: idx(1:,1:)
    real(8),    intent(in)    :: dy, dx, dt, cs, ah, cdiv
    real(8),    intent(in)    :: cosphi(:,0:), shear(0:), stretch(0:), div(0:)
    real(8),    intent(inout) :: eddyh(0:), eddyd(0:)
  
    !- local variables ---------------------------------------------------------
    integer(4) :: n, i, j, n2dl, n2du
    real(8)    :: st, l2s, cs2dxdy, ed, cdl, edmax, mindx, sh

    !- Parameter to control numerical stability (linear, in von Neumann sense):
    real(8), parameter :: MXFAC = 0.10_8, dy0 = 1852.0_8
  
    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)
  
    cs2dxdy = cs*cs*ah*dx*dy*(dy0/dy)**2
    mindx   = dx*minval(cosphi(1,1:imax-1))
    edmax   = (MXFAC/dt)/(one/mindx**2 + one/dy**2)

    ! k=1
    do n=n2dl,n2du
      if (casus(n) .or. fastmask(n)) then
        eddyh(n) = zero
        eddyd(n) = zero
        cycle
      endif

      i = ind(1,n)
      j = ind(2,n)
  
      l2s = cs2dxdy*cosphi(1,i)
      cdl = cdiv*l2s

      ! k=1 unrolled
      sh = qrt*( shear(n)           + shear(mm(1,i-1,j))                       &
               + shear(mm(1,i,j-1)) + shear(mm(1,i-1,j-1)) )
      
      st = stretch(n)
      ed = st*st + sh*sh
      eddyh(n) = min( l2s*sqrt( ed )                , edmax )
      eddyd(n) = min( cdl*sqrt( ed + div(n)*div(n) ), edmax )
    enddo

  end subroutine ice_smag
  
  !=============================================================================

  subroutine ice_deform (mm,ind,n2d,ddx,ddy,kh,casus,fastmask,                 &
                         u,v,shear,stretch,div,cosphi,idx)

    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain
  
    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    logical,    intent(in)    :: casus(0:), fastmask(0:)
    integer(4), intent(in)    :: n2d
    integer(4), intent(in)    :: mm(:,0:,0:), kh(0:), ind(:,:)
    integer(4), intent(inout) :: idx(1:,1:)
    real(8),    intent(in)    :: u(0:), v(0:), ddy, ddx, cosphi(:,0:)
    real(8),    intent(inout) :: shear(0:), stretch(0:), div(0:)
  
    !- local variables ---------------------------------------------------------
    integer(4) :: n, i, j, mi, mw, mn, me, ms, mse
    integer(4) :: n2dl, n2du
    real(8)    :: ddxi1, ddxi2, uterm, vterm
  
    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)
  
    !-   stretch = du/dx - dv/dy   ---------------------------------------------
    !-   div     = du/dx + dv/dy   ---------------------------------------------
    !-   shear   = du/dy + dv/dx   ---------------------------------------------

    ! k=1:
    do n=n2dl,n2du
      if (casus(n) .or. fastmask(n)) then
        stretch(n) = zero
        div(n)     = zero
        shear(n)   = zero
        cycle
      endif

      i = ind(1,n)
      j = ind(2,n)
  
      ddxi1 = ddx/cosphi(1,i)
      ddxi2 = ddx/cosphi(2,i)
      
      mi  = n
      mw  = mm(1,i,  j-1)
      mn  = mm(1,i-1,j  )
      me  = mm(1,i,  j+1)
      ms  = mm(1,i+1,j  )
      mse = mm(1,i+1,j+1)
  
      uterm = (u(mi)-u(mw))*ddxi1
      vterm = (v(mn)-v(mi))*ddy
  
      stretch(mi) = uterm - vterm
      div    (mi) = uterm + vterm
  
      if (me>0 .and. ms>0 .and. mse>0) then
        shear(mi) = (v(me)-v(mi))*ddxi2 + (u(mi)-u(ms))*ddy
      else
        shear(mi) = zero
      endif
    enddo

  end subroutine ice_deform
  
  !=============================================================================

end module ice_smagorinsky
