module ice_solver

  implicit none
  private
  public  :: solver
  private :: solvey, order, ndr2, nsf2, nnf2

  !- work arrays:
  integer(4), allocatable, save, private :: iwil(:), iwiu(:), iwjl(:), iwju(:)

  !- Magic Numbers:
  integer(4), parameter, private :: jluscale = 15
  integer(4), parameter, private :: nluscale =  4
  integer(4), parameter, private :: nscale   =  3

contains

!===============================================================================

  subroutine solver(np,s,a,ia,ja,l)
     use exits,     only : exitme
     use constants, only : ldebug
     implicit none
  
     integer(4), intent(in)    :: l, np
     integer(4), intent(in)    :: ia(:), ja(:)         ! (np+1), (9*np)
     real(8),    intent(in)    :: a(:)                 ! (9*np)
     real(8),    intent(inout) :: s(:)                 ! (np)
  
     integer(4) :: iflag
     logical    :: ncall

     ncall = (l == 1)

     call solvey(np,s,a,ia,ja,ncall,ldebug,iflag)

     if (iflag > 0) call exitme(66,'FATAL ERROR: cmod has gone unstable')
  
  end subroutine solver

!===============================================================================

  subroutine solvey(np,vf,a,ia,ja,ncall,lprint,iflag)
  
  !      the subroutine solvey solves by l-u-decomposition the linear
  !      system  aa*y = f  with square matrix aa, using the yale sparse
  !      matrix programs.
  !      fill-in is reduced by calling the yale subroutine 'order'.
  !
  !      the call of solvey :
  !
  !      call solvey(np,vf,a,ia,ja,ncall,lprint,iflag)
  !
  !      the parameters :
  !
  !      np      number of unknowns
  !      vf      real(8)-vector of right hand sides f and solution y if
  !              entering to or returning from solvey, respectively.
  !              dimension of vf is np.
  !      a       real(8)-vector containing the nonzero elements of the
  !              matrix aa.
  !              dimension of a :  number of nonzero elements
  !      ia      integer(4)-vector to the first nonzero elements of the rows
  !              of aa. The vector is of dimension np+1, with ia(np+1) 
  !              pointing to 
  !                 (position of the last nonzero element of aa in a) + 1
  !      ja      integer(4)-vector giving the j-indices of the compactly 
  !              stored nonzero elements of aa
  !      ncall   switching parameter :
  !               true  : first iteration, need to set up arrays.
  !               false : second or higher iteration, need not do a setup
  !      lprint  switch for debug printing
  !      iflag   return code :  iflag=0 : no errors,
  !                             iflag>0 : error return
  !
  !              directly called module private subroutines: ndr2, order
  !                         (yale university sparse matrix programs)
  !
  !
  ! Some notes on storage:
  !
  !           ( .1  .0  .2  .3  .0 )
  !           ( .0  .4  .0  .0  .0 )
  !      aa = ( .5  .0  .6  .7  .0 )
  !           ( .8  .0  .9 .10 .11 )
  !           ( .0  .0  .0 .12 .13 )
  !
  !  could be stored as:
  !
  !           i =  1   2   3   4   5   6   7   8   9  10  11  12  13
  !       ------+---------------------------------------------------
  !       ia(i) |  1   4   5   8  12  14
  !       ja(i) |  1   3   4   2   1   3   4   1   3   4   5   4   5
  !        a(i) | .1  .2  .3  .4  .5  .6  .7  .8  .9 .10 .11 .12 .13
  !
  !
     use dmi_mpi_global, only : iu06
     use io_subs,        only : flush_unit

     implicit none
  
     logical,    intent(in)    :: lprint, ncall
     integer(4), intent(in)    :: np
     integer(4), intent(in)    :: ia(:), ja(:)    ! (np+1), 9*np
     integer(4), intent(out)   :: iflag
     real(8),    intent(in)    :: a(:)            ! 
     real(8),    intent(inout) :: vf(:)           ! (np)
  
     integer(4)     :: i,ii,if1,if2,np2,nlo,nup,jlumax,nlumax,nn
     character(256) :: cfmt

     !- vars for work array allocation -----------------------------------------
     integer(4),              save :: npsave = 0, jlusave = 0
     integer(4), allocatable, save :: iwrk1(:), iwrk2(:)

     !- find some basic numbers ------------------------------------------------
     nlo = 0
     nup = 0
     do i=1,np
       do ii=ia(i),ia(i+1)-1
         if (ja(ii) < i) nup = nup + 1
         if (ja(ii) > i) nlo = nlo + 1
       enddo
     enddo
     nlumax = max(nlo,nup)
     jlumax = jluscale*nlumax
     if (lprint) then
       write(iu06,*) 'np:          ', np
       write(iu06,*) 'lower, upper:', nlo, nup
       write(iu06,*) 'non-zeroes:  ', ia(np+1)-1    !    <= np+nlo+nup
       write(iu06,*) 'jlumax:      ', jlumax
       call flush_unit(iu06)
     endif
     
     !- possibly allocate or re-allocate work arrays ---------------------------
     if (np > npsave .or. jlumax > jlusave) then
       if (npsave > 0) deallocate( iwrk1, iwrk2, iwil, iwiu, iwjl, iwju )
       ! FIXME: until it has been fixed properly, use this empirical number:
       allocate( iwrk1(1:np),  iwrk2(1:np),                                    &
                 iwil(1:np+1), iwiu(1:np+1),                                   &
                 iwjl(jlumax), iwju(jlumax) )
       npsave  = np
       jlusave = jlumax
     endif

     !- let's do it: -----------------------------------------------------------
     iflag = 0
  
     !- fill-in reducing by ORDER() --------------------------------------------
     if (ncall) then
       do i = 1,np
         iwrk1(i) = i
       enddo

       iflag = -1
       np2   = 0
       nn    = nscale*np + nluscale*nlumax
       do while (iflag /= 0)
         if (lprint .and. np2 > 0) then
           write(iu06,*) 'trying =', nn
           call flush_unit(iu06)
         endif
         call order(np,ia,ja,iwrk1,iwrk2,nn,iflag)
         if (np2 >= 3) exit
         if (iflag /= 0) then
           np2 = np2 + 1
           nn  = nn + np2*nlumax
         endif
       enddo
       if (lprint .and. iflag == 0) write(iu06,*) 'nn =', nn, ' should do it'
       if (iflag /= 0) write(iu06,*) 'nn, np2 =', nn, np2, ' is not sufficient'
     endif
  
     !- could not find a reasonable size of work arrays ------------------------
     if (iflag /= 0) then
       write(iu06,*) 'solvey: error return from subroutine order'
       if1 = (iflag-1)/np
       if2 = iflag - np*if1
       if (if1 == 9) then
         write(iu06,*) ' order: insufficient storage'
       else
         write(iu06,*) ' order: null row in a'
       endif
       if (iflag < 100000000) then
         cfmt ="(/' ERROR return FROM SOLVEY, IFLAG=',i8,' =',i2,'*np +',i5/)"
         write(iu06,trim(cfmt)) iflag , if1 , if2
       else
         write(iu06,"(/' ERROR return FROM SOLVEY, IFLAG=')")
         write(iu06,*) iflag
         write(iu06,"(' =',i4,' * np +',i5/)") if1 , if2
       endif
     endif
  
     if (iflag == 0) then
       !- solve the system -----------------------------------------------------
       call ndr2(np, iwrk1, iwrk2, ia, ja, a, vf, ncall, iflag, lprint, jlumax)
     endif
  
     if (iflag /= 0) then
       !- traps ----------------------------------------------------------------
       if1 = (iflag-1)/np
       if2 = iflag - np*if1
  
       if (iflag < 100000000) then
         cfmt ="(/' ERROR return FROM SOLVEY, IFLAG=',i8,' =',i2,' * np +',i5/)"
         write(iu06,trim(cfmt)) iflag , if1 , if2
       else
         write(iu06,"(/' ERROR return FROM SOLVEY, IFLAG=')")
         write(iu06,*) iflag
         write(iu06,"(' =',i4,' * np +',i5/)") if1 , if2
       endif
  
       if (if1 == 3 .or. if1 == 6) then
         write(iu06,*) ' error in subroutine nsf2: insufficient storage'
       elseif (if1 == 7) then
         write(iu06,*) ' error in subroutine nnf2: insufficient storage'
       elseif (if1 == 1) then
         write(iu06,*) ' error in subroutine nsf2: null row in a'
       elseif (if1 == 2) then
         write(iu06,*) ' error in subroutine nsf2: duplicate entry in a'
       elseif (if1 == 5) then
         write(iu06,*) ' error in subroutine nsf2: null pivot'
       elseif (if1 == 8) then
         write(iu06,*) ' error in subroutine nnf2: zero pivot'
       endif
     endif

  end subroutine solvey

!===============================================================================

  subroutine order(n,ia,ja,p,ip,maxwrk,flag)
     implicit none
  
     ! Dummy arguments
     integer(4), intent(in)    :: maxwrk, n
     integer(4), intent(in)    :: ia(:), ja(:)
     integer(4), intent(out)   :: flag
     integer(4), intent(inout) :: p(:)
     integer(4), intent(out)   :: ip(:)

     ! So far we stick to this (passed bounds/stack tests) 
     integer(4) :: lv(maxwrk), vv(maxwrk)

  
     ! Local variables
     integer(4) :: dmin, dthr, i, ipvk, j, jmax, jmin, k, kmax, li,            &
                   lj, lk, lli, llj, llk, ni, s, sfs, tmp, vi, vj, vk, vl
     logical    :: notfound, again, bigloop
     intrinsic  :: max0 , min0
   
     ! dummy value to start doing the order stuff
     flag = -1

     orderstuff: do while (flag < 0)
       ! prepare for no errors and termination of doing the stuff
       flag = 0
  
       ! * initialize free storage
       vk = 1
       do s = n,maxwrk
         lv(s) = s + 1
       enddo
       lv(maxwrk) = 0
       sfs = 1
       ! initialize ordering, degree, and adjacency
       do k = 1,n
         vk     = p(k)
         ip(vk) = k
         vv(k)  = n + 1
         lv(k)  = k
       enddo
       sfs = sfs + n
    
       ! initialize nonzero structure for every vertex vk
       ! for every vertex vj adjacent to vk
       do vk = 1,n
         jmin = ia(vk)
         jmax = ia(vk+1) - 1
!        if (jmin > jmax) then
!          ! error: null row in a
!          flag = n + vk
!          exit orderstuff
!        endif
         do j = jmin,jmax
           vj = ja(j)
           if (vj /= vk) then
             ! search for vj in adjacency of vk
             llk = vk
             again = .true.
             do while (again)
               lk  = llk
               llk = lv(lk)
               if (vv(llk) >= vj) again = .false.
             enddo
             if (vv(llk) /= vj) then
               ! insert vj in adjacency of vk
               vv(vk)  = vv(vk) + 1
!              if (sfs == 0) then
!                ! error: insufficient storage
!                flag = 9*n + vk
!                exit orderstuff
!              endif
               llk     = sfs
               sfs     = lv(sfs)
               vv(llk) = vj
               lv(llk) = lv(lk)
               lv(lk)  = llk
               ! search for vk in adjacency of vj
               llj = vj
               again = .true.
               do while (again)
                 lj  = llj
                 llj = lv(lj)
                 if (vv(llj) >= vk) again = .false.
               enddo
               if (vv(llj) /= vk) then
                 ! insert vk in adjacency of vj
                 vv(vj)  = vv(vj) + 1
!                if (sfs == 0) then
!                  ! error: insufficient storage
!                  flag = 9*n + vk
!                  exit orderstuff
!                endif
                 llj     = sfs
                 sfs     = lv(sfs)
                 vv(llj) = vk
                 lv(llj) = lv(lj)
                 lv(lj)  = llj
               endif
             endif
           endif
         enddo
       enddo
  
       ! minimum degree algorithm with threshold search
       ! initialize vertex count and thresholds for search
       i = 0
       jmin = 1
       dthr = 0
       dmin = n + n
       ! while uneliminated vertices exist
       ! search for vertex vi of minimum degree
       bigloop = .true.
       do while (bigloop)
         notfound = .true.
         do while (notfound)
           jmin = max0(jmin,i+1)
           jloop: do j = jmin,n   ! exit value of j used below
             vi = p(j)
             if (vv(vi) <= dthr) then
               notfound = .false.
               exit jloop
             endif
             dmin = min0(dmin,vv(vi))
           enddo jloop
           if (notfound) then
             jmin = 1
             dthr = dmin
             dmin = n + n
           endif
         enddo
    
         ! number vertex vi of minimum degree
         ! In the following, j equals exit value from jloop above
         jmin   = j
         i      = i + 1
         vj     = p(i)
         p(j)   = vj
         ip(vj) = j
         p(i)   = vi
         ip(vi) = i
         ni     = 1
         ! delete eliminated vertices from adjacency of vi
         ! for every vertex vk adjacent to vi
         lli  = vi
         kmax = (vv(vi)-ni) - n
         do k = 1,kmax
           again = .true.
           do while (again)
             li  = lli
             lli = lv(li)
             vk  = vv(lli)
             ! if vk eliminated, then delete from adjacency of vi
             if (ip(vk) <= i) then
               lv(li)  = lv(lli)
               lv(lli) = sfs
               sfs     = lli
               lli     = li
               again   = .true.
             else
               again   = .false.
             endif
           enddo
         enddo
         ! eliminate vertex vi
         ! for every vertex vk adjacent to vi
         lli  = vi
         kmax = (vv(vi)-ni) - n
         do k = 1,kmax
           li  = lli
           lli = lv(li)
           vk  = vv(lli)
           ! merge adjacency of vi into adjacency of vk
           ! for every vertex vj adjacent to vi
           llk  = vk
           lj   = vi
           jmax = (vv(vi)-ni) - n
           do j = 1,jmax
             lj = lv(lj)
             vj = vv(lj)
             if (vj /= vk) then
               notfound = .true.
               ! search for vj in adjacency of vk
               do while (notfound)
                 notfound = .false. 
  
                 lk  = llk
                 llk = lv(lk)
                 vl  = vv(llk)
                 if (vj > vl) then
                   ! ... while deleting eliminated vertices
                   if (ip(vl) <= i) then
                     lv(lk)  = lv(llk)
                     lv(llk) = sfs
                     sfs     = llk
                     llk     = lk
                   endif
                   notfound = .true.
                 ! insert vj in adjacency of vk
                 elseif (vj /= vl) then
                   vv(vk)  = vv(vk) + 1
!                  if (sfs == 0) then
!                    ! error: insufficient storage
!                    flag = 9*n + vk
!                    exit orderstuff
!                  endif
                   llk     = sfs
                   sfs     = lv(sfs)
                   vv(llk) = vj
                   lv(llk) = lv(lk)
                   lv(lk)  = llk
                 endif
  
               enddo
             endif
           enddo
  
           ! if vk of minimal degree, then number vertex vk, ...
           if (vv(vk) <= vv(vi)) then
             i       = i + 1
             j       = ip(vk)
             vj      = p(i)
             p(j)    = vj
             ip(vj)  = j
             p(i)    = vk
             ip(vk)  = i
             ni      = ni + 1
             ! ... recover storage for adjacency of vk,
             tmp     = lv(vk)
             lv(vk)  = sfs
             sfs     = tmp
             ! ... and delete vk from adjacency of vi
             lv(li)  = lv(lli)
             lv(lli) = sfs
             sfs     = lli
             lli     = li
           endif
         enddo
      
         ! update degrees of uneliminated vertices adjacent to vi
         ! for every vertex vk in adjacency of vi
         li   = vi
         kmax = (vv(vi)-ni) - n
         do k = 1,kmax
           li     = lv(li)
           vk     = vv(li)
           ! update degree of vk and thresholds for cyclic search
           vv(vk) = vv(vk) - ni
           if (vv(vk) < dmin) then
             if (vv(vk) > dthr) then
               dmin = vv(vk)
             else
               dmin = dthr
               dthr = vv(vk)
               ipvk = ip(vk)
               jmin = min0(jmin,ipvk)
             endif
           endif
         enddo
    
         ! recover storage for adjacency of vi
         tmp    = lv(vi)
         lv(vi) = sfs
         sfs    = tmp
  
         bigloop = (i < n)
       enddo  ! bigloop
  
     enddo orderstuff

  end subroutine order

!===============================================================================

  ! driver for subroutines for solving sparse nonsymmetric systems of
  ! linear equations (uncompressed pointer storage)
  subroutine ndr2(n,c,ic,ia,ja,a,z,ncall,flag,lprint,jlumax)

     use dmi_mpi_global, only : iu06
     use io_subs,        only : flush_unit

     implicit none
  
     logical,    intent(in)    :: lprint, ncall
     integer(4), intent(in)    :: n, jlumax
     integer(4), intent(in)    :: ia(:), ja(:), c(:), ic(:)
     integer(4), intent(out)   :: flag
     real(8),    intent(in)    :: a(:)
     real(8),    intent(inout) :: z(:)
  
     integer(4) :: jlmax, jumax, np1
  
     np1    = n + 1
     flag   = 0
  
     ! call nsf if flag is set
     if (ncall) then
       if (lprint) then
         write(iu06,*) "calling nsf2:"
         write(iu06,*) "  First guess jlmax=jumax :", jlumax
         call flush_unit(iu06)
       endif
       call nsf2(n,c,ic,ia,ja,iwil,iwjl,iwiu,iwju,flag)
     endif
  
     if (flag == 0) then
       jlmax = iwil(np1) - 1
       jumax = iwiu(np1) - 1
       if (lprint .and. ncall) then
         write(iu06,*) "  Actual jlmax, jumax     :", jlmax, jumax
         call flush_unit(iu06)
       endif
  
       if (jlmax > jlumax .or. jumax >  jlumax) then
         ! error: insufficient storage
         flag = 7*n + 1
       else
         if (lprint) then
           write(iu06,*) "calling nnf2:"
           call flush_unit(iu06)
         endif
         call nnf2(n,c,ic,ia,ja,a,z,iwil,iwjl,iwiu,iwju,flag,max(jlmax,jumax))
       endif
     endif
  
  end subroutine ndr2

!===============================================================================

  ! symbolic ldu-factorization of a nonsymmetric sparse matrix
  ! (uncompressed pointer storage)
  !
  subroutine nsf2(n,r,ic,ia,ja,il,jl,iu,ju,flag)
     implicit none
  
     integer(4), intent(in)    :: n
     integer(4), intent(in)    :: ia(:), ja(:), ic(:), r(:)
     integer(4), intent(out)   :: flag, il(:), jl(:), iu(:), ju(:)
  
     integer(4) :: i , j , jlptr , jmax , jmin , juptr , k , m , qm , vj
     logical    :: search, notfound

     ! Local stack arrays:
     integer(4) :: q(n+1), im(n+1)
  
     ! initialize pointers
     jlptr = 0
     il(1) = 1
     juptr = 0
     iu(1) = 1
  
     ! assume no errors initially:
     flag = 0 
  
     ! for each row of l and u
     kloop: do k = 1,n
  
       ! set q to the reordered row of a
       q(n+1) = n + 1
       jmin = ia(r(k))
       jmax = ia(r(k)+1) - 1
!      if (jmin > jmax) then
!        ! error: null row in a
!        flag = n + r(k)
!        exit kloop
!      endif
       do j = jmin,jmax
         vj = ic(ja(j))
         qm = n + 1
         notfound = .true.
         do while (notfound)
           m  = qm
           qm = q(m)
           notfound = (qm < vj)
         enddo
!        if (qm == vj) then
!          ! error: duplicate entry in a
!          flag = 2*n + r(k)
!          exit kloop
!        endif
         q(m)  = vj
         q(vj) = qm
       enddo
  
       ! for each entry in the lower triangle
       i = n + 1
       search = .true.
       do while (search)
         i = q(i)
         if (i >= k) then
  
           ! check for null pivot
!          if (i /= k) then
!            ! error: null pivot
!            flag = 5*n + k
!            exit kloop
!          endif
           ! remaining elements of q define structure of u(k, )
           notfound = .true.
           do while (notfound)
             i = q(i)
             if (i > n) then
               ! get ready for next row
               im(k)   = juptr
               il(k+1) = jlptr + 1
               iu(k+1) = juptr + 1
               notfound = .false.
             else
               juptr = juptr + 1
!              if (juptr > jlumax) then
!                ! error: insufficient storage for ju
!                flag = 6*n + k
!                exit kloop
!              endif
               ju(juptr) = i
             endif
           enddo
           search = .false.
         else
           ! l(k,i) will be nonzero, so add it to jl
           jlptr = jlptr + 1
!          if (jlptr > jlumax) then
!            ! error: insufficient storage for jl
!            flag = 3*n + k
!            exit kloop
!          endif
           jl(jlptr) = i
           qm = i
           ! inspect i-th row for fillin, adjust im if possible
           jmin = iu(i)
           jmax = im(i)
           do j = jmin,jmax
             vj = ju(j)
             if (vj == k) im(i) = j
             notfound = .true.
             do while (notfound)
               m  = qm
               qm = q(m)
               notfound = (qm < vj)
             enddo
             if (qm /= vj) then
               q(m)  = vj
               q(vj) = qm
               qm    = vj
             endif
           enddo
         endif
       enddo ! search
  
     enddo kloop
  
  
  end subroutine nsf2

!===============================================================================

  ! numeric ldu-factorization of sparse nonsymmetric matrix and
  ! solution of system of linear equations (uncompressed pointer storage)
  
  subroutine nnf2(n,c,ic,ia,ja,a,z,il,jl,iu,ju,flag,jlumax)
     use constants, only : zero, one
     implicit none
  
     integer(4), intent(in)    :: n, jlumax
     integer(4), intent(out)   :: flag
     integer(4), intent(in)    :: ia(:), ja(:), c(:), ic(:)
     integer(4), intent(in)    :: il(:), iu(:), jl(:), ju(:)
     real(8),    intent(in)    :: a(:)
     real(8),    intent(inout) :: z(:)

     ! Local stack arrays:  (experiments with +1 for safety)
     real(8) :: row(n+1), tmp(n+1)
     real(8) :: u(jlumax+1)
  
     ! Local variables
     real(8)    :: dk, li, ssum, term
     integer(4) :: i, j, k

     real(8), parameter :: tiny_tiny = 1.0e-30_8

     ! assume no errors initially:
     flag = 0
  
     ! for each row
     kloop: do k = 1,n
       ! set the initial structure of row
       !
       ! if l(k,m) /= 0, row(m) = 0
       do j = il(k),il(k+1)-1
         row(jl(j)) = zero
       enddo
       row(k) = zero
       ! if u(k,m) /= 0, row(m) = 0
       do j = iu(k),iu(k+1)-1
         row(ju(j)) = zero
       enddo
       ! set row to k-th row of reordered a
       do j = ia(c(k)),ia(c(k)+1)-1
         row(ic(ja(j))) = a(j)
       enddo
       ! initialize sum
       ssum = z(c(k))
  
       ! assign the k-th row of l and adjust row, sum
       do i = il(k),il(k+1)-1
         li   = -row(jl(i))
         ssum = ssum + li*tmp(jl(i))
         do j = iu(jl(i)),iu(jl(i)+1)-1
           term = li*u(j)
           if (abs(term) > tiny_tiny) row(ju(j)) = row(ju(j)) + term
         enddo
       enddo
  
       ! assign diagonal d and k-th row of u, set tmp(k)
!      if (row(k) == zero) then
!        ! error: zero pivot
!        flag = 8*n + k   
!        exit kloop
!      else
         dk     = one/row(k)
         tmp(k) = ssum*dk
         do j = iu(k),iu(k+1)-1
           u(j) = row(ju(j))*dk
         enddo
!      endif
     enddo kloop
  
     ! solve u x = tmp by back substitution
     if (flag == 0) then
       k = n
       do i = 1,n
         ssum = tmp(k)
         do j = iu(k),iu(k+1)-1
           ssum = ssum - u(j)*tmp(ju(j))
         enddo
         tmp(k)  = ssum
         z(c(k)) = ssum
         k = k - 1
       enddo
     endif
  
  end subroutine nnf2

!===============================================================================
  
end module ice_solver
