module icedyn_drift

  use constants,         only : zero, one, hgr, fastice_test!, two, CDIV, CSMAG
  use cmod_local_arrays, only : uice_o, vice_o, fastmask, shear, stretch, div, &
                                eddyh, eddyd

  !- directives:
  implicit none
  private

  !- public arrays:
  ! none

  !- public stuff:
  public :: icedyn, icedyn_ft, ice_vel_drift, calc_fastmask

  !- private stuff:
  private :: iota_decay

  ! private parameters:
! real(8), parameter :: icdiv = CDIV, ics = CSMAG, iah = 1.0_8
  real(8), parameter :: icdiv = 1.0_8, ics = 50.0_8, iah = 1.0_8
  real(8), parameter :: p0 = 5.0_8 ! 10.0_8

  ! temporary test of naive fastice model:
  !real(8), parameter :: fih0 = 10.0_8, fic0 = 0.8_8, fit0 = 0.1_8
  real(8), parameter :: fih0 = 5.0_8, fic0 = 0.8_8, fit0 = 0.3_8

  ! iota model: ice deformation rate: coefficient
  real(8), parameter :: idc = 1.e-8_8

contains
  
  !-----------------------------------------------------------------------------

  subroutine icedyn(m,n2d,ind,khh,idx,imax,dx,dy,cosphi,ft,dt,uwat,vwat,h,zeta,&
                    wx,wy,casus,ice,uice,vice,ia,mmk,nud,nvd,uvdam,z0srf,      &
                    iotamodel, iota)
                    ! pl)

    use constants,         only : g, rhow, rhoi, cdrag, qrt, half, rhoa, itmax
    use ice_smagorinsky,   only : ice_smag, ice_deform
    use icedyn_nest,       only : ice_mom_c_f
    use cmod_arrays,       only : wrk, wrk2, mm1k, kh
    use cmod_local_arrays, only : kh_l, mm1k_l, ueis_l, veis_l
    use cmod_params,       only : mainarea, nestingfrom, nestingto, znesting,  &
                                  iga, jga, krz
    use cmod_timer,        only : timer
    use dmi_omp,           only : domp_get_domain
    use dmi_mpi,           only : dmpi_halo, dmpi_gather, nftable, mctable,    &
                                  mpi_rank

    implicit none
 
    integer(4), intent(in) :: ia, n2d, m(0:,0:), ind(1:,1:), mmk(1:,0:,0:)
    integer(4), intent(in) :: nud, nvd, khh(0:), imax
    real(8),    intent(in) :: cosphi(1:,0:), ft(0:)
    real(8),    intent(in) :: wx(0:), wy(0:), dx, dy, dt !, c
    real(8),    intent(in) :: uwat(0:), vwat(0:), zeta(0:), h(0:)
    real(8),    intent(in) :: ice(1:,0:), z0srf(0:)
    real(8),    intent(in) :: uvdam(1:,0:) ! 1.0/0.0 mask for dams
    !real(8),    intent(in) :: pl(0:)
    logical,    intent(in) :: casus(0:), iotamodel

    integer(4), intent(inout) :: idx(1:,1:)
    real(8),    intent(inout) :: uice(0:), vice(0:), iota(1:,0:)

    logical    :: cyclenext
    integer(4) :: n, i, j, nl, nu, ms, me, mn, mw, mne, mse, msw, it
    integer(4) :: nz1, nz2, ii, iia, iao
    !convective: integer(4) :: i1, i2, i3, i4, j1, j2, j3, j4
    real(8)    :: duv, uicen, vicen, ddx, ddy, bot, top, stab, hhh, dwu, dwv
    real(8)    :: cdrhow, dtcdrwre, dtcdrwra, dtdri, dtddx, dtddy, tx, tau, p
    real(8)    :: hu, hv, cv, cu, ehmi, ehww, ehnn, uw, vw, wax, way, duvw
    real(8)    :: utau, vtau, uedd, vedd, ucor, vcor, usrf, vsrf, dau, dav, dd
    real(8)    :: cn, ce, cs, tn, te, ts, damp, gnablaz
    !conv/lapl: real(8)    :: ucon, vcon, urd, vrd, aurd, avrd, ulap, vlap, dtd

    real(8), parameter :: ddlim  = 0.00001_8
    real(8), parameter :: duv_m2 = (0.05_8)**2   ! [m^2/s^2]
    real(8), parameter :: cdrhoa = 0.0029_8*rhoa
    !lapl: real(8), parameter :: laplacescale = 100000.0_8

    ! empirical damping of the accelleration: 
    real(8), parameter :: cuvh = 0.20_8, cuvl = 0.05_8
    real(8), parameter :: huvh = 0.03_8, huvl = 0.001_8
    real(8), parameter :: damp0 = one/(cuvh-cuvl)/(huvh-huvl)

    !dtd      = dt/laplacescale
    ddx      = one/dx
    ddy      = one/dy
    dtddx    = dt/dx
    dtddy    = dt/dy
    cdrhow   = cdrag*rhow
    dtdri    = dt/rhoi
    dtcdrwre = dtdri*cdrhow
    dtcdrwra = dtdri*cdrhoa

!$OMP PARALLEL DEFAULT(shared)                                                 &
!$OMP private(n,nl,nu,i,j,me,ms,mn,mw,mne,mse,msw,it,cyclenext,ehmi,ehnn,ehww, &
!$OMP         uicen,vicen,uw,vw,hu,hv,cu,cv,tx,bot,top,hhh,stab,wax,way,duvw,  &
!$OMP         duv,utau,usrf,ucor,uedd,vtau,vsrf,vcor,vedd,tau,dau,dav,dwu,dwv, &
!$OMP         dd,p,cn,ce,cs,tn,te,ts)
    call domp_get_domain(khh, 1, n2d, nl, nu, idx)

    call calc_fastmask(iotamodel, n2d, khh, idx, iota, h, ice, mmk, ind)
    call ice_deform (mmk,ind,n2d,ddx,ddy,khh,casus,fastmask,                   &
                     uice,vice,shear,stretch,div,cosphi,idx)

    call timer(1,'MPI_halo')
!$OMP BARRIER
!   barrier/halo-swap here due to use of shear,stretch,div and fastmask below
    if (iotamodel.or.fastice_test) call dmpi_halo(mmk,ia,fastmask)
    call dmpi_halo(1,mmk,khh,ia,1,1,1,6,shear,stretch,div)
!$OMP BARRIER
    call timer(2,'MPI_halo')

    do n=nl,nu
      uice_o(n) = uice(n)
      vice_o(n) = vice(n)
    enddo
    call ice_smag (imax,mmk,ind,n2d,dx,dy,dt,ics,iah,khh,casus,fastmask,       &
                   shear,stretch,cosphi,eddyh,eddyd,icdiv,idx,div)

    call timer(1,'MPI_halo')
!$OMP BARRIER
!   barrier/halo-swap here due to use of u/vice_o iand eddyh/d below
    call dmpi_halo(1,mmk,khh,ia,1,1,1,6,uice_o,vice_o,eddyh,eddyd)
!$OMP BARRIER
    call timer(2,'MPI_halo')

! conv/lapl:  i1,j1,i2,j2,i3,j3,i4,j4,ucon,ulap,vcon,vlap,urd,vrd,aurd,avrd
    ! calculate uice -----------------------------------------------------------
    do n=nl,nu
      i=ind(1,n)
      j=ind(2,n)

      uice(n) = zero

      ! find out if there's anything at all to do:
      me = m(i,j+1)
      cyclenext = .true.
      if (me > 0) then
        if ((.not.(casus(n).and.casus(me))) .and.                              &
            (.not.fastmask(n)) .and. (.not.fastmask(me)) .and.                 &
            (h(n) > hgr) .and. (h(me) > hgr)) cyclenext = .false.
      endif
      ! handle u-dams
      if (nud > 0) cyclenext = (uvdam(1,n) <= zero) .or. cyclenext

      if (cyclenext) cycle

      ! lets continue ...
      mn  = m(i-1,j  )
      mne = m(i-1,j+1)
      ms  = m(i+1,j  )
      mse = m(i+1,j+1)
      tn  = ice(1,n)
      te  = ice(1,me)
      cn  = ice(2,n)
      ce  = ice(2,me)

      ! extract quantities centred at u point:
      uicen = uice_o(n)
      vicen = qrt*(vice_o(n) + vice_o(me) + vice_o(mn) + vice_o(mne))
      uw    = uwat(n)
      vw    = qrt*(vwat(n) + vwat(me) + vwat(mn) + vwat(mne))
      wax   = half*(wx(n)+wx(me))
      way   = half*(wy(n)+wy(me))
      tx    = dtddx/cosphi(1,i)

      ! current and wind drag terms:
      dwu  = uw -uicen
      dau  = wax-uicen
      duv  = sqrt(duv_m2 + (vw -vicen)**2 + dwu**2)
      duvw = sqrt(duv_m2 + (way-vicen)**2 + dau**2)
      bot  = dtcdrwre*duv
      top  = dtcdrwra*duvw
      tau  = bot*dwu + top*dau
      hu   = half*(tn+te+z0srf(n)+z0srf(me))

      ! extract upwind-centred compactness:
      if (tau >= zero) then
        cu = cn
      else
        cu = ce
      endif
      
      ! experimental damping:
      if (cu > cuvl .and. hu > huvl) then
        damp = (min(cu,cuvh) - cuvl)*(min(hu,huvh) - huvl)*damp0
        hhh  = damp*cu/hu
      else
        damp = zero
        hhh  = zero
      endif

      ! down-stream centred strength:
      if (uicen >= zero) then
        p = dt*p0*te*ce**2
      else
        p = dt*p0*tn*cn**2
      endif

      ! current and wind drag:
      stab = one + p                                                           &
                 + (dtcdrwre*(dwu**2/duv+duv) + dtcdrwra*(dau**2/duvw+duvw))*hhh
      utau = hhh*tau

      ! nabla(z) term:
      gnablaz = g*(zeta(me)-zeta(n))*tx
      usrf    = gnablaz*damp

      ! air pressure term:
      !  fixme: why not ?
      ! usrf = usrf + (pl(me)-pl(n))*tx/rhoi

      ! Coriolis:
      ucor = ft(i)*vicen
      
      ! convective terms using vector upwind scheme:
      !  fixme: this might be slightly overkill :)
!     urd  = uicen*tx
!     vrd  = vicen*dtddy
!     aurd = abs(urd)
!     avrd = abs(vrd)
!     if (urd >= zero) then
!       j2 = j-1
!     else
!       j2 = j+1
!     endif
!     if (vrd >= zero) then
!       i4 = i+1
!     else
!       i4 = i-1
!     endif
!     if (aurd >= avrd) then
!       i1 = i
!       j3 = j2
!     else
!       i1 = i4
!       j3 = j
!     endif
!     ucon = aurd*(uice_o(m(i1,j))-uice_o(m(i1,j2)))                           &
!          + avrd*(uice_o(m(i,j3))-uice_o(m(i4,j3)))

      ! Laplacian background diffusion:
!     ulap = uice_o(me) + uice_o(m(i,j-1)) - two*uicen
!     if (mne > 0 .and. mn > 0) ulap = ulap + uice_o(mn) - uicen
!     if (mse > 0 .and. ms > 0) ulap = ulap + uice_o(ms) - uicen
!     ulap = dtd*ulap

      ! Eddy terms:
      ehmi = eddyh(n) + eddyh(me) + eddyh(ms) + eddyh(mse)
      ehnn = eddyh(n) + eddyh(me) + eddyh(mn) + eddyh(mne)
      uedd =   (eddyh(me)*stretch(me)-eddyh(n)*stretch(n))*tx                  &
             + (ehnn     *shear  (mn)-ehmi     *shear (n))*dtddy               &
             + (eddyd(me)*div    (me)-eddyd(n)*div    (n))*tx 

      ! first guess:
      dd = (utau - usrf + ucor + uedd - p*uicen)/stab
      uicen = uicen + dd

      ! further iterations:
      it = 2
      do while (it <= itmax .and. abs(dd) > ddlim)
        dwu  = uw -uicen
        dau  = wax-uicen
        duv  = sqrt(duv_m2 + (vw -vicen)**2 + dwu**2)
        duvw = sqrt(duv_m2 + (way-vicen)**2 + dau**2)
        bot  = dtcdrwre*duv
        top  = dtcdrwra*duvw
        tau  = bot*dwu + top*dau
        if (tau >= zero) then
          cu = cn
        else
          cu = ce
        endif

        ! experimental damping:
        if (cu > cuvl .and. hu > huvl) then
          damp = (min(cu,cuvh) - cuvl)*(min(hu,huvh) - huvl)*damp0
          hhh  = damp*cu/hu
        else
          damp = zero
          hhh  = zero
        endif
        usrf = gnablaz*damp

        if (uicen >= zero) then
          p = dt*p0*te*ce**2
        else
          p = dt*p0*tn*cn**2
        endif

        utau = hhh*tau
        stab = one + p                                                         &
                   + (dtcdrwre*(dwu**2/duv+duv)+dtcdrwra*(dau**2/duvw+duvw))*hhh

        it = it + 1
        dd = (uice_o(n) - uicen + utau - usrf + ucor + uedd - p*uicen)/stab
        uicen = uicen + dd
      enddo

      ! finally, new u vel:
      uice(n) = uicen
    enddo

    ! calculate vice -----------------------------------------------------------
    do n=nl,nu
      i=ind(1,n)
      j=ind(2,n)

      vice(n) = zero

      ! find out if there's anything at all to do:
      ms = m(i+1,j)
      cyclenext = .true.
      if (ms > 0) then
        if ((.not.(casus(n).and.casus(ms))) .and.                              &
            (.not.fastmask(n)) .and. (.not.fastmask(ms)) .and.                 &
            (h(n) > hgr) .and. (h(ms) > hgr)) cyclenext = .false.
      endif
      ! handle v-dams
      if (nvd > 0) cyclenext = (uvdam(2,n) <= zero) .or. cyclenext

      if (cyclenext) cycle

      ! lets continue ...
      me  = m(i  ,j+1)
      mse = m(i+1,j+1)
      mw  = m(i,  j-1)
      msw = m(i+1,j-1)
      tn = ice(1,n)
      ts = ice(1,ms)
      cn = ice(2,n)
      cs = ice(2,ms)

      ! extract quantities centred at v point:
      vicen = vice_o(n)
      uicen = qrt*(uice_o(n) + uice_o(mw) + uice_o(ms) + uice_o(msw))
      vw    = vwat(n)
      uw    = qrt*(uwat(n) + uwat(mw) + uwat(ms) + uwat(msw))
      wax   = half*(wx(n)+wx(ms))
      way   = half*(wy(n)+wy(ms))
      tx    = dtddx/cosphi(2,i)

      ! current and wind drag terms:
      dwv  = vw- vicen
      dav  = way-vicen
      duv  = sqrt(duv_m2 + (uw- uicen)**2 + dwv**2)
      duvw = sqrt(duv_m2 + (wax-uicen)**2 + dav**2)
      bot  = dtcdrwre*duv
      top  = dtcdrwra*duvw
      tau  = bot*dwv + top*dav
      hv   = half*(tn+ts+z0srf(n)+z0srf(ms))

      ! extract upwind-centred compactness:
      if (tau >= zero) then
        cv = cs
      else
        cv = cn
      endif

      ! experimental damping:
      if (cv > cuvl .and. hv > huvl) then
        damp = (min(cv,cuvh) - cuvl)*(min(hv,huvh) - huvl)*damp0
        hhh  = damp*cv/hv
      else
        damp = zero
        hhh  = zero
      endif

      ! down-stream centred strength:
      if (vicen >= zero) then
        p = dt*p0*tn*cn**2
      else
        p = dt*p0*ts*cs**2
      endif

      ! current and wind drag:
      stab = one + p                                                           &
                 + (dtcdrwre*(dwv**2/duv+duv) + dtcdrwra*(dav**2/duvw+duvw))*hhh
      vtau = hhh*tau

      ! nabla(z) term:
      gnablaz = g*(zeta(n)-zeta(ms))*dtddy
      vsrf    = gnablaz*damp

      ! air pressure term:
      !  fixme: why not ?
      ! vsrf = vsrf + (pl(n)-pl(ms))*dtddy/rhoi

      ! Coriolis:
      vcor = -ft(i)*uicen

      ! convective terms using vector upwind scheme:
      !  fixme: this might be slightly overkill :)
!     urd  = uicen*tx
!     vrd  = vicen*dtddy
!     aurd = abs(urd)
!     avrd = abs(vrd)
!     if (vrd >= zero) then
!       i2 = i+1
!     else
!       i2 = i-1
!     endif
!     if (urd >= zero) then
!       j4 = j-1
!     else
!       j4 = j+1
!     endif
!     if (avrd >= aurd) then
!       i3 = i2
!       j1 = j
!     else
!       i3 = i
!       j1 = j4
!     endif
!     vcon = avrd*(vice_o(m(i,j1))-vice_o(m(i2,j1)))                           &
!          + aurd*(vice_o(m(i3,j))-vice_o(m(i3,j4)))

      ! Laplacian smoothing term using full-slip condition for lateral terms:
!     vlap = vice_o(m(i-1,j)) + vice_o(ms) - two*vicen
!     if (me > 0 .and. mse > 0) vlap = vlap + vice_o(me) - vicen
!     if (mw > 0 .and. msw > 0) vlap = vlap + vice_o(mw) - vicen
!     vlap = dtd*vlap

      ! Horizontal eddy viscosity:
      ehmi = eddyh(n) + eddyh(me) + eddyh(ms) + eddyh(mse)
      ehww = eddyh(n) + eddyh(mw) + eddyh(ms) + eddyh(msw)
      vedd =   (ehmi     *shear (n)-ehww     *shear  (mw))*tx                  &
             - (eddyh(n)*stretch(n)-eddyh(ms)*stretch(ms))*dtddy               &
             + (eddyd(n)*div    (n)-eddyd(ms)*div    (ms))*dtddy

      ! first guess:
      dd = (vtau - vsrf + vcor + vedd - p*vicen)/stab
      vicen = vicen + dd

      ! further iterations:
      it = 2
      do while (it <= itmax .and. abs(dd) > ddlim)
        dwv  = vw- vicen
        dav  = way-vicen
        duv  = sqrt(duv_m2 + (uw- uicen)**2 + dwv**2)
        duvw = sqrt(duv_m2 + (wax-uicen)**2 + dav**2)
        bot  = dtcdrwre*duv
        top  = dtcdrwra*duvw
        tau  = bot*dwv + top*dav
        if (tau >= zero) then
          cv = cs
        else
          cv = cn
        endif
        hhh  = cv/hv

        ! experimental damping:
        if (cv > cuvl .and. hv > huvl) then
          damp = (min(cv,cuvh) - cuvl)*(min(hv,huvh) - huvl)*damp0
          hhh  = damp*cv/hv
        else
          damp = zero
          hhh  = zero
        endif
        vsrf = gnablaz*damp

        if (vicen >= zero) then
          p = dt*p0*tn*cn**2
        else
          p = dt*p0*ts*cs**2
        endif

        vtau = hhh*tau
        stab = one + p                                                         &
                   + (dtcdrwre*(dwv**2/duv+duv)+dtcdrwra*(dav**2/duvw+duvw))*hhh

        it = it + 1
        dd = (vice_o(n) - vicen + vtau - vsrf + vcor + vedd - p*vicen)/stab
        vicen = vicen + dd
      enddo

      ! finally, new v vel:
      vice(n) = vicen
    enddo

    ! for drift ice, update with deformation term (in case of a iota model):
    if (iotamodel) call iota_decay (nl,nu,dt,idc,casus,h,div,shear,stretch,iota)
!$OMP END PARALLEL

    if (ia == mainarea) then
      ! do the VAB/VOB stuff
!FIXME
    else
      ! do c_f nesting
      do iao=1,nestingfrom(ia)%p(0)
        !  ia:  fine grid
        !  iia: coarse grid
        iia = nestingfrom(ia)%p(iao)
        ii  = 1
        do while (ia /= nestingto(iia)%p(ii))
          ii = ii + 1
        enddo
        !  first, gather coarse grid u/vice 
        nz1 = znesting(ia)%p(iao,1)
        nz2 = znesting(ia)%p(iao,2)

        call timer(1,'MPI_nest_f')
        call dmpi_gather(ia, iia, nz1, nz2, krz(ia)%p,                         &
                         nftable(mpi_rank+1,iia,ia), mctable(iia,ia),          &
                         kh_l(iia)%p, kh(iia)%p, (1),                          &
                         mm1k_l(iia)%p, mm1k(iia)%p, mm1k(ia)%p,               &
                         iga(ia)%p(:,iao), jga(ia)%p(:,iao),                   &
                         ueis_l(iia)%p, wrk(iia)%p, veis_l(iia)%p, wrk2(iia)%p)
        call timer(2,'MPI_nest_f')

        call ice_mom_c_f (mmk,dt,ft,uice,vice,h,zeta,casus,ice,dx,dy,          &
                          cosphi,z0srf,uwat,vwat,wx,wy,uice_o,vice_o,fastmask, &
                          iga(ia)%p(:,iao),jga(ia)%p(:,iao),krz(ia)%p,         &
                          mm1k(iia)%p,wrk(iia)%p,wrk2(iia)%p,nz1,nz2)
      enddo
    endif
  end subroutine icedyn

  !-----------------------------------------------------------------------------

  subroutine iota_decay (nl,nu,dt,c,casus,h,div,shear,stretch,iota)
    !  Implementation notes:
    !
    !        iota_new = iota_old * exp( -c*abs(epsilon dot)*dt )
    !        L_new    =    L_old * exp( -c*abs(epsilon dot)*dt )
    !
    !        epsilon dot = sqrt( eps_I**2 + eps_II**2 )
    !        eps_I **2   = div**2 = (d(uice)/dx)**2 + (d(vice)/dy)**2
    !
    !        eps_II **2  = stretch**2 + shear**2 = eps_IIc **2  + eps_IIn **2
    !
    !        which (wrt to our Arakawa C-grid) has a central term:
    !
    !        stretch**2  = (d(uice)/dx - d(vice)/dy)**2
    !
    !        and a nodal term:
    !
    !        shear**2    = (d(uice)/dy + d(vice)/dx)**2
    !
    !   FIXME: As a starting point, we are sloppy and ignore the 
    !          spherical curvature term, i.e:
    !          eps_I = 1/(R cos(lat)) du/d(long) + 1/R dv/d(lat) -1/R tan(lat) v
    !          tan(lat) = sin(lat) 1/cos(lat) = -d(cos(lat))/d(lat) 1/cos(lat)
    !          We should be aware that we might need to do this properly at a
    !          later stage. 
    !
    !        Furthermore, we are lazy enough to merge 
    !          div**2 + stretch**2 = 2.0*( (d(uice)/dx)**2 + (d(vice)/dy)**2 )

    implicit none
 
    logical,    intent(in)    :: casus(0:)
    integer(4), intent(in)    :: nl, nu
    real(8),    intent(in)    :: h(0:), c, dt
    real(8),    intent(in)    :: div(0:), shear(0:), stretch(0:)
    real(8),    intent(inout) :: iota(1:,0:)

    integer(4) :: n
    real(8)    :: e2, e, cdt, arg 
    real(8), parameter :: MaxArg = 20.0_8

    cdt = c*dt

    do n=nl,nu
      if (fastmask(n)) cycle

      e  = one

      if ( (.not.casus(n)) .and. (.not.fastmask(n)) .and. (h(n) > hgr) ) then
        e2 = div(n)**2 + stretch(n)**2 + shear(n)**2

        arg = cdt*sqrt(e2)
        if (arg < MaxArg) then
          e = exp(-arg)
        else
          e = zero
        endif
      endif

      iota(1:2,n) = iota(1:2,n)*e
    enddo

  end subroutine iota_decay

  !-----------------------------------------------------------------------------

  subroutine icedyn_ft(n2d, n2dhalo, kh, idx)
    use dmi_omp, only : domp_get_domain
    implicit none
    integer(4), intent(in)    :: n2d, n2dhalo, kh(0:)
    integer(4), intent(inout) :: idx(1:,1:)

    integer(4) :: nl, nu, n

    ! NUMA-initialize all arrays:

!$OMP MASTER
    uice_o(0)   = zero
    vice_o(0)   = zero
    fastmask(0) = .true.   ! "fast ice" on land
    shear(0)    = zero
    stretch(0)  = zero
    div(0)      = zero
    eddyh(0)    = zero
    eddyd(0)    = zero
    do n=n2d+1,n2d+n2dhalo
      uice_o(n)   = zero
      vice_o(n)   = zero
      fastmask(n) = .false.
      shear(n)    = zero
      stretch(n)  = zero
      div(n)      = zero
      eddyh(n)    = zero
      eddyd(n)    = zero
    enddo
!$OMP END MASTER

    call domp_get_domain(kh, 1, n2d, nl, nu, idx)

    do n=nl,nu
      uice_o(n)   = zero
      vice_o(n)   = zero
      fastmask(n) = .false.
      shear(n)    = zero
      stretch(n)  = zero
      div(n)      = zero
      eddyh(n)    = zero
      eddyd(n)    = zero
    enddo

  end subroutine icedyn_ft

  !-----------------------------------------------------------------------------

  subroutine calc_fastmask(iotamodel, n2d, kh, idx, iota, h, ice, mmk, ind)
    use dmi_omp, only : domp_get_domain
    implicit none
    logical,    intent(in)    :: iotamodel
    integer(4), intent(in)    :: n2d, kh(0:), mmk(1:,0:,0:), ind(1:,1:)
    real(8),    intent(in)    :: iota(1:,0:), h(0:), ice(1:,0:)
    integer(4), intent(inout) :: idx(1:,1:)
    
    integer(4) :: n,nl,nu,k
    real(8)    :: d
 
!$OMP MASTER
    fastmask(0) = .true.
!$OMP END MASTER
    call domp_get_domain(kh, 1, n2d, nl, nu, idx)

    if (fastice_test) then
      do n=nl,nu
        d = h(n)
        do k=2,kh(n)
          d = d + h(mmk(k,ind(1,n),ind(2,n)))
          if (d >= fih0) exit
        enddo
        fastmask(n) = (ice(1,n) >= fit0 .and. ice(2,n) >= fic0 .and. d < fih0)
      enddo
    elseif (iotamodel) then
      do n=nl,nu
        fastmask(n) = (iota(1,n) > one)
      enddo
    endif
  end subroutine calc_fastmask

  !-----------------------------------------------------------------------------

  subroutine ice_vel_drift(n2d, mmk, ind, kh, idx, casus, uice, vice)
    use dmi_omp, only : domp_get_domain
    implicit none

    integer(4), intent(in)    :: n2d, mmk(1:,0:,0:), ind(1:,1:),kh(0:)
    logical,    intent(in)    :: casus(0:)
    integer(4), intent(inout) :: idx(1:,1:)
    real(8),    intent(inout) :: uice(0:), vice(0:)

    integer(4) :: nn, nl, nu, i, j

    call domp_get_domain(kh, 1, n2d, nl, nu, idx)

    do nn=nl,nu
      i = ind(1,nn)
      j = ind(2,nn)
      if (casus(nn).and.casus(mmk(1,i,j+1))) uice(nn) = zero
      if (casus(nn).and.casus(mmk(1,i+1,j))) vice(nn) = zero
    enddo

  end subroutine ice_vel_drift

end module icedyn_drift
