module icedyn_nest
  use constants, only : zero, one
  !-----------------------------------------------------------------------------
  !  a collection of methods to do the nesting stuff for ice vel
  !-----------------------------------------------------------------------------

  implicit none
  private
  
  public  :: ice_mom_f_c, ice_mom_c_f

contains 

! ------------------------------------------------------------------------------

  subroutine ice_mom_f_c (meg, ugn, vgn, krug, krvg, mef, igaf, jgaf,          &
                          ufn, vfn, nu1, nu2, nv1, nv2 )
  !-----------------------------------------------------------------------------
  ! Use fine grid velocity components to 
  ! obtain coarse grid velocity components at border.
  !-----------------------------------------------------------------------------
  
    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain
  
    !- directives --------------------------------------------------------------
    implicit none
  
  
    !- arguments ---------------------------------------------------------------
    !  indices:
    integer(4), intent(in) :: nu1, nu2, nv1, nv2, krug(:,0:), krvg(:,0:)
    integer(4), intent(in) :: meg(0:,0:), mef(0:,0:), igaf(:), jgaf(:)
    real(8), intent(inout) :: ugn(0:), vgn(0:)
    real(8), intent(in)    :: ufn(0:), vfn(0:)
  
    !- local vars --------------------------------------------------------------
    integer(4) :: ira, i, j, ii, jj, iii, jjj, iff, jf, i1, j1, iioff, jjoff
    integer(4) :: mi, mss, mee, mww, is, je, jw, inn, mnn, nl, nll, nlu
    integer(4) :: iprev, jprev, inext, jnext, nlowu, nlowv, nul, nuu, nvl, nvu
    logical    :: calc_nw
    real(8)    :: vf, uf, hs
  

    !- Determine the decomposition of the borders ------------------------------
    nlowu = max(1,nu1)
    nlowv = max(1,nv1)
    if (nu2 >= nlowu .and. nv2 >= nlowv) then
      ! points on both u- and v-brds
      nl = (nu2+1-nlowu) + (nv2+1-nlowv)
      call domp_get_domain(1, nl, nll, nlu)
      nul = max(nlowu, nll-1+nlowu)
      nuu = min(nu2,   nlu-1+nlowu)
      nvl = max(nlowv, nll-nu2-1+nlowu-1+nlowv)
      nvu = min(nv2  , nlu-nu2-1+nlowu-1+nlowv)
    elseif (nu2 < nlowu .and. nv2 < nlowv) then
      ! nothing to do here ... let's get out
      return
    elseif (nu2 < nlowu) then
      ! only v-brds
      call domp_get_domain(nlowv, nv2, nvl, nvu)
      nul = 1
      nuu = -1
    elseif (nv2 < nlowv) then
      ! only u-brds
      call domp_get_domain(nlowu, nu2, nul, nuu)
      nvl = 1
      nvu = -1
    endif

    !- Do W and E borders ------------------------------------------------------
    uloop: do ira=nul,nuu
      if (krug(3,ira) == 0) cycle uloop
      i = krug(1,ira)
      j = krug(2,ira)

      !- Check if v at north border is calculated by previous or will be -------
      !- calculated by next border point                                 -------
      calc_nw = .true.
      if (ira < nu2) then
        iprev = krug(1,ira-1)
        jprev = krug(2,ira-1)
        inext = krug(1,ira+1)
        jnext = krug(2,ira+1)
        if (iprev == i-1 .and. jprev == j) calc_nw = .false.
        if (inext == i-1 .and. jnext == j) calc_nw = .false.
      endif  

      if (krug(3,ira) == 3) j=j+1
      if (meg(i,j) == 0) cycle uloop
  
      ii = igaf(2)*(i-igaf(1))+1
      jj = jgaf(2)*(j-jgaf(1))+1
  
      !- coordinates for vel.comp. 
      if (krug(3,ira) == 3) then
        !- east brd as seen from coarse grid:
        j1    = j
        jjoff = jgaf(2)-1
      elseif (krug(3,ira) == 1) then
        !- west brd as seen from coarse grid:
        j1    = j-1
        jjoff = -1
      endif
      inn = max(i-1,1)
      is  = i

      !- U:
      mi = meg(i,j1)
      if (mi > 0 .and. meg(i,j1+1) > 0) then
        hs  = zero
        uf  = zero
        jf  = jj + jjoff
        do iii=0,igaf(2)-1
          iff = ii + iii
          if (mef(iff,jf) > 0 .and. mef(iff,jf+1) > 0) then
            hs  = hs + one
            uf  = uf + ufn(mef(iff,jf))
          endif
        enddo
        if (hs > zero) then
          ugn(mi) = uf/hs
        else
          ugn(mi) = zero
        endif
      else
        ugn(mi) = zero
      endif
  
      !- Treatment of S and N faces is only needed, if the actual point
      !- is not a corner!
      if (krug(4,ira) == 0) then 
        !- Vsouth:
        mss = meg(is,j)
        if (mss > 0 .and. meg(is+1,j) > 0) then
          hs  = zero
          vf  = zero
          iff  = ii + igaf(2)-1
          do jjj=0,jgaf(2)-1
            jf = jj + jjj
            hs = hs + one
            vf = vf + vfn(mef(iff,jf))
          enddo
          if (hs > zero) then
            vgn(mss) = vf/hs
          else
            vgn(mss) = zero
          endif
        else
          vgn(mss) = zero
        endif
 
        if (calc_nw) then
          !- Vnorth:
          mnn = meg(inn,j)
          if (mnn > 0 .and. meg(inn+1,j) > 0) then
            hs  = zero
            vf  = zero
            iff  = max(ii-1,1)
            do jjj=0,jgaf(2)-1
              jf = jj + jjj
              hs = hs + one
              vf = vf + vfn(mef(iff,jf))
            enddo
            if (hs > zero) then
              vgn(mnn) = vf/hs
            else
              vgn(mnn) = zero
            endif
          else
            vgn(mnn) = zero
          endif
        endif

      endif
  
    enddo uloop


    !- Do N and S borders ------------------------------------------------------
    vloop: do ira=nvl,nvu
      if (krvg(3,ira) == 0) cycle vloop
      i = krvg(1,ira)
      j = krvg(2,ira)

      !- Check if u at west border is calculated by previous or will be --------
      !- calculated by next border point                                --------
      calc_nw = .true.
      if (ira < nv2) then
        iprev = krvg(1,ira-1)
        jprev = krvg(2,ira-1)
        inext = krvg(1,ira+1)
        jnext = krvg(2,ira+1)
        if (iprev == i .and. jprev == j-1) calc_nw = .false.
        if (inext == i .and. jnext == j-1) calc_nw = .false.
      endif

      if (krvg(3,ira) == 4) i=i+1
      if (meg(i,j) == 0) cycle vloop
  
      ii = igaf(2)*(i-igaf(1))+1
      jj = jgaf(2)*(j-jgaf(1))+1
  
      !- coordinates for vel.comp. 
      if (krvg(3,ira) == 2) then
        !- north brd as seen from coarse grid:
        i1    = i-1
        iioff = -1
      elseif (krvg(3,ira) == 4) then
        !- south brd as seen from coarse grid:
        i1    = i
        iioff = igaf(2)-1
      endif
      jw = max(j-1,1)
      je = j

      !- V:
      mi = meg(i1,j)
      if (mi > 0 .and. meg(i1+1,j) > 0) then
        hs  = zero
        vf  = zero
        iff  = ii + iioff
        do jjj=0,jgaf(2)-1
          jf = jj + jjj
          if (mef(iff,jf) > 0 .and. mef(iff+1,jf) > 0) then
            hs = hs + one
            vf = vf + vfn(mef(iff,jf))
          endif
        enddo
        if (hs > zero) then
          vgn(mi) = vf/hs
        else
          vgn(mi) = zero
        endif
      else
        vgn(mi) = zero
      endif
  
      !- Treatment of E and W faces is only needed, if the actual point
      !- is not a corner!
      if (krvg(4,ira) == 0) then 
        !- Ueast:
        mee = meg(i,je)
        if (mee > 0 .and. meg(i,je+1) > 0) then
          hs  = zero
          uf  = zero
          jf  = jj + jgaf(2)-1
          do iii=0,igaf(2)-1
            iff = ii + iii
            hs  = hs + one
            uf  = uf + ufn(mef(iff,jf))
          enddo
          if (hs >zero) then
            ugn(mee) = uf/hs
          else
            ugn(mee) = zero
          endif
        else
          ugn(mee) = zero
        endif
  
        if (calc_nw) then
          !- Uwest:
          mww = meg(i,jw)
          if (mww > 0 .and. meg(i,jw+1) > 0) then
            hs  = zero
            uf  = zero
            jf  = max(jj-1,1)
            do iii=0,igaf(2)-1
              iff = ii + iii
              hs  = hs + one
              uf  = uf + ufn(mef(iff,jf))
            enddo
            if (hs >zero) then
              ugn(mww) = uf/hs
            else
              ugn(mww) = zero
            endif
          else
            ugn(mww) = zero
          endif
        endif

      endif
  
    enddo vloop
  
  end subroutine ice_mom_f_c

! ------------------------------------------------------------------------------

  subroutine ice_mom_c_f (mmk,dt,ft,uice,vice,h,zeta,casus,ice,dx,dy,          &
                          cosphi,z0srf,uwat,vwat,wx,wy,uice_o,vice_o,fastmask, &
                          igaf,jgaf,krzf,mmkc,ucn,vcn,nz1,nz2)

    !- modules -----------------------------------------------------------------
    use constants, only : g, rhow, rhoi, cdrag, rhoa, two, half, qrt, itmax, hgr

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: mmk(1:,0:,0:)
    real(8),    intent(in) :: cosphi(1:,0:), ft(0:)
    real(8),    intent(in) :: wx(0:), wy(0:), dx, dy, dt
    real(8),    intent(in) :: uwat(0:), vwat(0:), zeta(0:), h(0:)
    real(8),    intent(in) :: ice(1:,0:), z0srf(0:), uice_o(0:), vice_o(0:)
    logical,    intent(in) :: casus(0:), fastmask(0:)

    integer(4), intent(in) :: mmkc(:,0:,0:)
    integer(4), intent(in) :: igaf(:),jgaf(:)
    integer(4), intent(in) :: krzf(:,0:)
    integer(4), intent(in) :: nz1, nz2
    real(8),    intent(in) :: ucn(0:),vcn(0:)

    real(8),    intent(inout) :: uice(0:), vice(0:)


    !- locals ------------------------------------------------------------------
    integer(4) :: ibf, bordertype, ncorner, it
    integer(4) :: mi, me, ms, mn, mne, mw, msw
    integer(4) :: i, j, ig, jg, igs, jge
    real(8)    :: vel1_c_f, vel3_c_f, vel4_c_f
    real(8)    :: dtddx, dtddy, cdrhow, dtdri, dtcdrwre, dtcdrwra
    real(8)    :: tn, te, ts, cn, ce, cs, uicen, vicen, uw, vw, wax, way, tx
    real(8)    :: hu, hv, cu, cv, hhh, p, stab, dd, bot, top
    real(8)    :: dwu, dwv, dau, dav,  duv, duvw, tau
    real(8)    :: utau, ucor, usrf, vtau, vcor, vsrf
    logical    :: cyclenext

    real(8), parameter :: ddlim  = 0.00001_8
    real(8), parameter :: duv_m2 = (0.05_8)**2   ! [m^2/s^2]
    real(8), parameter :: cdrhoa = 0.0029_8*rhoa
    real(8), parameter :: p0 = 5.0_8 ! 10.0_8


    !- initialise on every entry -----------------------------------------------
    dtddx    = dt/dx
    dtddy    = dt/dy
    cdrhow   = cdrag*rhow
    dtdri    = dt/rhoi
    dtcdrwre = dtdri*cdrhow
    dtcdrwra = dtdri*cdrhoa


    ! border point loop --------------------------------------------------------
    brdloop: do ibf=nz1,nz2

      ! border point index (i,j) in fine grid 
      bordertype = krzf(3,ibf)
      if (bordertype == 0) cycle brdloop
      i  = krzf(1,ibf)
      j  = krzf(2,ibf)
      mi = mmk(1,i,j)
      if (mi <= 0) cycle brdloop

      ! only do corner points once
      if (ibf > 1) then
        ! skip a corner point that has already been handled
        if (krzf(3,ibf-1) > 0  .and.                                           &
            i == krzf(1,ibf-1) .and. j == krzf(2,ibf-1)) cycle brdloop
      endif
      ncorner = 0
      if (ibf < nz2) then
        ! first-time identification of a convex corner point
        if (krzf(3,ibf+1) > 0  .and.                                           &
            i == krzf(1,ibf+1) .and. j == krzf(2,ibf+1)) then
          ! identify this corner by a code:
          !   unknown corner type:  < 0
          !   not a corner:           0
          !   N/W corner:             1
          !   N/E corner:             2
          !   S/E corner:             3
          !   S/W corner:             4
          ncorner = bordertype
        endif
      endif
  
      ! border point index (ig,jg) in coarse grid 
      ig = igaf(1) + (i-1)/igaf(2)
      jg = jgaf(1) + (j-1)/jgaf(2)
  
      ! assign coarse grid normal velocities at srf-level
      ! to the west of west border, north of north border, etc.
      ! three grid points (left, central, right)
      
      if (bordertype == 1) then
        ! West border 
        !   extract U(west) from coarse grid.
        jg  = jg - 1
        igs = igaf(1) + (i  )/igaf(2)
        vel1_c_f = ucn(mmkc(1,ig, jg))
        vel3_c_f = ucn(mmkc(1,igs,jg))
        if (ncorner == 1) then
          ! N/W corner point
          !   extract V(north) from coarse grid.
          vel4_c_f = vcn(mmkc(1,ig-1,jg+1))
        endif
  
      elseif (bordertype == 2) then
        ! North border
        !   extract V(north) from coarse grid.
        ig  = ig - 1
        jge = jgaf(1) + (j  )/jgaf(2)
        vel1_c_f = vcn(mmkc(1,ig,jg))
        vel3_c_f = vcn(mmkc(1,ig,jge))
        if (ncorner == 2) then
          ! N/E corner point
          !   finish un(mmk(1,i,j)) here by extraction from coarse,
          !   extract U(east) from coarse grid.
          uice(mi) = ucn(mmkc(1,ig+1,jg))
          vel4_c_f = ucn(mmkc(1,ig+1,jg))
        endif
  
      elseif (bordertype == 3) then
        ! East border 
        !   finish un(mmk(1,i,j)) here by extraction from coarse,
        !   extract U(east) from coarse grid.
        igs = igaf(1) + (i  )/igaf(2)
        uice(mi) = ucn(mmkc(1,ig, jg))
        vel1_c_f = ucn(mmkc(1,ig, jg))
        vel3_c_f = ucn(mmkc(1,igs,jg))
        if (ncorner == 3) then
          ! S/E corner point
          !   finish vn(mmk(1,i,j)) here by extraction from coarse,
          !   extract V(south) from coarse grid.
          vice(mi) = vcn(mmkc(1,ig,jg))
        endif
  
      elseif (bordertype == 4) then
        ! South border
        !   finish vn(mmk(1,i,j)) here by extraction from coarse,
        !   extract V(south) from coarse grid.
        jge = jgaf(1) + (j  )/jgaf(2)
        vice(mi) = vcn(mmkc(1,ig,jg))
        vel1_c_f = vcn(mmkc(1,ig,jg))
        vel3_c_f = vcn(mmkc(1,ig,jge))
        if (ncorner == 4) then
          ! S/W corner 
          !   extract U(west) from coarse grid.
          vel4_c_f = ucn(mmkc(1,ig,  jg-1))
        endif
  
      endif
  
      ! for east border, shift e/w-index one point westwards
      ! - shifted back again after u-equation
      if (bordertype == 3) j = j - 1
  
      ! --- u-equation  --------------------------------------------------------

      mi = mmk(1,i,j)
      me = mmk(1,i,j+1)
  
      !- skip trivial cases as well as N/E, S/E corners
      if (mi > 0 .and. me > 0 .and. ncorner /= 2 .and. ncorner /= 3) then

        uice(mi) = zero

        cyclenext = .true.
        if ((.not.(casus(mi).and.casus(me))) .and.                             &
            (.not.fastmask(mi)) .and. (.not.fastmask(me)) .and.                &
            (h(mi) > hgr) .and. (h(me) > hgr)) cyclenext = .false.
        if (.not.cyclenext) then

          ! lets continue ...
          mn  = mmk(1,i-1,j  )
          mne = mmk(1,i-1,j+1)
          ms  = mmk(1,i+1,j  )
          tn  = ice(1,mi)
          te  = ice(1,me)
          cn  = ice(2,mi)
          ce  = ice(2,me)

          ! extract quantities centred at u point:
          uicen = uice_o(mi)
          uw    = uwat(mi)
          wax   = half*(wx(mi)+wx(me))
          way   = half*(wy(mi)+wy(me))
          tx    = dtddx/cosphi(1,i)
          ! fixme: vw terms outside grid are strictly not correct here:
          !        we simply take 2*vwat(mi)
          if (bordertype == 1) then
            ! -- west border
            if (ncorner == 1) then
              ! -- N/W corner
              vicen = two*vel4_c_f
              vw    = two*vwat(mi)
            else
              vicen = vice_o(mn) + vice_o(mne)
              vw    = vwat(mn) + vwat(mne)
            endif
            vicen = qrt*(vice_o(mi) + vice_o(me) + vicen)
            vw    = qrt*(vwat(mi) + vwat(me) + vw)
          elseif (bordertype == 2) then
            ! -- north border
            vicen = qrt*(vice_o(mi) + vice_o(me) + vel1_c_f + vel3_c_f)
            vw    = qrt*(vwat(mi) + vwat(me) + two*vwat(mi))
          elseif (bordertype == 3 .or. bordertype == 4) then
            ! -- east or south border
            vicen = qrt*(vice_o(mi) + vice_o(me) + vice_o(mn) + vice_o(mne))
            vw    = qrt*(vwat(mi) + vwat(me) + vwat(mn) + vwat(mne))
          endif

          ! current and wind drag terms:
          dwu  = uw -uicen
          dau  = wax-uicen
          duv  = sqrt(duv_m2 + (vw -vicen)**2 + dwu**2)
          duvw = sqrt(duv_m2 + (way-vicen)**2 + dau**2)
          bot  = dtcdrwre*duv
          top  = dtcdrwra*duvw
          tau  = bot*dwu + top*dau
          hu   = half*(tn+te+z0srf(mi)+z0srf(me))

          ! extract upwind-centred compactness:
          if (tau >= zero) then
            cu = cn
          else
            cu = ce
          endif
          hhh = cu/hu

          ! down-stream centred strength:
          if (uicen >= zero) then
            p = dt*p0*te*ce**2
          else
            p = dt*p0*tn*cn**2
          endif

          ! current and wind drag:
          stab = one + p                                                       &
               + (dtcdrwre*(dwu**2/duv+duv) + dtcdrwra*(dau**2/duvw+duvw))*hhh
          utau = hhh*tau

          ! nabla(z) term:
          usrf = g*(zeta(me)-zeta(mi))*tx

          ! Coriolis:
          ucor = ft(i)*vicen

          ! first guess:
          dd = (utau - usrf + ucor - p*uicen)/stab
          uicen = uicen + dd

          ! further iterations:
          it = 2
          do while (it <= itmax .and. abs(dd) > ddlim)
            dwu  = uw -uicen
            dau  = wax-uicen
            duv  = sqrt(duv_m2 + (vw -vicen)**2 + dwu**2)
            duvw = sqrt(duv_m2 + (way-vicen)**2 + dau**2)
            bot  = dtcdrwre*duv
            top  = dtcdrwra*duvw
            tau  = bot*dwu + top*dau
            if (tau >= zero) then
              cu = cn
            else
              cu = ce
            endif
            hhh  = cu/hu

            if (uicen >= zero) then
              p = dt*p0*te*ce**2
            else
              p = dt*p0*tn*cn**2
            endif

            utau = hhh*tau
            stab = one + p                                                     &
                 + (dtcdrwre*(dwu**2/duv+duv)+dtcdrwra*(dau**2/duvw+duvw))*hhh

            it = it + 1
            dd = (uice_o(mi) - uicen + utau - usrf + ucor - p*uicen)/stab
            uicen = uicen + dd
          enddo

          ! finally, new u vel:
          uice(mi) = uicen
        endif
      endif
  
      ! for east border, shift back eastwards
      if (bordertype == 3) j = j + 1
  
      !----------------- end of u-equation -------------------------------------
  
      !----------------- v-equation --------------------------------------------
  
      ! for south border, shift n/s index one point northwards
      ! - shifted back again after v-equation
      if (bordertype == 4) i = i - 1
  
      mi = mmk(1,i,j)
      ms = mmk(1,i+1,j)

      !- skip trivial cases as well as S/E, S/W corners
      if (mi > 0 .and. ms > 0 .and. ncorner /= 3 .and. ncorner /=4) then
  
        vice(mi) = zero

        ! find out if there's anything at all to do:
        cyclenext = .true.
        if (ms > 0) then
          if ((.not.(casus(mi).and.casus(ms))) .and.                           &
              (.not.fastmask(mi)) .and. (.not.fastmask(ms)) .and.              &
              (h(mi) > hgr) .and. (h(ms) > hgr)) cyclenext = .false.
        endif
        if (.not.cyclenext) then

          ! lets continue ...
          me  = mmk(1,i  ,j+1)
          mw  = mmk(1,i,  j-1)
          msw = mmk(1,i+1,j-1)
          tn = ice(1,mi)
          ts = ice(1,ms)
          cn = ice(2,mi)
          cs = ice(2,ms)

          ! extract quantities centred at v point:
          vicen = vice_o(mi)
          vw    = vwat(mi)
          wax   = half*(wx(mi)+wx(ms))
          way   = half*(wy(mi)+wy(ms))
          tx    = dtddx/cosphi(2,i)
          ! fixme: uw terms outside grid are strictly not correct here:
          !        we simply take 2*uwat(mi)
          if (bordertype == 1) then
            ! -- west border
            uicen = qrt*(uice_o(mi) + uice_o(ms) + vel1_c_f + vel3_c_f)
            uw    = qrt*(uwat(mi) + uwat(ms) + two*uwat(mi))
          else
            ! -- north, east, south border
            uicen = qrt*(uice_o(mi) + uice_o(mw) + uice_o(ms) + uice_o(msw))
            uw    = qrt*(uwat(mi) + uwat(mw) + uwat(ms) + uwat(msw))
          endif

          ! current and wind drag terms:
          dwv  = vw- vicen
          dav  = way-vicen
          duv  = sqrt(duv_m2 + (uw- uicen)**2 + dwv**2)
          duvw = sqrt(duv_m2 + (wax-uicen)**2 + dav**2)
          bot  = dtcdrwre*duv
          top  = dtcdrwra*duvw
          tau  = bot*dwv + top*dav
          hv   = half*(tn+ts+z0srf(mi)+z0srf(ms))

          ! extract upwind-centred compactness:
          if (tau >= zero) then
            cv = cs
          else
            cv = cn
          endif
          hhh = cv/hv

          ! down-stream centred strength:
          if (vicen >= zero) then
            p = dt*p0*tn*cn**2
          else
            p = dt*p0*ts*cs**2
          endif

          ! current and wind drag:
          stab = one + p                                                       &
               + (dtcdrwre*(dwv**2/duv+duv) + dtcdrwra*(dav**2/duvw+duvw))*hhh
          vtau = hhh*tau

          ! nabla(z) term:
          vsrf = g*(zeta(mi)-zeta(ms))*dtddy

          ! Coriolis:
          vcor = -ft(i)*uicen

          ! first guess:
          dd = (vtau - vsrf + vcor - p*vicen)/stab
          vicen = vicen + dd

          ! further iterations:
          it = 2
          do while (it <= itmax .and. abs(dd) > ddlim)
            dwv  = vw- vicen
            dav  = way-vicen
            duv  = sqrt(duv_m2 + (uw- uicen)**2 + dwv**2)
            duvw = sqrt(duv_m2 + (wax-uicen)**2 + dav**2)
            bot  = dtcdrwre*duv
            top  = dtcdrwra*duvw
            tau  = bot*dwv + top*dav
            if (tau >= zero) then
              cv = cs
            else
              cv = cn
            endif
            hhh  = cv/hv

            if (vicen >= zero) then
              p = dt*p0*tn*cn**2
            else
              p = dt*p0*ts*cs**2
            endif

            vtau = hhh*tau
            stab = one + p                                                     &
                 + (dtcdrwre*(dwv**2/duv+duv)+dtcdrwra*(dav**2/duvw+duvw))*hhh

            it = it + 1
            dd = (vice_o(mi) - vicen + vtau - vsrf + vcor - p*vicen)/stab
            vicen = vicen + dd
          enddo

          ! finally, new v vel:
          vice(mi) = vicen
        endif
      endif
  
      ! for south border, shift back southwards
      if (bordertype == 4) i = i + 1
  
    !----------------- end of v-equation ---------------------------------------
  
    enddo brdloop
  
    !---------------------------------------------------------------------------
  
  end subroutine ice_mom_c_f

end module icedyn_nest
