module iflow
  use constants, only : zero, one, half, two

  !- directives:
  implicit none
  private

  !- allocatable local arrays:
  real(8), allocatable, public :: ee(:,:)
  real(8), allocatable, private, dimension(:,:) :: rin,rout,eu,ev,dxe,dye,     &
                                                   deu,dev,cx,cy

  !- methods:
  public  :: iflow_alloc  !< Method for allocating local arrays.
  public  :: iflow_ft     !< Method for doing numa aware init, i.e. first-touch
  public  :: iflow_int    !< Routine doing the 2D advection in conservative form
                          !  of ice and snow.
  public  :: iflow_a      !< Routine to do 2D advection in material derivative
                          !  form. To be used with intactness and length 
                          !  scale iota(1:2,0:)  FIXME: is this correct ?

  !- some parameters -----------------------------------------------------------
  real(8), parameter, private :: three = 3.0_8, onehalf = half
  real(8), parameter, private :: oneforth = onehalf*onehalf
  real(8), parameter, private :: onethird = one/three
  real(8), parameter, private :: twothird = two/three
  real(8), parameter, private :: onesixth = onehalf*onethird
  real(8), parameter, private :: tm12 = 1.0e-12_8

contains

  ! ----------------------------------------------------------------------------

subroutine iflow_int (me,n2d,ind,kh,idx,casus,u,v,ice,dxj,dyi,dt,              &
                      cosphi,nbpz,krz,rwze,nbpu,kru,nbpv,krv,nc,ia,mmk)

  !- modules -------------------------------------------------------------------
  use dmi_omp,    only : domp_get_domain
  use dmi_mpi,    only : dmpi_halo
  use cmod_timer, only : timer


  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in)    :: nc, ia, kh(0:), mmk(1:,0:,0:)
  integer(4), intent(in)    :: n2d,nbpz,nbpu,nbpv
  integer(4), intent(in)    :: me(0:,0:)                  ! (0:imax,0:jmax)
  integer(4), intent(in)    :: ind(:,:)                   ! (2,n2d)
  integer(4), intent(in)    :: krz(:,0:)                  ! (3,0:nbpz)
  integer(4), intent(in)    :: kru(:,0:)                  ! (3,0:nbpu)
  integer(4), intent(in)    :: krv(:,0:)                  ! (3,0:nbpv)
  integer(4), intent(inout) :: idx(1:,1:)
  logical, intent(in)       :: casus(0:)                  ! (0:n2d)
  real(8), intent(in)       :: u(0:), v(0:)               ! (0:n2d)
  real(8), intent(in)       :: rwze(:,0:)
  real(8), intent(in)       :: dyi(0:), dxj(0:)           ! (0:imax), (0:jmax)
  real(8), intent(in)       :: cosphi(:,0:)               ! (2,0:imax)
  real(8), intent(in)       :: dt
  real(8), intent(inout)    :: ice(:,0:)


  !- local vars ----------------------------------------------------------------
  integer(4) :: n,i,j,nuw,nue,nvn,nvs,xma,yma,xmb,ymb,ntyp
  real(8)    :: udye(1:nc),udyw(1:nc),vdxn(1:nc),vdxs(1:nc),aa(1:nc),bb(1:nc)
  real(8)    :: wdxdyout,wdxdyin,wdxdyein,wdxdyeout,sx,sy
  real(8)    :: wdxdye,div,ddd,emax(1:nc),emin(1:nc),pin,qin,pout,qout
  real(8)    :: hdt,fac,facx,facy, rwzin,rwzout, vol(1:nc)
  real(8)    :: dx,dy,dxdy,dtdx,dtdy,dtddx,dtddy,hdtddx,hdtddy
  integer(4) :: nl,nu,ic, mme, mms, mmi, mmn, mne, mse, mmw, msw


  dx  = dxj(1)   ! since dxj(0:jmax) is constant
  dy  = dyi(1)   ! since dyi(0:imax) is constant

  ! derived from dx, dy and dt and used throughout:
  hdt    = half*dt
  dxdy   = dx*dy
  dtddx  = dt/dx
  dtddy  = dt/dy
  hdtddx = hdt/dx
  hdtddy = hdt/dy
  dtdx   = dt*dx
  dtdy   = dt*dy

  !- some initialisations:
  eu(1:nc,0) = zero
  ev(1:nc,0) = zero
  ee(1:nc,0) = zero
  dxe(1:nc,0) = zero
  dye(1:nc,0) = zero
  deu(1:nc,0) = zero
  dev(1:nc,0) = zero
  rin (1:nc,0) = zero
  rout(1:nc,0) = zero
  cx(1:nc,0) = zero
  cy(1:nc,0) = zero

!$OMP PARALLEL DEFAULT (SHARED)                                                &
!$OMP          private (n,nl,nu,i,j,fac,mme,mmn,mms,mne,mmw,msw,mse,           &
!$OMP                   xma,yma,aa,xmb,ymb,bb)
  call domp_get_domain(kh, 1, n2d, nl, nu, idx)

  fac = hdtddy
  do n=nl, nu ! Loop #1
    i=ind(1,n)
    j=ind(2,n)

    eu(1:nc,n)=zero

    mme = me(i,j+1)
    if (mme == 0) cycle
    if (casus(n).and.casus(mme)) cycle

    if (u(n) >= zero) then
      eu(1:nc,n)=ice(1:nc,n)
      mmn = me(i-1,j)
      if (v(mmn) < zero) then
        eu(1:nc,n) = eu(1:nc,n) - fac*v(mmn)*(ice(1:nc,mmn)-ice(1:nc,n))
      endif
      if (v(n) > zero) then
        mms = me(i+1,j)
        eu(1:nc,n) = eu(1:nc,n) - fac*v(n)*(ice(1:nc,n)-ice(1:nc,mms))
      endif
    else
      eu(1:nc,n)=ice(1:nc,mme)
      mne =  me(i-1,j+1)
      if (v(mne) < zero) then
        eu(1:nc,n) = eu(1:nc,n) - fac*v(mne)*(ice(1:nc,mne)-ice(1:nc,mme))
      endif
      if (v(mme) > zero) then
        mse = me(i+1,j+1)
        eu(1:nc,n) = eu(1:nc,n) - fac*v(mme)*(ice(1:nc,mme)-ice(1:nc,mse))
      endif
    endif
  enddo

  do n=nl, nu ! Loop #2
    i=ind(1,n)
    j=ind(2,n)

    ev(1:nc,n)=zero

    mms = me(i+1,j)
    if (mms == 0) cycle
    if (casus(n).and.casus(mms)) cycle

    fac = hdtddx/cosphi(1,i)

    if (v(n) <= zero) then
      ev(1:nc,n)=ice(1:nc,n)
      mmw = me(i,j-1)
      if (u(mmw) > zero) then
        ev(1:nc,n) = ev(1:nc,n) - fac*u(mmw)*(ice(1:nc,n)-ice(1:nc,mmw))
      endif
      if (u(n) < zero) then
        mme = me(i,j+1) 
        ev(1:nc,n) = ev(1:nc,n) - fac*u(n)*(ice(1:nc,mme)-ice(1:nc,n))
      endif
    else
      ev(1:nc,n)=ice(1:nc,mms)
      msw = me(i+1,j-1) 
      if (u(msw) > zero) then
        ev(1:nc,n) = ev(1:nc,n) - fac*u(msw)*(ice(1:nc,mms)-ice(1:nc,msw))
      endif
      if (u(mms) < zero) then
        mse = me(i+1,j+1)
        ev(1:nc,n) = ev(1:nc,n) - fac*u(mms)*(ice(1:nc,mse)-ice(1:nc,mms))
      endif
    endif
  enddo

  do n=nl, nu ! Loop #5b
    if (casus(n)) then
      dxe(1:nc,n) = zero
      dye(1:nc,n) = zero
    else
      i=ind(1,n)
      j=ind(2,n)

      mme = me(i,j+1)
      if (mme > 0) then
        xma=1
        aa(1:nc)=ice(1:nc,mme)-ice(1:nc,n)
      else
        xma=0
        aa(1:nc)=zero
      endif
      mmw = me(i,j-1)
      if (mmw > 0) then
        xmb=1
        bb(1:nc)=ice(1:nc,n)-ice(1:nc,mmw)
      else
        xmb=0
        bb(1:nc)=zero
      endif
      if (xma*xmb == 1) then
        dxe(1:nc,n) = half*(aa(1:nc) + bb(1:nc))
      else
        dxe(1:nc,n) = aa(1:nc) + bb(1:nc)
      endif

      mmn = me(i-1,j)
      if (mmn > 0) then
        yma=1
        aa(1:nc)=ice(1:nc,mmn)-ice(1:nc,n)
      else
        yma=0
        aa(1:nc)=zero
      endif
      mms = me(i+1,j)
      if (mms > 0) then
        ymb=1
        bb(1:nc)=ice(1:nc,n)-ice(1:nc,mms)
      else
        ymb=0
        bb(1:nc)=zero
      endif
      if (yma*ymb == 1) then
        dye(1:nc,n) = half*(aa(1:nc) + bb(1:nc))
      else
        dye(1:nc,n) = aa(1:nc) + bb(1:nc)
      endif
    endif
  enddo
!$OMP END PARALLEL

  if (nbpz > 0) then
    do n=1,nbpz ! Loop #6
      i=krz(1,n)
      j=krz(2,n)
      mmi = me(i,j)
      if (mmi == 0) cycle
      ntyp=abs(krz(3,n))
      if     (ntyp == 3) then
        mmw = me(i,j-1)
        dxe(1:nc,mmi) = ice(1:nc,mmi)-ice(1:nc,mmw)
      elseif (ntyp == 1) then
        mme = me(i,j+1)
        dxe(1:nc,mmi) = ice(1:nc,mme)-ice(1:nc,mmi)
      elseif (ntyp == 4) then
        mmn = me(i-1,j)
        dye(1:nc,mmi) = ice(1:nc,mmn)-ice(1:nc,mmi)
      elseif (ntyp == 2) then
        mms = me(i+1,j)
        dye(1:nc,mmi) = ice(1:nc,mmi)-ice(1:nc,mms)
      endif
    enddo
  endif


  if (nbpu > 0) then
    do n=1,nbpu ! Loop #7
      i= kru(1,n)
      j= kru(2,n)
      ntyp=kru(3,n)
      if     (ntyp == 3) then
        if (me(i,j-1) > 0 .and. me(i,j) > 0) then
          dxe(1:nc,me(i,j))=ice(1:nc,me(i,j))-ice(1:nc,me(i,j-1))
        endif
      elseif (ntyp == 1) then
        if (me(i,j+1) > 0 .and. me(i,j+2) > 0) then
          dxe(1:nc,me(i,j+1))=ice(1:nc,me(i,j+2))-ice(1:nc,me(i,j+1))
        endif
      endif
    enddo
  endif


  if (nbpv > 0) then
    do n=1,nbpv ! Loop #8
      i= krv(1,n)
      j= krv(2,n)
      ntyp=krv(3,n)
      if     (ntyp == 4) then
        if (me(i-1,j) > 0 .and. me(i,j) > 0) then
          dye(1:nc,me(i,j))=ice(1:nc,me(i-1,j))-ice(1:nc,me(i,j))
        endif
      elseif (ntyp == 2) then
        if (me(i+1,j) > 0 .and. me(i+2,j) > 0) then
          dye(1:nc,me(i+1,j))=ice(1:nc,me(i+1,j))-ice(1:nc,me(i+2,j))
        endif
      endif
    enddo
  endif

! FIXME: consider:
! Optimization: Use the non-blocking halo swap routine.
!               Split the R/S/W calls, espeically pre-post RECV early.
!$OMP PARALLEL DEFAULT (SHARED)                                                &
!$OMP private(n,nl,nu,i,j,vol,facx,facy,sx,sy,mme,mmw,mmn,mms,mne,msw,mse)
  call timer(1,'MPI_halo')
  call dmpi_halo(1,mmk,kh,ia,1,1,1,6, eu,nc)
  call dmpi_halo(1,mmk,kh,ia,1,1,1,6, ev,nc)
  call dmpi_halo(1,mmk,kh,ia,1,1,1,6,dxe,nc)
  call dmpi_halo(1,mmk,kh,ia,1,1,1,6,dye,nc)
  call timer(2,'MPI_halo')
!$OMP BARRIER

  call domp_get_domain(kh, 1, n2d, nl, nu, idx)

  facy = dtddy
  do n=nl,nu ! Loop #9
    i=ind(1,n)
    j=ind(2,n)

    deu(1:nc,n) = zero

    mme = me(i,j+1)
    if (mme == 0) cycle
    if (casus(n).and.casus(mme)) cycle

    sx = dtddx*u(n)/cosphi(1,i)

    if (u(n) >= zero) then

      deu(1:nc,n) = onehalf*(one-sx)*dxe(1:nc,n)

      mmn = me(i-1,j)
      if (v(mmn) < zero) then
        sy = facy*v(mmn)

        deu(1:nc,n) = deu(1:nc,n)                                              &
        -sy*( (oneforth-onethird*sx)*(dxe(1:nc,mmn)-dxe(1:nc,n))               &
             -(oneforth+onesixth*sy)*(dye(1:nc,mmn)-dye(1:nc,n)))
      endif

      if (v(n) > zero) then
        sy = facy*v(n)

        deu(1:nc,n) = deu(1:nc,n)                                              &
        -sy*( (oneforth-onethird*sx)*(dxe(1:nc,n)-dxe(1:nc,mme))               &
             +(oneforth-onesixth*sy)*(dye(1:nc,n)-dye(1:nc,mme)))
      endif

    else

      deu(1:nc,n) = onehalf*(one+sx)*(-dxe(1:nc,mme))

      mne = me(i-1,j+1)
      if (v(mne) < zero) then
        sy = facy*v(mne)

        deu(1:nc,n) = deu(1:nc,n)                                              &
        -sy*(-(oneforth+onethird*sx)*(dxe(1:nc,mne)-dxe(1:nc,mme))             &
             -(oneforth+onesixth*sy)*(dye(1:nc,mne)-dye(1:nc,mme)))
      endif

      if (v(mme) > zero) then
        sy = facy*v(mme)

        mse = me(i+1,j+1)
        deu(1:nc,n) = deu(1:nc,n)                                              &
        -sy*(-(oneforth+onethird*sx)*(dxe(1:nc,mme)-dxe(1:nc,mse))             &
             +(oneforth-onesixth*sy)*(dye(1:nc,mme)-dye(1:nc,mse)))
      endif

    endif
  enddo


  do n=nl,nu ! Loop #10
    i=ind(1,n)
    j=ind(2,n)

    dev(1:nc,n) = zero

    mms = me(i+1,j)
    if(mms == 0) cycle
    if (casus(n).and.casus(mms)) cycle

    sy   = dtddy*v(n)
    facx = dtddx/cosphi(1,i)

    if (v(n) <= zero) then

      dev(1:nc,n) = onehalf*(one+sy)*(-dye(1:nc,n))

      mmw = me(i,j-1)
      if (u(mmw) > zero) then
        sx = facx*u(mmw)

        dev(1:nc,n) = dev(1:nc,n)                                              &
        -sx*(-(oneforth+onethird*sy)*(dye(1:nc,mmw)-dye(1:nc,n))               &
             +(oneforth-onesixth*sx)*(dxe(1:nc,mmw)-dxe(1:nc,n)))
      endif

      if (u(n) < zero) then
        sx = facx*u(n)

        dev(1:nc,n) = dev(1:nc,n)                                              &
        -sx*(-(oneforth+onethird*sy)*(dye(1:nc,n)-dye(1:nc,mms))               &
             -(oneforth+onesixth*sx)*(dxe(1:nc,n)-dxe(1:nc,mms)))
      endif

    else

      dev(1:nc,n) = onehalf*(one-sy)*dye(1:nc,mms)

      msw = me(i+1,j-1)
      if (u(msw) > zero) then
        sx = facx*u(msw)

        dev(1:nc,n) = dev(1:nc,n)                                              &
        -sx*( (oneforth-onethird*sy)*(dye(1:nc,msw)-dye(1:nc,mms))             &
             +(oneforth-onesixth*sx)*(dxe(1:nc,msw)-dxe(1:nc,mms)))
      endif

      if (u(mms) < zero) then
        sx = facx*u(mms)

        mse = me(i+1,j+1) 
        dev(1:nc,n) = dev(1:nc,n)                                              &
        -sx*( (oneforth-onethird*sy)*(dye(1:nc,mms)-dye(1:nc,mse))             &
             -(oneforth+onesixth*sx)*(dxe(1:nc,mms)-dxe(1:nc,mse)))
      endif

    endif
  enddo


  do n=nl, nu ! Loop #3
    i=ind(1,n)
    j=ind(2,n)

    if (casus(n)) then
      vol(1:nc) = zero
    else
      vol(1:nc) = ice(1:nc,n)*cosphi(1,i)*dxdy
    endif

    mmw = me(i,j-1)
    mmn = me(i-1,j)

    ee(1:nc,n)= vol(1:nc)                                                      &
        - dtdy*(u(n)*eu(1:nc,n)             - u(mmw)*eu(1:nc,mmw))             &
        + dtdx*(v(n)*ev(1:nc,n)*cosphi(2,i) - v(mmn)*ev(1:nc,mmn)*cosphi(2,i-1))
  enddo
!$OMP END PARALLEL


  if (nbpz > 0) then
    do n=1,nbpz ! Loop #4
      i=krz(1,n)
      j=krz(2,n)
      mmi = me(i,j)
      if (mmi == 0) cycle
      if ( n > 1 .and. i == krz(1,n-1) .and. j == krz(2,n-1) ) cycle
      rwzin =zero
      rwzout=zero
      nuw=me(i,j-1)
      nue=mmi
      nvn=me(i-1,j)
      nvs=mmi
      facx = one/(dx*cosphi(1,i))
      if (abs(krz(3,n)) == 3 .and. u(nue) > zero) then
        rwzin  = rwzin  + u(nue)*facx
      else
        rwzin  = rwzin  + min(u(nue),zero)*facx
        rwzout = rwzout + max(u(nue),zero)*facx
      endif
      if (abs(krz(3,n)) == 1 .and. u(nuw) < zero) then
        rwzin  = rwzin  - u(nuw)*facx
      else
        rwzin  = rwzin  - max(u(nuw),zero)*facx
        rwzout = rwzout - min(u(nuw),zero)*facx
      endif
      facy = one/(dy*cosphi(1,i))
      if (abs(krz(3,n)) == 2 .and. v(nvn) > zero) then
        rwzin  = rwzin  + cosphi(2,i-1)*v(nvn)*facy
      else
        rwzin  = rwzin  + cosphi(2,i-1)*min(v(nvn),zero)*facy
        rwzout = rwzout + cosphi(2,i-1)*max(v(nvn),zero)*facy
      endif
      if (abs(krz(3,n)) == 4 .and. v(nvs) < zero) then
        rwzin  = rwzin  - cosphi(2, i )*v(nvs)*facy
      else
        rwzin  = rwzin  - cosphi(2, i )*max(v(nvs),zero)*facy
        rwzout = rwzout - cosphi(2, i )*min(v(nvs),zero)*facy
      endif
      fac       = dxdy*cosphi(1,i)
      wdxdyin   = rwzin *fac
      wdxdyout  = rwzout*fac
      do ic=1,nc
        wdxdyein   = wdxdyin *ice(ic,mmi)
        wdxdyeout  = wdxdyout*rwze(ic,n)
        wdxdye     = wdxdyein + wdxdyeout
        ee(ic,mmi) = ee(ic,mmi) + dt*wdxdye
      enddo
    enddo
  endif


!$OMP PARALLEL DEFAULT (shared)                                                &
!$OMP private(n,nl,nu,i,j,udye,udyw,vdxn,vdxs,fac,facx,facy,                   &
!$OMP         div,ddd,emax,emin,pin,qin,pout,qout,ic,mme,mmw,mmn,mms)
  call domp_get_domain(kh, 1, n2d, nl, nu, idx)

  do n=nl, nu ! Loop #5a
    ee(1:nc,n) = ee(1:nc,n)/(cosphi(1,ind(1,n))*dxdy)
  enddo


!$OMP BARRIER
  call timer(1,'MPI_halo')
  call dmpi_halo(1,mmk,kh,ia,1,1,1,6,deu,nc)
  call dmpi_halo(1,mmk,kh,ia,1,1,1,6,dev,nc)
  call dmpi_halo(1,mmk,kh,ia,1,1,1,6, ee,nc)
  call timer(2,'MPI_halo')
!$OMP BARRIER


  do n=nl,nu ! Loop #11
    i = ind(1,n)
    j = ind(2,n)

    mmw = me(i,j-1)
    udye(1:nc) = u(n)*deu(1:nc,n)
    udyw(1:nc) = u(mmw)*deu(1:nc,mmw)
    
    mmn = me(i-1,j)
    vdxs(1:nc) = v(n)*dev(1:nc,n)*cosphi(2,i)
    vdxn(1:nc) = v(mmn)*dev(1:nc,mmn)*cosphi(2,i-1)

    fac  = one/(cosphi(1,i)*dxdy)
    facx = dtdy*fac
    facy = dtdx*fac

    do ic=1,nc
      rout(ic,n) = facx*(  max(udye(ic),zero) - min(udyw(ic),zero) )           &
                 + facy*(- min(vdxs(ic),zero) + max(vdxn(ic),zero) ) 

      rin (ic,n) = facx*(- min(udye(ic),zero) + max(udyw(ic),zero) )           &
                 + facy*(  max(vdxs(ic),zero) - min(vdxn(ic),zero) )
    enddo
  enddo

  do n=nl,nu ! Loop #12
    i=ind(1,n)
    j=ind(2,n)

    facx = dtddx/cosphi(1,i)
    facy = dtddy/cosphi(1,i)
    div = (u(n)-u(me(i,j-1)))*facx                                             &
        + (cosphi(2,i-1)*v(me(i-1,j))-cosphi(2,i)*v(n))*facy
    fac = max(one-div,zero)
    do ic=1,nc
      ddd      = fac*ice(ic,n)
      emax(ic) = max(ddd,ee(ic,n))
      emin(ic) = min(ddd,ee(ic,n))
    enddo

    mmn = me(i-1,j)
    if (mmn > 0) then
      div = (u(mmn)-u(me(i-1,j-1)))*dtddx/cosphi(1,i-1)                        &
          + (cosphi(2,i-2)*v(me(i-2,j))-cosphi(2,i-1)*v(mmn))                  &
           *dtddy/cosphi(1,i-1)
      fac = max(one-div,zero)
      do ic=1,nc
        ddd      = fac*ice(ic,mmn)
        emax(ic) = max(emax(ic),ddd,ee(ic,mmn))
        emin(ic) = min(emin(ic),ddd,ee(ic,mmn))
      enddo
    endif
    mms = me(i+1,j)
    if (mms > 0) then
      div = (u(mms)-u(me(i+1,j-1)))*dtddx/cosphi(1,i+1)                        &
          + (cosphi(2,i)*v(n)-cosphi(2,i+1)*v(mms))                            &
           *dtddy/cosphi(1,i+1)
      fac = max(one-div,zero)
      do ic=1,nc
        ddd      = fac*ice(ic,mms)
        emax(ic) = max(emax(ic),ddd,ee(ic,mms))
        emin(ic) = min(emin(ic),ddd,ee(ic,mms))
      enddo
    endif
    mmw = me(i,j-1)
    if (mmw > 0) then
      div = (u(mmw)-u(me(i,j-2)))*facx                                         &
          + (cosphi(2,i-1)*v(me(i-1,j-1))-cosphi(2,i)*v(mmw))*facy        
      fac = max(one-div,zero)
      do ic=1,nc
        ddd      = fac*ice(ic,mmw)
        emax(ic) = max(emax(ic),ddd,ee(ic,mmw))
        emin(ic) = min(emin(ic),ddd,ee(ic,mmw))
      enddo
    endif
    mme = me(i,j+1)
    if (mme > 0) then
      div = (u(mme)-u(n))*facx                                                 &
          + (cosphi(2,i-1)*v(me(i-1,j+1))-cosphi(2,i)*v(mme))*facy
      fac = max(one-div,zero)
      do ic=1,nc
        ddd      = fac*ice(ic,mme)
        emax(ic) = max(emax(ic),ddd,ee(ic,mme))
        emin(ic) = min(emin(ic),ddd,ee(ic,mme))
      enddo
    endif

    do ic=1,nc
      pin=rin(ic,n)
      !     pin>=0
      if (pin > tm12) then
        qin=emax(ic)-ee(ic,n)
        !     qin>=0
        rin(ic,n)=min(one,qin/pin)
      else
        rin(ic,n)=one
      endif

      pout=rout(ic,n)
      !     pout>=0
      if (pout > tm12) then
        qout=ee(ic,n)-emin(ic)
        !     qout>=0
        rout(ic,n)=min(one,qout/pout)
      else
        rout(ic,n)=one
      endif
    enddo

  enddo


!$OMP BARRIER
  call timer(1,'MPI_halo')
  call dmpi_halo(1,mmk,kh,ia,1,1,1,6,rout,nc)
  call dmpi_halo(1,mmk,kh,ia,1,1,1,6,rin ,nc)
  call timer(2,'MPI_halo')
!$OMP BARRIER


  do n=nl, nu ! Loop #13
    cx(1:nc,n) = zero
    mme = me(ind(1,n),ind(2,n)+1)
    if (mme == 0) cycle
    do ic=1,nc
      if (u(n)*deu(ic,n) >= zero) then
        cx(ic,n)=min(rin(ic,mme),rout(ic,n))
      else
        cx(ic,n)=min(rin(ic,n),rout(ic,mme))
      endif
    enddo
  enddo

  do n=nl, nu ! Loop #14
    cy(1:nc,n) = zero
    mms = me(ind(1,n)+1,ind(2,n))
    if(mms == 0) cycle
    do ic=1,nc
      if (v(n)*dev(ic,n) <= zero) then
        cy(ic,n)=min(rin(ic,mms),rout(ic,n))
      else
        cy(ic,n)=min(rin(ic,n),rout(ic,mms))
      endif
    enddo
  enddo


!$OMP BARRIER
  call timer(1,'MPI_halo')
  call dmpi_halo(1,mmk,kh,ia,1,1,1,6,cx,nc)
  call dmpi_halo(1,mmk,kh,ia,1,1,1,6,cy,nc)
  call timer(2,'MPI_halo')
!$OMP BARRIER


  do n=nl,nu ! Loop #15
    i=ind(1,n)
    j=ind(2,n)

    mmw = me(i,j-1)
    udye(1:nc) = u(n)*cx(1:nc,n)*deu(1:nc,n)
    udyw(1:nc) = u(mmw)*cx(1:nc,mmw)*deu(1:nc,mmw)

    mmn = me(i-1,j)
    vdxs(1:nc) = v(n)*cosphi(2,i)*cy(1:nc,n)*dev(1:nc,n)
    vdxn(1:nc) = v(mmn)*cosphi(2,i-1)*cy(1:nc,mmn)*dev(1:nc,mmn)

    fac  = one/(cosphi(1,i)*dxdy)
    facx = dtdy*fac
    facy = dtdx*fac

    ice(1:nc,n) = ee(1:nc,n) + facx*(-udye(1:nc)+udyw(1:nc))                   &
                             + facy*( vdxs(1:nc)-vdxn(1:nc))

  enddo
!$OMP END PARALLEL

end subroutine iflow_int

  ! ----------------------------------------------------------------------------

  subroutine iflow_alloc (n2dm, nc, ierr)
    implicit none

    integer(4), intent(in)    :: n2dm, nc
    integer(4), intent(inout) :: ierr

    allocate(ee(1:nc,0:n2dm), eu(1:nc,0:n2dm), ev(1:nc,0:n2dm),                &
             rin(1:nc,0:n2dm), rout(1:nc,0:n2dm),                              &
             dxe(1:nc,0:n2dm), dye(1:nc,0:n2dm),                               &
             deu(1:nc,0:n2dm), dev(1:nc,0:n2dm),                               &
             cx(1:nc,0:n2dm), cy(1:nc,0:n2dm), stat=ierr )

  end subroutine iflow_alloc

  ! ----------------------------------------------------------------------------

  subroutine iflow_ft (n2d, n2dhalo, kh, nc, idx)
    use dmi_omp, only : domp_get_domain
    implicit none

    integer(4), intent(in)    :: n2d, n2dhalo, kh(0:), nc
    integer(4), intent(inout) :: idx(1:,1:)

    integer(4) :: nl, nu, n

    ! NUMA-initialize all arrays:

!$OMP MASTER
    ee(1:nc,0)   = zero
    eu(1:nc,0)   = zero
    ev(1:nc,0)   = zero
    rin(1:nc,0)  = zero
    rout(1:nc,0) = zero
    dxe(1:nc,0)  = zero
    dye(1:nc,0)  = zero
    deu(1:nc,0)  = zero
    dev(1:nc,0)  = zero
    cx(1:nc,0)   = zero
    cy(1:nc,0)   = zero

    do n=n2d+1,n2d+n2dhalo
      ee(1:nc,n)   = zero
      eu(1:nc,n)   = zero
      ev(1:nc,n)   = zero
      rin(1:nc,n)  = zero
      rout(1:nc,n) = zero
      dxe(1:nc,n)  = zero
      dye(1:nc,n)  = zero
      deu(1:nc,n)  = zero
      dev(1:nc,n)  = zero
      cx(1:nc,n)   = zero
      cy(1:nc,n)   = zero
    enddo
!$OMP END MASTER

    call domp_get_domain(kh, 1, n2d, nl, nu, idx)

    do n=nl,nu
      ee(1:nc,n)   = zero
      eu(1:nc,n)   = zero
      ev(1:nc,n)   = zero
      rin(1:nc,n)  = zero
      rout(1:nc,n) = zero
      dxe(1:nc,n)  = zero
      dye(1:nc,n)  = zero
      deu(1:nc,n)  = zero
      dev(1:nc,n)  = zero
      cx(1:nc,n)   = zero
      cy(1:nc,n)   = zero
    enddo

  end subroutine iflow_ft

  ! ----------------------------------------------------------------------------

  subroutine iflow_a (m,n2d,kh,ind,casus,u,v,iota,dxj,dyi,dt,cosphi,nbpz,krz,  &
                      rwziota,nbpu,kru,nbpv,krv,nc,ia,mmk,fastmask)

    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain
    use dmi_mpi, only : dmpi_halo

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: nc, ia, mmk(1:,0:,0:), kh(0:)
    integer(4), intent(in) :: n2d,nbpz,nbpu,nbpv
    integer(4), intent(in) :: m(0:,0:)                   ! (0:imax,0:jmax)
    integer(4), intent(in) :: ind(:,:)                   ! (2,n2d)
    integer(4), intent(in) :: krz(:,0:)                  ! (3,0:nbpz)
    integer(4), intent(in) :: kru(:,0:)                  ! (3,0:nbpu)
    integer(4), intent(in) :: krv(:,0:)                  ! (3,0:nbpv)
    logical,    intent(in) :: casus(0:)                  ! (0:n2d)
    real(8),    intent(in) :: u(0:), v(0:)               ! (0:n2d)
    real(8),    intent(in) :: rwziota(:,0:)              ! (1:nc,0:nbpz)
    real(8),    intent(in) :: dyi(0:), dxj(0:)           ! (0:imax), (0:jmax)
    real(8),    intent(in) :: cosphi(:,0:)               ! (2,0:imax)
    real(8),    intent(in) :: dt
    real(8),    intent(inout) :: iota(:,0:)              ! (1:nc,0:n2d)
    logical,    intent(inout) :: fastmask(0:)            ! (0:n2d)

    integer(4) :: n, i, j, mi, nuw, nue, nvn, nvs, ntyp, ms, me, mw, mn
    integer(4) :: msw, mse, mne, mee, mss, nt, nl, nu
    real(8)    :: aa(1:nc), bb(1:nc), dd1(1:nc), dd2(1:nc)
    real(8)    :: emax, emin, pin, pout, qin, qout, sx, sy, dx, dy
    real(8)    :: fac, facx, facy, hdt, hdtddy, dxdy
    real(8)    :: dtdy, dtdx, dtddx, dtddy, vol, ee0, rwzout
    real(8)    :: udye, udyw, vdxs, vdxn


    dx = dxj(1)   ! since dxj(0:jmax) is constant
    dy = dyi(1)   ! since dyi(0:imax) is constant

    ! derived from dx, dy and dt and used throughout:
    hdt    = half*dt
    dxdy   = dx*dy
    dtddx  = dt/dx
    dtddy  = dt/dy
    hdtddy = hdt/dy
    dtdx   = dt*dx
    dtdy   = dt*dy

    !- some initialisations:
    eu(1:nc,0)   = zero
    ev(1:nc,0)   = zero
    ee(1:nc,0)   = zero
    dxe(1:nc,0)  = zero
    dye(1:nc,0)  = zero
    deu(1:nc,0)  = zero
    dev(1:nc,0)  = zero
    rin (1:nc,0) = zero
    rout(1:nc,0) = zero
    cx(1:nc,0)   = zero
    cy(1:nc,0)   = zero


!$OMP PARALLEL DEFAULT (SHARED)                                                &
!$OMP          private(n,nl,nu,i,j,me,mn,ms,mw,mne,mse,msw,fac,facx,aa,bb)
    call domp_get_domain(1,n2d, nl, nu)

    do n=nl,nu ! Loop #1
      i  = ind(1,n)
      j  = ind(2,n)

      eu(1:nc,n)= zero

      me = m(i,j+1)
      if (me == 0) cycle

      if (casus(n).and.casus(me)) cycle

      if (u(n) >= zero) then

        eu(1:nc,n)=iota(1:nc,n)

        mn = m(i-1,j)
        if (v(mn) < zero) then
          fac   = hdtddy*v(mn)
          eu(1:nc,n) = eu(1:nc,n) - fac*(iota(1:nc,mn)-iota(1:nc,n))
        endif

        ms = m(i+1,j)
        if (v(n) > zero) then
          fac   = hdtddy*v(n)
          eu(1:nc,n) = eu(1:nc,n) - fac*(iota(1:nc,n)-iota(1:nc,ms))
        endif

      else

        eu(1:nc,n)=iota(1:nc,me)

        mne = m(i-1,j+1)
        if (v(mne) < zero) then
          fac   = hdtddy*v(mne)
          eu(1:nc,n) = eu(1:nc,n) - fac*(iota(1:nc,mne)-iota(1:nc,me))
        endif

        mse = m(i+1,j+1)
        if (v(me) > zero) then
          fac   = hdtddy*v(me)
          eu(1:nc,n) = eu(1:nc,n) - fac*(iota(1:nc,me)-iota(1:nc,mse))
        endif

      endif
    enddo


    do n=nl,nu ! Loop #2
      i  = ind(1,n)
      j  = ind(2,n)

      ev(1:nc,n)= zero

      ms = m(i+1,j)
      if (ms == 0) cycle

      if (casus(n).and.casus(ms)) cycle
      facx = hdtddy/cosphi(1,i)

      if (v(n) <= zero) then

        ev(1:nc,n)=iota(1:nc,n)

        mw = m(i,j-1)
        if (u(mw) > zero) then
          fac   = facx*u(mw)
          ev(1:nc,n) = ev(1:nc,n) - fac*(iota(1:nc,n)-iota(1:nc,mw))
        endif

        me = m(i,j+1)
        if (u(n) < zero) then
          fac   = facx*u(n)
          ev(1:nc,n) = ev(1:nc,n) - fac*(iota(1:nc,me)-iota(1:nc,n))
        endif

      else

        ev(1:nc,n)=iota(1:nc,ms)

        msw = m(i+1,j-1)
        if (u(msw) > zero) then
          fac   = facx*u(msw)
          ev(1:nc,n) = ev(1:nc,n) - fac*(iota(1:nc,ms)-iota(1:nc,msw))
        endif

        mse = m(i+1,j+1)
        if (u(ms) < zero) then
          fac   = facx*u(ms)
          ev(1:nc,n) = ev(1:nc,n) - fac*(iota(1:nc,mse)-iota(1:nc,ms))
        endif

      endif
    enddo


    do n=nl,nu ! Loop #5b
      if (casus(n)) then
        dxe(1:nc,n)= zero
        dye(1:nc,n)= zero
      else
        i=ind(1,n)
        j=ind(2,n)

        me = m(i,j+1)
        if (me > 0) then
          aa(1:nc)=iota(1:nc,me)-iota(1:nc,n)
        else
          aa(1:nc)= zero
        endif
        mw = m(i,j-1)
        if (mw > 0) then
          bb(1:nc)=iota(1:nc,n)-iota(1:nc,mw)
        else
          bb(1:nc)= zero
        endif
        if (me > 0 .and. mw > 0) then
          dxe(1:nc,n) = half*(aa(1:nc)+bb(1:nc))
        else
          dxe(1:nc,n) = aa(1:nc) + bb(1:nc)
        endif

        mn = m(i-1,j)
        if (mn > 0) then
          aa(1:nc)=iota(1:nc,mn)-iota(1:nc,n)
        else
          aa(1:nc)= zero
        endif
        ms = m(i+1,j)
        if (ms > 0) then
          bb(1:nc)=iota(1:nc,n)-iota(1:nc,ms)
        else
          bb(1:nc)= zero
        endif
        if (mn > 0 .and. ms > 0) then
          dye(1:nc,n) = half*(aa(1:nc)+bb(1:nc))
        else
          dye(1:nc,n) = aa(1:nc) + bb(1:nc)
        endif
      endif
    enddo
!$OMP END PARALLEL


    if (nbpz > 0) then
      do n=1,nbpz ! Loop #6
        i=krz(1,n)
        j=krz(2,n)
        mi = m(i,j)
        if (mi <= 0) cycle
        ntyp=abs(krz(3,n))
        if (ntyp == 3) then
          mw = m(i,j-1)
          if (mw > 0) then
            dxe(1:nc,mi)=iota(1:nc,mi)-iota(1:nc,mw)
          else
            dxe(1:nc,mi)=iota(1:nc,mi)
          endif
        elseif (ntyp == 1) then
          me = m(i,j+1)
          if (me > 0) then
            dxe(1:nc,mi)=iota(1:nc,me)-iota(1:nc,mi)
          else
            dxe(1:nc,mi)=-iota(1:nc,mi)
          endif
        elseif (ntyp == 4) then
          mn = m(i-1,j)
          if (mn > 0) then
            dye(1:nc,mi)=iota(1:nc,mn)-iota(1:nc,mi)
          else
            dye(1:nc,mi)=-iota(1:nc,mi)
          endif
        elseif (ntyp == 2) then
          ms = m(i+1,j)
          if (ms > 0) then
            dye(1:nc,mi)=iota(1:nc,mi)-iota(1:nc,ms)
          else
            dye(1:nc,mi)=iota(1:nc,mi)
          endif
        endif
      enddo
    endif


    if (nbpu > 0) then
      do n=1,nbpu ! Loop #7
        i=kru(1,n)
        j=kru(2,n)
        ntyp=kru(3,n)
        if (ntyp == 3) then
          mi = m(i,j)
          mw = m(i,j-1)
          if (mi > 0 .and. mw > 0) then
            dxe(1:nc,mi)=iota(1:nc,mi)-iota(1:nc,mw)
          endif
        elseif (ntyp == 1) then
          me  = m(i,j+1)
          mee = m(i,j+2)
          if (me > 0 .and. mee > 0) then
            dxe(1:nc,me)=iota(1:nc,mee)-iota(1:nc,me)
          endif
        endif
      enddo
    endif


    if (nbpv > 0) then
      do n=1,nbpv ! Loop #8
        i= krv(1,n)
        j= krv(2,n)
        ntyp=krv(3,n)
        if (ntyp == 4) then
          mi = m(i,j)
          mn = m(i-1,j)
          if (mn > 0 .and. mi > 0) then
            dye(1:nc,mi)=iota(1:nc,mn)-iota(1:nc,mi)
          endif
        elseif (ntyp == 2) then
          ms  = m(i+1,j)
          mss = m(i+2,j)
          if (ms > 0 .and. mss > 0) then
            dye(1:nc,ms)=iota(1:nc,ms)-iota(1:nc,mss)
          endif
        endif
      enddo
    endif


!$OMP PARALLEL DEFAULT (SHARED)                                                &
!$OMP          private(n,nl,nu,i,j,me,mn,ms,mw,mne,mse,msw,fac,facx,facy,      &
!$OMP                  dd1,dd2,sx,sy,vol,ee0)
    call dmpi_halo(1,mmk,kh,ia,1,1,1,6, eu,nc)
    call dmpi_halo(1,mmk,kh,ia,1,1,1,6, ev,nc)
    call dmpi_halo(1,mmk,kh,ia,1,1,1,6,dxe,nc)
    call dmpi_halo(1,mmk,kh,ia,1,1,1,6,dye,nc)
!$OMP BARRIER


    call domp_get_domain(1,n2d, nl, nu)

    do n=nl,nu ! Loop #9
      i=ind(1,n)
      j=ind(2,n)

      ! initialize all four here for simplicity:
      deu (1:nc,n) = zero
      dev (1:nc,n) = zero

      me = m(i,j+1)
      if (me <= 0) cycle
      if (casus(n).and.casus(me)) cycle

      if (u(n) >= zero) then

        sx     = dtddx*u(n)/cosphi(1,i)  ! >= 0
        deu(1:nc,n) = half*(one-sx)*dxe(1:nc,n)

        mn = m(i-1,j)
        if (v(mn) < zero) then
          sy     = dtddy*v(mn)  ! < 0
          facy   = (half+onethird*sy)
          facx   = (half-twothird*sx)
          dd1(1:nc)   = facx*dxe(1:nc,mn) - facy*dye(1:nc,mn)
          dd2(1:nc)   = facx*dxe(1:nc,n)  - facy*dye(1:nc,n)
          deu(1:nc,n) = deu(1:nc,n) - half*sy*(dd1(1:nc)-dd2(1:nc))
        endif

        ms = m(i+1,j)
        if (v(n) > zero) then
          sy     = dtddy*v(n)  ! > 0
          facy   = (half-onethird*sy)
          facx   = (half-twothird*sx)
          dd1(1:nc)   = facx*dxe(1:nc,n)  + facy*dye(1:nc,n)
          dd2(1:nc)   = facx*dxe(1:nc,ms) + facy*dye(1:nc,ms)
          deu(1:nc,n) = deu(1:nc,n) - half*sy*(dd1(1:nc)-dd2(1:nc))
        endif

      else

        sx     = dtddx*u(n)/cosphi(1,i)  ! < 0
        deu(1:nc,n) = half*(one+sx)*(-dxe(1:nc,me))
  
        mne = m(i-1,j+1)
        if (v(mne) < zero) then
          sy     = dtddy*v(mne)  ! < 0
          facy   = (half+onethird*sy)
          facx   =-(half+twothird*sx)
          dd1(1:nc)   = facx*dxe(1:nc,mne) - facy*dye(1:nc,mne)
          dd2(1:nc)   = facx*dxe(1:nc,me)  - facy*dye(1:nc,me)
          deu(1:nc,n) = deu(1:nc,n) - half*sy*(dd1(1:nc)-dd2(1:nc))
        endif

        mse = m(i+1,j+1)
        if (v(me) > zero) then
          sy     = dtddy*v(me)  ! > 0
          facy   = (half-onethird*sy)
          facx   =-(half+twothird*sx)
          dd1(1:nc)   = facx*dxe(1:nc,me)  + facy*dye(1:nc,me)
          dd2(1:nc)   = facx*dxe(1:nc,mse) + facy*dye(1:nc,mse)
          deu(1:nc,n) = deu(1:nc,n) - half*sy*(dd1(1:nc)-dd2(1:nc))
        endif

      endif
    enddo


    do n=nl,nu ! Loop #10
      i  = ind(1,n)
      j  = ind(2,n)

      ms = m(i+1,j)
      if (ms == 0) cycle
      if (casus(n).and.casus(ms)) cycle

      if (v(n) <= zero) then

        sy     = dtddy*v(n)  ! <= 0
        dev(1:nc,n) = half*(one+sy)*(-dye(1:nc,n))

        mw = m(i,j-1)
        if (u(mw) > zero) then
          sx     = dtddx*u(mw)/cosphi(1,i)  ! > 0
          facx   = (half-onethird*sx)
          facy   =-(half+twothird*sy)
          dd1(1:nc)   = facy*dye(1:nc,mw) + facx*dxe(1:nc,mw)
          dd2(1:nc)   = facy*dye(1:nc,n)  + facx*dxe(1:nc,n)
          dev(1:nc,n) = dev(1:nc,n) - half*sx*(dd2(1:nc)-dd1(1:nc))
        endif

        me = m(i,j+1)
        if (u(n) < zero) then
          sx     = dtddx*u(n)/cosphi(1,i)  ! < 0
          facx   = (half+onethird*sx)
          facy   =-(half+twothird*sy)
          dd1(1:nc)   = facy*dye(1:nc,n)  - facx*dxe(1:nc,n)
          dd2(1:nc)   = facy*dye(1:nc,me) - facx*dxe(1:nc,me)
          dev(1:nc,n) = dev(1:nc,n) - half*sx*(dd2(1:nc)-dd1(1:nc))
        endif

      else

        sy     = dtddy*v(n)  ! > 0
        ms     = m(i+1,j)
        dev(1:nc,n) = half*(one-sy)*dye(1:nc,ms)

        msw = m(i+1,j-1)
        if (u(msw) > zero) then
          sx     = dtddx*u(msw)/cosphi(1,i)  ! > 0
          facx   = (half-onethird*sx)
          facy   = (half-twothird*sy)
          dd1(1:nc)   = facy*dye(1:nc,msw) + facx*dxe(1:nc,msw)
          dd2(1:nc)   = facy*dye(1:nc,ms)  + facx*dxe(1:nc,ms)
          dev(1:nc,n) = dev(1:nc,n) - half*sx*(dd2(1:nc)-dd1(1:nc))
        endif

        mse = m(i+1,j+1)
        if (u(ms) < zero) then
          sx     = dtddx*u(ms)/cosphi(1,i)  ! < 0
          facx   = (half+onethird*sx)
          facy   = (half-twothird*sy)
          dd1(1:nc)    = facy*dye(1:nc,ms)  - facx*dxe(1:nc,ms)
          dd2(1:nc)    = facy*dye(1:nc,mse) - facx*dxe(1:nc,mse)
          dev(1:nc,n) = dev(1:nc,n) - half*sx*(dd2(1:nc)-dd1(1:nc))
        endif

      endif
    enddo


    do n=nl,nu ! Loop #3
      i = ind(1,n)
      j = ind(2,n)
      if (casus(n)) then
        vol = zero
      else
        vol = dxdy*cosphi(1,i)
      endif

      facx = dtdx*cosphi(2,i)

      mw = m(i,j-1)
      mn = m(i-1,j)

      ee0   = - (u(n) - u(mw))*dtdy - (v(ms) - v(n))*facx

      ee(1:nc,n) = (vol - ee0)*iota(1:nc,n)                                    &
                 - (u(n)*eu(1:nc,n) - u(mw)*eu(1:nc,mw))*dtdy                  &
                 - (v(mn)*ev(1:nc,mn) - v(n)*ev(1:nc,n))*facx
    enddo
!$OMP END PARALLEL


    if (nbpz > 0) then
      do n=1,nbpz ! Loop #4
        i=krz(1,n)
        j=krz(2,n)
        if (n > 1 .and. i == krz(1,n-1) .and. j == krz(2,n-1)) cycle
        mi = m(i,j)
        if (mi <= 0) cycle
  
        rwzout= zero
        nuw= m(i,j-1)
        nue= mi
        nvn= m(i-1,j)
        nvs= mi

        facx = dx*cosphi(1,i)
        if (abs(krz(3,n)) == 3 .and. u(nue) > zero) then
        else
          rwzout=rwzout + max(u(nue), zero)/facx
        endif
        if (abs(krz(3,n)) == 1 .and. u(nuw) < zero) then
        else
          rwzout=rwzout - min(u(nuw), zero)/facx
        endif
        facy = cosphi(2,i-1)/(dy*cosphi(1,i))
        if (abs(krz(3,n)) == 2 .and. v(nvn) > zero) then
        else
          rwzout=rwzout + max(v(nvn), zero)*facy
        endif
        facy = cosphi(2,i)/(dy*cosphi(1,i))
        if (abs(krz(3,n)) == 4 .and. v(nvs) < zero) then
        else
          rwzout=rwzout - min(v(nvs), zero)*facy
        endif
        fac = dtdy*facx*rwzout
        ee(1:nc,mi) = ee(1:nc,mi) + fac*(rwziota(1:nc,n) - iota(1:nc,mi))
      enddo
    endif
    
!$OMP PARALLEL DEFAULT (SHARED)                                                &
!$OMP private(n,nl,nu,i,j,me,mn,ms,mw,nt,fac,emin,emax,pin,qin,pout,qout,      &
!$OMP         facx,facy,udye,udyw,vdxs,vdxn)
    call domp_get_domain(1,n2d, nl, nu)

    do n=nl,nu ! Loop #5a
      ee(1:nc,n) = ee(1:nc,n)/(dxdy*cosphi(1,ind(1,n)))
    enddo


!$OMP BARRIER
    call dmpi_halo(1,mmk,kh,ia,1,1,1,6,deu,nc)
    call dmpi_halo(1,mmk,kh,ia,1,1,1,6,dev,nc)
    call dmpi_halo(1,mmk,kh,ia,1,1,1,6, ee,nc)
!$OMP BARRIER


    do n=nl,nu ! Loop #11a
      i  = ind(1,n)
      j  = ind(2,n)

      rin (1:nc,n) = zero
      rout(1:nc,n) = zero

      me = m(i,j+1)
      if (me > 0 .and. .not.(casus(n).and.casus(me))) then
        do nt=1,nc
          fac = dtdy*u(n)*deu(nt,n)
          if (fac >= zero) then
            rout(nt,n) = rout(nt,n) + fac
          else
            rin(nt,n) = rin(nt,n) - fac
          endif
        enddo
      endif

      mw = m(i,j-1)
      if (mw > 0 .and. .not.(casus(n).and.casus(mw))) then
        do nt=1,nc
          fac = dtdy*u(mw)*deu(nt,mw)
          if (fac >= zero) then
            rin(nt,n) = rin(nt,n) + fac
          else
            rout(nt,n) = rout(nt,n) - fac
          endif
        enddo
      endif
    enddo


    do n=nl,nu ! Loop #11b
      i  = ind(1,n)
      j  = ind(2,n)

      ms = m(i+1,j)
      if (ms > 0 .and. .not.(casus(n).and.casus(ms))) then
        do nt=1,nc
          fac = dtdx*v(n)*cosphi(2,i)*dev(nt,n)
          if (fac >= zero) then
            rin(nt,n) = rin(nt,n) + fac
          else
            rout(nt,n) = rout(nt,n) - fac
          endif
        enddo
      endif

      mn = m(i-1,j)
      if (mn > 0 .and. .not.(casus(n).and.casus(mn))) then
        do nt=1,nc
          fac = dtdx*v(mn)*cosphi(2,i)*dev(nt,mn)
          if (fac >= zero) then
            rout(nt,n) = rout(nt,n) + fac
          else
            rin(nt,n) = rin(nt,n) - fac
          endif
        enddo
      endif
    enddo


    do n=nl,nu ! Loop #11c
      fac = dxdy*cosphi(1,ind(1,n))
      rin (1:nc,n) = rin (1:nc,n)/fac
      rout(1:nc,n) = rout(1:nc,n)/fac
    enddo


    do n=nl,nu ! Loop #12
      i=ind(1,n)
      j=ind(2,n)

      do nt=1,nc
        emax=max(iota(nt,n),ee(nt,n))
        emin=min(iota(nt,n),ee(nt,n))
        mn = m(i-1,j)
        if (mn > 0) then
          emax=max(emax,iota(nt,mn),ee(nt,mn))
          emin=min(emin,iota(nt,mn),ee(nt,mn))
        endif
        ms = m(i+1,j)
        if (ms > 0) then
          emax=max(emax,iota(nt,ms),ee(nt,ms))
          emin=min(emin,iota(nt,ms),ee(nt,ms))
        endif
        mw = m(i,j-1)
        if (mw > 0) then
          emax=max(emax,iota(nt,mw),ee(nt,mw))
          emin=min(emin,iota(nt,mw),ee(nt,mw))
        endif
        me = m(i,j+1)
        if (me > 0) then
          emax=max(emax,iota(nt,me),ee(nt,me))
          emin=min(emin,iota(nt,me),ee(nt,me))
        endif

        pin=rin(nt,n)
        if (pin > tm12) then
          qin=emax-ee(nt,n)
          rin(nt,n)=min(one,qin/pin)
        else
          rin(nt,n)=one
        endif

        pout=rout(nt,n)
        if (pout > tm12) then
          qout=ee(nt,n)-emin
          rout(nt,n)=min(one,qout/pout)
        else
          rout(nt,n)=one
        endif
      enddo
    enddo


!$OMP BARRIER
    call dmpi_halo(1,mmk,kh,ia,1,1,1,6,rout,nc)
    call dmpi_halo(1,mmk,kh,ia,1,1,1,6,rin ,nc)
!$OMP BARRIER


    do n=nl,nu ! Loop #13
      cx(1:nc,n) = zero
      me = m(ind(1,n),ind(2,n)+1)
      if (me == 0) cycle
      do nt=1,nc
        if (u(n)*deu(nt,n) >= zero) then
          cx(nt,n)=min(rin(nt,me),rout(nt,n))
        else
          cx(nt,n)=min(rin(nt,n),rout(nt,me))
        endif
      enddo
    enddo


    do n=nl,nu ! Loop #14
      cy(1:nc,n) = zero
      ms = m(ind(1,n)+1,ind(2,n))
      if (ms == 0) cycle
      do nt=1,nc
        if (v(n)*dev(nt,n) <= zero) then
          cy(nt,n)=min(rin(nt,ms),rout(nt,n))
        else
          cy(nt,n)=min(rin(nt,n),rout(nt,ms))
        endif
      enddo
    enddo


!$OMP BARRIER
    call dmpi_halo(1,mmk,kh,ia,1,1,1,6,cx,nc)
    call dmpi_halo(1,mmk,kh,ia,1,1,1,6,cy,nc)
!$OMP BARRIER


    do n=nl,nu ! Loop #15
      i = ind(1,n)
      j = ind(2,n)

      mw = m(i,j-1)
      mn = m(i-1,j)

      fac  = one/(cosphi(1,i)*dxdy)
      facx = dtdy*fac
      facy = dtdx*fac

      udye = u(n)*facx
      udyw = u(mw)*facx

      vdxs = v(n)*cosphi(2,i)*facy
      vdxn = v(mn)*cosphi(2,i-1)*facy

      iota(1:nc,n) = ee(1:nc,n) - udye*cx(1:nc,n )*deu(1:nc,n)                 &
                                + udyw*cx(1:nc,mw)*deu(1:nc,mw)                &
                                + vdxs*cy(1:nc,n )*dev(1:nc,n)                 &
                                - vdxn*cy(1:nc,mn)*dev(1:nc,mn)
      fastmask(n)  = (iota(1,n) > one)
    enddo
!$OMP END PARALLEL

  end subroutine iflow_a

  ! ----------------------------------------------------------------------------

end module iflow
