module io_control
  !-----------------------------------------------------------------------------
  ! Module for controlling the asynchronous IO
  !
  ! TODO: * Need to condider isend-wait if more than one fh is choosen
  !       * Need to write sponge files
  !       * IO-server should take care of the Validation
  !       * Headers in restart and output files
  !       * Read restart (and sponge in the restart files)
  !       * Metread? 
  !       * wlfiles?
  !-----------------------------------------------------------------------------
  use cmod_mem, only : cmr1
  implicit none

  !- Public modules variables and arrays
  logical,    public, save :: io_server, in_hbm, io_debug, io_readrestart
  integer(4), public, save :: io_ntask, io_maxfh, io_maxfh_coupler

  !- Private modules variables and arrays
  integer(4), parameter, private :: maxiotasks = 5 
  
  !- codes for io_signals
  integer(4), parameter, private :: io_finish         = 1 ! prepare to die
  integer(4), parameter, private :: io_open           = 2 ! open file
  integer(4), parameter, private :: io_write          = 3 ! write output
  integer(4), parameter, private :: io_read           = 4 ! read input
  integer(4), parameter, private :: io_close          = 5 ! close file
  integer(4), parameter, private :: io_writerestart   = 6 ! write restart
  integer(4), parameter, private :: io_validate       = 7 ! write validation 
  integer(4), parameter, private :: io_open_c         = 8 ! open coupler file
  integer(4), parameter, private :: io_write_c        = 9  ! open coupler file
  integer(4), parameter, private :: io_close_c        = 10 ! open coupler file
  integer(4), parameter, private :: io_writerestart_c = 11 ! write coupler rest

  !- arrays for ouput information
  integer(4), allocatable, private, save :: output_fields(:,:), output_freq(:),&
                                            output_fh(:,:), output_dims(:,:),  &
                                            restart_fields(:), restart_dims(:),&
                                            restart_fh(:)
  character(len=20), allocatable, private, save :: output_fname(:)

  !- arrays for coupler ouput information
  integer(4), allocatable, private, save :: output_fields_coupler(:,:),        &
                                            output_freq_coupler(:),            &
                                            output_fh_coupler(:,:),            &
                                            output_dims_coupler(:,:),          &
                                            restart_fields_coupler(:),         &
                                            restart_dims_coupler(:),           &
                                            restart_fh_coupler(:)
  character(len=20), allocatable, private, save :: output_fname_coupler(:)

  !- bookkepping arrays for IO-server
  integer(4), allocatable, private, save :: io_rank(:)
  integer(4), allocatable, private, save :: bufsize2d(:,:), bufsize3d(:,:)
  integer(4), allocatable, private, save :: sbufsize2d(:), sbufsize3d(:)
  integer(4),              private, save :: hbm_lb, hbm_ub
  real(8),                 private, save :: msg_io(2)
  logical,                 private, save :: wait_msgsend       = .false.
  logical,                 private, save :: wait_outscatter    = .false.
  logical,                 private, save :: wait_restscatter   = .false.
  logical,                 private, save :: wait_valscatter    = .false.
  logical,                 private, save :: wait_outscatter_c  = .false.
  logical,                 private, save :: wait_restscatter_c = .false.

  !- Public routines 
  public  :: ioc_write, ioc_open, ioc_close, ioc_writerestart, ios_loop,       &
             ios_init, ios_usage, ioc_finalize, ioc_validate, ioc_readrestart, &
             ioc_open_coupler, ioc_close_coupler, ioc_write_coupler,           &
             ioc_writerestart_coupler
  !- Private routines
  private :: ios_write, ios_open, ios_close, ios_writerestart, split_domain,   &
             prep, ios_finalize, ios_validate, getmyIOtask
             
contains

!===============================================================================

  subroutine ios_loop(coupler_active)
    !---------------------------------------------------------------------------
    ! Main loop for the IO-tasks (Master and slaves), HBM-task will do nothing.
    ! The io task communicate with the HBM-tasks with the real array 'msg' 
    ! which contains: msg(1) = float version of io_signal
    !                 msg(2) = extra info (e.g. dtmain or aufdat or izeits)
    !---------------------------------------------------------------------------

    !- modules
    use dmi_mpi_global, only : iu06
    use io_subs,        only : flush_unit
    use exits,          only : exitme
    use io_mpi,         only : my_rank, check4assignments, alloc_HBM_arrays,   &
                               alloc_IOS_arrays, create_mpitypes

    !- directives
    implicit none

    !- arguments 
    logical, intent(in) :: coupler_active

    !- locals
    logical           :: running 
    real(8)           :: msg(2), dtmain
    integer(4)        :: myassignment, izeits
    character(len=40) :: restart_name

    running = .true.
    if (.not. io_server) return !HBM-tasks have nothing to do in here
 
    if (in_hbm .and. coupler_active) then !- HBM tasks and coupler active
      call alloc_HBM_arrays(output_fields, output_dims, io_maxfh,              &
                            output_fields_coupler, output_dims_coupler,        &
                            io_maxfh_coupler)
      return !HBM-tasks have nothing more to do in here

    elseif (in_hbm) then !- HBM tasks and coupler not active
      call alloc_HBM_arrays(output_fields, output_dims, io_maxfh)
      return !HBM-tasks have nothing more to do in here

    elseif (coupler_active) then !- IO tasks and coupler active
      call prep()
      call alloc_IOS_arrays(output_fields, output_dims, io_maxfh, hbm_lb,    &
                            hbm_ub, sbufsize2d, sbufsize3d,                  &
                            output_fields_coupler, output_dims_coupler,      &
                            io_maxfh_coupler)
    
      !- create MPI-types for permuting the physics input and output data 
      call create_mpitypes(output_dims, output_fields, restart_dims,         &
                           io_maxfh, hbm_lb, hbm_ub, coupler_active)
      
      !- create MPI-types for permuting the coupler input and output data 
      call create_mpitypes(output_dims_coupler, output_fields_coupler,       &
                           restart_dims_coupler, io_maxfh_coupler, hbm_lb,   &
                           hbm_ub, coupler_active, .true.)
    else !- IO tasks and coupler not active
      call prep()
      call alloc_IOS_arrays(output_fields, output_dims, io_maxfh, hbm_lb,    &
                            hbm_ub, sbufsize2d, sbufsize3d)
 
      call create_mpitypes(output_dims, output_fields, restart_dims,         &
                           io_maxfh, hbm_lb, hbm_ub, coupler_active)
 
    endif

    if (my_rank==0) then 
      INF_LOOP: do while (running)
        !- get mess and inform slave
        call check4assignments(msg,.true.)
        myassignment = int(msg(1))
        select case(myassignment)
          case (io_finish)
            write (iu06,*) 'IOserver terminate - goodbye cruel world...   '
            call flush_unit(iu06)
            call ios_finalize()
            running   = .false.
          case (io_open)
            write (iu06,*) 'Got an open signal ...'
            call flush_unit(iu06)
            dtmain = msg(2)
            call ios_open(dtmain, output_freq, io_maxfh, output_fname,         &
                          output_fh)
          case (io_write)
            write (iu06,*) 'Got a write signal ...'
            call flush_unit(iu06)
            izeits = int(msg(2))
            call ios_write(izeits, io_maxfh, output_dims, output_fields,       &
                           output_freq, output_fh)
          case (io_writerestart)
            write (iu06,*) 'Got a writerestart signal ...'
            call flush_unit(iu06)
            !- revive aufdat?
            !- bcast aufdat, ?
            restart_name = 'restart_out_'
            call ios_writerestart(restart_name, restart_fh, restart_fields,    &
                                  restart_dims)
          case (io_validate)
            write (iu06,*) 'Got a validation signal ...'
            call flush_unit(iu06)
            call ios_validate()
          case (io_close)
            write (iu06,*) 'Got a close signal ...'
            call flush_unit(iu06)
            call ios_close(io_maxfh, output_fh, output_fname)
          case (io_open_c)
            write (iu06,*) 'Got an open coupler signal ...'
            call flush_unit(iu06)
            dtmain = msg(2)
            call ios_open(dtmain, output_freq_coupler, io_maxfh_coupler,       &
                          output_fname_coupler, output_fh_coupler) 
          case (io_write_c)
            write (iu06,*) 'Got a write coupler signal ...'
            call flush_unit(iu06)
            izeits = int(msg(2))
            call ios_write(izeits, io_maxfh_coupler, output_dims_coupler,      &
                           output_fields_coupler, output_freq_coupler,         &
                           output_fh_coupler)
         case (io_writerestart_c)
            write (iu06,*) 'Got a write coupler restart signal ...'
            call flush_unit(iu06)
            !- revive aufdat?
            !- bcast aufdat, ?
            restart_name = 'restart_c_out_'
            call ios_writerestart(restart_name, restart_fh_coupler,            &
                                  restart_fields_coupler, restart_dims_coupler)
          case (io_close_c)
            write (iu06,*) 'Got a close coupler signal ...'
            call flush_unit(iu06)
            call ios_close(io_maxfh_coupler, output_fh_coupler,                &
                           output_fname_coupler)
        end select
      enddo INF_LOOP
    else 
      ! slave IO servers
      INF_LOOP2: do while (running)
        !- get mess from master IO
        call check4assignments(msg,.false.)
        myassignment = int(msg(1))
        select case(myassignment)
          case (io_finish)
            write (iu06,*) 'IOserver terminate - goodbye cruel world...   '
            call flush_unit(iu06)
            call ios_finalize()
            running   = .false.
          case (io_open)
            write (iu06,*) 'Got an open signal ...'
            call flush_unit(iu06)
            dtmain = msg(2)
            call ios_open(dtmain, output_freq, io_maxfh, output_fname,         &
                          output_fh)
          case (io_write)
            write (iu06,*) 'Got a write signal ...'
            call flush_unit(iu06)
            izeits = int(msg(2))
            call ios_write(izeits, io_maxfh, output_dims, output_fields,       &
                           output_freq, output_fh)
          case (io_writerestart)
            write (iu06,*) 'Got a writerestart signal ...'
            call flush_unit(iu06)
            !- bcast aufdat, ?
            restart_name = 'restart_out_'
            call ios_writerestart(restart_name, restart_fh, restart_fields,    &
                                  restart_dims)
          case (io_validate)
            write (iu06,*) 'Got a validation signal, but I am not doing anyting'
            call flush_unit(iu06)
          case (io_close)
            write (iu06,*) 'Got a close signal ...'
            call flush_unit(iu06)
            call ios_close(io_maxfh, output_fh, output_fname)
         case (io_open_c)
            write (iu06,*) 'Got an open coupler signal ...'
            call flush_unit(iu06)
            dtmain = msg(2)
            call ios_open(dtmain, output_freq_coupler, io_maxfh_coupler,       &
                          output_fname_coupler, output_fh_coupler)
          case (io_write_c)
            write (iu06,*) 'Got a write coupler signal ...'
            call flush_unit(iu06)
            izeits = int(msg(2))
            call ios_write(izeits, io_maxfh_coupler, output_dims_coupler,      &
                           output_fields_coupler, output_freq_coupler,         &
                           output_fh_coupler)
         case (io_writerestart_c)
            write (iu06,*) 'Got a write coupler restart signal ...'
            call flush_unit(iu06)
            !- revive aufdat?
            !- bcast aufdat, ?
            restart_name = 'restart_c_out_'
            call ios_writerestart(restart_name, restart_fh_coupler,            &
                                  restart_fields_coupler, restart_dims_coupler)
          case (io_close_c)
            write (iu06,*) 'Got a close coupler signal ...'
            call flush_unit(iu06)
            call ios_close(io_maxfh_coupler, output_fh_coupler,                &
                           output_fname_coupler)
          end select
      enddo INF_LOOP2
    endif!MASTER/SLAVE

    call exitme(0, 'END TIME IN IO' )
  end subroutine ios_loop

!===============================================================================

  subroutine ios_init()
    !---------------------------------------------------------------------------
    ! setting up MPI-communication
    !---------------------------------------------------------------------------
    use io_mpi,         only : set_comm_models, inhbm, my_iotask, my_rank
    use dmi_mpi,        only : mpi_size
    use dmi_mpi_global, only : iu06

    implicit none

    !- Let's split the tasks into two groups HBM and IO
    call set_comm_models(io_rank, io_server, io_ntask)
    in_hbm = inhbm ! get the value from io_mpi

    if (.not.io_server) return

    !- If multiple IO-tasks, split the HBM-tasks among them
    if (in_hbm) then
      !- Which IO-tasks should I send to
      call getmyIOtask(mpi_size,io_ntask,my_rank,my_iotask)
      write (iu06,'(a33,i3)') 'I will send my results to IO-task', my_iotask
    else ! IO-task
      call split_domain(mpi_size, io_ntask, my_rank, hbm_lb, hbm_ub)
      write (iu06,'(A65,I8,I8)')                                               &
         'local ios_server will communicate with HBM ranks in the interval:',  &
         hbm_lb,hbm_ub
      my_iotask = -1
    endif
  end subroutine ios_init

!===============================================================================

  subroutine ios_open(dtmain, freq, maxfh, fname, fh)
    !---------------------------------------------------------------------------
    ! Subroutine for opening the outputfiles
    !---------------------------------------------------------------------------
    use dmi_mpi_global, only : iu06
    use cmod_params,    only : narea
    use io_subs,        only : flush_unit
    use io_mpi,         only : iompi_openfile, iompi_setfileview, comm_io
    
    implicit none

    !- args
    real(8),      intent(in)    :: dtmain
    integer(4),   intent(in)    :: maxfh
    character(*), intent(in)    :: fname(:)
    integer(4),   intent(inout) :: fh(:,:), freq(:)

    !- locals
    integer(4)    :: i, ia
    character(2)  :: ctmp
    character(40) :: tdnam 

    do i=1,maxfh
      freq(i)=nint(freq(i)/dtmain)
      do ia=1,narea
        write(ctmp,'(i2.2)') ia
        tdnam = trim(fname(i))//'.'//ctmp
        write (iu06,*) 'Opening file ', tdnam
        write (iu06,*) 'File will have freq', freq(i)
        call flush_unit(iu06)
        !- open the file
        call iompi_openfile(comm_io, tdnam, fh(ia,i), .false.)
        !- set the fileview
        call iompi_setfileview(fh(ia,i), i, ia, .false.)
      enddo
    enddo
 
    write (iu06,*) 'IO server has opened ouput file(s)'
    write (iu06,*) 
  end subroutine ios_open

!===============================================================================

  subroutine ios_write(izeits, maxfh, dims, fields, freq, fh)
    !---------------------------------------------------------------------------
    ! IO tasks goes in here to write gather fields from HBM tasks and write
    ! outout.
    !---------------------------------------------------------------------------

    !- modules
    use dmi_mpi_global, only : iu06
    use cmod_params,    only : narea
    use io_subs,        only : flush_unit
    use io_mpi,         only : gather

    !- directives
    implicit none

    !- args
    integer(4), intent(in) :: izeits, maxfh, dims(:,:), fields(:,:), fh(:,:),  &
                              freq(:)

    !- locals
    integer(4) :: j, ia

    do j=1,maxfh
      if (mod(izeits,freq(j)) == 0) then
        do ia=1,narea
          !- gather and write outputvars
          call gather(ia, j, dims(:,j), fields(:,j), bufsize2d(:,ia),          &
                      bufsize3d(:,ia), fh(ia,j), hbm_lb, hbm_ub, io_debug)
        enddo
      endif
    enddo
    write (iu06,*) 'IO server has written output file(s)'
    write (iu06,*) 
    call flush_unit(iu06)
  end subroutine ios_write

!===============================================================================

  subroutine ios_writerestart(filename, fh, fields, dims)
    !- modules
    use dmi_mpi_global, only : iu06
    use cmod_params,    only : narea
    use io_mpi,         only : gather, iompi_closefile, iompi_openfile,        &
                               iompi_setfileview, comm_io

    !- directives
    implicit none

    !- arguments
    character(len=40), intent(in)    :: filename
    integer(4),        intent(inout) :: fh(:)
    integer(4),        intent(in)    :: fields(:), dims(:)

    !- locals
    integer(4)        :: ia, dummy
    character(len=40) :: fname
    character(len=2)  :: ch2

    !- open the restart file, set fileview, write and close
    do ia=1,narea
      if (ia < 10) then
        write(ch2,'(2i1)') 0,ia
      elseif (ia < 100) then
        write(ch2,'(i2)') ia
      else
        !FIXME: ERROR
      endif
      fname = filename
      fname = trim(fname)//ch2

      !- open file
      call iompi_openfile(comm_io, fname, fh(ia), .false.)
      !- set fileview
      dummy = 1
      call iompi_setfileview(fh(ia), dummy, ia, .true.)
      !- gather and write
      call gather(ia, dummy, dims(:), fields(:), bufsize2d(:,ia),              &
                  bufsize3d(:,ia), fh(ia), hbm_lb, hbm_ub, io_debug)
      !- close
      call iompi_closefile(fh(ia))
    enddo
 
    write (iu06,*) 'IO server has written restart file(s)'
    write (iu06,*) 
  end subroutine ios_writerestart

!===============================================================================

  subroutine ios_close(maxfh, fh, fname)
    use dmi_mpi_global, only : iu06
    use cmod_params,    only : narea
    use io_mpi,         only : iompi_closefile

    implicit none

    !- args
    integer(4),   intent(in)    :: maxfh
    integer(4),   intent(inout) :: fh(:,:)
    character(*), intent(in)    :: fname(:)

    !- local vars
    integer(4)    :: i, ia
    character(2)  :: ctmp
    character(40) :: tdnam

    do ia=1,narea
      write(ctmp,'(i2.2)') ia
      do i=1,maxfh
        tdnam = trim(fname(i))//'.'//ctmp
        write (iu06,*) 'Closing file ', tdnam
        !- close the file
        call iompi_closefile(fh(ia,i))
      enddo
    enddo
 
    write (iu06,*) 'IO server has closed ouput file(s)'
    write (iu06,*) 
  end subroutine ios_close

!===============================================================================

  subroutine ios_validate()
    use io_mpi, only : iompi_validate
    implicit none

    logical :: ldummy ! only used by HBM-tasks

    call iompi_validate(ldummy)
  end subroutine ios_validate

!===============================================================================

  subroutine ios_finalize()
    use io_mpi, only : iompi_cleanup
    implicit none

    call iompi_cleanup(io_maxfh, io_maxfh_coupler, io_server)
  end subroutine ios_finalize

!===============================================================================

  subroutine ioc_open()
    !---------------------------------------------------------------------------
    ! Master HBM-task goes in here
    !---------------------------------------------------------------------------
    use cmod_miscvars,  only : dtmain
    use io_mpi,         only : comm_inter, my_rank, ireq_msg, msgtag, iompi_wait
    use dmi_mpi_global, only : iu06
    use dmi_mpi,        only : dmpi_isend
    
    implicit none

    integer(4) :: i

    do i=1,io_maxfh
      output_freq(i)=nint(output_freq(i)/dtmain)
      if (io_debug) write (iu06,*) 'output_freq(i) ', i, output_freq(i)
    enddo

    ! from master (rank0 in HBM group) to master (rank0 in IO group)
    if (my_rank == 0) then
      if (wait_msgsend) call iompi_wait(ireq_msg)
     
      msg_io(1) = real(io_open,8)
      msg_io(2) = dtmain

      if (io_debug) write (iu06,*) 'ioc_open - sending io_open' 
      call dmpi_isend(msg_io,2,0,msgtag,ireq_msg,comm_inter)
      wait_msgsend = .true.
    endif
  end subroutine ioc_open

!===============================================================================

  subroutine ioc_open_coupler()
    !---------------------------------------------------------------------------
    ! Master HBM-task goes in here
    !---------------------------------------------------------------------------
    use cmod_miscvars,  only : dtmain
    use io_mpi,         only : comm_inter, my_rank, ireq_msg, msgtag, iompi_wait
    use dmi_mpi_global, only : iu06
    use io_subs,        only : flush_unit
    use dmi_mpi,        only : dmpi_isend
    
    implicit none

    integer(4) :: i

    do i=1,io_maxfh_coupler
      output_freq_coupler(i)=nint(output_freq_coupler(i)/dtmain)
      if (io_debug) then
        write (iu06,*) 'output_freq_coupler(i) ', i, output_freq_coupler(i)
        call flush_unit(iu06)
      endif
    enddo

    ! from master (rank0 in HBM group) to master (rank0 in IO group)
    if (my_rank == 0) then
      if (wait_msgsend) call iompi_wait(ireq_msg)
     
      msg_io(1) = real(io_open_c,8)
      msg_io(2) = dtmain

      if (io_debug) write (iu06,*) 'ioc_open_coupler - sending signal' 
      call dmpi_isend(msg_io,2,0,msgtag,ireq_msg,comm_inter)
      wait_msgsend = .true.
    endif
  end subroutine ioc_open_coupler

!===============================================================================
  
  subroutine ioc_readrestart(aufdat)
    !---------------------------------------------------------------------------
    ! HBM-tasks goes in here to read its part of the restart files
    !---------------------------------------------------------------------------
    use cmod_params,       only : narea, mm1, mmx, nmx,kmx, iw2, iw3 
    use io_mpi,            only : iompi_readrestart
    use exits,             only : exitme
    use dmi_mpi_global,    only : iu06
    use cmod_arrays,       only : kh, idx
    use cmod_local_arrays, only : u_l, v_l, z_l, cmp_l, casus_l, ice_l, ueis_l,&
                                  veis_l, tsnei_l, t_soil_l, tke_l, diss_l,    &
                                  avv_l
    
    implicit none

    !- arguments 
    character(11), intent(in) :: aufdat

    !- local variables
    integer(4) :: ia
    character(len=10) :: fname
    character(len=2)  :: ch2

    if (narea>99) then
      call exitme(99,'IO module cannot handle more than 99 areas')
      !ERROR
    endif

    do ia = 1,narea
      write(ch2,'(i2.2)') ia
      fname  = 'restart_'//ch2
      write(iu06,*) 'IO module: Reading restart file:', fname

      call iompi_readrestart(ia, fname, mm1(ia)%p, kh(ia)%p, mmx(ia), nmx(ia), &
                             kmx(ia), iw2(ia), iw3(ia), restart_dims, aufdat,  &
                             u_l(ia)%p, v_l(ia)%p, z_l(ia)%p, cmp_l(ia)%p,     &
                             casus_l(ia)%p, ice_l(ia)%p, tsnei_l(ia)%p,        &
                             ueis_l(ia)%p, veis_l(ia)%p, t_soil_l(ia)%p,       &
                             tke_l(ia)%p, diss_l(ia)%p, avv_l(ia)%p, idx(ia)%p)
    enddo
  end subroutine ioc_readrestart

!===============================================================================

  subroutine ioc_write(izeits)
    !---------------------------------------------------------------------------
    ! HBM-tasks goes in here to inform the IO-tasks its time to write. 
    ! It also sends selected variables to the IO-tasks
    !---------------------------------------------------------------------------
    !- modules
    use dmi_mpi_global, only : iu06
    use dmi_mpi,        only : mpi_rank, dmpi_isend, dd
    use cmod_params,    only : narea
    use io_subs,        only : flush_unit
    use io_mpi,         only : scatter, comm_inter, ireq_msg, msgtag,          &
                               iompi_wait, ireq_scatter

    !- directives
    implicit none

    !- args
    integer(4), intent(in) :: izeits

    !- locals
    integer(4) :: j, ia, n3dl
    logical    :: io_sendmsg

    io_sendmsg = .true.

    if (io_debug) write (iu06,*) 'ioc_write - izeits', izeits
    call flush_unit(iu06)

    do j=1,io_maxfh
      if (mod(izeits,output_freq(j)) == 0) then
        if (io_debug) write (iu06,*) 'ioc_write - match on izeitz for FH:', j

        if (mpi_rank==0 .and. (io_sendmsg)) then
          if (wait_msgsend) call iompi_wait(ireq_msg)
          
          msg_io(1) = real(io_write,8)
          msg_io(2) = real(izeits,8)
          if (io_debug) write (iu06,*) 'ioc_write - sending write signal'
          call flush_unit(iu06)
          call dmpi_isend(msg_io,2,0,msgtag,ireq_msg,comm_inter)
          wait_msgsend = .true.

          !- only send one msg to IO
          io_sendmsg = .false.
        endif
        
        do ia=1,narea
          n3dl = dd(ia)%nwet3
          !- wait for previous send to be complete if there is a previous send
          !  and if the domain is not dry. If it is dry, it doesn't send
          if (wait_outscatter .and. n3dl>0) then
            call iompi_wait( ireq_scatter(1)%p(ia) )
          endif

          call scatter(ia,output_fields(:,j),output_dims(:,j),io_debug,.false.)
        enddo
        wait_outscatter = .true.
      endif !freq matches izeits
    enddo
  end subroutine ioc_write

!===============================================================================

  subroutine ioc_write_coupler(izeits, ia, ctracers3, ctracers2)
    !---------------------------------------------------------------------------
    ! HBM-tasks goes in here to inform the IO-tasks its time to write. 
    ! It also sends selected variables to the IO-tasks
    !---------------------------------------------------------------------------
    !- modules
    use dmi_mpi_global, only : iu06
    use dmi_mpi,        only : mpi_rank, dmpi_isend, dd
    use cmod_params,    only : narea
    use io_subs,        only : flush_unit
    use io_mpi,         only : scatter, comm_inter, ireq_msg, msgtag,          &
                               iompi_wait, ireq_scatter

    !- directives
    implicit none

    !- args
    integer(4), intent(in) :: izeits, ia
    real(8),    intent(in) :: ctracers3(:,0:), ctracers2(:,0:)

    !- locals
    integer(4) :: j, n3dl
    logical    :: io_sendmsg

    !- only send message once (not for every domain)
    if (ia == 1) then
      io_sendmsg = .true.
    else
      io_sendmsg = .false.
    endif

    if (io_debug .and. ia==1)  then 
      write (iu06,*) 'ioc_write_coupler- izeits', izeits
      call flush_unit(iu06)
    endif

    do j=1,io_maxfh_coupler
      if (mod(izeits,output_freq_coupler(j)) == 0) then
        if (io_debug) then 
          write (iu06,*) 'ioc_write_coupler - match on izeitz for FH:', j
        endif

        if (mpi_rank==0 .and. (io_sendmsg)) then
          if (wait_msgsend) call iompi_wait(ireq_msg)
        
          msg_io(1) = real(io_write_c,8)
          msg_io(2) = real(izeits,8)
          if (io_debug) then 
            write (iu06,*) 'ioc_write_coupler - sending write coupler signal'
            call flush_unit(iu06)
          endif
          call dmpi_isend(msg_io,2,0,msgtag,ireq_msg,comm_inter)
          wait_msgsend = .true.

          !- only send one msg to IO
          io_sendmsg = .false.
        endif
      
        n3dl = dd(ia)%nwet3
        !- wait for previous send to be complete if there is a previous send
        !  and if the domain is not dry. If it is dry, it doesn't send
        if (wait_outscatter_c .and. n3dl>0) then
          call iompi_wait( ireq_scatter(3)%p(ia) )
        endif

        call scatter(ia,output_fields_coupler(:,j),output_dims_coupler(:,j), &
                     io_debug,.false., ctracers3, ctracers2)
        !- next time we have to wait for the isend
        if (ia == narea)  wait_outscatter_c = .true.
      endif
    enddo
  end subroutine ioc_write_coupler

!===============================================================================

  subroutine ioc_close()
    !---------------------------------------------------------------------------
    ! Just send close signal to the IO-tasks
    !---------------------------------------------------------------------------
    !- modules
    use io_mpi,         only : comm_inter, ireq_msg, msgtag, iompi_wait,       &
                               ireq_scatter
    use dmi_mpi,        only : mpi_rank, dmpi_isend, dd
    use dmi_mpi_global, only : iu06
    use cmod_params,    only : narea

    !- directives
    implicit none

    !- locals
    real(8), parameter :: unused = -1.0_8
    integer(4)         :: ia

    !- Wait for previous scatter send of output
    if (wait_outscatter) then
      do ia=1,narea
        if (dd(ia)%nwet3 <=0) cycle ! nothing was sended on this domain
        call iompi_wait( ireq_scatter(1)%p(ia) )
      enddo
    endif

    ! from master (rank0 in HBM group) to master (rank0 in IO group)
    if (mpi_rank == 0) then
      if (wait_msgsend) call iompi_wait(ireq_msg)
 
      msg_io(1) = real(io_close,8)
      msg_io(2) = unused
      if (io_debug) write (iu06,*) 'ioc_close - sending io_close'
      call dmpi_isend(msg_io,2,0,msgtag,ireq_msg,comm_inter)
      wait_msgsend = .true.
    endif
  end subroutine ioc_close

!===============================================================================

  subroutine ioc_close_coupler()
    !---------------------------------------------------------------------------
    ! Just send close signal to the IO-tasks
    !---------------------------------------------------------------------------
    !- modules
    use io_mpi,         only : comm_inter, ireq_msg, msgtag, iompi_wait,       &
                               ireq_scatter
    use dmi_mpi,        only : mpi_rank, dmpi_isend, dd
    use dmi_mpi_global, only : iu06
    use cmod_params,    only : narea

    !- directives
    implicit none

    !- locals
    real(8), parameter :: unused = -1.0_8
    integer(4)         :: ia

    !- Wait for previous scatter send of output
    if (wait_outscatter_c) then
      do ia=1,narea
        if (dd(ia)%nwet3 <=0) cycle ! nothing was sended on this domain
        call iompi_wait( ireq_scatter(3)%p(ia) )
      enddo
    endif

    ! from master (rank0 in HBM group) to master (rank0 in IO group)
    if (mpi_rank == 0) then
      if (wait_msgsend) call iompi_wait(ireq_msg)
 
      msg_io(1) = real(io_close_c,8)
      msg_io(2) = unused
      if (io_debug) write (iu06,*) 'ioc_close_coupler - sending signal'
      call dmpi_isend(msg_io,2,0,msgtag,ireq_msg,comm_inter)
      wait_msgsend = .true.
    endif
  end subroutine ioc_close_coupler

!===============================================================================

  subroutine ioc_writerestart(tim)
    !---------------------------------------------------------------------------
    ! HBM-tasks goes in here to inform the IO-tasks its time to write restart. 
    ! It also sends selected variables to the IO-tasks
    !---------------------------------------------------------------------------
    !- modules
    use dmi_mpi_global, only : iu06
    use dmi_mpi,        only : dmpi_isend, dd
    use cmod_params,    only : narea
    use io_subs,        only : flush_unit
    use io_mpi,         only : scatter, comm_inter, my_rank, ireq_msg, msgtag, &
                               iompi_wait, ireq_scatter

    !- directives
    implicit none

    !- args 
    real(8), intent(in) :: tim

    !- locals
    integer(4) :: ia, n3dl

    if (io_debug) write (iu06,*) 'ioc_writerestart' 
    call flush_unit(iu06)
    if (my_rank==0) then
      if (wait_msgsend) call iompi_wait(ireq_msg)

      msg_io(1) = real(io_writerestart,8)
      msg_io(2) = tim
      if (io_debug) then
        write (iu06,*) 'ioc_writerestart - sending writerestart signal'
        call flush_unit(iu06)
      endif
      call dmpi_isend(msg_io,2,0,msgtag,ireq_msg,comm_inter)
      wait_msgsend = .true.
    endif
        
    do ia=1,narea
      n3dl = dd(ia)%nwet3
      !- wait for previous send to be complete if there is a previous send
      !  and if the domain is not dry. If it is dry, it doesn't send
      if (wait_restscatter .and. n3dl>0) then
        call iompi_wait(ireq_scatter(2)%p(ia))
      endif
      call scatter(ia,restart_fields(:),restart_dims(:),io_debug, .true.)
    enddo
    wait_restscatter = .true.
  end subroutine ioc_writerestart

!===============================================================================

  subroutine ioc_writerestart_coupler(tim, ia)
    !---------------------------------------------------------------------------
    ! HBM-tasks goes in here to inform the IO-tasks its time to write coupler
    ! restart. It also sends selected variables to the IO-tasks
    !---------------------------------------------------------------------------
    !- modules
    use dmi_mpi_global, only : iu06
    use dmi_mpi,        only : dmpi_isend, dd
    use cmod_params,    only : narea
    use io_subs,        only : flush_unit
    use io_mpi,         only : scatter, comm_inter, my_rank, ireq_msg, msgtag, &
                               iompi_wait, ireq_scatter

    !- directives
    implicit none

    !- args 
    integer(4), intent(in) :: ia
    real(8),    intent(in) :: tim

    !- locals
    integer(4) :: n3dl
    logical    :: io_sendmsg

    if (ia == 1) then
      io_sendmsg = .true.
    else
      io_sendmsg = .false.
    endif

    if (io_debug .and. ia == 1) then
      write (iu06,*) 'ioc_writerestart_coupler' 
      call flush_unit(iu06)
    endif

    if (my_rank==0 .and. io_sendmsg) then
      if (wait_msgsend) call iompi_wait(ireq_msg)

      msg_io(1) = real(io_writerestart_c,8)
      msg_io(2) = tim
      if (io_debug) then
        write (iu06,*)                                                         &
                'ioc_writerestart_coupler - sending writerestart coupler signal'
        call flush_unit(iu06)
      endif
      call dmpi_isend(msg_io,2,0,msgtag,ireq_msg,comm_inter)
      wait_msgsend = .true.
    endif
        
    n3dl = dd(ia)%nwet3
    !- wait for previous send to be complete if there is a previous send
    !  and if the domain is not dry. If it is dry, it doesn't send
    if (wait_restscatter_c .and. n3dl>0) then
      call iompi_wait(ireq_scatter(4)%p(ia))
    endif
    call scatter(ia,restart_fields_coupler(:),restart_dims_coupler(:),         &
                 io_debug, .true.)

    if (ia == narea) wait_restscatter_c = .true.
  end subroutine ioc_writerestart_coupler

!===============================================================================

  subroutine ioc_validate()
    use dmi_mpi_global, only : iu06
    use dmi_mpi,        only : dmpi_isend
    use io_subs,        only : flush_unit
    use io_mpi,         only : comm_inter, my_rank, ireq_msg, msgtag,          &
                               iompi_wait, iompi_validate, ireq_val
 
    !- directives
    implicit none

    !- locals
    real(8), parameter :: unused = -1.0_8

    logical    :: IhaveSend

    if (io_debug) write (iu06,*) 'ioc_validate' 
    call flush_unit(iu06)
    if (my_rank==0) then
      if (wait_msgsend) call iompi_wait(ireq_msg)

      msg_io(1) = real(io_validate,8)
      msg_io(2) = unused
      if (io_debug) then
        write (iu06,*) 'ioc_validate - sending validate signal'
        call flush_unit(iu06)
      endif
      call dmpi_isend(msg_io,2,0,msgtag,ireq_msg,comm_inter)
      wait_msgsend = .true.
    endif

    !- wait for previous send to be complete if there is a previous send
    if (wait_valscatter) call iompi_wait( ireq_val )
    call iompi_validate(IhaveSend)
    
    if (IhaveSend)  wait_valscatter = .true. 
  end subroutine ioc_validate

!===============================================================================

 subroutine ioc_finalize()
    !---------------------------------------------------------------------------
    ! Cleanup MPI and send finish signal to IO server
    !---------------------------------------------------------------------------
    use dmi_mpi,        only : dmpi_send, dd
    use io_mpi,         only : my_rank, comm_inter, iompi_cleanup, msgtag,     &
                               ireq_msg, iompi_wait, ireq_scatter, ireq_val
    use dmi_mpi_global, only : iu06
    use cmod_params,    only : narea
    implicit none

    !- local vars and parameter
    real(8), parameter :: unused = -1.0_8

    integer(4) :: ia

    if (.not.io_server) then
      call iompi_cleanup(0, 0, io_server)
      return
    endif

    !- Wait for previous scatter restart send
    if (wait_restscatter) then
      do ia=1,narea
        if (dd(ia)%nwet3 <= 0) cycle ! no need to wait 'cause we haven't sended
        call iompi_wait(ireq_scatter(2)%p(ia))
      enddo
      wait_restscatter = .false.
    endif
    if(wait_valscatter) call iompi_wait( ireq_val )

    if (my_rank==0) then
      if (wait_msgsend) call iompi_wait(ireq_msg)

      if (io_debug) write (iu06,*) 'ioc_finalize - sending io_finish'
      !- blocking call so we know IO is ready to finish
      msg_io(1) = real(io_finish,8)
      msg_io(2) = unused
      call dmpi_send(msg_io,2,0,msgtag,comm_inter)
      wait_msgsend = .true.
    else
      if (io_debug) write (iu06,*) 'ioc_finalize - slave model'
    endif
    
    call iompi_cleanup(io_maxfh, io_maxfh_coupler, io_server)
  end subroutine ioc_finalize

!===============================================================================

  subroutine ios_usage(iu06,coupler_active, coupler_3d, coupler_2d, ldefault)
    ! modules used
    use io_subs,     only : flush_unit
    use io_mpi,      only : color, first_model_task
    use io_miscsubs, only : f_tempdat, f_coupler,                              &
                            max_io_fields_def, set_predefined, set_restart
    use dmi_mpi,     only : mpi_rank, mpi_size
    use exits,       only : exitme

    implicit none

    !- Arguments
    integer(4), intent(in)           :: iu06
    logical,    intent(in)           :: coupler_active
    integer(4), intent(in)           :: coupler_3d, coupler_2d
    logical,    intent(in), optional :: ldefault
    
    !- Local variables and params
    character(5), parameter :: w = '(a57)'
    integer(4),  parameter :: uninitialized=-99999
    integer(4)    :: i, ios, j, f
    integer(4)    :: io_rank_in(maxiotasks)
    integer(4)    :: fields_in(max_io_fields_def), freq_in
    character(10) :: filename
   
    !- MAybe all these should go in module header instead
    namelist /iospec_general/     io_server, io_ntask, io_maxfh, io_debug,     &
                                  io_readrestart, io_maxfh_coupler
    namelist /iospec_rank/        io_rank_in
    namelist /iospec_out/         filename, fields_in,freq_in
    namelist /iospec_out_coupler/ filename, fields_in,freq_in

    !-----Defaults -------------------------------------------------------------
    io_server        = .false.
    io_ntask         = 0
    io_debug         = .false.
    io_readrestart   = .true.
    io_maxfh         = 1
    io_maxfh_coupler = 0
    do i = 1, maxiotasks 
      io_rank_in(i) = uninitialized 
    enddo
 
    if (present(ldefault)) then
      return
    endif

    !- open io.nml and read general specs
    open (unit=10, file='io.nml', status='old', iostat=ios)
    if (ios /= 0) then
      call exitme(ios,'Error: Cannot open file io.nml...aborting')
    endif
      
    read (unit=10, nml=iospec_general, iostat=ios)
    if (ios /= 0) then
      call exitme(ios,                                                         &
      'Error: Cannot read iospec_general in io.nml...aborting')
    endif

    if (io_server) then
      if (io_ntask>mpi_size-io_ntask) then
        write (iu06,'(a53)')                                                   &
                         'Error - More IO-tasks than HBM-tasks is not permitted'
        write (iu06,'(a10,i4)') '# HBM-task', mpi_size-io_ntask
        write (iu06,'(a9,i4)') '# IO-task', io_ntask 
        call flush_unit(iu06)
        call exitme(ios,'IO Error')
      endif
      if (io_ntask>maxiotasks) then
        call exitme(ios,'To many IO-tasks requested')
      endif
      ! default is one IO server
      if (io_ntask<=0) then
        write (iu06,w) 'Warning - Zero IO tasks specified, using default=1   '
        io_ntask=1
        io_rank_in(1)=mpi_size-1 ! default is to use last mpi task
      else
        !- read iospec
        read (unit=10, nml=iospec_rank, iostat=ios)
        if (ios /= 0) then
          write(iu06,w)'Warning: Cannot read iospec_rank in io.nml file ...  '
          write(iu06,w)'         ... aborting                                '
          call exitme(ios,'IO Error')
        endif
      endif
      if (io_maxfh <= 0) then
        call exitme(ios,'io_maxfh must be larger than zero in io.nml')
      endif
      if (io_maxfh > 1) then
        call exitme(ios,'Multiple output files is not implemented yet')
      endif

      !- allocate arrays for holding io-tasks specs   
      allocate(io_rank(io_ntask))
      !- allocate arrays for holding output information
      allocate (output_freq(io_maxfh), output_fname(io_maxfh),                 &
                output_dims(max_io_fields_def,io_maxfh),                       &
                output_fields(max_io_fields_def,io_maxfh))
      output_freq(:)     = uninitialized
      output_fields(:,:) = uninitialized
      output_dims(:,:)   = uninitialized
      output_fname(:)    = ' '
      
      !- allocate arrays for restart output
      i = max_io_fields_def 
      allocate(restart_dims(i), restart_fields(i) ) 
      restart_fields(:) = uninitialized
      restart_dims(:)   = uninitialized

      !- read iospec out for every fh (fields_in, freq_in, filename)
      !- fields in is save to io_flist()
      do i=1,io_maxfh
        filename     = 'default'
        freq_in      = uninitialized
        fields_in(:) = uninitialized
        read (unit=10, nml=iospec_out, iostat=ios)
        if (ios /= 0) then
          call exitme(ios,'Error: Cannot read iospec_out in io.nml')
        endif
        if (freq_in > 0) then
          output_freq(i)=freq_in ! adjust with dtmain in ioc_open
        else
          call exitme(5, 'Error reading output_freq from iospec_out')
        endif
        output_fname(i)=filename

        !- read what fields to handle

        !- first check if a predefinded schema is chooses
        if (fields_in(1) == f_tempdat) then 
          call set_predefined(fields_in(1), output_fields(:,i),output_dims(:,i))
        else ! read specified fields
          do j=1,max_io_fields_def
            f = fields_in(j)
            if (f <= -1) exit
            
            output_fields(j,i) = f
            
            ! 2d or 3d?
            if (f >= 200 .and. f<300) then
              output_dims(j,i)   = 2
            elseif (f >= 300 .and. f<400) then
              output_dims(j,i)   = 3
            else
              write(iu06,*) '  variable number does not exis or collective'
              write(iu06,*) '  number mixed with single variable specs'
              call flush_unit(iu06)
              call exitme(ios,'Error reading iospec_out in io.nml')
            endif          
          enddo
        endif
      enddo

      !- pre-defined restart output info
      call set_restart(restart_fields(:), restart_dims(:))

      !- coupler stuff
      if (coupler_active) then
        if (io_maxfh_coupler <= 0) then
          call exitme(ios,'io_maxfh_coupler must be larger than zero in io.nml')
        endif
 
        !- This is due to the namings of the variables in io_miscsubs.f90
        !  The coupler variables would be mixed up with the physics vars in
        !  the subroutine select_scatter_vars
        if (coupler_3d + coupler_2d > 99) then
          call exitme(1,'IO Error: too many coupler tracers for IO-server')
        endif

        ! the output_fields should be specified as by tracers number and the 
        ! 2d fields is speficied as the number above these
        ! i.e. 
        ! if tracers 1 and 5 should be outputted as well as 2d var 2, then
        ! the io.nml nml=iospec_out shuold be specified as:
        ! fields_in = 1, 5, <number_3d_tracsers> + 2

        !- allocate arrays for holding output information
        i = io_maxfh_coupler
        allocate (output_freq_coupler(i),output_fname_coupler(i),              &
                  output_dims_coupler(max_io_fields_def,i),                    &
                  output_fields_coupler(max_io_fields_def,i))
        output_freq_coupler(:)     = uninitialized
        output_fields_coupler(:,:) = uninitialized
        output_dims_coupler(:,:)   = uninitialized
        output_fname_coupler(:)    = ' '
        
        !- allocate arrays for restart output
        i = max_io_fields_def 
        allocate(restart_dims_coupler(i), restart_fields_coupler(i) )
        restart_fields_coupler(:) = uninitialized
        restart_dims_coupler(:)   = uninitialized

        do i=1,io_maxfh_coupler
          filename     = 'default_coupler'
          freq_in      = uninitialized
          fields_in(:) = uninitialized
          read (unit=10, nml=iospec_out_coupler, iostat=ios)
          if (ios /= 0) then
            call exitme(ios,                                                   &
                 'Error: Cannot read iospec_out_coupeler for coupler in io.nml')
          endif
          if (freq_in > 0) then
            output_freq_coupler(i)=freq_in ! adjust with dtmain in ioc_open
          else
            call exitme(5,                                                     &
            'Error reading output_freq_coupler from iospec_out_coupler')
          endif
          output_fname_coupler(i)=filename
  
          !- read what fields to handle
  
          !- first check if a predefinded schema is chooses
          if (fields_in(1) == f_coupler) then
            call set_predefined(fields_in(1), output_fields_coupler(:,i),      &
                           output_dims_coupler(:,i), coupler_2d, coupler_3d)
          else ! read specified fields
            do j=1,max_io_fields_def
              f = fields_in(j)
              if (f < 0) exit
              
              output_fields_coupler(j,i) = f
              
              ! 2d or 3d?
              if (f <= coupler_3d) then
                output_dims_coupler(j,i)   = 3
              elseif (f <= coupler_2d + coupler_3d) then
                output_dims_coupler(j,i)   = 2
              else
                write(iu06,*) ' To many coupler output variables specified'
                call flush_unit(iu06)
                call exitme(ios,'Error reading iospec_out in io.nml')
              endif          
            enddo
          endif
        enddo
        !- pre-defined restart output info for the coupler
        call set_restart(restart_fields_coupler(:), restart_dims_coupler(:),   &
                         coupler_2d, coupler_3d)
      endif
      close (unit=10)


      !- bookkeeping on iotasks
      do i=1,io_ntask
        if (io_rank_in(i) >= 0) then
          if (io_rank_in(i) < mpi_size .and. io_rank_in(i) >= 0) then
            io_rank(i)=io_rank_in(i)
          else
            call exitme(1,'ERROR - IO rank outside rank range - aborting')
          endif
        else
          ! default
          io_rank(i)=mpi_size-i
          write(iu06,*) 'WARNING: using default value for io_rank...'
        endif
      enddo
      color=0 ! default is a model_task
      do i=1,io_ntask
        if (mpi_rank == io_rank(i)) then 
          color=1
        endif
      enddo
      write(iu06,'(a13,i5,a16)') 'Running with ',io_ntask,' IO MPI task(s)'
      write(iu06,'(a13,i5,a17)') 'Running with ',mpi_size-io_ntask,            &
                                 ' HBM MPI task(s)'
      if (color==0) then
        write(iu06,'(a34)') 'I am part of the team of HBM tasks'
      else
        write(iu06,'(a33)') 'I am part of the team of IO tasks'
      endif
      first_model_task=0
      do i=1,io_ntask
        if (io_rank(i)==first_model_task) then
          first_model_task=first_model_task+1
        endif
      enddo
    else !if not IO-server
      write (iu06,w) 'Running without IO servers                            '
    endif
  end subroutine ios_usage

!===============================================================================

  subroutine split_domain(ds,dm,iam,low,up)
    implicit none

    !- Args
    integer(4), intent(in)  :: ds,dm,iam
    integer(4), intent(out) :: low,up

    !- locals
    real(8) :: dlen, riam

    dlen = real(ds,8)
    riam = real(iam,8)
    low  = floor((riam*dlen+0.5_8)/dm)
    up   = -1 + floor((riam*dlen+dlen+0.5_8)/dm)
  end subroutine split_domain

!===============================================================================

  subroutine getmyIOtask(ds,dm,iam,myio)
    ! THIS is a HACK to prevent the HBM-task of waiting fot the IO-task to 
    ! finish its "split_domain" routine
    implicit none

    !- Args
    integer(4), intent(in)  :: ds,dm,iam
    integer(4), intent(out) :: myio

    !- locals
    integer(4) :: i, low, up

    do i=0,io_ntask-0
      call split_domain(ds,dm,i, low, up)
      if (iam >= low .and. iam<=up) then
        myio = i
      endif
    enddo
  end subroutine getmyIOtask

!===============================================================================

  subroutine prep()
    use io_mpi,         only : my_rank
    use cmod_params,    only : narea
    use dmi_mpi_global, only : iu06
    use dmi_mpi,        only : mpi_tt
    use io_subs,        only : flush_unit

    implicit none
   
    !- local params and vars
    integer(4) :: it, ia, itp1, n2d, numhbm

    numhbm = hbm_ub-hbm_lb+1

    !- allocate some work arrays for the IO-tasks
    allocate(output_fh(narea,io_maxfh))
    allocate(restart_fh(narea))
    if (io_maxfh_coupler > 0) then
      ! coupler is active and should do output
      allocate(output_fh_coupler(narea,io_maxfh_coupler))
      allocate(restart_fh_coupler(narea))
    endif
   
    ! Allocate buffers and related arrays
    allocate (bufsize2d(1:numhbm,narea), bufsize3d(1:numhbm,narea))
    ! sum buffersizes, i.e. size of the global buffer for each IO task
    ! basically the block lengths bl(:,:)
    allocate(sbufsize2d(narea),sbufsize3d(narea))
 
    ! - Set the bufsizes for each HBM-tasks
    write (iu06,'(A32)') 'ios_server bufsize diagnostics: '
    do ia=1,narea
      do it=1,numhbm
        itp1 = hbm_lb + it
        bufsize2d(it,ia) = 0
        bufsize3d(it,ia) = 0
        if (mpi_tt(ia,itp1)%nwet3 > 0) then
          bufsize3d(it,ia) = mpi_tt(ia,itp1)%nwet3
          n2d = mpi_tt(ia,itp1)%up_ws - mpi_tt(ia,itp1)%low_ws + 1
          bufsize2d(it,ia) = n2d
        else
          bufsize3d(it,ia) = 0
          bufsize2d(it,ia) = 0
        endif
        write (iu06,'(5X,A36,I8,I8,I8,I8)')                                    &
                    'area/hbm_rank/bufsize2d/bufsize3d:',                      &
                     ia, it, bufsize2d(it,ia), bufsize3d(it,ia)
      enddo
    enddo   
  
    !- Set the total buffsizes for all HBM-tasks
    write (iu06,'(A36)') 'ios_server block-length diagnostics: '
    do ia=1,narea
      sbufsize2d(ia)=0
      sbufsize3d(ia)=0
      do it=1,numhbm
        sbufsize2d(ia) = sbufsize2d(ia)+bufsize2d(it,ia) 
        sbufsize3d(ia) = sbufsize3d(ia)+bufsize3d(it,ia) 
      enddo
      write (iu06,'(5X,A36,I8,I8,I8,I8)')                                      &
                  'ios/area/blocklenght2d/blocklenght3d:',                     &
                   my_rank,ia,sbufsize2d(ia),sbufsize3d(ia)
    enddo
    write (iu06,*) 

    call flush_unit(iu06)
  end subroutine prep

end module io_control
