module io_mpi
  !-----------------------------------------------------------------------------
  !
  !-----------------------------------------------------------------------------
  !- modules 
#if defined (MPI)
  use mpi
#endif
  use cmod_mem, only : cmr1, cmi1, cmi3

  implicit none

  !- Some definitions
  integer(4), parameter, public :: msgtag         = 100
  integer(4), parameter, public :: scattag        = 200
  integer(4), parameter, public :: valtag         = 300

  !- Public MPI related variables
  logical,    public, save      :: inhbm
  integer(4), public, save      :: my_rank, my_size, color, comm_model, comm_io
  integer(4), public, save      :: comm_inter, first_model_task
  integer(4), public, save      :: ireq_msg, ireq_val, my_iotask
  type(cmi1), allocatable, public, save :: ireq_scatter(:)

  !- Private module variables and parameters
  integer(4), private, allocatable, save :: iw2io(:), iw3io(:)
  !- Extra communicators
  integer(4), private, save :: comm_all, comm_dupworld

  !- more MPI variables
  integer(4), allocatable, private, save :: filetype(:,:), restart_filetype(:)
  integer(4), allocatable, private, save :: filetype_coupler(:,:)
  integer(4), allocatable, private, save :: restart_filetype_coupler(:)
  real(8),    allocatable, private, save :: rbuf(:), wbuf(:)
  type(cmi3), allocatable, private, save :: mo(:), mop(:)
  type(cmr1), allocatable, private, save :: tmpbuf(:,:)
  integer(4), allocatable, private, save :: n2dt(:), n3dlb(:), n3ds(:),        &
                                            n3dlen(:), n3dub(:)
  integer(4), allocatable, private, save :: wrk1(:), wrk2(:)
 
  !- Public routines
  public  :: set_comm_models, create_mpitypes, gather, scatter,                &
             check4assignments, iompi_openfile, iompi_closefile,               &
             iompi_setfileview, iompi_cleanup, iompi_validate,                 &
             iompi_readrestart, alloc_HBM_arrays, alloc_IOS_arrays

  !- Private routines
  private :: check_mpi_err
contains

#if defined (MPI)
!===============================================================================
  
  subroutine iompi_readrestart(ia, fname, mm1, kh, mmx, nmx, kmx, n2d, n3d,    &
                               vardim,aufdat, u, v, z, cmp, casus, ice, tsnei, &
                               ueis, veis, t_soil, tke, diss, avv, ompidx,     &
                               coupler2d)
    !- modules
    use mpi
    use dmi_mpi,        only : dd
    use cmod_params,    only : narea
    use exits,          only : exitme
    use dmi_mpi_global, only : iu06
    use dmi_omp,        only : domp_get_domain
    use io_miscsubs,    only : num_rfields3, num_rfields2,                     &
                               num_rfields3_coupler, num_rfields2_coupler
    
    implicit none

    !- arguments
    character(len=11), intent(in)    :: aufdat
    character(len=*),  intent(in)    :: fname
    integer(4),        intent(in)    :: ia, mmx, nmx, kmx, n2d, n3d
    integer(4),        intent(in)    :: vardim(:), mm1(0:,0:,:), kh(0:)
    real(8),           intent(out)   :: u(0:), v(0:), z(0:), tke(0:), diss(0:),&
                                        avv(0:), ueis(0:), veis(0:), ice(:,0:),&
                                        t_soil(:,:), tsnei(0:), cmp(:,0:)
    logical,           intent(out)   :: casus(0:)
    integer(4),        intent(inout) :: ompidx(:,:)
    real(8), optional, intent(out)   :: coupler2d(:,:)

    !- local variables
    real(8), parameter            :: land = 0.0_8

    integer(4) :: ial, i, j, k, ms, ns, ibuf, ibuf2, kb, jlh, jhh, ilh, ihh,   &
                  jl, jh, il,ih, buflen2d, buflen3d, up, low, nvar, ivar,      &
                  varoffset, fhandle, buflen, ftype, mi, ihalo2, ihalo3,       &
                  nhalo2, nhalo3, n2d_l, n3d_l, rfields2, rfields3, ncoupler2d
    real(8)    :: rdum
    logical    :: do_coupler
    
    integer(KIND=MPI_OFFSET_KIND) :: disp
    integer(4)                    :: mpistat(MPI_STATUS_SIZE)

    !- work arrays
    integer(4), allocatable, save :: kh_wrk(:), rbuf2local(:)

    !-------------------
    ! initialization and allcation
    !-------------------
    disp = 0

    if (present(coupler2d)) then
      do_coupler = .true.
      ncoupler2d = size(coupler2d)
      rfields2   = num_rfields2_coupler
      rfields3   = num_rfields3_coupler
    else
      do_coupler = .false.
      rfields2   = num_rfields2
      rfields3   = num_rfields3
    endif

    ! first time allocate enough buflength
    if (ia == 1) then
      buflen2d = 0
      buflen3d = 0
      do ial=1,narea
        buflen2d = max(buflen2d,dd(ial)%up_ws-dd(ial)%low_ws+1+dd(ial)%halo2)
        buflen3d = max(buflen3d, dd(ial)%nwet3 + dd(ial)%halo3)
      enddo
      !- Allocate work arrays
      buflen = buflen2d*rfields2 + buflen3d*rfields3
      allocate(wrk1(1:buflen), wrk2(1:buflen), rbuf2local(1:buflen),           &
               rbuf(1:buflen))
      !- need to store the permuted kh
      allocate(kh_wrk(1:buflen2d))
      wrk2(:) = 1
    endif

    !- some task information
    n2d_l  = dd(ia)%up_ws - dd(ia)%low_ws + 1
    n3d_l  = dd(ia)%nwet3
    nhalo2 = dd(ia)%halo2
    nhalo3 = dd(ia)%halo3

    if (n3d_l <= 0) then
      ! HACK!
      ! open, close, read, and fileview are collective calls for all hbm-tasks,
      ! so we still have to do this for an empty decompo. Otherwhise we 
      ! have to make a new communicator. 
      ! So, lets just let these tasks read a dummy real.

      call iompi_openfile(comm_model, fname, fhandle, .true. )
      call MPI_FILE_SET_VIEW(fhandle, disp,MPI_REAL8,MPI_REAL8,'native',       &
                             MPI_INFO_NULL, i)
      call check_mpi_err(i)
      call MPI_FILE_READ_ALL(fhandle,rdum,1,MPI_REAL8,mpistat,i)
      call check_mpi_err(i)
      call iompi_closefile(fhandle)

      if (ia==narea) then
        !also deallocate the work arrays before leaving
        deallocate(wrk1, kh_wrk, rbuf2local, wrk2, rbuf)
      endif
      return
    endif

    !- this tasks buffer lengths
    buflen2d = n2d_l + nhalo2
    buflen3d = n3d_l + nhalo3 
    buflen   = buflen2d*rfields2 + buflen3d*rfields3
 
    !- some index limits for the mpi-decompo
    !- halo limits
    jlh = dd(ia)%low_hj
    jhh = dd(ia)%up_hj
    ilh = dd(ia)%low_hi
    ihh = dd(ia)%up_hi
    !- domain limits
    jl = dd(ia)%low_j
    jh = dd(ia)%up_j
    il = dd(ia)%low_i
    ih = dd(ia)%up_i
 
    !-------------------
    ! Create index array for the fileview 
    ! (which parts of the file should this task read)
    !-------------------
    ibuf  = 0
    ns    = n2d
    ibuf2 = buflen2d
    do j=1,nmx
      do i=1, mmx
        mi = mm1(i,j,1)
        if (.not. mi > 0) cycle
        
        kb = kh(mi)
        
        !- Is this grid point outside the mpi-decompo for this task?
        if (j<jlh .or. j>jhh .or. i<ilh .or. i>ihh) then
          ns = ns + kb - 1 ! keep track of the subsurface grid points
          cycle
        endif
        
        ibuf = ibuf + 1
        wrk1(ibuf) = mi - 1
        do k=2,kb
          ns = ns + 1
          ibuf2 = ibuf2 + 1
          wrk1(ibuf2) = ns - 1
        enddo
      enddo
    enddo

    !- condiser filedisplay for all variables
    nvar = rfields2+rfields3
    !- first 3d variable (u) is already set
    low  = buflen3d+1
    varoffset = n3d

    do ivar = 2,nvar
      if (vardim(ivar) == 2) then
        up = low + buflen2d - 1
        wrk1(low:up) = wrk1(1:buflen2d) + varoffset
        varoffset = varoffset + n2d
      else
        up = low + buflen3d -1
        wrk1(low:up) = wrk1(1:buflen3d) + varoffset
        varoffset = varoffset + n3d
      endif
      low = up + 1
    enddo

    !-------------------
    ! Commit MPI filetype
    !-------------------
    call MPI_TYPE_INDEXED(buflen, wrk2(1:buflen), wrk1(1:buflen),              &
                          MPI_REAL8, ftype, i)
    call check_mpi_err(i)
    call MPI_TYPE_COMMIT(ftype,i)
    call check_mpi_err(i)

    !-------------------
    ! open file and set fileview
    !-------------------
    call iompi_openfile(comm_model, fname, fhandle, .true. )
    call MPI_FILE_SET_VIEW(fhandle, disp, MPI_REAL8, ftype, 'native',          &
                           MPI_INFO_NULL, i)
    call check_mpi_err(i)

    !-------------------
    ! read file and close the file
    !-------------------
    call MPI_FILE_READ_ALL(fhandle,rbuf(1:buflen),buflen,MPI_REAL8,mpistat,i)
    call check_mpi_err(i)
    call iompi_closefile(fhandle)
 
    call MPI_GET_COUNT(mpistat, MPI_REAL8, j, i)
    if (j /= buflen) then
      write(iu06,*) 'input buffer does not match array size'
      call exitme(66,'Cannot read restartfile')
    endif
    
    !-------------------
    ! Make a index array of where to put the buffer in the local data
    ! (will overwrite parts of the array wrk in this loop)
    !-------------------
    ms     = 0
    ibuf   = 0
    ns     = n2d_l     !excluding halo
    ibuf2  = buflen2d  !including halo
    ihalo2 = n3d_l
    ihalo3 = n3d_l + nhalo2
    do j=jlh,jhh
      do i=ilh, ihh
        mi = mm1(i,j,1)
        if (mi == 0) cycle
        kb = kh(mi)

        ibuf = ibuf + 1

        if (j>jh .or. j<jl .or. i>ih .or. i<il) then !it's a halo
          ihalo2 = ihalo2 + 1
          rbuf2local(ihalo2) = ibuf
          do k=2,kb
            ihalo3 = ihalo3 + 1
            ibuf2  = ibuf2 + 1
            rbuf2local(ihalo3) = ibuf2
          enddo

          cycle
        endif

        ms = ms + 1
        rbuf2local(ms) = ibuf
        kh_wrk(ms) = kb ! insted of kh which is not permuted
        wrk1(ms)    = ns ! store where the subsurface index start
        do k=2,kb
          ns    = ns + 1
          ibuf2 = ibuf2 + 1
          rbuf2local(ns) = ibuf2
        enddo      
      enddo
    enddo
 
    !-------------------
    ! Distribute readbuffer to local arrays  (threaded)
    ! (will also serve as a NUMA first touch)
    !-------------------

    !- land values
    u(0)        = land
    v(0)        = land 
    z(0)        = land
    cmp(2,0)    = land
    cmp(1,0)    = land
    rdum        = land
    casus(0)    = .true.
    ice(:,0)    = land 
    tsnei(0)    = land
    ueis(0)     = land
    veis(0)     = land
    tke(0)      = land
    diss(0)     = land
    avv(0)      = land
 
!$OMP PARALLEL DEFAULT (shared) PRIVATE(i,j,k,kb,low,up,rdum,ihalo2)
    call domp_get_domain(kh_wrk, 1, n2d_l, low, up, ompidx)

    ! surface data
    do i = low,up
      u(i)        = rbuf( rbuf2local(i) )
      v(i)        = rbuf( rbuf2local(i) + buflen3d)
      z(i)        = rbuf( rbuf2local(i) + 2*buflen3d)
      cmp(2,i)    = rbuf( rbuf2local(i) + 2*buflen3d+buflen2d)
      cmp(1,i)    = rbuf( rbuf2local(i) + 3*buflen3d+buflen2d)
      rdum        = rbuf( rbuf2local(i) + 4*buflen3d+buflen2d)
      if (rdum > 0) then
        casus(i)    = .true.
      else
        casus(i)    = .false.
      endif
      ice(1,i)    = rbuf( rbuf2local(i) + 4*buflen3d+2*buflen2d)
      ice(2,i)    = rbuf( rbuf2local(i) + 4*buflen3d+3*buflen2d)
      ice(3,i)    = rbuf( rbuf2local(i) + 4*buflen3d+4*buflen2d)
      tsnei(i)    = rbuf( rbuf2local(i) + 4*buflen3d+5*buflen2d)
      ueis(i)     = rbuf( rbuf2local(i) + 4*buflen3d+6*buflen2d)
      veis(i)     = rbuf( rbuf2local(i) + 4*buflen3d+7*buflen2d)
      t_soil(1,i) = rbuf( rbuf2local(i) + 4*buflen3d+8*buflen2d)
      t_soil(2,i) = rbuf( rbuf2local(i) + 4*buflen3d+9*buflen2d)
      t_soil(3,i) = rbuf( rbuf2local(i) + 4*buflen3d+10*buflen2d)
      tke(i)      = rbuf( rbuf2local(i) + 4*buflen3d+11*buflen2d)
      diss(i)     = rbuf( rbuf2local(i) + 5*buflen3d+11*buflen2d)
      avv(i)      = rbuf( rbuf2local(i) + 6*buflen3d+11*buflen2d)
    enddo
 
    !- subsurface data
    do i = low,up
      j = wrk1(i)
      kb = kh_wrk(i)
      do k=2,kb
        j = j + 1
        u(j+nhalo2)     = rbuf( rbuf2local(j) )
        v(j+nhalo2)     = rbuf( rbuf2local(j) + buflen3d)
        cmp(2,j+nhalo2) = rbuf( rbuf2local(j) + 2*buflen3d+buflen2d)
        cmp(1,j+nhalo2) = rbuf( rbuf2local(j) + 3*buflen3d+buflen2d)
        tke(j+nhalo2)   = rbuf( rbuf2local(j) + 4*buflen3d+11*buflen2d)
        diss(j+nhalo2)  = rbuf( rbuf2local(j) + 5*buflen3d+11*buflen2d)
        avv(j+nhalo2)   = rbuf( rbuf2local(j) + 6*buflen3d+11*buflen2d)
      enddo
    enddo

    ! the local array index should go from n2d_l+1 to n2d_l+nhalo2 and the
    ! rbuf2local index from n3d_l + 1 to n3d_l + nhalo2
    call domp_get_domain(n3d_l+1,n3d_l+nhalo2,low,up)
    j = n2d_l + low - (n3d_l+1)
    
    !- distribute halo2d
    do i=low,up
      j = j+1
      u(j)        = rbuf( rbuf2local(i) )
      v(j)        = rbuf( rbuf2local(i) + buflen3d)
      z(j)        = rbuf( rbuf2local(i) + 2*buflen3d)
      cmp(2,j)    = rbuf( rbuf2local(i) + 2*buflen3d+buflen2d)
      cmp(1,j)    = rbuf( rbuf2local(i) + 3*buflen3d+buflen2d)
      rdum        = rbuf( rbuf2local(i) + 4*buflen3d+buflen2d)
      if (rdum > 0) then
        casus(j)    = .true.
      else
        casus(j)    = .false.
      endif
      ice(1,j)    = rbuf( rbuf2local(i) + 4*buflen3d+2*buflen2d)
      ice(2,j)    = rbuf( rbuf2local(i) + 4*buflen3d+3*buflen2d)
      ice(3,j)    = rbuf( rbuf2local(i) + 4*buflen3d+4*buflen2d)
      tsnei(j)    = rbuf( rbuf2local(i) + 4*buflen3d+5*buflen2d)
      ueis(j)     = rbuf( rbuf2local(i) + 4*buflen3d+6*buflen2d)
      veis(j)     = rbuf( rbuf2local(i) + 4*buflen3d+7*buflen2d)
      !- tsoil has no halo
      tke(j)      = rbuf( rbuf2local(i) + 4*buflen3d+11*buflen2d)
      diss(j)     = rbuf( rbuf2local(i) + 5*buflen3d+11*buflen2d)
      avv(j)      = rbuf( rbuf2local(i) + 6*buflen3d+11*buflen2d)
    enddo
    
    !- distribute halo3d
    ! the local array index should go from n3d_l+nhalo2 +1 > n3d_l + nhalo3
    ! and the rbuf2local index from n3d_l + nhalo2 + 1 to n3d_l+nhalo3
    
    call domp_get_domain(n3d_l+nhalo2+1,buflen3d,low,up)
    j = low-1
    do i=low,up
      j = j + 1
      u(j)        = rbuf( rbuf2local(i) )
      v(j)        = rbuf( rbuf2local(i) + buflen3d)
      cmp(2,j)    = rbuf( rbuf2local(i) + 2*buflen3d+buflen2d)
      cmp(1,j)    = rbuf( rbuf2local(i) + 3*buflen3d+buflen2d)
      tke(j)      = rbuf( rbuf2local(i) + 4*buflen3d+11*buflen2d)
      diss(j)     = rbuf( rbuf2local(i) + 5*buflen3d+11*buflen2d)
      avv(j)      = rbuf( rbuf2local(i) + 6*buflen3d+11*buflen2d)
    enddo
!$OMP END PARALLEL
 
    !-------------------
    ! Deallocate work arrays and MPI types
    !-------------------
    call MPI_TYPE_FREE(ftype,i)
    call check_mpi_err(i)
    if (ia==narea) then
      deallocate(wrk1, kh_wrk, rbuf2local, wrk2, rbuf)
    endif

  end subroutine iompi_readrestart

!===============================================================================

  subroutine gather(ia, j, vardims, flist, bufsize2d, bufsize3d, fh, hbmlow,   &
                    hbmup, debug)
    !---------------------------------------------------------------------------
    ! IO-tasks goes in here and gather and writes the fields in flist to either
    ! a restart or ouput file.
    !---------------------------------------------------------------------------
    !- modules
    use io_miscsubs,    only : depermute_output, max_io_fields_def
    use dmi_mpi_global, only : iu06
    use dmi_mpi,        only : dmpi_recv
    use cmod_params,    only : narea
    use io_subs,        only : flush_unit

    !- direqtive
    implicit none

    !- args
    integer(4), intent(in) :: ia, vardims(:), flist(:), bufsize2d(1:),        &
                              bufsize3d(1:), j, fh, hbmlow, hbmup
    logical,    intent(in) :: debug

    !- locals
    integer(4) :: it, iof, offset, bsize, sbsize, ierr, tag, numhbm, ihbm, nf

    numhbm = hbmup-hbmlow+1
    offset = 1
    sbsize = 0
    do it=1,numhbm
      if (bufsize3d(it) <= 0) cycle
      ihbm   = hbmlow + it - 1
      bsize  = 0
      nf     = 0
      !- check buffer size
      do iof = 1,max_io_fields_def
        if (flist(iof) < 0) exit
        nf = nf + 1
        if (vardims(iof) == 2) then
          bsize = bsize + bufsize2d(it)
        elseif (vardims(iof) == 3) then
          bsize = bsize + bufsize3d(it)
        endif
      enddo
       
      if (debug) then
        if (ia==1 .and. it == 1) then
          write (iu06,*) 'ios_server must gather fields: ',flist(1:iof-1)
        endif
        write (iu06,*) 'io gather diagnostics: j,ia,hbm,offset,bsize: ',       &
                        j,ia,ihbm,offset,bsize
        call flush_unit(iu06)
      endif

      !- recv values from HBM-task #ihbm
      tag = scattag + ia
      call dmpi_recv(rbuf(offset:),bsize,ihbm,tag,comm_inter)

      !- move offset in the revieve buffer
      sbsize  = sbsize + bsize
      offset  = sbsize + 1
    enddo !HBM task
    
    if (debug) then
       write(iu06,*) 'Statistics for buf j,ia=', j, ia
       write(iu06,*) 'Min:', minval(rbuf(1:sbsize))
       write(iu06,*) 'Max:', maxval(rbuf(1:sbsize))
       call flush_unit(iu06)
    endif

    !- depermute the input buffer
    call depermute_output(ia,nf,mo(ia)%p,mop(ia)%p,vardims,iw2io(ia),iw3io(ia),&
                          hbmlow, hbmup, rbuf(1:sbsize), wbuf(1:sbsize) )

    call MPI_FILE_WRITE_ALL(fh, wbuf(1:sbsize), sbsize, MPI_REAL8,             &
                            MPI_STATUS_IGNORE, ierr)
    call check_mpi_err(ierr)
  end subroutine gather

!===============================================================================

  subroutine scatter(ia, flist, vardims, debug, lrestart, ctracers3, ctracers2)
    !---------------------------------------------------------------------------
    ! HBM-tasks goes in here and scatters its vars to the IO-tasks.
    !
    ! Note: Ugly solution with tmpbuf(ibuf,restart) and ireq but MPI fails when
    !       sending in a 2-d array for some reason. This is a work-around
    !---------------------------------------------------------------------------
    !- modules
    use mpi
    use io_miscsubs,        only : select_scatter_var, select_coupler_vars,    &
                                   max_io_fields_def
    use dmi_mpi_global,     only : iu06
    use dmi_mpi,            only : dd, mpi_rank, dmpi_send, dmpi_isend
    use cmod_params,        only : narea
    use io_subs,            only : flush_unit

    !- directives
    implicit none

    !- args
    integer(4), intent(in) :: ia, flist(:), vardims(:)
    logical,    intent(in) :: debug, lrestart
    real(8),    intent(in), optional :: ctracers2(:,0:), ctracers3(:,0:)

    !- locals
    integer(4) :: ial, iof,  bs, buflow, bufup, iof2, bs2, tag, ibuf
    logical    :: do_coupler

    do_coupler = .false.

    if (present(ctracers3)) do_coupler = .true.
    
    !- should we scatter restart or tempdat?
    if (do_coupler .and. lrestart) then
      !- coupler restart
      ibuf = 4
    elseif (do_coupler) then
      !- coupler output
      ibuf = 3
    elseif (lrestart) then
      !- physics restart
      ibuf = 2
    else
      !- physics output
      ibuf = 1
    endif

    if (dd(ia)%nwet3 <= 0) return 
!$OMP PARALLEL DEFAULT (shared) PRIVATE(iof, bs, buflow, bufup)
    bs     = 0
    buflow = 1
    bufup  = 0
    do iof=1,max_io_fields_def
      if (flist(iof) < 0) exit
      bufup  = bufup + n2dt(ia)
      ! must select variable to scatter
      if (do_coupler) then
        call  select_coupler_vars(tmpbuf(ibuf,ia)%p(:), flist(iof),            &
                                  buflow, bufup, 1, n2dt(ia), ctracers3,       &
                                  ctracers2)
 
        !call select_coupler_vars()
      else
        call select_scatter_var(ia,tmpbuf(ibuf,ia)%p(:),flist(iof), buflow,    &
                                bufup, 1, n2dt(ia))
      endif
      bs = bs + n2dt(ia)
      if (vardims(iof) == 3) then
        buflow = bufup + 1
        bufup  = bufup + n3dlen(ia)
        ! must select variable to scatter in 3d
        if (do_coupler) then
          call  select_coupler_vars(tmpbuf(ibuf,ia)%p(:), flist(iof),          &
                                    buflow, bufup, n3dlb(ia), n3dub(ia),       &
                                    ctracers3, ctracers2)
 
        else
          call select_scatter_var(ia, tmpbuf(ibuf,ia)%p(:),flist(iof), buflow, &
                                  bufup, n3dlb(ia), n3dub(ia)) 
        endif
        bs     = bs + n3ds(ia) -  n2dt(ia)
      endif
      buflow = bufup + 1
    enddo
!$OMP MASTER
    bs2  = bs
    iof2 = iof 
!$OMP END MASTER
!$OMP END PARALLEL

    if (debug) then
      write (iu06,*) 'ios_server SCATTER diagnostics ',                        &
         n2dt(ia), n3dlb(ia), n3dlen(ia), n3ds(ia), bs2
      write(iu06,*) 'Min, max: ', minval(tmpbuf(ibuf,ia)%p(1:bs2)),            &
                                  maxval(tmpbuf(ibuf,ia)%p(1:bs2))
      call flush_unit(iu06)
    endif

    tag = scattag + ia
    call dmpi_isend( tmpbuf(ibuf,ia)%p(1:bs2), bs2, my_iotask, tag,            &
                     ireq_scatter(ibuf)%p(ia), comm_inter )
  end subroutine scatter

!===============================================================================

  subroutine iompi_validate(sended)
    ! HBM-tasks calculates its statistics and send it to IO-task 0
    ! 
    ! IO-task 0 recieves these values and calulcate the global statistics
    !
    ! Note: Since we must be carefull of empty tasks we need some additional
    !       bookkepping (i.e. narea_recv() and num_recv_on_domain())
    !
    ! FIXME: This subroutine needs a thourougly verification...please be 
    !        carefull when using it

    !- modules
    use io_miscsubs,       only : localTracerStats
    use cmod_local_arrays, only : cmp_l, u_l, v_l, z_l, casus_l
    use dmi_mpi,           only : dmpi_isend, mpi_size, dmpi_recv,dd, mpi_tt
    use cmod_params,       only : narea
    use dmi_mpi_global,    only : iu06
    use io_subs,           only : flush_unit
    use cmod_miscsubs,     only : sum_of_values
    
    implicit none

    !- args
    logical, intent(out) :: sended

    !- locals params
    real(8),       parameter :: zero = 0.0_8, one=1.0_8
    integer(4),    parameter :: numstats = 14
    character(8),  parameter :: fmti1 = '(a33,i5)'
    character(12), parameter :: fmtr1 = '(a33,f16.13)'
    character(22), parameter :: fmtr2 = '(a33,f16.13,5x,f16.13)'
   
    !- local allocatable array
    real(8),    allocatable, save :: srbuf(:)
    integer(4), allocatable, save :: num_recv_on_domain(:), narea_recv(:)
    
    !- more locals
    real(8)    :: smean, smin, smax, tmean, tmin, tmax, zmean, zmin, zmax
    real(8)    :: srms, sstd, trms, tstd, umin, vmin, umax, vmax, partsum(0:2)

    !- global means, max and mins
    real(8)    :: gmean(3), gmin(5), gmax(5), frac, tmp1, tmp2, nice,          &
                  min_all(5,narea,0:mpi_size), max_all(5,narea,0:mpi_size),    &
                  wmean_all(3,narea, 0:mpi_size), icecovered(narea,0:mpi_size)
                 
    !- counters etc. 
    integer(4) :: n2d, ia, i, ihbm, ii, ir(narea)
    
    logical, save  :: firsttime = .true.

    if (firsttime) then
      firsttime = .false.
      allocate(srbuf(narea*numstats), num_recv_on_domain(narea),               &
               narea_recv(1:mpi_size))
      srbuf(:) = zero 

      !- is there any empty domain in any task?
      num_recv_on_domain(:) = 0
      narea_recv(:)         = 0
      do ihbm=1,mpi_size
        do ia=1,narea
          if (mpi_tt(ia,ihbm)%nwet3 > 0) then
            narea_recv(ihbm)         = narea_recv(ihbm) + 1
            num_recv_on_domain(ia) = num_recv_on_domain(ia) + 1
          endif
        enddo
      enddo
    endif
  
    !- init
    wmean_all(:,:,:) = zero
    min_all(:,:,:)   = zero
    max_all(:,:,:)   = zero
    partsum(:)       = zero
   
    if (inhbm) then
      ii = 0
      do ia=1,narea
        ! it need to count how many areas we are sending...they could be empty 
        if (dd(ia)%nwet3 <= 0) cycle

        ii = ii + 1
        n2d = dd(ia)%up_ws - dd(ia)%low_ws + 1
        !- temp, salt, u, v, z
        call localTracerStats(ia,cmp_l(ia)%p(1,:),tmin,tmax, .true., tmean)
        call localTracerStats(ia,cmp_l(ia)%p(2,:),smin,smax, .true., smean)
        call localTracerStats(ia,u_l(ia)%p(:),umin,umax, .true.)
        call localTracerStats(ia,v_l(ia)%p(:),vmin,vmax, .true.)
        call localTracerStats(ia,z_l(ia)%p(:),zmin,zmax, .false., zmean)
        nice = zero
        do i=1,n2d
          if (.not.casus_l(ia)%p(i)) nice = nice + one
        enddo
 
        !- lets pack into buffer and send
        i = numstats*(ii-1)
        srbuf(1+i)  = tmean
        srbuf(2+i)  = tmin
        srbuf(3+i)  = tmax
        srbuf(4+i)  = smean
        srbuf(5+i)  = smin
        srbuf(6+i)  = smax
        srbuf(7+i)  = umin
        srbuf(8+i)  = umax
        srbuf(9+i)  = vmin
        srbuf(10+i) = vmax
        srbuf(11+i) = zmean
        srbuf(12+i) = zmin
        srbuf(13+i) = zmax
        srbuf(14+i) = nice
      enddo
      i = numstats*ii
      if (i > 0) then ! are there any wet domains in this task?
        sended = .true.
        call dmpi_isend(srbuf(1:i), i, 0, valtag, ireq_val, comm_inter )
      else
        sended = .false. ! I have no wet domains and don't send anything
      endif
    
    else !- iam ioserver -------------------------------------------------------
      sended = .false.
      ir(:) = 0
      do ihbm=0,mpi_size-1
        i = numstats*narea_recv(ihbm+1)
        if (i <= 0) cycle ! all domains are empty on this task
        
        ! recv
        call dmpi_recv(srbuf(1:i), i, ihbm, valtag, comm_inter)
        
        !- collect means, mins and maxs from all tasks and domains
        ii = 0 ! counter for domains
        do ia=1,narea
          ! test if dry
          if (mpi_tt(ia,ihbm+1)%nwet3 > 0) then
            ! update counters
            ir(ia) = ir(ia) + 1 ! how many hbm-tasks sends this domain
            ii = ii + 1         ! to keep track on how many which domain I am
            i = numstats*(ii-1) ! to keep track where in the rbuf I should look

            tmean = srbuf(1+i) 
            smean = srbuf(4+i) 
            zmean = srbuf(11+i)
            wmean_all(1,ia,ir(ia)) = tmean
            wmean_all(2,ia,ir(ia)) = smean
            wmean_all(3,ia,ir(ia)) = zmean
            min_all(1,ia,ir(ia))   = srbuf(2+i)
            min_all(2,ia,ir(ia))   = srbuf(5+i)
            min_all(3,ia,ir(ia))   = srbuf(7+i)
            min_all(4,ia,ir(ia))   = srbuf(9+i)
            min_all(5,ia,ir(ia))   = srbuf(12+i)
            max_all(1,ia,ir(ia))   = srbuf(3+i)
            max_all(2,ia,ir(ia))   = srbuf(6+i)
            max_all(3,ia,ir(ia))   = srbuf(8+i)
            max_all(4,ia,ir(ia))   = srbuf(10+i)
            max_all(5,ia,ir(ia))   = srbuf(13+i)
            icecovered(ia,ir(ia))  = srbuf(14+i)
          endif
        enddo
      enddo
      !- calculate and write statistics
      write(iu06,'(a18)') 'Validation prints:'
      call flush_unit(iu06)

      do ia=1,narea
        write(iu06,'(a72)')                                                    &
      '------------------------------------------------------------------------'
        write(iu06,fmti1) 'Statistics for domain:           ', ia
        !- sum_of_values expects an land-point at pos one so leave it empty
        
        !- sum up temp, salt and z mean from all tasks
        i = num_recv_on_domain(ia)
        call sum_of_values(i,wmean_all(1,ia,0:i),gmean(1))
        call sum_of_values(i,wmean_all(2,ia,0:i),gmean(2))
        call sum_of_values(i,wmean_all(3,ia,0:i),gmean(3))
        gmin(1) = minval(min_all(1,ia,1:i))
        gmax(1) = maxval(max_all(1,ia,1:i))
        gmin(2) = minval(min_all(2,ia,1:i))
        gmax(2) = maxval(max_all(2,ia,1:i))
        gmin(3) = minval(min_all(3,ia,1:i))
        gmax(3) = maxval(max_all(3,ia,1:i))
        gmin(4) = minval(min_all(4,ia,1:i))
        gmax(4) = maxval(max_all(4,ia,1:i))
        gmin(5) = minval(min_all(5,ia,1:i))
        gmax(5) = maxval(max_all(5,ia,1:i))
        call sum_of_values(i,icecovered(ia,0:i),nice)
         
        write(iu06,fmtr1) 'Average value for salinity:      ', gmean(2)
        !write(iu06,fmtr2) 'RMS and STD for salinity:        ', srms, sstd
        write(iu06,fmtr1) 'Average value for temperature:   ', gmean(1)
        !write(iu06,fmtr2) 'RMS and STD for temperature:     ', trms, tstd
        write(iu06,fmtr2) 'Min and max for salinity:        ', gmin(2), gmax(2)
        write(iu06,fmtr2) 'Min and max for temperature:     ', gmin(1), gmax(1)
        write(iu06,fmtr2) 'Min and max for u:               ',                 &
        gmin(3), gmax(3)
        write(iu06,fmtr2) 'Min and max for v:               ',                 &
        gmin(4), gmax(4)
        write(iu06,fmtr1) 'Average value for z:             ', gmean(3)
        write(iu06,fmtr2) 'Min and max for z:               ', gmin(5), gmax(5)
        write(iu06,'(a41,i8)') 'No. of partially ice-covered grid points:',    &
        int(nice)
      enddo
      write(iu06,'(a72)')                                                      &
      '------------------------------------------------------------------------'
      call flush_unit(iu06)
    endif
 
  end subroutine iompi_validate

!===============================================================================

  subroutine iompi_openfile(comm, tdnam, fh, rdonly)
    !---------------------------------------------------------------------------
    ! Simple wrapper for MPI_FILE_OPEN
    !---------------------------------------------------------------------------
    use mpi
    implicit none

    !- args
    integer(4),    intent(in)  :: comm
    integer(4),    intent(out) :: fh
    character(*),  intent(in)  :: tdnam 
    logical,       intent(in)  :: rdonly

    !- locals
    integer(4) :: ierr

    if (rdonly) then
      call MPI_FILE_OPEN(comm, tdnam, MPI_MODE_RDONLY, MPI_INFO_NULL, fh, ierr)
    else
      call MPI_FILE_OPEN(comm, tdnam, MPI_MODE_WRONLY+MPI_MODE_CREATE,         &
                         MPI_INFO_NULL, fh, ierr)
    endif
    call check_mpi_err(ierr)

  end subroutine iompi_openfile

!===============================================================================

  subroutine iompi_setfileview(fh, i, ia, restart)
    !- modules
    use mpi

    implicit none

    !- args
    integer(4), intent(in)           :: i, ia
    integer(4), intent(inout)        :: fh
    logical, intent(in)              :: restart

    !- locals
    integer(4)                    :: ierr
    integer(KIND=MPI_OFFSET_KIND) :: disp
 
    disp = 0

    if (restart) then
      !- set restart filetype
      call MPI_FILE_SET_VIEW(fh,disp,MPI_REAL8,restart_filetype(ia),           &
                             'native', MPI_INFO_NULL, ierr)
    else
      call MPI_FILE_SET_VIEW(fh, disp, MPI_REAL8, filetype(i,ia),              &
                             'native', MPI_INFO_NULL, ierr)
    endif
    !FIXME: Why do we get an error here with external32??
    call check_mpi_err(ierr)
  end subroutine iompi_setfileview

!===============================================================================

  subroutine iompi_closefile(fh)
    !---------------------------------------------------------------------------
    ! Simple wrapper for MPI_CLOSE_FILE
    !---------------------------------------------------------------------------
    use mpi
    implicit none

    !- args
    integer(4), intent(inout) :: fh

    !- locals
    integer(4) :: ierr
 
    call MPI_FILE_CLOSE(fh,ierr) 
    call check_mpi_err(ierr)
  end subroutine iompi_closefile

!===============================================================================

  subroutine check4assignments(msg, masterIO)
    use mpi
    implicit none

    !- args
    real(8), intent(out) :: msg(:)
    logical, intent(in)  :: masterIO
    !- locals
    real(8), parameter :: noaction = -1.0_8

    integer(4) :: ierr, rrank
    logical    :: lflag
    integer(4) :: istat(MPI_STATUS_SIZE)

    if (masterIO) then
      !- probe for new msg
      call mpi_iprobe(MPI_ANY_SOURCE,msgtag,comm_inter,lflag,istat,ierr)
      call check_mpi_err(ierr)

      if (lflag) then
        rrank = istat(mpi_source)
        !- get new msg from HBM
        call MPI_RECV(msg,2,MPI_REAL8,rrank,msgtag,comm_inter,istat,ierr)
        call check_mpi_err(ierr)
        !- inform slave IO
        call MPI_BCAST(msg,2,MPI_REAL8,0,comm_io,ierr)
        call check_mpi_err(ierr)
      else
        msg(:) = noaction
      endif
    else ! slave IO
      !- get msg from MASTER IO
      call MPI_BCAST(msg,2,MPI_REAL8,0,comm_io,ierr)
      call check_mpi_err(ierr)
    endif
  end subroutine check4assignments

!===============================================================================

  subroutine set_comm_models(io_rank, io_server, io_ntask)
    !---------------------------------------------------------------------------
    !
    !---------------------------------------------------------------------------
    !- modules
    use dmi_mpi, only : mpi_rank, mpi_size, dmpi_reinit, dmpi_reset_size

    !- directives
    implicit none

    !- arguments
    integer(4), intent(in) :: io_rank(:), io_ntask
    logical,    intent(in) :: io_server

    !- locals
    integer(4) :: i, ierr

    inhbm = .true.

    if (io_server) then
      call MPI_COMM_DUP(MPI_COMM_WORLD,comm_dupworld,ierr) 
      call check_mpi_err(ierr)
      call mpi_comm_split(comm_dupworld,color,mpi_rank,comm_all,ierr) 
      call check_mpi_err(ierr)
      if (color==0) then
        inhbm = .true.
        call MPI_COMM_DUP(comm_all,comm_model,ierr) 
        ! FIXME - bug if io_rank is reverse sorted - MUST be first task
        call MPI_INTERCOMM_CREATE(comm_model,0,comm_dupworld,io_rank(1),0,     &
                                  comm_inter,ierr)
      else
        inhbm = .false.
        call MPI_COMM_DUP(comm_all,comm_io,ierr)
        call MPI_INTERCOMM_CREATE(comm_io,0,comm_dupworld,first_model_task,0,  &
                                  comm_inter,ierr) 
      endif
      call check_mpi_err(ierr)
      if (inhbm) then
        call MPI_COMM_SIZE(comm_inter, my_size, ierr) 
        call check_mpi_err(ierr)
        call MPI_COMM_RANK(comm_inter, my_rank, ierr) 
        call check_mpi_err(ierr)
        call dmpi_reinit(my_size,my_rank,comm_model)
      else
        !FIXME: is it OK to reset mpi_rank and comm for IO?
        !call dmpi_reset_size(mpi_size-io_ntask)
        call MPI_COMM_RANK(comm_inter, my_rank, ierr) 
        call dmpi_reinit(mpi_size-io_ntask,my_rank,comm_io)
        call check_mpi_err(ierr)
      endif
    else !no IO-server
      call dmpi_reinit(comm_model)
      return
    endif
    call MPI_BARRIER(MPI_COMM_WORLD,ierr)
 
  end subroutine set_comm_models

!===============================================================================

  subroutine create_mpitypes(vardim, varlist, restartdims,maxfh, hbm_lb,       &
                             hbm_ub, coupler_active, do_coupler)
    !---------------------------------------------------------------------------
    ! Allocate and create  MPI_TYPES for depermuting the buffer before output
    !---------------------------------------------------------------------------
    use mpi
    use io_miscsubs, only : set_mo, calc_fileview, max_io_fields_def,          &
                            num_rfields2, num_rfields3, num_rfields2_coupler,  &
                            num_rfields3_coupler
    use dmi_mpi,     only : mpi_tt
    use cmod_params, only : narea, mm1, mmx,nmx,kmx,iw2,iw3

    implicit none

    !- arguments
    integer(4),intent(in) :: vardim(1:,1:), varlist(:,:), restartdims(1:),     &
                             maxfh, hbm_lb, hbm_ub
    logical,   intent(in)           :: coupler_active
    logical,   intent(in), optional :: do_coupler

    !- locals
    integer(4) :: ia, i, iof, nf, ierr, nblocs, fsize, ftype
    logical    :: cleanup

    !- MPI 
    integer(kind=MPI_ADDRESS_KIND) :: lb, ub

    cleanup = .false.

    !- Calculate the index arrays mo(=part of perm mm1) and mop(=perm mm1)
    do ia=1,narea
      call set_mo(ia, mm1(ia)%p, mo(ia)%p, mop(ia)%p, iw2io(ia), iw2(ia),      &
                  hbm_lb, hbm_ub)
    enddo
    
    !- create the MPI types
    do ia=1,narea
      do i=1,maxfh
        nf     = 0
        fsize  = 0
        do iof = 1,max_io_fields_def
          if ( varlist(iof,i) < 0 ) exit
          nf = nf + 1
          if (vardim(iof,i) == 2) then
            fsize = fsize + iw2(ia)
          elseif (vardim(iof,i) == 3) then
            fsize = fsize + iw3(ia)
          else
            !ERROR
          endif
        enddo
        
        !- Commit output fileview
        call calc_fileview(ia, nf,  mo(ia)%p, mop(ia)%p, wrk1, wrk2,           &
                            vardim(:,i), iw2io(ia), iw3io(ia), iw2(ia),        &
                            iw3(ia), hbm_lb, hbm_ub, nblocs )
        call MPI_TYPE_INDEXED(nblocs, wrk1(1:nblocs), wrk2(1:nblocs),          &
                               MPI_REAL8,ftype, ierr)
        call check_mpi_err(ierr)
        !- set extend of the fileview to the whole file so each IO-task uses
        !  correct stepping for each write...not necessary for restart since
        !  the do not append data
        lb = 0
        ub = fsize*8
        if (present(do_coupler)) then
          call MPI_Type_create_resized(ftype,lb, ub,filetype_coupler(i,ia),ierr)
          call check_mpi_err(ierr)
          call MPI_TYPE_COMMIT(filetype_coupler(i,ia),ierr)
          nf = num_rfields2_coupler+num_rfields3_coupler
        else
          call MPI_Type_create_resized(ftype,lb, ub, filetype(i,ia),ierr)
          call check_mpi_err(ierr)
          call MPI_TYPE_COMMIT(filetype(i,ia),ierr)
          nf = num_rfields2+num_rfields3
        endif
      enddo
      call check_mpi_err(ierr)

      !- Calculate restart fileview
      call calc_fileview(ia, nf, mo(ia)%p, mop(ia)%p, wrk1, wrk2,              &
                          restartdims(:), iw2io(ia), iw3io(ia), iw2(ia),       &
                          iw3(ia), hbm_lb, hbm_ub, nblocs)

      !- Commit restart fileview
      if (present(do_coupler)) then
        call MPI_TYPE_INDEXED(nblocs, wrk1(1:nblocs), wrk2(1:nblocs),          &
                              MPI_REAL8, restart_filetype_coupler(ia), ierr)
        call check_mpi_err(ierr)
        call MPI_TYPE_COMMIT(restart_filetype_coupler(ia),ierr)
        call check_mpi_err(ierr)
        !- cleanup (io-server doesnt need the wrk arrays anymore)
        cleanup = .true.
      else
        call MPI_TYPE_INDEXED(nblocs, wrk1(1:nblocs), wrk2(1:nblocs),          &
                              MPI_REAL8, restart_filetype(ia), ierr)
        call check_mpi_err(ierr)
        call MPI_TYPE_COMMIT(restart_filetype(ia),ierr)
        call check_mpi_err(ierr)
        !- cleanup (io-server doesnt need the wrk arrays anymore)
        if (.not. coupler_active) cleanup = .true.
      endif 
      ! FREE temporary ftype
      call MPI_TYPE_FREE(ftype,ierr)
      call check_mpi_err(ierr)
    enddo

    !- cleanup
    if (cleanup) deallocate(wrk1, wrk2)
  end subroutine create_mpitypes

!===============================================================================

  subroutine alloc_HBM_arrays(flist, fdims, maxfh, flist_c, fdims_c, maxfh_c)
    !- modules
    use dmi_mpi,     only : dd
    use io_miscsubs, only : max_io_fields_def, num_rfields2, num_rfields3,     &
                            num_rfields2_coupler, num_rfields3_coupler
    use cmod_params, only : narea

    implicit none
  
    !- arguments
    integer(4), intent(in)           :: flist(:,:), fdims(:,:), maxfh
    integer(4), intent(in), optional :: flist_c(:,:), fdims_c(:,:), maxfh_c

    !- locals
    integer(4) :: i, ia, iof, nbufs, bs 
    integer(4) :: max_out3, max_out2, max_out3_c, max_out2_c, nf2, nf3
    logical    :: coupler_active
   
    !- defaults
    coupler_active = .false.
    nbufs          = 2

    if (present(flist_c) .and. present(fdims_c) )  coupler_active = .true.

    !- calculate the size needed on buffers
    max_out3   = 0
    max_out2   = 0
    max_out3_c = 0
    max_out2_c = 0
    do i = 1,maxfh
      nf2 = 0
      nf3 = 0
      do iof = 1, max_io_fields_def
        if (flist(iof,i) <= -1) exit

        if (fdims(iof,i) == 2) then
          nf2 = nf2 + 1
        elseif (fdims(iof,i) == 3) then
          nf3 = nf3 + 1
        endif
      enddo
      max_out3 = max(max_out3,nf3)
      max_out2 = max(max_out2,nf2)
    enddo

    if (coupler_active) then
      !- two more buffers needed (one for restart, one for results)
      nbufs = 4
    
      do i = 1,maxfh_c
        nf2 = 0
        nf3 = 0
        do iof = 1, max_io_fields_def
          if (flist_c(iof,i) <= -1) exit
  
          if (fdims_c(iof,i) == 2) then
            nf2 = nf2 + 1
          elseif (fdims_c(iof,i) == 3) then
            nf3 = nf3 + 1
          endif
        enddo
        max_out3_c = max(max_out3_c,nf3)
        max_out2_c = max(max_out2_c,nf2)
      enddo
    endif
 
    !------------
    ! init/alloc arrays for the hbm-tasks
    !------------
    allocate (n2dt(narea), n3dlb(narea), n3ds(narea), n3dlen(narea),           &
              n3dub(narea), ireq_scatter(nbufs), tmpbuf(nbufs,narea))
    do i=1,nbufs
      allocate( ireq_scatter(i)%p(narea))
    enddo
    
    !- assign start and end positions of 2d and 3d vars
    do ia=1,narea
      if (dd(ia)%nwet3 > 0) then
        n2dt(ia)   = dd(ia)%up_ws - dd(ia)%low_ws + 1
        bs          = dd(ia)%nwet3
        n3dlb(ia)  = (n2dt(ia)+1) + dd(ia)%halo2
        n3dlen(ia) = bs - n2dt(ia)
        n3dub(ia)  = n3dlb(ia) + n3dlen(ia) - 1
        n3ds(ia)   = dd(ia)%nwet3
        
        !- output buffer
        bs = n3ds(ia)*max_out3+n2dt(ia)*max_out2
        allocate( tmpbuf(1,ia)%p(1:bs) )
        !- restart buffer
        bs = n3ds(ia)*num_rfields3+n2dt(ia)*num_rfields2
        allocate( tmpbuf(2,ia)%p(1:bs) )
        
        if (coupler_active) then
          !- output buffer
          bs = n3ds(ia)*max_out3_c+n2dt(ia)*max_out2_c
          allocate( tmpbuf(3,ia)%p(1:bs) )
          !- restart buffer
          bs = n3ds(ia)*num_rfields3_coupler+n2dt(ia)*num_rfields2_coupler
          allocate( tmpbuf(4,ia)%p(1:bs) )
        endif
      endif
    enddo
  end subroutine alloc_HBM_arrays

!===============================================================================

  subroutine alloc_IOS_arrays(flist, fdims, maxfh, hbm_lb, hbm_ub, sbufsize2d, &
                              sbufsize3d, flist_c, fdims_c, maxfh_c)

    !- modules
    use dmi_mpi,     only : mpi_tt
    use io_miscsubs, only : max_io_fields_def, num_rfields2, num_rfields3,     &
                            num_rfields2_coupler, num_rfields3_coupler
    use cmod_params, only : narea, mmx, nmx, kmx
    
    implicit none
  
    !- arguments
    integer(4), intent(in)           :: flist(:,:), fdims(:,:), maxfh
    integer(4), intent(in)           :: hbm_lb, hbm_ub
    integer(4), intent(in)           :: sbufsize2d(1:), sbufsize3d(1:)
    integer(4), intent(in), optional :: flist_c(:,:), fdims_c(:,:), maxfh_c

    !- locals
    integer(4), parameter :: uninitialized=-99999.0
    
    integer(4) :: i,ihbm, ia, iof, nbufs, bs,  hbmlen, n3dl
    integer(4) :: n2dl, mxbuflen, buflen, mmax, nmax, kmax
    integer(4) :: max_out3, max_out2, max_out3_c, max_out2_c, nf2, nf3
    logical    :: coupler_active

    !- defaults
    coupler_active = .false.
    nbufs          = 2

    if (present(flist_c) .and. present(fdims_c) )  coupler_active = .true.

    !- calculate the size needed on buffers
    max_out3   = 0
    max_out2   = 0
    max_out3_c = 0
    max_out2_c = 0
    do i = 1,maxfh
      nf2 = 0
      nf3 = 0 
      do iof = 1, max_io_fields_def
       if (flist(iof,i) <= -1) exit

        if (fdims(iof,i) == 2) then
          nf2 = nf2 + 1
        elseif (fdims(iof,i) == 3) then
          nf3 = nf3 + 1
        endif
      enddo
      max_out3 = max(max_out3,nf3)
      max_out2 = max(max_out2,nf2)
    enddo

    if (coupler_active) then
      !- two more buffers needed (one for restart, one for results)
      nbufs = 4
    
      do i = 1,maxfh_c
        nf2 = 0
        nf3 = 0
        do iof = 1, max_io_fields_def
          if (flist_c(iof,i) <= -1) exit
  
          if (fdims_c(iof,i) == 2) then
            nf2 = nf2 + 1
          elseif (fdims_c(iof,i) == 3) then
            nf3 = nf3 + 1
          endif
        enddo
        max_out3_c = max(max_out3_c,nf3)
        max_out2_c = max(max_out2_c,nf2)
      enddo
    endif
 
    !------------
    ! init/alloc arrays for IO-tasks
    !------------
    hbmlen = hbm_ub - hbm_lb + 1

    !- allocate some arrays
    allocate( mo(narea), mop(narea) )
    allocate( iw2io(narea), iw3io(narea) )
    allocate( filetype(maxfh,narea), restart_filetype(narea) )
    
    if (coupler_active) then
      allocate( filetype_coupler(maxfh_c,narea),                               &
                restart_filetype_coupler(narea) )
    endif

    !- get the number of 2d and 3d points this IO- taks handles
    do ia=1,narea
      iw2io(ia) = 0
      iw3io(ia) = 0
      do i=1,hbmlen
        ihbm  = hbm_lb + i
        n3dl  = mpi_tt(ia,ihbm)%nwet3
        if (n3dl > 0) then
          n2dl  = mpi_tt(ia,ihbm)%up_ws - mpi_tt(ia,ihbm)%low_ws + 1
          iw2io(ia) = iw2io(ia) + n2dl
          iw3io(ia) = iw3io(ia) + n3dl
        endif
      enddo
    enddo

    !- number of hbmtasks to handle
    hbmlen = hbm_ub-hbm_lb+1

    !- allocate the work arrays for construction of the file view

    !FIXME: only works for i-slices now
    nf2 = max(num_rfields2, num_rfields2_coupler, max_out2, max_out2_c)
    nf3 = max(num_rfields3, num_rfields3_coupler, max_out3, max_out3_c)
    mxbuflen = maxval(nmx(1:narea))*nf2 + maxval(nmx(1:narea))*nf3*2
    allocate( wrk1(1:mxbuflen), wrk2(1:mxbuflen) )

    !- allocate recieve and write buffer
    mxbuflen = 0
    do ia=1,narea !- which region has the largerst buflen
      buflen = sbufsize3d(ia)*nf3 + sbufsize2d(ia)*nf2
      mxbuflen = max(mxbuflen,buflen)
    enddo
    
    allocate( rbuf(1:mxbuflen), wbuf(1:mxbuflen)  )
    rbuf(:)=uninitialized
    wbuf(:)=uninitialized

    !- allocate the work arrays
    mmax = maxval(mmx(1:narea))
    nmax = maxval(nmx(1:narea))
    kmax = maxval(kmx(1:narea))
    do ia=1,narea
      allocate(  mo(ia)%p(mmx(ia),nmx(ia),kmx(ia)) )
      allocate( mop(ia)%p(mmx(ia),nmx(ia),kmx(ia)) )
    enddo
  end subroutine alloc_IOS_arrays

!===============================================================================

  subroutine check_mpi_err(ierr)
    use mpi
    use exits, only : exitme
    implicit none

    integer(4), intent(in) :: ierr
    if (ierr /= MPI_SUCCESS) then
      call exitme(ierr,'MPI Error in io_mpi.F90')
    endif
  end subroutine check_mpi_err

!===============================================================================

  subroutine iompi_wait(ireq)
    use mpi
    implicit none

    !- args
    integer(4), intent(inout) :: ireq

    !- locals
    integer(4) :: mpistat(MPI_STATUS_SIZE), ierr
 
    call MPI_WAIT(ireq,mpistat,ierr)
    call check_mpi_err(ierr)
  end subroutine iompi_wait

!===============================================================================
 
  subroutine iompi_cleanup(maxfh, maxfh_c, ioserver)
    use cmod_params, only : narea
    implicit none

    !- arge
    integer(4), intent(in) :: maxfh, maxfh_c
    logical,    intent(in) :: ioserver

    !- locals
    integer(4) :: i, ia, ierr

    if (.not. ioserver) then
      call MPI_COMM_FREE(comm_model,ierr)
      call check_mpi_err(ierr)
      return
    endif
     
    !- make sure everyone is here
    call MPI_BARRIER(MPI_COMM_WORLD,ierr)
    
    call MPI_COMM_FREE(comm_dupworld,ierr)
    call check_mpi_err(ierr)
    call MPI_COMM_FREE(comm_all,ierr)
    call check_mpi_err(ierr)
    call MPI_COMM_FREE(comm_inter,ierr)
    call check_mpi_err(ierr)

    if (.not.inhbm) then
      call MPI_COMM_FREE(comm_io,ierr)
      call check_mpi_err(ierr)
      do ia = 1,narea
        call MPI_TYPE_FREE(restart_filetype(ia),ierr)
        call check_mpi_err(ierr)
        do i=1,maxfh
          call MPI_TYPE_FREE(filetype(i,ia),ierr)
          call check_mpi_err(ierr)
        enddo
      enddo
      
      !- free coupler types
      if (maxfh_c>0) then
        do ia = 1,narea
          call MPI_TYPE_FREE(restart_filetype_coupler(ia),ierr)
          call check_mpi_err(ierr)
          do i=1,maxfh_c
            call MPI_TYPE_FREE(filetype_coupler(i,ia),ierr)
            call check_mpi_err(ierr)
          enddo
        enddo
      endif

    endif
  end subroutine iompi_cleanup

!===============================================================================
#else
!-------------------------------------------------------------------------------
! The subroutines below are interfaces which are used when running without MPI.
! Obviously, the ioserver is not activated in this case.
!-------------------------------------------------------------------------------
  subroutine iompi_readrestart(ia, fname, mm1, kh, mmx, nmx, kmx, n2d, n3d,    &
                               vardim,aufdat, u, v, z, cmp, casus, ice, tsnei, &
                               ueis, veis, t_soil, tke, diss, avv, ompidx,     &
                               coupler2d)
    implicit none

    !- arguments
    character(len=11), intent(in) :: aufdat
    character(len=*),  intent(in) :: fname
    integer(4),        intent(in) :: ia, mmx, nmx, kmx, n2d, n3d
    integer(4),        intent(in) :: vardim(:), mm1(0:,0:,:), kh(0:)
    real(8),           intent(in) :: u(0:), v(0:), z(0:), tke(0:), diss(0:),  &
                                     avv(0:), ueis(0:), veis(0:), ice(:,0:),  &
                                     t_soil(:,:), tsnei(0:), cmp(:,0:)
    integer(4),      intent(inout) :: ompidx(:,:)
    logical,           intent(in) :: casus(0:)
    real(8), optional, intent(in) :: coupler2d(:,:)
  end subroutine iompi_readrestart
  !---------
  subroutine set_comm_models(io_rank, io_server, io_ntask)
    use dmi_mpi, only : dmpi_reinit
    implicit none
    integer(4), intent(in) :: io_rank(:), io_ntask
    logical,    intent(in) :: io_server
    integer(4) :: dummy
    inhbm = .true.
    call dmpi_reinit(dummy)
  end subroutine set_comm_models
  !---------
  subroutine create_mpitypes(vardim, varlist, restartdims,maxfh, hbm_lb,       &
                             hbm_ub, coupler_active, do_coupler)
    implicit none
    integer(4), intent(in) :: vardim(1:,1:), varlist(:,:), restartdims(1:),    &
                              maxfh, hbm_lb, hbm_ub
    logical,    intent(in) :: coupler_active
    logical,    intent(in), optional :: do_coupler

  end subroutine create_mpitypes
  !---------
  subroutine gather(ia, j, vardims, flist, bufsize2d, bufsize3d, fh, hbmlow,   &
                    hbmup, debug)
    implicit none
    integer(4), intent(in) :: ia, vardims(:), flist(:), bufsize2d(1:),         &
                              bufsize3d(1:), j, fh, hbmlow, hbmup
    logical,    intent(in) :: debug
  end subroutine gather
  !---------
  subroutine scatter(ia, flist, vardims, debug, lrestart, ctracers3, ctracers2)
    implicit none
    integer(4), intent(in) :: ia, flist(:), vardims(:)
    logical,    intent(in) :: debug, lrestart
    real(8),    intent(in), optional :: ctracers2(:,:), ctracers3(:,:)
  end subroutine scatter
  !---------
  subroutine iompi_openfile(comm, tdnam, fh, rdonly)
    implicit none
    integer(4), intent(in) :: comm
    integer(4), intent(in) :: fh
    character(40), intent(in) :: tdnam
    logical, intent(in) :: rdonly
  end subroutine iompi_openfile
  !---------
  subroutine iompi_setfileview(fh,i,ia,restart)
    implicit none
    integer(4), intent(in) :: fh, i, ia
    logical,    intent(in) :: restart
  end subroutine iompi_setfileview
  !---------
  subroutine iompi_closefile(fh)
    implicit none
    integer(4), intent(in) :: fh
  end subroutine iompi_closefile
  !---------
  subroutine check4assignments(msg, masterIO)
    implicit none
    real(8), intent(in) :: msg(:)
    logical,    intent(in) :: masterIO
  end subroutine check4assignments
  !---------
  subroutine iompi_wait(ireq)
    implicit none
    integer(4), intent(in) :: ireq
  end subroutine iompi_wait
  !---------
  subroutine iompi_cleanup(maxfh, maxfh_c, ioserver)
    implicit none
    integer(4), intent(in) :: maxfh, maxfh_c
    logical,    intent(in) :: ioserver
  end subroutine iompi_cleanup
  !---------
  subroutine alloc_HBM_arrays(flist, fdims, maxfh, flist_c, fdims_c, maxfh_c)
    implicit none
    integer(4), intent(in)           :: flist(:,:), fdims(:,:), maxfh
    integer(4), intent(in), optional :: flist_c(:,:), fdims_c(:,:), maxfh_c
  end subroutine alloc_HBM_arrays
  !--------- 
  subroutine alloc_IOS_arrays(flist, fdims, maxfh, hbm_lb, hbm_ub, sbufsize2d, &
                             sbufsize3d, flist_c, fdims_c, maxfh_c)
    implicit none
    integer(4), intent(in)           :: flist(:,:), fdims(:,:), maxfh
    integer(4), intent(in)           :: hbm_lb, hbm_ub
    integer(4), intent(in)           :: sbufsize2d(1:), sbufsize3d(1:)
    integer(4), intent(in), optional :: flist_c(:,:), fdims_c(:,:), maxfh_c
  end subroutine alloc_IOS_arrays
  !---------
 
  subroutine iompi_validate(sended)
    implicit none
    logical, intent(out) :: sended
    sended = .false.
  end subroutine
  !---------
  subroutine check_mpi_err()
    implicit none
  end subroutine check_mpi_err
  !---------
#endif

end module io_mpi
