module masscorrection
  !  Wrapper for the mass correction to be applied in with tracer advection
  !  when drying is present.
  !  The present implementation attemps to keep the increased memory usage 
  !  and computational load low by using temporary compressed work arrays.
  !  This means that OpenMP is not supported here, but MPI is if supported by
  !  the caller. I.e. it _MUST_ be called outside OpenMP parallel sections, and
  !  and masscorr_init() should be called on the same MPI-task where the 
  !  masscorr() is later applied but then there is no need for explicit 
  !  MPI-communication (gather/scatter/haloswap).

  implicit none
  private

  public :: masscorr_firstentry, masscorr_init, masscorr

  integer(4), save,              private :: ntr, ndry_max
  integer(4), save, allocatable, private :: idry(:), tndry(:,:)
  real(8),    save, allocatable, private :: buf(:,:), cor(:)

contains

  !-----------------------------------------------------------------------------

  subroutine masscorr_firstentry( nthr, ierr )
    implicit none
    integer, intent(in)    :: nthr
    integer, intent(inout) :: ierr
    ntr = nthr 
    allocate ( tndry(1:nthr,1:2), stat=ierr ) 
    if (ierr /= 0) return
    ndry_max = 0
  end subroutine masscorr_firstentry

  !-----------------------------------------------------------------------------

  subroutine masscorr_init ( n2d, nc, cmp, hnew, hold, hx, hy, hstab, u, v,    &
                             dt, dx, dy, cosphi, ind, m, kh, idx )

    use constants, only : hgr, heps, half
    use dmi_omp,   only : domp_get_domain, domp_get_thread_no

    implicit none

    integer(4), intent(in)    :: n2d, nc, ind(1:,1:), m(1:,0:,0:), kh(0:)
    integer(4), intent(inout) :: idx(1:,1:)
    real(8),    intent(in)    :: hnew(0:), hold(0:), cmp(1:,0:), u(0:), v(0:)
    real(8),    intent(in)    :: cosphi(1:,0:), hx(0:), hy(0:)
    real(8),    intent(in)    :: hstab, dt, dx, dy
    
    real(8),    parameter :: hdry = hgr + heps
    integer(4)            :: ms, mw, mn, ii, i, j, tnum, itr, n2dl, n2du, ndry

    call domp_get_thread_no( tnum )
    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)

    !  find No. of dried points in this area/task:
    tndry(tnum,1) = 0
    do ms=n2dl,n2du
      if (hnew(ms) - hstab <= hdry) tndry(tnum,1) = tndry(tnum,1) + 1
    enddo

!$OMP BARRIER
!$OMP MASTER
    ndry       = tndry(1,1)
    tndry(1,2) = 0
    do itr=2,ntr
      ndry         = ndry + tndry(itr,1)
      tndry(itr,2) = tndry(itr-1,2) + tndry(itr-1,1)
    enddo
    !  poor man's re-allocator:
    if (ndry > ndry_max) then
      if (ndry_max > 0) deallocate( buf, cor, idry )
      allocate( buf(1:nc,1:ndry), cor(1:ndry), idry(1:ndry) )
      ndry_max = ndry
    endif
!$OMP END MASTER
!$OMP BARRIER

    !  now, fill the work arrays:
    if (tndry(tnum,1) > 0) then
      ii = tndry(tnum,2)
      do ms=n2dl,n2du
        if (hnew(ms) - hstab > hdry) cycle
  
        i  = ind(1,ms)
        j  = ind(2,ms)
        mw = m(1,i,  j-1)
        mn = m(1,i-1,j  )

        ! save index:
        ii = ii + 1
        idry(ii) = ms

        ! save components:
        buf(1:nc,ii) = cmp(1:nc,ms)
       
        ! save correction:
        cor(ii) = (  (hnew(ms) - hold(ms))/dt                                  &
                   + ( ( hx(ms)*u(ms)                                          &
                        -hx(mw)*u(mw))/dx                                      &
                      +( hy(mn)*v(mn)*cosphi(2,i-1)                            &
                        -hy(ms)*v(ms)*cosphi(2,i)  )/dy)/cosphi(1,i) )         &
                  *half*dt/hnew(ms)
      enddo 
    endif
  end subroutine masscorr_init

  !-----------------------------------------------------------------------------

  subroutine masscorr (nc, cmp)

    use constants, only : one
    use dmi_omp,   only : domp_get_thread_no

    implicit none

    integer(4), intent(in)    :: nc
    real(8),    intent(inout) :: cmp(1:,0:)

    integer(4) :: id, mm, tnum
    real(8)    :: mcor

    call domp_get_thread_no( tnum )

    !  do the mass correction:
    if (tndry(tnum,1) > 0) then
      do id=tndry(tnum,2)+1,tndry(tnum,2)+tndry(tnum,1)
        mm   = idry(id)
        mcor = cor(id)
        cmp(1:nc,mm) = (cmp(1:nc,mm) + mcor*buf(1:nc,id))/(one - mcor)
      enddo
    endif
  end subroutine masscorr

  !-----------------------------------------------------------------------------

end module masscorrection
