module masseqs
  use constants, only : zero, one
  !-----------------------------------------------------------------------------
  ! wrapping the mass equation solver
  !-----------------------------------------------------------------------------

  implicit none
  private

  public :: masseqs_solver_z, masseqs_solver_dyn

contains

subroutine masseqs_solver_z ( iwet2,mm,ind,dt,dx,dy,covo,covu,                 &
                              z,w,zn,un,vn,huti,hvti,kh,idx)  
  !-----------------------------------------------------------------------------
  ! mass (or continuity) equation solver
  !-----------------------------------------------------------------------------

  !- modules -------------------------------------------------------------------
  use dmi_omp,     only : domp_get_domain
  use cmod_params, only : kmax
 
  !- directives ----------------------------------------------------------------
  implicit none
 
  !- arguments -----------------------------------------------------------------
  integer(4), intent(in)                :: iwet2
  integer(4), intent(in)                :: mm(:,0:,0:)          ! (ke,0:mp,0:np)
  integer(4), intent(in)                :: ind(:,:)
  integer(4), intent(in)                :: kh(0:)
  real(8), intent(in)                   :: dt,dy
  real(8), intent(in),    dimension(:)  :: dx,covo,covu         ! (1:m)
  real(8), intent(in),    dimension(0:) :: z
  real(8), intent(in),    dimension(0:) :: un,vn,huti,hvti
  real(8), intent(inout), dimension(0:) :: zn
  real(8), intent(inout), dimension(0:) :: w
  integer(4), intent(inout)             :: idx(:,:)             ! (2,iwet2)
 
  !- local vars ----------------------------------------------------------------
  integer(4) :: i, j, k, kb, nsrf, nl, nu, mm2, mw2, mn2, mw, mn
  real(8)    :: wup, div
  real(8)    :: ddx,covody,covudy,ddy
  real(8), dimension(kmax) :: uwest_c, vnrth_c, div_c

  ! ----------------------------------------------------------------------------
  ddy = one/dy

  call domp_get_domain (kh, 1, iwet2, nl, nu, idx)

  do nsrf=nl,nu
    i  = ind(1,nsrf)
    j  = ind(2,nsrf)

    ddx    = one/dx(i)
    covody = covo(i)*ddy
    covudy = covu(i)*ddy
    kb     = kh(nsrf)

    mw = mm(1,i,j-1)
    mn = mm(1,i-1,j)

    if (kb     > 1) mm2 = mm(2,i,j)-2
    if (kh(mw) > 1) mw2 = mm(2,i,j-1)-2
    if (kh(mn) > 1) mn2 = mm(2,i-1,j)-2

    ! find u west of grid cell
    uwest_c(1) = un(mw)*huti(mw)
    do k=2,min(kb,kh(mw))
      uwest_c(k) = un(mw2+k)*huti(mw2+k)
    enddo
    do k=min(kb,kh(mw))+1,kb
      uwest_c(k) = zero
    enddo
    ! find v north of grid cell
    vnrth_c(1) = vn(mn)*hvti(mn)*covody
    do k=2,min(kb,kh(mn))
      vnrth_c(k) = vn(mn2+k)*hvti(mn2+k)*covody
    enddo
    do k=min(kb,kh(mn))+1,kb
      vnrth_c(k) = zero
    enddo

    ! do layer-integration and depth-integration of mass equation:
    k        = 1
    div      = (un(nsrf)*huti(nsrf) - uwest_c(k))*ddx                          &
             + vnrth_c(k) - vn(nsrf)*hvti(nsrf)*covudy
    div_c(k) = div
    wup      = -div
    do k=2,kb
      div      = (un(mm2+k)*huti(mm2+k) - uwest_c(k))*ddx                      &
               + vnrth_c(k) - vn(mm2+k)*hvti(mm2+k)*covudy
      div_c(k) = div
      wup      = wup - div
    enddo

    ! update surface elevation:
    zn(nsrf) = z(nsrf) + wup*dt

    ! distribute to diagnostic variable w:
    w(nsrf) = wup
    do k=2,kb
      w(mm2+k) = wup + div_c(k-1)
      wup      = w(mm2+k)
    enddo
  enddo

  ! ----------------------------------------------------------------------------

end subroutine masseqs_solver_z

!===============================================================================

subroutine masseqs_solver_dyn (ia,kmax,mmk,ind,n2d,n3d,khu,khv,kh,dx,dy,dt,    &
                               covo,covu,hu,hv,un,vn,z,zn,ti,h,h0,hn,ws,avv)

! ------------------------------------------------------------------------------
!  This subroutine describes the equation of continuity with special dynamical
!  vertical coordinates. The physical/mathematical basics you can find in the
!  article:
!  "A class of hybrid vertical coordinates for ocean circulation modelling" 
!  by E.Kleine, BSH, 2003
! ------------------------------------------------------------------------------


!  In this subroutine the layer thicknesses h(0:n3d) are calculated new. 
!  Because in this code the positions of the vertical layers are dynamic (and
!  in general not horizontal), the vertical velocity is divided in three
!  components. One component depends on the horizontal velocity (these terms 
!  exists, because the layers are not horizontal anymore), one which describes 
!  just the velocity of the movement of the layers itself - this one merges in 
!  h(0:n3d) which is the new calculated layer thickness - and one which 
!  describes the velocity of the water which moves between two layers (the so 
!  called pseudo vertical velocity ws(0:n3d).
!  Because the general vertical coordinates provides a degree of freedom, there 
!  is another differential equation which is introduced in the above named 
!  literature by E.Kleine, so that the equation which is to be solved is of the
!  form (1+gamma)d/dt*(dz/ds)+d/ds(d/ds(mu/(dz/ds))) + 
!  Grad_s(v * dz/ds - vi * Grad_s dz/ds ) = 0 


    !- modules  ----------------------------------------------------------------
    use constants,      only : two, hgr, wghtz, wghthar, wghteta, wghtturb,    &
                               gammadt
    use dmi_mpi_global, only : iu06
    use dmi_omp,        only : domp_get_domain
    use exits,          only : exitme
    
    
    !- directives --------------------------------------------------------------
    implicit none
    
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)       :: ia
    integer(4), intent(in)       :: kmax
    integer(4), intent(in)       :: mmk(:,0:,0:)
    integer(4), intent(in)       :: ind(:,:)
    integer(4), intent(in)       :: n2d,n3d
    integer(4), intent(in)       :: khu(0:),khv(0:)
    integer(4), intent(in)       :: kh(0:)
    real(8), intent(in)          :: dy,dt
    real(8), intent(in)          :: dx(:),covo(:),covu(:)
    real(8), intent(in)          :: hu(0:),hv(0:),un(0:),vn(0:)
    real(8), intent(in)          :: avv(0:)
    real(8), intent(in)          :: z(0:),ti(0:)
    real(8), intent(in)          :: h(0:),h0(0:)
    real(8), intent(inout)       :: hn(0:)
    real(8), intent(inout)       :: ws(0:), zn(0:)
    
    
    !- local vars --------------------------------------------------------------
    real(8),dimension(1:kmax)   :: h_tmp, h_inc
    real(8),dimension(1:kmax)   :: div,diverg,wpsd
    real(8),dimension(1:kmax)   :: a_h,b_h,c_h,e_h
    real(8),dimension(1:kmax+1) :: r_h,z_h
    integer(4)           :: nn,kb
    integer(4)           :: i,j,k
    integer(4)           :: n2dl, n2du
    integer(4)           :: kbue,kbuw,kbvn,kbvs
    integer(4)           :: iterat,it_max
    integer(4)           :: kbw,kbe,kbn,kbs
    integer(4)           :: mi
    real(8)              :: divk, dxi, covoidy, covuidy
    real(8)              :: depth_n,w_pseudo
    real(8)              :: hue,huw,hvn,hvs
    real(8)              :: norm,scal
    real(8)              :: dh, fac
    real(8)              :: length_a,length_b
    real(8)              :: h_tmp_a, h_tmp_b
    
    !- local parameters --------------------------------------------------------
    real(8),parameter    ::  scal_max = 1.0e+6_8, norm_min = 1.0e-9_8


      call domp_get_domain (n3d, kh, 1, n2d, n2dl, n2du)
          
              
      do nn = n2dl, n2du
        i = ind(1,nn)
        j = ind(2,nn)

        kb    = kh(nn)
        dxi   = dx(i)
        covoidy = covo(i)/dy
        covuidy = covu(i)/dy
   
        h_inc(:) = zero
        div(:)   = zero
        diverg(:)= zero
        wpsd(:)  = zero

        depth_n = zero
      
        !- Calculate v * dz/ds and Grad_s * v * dz/ds, i.e. the component of 
        !- vertical velocity which depends on the horizontal velocity resp. the 
        !- horizontal water flux. 
        do k=1,kb
          mi = mmk(k, i, j)
          hue=hu(mi)*un(mi)
          huw=hu(mmk(k,i,j-1))*un(mmk(k,i,j-1))
          hvn=hv(mmk(k,i-1,j))*vn(mmk(k,i-1,j))
          hvs=hv(mi)*vn(mi)
          divk=(hue-huw)/dxi+(hvn*covoidy-hvs*covuidy)

          depth_n = depth_n + h(mi) - dt*divk
          div(k)    = divk          
          diverg(k) = divk          
          h_tmp(k)  = h(mi)
        end do     

        !- vertical redistribution of divergence so as to smooth layer 
        !- thicknesses wghthar*dh/dx(i) stands for Grad_s 
        !- (velocity * length * Grad_s (dz/ds)), whereas the inner Grad_s is 
        !- already in the factor dh, since dh = dh+h(mm(i,j,k))-h(mm(i,j+1,k))
        !- depends of course direcly on the space between two cells, which is 
        !- described by dx or dy.

        ! Grad_s(dz/ds) conform to ds*Grad_s(dz/ds) in descrete.
        ! dh = lenght * Grad_s (dz/ds)

        kbue = khu(mmk(1,i,j))
        dh = zero
        do k=1,kbue-1
          dh=dh+h(mmk(k,i,j))-h(mmk(k,i,j+1))
          div( k )=div( k )+wghthar*dh/dxi
          div(k+1)=div(k+1)-wghthar*dh/dxi
        end do
        kbuw=khu(mmk(1,i,j-1))
        dh = zero
        do k=1,kbuw-1
          dh=dh+h(mmk(k,i,j))-h(mmk(k,i,j-1))
          div( k )=div( k )+wghthar*dh/dxi
          div(k+1)=div(k+1)-wghthar*dh/dxi
        end do
        kbvn=khv(mmk(1,i-1,j))
        dh = zero
        do k=1,kbvn-1
          dh=dh+h(mmk(k,i,j))-h(mmk(k,i-1,j))
          div( k )=div( k )+wghthar*dh*covoidy
          div(k+1)=div(k+1)-wghthar*dh*covoidy
        end do
        kbvs=khv(mmk(1,i,j))
        dh = zero
        do k=1,kbvs-1
          dh=dh+h(mmk(k,i,j))-h(mmk(k,i+1,j))
          div( k )=div( k )+wghthar*dh*covuidy
          div(k+1)=div(k+1)-wghthar*dh*covuidy
        end do


        !- levelling at bottom steps, this part smoothes the water-transfer at 
        !- bottom steps, i.e. there where a water column is deeper than the 
        !- neighbour-water-column.

        kbw = kh(mmk(1,i,j-1))
        if (kb > kbw .and. kbw > 0) then
          dh = ti(mmk(1,i,j-1))+z(mmk(1,i,j-1))
          do k=1,kbw
            dh = dh-h(mmk(k,i,j))
          end do
          dh = max(dh,zero)
          div(kbw)   = div(kbw)   - gammadt*dh
          div(kbw+1) = div(kbw+1) + gammadt*dh
        end if

        kbe = kh(mmk(1,i,j+1))
        if (kb > kbe .and. kbe > 0) then
          dh = ti(mmk(1,i,j+1))+z(mmk(1,i,j+1))
          do k=1,kbe
            dh = dh-h(mmk(k,i,j))
          end do
          dh = max(dh,zero)
          div(kbe)   = div(kbe)   - gammadt*dh
          div(kbe+1) = div(kbe+1) + gammadt*dh
        end if

        kbn = kh(mmk(1,i-1,j))
        if (kb > kbn .and. kbn > 0) then
          dh = ti(mmk(1,i-1,j))+z(mmk(1,i-1,j))
          do k=1,kbn
            dh = dh-h(mmk(k,i,j))
          end do
          dh = max(dh,zero)
          div(kbn)   = div(kbn)   - gammadt*dh
          div(kbn+1) = div(kbn+1) + gammadt*dh
        end if

        kbs = kh(mmk(1,i+1,j))
        if (kb > kbs .and. kbs > 0) then
          dh = ti(mmk(1,i+1,j))+z(mmk(1,i+1,j))
          do k=1,kbs
            dh = dh-h(mmk(k,i,j))
          end do
          dh = max(dh,zero)
          div(kbs)   = div(kbs)   - gammadt*dh
          div(kbs+1) = div(kbs+1) + gammadt*dh
        end if

        norm = one
        scal = zero
        iterat = 0
      
        a_h(1:kmax)   = zero
        b_h(1:kmax)   = zero
        c_h(1:kmax)   = zero
        e_h(1:kmax)   = zero
        r_h(1:kmax+1) = zero
        z_h(1:kmax+1) = zero
      
        it_max = 25 * kb

        !- Now the last term of the equation which is to be solved will be 
        !- integrated:
        !- dt*(wghteta+(wghtturb*avv))*(length_b/h_tmp( k )-length_a/h_tmp(k-1))
        !- is the parametrisation of the term d/ds * (d/ds ( mu / (dz/ds) ) ). 
        !- This parametrisation prevents the case that one layer disappears, 
        !- i.e. the layer thickness becomes zero, because this term trys always 
        !- to push the thicknesses in the origin state.

        !- To solve the whole equation of continuity over the vertical the 
        !- Newton iteration is used.

        if (depth_n > zero) then

          iteration: do while ( norm > norm_min .and. scal <= scal_max )

            do k=1,kb
              c_h(k) = one+wghtz
              e_h(k) = (one+wghtz)*(h(mmk(k,i,j))-h_tmp(k)) - dt*div(k)
            end do

            do k=1, kb
              e_h(1) = e_h(1) - dt*wghtz*div(k)
            enddo

            do k=2,kb
              !- the layer thicknesses h0 act like a reference thickness or
              !- resting state
              length_a = h0(mmk(k-1,i,j))
              length_b = h0(mmk(k,i,j))

              h_tmp_a  = h_tmp(k-1)
              h_tmp_b  = h_tmp(k)

              fac = wghteta + wghtturb*avv(mmk(k,i,j))
    
              a_h( k ) =        - dt*fac*length_a/(h_tmp_a*h_tmp_a)
              c_h( k ) = c_h(k) + dt*fac*length_b/(h_tmp_b*h_tmp_b)
              e_h( k ) = e_h(k) + dt*fac*( length_b/h_tmp( k )                 &
                                - length_a/h_tmp(k-1) )

              b_h(k-1) =         - dt*fac*length_b/(h_tmp_b*h_tmp_b)
              c_h(k-1) = c_h(k-1)+ dt*fac*length_a/(h_tmp_a*h_tmp_a)
              e_h(k-1) = e_h(k-1)+ dt*fac*( length_a/h_tmp(k-1)                &
                                 - length_b/h_tmp( k ) )
            end do


            !- line inversion (factorization) method to compute iterative 
            !- increments of cell thickness
            !- solved by using the Thomas algorithm the system of equations 
            !- c_h(k)*h_inc(k)+a_h(k)*h_inc(k-1)+b_h(k)*h_inc(k+1)=e_h(k) to
            !- get the  increment h_inc of the layer thicknesses to calculate 
            !- the new h(:) (hn(:)). 

            do k=1,kb
              r_h(k+1)=(e_h(k)-a_h(k)*r_h(k))/(c_h(k)+a_h(k)*z_h(k))
              z_h(k+1)=-b_h(k)/(c_h(k)+a_h(k)*z_h(k))
            end do

            h_inc(kb)=r_h(kb+1)
            do k=kb-1,1,-1
              h_inc(k)=z_h(k+1)*h_inc(k+1)+r_h(k+1)
            end do

            norm = zero
            scal = one
            do k=1,kb
              norm = max(norm,abs(h_inc(k)))
              scal = max(scal,-two*(h_inc(k)/h_tmp(k)))
            end do

            do k=1,kb
              h_inc(k) = h_inc(k) / scal
              h_tmp(k) = h_tmp(k) + h_inc(k)
            enddo

            iterat = iterat + 1
            if ( iterat > it_max ) call exitme(1,'Undue number of iterations')

          end do iteration

        elseif ( depth_n <= zero ) then
          !- In this case the water column would run originally dry - so set
          !- all layersthicknesses to 5 cm and write a WARNING to logfile.
          !- Most probably a coarse grid cell fell dry, which is also covered 
          !- by a finer grid, so that the real cell thickness (at the end of the
          !- timestep comes from fine grid anyway! 
          write(iu06,*) 'WARNING: In Grid ', ia, ' a column ran empty: line: ',&
                         i, ' column: ', j,' Number of layers: ', kb

          do k=1,kb
            h_tmp(k) = hgr
          enddo
        endif

        !- Calculate the pseudo vertical velocity by using the new layer 
        !- thicknesses.
        w_pseudo = zero
        do k=kb,1,-1
          w_pseudo = w_pseudo - (diverg(k) + (h_tmp(k)-h(mmk(k,i,j)))/dt )
          wpsd(k) = w_pseudo
        end do

        dh = zero
        do k=1,kb
          mi = mmk(k, i, j)
          hn(mi) = h_tmp(k)
          ws(mi) = wpsd(k)
          dh = dh + h_tmp(k)
        end do
        ws(mmk(1,i,j)) = zero
      
        !- Calculate the new sea level zn.      
        zn(mmk(1,i,j))= dh - ti(mmk(1,i,j))
            
      end do
      

end subroutine masseqs_solver_dyn


end module masseqs
