module metsubs
  use constants, only : zero, qrt, half, one

  implicit none
  private

  integer(4), private, save     :: lunmet, iwna2, lunprecip
  real(8), allocatable, private :: NOAmet(:)

  private :: get_n2d_noa, meteo_stat

  public  :: metinit, metread, metsave, metnull, metclose, wspang,             &
             meteo_interp_t

contains


!-------------------------------------------------------------------------------
!     open the met-forcings file
!-------------------------------------------------------------------------------
subroutine metinit(mdatei)

  use dmi_mpi_global, only : iu06
  use exits,          only : exitme
  use io_subs,        only : io_new_unit
  use constants,      only : ReadNOA, precevap

  implicit none

  character(11), intent(inout) :: mdatei

  integer(4)    :: ios
  character(11) :: pdatei


  !- atmospherical forcing file ------------------------------------------------
  lunmet = io_new_unit()
  open(lunmet,file=mdatei,form='unformatted',status='old',iostat=ios)
  if (ios /= 0) call exitme(1, 'Could not open '//mdatei)

  read(lunmet,iostat=ios) mdatei
  if (ios /= 0) then
    write(iu06,*) 'I/O problem - iostat is:',ios
    call exitme(1, 'Could not read mdatei in met data file')
  endif
  write(iu06,'(a7,a11)') 'mdatei=',mdatei

  !- do we need to skip the NOA data in met file?
  if (ReadNOA) then
    call get_n2d_noa() 
    allocate( NOAmet(0:iwna2) )
  endif


  !- precipitation file --------------------------------------------------------
  if (precevap) then
    pdatei = 'P'//trim(mdatei(2:11))
    lunprecip = io_new_unit()
    open(lunprecip,file=pdatei,form='unformatted',status='old',iostat=ios)
    if (ios /= 0) call exitme(1, 'Could not open '//pdatei)

    read(lunprecip,iostat=ios) pdatei
    if (ios /= 0) then
      write(iu06,*) 'I/O problem - iostat is:',ios
      call exitme(1,'Could not read pdatei in precip file')
    endif
    write(iu06,'(a7,a11)') 'pdatei=',pdatei
  endif

end subroutine metinit


!-------------------------------------------------------------------------------
!     read the met-forcings
!-------------------------------------------------------------------------------
subroutine metread(cmet,tmet,narea,pl,wu,wv,atemp,humid,cloud,precip)

  use dmi_mpi_global, only : iu06
  use exits,          only : exitme
  use cmod_mem,       only : cmr1
  use cmod_date,      only : jultimea
  use constants,      only : ReadNOA, precevap
  use cmod_miscvars,  only : lvalidate, printmetstats

  implicit none

  character(19), intent(out)  :: cmet
  real(8),      intent(out)   :: tmet
  integer(4),   intent(in)    :: narea
  type (cmr1),  intent(inout) :: atemp(:), humid(:), cloud(:), wu(:), wv(:)
  type (cmr1),  intent(inout) :: precip(:), pl(:)

  integer(4)    :: ia, ios, n2d
  character(19) :: cprec

  
  !- First, read the met forcing -----------------------------------------------
  read(lunmet,iostat=ios) cmet
  if (ios /= 0) then
      write(iu06,*) 'Error: Reading cmet'
      write(iu06,*) 'Error: The iostat failure was:: ', ios
      call exitme(1,'Meteorological file truncated')
  endif
  call jultimea (cmet,tmet)

  write(iu06,'(a36,a)') 'Atmospheric forcing read, valid at: ',cmet

  !- read met forcing, with or without NOAMOD data first:
  if (ReadNOA) then
    read(lunmet,iostat=ios) NOAmet,(pl(ia)%p,    ia=1,narea)
    if (ios /= 0) then
      write(iu06,*) 'Error: Reading NOAmet - pl'
      write(iu06,*) 'Error: The iostat failure was:: ', ios
      call exitme(1,'Could not read or read past eof met data file')
    endif
    read(lunmet,iostat=ios) NOAmet,(wu(ia)%p,    ia=1,narea)
    if (ios /= 0) then
      write(iu06,*) 'Error: Reading NOAmet - wu'
      write(iu06,*) 'Error: The iostat failure was:: ', ios
      call exitme(1,'Could not read or read past eof met data file')
    endif
    read(lunmet,iostat=ios) NOAmet,(wv(ia)%p,    ia=1,narea)
    if (ios /= 0) then
      write(iu06,*) 'Error: Reading NOAmet wv'
      write(iu06,*) 'Error: The iostat failure was:: ', ios
      call exitme(1,'Could not read or read past eof met data file')
    endif
    read(lunmet,iostat=ios) NOAmet,(atemp(ia)%p, ia=1,narea)
    if (ios /= 0) then
      write(iu06,*) 'Error: Reading NOAmet atemp'
      write(iu06,*) 'Error: The iostat failure was:: ', ios
      call exitme(1,'Could not read or read past eof met data file')
    endif
    read(lunmet,iostat=ios) NOAmet,(humid(ia)%p, ia=1,narea)
    if (ios /= 0) then
      write(iu06,*) 'Error: Reading NOAmet humid'
      write(iu06,*) 'Error: The iostat failure was:: ', ios
      call exitme(1,'Could not read or read past eof met data file')
    endif
    read(lunmet,iostat=ios) NOAmet,(cloud(ia)%p, ia=1,narea)
    if (ios /= 0) then
      write(iu06,*) 'Error: Reading NOAmet cloud'
      write(iu06,*) 'Error: The iostat failure was:: ', ios
      call exitme(1,'Could not read or read past eof met data file')
    endif
  else
    read(lunmet,iostat=ios) (pl(ia)%p,    ia=1,narea)
    if (ios /= 0) then
      write(iu06,*) 'Error: pl'
      write(iu06,*) 'Error: The iostat failure was:: ', ios
      call exitme(1,'Could not read or read past eof met data file')
    endif
    read(lunmet,iostat=ios) (wu(ia)%p,    ia=1,narea)
    if (ios /= 0) then
      write(iu06,*) 'Error: wu'
      write(iu06,*) 'Error: The iostat failure was:: ', ios
      call exitme(1,'Could not read or read past eof met data file')
    endif
    read(lunmet,iostat=ios) (wv(ia)%p,    ia=1,narea)
    if (ios /= 0) then
      write(iu06,*) 'Error: wv'
      write(iu06,*) 'Error: The iostat failure was:: ', ios
      call exitme(1,'Could not read or read past eof met data file')
    endif
    if (ios /= 0) call exitme(1,'Could not read or read past eof met data file')
    read(lunmet,iostat=ios) (atemp(ia)%p, ia=1,narea)
    if (ios /= 0) then
      write(iu06,*) 'Error: atemp'
      write(iu06,*) 'Error: The iostat failure was:: ', ios
      call exitme(1,'Could not read or read past eof met data file')
    endif
    read(lunmet,iostat=ios) (humid(ia)%p, ia=1,narea)
    if (ios /= 0) then
      write(iu06,*) 'Error: humid'
      write(iu06,*) 'Error: The iostat failure was:: ', ios
      call exitme(1,'Could not read or read past eof met data file')
    endif
    read(lunmet,iostat=ios) (cloud(ia)%p, ia=1,narea)
    if (ios /= 0) then
      write(iu06,*) 'Error: cloud'
      write(iu06,*) 'Error: The iostat failure was:: ', ios
      call exitme(1,'Could not read or read past eof met data file')
    endif
  endif
 

  !- Then, read the precipitation ----------------------------------------------
  if (precevap) then
    read(lunprecip,iostat=ios) cprec
    if (ios /= 0) call exitme(1,'Precipitation file truncated')
    
    write(iu06,'(a30,a)') 'Precipitation read, valid at: ',cprec

    read(lunprecip,iostat=ios) (precip(ia)%p, ia=1,narea)
    if (ios /= 0) call exitme(1,'Could not read or read past eof, precip file')
  endif


  !- Print stats ---------------------------------------------------------------
  !   Comments: 
  !      -  It is safe to print from here since we are on the I/O task.
  !      -  FIXME: An extra option printmetstats is needed at present so that 
  !         we not brake the nightlybuild before references have been made 
  !         ready; but due to the sloppyness of the caretaker this might take
  !         a while before we are there, sorry for that.
  if (lvalidate .and. printmetstats) then
    do ia=1,narea
      n2d = size( wu(ia)%p ) - 1
      call meteo_stat( n2d, ia, iu06, wu(ia)%p, wv(ia)%p, pl(ia)%p,            &
                       atemp(ia)%p, humid(ia)%p, cloud(ia)%p,                  &
                       precip(ia)%p, precevap)
    enddo
  endif

end subroutine metread


!-------------------------------------------------------------------------------
!     save old met-forcings 
!-------------------------------------------------------------------------------
subroutine metsave(narea, iwet2, st1, dohalo, dd,                              &
                          pln, wun, wvn, atempn, humidn, cloudn,               &
                          pl,  wu,  wv,  atemp,  humid,  cloud,                &
                          precipn, precip                       )
  
  !- modules -------------------------------------------------------------------
  use dmi_mpi,   only : distribution_data
  use dmi_omp,   only : domp_get_domain
  use cmod_mem,  only : cmr1
  use constants, only : precevap

  implicit none

  type (distribution_data), intent(in):: dd(:)
  logical,     intent(in)    :: st1, dohalo
  integer(4),  intent(in)    :: narea, iwet2(:)
  type (cmr1), intent(in)    :: atempn(:),humidn(:),cloudn(:),wun(:),wvn(:)
  type (cmr1), intent(inout) :: atemp (:),humid (:),cloud (:),wu (:),wv (:)
  type (cmr1), intent(in)    :: precipn(:), pln(:)
  type (cmr1), intent(inout) :: precip (:), pl (:)

  integer(4) :: ia, ns, n2dl, n2du, idx1

  if (st1) then
    idx1 = 1
  else
    idx1 = 0
  endif

!$OMP MASTER
  do ia=1,narea
    if (iwet2(ia) <= 0) cycle
    pl(ia)%p(idx1)    = pln(ia)%p(idx1)
    wu(ia)%p(idx1)    = wun(ia)%p(idx1)
    wv(ia)%p(idx1)    = wvn(ia)%p(idx1)
    atemp(ia)%p(idx1) = atempn(ia)%p(idx1)
    humid(ia)%p(idx1) = humidn(ia)%p(idx1)
    cloud(ia)%p(idx1) = cloudn(ia)%p(idx1)

    !- set the halo zone:
    if (dohalo) then
      n2dl = iwet2(ia) + 1
      n2du = iwet2(ia) + dd(ia)%halo2
      do ns=n2dl+idx1,n2du+idx1
        pl(ia)%p(ns) = pln(ia)%p(ns)
        wu(ia)%p(ns) = wun(ia)%p(ns)
        wv(ia)%p(ns) = wvn(ia)%p(ns)
      enddo
    endif
  enddo
  if (precevap) then
    do ia=1,narea
      if (iwet2(ia) <= 0) cycle
      precip(ia)%p(idx1) = precipn(ia)%p(idx1)
    enddo
  endif
!$OMP END MASTER

  do ia=1,narea
    if (iwet2(ia) <= 0) cycle
    call domp_get_domain (1, iwet2(ia), n2dl, n2du)

    do ns=n2dl+idx1,n2du+idx1
      pl(ia)%p(ns)    = pln(ia)%p(ns)
      wu(ia)%p(ns)    = wun(ia)%p(ns)
      wv(ia)%p(ns)    = wvn(ia)%p(ns)
      atemp(ia)%p(ns) = atempn(ia)%p(ns)
      humid(ia)%p(ns) = humidn(ia)%p(ns)
      cloud(ia)%p(ns) = cloudn(ia)%p(ns)
    enddo
  enddo

  if (precevap) then
    do ia=1,narea
      if (iwet2(ia) <= 0) cycle
      call domp_get_domain (1, iwet2(ia), n2dl, n2du)
      do ns=n2dl+idx1,n2du+idx1
        precip(ia)%p(ns) = precipn(ia)%p(ns)
      enddo
    enddo
  endif

end subroutine metsave


!-------------------------------------------------------------------------------
!     set up "neutral" met-forcings
!-------------------------------------------------------------------------------
subroutine metnull(pl,wu,wv,atemp,humid,cloud,precip)

  use constants, only : pair, precevap

  implicit none

  real(8), intent(out), dimension(:) :: atemp, humid, cloud, wu, wv, pl, precip

  real(8), parameter :: pneutral = pair*100.0_8
  real(8), parameter :: wneutral = 0.0_8
  real(8), parameter :: tneutral = 10.0_8
  real(8), parameter :: hneutral = 5.0_8
  real(8), parameter :: cneutral = 0.5_8
  real(8), parameter :: precipneutral = 0.0_8


  pl(:) = pneutral

  wu(:) = wneutral
  wv(:) = wneutral

  atemp(:) = tneutral
  humid(:) = hneutral
  cloud(:) = cneutral

  if (precevap) precip(:) =  precipneutral
 
end subroutine metnull


!-------------------------------------------------------------------------------
!     close the met-forcings file
!-------------------------------------------------------------------------------
subroutine metclose()
  use constants, only : precevap

  implicit none

  close(lunmet)

  if (precevap) close(lunprecip)
end subroutine metclose


!-------------------------------------------------------------------------------
!     get No of points in NOAMOD
!-------------------------------------------------------------------------------
subroutine get_n2d_noa()

  use exits,   only : exitme
  use io_subs, only : io_new_unit

  implicit none

  integer(4)              :: nna, mna, lun, ios
  integer(4), allocatable :: mnoa(:,:)
  character(3)            :: ch3

  !- Get NoA grid parameters ---------------------------------------------------
  lun = io_new_unit()
  open(lun,file='data_surge',status='old',iostat=ios)
  if (ios /= 0) call exitme('ERROR: Cannot open data_surge')
  read(lun,'(2X,A3)',iostat=ios) ch3
  if (ios /= 0) call exitme('ERROR: Cannot read or eof in data_surge')
  do while (ch3 /= 'DIM')
    read(lun,'(2X,A3)',iostat=ios) ch3
    if (ios /= 0) call exitme('ERROR: Cannot read or eof in data_surge')
  enddo
  read(lun,'(14X,I4)',iostat=ios) nna
  if (ios /= 0) call exitme('ERROR: Cannot read or eof in data_surge')
  read(lun,'(14X,I4)',iostat=ios) mna
  if (ios /= 0) call exitme('ERROR: Cannot read or eof in data_surge')
  close(lun)

  !- Obtain No. of wet points in NoA model -------------------------------------
  allocate( mnoa(mna,nna) )
  lun = io_new_unit()
  open(lun,file='surge_model_depth',status='old',iostat=ios)
  if (ios /= 0) call exitme('ERROR: Cannot open surge_model_depth')
  read(lun,*,iostat=ios) mnoa
  if (ios /= 0) call exitme('ERROR: Cannot read or eof in surge_model_depth')
  close(lun)

  !- Get No. of surface wet points in NoA model --------------------------------
  iwna2 = maxval(mnoa(:,:))
  deallocate( mnoa )

end subroutine get_n2d_noa

!-------------------------------------------------------------------------------
! Calculation of surface wind forcing
! Original windstress formulation
!-------------------------------------------------------------------------------
subroutine wspang(iwet2,me,ind,taux,tauy,wu,wv,u,v,tair,tsea)

  !- modules -------------------------------------------------------------------
  use dmi_omp,   only : domp_get_domain
  use constants, only : rhoa, CDA, CDB, CDC, CD1A, CD1B, CD1C

  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in)                 :: iwet2, me(0:,0:), ind(:,:)
  real(8),    intent(in),  dimension(0:) :: wu, wv, u, v, tair
  real(8),    intent(in)                 :: tsea(1:,0:)
  real(8),    intent(out), dimension(0:) :: taux, tauy
  
  !- local vars:
  integer(4) :: n2dl, n2du, nn, i, j, ne, ns
  real(8)    :: wuwa, wvwa, wmagu, wmagv, wvu, wuv, tdiff, cdx, cdy
  real(8)    :: weffu, weffv, wemagu, wemagv, weffvu, weffuv
  real(8)    :: wucut, wvcut, tdiffu, tdiffv
  real(8), parameter :: wlo = 2.5_8, wup = 32.5_8, tdiffmax = 5.0_8

  !- parameters:
  !
  !     drag-coefficient a la 'CHARNOCK' 
  !        cd = 0.001e0*(0.7e0 + 0.09e0*betwind)
  !     alternative drag-coefficient
  !        cd = 1.e-3*(0.5e0 + 0.07e0*betwind)
  !
  !     General formulation:
  !        cd = CD0*(CDA + CDB*betwind)
  real(8), parameter :: CD0 = 0.001_8, CD0RHOA = CD0*rhoa
  real(8)            :: CD0RHOACDA, CD0RHOACDB, CD0RHOACDC
  real(8)            :: CD0RHOACD1A, CD0RHOACD1B, CD0RHOACD1C

  CD0RHOACDA = CD0RHOA*CDA
  CD0RHOACDB = CD0RHOA*CDB
  CD0RHOACDC = CD0RHOA*CDC

  CD0RHOACD1A = CD0RHOA*CD1A
  CD0RHOACD1B = CD0RHOA*CD1B
  CD0RHOACD1C = CD0RHOA*CD1C

!$OMP MASTER
  !- play it safe and set tau to zero on land:
  taux(0) = zero
  tauy(0) = zero
!$OMP END MASTER

  call domp_get_domain (1, iwet2, n2dl, n2du)


  do nn=n2dl,n2du

    i = ind(1,nn)
    j = ind(2,nn)

    ne = me(i,j+1)
    ns = me(i+1,j)

    !- wind, current time step, correctly centered at u/v points: 
    wuwa = half*(wu(nn) + wu(ne))
    wvwa = half*(wv(nn) + wv(ns))

    !- Effective wind relative to current:
    weffu = wuwa - u(nn)
    weffv = wvwa - v(nn)

    !- Wind magnitude centered at u and v points, resp.:
    wvu   = half*(wv(nn) + wv(ne))
    wmagu = sqrt(wuwa*wuwa + wvu*wvu)
    wucut = max(wlo,min(wup,wmagu))
    wuv   = half*(wu(nn) + wu(ns))
    wmagv = sqrt(wuv*wuv + wvwa*wvwa)
    wvcut = max(wlo,min(wup,wmagv))

    !- Magnitude of effective wind, centred at u and v points, resp.:
    weffvu = wvu - qrt*(v(nn)+v(ne)+v(me(i-1,j))+v(me(i-1,j+1)))
    wemagu = sqrt(weffu*weffu + weffvu*weffvu)
    weffuv = wuv - qrt*(u(nn)+u(ns)+u(me(i,j-1))+u(me(i+1,j-1)))
    wemagv = sqrt(weffuv*weffuv + weffv*weffv)
    
    !- wind stress calculation, 
    !    Cd  = Cda + Cdb*|W| + Cdc*|W|**2 
    !          + (Cd1a + Cd1b*|W| + Cd1c*|W|**2)*(Tsea - Tair)
    !    tau = Cd*rhoair*|W-u|*(W-u) :
    tdiff  = tsea(1,nn) - tair(nn)
!FIXME: proper space centering of tdiff at u and v points requires
!       access to tsea and tair in the halo-zone.
!   tdiffu = half*(tsea(1,ne) - tair(ne) + tdiff)
    tdiffu = tdiff
    tdiffu = max(-tdiffmax,min(tdiffmax,tdiffu))
!   tdiffv = half*(tsea(1,ns) - tair(ns) + tdiff)
    tdiffv = tdiff
    tdiffv = max(-tdiffmax,min(tdiffmax,tdiffv))
    cdx   = (CD0RHOACDA +(CD0RHOACDB +CD0RHOACDC*wmagu )*wmagu)                &
          + (CD0RHOACD1A+(CD0RHOACD1B+CD0RHOACD1C*wucut)*wucut)*tdiffu
    cdy   = (CD0RHOACDA +(CD0RHOACDB +CD0RHOACDC*wmagv )*wmagv)                &
          + (CD0RHOACD1A+(CD0RHOACD1B+CD0RHOACD1C*wvcut)*wvcut)*tdiffv
    taux(nn) = cdx*wemagu*weffu
    tauy(nn) = cdy*wemagv*weffv

  enddo

end subroutine wspang

  subroutine meteo_interp_t (n2d, n2dh,                                        &
                             atempa, atempn, atemp, humida, humidn, humid,     &
                             clouda, cloudn, cloud, wua, wun, wu, wva, wvn,wv, &
                             pla, pln, pl, precipo, precipn, precip )
 
    !--- time interpolation for meteo forcing ----------------------------------
    !
    ! atemp - air temperature
    ! humid - specific humidity
    ! cloud - total cloud cover
    ! wu    - zonal wind component
    ! wv    - meridional wind component
    ! pl    - air pressure at sea level
    !
    ! *a    - old meteo time step
    ! *n    - new meteo time step
    ! *     - meteo data at ocean model time step
    !---------------------------------------------------------------------------
       
    !- modules -----------------------------------------------------------------
    use constants, only: rzeite, zzeits, precevap
    use dmi_omp,   only: domp_get_domain
       
    !- directives --------------------------------------------------------------
    implicit none
       
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)            :: n2d, n2dh
    real(8),    intent(in)            :: atempa(0:), atempn(0:)
    real(8),    intent(in)            :: humida(0:), humidn(0:)
    real(8),    intent(in)            :: clouda(0:), cloudn(0:)
    real(8),    intent(in)            :: wua(0:), wun(0:)
    real(8),    intent(in)            :: wva(0:), wvn(0:)
    real(8),    intent(in)            :: pla(0:), pln(0:)
    real(8),    intent(in),  optional :: precipo(0:), precipn(0:)
    real(8),    intent(out)           :: atemp(0:)
    real(8),    intent(out)           :: humid(0:)
    real(8),    intent(out)           :: cloud(0:)
    real(8),    intent(out)           :: wu(0:)
    real(8),    intent(out)           :: wv(0:)
    real(8),    intent(out)           :: pl(0:)
    real(8),    intent(out), optional :: precip(0:) 
       
    !- local vars --------------------------------------------------------------
    integer(4)         :: n2dl, n2du, nn, nl, nu
    real(8)            :: fac
 
    fac = zzeits/rzeite
       
!$OMP MASTER
    !- set the values to zero on land:
    atemp(0) = zero
    humid(0) = zero
    cloud(0) = zero
    wu(0)    = zero
    wv(0)    = zero
    pl(0)    = zero
    if( precevap ) precip(0) = zero

    !- set the halo zone:
    nl = n2d + 1
    nu = n2d + n2dh
    wu(nl:nu) = wua(nl:nu) + fac*(wun(nl:nu)-wua(nl:nu))
    wv(nl:nu) = wva(nl:nu) + fac*(wvn(nl:nu)-wva(nl:nu))
    pl(nl:nu) = pla(nl:nu) + fac*(pln(nl:nu)-pla(nl:nu))
!$OMP END MASTER
       
    call domp_get_domain (1, n2d, n2dl, n2du)
        
    do nn = n2dl, n2du
      atemp(nn) = atempa(nn) + fac*(atempn(nn)-atempa(nn))
      humid(nn) = humida(nn) + fac*(humidn(nn)-humida(nn))
      cloud(nn) = clouda(nn) + fac*(cloudn(nn)-clouda(nn))
      wu(nn)    = wua(nn)    + fac*(wun(nn)-wua(nn))
      wv(nn)    = wva(nn)    + fac*(wvn(nn)-wva(nn))
      pl(nn)    = pla(nn)    + fac*(pln(nn)-pla(nn))
      if (precevap) precip(nn) = precipo(nn) + fac*(precipn(nn)-precipo(nn))
    enddo
       
  end subroutine meteo_interp_t

  !-----------------------------------------------------------------------------
  ! Print stats of the meteo vars
  !-----------------------------------------------------------------------------
  subroutine meteo_stat( n2d, ia, iu06, wu, wv, pl, atemp, humid, cloud,       &
                         precip, precevap)

    use io_subs, only : flush_unit

    implicit none

    integer(4), intent(in) :: n2d, ia, iu06
    real(8),    intent(in) :: wu(0:), wv(0:), pl(0:)
    real(8),    intent(in) :: atemp(0:), humid(0:), cloud(0:)
    logical,    intent(in) :: precevap
    real(8),    intent(in) :: precip(0:)

    integer(4) :: i
    real(8)    :: us, vs, ps, as, hs, cs, fac1, fac2

    character(8),  parameter :: fmti1 = '(a33,i5)'
    character(12), parameter :: fmtr1 = '(a33,f16.13)'
    character(22), parameter :: fmtr2 = '(a33,f16.13,5x,f16.13)'
    character(12), parameter :: fmtw1 = '(a33,f16.12)'
    character(22), parameter :: fmtw2 = '(a33,f16.12,5x,f16.12)'
    character(12), parameter :: fmtp1 = '(a33,f16.9 )'
    character(22), parameter :: fmtp2 = '(a33,f16.9 ,5x,f16.9 )'

    write(iu06,'(a72)')                                                        &
      '------------------------------------------------------------------------'
    write(iu06,fmti1) 'Meteo statistics for domain:     ', ia

    us = zero
    vs = zero
    ps = zero
    as = zero
    hs = zero
    cs = zero
    do i=1,n2d
      fac2  = real(i,8)
      fac1  = one/(fac2 + one)
      us = (us*fac2 + wu(i))*fac1  
      vs = (vs*fac2 + wv(i))*fac1  
      ps = (ps*fac2 + pl(i))*fac1  
      as = (as*fac2 + atemp(i))*fac1  
      hs = (hs*fac2 + humid(i))*fac1  
      cs = (cs*fac2 + cloud(i))*fac1  
    enddo

    write(iu06,fmtw1) 'Average value for wu:            ', us
    write(iu06,fmtw1) 'Average value for wv:            ', vs
    write(iu06,fmtp1) 'Average value for pl:            ', ps
    write(iu06,fmtw1) 'Average value for atemp:         ', as
    write(iu06,fmtr1) 'Average value for humid:         ', hs
    write(iu06,fmtr1) 'Average value for cloud:         ', cs

    write(iu06,fmtw2) 'Min and max for wu:              ',                     &
                      minval(wu(1:)), maxval(wu(1:))
    write(iu06,fmtw2) 'Min and max for wv:              ',                     &
                      minval(wv(1:)), maxval(wv(1:))
    write(iu06,fmtp2) 'Min and max for pl:              ',                     &
                      minval(pl(1:)), maxval(pl(1:))
    write(iu06,fmtw2) 'Min and max for atemp:           ',                     &
                      minval(atemp(1:)), maxval(atemp(1:))
    write(iu06,fmtr2) 'Min and max for humid:           ',                     &
                      minval(humid(1:)), maxval(humid(1:))
    write(iu06,fmtr2) 'Min and max for cloud:           ',                     &
                      minval(cloud(1:)), maxval(cloud(1:))

    if (precevap) then
      ps = zero
      do i=1,n2d
        fac2  = real(i,8)
        fac1  = one/(fac2 + one)
        ps = (ps*fac2 + precip(i))*fac1
      enddo
      
      write(iu06,fmtr1) 'Average value for precip:        ', ps
      write(iu06,fmtr2) 'Min and max for precip:          ',                   &
                        minval(precip(1:)), maxval(precip(1:))
    endif

    write(iu06,'(a72)')                                                        &
      '------------------------------------------------------------------------'
    call flush_unit(iu06)

  end subroutine meteo_stat

end module metsubs
