module mom_heights_wrap

  implicit none
  private

  public :: mom_heights, mom_heights_srf, mom_heights_dyn

  real(8), parameter, private :: zero = 0.0_8, half = 0.5_8

contains

subroutine mom_heights( iwet2, mmk, kh, ind, idx, hz, hu, hv )
!-------------------------------------------------------------------------------
!
! First-time setup of hu & hv. 
!
!-------------------------------------------------------------------------------

  !- modules -------------------------------------------------------------------
  use dmi_omp,   only : domp_get_domain
  use constants, only : botfit

  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in)    :: iwet2
  integer(4), intent(in)    :: kh(0:), ind(:,:)
  integer(4), intent(in)    :: mmk(:,0:,0:)          ! (ke,0:mp,0:np)
  real(8), intent(in)       :: hz(0:)        
  integer(4), intent(inout) :: idx(1:,1:)
  real(8), intent(inout)    :: hu(0:), hv(0:)

  !- local vars ----------------------------------------------------------------
  integer(4) :: i, j, k, mi, ee, ss, nsrf, nl, nu, ii, jj, kbot, mbot, khb

!$OMP BARRIER
!$OMP MASTER
  !- initialize on-land values:
  hu(0) = zero
  hv(0) = zero
!$OMP END MASTER
  
  call domp_get_domain(kh, 1, iwet2, nl, nu, idx)

  !- run through all points and calculate layer-height at the
  !  momentum points on the staggered (Arakawa C) grid:
  do nsrf=nl,nu
    i=ind(1,nsrf)
    j=ind(2,nsrf)
    do k=1,kh(nsrf)
      mi = mmk(k,i,j)
      ee = mmk(k,i,j+1)
      ss = mmk(k,i+1,j)
      
      !- X-mom points:
      if (mi > 0 .and. ee > 0) then
        hu(mi) = half*(hz(mi)+hz(ee))
      else
        hu(mi) = zero
      endif

      !- Y-mom points:
      if (mi > 0 .and. ss > 0) then
        hv(mi) = half*(hz(mi)+hz(ss))
      else
        hv(mi) = zero
      endif
    enddo
  enddo

  if (botfit) then
    do nsrf=nl,nu
      i = ind(1,nsrf)
      j = ind(2,nsrf)

      ee = mmk(1,i,j+1)
      if (ee > 0) then
        kbot = max( kh(nsrf), kh(ee) )
        khb  = min( kh(nsrf), kh(ee) )
        if (kbot > khb) then
          mbot = mmk(khb,i,j)
          if (kh(nsrf) > kh(ee) ) then
            jj = j
          else 
            jj = j+1
          endif
          do k=kh(nsrf)+1,kbot
            hu(mbot) = hu(mbot) + half*hz(mmk(k,i,jj))
          enddo
        endif
      endif

      ss = mmk(1,i+1,j)
      if (ss > 0) then
        kbot = max( kh(nsrf), kh(ss) )
        khb  = min( kh(nsrf), kh(ss) )
        if (kbot > khb) then
          mbot = mmk(khb,i,j)
          if (kh(nsrf) > kh(ss) ) then
            ii = i
          else
            ii = i+1
          endif
          do k=kh(nsrf)+1,kbot
            hv(mbot) = hv(mbot) + half*hz(mmk(k,ii,j))
          enddo
        endif
      endif
    enddo
  endif

  !- That's all folks! ---------------------------------------------------------

end subroutine mom_heights

!===============================================================================
  
subroutine mom_heights_srf( n2d, mmk, kh, ind, hz, hu, hv )
!-------------------------------------------------------------------------------
!
! Update hu & hv at surface points. 
! (we do not need updating below surface)
!
!-------------------------------------------------------------------------------

  !- modules -------------------------------------------------------------------
  use dmi_omp,   only : domp_get_domain
  use constants, only : botfit

  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in)    :: n2d, mmk(1:,0:,0:), kh(0:), ind(:,:)
  real(8),    intent(in)    :: hz(0:)
  real(8),    intent(inout) :: hu(0:), hv(0:)

  !- local vars ----------------------------------------------------------------
  integer(4) :: i, j, nsrf, n2dl, n2du, ee, ss, kbot, khb, mbot, ii, jj, k

  call domp_get_domain (1, n2d, n2dl, n2du)

  !- run through all surface points and calculate surface-layer-height at the
  !  momentum points on the staggered (Arakawa C) grid:

  !- X-mom points:
  do nsrf=n2dl,n2du
    i  = ind(1,nsrf)
    j  = ind(2,nsrf)
    ee = mmk(1,i,j+1)
    if (ee == 0) cycle
    hu(nsrf) = half*(hz(nsrf) + hz(ee))
  enddo

  !- Y-mom points:
  do nsrf=n2dl,n2du
    i  = ind(1,nsrf)
    j  = ind(2,nsrf)
    ss = mmk(1,i+1,j)
    if (ss == 0) cycle
    hv(nsrf) = half*(hz(nsrf) + hz(ss)) 
  enddo

  if (botfit) then
    do nsrf=n2dl,n2du
      i = ind(1,nsrf)
      j = ind(2,nsrf)

      ee = mmk(1,i,j+1)
      if (ee > 0) then
        kbot = max( kh(nsrf), kh(ee) )
        khb  = min( kh(nsrf), kh(ee) )
        if (kbot > khb .and. khb == 1) then
          mbot = mmk(khb,i,j)
          if (kh(nsrf) > kh(ee) ) then
            jj = j
          else 
            jj = j+1
          endif
          do k=kh(nsrf)+1,kbot
            hu(mbot) = hu(mbot) + half*hz(mmk(k,i,jj))
          enddo
        endif
      endif

      ss = mmk(1,i+1,j)
      if (ss > 0) then
        kbot = max( kh(nsrf), kh(ss) )
        khb  = min( kh(nsrf), kh(ss) )
        if (kbot > khb .and. khb == 1) then
          mbot = mmk(khb,i,j)
          if (kh(nsrf) > kh(ss) ) then
            ii = i
          else
            ii = i+1
          endif
          do k=kh(nsrf)+1,kbot
            hv(mbot) = hv(mbot) + half*hz(mmk(k,ii,j))
          enddo
        endif
      endif
    enddo
  endif

  !- That's all folks! ---------------------------------------------------------

end subroutine mom_heights_srf

!===============================================================================

subroutine mom_heights_dyn( iwet2, mmk, kh, khu, khv, ind, idx, hz, hu, hv )

  !- modules -------------------------------------------------------------------
  use dmi_omp, only : domp_get_domain

  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in)    :: iwet2
  integer(4), intent(in)    :: kh(0:), khu(0:), khv(0:)
  integer(4), intent(in)    :: ind(:,:)
  integer(4), intent(in)    :: mmk(:,0:,0:)          ! (ke,0:mp,0:np)
  real(8), intent(in)       :: hz(0:)        
  integer(4), intent(inout) :: idx(1:,1:)
  real(8), intent(inout)    :: hu(0:), hv(0:)


  !- local vars ----------------------------------------------------------------
  integer(4)         :: i, j, k, mi, ee, ss, nsrf, nl, nu
  integer(4)         :: miubot, mivbot, kb
  
  call domp_get_domain(kh, 1, iwet2, nl, nu, idx)
  
  !- run through all points and calculate layer-height at the
  !  momentum points on the staggered (Arakawa C) grid:
  do nsrf=nl,nu
    i=ind(1,nsrf)
    j=ind(2,nsrf)
    
    mi = mmk(1,i,j)
    ee = mmk(1,i,j+1)
    ss = mmk(1,i+1,j)
    if (khu(mi) > 0) miubot = mmk(khu(mi),i,j)
    if (khv(mi) > 0) mivbot = mmk(khv(mi),i,j)
    
    kb = max ( kh(mi), kh(ee), kh(ss) )
        
    do k=1,kb
      mi = mmk(k,i,j)
      ee = mmk(k,i,j+1)
      ss = mmk(k,i+1,j)
      
      !- X-mom points:
      if (mi > 0 .and. ee > 0) then
        hu(mi) = half*(hz(mi)+hz(ee))
      elseif (khu(nsrf) > 0) then
        if (mi > 0 .and. ee == 0) then
          hu(miubot) = hu(miubot)+half*hz(mi)
          hu(mi) = zero
        elseif (mi == 0 .and. ee > 0) then
          hu(miubot) = hu(miubot)+half*hz(ee)
        endif
      else
        hu(mi) = zero
      endif

      !- Y-mom points:
      if (mi > 0 .and. ss > 0) then
        hv(mi) = half*(hz(mi)+hz(ss))
      elseif (khv(nsrf) > 0) then
        if (mi > 0 .and. ss == 0) then
          hv(mivbot) = hv(mivbot)+half*hz(mi)
          hv(mi) = zero
        elseif (mi == 0 .and. ss > 0) then
          hv(mivbot) = hv(mivbot)+half*hz(ss)
        endif
      else
        hv(mi) = zero
      endif
    
    enddo
       
  enddo

end subroutine mom_heights_dyn


end module mom_heights_wrap
