module mpi_task_permute
  !-----------------------------------------------------------------------------
  !  This module contains methods and data to do mpi task permutation
  !-----------------------------------------------------------------------------

  use exits,     only : exitme
  use constants, only : zero

  implicit none
  private 

  public :: task_permute, task_permute_all,                                    &
            task_local_arrays, task_global2local_met, task_local_khuv,         &
            task_local_global_compare, task_permute_index, task_local_numa_ft
contains

  !=============================================================================

  subroutine task_permute ( ia, mm, iwet3, iwet2, kh, h,                       &
                            u, v, cmp, ntracers, avv, tke, diss, z, tsoil,     &
                            casus, ice, ts, ui, vi, ti, coldst, lastasstime,   &
                            errvar, kalman, coupler_active, iotamodel, iota )
    !---------------------------------------------------------------------------
    ! Here we do the permutation according to the task-decomposition for the 3D 
    ! vars and some 2D vars, e.g. surface elevation. This is done only once 
    ! during initialisation of the model. 
    !
    ! For portions of the area not covered by the actual MPI task's halo zone, 
    ! information stored in the corresponding sections of the arrays on entry 
    ! is lost on exit, being replaced by "nice" values (like 0).
    !---------------------------------------------------------------------------

    !- modules -----------------------------------------------------------------
    use dmi_mpi, only : dd, mpi_tt, mpi_rank, mpi_size

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: ia
    integer(4), intent(in)    :: iwet2       ! number of wetpoints in surface
    integer(4), intent(in)    :: iwet3       ! number of wetpoints in total
    integer(4), intent(in)    :: mm(0:,0:,:) ! index for wetpoints into 1D
    real(8),    intent(inout) :: h(0:), u(0:), v(0:), cmp(:,0:)
    real(8),    intent(inout) :: avv(0:), tke(0:), diss(0:)
    real(8),    intent(inout) :: z(0:)          ! 2D surface (0:iwet2)
    real(8),    intent(inout) :: ti(0:)         ! (0:iwet2)
    real(8),    intent(inout) :: tsoil(:,:)     ! (1:3,1:iwet2)
    logical,    intent(inout) :: casus(0:)
    real(8),    intent(inout) :: ice(:,0:), ts(0:), ui(0:), vi(0:), iota(:,0:)
    integer(4), intent(in)    :: kh(0:)
    real(8),    intent(inout) :: lastasstime(0:), errvar(0:)
    logical,    intent(in)    :: coldst, kalman, coupler_active, iotamodel
    integer(4), intent(in)    :: ntracers

    !- local vars --------------------------------------------------------------
    integer(4)                            :: i,j,k,mc,kb,ns,ms,nc,iam,ncstart
    integer(4)                            :: il,iu,jl,ju,iw,ihn,hnt,ita,ncstart2
    integer(4)                            :: ncmax
    logical,    allocatable               :: casus_n(:)
    logical                               :: myneighbour
    real(8),    allocatable, dimension(:) :: h_n, u_n, v_n, z_n, ti_n
    real(8),    allocatable, dimension(:) :: turb1, turb2, turb3
    real(8),    allocatable               :: tsoil_n(:,:)
    real(8),    allocatable, dimension(:) :: ts_n, ui_n, vi_n, ssttime, errv
    real(8),    allocatable               :: ice_n(:,:), cmp_n(:,:), iota_n(:,:)
    real(8), parameter                    :: errv0 = 0.4_8

    !- Return if this task has nothing to do in this area ----------------------
    il = dd(ia)%low_i
    iu = dd(ia)%up_i
    jl = dd(ia)%low_j
    ju = dd(ia)%up_j
    if (max(il, iu, jl, ju) == 0) return
    
    iam = mpi_rank + 1

    if (coupler_active) then
      ncmax = 2
    else
      ncmax = ntracers
    endif

    !- Allocate a chunk of temp-storage ----------------------------------------
    allocate( cmp_n(1:ncmax,0:iwet3), tsoil_n(3,iwet2), h_n(0:iwet3),          &
              u_n(0:iwet3), v_n(0:iwet3), casus_n(0:iwet2), z_n(0:iwet2),      &
              ui_n(0:iwet2), vi_n(0:iwet2), ti_n(0:iwet2)   )
    if (.not.coldst) then
      allocate( ice_n(1:3,0:iwet2), ts_n(0:iwet2),                             &
                turb1(0:iwet3), turb2(0:iwet3), turb3(0:iwet3) )
      if (iotamodel) allocate( iota_n(1:3,0:iwet2) )
    endif
    if (kalman) allocate( ssttime(0:iwet2), errv(0:iwet3) )


    !- First, initialize to "nice" values and keep the on-land values ----------

    ! wet-point values:
    z_n(1:)           = zero
    ti_n(1:)          = zero
    h_n(1:)           = zero
    u_n(1:)           = zero
    v_n(1:)           = zero
    ui_n(1:)          = zero
    vi_n(1:)          = zero
    cmp_n(1:ncmax,1:) = zero
    casus_n(1:)       = .true.
    if (.not.coldst) then
      turb1(1:) = zero
      turb2(1:) = zero
      turb3(1:) = zero
      ice_n(:,1:) = zero
      ts_n(1:)    = zero
      if (iotamodel) iota_n(:,1:) = zero
    endif
    if (kalman) then
      ssttime(0:) = zero
      errv(0:)    = errv0
    endif

    ! on-land values:
    z_n(0)           = z(0)
    ti_n(0)          = ti(0)
    h_n(0)           = h(0)
    u_n(0)           = u(0)
    v_n(0)           = v(0)
    ui_n(0)          = ui(0)
    vi_n(0)          = vi(0)
    cmp_n(1:ncmax,0) = cmp(1:ncmax,0)
    casus_n(0)       = casus(0)
    if (.not.coldst) then
      turb1(0) = avv(0)
      turb2(0) = tke(0)
      turb3(0) = diss(0)
      ice_n(:,0) = ice(:,0)
      ts_n(0)    = ts(0)
      if (iotamodel) iota_n(:,0) = iota(:,0)
    endif

    !- Permute the index array itself and all other arrays ---------------------
    !    ns: surface index according to new permuation
    !    ms: surface index according to old permuation
    !    nc: index in the water column below surface, new permutation
    !    mc: index in the water column below surface, old permutation

    !- where should I start counting for this task?
    nc = 0
    do ita=1,mpi_size
      nc = max(nc, mpi_tt(ia,ita)%up_ws)
    enddo
    ncstart2 = nc
    do ita=1,iam-1
      ! run through this tasks region:
      do j=mpi_tt(ia,ita)%low_j,mpi_tt(ia,ita)%up_j
        do i=mpi_tt(ia,ita)%low_i,mpi_tt(ia,ita)%up_i
          ms = mm(i,j,1) 
          if (ms <= 0) cycle
          do k=2,kh(ms)
            nc = nc + 1
          enddo
        enddo
      enddo
    enddo
    ncstart = nc
    !- surface index:
    ns = dd(ia)%low_ws - 1

    !- do this task's region:
    do j=jl,ju
      do i=il,iu
        ms = mm(i,j,1) 
        if (ms <= 0) cycle

        ! handle surface points:
        ns = ns + 1
        kb = kh(ms)

        z_n(ns)     = z(ms)
        ti_n(ns)    = ti(ms)
        h_n(ns)     = h(ms)
        u_n(ns)     = u(ms)
        v_n(ns)     = v(ms)
        ui_n(ns)    = ui(ms)
        vi_n(ns)    = vi(ms)
        casus_n(ns) = casus(ms)
        cmp_n(1:ncmax,ns) = cmp(1:ncmax,ms)
        if (.not.coldst) then
          turb1(ns) = avv(ms)
          turb2(ns) = tke(ms)
          turb3(ns) = diss(ms)
          ice_n(:,ns) = ice(:,ms)
          ts_n(ns)    = ts(ms)
          if (iotamodel) iota_n(:,ns) = iota(:,ms)
        endif

        tsoil_n(1:3,ns) = tsoil(1:3,ms)

        if (kalman) then
          ssttime(ns) = lastasstime(ms)
          errv(ns)    = errvar(ms)
        endif

        ! handle sub-surface points:
        do k=2,kb
          mc = mm(i,j,k)
          nc = nc + 1
          h_n(nc)     = h(mc)
          u_n(nc)     = u(mc)
          v_n(nc)     = v(mc)
          cmp_n(1:ncmax,nc) = cmp(1:ncmax,mc)
          if (.not.coldst) then
            turb1(nc) = avv(mc)
            turb2(nc) = tke(mc)
            turb3(nc) = diss(mc)
          endif
          if (kalman) errv(nc) = errvar(mc)
        enddo
      enddo
    enddo

    !- A little validation would be nice to have -------------------------------
    iw = dd(ia)%up_ws
    if (ns /= iw) then
      call exitme('task_permute not able to reconstruct surface indices')
    endif
    iw = dd(ia)%nwet3 + ncstart - (dd(ia)%up_ws - dd(ia)%low_ws + 1)
    if (nc /= iw) then
      call exitme('task_permute not able to reconstruct indices')
    endif

    !- Set up ALL arrays in the halo zone, too ---------------------------------
    !
    !    task #T and numbers on its eight halo-neighbours
    !
    !       5    2    6
    !
    !       1    T    3
    !
    !       8    4    7
    !
    !    i.e. 1 is west neighbour and 7 is south-east neighbour.
    !
    !    ... apply the indexing of neighbour tasks, count index for ALL tasks
    !    ... both index arrays and data arrays MUST be valid in the halo zone
    nc = ncstart2
    ns = 0
    do ita=1,mpi_size
      ! is this task one of my neighbours?
      myneighbour = .false.
      do ihn=1,8
        hnt = mpi_tt(ia,iam)%hnb(ihn)
        if (hnt == ita) then
          myneighbour = .true.
          exit
        endif
      enddo

      ! run through this tasks region:
      do j=mpi_tt(ia,ita)%low_j,mpi_tt(ia,ita)%up_j
        do i=mpi_tt(ia,ita)%low_i,mpi_tt(ia,ita)%up_i
          ms = mm(i,j,1) 
          if (ms <= 0) cycle

          ! count surface index:
          ns = ns + 1
          kb = kh(ms)

          ! skip permuting unused array sections
          if (.not.myneighbour) then
            ! ... but always count index
            do k=2,kb
              nc = nc + 1
            enddo
          else
            ! first, handle surface points:
            z_n(ns)     = z(ms)
            ti_n(ns)    = ti(ms)
            h_n(ns)     = h(ms)
            u_n(ns)     = u(ms)
            v_n(ns)     = v(ms)
            ui_n(ns)    = ui(ms)
            vi_n(ns)    = vi(ms)
            casus_n(ns) = casus(ms)
            cmp_n(1:ncmax,ns) = cmp(1:ncmax,ms)
            if (.not.coldst) then
              turb1(ns) = avv(ms)
              turb2(ns) = tke(ms)
              turb3(ns) = diss(ms)
              ice_n(:,ns) = ice(:,ms)
              ts_n(ns) = ts(ms)
              if (iotamodel) iota_n(:,ns) = iota(:,ms)
            endif

            tsoil_n(1:3,ns) = tsoil(1:3,ms)

            ! handle sub-surface points:
            do k=2,kb
              mc = mm(i,j,k)
              nc = nc + 1
              h_n(nc)     = h(mc)
              u_n(nc)     = u(mc)
              v_n(nc)     = v(mc)
              cmp_n(1:ncmax,nc) = cmp(1:ncmax,mc)
              if (.not.coldst) then
                turb1(nc) = avv(mc)
                turb2(nc) = tke(mc)
                turb3(nc) = diss(mc)
              endif
            enddo
          endif
        enddo
      enddo
      
    enddo

    !- transfer back to original arrays, include the halo and everything -------
    z(:)     = z_n(:)
    ti(:)    = ti_n(:)
    h(:)     = h_n(:)
    u(:)     = u_n(:)
    v(:)     = v_n(:)
    ui(:)    = ui_n(:)
    vi(:)    = vi_n(:)
    casus(:) = casus_n(:)
    cmp(1:ncmax,:) = cmp_n(1:ncmax,:)
    if (.not.coldst) then
      avv(:) = turb1(:)
      tke(:)  = turb2(:)
      diss(:) = turb3(:)
      ice(:,:) = ice_n(:,:)
      ts(:)    = ts_n(:)
      if (iotamodel) iota(:,:) = iota_n(:,:)
    endif

    tsoil(:,:) = tsoil_n(:,:)

    if (kalman) then
      lastasstime(:) = ssttime(:)
      errvar(:)      = errv(:)
    endif

    !- clean up, please --------------------------------------------------------
    deallocate( cmp_n, u_n, v_n, z_n, h_n, tsoil_n, ti_n, casus_n, ui_n, vi_n )
    if (.not.coldst) then
      deallocate( ice_n, ts_n, turb1, turb2, turb3 )
      if (iotamodel) deallocate( iota_n )
    endif
    if (kalman) deallocate( ssttime, errv )

  end subroutine task_permute

  !=============================================================================

  subroutine task_local_arrays (ia, coldst, coupler_active, ntracers,          &
                                mm_g, mm_l, mmk_l, me_l, ind_l, kh_g, kh_l,    &
                                hz_g, hz_l, h_g, h_l, h_old_l, h_new_l,        &
                                z_g, elev_ref_g, z_l, elev_ref_l, zn_l,        &
                                ti_g, ti_l, t_s_g, t_s_l,                      &
                                u_g, v_g, u_l, v_l, w_l, cmp_g, cmp_l,         &
                                avv_g, avv_l, tke_g, tke_l, diss_g, diss_l,    &
                                casus_g, casus_l, tsn_g, tsn_l, ice_g, ice_l,  &
                                ueis_g, ueis_l, veis_g, veis_l,                &
                                iotamodel, iota_g, iota_l )

    !- modules -----------------------------------------------------------------
    use dmi_mpi,   only : dd
    use constants, only : dyn, elevation_ref

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: ia, ntracers
    logical, intent(in)    :: coldst, coupler_active, iotamodel
    integer(4), intent(in) :: mm_g(0:,0:,1:)         ! global (0:mi+1,0:ni+1,ki)
    integer(4), intent(in) :: kh_g(0:)               ! global (0:iwet2)
    logical, intent(in)    :: casus_g(0:)            ! global (0:iwet2)
    real(8), intent(in)    :: hz_g(0:), avv_g(0:)    ! global (0:iwet3)
    real(8), intent(in)    :: h_g(0:)                ! global (0:iwet3)
    real(8), intent(in)    :: u_g(0:), v_g(0:)       ! global (0:iwet3)
    real(8), intent(in)    :: cmp_g(:,0:)            ! global (ntracers,0:iwet3)
    real(8), intent(in)    :: tke_g(0:), diss_g(0:)  ! global (0:iwet3)
    real(8), intent(in)    :: tsn_g(0:)              ! global (0:iwet2)
    real(8), intent(in)    :: ti_g(0:)               ! global (0:iwet2)
    real(8), intent(in)    :: z_g(0:), elev_ref_g(0:)! global (0:iwet2)
    real(8), intent(in)    :: ueis_g(0:), veis_g(0:) ! global (0:iwet2)
    real(8), intent(in)    :: ice_g(:,0:)            ! global (1:3,0:iwet2)
    real(8), intent(in)    :: t_s_g(:,:)             ! global (1:3,1:iwet2)
    real(8), intent(in)    :: iota_g(:,0:)           ! global (1:3,0:iwet2)

    integer(4), intent(out) :: me_l(0:,0:)         ! local (0:mi+1,0:ni+1)
    integer(4), intent(out) :: mm_l(0:,0:,1:)      ! local (0:mi+1,0:ni+1,ki)
    integer(4), intent(out) :: mmk_l(1:,0:,0:)     ! local (ki,0:mi+1,0:ni+1)
    integer(4), intent(out) :: ind_l(:,:)          ! local (1:2,1:iwet2_l+halo)
    integer(4), intent(out) :: kh_l(0:)            ! local (0:iwet2_l+halo)
    logical, intent(out)    :: casus_l(0:)         ! local (0:iwet2_l+halo)
    real(8), intent(out)    :: hz_l(0:), avv_l(0:) ! local (0:iwet3_l+halo)
    real(8), intent(out)    :: h_l(0:)             ! local (0:iwet3_l+halo)
    real(8), intent(out)    :: z_l(0:), zn_l(0:)   ! local (0:iwet2_l+halo)
    real(8), intent(out)    :: elev_ref_l(0:)      ! local (0:iwet2_l+halo)
    real(8), intent(out)    :: h_old_l(0:), h_new_l(0:) ! local (0:iwet3_l+halo)
    real(8), intent(out)    :: u_l(0:), v_l(0:)    ! local (0:iwet3_l+halo)
    real(8), intent(out)    :: w_l(0:)             ! local (0:iwet3_l)
    real(8), intent(out)    :: cmp_l(:,0:)     ! local (ntracers,0:iwet3_l+halo)
    real(8), intent(out)    :: tke_l(0:), diss_l(0:)  ! local (0:iwet3_l)
    real(8), intent(out)    :: tsn_l(0:)           ! local (0:iwet2_l)
    real(8), intent(out)    :: ti_l(0:)            ! local (0:iwet2_l(+halo))
    real(8), intent(out)    :: ueis_l(0:), veis_l(0:) ! local (0:iwet2_l+halo)
    real(8), intent(out)    :: ice_l(:,0:)         ! local (1:3,0:iwet2_l+halo)
    real(8), intent(out)    :: t_s_l(:,:)          ! local (1:3,1:iwet2_l)
    real(8), intent(out)    :: iota_l(:,0:)        ! local (1:3,0:iwet2_l+halo)

!FIXME: The INTENT(OUT) args indicated as (0:iwet3_l) are not initialized in
!       the halo zone; they should not be used there, but eventually we
!       should assign some nice values

    !- local vars --------------------------------------------------------------
    integer(4) :: il, iu, jl, ju, ilh, iuh, jlh, juh, i, j, k
    integer(4) :: ns, nc, ms, mc, kb, nwet2, ncmax

    if (coupler_active) then
      ncmax = 2
    else
      ncmax = ntracers
    endif

    !- Initialise with nice and useful values ----------------------------------
    me_l(:,:)    = 0
    mm_l(:,:,:)  = 0
    mmk_l(:,:,:) = 0
    kh_l(0)      = 0
    hz_l(0)      = hz_g(0)
    h_l(0)       = h_g(0)
    z_l(0)       = z_g(0)
    elev_ref_l(0)= elev_ref_g(0)
    zn_l(:)      = zero
    h_old_l(0)   = hz_g(0)
    h_new_l(0)   = hz_g(0)
    u_l(0)       = u_g(0)
    v_l(0)       = v_g(0)
    w_l(:)       = zero
    avv_l(0)     = avv_g(0)
    tke_l(0)     = tke_g(0)
    diss_l(0)    = diss_g(0)
    casus_l(0)   = casus_g(0)
    tsn_l(0)     = tsn_g(0)
    ti_l(0)      = ti_g(0)
    ueis_l(0)    = ueis_g(0)
    veis_l(0)    = veis_g(0)
    ice_l(:,0)   = ice_g(:,0)
    cmp_l(1:ncmax,0) = cmp_g(1:ncmax,0)
    iota_l(:,0)  = iota_g(:,0)

    !- Return if this task has nothing to do in this area ----------------------
    il = dd(ia)%low_i
    iu = dd(ia)%up_i
    jl = dd(ia)%low_j
    ju = dd(ia)%up_j
    if (max(il, iu, jl, ju) == 0) return
    !  incl. halo zone:
    ilh = dd(ia)%low_hi
    iuh = dd(ia)%up_hi
    jlh = dd(ia)%low_hj
    juh = dd(ia)%up_hj


    !- do this task's region ---------------------------------------------------
    !    ns: surface index according to local permuation
    !    ms: surface index according to global permuation
    !    nc: index in the water column below surface, local permutation
    !    mc: index in the water column below surface, global permutation
    nwet2 = dd(ia)%up_ws - dd(ia)%low_ws + 1
    ns    = 0
    nc    = nwet2 + dd(ia)%halo2
    do j=jl,ju
      do i=il,iu
        ms = mm_g(i,j,1) 
        if (ms <= 0) cycle

        ! handle surface points:
        ns = ns + 1
        kb = kh_g(ms)

        ind_l(1,ns) = i
        ind_l(2,ns) = j
        kh_l(ns)    = kb

        mm_l(i,j,1)  = ns
        mmk_l(1,i,j) = ns
        casus_l(ns)  = casus_g(ms)
        z_l(ns)      = z_g(ms)
        h_l(ns)      = h_g(ms)
        hz_l(ns)     = hz_g(ms)
        h_old_l(ns)  = hz_g(ms)
        h_new_l(ns)  = hz_g(ms)
        u_l(ns)      = u_g(ms)
        v_l(ns)      = v_g(ms)
        t_s_l(:,ns)  = t_s_g(:,ms)
        ti_l(ns)     = ti_g(ms)
        ueis_l(ns)   = ueis_g(ms)
        veis_l(ns)   = veis_g(ms)
        cmp_l(1:ncmax,ns) = cmp_g(1:ncmax,ms)

        if (.not.coldst) then
          avv_l(ns) = avv_g(ms)
          tke_l(ns)  = tke_g(ms)
          diss_l(ns) = diss_g(ms)
          tsn_l(ns)   = tsn_g(ms)
          ice_l(:,ns) = ice_g(:,ms)
          if (iotamodel) iota_l(:,ns) = iota_g(:,ms)
        endif

        if (elevation_ref) elev_ref_l(ns) = elev_ref_g(ms)

        ! handle sub-surface points:
        do k=2,kb
          nc = nc + 1
          mc = mm_g(i,j,k)

          mm_l(i,j,k)  = nc
          mmk_l(k,i,j) = nc
          h_l(nc)      = h_g(mc)
          hz_l(nc)     = hz_g(mc)
          h_old_l(nc)  = hz_g(mc)
          h_new_l(nc)  = hz_g(mc)
          u_l(nc)      = u_g(mc)
          v_l(nc)      = v_g(mc)
          cmp_l(1:ncmax,nc) = cmp_g(1:ncmax,mc)
          if (.not.coldst) then
            avv_l(nc) = avv_g(mc)
            tke_l(nc)  = tke_g(mc)
            diss_l(nc) = diss_g(mc)
          endif
        enddo
      enddo
    enddo


    !- attempt to validate what we did above -----------------------------------
    if (maxval(mm_l(:,:,:)) /= dd(ia)%nwet3 + dd(ia)%halo2 .and.               &
        nc > nwet2 + dd(ia)%halo2) then
      call exitme('task_local_arrays not able to reconstruct No. of wet points')
    endif
    if (maxval(mm_l(:,:,1)) /= nwet2) then
      call exitme('task_local_arrays not able to reconstruct surface points')
    endif


    !- add the halo at the tail of this task's data ----------------------------
    ns = nwet2
    nc = dd(ia)%nwet3 + dd(ia)%halo2
    !    west + n/w + s/w:
    do j=jlh,jl-1
      do i=ilh,iuh
        ms = mm_g(i,j,1) 
        if (ms <= 0) cycle
        ns = ns + 1
        kb = kh_g(ms)
        ind_l(1,ns)  = i
        ind_l(2,ns)  = j
        kh_l(ns)     = kb
        mm_l(i,j,1)  = ns
        mmk_l(1,i,j) = ns
        casus_l(ns)  = casus_g(ms)
        z_l(ns)      = z_g(ms)
        h_l(ns)      = h_g(ms)
        hz_l(ns)     = hz_g(ms)
        h_old_l(ns)  = hz_g(ms)
        h_new_l(ns)  = hz_g(ms)
        u_l(ns)      = u_g(ms)
        v_l(ns)      = v_g(ms)
        ueis_l(ns)   = ueis_g(ms)
        veis_l(ns)   = veis_g(ms)
        ice_l(:,ns)  = ice_g(:,ms)
        tsn_l(ns)    = tsn_g(ms)
        if (dyn)       ti_l(ns)      = ti_g(ms)
        if (iotamodel) iota_l(:,ns)  = iota_g(:,ms)

        if (elevation_ref) elev_ref_l(ns) = elev_ref_g(ms)

        do k=2,kb
          nc = nc + 1
          mc = mm_g(i,j,k)
          mm_l(i,j,k)  = nc
          mmk_l(k,i,j) = nc
          h_l(nc)      = h_g(mc)
          hz_l(nc)     = hz_g(mc)
          h_old_l(nc)  = hz_g(mc)
          h_new_l(nc)  = hz_g(mc)
          u_l(nc)      = u_g(mc)
          v_l(nc)      = v_g(mc)
        enddo
      enddo
    enddo
    !    east + n/e + s/e:
    do j=ju+1,juh
      do i=ilh,iuh
        ms = mm_g(i,j,1) 
        if (ms <= 0) cycle
        ns = ns + 1
        kb = kh_g(ms)
        ind_l(1,ns)  = i
        ind_l(2,ns)  = j
        kh_l(ns)     = kb
        mm_l(i,j,1)  = ns
        mmk_l(1,i,j) = ns
        casus_l(ns)  = casus_g(ms)
        z_l(ns)      = z_g(ms)
        h_l(ns)      = h_g(ms)
        hz_l(ns)     = hz_g(ms)
        h_old_l(ns)  = hz_g(ms)
        h_new_l(ns)  = hz_g(ms)
        u_l(ns)      = u_g(ms)
        v_l(ns)      = v_g(ms)
        ueis_l(ns)   = ueis_g(ms)
        veis_l(ns)   = veis_g(ms)
        ice_l(:,ns)  = ice_g(:,ms)
        tsn_l(ns)    = tsn_g(ms)
        if (dyn)       ti_l(ns)      = ti_g(ms)
        if (iotamodel) iota_l(:,ns)  = iota_g(:,ms)

        if (elevation_ref) elev_ref_l(ns) = elev_ref_g(ms)

        do k=2,kb
          nc = nc + 1
          mc = mm_g(i,j,k)
          mm_l(i,j,k)  = nc
          mmk_l(k,i,j) = nc
          h_l(nc)      = h_g(mc)
          hz_l(nc)     = hz_g(mc)
          h_old_l(nc)  = hz_g(mc)
          h_new_l(nc)  = hz_g(mc)
          u_l(nc)      = u_g(mc)
          v_l(nc)      = v_g(mc)
        enddo
      enddo
    enddo
    !    north + south:
    do j=jl,ju
      !  north:
      do i=ilh,il-1
        ms = mm_g(i,j,1) 
        if (ms <= 0) cycle
        ns = ns + 1
        kb = kh_g(ms)
        ind_l(1,ns)  = i
        ind_l(2,ns)  = j
        kh_l(ns)     = kb
        mm_l(i,j,1)  = ns
        mmk_l(1,i,j) = ns
        casus_l(ns)  = casus_g(ms)
        z_l(ns)      = z_g(ms)
        h_l(ns)      = h_g(ms)
        hz_l(ns)     = hz_g(ms)
        h_old_l(ns)  = hz_g(ms)
        h_new_l(ns)  = hz_g(ms)
        u_l(ns)      = u_g(ms)
        v_l(ns)      = v_g(ms)
        ueis_l(ns)   = ueis_g(ms)
        veis_l(ns)   = veis_g(ms)
        ice_l(:,ns)  = ice_g(:,ms)
        tsn_l(ns)    = tsn_g(ms)
        if (dyn)       ti_l(ns)      = ti_g(ms)
        if (iotamodel) iota_l(:,ns)  = iota_g(:,ms)

        if (elevation_ref) elev_ref_l(ns) = elev_ref_g(ms)

        do k=2,kb
          nc = nc + 1
          mc = mm_g(i,j,k)
          mm_l(i,j,k)  = nc
          mmk_l(k,i,j) = nc
          h_l(nc)      = h_g(mc)
          hz_l(nc)     = hz_g(mc)
          h_old_l(nc)  = hz_g(mc)
          h_new_l(nc)  = hz_g(mc)
          u_l(nc)      = u_g(mc)
          v_l(nc)      = v_g(mc)
        enddo
      enddo
      !  south:
      do i=iu+1,iuh
        ms = mm_g(i,j,1) 
        if (ms <= 0) cycle
        ns = ns + 1
        kb = kh_g(ms)
        ind_l(1,ns)  = i
        ind_l(2,ns)  = j
        kh_l(ns)     = kb
        mm_l(i,j,1)  = ns
        mmk_l(1,i,j) = ns
        casus_l(ns)  = casus_g(ms)
        z_l(ns)      = z_g(ms)
        h_l(ns)      = h_g(ms)
        hz_l(ns)     = hz_g(ms)
        h_old_l(ns)  = hz_g(ms)
        h_new_l(ns)  = hz_g(ms)
        u_l(ns)      = u_g(ms)
        v_l(ns)      = v_g(ms)
        ueis_l(ns)   = ueis_g(ms)
        veis_l(ns)   = veis_g(ms)
        ice_l(:,ns)  = ice_g(:,ms)
        tsn_l(ns)    = tsn_g(ms)
        if (dyn)       ti_l(ns)      = ti_g(ms)
        if (iotamodel) iota_l(:,ns)  = iota_g(:,ms)

        if (elevation_ref) elev_ref_l(ns) = elev_ref_g(ms)

        do k=2,kb
          nc = nc + 1
          mc = mm_g(i,j,k)
          mm_l(i,j,k)  = nc
          mmk_l(k,i,j) = nc
          h_l(nc)      = h_g(mc)
          hz_l(nc)     = hz_g(mc)
          h_old_l(nc)  = hz_g(mc)
          h_new_l(nc)  = hz_g(mc)
          u_l(nc)      = u_g(mc)
          v_l(nc)      = v_g(mc)
        enddo
      enddo
    enddo


    !- surface index -----------------------------------------------------------
    me_l(:,:) = mm_l(:,:,1)


    !- attempt to validate what we did in the halo zone ------------------------
    if (maxval(mm_l(:,:,:)) /= dd(ia)%nwet3 + dd(ia)%halo3) then
      call exitme('task_local_arrays cannot reconstruct wet points in halo')
    endif
    if (maxval(mm_l(:,:,1)) /= nwet2 + dd(ia)%halo2) then
      call exitme('task_local_arrays cannot reconstruct surface points in halo')
    endif


  end subroutine task_local_arrays

  !=============================================================================
  
  subroutine task_local_khuv(iwet2, mmk, ind, kh, khu, khv, idx, halo2d)
  
    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    !  khu, khv need INTENT(INOUT) due to assignment of fractional segments only
    integer(4), intent(in)    :: iwet2, mmk(:,0:,0:), ind(:,:), kh(0:), halo2d
    integer(4), intent(inout) :: khu(0:), khv(0:) 
    integer(4), intent(inout) :: idx(:,:)
  
    !- local vars --------------------------------------------------------------
    integer(4) :: nl, nu, nsrf, i, j
  
!$OMP BARRIER
!$OMP MASTER
    !- initialize on-land values:
    khu(0) = 0
    khv(0) = 0

    !- initialise the halo:
    if (iwet2 > 0 .and. halo2d > 0) then
      do nsrf=iwet2+1,iwet2+halo2d
        i = ind(1,nsrf)
        j = ind(2,nsrf)
        khu(nsrf) = min(kh(nsrf),kh(mmk(1,i,j+1)))
        khv(nsrf) = min(kh(nsrf),kh(mmk(1,i+1,j)))
      enddo
    endif
!$OMP END MASTER

    call domp_get_domain(kh, 1, iwet2, nl, nu, idx)

    do nsrf=nl,nu
      i = ind(1,nsrf)
      j = ind(2,nsrf)
      khu(nsrf) = min(kh(nsrf),kh(mmk(1,i,j+1)))
      khv(nsrf) = min(kh(nsrf),kh(mmk(1,i+1,j)))
    enddo

  end subroutine task_local_khuv

  !=============================================================================

  subroutine task_local_global_compare (ia, mmk_g, mmk_l, kh_g, kh_l,          &
                                        a_g, a_l, a_dim, checkidx, txt)

    !- modules -----------------------------------------------------------------
    use dmi_mpi_global, only : iu06
    use dmi_mpi,        only : dd
    use io_subs,        only : flush_unit

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: ia, a_dim
    integer(4), intent(in) :: mmk_g(1:,0:,0:), kh_g(0:)
    integer(4), intent(in) :: mmk_l(1:,0:,0:), kh_l(0:)
    real(8), intent(in)    :: a_g(0:)
    real(8), intent(in)    :: a_l(0:)
    logical, intent(in)    :: checkidx
    character(*), intent(in), optional :: txt

    !- local vars --------------------------------------------------------------
    integer(4) :: il, iu, jl, ju, ilh, iuh, jlh, juh, i, j, k, kg, kl, gs, ls
    logical    :: checkdata


    !- Return if this task has nothing to do in this area ----------------------
    il = dd(ia)%low_i
    iu = dd(ia)%up_i
    jl = dd(ia)%low_j
    ju = dd(ia)%up_j
    if (max(il, iu, jl, ju) == 0) return
    !  incl. halo zone:
    ilh = dd(ia)%low_hi
    iuh = dd(ia)%up_hi
    jlh = dd(ia)%low_hj
    juh = dd(ia)%up_hj


    !- At first entry, check the indices ---------------------------------------
    checkdata = .true.
    if (checkidx) then
      do j=jlh,juh
        do i=ilh,iuh
          gs = mmk_g(1,i,j) 
          ls = mmk_l(1,i,j) 

          if (gs <= 0 .and. ls > 0) then
            !-  ooopppssss, I'm trapped, don't know what to do ?!??
            write(iu06,*) 'global land, local water at', ia, i, j
            checkdata = .false.
          endif
          
          if (gs <= 0) cycle

          if (ls <= 0) then
            !- got trapped again ...
            write(iu06,*) 'global water, local land at', ia, i, j
            checkdata = .false.
          endif

          kg = kh_g(gs)
          kl = kh_l(ls)
          if (kl /= kg) then
            !- wha's'appanin' ?
            write(iu06,*) 'number of layers differs at', ia, i, j
            write(iu06,*) 'global, local:', kg, kl
            checkdata = .false.
          endif
        enddo
      enddo    
    endif


    !- No need to compare data if indices differ -------------------------------
    if (.not.checkdata) then
      call flush_unit(iu06)
      return
    endif


    !- Compare land values -----------------------------------------------------
    if (a_l(0) /= a_g(0)) then
      if (present(txt)) then
        write(iu06,*) trim(txt), ': local differs from global at land, area:',ia
      else
        write(iu06,*) 'local differs from global at land, area:',ia
      endif
    endif


    !- do this task's region ---------------------------------------------------
    do j=jlh,juh
      do i=ilh,iuh
        gs = mmk_g(1,i,j) 
        ls = mmk_l(1,i,j) 

        if (gs <= 0) cycle

        kg = kh_g(gs)
          
        ! handle surface points:
        if (a_g(gs) /= a_l(ls)) then
          if (present(txt)) then
            write(iu06,*) trim(txt),': local differs from global at',ia,i,j
          else
            write(iu06,*) 'local differs from global at',ia,i,j
          endif
        endif

        ! handle sub-surface points:
        if (a_dim == 3) then
          do k=2,kg
            if (a_g(mmk_g(k,i,j)) /= a_l(mmk_l(k,i,j))) then
              if (present(txt)) then
                write(iu06,*)trim(txt),': local differs from global at',ia,i,j,k
              else
                write(iu06,*) 'local differs from global at',ia,i,j,k
              endif
            endif
          enddo
        endif
      enddo
    enddo

    call flush_unit(iu06)

  end subroutine task_local_global_compare

  !=============================================================================

  subroutine task_global2local_met ( ia, mmx_g, nmx_g, precevap, me_g, me_l,   &
                                     wu_g, wv_g, pa_g, ta_g, cl_g, hu_g, pr_g, &
                                     wu_l, wv_l, pa_l, ta_l, cl_l, hu_l, pr_l )
    !---------------------------------------------------------------------------
    ! Here we do the extraction of meteorological variables from the global
    ! input arrays, which are in the old permutation order, to the local arrays,
    ! which are permuted to the task ordering.
    !---------------------------------------------------------------------------

    !- modules -----------------------------------------------------------------
     use dmi_mpi, only : dd

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)              :: ia, mmx_g, nmx_g
    logical, intent(in)                 :: precevap
    integer(4), intent(in)              :: me_g(0:,0:)
    integer(4), intent(in)              :: me_l(0:,0:)
    real(8), intent(in),  dimension(0:) :: wu_g,wv_g,pa_g,ta_g,cl_g,hu_g,pr_g
    real(8), intent(out), dimension(0:) :: wu_l,wv_l,pa_l,ta_l,cl_l,hu_l,pr_l

    !- local vars --------------------------------------------------------------
    integer(4) :: il, iu, jl, ju, i, j, ns, ms, is, ilh, iuh, jlh, juh


    !- Return if this task has nothing to do in this area ----------------------
    il = dd(ia)%low_i
    iu = dd(ia)%up_i
    jl = dd(ia)%low_j
    ju = dd(ia)%up_j
    if (max(il, iu, jl, ju) == 0) return


    !- Extract to local, task permuted arrays ----------------------------------
    !   ns: surface index according to new local task permuation
    !   ms: surface index according to new global task permuation
    !   is: surface index according to old input permuation (need be calculated)
    !
    ! include the halo for for vector components wu, wv and air pressure:
    ilh = dd(ia)%low_hi
    iuh = dd(ia)%up_hi
    jlh = dd(ia)%low_hj
    juh = dd(ia)%up_hj
    is  = 0
    do j=1,nmx_g
      do i=1,mmx_g
        ms = me_g(i,j)
        if (ms <= 0) cycle
        is = is + 1
        if (j < jlh .or. j > juh .or. i < ilh .or. i > iuh) cycle
        ns = me_l(i,j)
        if (ns > 0) then
          wu_l(ns) = wu_g(is)
          wv_l(ns) = wv_g(is)
          pa_l(ns) = pa_g(is)
        endif
      enddo
    enddo


    !- Extract to local, task permuted arrays ----------------------------------
    !   ns: surface index according to new local task permuation
    !   ms: surface index according to new global task permuation
    !   is: surface index according to old input permuation (need be calculated)
    !
    ! do not include halo for other scalar meteo vars (airtemp, cloudness,
    ! humidity and possibly precipitation):
    is = 0
    do j=1,nmx_g
      do i=1,mmx_g
        ms = me_g(i,j)
        if (ms <= 0) cycle
        is = is + 1
        if (j < jl .or. j > ju .or. i < il .or. i > iu) cycle
        ns = me_l(i,j)
        if (ns > 0) then
          ta_l(ns) = ta_g(is)
          cl_l(ns) = cl_g(is)
          hu_l(ns) = hu_g(is)
          if (precevap) pr_l(ns) = pr_g(is)
        endif
      enddo
    enddo


    !- set the on-land values --------------------------------------------------
    wu_l(0) = wu_g(0)
    wv_l(0) = wv_g(0)
    pa_l(0) = pa_g(0)
    ta_l(0) = ta_g(0)
    cl_l(0) = cl_g(0)
    hu_l(0) = hu_g(0)
    if (precevap) pr_l(0) = pr_g(0)
    

    !- That's all, folks! ------------------------------------------------------

  end subroutine task_global2local_met

  !=============================================================================

  subroutine task_permute_index ( ia, mi, ni, ki, mm, me, mmk, iwet3, iwet2,   &
                                  kh, khu, khv, hz, ind )
    !---------------------------------------------------------------------------
    ! do task permutation of indices and a few data sets needed globally
    !---------------------------------------------------------------------------

    !- modules -----------------------------------------------------------------
    use dmi_mpi, only : mpi_tt, mpi_size

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: ia, mi, ni, ki
    integer(4), intent(in)    :: iwet2          ! number of wetpoints in surface
    integer(4), intent(in)    :: iwet3          ! number of wetpoints in total
    integer(4), intent(inout) :: kh(0:), khu(0:), khv(0:)
    integer(4), intent(inout) :: mm(0:,0:,:)    ! index for wetpoints into 1D
    integer(4), intent(out)   :: mmk(:,0:,0:)
    integer(4), intent(out)   :: me(0:,0:)      ! index for surface wetpoints
    integer(4), intent(out)   :: ind(:,:)       ! i,j index pairs for wetpoints
    real(8),    intent(inout) :: hz(0:)

    !- local vars --------------------------------------------------------------
    integer(4)              :: i,j,k,mc,ns,ms,nc,ita,kb
    integer(4), allocatable :: mm_n(:,:,:), kh_n(:), khu_n(:), khv_n(:)
    real(8),    allocatable :: hz_n(:)

    !- Allocate a chunk of temp-storage ----------------------------------------
    allocate( mm_n(0:mi+1,0:ni+1,ki), hz_n(0:iwet3), kh_n(0:iwet2),            &
              khu_n(0:iwet2), khv_n(0:iwet2) )

    !- First, initialize to "nice" values and keep the on-land values ----------
    mm_n(:,:,:) = 0

    ! wet-point values:
    hz_n(1:)    = zero
    kh_n(1:)    = 0
    khu_n(1:)   = 0
    khv_n(1:)   = 0

    ! on-land values:
    hz_n(0)    = hz(0)
    kh_n(0)    = kh(0)
    khu_n(0)   = khu(0)
    khv_n(0)   = khv(0)

    !- Permute the index array itself ------------------------------------------
    !    ns: surface index according to new permuation
    !    ms: surface index according to old permuation
    !    nc: index in the water column below surface, new permutation
    !    mc: index in the water column below surface, old permutation
    ns = 0
    nc = iwet2
    do ita=1,mpi_size
      ! run through this tasks region:
      do j=mpi_tt(ia,ita)%low_j,mpi_tt(ia,ita)%up_j
        do i=mpi_tt(ia,ita)%low_i,mpi_tt(ia,ita)%up_i
          ms = mm(i,j,1) 
          if (ms <= 0) cycle

          ! handle surface points:
          ns = ns + 1
          kb = kh(ms)

          mm_n(i,j,1) = ns
          hz_n(ns)    = hz(ms)
          kh_n(ns)    = kb
          khu_n(ns)   = khu(ms)
          khv_n(ns)   = khv(ms)

          ! handle sub-surface points:
          do k=2,kb
            mc = mm(i,j,k)
            nc = nc + 1
            mm_n(i,j,k) = nc
            hz_n(nc)    = hz(mc)
          enddo
        enddo
      enddo
    enddo

    !- A little validation would be nice to have -------------------------------
    if (ns /= iwet2) then
      call exitme('task_permute_index not able to reconstruct surface indices')
    endif
    if (nc /= iwet3) then
      call exitme('task_permute_index cannot reconstruct wet point indices')
    endif

    !- transfer back to original arrays, include the halo and everything -------
    do j=0,ni+1
      do i=0,mi+1
        me(i,j) = mm_n(i,j,1)
        do k=1,ki
          nc         = mm_n(i,j,k)
          mm(i,j,k)  = nc
          mmk(k,i,j) = nc
        enddo
      enddo
    enddo

    do ns=0,iwet2
      kh(ns)    = kh_n(ns)
      khu(ns)   = khu_n(ns)
      khv(ns)   = khv_n(ns)
      hz(ns)    = hz_n(ns)
    enddo
    do nc=iwet2+1,iwet3
      hz(nc)    = hz_n(nc)
    enddo

    do j=1,ni
      do i=1,mi
        ns = me(i,j)
        if (ns > 0) then
          ind(1,ns) = i
          ind(2,ns) = j
        endif
      enddo
    enddo

    !- clean up, please --------------------------------------------------------
    deallocate( mm_n,hz_n,kh_n,khu_n,khv_n )

  end subroutine task_permute_index

  !=============================================================================

  subroutine task_local_numa_ft(ia, kh_g, mm_g, nc, ni, ns, iwet2,             &
                                mmk, mm, me, ind, kh, khu, khv,                &
                                u, v, hz, h, h_old, h_new, cmp, avv, tke, diss,&
                                un, vn, ui, vi, wi, hu, hv, rho, press, geopot,&
                                dispv, eddyh, smu, smv, sm1, sm2, sm3,         &
                                z, elev_ref, zn, w, taux, tauy, casus,         &
                                ueis, veis, ice, z0s, uvdam, tsn, ti, tsoil,   &
                                ompidx, nio, iotamodel, iota, iotac)
    !---------------------------------------------------------------------------
    ! Do a first-touch on some major task-local arrays.
    !   Vars appended with _g are task global vars.
    !   If nothing else is mentioned, all other vars are local.
    !---------------------------------------------------------------------------

    !- modules -----------------------------------------------------------------
    use constants, only : dyn, elevation_ref
    use dmi_mpi,   only : dd
    use dmi_omp,   only : domp_get_domain

    !- arguments ---------------------------------------------------------------
    ! input:
    logical,    intent(in)  :: iotamodel
    integer(4), intent(in)  :: kh_g(0:), mm_g(0:,0:,1:)
    integer(4), intent(in)  :: iwet2, nc, ni, ns, ia, nio
    ! index arrays:
    integer(4), intent(inout) :: mm(0:,0:,1:), mmk(1:,0:,0:), ind(1:,1:)
    integer(4), intent(inout) :: me(0:,0:), kh(0:), khu(0:), khv(0:)
    integer(4), intent(inout) :: ompidx(:,:)
    ! 3D vars:
    real(8),    intent(inout) :: u(0:), v(0:), hz(0:), h(0:)
    real(8),    intent(inout) :: h_new(0:), h_old(0:)
    real(8),    intent(inout) :: un(0:), vn(0:), w(0:), hu(0:), hv(0:)
    real(8),    intent(inout) :: ui(0:), vi(0:), wi(0:)
    real(8),    intent(inout) :: smu(0:), smv(0:)
    real(8),    intent(inout) :: sm1(0:), sm2(0:), sm3(0:), eddyh(0:)
    real(8),    intent(inout) :: tke(0:), diss(0:), avv(0:), rho(0:), press(0:)
    real(8),    intent(inout) :: cmp(1:,0:), dispv(1:,0:)
    real(8),    intent(inout) :: geopot(0:)
    ! surface vars:
    logical,    intent(inout) :: casus(0:)
    real(8),    intent(inout) :: z(0:), elev_ref(0:), zn(0:), z0s(0:)
    real(8),    intent(inout) :: taux(0:), tauy(0:)
    real(8),    intent(inout) :: ueis(0:), veis(0:)
    real(8),    intent(inout) :: ice(1:,0:)   ! local (1:ni,0:iwet2_l+halo)
    real(8),    intent(inout) :: uvdam(1:,0:)
    real(8),    intent(inout) :: iota(1:,0:)
    real(8),    intent(inout) :: iotac(0:)
    ! surface vars, not def in halo:
    real(8),    intent(inout) :: tsn(0:), ti(0:)
    real(8),    intent(inout) :: tsoil(1:,1:) ! local (1:ns,1:iwet2_l)

!   TODO: 
!   consider dxx, dyy, cosphi, ...
! plus vars from meteo file, ...


    !- local variables ---------------------------------------------------------
    integer(4) :: kh_fake(0:iwet2), ind_fake(1:2,1:iwet2)
    integer(4) :: mp, np, nl, nu, k, ng, i, j, il, iu, jl, ju, noff


    !- Return if this task has nothing to do in this area ----------------------
    il = dd(ia)%low_i
    iu = dd(ia)%up_i
    jl = dd(ia)%low_j
    ju = dd(ia)%up_j
    if (max(il, iu, jl, ju) == 0) return

    ! set up kh_fake and ind_fake ----------------------------------------------
    kh_fake(0) = 0
    np         = 0
    do j=jl,ju
      do i=il,iu
        ng = mm_g(i,j,1)
        if (ng > 0) then
          np = np + 1
          kh_fake(np)    = kh_g(ng)
          ind_fake(1,np) = i
          ind_fake(2,np) = j
        endif
      enddo
    enddo

    ! get domain decomposition -------------------------------------------------
    call domp_get_domain(kh_fake, 1, iwet2, nl, nu, ompidx)

    ! get off-set for subsurface layers (for vars def in halo) -----------------
    noff = iwet2 + dd(ia)%halo2 
    np   = 0
    do j=jl,ju
      do i=il,iu
        if (mm_g(i,j,1) > 0) then
          np = np + 1
          if (np >= nl) exit
          do k=2,kh_fake(np)
            noff = noff  + 1
          enddo
        endif
      enddo
    enddo

    ! Do the NUMA first-touch on wet points on this task -----------------------
    !   **ONLY** on wet points inside task; we don't care 'bout land points and 
    !   points outside task, e.g. in halo zone
    mp = noff
    do np=nl,nu
      i = ind_fake(1,np)
      j = ind_fake(2,np)

      ! index arrays:
      mmk(1:,i,j) = 0
      mm(i,j,1:)  = 0
      me(i,j)     = 0
      ind(1:2,np) = 0  
      kh(np)      = 0
      khu(np)     = 0
      khv(np)     = 0

      ! surface vars:
      casus(np)      = .true.
      z0s(np)        = zero 
      z(np)          = zero 
      zn(np)         = zero 
      ti(np)         = zero 
      tsn(np)        = zero 
      taux(np)       = zero 
      tauy(np)       = zero 
      ueis(np)       = zero 
      veis(np)       = zero 
      ice(1:ni,np)   = zero
      uvdam(1:2,np)  = zero
      tsoil(1:ns,np) = zero
      if (elevation_ref) elev_ref(np) = zero
      if (iotamodel) then
        iota(1:nio,np) = zero
        iotac(np) = zero
      end if

      ! 3D vars:
      u(np)          = zero
      v(np)          = zero
      w(np)          = zero
      un(np)         = zero
      vn(np)         = zero
      ui(np)         = zero
      vi(np)         = zero
      wi(np)         = zero
      h(np)          = zero
      hz(np)         = zero
      h_old(np)      = zero
      h_new(np)      = zero
      hu(np)         = zero
      hv(np)         = zero
      avv(np)        = zero
      tke(np)        = zero
      diss(np)       = zero
      rho(np)        = zero
      press(np)      = zero
      smu(np)        = zero
      smv(np)        = zero
      sm1(np)        = zero
      sm2(np)        = zero
      sm3(np)        = zero
      eddyh(np)      = zero
      cmp(1:nc,np)   = zero
      dispv(1:,np)   = zero
      if (dyn) geopot(np) = zero

      ! vars defined in the halo
      !   (need another mp offset for those not def. in halo zone)
      do k=2,kh_fake(np)
        mp = mp + 1
        u(mp)          = zero
        v(mp)          = zero
        w(mp)          = zero
        un(mp)         = zero
        vn(mp)         = zero
        ui(mp)         = zero
        vi(mp)         = zero
        wi(mp)         = zero
        h(mp)          = zero
        hz(mp)         = zero
        h_old(mp)      = zero
        h_new(mp)      = zero
        hu(mp)         = zero
        hv(mp)         = zero
        avv(mp)        = zero
        tke(mp)        = zero
        diss(mp)       = zero
        rho(mp)        = zero
        press(mp)      = zero
        smu(mp)        = zero
        smv(mp)        = zero
        sm1(mp)        = zero
        sm2(mp)        = zero
        sm3(mp)        = zero
        eddyh(mp)      = zero
        cmp(1:nc,mp)   = zero
        dispv(1:,mp)   = zero
        if (dyn) geopot(mp) = zero
      enddo
    enddo
    
  end subroutine task_local_numa_ft

  !=============================================================================
  
  subroutine task_permute_all( ia, mi, ni, ki,iwet3, iwet2, mm_g, mm_l, mmk_g, &
                               mmk_l, me_g, me_l, ind_g, ind_l, kh_g, kh_l,    &
                               hz_g, hz_l, h_g, h_l, h_old_l, h_new_l, ti_g,   &
                               ti_l, elev_ref_g, elev_ref_l, khu, khv)
 
    !- modules -----------------------------------------------------------------
    use dmi_mpi,   only : dd
    use constants, only : elevation_ref
    use dmi_mpi,   only : mpi_tt, mpi_size

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: ia, mi, ni, ki
    integer(4), intent(in)    :: iwet2          ! number of wetpoints in surface
    integer(4), intent(in)    :: iwet3          ! number of wetpoints in total
    integer(4), intent(inout) :: mm_g(0:,0:,1:)   ! global (0:mi+1,0:ni+1,ki)
    integer(4), intent(out)   :: mmk_g(:,0:,0:)
    integer(4), intent(out)   :: me_g(0:,0:)      ! index for surface wetpoints
    integer(4), intent(out)   :: ind_g(:,:)       ! i,j index pair for wetpoints
    integer(4), intent(inout) :: kh_g(0:)             ! global (0:iwet2)
    integer(4), intent(inout) :: khu(0:), khv(0:)
    real(8), intent(inout)    :: hz_g(0:)             ! global (0:iwet3)
    real(8), intent(inout)    :: h_g(0:)              ! global (0:iwet3)
    real(8), intent(inout)    :: ti_g(0:)             ! global (0:iwet2)
    real(8), intent(in)       :: elev_ref_g(0:)       ! global (0:iwet2)
    
    integer(4), intent(out) :: me_l(0:,0:)         ! local (0:mi+1,0:ni+1)
    integer(4), intent(out) :: mm_l(0:,0:,1:)      ! local (0:mi+1,0:ni+1,ki)
    integer(4), intent(out) :: mmk_l(1:,0:,0:)     ! local (ki,0:mi+1,0:ni+1)
    integer(4), intent(out) :: ind_l(:,:)          ! local (1:2,1:iwet2_l+halo)
    integer(4), intent(out) :: kh_l(0:)            ! local (0:iwet2_l+halo)
    real(8), intent(out)    :: hz_l(0:)            ! local (0:iwet3_l+halo)
    real(8), intent(out)    :: h_l(0:)             ! local (0:iwet3_l+halo)
    real(8), intent(out)    :: h_old_l(0:), h_new_l(0:) ! local (0:iwet3_l+halo)
    real(8), intent(out)    :: ti_l(0:)             ! global (0:iwet2)
    real(8), intent(out)    :: elev_ref_l(0:)       ! global (0:iwet2)

    !- local vars --------------------------------------------------------------
    integer(4) :: il, iu, jl, ju, ilh, iuh, jlh, juh, i, j, k
    integer(4) :: ns, nc, ms, mc, kb, nwet2, ncmax
    integer(4) :: ita

    !- local arrays ------------------------------------------------------------
    integer(4), allocatable :: mm_n(:,:,:), kh_n(:), khu_n(:), khv_n(:)
    real(8),    allocatable :: hz_n(:), ti_n(:), h_n(:)

    ncmax = 2
    
    !---------------------------------------------------------------------------
    ! First permute the global indecies and datasets
    !---------------------------------------------------------------------------
    
    !- Allocate a chunk of temp-storage ----------------------------------------
    allocate( mm_n(0:mi+1,0:ni+1,ki), hz_n(0:iwet3), kh_n(0:iwet2),            &
              khu_n(0:iwet2), khv_n(0:iwet2), ti_n(0:iwet2), h_n(0:iwet3) )

    !- First, initialize to "nice" values and keep the on-land values ----------
    mm_n(:,:,:) = 0

    ! wet-point values:
    ti_n(1:)    = zero
    h_n(1:)     = zero
    hz_n(1:)    = zero
    kh_n(1:)    = 0
    khu_n(1:)   = 0
    khv_n(1:)   = 0

    ! on-land values:
    ti_n(0)    = ti_g(0) 
    h_n(0)     = h_g(0)
    hz_n(0)    = hz_g(0)
    kh_n(0)    = kh_g(0)
    khu_n(0)   = khu(0)
    khv_n(0)   = khv(0)

    !- Permute the index array itself ------------------------------------------
    !    ns: surface index according to new permuation
    !    ms: surface index according to old permuation
    !    nc: index in the water column below surface, new permutation
    !    mc: index in the water column below surface, old permutation
    ns = 0
    nc = iwet2
    do ita=1,mpi_size
      ! run through this tasks region:
      do j=mpi_tt(ia,ita)%low_j,mpi_tt(ia,ita)%up_j
        do i=mpi_tt(ia,ita)%low_i,mpi_tt(ia,ita)%up_i
          ms = mm_g(i,j,1) 
          if (ms <= 0) cycle

          ! handle surface points:
          ns = ns + 1
          kb = kh_g(ms)

          mm_n(i,j,1) = ns
          hz_n(ns)    = hz_g(ms)
          ti_n(ns)    = ti_g(ms)
          h_n(ns)     = h_g(ms)
          kh_n(ns)    = kb
          khu_n(ns)   = khu(ms)
          khv_n(ns)   = khv(ms)

          ! handle sub-surface points:
          do k=2,kb
            mc = mm_g(i,j,k)
            nc = nc + 1
            mm_n(i,j,k) = nc
            hz_n(nc)    = hz_g(mc)
            h_n(nc)     = h_g(mc)
          enddo
        enddo
      enddo
    enddo

    !- transfer back to original arrays, include the halo and everything -------
    do j=0,ni+1
      do i=0,mi+1
        me_g(i,j) = mm_n(i,j,1)
        do k=1,ki
          nc           = mm_n(i,j,k)
          mm_g(i,j,k)  = nc
          mmk_g(k,i,j) = nc
        enddo
      enddo
    enddo

    do ns=0,iwet2
      ti_g(ns)  = ti_n(ns)
      h_g(ns)   = h_n(ns)
      kh_g(ns)  = kh_n(ns)
      khu(ns)   = khu_n(ns)
      khv(ns)   = khv_n(ns)
      hz_g(ns)  = hz_n(ns)
    enddo
    do nc=iwet2+1,iwet3
      hz_g(nc)    = hz_n(nc)
      h_g(nc)     = h_n(nc)
    enddo

    do j=1,ni
      do i=1,mi
        ns = me_g(i,j)
        if (ns > 0) then
          ind_g(1,ns) = i
          ind_g(2,ns) = j
        endif
      enddo
    enddo

    !- clean up, please --------------------------------------------------------
    deallocate( mm_n,hz_n,kh_n,khu_n,khv_n, ti_n, h_n )

    !---------------------------------------------------------------------------
    ! Assign the local version of the arrays
    !---------------------------------------------------------------------------

    !- Initialise with nice and useful values ----------------------------------
    me_l(:,:)    = 0
    mm_l(:,:,:)  = 0
    mmk_l(:,:,:) = 0
    kh_l(0)      = 0
    hz_l(0)      = hz_g(0)
    h_l(0)       = h_g(0)
    h_old_l(0)   = hz_g(0)
    h_new_l(0)   = hz_g(0)
    ti_l(0)      = ti_g(0)

    !- Return if this task has nothing to do in this area ----------------------
    il = dd(ia)%low_i
    iu = dd(ia)%up_i
    jl = dd(ia)%low_j
    ju = dd(ia)%up_j
    if (max(il, iu, jl, ju) == 0) return
    !  incl. halo zone:
    ilh = dd(ia)%low_hi
    iuh = dd(ia)%up_hi
    jlh = dd(ia)%low_hj
    juh = dd(ia)%up_hj

    !- do this task's region ---------------------------------------------------
    !    ns: surface index according to local permuation
    !    ms: surface index according to global permuation
    !    nc: index in the water column below surface, local permutation
    !    mc: index in the water column below surface, global permutation
    nwet2 = dd(ia)%up_ws - dd(ia)%low_ws + 1
    ns    = 0
    nc    = nwet2 + dd(ia)%halo2
    do j=jl,ju
      do i=il,iu
        ms = mm_g(i,j,1) 
        if (ms <= 0) cycle

        ! handle surface points:
        ns = ns + 1
        kb = kh_g(ms)

        ind_l(1,ns)  = i
        ind_l(2,ns)  = j
        kh_l(ns)     = kb
        mm_l(i,j,1)  = ns
        mmk_l(1,i,j) = ns
        h_l(ns)      = h_g(ms)
        hz_l(ns)     = hz_g(ms)
        h_old_l(ns)  = hz_g(ms)
        h_new_l(ns)  = hz_g(ms)
        ti_l(ns)     = ti_g(ms)
 
        if (elevation_ref) elev_ref_l(ns) = elev_ref_g(ms)

        ! handle sub-surface points:
        do k=2,kb
          nc = nc + 1
          mc = mm_g(i,j,k)

          mm_l(i,j,k)  = nc
          mmk_l(k,i,j) = nc
          h_l(nc)      = h_g(mc)
          hz_l(nc)     = hz_g(mc)
          h_old_l(nc)  = hz_g(mc)
          h_new_l(nc)  = hz_g(mc)
        enddo
      enddo
    enddo

    !- add the halo at the tail of this task's data ----------------------------
    ns = nwet2
    nc = dd(ia)%nwet3 + dd(ia)%halo2
    !    west + n/w + s/w:
    do j=jlh,jl-1
      do i=ilh,iuh
        ms = mm_g(i,j,1) 
        if (ms <= 0) cycle
        ns = ns + 1
        kb = kh_g(ms)
        ind_l(1,ns)  = i
        ind_l(2,ns)  = j
        kh_l(ns)     = kb
        mm_l(i,j,1)  = ns
        mmk_l(1,i,j) = ns
        h_l(ns)      = h_g(ms)
        hz_l(ns)     = hz_g(ms)
        h_old_l(ns)  = hz_g(ms)
        h_new_l(ns)  = hz_g(ms) 

        if (elevation_ref) elev_ref_l(ns) = elev_ref_g(ms)

        do k=2,kb
          nc = nc + 1
          mc = mm_g(i,j,k)
          mm_l(i,j,k)  = nc
          mmk_l(k,i,j) = nc
          h_l(nc)      = h_g(mc)
          hz_l(nc)     = hz_g(mc)
          h_old_l(nc)  = hz_g(mc)
          h_new_l(nc)  = hz_g(mc)
        enddo
      enddo
    enddo
    !    east + n/e + s/e:
    do j=ju+1,juh
      do i=ilh,iuh
        ms = mm_g(i,j,1) 
        if (ms <= 0) cycle
        ns = ns + 1
        kb = kh_g(ms)
        ind_l(1,ns)  = i
        ind_l(2,ns)  = j
        kh_l(ns)     = kb
        mm_l(i,j,1)  = ns
        mmk_l(1,i,j) = ns
        h_l(ns)      = h_g(ms)
        hz_l(ns)     = hz_g(ms)
        h_old_l(ns)  = hz_g(ms)
        h_new_l(ns)  = hz_g(ms)
        
        if (elevation_ref) elev_ref_l(ns) = elev_ref_g(ms)

        do k=2,kb
          nc = nc + 1
          mc = mm_g(i,j,k)
          mm_l(i,j,k)  = nc
          mmk_l(k,i,j) = nc
          h_l(nc)      = h_g(mc)
          hz_l(nc)     = hz_g(mc)
          h_old_l(nc)  = hz_g(mc)
          h_new_l(nc)  = hz_g(mc)
        enddo
      enddo
    enddo
    !    north + south:
    do j=jl,ju
      !  north:
      do i=ilh,il-1
        ms = mm_g(i,j,1) 
        if (ms <= 0) cycle
        ns = ns + 1
        kb = kh_g(ms)
        ind_l(1,ns)  = i
        ind_l(2,ns)  = j
        kh_l(ns)     = kb
        mm_l(i,j,1)  = ns
        mmk_l(1,i,j) = ns
        h_l(ns)      = h_g(ms)
        hz_l(ns)     = hz_g(ms)
        h_old_l(ns)  = hz_g(ms)
        h_new_l(ns)  = hz_g(ms)
        
        if (elevation_ref) elev_ref_l(ns) = elev_ref_g(ms)

        do k=2,kb
          nc = nc + 1
          mc = mm_g(i,j,k)
          mm_l(i,j,k)  = nc
          mmk_l(k,i,j) = nc
          h_l(nc)      = h_g(mc)
          hz_l(nc)     = hz_g(mc)
          h_old_l(nc)  = hz_g(mc)
          h_new_l(nc)  = hz_g(mc)
        enddo
      enddo
      !  south:
      do i=iu+1,iuh
        ms = mm_g(i,j,1) 
        if (ms <= 0) cycle
        ns = ns + 1
        kb = kh_g(ms)
        ind_l(1,ns)  = i
        ind_l(2,ns)  = j
        kh_l(ns)     = kb
        mm_l(i,j,1)  = ns
        mmk_l(1,i,j) = ns
        h_l(ns)      = h_g(ms)
        hz_l(ns)     = hz_g(ms)
        h_old_l(ns)  = hz_g(ms)
        h_new_l(ns)  = hz_g(ms)

        if (elevation_ref) elev_ref_l(ns) = elev_ref_g(ms)

        do k=2,kb
          nc = nc + 1
          mc = mm_g(i,j,k)
          mm_l(i,j,k)  = nc
          mmk_l(k,i,j) = nc
          h_l(nc)      = h_g(mc)
          hz_l(nc)     = hz_g(mc)
          h_old_l(nc)  = hz_g(mc)
          h_new_l(nc)  = hz_g(mc)
        enddo
      enddo
    enddo

    !- surface index -----------------------------------------------------------
    me_l(:,:) = mm_l(:,:,1)

  end subroutine task_permute_all

end module mpi_task_permute
