module nesting
  use constants, only : zero, one, two, four, qrt, half
  !-----------------------------------------------------------------------------
  !  a collection of methods to do the nesting stuff
  !-----------------------------------------------------------------------------

  implicit none
  private
  
  public  :: copy_f_g, copy_f_g_sfc, mom_c_f, mom_f_c, w_c_f, checknesting,    &
             copy_f_g_ice, brdcopy, copy_f_g_vol, rand_z, masseqs_c_f_dyn,     &
             mom_drying_bnd_f_c

contains 

! ------------------------------------------------------------------------------

  subroutine copy_f_g ( mf,nf,mmkf,mg0,mg,ng0,ng,kgmx,mmkg,igaf,jgaf,dxf,dyf,  &
                        hf,qf,qg,kfg,ncomp)
    ! subroutine for extracting fine grid, please refer to iga,jga 
  
    !- modules -----------------------------------------------------------------
    use dmi_omp,   only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: ncomp
    integer(4), intent(in) :: mg0, mg, ng0, ng
    integer(4), intent(in) :: mf, nf
    integer(4), intent(in) :: kgmx
    integer(4), intent(in) :: mmkg(:,0:,0:),mmkf(:,0:,0:)
    integer(4), intent(in) :: igaf(:),jgaf(:), kfg(1:)
    !     coarse grid field:
    real(8), intent(inout) :: qg(1:,0:)
    !     fine grid field:
    real(8), intent(in)    :: qf(1:,0:)
    real(8), intent(in)    :: dxf(:), hf(0:)
    real(8), intent(in)    :: dyf
  
    
    !- local vars --------------------------------------------------------------
    !  coarse grid indices:
    integer(4) :: i, j, ia, ja, ie, je, k, jlower, jupper, mmc
    !  fine grid indices:
    integer(4) :: ifa, jfa, ife, jfe, iff, jf, k0(kgmx), kf, idx
    !  temporary quatities:
    real(8)    :: vf_s, qf_s(1:ncomp), hhf, vol
  

    ! fine  -->  coarse  -------------------------------------------------------
    ! Note the integer(4) divisions.
    ! We find the coarse grid pts to be corrected below.
    ia = igaf(1) + (igaf(2)/2+igaf(3))/igaf(2)
    ie = igaf(1) + (igaf(2)/2+igaf(3)+mf-1)/igaf(2)
    ja = jgaf(1) + (jgaf(2)/2+jgaf(3))/jgaf(2)
    je = jgaf(1) + (jgaf(2)/2+jgaf(3)+nf-1)/jgaf(2)

    !- let's set up the k_fine offset ------------------------------------------
    k0(1) = 0
    do k=2,kgmx
      k0(k) = k0(k-1) + kfg(k-1)
    enddo
    
    call domp_get_domain(max(ng0,ja), min(ng,je), jlower, jupper)
    
    do j=jlower,jupper
  
      jfa = jgaf(2)*(j-jgaf(1)) + 1
      jfe = jgaf(2)*(j-jgaf(1)+1)
  
      do i=max(mg0,ia),min(mg,ie)
  
        ifa = igaf(2)*(i-igaf(1)) + 1
        ife = igaf(2)*(i-igaf(1)+1)

        do k=1,kgmx
          mmc = mmkg(k,i,j)
          if (mmc == 0) exit ! k-loop
    
          vf_s    = zero
          qf_s(:) = zero
    
          do jf=jfa,jfe
            do iff=ifa,ife
              do kf=max(k0(k)+1,2),k0(k)+kfg(k)
                idx = mmkf(kf,iff,jf)
                if (idx <= 0) exit  ! kf-loop
    
                hhf  = hf(idx)
                vol  = dxf(iff)*dyf*hhf
                vf_s = vf_s + vol
                qf_s(1:ncomp) = qf_s(1:ncomp) + vol*qf(1:ncomp,idx)
              enddo
            enddo
          enddo
    
          ! un-roll kf==1 (i.e. k0(k)==0)
          if (k0(k) == 0) then
            kf = 1
            do jf=jfa,jfe
              do iff=ifa,ife
                idx = mmkf(kf,iff,jf)
                if (idx <= 0) cycle  ! if-loop
    
                hhf  = hf(idx)
                vol  = dxf(iff)*dyf*hhf
                vf_s = vf_s + vol
                qf_s(1:ncomp) = qf_s(1:ncomp) + vol*qf(1:ncomp,idx)
              enddo
            enddo
          endif
    
          if (vf_s <= zero) exit  ! k-loop

          qg(1:ncomp,mmc) = qf_s(1:ncomp)/vf_s
 
        enddo  ! k-loop
 
      enddo
    enddo
  
    !- That's all folks! -------------------------------------------------------
  
  end subroutine copy_f_g

! ------------------------------------------------------------------------------

  subroutine copy_f_g_sfc(mf,nf,mmkf,mg0,mg,ng0,ng,mmkg,igaf,jgaf,dxf,dyf,zf,zg)
    !---------------------------------------------------------------------------
  
    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain
  
    !- directives --------------------------------------------------------------
    implicit none
  
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: mg0, mg, ng0, ng
    integer(4), intent(in) :: mf, nf
    integer(4), intent(in) :: mmkg(:,0:,0:),mmkf(:,0:,0:)
    integer(4), intent(in) :: igaf(:),jgaf(:)
    real(8), intent(in)    :: dxf(:), dyf, zf(0:)
    real(8), intent(inout) :: zg(0:)
  
    
    !- local vars --------------------------------------------------------------
    !  coarse grid indices:
    integer(4) :: i, j, ia, ja, ie, je, idxg, jlower, jupper
    !  fine grid indices:
    integer(4) :: ifa, jfa, ife, jfe, iff, jf, idxf
    !  temporary quatities:
    real(8)    :: af_s, zf_s, aaf, zzf
  
  
    !- find extreme indices:
    ia = igaf(1) + (igaf(2)/2+igaf(3))/igaf(2)
    ie = igaf(1) + (igaf(2)/2+igaf(3)+mf-1)/igaf(2)
    ja = jgaf(1) + (jgaf(2)/2+jgaf(3))/jgaf(2)
    je = jgaf(1) + (jgaf(2)/2+jgaf(3)+nf-1)/jgaf(2)
  
    call domp_get_domain(max(ng0,ja), min(ng,je), jlower, jupper)

    do j=jlower,jupper
      do i=max(mg0,ia),min(mg,ie)
  
        idxg = mmkg(1,i,j)
        if (idxg <= 0) cycle
  
        ifa = igaf(2)*(i-igaf(1)) + 1
        ife = igaf(2)*(i-igaf(1)+1)
        jfa = jgaf(2)*(j-jgaf(1)) + 1
        jfe = jgaf(2)*(j-jgaf(1)+1)
  
        af_s = zero
        zf_s = zero
        do jf=jfa,jfe
          do iff=ifa,ife
  
            idxf = mmkf(1,iff,jf)
            if (idxf > 0) then
              zzf  = zf(idxf)
              aaf  = dxf(iff)*dyf
              af_s = af_s + aaf
              zf_s = zf_s + aaf*zzf
            endif
  
          enddo
        enddo
  
        if (af_s <= zero) cycle
  
        zg(idxg) = zf_s/af_s
  
      enddo
    enddo
  
  end subroutine copy_f_g_sfc

! ------------------------------------------------------------------------------

  subroutine copy_f_g_vol ( mg0,mg,ng0,ng,khg,mmkg,mf,nf,kfmx,mmkf,igaf,jgaf,  &
                            kfg,dxf,dyf,hg,hzg,hf,hzf,tig,zg)
    ! subroutine for extracting fine grid, please refer to iga,jga 
  
    !- modules -----------------------------------------------------------------
    use constants, only : hgr
    use dmi_omp, only   : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: mg0, mg, ng0, ng
    integer(4), intent(in) :: mf, nf, kfmx
    integer(4), intent(in) :: khg(0:)
    integer(4), intent(in) :: mmkg(:,0:,0:),mmkf(:,0:,0:)
    integer(4), intent(in) :: igaf(:),jgaf(:), kfg(1:)
    !     coarse grid field:
    real(8), intent(in)    :: hzg(0:), tig(0:)
    real(8), intent(inout) :: hg(0:), zg(0:)
    !     fine grid field:
    real(8), intent(in)    :: dxf(:)
    real(8), intent(in)    :: hf(0:), hzf(0:)
    real(8), intent(in)    :: dyf
  
    
    !- local vars --------------------------------------------------------------
    !  coarse grid indices:
    integer(4) :: i, j, ia, ja, ie, je, k, jlower, jupper, mmc, kbg, idxg
    !  fine grid indices:
    integer(4) :: ifa, jfa, ife, jfe, iff, jf, kf, kk, idx
    !  temporary quatities:
    real(8)    :: diff, hdiff, area_f, wheight, celldiff
    real(8)    :: voldiff(1:kfmx)
  

    !- find extreme indices:
    ia = igaf(1) + (igaf(2)/2+igaf(3))/igaf(2)
    ie = igaf(1) + (igaf(2)/2+igaf(3)+mf-1)/igaf(2)
    ja = jgaf(1) + (jgaf(2)/2+jgaf(3))/jgaf(2)
    je = jgaf(1) + (jgaf(2)/2+jgaf(3)+nf-1)/jgaf(2)
  
    call domp_get_domain(max(ng0,ja), min(ng,je), jlower, jupper)

    do j=jlower,jupper
      do i=max(mg0,ia),min(mg,ie)
  
        idxg = mmkg(1,i,j)
        if (idxg <= 0) cycle

        kbg        = khg(mmkg(1,i,j))
        wheight    = zero
        diff       = zero
        voldiff(:) = zero
        area_f     = zero

        ifa = igaf(2)*(i-igaf(1)) + 1
        ife = igaf(2)*(i-igaf(1)+1)
        jfa = jgaf(2)*(j-jgaf(1)) + 1
        jfe = jgaf(2)*(j-jgaf(1)+1)

        do jf=jfa,jfe
          do iff=ifa,ife

            !- unroll k = 1 to test if a copy is needed
            kf = 1
            idx = mmkf(kf,iff,jf)            
            if (idx > 0) then
              hdiff = hf(idx) - hzf(idx)
              voldiff(kf) = voldiff(kf) + dxf(iff)*dyf*hdiff
              area_f      = area_f + dxf(iff)*dyf
            endif

            do kf = 2,kfmx
              idx = mmkf(kf,iff,jf)
              if (idx <= 0) exit  ! kf-loop

              hdiff = hf(idx) - hzf(idx)
              voldiff(kf) = voldiff(kf) + dxf(iff)*dyf*hdiff
            enddo
          enddo
        enddo

        if ( area_f > zero ) then
          kk = 1
          do k = 1, kbg-1
            mmc = mmkg(k,i,j)
            celldiff = zero
            do kf = kk, kk-1+kfg(k)
              celldiff   = celldiff + voldiff(kf)
            enddo
            diff   = celldiff/area_f
            hg(mmc) = max(hzg(mmc) + diff, hgr)
            wheight = wheight + hg(mmc)
            kk = kk+kfg(k)
          enddo

          !- unroll k = kbg
          k = kbg
          mmc = mmkg(k,i,j)
          celldiff = zero
          do kf = kk, kfmx
            celldiff   = celldiff + voldiff(kf)
          enddo
          diff   = celldiff/area_f
          hg(mmc) = max(hzg(mmc) + diff, hgr)
          wheight = wheight + hg(mmc)

          zg(idxg) = wheight - tig(idxg)
          
        endif
      enddo
    enddo
    
  end subroutine copy_f_g_vol
  
! ------------------------------------------------------------------------------

  subroutine copy_f_g_ice(mf,nf,mef,mg0,mg,ng0,ng,mmk,igaf,jgaf,dxf,dyf,h,     &
                          icef,iceg,casusg,ts,tsnei)
    !---------------------------------------------------------------------------
  
    !- modules -----------------------------------------------------------------
    use constants, only : rhow, rhoi, cpw, cfreeze, Lice, nicepar
    use dmi_omp,   only : domp_get_domain
  
    !- directives --------------------------------------------------------------
    implicit none
  
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: mg0, mg, ng0, ng
    integer(4), intent(in) :: mf, nf
    integer(4), intent(in) :: mmk(:,0:,0:)
    integer(4), intent(in) :: mef(0:,0:)
    integer(4), intent(in) :: igaf(:),jgaf(:)
    real(8), intent(in)    :: dxf(:), dyf, icef(:,0:)
    real(8), intent(in)    :: h(0:)
    real(8), intent(inout) :: iceg(:,0:), ts(:,0:), tsnei(0:)
    logical, intent(inout) :: casusg(0:)
  
    
    !- local vars --------------------------------------------------------------
    !  coarse grid indices:
    integer(4) :: i, j, ia, ja, ie, je, idxg, jlower, jupper
    !  fine grid indices:
    integer(4) :: ifa, jfa, ife, jfe, iff, jf, idxf
    !  temporary quatities:
    real(8)    :: af_s, if_s(nicepar), aaf
    real(8)    :: tmelt, heat, hc

    !FIXME: make these global such that we for sure use the same everywhere:
    real(8), parameter :: MinIce   = 1.e-6_8
    real(8), parameter :: icecomin = 1.e-3_8
    ! FIXME: add a min value for snowdepth when this var comes into play

    real(8), parameter :: RhoCp = rhow*cpw, lhfRhoi = Lice*rhoi
    
      
    !- find extreme indices:
    ia = igaf(1) + (igaf(2)/2+igaf(3))/igaf(2)
    ie = igaf(1) + (igaf(2)/2+igaf(3)+mf-1)/igaf(2)
    ja = jgaf(1) + (jgaf(2)/2+jgaf(3))/jgaf(2)
    je = jgaf(1) + (jgaf(2)/2+jgaf(3)+nf-1)/jgaf(2)
  
    call domp_get_domain(max(ng0,ja), min(ng,je), jlower, jupper)

    do j=jlower,jupper
      do i=max(mg0,ia),min(mg,ie)
  
        idxg = mmk(1,i,j)
        if (idxg <= 0) cycle
  
        ifa = igaf(2)*(i-igaf(1)) + 1
        ife = igaf(2)*(i-igaf(1)+1)
        jfa = jgaf(2)*(j-jgaf(1)) + 1
        jfe = jgaf(2)*(j-jgaf(1)+1)
  
        af_s    = zero
        if_s(:) = zero
        do jf=jfa,jfe
          do iff=ifa,ife
  
            idxf = mef(iff,jf)
            if (idxf > 0) then
              aaf     = dxf(iff)*dyf
              af_s    = af_s + aaf
              if_s(:) = if_s(:) + aaf*icef(:,idxf)
            endif
  
          enddo
        enddo
  
        if (af_s <= zero) cycle
  
        ! first guess:
        iceg(:,idxg) = if_s(:)/af_s
 
        ! make sure to have ice-model-values consistent with 
        ! thermodynamics (heat budget):
        tmelt = -cfreeze*ts(2,idxg)
        hc    = h(idxg) 

        if (iceg(1,idxg) > zero .and. casusg(idxg)) then
          heat = RhoCp*hc*(ts(1,idxg)-tmelt) - iceg(1,idxg)*lhfRhoi
          if (heat >= zero) then
            casusg(idxg) = .true.
            ts(1,idxg)   = heat/(RhoCp*hc+rhoi*cpw*iceg(1,idxg)) + tmelt
            iceg(1,idxg) = zero
          else
            casusg(idxg) = .false.
            ts(1,idxg)   = tmelt
            iceg(1,idxg) = -heat/lhfRhoi
          endif
        endif

        if (iceg(1,idxg) < MinIce .and. .not.casusg(idxg)) then
          casusg(idxg) = .true.
          ts(1,idxg)   = max(tmelt, ts(1,idxg)-iceg(1,idxg)*lhfRhoi/(RhoCp*hc) )
          iceg(1,idxg) = zero
        endif

        if (casusg(idxg)) then
          iceg(:,idxg) = zero
          tsnei(idxg)  = tmelt
        else
          iceg(nicepar,idxg) = zero
          iceg(2,idxg)       = max(min(iceg(2,idxg),1.0_8),icecomin)
          tsnei(idxg)        = min(tsnei(idxg),tmelt)
        endif

      enddo
    enddo
  
  end subroutine copy_f_g_ice

! ------------------------------------------------------------------------------

  subroutine mom_c_f(n,ke,kdry,mmk,me,dt,ft,rt,ty,tx,gty,gtx,z,u,v,hz,un,vn,   &
                     taux,tauy,pl,press,geopot,rho,avv,casus,ice,ueis,veis,kc, &
                     igaf,jgaf,krzf,mmkc,kfg,ucn,vcn,nz1,nz2,z0srf,            &
                     ilw,iup,jlw,jup)
  !-----------------------------------------------------------------------------
  ! Use momentum equation and coarse grid velocity components to 
  ! obtain fine grid velocity components at border.
  !
  !
  ! The method works with the following border types (as seen from fine grid):
  !   unknown border type:  < 0, no action
  !   not a border:         0, no action
  !   W border:             1
  !   N border:             2
  !   E border:             3
  !   S border:             4
  ! Each border must be enumerated in a specific order in increments of 1:
  !   W: from S to N
  !   N: from W to E
  !   E: from N to S
  !   S: from E to W
  ! Each border must be fully connected in the specifications.
  ! There can be more than one border of each type.
  ! 
  ! Convex corners (as seen from fine grid) need special treatment. The corner 
  ! point (in fine grid) must be defined on BOTH of the joined borders, 
  ! enumerated by an increment of 1 clockwise, eg 
  !   for a W/N corner, where the W border ends at brd-point No. 7, the N border
  !   must start at brd-point No. 8.
  ! 
  ! Concave corners (as seen from fine grid) need no special treatment with the 
  ! present method.
  !
  !
  ! Initiated Nov. 28, 2005, by Per Berg, COI/DMI.
  !-----------------------------------------------------------------------------
  
    !- modules -----------------------------------------------------------------
    use constants,    only : dyn, hgr, rhow, rhoi, cdrag 
    use drying_cases, only : drying_case
    use dmi_omp,      only : domp_get_domain
  
    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: n               ! north-south dimension 
    integer(4), intent(in)    :: ke              ! number of layers
    integer(4), intent(in)    :: kdry
    integer(4), intent(in)    :: mmk(:,0:,0:)    ! index
    integer(4), intent(in)    :: me(0:,0:)       ! index
    real(8),    intent(in)    :: dt              ! zeitschritt
    real(8),    intent(in)    :: ft(:)           ! ft(i):f(i)*dt
    real(8),    intent(in)    :: rt              ! rt: r*dt
    real(8),    intent(in)    :: ty              ! ty: dt/dy
    real(8),    intent(in)    :: tx(:)           ! tx(i): dt/dx(i)
    real(8),    intent(in)    :: gty             ! gty:g*dt/dy
    real(8),    intent(in)    :: gtx(:)          ! gtx(i):g*dt/dx(i)
    real(8),    intent(in)    :: z(0:)
    real(8),    intent(in)    :: u(0:)
    real(8),    intent(in)    :: v(0:)
    real(8),    intent(in)    :: hz(0:)
    real(8),    intent(inout) :: un(0:)
    real(8),    intent(inout) :: vn(0:)
    real(8),    intent(in)    :: taux(0:)
    real(8),    intent(in)    :: tauy(0:)
    real(8),    intent(in)    :: pl(0:)
    logical,    intent(in)    :: casus(0:)
    real(8),    intent(in)    :: ice(1:,0:)
    real(8),    intent(in)    :: ueis(0:)
    real(8),    intent(in)    :: veis(0:)
    real(8),    intent(in)    :: press(0:)
    real(8),    intent(in)    :: geopot(0:)
    real(8),    intent(in)    :: rho(0:)
    real(8),    intent(in)    :: avv(0:)
    real(8),    intent(in)    :: z0srf(0:)
    integer(4), intent(in)    :: igaf(:),jgaf(:)
    integer(4), intent(in)    :: krzf(:,0:)
    integer(4), intent(in)    :: kc
    integer(4), intent(in)    :: kfg(:)
    integer(4), intent(in)    :: mmkc(:,0:,0:)
    real(8),    intent(in)    :: ucn(0:),vcn(0:)
    integer(4), intent(in)    :: nz1, nz2, ilw, iup, jlw, jup
  
    !- local vars --------------------------------------------------------------
    integer(4) :: i,j,kev,keu,k,ibf,ig,jg,kgr,kk,jgw,jge,ign,igs, nl, nu
    integer(4) :: mi,ee,ww,nn,ss,ne,se,sw,nw
    integer(4) :: mie,eee,sse, ncorner, bordertype
  
    real(8) :: tutop,decfu,vnn,vne,eiscol,eiscor,eiscox
    real(8) :: tvtop,eiscoo,eiscou,eiscoy,diffuo,diffuu,diffvo
    real(8) :: diffvu,diffv,decfv,vlu,vzet,vsrf,vss,vww,vse,vsw,vnw,ust
    real(8) :: ustar,vaus,vcon,vpress,sdv
    real(8) :: sdvkev,sdvlum
    real(8) :: diffu,difful,diffur,diffvl,diffvr,diffuv
    real(8) :: ulu,uzet,usrf,urd,vrd,vmi,vee,umi,uee,uww,tmp,tmp2,tmp3,fac
    real(8) :: unn,uss,une,u_se,usw,unw,vst,vstar,ucon,uaus,upress,aurd,avrd
    real(8) :: sdu,sdukeu,hrt,sdulum, zm, zu, zr
    real(8) :: dtd, drhow, dtdr, txi, fti, rth, cdri
    real(8) :: velcoarse1, velcoarse2, velcoarse3, velcoarse4, velcoarse5
  
    real(8) :: tm(ke), vel1_c_f(ke), vel2_c_f(ke), vel3_c_f(ke), vel4_c_f(ke), &
               vel5_c_f(ke), ec(ke), ea(ke+1), alpha(ke), beta(ke), eb(ke)

    real(8), parameter :: laplacefac = 0.001_8
  
    !- intrinsic  --------------------------------------------------------------
    intrinsic  :: sqrt, min, max, abs
  
    !- initialise on every entry -----------------------------------------------
    dtd   = dt*laplacefac
    drhow = one/rhow
    dtdr  = dt/rhow
    rth   = rt*half
    cdri  = cdrag*rhoi
 
    !Fixme: we need to consider an even wet point decompo instead 
    call domp_get_domain(nz1, nz2, nl, nu)

    brdloop: do ibf=nl,nu

      ! border point index (i,j) in fine grid 
      bordertype = krzf(3,ibf)
      if (bordertype == 0) cycle brdloop
      i  = krzf(1,ibf)
      j  = krzf(2,ibf)
      if (mmk(1,i,j) <= 0) cycle brdloop
      if (i < ilw .or. i > iup .or. j < jlw .or. j > jup) cycle

      ! only do corner points once
      if (ibf > 1) then
        ! skip a corner point that has already been handled
        if (krzf(3,ibf-1) > 0  .and.                                           &
            i == krzf(1,ibf-1) .and. j == krzf(2,ibf-1)) cycle
      endif
      ncorner = 0
      if (ibf < nu) then
        ! first-time identification of a convex corner point
        if (krzf(3,ibf+1) > 0  .and.                                           &
            i == krzf(1,ibf+1) .and. j == krzf(2,ibf+1)) then
          ! identify this corner by a code:
          !   unknown corner type:  < 0
          !   not a corner:           0
          !   N/W corner:             1
          !   N/E corner:             2
          !   S/E corner:             3
          !   S/W corner:             4
          ncorner = bordertype
        endif
      endif
  
      ! border point index (ig,jg) in coarse grid 
      ig = igaf(1) + (i-1)/igaf(2)
      jg = jgaf(1) + (j-1)/jgaf(2)
  
      ! assign coarse grid normal velocities at all k-levels
      ! to the west of west border, north of north border, etc.
      ! three grid points (left, central, right)
      
      if (bordertype == 1) then
        ! West border 
        !   extract U(west) from coarse grid.
        jg  = jg - 1
        ign = igaf(1) + (i-2)/igaf(2)
        igs = igaf(1) + (i  )/igaf(2)
        k   = 0
        vel1_c_f(:) = zero
        vel2_c_f(:) = zero
        vel3_c_f(:) = zero
        kgr1: do kgr=1,kc
          velcoarse1 = ucn(mmkc(kgr,ig, jg))
          velcoarse2 = ucn(mmkc(kgr,ign,jg))
          velcoarse3 = ucn(mmkc(kgr,igs,jg))
          do kk=1,kfg(kgr)
            k = k+1
            if (mmk(k,i,j) <= 0) exit kgr1
            vel1_c_f(k) = velcoarse1
            vel2_c_f(k) = velcoarse2
            vel3_c_f(k) = velcoarse3
          enddo
        enddo kgr1
        if (ncorner == 1) then
          ! N/W corner point
          !   extract V(north) from coarse grid.
          k = 0
          vel4_c_f(:) = zero
          vel5_c_f(:) = zero
          kgr1c: do kgr=1,kc
            velcoarse4 = vcn(mmkc(kgr,ig-1,jg+1))
            velcoarse5 = vcn(mmkc(kgr,ig-1,jg))
            do kk=1,kfg(kgr)
              k = k+1
              if (mmk(k,i,j) <= 0) exit kgr1c
              vel4_c_f(k) = velcoarse4
              vel5_c_f(k) = velcoarse5
            enddo
          enddo kgr1c
        endif
  
      elseif (bordertype == 2) then
        ! North border
        !   extract V(north) from coarse grid.
        ig  = ig - 1
        jgw = jgaf(1) + (j-2)/jgaf(2)
        jge = jgaf(1) + (j  )/jgaf(2)
        k   = 0
        vel1_c_f(:) = zero
        vel2_c_f(:) = zero
        vel3_c_f(:) = zero
        kgr2: do kgr=1,kc
          velcoarse1 = vcn(mmkc(kgr,ig,jg))
          velcoarse2 = vcn(mmkc(kgr,ig,jgw))
          velcoarse3 = vcn(mmkc(kgr,ig,jge))
          do kk=1,kfg(kgr)
            k = k+1
            if (mmk(k,i,j) <= 0) exit kgr2
            vel1_c_f(k) = velcoarse1
            vel2_c_f(k) = velcoarse2
            vel3_c_f(k) = velcoarse3
          enddo
        enddo kgr2
        if (ncorner == 2) then
          ! N/E corner point
          !   finish un(mmk(k,i,j)) here by extraction from coarse,
          !   extract U(east) from coarse grid.
          k = 0
          vel4_c_f(:) = zero
          kgr2c: do kgr=1,kc
            velcoarse4   = ucn(mmkc(kgr,ig+1,jg))
            do kk=1,kfg(kgr)
              k  = k+1
              mi = mmk(k,i,j)
              if (mi <= 0) exit kgr2c
              un(mi)      = velcoarse4
              vel4_c_f(k) = velcoarse4
            enddo
          enddo kgr2c
        endif
  
      elseif (bordertype == 3) then
        ! East border 
        !   finish un(mmk(k,i,j)) here by extraction from coarse,
        !   extract U(east) from coarse grid.
        ign = igaf(1) + (i-2)/igaf(2)
        igs = igaf(1) + (i  )/igaf(2)
        k   = 0
        vel1_c_f(:) = zero
        vel2_c_f(:) = zero
        vel3_c_f(:) = zero
        kgr3: do kgr=1,kc
          velcoarse1 = ucn(mmkc(kgr,ig, jg))
          velcoarse2 = ucn(mmkc(kgr,ign,jg))
          velcoarse3 = ucn(mmkc(kgr,igs,jg))
          do kk=1,kfg(kgr)
            k  = k+1
            mi = mmk(k,i,j)
            if (mi <= 0) exit kgr3
            un(mi)      = velcoarse1
            vel1_c_f(k) = velcoarse1
            vel2_c_f(k) = velcoarse2
            vel3_c_f(k) = velcoarse3
          enddo
        enddo kgr3
        if (ncorner == 3) then
          ! S/E corner point
          !   finish vn(mmk(k,i,j)) here by extraction from coarse,
          !   extract V(south) from coarse grid.
          k = 0
          kgr3c: do kgr=1,kc
            velcoarse4 = vcn(mmkc(kgr,ig,jg))
            do kk=1,kfg(kgr)
              k  = k+1
              mi = mmk(k,i,j)
              if (mi <= 0) exit kgr3c
              vn(mi) = velcoarse4
            enddo
          enddo kgr3c
        endif
  
      elseif (bordertype == 4) then
        ! South border
        !   finish vn(mmk(k,i,j)) here by extraction from coarse,
        !   extract V(south) from coarse grid.
        jgw = jgaf(1) + (j-2)/jgaf(2)
        jge = jgaf(1) + (j  )/jgaf(2)
        k   = 0
        vel1_c_f(:) = zero
        vel2_c_f(:) = zero
        vel3_c_f(:) = zero
        kgr4: do kgr=1,kc
          velcoarse1 = vcn(mmkc(kgr,ig,jg))
          velcoarse2 = vcn(mmkc(kgr,ig,jgw))
          velcoarse3 = vcn(mmkc(kgr,ig,jge))
          do kk=1,kfg(kgr)
            k  = k+1
            mi = mmk(k,i,j)
            if (mi <= 0) exit kgr4
            vn(mi)   = velcoarse1
            vel1_c_f(k) = velcoarse1
            vel2_c_f(k) = velcoarse2
            vel3_c_f(k) = velcoarse3
          enddo
        enddo kgr4
        if (ncorner == bordertype) then
          ! S/W corner 
          !   extract U(west) from coarse grid.
          k = 0
          vel4_c_f(:) = zero
          vel5_c_f(:) = zero
          kgr4c: do kgr=1,kc
            velcoarse4 = ucn(mmkc(kgr,ig,  jg-1))
            velcoarse5 = ucn(mmkc(kgr,ig+1,jg-1))
            do kk=1,kfg(kgr)
              k  = k+1
              if (mmk(k,i,j) <= 0) exit kgr4c
              vel4_c_f(k) = velcoarse4
              vel5_c_f(k) = velcoarse5
            enddo
          enddo kgr4c
        endif
  
      endif
  
      ! for east border, shift e/w-index one point westwards
      ! - shifted back again after u-equation
  
      if (bordertype == 3) j = j - 1
  
      ! --- u-equation  --------------------------------------------------------

      keu=0
      do k=1,ke
        mi = mmk(k,i,j)
        ee = mmk(k,i,j+1)
        if (mi <= 0 .or. ee <= 0) exit
        keu   = k
        tm(k) = half*(hz(mi)+hz(ee))
      enddo

      mi = mmk(1,i,j)
      if (keu == 0) un(mi) = zero
      ee = mmk(1,i,j+1)
  
      !- skip trivial cases as well as N/E, S/E corners
      if (keu > 0 .and. ncorner /= 2 .and. ncorner /= 3) then
  
        ! assert(keu>0)
  
        ! --- oberflaechenspannung (used in eb(:) and ec(:) computations) ------
        tmp = half*(z0srf(mi) + z0srf(ee))
        if (casus(me(i,j)).and.casus(ee)) then
          tutop = taux(mi)*dtdr/(tm(1)+tmp)
        else
          ! the lines below are not symm.
          mie = me(i,j  )
          eee = me(i,j+1)
          eiscol = min(max(ice(2,mie),zero),one)
          eiscor = min(max(ice(2,eee),zero),one)
          eiscox = half*(eiscol+eiscor)
          diffu  = ueis(mie)-u(mi)
          difful = diffu + ueis(me(i,j-1)) - u(mmk(1,i,j-1))
          diffur = diffu + ueis(eee)       - u(ee)
          diffvl = veis(me(i-1,j  )) - v(mmk(1,i-1,j)) + veis(mie) - v(mi)
          diffvr = veis(me(i-1,j+1)) - v(mmk(1,i-1,j+1)) + veis(eee) - v(ee)
          diffuv = qrt*( eiscol*sqrt(difful**2+diffvl**2)                      &
                        +eiscor*sqrt(diffur**2+diffvr**2))

          fac = dtdr/(tm(1)+tmp)

          decfu  = cdri*diffuv*fac
          tutop  = (cdri*diffuv*diffu+(one-eiscox)*taux(mi))*fac
        endif
  
        ! begin ea(:), ec(:) computation ---------------------------------------
        ea=zero ! i.e. ea(1)=0
        ec=zero
        do k=1,keu-2,2
          tmp     = tm(k)
          tmp2    = tm(k+1)
          ea(k+1) = dt*(avv(mmk(k+1,i,j))+avv(mmk(k+1,i,j+1)))/(tmp+tmp2)
          ec(k)   = one+(ea(k)+ea(k+1))/tmp
          tmp3    = tm(k+2)
          ea(k+2) = dt*(avv(mmk(k+2,i,j))+avv(mmk(k+2,i,j+1)))/(tmp2+tmp3)
          ec(k+1) = one+(ea(k+1)+ea(k+2))/tmp2
        enddo
        if ((mod(keu,2) /= 0).and.(keu > 0)) then
          ec(keu) = one+(ea(keu)+ea(keu+1))/tm(keu)
        else
          ea(keu)   = dt*(avv(mmk(keu,i,j))+avv(mmk(keu,i,j+1)))               &
                        /(tm(keu-1)+tm(keu))
          ec(keu-1) = one+(ea(keu-1)+ea(keu))/tm(keu-1)
          ec(keu)   = one+(ea(keu)+ea(keu+1))/tm(keu)
        endif
        ! end ea(:) computation ------------------------------------------------
  
        ! begin eb(:) computation ----------------------------------------------
  
        ! --- luftdruckterm  ---------------------------------------------------
        ulu = (pl(ee)-pl(mi))*tx(i)*drhow ! not symm., used in eb(:) computation

        ! zeta:
        zm = z(mi)
        zr = z(ee)

        ! pressure term:
        uzet = gtx(i)*(zr-zm)

        ! collect different surface terms:
        usrf = ulu + uzet
  
        txi = tx(i)
        fti = ft(i)
        eb  = zero

        ! --- calculate eb(k) --------------------------------------------------
        keuloop: do k=1,keu
          nw = mmk(k,i-1,j-1)
          ww = mmk(k,i  ,j-1)
          sw = mmk(k,i+1,j-1)
          nn = mmk(k,i-1,j)
          mi = mmk(k,i  ,j)
          ss = mmk(k,i+1,j)
          ne = mmk(k,i-1,j+1)
          ee = mmk(k,i  ,j+1)
          if (ee == 0) exit keuloop
          se = mmk(k,i+1,j+1)
  
          ! -- default nine-point u stencil
          usw = u(sw)
          uss = u(ss)
          u_se = u(se)
          uww = u(ww)
          umi = u(mi)
          uee = u(ee)
          unw = u(nw)
          unn = u(nn)
          une = u(ne)
  
          ! -- default four-point v stencil around umi
          vmi = v(mi)
          vee = v(ee)
          vnn = v(nn)
          vne = v(ne)
  
          ! -- overrule defaults according to border and corner
          if (bordertype == 1) then
            ! -- west border
            unw = vel2_c_f(k)
            uww = vel1_c_f(k)
            usw = vel3_c_f(k)
            if (ncorner == 1) then
              ! -- N/W corner
              vnn = vel4_c_f(k)
              vne = vel4_c_f(k)
              une = uee
              unn = umi
            endif
  
          elseif (bordertype == 2) then
            ! -- north border
            vnn = vel1_c_f(k)
            vne = vel3_c_f(k)
            une = uee
            unn = umi
            unw = uww
  
          elseif (bordertype == 3) then
            ! -- east border
            une  = vel2_c_f(k) 
            uee  = vel1_c_f(k) 
            u_se = vel3_c_f(k)
  
          elseif (bordertype == 4) then
            ! -- south border
            vmi = vel1_c_f(k)
            vee = vel3_c_f(k)
            u_se = uee
            uss = umi
            usw = uww
            if (ncorner == 4) then
              unw = vel4_c_f(k)
              uww = vel4_c_f(k)
              usw = vel5_c_f(k)
            endif
  
          endif
  
          vst = qrt*(vnn+vne+vmi+vee) ! not symm.
  
          !---  barokliner druck -----------------------------------------------
          upress = (press(ee)-press(mi))*txi/(half*(rho(ee)+rho(mi)))
  
          ! convective terms using vector upwind scheme ------------------------
          urd  = umi*txi
          vrd  = vst*ty
          aurd = abs(urd)
          avrd = abs(vrd)

          if (aurd >= avrd) then
            if (urd >= zero) then
              if (vrd >= zero) then
                ucon = aurd*(umi-uww) + avrd*(uww-usw)
              else
                ucon = aurd*(umi-uww) + avrd*(uww-unw)
              endif
            else
              if (vrd >= zero) then
                ucon = aurd*(umi-uee) + avrd*(uee-u_se)
              else
                ucon = aurd*(umi-uee) + avrd*(uee-une)
              endif
            endif
          else
            if (vrd >= zero) then
              if (urd >= zero) then
                ucon = aurd*(uss-usw) + avrd*(umi-uss)
              else
                ucon = aurd*(uss-u_se) + avrd*(umi-uss)
              endif
            else
              if (urd >= zero) then
                ucon = aurd*(unn-usw) + avrd*(umi-unn)
              else
                ucon = aurd*(unn-u_se) + avrd*(umi-unn)
              endif
            endif
          endif
  
          !---  austauschterm --------------------------------------------------
          uaus = uee + uww - four*umi
          if (ne > 0 .and. nn > 0) then
            uaus = uaus + unn
          else
            uaus = uaus + umi
          endif
          if (se > 0 .and. ss > 0) then
            uaus = uaus + uss
          else
            uaus = uaus + umi
          endif
          ! not symm.
          uaus = dtd*uaus
  
          !- RHS of x-mom-eq stored in eb(:) -----------------------------------
          eb(k)=umi+fti*vst-ucon+uaus-usrf-upress
        enddo keuloop

        if (dyn) then
          !- The geopotential term
          do k=1,keu
            mi = mmk(k,i  ,j  )
            ee = mmk(k,i  ,j+1)
            eb(k) = eb(k) - (geopot(ee)-geopot(mi))*txi
          enddo
        endif
  
        vstar = vst
  
        sdukeu  = tm(keu)
        hrt     = sdukeu + hgr
        ! the lines below are not symm.
        ustar   = u(mmk(keu,i,j))
        tmp     = rth/hrt*sqrt(ustar*ustar + vstar*vstar)
        eb(keu) = eb(keu) - tmp*ustar
        ec(keu) = ec(keu) + tmp
  
        eb(1) = eb(1) + tutop
        if (.not.(casus(me(i,j)).and.casus(me(i,j+1)))) then
          eb(1) = eb(1) + half*decfu*u(mmk(1,i,j))
          ec(1) = ec(1) + half*decfu
        endif
        ! end of eb(:), ec(:) computation --------------------------------------
  
        ! begin alpha, beta computation ----------------------------------------
        alpha = zero
        beta  = zero
        alpha(keu) = eb(keu)
        beta(keu)  = ec(keu)
        do k=keu-1,1,-1
          sdu      = ea(k+1)/(tm(k)*beta(k+1))
          sdulum   = ea(k+1)/tm(k+1)
          alpha(k) = eb(k) + sdu*alpha(k+1)
          beta(k)  = ec(k) - sdu*sdulum
        enddo
        ! end alpha, beta computation ------------------------------------------
  
        ! begin un(:) computation ----------------------------------------------
        mi = mmk(1,i,j)
        un(mi) = alpha(1)/beta(1)
        do k=2,keu
          sdu = tm(k)
          un(mmk(k,i,j)) = (alpha(k)+ea(k)/sdu*un(mmk(k-1,i,j)))/beta(k)
        enddo

        !--- handle a drying situation -----------------------------------------
        if (keu <= kdry) then
          ee = mmk(1,i,j+1)
          call drying_case(mi, ee, un(mi), hz(mi), hz(ee), z) 
          if (kdry > 1) then
            do k = 2,min(keu,kdry)
              call drying_case(mi,ee,un(mmk(k,i,j)),hz(mmk(k,i,j)),            &
                               hz(mmk(1,i,j+1)),z)
            enddo
          endif
        endif
        ! end un(:) computation ------------------------------------------------
      endif
  
      ! for east border, shift back eastwards
  
      if (bordertype == 3) j = j + 1
  
      !----------------- end of u-equation -------------------------------------
  
      !----------------- v-equation --------------------------------------------
  
      ! for south border, shift n/s index one point northwards
      ! - shifted back again after v-equation
  
      if (bordertype == 4) i = i - 1
  
      kev = 0
      do k=1,ke
        mi = mmk(k,i,j)
        ss = mmk(k,i+1,j)
        if (mi <= 0 .or. ss <= 0) exit
        kev   = k
        tm(k) = half*(hz(mi)+hz(ss))
      enddo
  
      mi = mmk(1,i,j)
      if (kev == 0) vn(mi) = zero
      ss = mmk(1,i+1,j)

      !- skip trivial cases as well as S/E, S/W corners
      if (kev > 0 .and. ncorner /= 3 .and. ncorner /=4) then
  
        ! assert(kev >0)
  
        tmp = half*(z0srf(mi) + z0srf(ss))
        if (casus(me(i,j)).and.casus(ss)) then
          tvtop = tauy(mi)*dtdr/(tm(1)+tmp)
        else
          mie = me(i,  j)
          sse = me(i+1,j)
          eiscoo = min(max(ice(2,mie),zero),one)
          eiscou = min(max(ice(2,sse),zero),one)
          eiscoy = half*(eiscoo+eiscou)
          diffuo = ueis(me(i,  j-1)) - u(mmk(1,i,  j-1)) + ueis(mie) - u(mi)
          diffuu = ueis(me(i+1,j-1)) - u(mmk(1,i+1,j-1)) + ueis(sse) - u(ss)
          diffv  = veis(mie)-v(mi)
          diffvo = diffv + veis(me(i-1,j)) - v(mmk(1,i-1,j))
          diffvu = diffv + veis(sse)       - v(ss)
          diffuv = qrt*( eiscoo*sqrt(diffuo**2+diffvo**2)                      &
                        +eiscou*sqrt(diffuu**2+diffvu**2))

          fac = dtdr/(tm(1)+tmp)

          decfv  = cdri*diffuv*fac
          tvtop  = (cdri*diffuv*diffv+(one-eiscoy)*tauy(mi))*fac
        endif
  
        ! begin ea(:), ec(:) computation ---------------------------------------
        ea = zero ! i.e. ea(1)=0
        ec = zero
        do k=1,kev-2,2
          tmp     = tm(k)
          tmp2    = tm(k+1)
          ea(k+1) = dt*(avv(mmk(k+1,i,j))+avv(mmk(k+1,i+1,j)))/(tmp+tmp2)
          ec(k)   = one+(ea(k)+ea(k+1))/tmp
          tmp3    = tm(k+2)
          ea(k+2) = dt*(avv(mmk(k+2,i,j))+avv(mmk(k+2,i+1,j)))/(tmp2+tmp3)
          ec(k+1) = one+(ea(k+1)+ea(k+2))/tmp2
        enddo
        if ((mod(kev,2) /= 0).and.(kev > 0)) then
          ec(kev) = one+(ea(kev)+ea(kev+1))/tm(kev)
        else
          ea(kev)   = dt*(avv(mmk(kev,i,j))+avv(mmk(kev,i+1,j)))               &
                        /(tm(kev-1)+tm(kev))
          ec(kev-1) = one+(ea(kev-1)+ea(kev))/tm(kev-1)
          ec(kev)   = one+(ea(kev)+ea(kev+1))/tm(kev)
        endif
        ! end ea(:) computation ------------------------------------------------
  
        ! begin eb(:) computation ----------------------------------------------
  
        ! --- luftdruckterm  ---------------------------------------------------
        vlu = (pl(mi)-pl(ss))*ty*drhow ! used in eb(:) computation
  
        ! zeta:
        zm = z(mi)
        zu = z(ss)

        ! pressure term:
        vzet = gty*(zm-zu)

        ! collect different surface terms:
        vsrf = vlu + vzet

        txi = tx(i)
        fti = ft(i)
        eb  = zero
  
        ! --- calculate eb(k) --------------------------------------------------
        kevloop: do k=1,kev
          nw = mmk(k,i-1,j-1)
          ww = mmk(k,i  ,j-1)
          sw = mmk(k,i+1,j-1)
          nn = mmk(k,i-1,j)
          mi = mmk(k,i  ,j)
          ss = mmk(k,i+1,j)
          if (ss == 0) exit kevloop
          ne = mmk(k,i-1,min(j+1,n))
          ee = mmk(k,i  ,min(j+1,n))
          se = mmk(k,i+1,min(j+1,n))
  
          ! -- default nine-point v stencil
          vsw = v(sw)
          vss = v(ss)
          vse = v(se)
          vww = v(ww)
          vmi = v(mi)
          vee = v(ee)
          vnw = v(nw)
          vnn = v(nn)
          vne = v(ne)
  
          ! -- default four-point u stencil around vmi
          uww = u(ww)
          usw = u(sw)
          uss = u(ss)
          umi = u(mi)
  
          ! -- overrule defaults according to border and corner
          if (bordertype == 1) then
            ! -- west border
            uww = vel1_c_f(k)
            usw = vel3_c_f(k)
            vnw = vnn
            vww = vmi
            vsw = vss
            if (ncorner == 1) then
              ! -- N/W corner
              vnn = vel4_c_f(k)
              vne = vel4_c_f(k)
              vnw = vel5_c_f(k)
            endif
  
          elseif (bordertype == 2) then
            ! -- north border
            vne = vel3_c_f(k)
            vnn = vel1_c_f(k)
            vnw = vel2_c_f(k)
            if (ncorner == 2) then
              ! -- N/E corner
              umi = vel4_c_f(k)
              uss = vel4_c_f(k)
            endif
  
          elseif (bordertype == 3) then
            ! -- east border
            umi = vel1_c_f(k)
            uss = vel3_c_f(k)
            vne = vnn
            vee = vmi
            vse = vss
  
          elseif (bordertype == 4) then
            ! -- south border
            vse = vel3_c_f(k)
            vss = vel1_c_f(k)
            vsw = vel2_c_f(k)
  
          endif
  
          ust=qrt*(uww+umi+usw+uss)
  
          !---  barokliner druck -----------------------------------------------
          vpress=(press(mi)-press(ss))*ty/(half*(rho(mi)+rho(ss)))
  
          ! convective terms using vector upwind scheme:
          urd  = ust*txi
          vrd  = vmi*ty
          aurd = abs(urd)
          avrd = abs(vrd)

          if (avrd >= aurd) then
            if (vrd >= zero) then
              if (urd >= zero) then
                vcon = avrd*(vmi-vss) + aurd*(vss-vsw)
              else
                vcon = avrd*(vmi-vss) + aurd*(vss-vse)
              endif
            else
              if (urd >= zero) then
                vcon = avrd*(vmi-vnn) + aurd*(vnn-vnw)
              else
                vcon = avrd*(vmi-vnn) + aurd*(vnn-vne)
              endif
            endif
          else
            if (urd >= zero) then
              if (vrd >= zero) then
                vcon = avrd*(vww-vsw) + aurd*(vmi-vww)
              else
                vcon = avrd*(vww-vnw) + aurd*(vmi-vww)
              endif
            else
              if (vrd >= zero) then
                vcon = avrd*(vee-vse) + aurd*(vmi-vee)
              else
                vcon = avrd*(vee-vne) + aurd*(vmi-vee)
              endif
            endif
          endif
  
          !---  austauschterm --------------------------------------------------
          vaus = vnn + vss - four*vmi
          if (ee > 0 .and. se > 0) then
            vaus = vaus + vee
          else
            vaus = vaus + vmi
          endif
          if (ww > 0 .and. sw > 0) then
            vaus = vaus + vww
          else
            vaus = vaus + vmi
          endif
          ! not symm.
          vaus = dtd*vaus
  
          !- RHS of y-mom-eq stored in eb(:) -----------------------------------
          eb(k)=vmi-fti*ust-vcon+vaus-vsrf-vpress
        enddo kevloop

        if (dyn) then
          !- The geopotential term
          do k=1,kev
            mi = mmk(k,i  ,j  )
            ss = mmk(k,i+1,j  )
            eb(k) = eb(k) - (geopot(mi)-geopot(ss))*ty
          enddo
        endif

        ustar = ust
  
        sdvkev  = tm(kev)
        hrt     = sdvkev+hgr
        vstar   = v(mmk(kev,i,j))
        tmp     = rth/hrt*sqrt(vstar*vstar + ustar*ustar) 
        eb(kev) = eb(kev) - tmp*vstar
        ec(kev) = ec(kev) + tmp
  
        eb(1) = eb(1)+tvtop
        if (.not.(casus(me(i,j)).and.casus(me(i+1,j)))) then
          eb(1) = eb(1) + half*decfv*v(mmk(1,i,j))
          ec(1) = ec(1) + half*decfv
        endif
  
        ! begin alpha, beta computation ----------------------------------------
        alpha = zero
        beta  = zero
        alpha(kev) = eb(kev)
        beta(kev)  = ec(kev)
        do k=kev-1,1,-1
          sdv      = ea(k+1)/(tm(k)*beta(k+1))
          sdvlum   = ea(k+1)/tm(k+1)
          alpha(k) = eb(k) + sdv*alpha(k+1)
          beta(k)  = ec(k) - sdv*sdvlum
        enddo
        ! end alpha, beta computation ------------------------------------------
  
        ! begin vn(:) computation ----------------------------------------------
        mi = mmk(1,i,j)
        vn(mi) = alpha(1)/beta(1)
        do k=2,kev
          sdv = tm(k)
          vn(mmk(k,i,j)) = (alpha(k)+ea(k)/sdv*vn(mmk(k-1,i,j)))/beta(k)
        enddo
  
        !--- handle a drying situation -----------------------------------------
        if (kev <= kdry) then
          ss = mmk(1,i+1,j)
          call drying_case(ss, mi, vn(mi), hz(ss), hz(mi), z) 
          if (kdry > 1) then
            do k = 2,min(kev,kdry)
              call drying_case(ss,mi,vn(mmk(1,i,j)),hz(mmk(1,i,j)),            &
                               hz(mmk(1,i+1,j)),z)
            enddo
          endif
        endif
        ! end vn(:) computation ------------------------------------------------
      endif
  
      ! for south border, shift back southwards
  
      if (bordertype == 4) i = i + 1
  
    !----------------- end of v-equation ---------------------------------------
  
    enddo brdloop
  
    !---------------------------------------------------------------------------
  
  end subroutine mom_c_f

!-------------------------------------------------------------------------------

  subroutine mom_f_c   ( kg,mmkg,ugn,vgn,hug,hvg,krug,krvg,mmkf,igaf,jgaf,     &
                         kfg, ufn, vfn, huf, hvf, nu1, nu2, nv1, nv2 )
  !-----------------------------------------------------------------------------
  ! Use fine grid velocity components to 
  ! obtain coarse grid velocity components at border.
  !
  !
  ! The method works with the following border types (as seen from coarse grid):
  !   unknown border type:  < 0, no action
  !   not a border:         0, no action
  !   W border:             1
  !   N border:             2
  !   E border:             3
  !   S border:             4
  ! Each border must be enumerated in a specific order in increments of 1:
  !   E: from S to N
  !   S: from W to E
  !   W: from N to S
  !   N: from E to W
  ! There can be more than one border of each type.
  ! Each border must be fully connected in the specifications. The border points
  ! must be specified such that all coarse grid momentum points one point inside
  ! the border are covered with start and end either at a land point or 
  ! continuing at/from a border point of another border type (ie a corner).
  !
  ! DO NOT DUPLICATE POINTS TO MODEL NESTING CORNERS, i.e.
  !
  ! * concave corners points (as seen from coarse grid) should NOT appear in the
  !   specification since this would complicate this nesting procedure more than
  !   intended.
  !
  ! * convex corners (as seen from coarse grid) should be specified in only one 
  !   of the joining borders.
  !
  !
  ! Algorithm:
  !   At any point: fac = dxf/dxc, Ac = Af = A, Tc = Tf = T
  !
  !   Seen from fine grid
  !     area:       A = SUM( hf_i*dxf ) 
  !     transport:  T = SUM( hf_i*uf_i*dxf ) 
  !
  !   Seen from coarse grid
  !     hc = A/dxc = fac * SUM( hf_i )
  !     uc = T/A = T/(hc*dxc) = SUM( hf_i*uf_i*dxf ) / SUM( hf_i*dxf )
  ! 
  !-----------------------------------------------------------------------------
  
    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain
  
    !- directives --------------------------------------------------------------
    implicit none
  
  
    !- arguments ---------------------------------------------------------------
    !  indices:
    integer(4), intent(in) :: kg, nu1, nu2, nv1, nv2
    integer(4), intent(in) :: mmkg(:,0:,0:)
    integer(4), intent(in) :: mmkf(:,0:,0:)
    integer(4), intent(in) :: igaf(:),jgaf(:)
    integer(4), intent(in) :: kfg(1:)
    real(8), intent(inout) :: hug(0:),hvg(0:)
    real(8), intent(in)    :: huf(0:),hvf(0:)
    real(8), intent(inout) :: ugn(0:),vgn(0:)
    real(8), intent(in)    :: ufn(0:),vfn(0:)
  
    integer(4), intent(in) :: krug(:,0:), krvg(:,0:)
  
  
    !- local vars --------------------------------------------------------------
    integer(4) :: ira, i, j, ii, jj, iii, jjj, iff, jf, i1, j1, k, iioff, jjoff
    integer(4) :: kk, kkf, mi, mss, mee, mww, is, je, jw, inn, mnn, nl, nll, nlu
    integer(4) :: iprev, jprev, inext, jnext, nul, nuu, nvl, nvu, nlowu, nlowv
    logical    :: calc_nw
    real(8)    :: fac, vf, uf, hsum
  
! FIXME: hug/hvg might really only modify in top layer. Check this!


    !- Determine the decomposition of the borders ------------------------------
    nlowu = max(1,nu1)
    nlowv = max(1,nv1)
    if (nu2 >= nlowu .and. nv2 >= nlowv) then
      ! points on both u- and v-brds
      nl = (nu2+1-nlowu) + (nv2+1-nlowv)
      call domp_get_domain(1, nl, nll, nlu)
      nul = max(nlowu, nll-1+nlowu)
      nuu = min(nu2,   nlu-1+nlowu)
      nvl = max(nlowv, nll-nu2-1+nlowu-1+nlowv)
      nvu = min(nv2  , nlu-nu2-1+nlowu-1+nlowv)
    elseif (nu2 < nlowu .and. nv2 < nlowv) then
      ! nothing to do here ... let's get out
      return
    elseif (nu2 < nlowu) then
      ! only v-brds
      call domp_get_domain(nlowv, nv2, nvl, nvu)
      nul = 1
      nuu = -1
    elseif (nv2 < nlowv) then
      ! only u-brds
      call domp_get_domain(nlowu, nu2, nul, nuu)
      nvl = 1
      nvu = -1
    endif


    !- Do W and E borders ------------------------------------------------------
    uloop: do ira=nul,nuu
      if (krug(3,ira) == 0) cycle uloop
      i = krug(1,ira)
      j = krug(2,ira)

      !- Check if v at north border is calculated by previous or will be -------
      !- calculated by next border point                                 -------
      calc_nw = .true.
      if (ira < nu2) then
        iprev = krug(1,ira-1)
        jprev = krug(2,ira-1)
        inext = krug(1,ira+1)
        jnext = krug(2,ira+1)
        if (iprev == i-1 .and. jprev == j) calc_nw = .false.
        if (inext == i-1 .and. jnext == j) calc_nw = .false.
      endif  

      if (krug(3,ira) == 3) j=j+1
      if (mmkg(1,i,j) == 0) cycle uloop
  
      ii = igaf(2)*(i-igaf(1))+1
      jj = jgaf(2)*(j-jgaf(1))+1
  
      !- coordinates for vel.comp. 
      if (krug(3,ira) == 3) then
        !- east brd as seen from coarse grid:
        j1    = j
        jjoff = jgaf(2)-1
      elseif (krug(3,ira) == 1) then
        !- west brd as seen from coarse grid:
        j1    = j-1
        jjoff = -1
      endif
      inn = max(i-1,1)
      is  = i

      !- update coarse grid Vs, Vn and U at border:
      kk = 1
      ukloop: do k=1,kg

        !- U:
        mi = mmkg(k,i,j1)
        if (hug(mi) > zero) then
          uf  = zero
          hsum= zero
          fac = one/real(igaf(2),8)
          jf  = jj + jjoff
          do kkf=kk,kk-1+kfg(k)
            do iii=0,igaf(2)-1
              iff  = ii + iii
              uf   = uf + ufn(mmkf(kkf,iff,jf))*huf(mmkf(kkf,iff,jf))
              hsum = hsum + huf(mmkf(kkf,iff,jf))
            enddo
          enddo
          if (hsum > zero) then
            hug(mi) = hsum*fac
            ugn(mi) = uf/hsum
          else
            ugn(mi) = zero
          endif
        endif
  
        !- Treatment of S and N faces is only needed, if the actual point
        !- is not a corner!
        if (krug(4,ira) == 0) then 
          !- Vsouth:
          mss = mmkg(k,is,j)
          if (hvg(mss) > zero) then
            vf  = zero
            hsum= zero
            fac = one/real(jgaf(2),8)
            iff  = ii + igaf(2)-1
            do kkf=kk,kk-1+kfg(k)
              do jjj=0,jgaf(2)-1
                jf   = jj + jjj
                vf   = vf + vfn(mmkf(kkf,iff,jf))*hvf(mmkf(kkf,iff,jf))
                hsum = hsum + hvf(mmkf(kkf,iff,jf))
              enddo
            enddo
            if (hsum > zero) then
              hvg(mss) = hsum*fac
              vgn(mss) = vf/hsum
            else
              vgn(mss) = zero
            endif
          endif
 
          if (calc_nw) then
            !- Vnorth:
            mnn = mmkg(k,inn,j)
            if (hvg(mnn) > zero) then
              vf  = zero
              hsum= zero
              fac = one/real(jgaf(2),8)
              iff  = max(ii-1,1)
              do kkf=kk,kk-1+kfg(k)
                do jjj=0,jgaf(2)-1
                  jf   = jj + jjj
                  vf   = vf + vfn(mmkf(kkf,iff,jf))*hvf(mmkf(kkf,iff,jf))
                  hsum = hsum + hvf(mmkf(kkf,iff,jf))
                enddo
              enddo
              if (hsum > zero) then
                hvg(mnn) = hsum*fac
                vgn(mnn) = vf/hsum
              else
                vgn(mnn) = zero
              endif
            endif
          endif

        endif
        kk = kk+kfg(k)
      enddo ukloop
  
    enddo uloop


    !- Do N and S borders ------------------------------------------------------
    vloop: do ira=nvl,nvu
      if (krvg(3,ira) == 0) cycle vloop
      i = krvg(1,ira)
      j = krvg(2,ira)

      !- Check if u at west border is calculated by previous or will be --------
      !- calculated by next border point                                --------
      calc_nw = .true.
      if (ira < nv2) then
        iprev = krvg(1,ira-1)
        jprev = krvg(2,ira-1)
        inext = krvg(1,ira+1)
        jnext = krvg(2,ira+1)
        if (iprev == i .and. jprev == j-1) calc_nw = .false.
        if (inext == i .and. jnext == j-1) calc_nw = .false.
      endif

      if (krvg(3,ira) == 4) i=i+1
      if (mmkg(1,i,j) == 0) cycle vloop
  
      ii = igaf(2)*(i-igaf(1))+1
      jj = jgaf(2)*(j-jgaf(1))+1
  
      !- coordinates for vel.comp. 
      if (krvg(3,ira) == 2) then
        !- north brd as seen from coarse grid:
        i1    = i-1
        iioff = -1
      elseif (krvg(3,ira) == 4) then
        !- south brd as seen from coarse grid:
        i1    = i
        iioff = igaf(2)-1
      endif
      jw = max(j-1,1)
      je = j

      !- update coarse grid Ue, Uw and V at border:
      kk = 1
      vkloop: do k=1,kg

        !- V:
        mi = mmkg(k,i1,j)
        if (hvg(mi) > zero) then
          vf  = zero
          hsum= zero
          fac = one/real(jgaf(2),8)
          iff  = ii + iioff
          do kkf=kk,kk-1+kfg(k)
            do jjj=0,jgaf(2)-1
              jf   = jj + jjj
              vf   = vf + vfn(mmkf(kkf,iff,jf))*hvf(mmkf(kkf,iff,jf))
              hsum = hsum + hvf(mmkf(kkf,iff,jf))
            enddo
          enddo
          if (hsum > zero) then
            hvg(mi) = hsum*fac
            vgn(mi) = vf/hsum
          else
            vgn(mi) = zero
          endif
        endif
  
        !- Treatment of E and W faces is only needed, if the actual point
        !- is not a corner!
        if (krvg(4,ira) == 0) then 
          !- Ueast:
          mee = mmkg(k,i,je)
          if (hug(mee) > zero) then
            uf  = zero
            hsum= zero
            fac = one/real(igaf(2),8)
            jf  = jj + jgaf(2)-1
            do kkf=kk,kk-1+kfg(k)
              do iii=0,igaf(2)-1
                iff  = ii + iii
                uf   = uf + ufn(mmkf(kkf,iff,jf))*huf(mmkf(kkf,iff,jf))
                hsum = hsum + huf(mmkf(kkf,iff,jf))
              enddo
            enddo
            if (hsum >zero) then
              hug(mee) = hsum*fac
              ugn(mee) = uf/hsum
            else
              ugn(mee) = zero
            endif
          endif
  
          if (calc_nw) then
            !- Uwest:
            mww = mmkg(k,i,jw)
            if (hug(mww) > zero) then
              uf  = zero
              hsum= zero
              fac = one/real(igaf(2),8)
              jf  = max(jj-1,1)
              do kkf=kk,kk-1+kfg(k)
                do iii=0,igaf(2)-1
                  iff  = ii + iii
                  uf   = uf + ufn(mmkf(kkf,iff,jf))*huf(mmkf(kkf,iff,jf))
                  hsum = hsum + huf(mmkf(kkf,iff,jf))
                enddo
              enddo
              if (hsum >zero) then
                hug(mww) = hsum*fac
                ugn(mww) = uf/hsum
              else
                ugn(mww) = zero
              endif
            endif
          endif

        endif
        kk = kk+kfg(k)
      enddo vkloop
  
    enddo vloop
  
  end subroutine mom_f_c

! ------------------------------------------------------------------------------

  subroutine rand_z (kg,mmkg,ugn,vgn,hug,hvg,                                  &
                     kf,mmkf,igaf,jgaf,zf,zfn,ufn,vfn,hzf,huf,hvf,krzf,        &
                     rwzf,dtf,dxf,dyf,covof,covuf,nz1,nz2)
    !---------------------------------------------------------------------------
    ! subroutine for coupling of elevation from coarse to fine grid 
    ! when using fully dynamical nesting.
    !---------------------------------------------------------------------------
  
    !- modules -----------------------------------------------------------------
    use constants, only    : hgr
    use dmi_omp,   only    : domp_get_domain
    

    !- directives --------------------------------------------------------------
    implicit none
  
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: kg, nz1, nz2
    integer(4), intent(in) :: mmkg(:,0:,0:)
    integer(4), intent(in) :: kf
    integer(4), intent(in) :: mmkf(:,0:,0:)
    integer(4), intent(in) :: igaf(:),jgaf(:)
    !  coarse grid fields:
    real(8), intent(in)    :: ugn(0:),vgn(0:),hug(0:),hvg(0:)
    !  fine grid fields:
    real(8), intent(in)    :: zf(0:),ufn(0:),vfn(0:),huf(0:),hvf(0:),hzf(0:),  &
                              dtf,dxf(:),dyf,covof(:),covuf(:)
    real(8), intent(inout) :: zfn(0:)
    !  fields for border positions and values:
    integer(4), intent(in) :: krzf(:,0:)
    real(8), intent(inout) :: rwzf(0:)
  
  
    !- local vars --------------------------------------------------------------
    !  simple vars:
    integer(4) :: iw, i, j, ig, jg, igr, jgr, ii, jj, igrr, jgrr, nn, k, ns
    integer(4) :: il, ir, io, iu, nu, nl
    real(8)    :: trug, truf, trvg, trvf, dtrdx, dtrdy
  
    !
    !   west- and east-border
    !

    !Fixme: we need to consider an even wet point decompo instead 
    call domp_get_domain(nz1, nz2, nl, nu)
    weloop: do iw=nl,nu
  
      if (krzf(3,iw) == 0) cycle weloop
      if (krzf(3,iw) == 2) cycle weloop
      if (krzf(3,iw) == 4) cycle weloop
      i=krzf(1,iw)
      j=krzf(2,iw)
      ns = mmkf(1,i,j)
      if (ns <= 0) cycle weloop

      ig=int((i-1)/igaf(2))
      jg=int((j-1)/jgaf(2))
      igr=igaf(1)+ig
      jgr=jgaf(1)+jg
  
      if (krzf(3,iw) == 1) then
        jgrr = jgr-1
        jj   = j
      elseif (krzf(3,iw) == 3) then
        jgrr = jgr
        jj   = j-1
      endif
  
      trug=zero
      do k=1,kg
        nn   = mmkg(k,igr,jgrr)
        trug = trug + hug(nn)*ugn(nn)
      enddo
  
      truf=zero
      do k=1,kf
        nn   = mmkf(k,i,jj)
        truf = truf+huf(nn)*ufn(nn)
      enddo
  
      if (krzf(3,iw) == 1) then
        dtrdx = (truf-trug)/dxf(i)
      end if
      if (krzf(3,iw) == 3) then
        dtrdx = (trug-truf)/dxf(i)
      end if
  
      dtrdy=zero
      do k=1,kf
        io    = mmkf(k,i-1,j)
        iu    = mmkf(k, i ,j)
        dtrdy = dtrdy + (covof(i)*hvf(io)*vfn(io)-covuf(i)*hvf(iu)*vfn(iu))/dyf
      enddo
  
      zfn(ns)  = max( -(hzf(ns)-zf(ns))+hgr, zf(ns) - dtf*(dtrdx+dtrdy) )
      rwzf(iw) = zfn(ns)
    enddo weloop
   
    !
    !   nord- and south-border
    !
    nsloop: do iw=nl,nu
  
      if (krzf(3,iw) == 0) cycle nsloop
      if (krzf(3,iw) == 1) cycle nsloop
      if (krzf(3,iw) == 3) cycle nsloop
      i=krzf(1,iw)
      j=krzf(2,iw)
      ns = mmkf(1,i,j)
      if (ns <= 0) cycle nsloop

      ig=int((i-1)/igaf(2))
      jg=int((j-1)/jgaf(2))
      igr=igaf(1)+ig
      jgr=jgaf(1)+jg
  
      if (krzf(3,iw) == 2) then
        igrr = igr-1
        ii   = i
      elseif (krzf(3,iw) == 4) then
        igrr = igr
        ii   = i-1
      endif

      trvg=zero
      do k=1,kg
        nn   = mmkg(k,igrr,jgr)
        trvg = trvg + hvg(nn)*vgn(nn)
      enddo
  
      trvf=zero
      do k=1,kf
        nn   = mmkf(k, ii ,j)
        trvf = trvf+hvf(nn)*vfn(nn)
      enddo
  
      if (krzf(3,iw) == 2) then
        dtrdy = (covof(i)*trvg-covuf(i)*trvf)/dyf
      end if
      if (krzf(3,iw) == 4) then
        dtrdy = (covof(i)*trvf-covuf(i)*trvg)/dyf
      end if
  
      dtrdx=zero
      do k=1,kf
        il    = mmkf(k,i,j-1)
        ir    = mmkf(k,i, j )
        dtrdx = dtrdx+(huf(ir)*ufn(ir)-huf(il)*ufn(il))/dxf(i)
      enddo
  
      zfn(ns)  = max( -(hzf(ns)-zf(ns))+hgr, zf(ns) - dtf*(dtrdx+dtrdy) )
      rwzf(iw) = zfn(ns)
    enddo nsloop
   
  end subroutine rand_z

! ------------------------------------------------------------------------------

  subroutine w_c_f ( krzf,mmk,kc,mmkc,kfg,igaf,jgaf,zo,zn,w,u,v,uc,vc,         &
                     covo,covu,dx,dy,dt,huti,hvti, nz1, nz2)

    !---------------------------------------------------------------------------
    ! Update w in fine grid along N and W borders.
    ! (w is not really used for anything, only cosmetic)
    ! This routine only works 100% correctly if depth @ border and depth just
    ! outside border are equal.
    !
    ! Initiated 29 Nov, 2005, by Per Berg, COI,DMI
    !---------------------------------------------------------------------------

    use dmi_omp, only : domp_get_domain
    implicit none

    integer(4), intent(in) :: kc, nz1, nz2
    integer(4), intent(in) :: krzf(:,0:),mmkc(:,0:,0:),mmk(:,0:,0:)
    integer(4), intent(in) :: kfg(:),igaf(:),jgaf(:)
    real(8), intent(in)    :: zo(0:),zn(0:)
    real(8), intent(in)    :: u(0:),v(0:),huti(0:),hvti(0:)
    real(8), intent(in)    :: uc(0:),vc(0:)
    real(8), intent(in)    :: dx(:),covo(:),covu(:),dy,dt
    real(8), intent(inout) :: w(0:)

    integer(4) :: ibf, i, j, ig, jg, k, kgr, kk, mi, nzl, nzu
    real(8)    :: ddx,covudy,covody,wnew,uwest,ueast,vnorth,vsouth

    !---------------------------------------------------------------------------

    call domp_get_domain(nz1, nz2, nzl, nzu)

    brdloop: do ibf=nzl,nzu
      if (krzf(3,ibf) == 0) cycle brdloop
      if (krzf(3,ibf) == 3) cycle brdloop
      if (krzf(3,ibf) == 4) cycle brdloop
      i  = krzf(1,ibf)
      j  = krzf(2,ibf)
      mi = mmk(1,i,j)
      if ( mi <= 0) cycle brdloop
  
      ig = igaf(1) + int((i-1)/igaf(2))
      jg = jgaf(1) + int((j-1)/jgaf(2))
  
      ddx    = one/dx(i)
      covody = covo(i)/dy
      covudy = covu(i)/dy
  
      wnew = (zn(mi) - zo(mi))/dt

      if     (krzf(3,ibf) == 2) then
        ! North border
        !   extract V(north) from coarse grid,
        ig = ig - 1
        k  = 0
        do kgr=1,kc
          vnorth = vc(mmkc(kgr,ig,jg))*covody
          do kk=1,kfg(kgr)
            k = k+1
            mi = mmk(k,i,j)
            if (mi <= 0) cycle brdloop 
            w(mi)  = wnew
            uwest  = u(mmk(k,i,j-1))*huti(mmk(k,i,j-1))
            ueast  = u(mi)*huti(mi)
            vsouth = v(mi)*covudy
            wnew   = wnew + (ueast-uwest)*ddx + (vnorth-vsouth)*hvti(mi)
          enddo
        enddo

      elseif (krzf(3,ibf) == 1) then
        ! West border
        !   extract U(west) from coarse grid,
        jg = jg - 1
        k  = 0
        do kgr=1,kc
          uwest = uc(mmkc(kgr,ig,jg))
          do kk=1,kfg(kgr)
            k = k+1
            mi = mmk(k,i,j)
            if (mi <= 0) cycle brdloop
            w(mi)  = wnew
            vsouth = v(mi)*hvti(mi)*covudy
            vnorth = v(mmk(k,i-1,j))*hvti(mmk(k,i-1,j))*covody
            ueast  = u(mi)
            wnew   = wnew + (ueast-uwest)*ddx*huti(mi) + (vnorth-vsouth)
          enddo
        enddo
  
      endif

    enddo brdloop

  end subroutine w_c_f

! ------------------------------------------------------------------------------

  subroutine checknesting()
    !---------------------------------------------------------------------------
    ! Make sure that we only treat nesting border for areas that are on this MPI
    ! task. Both inclosing and enclosing nested domains must be handled on the
    ! same MPI task; if they are not (i.e. if one of them have no wet points on 
    ! this task) then we fake the upper limit of the border point number to be
    ! lower than lower limit in order to escape loops.
    !---------------------------------------------------------------------------
!FIXME: This relies on the original - and strict - assumption that corresponding
!       nesting points in coarse and fine grid are on the same mpi task.
!       We will have to generalize - and loosen up on - these restrictions.
!        ... work in progress ...
  
    use dmi_mpi,        only : dd, mpi_size, nctable, nftable, mpi_tt,         &
                               mpi_rank, uctable, vctable, uftable, vftable,   &
                               cctable, cftable, uhtable, vhtable, mctable
    use cmod_params,    only : narea, nestinglevels, nestingto, iga, jga, krz, &
                               nestingfrom, enclosing, mmx, nmx,               &
                               znesting, unesting, vnesting,                   &
                               un_c_mpi, vn_c_mpi, un_f_mpi, vn_f_mpi
    use cmod_arrays,    only : kru, krv
  
    implicit none
  
    integer(4) :: ia, ii, iia, iif, n, n1, n2, i, j, ic, jc, ig, jg, it, iao
    integer(4) :: il, iu, jl, ju, ig1, ig2, jg1, jg2, iff, jff, iii, jjj
    integer(4) :: is, ie, js, je, ign, igs, jgw, jge, bordertype, ncorner

    !  initialise:
    do ia=1,narea
      un_c_mpi(ia)%p(:,:) = unesting(ia)%p(:,:)
      vn_c_mpi(ia)%p(:,:) = vnesting(ia)%p(:,:)
      un_f_mpi(ia)%p(:,:) = unesting(ia)%p(:,:)
      vn_f_mpi(ia)%p(:,:) = vnesting(ia)%p(:,:)
    enddo
    nctable(:,:,:) = .false.
    nftable(:,:,:) = .false.
    uctable(:,:,:) = .false.
    uftable(:,:,:) = .false.
    vctable(:,:,:) = .false.
    vftable(:,:,:) = .false.

    cctable(:,:) = .false.
    cftable(:,:) = .false.
    uhtable(:,:) = .false.
    vhtable(:,:) = .false.
    mctable(:,:) = .false.
  
    !  run through each area:
    do ia=1,narea
      do ii=1,nestinglevels(ia)
        !  iia: fine grid
        !  ia:  coarse grid
        iia = nestingto(ia)%p(ii)
        iif = 1
        do while(ia /= nestingfrom(iia)%p(iif))
          iif = iif + 1
        enddo
  
        !  Do any of these two areas have anything to contribute to this 
        !  nesting as seen from this MPI task?
        !   - if not, make sure that they do not attemp to nest,
        !     neither z-brds or u/v-bdrs:

        iao = enclosing(iia,ia)
        ig1 = iga(iia)%p(1,iao)
        ig2 = iga(iia)%p(2,iao)
        jg1 = jga(iia)%p(1,iao)
        jg2 = jga(iia)%p(2,iao)

        !  run through all tasks and setup z-nesting MPI-tables as seen from 
        !  coarse and fine grid 
        n1 = znesting(iia)%p(iif,1)
        n2 = znesting(iia)%p(iif,2)
        do it=1,mpi_size
          do n=n1,n2
            i  = krz(iia)%p(1,n)
            j  = krz(iia)%p(2,n)
            ic = ig1+int((i-1)/ig2)
            jc = jg1+int((j-1)/jg2)

            !  z brd seen from coarse ------------------------------------------
            if (.not.nctable(it,ia,iia)) then
              ig = ic
              jg = jc
              if     (krz(iia)%p(3,n) == 1) then
                jg = jc-1
              elseif (krz(iia)%p(3,n) == 2) then
                ig = ic-1
              elseif (krz(iia)%p(3,n) == 3) then
                jg = jc+1
              elseif (krz(iia)%p(3,n) == 4) then
                ig = ic+1
              endif
              il = mpi_tt(ia,it)%low_i
              iu = mpi_tt(ia,it)%up_i
              jl = mpi_tt(ia,it)%low_j
              ju = mpi_tt(ia,it)%up_j
              if ((il <= ig .and. ig <= iu) .and.                              &
                  (jl <= jg .and. jg <= ju) ) then
                ! found coarse zbrd
                nctable(it,ia,iia) = .true.
              endif
            endif

            !  z brd seen from fine --------------------------------------------
            if (.not.nftable(it,ia,iia)) then
              il = mpi_tt(iia,it)%low_i
              iu = mpi_tt(iia,it)%up_i
              jl = mpi_tt(iia,it)%low_j
              ju = mpi_tt(iia,it)%up_j
              if ((il <= i .and. i <= iu) .and. (jl <= j .and. j <= ju) ) then
                ! found fine zbdr
                nftable(it,ia,iia) = .true.
              endif
            endif

            if (nftable(it,ia,iia) .and. nctable(it,ia,iia)) exit
          enddo

          !  coarse grid dims:
          il = mpi_tt(ia,it)%low_i
          iu = mpi_tt(ia,it)%up_i
          jl = mpi_tt(ia,it)%low_j
          ju = mpi_tt(ia,it)%up_j
          brdloop: do n=n1,n2
            ! border point index (i,j) in fine grid
            bordertype = krz(iia)%p(3,n)
            if (bordertype == 0) cycle brdloop
            i = krz(iia)%p(1,n)
            j = krz(iia)%p(2,n)

            ! only do corner points once
            if (n > 1) then
              ! skip a corner point that has already been handled
              if (krz(iia)%p(3,n-1) > 0  .and.                                 &
                  i == krz(iia)%p(1,n-1) .and. j == krz(iia)%p(2,n-1))         &
                cycle brdloop
            endif
            ncorner = 0
            if (n < n2) then
              ! first-time identification of a convex corner point
              if (krz(iia)%p(3,n+1) > 0  .and.                                 &
                  i == krz(iia)%p(1,n+1) .and. j == krz(iia)%p(2,n+1)) then
                ! identify this corner by a code:
                !   unknown corner type:  < 0
                !   not a corner:           0
                !   N/W corner:             1
                !   N/E corner:             2
                !   S/E corner:             3
                !   S/W corner:             4
                ncorner = bordertype
              endif
            endif

            ! border point index (ic,jc) in coarse grid
            ic = ig1 + int((i-1)/ig2)
            jc = jg1 + int((j-1)/jg2)

            ! To assign coarse grid normal velocities at all k-levels
            ! to the west of west border, north of north border, etc.
            ! three grid points (left, central, right)
            if (bordertype == 1) then
              ! West border
              !   extract U(west) from coarse grid.
              jc  = jc - 1
              ign = ig1 + int((i-2)/ig2)
              igs = ig1 + int((i  )/ig2)
              if (jl <= jg  .and. jg  <= ju) then
                if ((il <= ig  .and. ig  <= iu) .or.                           &
                    (il <= ign .and. ign <= iu) .or.                           &
                    (il <= igs .and. igs <= iu)      ) then
                  mctable(ia,iia) = .true.
                  exit brdloop
                endif
              endif
              if (ncorner == bordertype) then
                ! N/W corner point
                !   extract V(north) from coarse grid.
                if (il <= ig-1 .and. ig-1 <= iu) then
                  if ((jl <= jg   .and. jg   <= ju) .or.                       &
                      (jl <= jg+1 .and. jg+1 <= ju)      ) then
                    mctable(ia,iia) = .true.
                    exit brdloop
                  endif
                endif
              endif

            elseif (bordertype == 2) then
              ! North border
              !   extract V(north) from coarse grid.
              ig  = ig - 1
              jgw = jg1 + (j-2)/jg2
              jge = jg1 + (j  )/jg2
              if (il <= ig .and. ig <= iu) then
                if ((jl <= jg  .and. jg  <= ju) .or.                           &
                    (jl <= jgw .and. jgw <= ju) .or.                           &
                    (jl <= jge .and. jge <= ju)      ) then
                  mctable(ia,iia) = .true.
                  exit brdloop
                endif
              endif
              if (ncorner == bordertype) then
                ! N/E corner point
                !   extract U(east) from coarse grid.
                if ((jl <= jg   .and. jg   <= ju) .and.                        &
                    (il <= ig+1 .and. ig+1 <= iu)       ) then
                  mctable(ia,iia) = .true.
                  exit brdloop
                endif
              endif

            elseif (bordertype == 3) then
              ! East border
              !   extract U(east) from coarse grid.
              ign   = ig1 + (i-2)/ig2
              igs   = ig1 + (i  )/ig2
              if (jl <= jg  .and. jg  <= ju) then
                if ((il <= ig  .and. ig  <= iu) .or.                           &
                    (il <= ign .and. ign <= iu) .or.                           &
                    (il <= igs .and. igs <= iu)      ) then
                  mctable(ia,iia) = .true.
                  exit brdloop
                endif
              endif
              ! S/E corner point
              !   extract V(south) from coarse grid.
              ! i.e. no special treatment.

            elseif (bordertype == 4) then
              ! South border
              !   extract V(south) from coarse grid.
              jgw   = jg1 + (j-2)/jg2
              jge   = jg1 + (j  )/jg2
              if (il <= ig .and. ig <= iu) then
                if ((jl <= jg  .and. jg  <= ju) .or.                           &
                    (jl <= jgw .and. jgw <= ju) .or.                           &
                    (jl <= jge .and. jge <= ju)      ) then
                  mctable(ia,iia) = .true.
                  exit brdloop
                endif
              endif
              if (ncorner == bordertype) then
                ! S/W corner
                !   extract U(west) from coarse grid.
                if ((jl <= jg-1 .and. jg-1 <= ju)) then
                  if ((il <= ig   .and. ig   <= iu) .or.                       &
                      (il <= ig+1 .and. ig+1 <= iu)     ) then
                    mctable(ia,iia) = .true.
                    exit brdloop
                  endif
                endif
              endif

            endif ! bordertype

          enddo brdloop

        enddo ! it



        !  run through all tasks and setup u-nesting MPI-tables as seen from
        !  both coarse and fine grid 
        n1 = un_c_mpi(ia)%p(ii,1)
        n2 = un_c_mpi(ia)%p(ii,2)
        do it=1,mpi_size
          do n=n1,n2
            ig = kru(ia)%p(1,n)
            jg = kru(ia)%p(2,n)
            if (kru(ia)%p(3,n) == 3) jg = jg+1

            !  u brd seen from coarse ------------------------------------------
            if (.not.uctable(it,ia,iia)) then
              il = mpi_tt(ia,it)%low_i
              iu = mpi_tt(ia,it)%up_i
              if (il <= ig .and. ig <= iu) then
                jl = mpi_tt(ia,it)%low_j
                ju = mpi_tt(ia,it)%up_j
                if (jl <= jg .and. jg <= ju) then
                  ! found coarse ubrd
                  uctable(it,ia,iia) = .true.
                endif
              endif
            endif

            !  u brd seen from fine --------------------------------------------
            if (.not.uftable(it,ia,iia)) then
              il = mpi_tt(iia,it)%low_i
              iu = mpi_tt(iia,it)%up_i
              jl = mpi_tt(iia,it)%low_j
              ju = mpi_tt(iia,it)%up_j
              i  = ig2*(ig-ig1) + 1
              j  = jg2*(jg-jg1) + 1
              do jjj=0,jg2-1
                jff = j + jjj
                if (jl > jff .or. jff > ju) cycle
                do iii=0,ig2-1
                  iff = i + iii
                  if (il <= iff .and. iff <= iu) then
                    ! found fine ubdr
                    uftable(it,ia,iia) = .true.
                    exit
                  endif
                enddo
              enddo
            endif

            !  u brd seen from fine, mom-f-c nesting ---------------------------
            if (.not.uhtable(ia,iia)) then
              il = mpi_tt(iia,it)%low_i
              iu = mpi_tt(iia,it)%up_i
              jl = mpi_tt(iia,it)%low_j
              ju = mpi_tt(iia,it)%up_j
              i  = ig2*(ig-ig1) + 1
              j  = jg2*(jg-jg1) + 1
              ujloop: do jjj=-1,jg2
                jff = j + jjj
                if (jl > jff .or. jff > ju) cycle
                do iii=-1,ig2
                  iff = i + iii
                  if (il <= iff .and. iff <= iu) then
                    ! found fine ubdr
                    uhtable(ia,iia) = .true.
                    exit ujloop
                  endif
                enddo
              enddo ujloop
            endif

            if (uftable(it,ia,iia) .and. uctable(it,ia,iia) .and.              &
                uhtable(ia,iia)) exit
          enddo
        enddo
        if (.not.uctable(mpi_rank+1,ia,iia)) un_c_mpi(ia)%p(ii,2) = -1


        !  run through all tasks and setup v-nesting MPI-tables as seen from
        !  both coarse and fine grid
        n1 = vn_c_mpi(ia)%p(ii,1)
        n2 = vn_c_mpi(ia)%p(ii,2)
        do it=1,mpi_size
          do n=n1,n2
            ig = krv(ia)%p(1,n)
            if (krv(ia)%p(3,n) == 4) ig = ig+1
            jg = krv(ia)%p(2,n)

            !  v brd seen from coarse ------------------------------------------
            if (.not.vctable(it,ia,iia)) then
              jl = mpi_tt(ia,it)%low_j
              ju = mpi_tt(ia,it)%up_j
              if (jl <= jg .and. jg <= ju) then
                il = mpi_tt(ia,it)%low_i
                iu = mpi_tt(ia,it)%up_i
                if (il <= ig .and. ig <= iu) then
                  ! found coarse vbrd
                  vctable(it,ia,iia) = .true.
                endif
              endif
            endif

            !  v brd seen from fine --------------------------------------------
            if (.not.vftable(it,ia,iia)) then
              il = mpi_tt(iia,it)%low_i
              iu = mpi_tt(iia,it)%up_i
              jl = mpi_tt(iia,it)%low_j
              ju = mpi_tt(iia,it)%up_j
              i  = ig2*(ig-ig1) + 1
              j  = jg2*(jg-jg1) + 1
              do jjj=0,jg2-1
                jff = j + jjj
                if (jl > jff .or. jff > ju) cycle
                do iii=0,ig2-1
                  iff = i + iii
                  if (il <= iff .and. iff <= iu) then
                    ! found fine vbdr
                  vftable(it,ia,iia) = .true.
                    exit
                  endif
                enddo
              enddo
            endif

            !  v brd seen from fine, mom-f-c nesting ---------------------------
            if (.not.vhtable(ia,iia)) then
              il = mpi_tt(iia,it)%low_i
              iu = mpi_tt(iia,it)%up_i
              jl = mpi_tt(iia,it)%low_j
              ju = mpi_tt(iia,it)%up_j
              i  = ig2*(ig-ig1) + 1
              j  = jg2*(jg-jg1) + 1
              vjloop: do jjj=-1,jg2
                jff = j + jjj
                if (jl > jff .or. jff > ju) cycle
                do iii=-1,ig2
                  iff = i + iii
                  if (il <= iff .and. iff <= iu) then
                    ! found fine ubdr
                    vhtable(ia,iia) = .true.
                    exit vjloop
                  endif
                enddo
              enddo vjloop
            endif

            if (vftable(it,ia,iia) .and. vctable(it,ia,iia) .and.              &
                vhtable(ia,iia)) exit
          enddo
        enddo
        if (.not.vctable(mpi_rank+1,ia,iia)) vn_c_mpi(ia)%p(ii,2) = -1


        !  setup copy-cmp-MPI-tables as seen from both coarse and fine grid ----
        !
        is = ig1 + (ig2/2+iga(iia)%p(3,iao))/ig2
        ie = ig1 + (ig2/2+iga(iia)%p(3,iao)+mmx(iia)-1)/ig2
        js = jg1 + (jg2/2+jga(iia)%p(3,iao))/jg2
        je = jg1 + (jg2/2+jga(iia)%p(3,iao)+nmx(iia)-1)/jg2
        !
        !  first, the coarse grid; is it present on this task ?
        il = dd(ia)%low_i
        iu = dd(ia)%up_i
        jl = dd(ia)%low_j
        ju = dd(ia)%up_j
        jgloop: do jg=max(js,jl),min(je,ju)
          do ig=max(is,il),min(ie,iu)
            cctable(ia,iia) = .true.
            exit jgloop
          enddo
          !NOTE: DO loop does not loop because of unconditional EXIT statement
        enddo jgloop  
        !
        !  then, the fine grid; is it present on any task ?
        if (dd(iia)%nwet3 > 0) then
          taskloop: do it=1,mpi_size
            do jg=max(js,mpi_tt(ia,it)%low_j),min(je,mpi_tt(ia,it)%up_j)
              do ig=max(is,mpi_tt(ia,it)%low_i),min(ie,mpi_tt(ia,it)%up_i)
                i = ig2*(ig-ig1) + 1
                j = jg2*(jg-jg1) + 1
                do jjj=0,jg2-1
                  jff = j + jjj
                  if (jff < dd(iia)%low_j .or. jff > dd(iia)%up_j) cycle
                  do iii=0,ig2-1
                    iff = i + iii
                    if (iff < dd(iia)%low_i .or. iff > dd(iia)%up_i) cycle
                      cftable(ia,iia) = .true.
                      exit taskloop
                  enddo  !NOTE: Last statement of DO loop body is an
                         !      unconditional EXIT statement
                enddo
              enddo
            enddo 
          enddo taskloop
        endif

      enddo
    enddo
  
  end subroutine checknesting

!===============================================================================

  subroutine brdcopy ( ia, kc,mmck,kruc,krvc,cmp_c,mmfkg,igaf,jgaf,cmp_f,kfg,  &
                       hf,dxf,dyf,nu1,nu2,nv1,nv2,ntracers )
    !---------------------------------------------------------------------------
    !
    !  Tracer values from fine grid to coarse grid at borders
    !
    !---------------------------------------------------------------------------

    use dmi_omp, only : domp_get_domain
    use dmi_mpi, only : dd

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)  :: ia, kc, nu1, nu2, nv1, nv2, ntracers
    integer(4), intent(in)  :: mmck(:,0:,0:)
    integer(4), intent(in)  :: mmfkg(:,0:,0:)   ! global
    integer(4), intent(in)  :: igaf(:), jgaf(:)

    !     fine grid fields
    real(8), intent(in)  :: hf(0:)       ! global
    real(8), intent(in)  :: cmp_f(:,0:)  ! global
    real(8), intent(in)  :: dxf(:), dyf

    !     fields for bnd data
    integer(4), intent(in)    :: kruc(:,0:)
    integer(4), intent(in)    :: krvc(:,0:)
    real(8),    intent(inout) :: cmp_c(:,0:)
    integer(4), intent(in)    :: kfg(:)


    !- local vars --------------------------------------------------------------
    integer(4) :: ira, i, j, ii, jj, iii, jjj, iff, jf
    integer(4) :: k, kff, kff0, idx, idxc, il, iu, jl, ju 
    integer(4) :: nlowu, nlowv, nl, nll, nlu, nul, nuu, nvl, nvu
    real(8)    :: r, rwt(ntracers), vol

    il = dd(ia)%low_i
    iu = dd(ia)%up_i
    jl = dd(ia)%low_j
    ju = dd(ia)%up_j

    !- Determine the decomposition of the borders ------------------------------
    nlowu = max(1,nu1)
    nlowv = max(1,nv1)
    if (nu2 >= nlowu .and. nv2 >= nlowv) then
      ! points on both E/W- and N/S-brds
      nl = (nu2+1-nlowu) + (nv2+1-nlowv)
      call domp_get_domain(1, nl, nll, nlu)
      nul = max(nlowu, nll-1+nlowu)
      nuu = min(nu2,   nlu-1+nlowu)
      nvl = max(nlowv, nll-nu2-1+nlowu-1+nlowv)
      nvu = min(nv2  , nlu-nu2-1+nlowu-1+nlowv)
    elseif (nu2 < nlowu .and. nv2 < nlowv) then
      ! nothing to do here ... let's get out
      return
    elseif (nu2 < nlowu) then
      ! only N/S-brds
      call domp_get_domain(nlowv, nv2, nvl, nvu)
      nul = 1
      nuu = -1
    elseif (nv2 < nlowv) then
      ! only E/W-brds
      call domp_get_domain(nlowu, nu2, nul, nuu)
      nvl = 1
      nvu = -1
    endif


    !- Do W and E borders ------------------------------------------------------
    do ira=nul,nuu
      i = kruc(1,ira)
      j = kruc(2,ira)

      if (kruc(3,ira) == 3) j = j+1

      if (i < il .or. i > iu .or. j < jl .or. j > ju) cycle

      ii = igaf(2)*(i-igaf(1))+1
      jj = jgaf(2)*(j-jgaf(1))+1

      kff0 = 1
      do k=1,kc
        idxc = mmck(k,i,j)
        if (idxc <= 0) exit

        r      = zero
        rwt(:) = zero

        do jjj=0,jgaf(2)-1
          jf = jj+jjj
          do iii=0,igaf(2)-1
            iff = ii+iii
            do kff = kff0,kff0+kfg(k)-1
              idx = mmfkg(kff,iff,jf)
              if (idx <= 0) exit

              vol    = dxf(iff)*dyf*hf(idx)
              r      = r + vol
              rwt(:) = rwt(:) + cmp_f(:,idx)*vol
            enddo
          enddo
        enddo

        kff0 = kff0+kfg(k)

        cmp_c(:,idxc) = rwt(:)/r
      enddo
    enddo

    !- Do N and S borders ------------------------------------------------------
    do ira=nvl,nvu
      i = krvc(1,ira)
      j = krvc(2,ira)

      if (krvc(3,ira) == 4) i = i+1

      if (i < il .or. i > iu .or. j < jl .or. j > ju) cycle

      ii = igaf(2)*(i-igaf(1))+1
      jj = jgaf(2)*(j-jgaf(1))+1

      kff0 = 1
      do k=1,kc
        idxc = mmck(k,i,j)
        if (idxc <= 0) exit

        r      = zero
        rwt(:) = zero

        do jjj=0,jgaf(2)-1
          jf = jj+jjj
          do iii=0,igaf(2)-1
            iff = ii+iii
            do kff = kff0,kff0+kfg(k)-1
              idx = mmfkg(kff,iff,jf)
              if (idx <= 0) exit
              
              vol    = dxf(iff)*dyf*hf(idx)
              r      = r + vol
              rwt(:) = rwt(:) + cmp_f(:,idx)*vol
            enddo
          enddo
        enddo

        kff0 = kff0+kfg(k)

        cmp_c(:,idxc) = rwt(:)/r
      enddo
    enddo


  end subroutine brdcopy

!===============================================================================

  subroutine masseqs_c_f_dyn (ia,nz1,nz2,kh_c,kmax_c,mmk_c,hu_c,hv_c,un_c,     &
                              vn_c,kfg,iga,jga,kmax,mmk,krz,khu,khv,kh,dx,dy,  &
                              dt,covo,covu,hu,hv,un,vn,z,zn,ti,h,h0,hn,ws,avv)

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !!! FIXME: This routine is only tested intensively in the case that the    !!!
  !!!        vertical resolution in the fine grid is exactly the same than   !!!
  !!!        the vertical resolution in the coarse grid!                     !!!
  !!!        Technically it works also in other cases - but results are      !!!
  !!!        maybe not as intended.                                          !!!
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  !-----------------------------------------------------------------------------
  ! Use momentum equation and coarse grid velocity components to 
  ! obtain fine grid velocity components at border.
  !
  !
  ! The method works with the following border types (as seen from fine grid):
  !   unknown border type:  < 0, no action
  !   not a border:         0, no action
  !   W border:             1
  !   N border:             2
  !   E border:             3
  !   S border:             4
  ! Each border must be enumerated in a specific order in increments of 1:
  !   W: from S to N
  !   N: from W to E
  !   E: from N to S
  !   S: from E to W
  ! Each border must be fully connected in the specifications.
  ! There can be more than one border of each type.
  ! 
  ! Convex corners (as seen from fine grid) need special treatment. The corner 
  ! point (in fine grid) must be defined on BOTH of the joined borders, 
  ! enumerated by an increment of 1 clockwise, eg 
  !   for a W/N corner, where the W border ends at brd-point No. 7, the N border
  !   must start at brd-point No. 8.
  ! 
  ! Concave corners (as seen from fine grid) need no special treatment with the 
  ! present method.
  !
  !
  ! Initiated Nov. 28, 2005, by Per Berg, COI/DMI.
  !-----------------------------------------------------------------------------

      !- modules  --------------------------------------------------------------
      use constants,      only : hgr, wghtz, wghthar, wghteta, wghtturb, gammadt
      use dmi_mpi_global, only : iu06
      use dmi_omp,        only : domp_get_domain
      use exits,          only : exitme
    
    
      !- directives ------------------------------------------------------------
      implicit none
    
    
      !- arguments -------------------------------------------------------------
      integer(4), intent(in)       :: ia
      integer(4), intent(in)       :: nz1,nz2
      integer(4), intent(in)       :: kmax_c, kmax
      integer(4), intent(in)       :: kfg(:)
      integer(4), intent(in)       :: iga(:),jga(:)
      integer(4), intent(in)       :: mmk_c(:,0:,0:), mmk(:,0:,0:)
      integer(4), intent(in)       :: krz(:,0:)
      integer(4), intent(in)       :: khu(0:),khv(0:)
      integer(4), intent(in)       :: kh_c(0:),kh(0:)
      real(8), intent(in)          :: dy,dt
      real(8), intent(in)          :: dx(:),covo(:),covu(:)
      real(8), intent(in)          :: hu_c(0:),hv_c(0:),hu(0:),hv(0:)
      real(8), intent(in)          :: un_c(0:),vn_c(0:),un(0:),vn(0:)
      real(8), intent(in)          :: avv(0:)
      real(8), intent(in)          :: ti(0:)
      real(8), intent(in)          :: z(0:)
      real(8), intent(in)          :: h(0:)
      real(8), intent(in)          :: h0(0:)
      real(8), intent(inout)       :: hn(0:)
      real(8), intent(inout)       :: ws(0:), zn(0:)
    
      !- local vars ------------------------------------------------------------
      real(8),dimension(1:kmax)   :: h_tmp, h_inc
      real(8),dimension(1:kmax)   :: div,diverg,wpsd
      real(8),dimension(1:kmax)   :: a_h,b_h,c_h,e_h
      real(8),dimension(1:kmax)   :: huwest,hueast,hvnorth,hvsouth
      real(8),dimension(1:kmax)   :: uwest,ueast,vnorth,vsouth
      real(8),dimension(1:kmax+1) :: r_h,z_h
      real(8),dimension(1:kmax_c) :: h_r
      integer(4)           :: nnz,nn, ncorner, nzl, nzu
      integer(4)           :: kb_c, kb
      integer(4)           :: i,j,kc,k,kk
      integer(4)           :: igr,jgr,ig,jg  
      integer(4)           :: kbue,kbuw,kbvn,kbvs
      integer(4)           :: iterat,it_max
      integer(4)           :: kbw,kbe,kbn,kbs
      integer(4)           :: mi
      real(8)              :: divk, dxi, covoidy, covuidy
      real(8)              :: depth_n,w_pseudo
      real(8)              :: hue,huw,hvn,hvs
      real(8)              :: norm,scal
      real(8)              :: dh,fac
      real(8)              :: length_a,length_b
      real(8)              :: h_tmp_a, h_tmp_b
    
      !- local parameters ------------------------------------------------------
      real(8),parameter    ::  scal_max = 1.0e+6_8, norm_min = 1.0e-9_8

      call domp_get_domain(nz1, nz2, nzl, nzu)
                        
      do nnz = nzl, nzu
        i = krz(1,nnz)
        j = krz(2,nnz)

        if (mmk(1,i,j) <= 0) cycle

        ! only do corner points once
        if (nnz > 1) then
          ! skip a corner point that has already been handled
          if (krz(3,nnz-1) > 0  .and.                                          &
              i == krz(1,nnz-1) .and. j == krz(2,nnz-1)) cycle
        endif

        ncorner = 0
        if (nnz < nz2) then
          ! first-time identification of a convex corner point
          if (krz(3,nnz+1) > 0  .and.                                          &
              i == krz(1,nnz+1) .and. j == krz(2,nnz+1)) then
            ! identify this corner by a code:
            !   unknown corner type:  < 0
            !   not a corner:           0
            !   N/W corner:             1
            !   N/E corner:             2
            !   S/E corner:             3
            !   S/W corner:             4
            ncorner = krz(3,nnz)
          endif
        endif

        nn = mmk(1,i,j)
        kb = kh(nn)
        dxi   = dx(i)
        covoidy = covo(i)/dy
        covuidy = covu(i)/dy

        ig=int((i-1)/iga(2))
        jg=int((j-1)/jga(2))
        igr=iga(1)+ig
        jgr=jga(1)+jg

        kb_c = kh_c(mmk_c(1,igr,jgr))

        uwest(:)     = zero
        ueast(:)     = zero
        vnorth(:)    = zero
        vsouth(:)    = zero

        kk = 1
        do kc = 1, kb_c
          h_r(kc) = zero
          do k = kk, kk-1+kfg(kc)
            if(h(mmk(k,i,j)) >= zero) h_r(kc) = h_r(kc) + h(mmk(k,i,j))
          enddo

          do k = kk, min(kk-1+kfg(kc),kb)
            fac = h(mmk(k,i,j)) / h_r(kc)

            huwest(k)  = hu_c(mmk_c(kc,igr,jgr-1))*fac
            uwest(k)   = un_c(mmk_c(kc,igr,jgr-1))

            hvnorth(k) = hv_c(mmk_c(kc,igr-1,jgr))*fac
            vnorth(k)  = vn_c(mmk_c(kc,igr-1,jgr))

            hueast(k)  = hu_c(mmk_c(kc,igr,jgr))*fac
            ueast(k)   = un_c(mmk_c(kc,igr,jgr))

            hvsouth(k) = hv_c(mmk_c(kc,igr,jgr))*fac
            vsouth(k)  = vn_c(mmk_c(kc,igr,jgr))
          end do

          kk = kk + kfg(kc)
        end do  

   
        h_inc(:) = zero
        div(:)   = zero
        diverg(:)= zero
        wpsd(:)  = zero

        depth_n = zero
      
        !- Calculate v * dz/ds and Grad_s * v * dz/ds, i.e. the component of 
        !- vertical velocity which depends on the horizontal velocity resp. the 
        !- horizontal water flux. 

        !- Set the default values
        do k=1,kb
          mi = mmk(k, i, j)

          !-- default four-point stencil around actual grid point
          hue=hu(mi)*un(mi)
          huw=hu(mmk(k,i,j-1))*un(mmk(k,i,j-1))
          hvn=hv(mmk(k,i-1,j))*vn(mmk(k,i-1,j))
          hvs=hv(mi)*vn(mi)

          ! -- overrule defaults according to border and corner
          !at points at western boundary
          if (krz(3,nnz) == 1) then
            huw=uwest(k)*huwest(k)
            if(ncorner > 0) then
              hvn=vnorth(k)*hvnorth(k)
            endif

          !at points at northern boundary
          elseif (krz(3,nnz) == 2) then
            hvn=vnorth(k)*hvnorth(k)
            if(ncorner > 0) then
              hue=ueast(k)*hueast(k)
            endif

          !at points at eastern boundary 
          elseif (krz(3,nnz) == 3) then
            hue=ueast(k)*hueast(k)
            if(ncorner > 0) then
              hvs=vsouth(k)*hvsouth(k)
            endif

          !at points at southern boundary 
          elseif (krz(3,nnz) == 4) then
            hvs=vsouth(k)*hvsouth(k)
            if(ncorner > 0) then
              huw=uwest(k)*huwest(k)
            endif
          endif

          divk=(hue-huw)/dxi+(hvn*covoidy-hvs*covuidy)

          depth_n   = depth_n + h(mi) - dt*divk
          div(k)    = divk
          diverg(k) = divk          
          h_tmp(k)  = h(mi)
        enddo


        !- vertical redistribution of divergence so as to smooth layer 
        !- thicknesses wghthar*dh/dx(i) stands for Grad_s 
        !- (velocity * length * Grad_s (dz/ds)), whereas the inner Grad_s is 
        !- already in the factor dh, since dh = dh+h(mm(i,j,k))-h(mm(i,j+1,k))
        !- depends of course direcly on the space between two cells, which is 
        !- described by dx or dy.

        ! Grad_s(dz/ds) conform to ds*Grad_s(dz/ds) in descrete.
        ! dh = lenght * Grad_s (dz/ds)

        kbue = khu(mmk(1,i,j))
        dh = zero
        do k=1,kbue-1
          dh=dh+h(mmk(k,i,j))-h(mmk(k,i,j+1))
          div( k )=div( k )+wghthar*dh/dxi
          div(k+1)=div(k+1)-wghthar*dh/dxi
        end do
        kbuw=khu(mmk(1,i,j-1))
        dh = zero
        do k=1,kbuw-1
          dh=dh+h(mmk(k,i,j))-h(mmk(k,i,j-1))
          div( k )=div( k )+wghthar*dh/dxi
          div(k+1)=div(k+1)-wghthar*dh/dxi
        end do
        kbvn=khv(mmk(1,i-1,j))
        dh = zero
        do k=1,kbvn-1
          dh=dh+h(mmk(k,i,j))-h(mmk(k,i-1,j))
          div( k )=div( k )+wghthar*dh*covoidy
          div(k+1)=div(k+1)-wghthar*dh*covoidy
        end do
        kbvs=khv(mmk(1,i,j))
        dh = zero
        do k=1,kbvs-1
          dh=dh+h(mmk(k,i,j))-h(mmk(k,i+1,j))
          div( k )=div( k )+wghthar*dh*covuidy
          div(k+1)=div(k+1)-wghthar*dh*covuidy
        end do


        !- levelling at bottom steps, this part smoothes the water-transfer at 
        !- bottom steps, i.e. there where a water column is deeper than the 
        !- neighbour-water-column.

        kbw = kh(mmk(1,i,j-1))
        if (kb > kbw .and. kbw > 0) then
          dh = ti(mmk(1,i,j-1))+z(mmk(1,i,j-1))
          do k=1,kbw
            dh = dh-h(mmk(k,i,j))
          end do
          dh = max(dh,zero)
          div(kbw)   = div(kbw)   - gammadt*dh
          div(kbw+1) = div(kbw+1) + gammadt*dh
        end if

        kbe = kh(mmk(1,i,j+1))
        if (kb > kbe .and. kbe > 0) then
          dh = ti(mmk(1,i,j+1))+z(mmk(1,i,j+1))
          do k=1,kbe
            dh = dh-h(mmk(k,i,j))
          end do
          dh = max(dh,zero)
          div(kbe)   = div(kbe)   - gammadt*dh
          div(kbe+1) = div(kbe+1) + gammadt*dh
        end if

        kbn = kh(mmk(1,i-1,j))
        if (kb > kbn .and. kbn > 0) then
          dh = ti(mmk(1,i-1,j))+z(mmk(1,i-1,j))
          do k=1,kbn
            dh = dh-h(mmk(k,i,j))
          end do
          dh = max(dh,zero)
          div(kbn)   = div(kbn)   - gammadt*dh
          div(kbn+1) = div(kbn+1) + gammadt*dh
        end if

        kbs = kh(mmk(1,i+1,j))
        if (kb > kbs .and. kbs > 0) then
          dh = ti(mmk(1,i+1,j))+z(mmk(1,i+1,j))
          do k=1,kbs
            dh = dh-h(mmk(k,i,j))
          end do
          dh = max(dh,zero)
          div(kbs)   = div(kbs)   - gammadt*dh
          div(kbs+1) = div(kbs+1) + gammadt*dh
        end if

        norm = one
        scal = zero
        iterat = 0
      
        a_h(1:kmax)   = zero
        b_h(1:kmax)   = zero
        c_h(1:kmax)   = zero
        e_h(1:kmax)   = zero
        r_h(1:kmax+1) = zero
        z_h(1:kmax+1) = zero
      
        it_max = 25 * kb

        !- Now the last term of the equation which is to be solved will be 
        !- integrated:
        !- dt*(wghteta+(wghtturb*avv))*(length_b/h_tmp( k )-length_a/h_tmp(k-1))
        !- is the parametrisation of the term d/ds * (d/ds ( mu / (dz/ds) ) ). 
        !- This parametrisation prevents the case that one layer disappears, 
        !- i.e. the layer thickness becomes zero, because this term trys always 
        !- to push the thicknesses in the origin state.

        !- To solve the whole equation of continuity over the vertical the 
        !- Newton iteration is used.

        if (depth_n > zero) then
          iteration: do while ( norm > norm_min .and. scal <= scal_max )

            do k=1,kb
              c_h(k) = one+wghtz
              e_h(k) = (one+wghtz)*(h(mmk(k,i,j))-h_tmp(k)) - dt*div(k)
            end do

            do k=1, kb
              e_h(1) = e_h(1) - dt*wghtz*div(k)
            enddo

            do k=2,kb
              !- the layer thicknesses h0 act like a reference thickness or
              !- resting state
              length_a = h0(mmk(k-1,i,j))
              length_b = h0(mmk(k,i,j))

              h_tmp_a  = h_tmp(k-1)
              h_tmp_b  = h_tmp(k)
    
              fac = wghteta + wghtturb*avv(mmk(k,i,j))
    
              a_h( k ) =        - dt*fac*length_a/(h_tmp_a*h_tmp_a)
              c_h( k ) = c_h(k) + dt*fac*length_b/(h_tmp_b*h_tmp_b)
              e_h( k ) = e_h(k) + dt*fac*( length_b/h_tmp( k )                 &
                                - length_a/h_tmp(k-1) )

              b_h(k-1) =         - dt*fac*length_b/(h_tmp_b*h_tmp_b)
              c_h(k-1) = c_h(k-1)+ dt*fac*length_a/(h_tmp_a*h_tmp_a)
              e_h(k-1) = e_h(k-1)+ dt*fac*( length_a/h_tmp(k-1)                &
                                 - length_b/h_tmp( k ) )
            end do


            !- line inversion (factorization) method to compute iterative 
            !- increments of cell thickness
            !- solved by using the Thomas algorithm the system of equations 
            !- c_h(k)*h_inc(k)+a_h(k)*h_inc(k-1)+b_h(k)*h_inc(k+1)=e_h(k) to 
            !- get the  increment h_inc of the layer thicknesses to calculate 
            !- the new h(:) ((hn(:)). 

            do k=1,kb
              r_h(k+1)=(e_h(k)-a_h(k)*r_h(k))/(c_h(k)+a_h(k)*z_h(k))
              z_h(k+1)=-b_h(k)/(c_h(k)+a_h(k)*z_h(k))
            end do

            h_inc(kb)=r_h(kb+1)
            do k=kb-1,1,-1
              h_inc(k)=z_h(k+1)*h_inc(k+1)+r_h(k+1)
            end do

            norm = zero
            scal = one
            do k=1,kb
              norm = max(norm,abs(h_inc(k)))
              scal = max(scal,-two*(h_inc(k)/h_tmp(k)))
            end do

            do k=1,kb
              h_inc(k) = h_inc(k) / scal
              h_tmp(k) = h_tmp(k) + h_inc(k)
            enddo

            iterat = iterat + 1
            if ( iterat > it_max ) call exitme(1,'Undue number of iterations')

          end do iteration
    
        elseif ( depth_n <= zero ) then
          !- In this case the water column would run originally dry - so set
          !- all layersthicknesses to 5 cm and write a WARNING to logfile.
          !- Most probably a coarse grid cell fell dry, which is also covered 
          !- by a finer grid, so that the real cell thickness (at the end of the
          !- timestep comes from fine grid anyway! 
          write(iu06,*) 'WARNING: In Grid ', ia, ' a column ran empty: line: ',&
                         i, ' column: ', j,' Number of layers: ', kb

          do k=1,kb
            h_tmp(k) = hgr
          enddo
        endif

        !- Calculate the pseudo vertical velocity by using the new layer 
        !- thicknesses.
        w_pseudo = zero
        do k=kb,1,-1
          w_pseudo = w_pseudo - (diverg(k) + (h_tmp(k)-h(mmk(k,i,j)))/dt )
          wpsd(k) = w_pseudo
        end do

        dh = zero
        do k=1,kb
          mi = mmk(k, i, j)
          hn(mi) = h_tmp(k)
          ws(mi) = wpsd(k)
          dh = dh + h_tmp(k)
        end do
        ws(mmk(1,i,j)) = zero
          
        !- Calculate the new sea level zn.      
        zn(mmk(1,i,j))= dh - ti(mmk(1,i,j))
            
      end do

end subroutine masseqs_c_f_dyn

!===============================================================================

subroutine mom_drying_bnd_f_c (nz1,nz2,krz,mmkf,khf,kfg,iga,jga,hzf,mmkc,      &
                               hzc,zc,unc,vnc)

      !- modules ---------------------------------------------------------------
      use constants,    only : hgr
      use dmi_omp,      only : domp_get_domain
      use drying_cases, only : drying_case
      use exits,        only : exitme

      !- directives ------------------------------------------------------------
      implicit none
  
      integer(4), intent(in) :: nz1,nz2
      integer(4), intent(in) :: krz(:,0:)
      integer(4), intent(in) :: mmkf(:,0:,0:)
      integer(4), intent(in) :: khf(0:)
      integer(4), intent(in) :: kfg(:)
      integer(4), intent(in) :: iga(:),jga(:)
      integer(4), intent(in) :: mmkc(:,0:,0:)
      real(8), intent(in)    :: hzf(0:)
      real(8), intent(in)    :: hzc(0:), zc(0:)
      real(8), intent(inout) :: unc(0:),vnc(0:)
  
      !- local vars ------------------------------------------------------------
      integer(4)  :: nnz, kb, i, j, k, kc, kcf, nzl, nzu
      integer(4)  :: ig,jg,igr,jgr, mic, eec, ssc, gam
      
      !- local parameters ------------------------------------------------------
      real(8), parameter :: twohgr = two*hgr
      
      if(iga(2) == jga(2) ) then
        gam = iga(2)
      else
        gam = iga(2)*jga(2)
      endif

      call domp_get_domain(nz1,nz2/gam, nzl, nzu)

      do nnz = (nzl-1)*gam+1, nzu*gam
        i = krz(1,nnz)
        j = krz(2,nnz)

        kb = khf(mmkf(1,i,j))

        do k = 1,kb
          if( hzf(mmkf(k,i,j)) <= twohgr) then
            ig=int((i-1)/iga(2))
            jg=int((j-1)/jga(2))
            igr=iga(1)+ig
            jgr=jga(1)+jg

            kc  = 0
            kcf = 0
            do while(kcf < k)
              kc  = kc + 1 
              kcf = kcf + kfg(kc)
            enddo

            if(krz(3,nnz) == 1) then
              jgr = jgr-1
              mic = mmkc(kc,igr,jgr)
              eec = mmkc(kc,igr,jgr+1)
              if (mic > 0 .and. eec > 0) then
                call drying_case(mmkc(1,igr,jgr), mmkc(1,igr,jgr+1), unc(mic), &
                                 hzc(mic), hzf(mmkf(k,i,j)), zc)
              endif
            elseif( krz(3,nnz) == 2) then
              igr = igr-1
              mic = mmkc(kc,igr,jgr)
              ssc = mmkc(kc,igr+1,jgr)
              if (mic > 0 .and. ssc > 0) then
                call drying_case(mmkc(1,igr+1,jgr), mmkc(1,igr,jgr), vnc(mic), &
                                 hzf(mmkf(k,i,j)), hzc(mic), zc) 
              endif
            elseif( krz(3,nnz) == 3) then
              mic = mmkc(kc,igr,jgr)
              eec = mmkc(kc,igr,jgr+1)
              if (mic > 0 .and. eec > 0) then
                call drying_case(mmkc(1,igr,jgr), mmkc(1,igr,jgr+1), unc(mic), &
                                 hzf(mmkf(k,i,j)), hzc(eec), zc)
              endif
            elseif( krz(3,nnz) == 4) then
              mic = mmkc(kc,igr,jgr)
              ssc = mmkc(kc,igr+1,jgr)
              if (mic > 0 .and. ssc > 0) then
                call drying_case(mmkc(1,igr+1,jgr), mmkc(1,igr,jgr), vnc(mic), &
                                 hzc(ssc), hzf(mmkf(k,i,j)), zc) 
              endif
            else
              call exitme(1,'Wrong boundary type in mom_drying_bnd_f_c')
            endif
          endif
        enddo
      enddo      

end subroutine mom_drying_bnd_f_c

end module nesting
