module netcdf_interface

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !! This module should be the interface between the HBM-Code and the netcdf- !!
  !! libraries, which means that all functions etc. which come from netcdf    !!
  !! were processed in this file.                                             !!
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  !- directives ----------------------------------------------------------------
  implicit none

  !- interfaces ----------------------------------------------------------------
  interface ncdf_set_3D_variables
    module procedure ncdf_set_3D_variables_int
    module procedure ncdf_set_3D_variables_log
  end interface
  
  interface ncdf_set_variable_attr
    module procedure ncdf_set_variable_attr_int
    module procedure ncdf_set_variable_attr_byte
  end interface
    
  !- public routines -----------------------------------------------------------
  public :: ncdf_create_new, ncdf_open, ncdf_end_definitions, ncdf_close_file, &
            ncdf_set_global_attr, ncdf_set_dimensions, ncdf_set_dim_attr,      &
            ncdf_set_3D_variables,ncdf_set_4D_variables,ncdf_set_variable_attr,&
            ncdf_save_time, ncdf_save_3D_vars, ncdf_save_4D_vars,              &
            ncdf_request_varid, ncdf_read_var, ncdf_set_bio_var_attrib,        &
            ncdf_save_dim_vars,ncdf_save_4D_biovars, ncdf_set_sed_var_attrib,  &
            ncdf_set_chl_var_attrib

  !- private routines ----------------------------------------------------------
  private :: ncdf_check_status, ncdf_set_attrib, ncdf_set_3D_variables_int,    &
             ncdf_set_3D_variables_log, ncdf_set_variable_attr_int,            &
             ncdf_set_variable_attr_byte, ncdf_set_attrib_bio
  
contains

  subroutine ncdf_check_status (status)
    !---------------------------------------------------------------------------
    ! wrap calls of NF90 functions into this poor-mans' errorhandler
    !---------------------------------------------------------------------------

    use dmi_mpi_global, only : iu06
    use netcdf        , only : nf90_noerr, nf90_strerror
  
    implicit none

    integer(4), intent(in) :: status

    if (status /= nf90_noerr) then
      write(iu06,*) 'ERROR: '//trim(nf90_strerror(status)) 
    endif 

  end subroutine ncdf_check_status

  !=============================================================================

  subroutine ncdf_open(filename,ncid)
  
    use netcdf, only: nf90_open, nf90_nowrite

    implicit none
  
    character(60),intent(in) :: filename
    integer(4),intent(out)   :: ncid
  
    call ncdf_check_status(nf90_open(trim(filename),nf90_nowrite,ncid)) 

  end subroutine ncdf_open

  !=============================================================================

  subroutine ncdf_set_attrib(ncid, varid,                                      &
                                   standard_name,                              &
                                   units,                                      &
                                   unit_long,                                  &
                                   axis,                                       &
                                   long_name,                                  &
                                   positive,                                   &
                                   scale_factor,                               &
                                   add_offset,                                 &
                                   FillValue,                                  &
                                   FillValue_byte,                             &
                                   missing_value,                              &
                                   missing_value_byte )
  !-----------------------------------------------------------------------------
  !  Set attributes for an item varid in netCDF file ncid. 
  !
  !  The list of recognized attributes is easily expandable, if more attributes
  !  are required.
  !-----------------------------------------------------------------------------

    !- modules
    use netcdf, only: nf90_put_att

    !- directives:
    implicit none

    !- arguments:
    integer,                intent(in) :: ncid, varid
    character(*), optional, intent(in) :: standard_name, units, long_name,     &
                                          unit_long
    character(*), optional, intent(in) :: positive, axis
    real(8),      optional, intent(in) :: add_offset
    real(8),      optional, intent(in) :: scale_factor
    integer(4),   optional, intent(in) :: FillValue, missing_value
    integer(1),   optional, intent(in) :: FillValue_byte, missing_value_byte

    if (present(standard_name)) then
      call ncdf_check_status( nf90_put_att( ncid, varid, 'standard_name',      &
                                            trim(standard_name)) )
    endif

    if (present(long_name)) then
      call ncdf_check_status( nf90_put_att(ncid, varid, 'long_name',           &
                                           trim(long_name)) )
    endif

    if (present(units)) then
      call ncdf_check_status( nf90_put_att(ncid,varid,'units',trim(units)) )
    endif

    if (present(unit_long)) then
      call ncdf_check_status( nf90_put_att(ncid, varid, 'unit_long',           &
                                           trim(unit_long)) )
    endif

    if (present(axis)) then
      call ncdf_check_status( nf90_put_att(ncid,varid,'axis',trim(axis)) )
    endif

    if (present(positive)) then
      call ncdf_check_status( nf90_put_att(ncid, varid, 'positive',            &
                                           trim(positive)) )
    endif

    if (present(scale_factor)) then
      call ncdf_check_status( nf90_put_att(ncid, varid, 'scale_factor',        &
                                           scale_factor) )
    endif

    if (present(add_offset)) then
      call ncdf_check_status( nf90_put_att(ncid,varid,'add_offset',add_offset) )
    endif

    if (present(FillValue)) then
      call ncdf_check_status( nf90_put_att(ncid,varid,'_FillValue',FillValue) )
    endif

    if (present(FillValue_byte)) then
      call ncdf_check_status( nf90_put_att(ncid, varid, '_FillValue',          &
                                           FillValue_byte) )
    endif

    if (present(missing_value)) then
      call ncdf_check_status(nf90_put_att(ncid, varid, 'missing_value',        &
                                          missing_value))
    endif

    if (present(missing_value_byte)) then
      call ncdf_check_status(nf90_put_att(ncid, varid, 'missing_value',        &
                                          missing_value_byte))
    endif

  end subroutine ncdf_set_attrib

  !=============================================================================

  subroutine ncdf_create_new( filename, ncid )
  !-----------------------------------------------------------------------------
  ! interface to create a new netCDF-file
  !-----------------------------------------------------------------------------

    use netcdf, only: nf90_create, NF90_NOCLOBBER

    implicit none

    character(40), intent(in) :: filename 
    integer(4),   intent(out) :: ncid

    call ncdf_check_status( nf90_create(trim(filename), NF90_NOCLOBBER, ncid) )

  end subroutine ncdf_create_new
  
  !=============================================================================

  subroutine ncdf_end_definitions( ncid )
  !-----------------------------------------------------------------------------
  ! intertace to end NetCDF definitions
  !-----------------------------------------------------------------------------

    use netcdf, only: nf90_enddef

    implicit none

    integer(4),    intent(in) :: ncid

    call ncdf_check_status( nf90_enddef(ncid) )

  end subroutine ncdf_end_definitions

  !=============================================================================

  subroutine ncdf_close_file( ncid )
  !-----------------------------------------------------------------------------
  ! interface to close a netCDF-file
  !-----------------------------------------------------------------------------

    use netcdf, only: nf90_close

    implicit none

    integer(4),   intent(in) :: ncid

    call ncdf_check_status( nf90_close(ncid) )

  end subroutine ncdf_close_file
  
  !=============================================================================
  
  subroutine ncdf_set_global_attr ( ncid,lonmin,lonmax,latmin,latmax,zmin,zmax )
  !-----------------------------------------------------------------------------
  ! The global attributes of every netcdf-file is set here
  ! Every file should be a 4D-file with latitude, longitude, depth and time
  ! Is only one point or one layer saved in a file, so the regarding variable
  ! should have the same min and max (the only value, which should be saved
  !-----------------------------------------------------------------------------
 
    use netcdf, only: nf90_put_att, NF90_GLOBAL 

    implicit none

    integer(4), intent(in)        :: ncid
    real(8), intent(in)           :: lonmin,lonmax,latmin,latmax
    real(8), intent(in)           :: zmin,zmax

    integer, parameter  :: varid = NF90_GLOBAL


    !- Describe convention:
    call ncdf_check_status( nf90_put_att(ncid,varid,'Conventions','CF-1.0') )

!    !- Promote the department:
!    call ncdf_check_status( nf90_put_att(ncid,varid,'references',             &
!                            'http://www.bsh.de') )
!    call ncdf_check_status( nf90_put_att(ncid,varid,'contact',                &
!                            'thorger.bruening@bsh.de') )

    !- Origin of the data:
    call ncdf_check_status( nf90_put_att(ncid,varid,'source',                  &
                            'HIROMB-BOOS-model output') )
 
    !- Domain extent:
    call ncdf_check_status( nf90_put_att(ncid,varid,'longitude_min',lonmin ) )
    call ncdf_check_status( nf90_put_att(ncid,varid,'longitude_max',lonmax ) )
    call ncdf_check_status( nf90_put_att(ncid,varid,'latitude_min', latmin ) )
    call ncdf_check_status( nf90_put_att(ncid,varid,'latitude_max', latmax ) )
    call ncdf_check_status( nf90_put_att(ncid,varid,'depth_min',    zmin   ) )
    call ncdf_check_status( nf90_put_att(ncid,varid,'depth_max',    zmax   ) )

  end subroutine ncdf_set_global_attr

  !=============================================================================

  subroutine ncdf_set_dimensions (ncid, nlon, lonid,  nlat, latid,ndpth,dpthid,&
                                        tmid )
  !-----------------------------------------------------------------------------
  ! Set the spactial coordinate and time dimensions - all fields should be 4D  
  ! fields (longitude, latitude, depth and time)
  !
  ! EXAMPLE:
  !
  ! dimensions:
  !  lat=73;
  !  lon=96;
  !  depth=4;
  ! variables:
  !  int salinity(time,depth,lat,lon);
  !  double depth(depth);
  !  double lat(lat);
  !  double lon(lon);
  !  double time(time);
  ! data:
  !  ...
  !-----------------------------------------------------------------------------
 
    use netcdf, only: nf90_def_dim, NF90_UNLIMITED

    implicit none

    integer(4), intent(in)            :: ncid, nlon, nlat, ndpth

    integer(4), intent(out)           :: lonid, latid, tmid, dpthid

    call ncdf_check_status( nf90_def_dim(ncid, "lon",   nlon,          lonid) )
    call ncdf_check_status( nf90_def_dim(ncid, "lat",   nlat,          latid) )
    call ncdf_check_status( nf90_def_dim(ncid, "depth", ndpth,        dpthid) )
    call ncdf_check_status( nf90_def_dim(ncid, "time",  NF90_UNLIMITED, tmid) )

  end subroutine ncdf_set_dimensions

  !=============================================================================

  subroutine ncdf_set_dim_vars (ncid, lonid, lonvid, latid, latvid, dpthid,    &
                                      dpthvid, tmid, tmvid )
  !-----------------------------------------------------------------------------
  ! Define the spacial and temporal coordinates (independent) variables
  !-----------------------------------------------------------------------------
 
    use netcdf, only: nf90_def_var, NF90_DOUBLE

    implicit none

    integer(4), intent(in)            :: ncid
    integer(4), intent(in)            :: lonid, latid, dpthid, tmid
    integer(4), intent(out)           :: lonvid, latvid, dpthvid, tmvid

    call ncdf_check_status( nf90_def_var(ncid,"lon", NF90_DOUBLE,lonid, lonvid))
    call ncdf_check_status( nf90_def_var(ncid,"lat", NF90_DOUBLE,latid, latvid))
    call ncdf_check_status( nf90_def_var(ncid, "depth", NF90_DOUBLE, dpthid,   &
                                         dpthvid))
    call ncdf_check_status( nf90_def_var(ncid,"time",NF90_DOUBLE,tmid,   tmvid))

  end subroutine ncdf_set_dim_vars 

  !=============================================================================

  subroutine ncdf_set_dim_attr (ncid, lonvid, latvid, dpthvid, tmvid )
  !-----------------------------------------------------------------------------
  ! Define attributes for spatial and temporal coordinate (independent) 
  ! variables
  !-----------------------------------------------------------------------------

    implicit none

    integer(4), intent(in)           :: ncid
    integer(4), intent(in)           :: lonvid, dpthvid, latvid, tmvid
    
    call ncdf_set_attrib(ncid, lonvid,                                         &
                             axis          = 'x',                              &
                             standard_name = 'longitude',                      &
                             units         = 'degrees_east',                   &
                             unit_long     = 'Degrees East',                   &
                             long_name     = 'longitude'      )
                           
    call ncdf_set_attrib(ncid, latvid,                                         &
                             axis          = 'y',                              &
                             standard_name = 'latitude',                       &
                             units         = 'degrees_north',                  &
                             unit_long     = 'Degrees North',                  &
                             long_name     = 'latitude'       )
  
    call ncdf_set_attrib(ncid, dpthvid,                                        &
                             axis          = 'z',                              &
                             standard_name = 'depth',                          &
                             units         = 'm',                              &
                             unit_long     = 'meters',                         &
                             long_name     = 'depth',                          &
                             positive      = 'down'           )

    call ncdf_set_attrib(ncid, tmvid,                                          &
                             axis          = 't',                              &
                             standard_name = 'time',                           &
                             long_name     = 'time',                           &
                             units         = 'days since 1900-01-01 00:00:00' )


  end subroutine ncdf_set_dim_attr

  !=============================================================================

  subroutine ncdf_save_dim_vars(ncid, lonvid, rlon, latvid, rlat, dpthvid, zlev)
  !-----------------------------------------------------------------------------
  ! Save the spatial dimension variables in the NetCDF data file
  !-----------------------------------------------------------------------------
 
    use netcdf, only: nf90_put_var, nf90_sync

    implicit none
    
    integer(4), intent(in)           :: ncid
    integer(4), intent(in)           :: lonvid, latvid, dpthvid
    real(8), intent(in)              :: rlon(:), rlat(:), zlev(:)

    call ncdf_check_status( nf90_put_var(ncid,  lonvid, rlon) )
    call ncdf_check_status( nf90_put_var(ncid,  latvid, rlat) )
    call ncdf_check_status( nf90_put_var(ncid, dpthvid, zlev) )  
 
    call ncdf_check_status( nf90_sync(ncid) )

  end subroutine ncdf_save_dim_vars
  
  !=============================================================================

  subroutine ncdf_set_3D_variables_int (ncid, varname, lonvid, latvid, tmvid,  &
                                        varid) 
  !-----------------------------------------------------------------------------
  ! Define the dependent 3D-variables
  !-----------------------------------------------------------------------------
 
    use netcdf, only: nf90_def_var, NF90_INT

    implicit none

    integer(4), intent(in)            :: ncid
    character(*), intent(in)          :: varname
    integer(4), intent(in)            :: lonvid, latvid, tmvid
    integer(4), intent(out)           :: varid

    integer(4), dimension(1:3)  :: dims3D

    dims3D(1) = lonvid 
    dims3D(2) = latvid
    dims3D(3) = tmvid

    call ncdf_check_status( nf90_def_var(ncid,varname, NF90_INT,dims3D,varid)) 
 
  end subroutine ncdf_set_3D_variables_int

  !=============================================================================

  subroutine ncdf_set_3D_variables_log (ncid, varname, lonvid, latvid, tmvid,  &
                                        varid, logbyt) 
  !-----------------------------------------------------------------------------
  ! Define the dependent 3D-variables
  !-----------------------------------------------------------------------------
 
    use netcdf, only: nf90_def_var, NF90_BYTE

    implicit none

    integer(4), intent(in)            :: ncid
    character(*), intent(in)          :: varname
    integer(4), intent(in)            :: lonvid, latvid, tmvid
    logical, intent(in)               :: logbyt
    integer(4), intent(out)           :: varid

    integer(4), dimension(1:3)  :: dims3D

    dims3D(1) = lonvid 
    dims3D(2) = latvid
    dims3D(3) = tmvid

    call ncdf_check_status( nf90_def_var(ncid,varname, NF90_BYTE,dims3D,varid)) 
 
  end subroutine ncdf_set_3D_variables_log
  
  !=============================================================================

  subroutine ncdf_set_4D_variables (ncid, varname, lonvid,latvid,dpthvid,tmvid,&
                                    varid ) 
  !-----------------------------------------------------------------------------
  ! Define the dependent 4D-variables
  !-----------------------------------------------------------------------------
 
    use netcdf, only: nf90_def_var, NF90_INT

    implicit none

    integer(4), intent(in)            :: ncid
    character(*), intent(in)          :: varname
    integer(4), intent(in)            :: lonvid, latvid, dpthvid, tmvid
    integer(4), intent(out)           :: varid

    integer(4), dimension(1:4) :: dims4D

    dims4D(1) = lonvid 
    dims4D(2) = latvid
    dims4D(3) = dpthvid
    dims4D(4) = tmvid

    call ncdf_check_status( nf90_def_var(ncid,varname,NF90_INT,dims4D,varid) )
  
  end subroutine ncdf_set_4D_variables

  !=============================================================================

  subroutine ncdf_set_variable_attr_int (ncid,varid,stand_name,unit,long_unit, &
                                     long_name, delete_value, scale )
  !-----------------------------------------------------------------------------
  ! Define attributes for the dependent wind-variables
  !-----------------------------------------------------------------------------
  
    implicit none

    integer(4), intent(in)        :: ncid, varid
    character(*), intent(in)      :: stand_name, unit, long_unit, long_name
    integer(4), intent(in)        :: delete_value
    real(8), intent(in)           :: scale
  
    call ncdf_set_attrib(ncid, varid,                                          &
                         standard_name = stand_name,                           &
                         units         = unit,                                 &
                         unit_long     = long_unit,                            &
                         long_name     = long_name,                            &
                         FillValue     = delete_value,                         &
                         missing_value = delete_value,                         &
                         scale_factor  = scale  )

  end subroutine ncdf_set_variable_attr_int

  !=============================================================================

  subroutine ncdf_set_variable_attr_byte(ncid,varid,stand_name,unit,long_name, &
                                         delete_value )
  !-----------------------------------------------------------------------------
  ! Define attributes for the dependent wind-variables
  !-----------------------------------------------------------------------------
  
    implicit none

    integer(4), intent(in)        :: ncid, varid
    character(*), intent(in)      :: stand_name, unit, long_name
    integer(4), intent(in)        :: delete_value
  
    call ncdf_set_attrib(ncid, varid,                                          &
                         standard_name       = stand_name,                     &
                         units               = unit,                           &
                         long_name           = long_name,                      &
                         FillValue_byte      = int(delete_value, kind = 1),    &
                         missing_value_byte  = int(delete_value, kind = 1)  )

  end subroutine ncdf_set_variable_attr_byte

  !=============================================================================

  subroutine ncdf_save_time (ncid, tmvid, itime, counter )
  !-----------------------------------------------------------------------------
  ! Save actual time in the NetCDF data file
  !-----------------------------------------------------------------------------
 
    use netcdf, only: nf90_put_var

    implicit none

    integer(4), intent(in)        :: ncid
    integer(4), intent(in)        :: tmvid
    real(8), intent(in)           :: itime
    integer(4), intent(in)        :: counter 
  
    call ncdf_check_status( nf90_put_var(ncid,tmvid,itime,start=(/ counter /)) )

  end subroutine ncdf_save_time

  !=============================================================================

  subroutine ncdf_save_3D_vars (ncid, id, tim, dummy_3D )
  !-----------------------------------------------------------------------------
  ! Save actual dependent 3D-variables in the NetCDF data file
  !-----------------------------------------------------------------------------

    use netcdf, only: nf90_put_var, nf90_sync

    implicit none

    integer(4), intent(in)       :: ncid
    integer(4), intent(in)       :: id
    integer(4), intent(in)       :: tim
    integer(4), intent(in)       :: dummy_3D(:,:)

    integer(4)           :: start3(1:3)

    start3(1) = 1
    start3(2) = 1
    start3(3) = tim

    call ncdf_check_status(nf90_put_var(ncid,id,dummy_3D(:,:),start=start3) )
  
    !- Syncronize:
    call ncdf_check_status( nf90_sync(ncid) )

  end subroutine ncdf_save_3D_vars

  !=============================================================================

  subroutine ncdf_save_4D_vars (ncid, id, tim, dummy_4D )
  !-----------------------------------------------------------------------------
  ! Save actual dependent 4D-variables in the NetCDF data file
  !-----------------------------------------------------------------------------
 
    use netcdf, only: nf90_put_var, nf90_sync

    implicit none

    integer(4), intent(in)       :: ncid
    integer(4), intent(in)       :: id
    integer(4), intent(in)       :: tim
    integer(4), intent(in)       :: dummy_4D(:,:,:)

    integer(4)            :: start3(1:4)

    start3(1) = 1
    start3(2) = 1
    start3(3) = 1
    start3(4) = tim

    call ncdf_check_status(nf90_put_var(ncid,id,dummy_4D(:,:,:),start=start3) )

    !- Syncronize:
    call ncdf_check_status( nf90_sync(ncid) )

  end subroutine ncdf_save_4D_vars
 
  !=============================================================================

  subroutine ncdf_request_varid(ncid,stand_name,varid)
 
    use netcdf, only: nf90_inq_varid
 
    implicit none
    
    character(100), intent(in) :: stand_name
    integer(4),     intent(in) :: ncid
    
    integer(4),intent(out):: varid
    
    call ncdf_check_status(nf90_inq_varid(ncid,trim(stand_name),varid))
    
  end subroutine ncdf_request_varid
  
  !=============================================================================

  subroutine ncdf_read_var(ncid,varid,time,init_data, mmx, nmx, kmx)

    use netcdf, only: nf90_get_var
    
    integer(4),intent(in) :: ncid, mmx, nmx, kmx
    integer(4),intent(in) :: varid
    integer(4),intent(in) :: time
    real(8),intent(out)   :: init_data(:,:,:,:)
 
    integer(4)            :: start(1:4), count(1:4)

    start(1) = 1
    start(2) = 1
    start(3) = 1
    start(4) = 1

    count(1) = nmx
    count(2) = mmx
    count(3) = kmx
    count(4) = time
    

    call ncdf_check_status(nf90_get_var(ncid,varid,init_data,start=start,      &
                           count=count))

  end subroutine ncdf_read_var

  !=============================================================================

   subroutine ncdf_set_bio_var_attrib(ncid, varid, ntbio)
 
     use exits,            only : exitme 
     use io_subs,          only : io_new_unit

     implicit none
 
     integer(4), intent(in) :: ncid, ntbio, varid(ntbio)
 
     character(17)  :: unit(ntbio)
     character(120) :: std_name(ntbio), standard_name
     character(12)  :: fnam_nml
     integer(4)     :: missing_value  
     integer(4)     :: lun,tr,ios
 
     !- get var units
     fnam_nml='units'
     lun = io_new_unit()
     open(lun,file=fnam_nml,form="formatted",status="old",iostat=ios)
     if (ios /= 0) call exitme(8,'Could not open '//trim(fnam_nml))

     do tr=1,ntbio 
       read(lun,*) unit(tr)
     enddo
     close(lun)
  
     !- Get variable names
     fnam_nml='nc_std_names'
     lun = io_new_unit()
     open(lun,file=fnam_nml,form="formatted",status="old",iostat=ios)
     if (ios /= 0) call exitme(8,'Could not open '//trim(fnam_nml))

     do tr=1,ntbio 
       read(lun,*) std_name(tr)
     enddo
     close(lun)
  
     do tr=1,ntbio
      call ncdf_set_attrib_bio(ncid,varid(tr),standard_name=trim(std_name(tr)),&
                            units=trim(unit(tr)), missing_value=-999)  
     enddo

  end subroutine ncdf_set_bio_var_attrib

  !=============================================================================

  subroutine ncdf_set_sed_var_attrib(ncid, varid)
 
    use exits,            only : exitme 
    use io_subs,          only : io_new_unit
   
    implicit none
 
    integer(4), intent(in) :: ncid, varid
 
    character(17)  :: unit
    character(120) :: name
    integer(4)     :: missing_value  
    integer(4)     :: lun,tr,ios
 
    !- get var units
    unit="micromolN/m3"
    name="mole_concentration_of_nitrogen_in_the_sediment"
  
    call ncdf_set_attrib_bio(ncid,varid,standard_name=trim(name),              &
                            units=trim(unit), missing_value=-999)  
 
  end subroutine ncdf_set_sed_var_attrib

  !=============================================================================

  subroutine ncdf_set_chl_var_attrib(ncid, varid)
 
    use exits,            only : exitme 
    use io_subs,          only : io_new_unit

    implicit none
 
    integer(4), intent(in) :: ncid, varid

    character(17)  :: unit
    character(120) :: name
    integer(4)     :: missing_value
    integer(4)     :: lun,tr,ios
 
    !- get var units
    unit="microgChl/m3"  
    name="mass_concentration_of_chlorophyll_in_seawater"
  
    call ncdf_set_attrib_bio(ncid,varid,standard_name=trim(name),              &
                            units=trim(unit), missing_value=-999)  
 
  end subroutine ncdf_set_chl_var_attrib

  !=============================================================================

  subroutine ncdf_save_4D_biovars (ncid, id, tim, dummy_4D)
  !-----------------------------------------------------------------------------
  ! Save actual dependent 4D-variables in the NetCDF data file
  !-----------------------------------------------------------------------------
 
    use netcdf, only: nf90_put_var, nf90_sync

    implicit none

    integer(4), intent(in)       :: ncid
    integer(4), intent(in)       :: id
    integer(4), intent(in)       :: tim
    integer(4), intent(in)          :: dummy_4D(:,:,:)

    integer(4)            :: start3(1:4)
  
    start3(1) = 1
    start3(2) = 1
    start3(3) = 1
    start3(4) = tim
    
    call ncdf_check_status(nf90_put_var(ncid,id,dummy_4D(:,:,:),start=start3))

    !- Syncronize:
    call ncdf_check_status( nf90_sync(ncid) )

  end subroutine ncdf_save_4D_biovars
 
  !=============================================================================

  subroutine ncdf_set_attrib_bio(ncid, varid,                                  &
                                   standard_name,                              &
                                   units,                                      &
                                   unit_long,                                  &
                                   axis,                                       &
                                   long_name,                                  &
                                   positive,                                   &
                                   scale_factor,                               &
                                   add_offset,                                 &
                                   FillValue,                                  &
                                   FillValue_byte,                             &
                                   missing_value,                              &
                                   missing_value_byte )
  !-----------------------------------------------------------------------------
  !  Set attributes for an item varid in netCDF file ncid. 
  !
  !  The list of recognized attributes is easily expandable, if more attributes
  !  are required.
  !-----------------------------------------------------------------------------

    !- modules
    use netcdf, only: nf90_put_att

    !- directives:
    implicit none

    !- arguments:
    integer,                intent(in) :: ncid, varid
    character(*), optional, intent(in) :: standard_name, units, long_name,     &
                                          unit_long
    character(*), optional, intent(in) :: positive, axis
    real(8),      optional, intent(in) :: add_offset
    real(8),      optional, intent(in) :: scale_factor
    integer(4),   optional, intent(in) :: FillValue, missing_value
    integer(1),   optional, intent(in) :: FillValue_byte, missing_value_byte

    if (present(standard_name)) then
      call ncdf_check_status( nf90_put_att( ncid, varid, 'standard_name',      &
                                            trim(standard_name)) )
    endif

    if (present(long_name)) then
      call ncdf_check_status( nf90_put_att(ncid, varid, 'long_name',           &
                                           trim(long_name)) )
    endif

    if (present(units)) then
      call ncdf_check_status( nf90_put_att(ncid,varid,'units',trim(units)) )
    endif

    if (present(unit_long)) then
      call ncdf_check_status( nf90_put_att(ncid, varid, 'unit_long',           &
                                           trim(unit_long)) )
    endif

    if (present(axis)) then
      call ncdf_check_status( nf90_put_att(ncid,varid,'axis',trim(axis)) )
    endif

    if (present(positive)) then
      call ncdf_check_status( nf90_put_att(ncid, varid, 'positive',            &
                                           trim(positive)) )
    endif

    if (present(scale_factor)) then
      call ncdf_check_status( nf90_put_att(ncid, varid, 'scale_factor',        &
                                           scale_factor) )
    endif

    if (present(add_offset)) then
      call ncdf_check_status( nf90_put_att(ncid,varid,'add_offset',add_offset) )
    endif

    if (present(FillValue)) then
      call ncdf_check_status( nf90_put_att(ncid,varid,'_FillValue',FillValue) )
    endif

    if (present(FillValue_byte)) then
      call ncdf_check_status( nf90_put_att(ncid, varid, '_FillValue',          &
                                           FillValue_byte) )
    endif

    if (present(missing_value)) then
      call ncdf_check_status(nf90_put_att(ncid, varid, 'missing_value',        &
                                          missing_value))
    endif

    if (present(missing_value_byte)) then
      call ncdf_check_status(nf90_put_att(ncid, varid, 'missing_value',        &
                                          missing_value_byte))
    endif

  end subroutine ncdf_set_attrib_bio

  !=============================================================================

end module netcdf_interface
