module momeqs_wrapping
  use constants, only : zero, one, two, half, qrt

  implicit none
  private

  !- interfaces ----------------------------------------------------------------
  interface momeqs
    module procedure momeqs_default_openacc
  end interface
  
  public  :: momeqs
  private :: momeqs_default_openacc
  ! HACK TO ENSURE INLINING - IPA and craypat doesn't work, cf. trisolver.f90
  private :: trisolv

contains 

! ------------------------------------------------------------------------------
!
! momeqs: sets up and solves the momentum equations using a tridiagonal solver
!
!         a(k)*s(k-1) + b(k)*s(k) + c(k)*s(k+1) = d(k)
!
! Based on the same numerical scheme as the original bewgl, but implementation 
! of the method into the actual code is modified completely for clairity, 
! simplicity and very much indeed for performance efficiency.
! ------------------------------------------------------------------------------
subroutine momeqs_default_openacc (m, kdry, iwet2, mm, dt, dx, ft, rt,         &
                   ty, tx, gty, gtx, nud, nvd, uvdam, nweir, z, u, v, hz,      &
                   un, vn, hx, hy, khu, khv, taux, tauy, pl, press, geopot,    &
                   rho, avv, eddyh, shear, stretch, div, eddyd,                &
                   casus, ice, ueis, veis, ind, kh, z0srf, idx, widx, zweir,   &
                   ia, botfricmap)

  !- modules -------------------------------------------------------------------
  use cmod_params,  only : kmax
  use constants,    only : dyn, weirs, hgr, rhow, rhoi, cdrag 
  use drying_cases, only : drying_case
  use weireqs,      only : weir
  use botfricmaps,  only : get_simple_bf_map

  ! ----------------------------------------------------------------------------
  !    g:     gravity
  !    hgr:   thickness limit (0.05m)
  !    rhow:  reference density of sea water
  !    rhoi:  reference density of sea ice
  !    cdrag: drag coefficient
  !    kmax:  max No of layers in all areas
  ! ----------------------------------------------------------------------------

  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in) :: m          ! north-south dimension of the field 
  integer(4), intent(in) :: kdry       ! number of layers which could fall dry
  integer(4), intent(in) :: iwet2      ! number of wet points in me
  integer(4), intent(in) :: mm(:,0:,0:)! index
  real(8), intent(in)    :: dt         ! time step size
  real(8), intent(in)    :: dx(:)      ! dx(i): grid spacing in x-direction
  real(8), intent(in)    :: ft(:)      ! ft(i):f(i)*dt, f(i) = Coriolis param
  real(8), intent(in)    :: rt         ! rt: r*dt, r = bed roughness
  real(8), intent(in)    :: ty         ! ty: dt/dy
  real(8), intent(in)    :: tx(:)      ! tx(i): dt/dx(i)
  real(8), intent(in)    :: gty        ! gty:g*dt/dy
  real(8), intent(in)    :: gtx(:)     ! gtx(i):g*dt/dx(i)
  integer(4), intent(in) :: nud        ! No. of x-dams
  integer(4), intent(in) :: nvd        ! No. of y-dams
  real(8),    intent(in) :: uvdam(1:,0:) ! 1.0/0.0 mask for dams
  integer(4), intent(in) :: nweir      ! No. of weirs
  integer(4), intent(in) :: widx(:,0:) ! weir-index
  real(8), intent(in)    :: z(0:)      ! surface elevation
  real(8), intent(in)    :: u(0:)      ! u at old time level
  real(8), intent(in)    :: v(0:)      ! v at old time level
  real(8), intent(in)    :: hz(0:)     ! layer thickness
  real(8), intent(in)    :: hx(0:)     ! layer thickness @ x-mom point
  real(8), intent(in)    :: hy(0:)     ! layer thickness @ y-mom point
  integer(4), intent(in) :: kh(0:)     ! No of layers
  integer(4), intent(in) :: ind(:,:)   ! i,j indices
  integer(4), intent(in) :: khu(0:)    ! No of x-mom layers
  integer(4), intent(in) :: khv(0:)    ! No of y-mom layers
  real(8), intent(in)    :: taux(0:)   ! wind stress, x-dir
  real(8), intent(in)    :: tauy(0:)   ! wind stress, y-dir
  real(8), intent(in)    :: pl(0:)     ! air pressure
  logical, intent(in)    :: casus(0:)  ! water-ice mask
  real(8), intent(in)    :: ice(1:,0:) ! ice parameters
  real(8), intent(in)    :: ueis(0:)   ! ice velocity in x-dir
  real(8), intent(in)    :: veis(0:)   ! ice velocity in y-dir 
  real(8), intent(in)    :: press(0:)  ! baroclinic pressure
  real(8), intent(in)    :: rho(0:)    ! denisty
  real(8), intent(in)    :: geopot(0:)
  real(8), intent(in)    :: avv(0:)    ! vertical eddy viscosity
  real(8), intent(in)    :: eddyh(0:)  ! horizontal eddy visc., Smagorinsky
  real(8), intent(in)    :: eddyd(0:)  ! horizontal eddy visc., divergence
  real(8), intent(in)    :: shear(0:)  ! horizontal velocity shear
  real(8), intent(in)    :: stretch(0:)! horizontal velocity stretch
  real(8), intent(in)    :: div(0:)    ! horizontal velocity divergence
  real(8), intent(in)    :: z0srf(0:)  ! surface roughness length scale
  real(8), intent(in)    :: botfricmap(0:)

  !- Output variables: ---------------------------------------------------------
  !  unmodified locations must have their values maintained, therefore INOUT
  real(8), intent(inout) :: zweir(0:)  ! congestion height of the weir
  real(8), intent(inout) :: un(0:)     ! u at new time level (0:iwet3)
  real(8), intent(inout) :: vn(0:)     ! v at new time level (0:iwet3)
  integer(4), intent(inout) :: idx(:,:)
  integer(4), intent(in) :: ia

  !- local vars ----------------------------------------------------------------
  integer(4) :: i, j, kev, keu, k
  integer(4) :: mi, mie, eee, sse, ee, ww, nn, ss, ne, se, sw
  integer(4) :: nl, nu, nsurf, i1, i2, i3, i4, j1, j2, j3, j4
  integer(4) :: mmi, mme, mms
  integer(4) :: insurf, nsurfl, nsurfu
  real(8)    :: eiscol, eiscor, eiscox, eiscoo, eiscou, eiscoy
  real(8)    :: diffu, difful, diffur, diffvl, diffvr, diffuv
  real(8)    :: diffv, diffvo, diffvu, diffuo, diffuu
  real(8)    :: umi, uzet, ulu 
  real(8)    :: vmi, vzet, vlu 
  real(8)    :: ustar, vaus, vcon, vpress
  real(8)    :: vstar, uaus, ucon, upress
  real(8)    :: urd, vrd, aurd, avrd
  real(8)    :: dtd, dtdr, dxi, txi, fti, rth, hrt, ttx, tty
  real(8)    :: fac, dec, zu, zr, zm, tmp
  real(8)    :: ehmi, ehnn, ehww, fy, fxh, fxm, fyh, dam

  integer(4), parameter :: istride = 32 

  real(8), dimension(istride,kmax) :: a, b, c, d, s, tm
  real(8), dimension(istride)      :: usrf, vsrf
  real(8), dimension(istride,kmax) :: vst, ust, fv

  real(8), parameter :: zero = 0.0_8, half = 0.5_8, qrt = 0.25_8 
  real(8), parameter :: one = 1.0_8, two = 2.0_8
  real(8), parameter :: cdri = cdrag*rhoi, drhow  = one/rhow
  real(8)    :: e
  integer(4) :: ii

  !- initialise on every entry -------------------------------------------------
  dtd   = dt/1.e5
  dtdr  = dt/rhow
  rth   = rt*half
!  call get_simple_bf_map(ia,nsurf,botfricmap,rt,rth)
  fyh   = ty
  fy    = qrt*ty

!- BEGIN GPU -------------------------------------------------------------------

!$acc  parallel loop                                                           &
!$acc& private (a, aurd, avrd, b, c, d, dec, diffu, difful, dam,               &
!$acc&            diffuo, diffur, diffuu, diffuv, diffv, diffvl,               &
!$acc&            diffvo, diffvr, diffvu, dxi, e, ee, eee, ehmi,               &
!$acc&            ehnn, ehww, eiscol, eiscoo, eiscor, eiscou, eiscox,          &
!$acc&            eiscoy, fv, fac, fti, fxh, fxm, hrt, i, i1, i2,              &
!$acc&            i3, i4, ii, insurf, j, j1, j2, j3, j4, k, kev, mi,           &
!$acc&            mie, mme, mmi, mms, ne, nn, nsurf, nsurfl, nsurfu, s, se,    &
!$acc&            ss, sse, sw, tm, tmp, ttx, tty, txi, uaus, ucon, ulu, umi,   &
!$acc&            upress, urd, usrf, ust, ustar, uzet, vaus, vcon, vlu,        &
!$acc&            vmi, vpress, vrd, vsrf, vst, vstar, vzet, ww, zm, zr, zu)    &
!$acc copyin(ind,mm,hx,u,eddyh,stretch,shear,eddyd,div,z0srf,press,pl,ft)      &
!$acc copyin(casus,ice,ueis,veis,v,taux,tauy,avv,khv,khu,hy,z,rho)             &
!$acc copyin(tx,gtx,dx,uvdam,kdry,geopot)                                      &
!$acc copy(un,vn)
  do insurf=1,iwet2,istride
    nsurfl = insurf
    nsurfu = min(iwet2,nsurfl + istride -1)

!$acc loop
    do nsurf = nsurfl,nsurfu
      if (khu(nsurf) == 0) then
        ! land east of point, maintain zero velocity:
        un(nsurf) = zero 
      else
        tm(nsurf-nsurfl+1,1) = hx(nsurf)
        if (khu(nsurf) > 1) then
          i = ind(1,nsurf)
          j = ind(2,nsurf)
          mmi = mm(2,i,j)
          tm(nsurf-nsurfl+1,2:khu(nsurf)) = hx(mmi:mmi+khu(nsurf)-2)
        endif
        ! set up the vertical eddy viscosity terms, coefficient a
        a(nsurf-nsurfl+1,1) = zero
      endif
    enddo


!$acc loop
    ! set up the vertical eddy viscosity terms, coefficients a, b and c:
    do nsurf = nsurfl,nsurfu
      if (khu(nsurf) .le. 1) cycle
      i = ind(1,nsurf)
      j = ind(2,nsurf)
      mmi = mm(2,i,j  )
      mme = mm(2,i,j+1)
      do k=1,khu(nsurf)-1
        tmp    = one/tm(nsurf-nsurfl+1,k)
        a(nsurf-nsurfl+1,k+1) = -dt*(avv(mmi+k-1)+avv(mme+k-1))/               &
                        (tm(nsurf-nsurfl+1,k)+tm(nsurf-nsurfl+1,k+1))
        a(nsurf-nsurfl+1,k)   = a(nsurf-nsurfl+1,k)  *tmp
        c(nsurf-nsurfl+1,k)   = a(nsurf-nsurfl+1,k+1)*tmp
        b(nsurf-nsurfl+1,k)   = one - (a(nsurf-nsurfl+1,k)+c(nsurf-nsurfl+1,k))
      enddo
    enddo


!$acc loop
    do nsurf = nsurfl,nsurfu
      if (khu(nsurf) == 0) cycle
      i = ind(1,nsurf)
      j = ind(2,nsurf)
      mie = nsurf
      eee = mm(1,i,j+1)
      a(nsurf-nsurfl+1,khu(nsurf)) = a(nsurf-nsurfl+1,khu(nsurf))/             &
                                     tm(nsurf-nsurfl+1,khu(nsurf))
      b(nsurf-nsurfl+1,khu(nsurf)) = one - a(nsurf-nsurfl+1,khu(nsurf))
      c(nsurf-nsurfl+1,khu(nsurf)) = zero
      ! air pressure term:
      ulu = (pl(eee)-pl(mie))*tx(i)*drhow
      ! zeta:
      zm = z(mie)
      zr = z(eee)
      ! pressure term:
      uzet = gtx(i)*(zr-zm)
      ! collect different surface terms:
      usrf(nsurf-nsurfl+1) = ulu + uzet
    enddo


    do k = 1,kmax 
!$acc loop
      do nsurf = nsurfl,nsurfu
        if (khu(nsurf) == 0) cycle
        if (k.gt.khu(nsurf)) cycle
        i = ind(1,nsurf)
        j = ind(2,nsurf)
        ! do each layer, coefficient d:
        dxi = dx(i)
        txi = tx(i)
        fti = ft(i)
        fxh = txi
        ttx = two*txi
        ! extract commonly used indices:
        mi = mm(k,i  ,j  )
        nn = mm(k,i-1,j  )
        ss = mm(k,i+1,j  )
        ne = mm(k,i-1,j+1)
        ee = mm(k,i  ,j+1)
        se = mm(k,i+1,j+1)

        ! center value of u:
        umi = u(mi)

        ! v centred @ u point:
        vst(nsurf-nsurfl+1,k)  = qrt*(v(nn) + v(ne) + v(mi) + v(ee))

        ! baroclinic pressure:
        upress = (press(ee)-press(mi))*ttx/(rho(ee)+rho(mi))

        ! convective terms using vector upwind scheme:
        urd  = umi*txi
        vrd  = vst(nsurf-nsurfl+1,k)*ty
        aurd = abs(urd)
        avrd = abs(vrd)

        if (urd >= zero) then
          j2 = j-1
        else
          j2 = j+1
        endif
        if (vrd >= zero) then
          i4 = i+1
        else
          i4 = i-1
        endif
        if (aurd >= avrd) then
          i1 = i
          j3 = j2
        else
          i1 = i4
          j3 = j
        endif

        ucon = aurd*(u(mm(k,i1,j))-u(mm(k,i1,j2)))                             &
             + avrd*(u(mm(k,i,j3))-u(mm(k,i4,j3)))

        ! Laplacian smoothing term using full-slip condition for lateral terms:
        uaus = u(ee) + u(mm(k,i,j-1)) - two*umi
        if (ne > 0 .and. nn > 0) uaus = uaus + u(nn) - umi
        if (se > 0 .and. ss > 0) uaus = uaus + u(ss) - umi

        ! Horizontal eddy viscosity:
        ehmi = eddyh(mi) + eddyh(ee) + eddyh(ss) + eddyh(se)
        ehnn = eddyh(mi) + eddyh(ee) + eddyh(nn) + eddyh(ne)
        uaus = dtd*uaus                                                        &
             + ( (eddyh(ee)*stretch(ee)-eddyh(mi)*stretch(mi))*txi             &
                +(ehnn     *shear  (nn)-ehmi     *shear  (mi))*fy              &
                +(eddyd(ee)*div    (ee)-eddyd(mi)*div    (mi))*fxh )

        ! RHS of x-mom-eq stored in d:
        d(nsurf-nsurfl+1,k) = umi + fti*vst(nsurf-nsurfl+1,k) - ucon + uaus -  &
                              upress - usrf(nsurf-nsurfl+1)
      enddo

      if ( dyn ) then 
!$acc loop
        do nsurf = nsurfl,nsurfu
          if (khu(nsurf) == 0) cycle
          if (k.gt.khu(nsurf)) cycle
          i = ind(1,nsurf)
          j = ind(2,nsurf)
          mi = mm(k,i  ,j  )
          ee = mm(k,i  ,j+1)
          d(nsurf-nsurfl+1,k) = d(nsurf-nsurfl+1,k)                            &
                              - (geopot(ee)-geopot(mi))*txi
        enddo
      endif

    enddo

!$acc loop
    do nsurf = nsurfl,nsurfu
      if (khu(nsurf) == 0) cycle
      i = ind(1,nsurf)
      j = ind(2,nsurf)
      mie = nsurf
      eee = mm(1,i,j+1)

      ! bottom friction:
      vstar  = vst(nsurf-nsurfl+1,khu(nsurf))
      hrt    = tm(nsurf-nsurfl+1,khu(nsurf)) + hgr
      ustar  = u(mm(khu(nsurf),i,j))
      tmp    = rth/hrt*sqrt(ustar*ustar + vstar*vstar)
      d(nsurf-nsurfl+1,khu(nsurf)) = d(nsurf-nsurfl+1,khu(nsurf)) - tmp*ustar
      b(nsurf-nsurfl+1,khu(nsurf)) = b(nsurf-nsurfl+1,khu(nsurf)) + tmp

      ! wind and possibly ice drag applied at the surface boundary:
      tmp = half*(z0srf(mie) + z0srf(eee))
      if (casus(mie) .and. casus(eee)) then
        ! ice free; include wind drag:
        d(nsurf-nsurfl+1,1) = d(nsurf-nsurfl+1,1) +                            &
                              taux(mie)*dtdr/(tm(nsurf-nsurfl+1,1)+tmp)
      else
        ! ice covered; reduce wind drag and include ice drag:
        eiscol = min(max(ice(2,mie),zero),one)
        eiscor = min(max(ice(2,eee),zero),one)
        eiscox = half*(eiscol+eiscor)
        diffu  = ueis(mie)-u(mie)
        difful = diffu + ueis(mm(1,i,j-1)) - u(mm(1,i,j-1))
        diffur = diffu + ueis(eee)         - u(eee)
        diffvl = veis(mm(1,i-1,j  )) - v(mm(1,i-1,j  )) + veis(mie) - v(mie)
        diffvr = veis(mm(1,i-1,j+1)) - v(mm(1,i-1,j+1)) + veis(eee) - v(eee)
        diffuv = qrt*( eiscol*sqrt(difful**2+diffvl**2)                        &
                      +eiscor*sqrt(diffur**2+diffvr**2))
  
        fac = dtdr/(tm(nsurf-nsurfl+1,1)+tmp)
        dec = half*cdri*diffuv*fac
  
        d(nsurf-nsurfl+1,1) = d(nsurf-nsurfl+1,1)+                             &
                     (cdri*diffuv*diffu+(one-eiscox)*taux(mie))*fac + dec*u(mie)
        b(nsurf-nsurfl+1,1) = b(nsurf-nsurfl+1,1) + dec
      endif
    enddo


!$acc loop
    do nsurf = nsurfl,nsurfu
      if (khu(nsurf) == 0) cycle
      i = ind(1,nsurf)
      j = ind(2,nsurf)
      mie = nsurf
      eee = mm(1,i,j+1)
      ! solve the system defined above (manual inlining of trisolver):
      e    = b(nsurf-nsurfl+1,1)
      s(nsurf-nsurfl+1,1) = d(nsurf-nsurfl+1,1)/e
      do ii = 2, khu(nsurf)
        fv(nsurf-nsurfl+1,ii) = c(nsurf-nsurfl+1,ii-1)/e
        e    = b(nsurf-nsurfl+1,ii) - a(nsurf-nsurfl+1,ii)*fv(nsurf-nsurfl+1,ii)
        s(nsurf-nsurfl+1,ii) = (d(nsurf-nsurfl+1,ii) -                         &
                                a(nsurf-nsurfl+1,ii)*s(nsurf-nsurfl+1,ii-1))/e
      enddo
      !- backward sweep
      do ii = khu(nsurf)-1, 1, -1
        s(nsurf-nsurfl+1,ii) = s(nsurf-nsurfl+1,ii) -                          &
                               fv(nsurf-nsurfl+1,ii+1)*s(nsurf-nsurfl+1,ii+1)
      enddo

      un(mie) = s(nsurf-nsurfl+1,1)
      if (khu(nsurf)  > 1) then
        mmi = mm(2,i,j)
        do ii=2, khu(nsurf)
          un(ii+mmi-2) = s(nsurf-nsurfl+1,ii )
        enddo
      endif

      ! handle a drying situation:
      if (khu(nsurf) <= kdry) then
        call drying_case(mie,eee,un(mie),hz(mie),hz(eee),z)
        if (kdry > 1) then
          mme = mm(2,i,j+1)
          do ii = 2, khu(nsurf)
            call drying_case(mie,eee,un(mmi+ii-2),hz(mmi+ii-2),hz(mme+ii-2),z)
          enddo
        endif
      endif

      ! handle u-dams
      if (nud > 0) then
        dam = uvdam(1,mie)
        un(mie) = un(mie)*dam
        if (khu(nsurf) > 1) then
          un(mmi:mmi+khu(nsurf)-2) = un(mmi:mmi+khu(nsurf)-2)*dam
        endif
      endif
    enddo


!$acc loop
    do nsurf = nsurfl,nsurfu
      if (khv(nsurf) == 0) then
        ! land east of point, maintain zero velocity:
        vn(nsurf) = zero
      else
        tm(nsurf-nsurfl+1,1) = hy(nsurf)
        if (khv(nsurf) > 1) then
          i = ind(1,nsurf)
          j = ind(2,nsurf)
          mmi = mm(2,i,j)
          tm(nsurf-nsurfl+1,2:khv(nsurf)) = hy(mmi:mmi+khv(nsurf)-2)
        endif
        ! set up the vertical eddy viscosity terms, coefficients a
        a(nsurf-nsurfl+1,1) = zero
      endif
    enddo


!$acc loop
    ! set up the vertical eddy viscosity terms, coefficients a, b and c:
    do nsurf = nsurfl,nsurfu
      if (khv(nsurf) .le. 1) cycle
      i = ind(1,nsurf)
      j = ind(2,nsurf)
      mmi = mm(2,i,  j)
      mms = mm(2,i+1,j)
      do k=1,khv(nsurf)-1
        tmp    = one/tm(nsurf-nsurfl+1,k)
        a(nsurf-nsurfl+1,k+1) = -dt*(avv(mmi+k-1)+avv(mms+k-1))/               &
                       (tm(nsurf-nsurfl+1,k)+tm(nsurf-nsurfl+1,k+1))
        a(nsurf-nsurfl+1,k)   = a(nsurf-nsurfl+1,k)  *tmp
        c(nsurf-nsurfl+1,k)   = a(nsurf-nsurfl+1,k+1)*tmp
        b(nsurf-nsurfl+1,k)   = one - (a(nsurf-nsurfl+1,k)+c(nsurf-nsurfl+1,k))
      enddo
    enddo


!$acc loop
    do nsurf = nsurfl,nsurfu
      if (khv(nsurf) == 0) cycle
      i = ind(1,nsurf)
      j = ind(2,nsurf)
      mie = nsurf
      sse = mm(1,i+1,j)
      a(nsurf-nsurfl+1,khv(nsurf)) = a(nsurf-nsurfl+1,khv(nsurf))/             &
                                     tm(nsurf-nsurfl+1,khv(nsurf))
      b(nsurf-nsurfl+1,khv(nsurf)) = one - a(nsurf-nsurfl+1,khv(nsurf))
      c(nsurf-nsurfl+1,khv(nsurf)) = zero
      ! air pressure term:
      vlu = (pl(mie)-pl(sse))*ty*drhow 
      ! zeta:
      zm = z(mie)
      zu = z(sse)
      ! pressure term:
      vzet = gty*(zm-zu)
      ! collect different surface terms:
      vsrf(nsurf-nsurfl+1) = vlu + vzet
    enddo


    do k = 1,kmax 
!$acc loop
      do nsurf = nsurfl,nsurfu
        if (khv(nsurf) == 0) cycle
        if (k.gt.khv(nsurf)) cycle
        i = ind(1,nsurf)
        j = ind(2,nsurf)
        ! do each layer, coefficient d:
        dxi = dx(i)
        txi = tx(i)
        fti = ft(i)
        fxm = half*dt/(dxi+dx(min(i+1,m)))
        tty = two*ty
        ! extract commonly used indices:
        mi = mm(k,   i,j )
        ss = mm(k,i+1,j  )
        ee = mm(k,i  ,j+1)
        se = mm(k,i+1,j+1)
        ww = mm(k,i,  j-1)
        sw = mm(k,i+1,j-1)

        ! center value of v:
        vmi = v(mi)

        ! u centred @ v point:
        ust(nsurf-nsurfl+1,k) = qrt*(u(ww) + u(mi) + u(sw) + u(ss))

        ! baroclinic presssure:
        vpress = (press(mi)-press(ss))*tty/(rho(mi)+rho(ss))

        ! convective terms using vector upwind scheme:
        urd  = ust(nsurf-nsurfl+1,k)*txi
        vrd  = vmi*ty
        aurd = abs(urd)
        avrd = abs(vrd)

        if (vrd >= zero) then
          i2 = i+1
        else
          i2 = i-1
        endif
        if (urd >= zero) then
          j4 = j-1
        else
          j4 = j+1
        endif
        if (avrd >= aurd) then
          i3 = i2
          j1 = j
        else
          i3 = i
          j1 = j4
        endif

        vcon = avrd*(v(mm(k,i,j1))-v(mm(k,i2,j1)))                             &
             + aurd*(v(mm(k,i3,j))-v(mm(k,i3,j4)))

        ! Laplacian smoothing term using full-slip condition for lateral terms:
        vaus = v(mm(k,i-1,j)) + v(ss) - two*vmi
        if (ee > 0 .and. se > 0) vaus = vaus + v(ee) - vmi
        if (ww > 0 .and. sw > 0) vaus = vaus + v(ww) - vmi

        ! Horizontal eddy viscosity:
        ehmi = eddyh(mi) + eddyh(ee) + eddyh(ss) + eddyh(se)
        ehww = eddyh(mi) + eddyh(ww) + eddyh(ss) + eddyh(sw)
        vaus = dtd*vaus                                                        &
             + ( (ehmi     *shear  (mi)-ehww     *shear  (ww))*fxm             &
                -(eddyh(mi)*stretch(mi)-eddyh(ss)*stretch(ss))*ty              &
                +(eddyd(mi)*div    (mi)-eddyd(ss)*div    (ss))*fyh )

        !- RHS of y-mom-eq stored in d:
        d(nsurf-nsurfl+1,k) = vmi - fti*ust(nsurf-nsurfl+1,k) - vcon + vaus -  &
                              vpress - vsrf(nsurf-nsurfl+1)
      enddo

      if ( dyn ) then 
!$acc loop
        do nsurf = nsurfl,nsurfu
          if (khu(nsurf) == 0) cycle
          if (k.gt.khu(nsurf)) cycle
          i = ind(1,nsurf)
          j = ind(2,nsurf)
          mi = mm(k,i  ,j  )
          ss = mm(k,i+1,j  )
          d(nsurf-nsurfl+1,k) = d(nsurf-nsurfl+1,k) - (geopot(mi)-geopot(ss))*ty
        enddo
      endif

    enddo

!$acc loop
    do nsurf = nsurfl,nsurfu
      if (khv(nsurf) == 0) cycle
      i = ind(1,nsurf)
      j = ind(2,nsurf)
      mie = nsurf
      sse = mm(1,i+1,j)

      ! bottom friction:
      ustar  = ust(nsurf-nsurfl+1,khv(nsurf))
      hrt    = tm(nsurf-nsurfl+1,khv(nsurf))+hgr
      vstar  = v(mm(khv(nsurf),i,j))
      tmp    = rth/hrt*sqrt(vstar*vstar + ustar*ustar) 
      d(nsurf-nsurfl+1,khv(nsurf)) = d(nsurf-nsurfl+1,khv(nsurf)) - tmp*vstar
      b(nsurf-nsurfl+1,khv(nsurf)) = b(nsurf-nsurfl+1,khv(nsurf)) + tmp

      ! wind and possibly ice drag applied at the surface boundary:
      tmp = half*(z0srf(mie) + z0srf(sse))
      if (casus(mie) .and. casus(sse)) then
        ! ice free; include wind drag:
        d(nsurf-nsurfl+1,1) = d(nsurf-nsurfl+1,1) +                            &
                              tauy(mie)*dtdr/(tm(nsurf-nsurfl+1,1)+tmp)
      else
        ! ice covered; reduce wind drag and include ice drag:
        eiscoo = min(max(ice(2,mie),zero),one)
        eiscou = min(max(ice(2,sse),zero),one)
        eiscoy = half*(eiscoo+eiscou)
        diffuo = ueis(mm(1,i,  j-1)) - u(mm(1,i,  j-1)) + ueis(mie) - u(mie)
        diffuu = ueis(mm(1,i+1,j-1)) - u(mm(1,i+1,j-1)) + ueis(sse) - u(sse)
        diffv  = veis(mie)-v(mie)
        diffvo = diffv + veis(mm(1,i-1,j)) - v(mm(1,i-1,j))
        diffvu = diffv + veis(sse)         - v(sse)
        diffuv = qrt*( eiscoo*sqrt(diffuo**2+diffvo**2)                        &
                      +eiscou*sqrt(diffuu**2+diffvu**2))

        fac = dtdr/(tm(nsurf-nsurfl+1,1)+tmp)
        dec = half*cdri*diffuv*fac

        d(nsurf-nsurfl+1,1) = d(nsurf-nsurfl+1,1)+                             &
                              (cdri*diffuv*diffv+(one-eiscoy)*tauy(mie))*fac + &
                              dec*v(mie)
        b(nsurf-nsurfl+1,1) = b(nsurf-nsurfl+1,1) + dec
      endif
    enddo


!$acc loop
    do nsurf = nsurfl,nsurfu
      if (khv(nsurf) == 0) cycle
      i = ind(1,nsurf)
      j = ind(2,nsurf)
      mie = nsurf
      sse = mm(1,i+1,j)
      ! solve the system defined above (manual inlining of trisolver):
      e = b(nsurf-nsurfl+1,1)
      s(nsurf-nsurfl+1,1) = d(nsurf-nsurfl+1,1)/e
      do ii = 2, khv(nsurf)
        fv(nsurf-nsurfl+1,ii) = c(nsurf-nsurfl+1,ii-1)/e
        e    = b(nsurf-nsurfl+1,ii) - a(nsurf-nsurfl+1,ii)*fv(nsurf-nsurfl+1,ii)
        s(nsurf-nsurfl+1,ii) = (d(nsurf-nsurfl+1,ii) -                         &
                                a(nsurf-nsurfl+1,ii)*s(nsurf-nsurfl+1,ii-1))/e
      enddo
      !- backward sweep
      do ii = khv(nsurf)-1, 1, -1
        s(nsurf-nsurfl+1,ii) = s(nsurf-nsurfl+1,ii) -                          &
                               fv(nsurf-nsurfl+1,ii+1)*s(nsurf-nsurfl+1,ii+1)
      enddo

      vn(mie) = s(nsurf-nsurfl+1,1)
      if (khv(nsurf) > 1) then
        mmi = mm(2,i,j)
        do ii=2, khv(nsurf)
          vn(ii+mmi-2) = s(nsurf-nsurfl+1,ii )
        enddo
      endif

      ! handle a drying situation:
      if (khv(nsurf) <= kdry) then
        call drying_case(sse,mie,vn(mie),hz(sse),hz(mie),z)
        if (kdry > 1) then
          mms = mm(2,i+1,j)
          do ii = 2, khv(nsurf)
            call drying_case(sse,mie,vn(mmi+ii-2),hz(mms+ii-2),hz(mmi+ii-2),z)
          enddo
        endif
      endif

      ! handle v-dams
      if (nvd > 0) then
        dam = uvdam(2,mie)
        vn(mie) = vn(mie)*dam
        if (khv(nsurf) > 1) then
          vn(mmi:mmi+khv(nsurf)-2) = vn(mmi:mmi+khv(nsurf)-2)*dam
        endif
      endif
    enddo

  enddo
!$acc end parallel loop

!- END GPU ---------------------------------------------------------------------

  if (weirs) then
    call weir (nweir, widx, mm, ty, tx, z, zweir, un, vn)  
  endif

end subroutine momeqs_default_openacc 

  subroutine TriSolv(s, a, b, c, d, n)
  !-----------------------------------------------------------------------------
  !  standard tridiagonal solver
  !-----------------------------------------------------------------------------

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: n
    real(8), intent(out)   :: s(1:)
    real(8), intent(in)    :: a(1:), b(1:), c(1:), d(1:)

    !-local vars ---------------------------------------------------------------
    integer(4) :: i
    real(8)    :: e
    real(8)    :: f(n)

    !- forward sweep -----------------------------------------------------------
    e    = b(1)
    s(1) = d(1)/e
    do i = 2, n
      f(i) = c(i-1)/e
      e    = b(i) - a(i)*f(i)
      s(i) = (d(i) - a(i)*s(i-1))/e
    enddo

    !- backward sweep ----------------------------------------------------------
    do i = n-1, 1, -1
      s(i) = s(i) - f(i+1)*s(i+1)
    enddo

    !- That's it! --------------------------------------------------------------

  end subroutine TriSolv

end module momeqs_wrapping

! ------------------------------------------------------------------------------
