module output
  
  implicit none
  private
  
  public :: output_init, output_close, output_write
  
  
  contains
  
    subroutine output_init(aufdat,ctim,tim)
      ! Output-files will be opend in this file
    
      use cmod_miscvars,      only : itime, tidal
      use constants,          only : iptracers, calctrans
      use dmi_mpi,            only : mpi_io_serial, ice_serial
      use transport_variables,only : transdat_new
      use wlstat_wrap,        only : wlstat 
      use writetempdat,       only : tempdat_new
      use writegrib,          only : grb_init

      implicit none
    
      character(11), intent(in)           :: aufdat
      character(19), intent(in)           :: ctim  !<==  watch out here!
      real(8),       intent(in)           :: tim
 
      if (mpi_io_serial) then
        !- open binary HD output file w/start time as header -------------------
        call tempdat_new () 

        !. initialize grib files
        call grb_init ()

        !- open ascii sea level time series files, read station list file ------
        call wlstat(0,ctim,tim,itime)

        !- open binary file for inline transport output ------------------------
        if (calctrans) call transdat_new ()
      endif

    end subroutine output_init
  
  !============================================================================!
  
  subroutine output_write(izeits,aufdat,ctim,tim)
    ! Output-files will be closed in this file
    
    use cmod_arrays,        only : u, v, h, z, wu, wv, cmp, tke,               &
                                   diss, avv, dispv, wrk
    use cmod_local_arrays,  only : u_l, v_l, h_l, wi_l, z_l,                   &
                                   wu_l, wv_l, cmp_l, tke_l, diss_l, avv_l,    &
                                   dispv_l
    use cmod_miscvars,      only : izeite, itime, lmn1, lmn2, lmne, tidal
    use cmod_params,        only : narea, kmx
!   use compress,           only : compress_i4
    use constants,          only : iptracers, printturb, dyn, calctrans
    use dmi_mpi,            only : mpi_io_serial, dmpi_gather
    use dmi_mpi_global,     only : mpi_io_rank
    use transports,         only : transport_output
    use transport_variables,only : NDTTRANSOUT, transdat_new
    use turbout_wrap,       only : turbout
    use wlstat_wrap,        only : wlstat 
    use writetempdat,       only : tempdat_next, write_icefile, tempdat_write
    use writegrib,          only : grb_write, grb_close
 
    implicit none
    
    integer(4), intent(in)       :: izeits
    character(11), intent(in)    :: aufdat
    character(19), intent(in)    :: ctim  !<==  watch out here!
    real(8),       intent(in)    :: tim
    
    integer(4)                   :: ia       


    !- flow field varibles to tempdat file, etc.:
    if (mod(izeits,lmn2) == 0) then

      !- Gather to I/O task:
      do ia=1,narea
        call dmpi_gather(ia,kmx(ia),  u_l(ia)%p,  u(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,kmx(ia),  v_l(ia)%p,  v(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,kmx(ia), wi_l(ia)%p,wrk(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,(1),      z_l(ia)%p,  z(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,(1),     wu_l(ia)%p, wu(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,(1),     wv_l(ia)%p, wv(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,kmx(ia),cmp_l(ia)%p,cmp(ia)%p,2,mpi_io_rank)
      enddo
      if (dyn) then
        do ia=1,narea
          call dmpi_gather(ia,kmx(ia),h_l(ia)%p,h(ia)%p,mpi_io_rank)
        enddo
      endif          

! MPI   The following global arrays must be valid on the I/O task and need
! MPI   not be gathered:
! MPI          eisdi,eisco,snowd,tsnei,ueis,veis,casus

      if (mpi_io_serial) then
        !- first compress to i4:
!        call compress_i4 (izeits,lmn1)

        !- output to tempdat file:
!        call tempdat_write (ctim,tim,aufdat,izeits,lmn2) ! FIXME
        !- output to grib files:                  
        call grb_write (ctim,tim,izeits)
! FIXME testing
        call grb_close()
        stop
      endif

      !- output turbulence profiles:
      if (printturb) then 
        do ia=1,narea
          call dmpi_gather(ia,kmx(ia),  tke_l(ia)%p,  tke(ia)%p,mpi_io_rank)
          call dmpi_gather(ia,kmx(ia), diss_l(ia)%p, diss(ia)%p,mpi_io_rank)
          call dmpi_gather(ia,kmx(ia),  avv_l(ia)%p,  avv(ia)%p,mpi_io_rank)
          call dmpi_gather(ia,kmx(ia),dispv_l(ia)%p,dispv(ia)%p,2,mpi_io_rank)
        enddo
        if (mpi_io_serial)  call turbout(ctim)
      endif

      if (mpi_io_serial) then
        !- ice data to file:
        if (.not.tidal) then 
          if (mod(izeits,lmn1) == 0) then    ! hourly
            call write_icefile (ctim,tim) ! FIXME
          endif
        endif
 
     endif
    endif


    !- ascii sea level time series files:
    if (mod(izeits,lmne) == 0) then
      if (mod(izeits,lmn2) /= 0) then
        do ia=1,narea
          call dmpi_gather(ia,(1),z_l(ia)%p,z(ia)%p,mpi_io_rank)
        enddo
      endif
      if (mpi_io_serial) call wlstat(1,ctim,tim,itime)
    endif

    if (calctrans) then
      if (mod(izeits, NDTTRANSOUT) == 0) then
        call transport_output( narea, ctim )
      endif
    endif

    if (izeits == izeite) then
      !- tempdat files in numbered 6 or 12 hour chunks:
      !    With the option archive=.true. the first chunk goes to the 
      !    'archive' file, and subsequent chunks will go to 'tempdat.00', 
      !    'tempdat.01', ... .   
      !    With archive=.false., the first file is 'tempdat.00', and subsequent
      !    files are 'tempdat.01', 'tempdat.02', ... .
      !    If spinup=.true. there is new, updated header 'aufdat' in each file.
      !    If spinup=.false. there is one 'aufdat'-header in the first file
      !    corresponding to the start time, and the second and all subsequent
      !    files will have the 'aufdat'-header corresponding to the start of
      !    second 6 or 12 hr chunk.
      !    
      if (mpi_io_serial) then
        call tempdat_next () ! FIXME
        if (calctrans) call transdat_new () 
      endif
    endif
      
  end subroutine output_write

  !============================================================================!
  
  subroutine output_close(ctim,tim)
      ! Output-files will be closed in this file
    use constants,          only : iptracers, calctrans
    use dmi_mpi,            only : mpi_io_serial, ice_serial
    use transport_variables,only : transdat_close
    use wlstat_wrap,        only : wlstat 
    use writetempdat,       only : tempdat_close
    use writegrib,          only : grb_close
    implicit none
    character(19), intent(in)           :: ctim  !<==  watch out here!
    real(8),       intent(in)           :: tim

    if (mpi_io_serial) then
      call grb_close()
      call tempdat_close()
      call wlstat(2,ctim,tim)
      if (calctrans) call transdat_close()
    endif
  end subroutine output_close

end module output

