module perm

  use exits,    only : exitme

  implicit none
  private

  real(8), allocatable, save, public :: u_o(:),v_o(:),h_o(:),w_o(:)
  real(8), allocatable, save, public :: s_o(:),t_o(:)
  real(8), allocatable, save, public :: turb1_o(:),turb2_o(:),cmp_o(:,:)
  real(8), allocatable, save, public :: z_o(:),tsoil_o(:,:)
  real(8), allocatable, save, public :: ice_o(:,:),ui_o(:),vi_o(:),ts_o(:)
  logical, allocatable, save, public :: casus_o(:)

  public :: de_permute1, de_permute2, de_permute_2Dvars, de_permute_omp

  interface de_permute_omp
    module procedure de_permute1_omp
    module procedure de_permute_cmp_omp
  end interface

  interface de_permute_2Dvars
    module procedure de_perm_z
    module procedure de_perm_ice
    module procedure de_perm_tsoil
    module procedure de_perm_iota
  end interface

  private :: de_perm_z, de_perm_ice, de_perm_tsoil, de_perm_iota
  private :: de_permute1_omp, de_permute_cmp_omp

contains

  !=============================================================================

  subroutine de_permute1 ( mi,ni,ki,mm, a,a_o,b,b_o, c,d,c_o,d_o,w,w_o)
    !---------------------------------------------------------------------------
    ! This routine de-permutes before output.
    ! Takes 1 (a), 2 (a and b) or 5 (a, b, c, d and w) arrays.
    ! Only data arrays need being de-permuted to temporary arrays before 
    ! output, index arrays and original data are kept unchanged.
    !---------------------------------------------------------------------------

    !- modules -----------------------------------------------------------------
    ! none

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)        :: mi               ! 
    integer(4), intent(in)        :: ni               ! 
    integer(4), intent(in)        :: ki               ! number of layers
    integer(4), intent(in)        :: mm(0:,0:,:)      ! (0:mi+1,0:ni+1,ki) index
    real(8), intent(in)           :: a(0:)            !
    real(8), intent(in), optional :: b(0:)            !
    real(8), intent(in), optional :: c(0:)            !
    real(8), intent(in), optional :: d(0:)            !
    real(8), intent(in), optional :: w(0:)     
    real(8), intent(out)          :: a_o(0:)     
    real(8), intent(out),optional :: b_o(0:)     
    real(8), intent(out),optional :: c_o(0:)     
    real(8), intent(out),optional :: d_o(0:)     
    real(8), intent(out),optional :: w_o(0:)     


    !- local vars --------------------------------------------------------------
    integer(4) :: i,j,k,mc,depermcase,nc

    ! --------------------------------------------------------------------------

    ! find which the relevant case or bail out on traps:
    depermcase = 1
    if (present(b)) then
      if (.not.(present(b_o))) then
        call exitme('Error in de_permute1: Missing output array')
      endif
      depermcase = 2
      if (present(c)) then
        if (.not.(present(d) .and. present(w))) then
          call exitme('Error in de_permute1: Missing input array')
        endif
        if (.not.(present(c_o) .and. present(d_o) .and. present(w_o))) then
          call exitme('Error in de_permute1: Missing output arrays')
        endif
        depermcase = 3
      endif
    endif

    ! now, let's permute data arrays:
    if (depermcase == 1) then
      a_o(0) = a(0)
      mc = 0
      do k=1,ki
        do j=1,ni
          do i=1,mi
            nc = mm(i,j,k)
            if (nc == 0) cycle
            mc = mc + 1
            a_o(mc) = a(nc)
          enddo
        enddo
      enddo
    elseif (depermcase == 2) then
      a_o(0) = a(0)
      b_o(0) = b(0)
      mc = 0
      do k=1,ki
        do j=1,ni
          do i=1,mi
            nc = mm(i,j,k)
            if (nc == 0) cycle
            mc = mc + 1
            a_o(mc) = a(nc)
            b_o(mc) = b(nc)
          enddo
        enddo
      enddo
    elseif (depermcase == 3) then
      a_o(0) = a(0)
      b_o(0) = b(0)
      c_o(0) = c(0)
      d_o(0) = d(0)
      w_o(0) = w(0)
      mc = 0
      do k=1,ki
        do j=1,ni
          do i=1,mi
            nc = mm(i,j,k)
            if (nc == 0) cycle
            mc = mc + 1
            a_o(mc) = a(nc)
            b_o(mc) = b(nc)
            c_o(mc) = c(nc)
            d_o(mc) = d(nc)
            w_o(mc) = w(nc)
          enddo
        enddo
      enddo
    endif

    ! caught in a trap?
    if (mc /= maxval(mm)) then
      call exitme('Error in de_permute1: Unable to reconstruct No of wetpoints')
    endif

    !-  That's it!

  end subroutine de_permute1

  !=============================================================================

  subroutine de_permute2 ( mi,ni,ki,mm,turb1,turb2)
    !---------------------------------------------------------------------------
    ! This routine de-permutes turbulence vars before output to restart file.
    ! Due to different requirements with different turbulence schemes, we may 
    ! need permuting one, optionally two parameters.
    ! Only data arrays need being de-permuted to temporary arrrays before
    ! output, index arrays and original data are kept unchanged.
    !---------------------------------------------------------------------------

    !- modules -----------------------------------------------------------------
    ! none

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: mi               ! 
    integer(4), intent(in)    :: ni               ! 
    integer(4), intent(in)    :: ki               ! number of layers
    integer(4), intent(in)    :: mm(0:,0:,:)      ! (0:mi+1,0:ni+1,ki) index 
    real(8), intent(in)       :: turb1(0:)     
    real(8), intent(in), optional :: turb2(0:)     

    !- local vars --------------------------------------------------------------
    integer(4) :: i,j,k,mc,nc

    ! --------------------------------------------------------------------------

    ! now, let's permute data arrays:
    if (.not.allocated(turb1_o)) then
      call exitme('Error in de_permute2: turb1_o not allocated')
    endif
    turb1_o(0) = turb1(0)
    mc = 0
    do k=1,ki
      do j=1,ni
        do i=1,mi
          nc = mm(i,j,k)
          if (nc == 0) cycle
          mc = mc + 1
          turb1_o(mc) = turb1(nc)
        enddo
      enddo
    enddo
    if (present(turb2)) then
      ! yet another trap?
      if (.not.allocated(turb2_o)) then
        call exitme('Error in de_permute2: turb2_o not allocated')
      endif
      turb2_o(0) = turb2(0)
      mc = 0
      do k=1,ki
        do j=1,ni
          do i=1,mi
            nc = mm(i,j,k)
            if (nc == 0) cycle
            mc = mc + 1
            turb2_o(mc) = turb2(nc)
          enddo
        enddo
      enddo
    endif

    ! caught in a trap?
    if (mc /= maxval(mm)) then
      call exitme('Error in de_permute1: Unable to reconstruct No of wetpoints')
    endif

    !-  That's it!

  end subroutine de_permute2

  !=============================================================================

  subroutine de_perm_z( mi, ni, mm, z )

    !- modules -----------------------------------------------------------------
    ! none

    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: mi, ni, mm(0:,0:,:)
    real(8),    intent(in) :: z(0:)

    !- local vars --------------------------------------------------------------
    integer(4) :: mo, mn, i, j

    ! now, let's permute data array to old permutation:
    z_o(0) = z(0)
    mo = 0
    do j=1,ni
      do i=1,mi
        mn = mm(i,j,1)
        if (mn == 0) cycle
        mo = mo + 1
        z_o(mo) = z(mn)
      enddo
    enddo

  end subroutine de_perm_z

  !=============================================================================

  subroutine de_perm_ice( mi, ni, mm, ice, ts, ui, vi, casus )

    !- modules -----------------------------------------------------------------
    ! none

    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)           :: mi, ni, mm(0:,0:,:)
    real(8),    intent(in)           :: ice(:,0:), ts(0:), ui(0:), vi(0:)
    logical,    intent(in), optional :: casus(0:)

    !- local vars --------------------------------------------------------------
    integer(4) :: mo, mn, i, j

    ! now, let's permute data array to old permutation:
    ice_o(:,0) = ice(:,0)
    ts_o(0)    = ts(0)
    ui_o(0)    = ui(0)
    vi_o(0)    = vi(0)
    mo = 0
    do j=1,ni
      do i=1,mi
        mn = mm(i,j,1)
        if (mn == 0) cycle
        mo = mo + 1
        ice_o(:,mo) = ice(:,mn)
        ts_o(mo)    = ts(mn)
        ui_o(mo)    = ui(mn)
        vi_o(mo)    = vi(mn)
      enddo
    enddo
    if (present(casus)) then
      casus_o(0) = casus(0)
      mo = 0
      do j=1,ni
        do i=1,mi
          mn = mm(i,j,1)
          if (mn == 0) cycle
          mo = mo + 1
          casus_o(mo) = casus(mn)
        enddo
      enddo
    endif

  end subroutine de_perm_ice

  !=============================================================================

  subroutine de_perm_tsoil( mi, ni, mm, tsoil )

    !- modules -----------------------------------------------------------------
    ! none

    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: mi, ni, mm(0:,0:,:)
    real(8),    intent(in) :: tsoil(:,:)

    !- local vars --------------------------------------------------------------
    integer(4) :: mo, mn, i, j

    ! now, let's permute data array to old permutation:
    mo = 0
    do j=1,ni
      do i=1,mi
        mn = mm(i,j,1)
        if (mn == 0) cycle
        mo = mo + 1
        tsoil_o(:,mo) = tsoil(:,mn)
      enddo
    enddo

  end subroutine de_perm_tsoil

!=============================================================================

  subroutine de_perm_iota( mi, ni, mm, iotamodel, iota )

    !- modules -----------------------------------------------------------------
    ! none

    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)           :: mi, ni, mm(0:,0:,:)
    real(8),    intent(in)           :: iota(:,0:)
    logical,    intent(in)           :: iotamodel

    !- local vars --------------------------------------------------------------
    integer(4) :: mo, mn, i, j

    if (iotamodel) then
      ice_o(:,0) = iota(:,0)
      mo = 0
      do j=1,ni
        do i=1,mi
          mn = mm(i,j,1)
          if (mn == 0) cycle
          mo = mo + 1
          ice_o(:,mo) = iota(:,mn)
        enddo
      enddo
    end if

  end subroutine de_perm_iota

  !=============================================================================

  subroutine de_permute1_omp( iwet2, mmk, mop, ind, kh, ompidx,                &
                           a, a_o, b, b_o, c, d, c_o, d_o, w, w_o)
    !---------------------------------------------------------------------------
    ! This routine de-permutes before output.
    ! Takes 1 (a), 2 (a and b) or 5 (a, b, c, d and w) arrays.
    ! Only data arrays need being de-permuted to temporary arrays before 
    ! output, index arrays and original data are kept unchanged.
    !---------------------------------------------------------------------------

    !- modules -----------------------------------------------------------------
    use dmi_omp,          only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)        :: mmk(:,0:,0:)     ! (ki,0:mi+1,0:ni+1) index
    integer(4), intent(in)        :: mop(0:,0:,:)     ! old permutation
    integer(4), intent(in)        :: ind(:,:)         !
    integer(4), intent(in)        :: iwet2
    integer(4), intent(in)        :: kh(0:)
    integer(4), intent(inout)     :: ompidx(:,:)

    real(8), intent(in)             :: a(0:)            !
    real(8), intent(in),   optional :: b(0:)            !
    real(8), intent(in),   optional :: c(0:)            !
    real(8), intent(in),   optional :: d(0:)            !
    real(8), intent(in),   optional :: w(0:)     
    real(8), intent(inout)          :: a_o(0:)     
    real(8), intent(inout),optional :: b_o(0:)     
    real(8), intent(inout),optional :: c_o(0:)     
    real(8), intent(inout),optional :: d_o(0:)     
    real(8), intent(inout),optional :: w_o(0:)     


    !- local vars --------------------------------------------------------------
    integer(4) :: i,j,k,depermcase,ns,nl,nu,nc,mc

    ! --------------------------------------------------------------------------

    ! find which the relevant case or bail out on traps:
    depermcase = 1
    if (present(b)) then
      if (.not.(present(b_o))) then
        call exitme('Error in de_permute1_omp: Missing output array')
      endif
      depermcase = 2
      if (present(c)) then
        if (.not.(present(d) .and. present(w))) then
          call exitme('Error in de_permute1_omp: Missing input array')
        endif
        if (.not.(present(c_o) .and. present(d_o) .and. present(w_o))) then
          call exitme('Error in de_permute1_omp: Missing output arrays')
        endif
        depermcase = 3
      endif
    endif

    ! get domain decomposition:
    call domp_get_domain(kh, 1, iwet2, nl, nu, ompidx)

    ! now, let's permute data arrays:
    if (depermcase == 1) then
!$OMP MASTER
      a_o(0) = a(0)
!$OMP END MASTER
      do ns=nl,nu
        i = ind(1,ns)
        j = ind(2,ns)
        a_o(mop(i,j,1)) = a(ns)
        do k=2,kh(ns)
          a_o(mop(i,j,k)) = a(mmk(k,i,j))
        enddo
      enddo
    elseif (depermcase == 2) then
!$OMP MASTER
      a_o(0) = a(0)
      b_o(0) = b(0)
!$OMP END MASTER
      do ns=nl,nu
        i  = ind(1,ns)
        j  = ind(2,ns)
        a_o(mop(i,j,1)) = a(ns)
        b_o(mop(i,j,1)) = b(ns)
        do k=2,kh(ns)
          mc = mop(i,j,k)
          nc = mmk(k,i,j)
          a_o(mc) = a(nc)
          b_o(mc) = b(nc)
        enddo
      enddo
    elseif (depermcase == 3) then
!$OMP MASTER
      a_o(0) = a(0)
      b_o(0) = b(0)
      c_o(0) = c(0)
      d_o(0) = d(0)
      w_o(0) = w(0)
!$OMP END MASTER
      do ns=nl,nu
        i  = ind(1,ns)
        j  = ind(2,ns)
        a_o(mop(i,j,1)) = a(ns)
        b_o(mop(i,j,1)) = b(ns)
        c_o(mop(i,j,1)) = c(ns)
        d_o(mop(i,j,1)) = d(ns)
        w_o(mop(i,j,1)) = w(ns)
        do k=2,kh(ns)
          mc = mop(i,j,k)
          nc = mmk(k,i,j)
          a_o(mc) = a(nc)
          b_o(mc) = b(nc)
          c_o(mc) = c(nc)
          d_o(mc) = d(nc)
          w_o(mc) = w(nc)
        enddo
      enddo
    endif

    !-  That's it!

  end subroutine de_permute1_omp

  !=============================================================================

  subroutine de_permute_cmp_omp( iwet2, mmk, mop, ind, kh, ompidx,             &
                                 cmp, c1_o, c2_o, a, a_o, b, b_o, c, c_o)
    !---------------------------------------------------------------------------
    ! This routine de-permutes before output.
    ! Takes 1 two-component array (cmp), or 
    !       1 two-component array (cmp) and 3 single component arrays (a,b,c).
    ! Only data arrays need being de-permuted to temporary arrays before 
    ! output, index arrays and original data are kept unchanged.
    !---------------------------------------------------------------------------

    !- modules -----------------------------------------------------------------
    use dmi_omp,          only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)        :: mmk(:,0:,0:)     ! (ki,0:mi+1,0:ni+1) index
    integer(4), intent(in)        :: mop(0:,0:,:)     ! old permutation
    integer(4), intent(in)        :: ind(:,:)         !
    integer(4), intent(in)        :: iwet2
    integer(4), intent(in)        :: kh(0:)
    integer(4), intent(inout)     :: ompidx(:,:)

    real(8), intent(in)             :: cmp(1:,0:)
    real(8), intent(in),   optional :: a(0:)
    real(8), intent(in),   optional :: b(0:)
    real(8), intent(in),   optional :: c(0:)
    real(8), intent(inout)          :: c1_o(0:), c2_o(0:)
    real(8), intent(inout),optional :: a_o(0:)     
    real(8), intent(inout),optional :: b_o(0:)     
    real(8), intent(inout),optional :: c_o(0:)     


    !- local vars --------------------------------------------------------------
    integer(4) :: i,j,k,depermcase,ns,nl,nu,nc,mc

    ! --------------------------------------------------------------------------

    ! find which the relevant case or bail out on traps:
    depermcase = 1
    if (present(a)) then
      if (.not.(present(b) .and. present(c))) then
        call exitme('Error in de_permute_cmp_omp: Missing output array')
      endif
      if (.not.(present(a_o) .and. present(b_o) .and. present(c_o))) then
        call exitme('Error in de_permute_cmp_omp: Missing output array')
      endif
      depermcase = 2
    endif

    ! get domain decomposition:
    call domp_get_domain(kh, 1, iwet2, nl, nu, ompidx)

    ! now, let's permute data arrays:
    if (depermcase == 1) then
!$OMP MASTER
      c1_o(0) = cmp(1,0)
      c2_o(0) = cmp(2,0)
!$OMP END MASTER
      do ns=nl,nu
        i = ind(1,ns)
        j = ind(2,ns)
        c1_o(mop(i,j,1)) = cmp(1,ns)
        c2_o(mop(i,j,1)) = cmp(2,ns)
        do k=2,kh(ns)
          mc = mop(i,j,k)
          nc = mmk(k,i,j)
          c1_o(mc) = cmp(1,nc)
          c2_o(mc) = cmp(2,nc)
        enddo
      enddo
    elseif (depermcase == 2) then
!$OMP MASTER
      c1_o(0) = cmp(1,0)
      c2_o(0) = cmp(2,0)
      a_o(0) = a(0)
      b_o(0) = b(0)
      c_o(0) = c(0)
!$OMP END MASTER
      do ns=nl,nu
        i  = ind(1,ns)
        j  = ind(2,ns)
        c1_o(mop(i,j,1)) = cmp(1,ns)
        c2_o(mop(i,j,1)) = cmp(2,ns)
        a_o(mop(i,j,1)) = a(ns)
        b_o(mop(i,j,1)) = b(ns)
        c_o(mop(i,j,1)) = c(ns)
        do k=2,kh(ns)
          mc = mop(i,j,k)
          nc = mmk(k,i,j)
          c1_o(mc) = cmp(1,nc)
          c2_o(mc) = cmp(2,nc)
          a_o(mc) = a(nc)
          b_o(mc) = b(nc)
          c_o(mc) = c(nc)
        enddo
      enddo
    endif

    !-  That's it!

  end subroutine de_permute_cmp_omp

  !=============================================================================

end module perm

