module smagorinsky
  use constants, only : zero, one, qrt, half
  !-----------------------------------------------------------------------------
  !  This module wraps some of the routines for doing the Smagorinsky
  !  parameterization and horizontal defromation terms
  !-----------------------------------------------------------------------------

  implicit none
  private

  public  :: smag, deform, calc_uvterms

contains

  !=============================================================================

  subroutine smag (imax,kmax,mm,ind,n2d,dx,dy,dt,cs,ah,kh,                     &
                   shear,stretch,cosphi,eddyh,cdiv,idx,div)
    !---------------------------------------------------------------------------
    !  Calculate horizontal eddy viscosity from stretch and shear terms.
    !---------------------------------------------------------------------------
  
    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain
  
    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: imax, kmax, n2d
    integer(4), intent(in)    :: mm(:,0:,0:), kh(0:), ind(:,:)
    integer(4), intent(inout) :: idx(1:,1:)
    real(8),    intent(in)    :: dy, dx, dt, cs, ah, cdiv
    real(8),    intent(in)    :: cosphi(:,0:), shear(0:)
    real(8),    intent(inout) :: stretch(0:), div(0:), eddyh(0:)
  
    !- local variables ---------------------------------------------------------
    integer(4) :: n, i, j, k, kb, mi, mi2, mn2, mw2, nw2, n2dl, n2du
    real(8)    :: st, l2s, cs2dxdy, ed, cdl, edmax, mindx, dv, sh
    real(8)    :: shtmp(1:kmax), dvtmp(1:kmax)

    !- Parameter to control deep divergence:
    real(8), parameter :: CDFAC = 100.0_8

    !- Parameter to control numerical stability (linear, in von Neumann sense):
    real(8), parameter :: MXFAC = 0.10_8
  
    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)
  
    cs2dxdy = cs*cs*ah*dx*dy
    mindx   = dx*minval(cosphi(1,1:imax-1))
    edmax   = (MXFAC/dt)/(one/mindx**2 + one/dy**2)
  

    ! k=1
    do n=n2dl,n2du
      i = ind(1,n)
      j = ind(2,n)
  
      l2s = cs2dxdy*cosphi(1,i)
      cdl = cdiv*l2s

      ! k=1 unrolled
      sh = qrt*( shear(n)           + shear(mm(1,i-1,j))                       &
               + shear(mm(1,i,j-1)) + shear(mm(1,i-1,j-1)) )
      
      st = stretch(n)
      ed = st*st + sh*sh
      eddyh(n)   = min( l2s*sqrt( ed )        , edmax )
      stretch(n) = st*eddyh(n) 
      dv         = div(n)
      div(n)     = min( cdl*sqrt( ed + dv*dv ), edmax )*dv
    enddo

    ! k=2..kb
    do n=n2dl,n2du
      kb = kh(n)
 
      if (kb <= 1) cycle
      
      i = ind(1,n)
      j = ind(2,n)
  
      l2s = cs2dxdy*cosphi(1,i)
      cdl = cdiv*l2s

      mi2 = mm(2,i,  j  ) - 2
      mn2 = mm(2,i-1,j  ) - 2
      mw2 = mm(2,i,  j-1) - 2
      nw2 = mm(2,i-1,j-1) - 2

      !  extract shear terms:
      shtmp(2:kb) = shear(mi2+2:mi2+kb)
      do k=2,min(kb,kh(mm(1,i,j-1)))
        shtmp(k) = shtmp(k) + shear(mw2+k)
      enddo
      do k=2,min(kb,kh(mm(1,i-1,j)))
        shtmp(k) = shtmp(k) + shear(mn2+k)
      enddo
      do k=2,min(kb,kh(mm(1,i-1,j-1)))
        shtmp(k) = shtmp(k) + shear(nw2+k)
      enddo

      ! calc the eddy coefficients:
      do k=2,kb
        mi = mi2 + k
        sh = qrt*shtmp(k)
        st = stretch(mi)
        ed = st*st + sh*sh
        dv = div(mi)
        eddyh(mi) = min( l2s*sqrt( ed )        , edmax )
        dvtmp(k)  = min( cdl*sqrt( ed + dv*dv ), edmax )
      enddo
  
      ! extra stabilizer in deepest part:
      do k=max(kmax-1,3),min(kmax,kb)
        dvtmp(k) = min( CDFAC*dvtmp(k), edmax )
      enddo

      do k=2,kb
        mi = mi2 + k
        stretch(mi) = stretch(mi)*eddyh(mi)
        div(mi)     = div(mi)*dvtmp(k)
      enddo

    enddo
   
  end subroutine smag
  
  !=============================================================================

  subroutine deform (mm,ind,n2d,ddx,ddy,kh,u,v,shear,stretch,div,cosphi,idx)

    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain
  
    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: n2d
    integer(4), intent(in)    :: mm(:,0:,0:), kh(0:), ind(:,:)
    integer(4), intent(inout) :: idx(1:,1:)
    real(8),    intent(in)    :: u(0:), v(0:), ddy, ddx, cosphi(:,0:)
    real(8),    intent(inout) :: shear(0:), stretch(0:), div(0:)
  
    !- local variables ---------------------------------------------------------
    integer(4) :: n,i,j,k,kb,mi,mw,mn,me,ms,mse
    integer(4) :: n2dl, n2du, mi0, mn0, mw0, kshear, me0, ms0
    real(8)    :: ddxi1, ddxi2, uterm, vterm
  
    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)
  
    !-   stretch = du/dx - dv/dy   ---------------------------------------------
    !-   div     = du/dx + dv/dy   ---------------------------------------------
    !-   shear   = du/dy + dv/dx   ---------------------------------------------

    ! k=1 unrolled:
    do n=n2dl,n2du
      i = ind(1,n)
      j = ind(2,n)
  
      ddxi1 = ddx/cosphi(1,i)
      ddxi2 = ddx/cosphi(2,i)
      
      mi  = n
      mw  = mm(1,i,  j-1)
      mn  = mm(1,i-1,j  )
      me  = mm(1,i,  j+1)
      ms  = mm(1,i+1,j  )
      mse = mm(1,i+1,j+1)
  
      uterm = (u(mi)-u(mw))*ddxi1
      vterm = (v(mn)-v(mi))*ddy
  
      stretch(mi) = uterm - vterm
      div    (mi) = uterm + vterm
  
      if (me>0 .and. ms>0 .and. mse>0) then
        shear(mi) = (v(me)-v(mi))*ddxi2 + (u(mi)-u(ms))*ddy
      else
        shear(mi) = zero
      endif
    enddo

    do n=n2dl,n2du
      kb  = kh(n)
      if (kb < 2) cycle
      
      i = ind(1,n)
      j = ind(2,n)

      ddxi1 = ddx/cosphi(1,i)
      ddxi2 = ddx/cosphi(2,i)

      mi0 = mm(2,i,  j  ) - 2
      mn0 = mm(2,i-1,j  ) - 2
      mw0 = mm(2,i,  j-1) - 2
      me0 = mm(2,i,  j+1) - 2
      ms0 = mm(2,i+1,j  ) - 2

      ! stretch and divergence
      do k=2,kb
        mi = mi0 + k
        uterm =  u(mi)*ddxi1
        vterm = -v(mi)*ddy
        stretch(mi) = uterm - vterm
        div    (mi) = uterm + vterm
      enddo
      do k=2,min(kb,kh(mm(1,i,j-1)))
        mi = mi0 + k
        uterm = -u(mw0+k)*ddxi1
        stretch(mi) = stretch(mi) + uterm 
        div    (mi) = div    (mi) + uterm
      enddo
      do k=2,min(kb,kh(mm(1,i-1,j)))
        mi = mi0 + k
        vterm = v(mn0+k)*ddy
        stretch(mi) = stretch(mi) - vterm
        div    (mi) = div    (mi) + vterm
      enddo

      ! shear
      kshear = min(kb,kh(mm(1,i,j+1)),kh(mm(1,i+1,j)),kh(mm(1,i+1,j+1)))
      do k=2,kshear
        mi = mi0 + k
        shear(mi) = (v(me0+k)-v(mi))*ddxi2 + (u(mi)-u(ms0+k))*ddy
      enddo
      do k=max(2,kshear+1),kb
        shear(mi0+k) = zero
      enddo
    enddo
   
  end subroutine deform
  
  !=============================================================================

  subroutine calc_uvterms (imax,n2d,mm,dt,dx,ty,tx,ind,kh,khu,khv,idx,         &
                           shear,sm2,sm3,eddyh,smu,smv)
    !---------------------------------------------------------------------------
    !  Calculate u- and v-smagorinsky terms to be used directly in momeqs
    !---------------------------------------------------------------------------
  
    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain
  
    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: imax, n2d
    integer(4), intent(in)    :: mm(:,0:,0:), kh(0:), khu(0:), khv(0:), ind(:,:)
    integer(4), intent(inout) :: idx(1:,1:)
    real(8),    intent(in)    :: dx(:), dt, ty, tx(:)
    real(8),    intent(in)    :: shear(0:), eddyh(0:), sm2(0:), sm3(0:)
    real(8),    intent(inout) :: smu(0:), smv(0:)

    !- local variables ---------------------------------------------------------
    integer(4) :: n, i, j, k, keu, kev, nl, nu, mi, nn, ss, ne, ee, se, ww, sw
    real(8)    :: fy, txi, fxm, sm2tmp, sm3tmp, ehmi, ehnn, ehww, shrtmp

    fy = qrt*ty

    call domp_get_domain(kh, 1, n2d, nl, nu, idx)

    do n=nl,nu
      i = ind(1,n)
      j = ind(2,n)

      txi = tx(i)
      fxm = half*dt/(dx(i)+dx(min(i+1,imax)))

      keu = khu(n)
      kev = khv(n)

      do k=1,min(keu,kev)
        ! extract commonly used indices:
        mi = mm(k,i  ,j  )
        nn = mm(k,i-1,j  )
        ss = mm(k,i+1,j  )
        ne = mm(k,i-1,j+1)
        ee = mm(k,i  ,j+1)
        se = mm(k,i+1,j+1)
        ww = mm(k,i,  j-1)
        sw = mm(k,i+1,j-1)

        ehmi = eddyh(mi) + eddyh(ee) + eddyh(ss) + eddyh(se)
        ehnn = eddyh(mi) + eddyh(ee) + eddyh(nn) + eddyh(ne)
        ehww = eddyh(mi) + eddyh(ww) + eddyh(ss) + eddyh(sw)

        sm2tmp = sm2(mi)
        sm3tmp = sm3(mi)
        shrtmp = ehmi*shear(mi)

        ! Horizontal eddy viscosity, u:
        smu(mi) = (sm2(ee)-sm2tmp + sm3(ee)-sm3tmp)*txi                        &
                + (ehnn*shear(nn) - shrtmp)*fy

        ! Horizontal eddy viscosity, v:
        smv(mi) = (sm2(ss)-sm2tmp + sm3tmp-sm3(ss))*ty                         &
                + (shrtmp - ehww*shear(ww))*fxm
      enddo

      if (keu > kev) then
        do k=kev+1,keu
          ! extract commonly used indices:
          mi = mm(k,i  ,j  )
          nn = mm(k,i-1,j  )
          ss = mm(k,i+1,j  )
          ne = mm(k,i-1,j+1)
          ee = mm(k,i  ,j+1)
          se = mm(k,i+1,j+1)

          ehmi = eddyh(mi) + eddyh(ee) + eddyh(ss) + eddyh(se)
          ehnn = eddyh(mi) + eddyh(ee) + eddyh(nn) + eddyh(ne)

          sm2tmp = sm2(mi)
          sm3tmp = sm3(mi)

          ! Horizontal eddy viscosity, u:
          smu(mi) = (sm2(ee)-sm2tmp + sm3(ee)-sm3tmp)*txi                      &
                  + (ehnn*shear(nn) - ehmi*shear(mi))*fy
        enddo
      elseif (kev > keu) then
        do k=keu+1,kev
          ! extract commonly used indices:
          mi = mm(k,i  ,j  )
          ss = mm(k,i+1,j  )
          ee = mm(k,i  ,j+1)
          se = mm(k,i+1,j+1)
          ww = mm(k,i,  j-1)
          sw = mm(k,i+1,j-1)

          ehmi = eddyh(mi) + eddyh(ee) + eddyh(ss) + eddyh(se)
          ehww = eddyh(mi) + eddyh(ww) + eddyh(ss) + eddyh(sw)

          sm2tmp = sm2(mi)
          sm3tmp = sm3(mi)

          ! Horizontal eddy viscosity, v:
          smv(mi) = (sm2(ss)-sm2tmp + sm3tmp-sm3(ss))*ty                       &
                  + (ehmi*shear(mi) - ehww*shear(ww))*fxm
        enddo
      endif
    enddo

  end subroutine calc_uvterms

  !=============================================================================

end module smagorinsky
