module t10m_mod

  !- modules: ------------------------------------------------------------------
  use constants, only : g, vkarm, pi, half, zero, one, two, three, four, ten,  &
                        qrt, onethird, twothird

  !- implicit directives: ------------------------------------------------------
  implicit none
  private

  !- Routines: -----------------------------------------------------------------
  public  :: t10m, t10m_init

  !- Constants: ----------------------------------------------------------------
  !use these to compile "stand-alone, i.e. without the constants module:
! real(8), parameter, private :: g=9.81_8, vkarm=0.41_8
! real(8), parameter, private :: qrt = 0.25_8
! real(8), parameter, private :: pi=3.14159265358979323846_8
! real(8), parameter, private :: zero = 0.0_8, one = 1.0_8, half = 0.5_8
! real(8), parameter, private :: two = 2.0_8, three = 3.0_8
! real(8), parameter, private :: four = 4.0_8, ten = 10.0_8
! real(8), parameter, private :: onethird=1.0_8/3.0_8, twothird=2.0_8/3.0_8

  real(8), parameter, public :: qra = 0.61_8

  integer(4), parameter :: nitmin = 1, nitmax = 3

  real(8), parameter :: Beta = 1.2_8, von = vkarm, fdg = 1.0_8, Rgas = 287.1_8 
  real(8), parameter :: tdk = 273.16_8, grav = g
  real(8), parameter :: tsle = -2370.0_8, Le0 = 2501000.0_8, wetc0 = 0.622_8
  real(8), parameter :: vsa0 = 1.326e-5_8, vsa1 = 6.542e-3_8
  real(8), parameter :: vsa2 = 8.301e-6_8, vsa3 = -4.84e-9_8
  real(8), parameter :: b4 = 0.004_8
  real(8), parameter :: charnl = 0.011_8, charnh = 0.018_8, utup = 18.0_8
  real(8), parameter :: zetuhigh = 50.0_8
  real(8), parameter :: neutraldt = -0.0098_8, dter0 = 0.3_8, usr0 = 0.035_8
  real(8), parameter :: zouu = 0.011_8, zovu = 0.11_8, Ch10 = 0.00115_8
  real(8), parameter :: xlpow2 = 0.333_8
  real(8), parameter :: ug0 = 0.2_8
  real(8), parameter :: zoqmax = 0.000115_8, zoqfac = 0.000055_8, zoqpow = 0.6_8
  real(8), parameter :: cmax = 50.0_8, cslope = 0.35_8
  real(8), parameter :: c0 = 14.28_8, c1 = 8.525_8
  real(8), parameter :: p1  = 10.15_8, p2  = 1.5_8, p3 = 34.15_8
  real(8), parameter :: fifteen = 15.0_8, oneph = 1.5_8

  real(8), save :: sq3, a1, a2, Ribcu

    ! zu      wind speed measurement height (m)
    ! zt      air T measurement height (m)
    ! zq      air q measurement height (m)
    ! zi      PBL depth (m), usually 600m
  real(8), save :: zu, zt, zq, zi

contains

  !-----------------------------------------------------------------------------

  subroutine t10m (u, ts, t, qs, q, tzs, qzs)
    ! Get temperature @ std ht 10m.
    ! Based on the original cor30a.
    !
    !
    ! input:
    ! u       wind speed (m/s) at height zu (m)
    ! ts      bulk water temperature (C)
    ! t       bulk air temperature (C) at height zt
    ! qs      bulk water spec hum (kg/kg)
    ! q       bulk air spec hum (kg/kg) at height zq
    !
    !
    ! output:
    ! tzs     bulk air temperature (C) at standard height zs=10m
    ! qzs     humidity at standard height


    implicit none

    !- Arguments ---------------------------------------------------------------
    real(8), intent(in)  :: u, ts, t, Qs, Q
    real(8), intent(out) :: tzs, qzs

    !- Local vars --------------------------------------------------------------
    real(8)    :: Le, wetc
    real(8)    :: dt, dq
    real(8)    :: bf, CC, charn, Ct
    real(8)    :: L10, Ribu, rr, ta, ut, zet, zetu, zo10, zot10
    real(8)    :: zo, zot, zoq, L, usr, tsr, qsr
    real(8)    :: Cd, ug 
    real(8)    :: xx, visa, psiuo, psiq, psit, psik, psic, z, f
    real(8)    :: lzu10, lzt10, lzq10, dtjcl
    integer(4) :: i, nits

    ! air parameters -----------------------------------------------------------
    Le   = Le0 + tsle*ts
    visa = vsa0*(one + (vsa1 + (vsa2 + vsa3*t)*t)*t) 

    ! cool skin parameters -----------------------------------------------------
    wetc = wetc0*Le*qs/(Rgas*(ts+tdk)**2) 
     
    !***************   Begin bulk loop *****************************************
    ! first guess --------------------------------------------------------------
    dt    = ts - t + neutraldt*zt 
    dq    = qs - q 
    ta    = t + tdk 
    ut    = sqrt(u*u + qrt) 
    usr   = usr0*ut 
    zo10  = zouu*usr*usr/grav + zovu*visa/usr 
    xx    = von/log(ten/zo10)
    zot10 = ten*exp(-(von/Ch10)*xx) 
    lzu10 = log(zu/zo10)
    lzt10 = log(zt/zot10)
    lzq10 = log(zq/zot10)
    Cd    = (von/lzu10)**2 
    Ct    = von/lzt10 
    CC    = von*Ct/Cd 
    dtjcl = dter0
    Ribu  = -grav*zu*((dt-dtjcl)/ta + qra*dq)/ut**2 
    if (Ribu < zero) then 
      zetu = CC*Ribu/(one + Ribu/Ribcu) 
    else 
      zetu =    Ribu*(CC  + three*Ribu)
    endif 
    L10 = zu/zetu 
    if (zetu > zetuhigh) then 
      nits = nitmin 
    else
      nits = nitmax
    endif 

    ! wave parameters ----------------------------------------------------------
    if (ut > utup) then
      charn = charnh
    elseif (ut > ten) then
      charn = charnl + (ut-ten)/(utup-ten)*(charnh-charnl)
    else
      charn = charnl
    endif

    ! first guesses for usr, sr, qst:
    if (L10 <= zero) then
      ! wind
      z     = zu/L10
      xx    = sqrt(sqrt(one-fifteen*z))
      psik  = log((one+xx)*(one+xx)*(one+xx*xx)) - atan(xx) + a1
      xx    = (one - p1*z)**onethird
      psic  = p2*log(one+xx+xx*xx) - sq3*atan((one+two*xx)/sq3) + a2
      f     = z*z/(one+z*z)
      psiuo = (one-f)*psik + f*psic
      ! temperature
      z     = zt/L10
      xx    = sqrt(one-fifteen*z)
      psik  = two*log(one+xx) - log(four)
      xx    = (one - p3*z)**onethird
      psic  = oneph*log(one+xx+xx*xx) - sq3*atan((one+two*xx)/sq3) + a2
      f     = z*z/(one+z*z)
      psit  = (one-f)*psik + f*psic
      ! humidity
      z     = zq/L10
      xx    = sqrt(one-fifteen*z)
      psik  = two*log(one+xx) - log(four)
      xx    = (one - p3*z)**onethird
      psic  = oneph*log(one+xx+xx*xx) - sq3*atan((one+two*xx)/sq3) + a2
      f     = z*z/(one+z*z)
      psiq  = (one-f)*psik + f*psic
    else
      ! wind
      z     = zu/L10
      psiuo = -((one+z) + twothird*(z-c0)*exp(-min(cmax, cslope*z)) + c1)
      ! temperature
      z     = zt/L10
      xx    = (one+twothird*z)
      psit  = -(xx*sqrt(xx) + twothird*(z-c0)*exp(-min(cmax, cslope*z)) + c1)
      ! humidity
      z     = zq/L10
      xx    = (one+twothird*z)
      psiq  = -(xx*sqrt(xx) + twothird*(z-c0)*exp(-min(cmax, cslope*z)) + c1)
    endif
    usr = ut*von/(lzu10 - psiuo)
    tsr = -(dt-dtjcl)*von*fdg/(lzt10 - psit) 
    qsr = -(dq-wetc*dtjcl)*von*fdg/(lzq10 - psiq) 
        
    ! bulk loop ----------------------------------------------------------------
    do i=1,nits-1
      zet = von*grav*zu/ta*(tsr*(one+qra*q)+qra*ta*qsr)/(usr*usr)/(one+qra*q) 
      zo  = charn*usr*usr/grav+zovu*visa/usr
      rr  = zo*usr/visa 
      L   = zu/zet 
      if (L <= zero) then
        ! wind
        z     = zu/L
        xx    = sqrt(sqrt(one-fifteen*z))
        psik  = log((one+xx)*(one+xx)*(one+xx*xx)) - atan(xx) + a1
        xx    = (one - p1*z)**onethird
        psic  = p2*log(one+xx+xx*xx) - sq3*atan((one+two*xx)/sq3) + a2
        f     = z*z/(one+z*z)
        psiuo = (one-f)*psik + f*psic
        ! temperature
        z     = zt/L
        xx    = sqrt(one-fifteen*z)
        psik  = two*log(one+xx) - log(four)
        xx    = (one - p3*z)**onethird
        psic  = oneph*log(one+xx+xx*xx) - sq3*atan((one+two*xx)/sq3) + a2
        f     = z*z/(one+z*z)
        psit  = (one-f)*psik + f*psic
        ! humidity
        z     = zq/L
        xx    = sqrt(one-fifteen*z)
        psik  = two*log(one+xx) - log(four)
        xx    = (one - p3*z)**onethird
        psic  = oneph*log(one+xx+xx*xx) - sq3*atan((one+two*xx)/sq3) + a2
        f     = z*z/(one+z*z)
        psiq  = (one-f)*psik + f*psic
      else
        ! wind
        z     = zu/L
        psiuo = -((one+z) + twothird*(z-c0)*exp(-min(cmax, cslope*z)) + c1)
        ! temperature
        z     = zt/L
        xx    = (one+twothird*z)
        psit  = -(xx*sqrt(xx) + twothird*(z-c0)*exp(-min(cmax,cslope*z)) + c1)
        ! humidity
        z     = zq/L
        xx    = (one+twothird*z)
        psiq  = -(xx*sqrt(xx) + twothird*(z-c0)*exp(-min(cmax,cslope*z)) + c1)
      endif

      usr = ut*von/(lzu10 - psiuo) 
      qsr = -(dq-wetc*dtjcl)*von*fdg/(lzq10 - psiq) 
      tsr = -(dt-dtjcl)*von*fdg/(lzt10 - psit) 

      if (i == 1) then
        Bf = -grav*usr*(tsr + qra*ta*qsr)/ta
        if (Bf > zero) then
          ug = Beta*(Bf*zi)**xlpow2 
        else
          ug = ug0
        endif
        ut = sqrt(u*u+ug*ug) 
      endif
    enddo !bulk iter loop
    zet = von*grav*zu/ta*(tsr*(one+qra*q)+qra*ta*qsr)/(usr*usr)/(one+qra*q) 
    zo  = charn*usr*usr/grav+zovu*visa/usr
    rr  = zo*usr/visa 
    L   = zu/zet 
    zoq = min(zoqmax, zoqfac/rr**zoqpow) 
    zot = zoq 
    if (L <= zero) then
      ! temperature
      z    = zt/L
      xx   = sqrt(one-fifteen*z)
      psik = two*log(one+xx) - log(four)
      xx   = (one - p3*z)**onethird
      psic = oneph*log(one+xx+xx*xx) - sq3*atan((one+two*xx)/sq3) + a2
      f    = z*z/(one+z*z)
      psit = (one-f)*psik + f*psic
      ! humidity
      z    = zq/L
      xx   = sqrt(one-fifteen*z)
      psik = two*log(one+xx) - log(four)
      xx   = (one - p3*z)**onethird
      psic = oneph*log(one+xx+xx*xx) - sq3*atan((one+two*xx)/sq3) + a2
      f    = z*z/(one+z*z)
      psiq = (one-f)*psik + f*psic
    else
      ! temperature
      z    = zt/L
      xx   = (one+twothird*z)
      psit = -(xx*sqrt(xx) + twothird*(z-c0)*exp(-min(cmax,cslope*z)) + c1)
      ! humidity
      z    = zq/L
      xx   = (one+twothird*z)
      psiq = -(xx*sqrt(xx) + twothird*(z-c0)*exp(-min(cmax,cslope*z)) + c1)
    endif
    !******************* End bulk loop *****************************************
    
    ! Calculate and return the temperature value at std ht (10m) ---------------
    tzs = ts -      dtjcl -      (dt-dtjcl)*fdg*(log(ten/zot)-psit)/(lzt10-psit)
    qzs = qs - wetc*dtjcl - (dq-wetc*dtjcl)*fdg*(log(ten/zoq)-psiq)/(lzq10-psiq)

  end subroutine t10m

  !-----------------------------------------------------------------------------

  subroutine t10m_init()

    implicit none

    sq3 = sqrt(three)
    a1  = pi*half - three*log(two)
    a2  = pi/sq3 - log(three*sq3)

    zu = ten
    zq = two
    zt = two
    zi = 600.0_8

    Ribcu = -zu/(zi*b4*Beta**3) 

  end subroutine t10m_init

  !-----------------------------------------------------------------------------

end module t10m_mod
