module tflow
  use constants, only : qrt, half, zero, one, two,                             &
                        onethird, twothird, fourthird
  use nc_mod,    only : nc
  implicit none
  private

  !- Private work arrays -------------------------------------------------------
  !
  !  Now 9 work arrays, previous it was 15.
  !  Substitution as compared to ealier versions of tflow:
  !    tu/v/w  --> t1/2/3
  !    dx/y/zt --> t4/5/6
  !    dtu/v/w --> t1/2/3
  !    rin/out --> t7/8
  !    cx/y/z  --> t4/5/6
  !
  !  Computational flowchart (t,u,v,w ao args are implicit):
  !  A: c_tu/v/w --> t1/2/3 --> halo(t1/2) --> c_tt --> tt (release t1/2/3)
  !  B: c_delta --> t4/5/6 --> halo(t4/5/6) --> c_dt --> t1/2/3 (release t4/5/6)
  !  C: tt, t1/2/3 --> c_rin_rout --> t7/8 
  !  D: t1/2/3, t7/8 --> c_cx_cy_cz --> t1/2/3 (release t7/8)
  !  E: tt, t1/2/3, t4/5/6 --> t (release tt, t1/2/3/4/5/6/7)
  !
  !-----------------------------------------------------------------------------
  real(8), allocatable, private :: tt(:,:), t1(:,:), t2(:,:), t3(:,:)
  real(8), allocatable, private :: t4(:,:), t5(:,:), t6(:,:), t7(:,:), t8(:,:)
  
  !- Other private module data -------------------------------------------------
  real(8),              private :: dx, dy, dt, dxdy, ddxdy, dtdx, dtdy, dtdxdy
  integer(4), save,     private :: n3dmax = 0

  public  :: tflow_int, tflow_alloc, tflow_ft
  private :: c_rin_rout, copy_t2tt, diffusion_hx, diffusion_vi, diffusion_vx,  &
             c_tu, c_tv, c_tw, c_tt, c_delta, c_dt, c_cx_cy_cz, advection

contains
    
  ! ----------------------------------------------------------------------------

  subroutine c_rin_rout (kmax,m,ind,n2d,kh,hx,hy,h_new,t,u,v,w,cosphi,idx)

    ! NOTE: it modifies the global module variables: t7, t8  
    ! and it uses tt, t1, t2, t3 

    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: kmax, n2d
    integer(4), intent(in)    :: m(:,0:,0:), ind(:,:), kh(0:)
    real(8),    intent(in)    :: u(0:), v(0:), w(0:), cosphi(:,0:)
    real(8),    intent(in)    :: hx(0:), hy(0:), h_new(0:), t(:,0:)
    integer(4), intent(inout) :: idx(1:,1:)

    !- local variables ---------------------------------------------------------
    integer(4) :: n, k, kb, i, j, ic, mi, me, ms, mu, mw, mn, md, n2dl, n2du
    integer(4) :: mi0, mw0, mn0, me0, ms0
    real(8)    :: pin, qin, pout, qout
    real(8)    :: fac0, humi, humw, hvmi, hvmn, hwmi, hwmd
    real(8)    :: facx, facy1, facy2, dh, tmin, tmax
    real(8)    :: tmint(nc,kmax), tmaxt(nc,kmax)
    real(8), parameter :: small = 1.e-12_8

    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)

    ! k=1 unrolled:
    do n = n2dl,n2du
      kb = kh(n)
      if (kb < 1) cycle
      i = ind(1,n)
      j = ind(2,n)

      fac0  = one/cosphi(1,i)
      facx  = fac0/dx
      facy1 = fac0*cosphi(2,i  )/dy
      facy2 = fac0*cosphi(2,i-1)/dy

      mi = n
      mw = m(1,i,  j-1)
      mn = m(1,i-1,j  )
      if (kb > 1) then
        md = m(2,i,j)
      else
        md = 0
      endif
      dh   = dt/h_new(mi)
      humi = hx(mi)*u(mi)*facx *dh
      humw = hx(mw)*u(mw)*facx *dh
      hvmi = hy(mi)*v(mi)*facy1*dh
      hvmn = hy(mn)*v(mn)*facy2*dh
      hwmi =        w(mi)*      dh
      hwmd =        w(md)*      dh
      do ic=1,nc
        t8(ic,mi) =   max(humi*t1(ic,mi),zero) - min(humw*t1(ic,mw),zero)      &
                    - min(hvmi*t2(ic,mi),zero) + max(hvmn*t2(ic,mn),zero)      &
                    + max(hwmi*t3(ic,mi),zero) - min(hwmd*t3(ic,md),zero)
        t7(ic,mi) = - min(humi*t1(ic,mi),zero) + max(humw*t1(ic,mw),zero)      &
                    + max(hvmi*t2(ic,mi),zero) - min(hvmn*t2(ic,mn),zero)      &
                    - min(hwmi*t3(ic,mi),zero) + max(hwmd*t3(ic,md),zero)
      enddo
    enddo

    do n = n2dl,n2du
      kb = kh(n)
      if (kb <= 1) cycle
      i = ind(1,n)
      j = ind(2,n)

      fac0  = one/cosphi(1,i)
      facx  = fac0/dx
      facy1 = fac0*cosphi(2,i  )/dy
      facy2 = fac0*cosphi(2,i-1)/dy

      !  column offsets:
      mi0 = m(2,i,  j  ) - 2
      mw0 = m(2,i,  j-1) - 2
      mn0 = m(2,i-1,j  ) - 2
      do k=2,kb-1
        mi = mi0 + k
        md = mi + 1
        dh   = dt/h_new(mi)
        humi = hx(mi)*u(mi)*facx *dh
        hvmi = hy(mi)*v(mi)*facy1*dh
        hwmi =        w(mi)*      dh
        hwmd =        w(md)*      dh
        do ic=1,nc
          t8(ic,mi)=   max(humi*t1(ic,mi),zero)                                &
                     - min(hvmi*t2(ic,mi),zero)                                &
                     + max(hwmi*t3(ic,mi),zero) - min(hwmd*t3(ic,md),zero)
          t7(ic,mi)= - min(humi*t1(ic,mi),zero)                                &
                     + max(hvmi*t2(ic,mi),zero)                                &
                     - min(hwmi*t3(ic,mi),zero) + max(hwmd*t3(ic,md),zero)
        enddo
      enddo
      do k=2,min(kb-1,kh(m(1,i,j-1)))
        mi = mi0 + k
        mw = mw0 + k
        humw = hx(mw)*u(mw)*facx *dt/h_new(mi)
        do ic=1,nc
          t8(ic,mi)= t8(ic,mi)                  - min(humw*t1(ic,mw),zero) 
          t7(ic,mi)= t7(ic,mi)                  + max(humw*t1(ic,mw),zero)
        enddo
      enddo
      do k=2,min(kb-1,kh(m(1,i-1,j)))
        mi = mi0 + k
        mn = mn0 + k
        hvmn = hy(mn)*v(mn)*facy2*dt/h_new(mi)
        do ic=1,nc
          t8(ic,mi)= t8(ic,mi)                  + max(hvmn*t2(ic,mn),zero)
          t7(ic,mi)= t7(ic,mi)                  - min(hvmn*t2(ic,mn),zero)
        enddo
      enddo
      ! k=kb
      mi = m(kb,i,  j  )
      mw = m(kb,i,  j-1)
      mn = m(kb,i-1,j  )
      dh   = dt/h_new(mi)
      humi = hx(mi)*u(mi)*facx *dh
      humw = hx(mw)*u(mw)*facx *dh
      hvmi = hy(mi)*v(mi)*facy1*dh
      hvmn = hy(mn)*v(mn)*facy2*dh
      hwmi =        w(mi)*      dh
      do ic=1,nc
        t8(ic,mi) =   max(humi*t1(ic,mi),zero) - min(humw*t1(ic,mw),zero)      &
                    - min(hvmi*t2(ic,mi),zero) + max(hvmn*t2(ic,mn),zero)      &
                    + max(hwmi*t3(ic,mi),zero)
        t7(ic,mi) = - min(humi*t1(ic,mi),zero) + max(humw*t1(ic,mw),zero)      &
                    + max(hvmi*t2(ic,mi),zero) - min(hvmn*t2(ic,mn),zero)      &
                    - min(hwmi*t3(ic,mi),zero)
      enddo
    enddo
    
    ! limiter loop:
    do n = n2dl,n2du
      kb = kh(n)
      if (kb < 1) cycle 
      i = ind(1,n)
      j = ind(2,n)

      ! k=1
      mi = n
      me = m(1,i  ,j+1)
      mw = m(1,i  ,j-1)
      ms = m(1,i+1,j  )
      mn = m(1,i-1,j  )
      if (kmax > 1) then
        md = m(2,i,j)
      else
        md = 0
      endif
      do ic=1,nc
        tmax=max(t(ic,mi),tt(ic,mi))
        tmin=min(t(ic,mi),tt(ic,mi))
        if (mn > 0) then
          tmax=max(tmax,t(ic,mn),tt(ic,mn))
          tmin=min(tmin,t(ic,mn),tt(ic,mn))
        endif
        if (ms > 0) then
          tmax=max(tmax,t(ic,ms),tt(ic,ms))
          tmin=min(tmin,t(ic,ms),tt(ic,ms))
        endif
        if (mw > 0) then
          tmax=max(tmax,t(ic,mw),tt(ic,mw))
          tmin=min(tmin,t(ic,mw),tt(ic,mw))
        endif
        if (me > 0) then
          tmax=max(tmax,t(ic,me),tt(ic,me))
          tmin=min(tmin,t(ic,me),tt(ic,me))
        endif
        if (md > 0) then
          tmax=max(tmax,t(ic,md),tt(ic,md))
          tmin=min(tmin,t(ic,md),tt(ic,md))
        endif
        pin=t7 (ic,mi)
        !     pin>=0
        if (pin > small) then
          qin=tmax-tt(ic,mi)
        !     qin>=0
          t7(ic,mi)=min(one,qin/pin)
        else
          t7(ic,mi)=one
        endif
        pout=t8(ic,mi)
        !     pout>=0
        if (pout > small) then
          qout=tt(ic,mi)-tmin
        !     qout>=0
          t8(ic,mi)=min(one,qout/pout)
        else
          t8(ic,mi)=one
        endif
      enddo

      if (kb <= 1) cycle

      mi0 = m(2,i,j) - 2

      do k=2,kb
        mi = mi0 + k
        do ic=1,nc
          tmaxt(ic,k) = max(t(ic,mi),tt(ic,mi))
          tmint(ic,k) = min(t(ic,mi),tt(ic,mi))
        enddo
      enddo

      do k=2,kb-1
        md = mi0 + k + 1
        do ic=1,nc
          tmaxt(ic,k) = max(tmaxt(ic,k),t(ic,md),tt(ic,md))
          tmint(ic,k) = min(tmint(ic,k),t(ic,md),tt(ic,md))
        enddo
      enddo
      
      mu = n
      do ic=1,nc
        tmaxt(ic,2) = max(tmaxt(ic,2),t(ic,mu),tt(ic,mu))
        tmint(ic,2) = min(tmint(ic,2),t(ic,mu),tt(ic,mu))
      enddo
      do k=3,kb
        mi = mi0 + k
        mu = mi - 1
        do ic=1,nc
          tmaxt(ic,k) = max(tmaxt(ic,k),t(ic,mu),tt(ic,mu))
          tmint(ic,k) = min(tmint(ic,k),t(ic,mu),tt(ic,mu))
        enddo
      enddo

      me0 = m(2,i,  j+1) - 2
      mw0 = m(2,i,  j-1) - 2
      ms0 = m(2,i+1,j  ) - 2
      mn0 = m(2,i-1,j  ) - 2

      do k=2,min(kb,kh(m(1,i,j+1)))
        me = me0 + k
        do ic=1,nc
          tmaxt(ic,k) = max(tmaxt(ic,k),t(ic,me),tt(ic,me))
          tmint(ic,k) = min(tmint(ic,k),t(ic,me),tt(ic,me))
        enddo
      enddo
      do k=2,min(kb,kh(m(1,i,j-1)))
        mw = mw0 + k
        do ic=1,nc
          tmaxt(ic,k) = max(tmaxt(ic,k),t(ic,mw),tt(ic,mw))
          tmint(ic,k) = min(tmint(ic,k),t(ic,mw),tt(ic,mw))
        enddo
      enddo
      do k=2,min(kb,kh(m(1,i+1,j)))
        ms = ms0 + k
        do ic=1,nc
          tmaxt(ic,k) = max(tmaxt(ic,k),t(ic,ms),tt(ic,ms))
          tmint(ic,k) = min(tmint(ic,k),t(ic,ms),tt(ic,ms))
        enddo
      enddo
      do k=2,min(kb,kh(m(1,i-1,j)))
        mn = mn0 + k
        do ic=1,nc
          tmaxt(ic,k) = max(tmaxt(ic,k),t(ic,mn),tt(ic,mn))
          tmint(ic,k) = min(tmint(ic,k),t(ic,mn),tt(ic,mn))
        enddo
      enddo

      do k=2,kb
        mi = mi0 + k
        do ic=1,nc
          pin = t7(ic,mi)
          !     pin>=0
          if (pin > small) then
            qin = tmaxt(ic,k) - tt(ic,mi)
            !     qin>=0
            t7(ic,mi) = min(one,qin/pin)
          else
            t7(ic,mi) = one
          endif
          pout = t8(ic,mi)
          !     pout>=0
          if (pout > small) then
            qout = tt(ic,mi) - tmint(ic,k)
            !     qout>=0
            t8(ic,mi) = min(one,qout/pout)
          else
            t8(ic,mi) = one
          endif
        enddo
      enddo

    enddo

  end subroutine c_rin_rout

  ! ----------------------------------------------------------------------------

  subroutine c_cx_cy_cz (m,ind,n2d,kh,u,v,w,idx)

    ! NOTE: it modifies the global module variables: t4, t5, t6
    ! and it uses t7, t8, t1, t2, t3 

    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: n2d
    integer(4), intent(in)    :: m(:,0:,0:), ind(:,:), kh(0:)
    real(8),    intent(in)    :: u(0:), v(0:), w(0:)
    integer(4), intent(inout) :: idx(1:,1:)

    !- local variables ---------------------------------------------------------
    integer(4) :: n, k, kb, i, j, ic, mi, me, ms, md
    integer(4) :: n2dl, n2du, mi0, me0, ms0, kbe, kbs

    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)

    ! compute t4, t5, t6

    ! k=1 unrolled
    do n = n2dl,n2du
      if (kh(n) == 0) cycle
      i = ind(1,n)
      j = ind(2,n)

      mi = n
      me = m(1,i,  j+1)
      ms = m(1,i+1,j  )

      t6(1:nc,mi) = zero
        
      do ic=1,nc
        if (u(mi)*t1 (ic,mi) >= zero) then
          t4(ic,mi) = min(t7(ic,me),t8(ic,mi))
        else
          t4(ic,mi) = min(t7(ic,mi),t8(ic,me))
        endif
        if (v(mi)*t2 (ic,mi) <= zero) then
          t5(ic,mi) = min(t7(ic,ms),t8(ic,mi))
        else
          t5(ic,mi) = min(t7(ic,mi),t8(ic,ms))
        endif
      enddo
    enddo

    do n = n2dl,n2du
      kb = kh(n)
      if (kb <= 1) cycle
      i = ind(1,n)
      j = ind(2,n)

      mi0 = m(2,i  ,j  ) - 2
      me0 = m(2,i  ,j+1) - 2
      ms0 = m(2,i+1,j  ) - 2

      mi = n
      md = m(2,i,j)
      do ic=1,nc
        if (w(md)*t3 (ic,md) <= zero) then
          t6(ic,md) = min(t7(ic,md),t8(ic,mi))
        else
          t6(ic,md) = min(t7(ic,mi),t8(ic,md))
        endif
      enddo
      do k=2,kb-1
        mi = mi0 + k
        md = mi + 1
        do ic=1,nc
          if (w(md)*t3 (ic,md) <= zero) then
            t6(ic,md) = min(t7(ic,md),t8(ic,mi))
          else
            t6(ic,md) = min(t7(ic,mi),t8(ic,md))
          endif
        enddo
      enddo

      kbe = kh(m(1,i,j+1))
      do k=2,min(kb,kbe)
        mi = mi0 + k
        me = me0 + k
        do ic=1,nc
          if (u(mi)*t1 (ic,mi) >= zero) then
            t4(ic,mi) = min(t7(ic,me),t8(ic,mi))
          else
            t4(ic,mi) = min(t7(ic,mi),t8(ic,me))
          endif
        enddo
      enddo
      if (kb > kbe) then
        me = 0
        do k=max(2,kbe+1),kb
          mi = mi0 + k
          do ic=1,nc
            if (u(mi)*t1 (ic,mi) >= zero) then
              t4(ic,mi) = min(t7(ic,me),t8(ic,mi))
            else
              t4(ic,mi) = min(t7(ic,mi),t8(ic,me))
            endif
          enddo
        enddo
      endif

      kbs = kh(m(1,i+1,j))
      do k=2,min(kb,kbs)
        mi = mi0 + k
        ms = ms0 + k
        do ic=1,nc
          if (v(mi)*t2 (ic,mi) <= zero) then
            t5(ic,mi) = min(t7(ic,ms),t8(ic,mi))
          else
            t5(ic,mi) = min(t7(ic,mi),t8(ic,ms))
          endif
        enddo
      enddo
      if (kb > kbs) then
        ms = 0
        do k=max(2,kbs+1),kb
          mi = mi0 + k
          do ic=1,nc
            if (v(mi)*t2 (ic,mi) <= zero) then
              t5(ic,mi) = min(t7(ic,ms),t8(ic,mi))
            else
              t5(ic,mi) = min(t7(ic,mi),t8(ic,ms))
            endif
          enddo
        enddo
      endif

    enddo

  end subroutine c_cx_cy_cz

  !  ---------------------------------------------------------------------------

  subroutine c_dt (kmax,m,ind,n2d,advecstab,h_new,kh,khu,khv,u,v,w,cosphi,idx)

    ! NOTE: it modifies the global module variables: t1, t2, t3  and
    ! it uses t4, t5 ,t6 

    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: kmax,n2d
    integer(4), intent(in) :: m(:,0:,0:),ind(:,:)
    integer(4), intent(in) :: kh(0:),khu(0:),khv(0:)
    real(8), intent(in)    :: advecstab,h_new(0:)
    real(8), intent(in)    :: u(0:),v(0:),w(0:)
    real(8), intent(in)    :: cosphi(:,0:)
    integer(4), intent(inout) :: idx(1:,1:)

    !- local variables ---------------------------------------------------------
    real(8)    :: fac
    integer(4) :: n,k,kb,i,j,kbu,kbv,kbm1, mi, kp, mi0, md, me, mw, mn, ms
    integer(4) :: mse, mne, mde, mu, mue, msw, mds, mus
    real(8)    :: sx, sy, sz, dtddx, dtddy, dtddz, hdt, dtddx0
    real(8)    :: s1, s2, sxp, sxm, syp, sym, szm, szp
    integer(4) :: n2dl, n2du

    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)

    hdt    = half*dt
    dtddy  = hdt/dy
    dtddx0 = hdt/dx

    ! compute t1  --------------------------------------------------------------
    do n=n2dl, n2du
      kbu = khu(n)
      kb  = kh(n)
      i   = ind(1,n)
      j   = ind(2,n)

      ! make sure to have nice values throughout
      if (kb > 1 .and. kb > kbu) t1(1:nc,m(max(2,kbu+1),i,j):m(kb,i,j)) = zero
      if (kbu == 0) then
        t1(1:nc,n) = zero
        cycle
      endif

      dtddx = dtddx0/cosphi(1,i)

      ! unroll k=1:
      if (u(n) >= zero) then
    
        sx  = dtddx*u(n)
        sxm = (half-fourthird*sx)
        fac = half-sx
        t1(1:nc,n)=fac*t4(1:nc,n)
    
        mn = m(1,i-1,j)
        if (v(mn) < zero) then
          sy = dtddy*v(mn)
          s1 = sxm*sy
          s2 = (half+ twothird*sy)*sy
          t1(1:nc,n) = t1(1:nc,n)                                              &
                      - s1*(t4(1:nc,mn)-t4(1:nc,n))                            &
                      + s2*(t5(1:nc,mn)-t5(1:nc,n))
        endif
    
        if (v(n) > zero) then
          ms = m(1,i+1,j)
          sy = dtddy*v(n)
          s1 = sxm*sy
          s2 = (half- twothird*sy)*sy
          t1(1:nc,n) = t1(1:nc,n)                                              &
                      - s1*(t4(1:nc,n)-t4(1:nc,ms))                            &
                      - s2*(t5(1:nc,n)-t5(1:nc,ms))
        endif
    
        if (kmax > 1) then
          md = m(2,i,j)
          if (w(md) > zero) then
            dtddz = hdt/(h_new(n) - advecstab)
            sz = dtddz*w(md)
            s1 = sxm*sz
            s2 = (half- twothird*sz)*sz
            t1(1:nc,n) = t1(1:nc,n)                                            &
                        - s1*(t4(1:nc,n)-t4(1:nc,md))                          &
                        - s2*(t6(1:nc,n)-t6(1:nc,md))
          endif
        endif
    
      else ! u < 0.0
    
        sx  = dtddx*u(n)
        sxp = (half+fourthird*sx)
        fac = -half-sx
        me  = m(1,i,j+1)
        t1(1:nc,n)=fac*t4(1:nc,me)
    
        mne = m(1,i-1,j+1)
        if (v(mne) < zero) then
          sy = dtddy*v(mne)
          s1 = sxp*sy
          s2 = (half+ twothird*sy)*sy
          t1(1:nc,n) = t1(1:nc,n)                                              &
                      + s1*(t4(1:nc,mne)-t4(1:nc,me))                          &
                      - s2*(t5(1:nc,mne)-t5(1:nc,me))
        endif
    
        if (v(me) > zero) then
          mse = m(1,i+1,j+1)
          sy = dtddy*v(me)
          s1 = sxp*sy
          s2 = (half- twothird*sy)*sy
          t1(1:nc,n) = t1(1:nc,n)                                              &
                      + s1*(t4(1:nc,me)-t4(1:nc,mse))                          &
                      + s2*(t5(1:nc,me)-t5(1:nc,mse))
        endif
    
        if (kmax > 1) then
          mde = m(2,i,j+1)
          if (w(mde) > zero) then
            dtddz = hdt/(h_new(n) - advecstab)
            sz = dtddz*w(mde)
            s1 = sxp*sz
            s2 = (half- twothird*sz)*sz
            t1(1:nc,n) = t1(1:nc,n)                                            &
                        + s1*(t4 (1:nc,me)-t4(1:nc,mde))                       &
                        + s2*(t6 (1:nc,me)-t6(1:nc,mde))
          endif
        endif

      endif

      mi0 = m(2,i,j) - 2
      do k=2,kbu
        mi = mi0 + k

        if (u(mi) >= zero) then
    
          sx  = dtddx*u(mi)
          sxm = (half-fourthird*sx)
          fac = half-sx
          t1(1:nc,mi)=fac*t4(1:nc,mi)
    
          mn = m(k,i-1,j)
          if (v(mn) < zero) then
            sy = dtddy*v(mn)
            s1 = sxm*sy
            s2 = (half+ twothird*sy)*sy
            t1(1:nc,mi) = t1(1:nc,mi)                                          &
                        - s1*(t4(1:nc,mn)-t4(1:nc,mi))                         &
                        + s2*(t5(1:nc,mn)-t5(1:nc,mi))
          endif
    
          if (v(mi) > zero) then
            ms = m(k,i+1,j)
            sy = dtddy*v(mi)
            s1 = sxm*sy
            s2 = (half- twothird*sy)*sy
            t1(1:nc,mi) = t1(1:nc,mi)                                          &
                        - s1*(t4(1:nc,mi)-t4(1:nc,ms))                         &
                        - s2*(t5(1:nc,mi)-t5(1:nc,ms))
          endif
    
          if (w(mi) < zero) then
            mu = m(k-1,i,j)
            dtddz = hdt/h_new(mi)
            sz = dtddz*w(mi)
            s1 = sxm*sz
            s2 = (half+ twothird*sz)*sz
            t1(1:nc,mi) = t1(1:nc,mi)                                          &
                        - s1*(t4(1:nc,mu)-t4(1:nc,mi))                         &
                        + s2*(t6(1:nc,mu)-t6(1:nc,mi))
          endif
    
          if (k < kmax) then
            md = m(k+1,i,j)
            if (w(md) > zero) then
              dtddz = hdt/h_new(mi)
              sz = dtddz*w(md)
              s1 = sxm*sz
              s2 = (half- twothird*sz)*sz
              t1(1:nc,mi) = t1(1:nc,mi)                                        &
                          - s1*(t4(1:nc,mi)-t4(1:nc,md))                       &
                          - s2*(t6(1:nc,mi)-t6(1:nc,md))
            endif
          endif
    
        else ! u < 0.0
    
          sx  = dtddx*u(mi)
          sxp = (half+fourthird*sx)
          fac = -half-sx
          me  = m(k,i,j+1)
          t1(1:nc,mi)=fac*t4(1:nc,me)
    
          mne = m(k,i-1,j+1)
          if (v(mne) < zero) then
            sy = dtddy*v(mne)
            s1 = sxp*sy
            s2 = (half+ twothird*sy)*sy
            t1(1:nc,mi) = t1(1:nc,mi)                                          &
                        + s1*(t4(1:nc,mne)-t4(1:nc,me))                        &
                        - s2*(t5(1:nc,mne)-t5(1:nc,me))
          endif
    
          if (v(me) > zero) then
            sy = dtddy*v(me)
            s1 = sxp*sy
            s2 = (half- twothird*sy)*sy
            mse = m(k,i+1,j+1)
            t1(1:nc,mi) = t1(1:nc,mi)                                          &
                        + s1*(t4(1:nc,me)-t4(1:nc,mse))                        &
                        + s2*(t5(1:nc,me)-t5(1:nc,mse))
          endif
    
          if (w(me) < zero) then
            dtddz = hdt/h_new(mi)
            sz = dtddz*w(me)
            s1 = sxp*sz
            s2 = (half+ twothird*sz)*sz
            mue = m(k-1,i,j+1)
            t1(1:nc,mi) = t1(1:nc,mi)                                          &
                        + s1*(t4(1:nc,mue)-t4(1:nc,me))                        &
                        - s2*(t6(1:nc,mue)-t6(1:nc,me))
          endif
    
          if (k < kmax) then
            mde = m(k+1,i,j+1)
            if (w(mde) > zero) then
              dtddz = hdt/h_new(mi)
              sz = dtddz*w(mde)
              s1 = sxp*sz
              s2 = (half- twothird*sz)*sz
              t1(1:nc,mi) = t1(1:nc,mi)                                        &
                          + s1*(t4 (1:nc,me)-t4(1:nc,mde))                     &
                          + s2*(t6 (1:nc,me)-t6(1:nc,mde))
            endif
          endif
    
        endif
      enddo
    enddo
    
    ! compute t2  --------------------------------------------------------------
    do n=n2dl,n2du
      kbv = khv(n)
      kb  = kh(n)
      i   = ind(1,n)
      j   = ind(2,n)

      ! make sure to have nice values throughout
      if (kb > 1 .and. kb > kbv) t2(1:nc,m(max(2,kbv+1),i,j):m(kb,i,j)) = zero
      if (kbv == 0) then
        t2(1:nc,n) = zero
        cycle
      endif

      dtddx = dtddx0/cosphi(1,i)

      ! unroll k=1:
      if (v(n) <= zero) then
    
        sy  = dtddy*v(n)
        syp =-(half+fourthird*sy)
        fac = -half-sy
        t2(1:nc,n)=fac*t5(1:nc,n)

        mw = m(1,i,j-1)
        if (u(mw) > zero) then
          sx = dtddx*u(mw)
          s1 = syp*sx
          s2 = (half- twothird*sx)*sx
          t2(1:nc,n) = t2(1:nc,n)                                              &
                      - s1*(t5(1:nc,n)-t5(1:nc,mw))                            &
                      - s2*(t4(1:nc,n)-t4(1:nc,mw))
        endif
    
        if (u(n) < zero) then
          sx = dtddx*u(n)
          s1 = syp*sx
          s2 =-(half+ twothird*sx)*sx
          me = m(1,i,j+1)
          t2(1:nc,n) = t2(1:nc,n)                                              &
                      - s1*(t5(1:nc,me)-t5(1:nc,n))                            &
                      - s2*(t4(1:nc,me)-t4(1:nc,n))
        endif
    
        if (kmax > 1) then
          md = m(2,i,j)
          if (w(md) > zero) then
            dtddz = hdt/(h_new(n) - advecstab)
            sz = dtddz*w(md)
            s1 = syp*sz
            s2 = (half- twothird*sz)*sz
            t2(1:nc,n) = t2(1:nc,n)                                            &
                        - s1*(t5(1:nc,n)-t5(1:nc,md))                          &
                        - s2*(t6(1:nc,n)-t6(1:nc,md))
          endif
        endif
    
      else ! v > 0.0
    
        sy  = dtddy*v(n)
        sym = (half-fourthird*sy)
        fac = half-sy
        ms = m(1,i+1,j)
        t2(1:nc,n)=fac*t5(1:nc,ms)
    
        msw = m(1,i+1,j-1)
        if (u(msw) > zero) then
          sx = dtddx*u(msw)
          s1 = sym*sx
          s2 = (half- twothird*sx)*sx
          t2(1:nc,n) = t2(1:nc,n)                                              &
                      - s1*(t5(1:nc,ms)-t5(1:nc,msw))                          &
                      - s2*(t4(1:nc,ms)-t4(1:nc,msw))
        endif
    
        if (u(ms) < zero) then
          sx = dtddx*u(ms)
          s1 = sym*sx
          s2 =-(half+ twothird*sx)*sx
          mse = m(1,i+1,j+1)
          t2(1:nc,n) = t2(1:nc,n)                                              &
                      - s1*(t5(1:nc,mse)-t5(1:nc,ms))                          &
                      - s2*(t4(1:nc,mse)-t4(1:nc,ms))
        endif
    
        if (kmax > 1) then
          mds = m(2,i+1,j)
          if (w(mds) > zero) then
            dtddz = hdt/(h_new(n) - advecstab)
            sz = dtddz*w(mds)
            s1 = sym*sz
            s2 = (half- twothird*sz)*sz
            t2(1:nc,n) = t2(1:nc,n)                                            &
                        - s1*(t5(1:nc,ms)-t5(1:nc,mds))                        &
                        - s2*(t6(1:nc,ms)-t6(1:nc,mds))
          endif
        endif
    
      endif

      mi0 = m(2,i,j) - 2
      do k=2,kbv
        mi = mi0 + k

        if (v(mi) <= zero) then
    
          sy  = dtddy*v(mi)
          syp = -(half+fourthird*sy)
          fac = -half-sy
          t2(1:nc,mi)=fac*t5(1:nc,mi)

          mw = m(k,i,j-1)
          if (u(mw) > zero) then
            sx = dtddx*u(mw)
            s1 = syp*sx
            s2 = (half- twothird*sx)*sx
            t2(1:nc,mi) = t2(1:nc,mi)                                          &
                        - s1*(t5(1:nc,mi)-t5(1:nc,mw))                         &
                        - s2*(t4(1:nc,mi)-t4(1:nc,mw))
          endif
    
          if (u(mi) < zero) then
            sx = dtddx*u(mi)
            s1 = syp*sx
            s2 =-(half+ twothird*sx)*sx
            me = m(k,i,j+1)
            t2(1:nc,mi) = t2(1:nc,mi)                                          &
                        - s1*(t5(1:nc,me)-t5(1:nc,mi))                         &
                        - s2*(t4(1:nc,me)-t4(1:nc,mi))
          endif
    
          if (w(mi) < zero) then
            dtddz = hdt/h_new(mi)
            sz = dtddz*w(mi)
            s1 = syp*sz
            s2 =-(half+ twothird*sz)*sz
            mu = m(k-1,i,j)
            t2(1:nc,mi) = t2(1:nc,mi)                                          &
                        - s1*(t5(1:nc,mu)-t5(1:nc,mi))                         &
                        - s2*(t6(1:nc,mu)-t6(1:nc,mi))
          endif
    
          if (k < kmax) then
            md = m(k+1,i,j)
            if (w(md) > zero) then
              dtddz = hdt/h_new(mi)
              sz = dtddz*w(md)
              s1 = syp*sz
              s2 = (half- twothird*sz)*sz
              t2(1:nc,mi) = t2(1:nc,mi)                                        &
                          - s1*(t5(1:nc,mi)-t5(1:nc,md))                       &
                          - s2*(t6(1:nc,mi)-t6(1:nc,md))
            endif
          endif
    
        else ! v > 0.0
    
          sy  = dtddy*v(mi)
          sym = (half-fourthird*sy)
          fac = half-sy
          ms = m(k,i+1,j)
          t2(1:nc,mi)=fac*t5(1:nc,ms)
    
          msw = m(k,i+1,j-1)
          if (u(msw) > zero) then
            sx = dtddx*u(msw)
            s1 = sym*sx
            s2 = (half- twothird*sx)*sx
            t2(1:nc,mi) = t2(1:nc,mi)                                          &
                        - s1*(t5(1:nc,ms)-t5(1:nc,msw))                        &
                        - s2*(t4(1:nc,ms)-t4(1:nc,msw))
          endif
    
          if (u(ms) < zero) then
            sx = dtddx*u(ms)
            s1 = sym*sx
            s2 =-(half+ twothird*sx)*sx
            mse = m(k,i+1,j+1)
            t2(1:nc,mi) = t2(1:nc,mi)                                          &
                        - s1*(t5(1:nc,mse)-t5(1:nc,ms))                        &
                        - s2*(t4(1:nc,mse)-t4(1:nc,ms))
          endif
    
          if (w(ms) < zero) then
            dtddz = hdt/h_new(mi)
            sz = dtddz*w(ms)
            s1 = sym*sz
            s2 =-(half+ twothird*sz)*sz
            mus = m(k-1,i+1,j)
            t2(1:nc,mi) = t2(1:nc,mi)                                          &
                        - s1*(t5(1:nc,mus)-t5(1:nc,ms))                        &
                        - s2*(t6(1:nc,mus)-t6(1:nc,ms))
          endif
    
          if (k < kmax) then
            mds = m(k+1,i+1,j)
            if (w(mds) > zero) then
              dtddz = hdt/h_new(mi)
              sz = dtddz*w(mds)
              s1 = sym*sz
              s2 = (half- twothird*sz)*sz
              t2(1:nc,mi) = t2(1:nc,mi)                                        &
                          - s1*(t5(1:nc,ms)-t5(1:nc,mds))                      &
                          - s2*(t6(1:nc,ms)-t6(1:nc,mds))
            endif
          endif
    
        endif
      enddo
    enddo
    
    ! compute t3  --------------------------------------------------------------
    do n=n2dl,n2du
      kb = kh(n)

      ! make sure to have nice values throughout
      mi = n
      t3(1:nc,mi) = zero

      if (kb <= 1) cycle
      i = ind(1,n)
      j = ind(2,n)
      kbm1 = kb-1

      dtddx = dtddx0/cosphi(1,i)

      k  = 1
      kp = k + 1
      md = m(kp,i,j)

      if (w(md) <= zero) then
        sz  = w(md)*hdt/(h_new(mi) - advecstab)
        szp = -(half+fourthird*sz)
        fac = -half-sz
        t3(1:nc,md)=fac*t6(1:nc,mi)

        mw = m(k,i,j-1)
        if (u(mw) > zero) then
          sx = dtddx*u(mw)
          s1 = szp*sx
          s2 = (half- twothird*sx)*sx
          t3(1:nc,md) = t3(1:nc,md)                                            &
                 - s1*(t6(1:nc,mi)-t6(1:nc,mw)) - s2*(t4(1:nc,mi)-t4(1:nc,mw))
        endif
    
        if (u(mi) < zero) then
          me = m(k,i,j+1)
          sx = dtddx*u(mi)
          s1 = szp*sx
          s2 =-(half+ twothird*sx)*sx
          t3(1:nc,md) = t3(1:nc,md)                                            &
                 - s1*(t6(1:nc,me)-t6(1:nc,mi)) - s2*(t4(1:nc,me)-t4(1:nc,mi))
        endif
    
        mn = m(k,i-1,j)
        if (v(mn) < zero) then
          sy = dtddy*v(mn)
          s1 = szp*sy
          s2 =-(half+ twothird*sy)*sy
          t3(1:nc,md) = t3(1:nc,md)                                            &
                 - s1*(t6(1:nc,mn)-t6(1:nc,mi)) - s2*(t5(1:nc,mn)-t5(1:nc,mi))
        endif
    
        if (v(mi) > zero) then
          ms = m(k,i+1,j)
          sy = dtddy*v(mi)
          s1 = szp*sy
          s2 = (half- twothird*sy)*sy
          t3(1:nc,md) = t3(1:nc,md)                                            &
                 - s1*(t6(1:nc,mi)-t6(1:nc,ms)) - s2*(t5(1:nc,mi)-t5(1:nc,ms))
        endif
    
      else ! w > 0.0
        sz  = w(md)*hdt/(h_new(mi) - advecstab)
        szm = (half-fourthird*sz)
        fac = half-sz
        t3(1:nc,md) = fac*t6(1:nc,md)
    
        mw = m(kp,i,j-1)
        if (u(mw) > zero) then
          sx = dtddx*u(mw)
          s1 = szm*sx
          s2 = (half- twothird*sx)*sx
          t3(1:nc,md) = t3(1:nc,md)                                            &
                 - s1*(t6(1:nc,md)-t6(1:nc,mw)) - s2*(t4(1:nc,md)-t4(1:nc,mw))
        endif
    
        if (u(md) < zero) then
          me = m(kp,i,j+1)
          sx = dtddx*u(md)
          s1 = szm*sx
          s2 =-(half+ twothird*sx)*sx
          t3(1:nc,md) = t3(1:nc,md)                                            &
                 - s1*(t6(1:nc,me)-t6(1:nc,md)) - s2*(t4(1:nc,me)-t4(1:nc,md))
        endif
    
        mn = m(kp,i-1,j)
        if (v(mn) < zero) then
          sy = dtddy*v(mn)
          s1 = szm*sy
          s2 =-(half+ twothird*sy)*sy
          t3(1:nc,md) = t3(1:nc,md)                                            &
                 - s1*(t6(1:nc,mn)-t6(1:nc,md)) - s2*(t5(1:nc,mn)-t5(1:nc,md))
        endif
    
        if (v(md) > zero) then
          ms = m(kp,i+1,j)
          sy = dtddy*v(md)
          s1 = szm*sy
          s2 = (half- twothird*sy)*sy
          t3(1:nc,md) = t3(1:nc,md)                                            &
                 - s1*(t6(1:nc,md)-t6(1:nc,ms)) - s2*(t5(1:nc,md)-t5(1:nc,ms))
        endif
    
      endif

      mi0 = m(2,i,j) - 2
      do k=2,kbm1
        mi = mi0 + k
        kp = k + 1
        md = mi + 1

        if (w(md) <= zero) then
          sz  = w(md)*hdt/h_new(mi)
          szp = -(half+fourthird*sz)
          fac = -half-sz
          t3(1:nc,md) = fac*t6(1:nc,mi)

          mw = m(k,i,j-1)
          if (u(mw) > zero) then
            sx = dtddx*u(mw)
            s1 = szp*sx
            s2 = (half- twothird*sx)*sx
            t3(1:nc,md) = t3(1:nc,md)                                          &
                   - s1*(t6(1:nc,mi)-t6(1:nc,mw)) - s2*(t4(1:nc,mi)-t4(1:nc,mw))
          endif
    
          if (u(mi) < zero) then
            me = m(k,i,j+1)
            sx = dtddx*u(mi)
            s1 = szp*sx
            s2 =-(half+ twothird*sx)*sx
            t3(1:nc,md) = t3(1:nc,md)                                          &
                   - s1*(t6(1:nc,me)-t6(1:nc,mi)) - s2*(t4(1:nc,me)-t4(1:nc,mi))
          endif
    
          mn = m(k,i-1,j)
          if (v(mn) < zero) then
            sy = dtddy*v(mn)
            s1 = szp*sy
            s2 =-(half+ twothird*sy)*sy
            t3(1:nc,md) = t3(1:nc,md)                                          &
                   - s1*(t6(1:nc,mn)-t6(1:nc,mi)) - s2*(t5(1:nc,mn)-t5(1:nc,mi))
          endif
    
          if (v(mi) > zero) then
            ms = m(k,i+1,j)
            sy = dtddy*v(mi)
            s1 = szp*sy
            s2 = (half- twothird*sy)*sy
            t3(1:nc,md) = t3(1:nc,md)                                          &
                   - s1*(t6(1:nc,mi)-t6(1:nc,ms)) - s2*(t5(1:nc,mi)-t5(1:nc,ms))
          endif
    
        else ! w > 0.0
          sz  = w(md)*hdt/h_new(mi)
          szm = (half-fourthird*sz)
          fac = half-sz
          t3(1:nc,md) = fac*t6(1:nc,md)
    
          mw = m(kp,i,j-1)
          if (u(mw) > zero) then
            sx = dtddx*u(mw)
            s1 = szm*sx
            s2 = (half- twothird*sx)*sx
            t3(1:nc,md) = t3(1:nc,md)                                          &
                   - s1*(t6(1:nc,md)-t6(1:nc,mw)) - s2*(t4(1:nc,md)-t4(1:nc,mw))
          endif
    
          if (u(md) < zero) then
            me = m(kp,i,j+1)
            sx = dtddx*u(md)
            s1 = szm*sx
            s2 =-(half+ twothird*sx)*sx
            t3(1:nc,md) = t3(1:nc,md)                                          &
                   - s1*(t6(1:nc,me)-t6(1:nc,md)) - s2*(t4(1:nc,me)-t4(1:nc,md))
          endif
    
          mn = m(kp,i-1,j)
          if (v(mn) < zero) then
            sy = dtddy*v(mn)
            s1 = szm*sy
            s2 =-(half+ twothird*sy)*sy
            t3(1:nc,md) = t3(1:nc,md)                                          &
                   - s1*(t6(1:nc,mn)-t6(1:nc,md)) - s2*(t5(1:nc,mn)-t5(1:nc,md))
          endif
    
          if (v(md) > zero) then
            ms = m(kp,i+1,j)
            sy = dtddy*v(md)
            s1 = szm*sy
            s2 = (half- twothird*sy)*sy
            t3(1:nc,md) = t3(1:nc,md)                                          &
                   - s1*(t6(1:nc,md)-t6(1:nc,ms)) - s2*(t5(1:nc,md)-t5(1:nc,ms))
          endif
    
        endif
      enddo
    enddo
  end subroutine c_dt
    
  !  ---------------------------------------------------------------------------
    
  subroutine diffusion_hx(m,ind,n2d,kh,hn,hx,hy,t,eddyh,cosphi,idx,uvdam)
    ! NOTE: it uses the global module variable: tt
    ! NOTE: assumes eddyh and tt are zero on land (these were initialized to 0).

    !- modules -----------------------------------------------------------------
    use dmi_omp,   only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: n2d
    integer(4), intent(in)    :: m(:,0:,0:), ind(:,:), kh(0:)
    real(8),    intent(in)    :: hn(0:), hx(0:), hy(0:)
    real(8),    intent(in)    :: cosphi(:,0:), eddyh(0:), uvdam(:,0:)
    real(8),    intent(inout) :: t(:,0:)
    integer(4), intent(inout) :: idx(1:,1:)

    !  simple vars:
    integer(4) :: n, k, kb, i, j, mi, me, mn, mw, ms
    integer(4) :: n2dl, n2du, mi0, me0, mw0, ms0, mn0
    real(8)    :: ahx, ahy, dh, dp, fx0, fy0, fx, fy, dahe, dahw, dahn, dahs
    real(8)    :: dtl2, ede, edw, eds, edn, lf, ude, udw, vdn, vds

    real(8), parameter :: p2 = 0.2_8, p2s = p2*p2

    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)

    !- Horizontal diffusion ----------------------------------------------------
    !   The effective horizontal eddy diffusivity, D_effective, is obtained from
    !   the horizontal eddy viscosity, Eh, by a relaxation according to:
    !       D_effective = 0.25*Eh/(0.1 + Eh*dt/L**2)
    !   where L is the length scale applied in the Smagorinsky model,
    !       L**2 = Csmag**2 * dx * dy,    Csmag = 0.2
    !   This yields a dimensionless diffusivity with maximum value approx 
    !       0.25 * Csmag**2 = 0.01
    !
    !   FIXME: Consider to make the magic numbers 0.1, 0.25 and 0.2 configurable
    !          i.e. through the argument list. At least, 0.2 should be exactly
    !          the same as the Csmag value use elsewhere in model.
    ! 

    !- some constants:
    dtl2 = dt/(p2s*dx*dy)
    ahx  = qrt*dt/(dx*dx)
    ahy  = qrt*dt/(dy*dy)

    ! k=1 un-rolled:
    do n=n2dl,n2du
      kb = kh(n)
      if (kb < 1) cycle
      i = ind(1,n)
      j = ind(2,n)

      dp  = one/cosphi(1,i)
      fx0 = ahx*dp*dp
      fy0 = ahy*dp
      lf  = dtl2*dp

      mi = n
      mn = m(1,i-1,j  )
      mw = m(1,i,  j-1)
      me = m(1,i,  j+1)
      ms = m(1,i+1,j  )

      ude = uvdam(1,mi)
      udw = uvdam(1,mw)
      vdn = uvdam(2,mn)
      vds = uvdam(2,mi)

      dh   = one/hn(mi)

      fx   = fx0*dh
      ede  = (eddyh(me)+eddyh(mi))*ude
      dahe = fx*hx(mi)*ede/(p2 + lf*ede)
      edw  = (eddyh(mw)+eddyh(mi))*udw
      dahw = fx*hx(mw)*edw/(p2 + lf*edw)

      fy   = fy0*dh
      eds  = (eddyh(ms)+eddyh(mi))*vds
      dahs = fy*hy(mi)*cosphi(2,i  )*eds/(p2 + lf*eds)
      edn  = (eddyh(mn)+eddyh(mi))*vdn
      dahn = fy*hy(mn)*cosphi(2,i-1)*edn/(p2 + lf*edn)

      t(1:nc,mi) = t(1:nc,mi)                                                  &
              + dahe*(tt(1:nc,me)-tt(1:nc,mi)) - dahw*(tt(1:nc,mi)-tt(1:nc,mw))&
              + dahn*(tt(1:nc,mn)-tt(1:nc,mi)) - dahs*(tt(1:nc,mi)-tt(1:nc,ms)) 
    enddo

    do n=n2dl,n2du
      kb = kh(n)
      if (kb <= 1) cycle
      i = ind(1,n)
      j = ind(2,n)

      dp  = one/cosphi(1,i)
      fx0 = ahx*dp*dp
      fy0 = ahy*dp
      lf  = dtl2*dp

      ude = uvdam(1,n)
      udw = uvdam(1,m(1,i,j-1))
      vdn = uvdam(2,m(1,i-1,j))
      vds = uvdam(2,n)

      mi0 = m(2,i,j) - 2
      if (ude > zero) then
        me0 = m(2,i,j+1) - 2
        do k=2,min(kb,kh(m(1,i,j+1)))
          mi = mi0 + k
          me = me0 + k
          ede  = eddyh(me)+eddyh(mi)
          dahe = fx0*hx(mi)*ede/(p2 + lf*ede)/hn(mi)
          t(1:nc,mi) = t(1:nc,mi) + dahe*(tt(1:nc,me)-tt(1:nc,mi))
        enddo
      endif
      if (udw > zero) then
        mw0 = m(2,i,j-1) - 2
        do k=2,min(kb,kh(m(1,i,j-1)))
          mi = mi0 + k
          mw = mw0 + k
          edw  = eddyh(mw)+eddyh(mi)
          dahw = fx0*hx(mw)*edw/(p2 + lf*edw)/hn(mi)
          t(1:nc,mi) = t(1:nc,mi) - dahw*(tt(1:nc,mi)-tt(1:nc,mw))
        enddo
      endif
      if (vds > zero) then
        ms0 = m(2,i+1,j) - 2
        fy = fy0*cosphi(2,i)
        do k=2,min(kb,kh(m(1,i+1,j)))
          mi = mi0 + k
          ms = ms0 + k
          eds  = eddyh(ms)+eddyh(mi)
          dahs = fy*hy(mi)*eds/(p2 + lf*eds)/hn(mi)
          t(1:nc,mi) = t(1:nc,mi) - dahs*(tt(1:nc,mi)-tt(1:nc,ms))
        enddo
      endif
      if (vdn > zero) then
        mn0 = m(2,i-1,j) - 2
        fy = fy0*cosphi(2,i-1)
        do k=2,min(kb,kh(m(1,i-1,j)))
          mi = mi0 + k
          mn = mn0 + k
          edn  = eddyh(mn)+eddyh(mi)
          dahn = fy*hy(mn)*edn/(p2 + lf*edn)/hn(mi)
          t(1:nc,mi) = t(1:nc,mi) + dahn*(tt(1:nc,mn)-tt(1:nc,mi))
        enddo
      endif
    enddo
    !- Horizontal diffusion end ------------------------------------------------

  end subroutine diffusion_hx

  ! ----------------------------------------------------------------------------

  subroutine diffusion_vi(kmax,m,ind,n2d,kh,h,hn,t,avt,idx,q)
    !  inlined version over nc components

    !- modules -----------------------------------------------------------------
    use constants, only : cpw, rhow, hnull
    use dmi_omp,   only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)           :: kmax,n2d
    integer(4), intent(in)           :: m(:,0:,0:), ind(:,:), kh(0:)
    real(8),    intent(in)           :: h(0:), avt(:,0:), hn(0:)
    real(8),    intent(in), optional :: q(0:)
    real(8),    intent(inout)        :: t(:,0:)
    integer(4), intent(inout)        :: idx(1:,1:)

    !  automatic arrays:
    real(8) :: dtiavt(nc,kmax+1), hm(kmax+1)
    real(8) :: qa(nc,kmax+1),qb(nc,kmax+1),qc(nc,kmax+1),cn(nc)
    real(8) :: pgka(nc,kmax+1),pgkb(nc,kmax+1)
    
    !  simple vars:
    integer(4) :: n,k,kb,i,j,no,nu,kbp1,kp1,kbm2,ki
    integer(4) :: n2dl, n2du
    real(8)    :: dti2, fac, qfac

    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)

    !- some preparations:
    dti2 = two*dt
    if (present(q)) then
      qfac = dt/(cpw*rhow)
      t(1,n2dl:n2du) = t(1,n2dl:n2du) + qfac*q(n2dl:n2du)/h(n2dl:n2du)
    endif

    !- Vertical diffusion ------------------------------------------------------
    do n=n2dl,n2du
      if (h(n) < hnull) cycle

      kb=kh(n)
      if (kb <= 1) cycle
      i=ind(1,n)
      j=ind(2,n)
    
      !- Implicit part:
      hm(1) = hn(n)
      dtiavt(1:nc,1) = zero

      no = m(2,i,j) - 2

      if (nc > 2) then  !Expression in IF construct is constant when nc is.
        do k=2,kb
          nu = no + k
          hm(k) = hn(nu)
          fac   = dti2/(hm(k-1)+hm(k))
          dtiavt(1:2,k) = fac*avt(1:2,nu)
          dtiavt(3:nc,k) = fac*avt(3,nu)
        enddo
      else
        do k=2,kb
          nu = no + k
          hm(k) = hn(nu)
          fac   = dti2/(hm(k-1)+hm(k))
          dtiavt(1:2,k) = fac*avt(1:2,nu)
        enddo
      endif
      kbp1=kb+1
      dtiavt(1:nc,kbp1) = zero
      do k=1,kb
        kp1=k+1
        qa(1:nc,k) = dtiavt(1:nc,k  )/hm(k)
        qc(1:nc,k) = dtiavt(1:nc,kp1)/hm(k)
        qb(1:nc,k) = one + qa(1:nc,k) + qc(1:nc,k)
      enddo
      pgka(1:nc,1)=zero
      pgkb(1:nc,1)=zero
      k=1
      kp1=k+1
      cn(1:nc) = one/(qb(1:nc,k)-qa(1:nc,k)*pgka(1:nc,k))
      pgka(1:nc,kp1) = qc(1:nc,k)*cn(1:nc)
      pgkb(1:nc,kp1) = (qa(1:nc,k)*pgkb(1:nc,k)+t(1:nc,n))*cn(1:nc)

      do k=2,kb
        kp1=k+1
        cn(1:nc) = one/(qb(1:nc,k)-qa(1:nc,k)*pgka(1:nc,k))
        pgka(1:nc,kp1) = qc(1:nc,k)*cn(1:nc)
        pgkb(1:nc,kp1) = (qa(1:nc,k)*pgkb(1:nc,k)+t(1:nc,no+k))*cn(1:nc)
      enddo

      t(1:nc,no+kb) = pgkb(1:nc,kbp1)
      kbm2=kb-2
      do ki=1,kbm2
        k   = kb-ki
        kp1 = k+1
        t(1:nc,no+k) = pgka(1:nc,kp1)*t(1:nc,no+kp1)+ pgkb(1:nc,kp1)
      enddo
      k   = 1
      kp1 = k+1
      t(1:nc,n) = pgka(1:nc,kp1)*t(1:nc,no+kp1)+ pgkb(1:nc,kp1)
    enddo
    
    !- Vertical diffusion ended ------------------------------------------------

  end subroutine diffusion_vi

  ! ----------------------------------------------------------------------------

  subroutine diffusion_vx(kmax,m,ind,n2d,kh,h,hn,t,avt,idx,q)
    ! Experimenting with different tri-diagonal solvers.
    !   - there is one call of the solver for each tracer.
    !     Note that S and T has a diffusion coefficient of their own,
    !     while passive tracers can share the matrix. 
    !
    ! Also, experimenting with Neumann boundary conditions
    !   - so far only for the temperature tracer in which case the surface flux
    !     is determined by the surface heat flux q which was determinded in
    !     thermodyn().
    !     q is an optional input arg and if present we will use it for the BC.
    !

    !- modules -----------------------------------------------------------------
    use constants, only : cpw, rhow, hnull
    use dmi_omp,   only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)           :: kmax,n2d
    integer(4), intent(in)           :: m(:,0:,0:), ind(:,:), kh(0:)
    real(8),    intent(in)           :: h(0:), avt(:,0:), hn(0:)
    real(8),    intent(in), optional :: q(0:)
    real(8),    intent(inout)        :: t(:,0:)
    integer(4), intent(inout)        :: idx(1:,1:)

    !  automatic arrays:
    real(8) :: dtiavt(3,kmax+1), hm(kmax)
    real(8) :: a(kmax),b(kmax),c(kmax),d(kmax)
!   real(8) :: s(kmax)   ! for TriSolv
    
    !  simple vars:
    integer(4) :: n,k,kb,i,j,no,nu 
    integer(4) :: n2dl, n2du, ic
!    integer(4) :: info ! for ddtsvb
    real(8)    :: dti2, fac, qfac

    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)

    !- some constants:
    dti2 = two*dt
    if (present(q)) then
      qfac = dt/(cpw*rhow)
      t(1,n2dl:n2du) = t(1,n2dl:n2du) + qfac*q(n2dl:n2du)/h(n2dl:n2du)
    endif

    !- Vertical diffusion ------------------------------------------------------
    do n=n2dl,n2du
      if (h(n) < hnull) cycle

      kb=kh(n)
      if (kb <= 1) cycle
      i=ind(1,n)
      j=ind(2,n)
    
      !- Implicit part:
      hm(1) = hn(n)
      dtiavt(1:2,1) = zero
      dtiavt(3  ,1) = zero

      no = m(2,i,j) - 2

      do k=2,kb
        nu = no + k
        hm(k) = hn(nu)
        fac   = dti2/(hm(k-1)+hm(k))
        dtiavt(1:2,k) = fac*avt(1:2,nu)
        if (nc > 2) dtiavt(3,k) = fac*avt(3,nu) !Expression in logical IF is
                                                !constant when nc is constant.
      enddo
      dtiavt(1:2,kb+1) = zero
      dtiavt(3  ,kb+1) = zero

      ! S&T tracers:
      do ic=1,2
        k = 1 
        a(k) = -dtiavt(ic,k  )/hm(k)
        c(k) = -dtiavt(ic,k+1)/hm(k)
        b(k) = one - a(k) - c(k)
        d(k) = t(ic,n)
        do k=2,kb
          a(k) = -dtiavt(ic,k  )/hm(k)
          c(k) = -dtiavt(ic,k+1)/hm(k)
          b(k) = one - a(k) - c(k)
          d(k) = t(ic,no + k)
        enddo

! TriSolv interface:
!       call TriSolv(s, a, b, c, d, kb)
!       t(ic,n) = s(1)
!       t(ic,no+2:no+kb) = s(2:kb)

! or ddtsvb interface:
!       do k=1,kb-1
!         a(k) = a(k+1)
!       enddo
!       call ddtsvb (kb, 1, a, b , c , d, kb, info)
!       t(ic,n) = d(1)
!       t(ic,no+2:no+kb) = d(2:kb)
      enddo

      ! passive tracers:
      do ic=3,nc   !Loop may has zero iteration count for nc=2
        k = 1 
        a(k) = -dtiavt(3,k  )/hm(k)
        c(k) = -dtiavt(3,k+1)/hm(k)
        b(k) = one - a(k) - c(k)
        d(k) = t(ic,n)
        do k=2,kb
          a(k) = -dtiavt(3,k  )/hm(k)
          c(k) = -dtiavt(3,k+1)/hm(k)
          b(k) = one - a(k) - c(k)
          d(k) = t(ic,no + k)
        enddo

!PLEASE NOTE: The above will unavoidably give compiler warnings like
!             "Variable B set but never referenced!
!             "Variable D set but never referenced!
!             untill one of the codes below are activated.

! TriSolv interface:
!       call TriSolv(s, a, b, c, d, kb)
!       t(ic,n) = s(1)
!       t(ic,no+2:no+kb) = s(2:kb)

! or ddtsvb interface:
!       do k=1,kb-1
!         a(k) = a(k+1)
!       enddo
!       call ddtsvb (kb, 1, a, b , c , d, kb, info)
!       t(ic,n) = d(1)
!       t(ic,no+2:no+kb) = d(2:kb)
      enddo
      
    enddo
    
    !- Vertical diffusion ended ------------------------------------------------

  end subroutine diffusion_vx

  ! ----------------------------------------------------------------------------

  subroutine tflow_alloc (n3d,ntracers,ierr) 
    use exits, only : exitme

    implicit none

    integer(4), intent(in)    :: n3d
    integer(4), intent(in)    :: ntracers  ! Unused dummy variable NTRACERS
                                           ! untill activated below.
    integer(4), intent(inout) :: ierr

    if (n3d > n3dmax) then
      n3dmax = n3d  
    endif

!    if (nc==0) nc = ntracers ! MUST BE ACTIVE IF BUILDING WITH NC AS VARIABLE
    if (nc /= ntracers) then
      ! trap
      call exitme ('ERROR: Invalid number of components in advection/diffusion')
    endif

    allocate (tt(1:nc,0:n3d), t1(1:nc,0:n3d), t2(1:nc,0:n3d), t3(1:nc,0:n3d),  &
              t4(1:nc,0:n3d), t5(1:nc,0:n3d), t6(1:nc,0:n3d),                  &
              t8(1:nc,0:n3d), t7(1:nc,0:n3d), stat=ierr                       )
    if (ierr /= 0) return
  end subroutine tflow_alloc

  ! ----------------------------------------------------------------------------

  subroutine tflow_ft (m, n2d, ind, kh, idx, n3d)
    ! Let the domain with the largest Ir determine the numa first-touch
    ! of tflow tmp-arrays in the wet points of that domain on the present task.
    ! The remaining points (i.e. land point, halo points, points in the largest
    ! domain) of the arrays initialised with zeroes from the MASTER thread.

    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: n2d, n3d, m(:,0:,0:), ind(:,:), kh(0:)
    integer(4), intent(inout) :: idx(1:,1:)

    !- local vars --------------------------------------------------------------
    integer(4) :: n, n2dl, n2du, kb, ml, mu

    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)

    ! NUMA first-touch in surface:
    tt(1:nc,n2dl:n2du) = zero 
    t1(1:nc,n2dl:n2du) = zero 
    t2(1:nc,n2dl:n2du) = zero 
    t3(1:nc,n2dl:n2du) = zero 
    t4(1:nc,n2dl:n2du) = zero 
    t5(1:nc,n2dl:n2du) = zero 
    t6(1:nc,n2dl:n2du) = zero 
    t7(1:nc,n2dl:n2du) = zero 
    t8(1:nc,n2dl:n2du) = zero 

    ! NUMA first-touch in sub-surface:
    do n=n2dl,n2du
      kb = kh(n)
      if (kb <= 1) cycle

      ml = m(2,ind(1,n),ind(2,n))
      mu = ml + kb -2

      tt(1:nc,ml:mu) = zero 
      t1(1:nc,ml:mu) = zero 
      t2(1:nc,ml:mu) = zero 
      t3(1:nc,ml:mu) = zero 
      t4(1:nc,ml:mu) = zero 
      t5(1:nc,ml:mu) = zero 
      t6(1:nc,ml:mu) = zero 
      t7(1:nc,ml:mu) = zero 
      t8(1:nc,ml:mu) = zero 
    enddo
      
    ! make sure to have nice initial values in land point, in 
    ! and possibly the halo:
!$OMP MASTER
    tt(1:nc,0) = zero
    t1(1:nc,0) = zero
    t2(1:nc,0) = zero
    t3(1:nc,0) = zero
    t4(1:nc,0) = zero
    t5(1:nc,0) = zero
    t6(1:nc,0) = zero
    t7(1:nc,0) = zero
    t8(1:nc,0) = zero

    if (n3d < n3dmax) then
      ml = n3d + 1
      mu = n3dmax

      tt(1:nc,ml:mu) = zero 
      t1(1:nc,ml:mu) = zero 
      t2(1:nc,ml:mu) = zero 
      t3(1:nc,ml:mu) = zero 
      t4(1:nc,ml:mu) = zero 
      t5(1:nc,ml:mu) = zero 
      t6(1:nc,ml:mu) = zero 
      t7(1:nc,ml:mu) = zero 
      t8(1:nc,ml:mu) = zero 
    endif
!$OMP END MASTER

  end subroutine tflow_ft

  ! ----------------------------------------------------------------------------
    
  subroutine c_delta (m,ind,n2d,kh,khu,khv,h,nbpz,krz,nbpu,kru,nbpv,krv,t,idx)

    ! NOTE: it modifies the global module variables: t4, t5 ,t6 

    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: n2d, nbpz, nbpu, nbpv
    integer(4), intent(in)    :: m(:,0:,0:), ind(:,:), khu(0:), khv(0:), kh(0:)
    integer(4), intent(in)    :: krz(:,0:), kru(:,0:), krv(:,0:)
    real(8),    intent(in)    :: h(0:), t(:,0:)
    integer(4), intent(inout) :: idx(1:,1:)

    !  simple vars:
    integer(4) :: n, ntyp, k, i, j, mi, mm, mp, nbz, nbu, nbv
    integer(4) :: mi0, mm0, mp0, kmmax, kpmax, kb, kbuw, kbue, kbvn, kbvs
    integer(4) :: n2dl, n2du, nbpzl, nbpzu, nbpul, nbpuu, nbpvl, nbpvu
    real(8)    :: fac, qqa, qqb


    ! obtain domains for the n-loops -------------------------------------------
    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)


    ! compute t4  --------------------------------------------------------------
    ! k=1 unrolled
    do n = n2dl, n2du
      kb = kh(n)
      i  = ind(1,n)
      j  = ind(2,n)
      mm = m(1,i,j-1)
      mp = m(1,i,j+1)
      if (mp > 0 .and. mm > 0) then
        t4(1:nc,n) = half*(t(1:nc,mp)-t(1:nc,mm))
      elseif (mp > 0) then
        t4(1:nc,n) = t(1:nc,mp)-t(1:nc,n)
      elseif (mm > 0) then
        t4(1:nc,n) = t(1:nc,n)-t(1:nc,mm)
      else
        t4(1:nc,n) = zero
      endif
    enddo
    do n = n2dl, n2du
      kb = kh(n)
      if (kb < 2) cycle
      i = ind(1,n)
      j = ind(2,n)

      mi0 = m(2,i,j) - 2
      if (kh(m(1,i,j-1)) >= 2) then
        kmmax = kh(m(1,i,j-1))
        mm0   = m(2,i,j-1) - 2
      else
        kmmax = 0
      endif
      if (kh(m(1,i,j+1)) >= 2) then
        kpmax = kh(m(1,i,j+1))
        mp0   = m(2,i,j+1) - 2
      else
        kpmax = 0
      endif
      do k=2,min(kb,kpmax,kmmax)
        t4(1:nc,mi0+k) = half*(t(1:nc,mp0+k) - t(1:nc,mm0+k))
      enddo
      if     (kpmax > kmmax) then
        do k=max(2,kmmax+1),min(kb,kpmax)
          mi = mi0 + k
          t4(1:nc,mi) = t(1:nc,mp0+k)-t(1:nc,mi)
        enddo
      elseif (kmmax > kpmax) then
        do k=max(2,kpmax+1),min(kb,kmmax)
          mi = mi0 + k
          t4(1:nc,mi) = t(1:nc,mi)-t(1:nc,mm0+k)
        enddo
      endif
      do k=max(2,max(kpmax,kmmax)+1),kb
        t4(1:nc,mi0+k) = zero
      enddo
    enddo
    

    ! compute t5  --------------------------------------------------------------
    ! k=1 unrolled
    do n=n2dl, n2du
      kb = kh(n)
      i  = ind(1,n)
      j  = ind(2,n)
      mm = m(1,i-1,j)
      mp = m(1,i+1,j)
      if (mp > 0 .and. mm > 0) then
        t5(1:nc,n) = half*(t(1:nc,mm)-t(1:nc,mp))
      elseif (mm > 0) then
        t5(1:nc,n) = t(1:nc,mm)-t(1:nc,n)
      elseif (mp > 0) then
        t5(1:nc,n) = t(1:nc,n)-t(1:nc,mp)
      else
        t5(1:nc,n) = zero
      endif
    enddo
    do n = n2dl, n2du
      kb = kh(n)
      if (kb < 2) cycle
      i = ind(1,n)
      j = ind(2,n)

      mi0 = m(2,i,j) - 2
      if (kh(m(1,i-1,j)) >= 2) then
        kmmax = kh(m(1,i-1,j))
        mm0   = m(2,i-1,j) - 2
      else
        kmmax = 0
      endif
      if (kh(m(1,i+1,j)) >= 2) then
        kpmax = kh(m(1,i+1,j))
        mp0   = m(2,i+1,j) - 2
      else
        kpmax = 0
      endif
      do k=2,min(kb,kpmax,kmmax)
        t5(1:nc,mi0+k) = half*(t(1:nc,mm0+k) - t(1:nc,mp0+k))
      enddo
      if     (kpmax > kmmax) then
        do k=max(2,kmmax+1),min(kb,kpmax)
          mi = mi0 + k
          t5(1:nc,mi) = t(1:nc,mi)-t(1:nc,mp0+k)
        enddo
      elseif (kmmax > kpmax) then
        do k=max(2,kpmax+1),min(kb,kmmax)
          mi = mi0 + k
          t5(1:nc,mi) = t(1:nc,mm0+k)-t(1:nc,mi)
        enddo
      endif
      do k=max(2,max(kpmax,kmmax)+1),kb
        t5(1:nc,mi0+k) = zero
      enddo
    enddo
    

    ! compute t6  --------------------------------------------------------------
    do n=n2dl, n2du
      kb = kh(n)
      i = ind(1,n)
      j = ind(2,n)

      ! un-roll k=1:
      k = 1
      if (kb == 1) then
        t6(1:nc,n) = zero
        cycle
      endif
      mp = m(k+1,i,j)
      t6(1:nc,n) = t(1:nc,n)-t(1:nc,mp)

      ! loop except top and bottom:
      mm = n
      do mi=m(2,i,j),m(2,i,j)+kb-3
        mp = mi+1

        fac = one/(h(mm)+two*h(mi)+h(mp))
        qqa = (h(mm)+h(mi))*fac
        qqb = (h(mi)+h(mp))*fac

        t6(1:nc,mi) = (one-qqa)*(t(1:nc,mm)-t(1:nc,mi))                        &
                    + (one-qqb)*(t(1:nc,mi)-t(1:nc,mp))
        mm = mi
      enddo

      ! un-roll k=kb:
      k = kb 
      mi = m(k,  i,j)
      mm = m(k-1,i,j)
      t6(1:nc,mi) = t(1:nc,mm)-t(1:nc,mi)
    enddo

    
    ! compute boundary contributions to t4  and t5   ---------------------------
    ! - do this as the last actions in this routine to gather the barriers
    !   and push these as far as posible.
    if (nbpz > 0 .or. nbpu > 0 .or. nbpv > 0) then
!$OMP BARRIER

! fixme - consider fusion of the three decompos to get better balancing:
      if (nbpz > 0) then
        call domp_get_domain(1, nbpz, nbpzl, nbpzu)
        do nbz=nbpzl,nbpzu
          i = krz(1,nbz)
          j = krz(2,nbz)
          ntyp = abs(krz(3,nbz))
          if     (ntyp == 3) then
            kbuw = khu(m(1,i,j-1))
            if (kbuw > 0) then
              mi = m(1,i,j)
              t4(1:nc,mi) = t(1:nc,mi)-t(1:nc,m(1,i,j-1))
              do k=2,kbuw
                mi = m(2,i,j) + k - 2
                t4(1:nc,mi) = t(1:nc,mi)-t(1:nc,m(2,i,j-1)+k-2)
              enddo
            endif
          elseif (ntyp == 1) then
            kbue = khu(m(1,i, j))
            if (kbue > 0) then
              mi = m(1,i,j)
              t4(1:nc,mi) = t(1:nc,m(1,i,j+1))-t(1:nc,mi)
              do k=2,kbue
                mi = m(2,i,j) + k - 2
                t4(1:nc,mi) = t(1:nc,m(2,i,j+1)+k-2)-t(1:nc,mi)
              enddo
            endif
          elseif (ntyp == 4) then
            kbvn = khv(m(1,i-1,j))
            if (kbvn > 0) then
              mi = m(1,i,j)
              t5(1:nc,mi) = t(1:nc,m(1,i-1,j))-t(1:nc,mi)
              do k=2,kbvn
                mi = m(2,i,j) + k - 2
                t5(1:nc,mi) = t(1:nc,m(2,i-1,j)+k-2)-t(1:nc,mi)
              enddo
            endif
          elseif (ntyp == 2) then
            kbvs = khv(m(1,i, j))
            if (kbvs > 0) then
              mi = m(1,i,j)
              t5(1:nc,mi) = t(1:nc,mi)-t(1:nc,m(1,i+1,j))
              do k=2,kbvs
                mi = m(2,i,j) + k - 2
                t5(1:nc,mi) = t(1:nc,mi)-t(1:nc,m(2,i+1,j)+k-2)
              enddo
            endif
          endif
        enddo
      endif
    
      if (nbpu > 0) then
        call domp_get_domain(1, nbpu, nbpul, nbpuu)
        do nbu=nbpul,nbpuu
          i = kru(1,nbu)
          j = kru(2,nbu)
          ntyp = kru(3,nbu)
          if     (ntyp == 3) then
            kbuw = khu(m(1,i,j-1))
            if (kbuw > 0) then
              mi = m(1,i,j)
              t4(1:nc,mi) = t(1:nc,mi)-t(1:nc,m(1,i,j-1))
              do k=2,kbuw
                mi = m(2,i,j) + k - 2
                t4(1:nc,mi) = t(1:nc,mi)-t(1:nc,m(2,i,j-1)+k-2)
              enddo
            endif
          elseif (ntyp == 1) then
            kbue = khu(m(1,i,j+1))
            if (kbue > 0) then
              mp = m(1,i,j+1)
              t4(1:nc,mp) = t(1:nc,m(1,i,j+2))-t(1:nc,mp)
              do k=2,kbue
                mp = m(2,i,j+1) + k - 2
                t4(1:nc,mp) = t(1:nc,m(2,i,j+2)+k-2)-t(1:nc,mp)
              enddo
            endif
          endif
        enddo
      endif

      if (nbpv > 0) then
        call domp_get_domain(1, nbpv, nbpvl, nbpvu)
        do nbv=nbpvl,nbpvu
          i = krv(1,nbv)
          j = krv(2,nbv)
          ntyp = krv(3,nbv)
          if     (ntyp == 4) then
            kbvn = khv(m(1,i-1,j))
            if (kbvn > 0) then
              mi = m(1,i,j)
              t5(1:nc,mi) = t(1:nc,m(1,i-1,j))-t(1:nc,mi)
              do k=2,kbvn
                mi = m(2,i,j) + k - 2
                t5(1:nc,mi) = t(1:nc,m(2,i-1,j)+k-2)-t(1:nc,mi)
              enddo
            endif
          elseif (ntyp == 2) then
            kbvs = khv(m(1,i+1,j))
            if (kbvs > 0) then
              mp = m(1,i+1,j)
              t5(1:nc,mp) = t(1:nc,mp)-t(1:nc,m(1,i+2,j))
              do k=2,kbvs
                mp = m(2,i+1,j) + k - 2
                t5(1:nc,mp) = t(1:nc,mp)-t(1:nc,m(2,i+2,j)+k-2)
              enddo
            endif
          endif
        enddo
      endif
    endif

  end subroutine c_delta

  ! ----------------------------------------------------------------------------
    
  subroutine c_tt (kmax,m,ind,n2d,kh,khu,khv,hx,hy,h_old,h_new,                &
                   u,v,w,t,nbpz,krz,nbpq,krq,rwqk,rwzkt,rwqkt,cosphi,idx)
    
    ! NOTE: it modifies the global module variables: tt and uses t1, t2, t3

    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain
    
    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: kmax, n2d
    integer(4), intent(in)    :: m(:,0:,0:), ind(:,:), kh(0:), khu(0:), khv(0:)
    integer(4), intent(in)    :: nbpz, nbpq
    integer(4), intent(in)    :: krq(:,0:), krz(:,0:)
    real(8),    intent(in)    :: hx(0:), hy(0:), h_old(0:), h_new(0:)
    real(8),    intent(in)    :: u(0:), v(0:), w(0:), t(:,0:)
    real(8),    intent(in)    :: rwqk(0:,:), rwzkt(:,:,0:)
    real(8),    intent(in)    :: rwqkt(:,:,0:), cosphi(:,0:)
    integer(4), intent(inout) :: idx(1:,1:)
    
    !- local vars --------------------------------------------------------------
    !  automatic:
    real(8), dimension(kmax) :: rwzkin, rwzkout
    
    !  simple vars:
    integer(4) :: nuw, nue, nvn
    integer(4) :: kbuw, kbue, kbvn, kbvs
    integer(4) :: n, nvs, i, j, k, kb, mi, mn, mw, md, mi0, mn0, mw0
    integer(4) :: n2dl, n2du, nbpql, nbpqu, nbpzl, nbpzu
    real(8)    :: fac0, fac, hudy, hvdx, facx, facy1, facy2, factxy
    real(8)    :: uuu, vvv, wh
    real(8)    :: ws, ddx, ddy, ddt, hudy2, hvdx2, dtf0, dtdx1, dtdx2

    ! --------------------------------------------------------------------------
    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)
    
    do n=n2dl,n2du
      kb=kh(n)
      if (kb < 1) cycle
      i=ind(1,n)
      j=ind(2,n)
      fac0  = dxdy*cosphi(1,i)
      dtf0  = dt*fac0
      dtdx1 = dtdx*cosphi(2,i)
      dtdx2 = dtdx*cosphi(2,i-1)

      ! k=1 unrolled:
      mi = m(1,i,j)
      mw = m(1,i,j-1)
      mn = m(1,i-1,j)
      fac   = fac0*h_old(mi)
      hudy  = hx(mi)*u(mi)*dtdy
      hudy2 = hx(mw)*u(mw)*dtdy
      hvdx  = hy(mi)*v(mi)*dtdx1
      hvdx2 = hy(mn)*v(mn)*dtdx2

      tt(1:nc,mi) = t(1:nc,mi)*fac                                             &
                  -       (hudy *t1(1:nc,mi) - hudy2*t1(1:nc,mw))              &
                  -       (hvdx2*t2(1:nc,mn) - hvdx *t2(1:nc,mi))              &
                  -        dtf0*t3(1:nc,mi)
      if (kb > 1) then
        md = m(2,i,j)
        tt(1:nc,mi) = tt(1:nc,mi) + dtf0*t3(1:nc,md)

        mi0 = m(2,i,  j  ) - 2
        mn0 = m(2,i-1,j  ) - 2
        mw0 = m(2,i,  j-1) - 2
      else
        cycle
      endif

      do k=2,kb-1
        mi = mi0 + k
        md = mi + 1

        fac   = fac0*h_old(mi)
        hudy  = hx(mi)*u(mi)*dtdy
        hvdx  = hy(mi)*v(mi)*dtdx1

        tt(1:nc,mi) = t(1:nc,mi)*fac - hudy*t1(1:nc,mi) + hvdx*t2(1:nc,mi)    &
                    - dtf0*(t3(1:nc,mi) - t3(1:nc,md))
      enddo
      do k=2,min(kb-1,khu(m(1,i,j-1)))
        mi = mi0 + k
        mw = mw0 + k
        hudy2 = hx(mw)*u(mw)*dtdy
        tt(1:nc,mi) = tt(1:nc,mi) + hudy2*t1(1:nc,mw)
      enddo
      do k=2,min(kb-1,khv(m(1,i-1,j)))
        mi = mi0 + k
        mn = mn0 + k
        hvdx2 = hy(mn)*v(mn)*dtdx2
        tt(1:nc,mi) = tt(1:nc,mi) - hvdx2*t2(1:nc,mn)
      enddo

      ! k=kb unrolled
      mi = m(kb,i,  j  )
      mw = m(kb,i,  j-1)
      mn = m(kb,i-1,j  )
      fac   = fac0*h_old(mi)
      hudy  = hx(mi)*u(mi)*dtdy
      hudy2 = hx(mw)*u(mw)*dtdy
      hvdx  = hy(mi)*v(mi)*dtdx1
      hvdx2 = hy(mn)*v(mn)*dtdx2

      tt(1:nc,mi) = t(1:nc,mi)*fac                                             &
                  -       (hudy *t1(1:nc,mi) - hudy2*t1(1:nc,mw))              &
                  -       (hvdx2*t2(1:nc,mn) - hvdx *t2(1:nc,mi))              &
                  -        dtf0*t3(1:nc,mi)
      
    enddo
    

! fixme - consider fusion of the two decompos to get better balancing:
    if (nbpq > 0 .or. nbpz > 0) then
      ! due to barrier

      if (nbpq > 0) then
!$OMP BARRIER
        call domp_get_domain(1, nbpq, nbpql, nbpqu)
        do n=nbpql,nbpqu
          i   = krq(1,n)
          mi  = m(1,i,krq(2,n))
          fac = rwqk(n,1)*dtdxdy*cosphi(1,i)
          tt(1:nc,mi) = tt(1:nc,mi) + fac*rwqkt(1:nc,1,n)
        enddo
        do n=nbpql,nbpqu
          i = krq(1,n)
          j = krq(2,n)
          fac0 = dtdxdy*cosphi(1,i)
          do k=2,kh(m(1,i,j))
            fac = rwqk(n,k)*fac0
            mi  = m(2,i,j) + k - 2
            tt(1:nc,mi) = tt(1:nc,mi) + fac*rwqkt(1:nc,k,n)
          enddo
        enddo
      endif
    
      if (nbpz > 0) then
!$OMP BARRIER
        call domp_get_domain(1, nbpz, nbpzl, nbpzu)
        ddx = one/dx
        ddy = one/dy
        ddt = one/dt
        do n=nbpzl,nbpzu
          i = krz(1,n)
          j = krz(2,n)
          if (n>1 .and. i==krz(1,n-1) .and. j==krz(2,n-1)) cycle
          kb = kh(m(1,i,j))
          if (kb < 1) cycle
          kbuw = khu(m(1,i,j-1))
          kbue = khu(m(1,i, j ))
          kbvn = khv(m(1,i-1,j))
          kbvs = khv(m(1,i, j ))
          rwzkin (1:kb)=zero
          rwzkout(1:kb)=zero

          facx   = ddx/cosphi(1,i)
          facy1  = ddy*cosphi(2,i-1)/cosphi(1,i)
          facy2  = ddy*cosphi(2, i )/cosphi(1,i)
          factxy = dtdxdy*cosphi(1,i)

          ! k=1 unrolled:
          nuw = m(1,i,j-1)
          nue = m(1,i, j )
          nvn = m(1,i-1,j)
          nvs = m(1,i, j )
          if (1 > kbuw) nuw=0
          if (1 > kbue) nue=0
          if (1 > kbvn) nvn=0
          if (1 > kbvs) nvs=0
          fac = hx(nue)*facx
          if (abs(krz(3,n)) == 3 .and. u(nue) > zero) then
            !rwzkout(1,n)=rwzkout(1,n)
            rwzkin (1)=rwzkin (1) + fac*u(nue)
          else
            rwzkin (1)=rwzkin (1) + fac*min(u(nue),zero)
            rwzkout(1)=rwzkout(1) + fac*max(u(nue),zero)
          endif
          fac = hx(nuw)*facx
          if (abs(krz(3,n)) == 1 .and. u(nuw) < zero) then
            !rwzkout(1)=rwzkout(1)
            rwzkin (1)=rwzkin (1) - fac*u(nuw)
          else
            rwzkin (1)=rwzkin (1) - fac*max(u(nuw),zero)
            rwzkout(1)=rwzkout(1) - fac*min(u(nuw),zero)
          endif
          fac = hy(nvn)*facy1
          if (abs(krz(3,n)) == 2 .and. v(nvn) > zero) then
            !rwzkout(1)=rwzkout(1)
            rwzkin (1)=rwzkin (1) + fac*v(nvn)
          else
            rwzkin (1)=rwzkin (1) + fac*min(v(nvn),zero)
            rwzkout(1)=rwzkout(1) + fac*max(v(nvn),zero)
          endif
          fac = hy(nvs)*facy2
          if (abs(krz(3,n)) == 4 .and. v(nvs) < zero) then
            !rwzkout(1)=rwzkout(1)
            rwzkin (1)=rwzkin (1) - fac*v(nvs)
          else
            rwzkin (1)=rwzkin (1) - fac*max(v(nvs),zero)
            rwzkout(1)=rwzkout(1) - fac*min(v(nvs),zero)
          endif

          ! fixme: consider moving IF(abs(krz...) branches outside loop
          !        and merge with above IF branches
          do k=2,min(kb,kh(m(1,i,j+1)))
            nue = m(2,i,j) + k - 2
            fac = hx(nue)*facx
            uuu = u(nue)
            if (abs(krz(3,n)) == 3 .and. uuu > zero) then
              !rwzkout(k,n)=rwzkout(k,n)
              rwzkin (k)=rwzkin (k) + fac*uuu
            else
              rwzkin (k)=rwzkin (k) + fac*min(uuu,zero)
              rwzkout(k)=rwzkout(k) + fac*max(uuu,zero)
            endif
          enddo
          do k=2,min(kb,kh(m(1,i,j-1)))
            nuw = m(2,i,j-1) + k - 2
            fac = hx(nuw)*facx
            uuu = u(nuw)
            if (abs(krz(3,n)) == 1 .and. uuu < zero) then
              !rwzkout(k)=rwzkout(k)
              rwzkin (k)=rwzkin (k) - fac*uuu
            else
              rwzkin (k)=rwzkin (k) - fac*max(uuu,zero)
              rwzkout(k)=rwzkout(k) - fac*min(uuu,zero)
            endif
          enddo
          do k=2,min(kb,kh(m(1,i-1,j)))
            nvn = m(2,i-1,j) + k - 2
            fac = hy(nvn)*facy1
            vvv = v(nvn)
            if (abs(krz(3,n)) == 2 .and. vvv > zero) then
              !rwzkout(k)=rwzkout(k)
              rwzkin (k)=rwzkin (k) + fac*vvv
            else
              rwzkin (k)=rwzkin (k) + fac*min(vvv,zero)
              rwzkout(k)=rwzkout(k) + fac*max(vvv,zero)
            endif
          enddo
          do k=2,min(kb,kh(m(1,i+1,j)))
            nvs = m(2,i,j) + k - 2
            fac = hy(nvs)*facy2
            vvv = v(nvs)
            if (abs(krz(3,n)) == 4 .and. vvv < zero) then
              !rwzkout(k)=rwzkout(k)
              rwzkin (k)=rwzkin (k) - fac*vvv
            else
              rwzkin (k)=rwzkin (k) - fac*max(vvv,zero)
              rwzkout(k)=rwzkout(k) - fac*min(vvv,zero)
            endif
          enddo

          mi = m(1,i,j)
          wh = ddt*(h_new(mi)-h_old(mi))
          rwzkin (1) = rwzkin (1) + min(wh,zero)
          rwzkout(1) = rwzkout(1) + max(wh,zero)
          do k=2,kb
            mi = m(2,i,j) + k - 2
            wh = ddt*(h_new(mi)-h_old(mi))
            ws = w(mi)
            rwzkout(k) = rwzkout(k) + max(ws,zero) + max(wh,zero)
            rwzkin (k) = rwzkin (k) + min(ws,zero) + min(wh,zero)
          enddo
          do k=1,kb-1
            ws = w(m(2,i,j) + k - 1)
            rwzkin (k) = rwzkin (k) - max(ws,zero)
            rwzkout(k) = rwzkout(k) - min(ws,zero)
          enddo

          mi = m(1,i,j)
          tt(1:nc,mi) = tt(1:nc,mi)                                            &
                      + factxy*(  rwzkin (1)*t(1:nc,mi)                        &
                                + rwzkout(1)*rwzkt(1:nc,1,n) )
          do k=2,kb
            mi = m(2,i,j) + k - 2
            tt(1:nc,mi) = tt(1:nc,mi)                                          &
                        + factxy*(  rwzkin (k)*t(1:nc,mi)                      &
                                  + rwzkout(k)*rwzkt(1:nc,k,n) )
          enddo
        enddo
      endif

!$OMP BARRIER
    endif
    
    do n=n2dl,n2du
      tt(1:nc,n) = tt(1:nc,n)*ddxdy/cosphi(1,ind(1,n))/h_new(n)
    enddo
    do n=n2dl,n2du
      kb=kh(n)
      if (kb < 2) cycle
      i=ind(1,n)
      j=ind(2,n)
      fac0 = ddxdy/cosphi(1,i)
      do mi=m(2,i,j),m(2,i,j) + kb - 2
        tt(1:nc,mi) = tt(1:nc,mi)*fac0/h_new(mi)
      enddo
    enddo
    
  end subroutine c_tt

  ! ----------------------------------------------------------------------------
    
  subroutine c_tw (kmax,m,ind,n2d,kh,u,v,w,t,cosphi,idx)
      
    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: n2d, kmax, m(:,0:,0:), ind(:,:), kh(0:)
    real(8),    intent(in)    :: u(0:), v(0:), w(0:), t(:,0:), cosphi(:,0:)
    integer(4), intent(inout) :: idx(1:,1:)
    
    !- local vars --------------------------------------------------------------
    integer(4) :: n, i, j, k, kb, mi, nn, ss, ee, ww, nw, ne, sw, se
    integer(4) :: n2dl, n2du
    real(8)    :: dtddy, dtddx, facy, facx, mhdtddy, mhdtddx, dtddx0, wk
    real(8)    :: t3tmp(1:nc,1:kmax), wm(1:kmax)
    
    ! --------------------------------------------------------------------------
    
    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)

    dtddy   = dt/dy
    facy    = -onethird*dtddy
    mhdtddy = -half*dtddy
    dtddx0  = dt/dx

    do n=n2dl,n2du
      i  = ind(1,n)
      j  = ind(2,n)

      ! make sure to have nice values throughout
      t3(1:nc,n) = zero
      
      kb = kh(n)
      if (kb <= 1) cycle

      dtddx   = dtddx0/cosphi(1,i)
      facx    = -onethird*dtddx
      mhdtddx = -half*dtddx
      
      ! unroll k=1:
      mi = n
      nn = m(1,i-1,j  )
      ss = m(1,i+1,j  )
      ee = m(1,i,  j+1)
      ww = m(1,i,  j-1)
      nw = m(1,i-1,j-1)
      sw = m(1,i+1,j-1)
      se = m(1,i+1,j+1)
      ne = m(1,i-1,j+1)
    
      t3tmp(1:nc,1) = t(1:nc,mi)                                               &
        + mhdtddx*max(u(ww),zero)*(  t(1:nc,mi) - t(1:nc,ww)                   &
                          + facy*(  min(v(nn),zero)*(t(1:nc,nn)-t(1:nc,mi))    &
                                  + max(v(mi),zero)*(t(1:nc,mi)-t(1:nc,ss))    &
                                  - min(v(nw),zero)*(t(1:nc,nw)-t(1:nc,ww))    &
                                  - max(v(ww),zero)*(t(1:nc,ww)-t(1:nc,sw)) )) &
        + mhdtddx*min(u(mi),zero)*(  t(1:nc,ee) - t(1:nc,mi)                   &
                          + facy*(  min(v(ne),zero)*(t(1:nc,ne)-t(1:nc,ee))    &
                                  + max(v(ee),zero)*(t(1:nc,ee)-t(1:nc,se))    &
                                  - min(v(nn),zero)*(t(1:nc,nn)-t(1:nc,mi))    &
                                  - max(v(mi),zero)*(t(1:nc,mi)-t(1:nc,ss)) )) &
        + mhdtddy*min(v(nn),zero)*( t(1:nc,nn) - t(1:nc,mi)                    &
                          + facx*(  max(u(nw),zero)*(t(1:nc,nn)-t(1:nc,nw))    &
                                  + min(u(nn),zero)*(t(1:nc,ne)-t(1:nc,nn))    &
                                  - max(u(ww),zero)*(t(1:nc,mi)-t(1:nc,ww))    &
                                  - min(u(mi),zero)*(t(1:nc,ee)-t(1:nc,mi)) )) &
        + mhdtddy*max(v(mi),zero)*( t(1:nc,mi) - t(1:nc,ss)                    &
                          + facx*(  max(u(ww),zero)*(t(1:nc,mi)-t(1:nc,ww))    &
                                  + min(u(mi),zero)*(t(1:nc,ee)-t(1:nc,mi))    &
                                  - max(u(sw),zero)*(t(1:nc,ss)-t(1:nc,sw))    &
                                  - min(u(ss),zero)*(t(1:nc,se)-t(1:nc,ss)) ))
    

      ! do the water column, k=1:kb ...
      do k=2,kb
        t3tmp(1:nc,k) = t(1:nc,m(2,i,j)-2+k)
      enddo

      do k=2,min(kb,kh(m(1,i,j-1)))
        mi = m(2,i,  j  )-2+k
        ww = m(2,i,  j-1)-2+k
        wm(k)         = mhdtddx*max(u(ww),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) + wm(k)*(t(1:nc,mi)- t(1:nc,ww))
      enddo
      do k=2,min(kb,kh(m(1,i,j-1)),kh(m(1,i-1,j-1)))
        nw = m(2,i-1,j-1)-2+k
        ww = m(2,i,  j-1)-2+k
        wk = wm(k)*facy*min(v(nw),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) - wk*(t(1:nc,nw) - t(1:nc,ww))
      enddo
      do k=2,min(kb,kh(m(1,i,j-1)),kh(m(1,i+1,j-1)))
        sw = m(2,i+1,j-1)-2+k
        ww = m(2,i,  j-1)-2+k
        wk = wm(k)*facy*max(v(ww),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) - wk*(t(1:nc,ww) - t(1:nc,sw))
      enddo
      do k=2,min(kb,kh(m(1,i,j-1)),kh(m(1,i-1,j)))
        mi = m(2,i,  j  )-2+k
        nn = m(2,i-1,j  )-2+k
        wk = wm(k)*facy*min(v(nn),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) + wk*(t(1:nc,nn) - t(1:nc,mi))
      enddo
      do k=2,min(kb,kh(m(1,i,j-1)),kh(m(1,i+1,j)))
        mi = m(2,i,  j  )-2+k
        ss = m(2,i+1,j  )-2+k
        wk = wm(k)*facy*max(v(mi),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) + wk*(t(1:nc,mi) - t(1:nc,ss))
      enddo

      do k=2,min(kb,kh(m(1,i,j+1)))
        mi = m(2,i,  j  )-2+k
        ee = m(2,i,  j+1)-2+k
        wm(k)         = mhdtddx*min(u(mi),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) + wm(k)*(t(1:nc,ee)- t(1:nc,mi))
      enddo
      do k=2,min(kb,kh(m(1,i,j+1)),kh(m(1,i-1,j)))
        mi = m(2,i  ,j  )-2+k
        nn = m(2,i-1,j  )-2+k
        wk = wm(k)*facy*min(v(nn),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) - wk*(t(1:nc,nn) - t(1:nc,mi))
      enddo
      do k=2,min(kb,kh(m(1,i,j+1)),kh(m(1,i+1,j)))
        mi = m(2,i  ,j  )-2+k
        ss = m(2,i+1,j  )-2+k
        wk = wm(k)*facy*max(v(mi),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) - wk*(t(1:nc,mi) - t(1:nc,ss))
      enddo
      do k=2,min(kb,kh(m(1,i,j+1)),kh(m(1,i-1,j+1)))
        ee = m(2,i  ,j+1)-2+k
        ne = m(2,i-1,j+1)-2+k
        wk = wm(k)*facy*min(v(ne),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) + wk*(t(1:nc,ne) - t(1:nc,ee))
      enddo
      do k=2,min(kb,kh(m(1,i,j+1)),kh(m(1,i+1,j+1)))
        ee = m(2,i  ,j+1)-2+k
        se = m(2,i+1,j+1)-2+k
        wk = wm(k)*facy*max(v(ee),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) + wk*(t(1:nc,ee) - t(1:nc,se))
      enddo

      do k=2,min(kb,kh(m(1,i-1,j)))
        mi = m(2,i,  j  )-2+k
        nn = m(2,i-1,j  )-2+k
        wm(k)         = mhdtddy*min(v(nn),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) + wm(k)*(t(1:nc,nn)- t(1:nc,mi))
      enddo
      do k=2,min(kb,kh(m(1,i-1,j)),kh(m(1,i-1,j-1)))
        nw = m(2,i-1,j-1)-2+k
        nn = m(2,i-1,j  )-2+k
        wk = wm(k)*facx*max(u(nw),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) + wk*(t(1:nc,nn) - t(1:nc,nw))
      enddo
      do k=2,min(kb,kh(m(1,i-1,j)),kh(m(1,i-1,j+1)))
        ne = m(2,i-1,j+1)-2+k
        nn = m(2,i-1,j  )-2+k
        wk = wm(k)*facx*min(u(nn),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) + wk*(t(1:nc,ne) - t(1:nc,nn))
      enddo
      do k=2,min(kb,kh(m(1,i-1,j)),kh(m(1,i,j-1)))
        mi = m(2,i,  j  )-2+k
        ww = m(2,i,  j-1)-2+k
        wk = wm(k)*facx*max(u(ww),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) - wk*(t(1:nc,mi) - t(1:nc,ww))
      enddo
      do k=2,min(kb,kh(m(1,i-1,j)),kh(m(1,i,j+1)))
        mi = m(2,i,  j  )-2+k
        ee = m(2,i,  j+1)-2+k
        wk = wm(k)*facx*min(u(mi),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) - wk*(t(1:nc,ee) - t(1:nc,mi))
      enddo

      do k=2,min(kb,kh(m(1,i+1,j)))
        mi = m(2,i,  j  )-2+k
        ss = m(2,i+1,j  )-2+k
        wm(k)         = mhdtddy*max(v(mi),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) + wm(k)*(t(1:nc,mi)- t(1:nc,ss))
      enddo
      do k=2,min(kb,kh(m(1,i+1,j)),kh(m(1,i,j-1)))
        ww = m(2,i,  j-1)-2+k
        mi = m(2,i,  j  )-2+k
        wk = wm(k)*facx*max(u(ww),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) + wk*(t(1:nc,mi) - t(1:nc,ww))
      enddo
      do k=2,min(kb,kh(m(1,i+1,j)),kh(m(1,i,j+1)))
        ee = m(2,i,  j+1)-2+k
        mi = m(2,i,  j  )-2+k
        wk = wm(k)*facx*min(u(mi),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) + wk*(t(1:nc,ee) - t(1:nc,mi))
      enddo
      do k=2,min(kb,kh(m(1,i+1,j)),kh(m(1,i+1,j-1)))
        ss = m(2,i+1,j  )-2+k
        sw = m(2,i+1,j-1)-2+k
        wk = wm(k)*facx*max(u(sw),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) - wk*(t(1:nc,ss) - t(1:nc,sw))
      enddo
      do k=2,min(kb,kh(m(1,i+1,j)),kh(m(1,i+1,j+1)))
        ss = m(2,i+1,j  )-2+k
        se = m(2,i+1,j+1)-2+k
        wk = wm(k)*facx*min(u(ss),zero)
        t3tmp(1:nc,k) = t3tmp(1:nc,k) - wk*(t(1:nc,se) - t(1:nc,ss))
      enddo

      ! mult by w (here instead of in c_tt) and transfer to t3 array:
      do k=2,kb
        mi = m(2,i,j)-2+k                 
        wk = w(mi)
        t3(1:nc,mi) = max(wk,zero)*t3tmp(1:nc,k) + min(wk,zero)*t3tmp(1:nc,k-1)
      enddo
    enddo

  end subroutine c_tw

  ! ----------------------------------------------------------------------------

  subroutine c_tu (kmax,m,ind,n2d,advecstab,h_new,khu,u,v,w,t,kh,idx)
    
    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: kmax,n2d
    integer(4), intent(in) :: m(:,0:,0:),ind(:,:),khu(0:),kh(0:)
    real(8), intent(in)    :: advecstab, h_new(0:)
    real(8), intent(in)    :: u(0:),v(0:),w(0:)
    real(8), intent(in)    :: t(:,0:)
    integer(4), intent(inout) :: idx(1:,1:)
    
    !- local vars --------------------------------------------------------------
    integer(4) :: kbu, n, i, j, k, jj, mjj, mi, kb
    integer(4) :: n2dl,n2du
    real(8)    :: fac, dtddz, dtddy, facz, facy, mhdtddz, mhdtddy
    real(8), dimension(nc) :: tm, tp
    
    ! --------------------------------------------------------------------------
    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)

    dtddy   = dt/dy
    facy    = -onethird*dtddy
    mhdtddy = -half*dtddy

    do n=n2dl, n2du
      kbu = khu(n)
      kb  = kh(n)
      i   = ind(1,n)
      jj  = ind(2,n)

      ! make sure to have nice values throughout
      if (kb > 1 .and. kb > kbu) t1(1:nc,m(max(2,kbu+1),i,jj):m(kb,i,jj)) = zero
      if (kbu == 0) then
        t1(1:nc,n) = zero
        cycle
      endif

      ! First do k=1 -----------------------------------------------------------
      k = 1
      mjj = m(k,i,jj)
      ! branch on flow direction:
      if (u(mjj) >= zero) then 
        j = jj
      else
        j = jj + 1
      endif
      mi = m(k,i,j)

      t1(1:nc,mjj)=t(1:nc,mi)
    
      dtddz = dt/(h_new(mi) - advecstab)
      facz  = -onethird*dtddz
      mhdtddz = -half*dtddz

      if (k < kmax) then  ! k=1<kmax
        if (v(m(k,i-1,j)) < zero) then
          tm(1:nc)=t(1:nc,m(k,i-1,j))
          if (w(m(k+1,i-1,j)) > zero) then
            fac = facz*w(m(k+1,i-1,j))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k,i-1,j))-t(1:nc,m(k+1,i-1,j)))
          endif
          tp(1:nc)=t(1:nc,mi)
          if (w(m(k+1,i,j)) > zero) then
            fac = facz*w(m(k+1,i,j))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,mi)-t(1:nc,m(k+1,i,j)))
          endif
          fac = mhdtddy*v(m(k,i-1,j))
          t1(1:nc,mjj)=t1(1:nc,mjj) + fac*(tm(1:nc)-tp(1:nc))
        endif
    
        if (v(mi) > zero) then
          tm(1:nc)=t(1:nc,mi)
          if (w(m(k+1,i,j)) > zero) then
            fac = facz*w(m(k+1,i,j))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,mi)-t(1:nc,m(k+1,i,j)))
          endif
          tp(1:nc)=t(1:nc,m(k,i+1,j))
          if (w(m(k+1,i+1,j)) > zero) then
            fac = facz*w(m(k+1,i+1,j))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k,i+1,j))-t(1:nc,m(k+1,i+1,j)))
          endif
          fac = mhdtddy*v(mi)
          t1(1:nc,mjj)=t1(1:nc,mjj) + fac*(tm(1:nc)-tp(1:nc))
        endif
    
        if (w(m(k+1,i,j)) > zero) then
          tm(1:nc)=t(1:nc,mi)
          if (v(m(k,i-1,j)) < zero) then
            fac = facy*v(m(k,i-1,j))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k,i-1,j))-t(1:nc,mi))
          endif
          if (v(mi) > zero) then
            fac = facy*v(mi)
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,mi)-t(1:nc,m(k,i+1,j)))
          endif
          tp(1:nc)=t(1:nc,m(k+1,i,j))
          if (v(m(k+1,i-1,j)) < zero) then
            fac = facy*v(m(k+1,i-1,j))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k+1,i-1,j))-t(1:nc,m(k+1,i,j)))
          endif
          if (v(m(k+1,i,j)) > zero) then
            fac = facy*v(m(k+1,i,j))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k+1,i,j))-t(1:nc,m(k+1,i+1,j)))
          endif
          fac = mhdtddz*w(m(k+1,i,j))
          t1(1:nc,mjj)=t1(1:nc,mjj) + fac*(tm(1:nc)-tp(1:nc))
        endif

      else ! k=1=kmax
        if (v(m(k,i-1,j)) < zero) then
          fac = mhdtddy*v(m(k,i-1,j))
          t1(1:nc,mjj)=t1(1:nc,mjj) + fac*(t(1:nc,m(k,i-1,j))-t(1:nc,mi))
        endif
        if (v(mi) > zero) then
          fac = mhdtddy*v(mi)
          t1(1:nc,mjj)=t1(1:nc,mjj) + fac*(t(1:nc,mi)-t(1:nc,m(k,i+1,j)))
        endif

      endif
      ! This ends treatment of k=1 ---------------------------------------------

      ! Do all k in-between k=1 and kmax ---------------------------------------
      ! fixme vectorize, see c_dt
      do k=2,min(kbu,kmax-1)
        mjj = m(k,i,jj)

        ! branch on flow direction:
        if (u(mjj) >= zero) then 
          j = jj
        else
          j = jj + 1
        endif
    
        mi = m(k,i,j)

        t1(1:nc,mjj)=t(1:nc,mi)
    
        dtddz = dt/h_new(mi)
        facz  = -onethird*dtddz
        mhdtddz = -half*dtddz
    
        if (v(m(k,i-1,j)) < zero) then
          tm(1:nc)=t(1:nc,m(k,i-1,j))
          if (w(m(k,i-1,j)) < zero) then
            fac = facz*w(m(k,i-1,j))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k-1,i-1,j))-t(1:nc,m(k,i-1,j)))
          endif
          if (w(m(k+1,i-1,j)) > zero) then
            fac = facz*w(m(k+1,i-1,j))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k,i-1,j))-t(1:nc,m(k+1,i-1,j)))
          endif
          tp(1:nc)=t(1:nc,mi)
          if (w(mi) < zero) then
            fac = facz*w(mi)
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k-1,i,j))-t(1:nc,mi))
          endif
          if (w(m(k+1,i,j)) > zero) then
            fac = facz*w(m(k+1,i,j))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,mi)-t(1:nc,m(k+1,i,j)))
          endif
          fac = mhdtddy*v(m(k,i-1,j))
          t1(1:nc,mjj)=t1(1:nc,mjj) + fac*(tm(1:nc)-tp(1:nc))
        endif
    
        if (v(mi) > zero) then
          tm(1:nc)=t(1:nc,mi)
          if (w(mi) < zero) then
            fac = facz*w(mi)
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k-1,i,j))-t(1:nc,mi))
          endif
          if (w(m(k+1,i,j)) > zero) then
            fac = facz*w(m(k+1,i,j))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,mi)-t(1:nc,m(k+1,i,j)))
          endif
          tp(1:nc)=t(1:nc,m(k,i+1,j))
          if (w(m(k,i+1,j)) < zero) then
            fac = facz*w(m(k,i+1,j))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k-1,i+1,j))-t(1:nc,m(k,i+1,j)))
          endif
          if (w(m(k+1,i+1,j)) > zero) then
            fac = facz*w(m(k+1,i+1,j))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k,i+1,j))-t(1:nc,m(k+1,i+1,j)))
          endif
          fac = mhdtddy*v(mi)
          t1(1:nc,mjj)=t1(1:nc,mjj) + fac*(tm(1:nc)-tp(1:nc))
        endif
    
        if (w(mi) < zero) then
          tm(1:nc)=t(1:nc,m(k-1,i,j))
          if (v(m(k-1,i-1,j)) < zero) then
            fac = facy*v(m(k-1,i-1,j))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k-1,i-1,j))-t(1:nc,m(k-1,i,j)))
          endif
          if (v(m(k-1,i,j)) > zero) then
            fac = facy*v(m(k-1,i,j))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k-1,i,j))-t(1:nc,m(k-1,i+1,j)))
          endif
          tp(1:nc)=t(1:nc,mi)
          if (v(m(k,i-1,j)) < zero) then
            fac = facy*v(m(k,i-1,j))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k,i-1,j))-t(1:nc,mi))
          endif
          if (v(mi) > zero) then
            fac = facy*v(mi)
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,mi)-t(1:nc,m(k,i+1,j)))
          endif
          fac = mhdtddz*w(mi)
          t1(1:nc,mjj)=t1(1:nc,mjj) + fac*(tm(1:nc)-tp(1:nc))
        endif
    
        if (w(m(k+1,i,j)) > zero) then
          tm(1:nc)=t(1:nc,mi)
          if (v(m(k,i-1,j)) < zero) then
            fac = facy*v(m(k,i-1,j))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k,i-1,j))-t(1:nc,mi))
          endif
          if (v(mi) > zero) then
            fac = facy*v(mi)
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,mi)-t(1:nc,m(k,i+1,j)))
          endif
          tp(1:nc)=t(1:nc,m(k+1,i,j))
          if (v(m(k+1,i-1,j)) < zero) then
            fac = facy*v(m(k+1,i-1,j))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k+1,i-1,j))-t(1:nc,m(k+1,i,j)))
          endif
          if (v(m(k+1,i,j)) > zero) then
            fac = facy*v(m(k+1,i,j))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k+1,i,j))-t(1:nc,m(k+1,i+1,j)))
          endif
          fac = mhdtddz*w(m(k+1,i,j))
          t1(1:nc,mjj)=t1(1:nc,mjj) + fac*(tm(1:nc)-tp(1:nc))
        endif
    
      enddo ! k loop

      ! Possibly k=kbu=kmax>1 --------------------------------------------------
      k = kbu
      if (k == kmax .and. kmax > 1) then
        mjj = m(k,i,jj)

        ! branch on flow direction:
        if (u(mjj) >= zero) then 
          j = jj
        else
          j = jj + 1
        endif
    
        mi = m(k,i,j)

        t1(1:nc,mjj)=t(1:nc,mi)
    
        dtddz = dt/h_new(mi)
        facz  = -onethird*dtddz
        mhdtddz = -half*dtddz
    
        if (v(m(k,i-1,j)) < zero) then
          tm(1:nc)=t(1:nc,m(k,i-1,j))
          if (w(m(k,i-1,j)) < zero) then
            fac = facz*w(m(k,i-1,j))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k-1,i-1,j))-t(1:nc,m(k,i-1,j)))
          endif
          tp(1:nc)=t(1:nc,mi)
          if (w(mi) < zero) then
            fac = facz*w(mi)
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k-1,i,j))-t(1:nc,mi))
          endif
          fac = mhdtddy*v(m(k,i-1,j))
          t1(1:nc,mjj)=t1(1:nc,mjj) + fac*(tm(1:nc)-tp(1:nc))
        endif
        if (v(mi) > zero) then
          tm(1:nc)=t(1:nc,mi)
          if (w(mi) < zero) then
            fac = facz*w(mi)
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k-1,i,j))-t(1:nc,mi))
          endif
          tp(1:nc)=t(1:nc,m(k,i+1,j))
          if (w(m(k,i+1,j)) < zero) then
            fac = facz*w(m(k,i+1,j))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k-1,i+1,j))-t(1:nc,m(k,i+1,j)))
          endif
          fac = mhdtddy*v(mi)
          t1(1:nc,mjj)=t1(1:nc,mjj) + fac*(tm(1:nc)-tp(1:nc))
        endif
        if (w(mi) < zero) then
          tm(1:nc)=t(1:nc,m(k-1,i,j))
          if (v(m(k-1,i-1,j)) < zero) then
            fac = facy*v(m(k-1,i-1,j))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k-1,i-1,j))-t(1:nc,m(k-1,i,j)))
          endif
          if (v(m(k-1,i,j)) > zero) then
            fac = facy*v(m(k-1,i,j))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k-1,i,j))-t(1:nc,m(k-1,i+1,j)))
          endif
          tp(1:nc)=t(1:nc,mi)
          if (v(m(k,i-1,j)) < zero) then
            fac = facy*v(m(k,i-1,j))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k,i-1,j))-t(1:nc,mi))
          endif
          if (v(mi) > zero) then
            fac = facy*v(mi)
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,mi)-t(1:nc,m(k,i+1,j)))
          endif
          fac = mhdtddz*w(mi)
          t1(1:nc,mjj)=t1(1:nc,mjj) + fac*(tm(1:nc)-tp(1:nc))
        endif
      endif ! k=kbu=kmax

    enddo

  end subroutine c_tu

  ! ----------------------------------------------------------------------------

  subroutine c_tv(kmax,m,ind,n2d,advecstab,h_new,khv,u,v,w,t,cosphi,kh,idx)

    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: kmax,n2d
    integer(4), intent(in) :: m(:,0:,0:),ind(:,:),khv(0:),kh(0:)
    real(8), intent(in)    :: advecstab, h_new(0:)
    real(8), intent(in)    :: u(0:),v(0:),w(0:)
    real(8), intent(in)    :: t(:,0:)
    real(8), intent(in)    :: cosphi(:,0:)
    integer(4), intent(inout) :: idx(1:,1:)
    
    
    !- local vars --------------------------------------------------------------
    !  automatic:
    real(8), dimension(nc) :: tm, tp
    
    !  simple vars:
    integer(4) :: kbv, n, i, j, k, ii, mii, mi, kb
    integer(4) :: n2dl, n2du
    real(8)    :: fac, dtddz, dtddx, facz, facx, mhdtddz, mhdtddx, dtddx0
    
    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)

    dtddx0 = dt/dx

    do n=n2dl,n2du
      kbv = khv(n)
      kb  = kh(n)
      ii  = ind(1,n)
      j   = ind(2,n)

      ! make sure to have nice values throughout
      if (kb > 1 .and. kb > kbv) t2(1:nc,m(max(2,kbv+1),ii,j):m(kb,ii,j)) = zero
      if (kbv == 0) then
        t2(1:nc,n) = zero
        cycle
      endif

      dtddx   = dtddx0/cosphi(1,ii)
      facx    = -onethird*dtddx
      mhdtddx = -half*dtddx

      ! First do k=1 -----------------------------------------------------------
      k = 1
      mii = m(k,ii,j)
      ! branch on flow direction:
      if (v(mii) <= zero) then
        i = ii
      else
        i = ii + 1
      endif
      mi = m(k,i,j)

      t2(1:nc,mii)=t(1:nc,mi)
    
      dtddz = dt/(h_new(mi) - advecstab)
      facz  = -onethird*dtddz
      mhdtddz = -half*dtddz
    
      if (k < kmax) then  ! k=1<kmax
        if (u(m(k,i,j-1)) > zero) then
          tm(1:nc)=t(1:nc,m(k,i,j-1))
          if (w(m(k+1,i,j-1)) > zero) then
            fac = facz*w(m(k+1,i,j-1))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k,i,j-1))-t(1:nc,m(k+1,i,j-1)))
          endif
          tp(1:nc)=t(1:nc,mi)
          if (w(m(k+1,i,j)) > zero) then
            fac = facz*w(m(k+1,i,j))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,mi)-t(1:nc,m(k+1,i,j)))
          endif
          fac = mhdtddx*u(m(k,i,j-1))
          t2(1:nc,mii)=t2(1:nc,mii) + fac*(tp(1:nc)-tm(1:nc))
        endif
    
        if (u(mi) < zero) then
          tm(1:nc)=t(1:nc,mi)
          if (w(m(k+1,i,j)) > zero) then
            fac = facz*w(m(k+1,i,j))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,mi)-t(1:nc,m(k+1,i,j)))
          endif
          tp(1:nc)=t(1:nc,m(k,i,j+1))
          if (w(m(k+1,i,j+1)) > zero) then
            fac = facz*w(m(k+1,i,j+1))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k,i,j+1))-t(1:nc,m(k+1,i,j+1)))
          endif
          fac = mhdtddx*u(mi)
          t2(1:nc,mii)=t2(1:nc,mii) + fac*(tp(1:nc)-tm(1:nc))
        endif
    
        if (m(k+1,i,j) > 0) then
          tm(1:nc)=t(1:nc,mi)
          if (u(m(k,i,j-1)) > zero) then
            fac = facx*u(m(k,i,j-1))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,mi)-t(1:nc,m(k,i,j-1)))
          endif
          if (u(mi) < zero) then
            fac = facx*u(mi)
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k,i,j+1))-t(1:nc,mi))
          endif
          tp(1:nc)=t(1:nc,m(k+1,i,j))
          if (u(m(k+1,i,j-1)) > zero) then
            fac = facx*u(m(k+1,i,j-1))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k+1,i,j))-t(1:nc,m(k+1,i,j-1)))
          endif
          if (u(m(k+1,i,j)) < zero) then
            fac = facx*u(m(k+1,i,j))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k+1,i,j+1))-t(1:nc,m(k+1,i,j)))
          endif
          fac = mhdtddz*w(m(k+1,i,j))
          t2(1:nc,mii)=t2(1:nc,mii) + fac*(tm(1:nc)-tp(1:nc))
        endif

      else ! k=1=kmax
        if (u(m(k,i,j-1)) > zero) then
          fac = mhdtddx*u(m(k,i,j-1))
          t2(1:nc,mii)=t2(1:nc,mii) + fac*(t(1:nc,mi)-t(1:nc,m(k,i,j-1)))
        endif
        if (u(mi) < zero) then
          fac = mhdtddx*u(mi)
          t2(1:nc,mii)=t2(1:nc,mii) + fac*(t(1:nc,m(k,i,j+1))-t(1:nc,mi))
        endif

      endif
      ! This ends treatment of k=1 ---------------------------------------------
      
      ! Do all k in-between k=1 and kmax ---------------------------------------
      ! fixme vectorize, see c_dt
      do k=2,min(kbv,kmax-1)
        mii = m(k,ii,j)

        ! branch on flow direction:
        if (v(mii) <= zero) then
          i = ii
        else
          i = ii + 1
        endif
    
        mi = m(k,i,j)

        t2(1:nc,mii)=t(1:nc,mi)
    
        dtddz = dt/h_new(mi)
        facz  = -onethird*dtddz
        mhdtddz = -half*dtddz
    
        if (u(m(k,i,j-1)) > zero) then
          tm(1:nc)=t(1:nc,m(k,i,j-1))
          if (w(m(k,i,j-1)) < zero) then
            fac = facz*w(m(k,i,j-1))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k-1,i,j-1))-t(1:nc,m(k,i,j-1)))
          endif
          if (w(m(k+1,i,j-1)) > zero) then
            fac = facz*w(m(k+1,i,j-1))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k,i,j-1))-t(1:nc,m(k+1,i,j-1)))
          endif
          tp(1:nc)=t(1:nc,mi)
          if (w(mi) < zero) then
            fac = facz*w(mi)
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k-1,i,j))-t(1:nc,mi))
          endif
          if (w(m(k+1,i,j)) > zero) then
            fac = facz*w(m(k+1,i,j))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,mi)-t(1:nc,m(k+1,i,j)))
          endif
          fac = mhdtddx*u(m(k,i,j-1))
          t2(1:nc,mii)=t2(1:nc,mii) + fac*(tp(1:nc)-tm(1:nc))
        endif
    
        if (u(mi) < zero) then
          tm(1:nc)=t(1:nc,mi)
          if (w(mi) < zero) then
            fac = facz*w(mi)
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k-1,i,j))-t(1:nc,mi))
          endif
          if (w(m(k+1,i,j)) > zero) then
            fac = facz*w(m(k+1,i,j))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,mi)-t(1:nc,m(k+1,i,j)))
          endif
          tp(1:nc)=t(1:nc,m(k,i,j+1))
          if (w(m(k,i,j+1)) < zero) then
            fac = facz*w(m(k,i,j+1))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k-1,i,j+1))-t(1:nc,m(k,i,j+1)))
          endif
          if (w(m(k+1,i,j+1)) > zero) then
            fac = facz*w(m(k+1,i,j+1))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k,i,j+1))-t(1:nc,m(k+1,i,j+1)))
          endif
          fac = mhdtddx*u(mi)
          t2(1:nc,mii)=t2(1:nc,mii) + fac*(tp(1:nc)-tm(1:nc))
        endif
    
        if (w(mi) < zero) then
          tm(1:nc)=t(1:nc,m(k-1,i,j))
          if (u(m(k-1,i,j-1)) > zero) then
            fac = facx*u(m(k-1,i,j-1))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k-1,i,j))-t(1:nc,m(k-1,i,j-1)))
          endif
          if (u(m(k-1,i,j)) < zero) then
            fac = facx*u(m(k-1,i,j))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k-1,i,j+1))-t(1:nc,m(k-1,i,j)))
          endif
          tp(1:nc)=t(1:nc,mi)
          if (u(m(k,i,j-1)) > zero) then
            fac = facx*u(m(k,i,j-1))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,mi)-t(1:nc,m(k,i,j-1)))
          endif
          if (u(mi) < zero) then
            fac = facx*u(mi)
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k,i,j+1))-t(1:nc,mi))
          endif
          fac = mhdtddz*w(mi)
          t2(1:nc,mii)=t2(1:nc,mii) + fac*(tm(1:nc)-tp(1:nc))
        endif
    
        if (w(m(k+1,i,j)) > zero) then
          tm(1:nc)=t(1:nc,mi)
          if (u(m(k,i,j-1)) > zero) then
            fac = facx*u(m(k,i,j-1))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,mi)-t(1:nc,m(k,i,j-1)))
          endif
          if (u(mi) < zero) then
            fac = facx*u(mi)
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k,i,j+1))-t(1:nc,mi))
          endif
          tp(1:nc)=t(1:nc,m(k+1,i,j))
          if (u(m(k+1,i,j-1)) > zero) then
            fac = facx*u(m(k+1,i,j-1))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k+1,i,j))-t(1:nc,m(k+1,i,j-1)))
          endif
          if (u(m(k+1,i,j)) < zero) then
            fac = facx*u(m(k+1,i,j))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k+1,i,j+1))-t(1:nc,m(k+1,i,j)))
          endif
          fac = mhdtddz*w(m(k+1,i,j))
          t2(1:nc,mii)=t2(1:nc,mii) + fac*(tm(1:nc)-tp(1:nc))
        endif
    
      enddo ! k loop

      ! Possibly k=kbv=kmax>1 --------------------------------------------------
      k = kbv
      if (k == kmax .and. kmax > 1) then
        mii = m(k,ii,j)

        ! branch on flow direction:
        if (v(mii) <= zero) then
          i = ii
        else
          i = ii + 1
        endif
    
        mi = m(k,i,j)

        t2(1:nc,mii)=t(1:nc,mi)
    
        dtddz = dt/h_new(mi)
        facz  = -onethird*dtddz
        mhdtddz = -half*dtddz
    
        if (u(m(k,i,j-1)) > zero) then
          tm(1:nc)=t(1:nc,m(k,i,j-1))
          if (w(m(k,i,j-1)) < zero) then
            fac = facz*w(m(k,i,j-1))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k-1,i,j-1))-t(1:nc,m(k,i,j-1)))
          endif
          tp(1:nc)=t(1:nc,mi)
          if (w(mi) < zero) then
            fac = facz*w(mi)
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k-1,i,j))-t(1:nc,mi))
          endif
          fac = mhdtddx*u(m(k,i,j-1))
          t2(1:nc,mii)=t2(1:nc,mii) + fac*(tp(1:nc)-tm(1:nc))
        endif
        if (u(mi) < zero) then
          tm(1:nc)=t(1:nc,mi)
          if (w(mi) < zero) then
            fac = facz*w(mi)
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k-1,i,j))-t(1:nc,mi))
          endif
          tp(1:nc)=t(1:nc,m(k,i,j+1))
          if (w(m(k,i,j+1)) < zero) then
            fac = facz*w(m(k,i,j+1))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k-1,i,j+1))-t(1:nc,m(k,i,j+1)))
          endif
          fac = mhdtddx*u(mi)
          t2(1:nc,mii)=t2(1:nc,mii) + fac*(tp(1:nc)-tm(1:nc))
        endif
        if (w(mi) < zero) then
          tm(1:nc)=t(1:nc,m(k-1,i,j))
          if (u(m(k-1,i,j-1)) > zero) then
            fac = facx*u(m(k-1,i,j-1))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k-1,i,j))-t(1:nc,m(k-1,i,j-1)))
          endif
          if (u(m(k-1,i,j)) < zero) then
            fac = facx*u(m(k-1,i,j))
            tm(1:nc)=tm(1:nc) + fac*(t(1:nc,m(k-1,i,j+1))-t(1:nc,m(k-1,i,j)))
          endif
          tp(1:nc)=t(1:nc,mi)
          if (u(m(k,i,j-1)) > zero) then
            fac = facx*u(m(k,i,j-1))
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,mi)-t(1:nc,m(k,i,j-1)))
          endif
          if (u(mi) < zero) then
            fac = facx*u(mi)
            tp(1:nc)=tp(1:nc) + fac*(t(1:nc,m(k,i,j+1))-t(1:nc,mi))
          endif
          fac = mhdtddz*w(mi)
          t2(1:nc,mii)=t2(1:nc,mii) + fac*(tm(1:nc)-tp(1:nc))
        endif
      endif ! k=kbv=kmax

    enddo
    
  end subroutine c_tv

  ! ----------------------------------------------------------------------------

  subroutine advection(mmk,ind,n2d,kh,khu,khv,hx,hy,h_new,u,v,w,t,cosphi,idx)
    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain
    
    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: n2d
    integer(4), intent(in)    :: mmk(:,0:,0:), ind(:,:), kh(0:), khu(0:)
    integer(4), intent(in)    :: khv(0:)
    real(8),    intent(in)    :: hx(0:), hy(0:), h_new(0:)
    real(8),    intent(in)    :: u(0:), v(0:), w(0:)
    real(8),    intent(inout) :: t(:,0:)
    real(8),    intent(in)    :: cosphi(:,0:)
    integer(4), intent(inout) :: idx(1:,1:)
    
    !- local vars --------------------------------------------------------------
    !  simple vars:
    integer(4) :: n, i, j, k, kb, mi, mn, mw, md, mi0, mn0, mw0
    integer(4) :: n2dl,n2du
    real(8)    :: fac0, fac, fx00, fy01, fy02, fx1, fx2, fy1, fy2, fz1, fz2

    ! --------------------------------------------------------------------------
    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)

    ! compute t
    do n=n2dl,n2du
      kb = kh(n)
      if (kb < 1) cycle
      i = ind(1,n)
      j = ind(2,n)
      fac0 = dt/cosphi(1,i)
      fx00 = fac0/dx
      fy01 = fac0*cosphi(2,i-1)/dy
      fy02 = fac0*cosphi(2,i  )/dy

      ! k=1 unrolled:
      mi  = n
      mw  = mmk(1,i,j-1)
      mn  = mmk(1,i-1,j)
      fac = one/h_new(mi) 
      fx1 = fx00*hx(mi)*u(mi)
      fx2 = fx00*hx(mw)*u(mw)
      fy1 = fy01*hy(mn)*v(mn)
      fy2 = fy02*hy(mi)*v(mi)
      t(1:nc,mi) = tt(1:nc,mi)                                                 &
         + fac*(  fx2*t1(1:nc,mw)*t4(1:nc,mw) - fx1*t1(1:nc,mi)*t4(1:nc,mi)    &
                + fy2*t2(1:nc,mi)*t5(1:nc,mi) - fy1*t2(1:nc,mn)*t5(1:nc,mn))

      if (kb > 1) then
        md  = mmk(2,i,j)
        fz2 = dt*w(md)/h_new(mi)
        t(1:nc,mi) = t(1:nc,mi) + fz2*t3(1:nc,md)*t6(1:nc,md)
      endif
    enddo

    do n=n2dl,n2du
      kb = kh(n)
      if (kb <= 1) cycle

      i = ind(1,n)
      j = ind(2,n)
      fac0 = dt/cosphi(1,i)
      fx00 = fac0/dx
      fy01 = fac0*cosphi(2,i-1)/dy
      fy02 = fac0*cosphi(2,i  )/dy

      mi0 = mmk(2,i,  j  ) - 2
      mn0 = mmk(2,i-1,j  ) - 2
      mw0 = mmk(2,i  ,j-1) - 2
      do k=2,kb-1
        mi = mi0 + k
        md = mi + 1
        fac = one/h_new(mi) 
        fx1 = fx00*hx(mi)*u(mi)
        fy2 = fy02*hy(mi)*v(mi)
        fz1 = dt*w(mi)
        fz2 = dt*w(md)
        t(1:nc,mi) = tt(1:nc,mi)                                               &
         + fac*(                              - fx1*t1(1:nc,mi)*t4(1:nc,mi)    &
                + fy2*t2(1:nc,mi)*t5(1:nc,mi)                                  &
                + fz2*t3(1:nc,md)*t6(1:nc,md) - fz1*t3(1:nc,mi)*t6(1:nc,mi))
      enddo
      do k=2,min(kb-1,khu(mmk(1,i,j-1)))
        mi = mi0 + k
        mw = mw0 + k
        fx2 = fx00*hx(mw)*u(mw)/h_new(mi)
        t(1:nc,mi) = t(1:nc,mi) + fx2*t1(1:nc,mw)*t4(1:nc,mw)
      enddo
      do k=2,min(kb-1,khv(mmk(1,i-1,j)))
        mi  = mi0 + k
        mn  = mn0 + k
        fy1 = fy01*hy(mn)*v(mn)/h_new(mi)
        t(1:nc,mi) = t(1:nc,mi) - fy1*t2(1:nc,mn)*t5(1:nc,mn)
      enddo

      ! k=kb unrolled:
      mi  = mmk(kb,i,  j)
      mw  = mmk(kb,i,j-1)
      mn  = mmk(kb,i-1,j)
      fac = one/h_new(mi) 
      fx1 = fx00*hx(mi)*u(mi)
      fx2 = fx00*hx(mw)*u(mw)
      fy1 = fy01*hy(mn)*v(mn)
      fy2 = fy02*hy(mi)*v(mi)
      fz1 = dt*w(mi)
      t(1:nc,mi) = tt(1:nc,mi)                                                 &
         + fac*(  fx2*t1(1:nc,mw)*t4(1:nc,mw) - fx1*t1(1:nc,mi)*t4(1:nc,mi)    &
                + fy2*t2(1:nc,mi)*t5(1:nc,mi) - fy1*t2(1:nc,mn)*t5(1:nc,mn)    &
                                              - fz1*t3(1:nc,mi)*t6(1:nc,mi))
    enddo
  end subroutine advection 
    
  ! ----------------------------------------------------------------------------
    
  subroutine copy_t2tt (m,ind,n2d,n2dhalo,kh,t,idx)

    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: n2d, n2dhalo
    integer(4), intent(in) :: m(:,0:,0:), ind(:,:), kh(0:)
    real(8),    intent(in) :: t(1:,0:)
    integer(4), intent(inout) :: idx(1:,1:)

    !- local vars --------------------------------------------------------------
    integer(4) :: n, n2dl, n2du, kb, ml, mu

    call domp_get_domain(kh, 1, n2d, n2dl, n2du, idx)

    tt(1:nc,n2dl:n2du) = t(1:nc,n2dl:n2du)
    do n=n2dl,n2du
      kb = kh(n)
      if (kb > 1) then
        ml = m(2,ind(1,n),ind(2,n))
        mu = ml + kb - 2
        tt(1:nc,ml:mu) = t(1:nc,ml:mu)
      endif
    enddo

!$OMP MASTER
    ! handle the halo:
    if (n2dhalo > 0) then
      n2dl = n2d+1
      n2du = n2d+n2dhalo
      tt(1:nc,n2dl:n2du) = t(1:nc,n2dl:n2du)
      do n=n2dl,n2du
        kb = kh(n)
        if (kb > 1) then
          ml = m(2,ind(1,n),ind(2,n))
          mu = ml + kb - 2
          tt(1:nc,ml:mu) = t(1:nc,ml:mu)
        endif
      enddo
    endif
!$OMP END MASTER
  end subroutine copy_t2tt

  ! ----------------------------------------------------------------------------
    
  subroutine tflow_int (kmax,mmk,ind,n2d,dx_j,dy_i,advecstab,dt_in,kh,khu,khv, &
                        h,hx,hy,h_old,h_new,u,v,w,t,avt,eddyh,nbpz,krz,        &
                        nbpq,krq,rwqk,nbpu,kru,nbpv,krv,rwzkt,rwqkt,cosphi,    &
                        doadvec,dodiff,idx,halo2d,ia,dyn,uvdam,vdiff,q)

    !- modules -----------------------------------------------------------------
    use dmi_mpi,        only : dmpi_halo
    use cmod_timer,     only : timer
    use masscorrection, only : masscorr_init, masscorr
    
    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)           :: ia,kmax,n2d,halo2d
    integer(4), intent(in)           :: ind(:,:)
    integer(4), intent(in)           :: mmk(:,0:,0:)
    integer(4), intent(in)           :: kh(0:),khu(0:),khv(0:)
    integer(4), intent(in)           :: nbpz,nbpq,nbpu,nbpv
    integer(4), intent(in)           :: krq(:,0:),krz(:,0:),kru(:,0:),krv(:,0:)
    real(8),    intent(in)           :: dt_in, advecstab
    real(8),    intent(in)           :: dx_j(0:),dy_i(0:)
    real(8),    intent(in)           :: h(0:),hx(0:),hy(0:)
    real(8),    intent(in)           :: h_old(0:),h_new(0:)
    real(8),    intent(in)           :: avt(1:,0:), uvdam(:,0:)
    real(8),    intent(in)           :: u(0:),v(0:),w(0:),eddyh(0:)
    real(8),    intent(in)           :: rwqk(0:,:),rwzkt(:,:,0:)
    real(8),    intent(in)           :: rwqkt(:,:,0:),cosphi(:,0:)
    logical,    intent(in)           :: doadvec, dodiff, dyn
    real(8),    intent(inout)        :: t(1:,0:)
    integer(4), intent(inout)        :: idx(1:,1:)
    logical,    intent(in), optional :: vdiff
    real(8),    intent(in), optional :: q(0:)
    
    !- local vars --------------------------------------------------------------
    !  none

    !- obtain constant dx, dy, local dt and derived quantities: ----------------
    dx     = dx_j(1)
    dy     = dy_i(1)
    dt     = dt_in
    dxdy   = dx*dy
    ddxdy  = one/dxdy
    dtdx   = dt*dx
    dtdy   = dt*dy
    dtdxdy = dt*dx*dy

    if (doadvec) then ! else skip advection
    
      call timer(1,'MPI_haloTF')
      call dmpi_halo(kmax,mmk,kh,ia,1,1,1,1,t1,t2)
      call dmpi_halo(kmax,mmk,kh,ia,4,3,2,1,t4,t5,t6)
      call timer(2,'MPI_haloTF')

      if (.not.dyn) then
!$OMP BARRIER
        ! prepared mass-corrective step ----------------------------------------
        call masscorr_init (n2d, nc, t, h_new, h_old, hx, hy, advecstab, u, v, &
                            dt, dx, dy, cosphi, ind, mmk, kh, idx)
!$OMP BARRIER
      endif

      ! ------------------------------------------------------------------------
      ! begin advection
      ! ------------------------------------------------------------------------
! MPI: Halo swap already done outside tflow-int for: u,v,w,h,t

      ! compute t1 -------------------------------------------------------------
      call c_tu(kmax,mmk,ind,n2d,advecstab,h_new,khu,u,v,w,t,kh,idx)
    
      ! compute t2 -------------------------------------------------------------
      call c_tv(kmax,mmk,ind,n2d,advecstab,h_new,khv,u,v,w,t,cosphi,kh,idx)
!$OMP BARRIER
    
      call timer(1,'MPI_haloTF')
      call dmpi_halo(kmax,mmk,kh,ia,1,1,1,3,t1,t2)
      call dmpi_halo(kmax,mmk,kh,ia,1,1,1,4,t1,t2)
      call timer(2,'MPI_haloTF')

      ! compute t3 -------------------------------------------------------------
      call c_tw(kmax,mmk,ind,n2d,kh,u,v,w,t,cosphi,idx)

      ! compute t4, t5, t6  ----------------------------------------------------
      call c_delta(mmk,ind,n2d,kh,khu,khv,h,nbpz,krz,nbpu,kru,nbpv,krv,t,idx)
!$OMP BARRIER

      call timer(1,'MPI_haloTF')
      call dmpi_halo(kmax,mmk,kh,ia,4,3,2,3,t4,t5,t6)
      call dmpi_halo(kmax,mmk,kh,ia,4,3,2,4,t4,t5,t6)
      call dmpi_halo(kmax,mmk,kh,ia,1,1,1,2,t1,t2)
      call dmpi_halo(kmax,mmk,kh,ia,1,1,1,5,t1,t2)
      call dmpi_halo(kmax,mmk,kh,ia,1,2,1,1,t1,t2)
      call dmpi_halo(kmax,mmk,kh,ia,3,1,3,1,tt)
      call timer(2,'MPI_haloTF')

      ! compute tt -------------------------------------------------------------
      ! this computation relies on t1, t2 and t3, i.e. barrier here
      call c_tt (kmax,mmk,ind,n2d,kh,khu,khv,hx,hy,h_old,h_new,                &
                 u,v,w,t,nbpz,krz,nbpq,krq,rwqk,rwzkt,rwqkt,cosphi,idx)
!$OMP BARRIER

      call timer(1,'MPI_haloTF')
      call dmpi_halo(kmax,mmk,kh,ia,3,1,3,3,tt)
      call dmpi_halo(kmax,mmk,kh,ia,3,1,3,4,tt)
      call dmpi_halo(kmax,mmk,kh,ia,4,3,2,2,t4,t5,t6)
      call dmpi_halo(kmax,mmk,kh,ia,4,3,2,5,t4,t5,t6)
      call dmpi_halo(kmax,mmk,kh,ia,4,1,2,1,t7,t8)
      call timer(2,'MPI_haloTF')

      ! compute t1, t2, t3  ----------------------------------------------------
      ! this computation relies on t4, t5, t6, i.e. barrier here
      call c_dt (kmax,mmk,ind,n2d,advecstab,h_new,kh,khu,khv,u,v,w,cosphi,idx)
!$OMP BARRIER

      call timer(1,'MPI_haloTF')
      call dmpi_halo(kmax,mmk,kh,ia,1,2,1,3,t1,t2)
      call dmpi_halo(kmax,mmk,kh,ia,1,2,1,4,t1,t2)
      call dmpi_halo(kmax,mmk,kh,ia,1,2,1,2,t1,t2)
      call dmpi_halo(kmax,mmk,kh,ia,1,2,1,5,t1,t2)
      call dmpi_halo(kmax,mmk,kh,ia,3,1,3,2,tt)
      call dmpi_halo(kmax,mmk,kh,ia,3,1,3,5,tt)
      call dmpi_halo(kmax,mmk,kh,ia,1,1,1,1,t4,t5)
      call timer(2,'MPI_haloTF')

      ! compute t8, t7 ---------------------------------------------------------
      ! this computation relies on tt, t1, t2, t3, i.e. barrier here
      call c_rin_rout (kmax,mmk,ind,n2d,kh,hx,hy,h_new,t,u,v,w,cosphi,idx)
!$OMP BARRIER

      call timer(1,'MPI_haloTF')
      call dmpi_halo(kmax,mmk,kh,ia,4,1,2,3,t7,t8)
      call dmpi_halo(kmax,mmk,kh,ia,4,1,2,4,t7,t8)
      call dmpi_halo(kmax,mmk,kh,ia,4,1,2,2,t7,t8)
      call dmpi_halo(kmax,mmk,kh,ia,4,1,2,5,t7,t8)
      call timer(2,'MPI_haloTF')

      ! compute t4, t5, t6 ----------------------------------------------------
      ! this computation relies on t7, t8, i.e. barrier here
      call c_cx_cy_cz (mmk,ind,n2d,kh,u,v,w,idx)
!$OMP BARRIER

      call timer(1,'MPI_haloTF')
      call dmpi_halo(kmax,mmk,kh,ia,1,1,1,3,t4,t5)
      call dmpi_halo(kmax,mmk,kh,ia,1,1,1,4,t4,t5)
      call dmpi_halo(kmax,mmk,kh,ia,1,1,1,2,t4,t5)
      call dmpi_halo(kmax,mmk,kh,ia,1,1,1,5,t4,t5)
      call timer(2,'MPI_haloTF')

      ! Finally, compute t -----------------------------------------------------
      ! this computation relies on tt, t1, t2, t3, t4, t5, t6, i.e. barrier here
      call advection(mmk,ind,n2d,kh,khu,khv,hx,hy,h_new,u,v,w,t,cosphi,idx)

      if(.not.dyn) then
!$OMP BARRIER
        ! do mass-correction on t ----------------------------------------------
        call masscorr ( nc, t )
      endif

    endif
    
    if (dodiff .and. doadvec) then
      ! barrier here due to fusion of two OMP parallel sections in case both
      ! advection and diffusion is activated in the same call
!$OMP BARRIER
    endif

    if (dodiff .and. present(vdiff)) then
      if (vdiff) then
        ! inlined version
        call diffusion_vi(kmax,mmk,ind,n2d,kh,h,h_new,t,avt,idx,q)
        ! exlined version
        !call diffusion_vx(kmax,mmk,ind,n2d,kh,h,h_new,t,avt,idx,q)
      else
        ! avoid data race for t(:,:) in diffusion() by making a copy:
        call copy_t2tt(mmk,ind,n2d,halo2d,kh,t,idx)
!$OMP BARRIER
        call diffusion_hx(mmk,ind,n2d,kh,h_new,hx,hy,t,eddyh,cosphi,idx,uvdam)
      endif

    elseif (dodiff) then
! MPI: Halo swap already done outside tflow-int for: t
! MPI: Halo of tt(:,:) also assigned in copy_t2tt(), i.e. no need for halo swap
      ! avoid data race for t(:,:) in diffusion() by making a copy:
      call copy_t2tt(mmk,ind,n2d,halo2d,kh,t,idx)
!$OMP BARRIER

! MPI: Halo swap already done outside tflow-int for: eddyh
      call diffusion_hx(mmk,ind,n2d,kh,h_new,hx,hy,t,eddyh,cosphi,idx,uvdam)
      ! inlined version
      call diffusion_vi(kmax,mmk,ind,n2d,kh,h,h_new,t,avt,idx)
      ! exlined version
      !call diffusion_vx(kmax,mmk,ind,n2d,kh,h,h_new,t,avt,idx)
    endif

  end subroutine tflow_int 
    
end module tflow

