module thermodynamic
  use constants, only : zero, one, two, ten, centi, milli, half, c2k
 
  implicit none
  private

  public  :: thermodyn, thermodyn_x, copy_iota2iotac
  private :: q_b,q_c,q_e,q_i,q_s,ocean,sea_ice,freeze,cl_func, swpen, ocean_x, &
             sea_ice_x, cl_func_x2, q_e_xw, q_e_xi, q_latent,                  &
             freeze_x, q_s_x, update_iota

  interface q_latent
    module procedure q_e_xw
    module procedure q_e_xi
  end interface

  real(8), parameter    :: albedos = 0.40_8    !soil albedo, the value was 
                                               !tuned to give reasonable re-
                                               !sults for the SST in shallow 
                                               !water, observed values are in 
                                               !the order of 0.1 !!!
  real(8), parameter    :: albedoO = 0.07_8    !ocean albedo
  real(8), parameter    :: albedoi = 0.60_8    !ice albedo, bare ice
  real(8), parameter    :: kapice  = 1.40_8    !extinction coef. for ice [1/m]
  real(8), parameter    :: CL_ice  = 0.0012_8  !exchange coef., latent heat, ice
  real(8), parameter    :: Rgas    = 287.1_8   ! the gas constant
  real(8), parameter    :: aw = 7.5_8, bw = 235.0_8
  real(8), parameter    :: ai = 9.5_8, bi = 265.0_8, ci = 0.00055_8
  real(8), parameter    :: e1 = 6.107_8, p1 = 0.622_8, p2 = 0.378_8
  real(8), parameter    :: tl = -8.0_8, tu = 7.0_8, wlim = 25.0_8
  real(8), parameter    :: MinIce = 1.e-6_8   !<== cut-off ice thickness

  real(8), parameter    :: limgrd = -0.003_8  !limiter for the iota gradient 
  real(8), parameter    :: h_fast = 0.1_8     !relative ice thickness
  real(8), parameter    :: ti_lim = 17.0_8    !depth limiter (off/near-shore)
  real(8), parameter    :: ci_offs = 22.5_8   !offshore intactness growth
  real(8), parameter    :: ci_nears = 50.0_8  !nearshore intactness growth
  real(8), parameter    :: ci_shallow = 15.0_8!shallow water intactness growth
  real(8), parameter    :: cL = 5.0_8         !cohesion length scale growth
  real(8), parameter    :: ck = 0.02_8        !fragility growth
  real(8), parameter    :: gk = 101.0_8       !fragility decay param.
  real(8), parameter    :: gi = 0.0002_8      !intactness gradual damaging
  real(8), parameter    :: gL = 0.0002_8      !cohesion length scale:
                                              !gradual damaging
!FIXME: is hs_lim different from MinIce for a reason ?
  real(8), parameter    :: hs_lim = 0.00001_8 !minimum ice thickness
  real(8), parameter    :: hssmi = 0.05_8     !min of rescaled ice thickness
  real(8), parameter    :: hssma = 1.0_8      !max of rescaled ice thickness
  real(8), parameter    :: hssa  = 0.25_8     !hss scale
  real(8), parameter    :: hssml = 0.005_8    !hss increment

  !-----------------------------------------------------------------------------
  ! Simple snow-ice short-wave radiation model
  !  - inspired by the CCSM3 and Fig. 4 and surrounding text in the CICE manual
  !
  real(8), parameter :: a1_bare_ice = 0.58_8, a0_bare_ice = 0.51_8, hi0 = 0.5_8
  real(8), parameter :: a1_max_snow = 0.85_8, a0_max_snow = 0.72_8, hs0 = 0.1_8
  real(8), parameter :: hspatch = 0.02_8, tsrfmin = -1.0_8 !, snoco0 = 0.03_8

  ! snowfall model:
  real(8), parameter :: maxsnowf = 0.000001_8
  real(8), parameter :: snowpos  = 0.1_8
  
contains

  !=============================================================================
  subroutine thermodyn_x(pen_type,mm,n2d,ind,xkoor,ykoor,h,atemp,hum,cloudi,   &
                         wu,wv,casus,cmp,tsnei,t_soil,dt,ice,heatflux,update,  &
                         kbh,evap,divrad,divrad_s,rho,pl,idx,                  &
                         iotamodel,cosphi,ti,uvdam,iota,iotac)

    !- modules  ----------------------------------------------------------------
    use dmi_omp,           only : domp_get_domain
    use cmod_params,       only : nsoilpar, h_soil, c_soil
    use constants,         only : rhoi, rhosoil, cpw, cpsoil, cfreeze, Lice,   &
                                  zeitstu, coldst, precevap, Q0, hnull
    use cmod_miscvars,     only : qforc
    
    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    logical,    intent(in)    :: update, iotamodel
    integer(4), intent(in)    :: pen_type
    integer(4), intent(in)    :: n2d
    integer(4), intent(in)    :: mm(:,0:,0:)
    integer(4), intent(in)    :: ind(:,:),kbh(0:)
    real(8),    intent(in)    :: atemp(0:),hum(0:)
    real(8),    intent(in)    :: cloudi(0:),wu(0:),wv(0:)
    real(8),    intent(in)    :: xkoor(:),ykoor(:),h(0:),ti(0:),dt
    real(8),    intent(in)    :: cosphi(:,0:)
    real(8),    intent(in)    :: uvdam(1:,0:)
    real(8),    intent(in)    :: rho(0:),pl(0:)
    real(8),    intent(in)    :: iotac(0:)
    real(8),    intent(inout) :: cmp(1:,0:)
    real(8),    intent(inout) :: heatflux(0:)
    logical,    intent(inout) :: casus(0:)
    real(8),    intent(inout) :: tsnei(0:)
    real(8),    intent(inout) :: ice(1:,0:)
    real(8),    intent(inout) :: t_soil(:,:)
    real(8),    intent(inout) :: evap(0:)
    real(8),    intent(inout) :: divrad(0:)
    real(8),    intent(inout) :: divrad_s(0:)
    integer(4), intent(inout) :: idx(:,:)
    real(8),    intent(inout) :: iota(1:,0:)
    
    !- localvars ---------------------------------------------------------------
    real(8), parameter    :: lhfus = Lice, geothermheat = -0.06_8
    real(8), parameter    :: RhoCp_soil = rhosoil*cpsoil
    real(8), parameter    :: lhfRhoi = lhfus*rhoi
    real(8)               :: tmelt, heat, hour
    real(8)               :: hs, dh, hl, ta, phi, rlam , humid, uw, vw, wind
    real(8)               :: cloud, q
    real(8)               :: amaxs, abarei, albs, albi, hsi, fsnow, sfac
    real(8)               :: qi, qilead, qi_m, qi_soil, fac, twater, tp, tm
    real(8)               :: dd, evapo
    real(8)               :: xtsnei, xeisco, xeisdi, xsnowd, rhow, RhoCp, pair
    logical               :: xcasus
    integer(4)            :: nn, i, j, k, kb, mi, l, mm2
    integer(4)            :: n2dl, n2du
        
    !  arrays:
    real(8)               :: q_soil(nsoilpar+1)   
  
    !  parameterization of penetrating insolation: -----------------------------
    !    Meier (JGR Vol.106, 2001) say for the short wave radiation into 
    !    the sea:
    !      I(z) = Qi * ( Rsw*exp(-z/zeta1) + (1-Rsw)*exp(-z/zeta2) )
    !    where z is vertical coordinate, positive downwards, and 
    !      d(Tw)/dt + advection_of_Tw = diffusion_of_T + 1/(rho*cpw)*dI/dz
    !    In the above zeta1 < zeta2 and Rsw < 1, and therefore the rapidly 
    !    decaying fraction of the solar isolation is Rsw and the not-so-rapidly 
    !    decaying fraction is (1-Rsw).
    !
    !   FIXME:
    !   If a non-negletable portion of the insolation reaches the bottom,
    !   this portion is just lost in the present implementation. It
    !   would probably be more correct to reflect it at the bottom.
    !
    !    Others use the so-called modified Beer's law:
    !      I(z) = (1-beta)*Qi*exp(-alpha*z) + beta*Qi*delta(z)
    !      delta() being the Dirach's delta function
    !    ie with a very rapidly decaying portion, beta, (the infrared) and a 
    !    less rapidly decay fraction, 1-beta, (the visible).
    !
    !    The modified Beer's law is justified through numerous measurements in
    !    lakes, while Meier's formulation is justified in the Baltic Sea.
    !    With values of alpha and beta in range of
    !      alpha = 0.5 - 1.4 1/m     (ie z_half = 0.5 - 1.5 m)
    !      beta  = 0.2 - 0.6
    !    the discrepancy might not be very important for practical applications.
    !---------------------------------------------------------------------------
    real(8), parameter :: Rsw = 0.64_8, onemRsw = one-Rsw
    real(8), parameter :: zeta1 = 1.78_8, zeta2 = 3.26_8
    real(8), parameter :: rzeta1 = one/zeta1, rzeta2 = one/zeta2
    real(8), parameter :: MaxDepth = 15.0_8*zeta2  !<== cut-off exp fct
  

    call domp_get_domain(kbh, 1, n2d, n2dl, n2du, idx)
  
    !- set on every entry:
    q_soil(nsoilpar+1) = geothermheat
    
    if (update .or. (.not.coldst)) then
      ! make sure to have sensible ice-model-values:
      do nn=n2dl, n2du
        tmelt = -cfreeze*cmp(2,nn)
        hs    = h(nn)
        RhoCp = rho(nn)*cpw

        if (ice(1,nn) > zero .and. casus(nn)) then
          heat = RhoCp*hs*(cmp(1,nn)-tmelt) - ice(1,nn)*lhfRhoi
          if (heat >= zero) then
            casus(nn) = .true.
            cmp(1,nn) = heat/(RhoCp*hs+rhoi*cpw*ice(1,nn)) + tmelt
            ice(1,nn) = zero
          else
            casus(nn) = .false.
            cmp(1,nn) = tmelt
            ice(1,nn) = -heat/lhfRhoi
          endif
        endif
  
        if (ice(1,nn) < MinIce .and. .not.casus(nn)) then
          casus(nn) = .true.
          cmp(1,nn) = max(tmelt, cmp(1,nn) - ice(1,nn)*lhfRhoi/(RhoCp*hs) )
          ice(1,nn) = zero
        endif
  
        if (casus(nn)) then
          ice(1,nn) = zero
          ice(2,nn) = zero
          ice(3,nn) = zero
          tsnei(nn) = tmelt
          if (iotamodel) then
             iota(1,nn) = zero
             iota(2,nn) = zero
             iota(3,nn) = zero
          end if
        else
          ice(2,nn) = max(min(ice(2,nn),one),milli)
          ice(3,nn) = max(MinIce,min(ice(3,nn),ice(1,nn)))
          tsnei(nn) = min(tsnei(nn),tmelt)
        endif
      enddo
    endif ! update
  
    if (.not.update) then
      do nn=n2dl, n2du
        heatflux(nn) = zero
      enddo
    endif
  
    if (qforc) then
      heatflux(n2dl:n2du) = Q0
      return
    endif
 
    hour    = zeitstu

    if (pen_type == 1) then
      surfaceloop1: do nn=n2dl, n2du

        kb = kbh(nn)
       
        if (kb == 1 .and. h(nn) < hnull) then
          q = zero
          heatflux(nn) = q
          if (precevap) evap(nn) = zero
          cycle surfaceloop1
        endif
      
        i = ind(1,nn)
        j = ind(2,nn)
      
        phi   = ykoor(i)
        rlam  = xkoor(j)
      
        hs     = max(h(nn),hnull)
        rhow   = rho(nn)
        ta     = atemp(nn)
        humid  = hum(nn)*milli
        uw     = wu(nn)
        vw     = wv(nn)
        wind   = sqrt(uw*uw + vw*vw)
        pair   = pl(nn)
        cloud  = cloudi(nn)
        tmelt  = -cfreeze*cmp(2,nn)
        twater = cmp(1,nn)
        xeisdi = ice(1,nn)
        xeisco = ice(2,nn)
        xsnowd = ice(3,nn)
        xcasus = casus(nn)
        xtsnei = tsnei(nn)

        RhoCp = rhow*cpw

        call q_i (phi,rlam,hour,cloud,humid,qi)

        if (update) then
          !- check if surface layer is frozen:
          if (twater < tmelt) then
            call freeze_x(xcasus,rhow,twater,hs,xtsnei,tmelt,xeisdi,xeisco,    &
                          hnull,MinIce)
            xeisdi = ice(1,nn) + xeisdi*(one-ice(2,nn))
            xeisco = one - (one-ice(2,nn))/(one+xeisco)
          endif
          if (xcasus) then
            !  no ice cover:
            fac = (one-albedoO)*qi*dt/RhoCp
          else
            !  free fraction of partially ice-covered grid cell:
            fac = (one-xeisco)*(one-albedoO)*qi*dt/RhoCp
          endif
          tm  = Rsw*exp(-hs*rzeta1) + onemRsw*exp(-hs*rzeta2)
          twater = twater + fac*(one-tm)/hs
        endif

        if (xcasus) then
          !  no ice cover:
          call ocean_x (cloud,humid,wind,xcasus,ta,twater,rhow,pair,q,evapo)
        else
          if (ta < zero) then
            sfac   = one - xsnowd/max(xeisdi,MinIce)
            xsnowd = min(xsnowd + sfac*maxsnowf*snowpos*cloud*dt, xeisdi)
          endif
          hsi    = xsnowd + xeisdi
          fsnow  = xsnowd/(xsnowd+hspatch)
          amaxs  = a0_max_snow + (a0_max_snow-a1_max_snow)*max(xtsnei,tsrfmin)
          abarei = a0_bare_ice + (a0_bare_ice-a1_bare_ice)*max(xtsnei,tsrfmin)
          if (hsi < hs0) then
            albs = amaxs  + (albedoO-amaxs )*((hsi-hs0)/hs0)**2
          else
            albs = amaxs
          endif
          if (hsi < hi0) then
            albi = abarei + (albedoO-abarei)*((hsi-hi0)/hi0)**2
          else
            albi = abarei
          endif
          qi = (fsnow*albs + (one-fsnow)*albi)*qi
          twater = twater + xeisco*qi*exp(-kapice*hsi)*dt/RhoCp/hs
          qi = qi*(one-exp(-kapice*hsi))
          call sea_ice_x (cloud,humid,wind,ta,xcasus,twater,rhow,pair,hs,dh,   &
                   xtsnei,tmelt,xeisdi,xsnowd,xeisco,hnull,q,qi,evapo,dt,MinIce)
!FIXME: prevent usage of iota model here until it has been fixed.
!         if (iotamodel .and. update) then
!            ! FIXME: It should increase the performance to move the if
!            ! statement out of the do loop.
!            call update_iota(nn,mm,uvdam,ti,cosphi,xeisdi,dh,dt,iota,iotac,i,j)
!         endif
        endif

        !   Save the evaporation rate:
        if (precevap) evap(nn) = evapo
  
        !   Save heat flux through sea surface for use in turbulence model,
        !   excl. insolation and excl. ice fraction (if any).
        !   Don't know at present how to do turbulence under ice anyway.
        heatflux(nn) = q
  
        !   Next surface point if we do not need updating:
        if (.not. update) cycle surfaceloop1
  
        !   OK, we need updating, therefore transfer twater to cmp(1,nn)
        !   and xeisdi, xeisco, xsnowd to ice(1:3,nn) :
        cmp(1,nn) = twater
        ice(1,nn) = xeisdi
        ice(2,nn) = xeisco
        ice(3,nn) = xsnowd
        casus(nn) = xcasus
        tsnei(nn) = xtsnei
      
        !   Insolation: Handle the sub-surfaces layers here - srf was done above
        if (kb > 1) then
          dd  = hs 
          tp  = tm*fac
          mm2 = mm(2,i,j) - 2
          do k=2,kb
            mi        = mm2 + k
            dd        = dd + h(mi)
            tm        = fac*(Rsw*exp(-dd*rzeta1) + onemRsw*exp(-dd*rzeta2))
            cmp(1,mi) = cmp(1,mi) + (tp-tm)/h(mi)
            if (dd >= MaxDepth) exit
            tp        = tm
          enddo
        else
          ! for safe use below for all kb=1,2,...
          mm2 = nn - kb
        endif
  
        !   Heat exchange with three-layer sea bed, k=kb:
        mi = mm2 + kb
        q_soil(1) = c_soil(1)*(cmp(1,mi)-t_soil(1,nn))/h_soil(1)
        do l=2,nsoilpar
          q_soil(l)=c_soil(l)*(t_soil(l-1,nn)-t_soil(l,nn))/h_soil(l)
        enddo
        cmp(1,mi) = cmp(1,mi) - dt*q_soil(1)/(RhoCp*max(h(mi),hnull))
        fac = dt/RhoCp_soil
        do l=1,nsoilpar
          t_soil(l,nn) = t_soil(l,nn) + fac*(q_soil(l)-q_soil(l+1))/h_soil(l)
        enddo
  
      enddo surfaceloop1

    elseif ( pen_type == 2 ) then
      ! --- penetration of shortwave radiation  ---------------
      call swpen (mm,xkoor,ykoor,hour,h,kbh,n2dl,n2du,ind,divrad,divrad_s)
    
      surfaceloop2: do nn=n2dl, n2du
        i = ind(1,nn)
        j = ind(2,nn)
      
        phi   = ykoor(i)
        rlam  = xkoor(j)
      
        hs    = max(h(nn),hnull)
  
        ta    = atemp(nn)
  
        humid = hum(nn)*milli
  
        uw  = wu(nn)
        vw  = wv(nn)
  
        wind = sqrt(uw*uw + vw*vw)
   
        pair   = pl(nn)
        cloud  = cloudi(nn)
  
        ! With snowfall=0.0 we have:
!        ice(3,nn) = zero

        kb = kbh(nn)
       
        tmelt  = -cfreeze*cmp(2,nn)
        twater = cmp(1,nn)
        xeisdi = ice(1,nn)
        xeisco = ice(2,nn)
        xsnowd = ice(3,nn)
        xcasus = casus(nn)
        xtsnei = tsnei(nn)
        rhow   = rho(nn)
        RhoCp  = rhow*cpw
        call q_i (phi,rlam,hour,cloud,humid,qi)

        if (update) then
          !- check if surface layer is frozen:
          if (twater < tmelt) then
            call freeze_x(xcasus,rhow,twater,hs,xtsnei,tmelt,xeisdi,xeisco,    &
                          hnull,MinIce)
            xeisdi = ice(1,nn) + xeisdi*(one-ice(2,nn))
            xeisco = one - (one-ice(2,nn))/(one+xeisco)
          endif
          if (xcasus) then
            !  no ice cover:
            fac    = (one-albedoO)*qi*dt/RhoCp
          else
            !  free fraction of partially ice-covered grid cell:
            fac = (one-xeisco)*(one-albedoO)*qi*dt/RhoCp
          endif
          twater = twater + divrad(nn)*fac/hs
        endif

        if (xcasus) then
          !  no ice cover:
          qi = (one-albedoO)*qi
          call ocean_x (cloud,humid,wind,xcasus,ta,twater,rhow,pair,q,evapo)
        else
!          xsnowd = min(snoco0,xeisdi)
          if (ta < zero) then
            sfac   = one - xsnowd/max(xeisdi,MinIce)
            xsnowd = min(xsnowd + sfac*maxsnowf*snowpos*cloud*dt, xeisdi)
          endif
          hsi    = xsnowd + xeisdi
          fsnow  = xsnowd/(xsnowd+hspatch)
          amaxs  = a0_max_snow + (a0_max_snow-a1_max_snow)*max(xtsnei,tsrfmin)
          abarei = a0_bare_ice + (a0_bare_ice-a1_bare_ice)*max(xtsnei,tsrfmin)
          if (hsi < hs0) then
            albs = amaxs  + (albedoO-amaxs )*((hsi-hs0)/hs0)**2
          else
            albs = amaxs
          endif
          if (hsi < hi0) then
            albi = abarei + (albedoO-abarei)*((hsi-hi0)/hi0)**2
          else
            albi = abarei
          endif
          qilead = (fsnow*albs + (one-fsnow)*albi)*qi
          twater = twater + xeisco*qilead*exp(-kapice*hsi)*dt/RhoCp/hs
          qilead = qilead*(one-exp(-kapice*hsi))

          qi = (one-albedoO)*(one-xeisco)*qi
          call sea_ice_x (cloud,humid,wind,ta,xcasus,twater,rhow,pair,hs,dh,   &
                   xtsnei,tmelt,xeisdi,xsnowd,xeisco,hnull,q,qilead,evapo,dt,  &
                   MinIce)
          if (iotamodel .and. update) then
             ! FIXME: It should increase the performance to move the if
             ! statement out of the do loop.
             call update_iota(nn,mm,uvdam,ti,cosphi,xeisdi,dh,dt,iota,iotac,i,j)
          endif
        endif

        !   Save the evaporation rate:
        if (precevap) evap(nn) = evapo
  
        !   Save heat flux through sea surface for use in turbulence model,
        !   excl. insolation and excl. ice fraction (if any).
        !   Don't know at present how to do turbulence under ice anyway.
        heatflux(nn) = q
  
        !   Next surface point if we do not need updating:
        if (.not. update) cycle surfaceloop2
  
        !   OK, we need updating, therefore transfer twater to cmp(1,nn)
        !   and xeisdi, xeisco, xsnowd to ice(1:3,nn) :
        cmp(1,nn) = twater
        ice(1,nn) = xeisdi
        ice(2,nn) = xeisco
        ice(3,nn) = xsnowd
        casus(nn) = xcasus
        tsnei(nn) = xtsnei
      
        !   Insolation: Handle the sub-surfaces layers here - srf was done above
        if (kb > 1) then
          mm2 = mm(2,i,j) - 2
          do k=2,kb
            mi = mm2 + k
            hl = max(h(mi),hnull)
            ! short wave radiation absorbed within layer k      
            qi_m = divrad(mi)*qi
            ! heating of layer k due to short wave radiation     
            cmp(1,mi) = cmp(1,mi) + dt*qi_m/RhoCp/hl
          enddo
        else
          ! for safe use below for all kb=1,2,...
          mm2 = nn - kb
        endif

        !   Heat exchange with three-layer sea bed, k=kb:
        qi_soil = divrad_s(nn)*qi      ! this part heats the soil
        t_soil(1,nn) = t_soil(1,nn) + dt * qi_soil /cpsoil /rhosoil /h_soil(1)

        mi = mm2 + kb
        q_soil(1) = c_soil(1)*(cmp(1,mi)-t_soil(1,nn))/h_soil(1) 
        do l=2,nsoilpar
          q_soil(l)=c_soil(l)*(t_soil(l-1,nn)-t_soil(l,nn))/h_soil(l)
        enddo
        
        hl = max(h(mi),hnull)
        cmp(1,mi) = cmp(1,mi) - dt*q_soil(1)/RhoCp/hl
        do l=1,nsoilpar
          t_soil(l,nn)=t_soil(l,nn) +                                          &
                         dt * (q_soil(l)-q_soil(l+1)) /RhoCp_soil /h_soil(l)
        end do 

      enddo surfaceloop2

    endif
   
  end subroutine thermodyn_x

  !=============================================================================

  subroutine ocean_x (cloud,humid,wind,casus,ta,twas,rhow,pair,q,e)
  
    use constants, only : Lwater, ct10m
    use t10m_mod,  only : t10m, qra
    
    implicit none                                              
  
    logical, intent(in)    :: casus
    real(8), intent(in)    :: cloud, humid, wind, ta, twas, rhow, pair
    real(8), intent(out)   :: q, e
  
  
    real(8) :: qlin, qlout, dqldts, qs, dqsdts, qe, dqedts, cl, dlr, pha, rhoa
    real(8) :: ta10, qa10, esat, qsea, tac
  
    pha  = pair*centi
    dlr  = -one/(Lwater*rhow)
    rhoa = pair/Rgas/(ta + c2k)
    esat = min(e1*ten**(aw*twas/(bw+twas)),pha)
    qsea = max(p1*esat/(pha-p2*esat), humid)

    call q_b (twas,ta,cloud,qlin,qlout,dqldts,casus)

    if (ct10m) then
      ! experiment: cast to 10m Tair and Qair, plus virtual Tair correction:
      if (tl < ta-twas .and. ta-twas < tu .and. wind < wlim) then
        call t10m(wind, twas, ta, qsea, humid, ta10, qa10)
        qsea = max(qsea, qa10)
        esat = e1*ten**(aw*ta10/(bw+ta10))
        tac  = ta10 - qra*(ta10+c2k)*(p1*esat/(pha-p2*esat)-qa10)
      else
        ta10 = ta
        qa10 = humid
        tac  = ta
      endif
      ! experiment: Full Kara et al 2005 corrected:
      call cl_func_x2 (twas,wind,tac,cl)
      call q_s_x (twas,ta10,wind,rhoa,qs,dqsdts,cl)
      call q_latent (twas,qa10,wind,pha,rhoa,esat,qsea,qe,dqedts,cl)
    else
      ! default:
      call cl_func (twas,wind,ta,cl)
      call q_s_x (twas,ta,wind,rhoa,qs,dqsdts,cl)
      call q_latent (twas,humid,wind,pha,rhoa,esat,qsea,qe,dqedts,cl)
    endif

    q = qlin + qlout + qs + qe
    e = qe*dlr
  
  end subroutine ocean_x

  !=============================================================================

  subroutine sea_ice_x (cloud,humid,wind,ta,casus,twas,rhow,pair,hs,dheis,tsrf,&
                        tmelt,heis,hsnow,eisco,hnull,q,qi,e,dt,MinIce)
  
  
    use constants, only : rhoi, cpw, cpi, Lwater, Lice, ct10m, rhos
    use t10m_mod,  only : t10m, qra
  
    implicit none
  
    real(8), intent(in)    :: cloud,humid,wind,ta,dt,hs,hnull,tmelt,MinIce,qi
    real(8), intent(in)    :: rhow, pair
    real(8), intent(inout) :: twas,tsrf,heis,eisco,hsnow
    real(8), intent(out)   :: q, e, dheis
    logical, intent(out)   :: casus
  
  
    integer(4), parameter :: itmax = 3 ! itmax > 1
    integer(4)            :: iter
    real(8), parameter :: lhfus = Lice
    real(8), parameter :: RhoiCpi = rhoi*cpi, RhoiLh = rhoi*lhfus
    real(8)            :: qlin, qlout, dqldts, qs, dqsdts, qe, dqedts
    real(8)            :: qocean, heat, told, qcond, deltahbb, deltahbs
    real(8)            :: dqcdts, dqdts, qlead, deltahl, qblock, esat, qsea
    real(8)            :: twodt, fac, eisfr, hedco, ei, ew, cl, RhoCp
    real(8)            :: qleada, hinew, dtemp, dlrw, dlri, pha, rhoa
    real(8)            :: ta10, qa10, tac, snoco, dsnow, hsi, tmp

    real(8), parameter :: hedco0 = 1.0e-10_8
    
  
    !- Some initialisations: ---------------------------------------------------
    pha   = pair*centi
    rhoa  = pair/Rgas/(ta + c2k)
    dlrw  = -one/(Lwater*rhow)
    dlri  = -one/((Lwater+Lice)*rhow)
    esat  = min(e1*ten**(aw*twas/(bw+twas)),pha)
    qsea  = max(p1*esat/(pha-p2*esat), humid)
    RhoCp = rhow*cpw
    dheis = zero
    dsnow = zero
    casus = .false.  ! Assume ice covered
    eisfr = one - eisco
    dtemp = zero
  
    !- Ocean part: -------------------------------------------------------------
    heat   = RhoCp*hs*(twas-tmelt)
    qocean = max(heat,zero)
  
  
    !- Blocked fraction: -------------------------------------------------------
    fac   = RhoiCpi*heis + rhos*hsnow*cpi
    !fac   = RhoiCpi*heis + rhos*snoco0*cpi
    twodt = two*dt
    hedco = max(hedco0, heis/max(milli,eisco) )
    snoco = max(hedco0, min(hsnow,heis)/max(milli,eisco) )
    told  = tsrf
    hsi   = hsnow*rhos/rhoi
    
    do iter=1,itmax
      call q_b (tsrf,ta,cloud,qlin,qlout,dqldts,casus)
      call q_s_x (tsrf,ta,wind,rhoa,qs,dqsdts,CL_ice)
      call q_latent (tsrf,humid,wind,pha,rhoa,qe,dqedts)
      call q_c (tsrf,tmelt,hedco,snoco,qcond,dqcdts)
      
      qblock = qi + qlin + qlout + qs + qe + qcond
      dqdts  = dqldts + dqsdts + dqedts + dqcdts
      tsrf   = (fac*told + twodt*(qblock-dqdts*tsrf))/(fac - twodt*dqdts)
      tsrf   = min(tsrf,zero)
    enddo
    deltahbb = (qcond*dt-qocean)/RhoiLh
    if ( deltahbb > zero ) then
      dheis = dheis + deltahbb*eisco
    else
      tmp   = deltahbb*eisco/(heis+hsi)
      dheis = dheis + tmp*heis
      dsnow = dsnow + tmp*hsi
    endif
    deltahbs = -(dt*qblock - fac*(tsrf-told)*half)/RhoiLh
    tmp      = deltahbs*eisco
    !dheis    = dheis  + min(deltahbs*eisco,zero)
    dheis    = dheis  + min(hsi+tmp,zero)
    dsnow    = dsnow - min(hsnow,max(-rhoi/rhos*tmp,zero))
    ei       = qe*dlri
    hinew    = heis + dheis
    if (hinew <= MinIce) then
      ! melting from above:
      dtemp = -min(hinew,zero)*RhoiLh/(hs*RhoCp)
      hinew = zero
      hsnow = zero
    else
      ! heat consumed from ocean below:
      dtemp = -eisco*qocean/(RhoCp*hs)
    endif

  
    !- Free fraction: ----------------------------------------------------------
    !  Here casus=.true. in calls to heat exchange routines.
    !  qi was already taken into account before calling this routine, so the
    !  corresponding heat is included into qocean.
    call q_b (twas,ta,cloud,qlin,qlout,dqldts,.true.)
    if (ct10m) then
      ! experiment: cast to 10m Tair and Qair, plus virtual Tair correction:
      if (tl < ta-twas .and. ta-twas < tu .and. wind < wlim) then
        call t10m(wind, twas, ta, qsea, humid, ta10, qa10)
        qsea = max(qsea, qa10)
        esat = e1*ten**(aw*ta10/(bw+ta10))
        tac  = ta10 - qra*(ta10+c2k)*(p1*esat/(pha-p2*esat)-qa10)
      else
        ta10 = ta
        qa10 = humid
        tac  = ta
      endif
      call cl_func_x2 (twas,wind,tac,cl)
      call q_s_x (twas,ta10,wind,rhoa,qs,dqsdts,cl)
      call q_latent (twas,qa10,wind,pha,rhoa,esat,qsea,qe,dqedts,cl)
    else
      ! default:
      call cl_func (twas,wind,ta,cl)
      call q_s_x (twas,ta,wind,rhoa,qs,dqsdts,cl)
      call q_latent (twas,humid,wind,pha,rhoa,esat,qsea,qe,dqedts,cl)
    endif

    qleada  = (qlin + qlout + qs + qe)
    qlead   = qleada*dt + qocean
    deltahl = -qlead/RhoiLh
    dheis   = dheis + deltahl*eisfr
    ew      = qe*dlrw
    hinew   = hinew + deltahl*eisfr
    if (hinew <= MinIce) then
      ! melting from below:
      !dtemp = dtemp - min(hinew,zero)*RhoiLh/(hs*RhoCp)
      hinew = zero
    ! hsnow = zero
    !else
      ! heat consumed from ocean:
      !dtemp = dtemp - eisfr*qocean/(RhoCp*hs)
    endif

 
    !- Both blocked and free fractions: ----------------------------------------
    ! not used: q = eisco*qblock + eisfr*qlead/dt
    e = eisco*ei + eisfr*ew


    !- Determine new ice cover fractions: --------------------------------------
    if (deltahl > zero) then
      eisco = one - eisfr*hnull/(hnull+deltahl)
    endif
    if (dheis < zero) then
      eisco = eisco*heis/(heis-half*dheis)
    endif


    !- Modify if ice has melted: -----------------------------------------------
    if (hinew < MinIce) then
      ! net heat budget melted all ice:
      casus = .true.
      heis  = zero
      eisco = zero
      hsnow = zero
      tsrf  = tmelt
    else
      heis = hinew
      hsnow = max(MinIce,min(hinew, hsnow + dsnow))
    endif


    !- Any remains of heat budget after freezing/melting ice must be added to T:
    twas = twas + dtemp


    !- Return only free fraction for use in turbulence model: ------------------
    q = eisfr*qleada
  
  end subroutine sea_ice_x

  !=============================================================================

  subroutine thermodyn(pen_type,mm,n2d,ind,xkoor,ykoor,h,atemp,hum,cloudi,     &
                       wu,wv,casus,cmp,tsnei,t_soil,dt,ice,heatflux,update,kbh,&
                       evap,divrad,divrad_s,snow,cosphi,ti,uvdam,              &
                       iotamodel,iota,iotac)
    !- modules  ----------------------------------------------------------------
    use dmi_omp,       only : domp_get_domain
    use cmod_params,   only : nsoilpar, h_soil, c_soil
    use constants,     only : rhow, rhoi, rhosoil, cpw, cpsoil, cfreeze, Lice, &
                              zeitstu, coldst, precevap, Q0, hnull
    use cmod_miscvars, only : qforc
  
    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    logical,    intent(in)    :: update, snow, iotamodel
    integer(4), intent(in)    :: pen_type
    integer(4), intent(in)    :: n2d
    integer(4), intent(in)    :: mm(:,0:,0:)
    integer(4), intent(in)    :: ind(:,:),kbh(0:)
    real(8),    intent(in)    :: atemp(0:),hum(0:)
    real(8),    intent(in)    :: cloudi(0:),wu(0:),wv(0:)
    real(8),    intent(in)    :: xkoor(:),ykoor(:),h(0:),ti(0:),dt
    real(8),    intent(in)    :: cosphi(:,0:)
    real(8),    intent(in)    :: uvdam(1:,0:)
    real(8),    intent(in)    :: iotac(0:)
    real(8),    intent(inout) :: cmp(1:,0:)
    real(8),    intent(inout) :: heatflux(0:)
    logical,    intent(inout) :: casus(0:)
    real(8),    intent(inout) :: tsnei(0:)
    real(8),    intent(inout) :: ice(1:,0:)
    real(8),    intent(inout) :: t_soil(:,:)
    real(8),    intent(inout) :: evap(0:)
    real(8),    intent(inout) :: divrad(0:)
    real(8),    intent(inout) :: divrad_s(0:)
    real(8),    intent(inout) :: iota(1:,0:)
    
    !- localvars ---------------------------------------------------------------
    real(8), parameter    :: lhfus = Lice, geothermheat = -0.06_8
    real(8), parameter    :: RhoCp = rhow*cpw, RhoCp_soil = rhosoil*cpsoil
    real(8), parameter    :: lhfRhoi = lhfus*rhoi
    real(8)               :: tmelt, heat, hour
    real(8)               :: hs, hl, ta, phi, rlam , humid, uw, vw, wind
    real(8)               :: cloud, q, sfac, dh
    real(8)               :: qi, qi_soil, qi_m, fac, twater, tp, tm, dd, evapo
    real(8)               :: xtsnei, xeisco, xeisdi, xsnowd
    logical               :: xcasus
    integer(4)            :: nn, i, j, k, kb, mi, l, mm2
    integer(4)            :: n2dl, n2du
        
    !  arrays:
    real(8)               :: q_soil(nsoilpar+1)   

    real(8), parameter :: MinIce = 1.e-6_8 !<== cut-off ice thickness
  
    !  parameterization of penetrating insolation (ONLY pen_type == 1): --------
    !    Meier (JGR Vol.106, 2001) say for the short wave radiation into 
    !    the sea:
    !      I(z) = Qi * ( Rsw*exp(-z/zeta1) + (1-Rsw)*exp(-z/zeta2) )
    !    where z is vertical coordinate, positive downwards, and 
    !      d(Tw)/dt + advection_of_Tw = diffusion_of_T + 1/(rho*cpw)*dI/dz
    !    In the above zeta1 < zeta2 and Rsw < 1, and therefore the rapidly 
    !    decaying fraction of the solar isolation is Rsw and the not-so-rapidly 
    !    decaying fraction is (1-Rsw).
    !
    !    Others use the so-called modified Beer's law:
    !      I(z) = (1-beta)*Qi*exp(-alpha*z) + beta*Qi*delta(z)
    !      delta() being the Dirach's delta function
    !    ie with a very rapidly decaying portion, beta, (the infrared) and a 
    !    less rapidly decay fraction, 1-beta, (the visible).
    !
    !    The modified Beer's law is justified through numerous measurements in
    !    lakes, while Meier's formulation is justified in the Baltic Sea.
    !    With values of alpha and beta in range of
    !      alpha = 0.5 - 1.4 1/m     (ie z_half = 0.5 - 1.5 m)
    !      beta  = 0.2 - 0.6
    !    the discrepancy might not be very important for practical applications.
    !---------------------------------------------------------------------------
    real(8), parameter :: Rsw = 0.64_8, onemRsw = one-Rsw
    real(8), parameter :: zeta1 = 1.78_8, zeta2 = 3.26_8
    real(8), parameter :: rzeta1 = one/zeta1, rzeta2 = one/zeta2
    real(8), parameter :: MaxDepth = 15.0_8*zeta2  !<== cut-off exp fct

    call domp_get_domain (1, n2d, n2dl, n2du)

    !- set on every entry:
    q_soil(nsoilpar+1) = geothermheat
    
    if (update .or. (.not.coldst)) then
      ! make sure to have sensible ice-model-values:
      do nn=n2dl, n2du
        tmelt = -cfreeze*cmp(2,nn)
        hs    = h(nn)

        if (ice(1,nn) > zero .and. casus(nn)) then
          heat = RhoCp*hs*(cmp(1,nn)-tmelt) - ice(1,nn)*lhfRhoi
          if (heat >= zero) then
            casus(nn) = .true.
            cmp(1,nn) = heat/(RhoCp*hs+rhoi*cpw*ice(1,nn)) + tmelt
            ice(1,nn) = zero
          else
            casus(nn) = .false.
            cmp(1,nn) = tmelt
            ice(1,nn) = -heat/lhfRhoi
          endif
        endif
  
        if (ice(1,nn) < MinIce .and. .not.casus(nn)) then
          casus(nn) = .true.
          cmp(1,nn) = max(tmelt, cmp(1,nn) - ice(1,nn)*lhfRhoi/(RhoCp*hs) )
          ice(1,nn) = zero
        endif
  
        if (casus(nn)) then
          ice(1,nn) = zero
          ice(2,nn) = zero
          ice(3,nn) = zero
          tsnei(nn) = tmelt
          if (iotamodel) then
             iota(1,nn) = zero
             iota(2,nn) = zero
             iota(3,nn) = zero
          end if
        else
          ice(2,nn) = max(min(ice(2,nn),one),milli)
          ice(3,nn) = max(MinIce,min(ice(3,nn),ice(1,nn)))
          tsnei(nn) = min(tsnei(nn),tmelt)
        endif
      enddo
    endif ! update
  
    if (.not.update) then
      do nn=n2dl, n2du
        heatflux(nn) = zero
      enddo
    endif
  
    if (qforc) then
      if (update) then 
        do nn=n2dl, n2du
          heatflux(nn) = Q0
          hs           = max(h(nn),hnull)
          cmp(1,nn)    = cmp(1,nn) + dt*Q0/(RhoCp*hs)
        enddo
      else
        do nn=n2dl, n2du
          heatflux(nn) = Q0
        enddo
      endif
      return
    endif
 
    hour    = zeitstu
    
    if (pen_type == 1) then
      surfaceloop1: do nn=n2dl, n2du
        i = ind(1,nn)
        j = ind(2,nn)
      
        phi   = ykoor(i)
        rlam  = xkoor(j)
      
        hs    = max(h(nn),hnull)

        ta    = atemp(nn)
  
        humid = hum(nn)*milli
  
        uw  = wu(nn)
        vw  = wv(nn)
  
        wind = sqrt(uw*uw + vw*vw)
   
        cloud  = cloudi(nn)
  
        kb = kbh(nn)
       
        if (kb == 1 .and. h(nn) < hnull) then
          q = zero
          heatflux(nn) = q
          if (precevap) evap(nn) = zero
          cycle surfaceloop1
        endif
      
        tmelt  = -cfreeze*cmp(2,nn)
        twater = cmp(1,nn)
        xeisdi = ice(1,nn)
        xeisco = ice(2,nn)
        xsnowd = ice(3,nn)
        xcasus = casus(nn)
        xtsnei = tsnei(nn)
        call q_i (phi,rlam,hour,cloud,humid,qi)
        if (casus(nn)) then
          call ocean(cloud,humid,wind,ta,xcasus,twater,hs,                     &
                     xtsnei,tmelt,xeisdi,xeisco,hnull,q,qi,evapo,dt,MinIce)
        else
          if (.not.snow) then
            xsnowd = zero
          elseif ( ta < zero .and. xeisdi > zero) then
            sfac   = one - xsnowd/max(xeisdi,MinIce)
            xsnowd = min(xsnowd + sfac*maxsnowf*snowpos*cloud*dt, xeisdi)
          endif
          call sea_ice(cloud,humid,wind,ta,xcasus,twater,hs,                   &
                       xtsnei,tmelt,xeisdi,xsnowd,xeisco,hnull,q,qi,evapo,dt,dh)
          if (iotamodel .and. update) then
             ! FIXME: It should increase the performance to move the if
             ! statement out of the do loop.
             call update_iota(nn,mm,uvdam,ti,cosphi,xeisdi,dh,dt,iota,iotac,i,j)
          endif
        endif

        !   Save the evaporation rate:
        if (precevap) evap(nn) = evapo
  
        !   Save heat flux through sea surface for use in turbulence model,
        !   excl. insolation and excl. ice fraction (if any).
        !   Don't know at present how to do turbulence under ice anyway.
        heatflux(nn) = q - qi
  
        !   Next surface point if we do not need updating:
        if (.not. update) cycle surfaceloop1
  
        !   OK, we need updating, therefore transfer twater to cmp(1,nn)
        !   and xeisdi, xeisco, xsnowd to ice(1:3,nn) :
        cmp(1,nn) = twater
        ice(1,nn) = xeisdi
        ice(2,nn) = xeisco
        ice(3,nn) = xsnowd
        casus(nn) = xcasus
        tsnei(nn) = xtsnei
      
        !   Insolation. 
        !
        !   FIXME:
        !   If a non-negletable portion of the insolation reaches the bottom,
        !   this portion is just lost in the present implementation. It
        !   would probably be more correct to reflect it at the bottom.
        fac       = qi*dt/RhoCp
        tm        = fac*(Rsw*exp(-hs*rzeta1) + onemRsw*exp(-hs*rzeta2))
        cmp(1,nn) = cmp(1,nn) - tm/hs
        dd        = hs 
        tp        = tm
        if (kb > 1) then
          mm2 = mm(2,i,j) - 2
        else
          ! for safe use below for all kb=1,2,...
          mm2 = nn - kb
        endif
        do k=2,kb
          mi        = mm2 + k
          dd        = dd + h(mi)
          tm        = fac*(Rsw*exp(-dd*rzeta1) + onemRsw*exp(-dd*rzeta2))
          cmp(1,mi) = cmp(1,mi) + (tp-tm)/h(mi)
          if (dd >= MaxDepth) exit
          tp        = tm
        enddo
  
        !   Update ice in surface layer
        if (casus(nn)) then
          !- check if surface layer is frozen:
          if (cmp(1,nn) < tmelt) then
            call freeze(casus(nn),cmp(1,nn),hs,tsnei(nn),tmelt,                &
                        ice(1,nn),ice(2,nn),hnull,MinIce)
            ice(2,nn) = one - one/(one+ice(2,nn))
          endif
        else
          call freeze(casus(nn),cmp(1,nn),hs,xtsnei,tmelt,xeisdi,xeisco,       &
                      hnull,MinIce)
          ice(1,nn) = ice(1,nn) + xeisdi*(one-ice(2,nn))
          ice(2,nn) = one - (one-ice(2,nn))/(one+xeisco)
        endif
  
        !   Heat exchange with three-layer sea bed, k=kb:
        mi = mm2 + kb
        q_soil(1) = c_soil(1)*(cmp(1,mi)-t_soil(1,nn))/h_soil(1)
        do l=2,nsoilpar
          q_soil(l)=c_soil(l)*(t_soil(l-1,nn)-t_soil(l,nn))/h_soil(l)
        enddo
        fac = dt/(RhoCp*max(h(mi),hnull))
        cmp(1,mi) = cmp(1,mi) - fac*q_soil(1)
        fac = dt/RhoCp_soil
        do l=1,nsoilpar
          t_soil(l,nn) = t_soil(l,nn) + fac*(q_soil(l)-q_soil(l+1))/h_soil(l)
        enddo
  
      enddo surfaceloop1

    elseif (pen_type == 2) then

      ! --- penetration of shortwave radiation  ---------------
      call swpen (mm,xkoor,ykoor,hour,h,kbh,n2dl,n2du,ind,divrad,divrad_s)
    
      surfaceloop2: do nn=n2dl, n2du
        i = ind(1,nn)
        j = ind(2,nn)
      
        phi   = ykoor(i)
        rlam  = xkoor(j)
      
        hs    = max(h(nn),hnull)

        ta    = atemp(nn)
  
        humid = hum(nn)*milli
  
        uw  = wu(nn)
        vw  = wv(nn)
  
        wind = sqrt(uw*uw + vw*vw)
   
        cloud  = cloudi(nn)
  
        kb = kbh(nn)
       
        tmelt  = -cfreeze*cmp(2,nn)
        twater = cmp(1,nn)
        xeisdi = ice(1,nn)
        xeisco = ice(2,nn)
        xsnowd = ice(3,nn)
        xcasus = casus(nn)
        xtsnei = tsnei(nn)
        call q_i (phi,rlam,hour,cloud,humid,qi)
        if (casus(nn)) then
          call ocean(cloud,humid,wind,ta,xcasus,twater,hs,                     &
                     xtsnei,tmelt,xeisdi,xeisco,hnull,q,qi,evapo,dt,MinIce)
        else
          if (.not.snow) then
            xsnowd = zero
          elseif ( ta < zero .and. xeisdi > zero) then
            sfac   = one - xsnowd/max(xeisdi,MinIce)
            xsnowd = min(xsnowd + sfac*maxsnowf*snowpos*cloud*dt, xeisdi)
          endif
          call sea_ice(cloud,humid,wind,ta,xcasus,twater,hs,                   &
                       xtsnei,tmelt,xeisdi,xsnowd,xeisco,hnull,q,qi,evapo,dt,dh)
          if (iotamodel .and. update) then
             ! FIXME: It should increase the performance to move the if
             ! statement out of the do loop.
             call update_iota(nn,mm,uvdam,ti,cosphi,xeisdi,dh,dt,iota,iotac,i,j)
          endif
        endif

        !   Save the evaporation rate:
        if (precevap) evap(nn) = evapo
  
        !   Save heat flux through sea surface for use in turbulence model,
        !   excl. insolation and excl. ice fraction (if any).
        !   Don't know at present how to do turbulence under ice anyway.
        heatflux(nn) = q - qi
  
        !   Next surface point if we do not need updating:
        if (.not. update) cycle surfaceloop2
  
        !   OK, we need updating, therefore transfer twater to cmp(1,nn)
        !   and xeisdi, xeisco, xsnowd to ice(1:3,nn) :
        cmp(1,nn) = twater
        ice(1,nn) = xeisdi
        ice(2,nn) = xeisco
        ice(3,nn) = xsnowd
        casus(nn) = xcasus
        tsnei(nn) = xtsnei
   
        fac       = qi*dt/RhoCp
        cmp(1,mm(1,i,j)) = cmp(1,mm(1,i,j))-(one-divrad(nn))*fac/hs
        do k=2,kb
          hl = max(h(mm(k,i,j)),hnull)
          !short wave radiation absorbed within layer k      
          qi_m = divrad(mm(k,i,j))*qi
          !heating of layer k due to short wave radiation     
          cmp(1,mm(k,i,j)) = cmp(1,mm(k,i,j))+dt*qi_m/RhoCp/hl
        enddo
  
        !   Update ice in surface layer
        if (casus(nn)) then
          !- check if surface layer is frozen:
          if (cmp(1,nn) < tmelt) then
            call freeze(casus(nn),cmp(1,nn),hs,tsnei(nn),tmelt,                &
                        ice(1,nn),ice(2,nn),hnull,MinIce)
          ice(2,nn) = one - one/(one+ice(2,nn))
          endif
        else
          call freeze(casus(nn),cmp(1,nn),hs,xtsnei,tmelt,xeisdi,xeisco,       &
                      hnull,MinIce)
          ice(1,nn) = ice(1,nn) + xeisdi*(one-ice(2,nn))
          ice(2,nn) = one - (one-ice(2,nn))/(one+xeisco)
        endif
 
        !   Heat exchange with three-layer sea bed, k=kb:
        qi_soil = divrad_s(mm(1,i,j))*qi      ! this part heats the soil
        t_soil(1,nn) = t_soil(1,nn) + dt * qi_soil /cpsoil /rhosoil /h_soil(1)

        q_soil(1) = c_soil(1)*(cmp(1,mm(kb,i,j))-t_soil(1,nn))/h_soil(1) 
        do l=2,nsoilpar
          q_soil(l)=c_soil(l)*(t_soil(l-1,nn)-t_soil(l,nn))/h_soil(l)
        enddo
        
        hl = max(h(mm(kb,i,j)),hnull)
        cmp(1,mm(kb,i,j))=cmp(1,mm(kb,i,j)) - dt*q_soil(1)/RhoCp/hl
        do l=1,nsoilpar
          t_soil(l,nn)=t_soil(l,nn) +                                          &
                         dt * (q_soil(l)-q_soil(l+1)) /RhoCp_soil /h_soil(l)
        end do 
            
      enddo surfaceloop2
    endif
  
  end subroutine thermodyn

  !=============================================================================
  
  subroutine q_b (ts,ta,cloud,qlin,qlout,dqldts,casus)

    implicit none
  
    logical, intent(in)  :: casus
    real(8), intent(in)  :: ts,ta,cloud
    real(8), intent(out) :: qlin,qlout,dqldts
    
    !
    !  s.b.idso & r.d.jackson, thermal radiation from the atmosphere,
    !       j. geophys. res. 74, 5397-5403, 1969
    !
    real(8), parameter :: sigma = 5.67e-8_8
    real(8), parameter :: eps0  = 0.99_8
    real(8), parameter :: epsi  = 0.97_8, minusfour = -4.0_8
    real(8), parameter :: a1 = 0.26_8, a2 = 0.75_8, a3 = -7.7e-4_8
  
    real(8) :: eps, tsa, taa, tsa3
  
    tsa  = ts + c2k
    taa  = ta + c2k
    tsa3 = tsa*tsa*tsa
    
    if (casus) then
      eps    = eps0
      dqldts = zero
    else
      eps    = epsi
      dqldts = minusfour*eps*sigma*tsa3
    endif
    
    qlin = eps*sigma*taa*taa*taa*taa                                           &
         * (one  - a1*exp(a3*ta*ta)                                            &
    !    * (1.e0 - 0.75e0*cloud**2) )
         * (one  - a2*cloud)    )
  
    qlout = -eps*sigma*tsa3*tsa
    
   
  end subroutine q_b

  !=============================================================================

  subroutine q_c (ts,tmelt,heis,hsnow,qcond,dqcdts)
  
    implicit none
  
    real(8), intent(in)  :: ts,tmelt,heis,hsnow
    real(8), intent(out) :: qcond,dqcdts 
       
    real(8), parameter :: zappai = 2.03_8
    real(8), parameter :: zappas = 0.31_8
  
    real(8) :: cond
    
    cond   = zappai*zappas/(hsnow*zappai+heis*zappas)
  
    qcond  = cond*(tmelt-ts)
    dqcdts =-cond
  
  end subroutine q_c

  !=============================================================================

  subroutine cl_func (ts, wind, ta, cl)

    use constants, only : CLA, CLB, CLC, CL1A, CL1B, CL1C

    implicit none

    real(8), intent(in)  :: ts, wind, ta
    real(8), intent(out) :: cl

    real(8)            :: wcut, dtcut, wct2
    real(8), parameter :: wl = 3.0_8, wu = 27.5_8, tdiffmax = 5.0_8

    dtcut = max(-tdiffmax,min(tdiffmax,(ts - ta)))
    wcut  = max(wl,min(wu,wind))
    wct2  = min(wu,wind)
    cl    = milli*(  ( CLA + ( CLB +  CLC*wct2)*wct2)                          &
                   + (CL1A + (CL1B + CL1C/wcut)/wcut)*dtcut )

  end subroutine cl_func

  !=============================================================================

  subroutine cl_func_x2 (ts, wind, ta, cl)

    implicit none

    real(8), intent(in)  :: ts, wind, ta
    real(8), intent(out) :: cl

    real(8), parameter :: wl  = 1.2_8,  wu = 5.0_8,  wm = 40.0_8
    real(8), parameter :: tdl = 0.75_8, tn = -0.098_8
    real(8), parameter :: CLMIN = milli*0.01_8, dtp = tdl-tn, dtm = tdl+tn

    ! Low wind speed:
    !   Neutral coefficient:
    real(8), parameter :: Cl00n= 1.14086_8, Cl01n=-0.00312_8, Cl02n=-0.00093_8
    !   Unstable range:
    real(8), parameter :: Cl00u= 2.07700_8, Cl01u=-0.39330_8, Cl02u= 0.03971_8
    real(8), parameter :: Cl10u=-0.28990_8, Cl11u= 0.07350_8, Cl12u=-0.006267_8
    real(8), parameter :: Cl20u=-0.01954_8, Cl21u= 0.005483_8,Cl22u=-0.0004867_8
    !   neutral-unstable range:
    real(8), parameter :: Cl00m = Cl00u + (-Cl10u + Cl20u*tdl)*tdl
    real(8), parameter :: Cl01m = Cl01u + (-Cl11u + Cl21u*tdl)*tdl
    real(8), parameter :: Cl02m = Cl02u + (-Cl12u + Cl22u*tdl)*tdl
    !   Stable range:
    real(8), parameter :: Cl00s=-0.29250_8,  Cl01s= 0.54980_8,Cl02s=-0.05544_8
    real(8), parameter :: Cl10s= 0.07272_8,  Cl11s=-0.17400_8,Cl12s= 0.02489_8
    real(8), parameter :: Cl20s=-0.006958_8, Cl21s= 0.01637_8,Cl22s=-0.002618_8
    !   neutral-stable range:
    real(8), parameter :: Cl00p = Cl00s + (Cl10s + Cl20s*tdl)*tdl
    real(8), parameter :: Cl01p = Cl01s + (Cl11s + Cl21s*tdl)*tdl
    real(8), parameter :: Cl02p = Cl02s + (Cl12s + Cl22s*tdl)*tdl

    ! High wind speed:
    !   Neutral coefficient:
    real(8), parameter :: Ch00n= 1.07300_8,Ch01n= 0.005531_8,Ch02n= 0.00005433_8
    !   Unstable range:
    real(8), parameter :: Ch00u= 1.07400_8,Ch01u= 0.005579_8,Ch02u= 0.00005263_8
    real(8), parameter :: Ch10u= 0.006912_8,  Ch11u=-0.22440_8, Ch12u=-1.02700_8
    real(8), parameter :: Ch20u= 0.0001849_8, Ch21u=-0.002167_8,Ch22u=-0.10100_8
    !   neutral-unstable range:
    real(8), parameter :: Ch00m = Ch00u + (-Ch10u + Ch20u*tdl)*tdl
    real(8), parameter :: Ch01m = Ch01u
    real(8), parameter :: Ch02m = Ch02u
    real(8), parameter :: Ch11m =         (-Ch11u + Ch21u*tdl)*tdl
    real(8), parameter :: Ch12m =         (-Ch12u + Ch22u*tdl)*tdl
    !   Stable range:
    real(8), parameter :: Ch00s= 1.02300_8,Ch01s= 0.009657_8,Ch02s=-0.00002281_8
    real(8), parameter :: Ch10s=-0.002672_8, Ch11s= 0.21030_8,  Ch12s=-5.32900_8
    real(8), parameter :: Ch20s= 0.001546_8, Ch21s=-0.06228_8,  Ch22s= 0.50940_8
    !   neutral-stable range:
    real(8), parameter :: Ch00p = Ch00s + (Ch10s + Ch20s*tdl)*tdl
    real(8), parameter :: Ch01p = Ch01s
    real(8), parameter :: Ch02p = Ch02s
    real(8), parameter :: Ch11p =         (Ch11s + Ch21s*tdl)*tdl
    real(8), parameter :: Ch12p =         (Ch12s + Ch22s*tdl)*tdl
    real(8)            :: wa, td, tc, wi, q

    ! set CL according to actual regime:
    td = ta - ts
    if (wind > wu) then
      wa = min(wm,wind)
      wi = one/wa

      if (td > tdl) then
        tc = min(td,tu)
        cl = milli*(    (Ch00s + (Ch01s + Ch02s*wa)*wa)                        &
                    + ( (Ch10s + (Ch11s + Ch12s*wi)*wi)                        &
                       +(Ch20s + (Ch21s + Ch22s*wi)*wi)*tc)*tc )
        cl = max(CLMIN,cl)
      elseif (td < -tdl) then
        tc = max(td,tl)
        cl = milli*(    (Ch00u + (Ch01u + Ch02u*wa)*wa)                        &
                    + ( (Ch10u + (Ch11u + Ch12u*wi)*wi)                        &
                       +(Ch20u + (Ch21u + Ch22u*wi)*wi)*tc)*tc )
      elseif (td >= tn) then ! tn <= td <= tdl
        q = -(td-tdl)/dtp
        cl = milli*(    (Ch00n + (Ch01n + Ch02n*wa)*wa)*q                      &
                    + ( (Ch00p + (Ch01p + Ch02p*wa)*wa)                        &
                               + (Ch11p + Ch12p*wi)*wi  )*(one-q) )
      else ! -tdl <= td < tn
        q = (td+tdl)/dtm
        cl = milli*(    (Ch00n + (Ch01n + Ch02n*wa)*wa)*q                      &
                    + ( (Ch00m + (Ch01m + Ch02m*wa)*wa)                        &
                               + (Ch11m + Ch12m*wi)*wi  )*(one-q) )
      endif

    else
      wa = max(wl,wind)

      if (td > tdl) then
        tc = min(td,tu)
        cl = milli*(    (Cl00s + (Cl01s + Cl02s*wa)*wa)                        &
                    + ( (Cl10s + (Cl11s + Cl12s*wa)*wa)                        &
                       +(Cl20s + (Cl21s + Cl22s*wa)*wa)*tc)*tc )
        cl = max(CLMIN,cl)

      elseif (td < -tdl) then
        tc = max(td,tl)
        cl = milli*(    (Cl00u + (Cl01u + Cl02u*wa)*wa)                        &
                    + ( (Cl10u + (Cl11u + Cl12u*wa)*wa)                        &
                       +(Cl20u + (Cl21u + Cl22u*wa)*wa)*tc)*tc )

      elseif (td >= tn) then ! tn <= td <= tdl
        q = -(td-tdl)/dtp
        cl = milli*(    (Cl00n + (Cl01n + Cl02n*wa)*wa)*q                      &
                    +   (Cl00p + (Cl01p + Cl02p*wa)*wa)*(one-q) )

      else ! -tdl <= td < tn
        q = (td+tdl)/dtm
        cl = milli*(    (Cl00n + (Cl01n + Cl02n*wa)*wa)*q                      &
                    +   (Cl00m + (Cl01m + Cl02m*wa)*wa)*(one-q) )
      endif

    endif

  end subroutine cl_func_x2

  !=============================================================================

  subroutine q_e_xw (ts,humid,wind,pair,rhoa,esat,qsat,qe,dqedts,cl)
    ! latent heat, water
  
    use constants, only : Lwater
  
    implicit none
  
    real(8), intent(in)  :: ts, humid, wind, pair, rhoa, esat, qsat, cl
    real(8), intent(out) :: qe, dqedts
  
    !
    !  w.t.liu, k.b.katsaros, & j.a.businger, bulk parametrization
    !       of air-sea exchanges of heat and water vapor including
    !       the molecular constraints at the interface,
    !       j. atmos. sci. 36, 1722-1735, 1979
    !
    real(8) :: desatdts, dqsatdesat, pfac, qfac, bts

    bts  = bw+ts
   
    pfac = p1/(pair - p2*esat)
    qfac = -Lwater*rhoa*cl*wind
    
    qe = qfac*(qsat-humid)
    
    desatdts = log(ten)*aw*bw*esat/(bts*bts)

    dqsatdesat = (pair/p1)*pfac*pfac
    
    dqedts = qfac*dqsatdesat*desatdts
  
    ! make sure to have sensible values:
    if (qe >= zero) then
      qe     = zero
      dqedts = zero
    endif
  
  end subroutine q_e_xw

  !=============================================================================

  subroutine q_e_xi (ts,humid,wind,pair,rhoa,qe,dqedts)
    ! latent heat, ice
  
    use constants, only : Lwater, Lice
  
    implicit none
  
    real(8), intent(in)  :: ts, humid, wind, pair, rhoa
    real(8), intent(out) :: qe, dqedts
  
    !
    !  w.t.liu, k.b.katsaros, & j.a.businger, bulk parametrization
    !       of air-sea exchanges of heat and water vapor including
    !       the molecular constraints at the interface,
    !       j. atmos. sci. 36, 1722-1735, 1979
    !
    real(8), parameter :: ell   = Lice + Lwater
  
    real(8) :: esat, qsat, desatdts, dqsatdesat, pfac, qfac, bts, lrc

    bts  = bi+ts
    lrc  = ell*rhoa*ci

    esat = min (e1*ten**(ai*ts/bts),pair)
  
    pfac = p1/(pair - p2*esat)
    qfac = -lrc*wind
    
    qsat = pfac*esat
  
    qe = qfac*(qsat-humid)
    
    desatdts = log(ten)*ai*bi*esat/(bts*bts)
    
    dqsatdesat = (pair/p1)*pfac*pfac
    
    dqedts = qfac*dqsatdesat*desatdts
  
    ! make sure to have sensible values:
    if (qe >= zero) then
      qe     = zero
      dqedts = zero
    endif
  
  end subroutine q_e_xi

  !=============================================================================

  subroutine q_e (ts,humid,wind,qe,dqedts,casus,cl)
  
    use constants, only : rhoa, pair, Lwater, Lice
  
    implicit none
  
    logical, intent(in)  :: casus
    real(8), intent(in)  :: ts,humid,wind,cl
    real(8), intent(out) :: qe,dqedts
  
    !
    !  w.t.liu, k.b.katsaros, & j.a.businger, bulk parametrization
    !       of air-sea exchanges of heat and water vapor including
    !       the molecular constraints at the interface,
    !       j. atmos. sci. 36, 1722-1735, 1979
    !
    real(8), parameter :: lhvap = Lwater
    real(8), parameter :: lhfus = Lice
    real(8), parameter :: pluft = pair
  
    real(8) :: a, b, ce, ell, esat, qsat, desatdts, dqsatdesat, pfac, qfac

    if (casus) then
      a   = aw
      b   = bw
      ce  = cl
      ell = lhvap
    else
      a   = ai
      b   = bi
      ce  = ci
      ell = lhfus + lhvap
    endif
   
    esat = e1*ten**(a*ts/(b+ts))
    esat = min (esat,pluft)
  
    pfac = p1/(pluft-p2*esat)
    qfac = -ell*rhoa*ce*wind
    
           ! qsat = 0.622e0*esat/pluft/(1.d00-0.378d00*esat/pluft)
    qsat = pfac*esat
  
           !qe = -ell*rhoa*ce*(qsat-humid)*abs(wind)
    qe = qfac*(qsat-humid)
    
    desatdts = log(ten)*a*b*esat/((b+ts)*(b+ts))
    
           ! dqsatdesat = 0.622e0/pluft/(1.d00-0.378d00*esat/pluft)**2.d00
    dqsatdesat = (pluft/p1)*pfac*pfac
    
           !dqedts = -ell*rhoa*ce*abs(wind)*dqsatdesat*desatdts
    dqedts = qfac*dqsatdesat*desatdts
  
    ! make sure to have sensible values:
    if (qe >= zero) then
      qe     = zero
      dqedts = zero
    endif
  
  end subroutine q_e

  !=============================================================================

  subroutine q_i (phi,rlam,hour,cloud,humid,qi)
  
    use constants, only : pair, pi
    
    implicit none
  
    real(8), intent(in)  :: phi,rlam,hour,cloud,humid
    real(8), intent(out) :: qi
    
    !
    !  Formel von Zillman, siehe Maykut,
    !       Introduction to Ice in the Polar Oceans, 1985, S.80,81
    !
    real(8), parameter :: twopi     = two*pi
    real(8), parameter :: q0        = 1350.0_8
    real(8), parameter :: sindek    = 0.3987_8    ! sindek = sin (23.5 grad)
    real(8), parameter :: pluft     = pair 
    real(8), parameter :: ddayspryr = one/365.2422_8, hpd = 24.0_8
    real(8), parameter :: eighty = 80.0_8
    real(8), parameter :: p3 = 0.6_8, p4 = 1.2_8, p5 = 0.045_8
  
    real(8) :: hd24, dekl, ea, s
     
    hd24 = hour/hpd
    
    dekl = asin( sindek*sin((hd24-eighty)*ddayspryr*twopi) )
    
    s = sin(phi)*sin(dekl)                                                     &
      + cos(phi)*cos(dekl)*cos((hd24-half)*twopi + rlam)
    s = max(s,zero)
  
    ea = pluft*humid/(p1 + p2*humid)
    
    qi = q0*s*s/(p4*s + (one+s)*ea*milli + p5) 
    qi = qi*(one - p3*cloud)

  end subroutine q_i

  !=============================================================================
  
  subroutine q_s_x (ts,ta,wind,rhoa,qs,dqsdts,cl)
  
    use constants, only : cpa, CSCL
  
    implicit none
  
    real(8), intent(in)  :: ts, ta, wind, rhoa, cl
    real(8), intent(out) :: qs, dqsdts
  
    dqsdts = -cpa*rhoa*wind*CSCL*cl   ! ch = cl*CSCL
  
    qs = -dqsdts*(ta-ts)       ! qs = cpa*rhoa*ch*(ta-ts)*abs(wind)
  
  end subroutine q_s_x

  !=============================================================================
  
  subroutine q_s (ts,ta,wind,qs,dqsdts,cl)
  
    use constants, only : rhoa, cpa, CSCL
  
    implicit none
  
    real(8), intent(in)  :: ts,ta,wind,cl
    real(8), intent(out) :: qs,dqsdts
  
    !
    !  w.t.liu, k.b.katsaros, & j.a.businger, bulk parametrization
    !       of air-sea exchanges of heat and water vapor including
    !       the molecular constraints at the interface,
    !       j. atmos. sci. 36, 1722-1735, 1979
    !
    real(8), parameter :: cparhoach = -cpa*rhoa

    dqsdts = cparhoach*wind*CSCL*cl
  
    qs = dqsdts*(ts-ta)       ! qs = -cpa*rhoa*ch*(ts-ta)*abs(wind)
  
  end subroutine q_s

  !=============================================================================

  subroutine ocean (cloud,humid,wind,ta,casus,twas,hs,                         &
                    tsnei,tmelt,heis,eisco,hnull,q,qi,e,dt,MinIce)
  
    use constants, only : rhow, cpw, Lwater
    
    implicit none                                              
  
    real(8), intent(in)    :: cloud,humid,wind,ta,dt
    real(8), intent(in)    :: hs,hnull,MinIce,tmelt
    real(8), intent(inout) :: twas,tsnei,heis,eisco,qi
    real(8), intent(out)   :: q, e
    logical, intent(out)   :: casus
  
  
    real(8), parameter :: RhoCp = rhow*cpw, dlr = -one/(Lwater*rhow)
    real(8)            :: qlin, qlout, dqldts, qs, dqsdts, qe, dqedts, cl
  
    casus = .true.
  
    qi = qi*(one - albedoO)
    call q_b (twas,ta,cloud,qlin,qlout,dqldts,casus)
    call cl_func (twas,wind,ta,cl)
    call q_s (twas,ta,wind,qs,dqsdts,cl)
    call q_e (twas,humid,wind,qe,dqedts,casus,cl)
      
    q    = qi + qlin + qlout + qs + qe
    twas = twas + dt*q/(RhoCp*hs)
    e    = qe*dlr
  
    if (twas < tmelt) then
      call freeze(casus,twas,hs,tsnei,tmelt,heis,eisco,hnull,MinIce)
    endif
  
  end subroutine ocean

  !=============================================================================

  subroutine sea_ice (cloud,humid,wind,ta,casus,twas,hs,tsrf,tmelt,            &
                  heis,hsnow,eisco,hnull,q,qilead,e,dt,dheis)
  
  
    use constants, only : rhow, rhoi, rhos, cpw, cpi, Lwater, Lice
  
    implicit none
  
    real(8), intent(in)    :: cloud,humid,wind,ta,dt
    real(8), intent(in)    :: hs,hnull,tmelt
    real(8), intent(inout) :: twas,tsrf,heis,eisco,hsnow,qilead
    real(8), intent(out)   :: dheis
    real(8), intent(out)   :: q, e
    logical, intent(out)   :: casus
  
  
    integer(4), parameter :: itmax = 3 ! itmax > 1
    integer(4)            :: iter
    real(8), parameter :: lhfus = Lice, RhoCp = rhow*cpw
    real(8), parameter :: RhoiCpi = rhoi*cpi, RhoiLh = rhoi*lhfus
    real(8), parameter :: dlrw = -one/(Lwater*rhow)
    real(8), parameter :: dlri = -one/((Lwater+Lice)*rhow)
    real(8)            :: qi, qlin, qlout, dqldts, qs, dqsdts, qe, dqedts
    real(8)            :: qocean, heat, told, qcond, deltahbb, deltahbs
    real(8)            :: dqcdts, dqdts, qlead, deltahl, qblock
    real(8)            :: twodt, fac, eisfr, hedco, ei, ew, cl
    real(8)            :: dqblock, dqlead, qleada, hinew
    real(8)            :: dhsnow, hsi

    real(8), parameter :: hedco0 = 1.0e-10_8
    
  
    !- Some initialisations: ---------------------------------------------------
    dheis = zero
    dhsnow= zero
    casus = .false.  ! Assume ice covered
    eisfr = one - eisco
  
    !- Ocean part: -------------------------------------------------------------
    heat   = RhoCp*hs*(twas-tmelt)
    qocean = max(heat,zero)
  
  
    !- Blocked fraction: -------------------------------------------------------
    fac   = RhoiCpi*heis + rhos*cpi*hsnow
    twodt = two*dt
    hedco = max(hedco0, heis/max(milli,eisco) )
    told  = tsrf
    
    qi = qilead*(one - albedoi)
  
    do iter=1,itmax
      call q_b (tsrf,ta,cloud,qlin,qlout,dqldts,casus)
      cl = CL_ice
      call q_s (tsrf,ta,wind,qs,dqsdts,cl)
      call q_e (tsrf,humid,wind,qe,dqedts,casus,cl)
      call q_c (tsrf,tmelt,hedco,hsnow,qcond,dqcdts)
      
      qblock = qi + qlin + qlout + qs + qe + qcond
      dqdts  = dqldts + dqsdts + dqedts + dqcdts
      tsrf   = (fac*told + twodt*(qblock-dqdts*tsrf))                          &
              /(fac      - twodt*dqdts)
      tsrf   = min(tsrf,tmelt)
    enddo
    deltahbb = (qcond*dt-qocean)/RhoiLh
    if ( deltahbb > zero ) then
      dheis    = dheis + deltahbb*eisco
    else
      hsi   = hsnow*rhos/rhoi
      dheis = dheis + deltahbb*eisco*(heis/(heis+hsi))
      dhsnow= dhsnow+ deltahbb*eisco*(hsi/(heis+hsi))
    endif

    deltahbs = -(dt*qblock - fac*(tsrf-told)*half)/RhoiLh
    dheis    = dheis + min(rhos/rhoi*hsnow+deltahbs*eisco,zero)
    dhsnow   = dhsnow - min(hsnow,max(-rhoi/rhos*deltahbs*eisco,zero))
    ei       = qe*dlri
    hinew    = heis + dheis
    if (hinew > zero) then
      dqblock = -eisco*qocean
    else
      dqblock = -hinew*RhoiLh
      hinew   = zero
    endif

  
    !- Free fraction: ----------------------------------------------------------
    !  here casus=.true. in calls to heat exchange routines
    qilead  = qilead*(one - albedoO)
    call q_b (twas,ta,cloud,qlin,qlout,dqldts,.true.)
    call cl_func (twas,wind,ta,cl)
    call q_s (twas,ta,wind,qs,dqsdts,cl)
    call q_e (twas,humid,wind,qe,dqedts,.true.,cl)
    qleada  = (qilead + qlin + qlout + qs + qe)
    qlead   = qleada*dt + qocean
    deltahl = -qlead/RhoiLh
    dheis   = dheis + deltahl*eisfr
    ew      = qe*dlrw
    hinew   = hinew + deltahl*eisfr
    if (hinew > zero) then
      dqlead = -eisfr*qocean
    else
      dqlead = -hinew*RhoiLh
      hinew  = zero
    endif

 
    !- Both blocked and free fractions: ----------------------------------------
    ! not used: q = eisco*qblock + eisfr*qlead/dt
    e = eisco*ei + eisfr*ew


    !- Determine new ice cover fractions: --------------------------------------
    if (deltahl > zero) then
      eisco = one - eisfr*hnull/(hnull+deltahl)
    endif
    if (dheis < zero) then
      eisco = eisco*heis/(heis-half*dheis)
    endif
  
  
    !- Determine new ice thickness: --------------------------------------------
    heis = hinew
  
    !- Determine new snow thickness: -------------------------------------------
    hsnow = min(hinew, max(hsnow + dhsnow, zero))
  
    !- Modify if ice has melted: -----------------------------------------------
    if (heis <= zero) then
      ! net heat budget melted all ice:
      casus = .true.
      heis  = zero
      eisco = zero
      hsnow = zero
      tsrf  = tmelt
    endif


    !- Remains of heat budget to alter T: --------------------------------------
    twas = max(tmelt, twas + (dqblock + dqlead)/(hs*RhoCp) )


    !- Return only free fraction for use in turbulence model: ------------------
    qilead = (one-eisco)*qilead
    q      = (one-eisco)*qleada
  

  end subroutine sea_ice

  !=============================================================================
  
  subroutine freeze_x(casus,rhow,twas,hs,tsnei,tmelt,heis,eisco,hnull,MinIce)

    use constants, only : rhoi, cpw, Lice
  
    implicit none                                              

    real(8), intent(in)    :: hs,hnull,tmelt,MinIce,rhow
    real(8), intent(inout) :: twas,tsnei
    real(8), intent(out)   :: heis,eisco
    logical, intent(inout) :: casus


    real(8), parameter :: lri = Lice*rhoi
    real(8)            :: qdef


    qdef = min(zero,twas-tmelt)*hs*rhow*cpw
    heis = -qdef/lri

    if (heis >= MinIce) then
      casus = .false.
      twas  = tmelt
      tsnei = tmelt
    else
      heis  = zero
    endif
  
    eisco = heis/hnull

  end subroutine freeze_x

  !=============================================================================
  
  subroutine freeze(casus,twas,hs,tsnei,tmelt,heis,eisco,hnull,MinIce)

    use constants, only : rhow, rhoi, cpw, Lice
  
    implicit none                                              

    real(8), intent(in)    :: hs,hnull,tmelt,MinIce
    real(8), intent(inout) :: twas,tsnei
    real(8), intent(out)   :: heis,eisco
    logical, intent(inout) :: casus


    real(8), parameter :: lhfus = Lice, RhoCp = rhow*cpw
    real(8)            :: qdef


    qdef = min(zero,twas-tmelt)*hs*RhoCp
    heis = -qdef/(rhoi*lhfus)

    if (heis >= MinIce) then
      casus = .false.
      twas  = tmelt
      tsnei = tmelt
    else
      heis  = zero
    endif
  
    eisco = heis/hnull

  end subroutine freeze

  !=============================================================================

  subroutine swpen(mm,xkoor, ykoor, hour, h, kh, nl, nu, ind , divrad, divrad_s)

  !-----------------------------------------------------------------------------
  !Calculation of the attenuation of shortwave radition in a 
  !water column in dependence on the chlorophyll-a  concentration. This is an
  !implementation of the optical model published by Morel and Antoine (1994).

  !Reference:

  !Morel, A., D. Antoine (1994):
  !Heating Rate within the Upper Ocean in Relation to its Bio-optical State.
  !Journal of Physical Oceanography, 24, 1652-1665
  !-----------------------------------------------------------------------------
    
    !- modules -----------------------------------------------------------------
    use constants, only : pi

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)         :: mm(:,0:,0:)
    integer(4), intent(in)         :: kh(0:)
    integer(4), intent(in)         :: nl, nu
    integer(4), intent(in)         :: ind(:,:)
    real(8), intent(in)            :: xkoor(:)
    real(8), intent(in)            :: ykoor(:)
    real(8), intent(in)            :: hour
    real(8), intent(in)            :: h(0:)
    real(8), intent(inout)         :: divrad(0:)
    real(8), intent(inout)         :: divrad_s(0:)

    !- local vars --------------------------------------------------------------
    real(8)                        :: rad1, rad2, rad3
    real(8)                        :: transpar1, transpar2, transpar3
    real(8)                        :: absorb1, absorb2, absorb3
    real(8)                        :: cloro
    real(8)                        :: phi, rlam, hd24, dekl, sinsolar
    integer(4)                     :: i,j,k, nn, mi, kb
    
    !- local parameters --------------------------------------------------------
    real(8), parameter             :: twopi = two*pi
    real(8), parameter             :: sindek    = 0.3987_8 
                                         ! sindek = sin (23.5 grad)
    real(8), parameter             :: ddayspryr = one/365.2422_8
    real(8), parameter             :: min_cloro = 0.02_8, max_cloro = 60.0_8
    real(8), parameter             :: min_rad = 0.0001_8, max_exp = 20.0_8
    real(8), parameter             :: attenuation_len = 0.267_8
                                    !- attenuation length is adjusted in such a 
                                    !- way taht, when the sun is at zenith, 99%
                                    !- of the infrared radiation is absorbed 
                                    !- within a layer extending down to 1.23 m
    

    !- coefficients for optical model taken from Morel and Antoine (1994)
    real(8)  :: V1, V2, Z1, Z2
    real(8)  :: C, C2, C3, C4, C5 
    real(8)  :: frac_vis

    real(8), parameter :: hpd = 24.0_8, eighty = 80.0_8, frac_vis0 = 0.57_8
    
    real(8), parameter ::                                                      &
      V1_coef(6) = (/0.353_8,-0.047_8, 0.083_8, 0.047_8,-0.011_8,-0.009_8/),   &
      V2_coef(6) = (/0.647_8, 0.047_8,-0.083_8,-0.047_8, 0.011_8, 0.009_8/),   &
      Z1_coef(6) = (/1.662_8,-0.605_8, 0.128_8,-0.033_8,-0.051_8,-0.004_8/),   &
      Z2_coef(6) = (/8.541_8,-8.924_8, 4.020_8,-0.077_8,-0.536_8, 0.055_8/)
    
    ! frac_vis is the fraction of light in the shortwave verses the long wave.
    ! frac_vis=0.54 on sunny days and frac_vis = 0.60 on cloudy days.
    frac_vis = frac_vis0

!$OMP MASTER 
    divrad(0)   = zero
    divrad_s(0) = zero
!$OMP END MASTER 

    do nn = nl, nu
       
      kb = kh(nn)
       
      i = ind(1,nn)
      j = ind(2,nn)
      
      !- calculate the sinus of the solar elevation angle, which equals the 
      !- cosinus of the zenith angle, since zenith angle is sun elevation angle
      !- minus 90 degree
      phi  = ykoor(i)
      rlam = xkoor(j)
      hd24 = hour/hpd
      dekl = asin( sindek*sin((hd24-eighty)*ddayspryr*twopi) )
      sinsolar = sin(phi)*sin(dekl)                                            &
               + cos(phi)*cos(dekl)*cos((hd24-half)*twopi + rlam)
      sinsolar = max(sinsolar,zero)
            
      rad1  = one
      rad2  = one
      rad3  = one

      !- absorption of downward radiation
      do k=1,kb
        mi = mm(k,i,j)
        if (mi > 0) then
          divrad(mi) = zero
      
        !- chlorophyll-a from data could be read here, cloro = chla(i,j,k) -
        !- afterwards the cloro values of the water column should be saved for
        !- upward radiation!
        !- As a first implementation step we assume constant chlorophyll-a of 
        !- 2 mg m-3
       
          cloro = two

          C = log10(min(max(cloro,min_cloro),max_cloro))
        
          C2 = C  * C
          C3 = C2 * C
          C4 = C3 * C
          C5 = C4 * C
              
          V1 = V1_coef(1) + V1_coef(2) * C + V1_coef(3) * C2  +                &
               V1_coef(4) * C3 + V1_coef(5) * C4 + V1_coef(6) * C5
          V2 = V2_coef(1) + V2_coef(2) * C + V2_coef(3) * C2  +                &
               V2_coef(4) * C3 + V2_coef(5) * C4 + V2_coef(6) * C5
          Z1 = Z1_coef(1) + Z1_coef(2) * C + Z1_coef(3) * C2  +                &
               Z1_coef(4) * C3 + Z1_coef(5) * C4 + Z1_coef(6) * C5
          Z2 = Z2_coef(1) + Z2_coef(2) * C + Z2_coef(3) * C2  +                &
               Z2_coef(4) * C3 + Z2_coef(5) * C4 + Z2_coef(6) * C5

          if ((h(mi)/Z1) > max_exp) then
            transpar1 = zero
           else
            transpar1 = exp(-h(mi)/Z1)
          endif

          if ((h(mi)/Z2) > max_exp) then
            transpar2 = zero
          else
            transpar2 = exp(-h(mi)/Z2)
          endif

          if (sinsolar > zero) then
            if (h(mi)/((attenuation_len*sinsolar)) > max_exp) then
              transpar3 = zero
            else
              transpar3 = exp(-h(mi)/(attenuation_len*sinsolar))
            endif
          else
            transpar3 = zero
          endif


          absorb1 = one - transpar1
          absorb2 = one - transpar2
          absorb3 = one - transpar3

          divrad(mi)= frac_vis*(V1*rad1*absorb1+V2*rad2*absorb2) +             &
                      (one-frac_vis)*rad3*absorb3
          !- fraction of short wave radiation that is absorbed in layer kl 
          !- due to heating of water, biological and chemical processes

          rad1 = rad1*transpar1
          if (rad1 < min_rad) rad1 = zero
          rad2 = rad2*transpar2
          if (rad2 < min_rad) rad2 = zero
          rad3 = rad3*transpar3
          if (rad3 < min_rad) rad3 = zero

        endif
      enddo
      
      mi = mm(1,i,j)
      if (mi > 0) then
        divrad_s(mi) = zero

        divrad_s(mi) = (frac_vis*(V1*rad1+V2*rad2)+(one-frac_vis)*rad3) *      &
                       (one-albedos)
        !- absorption in soil
      endif

      !absorption of upward reflected radiation
      do k = kb,1,-1
        mi  = mm(k,i,j)
        if (mi > 0) then

!FIXME: Here something should disappear - calculation were made twice!!
          cloro = two

          C = log10(min(max(cloro,min_cloro),max_cloro))
        
          C2 = C  * C
          C3 = C2 * C
          C4 = C3 * C
          C5 = C4 * C
              
          V1 = V1_coef(1) + V1_coef(2) * C + V1_coef(3) * C2  +                &
               V1_coef(4) * C3 + V1_coef(5) * C4 + V1_coef(6) * C5
          V2 = V2_coef(1) + V2_coef(2) * C + V2_coef(3) * C2  +                &
               V2_coef(4) * C3 + V2_coef(5) * C4 + V2_coef(6) * C5
          Z1 = Z1_coef(1) + Z1_coef(2) * C + Z1_coef(3) * C2  +                &
               Z1_coef(4) * C3 + Z1_coef(5) * C4 + Z1_coef(6) * C5
          Z2 = Z2_coef(1) + Z2_coef(2) * C + Z2_coef(3) * C2  +                &
               Z2_coef(4) * C3 + Z2_coef(5) * C4 + Z2_coef(6) * C5

          if ((h(mi)/Z1) > max_exp) then
            transpar1 = zero
          else
            transpar1 = exp(-h(mi)/Z1)
          endif

          if ((h(mi)/Z2) > max_exp) then
            transpar2 = zero
          else
            transpar2 = exp(-h(mi)/Z2)
          endif

          if (sinsolar > zero) then
            if ((h(mi)/(attenuation_len*sinsolar)) > max_exp) then
              transpar3 = zero
            else
              transpar3 = exp(-h(mi)/(attenuation_len*sinsolar))
            endif
          else
            transpar3 = zero
          endif

          absorb1 = one-transpar1
          absorb2 = one-transpar2
          absorb3 = one-transpar3

          divrad(mi) = divrad(mi) + frac_vis*(V1*rad1*absorb1+V2*rad2*absorb2) &
                                  + (one-frac_vis)*rad3*absorb3

          rad1 = rad1*transpar1
          if (rad1 < min_rad) rad1 = zero
          rad2 = rad2*transpar2
          if (rad2 < min_rad) rad2 = zero
          rad3 = rad3*transpar3
          if (rad3 < min_rad) rad3 = zero

        endif
      enddo
    enddo

  end subroutine swpen
      
  !=============================================================================

  subroutine update_iota(nn,mm,uvdam,ti,cosphi,hs,dh,dt,iota,iotac,i,j)
    
    !- modules  ----------------------------------------------------------------
    use constants, only : qrt,iota0,iscale,smfi,smfk,dx_iota,dy_iota,ddys,     &
                          ddyssmi,ddyssmk
    
    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: nn, i, j
    integer(4), intent(in)    :: mm(:,0:,0:)
    real(8),    intent(in)    :: uvdam(1:,0:)
    real(8),    intent(in)    :: ti(0:)
    real(8),    intent(in)    :: cosphi(:,0:)
    real(8),    intent(in)    :: hs, dh, dt
    real(8),    intent(in)    :: iotac(0:)
    real(8),    intent(inout) :: iota(1:,0:)
    
    !- localvars ---------------------------------------------------------------
    integer(4) :: mnw, mnn, mne, mww, mi, mee, msw, mss, mse
    real(8)    :: onland, ddxs, ddxys, ddxssmi, ddxyssmi, ddxssmk, ddxyssmk

    integer(4) :: uvdne, uvdnn, uvdnw, uvdww, uvdee, uvdsw, uvdss, uvdse, im
    real(8)    :: deei, dwwi, dnni, dssi, dnei, dnwi, dsei, dswi
    real(8)    :: dcmp1, dcmp2, dcmp3, dcmp4, ci_ns
    real(8)    :: ni_i, ni_k, hss, iota1, iota2, iota3

    onland = iota0

    im = nint(iotac(nn)*iscale)
        
    mnw = mm(1,i-1,j-1)  
    mnn = mm(1,i-1,j)    
    mne = mm(1,i-1,j+1)  
    mww = mm(1,i,j-1)    
    mi  = nn
    mee = mm(1,i,j+1)    
    msw = mm(1,i+1,j-1)  
    mss = mm(1,i+1,j)    
    mse = mm(1,i+1,j+1)

    ! temporary storage of input:
    iota1 = iota(1,nn)
    iota2 = iota(2,nn)
    iota3 = iota(3,nn)
    
    !---------------------------------------------------------------------------
    uvdne = nint( (uvdam(1,mnn)*iotac(mne)                                     &
                 +(one-uvdam(1,mnn))*onland) * iscale)  
    uvdnw = nint( (uvdam(1,mnw)*iotac(mnw)                                     &
                 +(one-uvdam(1,mnw))*onland) * iscale)
    uvdsw = nint( (uvdam(2,msw)*iotac(msw)                                     &
                 +(one-uvdam(2,msw))*onland) * iscale)
    uvdse = nint( (uvdam(2,mss)*iotac(mse)                                     &
                 +(one-uvdam(2,mss))*onland) * iscale)
    
    uvdee = nint( (uvdam(1,mi )*iotac(mee)                                     &
                 +(one-uvdam(1,mi) )*onland) * iscale)
    uvdww = nint( (uvdam(1,mww)*iotac(mww)                                     &
                 +(one-uvdam(1,mww))*onland) * iscale)
    uvdnn = nint( (uvdam(2,mnn)*iotac(mnn)                                     &
                 +(one-uvdam(2,mnn))*onland) * iscale)
    uvdss = nint( (uvdam(2,mi )*iotac(mss)                                     &
                 +(one-uvdam(2,mi) )*onland) * iscale)
      
    dnei = real(uvdne - im, 8)
    dnwi = real(uvdnw - im, 8)
    dsei = real(uvdse - im, 8)
    dswi = real(uvdsw - im, 8)
    
    deei = real(uvdee - im, 8)
    dwwi = real(uvdww - im, 8)
    dnni = real(uvdnn - im, 8)
    dssi = real(uvdss - im, 8)

    ddxs     = half/(iscale*dx_iota*cosphi(1,i))
    ddxys    = half/(iscale*sqrt((dx_iota*cosphi(1,i))**2+dy_iota**2))
    ddxssmi  = smfi*ddxs
    ddxyssmi = smfi*ddxys
    ddxssmk  = smfk*ddxs
    ddxyssmk = smfk*ddxys

    ! (1.) Gradient for the iota equation:
    !---------------------------------------------------------------------------
    dcmp1 = (abs(deei)+abs(dwwi))*ddxs  + (deei+dwwi)*ddxssmi
    dcmp2 = (abs(dnni)+abs(dssi))*ddys  + (dnni+dssi)*ddyssmi
    dcmp3 = (abs(dnei)+abs(dswi))*ddxys + (dnei+dswi)*ddxyssmi
    dcmp4 = (abs(dnwi)+abs(dsei))*ddxys + (dsei+dnwi)*ddxyssmi
    if (dcmp1 < limgrd) dcmp1=zero
    if (dcmp2 < limgrd) dcmp2=zero
    if (dcmp3 < limgrd) dcmp3=zero
    if (dcmp4 < limgrd) dcmp4=zero
    ni_i = iota2 * qrt * (dcmp1+dcmp2+dcmp3+dcmp4) * dt

    ! (2.) Gradient for the Kappa equation:
    !---------------------------------------------------------------------------
    dcmp1 = (deei+dwwi)*ddxssmk  - (abs(deei)+abs(dwwi))*ddxs
    dcmp2 = (dnni+dssi)*ddyssmk  - (abs(dnni)+abs(dssi))*ddys
    dcmp3 = (dnei+dswi)*ddxyssmk - (abs(dnei)+abs(dswi))*ddxys
    dcmp4 = (dsei+dnwi)*ddxyssmk - (abs(dnwi)+abs(dsei))*ddxys
    if (dcmp1 > limgrd) dcmp1=zero
    if (dcmp2 > limgrd) dcmp2=zero
    if (dcmp3 > limgrd) dcmp3=zero
    if (dcmp4 > limgrd) dcmp4=zero
    ni_k = half * (abs(dcmp1)+abs(dcmp2)+abs(dcmp3)+abs(dcmp4)) * dt 

    ! (3.) Ice growing conditions 
    !---------------------------------------------------------------------------
    if (dh > zero) then
      ! (3.1) Intactness increase
      if (iota1 <= iota0 .and. ti(nn) > ti_lim) then
        iota1 = iota1 + (ci_offs * dh/h_fast)*ni_i
      elseif (iota1 <= iota0 .and. ti(nn) <= ti_lim) then
        ci_ns = ci_nears + (one-ti(nn)/ti_lim) * ci_shallow
        iota1 = iota1 + (ci_ns * dh/h_fast)*ni_i
      elseif (iota1 > iota0) then
        iota1 = iota1 + (dh/h_fast)*ni_i
      endif

      ! (3.2) Cohesion length scale increase
      iota2 = iota2 + dh*(cL*iota1+one) 
            
      ! (3.3) Fragility reduction due to the growth of new ice
      iota3 = iota3/(dh*gk/max(hs,hs_lim)+one)
    endif

    ! (4.) gradual damaging (always active)
    !---------------------------------------------------------------------------
    ! (4.1) increase of fragility
    hss = hssmi+hssml*min(hssma, hssa * hs)               
    iota3 = iota3 + ( ck/(iota1*hss/h_fast+one) ) * ni_k 

    ! (4.2) Reduction of intactness due to gradual damaging
    iota1 = max(zero, iota1 - dt * gi * iota3)      
    
    ! (4.3) Reduction of the length scale due to gradual damaging
    iota2 = iota2/(dt * gL * iota3 + one)


    ! transfer values to output:
    iota(1,nn) = iota1
    iota(2,nn) = iota2
    iota(3,nn) = iota3
    
  end subroutine update_iota
  
  !=============================================================================

  subroutine copy_iota2iotac (n2d, n2dhalo, kh, idx, iota, iotac)

    !- modules -----------------------------------------------------------------
    use dmi_omp,   only : domp_get_domain
    use constants, only : iota0

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: n2d, n2dhalo
    integer(4), intent(in)    :: kh(0:)
    integer(4), intent(inout) :: idx(:,:)
    real(8),    intent(in)    :: iota(1:,0:)
    real(8),    intent(out)   :: iotac(0:)

    !- local vars --------------------------------------------------------------
    integer(4)  :: nl, nu, nlh, nuh

    call domp_get_domain(kh, 1, n2d, nl, nu, idx)

    iotac(nl:nu) = iota(1,nl:nu)

!$OMP MASTER
    iotac(0) = iota0

    ! handle the halo:
    if (n2dhalo > 0) then
      nlh = n2d+1
      nuh = n2d+n2dhalo
      iotac(nlh:nuh) = iota(1,nlh:nuh)
    endif
!$OMP END MASTER

  end subroutine copy_iota2iotac

end module thermodynamic
