module tideparam
  ! Container for handling of parameterization of effects of tides.
  !
  !   How to:
  !
  !   In options.nml:
  !     &optionlist
  !       tidetype = $n
  !     /
  !        with $n > 0
  !
  ! tidetype ==
  !       0: none
  !       1: unresolved bottom shear 
  !      10: internal, baroclinic tides
  !     100: tidal drag, shallow seas   
  !            ... and combos hereof, e.g. 11 includes both 1 and 10.
  !            ref: Canuto et al. Part III (2010), Chapter 7.
  !
  !
  ! tideparam.nml defines the type map use
  ! 
  ! Simple segment maps of rms-tide-velocity and topographic roughness:
  !
  !   In tideparam.nml e.g.:
  !     &nsegnml
  !       nsegments = 0, 2, 0, 0
  !     /
  !     &ijsegnml
  !       ijsegment = 192, 208, 230, 396
  !       uh = 1.0, 40.0
  !     /
  !     &ijsegnml
  !       ijsegment = 209, 240, 233, 396
  !       uh = 1.0, 40.0
  !     /
  !
  !
  ! Full maps:
  !
  !   In tideparam.nml e.g.:
  !     &nsegnml
  !       nsegments = 0, -1, 0, 0
  !     /
  !     &tfname
  !       tidefile  = "name_of_file"
  !     /
  !

  use cmod_params, only : maxareas
  use constants,   only : pi, zero, half, two

  implicit none
  private

  interface get_tide_map
    module procedure read_nsegments ! starting point
    module procedure return_flag
    module procedure assign_map
    module procedure get_value
  end interface

  public  :: get_tide_map
  private :: read_nsegments, return_flag, assign_map, get_value

  real(8), save, public  :: cdtide    = 0.003_8
  real(8), save, public  :: zetatide  = 500.0_8
  real(8), save, private :: kappatide = two*pi/125000.0_8
  real(8), save, private :: qtide     = 0.25_8

  integer(4), save, private :: nsegments(maxareas), lun, ialast
  integer(4),       private :: ios

contains

  !-----------------------------------------------------------------------------

  subroutine read_nsegments (narea)
    use exits,          only : exitme
    use io_subs,        only : io_new_unit
    use dmi_mpi_global, only : iu06

    implicit none

    integer(4), intent(in) :: narea

    integer(4) :: ia

    namelist /nsegnml/ nsegments

    ialast = narea
    nsegments(:) = 0

    lun = io_new_unit()
    open (unit=lun, file='tideparam.nml', status='old', iostat=ios)
    if (ios /= 0) then
      !  apply default settings:
      write(iu06,'(a71)')                                                      &
      'Warning: File tideparam.nml not found. Continue with default settings'
    else
      read (unit=lun, nml=nsegnml, iostat=ios)
      if (ios /= 0) then
        write(iu06,*) 'Error: Cannot read basic numbers from tideparam.nml.'
        write(iu06,*) 'Error: The iostat failure was: ', ios
        call exitme(1)
      endif
      do ia=1,narea
        if (nsegments(ia) < -1) call exitme(1,'Invalid No. of segments')
      enddo
    endif
  end subroutine read_nsegments

  !-----------------------------------------------------------------------------

  subroutine return_flag (ia,flag)
    implicit none
    integer(4), intent(in)    :: ia
    logical,    intent(inout) :: flag
    flag = (nsegments(ia) > 0) .or. (nsegments(ia) == -1)
    !-   That's all, folks!
  end subroutine return_flag

  !-----------------------------------------------------------------------------

  subroutine assign_map (ia,iw2,mmx,nmx,mmk,tidemap)
    use exits,          only : exitme
    use io_subs,        only : io_new_unit
    use dmi_mpi_global, only : iu06

    implicit none

    integer(4), intent(in)    :: ia, iw2, mmx, nmx, mmk(1:,0:,0:)
    real(8),    intent(inout) :: tidemap(0:,1:)

    integer(4)     :: is, i, j, n, ijsegment(4), iflun
    character(256) :: tidefile
    real(8)        :: uh(2)

    namelist /ijsegnml/   ijsegment, uh
    namelist /tfname/     tidefile
    namelist /tideparams/ cdtide, zetatide, kappatide, qtide

    ! relative factor 0 everywhere per default:
    tidemap(0:,1:2) = zero

    ! read simple maps:
    if (nsegments(ia) > 0) then
      do is=1,nsegments(ia)
        read (unit=lun, nml=ijsegnml, iostat=ios)
        if (ios /= 0) then
          write(iu06,*) 'Error: Cannot read ij-segment from tideparam.nml.'
          write(iu06,*) 'Error: The iostat failure was: ', ios
          call exitme(1)
        endif    
        if (uh(1) < zero) call exitme(1,'Invalid utide')
        if (uh(2) < zero) call exitme(1,'Invalid hrms')
        do j=max(1,ijsegment(3)),min(ijsegment(4),nmx)
          do i=max(1,ijsegment(1)),min(ijsegment(2),mmx)
            n = mmk(1,i,j)
            if (n > 0 .and. n <= iw2) then
              tidemap(n,1:2) = uh(1:2)
            endif
          enddo
        enddo
      enddo
    endif

    ! read full maps:
    if (nsegments(ia) == -1) then
      read (unit=lun, nml=tfname, iostat=ios)
      if (ios /= 0) then
        write(iu06,*) 'Error: Cannot read tide file name from tideparam.nml.'
        write(iu06,*) 'Error: The iostat failure was: ', ios
        call exitme(1)
      endif    
      iflun = io_new_unit()
      open(iflun,file=trim(tidefile),form='unformatted',iostat=ios,            &
           action='read',status='old')
      if (ios /= 0) call exitme (1, 'Could not open '//trim(tidefile) )
      read(iflun,iostat=ios) tidemap
      if (ios /= 0) then
        write(iu06,*) 'I/O problem - iostat is:',ios
        call exitme (3, 'Could not read tidemap from '//trim(tidefile) )
      endif
      close(iflun)
    endif

    ! read params only once:
    if (ia == ialast) then
      read (unit=lun, nml=tideparams, iostat=ios)
      if (ios /= 0) then
        if (ios == -1) then
          write(iu06,*) 'Cannot find namelist tideparams in tideparam.nml'
          write(iu06,*) '    iostat was: ', ios
          write(iu06,*) '    continue with default values.'
        else
          write(iu06,*) 'Error: Cannot read tide params from tideparam.nml.'
          write(iu06,*) 'Error: The iostat failure was: ', ios
          call exitme(1)
        endif
      endif
      close(lun)
    endif
  end subroutine assign_map

  !-----------------------------------------------------------------------------

  subroutine get_value(ia, n, tidetype, tidemap, utide, iwtide, tt1, tt2, tt3)
    implicit none

    integer(4), intent(in)  :: ia, n, tidetype
    real(8),    intent(in)  :: tidemap(0:,1:)
    real(8),    intent(out) :: utide, iwtide
    logical,    intent(out) :: tt1, tt2
    logical,    intent(out), optional :: tt3

    real(8) :: hrms
    logical :: t_all

    if (nsegments(ia) > 0 .or. nsegments(ia) == -1) then
      t_all = (tidetype == 111)

      tt1 = (tidetype ==   1 .or. tidetype ==  11 .or. tidetype==101 .or. t_all)
      tt2 = (tidetype ==  10 .or. tidetype ==  11 .or. tidetype==110 .or. t_all)

      if (present(tt3)) then
        tt3 = (tidetype == 100 .or. tidetype == 101 .or. tidetype==110         &
                                                    .or. t_all)
      endif

      utide = tidemap(n,1)
      hrms  = tidemap(n,2)

      iwtide = half*qtide*kappatide*(hrms**2)/zetatide
    else
      tt1 = .false.
      tt2 = .false.
      if (present(tt3)) tt3 = .false.
      
      utide  = zero
      iwtide = zero
    endif
      
  end subroutine get_value

  !-----------------------------------------------------------------------------

end module tideparam
