module coupler
  implicit none
  private

  ! private scalars:
  integer(4), private, save :: coupler_tracers = 2
  integer(4), private, save :: coupler_offset  = 2
  integer(4), private, save :: coupler_NDT     = 2
  real(8),    private, save :: coupler_dt      = 60.0_8
  real(8),    private, save :: coupler_outdt   = 3600.0_8
  integer(4), private, save :: olun            = 0
  integer(4), private, save :: itime           = 1
  integer(4), private, save :: otime           = 1
  integer(4), private, save :: out_nt          = 60
  logical,    private, save :: coupler_mpi     = .false.

  character(13), private, save :: coupler_output = 'tracers_stats'

  real(8), parameter, private :: zero = 0.0_8, one = 1.0_8, two = 2.0_8
  real(8), parameter, private :: three = 3.0_8

  real(8), private, save, allocatable :: tracer_max(:,:), tracer_min(:,:)
  real(8), private, save, allocatable :: tracer_ave(:,:)

  ! public methods:
  public :: coupler_usage, coupler_init, coupler_dynamics, coupler_finish

  ! private methods:
  private :: find_mean_value

contains

  !-----------------------------------------------------------------------------

  subroutine coupler_usage(iu06,nt,nt2d,irc)
    implicit none
    integer(4), intent(in)  :: iu06
    integer(4), intent(out) :: nt    ! number of traces to be inc. in cmp array 
    integer(4), intent(out) :: nt2d  ! number of 2d fields
    integer(4), intent(out) :: irc   ! return code; 0 means 'no errors'

    ! should read more specs from a namlelist, but for the sake of this demo 
    ! we just use the default values


    ! transfer characteristics of this sub model to cmod:
    nt            = coupler_tracers
    nt2d          = 0

    ! FIXME consider sub. the * below with some proper format string
    write(iu06,*) 'coupler_usage succesfully called ...'
    write(iu06,*) '  coupler_tracers:', coupler_tracers
    write(iu06,*) '  coupler_NDT:    ', coupler_NDT
    write(iu06,*) '  coupler_dt:     ', coupler_dt
    write(iu06,*) '  coupler_outdt:  ', coupler_outdt
    write(iu06,*) '  coupler_output: ', coupler_output

    ! some private vars:
    out_nt = coupler_outdt/coupler_dt

    ! return code:
    irc = 0

  end subroutine coupler_usage

  !-----------------------------------------------------------------------------

  subroutine coupler_init(ia, narea, iu06, mainarea, ncoff,                    &
                          mmx, nmx, kmx, nq, nz, mmk, kh, hz,                  &
                          components, rho, bndq, bndzk, on_IO_task, irc )
    !
    !Called from main as:
    !
    !if (mpi_io_serial) then
    !  do ia=1,narea
    !    call coupler_init(ia, narea, iu06, (ia==mainarea), ncoff,             &
    !                      mmx(ia),nmx(ia),kmx(ia),nq(ia),nz(ia),mm1k(ia)%p,   &
    !                      cmp(ia)%p, bndq(ia)%p, bndzk, mpi_io_serial, irc )
    !    if (irc /= 0) call exitme(irc, 'coupler_init failed')
    !  enddo
    !endif
    !
    implicit none

    ! arguments ----------------------------------------------------------------
    integer(4), intent(in)    :: ia, narea, iu06, mmx, nmx, kmx, nq, nz, ncoff
    logical,    intent(in)    :: mainarea
    integer(4), intent(in)    :: mmk(:,0:,0:), kh(0:)
    real(8),    intent(in)    :: hz(0:), rho(0:)
    real(8),    intent(inout) :: components(:,0:)
    real(8),    intent(inout) :: bndq(:,:,0:), bndzk(:,:,0:)
    logical,    intent(in)    :: on_IO_task
    integer(4), intent(out)   :: irc

    ! local vars ---------------------------------------------------------------
    integer(4)                :: nct, i, ios, nlow, nup
    integer(4)                :: n2d, n3d
    integer(4), parameter     :: io_min = 7, io_max = 130
    logical, save             :: FirstEntry = .true.
    logical                   :: file_is_open


    ! return code (assume no errors): ------------------------------------------
    irc = 0

    ! set river loadings -------------------------------------------------------
    !   ... laodings must be defined on each local task
    ! tracer #1 should be 1 all over:
    bndq(3,1:kmx,0:nq) = one
    ! tracer #2 has river inflows of 3:
    bndq(4,1:kmx,0:nq) = three

    ! return if not on IO task -------------------------------------------------
    if (.not.on_IO_task) return

    ! some short-hand notations ------------------------------------------------
    nct  = coupler_tracers
    nlow = ncoff + 1
    nup  = ncoff + nct
    n2d  = maxval(mmk(1,1:mmx,1:nmx))
    n3d  = maxval(mmk(1:kmx,1:mmx,1:nmx))

    ! on first entry, open a unit for output -----------------------------------
    if (FirstEntry) then
      FirstEntry = .false.

      write(iu06,'(a33,i4,a8)') 'Initialising coupler module with ',           &
                                nct,' tracers'

      coupler_offset = ncoff
      write(iu06,'(a33,i4)')    '                          offset ', ncoff

      file_is_open = .true.
      i = io_min-1
      do while (file_is_open)
        i = i+1
        if (i > io_max) then
          write(iu06,*) 'COUPLER ERROR: Could not find new I/O unit number'
          irc = 1
          return
        endif
        inquire (unit=i,opened=file_is_open)
      enddo
      olun = i

      open(olun,file=coupler_output,status='replace',iostat=ios)
      if (ios /= 0) then
        irc = ios
        write(iu06,*) 'COUPLER ERROR: Could not open '//coupler_output
        return
      endif

      if (nct < 1) then
        irc = 1
        write(iu06,*) 'COUPLER ERROR: Invalid No. of tracers:', nct
        return
      endif

      if (ncoff < 0) then  ! allowing for S/T, but be careful ...
        if (ncoff < 2) then
          write(iu06,*) 'COUPLER WARNING: This submodel accesses S/T.'
        endif
        irc = 1
        write(iu06,*) 'COUPLER ERROR: Invalid tracer offset No.:', ncoff
        return
      endif
    endif    

    ! print log logfile --------------------------------------------------------
    write(iu06,'(a20,i4)') '  Initialising area ',ia


    ! allocate local module arrays (if any) ------------------------------------
    if (ia == 1) then
      allocate(tracer_min(narea,nct), tracer_max(narea,nct),                   &
               tracer_ave(narea,nct), STAT=irc)
      if (irc /= 0) then
        write(iu06,*) 'COUPLER ERROR: Could not allocate local arrays'
        return
      endif
    endif


    ! initialize the tracers, one everywhere -----------------------------------
    components(nlow:nup,0:) = one
    do i=1,nct
      tracer_min(ia,i) = minval( components(ncoff+i,1:) )
      tracer_max(ia,i) = maxval( components(ncoff+i,1:) )
      call find_mean_value(i, ncoff, n3d, components, tracer_ave(ia,i))
    enddo


    ! set boundary values ------------------------------------------------------
    if (mainarea) then 
      ! tracer #1 should be 1 all over:
      bndzk(3,1:kmx,0:nz) = one
      ! tracer #2 has OBC of 2:
      bndzk(4,1:kmx,0:nz) = two
    endif


    ! That's it! ---------------------------------------------------------------
    write(iu06,'(a49,i4)')'... done coupler module initialisations for area ',ia
 
  end subroutine coupler_init

  !-----------------------------------------------------------------------------

  subroutine coupler_dynamics(ia, iu06, mmx, nmx, kmx, mmk, kh, hz,            &
                              components, dispv, rho, u, v, z, casus,          &
                              atemp, humid, cloud, wu, wv, pl, precip,         &
                              update_counters, idx, irc)
  !
  !Called from main something like this:
  !
  !
  ! ! mpi coupler:
  ! do ia=1,narea
  !   call coupler_dynamics(ia, iu06, ..., cmp_l(ia)%p, ..., (ia==narea), irc)
  !   if (irc /= 0) call exitme(irc, 'coupler_dynamics failed')
  ! enddo
  !

    implicit none

    ! arguments ----------------------------------------------------------------
    integer(4), intent(in)    :: ia, iu06, mmx, nmx, kmx, mmk(:,0:,0:), kh(0:)
    real(8),    intent(in)    :: hz(0:)
    real(8),    intent(inout) :: components(:,0:), dispv(:,0:)
    real(8),    intent(in)    :: u(0:), v(0:), z(0:), rho(0:)
    logical,    intent(in)    :: casus(0:)
    real(8),    intent(in)    :: atemp(0:), humid(0:), cloud(0:)
    real(8),    intent(in)    :: wu(0:), wv(0:), pl(0:), precip(0:)
    logical,    intent(in)    :: update_counters
    integer(4), intent(inout) :: idx(1:,1:)
    integer(4), intent(out)   :: irc


    ! local vars ---------------------------------------------------------------
    integer(4) :: i, n2d, n3d
    

    ! return code (assume no errors) -------------------------------------------
    irc = 0


    ! extract No of wet points -------------------------------------------------
    n2d  = maxval(mmk(1,1:mmx,1:nmx))
    n3d  = maxval(mmk(1:kmx,1:mmx,1:nmx))


    ! do the sub model dynamics ------------------------------------------------
    if (mod(itime,coupler_NDT) == 0) then
      do i=1,coupler_tracers
        tracer_min(ia,i) = minval( components(coupler_offset+i,1:) )
        tracer_max(ia,i) = maxval( components(coupler_offset+i,1:) )
        call find_mean_value(i,coupler_offset,n3d,components,tracer_ave(ia,i))
      enddo
    endif


    ! do the sub model outout --------------------------------------------------
    if (mod(otime,out_nt) == 0) then
      ! print the stats to unit olun
      do i=1,coupler_tracers
        !FIXME * with a proper format string
        write(olun,*) 'Area:',ia,' tracer No.:', i,' min:', tracer_min(ia,i)
        write(olun,*) 'Area:',ia,' tracer No.:', i,' ave:', tracer_ave(ia,i)
        write(olun,*) 'Area:',ia,' tracer No.:', i,' max:', tracer_max(ia,i)
      enddo
    endif


    ! update time step counters ------------------------------------------------
    if (update_counters) then
      if (mod(itime,coupler_NDT) == 0) then
        itime = 1
      else
        itime = itime + 1 
      endif  
      if (mod(otime,out_nt) == 0) then
        otime = 1
      else
        otime = otime + 1 
      endif  
    endif

  end subroutine coupler_dynamics

  !-----------------------------------------------------------------------------

  subroutine coupler_finish(iu06, irc)
    implicit none

    integer(4), intent(in)   :: iu06
    integer(4), intent(out)  :: irc

    ! return code (assume no errors):
    irc = 0

    ! close the units
    close (olun)

    ! deallocate
    deallocate(tracer_min, tracer_max, tracer_ave, STAT=irc)
    if (irc /= 0) then
      write(iu06,*) 'COUPLER ERROR: Could not deallocate local arrays'
      return
    endif

  end subroutine coupler_finish

  !-----------------------------------------------------------------------------
  ! Below here goes all the sub-model private subroutines:
  ! These guys could also be placed in seperate files in case this is desired, 
  ! cf. EXTRA_SOURCES in Makefile.am
  !-----------------------------------------------------------------------------

  subroutine find_mean_value(ic, icoff, n3d, cmp, mean)
    implicit none
    integer(4), intent(in)  :: ic, icoff, n3d
    real(8),    intent(in)  :: cmp(:,0:)
    real(8),    intent(out) :: mean

    integer(4) :: i
    real(8)    :: cmean, fac1, fac2

    cmean = zero
    do i=0,n3d-1
      fac2  = real(i,8)
      fac1  = one/(fac2 + one)
      cmean = (cmean*fac2 + cmp(ic+icoff,i+1) )*fac1
    enddo

    mean = cmean

  end subroutine find_mean_value

  !-----------------------------------------------------------------------------

end module coupler

