module transport_kernel
  !-----------------------------------------------------------------------------
  !- In this module all kernel-subroutines are saved which are needed for 
  !- inline transport calculation
  !-----------------------------------------------------------------------------

  implicit none
  private

  public  :: water_trans, transform, calculate_mean, transects
  private :: add_trans

  real(8), parameter, private :: zero = 0.0_8, centi = 0.01_8, two = 2.0_8

contains

  subroutine water_trans(n2d, ind, mmk, kh, hu, hv, u, v, cmp, utr, vtr,       &
                         temputr, tempvtr, saltutr, saltvtr, transupt, transvpt)

    !- modules -----------------------------------------------------------------
    use constants, only : rhow, cpw
    use dmi_omp,   only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: n2d
    integer(4), intent(in) :: ind(:,:)
    integer(4), intent(in) :: mmk(:,0:,0:)
    integer(4), intent(in) :: kh(0:)
    real(8), intent(in)    :: hu(0:), hv(0:)
    real(8), intent(in)    :: u(0:), v(0:)
    real(8), intent(in)    :: cmp(1:,0:)
    real(8), intent(inout) :: utr(0:), vtr(0:)
    real(8), intent(inout) :: temputr(0:), tempvtr(0:)
    real(8), intent(inout) :: saltutr(0:), saltvtr(0:)
    logical, intent(in)    :: transupt(0:), transvpt(0:)

    !- local vars --------------------------------------------------------------
    integer(4)   :: nsrf, nl, nu
    integer(4)   :: i, j, k, mi
    real(8)      :: tw, te, ts

    !- parameters --------------------------------------------------------------
    real(8), parameter :: rg = rhow/1000.0_8
    real(8), parameter :: cpr = cpw*rhow 
    real(8), parameter :: C2K = 273.15_8      ! Celcius to Kelvin

    call  domp_get_domain(1,n2d, nl, nu)

    do nsrf=nl,nu
      if (transupt(nsrf)) then
        i = ind(1,nsrf)
        j = ind(2,nsrf)
        do k=1,kh(nsrf)
          mi = mmk(k,i,j)

          if(hu(mi) > centi) then
            utr(mi)=u(mi)*hu(mi)
          else
            utr(mi)=zero
          endif

          tw=cmp(1,mi)+C2K
          te=cmp(1,mmk(k,i,j+1))+C2K

          if(u(mi) > zero) then
            saltutr(mi)=utr(mi)*cmp(2,mi)*rg
            temputr(mi)=utr(mi)*tw*cpr
          else
            saltutr(mi)=utr(mi)*cmp(2,mmk(k,i,j+1))*rg
            temputr(mi)=utr(mi)*te*cpr
          endif
        enddo
      endif

      if (transvpt(nsrf)) then
        i = ind(1,nsrf)
        j = ind(2,nsrf)
        do k=1,kh(nsrf)
          mi = mmk(k,i,j)

          if(hv(mi) > centi) then
            vtr(mi)=v(mi)*hv(mi)
          else
            vtr(mi)=zero
          endif

          tw=cmp(1,mi)+C2K
          ts=cmp(1,mmk(k,i+1,j))+C2K

          if(v(mi) > zero) then
            saltvtr(mi)=vtr(mi)*cmp(2,mmk(k,i+1,j))*rg
            tempvtr(mi)=vtr(mi)*ts*cpr
          else
            saltvtr(mi)=vtr(mi)*cmp(2,mi)*rg
            tempvtr(mi)=vtr(mi)*tw*cpr
          endif
        enddo
      endif

    enddo

  end subroutine water_trans


  !=============================================================================


  subroutine transform( n2d, ind, mmk, khu, khv, ti, z, hu, hv, ktrans,hztrans,&
                        transupt, transvpt, utr, vtr, temputr, tempvtr,        &
                        saltutr, saltvtr, utrans, vtrans,temputrans,tempvtrans,&
                        saltutrans, saltvtrans )

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!  This subroutine calculates (transforms the model layer transports to) !!
    !!  the total transports for fixed z-layers, which are given from outside !!
    !!  (e.g. NOOS/BOOS-layers). These layers are in a z-co-ordinate system   !!
    !!  with mean water level as reference level 0, whereas the first layer   !!
    !!  includes the variable surface fluctuation.                            !!
    !!  At present the external given layers has to be coarser than the model !!
    !!  layers - therfore the transformation is always a compression today!   !!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    !- modules -----------------------------------------------------------------
    use dmi_omp,   only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: n2d
    integer(4), intent(in)    :: khu(0:), khv(0:)
    integer(4), intent(in)    :: ind(:,:)
    integer(4), intent(in)    :: mmk(:,0:,0:)
    integer(4), intent(in)    :: ktrans
    logical, intent(in)       :: transupt(0:), transvpt(0:)
    real(8), intent(in)       :: ti(0:)
    real(8), intent(in)       :: z(0:)
    real(8), intent(in)       :: hu(0:), hv(0:)
    real(8), intent(in)       :: hztrans(0:)
    real(8), intent(inout)    :: utr(0:), vtr(0:)
    real(8), intent(inout)    :: temputr(0:), tempvtr(0:)
    real(8), intent(inout)    :: saltutr(0:), saltvtr(0:)
    real(8), intent(inout)    :: utrans(:,0:), vtrans(:,0:) 
    real(8), intent(inout)    :: temputrans(:,0:), tempvtrans(:,0:) 
    real(8), intent(inout)    :: saltutrans(:,0:), saltvtrans(:,0:) 

    !- local vars --------------------------------------------------------------
    integer(4)   :: i, j, k, kt
    integer(4)   :: mi, ee, ss
    integer(4)   :: nsrf, nl, nu
    integer(4)   :: ktop, kbot, kbott
    real(8)      :: tiu, tiv
    real(8)      :: ztop, zbot, ztopt, zbott
    real(8)      :: zz1, zz2, fac1, fac2
    real(8)      :: hutemp, hvtemp

    !- parameters --------------------------------------------------------------
    real(8), parameter :: tiny_val = 0.001_8

    call  domp_get_domain(1,n2d, nl, nu)

    do nsrf=nl,nu

      i = ind(1,nsrf)
      j = ind(2,nsrf)
 
      mi = mmk(1,i,j)
      ee = mmk(1,i,j+1)
      ss = mmk(1,i+1,j)

      if(mi > 0 .and. ee > 0 .and. ss > 0) then
        tiu = (ti(mi) + ti(ee))/two
        tiv = (ti(mi) + ti(ss))/two
 
        if (transupt(nsrf)) then

          hutemp = hu(mmk(1,i,j))

          !- Calculate the transports in u (eastward) direction
          ztop = zero
          zbot = zero
          ztopt= (z(mmk(1,i,j))+z(mmk(1,i,j+1)))/two
          zbott= zero
          ktop = 1
          kbot = khu(mmk(1,i,j))
          kbott= ktrans

          !- vertical loop over the (coarser !!) external given layers
          do kt = 1,kbott
            zbott = ztopt + hztrans(kt)

            !- vertical loop over the (finer !!) model layers
            do k = ktop,kbot
              mi = mmk(k,i,j)
              zbot = ztop + hutemp
    
              if( ztop <= (tiu-tiny_val) ) then
                if( zbot <= zbott) then 
                  utrans(kt,nsrf)     = utrans(kt,nsrf) + utr(mi)
                  temputrans(kt,nsrf) = temputrans(kt,nsrf) + temputr(mi)
                  saltutrans(kt,nsrf) = saltutrans(kt,nsrf) + saltutr(mi)
                  ztop=zbot
                  if(k == kbot) then
                    ktop = kbot + 1
                  else
                    hutemp = hu(mmk(k+1,i,j))
                  endif
                else
                  zz1  = zbott - ztop
                  zz2  = zbot - zbott
                  fac1 = zz1/hutemp
                  fac2 = zz2/hutemp
                  utrans(kt,nsrf)     = utrans(kt,nsrf) + utr(mi)*fac1
                  temputrans(kt,nsrf) = temputrans(kt,nsrf) + temputr(mi)*fac1
                  saltutrans(kt,nsrf) = saltutrans(kt,nsrf) + saltutr(mi)*fac1
                  ztop = zbott
                  ktop = k
                  utr(mi)     = utr(mi)*fac2
                  temputr(mi) = temputr(mi)*fac2
                  saltutr(mi) = saltutr(mi)*fac2
                  hutemp = zz2
                  exit
                endif
              endif
            enddo

            ztopt = zbott
          enddo

        endif


        if (transvpt(nsrf)) then

          hvtemp = hv(mmk(1,i,j))

          !- Calculate the transports in v (northward) direction
          ztop = zero
          zbot = zero
          ztopt= (z(mmk(1,i,j))+z(mmk(1,i+1,j)))/two
          zbott= zero
          ktop = 1
          kbot = khv(mmk(1,i,j))
          kbott= ktrans

          !- vertical loop over the (coarser !!) external given layers
          do kt = 1,kbott
            zbott = ztopt + hztrans(kt)

            !- vertical loop over the (finer !!) model layers
            do k = ktop,kbot
              mi = mmk(k,i,j)
              zbot = ztop + hvtemp
              if( ztop <= (tiv-tiny_val) ) then
                if( zbot <= zbott) then 
                  vtrans(kt,nsrf)     = vtrans(kt,nsrf) + vtr(mi)
                  tempvtrans(kt,nsrf) = tempvtrans(kt,nsrf) + tempvtr(mi)
                  saltvtrans(kt,nsrf) = saltvtrans(kt,nsrf) + saltvtr(mi)
                  ztop=zbot
                  if(k == kbot) then
                    ktop = kbot + 1
                  else
                    hvtemp = hv(mmk(k+1,i,j))
                  endif
                else
                  zz1  = zbott - ztop
                  zz2  = zbot - zbott
                  fac1 = zz1/hvtemp
                  fac2 = zz2/hvtemp
                  vtrans(kt,nsrf)     = vtrans(kt,nsrf) + vtr(mi)*fac1
                  tempvtrans(kt,nsrf) = tempvtrans(kt,nsrf) + tempvtr(mi)*fac1
                  saltvtrans(kt,nsrf) = saltvtrans(kt,nsrf) + saltvtr(mi)*fac1
                  ztop = zbott
                  ktop = k
                  vtr(mi)     = vtr(mi)*fac2
                  tempvtr(mi) = tempvtr(mi)*fac2
                  saltvtr(mi) = saltvtr(mi)*fac2
                  hvtemp = zz2
                  exit
                endif
              endif
            enddo

            ztopt = zbott
          enddo
        endif

      endif

    enddo

  end subroutine transform


  !=============================================================================


  subroutine calculate_mean ( n2d, kmax, rtsteps, utrans, vtrans, temputrans,  &
                              tempvtrans, saltutrans, saltvtrans, transupt,    &
                              transvpt )

    !- modules -----------------------------------------------------------------
    use dmi_omp,   only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)  :: n2d
    integer(4), intent(in)  :: kmax
    logical, intent(in)     :: transupt(0:), transvpt(0:)
    real(8), intent(in)     :: rtsteps
    real(8), intent(inout)  :: utrans(:,0:)
    real(8), intent(inout)  :: vtrans(:,0:)
    real(8), intent(inout)  :: temputrans(:,0:)
    real(8), intent(inout)  :: tempvtrans(:,0:)
    real(8), intent(inout)  :: saltutrans(:,0:)
    real(8), intent(inout)  :: saltvtrans(:,0:)
    
    !- local vars --------------------------------------------------------------
    integer(4) :: nn, kt, nl, nu

    !- local parameters --------------------------------------------------------
    real(8), parameter :: tiny_val = 1.e-9_8

    call  domp_get_domain(1,n2d, nl, nu)

    do nn = nl, nu
      if (transupt(nn) .or. transvpt(nn)) then

        do kt = 1,kmax
          utrans(kt,nn) = utrans(kt,nn)/rtsteps
          vtrans(kt,nn) = vtrans(kt,nn)/rtsteps
          temputrans(kt,nn) = temputrans(kt,nn)/rtsteps
          tempvtrans(kt,nn) = tempvtrans(kt,nn)/rtsteps
          saltutrans(kt,nn) = saltutrans(kt,nn)/rtsteps
          saltvtrans(kt,nn) = saltvtrans(kt,nn)/rtsteps
          if(abs(utrans(kt,nn)) <= tiny_val) utrans(kt,nn) = zero
          if(abs(vtrans(kt,nn)) <= tiny_val) vtrans(kt,nn) = zero
          if(abs(temputrans(kt,nn)) <= tiny_val) temputrans(kt,nn) = zero
          if(abs(tempvtrans(kt,nn)) <= tiny_val) tempvtrans(kt,nn) = zero
          if(abs(saltutrans(kt,nn)) <= tiny_val) saltutrans(kt,nn) = zero
          if(abs(saltvtrans(kt,nn)) <= tiny_val) saltvtrans(kt,nn) = zero
        enddo
      endif
    enddo    

  end subroutine calculate_mean


  !=============================================================================


  subroutine transects ( narea, iw2, ind, dx, dy, covu, kmax, ntrans, grid,    &
                         xdir, ydir, ctru, ctrv, utrans, vtrans, temputrans,   &
                         tempvtrans, saltutrans, saltvtrans, wtrans, ttrans,   &
                         strans )

    !- modules -----------------------------------------------------------------
    use cmod_mem,      only : cmi2, cmr1, cmr2, cml2

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: narea
    integer(4), intent(in) :: iw2(:)
    integer(4), intent(in) :: kmax
    integer(4), intent(in) :: ntrans
    logical, intent(in)    :: grid(:,:)
    real(8), intent(in)    :: dy(:)
    real(8), intent(in)    :: xdir(:), ydir(:)
    type(cmr1), intent(in) :: dx(:), covu(:)
    type(cmr2), intent(in) :: utrans(:), vtrans(:)
    type(cmr2), intent(in) :: temputrans(:), tempvtrans(:)
    type(cmr2), intent(in) :: saltutrans(:), saltvtrans(:)
    type(cmi2), intent(in) :: ind(:)
    type(cml2), intent(in) :: ctru(:), ctrv(:)
    real(8), intent(inout) :: wtrans(0:,:,:)
    real(8), intent(inout) :: ttrans(0:,:,:)
    real(8), intent(inout) :: strans(0:,:,:)

    !- local vars --------------------------------------------------------------
    integer(4)   :: ia, ntr, kt

    !- Initialising the output arrays ------------------------------------------
    wtrans(0:,:,:) = zero
    ttrans(0:,:,:) = zero
    strans(0:,:,:) = zero

    do ntr = 1, ntrans
      do ia = 1, narea
        if ( grid(ntr,ia) ) then
          call add_trans ( iw2(ia), ind(ia)%p, dx(ia)%p, dy(ia), covu(ia)%p,   &
                           ntr, kmax, xdir(ntr), ydir(ntr), ctru(ia)%p,        &
                           ctrv(ia)%p, utrans(ia)%p, vtrans(ia)%p,             &
                           temputrans(ia)%p, tempvtrans(ia)%p,                 &
                           saltutrans(ia)%p, saltvtrans(ia)%p,                 &
                           wtrans(0:,:,ntr), ttrans(0:,:,ntr),strans(0:,:,ntr) )
        endif
      enddo

      do kt=1,kmax
        wtrans(kt,1,ntr) = wtrans(kt,2,ntr) + wtrans(kt,3,ntr)
        ttrans(kt,1,ntr) = ttrans(kt,2,ntr) + ttrans(kt,3,ntr)
        strans(kt,1,ntr) = strans(kt,2,ntr) + strans(kt,3,ntr)
      enddo       
      
      do kt=1,kmax
        wtrans(0,1:3,ntr) = wtrans(0,1:3,ntr) + wtrans(kt,1:3,ntr)
        ttrans(0,1:3,ntr) = ttrans(0,1:3,ntr) + ttrans(kt,1:3,ntr)
        strans(0,1:3,ntr) = strans(0,1:3,ntr) + strans(kt,1:3,ntr)
      enddo

    enddo

  end subroutine transects


  !=============================================================================


  subroutine add_trans ( n2d, ind, dx, dy, covu, ntr, kmax, xdir, ydir, ctru,  &
                         ctrv, utrans, vtrans, temputrans, tempvtrans,         &
                         saltutrans, saltvtrans, wtrans, ttrans, strans )

  !- modules -------------------------------------------------------------------
  ! none

  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in) :: n2d
  integer(4), intent(in) :: ind(:,:)
  integer(4), intent(in) :: ntr
  integer(4), intent(in) :: kmax
  logical, intent(in)    :: ctru(0:,:), ctrv(0:,:)
  real(8), intent(in)    :: dy
  real(8), intent(in)    :: dx(:), covu(:)
  real(8), intent(in)    :: xdir, ydir
  real(8), intent(in)    :: utrans(:,0:), vtrans(:,0:)
  real(8), intent(in)    :: temputrans(:,0:), tempvtrans(:,0:)
  real(8), intent(in)    :: saltutrans(:,0:), saltvtrans(:,0:)
  real(8), intent(inout) :: wtrans(0:,:)
  real(8), intent(inout) :: ttrans(0:,:)
  real(8), intent(inout) :: strans(0:,:)

  !- local vars ----------------------------------------------------------------
  integer(4) :: nn, i, kt
  real(8)    :: factoru, factorv

  do nn = 1, n2d
    i = ind(1,nn)

     if ( ctru(nn,ntr) ) then
       do kt = 1,kmax
         factoru = dy*xdir
         if( utrans(kt,nn)*xdir > zero) then
           wtrans(kt,2) = wtrans(kt,2) + utrans(kt,nn)*factoru
           ttrans(kt,2) = ttrans(kt,2) + temputrans(kt,nn)*factoru
           strans(kt,2) = strans(kt,2) + saltutrans(kt,nn)*factoru
         else
           wtrans(kt,3) = wtrans(kt,3) + utrans(kt,nn)*factoru
           ttrans(kt,3) = ttrans(kt,3) + temputrans(kt,nn)*factoru
           strans(kt,3) = strans(kt,3) + saltutrans(kt,nn)*factoru
         endif
       enddo       
     endif

     if ( ctrv(nn,ntr) ) then
       do kt = 1,kmax
         factorv = dx(i)*covu(i)*ydir            
         if( vtrans(kt,nn)*ydir > zero) then
           wtrans(kt,2) = wtrans(kt,2) + vtrans(kt,nn)*factorv
           ttrans(kt,2) = ttrans(kt,2) + tempvtrans(kt,nn)*factorv
           strans(kt,2) = strans(kt,2) + saltvtrans(kt,nn)*factorv
         else
           wtrans(kt,3) = wtrans(kt,3) + vtrans(kt,nn)*factorv
           ttrans(kt,3) = ttrans(kt,3) + tempvtrans(kt,nn)*factorv
           strans(kt,3) = strans(kt,3) + saltvtrans(kt,nn)*factorv
         endif
       enddo       
     endif

   enddo

  end subroutine add_trans

end module transport_kernel
