module transport_variables
  !-----------------------------------------------------------------------------
  !- All fields and parameters which are needed for inline transport calculation
  !- are introduced and also allocates in this module
  !-----------------------------------------------------------------------------

  !- Modules -------------------------------------------------------------------
  use cmod_mem, only : cmi1, cml1, cml2, cmr1, cmr2, cmr3

  !- directives ----------------------------------------------------------------
  implicit none
  private
  
  !- Transport file-names ------------------------------------------------------
  character(256), allocatable, save, public :: transfile(:)

  !- Time indices --------------------------------------------------------------
  integer(4), save, public :: izeittrans

  !- Transport namelist parameter ----------------------------------------------
  integer(4), save, public :: ntransdat
  integer(4), save, public :: NDTTRANS   
  integer(4), save, public :: NDTTRANSOUT

  !- Indices -------------------------------------------------------------------
  integer(4), allocatable, save, public :: ntrans(:)
  integer(4), allocatable, save, public :: ktransmax(:)
  type(cmi1), allocatable, save, public     :: ktrans(:)
  type(cmr1), allocatable, save, public     :: hztrans(:)
  type(cmr1), allocatable, save, public     :: xdirectfac(:), ydirectfac(:)
  type(cml2), allocatable, save, public     :: grid(:)
  type(cml2), allocatable, save, public     :: ctru(:,:), ctrv(:,:)
  
  !- The output variables ------------------------------------------------------
  integer(4), save, public          :: luntrans
  character(40), save, public       :: transdnam
  type(cmr3), allocatable, save, public :: wtrans(:), ttrans(:), strans(:)  

  !- The global arrays ---------------------------------------------------------
  type(cml1), allocatable, save, public :: transupt(:), transvpt(:)
  type(cmr1), allocatable, save, public :: utr(:), vtr(:)
  type(cmr1), allocatable, save, public :: temputr(:), tempvtr(:)
  type(cmr1), allocatable, save, public :: saltutr(:), saltvtr(:)
  type(cmr2), allocatable, save, public :: utrans(:,:), vtrans(:,:)
  type(cmr2), allocatable, save, public :: temputrans(:,:), tempvtrans(:,:)
  type(cmr2), allocatable, save, public :: saltutrans(:,:), saltvtrans(:,:) 
  
  !- The local arrays ----------------------------------------------------------
  type(cml1), allocatable, save, public :: transupt_l(:), transvpt_l(:)
  type(cmr1), allocatable, save, public :: utr_l(:), vtr_l(:)
  type(cmr1), allocatable, save, public :: temputr_l(:), tempvtr_l(:)
  type(cmr1), allocatable, save, public :: saltutr_l(:), saltvtr_l(:)
  type(cmr2), allocatable, save, public :: utrans_l(:,:), vtrans_l(:,:)
  type(cmr2), allocatable, save, public :: temputrans_l(:,:), tempvtrans_l(:,:)
  type(cmr2), allocatable, save, public :: saltutrans_l(:,:), saltvtrans_l(:,:)

  !-subroutines/functions ------------------------------------------------------
  public  ::  transport_usage, transport_parameters, alloc_global, alloc_local,&
              transdat_new, transdat_close
  private ::  task_global2local_trans

  real(8), parameter, private :: zero = 0.0_8

  contains

    subroutine transport_usage ( )
    
      !- modules ---------------------------------------------------------------
      use dmi_mpi_global, only : iu06
      use exits,          only : exitme
      use io_subs,        only : io_new_unit

      !- dircetives ------------------------------------------------------------
      implicit none
  
      !- arguments -------------------------------------------------------------
      ! none
      
      !- local vars ------------------------------------------------------------
      integer(4)     :: lun, ios, itransdat
      character(256) :: transfile_in
      
      !- local parameters ------------------------------------------------------
      character(5), parameter  :: w = '(a55)'
      character(12), parameter :: wi = '(a13,i2,a43)'

      namelist /translist/   ntransdat, NDTTRANS, NDTTRANSOUT
      namelist /transfn/     transfile_in

      !-----Defaults -----------------------------------------------------------
      ntransdat    = 1
      NDTTRANS     = 2
      NDTTRANSOUT  = 10
      transfile_in = 'data_trans'

      !-----Arguments from NAMELIST --------------------------------------------
      lun = io_new_unit()
      open (unit=lun, file='transports.nml', status='old', iostat=ios)
      if (ios /= 0) then
        write(iu06,w)  'Warning: Cannot open file transports.nml ... exit      '
        write(iu06,w)  '         ... continue with default transport-options.  '
        write(iu06,'(a25,i5)') 'I/O problem - iostat is: ',ios

        allocate( transfile(ntransdat) )
        transfile(ntransdat) = transfile_in
      else
        read (unit=lun, nml=translist, iostat=ios)
        if (ios /= 0) then
          write(iu06,w)'Warning: Cannot read translist in transports.nml file..'
          write(iu06,w)'           ... continue with default transport-options.'
          write(iu06,'(a25,i5)') 'I/O problem - iostat is: ',ios
          allocate( transfile(ntransdat) )
          transfile(ntransdat) = transfile_in
        else
          allocate ( transfile(ntransdat) )
          do itransdat = 1,ntransdat
            read (unit=lun, nml=transfn, iostat=ios)
            if ( ios /= 0 ) then
              write(iu06,*) 'Error: Cannot read file names from transfn file ',&
                            '...exit'
              write(iu06,*) 'Error: The iostat failure was:: ', ios
              call exitme(1)
            else
              transfile(itransdat)   = trim(transfile_in)
            endif
          enddo
        endif
      endif
      close (unit=lun)

      !-----Information for the logifile ---------------------------------------
      write(iu06,wi)  'Transports of', ntransdat,                              &
                                   ' transect-set(s) will be calculated inline!'

      if ( NDTTRANSOUT /= NDTTRANS*(NDTTRANSOUT/NDTTRANS)) then
        write(iu06,w) 'ERROR: Time step size for transport output must be an..'
        write(iu06,w) '       ...integer multiple of the time step size for ..'
        write(iu06,w) '       ...transport calculation                        '
        call exitme(1,' ... bailing out')
      endif

    end subroutine transport_usage
  

!===============================================================================
  
  
    subroutine transport_parameters( narea )
    
      !- modules ---------------------------------------------------------------
      use cmod_params, only : iw2, me1
      use exits,       only : exitme
      use io_subs,     only : io_new_unit

      !- dircetives ------------------------------------------------------------
      implicit none

      !- arguments -------------------------------------------------------------
      integer(4), intent(in) :: narea

      !- local vars ------------------------------------------------------------
      character(3)  :: ch3
      integer(4)    :: ia, idx
      integer(4)    :: lun, ios
      integer(4)    :: itransdat, ntr, itr, i, j, iu, iv
      integer(4)    :: xdirINT, ydirINT
      integer(4), allocatable :: iwsfpoints(:)
     
      !- First allocate the first transport parameter arrays -------------------
      allocate ( ntrans(ntransdat), ktransmax(ntransdat), hztrans(ntransdat),  &
                 ktrans(ntransdat),xdirectfac(ntransdat),ydirectfac(ntransdat),&
                 grid(ntransdat), ctru(narea,ntransdat), ctrv(narea,ntransdat),&
                 transupt(narea), transvpt(narea) )

      !- Set izeittrans to zero ------------------------------------------------
      izeittrans = 0    
      
      !- Go through all transport datasets -------------------------------------
      !-  (e.g. NOOS/BOOS-transects are one dataset ----------------------------
      do itransdat = 1, ntransdat
        lun = io_new_unit( )
        open (unit=lun, file=trim(transfile(itransdat)),status='old',iostat=ios)
        if (ios /= 0) then
          call exitme(1,'Could not open '//trim(transfile(itransdat)) )
        endif
        read(lun,'(2X,A3)',iostat=ios) ch3
        if (ios /= 0) then
          call exitme(1,'Error while reading '// trim(transfile(itransdat)) )
        endif

        !- Read the basic informations of this dataset -------------------------
        do while (ch3 /= 'TRA')
          read(lun,'(2X,A3)',iostat=ios) ch3
          if (ios /= 0) then
            call exitme(1,'Error while reading '// trim(transfile(itransdat)) )
          endif
        enddo
        read(lun,'(14X,I3)',iostat=ios) ntrans(itransdat)
        if (ios /= 0) then
          call exitme(1,'Error while reading '// trim(transfile(itransdat)) )
        endif
        read(lun,'(14X,I3)',iostat=ios) ktransmax(itransdat)
        if (ios /= 0) then
          call exitme(1,'Error while reading '// trim(transfile(itransdat)) )
        endif

        !- Allocate some more transport parameter arrays -----------------------
        allocate( hztrans(itransdat)%p(0:ktransmax(itransdat)),                &
                  ktrans(itransdat)%p(1:ntrans(itransdat)),                    &
                  xdirectfac(itransdat)%p(1:ntrans(itransdat)),                &
                  ydirectfac(itransdat)%p(1:ntrans(itransdat)),                &
                  grid(itransdat)%p(1:ntrans(itransdat),1:narea)  )

        do ia = 1, narea
          allocate ( ctru(ia,itransdat)%p(0:iw2(ia),ntrans(itransdat)),        &
                     ctrv(ia,itransdat)%p(0:iw2(ia),ntrans(itransdat)),        &
                     transupt(ia)%p(0:iw2(ia)), transvpt(ia)%p(0:iw2(ia)))
        enddo

        !- Allocate local arrays -----------------------------------------------
        allocate( iwsfpoints(1:ntrans(itransdat)) )

        !- Initialise some arrays (Better safe than sorry) ---------------------
        grid(itransdat)%p(:,:) = .false.
        do ia = 1, narea
          ctru(ia,itransdat)%p(0:,:) = .false.
          ctrv(ia,itransdat)%p(0:,:) = .false.
          transupt(ia)%p(:) = .false.
          transvpt(ia)%p(:) = .false.
        enddo

        !- Read the values of hztrans this dataset -----------------------------
        read(lun,'(2X,A3)',iostat=ios) ch3
        if (ios /= 0) then
          call exitme(1,'Error while reading '// trim(transfile(itransdat)) )
        endif
        do while (ch3 /= 'TRA')
          read(lun,'(2X,A3)',iostat=ios) ch3
          if (ios /= 0) then
            call exitme(1,'Error while reading '// trim(transfile(itransdat)) )
          endif
        enddo
        hztrans(itransdat)%p(0) = zero
        do i = 1, ktransmax(itransdat)
          read(lun,'(14X,f8.3)',iostat=ios) hztrans(itransdat)%p(i)
          if (ios /= 0) then
            call exitme(1,'Error while reading '// trim(transfile(itransdat)) )
          endif
        enddo

        !- Read the information of the transects(direction,surface-points etc.)-
        read(lun,'(2X,A3)',iostat=ios) ch3
        if (ios /= 0) then
          call exitme(1,'Error while reading '// trim(transfile(itransdat)) )
        endif
        do while (ch3 /= 'TRA')
          read(lun,'(2X,A3)',iostat=ios) ch3
          if (ios /= 0) then
            call exitme(1,'Error while reading '// trim(transfile(itransdat)) )
          endif
        enddo
        do  ntr = 1, ntrans(itransdat)
          read(lun,'(33X,2i2,10X,i4,8X,i3)',iostat=ios) xdirINT, ydirINT,      &
                                                        iwsfpoints(ntr),       &
                                                        ktrans(itransdat)%p(ntr)
          if (ios /= 0) then
            call exitme(1,'Error while reading '// trim(transfile(itransdat)) )
          else
            xdirectfac(itransdat)%p(ntr) = real(xdirINT, kind=8)
            ydirectfac(itransdat)%p(ntr) = real(ydirINT, kind=8)
            if((xdirINT /= 1 .and. xdirINT /= -1) .or.                         &
               (ydirINT /= 1 .and. ydirINT /= -1) ) then
              call exitme(1,'Directions must be positive(1) or negative(-1)')
            endif

            do itr = 1,iwsfpoints(ntr)
              read(lun,'(18X,3I5,2I3)',iostat=ios) ia, i, j, iu, iv
              if (ios /= 0) then
                call exitme(1,'Error while reading '//                         &
                               trim(transfile(itransdat)))
              else
                if(.not. grid(itransdat)%p(ntr,ia)) then
                  grid(itransdat)%p(ntr,ia) = .true.
                endif

                idx = me1(ia)%p(i,j)

                if(iu == 1) then
                  ctru(ia,itransdat)%p(idx,ntr) = .true.
                  transupt(ia)%p(idx) = .true.
                endif
                if(iv == 1) then
                  ctrv(ia,itransdat)%p(idx,ntr) = .true.
                  transvpt(ia)%p(idx) = .true.
                endif
              endif
            enddo
          endif
        enddo

      enddo

    end subroutine transport_parameters
  

!===============================================================================

    subroutine transdat_new( )

      !- modules ---------------------------------------------------------------
      use exits,       only : exitme
      use io_subs,     only : io_new_unit

      !- dircetives ------------------------------------------------------------
      implicit none

      !- arguments -------------------------------------------------------------
      ! none

      !- local vars ------------------------------------------------------------
      logical, save    :: FirstEntry = .true.
      integer(4), save :: itransd
      character(2)     :: ctmp
      integer(4)       :: ios

      if (FirstEntry) then
        FirstEntry = .false.
        luntrans = io_new_unit( )
        itransd = 0
        write(ctmp,'(i2.2)') itransd
        transdnam = 'transdat.'//ctmp
        open(luntrans,file=transdnam,form='unformatted',status='replace',      &
             iostat=ios)
        if (ios /= 0) call exitme(1, 'Cannot open '//trim(transdnam))
      else
        close(luntrans)
        luntrans = io_new_unit( )
        itransd = itransd + 1
        write(ctmp,'(i2.2)') itransd
        transdnam = 'transdat.'//ctmp
        open(luntrans,file=transdnam,form='unformatted',status='replace',      &
             iostat=ios)
        if (ios /= 0) call exitme(1, 'Cannot open '//trim(transdnam))
      endif

   end subroutine transdat_new


!===============================================================================


    subroutine transdat_close( )

      !- modules ---------------------------------------------------------------
      ! none

      !- dircetives ------------------------------------------------------------
      implicit none

      !- arguments -------------------------------------------------------------
      ! none

      !- local vars ------------------------------------------------------------
      ! none

      close(luntrans)

    end subroutine transdat_close
 
 
 !==============================================================================


    subroutine alloc_global( narea )
    
      !- modules ---------------------------------------------------------------
      use cmod_params,   only : iw2, iw3, mainarea

      !- dircetives ------------------------------------------------------------
      implicit none

      !- arguments -------------------------------------------------------------
      integer(4), intent(in) :: narea

      !- local vars ------------------------------------------------------------
      integer(4) :: ia, itr, ktr, n2d, n3d
      
      allocate ( utr(narea), vtr(narea), temputr(narea), tempvtr(narea),       &
                 saltutr(narea), saltvtr(narea), utrans(narea,ntransdat),      &
                 vtrans(narea,ntransdat), temputrans(narea,ntransdat),         &
                 tempvtrans(narea,ntransdat), saltutrans(narea,ntransdat),     &
                 saltvtrans(narea,ntransdat), wtrans(ntransdat),               &
                 ttrans(ntransdat), strans(ntransdat) )

      do ia = 1, narea
        n2d = iw2(ia)
        n3d = iw3(ia)

        allocate( utr(ia)%p(0:n3d), vtr(ia)%p(0:n3d), temputr(ia)%p(0:n3d),    &
                  tempvtr(ia)%p(0:n3d),saltutr(ia)%p(0:n3d),                   &
                  saltvtr(ia)%p(0:n3d) )

        do itr = 1, ntransdat
          ktr = ktransmax(itr)
          if ( ia == mainarea ) then
            allocate ( wtrans(itr)%p(0:ktr,1:3,ntrans(itr)),                   &
                       ttrans(itr)%p(0:ktr,1:3,ntrans(itr)),                   &
                       strans(itr)%p(0:ktr,1:3,ntrans(itr)) )
          endif
          allocate ( utrans(ia,itr)%p(ktr,0:n2d), vtrans(ia,itr)%p(ktr,0:n2d), &
                     temputrans(ia,itr)%p(ktr,0:n2d),                          &
                     tempvtrans(ia,itr)%p(ktr,0:n2d),                          &
                     saltutrans(ia,itr)%p(ktr,0:n2d),                          &
                     saltvtrans(ia,itr)%p(ktr,0:n2d)   )
        enddo
      enddo
     
    end subroutine alloc_global  


!===============================================================================


    subroutine alloc_local( narea )
    
      !- modules ---------------------------------------------------------------
      use cmod_params,       only : me1, mmx, nmx
      use cmod_local_arrays, only : me_l
      use dmi_mpi,           only : dd
 
      !- dircetives ------------------------------------------------------------
      implicit none

      !- arguments -------------------------------------------------------------
      integer(4), intent(in) :: narea
      
      !- local vars ------------------------------------------------------------
      integer(4) :: ia, itr, ktr, n2d, n3s


      allocate ( transupt_l(narea), transvpt_l(narea), utr_l(narea),           &
                 vtr_l(narea), temputr_l(narea),tempvtr_l(narea),              &
                 saltutr_l(narea), saltvtr_l(narea), utrans_l(narea,ntransdat),&
                 vtrans_l(narea,ntransdat), temputrans_l(narea,ntransdat),     &
                 tempvtrans_l(narea,ntransdat), saltutrans_l(narea,ntransdat), &
                 saltvtrans_l(narea,ntransdat) )

      do ia = 1, narea
         
        !- Allocate the local arrays if MPI is used ----------------------------
        n2d = dd(ia)%iwet2
        n3s = dd(ia)%nwet3 + dd(ia)%halo3

        allocate( transupt_l(ia)%p(0:n2d),transvpt_l(ia)%p(0:n2d),             &
                  utr_l(ia)%p(0:n3s), vtr_l(ia)%p(0:n3s),                      &
                  temputr_l(ia)%p(0:n3s), tempvtr_l(ia)%p(0:n3s),              &
                  saltutr_l(ia)%p(0:n3s), saltvtr_l(ia)%p(0:n3s)   )

        do itr = 1, ntransdat
          ktr = ktransmax(itr)
          allocate ( utrans_l(ia,itr)%p(ktr,0:n2d),                            &
                     vtrans_l(ia,itr)%p(ktr,0:n2d),                            &
                     temputrans_l(ia,itr)%p(ktr,0:n2d),                        &
                     tempvtrans_l(ia,itr)%p(ktr,0:n2d),                        &
                     saltutrans_l(ia,itr)%p(ktr,0:n2d),                        &
                     saltvtrans_l(ia,itr)%p(ktr,0:n2d)  )
        enddo

        call task_global2local_trans ( ia, mmx(ia), nmx(ia), me1(ia)%p,        &
                                       me_l(ia)%p, transupt(ia)%p,             &
                                       transvpt(ia)%p, transupt_l(ia)%p,       &
                                       transvpt_l(ia)%p )

      enddo

    end subroutine alloc_local
    

!===============================================================================


    subroutine task_global2local_trans ( ia, mmx_g,nmx_g,me_g, me_l,transupt_g,&
                                       transvpt_g, transupt_l, transvpt_l )
      !-------------------------------------------------------------------------
      ! Here we do the extraction of meteorological variables from the global
      ! input arrays, which are in the old permutation order, to the local 
      ! arrays, which are permuted to the task ordering.
      !-------------------------------------------------------------------------

      !- modules ---------------------------------------------------------------
      use dmi_mpi, only : dd

      !- directives ------------------------------------------------------------
      implicit none

      !- arguments -------------------------------------------------------------
      integer(4), intent(in)              :: ia, mmx_g, nmx_g
      integer(4), intent(in)              :: me_g(0:,0:)
      integer(4), intent(in)              :: me_l(0:,0:)
      logical, intent(in),  dimension(0:) :: transupt_g,transvpt_g
      logical, intent(out), dimension(0:) :: transupt_l,transvpt_l

      !- local vars ------------------------------------------------------------
      integer(4) :: il, iu, jl, ju, i, j, ns, ms, is


      !- Return if this task has nothing to do in this area --------------------
      il = dd(ia)%low_i
      iu = dd(ia)%up_i
      jl = dd(ia)%low_j
      ju = dd(ia)%up_j
      if (max(il, iu, jl, ju) == 0) return


      is  = 0
      do j=1,nmx_g
        do i=1,mmx_g
          ms = me_g(i,j)
          if (ms <= 0) cycle
          is = is + 1
          if (j < jl .or. j > ju .or. i < il .or. i > iu) cycle
          ns = me_l(i,j)
          if (ns > 0) then
            transupt_l(ns) = transupt_g(is)
            transvpt_l(ns) = transvpt_g(is)
          endif
        enddo
      enddo

      !- set the on-land values ------------------------------------------------
      transupt_l(0) = transupt_g(0)
      transvpt_l(0) = transvpt_g(0)

    end subroutine task_global2local_trans

    
end module transport_variables
