module transports
  !-----------------------------------------------------------------------------
  !- Calculation of transports (water-, heat- and salt-transports)
  !-----------------------------------------------------------------------------

  implicit none
  private
  
  public  :: transport_init, transport_calc, transport_output
  
contains

  subroutine transport_init( narea )
  
    !- modules -----------------------------------------------------------------
    use cmod_local_arrays,   only : iw2_l, iw3_l
    use transport_variables, only : ntransdat, ktransmax,                      &
                                    utr_l, vtr_l, temputr_l, tempvtr_l,        &
                                    saltutr_l, saltvtr_l, utrans_l, vtrans_l,  &
                                    temputrans_l, tempvtrans_l, saltutrans_l,  &
                                    saltvtrans_l,                              &
                                    transport_usage, transport_parameters,     &
                                    alloc_global, alloc_local
    use transports2zero,     only : modtrans2zero, transecttrans2zero

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: narea
    
    !- local variables ---------------------------------------------------------
    integer(4) :: ia, itr
    
    call transport_usage (   )
  
    call transport_parameters ( narea )
  
    !- Allocate the additional transport-arrays
    !- the global ones
    call alloc_global ( narea )
    !- and the local ones
    call alloc_local ( narea )

!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia, itr)
    !- Initialise the global variables    
    do ia = 1, narea
      !- Initialise variables    
      call modtrans2zero( iw3_l(ia), utr_l(ia)%p, vtr_l(ia)%p, temputr_l(ia)%p,&
                          tempvtr_l(ia)%p, saltutr_l(ia)%p, saltvtr_l(ia)%p )
      
      do itr = 1,ntransdat
        call transecttrans2zero (iw2_l(ia), ktransmax(itr), utrans_l(ia,itr)%p,&
                                 vtrans_l(ia,itr)%p, temputrans_l(ia,itr)%p,   &
                                 tempvtrans_l(ia,itr)%p,saltutrans_l(ia,itr)%p,&
                                 saltvtrans_l(ia,itr)%p )
      enddo
    enddo
!$OMP END PARALLEL

  end subroutine transport_init


!===============================================================================


  subroutine transport_calc( )
  
    !- modules -----------------------------------------------------------------
    use cmod_local_arrays,   only : mm1k_l, iw2_l, mm1k_l, ind_l, kh_l,        &
                                    khu_l, khv_l, hu_l, hv_l, z_l, u_l, v_l,   &
                                    cmp_l, ti_l
    use cmod_params,         only : narea
    use transport_kernel,    only : water_trans, transform
    use transport_variables, only : izeittrans, ntransdat, ktransmax, hztrans, &
                                    transupt_l, transvpt_l,                    &
                                    utr_l, vtr_l, temputr_l, tempvtr_l,        &
                                    saltutr_l, saltvtr_l, utrans_l, vtrans_l,  &
                                    temputrans_l, tempvtrans_l, saltutrans_l,  &
                                    saltvtrans_l

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    ! none
    
    !- local vars --------------------------------------------------------------
    integer(4) :: ia, itr
    
    izeittrans = izeittrans + 1

!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia)
    do ia=1,narea
      call water_trans( iw2_l(ia), ind_l(ia)%p, mm1k_l(ia)%p, kh_l(ia)%p,      &
                        hu_l(ia)%p, hv_l(ia)%p, u_l(ia)%p, v_l(ia)%p,          &
                        cmp_l(ia)%p, utr_l(ia)%p, vtr_l(ia)%p, temputr_l(ia)%p,&
                        tempvtr_l(ia)%p, saltutr_l(ia)%p, saltvtr_l(ia)%p,     &
                        transupt_l(ia)%p, transvpt_l(ia)%p )
    enddo
!$OMP END PARALLEL

!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia,itr)
    do itr=1,ntransdat
      do ia=1,narea
        call transform( iw2_l(ia), ind_l(ia)%p, mm1k_l(ia)%p,                  &
                        khu_l(ia)%p, khv_l(ia)%p, ti_l(ia)%p, z_l(ia)%p,       &
                        hu_l(ia)%p, hv_l(ia)%p, ktransmax(itr),                &
                        hztrans(itr)%p, transupt_l(ia)%p, transvpt_l(ia)%p,    &
                        utr_l(ia)%p, vtr_l(ia)%p, temputr_l(ia)%p,             &
                        tempvtr_l(ia)%p, saltutr_l(ia)%p, saltvtr_l(ia)%p,     &
                        utrans_l(ia,itr)%p, vtrans_l(ia,itr)%p,                &
                        temputrans_l(ia,itr)%p, tempvtrans_l(ia,itr)%p,        &
                        saltutrans_l(ia,itr)%p, saltvtrans_l(ia,itr)%p  )
      enddo
    enddo
!$OMP END PARALLEL
 
  end subroutine transport_calc
     

!===============================================================================


  subroutine transport_output ( narea, ctim )

    !- modules -----------------------------------------------------------------
    use cmod_arrays,         only : ind , dx, dy, covu
    use cmod_params,         only : iw2
    use cmod_local_arrays,   only : iw2_l
    use dmi_mpi,             only : dmpi_gather, mpi_io_serial
    use dmi_mpi_global,      only : mpi_io_rank
    use exits,               only : exitme
    use transport_kernel,    only : calculate_mean, transects
    use transport_variables, only : ntransdat, izeittrans, ntrans, ktransmax,  &
                                    transupt_l, transvpt_l, grid, ctru, ctrv,  &
                                    xdirectfac, ydirectfac, transdnam,luntrans,&
                                    utrans, vtrans, temputrans, tempvtrans,    &
                                    saltutrans, saltvtrans, utrans_l, vtrans_l,&
                                    temputrans_l, tempvtrans_l, saltutrans_l,  &
                                    saltvtrans_l, wtrans, ttrans, strans
    use transports2zero,     only : transecttrans2zero

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: narea
    character(19), intent(in) :: ctim

    !- local vars --------------------------------------------------------------
    integer(4) :: itr, ia, kt, ios
    real(8)    :: rtsteps


    rtsteps = real(izeittrans, kind = 8 )

!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia,itr)
    do itr = 1, ntransdat
      do ia = 1, narea
        call calculate_mean (iw2_l(ia), ktransmax(itr), rtsteps,               &
                             utrans_l(ia,itr)%p, vtrans_l(ia,itr)%p,           &
                             temputrans_l(ia,itr)%p, tempvtrans_l(ia,itr)%p,   &
                             saltutrans_l(ia,itr)%p, saltvtrans_l(ia,itr)%p,   &
                             transupt_l(ia)%p, transvpt_l(ia)%p)
      enddo
    enddo
!$OMP END PARALLEL


    do itr = 1, ntransdat

     !- Gather needed variables for output
!FIXME: A new gather-subroutine for transport output is needed!!
      do ia = 1, narea
        do kt = 1, ktransmax(itr)
          call dmpi_gather(ia,(1), utrans_l(ia,itr)%p(kt,:),                   &
                                   utrans(ia,itr)%p(kt,:), mpi_io_rank)
          call dmpi_gather(ia,(1), vtrans_l(ia,itr)%p(kt,:),                   &
                                   vtrans(ia,itr)%p(kt,:), mpi_io_rank)
          call dmpi_gather(ia,(1), temputrans_l(ia,itr)%p(kt,:),               &
                                   temputrans(ia,itr)%p(kt,:), mpi_io_rank)
          call dmpi_gather(ia,(1), tempvtrans_l(ia,itr)%p(kt,:),               &
                                   tempvtrans(ia,itr)%p(kt,:), mpi_io_rank)
          call dmpi_gather(ia,(1), saltutrans_l(ia,itr)%p(kt,:),               &
                                   saltutrans(ia,itr)%p(kt,:), mpi_io_rank)
          call dmpi_gather(ia,(1), saltvtrans_l(ia,itr)%p(kt,:),               &
                                   saltvtrans(ia,itr)%p(kt,:), mpi_io_rank)
        enddo
      enddo      

      if (mpi_io_serial) then
        call transects ( narea, iw2, ind, dx, dy, covu, ktransmax(itr),        &
                         ntrans(itr), grid(itr)%p, xdirectfac(itr)%p,          &
                         ydirectfac(itr)%p, ctru(:,itr), ctrv(:,itr),          &
                         utrans(:,itr), vtrans(:,itr), temputrans(:,itr),      &
                         tempvtrans(:,itr), saltutrans(:,itr),                 &
                         saltvtrans(:,itr), wtrans(itr)%p, ttrans(itr)%p,      &
                         strans(itr)%p )

        !- Write the transect transports to output-file
        write(luntrans, iostat=ios) ctim, itr, ntrans(itr), ktransmax(itr)
        if(ios /= 0) then
          call exitme(1,'Could not write header in '//trim(transdnam))
        endif
        write(luntrans, iostat=ios) wtrans(itr)%p, ttrans(itr)%p, strans(itr)%p
        if(ios /= 0) then
          call exitme(1,'Could not write data in '//trim(transdnam))
        endif

      endif

      !- Set everything to zero again
      izeittrans = 0
      do ia = 1, narea
        call transecttrans2zero (iw2_l(ia), ktransmax(itr), utrans_l(ia,itr)%p,&
                                 vtrans_l(ia,itr)%p, temputrans_l(ia,itr)%p,   &
                                 tempvtrans_l(ia,itr)%p,saltutrans_l(ia,itr)%p,&
                                 saltvtrans_l(ia,itr)%p )
      enddo

    enddo

  end subroutine transport_output

end module transports
