module turbmodels
  
  !-----------------------------------------------------------------------------
  !
  ! Module containing constants and routines for turbulence models
  !
  ! Written for use in cmod at DMI
  ! by Per Berg, DMI
  !-----------------------------------------------------------------------------
  
  !- Global constants ----------------------------------------------------------
  use constants, only : g, rhow, rhoi, cpw, cdrag, zero, one, two, half,       &
                        five, six, ten, tenth

  !- Implicit directives -------------------------------------------------------
  implicit none
  private
  
  !- Control constants ---------------------------------------------------------
  !
  ! pdiffass ==
  !       .false. : use stability function for passive tracers from 
  !                 Canuto et. al., Part II (2002) with sign bug corrected.
  !                 (default). 
  !       .true.  : Optionally, use same diffusivity for passive tracer as
  !                 for salinity. 
  !
  ! prodtype == 
  !       1: Only resolved shear production.
  !       2: as 1, plus shear due to internal waves.
  !       3: as 1, plus shear due to Langmuir circulations.
  !       4: all three production terms
  !
  ! srfprod ==
  !       .false. : do not estimate production in the surface layer (default)
  !       .true.  : estimate production in the surface layer from below
  !
  ! tidetype ==
  !       0: none
  !       1: unresolved bottom shear 
  !      10: internal, baroclinic tides
  !     100: tidal drag, shallow seas   
  !            ... and combos hereof, e.g. 11 includes both 1 and 10.
  !            ref: Canuto et al. Part III (2010), Chapter 7.
  !
  logical,    save, public  :: pdiffass = .true. ! Passive tracer DIFFusion 
                                                 ! AS Salinity.
  integer(4), save, public  :: prodtype      = 2 ! resolved shear and shear due
                                                 ! to internal waves by default
  logical,    save, public  :: srfprod = .false.
  integer(4), save, private :: kb0 = 2

  integer(4), save, public  :: tidetype      = 0 ! no parameterization of 
                                                 ! unresolved production due to 
                                                 ! tides, by default

  !- Physical Constants --------------------------------------------------------
  real(8), parameter, private :: rho0       = rhow,                            &
                                 rhoice     = rhoi,                            &
                                 Cdice      = cdrag,                           &
                                 laminarvis = 0.000001_8,                      &
                                 Cpheat     = cpw,                             &
                                !alpha20C   = 0.0002_8,                        &
                                 vkarm      = 0.41_8

  !- Thermal Diffusivity -------------------------------------------------------
  !  thermd = k/(Cpheat*rho0)   [m^2/s] 
  !         approx. 1.35e-7 m^2/s
  !  
  !  with thermal conductivity:   
  !           k = 0.561 W/m/K @ 273 K
  !           k = 0.673 W/m/K @ 353 K
  !  specific heat:
  !           Cpheat = 4217 J/kg/C @  0 C
  !           Cpheat = 4186 J/kg/C @ 15 C
  real(8), parameter, private :: thermd = 0.58_8/(rho0*Cpheat)

  !- Constants related to turbulence models-------------------------------------
  real(8), parameter, private :: kappa=vkarm,                                  &
                                 ! "Classic" value of cmu0:
                                 ! cmu0 = 0.5562_8,          &
                                 ! Burchard&Bolding used with PartI:
                                 ! cmu0 = 0.5270_8,          &
                                 ! New value by P.Berg (DMI) used with PartII:
                                 cmu0 = 0.5234_8,                              &
                                 cmu04 = cmu0**4,                              &
                                 !----------------------------------------------
                                 ! Values for sigma_diss 
                                 ! Burchard & Bolding (2001):
                                 ! sigma_diss=1.02_8,                     &
                                 ! Rodi (1987), Launder&Spalding(1974),
                                 ! Meier (2001), Burchard&Baumert (1995): 
                                 sigma_diss=1.30_8,                            &
                                 ! Axell&Liungman (2001), Axell (2002):
                                 ! sigma_diss=1.08_8,                     &
                                 !----------------------------------------------
                                 !ceps1 = 1.44_8, ceps2 = 1.92_8,              &
                                 tke_min = 0.0000001_8,                        &
                                 diss_min = cmu04*tke_min**2/laminarvis,       &
                                 eddydif_min = laminarvis*0.01_8,              &
                                 !Rt_c = -1.0_8,                               &
                                 !Rt_min = -3.0_8, Rt_max = 5.0_8,             &
                                 Rg_max = 50.0_8,                              &
                                 !Rg_crit = 1.0_8,                             &
                                 F0 = 0.0009_8, Clc = 0.15_8,                  &
                                 !----------------------------------------------
                                 ! For k-eps model, use this:
                                 !sigma_tke_kepsilon = 1.0_8,                  &
                                 !Lkeps = 0.1_8,                               &
                                 ! alp_keps = -4.9_8,                     &
                                 !----------------------------------------------
                                 ! For k-omega model, use this:
                                 ! Umlauf et al (2003):
                                 sigma_tke_komega = 2.0_8,                     &
                                 sigma_omega      = 2.0_8,                     &
                                 ! Umlauf et al (2003):
                                 ! Lkome = 0.25_8, alp_kome = -2.7_8,     &
                                 ! HIRMOB:
                                 ! sigma_tke_komega = 1.30_8,             &
                                 ! sigma_omega      = 1.552_8,            &
                                 ! P.Berg (DMI):
                                 Lkome = 0.25_8, alp_kome = -2.5_8,            &
                                 sigma_tke = sigma_tke_komega,                 &
                                 !----------------------------------------------
                                 ! Umlauf et al (2003):
                                 !   W88:
                                 ! comega1 = 0.56_8, comega2 = 0.83_8,    &
                                 !   W98:
                                 ! comega1 = 0.52_8, comega2 = 0.8_8,     &
                                 ! HIROMB:
                                 ! comega1 = 0.500_8, comega2 = 0.833_8,  &
                                 ! P.Berg (DMI):
                                 comega1 = 0.53_8, comega2 = 0.84_8,           &
                                 !----------------------------------------------
                                 z0b = 0.1_8, z0s = 0.1_8,                     &
                                 !----------------------------------------------
                                 ! Use this Neumann bc, or bcs derived 
                                 ! Neumann bc:
                                 mwindN = 100.0_8,                             &
                                 ! Use this Dirichlet bc:
                                 !mwindD = 1.0_8,                              &

                                 mwind  = mwindN,                              &
                                 !----------------------------------------------
                                 ! Values for ceps3 during
                                 ! stable stratification:   ceps3stable
                                 ! unstable stratification: ceps3unstable
                                 ! Umlauf et al (2003):
                                 !ceps3stable = -0.623_8,                      &
                                 ! Burchard & Bolding (2001):
                                 ! ceps3stable = -0.629_8*ceps1,          &
                                 ! simple calc by P.Berg (2006):
                                 ! ceps3stable = 1.67_8,                  &
                                 ! Axell&Liungman (2001): 
                                 ! ceps3stable = -1.10_8,                 &
                                 ! Meier (2001): 
                                 ! ceps3stable = 0.0_8,                   &
                                 ! Axell&Liungman (2001), Axell (2002),
                                 ! ceps3unstable = 1.00_8
                                 ! Meier (2001), Rodi (1987):
                                 !ceps3unstable = ceps1,                       &
                                 !----------------------------------------------
                                 ! Values for ceps3 during stable 
                                 ! stratification and patchy/non-patchy 
                                 ! turbulence:
                                 !ceps3nonpatchy =                             &
                                 !  ceps2 - (ceps2-ceps1)/0.18_8,              &
                                 !ceps3patchy    =                             &
                                 !  ceps2 - (ceps2-ceps1)/0.09_8,              &
                                 !----------------------------------------------
                                 ! Values for comega3 during
                                 ! stable stratification:   comega3stable
                                 ! unstable stratification: comega3unstable
                                 ! Umlauf et al (2003):
                                 !comega3stable = ceps3stable - 1.0_8,         &
                                 ! HIROMB:
                                 ! comega3unstable = 1.0_8,               &
                                 ! here we take:
                                 comega3unstable = 0.0_8,                      &
                                 !----------------------------------------------
                                 ! Values for ceps3 during stable 
                                 ! stratification and patchy/non-patchy 
                                 ! turbulence:
                                 comega3nonpatchy =                            &
                                   comega2 - (comega2-comega1)/0.18_8,         &
                                 comega3patchy    =                            &
                                   comega2 - (comega2-comega1)/0.09_8
                                 !----------------------------------------------
                                 
  !- Constants in stability functions Part I by Canuto et al (2001)
  !real(8), parameter, private :: l1 = 0.127_8,  l2 = 0.00336_8,               &
  !                               l3 = 0.0906_8, l4 = 0.101_8,                 &
  !                               l5 = 11.2_8,   l6 = 0.4_8,                   &
  !                               l7 = 0.0_8,     l8 = 0.318_8,                &
  !                               s0 = 1.5_8*l1*l5**2,                         &
  !                               s1 = -l4*(l6+l7)                             &
  !                                    +2.0_8*l4*l5*(l1-l2/3.0_8-l3)           &
  !                                    +1.5_8*l1*l5*l8,                        &
  !                               s2 = -3.0_8*l1*(l6**2-l7**2)/8.0_8,          &
  !                               s4 = 2.0_8*l5,                               &
  !                               s5 = 2.0_8*l4,                               &
  !                               s6 = 2.0_8*l5*(3.0_8*l3**2-l2**2)/3.0_8      &
  !                                    -0.5_8*l5*l1*(3.0_8*l3-l2)              &
  !                                    +0.75_8*l1*(l6-l7),                     &
  !                               dd0 = 3.0_8*l5**2,                           &
  !                               dd1 = l5*(7.0_8*l4+3.0_8*l8),                &
  !                               dd2 = l5**2*(3.0_8*l3**2-l2**2)              &
  !                                     -0.75_8*(l6**2-l7**2),                 &
  !                               dd3 = l4*(4.0_8*l4+3.0_8*l8),                &
  !                               dd4 = l4*(l2*l6-3.0_8*l3*l7                  &
  !                                              -l5*(l2**2-l3**2))            &
  !                                     +l5*l8*(3.0_8*l3**2-l2**2),            &
  !                               dd5 = 0.25_8*(l2**2-3.0_8*l3**2)             &
  !                                           *(l6**2-l7**2)

  !- Constants in stability functions Part II by Canuto et al (2002)
  real(8), parameter, private :: pi1 = 0.08372_8,   pi4 = pi1,                 &
                                 pi2 = 1.0_8/3.0_8, pi3 = 0.72_8,              &
                                 !pi5 = pi3,                                   &
                                 !p1  = 0.832_8,     p2  = 0.545_8,            &
                                 !p3  = 0.2093_8,    p4  = 0.0323_8,           &
                                 !p5  = 0.0538_8,    p6  = 0.0698_8,           &
                                 !p7  = 1.6666_8,    p8  = p1,                 &
                                 !p9  = 0.2511_8,    p10 = p5,                 &
                                 !p11 = 0.1163_8,                              &
                                 a1  = 0.3022_8,    a2  = 0.2986_8,            &
                                 a3  = 0.064780_8,  a4  =-3.99459_8,           &
                                 a5  =-1.8493_8,                               &
                                 b1  =-0.0625_8,    b2  =-0.1163_8,            &
                                 b3  = 0.5702_8,    b4  =-0.9689_8,            &
                                 b5  =-2.0930_8,    b6  =-0.0538_8,            &
                                 b7  =-0.13488_8,                              &
                                 d1  = 0.03201_8,   d2  = 0.0318_8,            &
                                 d3  = 0.00686_8,   d4  =-0.0289_8,            &
                                 d5  =-0.04272_8,   d6  =-0.019780_8,          &
                                 d7  =-0.0028750_8, d8  =-0.41319_8,           &
                                 d9  =-0.1912_8,    d10 = 1.1773_8,            &
                                 d11 = 1.1612_8,    d12 = 0.2523_8,            &
                                 d13 = 1.1857_8,    d14 =-10.7721_8,           &
                                 d15 =-4.9871_8
  real(8), parameter, private :: Rrho_crit = 0.62_8, ccrit = 0.85_8

  
  !- Other constants -----------------------------------------------------------
  real(8), parameter, private :: tinyv=1.0e-15_8
  real(8), parameter, private :: pow1d3 = 0.33333_8, pow2d3 = 0.66667_8
  real(8), parameter, private :: ww = 0.54_8 ! = (1/1.23)**3 from Section 6.3.2,
                                           !   Fischer et al, "Mixing in Inland
                                           !   and Coastal Waters", 
                                           !   Academic Press, 1979.

  !- Public routines -----------------------------------------------------------
  public  :: TurbModel, InitTurb
 
  
  !- Public routines -----------------------------------------------------------
  private :: Production, KineticEnergy, Frequency, LowerLimit, ZeroProduction, &
             BoundaryCondtions, PartIIDiff, CheckRgCrit, BuoyancyProduction,   &
             TriSolv, densit
contains

  ! HACK TO ENSURE INLINING - IPA and craypat doesn't work, cf. trisolver.f90
  subroutine TriSolv(s, a, b, c, d, n)
  !-----------------------------------------------------------------------------
  !  standard tridiagonal solver
  !-----------------------------------------------------------------------------

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: n
    real(8), intent(out)   :: s(1:)
    real(8), intent(in)    :: a(1:), b(1:), c(1:), d(1:)

    !-local vars ---------------------------------------------------------------
    integer(4)           :: i
    real(8)              :: e
    real(8)              :: f(n)

    !- forward sweep -----------------------------------------------------------
    e    = b(1)
    s(1) = d(1)/e
    do i = 2, n
      f(i) = c(i-1)/e
      e    = b(i) - a(i)*f(i)
      s(i) = (d(i) - a(i)*s(i-1))/e
    enddo

    !- backward sweep ----------------------------------------------------------
    do i = n-1, 1, -1
      s(i) = s(i) - f(i+1)*s(i+1)
    enddo

    !- That's it! --------------------------------------------------------------

  end subroutine TriSolv

  !-----------------------------------------------------------------------------
  ! Simple k-e turbulence model
  !
  ! The indexing of the grid cells is from surface to bottom.
  ! The surface index is 1 and the bottom index kb. 
  ! 
  ! Global variables as input and output of the model:
  !
  ! VECTORS:
  !
  ! Pstot: Total shear production, located at cell interfaces. Ranging 
  !        from 2 to kb.
  !
  ! Ps: Resolved shear production of turbulence, located at cell 
  !     interface. Ranging from 2 to kb.
  ! 
  ! Piw: Shear production due to internal waves, located at cell
  !      interfaces. Ranging from 2 to kb.
  !
  ! Plc: Shear production due to Langmuir circulations, located at cell
  !      interfaces. Ranging from 2 to kb.
  !
  ! Pb: Buoyancy production of turbulence, located at cell interfaces.
  !     Ranging from 2 to kb.
  !
  ! u:  x-component of velocity vector, located at scalar points.
  !     Ranging from 1 to kb.
  ! 
  ! v:  y-component of velocity vector, located at scalar points.
  !     Ranging from 1 to kb.
  !
  ! N2: Buoyancy frequency squared, defined as -g/rho0*drho/dz, where 
  !     g is gravitational acceleration, rho0 a reference density 
  !     and rho the actual density. N2 is located at cell interfaces.
  !     Ranging from 1 to kb.
  ! 
  ! Sh2: Shear squared. Located at cell interfaces. Ranging from 1 to kb
  !
  ! Rg: Gradient Richardson number Rg = N2/Sh2. Located at cell 
  !     interfaces. Ranging from 1 to kb.
  !
  ! Rt: Turbulent Richardson number Rt = N2*(k/eps)**2. Located at cell 
  !     interfaces. Ranging from 1 to kb.
  !
  ! cmu: Stability function. Located at cell interfaces. Ranging from
  !      1 to kb.
  ! 
  ! sigma_t: Prandtl-Schmidt number. Located at cell interfaces. Ranging
  !          from 1 to kb.
  ! 
  ! tke: Turbulent kinetic energy, located at scalar points. Ranging
  !      from 1 to kb.
  ! 
  ! tke_old: TKE from former time step which for stability reasons is
  !          put into the dissipation-equation.
  !
  ! diss: Dissipation rate of TKE, located at scalar points. Ranging
  !       from 1 to kb.
  !
  ! eddyvis: Eddy viscosity, located at cell interfaces. Ranging from
  !          1 to kb. 
  !
  ! eddydif: Eddy diffusivity for T (and S), located at cell interfaces.
  !          Ranging from 1 to kb. 
  !
  ! Ks: Eddy diffusivity for S, located at cell interfaces.
  !     Ranging from 1 to kb. 
  !
  ! rT: Temperature gradient, g*alphaT*dT/dz, located at cell interfaces.
  !     Ranging from 1 to kb. 
  !
  ! rS: Salinity gradient, g*alphaS*dS/dz, located at cell interfaces.
  !     Ranging from 1 to kb. 
  !
  ! dz: Local distance between centers of scalar grid cells in m, 
  !     ranging from 2 to kb.
  !
  ! dh: Local thickness of scalar grid cells in m, ranging from 1 to kb.
  !
  ! SCALARS:
  !
  ! u_taus: Shear velocity in m/s at the surface. 
  !
  ! u_taub: Shear velocity in m/s at the bottom.
  !
  ! Bflx: Buoyancy flux through surface (in m**2/s**3).
  !
  !-----------------------------------------------------------------------------
  

  subroutine TurbModel (kx, iw2, mm, ind, kh, dt, rdt,                         &
                        casus, ice, uice, vice, Q,                             &
                        wu, wv, tauxs, tauys, hz, rho, u3D, v3D,               &
                        avv, avtse, tke3D, diss3D, z0srf, cmp, ompidx, tracers,&
                        ia, tmap)
    !---------------------------------------------------------------------------
    !
    ! Interface between cmod and turbulence models
    !
    ! Author:  Per Berg, September 2004.
    !
    !---------------------------------------------------------------------------

    !- modules -----------------------------------------------------------------
    use dmi_omp,   only : domp_get_domain
    use constants, only : cfreeze, saltmax, wtmpmax, cmaxdens, Tintercept
    use tideparam, only : get_tide_map, cdtide, zetatide

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)                     :: kx, iw2, ia
    integer(4), intent(in)                     :: mm(:,0:,0:)
    integer(4), intent(in)                     :: ind(:,:), kh(0:)
    real(8),    intent(in)                     :: dt, rdt
    logical,    intent(in)                     :: tracers
    logical,    intent(in),    dimension(0:)   :: casus
    real(8),    intent(in),    dimension(0:)   :: uice, vice, Q
    real(8),    intent(in),    dimension(0:)   :: wu, wv
    real(8),    intent(in),    dimension(0:)   :: tauxs, tauys
    real(8),    intent(in),    dimension(0:)   :: hz, rho, u3D, v3D
    real(8),    intent(in)                     :: cmp(1:,0:), ice(1:,0:)
    real(8),    intent(in)                     :: tmap(0:,1:)
    real(8),    intent(inout), dimension(0:)   :: avv, z0srf
    real(8),    intent(inout)                  :: avtse(1:,0:)
    real(8),    intent(inout), dimension(0:)   :: tke3D, diss3D
    integer(4), intent(inout)                  :: ompidx(1:,1:)

    !- local vars --------------------------------------------------------------
    integer(4)               :: i, j, ii, k, kb, idx, idm, kmix
    real(8), dimension(2:kx) :: Pstot, Ps, Piw, Plc, Pb, dz, ddz, buoyfreq
    real(8), dimension(1:kx) :: eddyvis, eddydif, N2, Sh2
    real(8), dimension(1:kx) :: tke, tke_old, diss, u, v, dh, Ks, rT, rS
    real(8), dimension(2:kx) :: ppk, spk, smk, tpk, tmk
    logical                  :: no_ice
    real(8)                  :: iceconc, Ua, Va, depth, rough, q3, W3
    real(8)                  :: u_taus, u_taub, taus, tauice, Bflx, H, Ws
    real(8)                  :: pp, ss, tt
    real(8)                  :: dtp, dtm, dsp, dsm, dz_min
    real(8)                  :: alphas, alphat, RHS, ep0, om0
    real(8), parameter       :: gdrho0 = -g/rho0 
    real(8), parameter       :: delta = 0.01_8, utau0 = 0.001_8
    real(8), parameter       :: gd2drho0 = -gdrho0/(2.0_8*delta)
    real(8), parameter       :: d2drho0 = one/(2.0_8*delta*rho0)
    real(8)                  :: eddydif_max, eddyvis_max, tmaxdens, tfreeze
    real(8)                  :: const, buoyfreqmean, sum1, sum2
    integer(4)               :: n2dl, n2du, mm2, mmw, mmn, mw2, mn2
    real(8)                  :: utide, ustr2, dtide, iwtide 
    logical                  :: tt1, tt2, tt3
  

    !- set loop invariants -----------------------------------------------------
    rough   = rdt/dt

    call domp_get_domain (kh, 1, iw2, n2dl, n2du, ompidx)

    !- run through each water column -------------------------------------------
    wetpointloop: do ii=n2dl,n2du

      !- obtain indices --------------------------------------------------------
      kb = kh(ii)
      if (kb < 1) cycle wetpointloop 

      i = ind(1,ii)
      j = ind(2,ii)


      !- extract 1D column vars from 3D vars -----------------------------------
      !- simple extraction of turbulence vars:
      tke(1)     = tke3D(ii)
      diss(1)    = diss3D(ii)
      eddyvis(1) = avv(ii)
      eddydif(1) = avtse(1,ii) - thermd
      if (kb > 1) then
        mm2 = mm(2,i,j)   

        tke (2:kb) = tke3D (mm2:mm2+kb-2)
        diss(2:kb) = diss3D(mm2:mm2+kb-2)

        eddyvis(2:kb) = avv(mm2:mm2+kb-2)
        eddydif(2:kb) = avtse(1,mm2:mm2+kb-2) - thermd
      endif
      tke_old(1:kb) = tke(1:kb)

      !- cell-centred velocity components: 
      mmw  = mm(1,i,j-1)
      mmn  = mm(1,i-1,j)
      u(1) = half*(u3D(ii) + u3D(mmw)) 
      v(1) = half*(v3D(mmn) + v3D(ii))
      if (kb > 1) then
        mw2 = mm(2,i,j-1)
        mn2 = mm(2,i-1,j)

        u(2:kb) = half*u3D(mm2:mm2+kb-2)
        v(2:kb) = half*v3D(mm2:mm2+kb-2)
        
        do k=2,min(kb,kh(mmw))
          u(k) = u(k) + half*u3D(mw2+k-2)
        enddo
        do k=2,min(kb,kh(mmn))
          v(k) = v(k) + half*v3D(mn2+k-2)
        enddo
      endif

      !- extract grid spacing, cell height
      !  and pp, sp. tp, sm, tm, dh
      dh(1) = max( hz(ii), z0s )
      pp    = tenth*one
      if (kb > 1) then
        ! k=2 unrolled:
        tmk(2) = cmp(1,ii)
        tpk(2) = cmp(1,mm2)
        smk(2) = cmp(2,ii)
        spk(2) = cmp(2,mm2)

        pp     = pp + tenth*hz(ii)
        ppk(2) = pp

        dh(2)  = hz(mm2)
        dz(2)  = half*(dh(2) + dh(1))
        ddz(2) = one/dz(2)

        do k=3,kb
          idm = mm2 + k - 3
          idx = mm2 + k - 2

          tmk(k) = cmp(1,idm)
          tpk(k) = cmp(1,idx)
          smk(k) = cmp(2,idm)
          spk(k) = cmp(2,idx)

          pp     = pp + tenth*hz(idm)
          ppk(k) = pp
        enddo
        do k=3,kb
          dh(k)  = hz(mm2 + k - 2)
          dz(k)  = half*(dh(k) + dh(k-1))
          ddz(k) = one/dz(k)
        enddo
      endif

      !- buoyancy frequency squared:
      N2(1) = zero      
      do k=2,kb
        pp    = ppk(k)
        N2(k) = (densit(smk(k),tmk(k),pp) - densit(spk(k),tpk(k),pp))          &
                *gdrho0*ddz(k)
      enddo

      !- maxium diffusivities for this water column:
      if (kb >= 2) then
        dz_min = minval(dz(2:kb))
      else
        dz_min = dh(1)
      endif
      eddyvis_max = ten*dz_min*dz_min/dt
      eddydif_max = eddyvis_max
     
      !- gradient Richardson number and mean shear:
      Sh2(1) = zero
      do k=2,kb
        Sh2(k) = ((u(k-1)-u(k))*ddz(k))**2 + ((v(k-1)-v(k))*ddz(k))**2
      enddo

      !- T and S gradients:
      Ks(1) = avtse(2,ii)
      if (kb > 1) Ks(2:kb) = avtse(2,mm2:mm2+kb-2)

      rT(1) = zero
      rS(1) = zero

      do k=2,kb
        pp = ppk(k)
        tt = half*(tmk(k)+tpk(k))
        ss = half*(smk(k)+spk(k))
          
        ! local thermal expansion coefficient:
        dtp    = min( max(-cfreeze*ss, tt + delta), wtmpmax)
        dtm    = min( max(-cfreeze*ss, tt - delta), wtmpmax)
        alphat = -(densit(ss,dtp,pp)-densit(ss,dtm,pp))*gd2drho0

        ! local haline contraction coefficient:
        dsp    = min( max(zero, ss + delta), saltmax)
        dsm    = min( max(zero, ss - delta), saltmax)
        alphas =  (densit(dsp,tt,pp)-densit(dsm,tt,pp))*gd2drho0

        ! local gradients:
        rT(k) = alphat*(tmk(k)-tpk(k))*ddz(k)
        rS(k) = alphas*(smk(k)-spk(k))*ddz(k)
      enddo
      !-------------------------------------------------------------------------


      !- extract 0D surface/bottom vars from 2D/3D vars ------------------------

      !- get parameters for effects from tides:
      if (tidetype > 0) then
        call get_tide_map(ia, ii, tidetype, tmap, utide, iwtide, tt1, tt2, tt3 )
      else
        tt3 = .false.
      endif
      
      !- wind components:
      Ua = wu(ii)
      Va = wv(ii)

      !- ice parameters:
      no_ice  = casus(ii)
      iceconc = ice(2,ii)

      !- water depth:
      depth = zero
      do k=1,kb
        depth = depth + dh(k)
      enddo
      
      !- friction velocity, surface:
      taus = half*sqrt( (tauxs(mmw)+tauxs(ii))**2 + (tauys(mmn)+tauys(ii))**2 )
      if (.not.no_ice) then
        !- treat ice conditions:
        tauice = Cdice*rhoice*                                                 &
                 (  (half*(uice(mmw) + uice(ii)) - u(1))**2                    &
                  + (half*(vice(mmn) + vice(ii)) - v(1))**2 )
        taus   = (one-iceconc)*taus + iceconc*tauice
      end if
      u_taus = sqrt( taus/rho0 )

      !- friction velocity, bottom:
      if (tt3) then
        ! include tidal drag, shallow seas:
        ustr2  = u(kb)**2+v(kb)**2
        u_taub = sqrt( rough*max(sqrt(ustr2)*sqrt(ustr2+utide*utide), utau0) )
      else  
        u_taub = sqrt( rough*max(u(kb)**2+v(kb)**2 , utau0) )
      endif

      !- determine buoyancy range in S-T space:
      tt       = cmp(1,ii)
      ss       = cmp(2,ii)
      tmaxdens = -cmaxdens*ss + Tintercept
      tfreeze  = -cfreeze*ss

      ! local thermal expansion coefficient:
      pp     = tenth*(one+half*dh(1))
      dtp    = min( tt + delta, wtmpmax)
      dtm    = min( max(-cfreeze*ss, tt - delta), wtmpmax)
      alphat = -(densit(ss,dtp,pp)-densit(ss,dtm,pp))*d2drho0

      !- buoyancy flux:
      if (tfreeze < tt .and. tt < tmaxdens) then
        Bflx = g*alphat*max(Q(ii),zero)/(rho(ii)*Cpheat)
      else
        Bflx = g*alphat*min(Q(ii),zero)/(rho(ii)*Cpheat)
      endif

      !- determine mixed layer depth:
      H    = zero
      kmix = -1
      if (Bflx < zero) then
        do k=2,kb
          if (N2(k) < zero) then
            H    = H + dh(k)
            kmix = k
          else
            exit
          endif
        enddo
      endif
      if (H > zero) then
        H = H + dh(1) - dh(kmix)
        kmix = kmix - 1
      endif
      W3 = max( -Bflx*H, zero )

      !- combined velocity scale (wind and heatflux) cubed:
      q3 = u_taus**3 + ww*W3

      !- convective velocity scale:
      Ws = W3**pow1d3
      !-------------------------------------------------------------------------


      !- set the boundary conditions for tke and diss --------------------------
      call BoundaryCondtions(tke, diss, q3, u_taub, kb, eddyvis(1),            &
                             RHS, ep0, om0, z0srf(ii))


      !- regular 1D column -----------------------------------------------------
      if (kb > kb0) then

        !- Buoyancy production of turbulence:
        call BuoyancyProduction(Pb, tke, diss, Bflx, kb, rT, eddydif, rS,      &
                                Ks, H, Ws, kmix)
  
        !- Get shear production terms
        if (prodtype == 2 .or. prodtype == 4) then
          !- Shear production due to unresolved internal waves:
          sum1 = zero
          sum2 = zero
          do k = 2,kb
            buoyfreq(k) = sqrt(max(N2(k),zero))
            sum1 = sum1 + buoyfreq(k)*dz(k)
            sum2 = sum2 + dz(k)
          enddo
          buoyfreqmean = sum1/sum2

          if (buoyfreqmean > zero) then
            const     = F0/(rho0 * depth * buoyfreqmean)
            Piw(2:kb) = const*buoyfreq(2:kb)
          else
            Piw(2:kb) = zero
          endif
        else
          Piw(2:kb) = zero
        endif

        !- parameterization of effects from tides:
        if (tidetype > 0) then
          if (tt1) then
            ! unresolved bottom shear:
            dtide = depth
            do k=2,kb
              dtide  = dtide - dh(k-1)
              ustr2  = cdtide*utide*sqrt(utide**2                              &
                                         + (half*(u(k)+u(k-1)))**2             &
                                         + (half*(v(k)+v(k-1)))**2)
              Sh2(k) = Sh2(k) + ustr2/(vkarm*dtide)**2
            enddo
          endif
          if (tt2) then
            ! internal baroclinic tides:
            dtide = utide*utide*iwtide/(one-exp(-depth/zetatide))
            sum2  = -depth
            do k=2,kb
              sum2   = sum2 + dz(k)
              Piw(k) = Piw(k) + buoyfreq(k)*dtide*exp(sum2/zetatide)
            enddo
          endif
          ! tidal drag, shallow seas: handled in BoundaryConditions
        endif

        ! production terms:
        call Production(Pstot, Ps, Piw, Plc, no_ice, iceconc, Ua, Va,          &
                        eddyvis, N2, Sh2, dh, kb)


        !- Solve for turbulent kinetic energy
        call KineticEnergy(tke, eddyvis, Pstot, Pb, diss, dz, dh, dt, kb, RHS)


        !- Dissipation based on turbulence model
        !- two-equation turbulence model (k-omega model)
        call Frequency(diss, eddyvis, Pstot, Pb, tke,                          &
                       tke_old, dz, dh, N2, dt, kb, Ps, Piw, om0)

      elseif (kb == 2) then
        !- assign all prod terms with zeroes:
        call ZeroProduction(Pb, Pstot, Ps, Piw, Plc, kb)
      endif


      !- ensure sensible magnitudes of both tke and diss -----------------------
      call LowerLimit(tke, diss, kb)


      !- calculate the eddy viscosity and eddy diffusivity ---------------------
!dir$ inline
      call PartIIDiff(eddyvis, eddydif, Ks, tke, diss, rT, rS, Sh2, kb,        &
                      eddydif_max, eddyvis_max, Piw)


      !- transfer quatities to 3D arrays ---------------------------------------
      tke3D (ii)  = tke (1)
      diss3D(ii)  = diss(1)
      avv(ii)     = eddyvis(1)
      avtse(1,ii) = eddydif(1) + thermd
      if (kb > 1) then
        tke3D (mm2:mm2+kb-2) = tke (2:kb)
        diss3D(mm2:mm2+kb-2) = diss(2:kb)

        avv(mm2:mm2+kb-2)     = eddyvis(2:kb)
        avtse(1,mm2:mm2+kb-2) = eddydif(2:kb) + thermd
      endif
      avtse(2,ii) = Ks(1)
      if (kb > 1) avtse(2,mm2:mm2+kb-2) = Ks(2:kb)
      if (tracers) then
        if (pdiffass) then
          avtse(3,ii)           = avtse(2,ii)
          avtse(3,mm2:mm2+kb-2) = avtse(2,mm2:mm2+kb-2)
        else
          if (abs(rT(1) + rS(1)) < tinyv) then
            avtse(3,ii) = Ks(1)
          else
            avtse(3,ii) = (rT(1)*eddydif(1) + rS(1)*Ks(1))/(rT(1)+rS(1))
          endif
          do k=2,kb
            if (abs(rT(k) + rS(k)) < tinyv) then
              avtse(3,mm2+k-2) = Ks(k)
            else
              avtse(3,mm2+k-2) = (rT(k)*eddydif(k)+rS(k)*Ks(k))/(rT(k)+rS(k))
            endif
          enddo
        endif
      endif
      

    enddo wetpointloop


    !- That's all folks! -------------------------------------------------------

  end subroutine TurbModel


  subroutine InitTurb(coldstart, kx, iw2, mmk, ind, kh, h,                     &
                      rho, u3D, v3D, avv, avtse, tke3D, diss3D, z0srf,         &
                      cmp, tracers, ompidx, dt, ia, tmap)
    !---------------------------------------------------------------------------
    !
    ! Initialise TKE and TKEDISS as well as eddy viscosity and diffusion
    !
    ! Author:  Per Berg, September 2004.
    !
    !---------------------------------------------------------------------------

    !- modules -----------------------------------------------------------------
    use dmi_omp,   only : domp_get_domain
    use constants, only : cfreeze, saltmax, wtmpmax
    use tideparam, only : get_tide_map, cdtide, zetatide

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    logical,    intent(in)                   :: coldstart
    integer(4), intent(in)                   :: kx, iw2, ia
    integer(4), intent(in)                   :: mmk(:,0:,0:)
    integer(4), intent(in)                   :: ind(:,:), kh(0:)
    real(8),    intent(in)                   :: dt
    real(8),    intent(in)                   :: cmp(1:,0:) 
    real(8),    intent(in)                   :: tmap(0:,1:)
    real(8),    intent(in),    dimension(0:) :: h, rho, u3D, v3D
    real(8),    intent(inout), dimension(0:) :: tke3D, diss3D, avv
    real(8),    intent(inout)                :: z0srf(0:)
    real(8),    intent(inout)                :: avtse(1:,0:)
    logical,    intent(in)                   :: tracers
    integer(4), intent(inout)                :: ompidx(1:,1:)

    !- local vars --------------------------------------------------------------
    integer(4)               :: i, j, k, kb, ii, idx, idm
    integer(4)               :: n2dl, n2du
    real(8), dimension(1:kx) :: eddyvis, eddydif, N2, tke, diss, Sh2
    real(8), dimension(1:kx) :: Ks, rT, rS, dh
    real(8), dimension(2:kx) :: dz, Piw, buoy_freq
    real(8)                  :: depth, sum1, sum2, dz_min
    real(8)                  :: pp, ss, tt, tp, tm, sp, sm
    real(8)                  :: alphas, alphat, dtp, dtm, dsp, dsm
    real(8), parameter       :: gdrho0 = -g/rho0
    real(8), parameter       :: delta = 0.01_8, twodel =0.02_8
    real(8)                  :: eddydif_max, eddyvis_max
    real(8)                  :: utide, ustr2, dtide, iwtide 
    logical                  :: tt1, tt2


!$OMP BARRIER
!$OMP MASTER
    !- initialise as minimum values for coldstarted simulation -----------------
    if (coldstart) then
      !  play safe and initialise all points during a coldstart,
      !  (for hot starts, tke and diss and avv are read from restart file)
      tke3D(:)    = tke_min
      diss3D(:)   = diss_min
      avv(:)      = laminarvis
      avtse(1,:)  = eddydif_min + thermd
      avtse(2:,:) = eddydif_min
      z0srf(:)    = z0s
    else
      !  initialise at land during a hotstart
      avtse(1,0)  = eddydif_min + thermd
      avtse(2:,0) = eddydif_min
      z0srf(0)    = z0s
    endif
    if (srfprod) then
      kb0 = 1
    else
      kb0 = 2
    endif
!$OMP END MASTER
!$OMP BARRIER


    !- set OMP vars ------------------------------------------------------------
    call domp_get_domain (kh, 1, iw2, n2dl, n2du, ompidx)


    !- eddy viscosity and diffusion based on initial tke and diss --------------
    wetpointloop: do ii=n2dl,n2du

      kb = kh(ii)
      if (kb < 1) cycle wetpointloop

      !- indices:
      i  = ind(1,ii)
      j  = ind(2,ii)

      !- extract 1D column vars from 3D vars:
      do k=1,kb
        idx = mmk(k,i,j)

        tke(k)  = tke3D(idx)
        diss(k) = diss3D(idx)

        eddyvis(k) = avv(idx)
      enddo

      !- buoyancy frequency squared:
      depth = max( h(mmk(1,i,j)), z0s )
      dh(1) = depth
      N2(1) = zero      
      do k=2,kb
        idx = mmk(k,i,j)
        idm = mmk(k-1,i,j)

        dh(k) = h(idx)
        depth = depth + dh(k)
        dz(k) = half*(dh(k) + dh(k-1))
        N2(k) = gdrho0*(rho(idm)-rho(idx))/dz(k)
      enddo

      !- maxium diffusivities for this water column:
      if (kb >= 2) then
        dz_min = minval(dz(2:kb))
      else
        dz_min = dh(1)
      endif
      eddyvis_max = ten*dz_min*dz_min/dt
      eddydif_max = eddyvis_max
     
      !- Gradient Richardson number:
      Sh2(1) = zero
      do k=2,kb
        idx = mmk(k,i,j)

        Sh2(k) =                                                               &
                (half*( u3D(mmk(k-1,i,j))   + u3D(mmk(k-1,i,j-1))              &
                       -u3D(idx)            - u3D(mmk(k,i,j-1))  )/dz(k))**2   &
              + (half*( v3D(mmk(k-1,i-1,j)) + v3D(mmk(k-1,i,j))                &
                       -v3D(mmk(k,i-1,j  )) - v3D(idx)          )/dz(k))**2
      enddo

      !- T and S gradients:
      pp = tenth*one

      rT(1) = zero
      rS(1) = zero
      do k=2,kb
        idx = mmk(k,i,j)
        idm = mmk(k-1,i,j)

        tm = cmp(1,idm)
        tp = cmp(1,idx)
        sm = cmp(2,idm)
        sp = cmp(2,idx)
        tt = min(wtmpmax, max(tinyv, half*(tm+tp)))
        ss = min(saltmax, max(tinyv, half*(sm+sp)))
        pp = pp + tenth*h(idm)

        ! local thermal expansion coefficient:
        dtp    = min( max(-cfreeze*ss, tt + delta), wtmpmax)
        dtm    = min( max(-cfreeze*ss, tt - delta), wtmpmax)
        alphat = -(densit(ss,dtp,pp)-densit(ss,dtm,pp))/(twodel*rho0)

        ! local haline contraction coefficient:
        dsp    = min( max(zero, ss + delta), saltmax)
        dsm    = min( max(zero, ss - delta), saltmax)
        alphas =  (densit(dsp,tt,pp)-densit(dsm,tt,pp))/(twodel*rho0)

        rT(k) = g*alphat*(tm-tp)/dz(k)
        rS(k) = g*alphas*(sm-sp)/dz(k)
      enddo

      !- calculate the eddy viscosity and eddy diffusivity
      !  Shear production due to unresolved internal waves
      Piw(2:kb) = zero
      if (kb > kb0 .and. (prodtype == 2 .or. prodtype == 4)) then
        sum1 = zero
        sum2 = zero
        do k = 2, kb
          buoy_freq(k) = sqrt(max(N2(k),zero))
          sum1 = sum1 + buoy_freq(k)*dz(k)
          sum2 = sum2 + dz(k)
        enddo 
        sum1 = sum1/sum2
        if (sum1 > zero) then
          sum2      = F0/(rho0 * depth * sum1) 
          Piw(2:kb) = sum2 * buoy_freq(2:kb)
        endif 
      endif

      !- parameterization of effects from tides:
      if (tidetype > 0) then
        call get_tide_map(ia, ii, tidetype, tmap, utide, iwtide, tt1, tt2 )
        if (tt1) then
          ! unresolved bottom shear:
          dtide = depth
          do k=2,kb
            dtide  = dtide - dh(k-1)
            ustr2  = cdtide*utide*sqrt(utide**2                                &
                           +( half*(u3D(idx)+u3D(mmk(k,i,j-1)))                &
                             +half*(u3D(mmk(k-1,i,j))+u3D(mmk(k-1,i,j-1))))**2 &
                           +( half*(v3D(idx)+v3D(mmk(k,i-1,j)))                &
                             +half*(v3D(mmk(k-1,i,j))+v3D(mmk(k-1,i-1,j))))**2)
            Sh2(k) = Sh2(k) + ustr2/(vkarm*dtide)**2
          enddo
        endif
        if (tt2) then
          ! internal baroclinic tides:
          dtide = utide*utide*iwtide/(one-exp(-depth/zetatide))
          sum2  = -depth
          do k=2,kb
            sum2   = sum2 + dz(k)
            Piw(k) = Piw(k) + buoy_freq(k)*dtide*exp(sum2/zetatide)
          enddo
        endif
        ! tidal drag, shallow seas:
        !   not used during initialization (requires dynamics).
      endif

!dir$ inline
      call PartIIDiff(eddyvis, eddydif, Ks, tke, diss, rT, rS, Sh2, kb,        &
                      eddydif_max, eddyvis_max, Piw)

      !- transfer to 3D arrays
      if (coldstart) then
        avv(mmk(1,i,j)) = laminarvis
        avtse(1,mmk(1,i,j)) = eddydif_min + thermd
        do k=2,kb
          idx = mmk(k,i,j)
          avv(idx) = eddyvis(k)
          avtse(1,idx) = eddydif(k) + thermd
        enddo
      else
        ! don't update avv(:) for a hotstart (it's read from restartfile)
        avtse(1,mmk(1,i,j)) = eddydif_min + thermd
        do k=2,kb
          avtse(1,mmk(k,i,j)) = eddydif(k) + thermd
        enddo
      endif
      avtse(2,mmk(1,i,j)) = eddydif_min
      do k=2,kb
        avtse(2,mmk(k,i,j)) = Ks(k)
      enddo
      if (tracers) then
        avtse(3,mmk(1,i,j)) = eddydif_min
        if (pdiffass) then
          do k=1,kb
            avtse(3,mmk(k,i,j)) = Ks(k)
          enddo
        else
          do k=1,kb
            if (abs(rT(k) + rS(k)) < tinyv) then
              avtse(3,mmk(k,i,j)) = Ks(k)
            else
              avtse(3,mmk(k,i,j)) = (rT(k)*eddydif(k)+rS(k)*Ks(k))/(rT(k)+rS(k))
            endif
          enddo
        endif
      endif

    enddo wetpointloop


    !- That's it! --------------------------------------------------------------

  end subroutine InitTurb





  subroutine BoundaryCondtions(tke, diss, q3, u_taub, kb, eddy0,               &
                               RHS, ep0, om0, z0srf )
    !---------------------------------------------------------------------------
    !
    ! Set top and bottom BCs for TKE and TKEDISS:
    !   based on bedfriction at bottom
    !   based on wind and buoyancy flux at top 
    !
    !   i.e. wind+flux "wins" over bedroughness in 1-layer cases
    !
    ! Author:  Per Berg, September 2004.
    !
    !---------------------------------------------------------------------------
    
    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: kb
    real(8),    intent(inout) :: tke(1:), diss(1:)
    real(8),    intent(in)    :: q3, u_taub
    real(8),    intent(inout) :: eddy0
    real(8),    intent(out)   :: RHS, ep0, om0, z0srf
    
    !- local vars --------------------------------------------------------------
    real(8) :: z, k0, zc, sqk0
    real(8), parameter :: B0 = cmu0*Lkome,                                     &
                          B1 = sigma_tke_komega/(-alp_kome*Lkome*cmu0),        &
                          B2 = (1.0_8-0.5_8*alp_kome)/Lkome/sigma_omega,       &
                          B3 = cmu04*(1.0_8-1.5_8*alp_kome)/sigma_diss,        &
                          beta = 0.5_8

  
    !- based on bedfriction at bottom ------------------------------------------
    z = kappa*z0b
    tke( kb) = (u_taub/cmu0)**2
    diss(kb) = ((cmu0*sqrt(tke(kb)))**3) / z
    if (tke(kb)  <  tke_min) tke(kb)  = tke_min
    if (diss(kb) < diss_min) diss(kb) = diss_min


    !- based on wind and buoyancy flux at top ----------------------------------
    !  increased roughness for high wind speeds and large buoyancy fluxes:
    zc = max( z0s, beta*sqrt(rhow)*(q3**pow1d3) )

    !  Neumann for k, Dirichlet for eps and omega:
    RHS   = mwind*q3
    k0    = (B1*RHS)**pow2d3
    sqk0  = sqrt(k0)
    ep0   = cmu04*k0*sqk0/(B0*zc)
    eddy0 = sqk0*B0*zc

    if (kb <= kb0) then
      !  Dirichlet:
      tke(1)  = max( k0,  tke_min)
      diss(1) = max(ep0, diss_min)
    endif

    !  Neumann for eps and omega as well:
    ep0 = B3*k0*k0/zc
    om0 = B2*k0/zc

    !  return the surface roughness length:
    z0srf = zc

    !- That's it! --------------------------------------------------------------

  end subroutine BoundaryCondtions





  subroutine Production(Pstot, Ps, Piw, Plc, no_ice, iceconc, Ua, Va,          &
                        eddyvis, N2, Sh2, dh, kb)
    !---------------------------------------------------------------------------
    !
    ! Calculates the shear production of TKE
    !
    ! Inputs:        Ua, Va:  wind velocity components
    !                iceconc: ice concentration
    !                eddyvis: Eddy viscosity. 
    !                N2:      Buoyancy frequency squared. 
    !                Sh2:     Mean shear squared.
    !                dh:      Thickness of grid cells
    !                kb:      Number of active cells.
    !                Piw:     Shear due to internal waves.
    !
    ! Outputs:       Pstot:   Total shear production.
    !                Ps:      Resolved shear production.
    !                Plc:     Shear due to Langmuir circulations. 
    !
    ! Reference for internal waves and Langmuir circulations:
    !   Axell, L.B., J. Geopys. Res., 107, C11, 3204,
    !   doi: 10.1029/2001JC000922, 2002.
    !
    ! Author:  Per Berg, September 2004.
    !
    !---------------------------------------------------------------------------
    
    !- modules -----------------------------------------------------------------
    use constants, only : pi

    !- directives --------------------------------------------------------------
    implicit none
  
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)  :: kb
    real(8),    intent(in)  :: iceconc, Ua, Va
    real(8),    intent(out) :: Ps(2:), Plc(2:), Pstot(2:)
    real(8),    intent(in)  :: eddyvis(1:), N2(1:), Sh2(1:), dh(1:)
    real(8),    intent(in)  :: Piw(2:)
    logical,    intent(in)  :: no_ice
    
    !- local vars --------------------------------------------------------------
    real(8)    :: Wlc3(2:kb)
    real(8)    :: z(1:kb+1)
    real(8)    :: Wa, Ws, Ep, Hlc, const
    integer(4) :: k, klc
    real(8), parameter :: wsc = 0.016_8
  
  
    !- Resolved shear production of turbulence ---------------------------------
    Ps(2:kb) = eddyvis(2:kb)*Sh2(2:kb)
  
    
    !- Shear production due to Langmuir circulations ---------------------------
    if (prodtype == 3 .or. prodtype == 4) then

      ! Calculate z-coordinates of cell boundaries
      z(1) = zero
      do k = 2, kb + 1
        z(k) = z(k-1) + dh(k-1)
      enddo
    
      ! Calculate Stokes drift Ws
      Wa = sqrt(Ua*Ua + Va*Va)    ! wind speed
      Ws = wsc*Wa                 ! Stokes drift at the surface
    
      ! Calculate maximum depth of Langmuir circulations, Hlc
      klc   = 1
      k     = 2
      Ep    = max(N2(k),zero)*z(k)*dh(k-1)
      const = Ws*Ws
      do while (Ep <= const .and. k < kb)
        klc = k
        k   = k + 1
        Ep  = Ep + (max(N2(k-1),zero)*z(k-1)+max(N2(k),zero)*z(k))*dh(k-1)
      enddo
      Hlc = z(klc)
      
      ! Calculate vertical velocity Wlc to third power
      ! Include effects of ice. 
      ! Note: If no_ice =.TRUE., iceconc is not necessarily zero
      if (no_ice) then
        const = one/Hlc
      else
        const = (one - iceconc)/Hlc
      endif
      const = const*(Clc*Ws)**3
      do k = 2, klc
        Wlc3(k) = const*(sin(pi*z(k)/Hlc))**3
      enddo
      if (klc < kb) Wlc3(klc+1:kb) = zero
    
      ! Calculate production term Plc
      Plc(2:kb) = Wlc3(2:kb)
    else
      Plc(2:kb) = zero
    endif
    
  
    !- Total shear production is the sum of Ps, Piw and Plc --------------------
    Pstot(2:kb) = Ps(2:kb) + Piw(2:kb) + Plc(2:kb)
  
    
    !- That's it ! -------------------------------------------------------------
  
  end subroutine Production
  
  
  
  
  
  subroutine BuoyancyProduction (Pb,tke,diss,Bflx,kb,rT,Kh,rS,Ks,H,Ws,kmix)

    !---------------------------------------------------------------------------
    !
    ! Calculates the buoyancy production of turbulence.
    ! Extended buoyancy production during convection modelled by TOM inclusion.
    ! Parameterization of TOM during convection generated by surface heat flux.
    !
    ! FIXME: include at a later stage a surface salinity flux -S0*(E-P)
    !
    !
    ! Inputs:        tke:     Turbulent kinetic energy.
    !                diss:    Dissipation.
    !                Bflx:    surface buoyancy flux.
    !                rT:      Temperature gradient, g*alphaT*dT/dz
    !                rS:      Salinity gradient, g*alphaS*dS/dz
    !                Kh:      Heat (and salt) diffusivity.
    !                Ks:      Salt diffusivity.
    !                dh:      Thickness of grid cells
    !                kb:      Number of active cells.
    !
    ! Outputs:       Pb:      Buoyancy production.
    !
    ! Author:  Per Berg, May 3, 2007.
    !
    !---------------------------------------------------------------------------

    !- modules -----------------------------------------------------------------
    use constants, only : tom

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)  :: kb, kmix
    real(8),    intent(in)  :: Bflx, H, Ws
    real(8),    intent(in)  :: tke(1:), diss(1:), rT(1:), Kh(1:)
    real(8),    intent(in)  :: rS(1:), Ks(1:)
    real(8),    intent(out) :: Pb(2:)
    
    !- locals vars -------------------------------------------------------------
    real(8)    :: gr, c1, c2, a1, a2, tau
    integer(4) :: k

    !- local parameters --------------------------------------------------------
    real(8), parameter :: gam = 1.2_8 ! entrainment parameter
    real(8), parameter :: tvd = 7.8_8 ! temperature variance dissipation const
    real(8), parameter :: tmf = 0.4_8 ! third moment factor
    real(8), parameter :: tsc = 1.56_8! time scale constant


    !- branch on choice of stability functions:
    !- standard non-convective terms:
    Pb(2:kb) = Ks(2:kb)*rS(2:kb) - Kh(2:kb)*rT(2:kb)

    !- include Third-Order-Moments if selected:
    if (tom) then
      !- calculate convective buoyancy production:
      if (H > zero .and. Ws > zero) then
      
        !  heat flux gradient:
        gr = (-Bflx/H)

        !  derived coefficients:
        c1 = half*tvd*(-Bflx)/(Ws*tsc)
        c2 = half*gam*gr
        a1 = tmf*Ws*c2
        a2 = c1*c2

        !  add TOM terms throughout mixed layer:
        do k=2,kmix
          tau   = tke(k-1)/diss(k-1) + tke(k)/diss(k)
          Pb(k) = Pb(k) + tau*(a1 + a2*tau)
        enddo
      endif
    endif
  
  end subroutine BuoyancyProduction
  
  
  
  
  
  subroutine KineticEnergy(tke,eddyvis,Pstot,Pb,diss,dz,dh,dt,kb,RHS)
    !---------------------------------------------------------------------------
    !
    ! Solves the differential equation for the TKE
    !
    ! Inputs:        eddyvis:   Eddy viscosity. 
    !                tke:       Turbulent kinetic energy.
    !                Pstot:     Total shear production.
    !                Pb:        Buoyancy production. 
    !                diss:      Dissipation. 
    !                dz:        Distance between centres of grid cells.
    !                dh:        Thickness of grid cells. 
    !                dt:        Time step. 
    !                kb:        Number of active cells. 
    !
    ! Outputs:       tke:       Turbulent kinetic energy.
    !                diss:      Dissipation, bottom and top BC
    !
    ! Author:  Per Berg, September 2004.
    !
    !---------------------------------------------------------------------------
!    use trisolver, only : trisolv
    
    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: kb
    real(8),    intent(inout) :: tke(1:), diss(1:)
    real(8),    intent(in)    :: Pstot(2:), Pb(2:), dz(2:)
    real(8),    intent(in)    :: eddyvis(1:), dh(1:)
    real(8),    intent(in)    :: dt
    real(8),    intent(in)    :: RHS
    
    !- local vars --------------------------------------------------------------
    real(8)    :: pplus(2:kb-1), pminus(2:kb-1)
    real(8)    :: a(1:kb), b(1:kb), c(1:kb), d(1:kb)
    real(8)    :: fac, Psc, Pbc, pp, pm
    integer(4) :: k
    real(8), parameter :: cm = one
    
  
    !- Quasi-implicit treatment of rhs of tke-equation -------------------------
    !  centred at scalar points
    fac = half
    do k = 2, kb-1
      Psc = fac*(Pstot(k) + Pstot(k+1))
      Pbc = fac*(Pb(k)    + Pb(k+1))
      if (Psc + Pbc > zero) then
        pplus(k)  = Psc + Pbc
        pminus(k) = cm*diss(k)
      else
        pplus(k)  = Psc
        pminus(k) = cm*diss(k) - Pbc
      endif
    enddo
  
  
    !- Set up tri-diagonal system ----------------------------------------------
    fac = dt/sigma_tke
    do k = 2, kb-1
      a(k) = -fac*eddyvis(k  )/(dz(k  )*dh(k))
    enddo
    do k = 2, kb-1
      c(k) = -fac*eddyvis(k+1)/(dz(k+1)*dh(k))
    enddo
    do k = 2, kb-1
      b(k) = one - a(k) - c(k) + dt*pminus(k)/tke(k)
      d(k) = tke(k) + dt*pplus(k)
    enddo
    
  
    !- Surface BC for tke ------------------------------------------------------
    !  Dirichlet:
    !  c(1) = zero
    !  b(1) = one 
    !  d(1) = tke(1)       ! already updated in BoundaryCondtions
    !  Neumann:
    c(1) = -fac*eddyvis(2)/(dz(2)*dh(1))
    if (srfprod) then
      Psc = Pstot(2)
      Pbc = Pb(2)
      if (Psc + Pbc > zero) then
        pp = Psc + Pbc
        pm = cm*diss(1)
      else
        pp = Psc
        pm = cm*diss(1) - Pbc
      endif
      b(1) = one - c(1) + dt*pm/tke(1)
      d(1) = tke(1) + dt*RHS/dh(1) + dt*pp
    else
      b(1) = one - c(1)
      d(1) = tke(1) + dt*RHS/dh(1) 
    endif
  
    !- Bottom BC for tke -------------------------------------------------------
    a(kb) = zero
    b(kb) = one
    d(kb) = tke(kb)     ! already updated in BoundaryCondtions
  
  
    !- Solve tri-diagonal system for tke ---------------------------------------
!dir$ inline
    call TriSolv (tke(1:kb), a(1:kb), b(1:kb), c(1:kb), d(1:kb), kb)
    
  
    ! FIXME:
    !- make sure that tke is not less than tke_min -----------------------------
    !  not required for the k-eps model, but
    !  might be required for k-model or for k-omega ??? dono yet ???
    !do k = 1, kb
    !  tke(k) = max(tke(k), tke_min)
    !enddo
  
  
    !- That's it ! -------------------------------------------------------------
  
  end subroutine KineticEnergy
  
  
  
  
  
  subroutine Frequency(diss, eddyvis, Pstot, Pb, tke, tke_old,                 &
                       dz, dh, N2, dt, kb, Ps, Piw, om0)
    !---------------------------------------------------------------------------
    !
    ! Solves the differential equation for the frequency of TKE
    !
    ! Inputs:        eddyvis:   Eddy viscosity.
    !                Pstot:     Total shear production.
    !                Ps:        Resolved shear production.
    !                Piw:       Production from internal waves.
    !                Pb:        Buoyancy production.
    !                tke:       Turbulent kinetic energy.
    !                tke_old:   tke from former time step.
    !                dz:        Distance between centres of grid cells. 
    !                dh:        Thickness of grid cells. 
    !                N2:        Buoyancy frequency. 
    !                dt:        Time step.
    !                kb:        Number of active cells.
    !
    ! Outputs:       diss:      Dissipation.
    !
    ! Author:  Per Berg, July 2006.
    !
    !---------------------------------------------------------------------------
!    use trisolver, only : trisolv
    
    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: kb
    real(8),    intent(inout) :: diss(1:)
    real(8),    intent(in)    :: tke(1:), tke_old(1:), eddyvis(1:), dh(1:)
    real(8),    intent(in)    :: N2(1:)
    real(8),    intent(in)    :: Pstot(2:), Pb(2:), dz(2:), Ps(2:), Piw(2:)
    real(8),    intent(in)    :: dt, om0
    
    !- local vars --------------------------------------------------------------
    real(8)    :: pplus(2:kb), pminus(2:kb), comega3(2:kb)
    real(8)    :: omega(1:kb), a(1:kb), b(1:kb), c(1:kb), d(1:kb)
    real(8)    :: fac, Psc, Pbc, pp, pm
    integer(4) :: k
  
  
    !- Obtain frequency from tke and diss --------------------------------------
    omega(1:kb) = diss(1:kb)/(tke_old(1:kb)*cmu04)
  
  
    !- Set "correct" value of comega3 ------------------------------------------
    ! Many possibilities:
    !   Just as in the case of ceps3 in the k-epsilon model, comega3 
    !   depends only on comega1 and comega2 in stable stratification, 
    !   if the steady-state flux Richardson number is known. The value 
    !   for comega3 during unstable situations is less known. 
    !
    !   HIROMB model suggests:
    !     comega3 = -1.25 for non-patchy turbulence
    !     comega3 = -3.33 for patchy turbulence
    !
    !   and Umlauf et al (2003) suggests 
    !     comega3 = ceps3 - 1 
    !
    !   during stable stratification.
    !
    !   During unstable situations, we apply 
    !     comega3 = ceps3 - 1    where ceps3 is usually set to 1.
    !---------------------------------------------------------------------------
    do k = 2, kb
      if (N2(k) < zero) then          ! unstable stratification
        comega3(k) = comega3unstable
      elseif (Ps(k) >= Piw(k)) then   ! stable stratification, 
                                      ! non-patchy turbulence
        comega3(k) = comega3nonpatchy ! corresponds to Rf = 0.16
      else                            ! stable stratification, 
                                      ! patchy turbulence
        comega3(k) = comega3patchy    ! corresponds to Rf = 0.16/2 = 0.08
      endif
    enddo
    
  
    !- quasi-implicit treatment of rhs of omega-equation -----------------------
    !  centred at scalar points
    do k = 2,kb-1
      Psc = half*comega1   *(Pstot(k) + Pstot(k+1))
      Pbc = half*comega3(k)*(Pb(k)    + Pb(k+1)   )
      fac = omega(k)/tke_old(k)
      if (Psc + Pbc > zero) then
        pplus(k)  = (Psc + Pbc)*fac
        pminus(k) = comega2*diss(k)/tke_old(k)
      else
        pplus(k)  = Psc*fac
        pminus(k) = (comega2*diss(k) - Pbc)/tke_old(k)
      endif
    enddo
  
  
    !- Set up tri-diagonal system ----------------------------------------------
    fac = dt/sigma_omega
    do k = 2, kb-1
      a(k) = -fac*eddyvis(k)  /(dz(k  )*dh(k))
    enddo
    do k = 2, kb-1
      c(k) = -fac*eddyvis(k+1)/(dz(k+1)*dh(k))
    enddo
    do k = 2, kb-1
      b(k) = one - a(k) - c(k) + dt*pminus(k)
      d(k) = omega(k) + dt*pplus(k)
    enddo
  
    
    !- Surface BC for omega ----------------------------------------------------
    !  Dirichlet:
    !  c(1) = zero
    !  b(1) = one 
    !  d(1) = diss(1)/(tke(1)*cmu04)      ! from BoundaryConditions
    !  Dirichlet for omega based on Neuman for k:
    !  a(1) = -fac*eddyvis(1)*two/dh(1)**2
    !  c(1) = -fac*eddyvis(2)/(dz(2)*dh(1))
    !  b(1) = one - a(1) - c(1)
    !  d(1) = omega(1) - a(1)*om0
    !  Neumann:
    c(1) = -fac*eddyvis(2)/(dz(2)*dh(1))
    if (srfprod) then
      Psc = comega1   *Pstot(2)
      Pbc = comega3(k)*Pb(2)
      if (Psc + Pbc > zero) then
        pp = (Psc + Pbc)*omega(1)/tke_old(1)
        pm = comega2*diss(1)/tke_old(1)
      else
        pp = Psc*omega(1)/tke_old(1)
        pm = (comega2*diss(1) - Pbc)/tke_old(1)
      endif
      b(1) = one - c(1) + dt*pm
      d(1) = omega(1) + dt*om0/dh(1) + dt*pp
    else
      b(1) = one - c(1)
      d(1) = omega(1) + dt*om0/dh(1)
    endif
  
    !- Bottom BC for omega -----------------------------------------------------
    a(kb) = zero
    b(kb) = one
    d(kb) = diss(kb)/(tke(kb)*cmu04)   ! from BoundaryConditions
  
  
    !- Solve the tri-diagonal system for omega ---------------------------------
!dir$ inline
    call TriSolv(omega(1:kb), a(1:kb), b(1:kb), c(1:kb), d(1:kb), kb)
    
  
    !- Calc dissipation rate ---------------------------------------------------
    diss(1:kb) = cmu04*tke(1:kb)*omega(1:kb)
  
  
    !- That's it! --------------------------------------------------------------
  
  end subroutine Frequency
  
  
  
  
  
  subroutine LowerLimit(tke, diss, kb)
    !---------------------------------------------------------------------------
    !
    ! Sets lower limits on TKE and dissipation.
    !
    ! Inputs:           kb:     Number of active cells. 
    !
    ! Outputs:          tke:    Turbulent kinetic energy.
    !                   diss:   Dissipation.
    !
    ! Author:  Per Berg, September 2004.
    !
    !---------------------------------------------------------------------------
    
    !- directives --------------------------------------------------------------
    implicit none
    
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: kb
    real(8),    intent(inout) :: tke(1:), diss(1:)
    
    !- local vars --------------------------------------------------------------
    integer(4) :: k
    
  
    !- Make sure that tke and diss have reasonable lower limits ----------------
    do k = 1, kb
      if (tke(k)  <  tke_min) tke(k)  =  tke_min
      if (diss(k) < diss_min) diss(k) = diss_min
    enddo
    
   
    !- That's it! --------------------------------------------------------------
  
  end subroutine LowerLimit
  
  
  
  
  
  subroutine PartIIDiff( Km, Kh, Ks, tke, diss, rT, rS, Sh2, kb,               &
                         eddydif_max, eddyvis_max, Piw)
    !---------------------------------------------------------------------------
    !
    ! Calculates the stability functions, the turbulent Prandtl-Schmidt
    ! number and the resulting eddy viscosity and eddy diffusivities.
    !
    !
    ! Inputs:        tke:       Turbulent kinetic energy.
    !                diss:      Dissipation.
    !                rT:        Temperature gradient, g*alphaT*dT/dz
    !                rS:        Salinity gradient, g*alphaS*dS/dz
    !                Sh2:       Mean shear.
    !                kb:        Number of active cells.
    !                Piw:       Shear production due to internal waves.
    !
    ! Outputs:       Km:        Momentum diffusivity.
    !                Kh:        Heat diffusivity.
    !                Ks:        Salt diffusivity.
    !
    ! Author:  Per Berg, November 2004.
    !
    !---------------------------------------------------------------------------

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: kb
    real(8),    intent(inout) :: Km(1:kb), Kh(1:kb), Ks(1:kb)
    real(8),    intent(in)    :: tke(1:kb), diss(1:kb)
    real(8),    intent(in)    :: rT(1:kb), rS(1:kb), Sh2(1:kb)
    real(8),    intent(in)    :: Piw(2:kb)
    real(8),    intent(in)    :: eddydif_max, eddyvis_max

    !- locals vars -------------------------------------------------------------
    real(8)    :: Rg_t, Km0
    integer(4) :: k
    logical    :: CheckRg(1:kb)
    real(8)    :: PiwKm(1:kb), Sh2_t(1:kb), t2(1:kb), ss(1:kb)
    real(8), dimension(1:kb) :: y, n, c, D, Nm, Nh, Ns, fc
    real(8), parameter :: y0 = 0.16_8, D0 = 24.0_8
    real(8), parameter :: nn1 = 8.0_8, nn2 = 12.0_8, nn3 = 75.0_8
    real(8), parameter :: ff1 = 4.0_8, ff2 = 60.0_8, ff3 = 15.0_8

    !- skip one-layers ---------------------------------------------------------
    if (kb < 2) return
    Km0 = Km(1)

    ! local arrays assigned dummy values at k=1 for better vectorization

    !- some initial settings ---------------------------------------------------
    !  Prod. due to internal waves div. by mom.diff., PiwKm
    PiwKm(1)    = zero
    PiwKm(2:kb) = Piw(2:kb)/Km(2:kb)
    !  total shear, Sh2_t
    Sh2_t(1)    = zero
    Sh2_t(2:kb) = Sh2(2:kb) + PiwKm(2:kb)
    !  dynamical turbulent time scale squared, t2
    !  eddy diffusivity normalised w/ structure function=1, ss
    t2(1) = zero
    ss(1) = zero
    do k=2,kb
      t2(k) = (tke(k-1)/diss(k-1) + tke(k)/diss(k))**2
      ss(k) = tke(k-1)**2/diss(k-1) + tke(k)**2/diss(k)
    enddo
    !  check if critical gradient Richardson number is exceeded, CheckRg
    CheckRg(1) = .true.
    do k=2,kb
      Rg_t = min( (rT(k) - rS(k))/max(Sh2_t(k),tinyv), Rg_max )
      call CheckRgCrit(rT(k), rS(k), Rg_t, CheckRg(k))
    enddo

    !- find some coefficients --------------------------------------------------
    where (CheckRg)
      D = zero
    elsewhere 
      n = -pi2*pi3*t2*rT
      c =  pi3*pi3*t2*rS
      y =       y0*t2*Sh2_t

      D = D0                                                                   &
        + ((d1*n + d2*c +  d8)*n + (d3*c +  d9)*c + d13)*y                     &
        + ((d4*n + d5*c + d10)*n + (d6*c + d11)*c + d14)*n                     &
        + ((d7*c + d12)*c + d15)*c
    endwhere

    where (D >= zero .and. D < tinyv)
      D = tinyv
    elsewhere (D < zero .and. D > -tinyv)
      D = -tinyv
    endwhere

    !- find vertical diffusivities ---------------------------------------------
    where (CheckRg)
      !- no mixing if critical Rg is exceeded: 
      Km = laminarvis
      Kh = eddydif_min
      Ks = eddydif_min

    elsewhere
      !- mixing if critical Rg is not exceeded: 

      !- momentum:
      Nm = nn1*(nn2 + (a1*n + a2*c + a4)*n + (a3*c + a5)*c)/nn3

      !- heat:
      fc = ff1*(ff2 + b3*y + b4*c +b5*n)/ff3
      Nh = pi4*(one + b1*c + b2*n)*fc

      !- salinity:
      Ns = pi1*(one + b6*c + b7*n)*fc
         
      !- vertical diffusivities:
      Km = min( max(ss*Nm/D, laminarvis),  eddyvis_max )
      Kh = min( max(ss*Nh/D, eddydif_min), eddydif_max )
      Ks = min( max(ss*Ns/D, eddydif_min), eddydif_max )
    endwhere

    Km(1) = Km0 
  end subroutine PartIIDiff





  subroutine CheckRgCrit( rT, rS, Rg, CRg )
    !---------------------------------------------------------------------------
    !
    ! check if critical Rg is exceeded.
    !
    !---------------------------------------------------------------------------

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    real(8), intent(in)  :: rT, rS, Rg
    logical, intent(out) :: CRg

    !- local vars --------------------------------------------------------------
    real(8)              :: f, f1, f2, rT2, rTS, rS2
    real(8), parameter   :: rsmall = 0.00000001_8   ! rather small value
    real(8), parameter   :: A0 = -6.25_8*pi2*pi3, B0 = 6.25_8*pi3*pi3
    real(8), parameter   :: AB = A0*B0


    !- check for obvious Rg value ----------------------------------------------
    if (Rg < one) then
      CRg = .false.
    
    else

      !- treat small rT:
      if (abs(rT) < rsmall) then
        f1 =  two*a3*B0 + five*pi1*b3*b6 - six*d3*B0
        f2 = (six*d7*B0 - five*pi1*b4*b6)*B0

      !- treat larger rT:
      elseif (abs(rS) < rsmall) then
        f1 =  two*a1*A0 - five*pi4*b3*b2 - six*d1*A0
        f2 = (six*d4*A0 + five*pi4*b2*b5)*A0
 
      !- super-critical density ratio:
      elseif (rS/rT > Rrho_crit) then
        f1 = ten      ! dummy values to make f larger than Rg
        f2 = one 

      !- general case:
      else
        rT2 = rT**2
        rTS = rT*rS
        rS2 = rS**2
        f1 = (two*a1*A0-five*pi4*b3*b2-six*d1*A0)*A0*rT2                       &
           + (two*a2*AB-five*pi4*b3*b1*B0+five*pi1*b3*b7*A0-six*d2*AB)*rTS     &
           + (two*a3*B0+five*pi1*b3*b6-six*d3*B0)*B0*rS2
        f2 = (six*d4*A0+five*pi4*b2*b5)*A0*A0*rT2*rT                           &
           + (six*d5*AB+five*pi4*(b1*b5+b2*b4)*B0-five*pi1*b5*b7*A0)*A0*rTS*rT &
           + (six*d6*AB-five*pi1*(b6*b5+b7*b4)*A0+five*pi4*b1*b4*B0)*B0*rTS*rS &
           + (six*d7*B0-five*pi1*b4*b6)*B0*B0*rS2*rS

      endif
 
      !- compare Rg to f:
      if (f2 >= zero .and. f2 < tinyv) then
        f2 = tinyv
      elseif (f2 < zero .and. f2 > -tinyv) then
        f2 = -tinyv
      endif
      f = ccrit*f1/f2
      if (f > one) then
        CRg = (Rg >= f)
      else
        CRg = .false.
      endif

    endif

  end subroutine CheckRgCrit

  subroutine ZeroProduction( Pb, Pstot, Ps, Piw, Plc, kb )
    !---------------------------------------------------------------------------
    !
    ! set production terms equal to zero
    !
    !---------------------------------------------------------------------------

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)  :: kb
    real(8),    intent(out) :: Pb(2:), Pstot(2:), Ps(2:), Piw(2:), Plc(2:)

    !- assign with zeroes ------------------------------------------------------
    Pb(2:kb)    = zero
    Pstot(2:kb) = zero
    Ps(2:kb)    = zero
    Piw(2:kb)   = zero
    Plc(2:kb)   = zero

  end subroutine ZeroProduction
  
  real(8) function densit (s,t,p)
  !-----------------------------------------------------------------------------
  !
  ! Function that calculates the density of sea water using
  ! the UNESCO equation of state.
  !
  ! Comments
  ! Re-written from the original F77 function in cmodlib 
  ! to F90 such that it is ready for in-lining if we choose to do so
  !
  ! Then, optimised with respect to No. of floating point multiplication
  !   (from 46 to 39)
  !
  !-----------------------------------------------------------------------------
  
    implicit none
  
    ! salinity, temperature, pressure (in bar):
    real(8), intent(in) :: s, t, p 
  
  
    real(8) :: sq, rhonul, sbm
    !**********      coefficients   ********************************************
    real(8), parameter :: a0 = 999.842594_8  ,                                 &
                          a1 =   6.793952e-2_8 ,                               &
                          a2 =  -9.095290e-3_8 ,                               &
                          a3 =   1.001685e-4_8 ,                               &
                          a4 =  -1.120083e-6_8 ,                               &
                          a5 =   6.536332e-9_8 ,                               &
                          b0 =  +8.24493e-1_8 ,                                &
                          b1 =  -4.0899e-3_8 ,                                 &
                          b2 =  +7.6438e-5_8 ,                                 &
                          b3 =  -8.2467e-7_8 ,                                 &
                          b4 =  +5.3875e-9_8 ,                                 &
                          c0 =  -5.72466e-3_8 ,                                &
                          c1 =  +1.0227e-4_8 ,                                 &
                          c2 =  -1.6546e-6_8 ,                                 &
                          d0 =  +4.8314e-4_8 ,                                 &
                          e0 = 19652.21_8      ,                               &
                          e1 =   148.4206_8    ,                               &
                          e2 =    -2.327105_8  ,                               &
                          e3 =     1.360477e-2_8 ,                             &
                          e4 =    -5.155288e-5_8                       
  
    real(8), parameter :: f0 =    54.6746_8    ,                               &
                          f1 =    -0.603459_8  ,                               &
                          f2 =     1.09987e-2_8 ,                              &
                          f3 =    -6.1670e-5_8 ,                               &
                          g0 =     7.944e-2_8 ,                                &
                          g1 =     1.6483e-2_8 ,                               &
                          g2 =    -5.3009e-4_8 ,                               &
                          h0 =  3.239908_8     ,                               &
                          h1 =  1.43713e-3_8 ,                                 &
                          h2 =  1.16092e-4_8 ,                                 &
                          h3 = -5.77905e-7_8 ,                                 &
                          i0 =  2.2838e-3_8 ,                                  &
                          i1 = -1.0981e-5_8 ,                                  &
                          i2 = -1.6078e-6_8 ,                                  &
                          j0 =  1.91075e-4_8 ,                                 &
                          k0 =  8.50935e-5_8 ,                                 &
                          k1 = -6.12293e-6_8 ,                                 &
                          k2 =  5.2787e-8_8 ,                                  &
                          l0 = -9.9348e-7_8 ,                                  &
                          l1 =  2.0816e-8_8 ,                                  &
                          l2 =  9.1697e-10_8
    !***************************************************************************
    sq = sqrt(s)
  
    rhonul =  a0+s*(b0+c0*sq+d0*s) +                                           &
             (a1+s*(b1+c1*sq)      +                                           &
             (a2+s*(b2+c2*sq)      +                                           &
             (a3+s*b3              +                                           &
             (a4+s*b4              +                                           &
              a5                    *t)*t)*t)*t)*t
  
    sbm    =  e0+s*(f0+g0*sq)+p*(h0+s*(i0+j0*sq)+p*(k0+l0*s)) +                &
             (e1+s*(f1+g1*sq)+p*(h1+s* i1       +p*(k1+l1*s)) +                &
             (e2+s*(f2+g2*sq)+p*(h2+s* i2       +p*(k2+l2*s)) +                &
             (e3+s* f3       +p* h3                           +                &
              e4                                               *t)*t)*t)*t
             
    densit = rhonul/(one-p/sbm)
  
  end function densit
  
end module turbmodels
