module turbout_wrap 

  implicit none
  private
  public :: turbout

  real(8), parameter, private :: zero = 0.0_8, half = 0.5_8

contains

  subroutine turbout(ctim)
    !---------------------------------------------------------------------------
    !  extract and print profiles of turbulence variables
    !---------------------------------------------------------------------------

    !- modules -----------------------------------------------------------------
    use dmi_mpi_global, only : iu06
    use exits,          only : exitme
    use io_subs,        only : io_new_unit
    use cmod_params,    only : narea, mmx, nmx, mm1, hz
    use constants,      only : gmaequ, grarad, printturb
    use cmod_arrays,    only : dxk, dyk, xkoor, ykoor, kh, tke, diss, avv, dispv

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    character(19), intent(in)   :: ctim

    !- local vars --------------------------------------------------------------
    integer(4)                 :: jy1, jx1, ios, ia, i, k, iw
    integer(4), parameter      :: nsm = 200
    integer(4), save           :: lun, ns
    integer(4), save, allocatable :: im(:), jn(:), nst(:), ast(:)
    logical, save              :: firstentry = .true.
    real(8), parameter         :: grad2min = 60.0_8
    real(8), parameter         :: m2g2m = grad2min/gmaequ
    real(8), parameter         :: r2g2m = grad2min/grarad
    real(8)                    :: dp, hdz
    real(8), allocatable       :: xs(:), ys(:), x0(:), x9(:), y0(:), y9(:)
    real(8), allocatable       :: dxa(:), dya(:)
    character(len=100)         :: string


    !- on first entry, open the file, allocate, etc ----------------------------
    if (firstentry .and. printturb) then
      ! I shall say this only once:
      firstentry = .false.

      ! set up stations:
      lun = io_new_unit() 
      open(lun,file='turbstations.txt',status='old',iostat=ios)
      if (ios /= 0) call exitme (1, 'Could not open turbstations.txt' )
      allocate(xs(nsm),ys(nsm),nst(nsm),                                       &
               x0(narea),x9(narea),y0(narea),y9(narea),dxa(narea),dya(narea))
      read(lun,*) string
      do while (string(1:1) == '#')
        read(lun,*) string
      enddo
      backspace(lun)

      do i=1,nsm
        read(lun,*,iostat=ios) nst(i),jy1,jx1
        if (ios /= 0) exit
        ! position in minutes:
        xs(i) = grad2min*(jx1/100) + mod(jx1,100)
        ys(i) = grad2min*(jy1/100) + mod(jy1,100)
      enddo
      close(lun)
      ns = i-1
      if (ns <= 0) then
        write(iu06,*) 'warning: no stations were read from turbstations.txt, '
        write(iu06,*) '         ... continue without. '
        printturb = .false.
      else
        write(iu06,'(a25,i3,a32)') 'turbulence parameters at ', ns,            &
                                   ' stations, output to turbout.txt'
      endif

      if (printturb) then
        ! cmod grid in minutes:
        do ia=1,narea
          dxa(ia) = m2g2m*dxk(ia)
          dya(ia) = m2g2m*dyk(ia)
          x0(ia)  = r2g2m*xkoor(ia)%p(1)
          y0(ia)  = r2g2m*ykoor(ia)%p(1)
          x9(ia)  = x0(ia) + (nmx(ia)-1)*dxa(ia)
          y9(ia)  = y0(ia) - (mmx(ia)-1)*dya(ia)
        enddo

        ! find indices:
        allocate( im(ns), jn(ns), ast(ns) )
        im (:) = 0
        jn (:) = 0
        ast(:) = 0
        do i=1,ns
          arealoop: do ia=narea,1,-1
            if (xs(i) <= x9(ia)-dxa(ia)*half  .and.                            &
                xs(i) >= x0(ia)+dxa(ia)*half  .and.                            &
                ys(i) <= y0(ia)-dya(ia)*half  .and.                            &
                ys(i) >= y9(ia)+dya(ia)*half       ) then
              ast(i) = ia
              im(i)  = int((y0(ia)-ys(i))/dya(ia)) + 1
              jn(i)  = int((xs(i)-x0(ia))/dxa(ia)) + 1
              exit arealoop    ! found, next station
            endif
          enddo arealoop
        enddo
  
        ! open output file:
        lun = io_new_unit() 
        open(lun,file='turbout.txt',status='replace',iostat=ios)
        if (ios /= 0) call exitme (1, 'Could not open turbout.txt' )
      endif

      ! clean up:
      deallocate( x0, x9, y0, y9, dxa, dya, xs, ys )
    endif ! on first entry


    !- on every entry, extract data at each station ----------------------------
    if (printturb) then
      do i=1,ns
        ia = ast(i)
        if (ia == 0) cycle
        dp = zero
        do k=1,kh(ia)%p(mm1(ia)%p(im(i),jn(i),1))
          iw  = mm1(ia)%p(im(i),jn(i),k)
          hdz = half*hz(ia)%p(iw)
          dp  = dp + hdz
          write(lun,'(a4,5a2,1x,i5.5,1x,f6.1,5(1x,e9.2))')                     &
            ctim(1:4), ctim(6:7), ctim(9:10),                                  &
            ctim(12:13), ctim(15:16), ctim(18:19),                             &
            nst(i), dp,                                                        &
            tke(ia)%p(iw), diss(ia)%p(iw),                                     &
            avv(ia)%p(iw), dispv(ia)%p(1,iw), dispv(ia)%p(2,iw)
          dp = dp + hdz
        enddo
      enddo
    endif

    
  end subroutine turbout
end module turbout_wrap 
