module wlstat_wrap 

  implicit none
  private

  public :: wlstat, forward

  real(8), parameter, private :: zero = 0.0_8

contains

subroutine forward(iu,jtime,do_backspace)
  use dmi_mpi_global, only : iu06
  use exits,          only : exitme

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in)           :: iu
  real(8),    intent(in)           :: jtime
  logical,    intent(in), optional :: do_backspace

  !- local vars ----------------------------------------------------------------
  integer(4)         :: ios
  real(8)            :: tfil
  real(8), parameter :: eps = 0.000001_8
  logical            :: readmore
  character(256)     :: fnam

  !-----------------------------------------------------------------------------
  !
  !       forward input file until the requested julian time
  !       written in cols 21-35 in precision better than 1 sec.
  !
  !-----------------------------------------------------------------------------

  inquire(unit=iu,name=fnam)

  tfil     = zero
  readmore = .true.
  do while(readmore)
    read(iu,fmt='(20x,f15.8)',iostat=ios) tfil
    readmore = (tfil < jtime - eps) .and. (ios == 0)
  enddo

  if (ios > 0) then
    call exitme( 'Error: Reading file '//trim(fnam)//' failed.' )
  elseif (ios < 0) then
    write(iu06,*) 'Warning: Reading file '//trim(fnam)//' past eof'
  endif

  !- Optionally, turn off the back-spacing:
  if (present(do_backspace)) then
    if (do_backspace) then
      backspace(iu)
    endif
  else
    backspace(iu)
  endif

  if (tfil > zero) then
    write(iu06,*) 'file '//trim(fnam)//' forwarded until jday= ',tfil
  else
    write(iu06,*) 'file '//trim(fnam)//' is not forwarded'
  endif

end subroutine forward

subroutine wlstat (iopt,ctim,tim,itime,wls)
  !-----------------------------------------------------------------------------
  !
  !       wlstat()
  !
  !       writes sea level at tabulated stations to two output files
  !
  !       iopt = 0 : read station list
  !                  open output files
  !
  !       iopt = 1 : write data
  !
  !       iopt = 2 : close output files
  !
  !-----------------------------------------------------------------------------

  !- modules -------------------------------------------------------------------
  use dmi_mpi_global, only : iu06
  use cmod_params, only    : mmx, nmx, me1, narea
  use cmod_arrays, only    : z
  use io_subs,     only    : io_new_unit, flush_unit

  !- directives ----------------------------------------------------------------
  implicit  none

  !- arguments -----------------------------------------------------------------
  integer(4),    intent(in)            :: iopt
  character(19), intent(in)            :: ctim  !<==  watch out here!
  real(8),       intent(in)            :: tim
  integer(4),    intent(in),  optional :: itime
  logical,       intent(out), optional :: wls

  !- local vars ----------------------------------------------------------------
  character(32)              :: file1, file2
  character(80)              :: line
  character(23), save        :: fmt1 = '(1X,a19,f15.8,i6,   i6)'
  character(23), save        :: fmt2 = '(1X,a19,f15.8,i6,   i6)'
  logical                    :: notfound, lfile
  logical, save              :: wlout = .true.
  integer(4)                 :: lun, n, ie, ip, ios, igrd, mx, nx
  integer(4), save           :: lun1, lun2, lun3, lun4, nwl, ii(2)
  real(8), parameter         :: iscale = 1000.0_8
  integer(4), allocatable    :: iwl(:), jwl(:)
  integer(4), allocatable, save :: igrid(:), ifile(:), ieta(:), ipeg(:), indx(:)


  !---- we have had an error, station list undefined ---------------------------
  if (.not.wlout) then
    if (present(wls)) wls = .false.
    return
  endif
  if (iopt < 0 .or. iopt > 2) then
    wlout = .false.
    if (present(wls)) wls = .false.
    return
  endif

  !---- read station list (pos,grid,outfile) at first call only ----------------
  if (iopt == 0) then
    lun=io_new_unit()
    open(lun,file='wlstat.dat',status='old',iostat=ios)
    if (ios /= 0) then
      write(iu06,*) 'No wl station list file: wlstat.dat'
      wlout = .false.
      if (present(wls)) wls = wlout
      return
    endif
    if (present(wls)) wls = .true.

    notfound = .true.
    do while (notfound)
      read(lun,'(a80)',iostat=ios) line
      if (ios /= 0) then
        write(iu06,*) 'Could not read wl station list file: wlstat.dat '
        wlout = .false.
        if (present(wls)) wls = wlout
        return
      endif
      notfound = (line(2:8) /= 'WLFILE1') 
    enddo
    read(line(11:42),'(a)') file1

    notfound = .true.
    do while (notfound)
      read(lun,'(a80)',iostat=ios) line
      if (ios /= 0) then
        write(iu06,*) 'Could not read wl station list file: wlstat.dat '
        wlout = .false.
        if (present(wls)) wls = wlout
        return
      endif
      notfound = (line(2:8) /= 'WLFILE2') 
    enddo
    read(line(11:42),'(a)') file2

    notfound = .true.
    do while (notfound)
      read(lun,'(a80)',iostat=ios) line
      if (ios /= 0) then
        write(iu06,*) 'Could not read wl station list file: wlstat.dat '
        wlout = .false.
        if (present(wls)) wls = wlout
        return
      endif
      notfound = (line(2:15) /= 'NO_OF_STATIONS')
    enddo
    read(line(17:20),'(i4)') nwl

    allocate(iwl(nwl),jwl(nwl),igrid(nwl),ifile(nwl),indx(nwl))

    notfound = .true.
    do while (notfound)
      read(lun,'(a80)',iostat=ios) line
      if (ios /= 0) then
        write(iu06,*) 'Could not read wl station list file: wlstat.dat '
        wlout = .false.
        if (present(wls)) wls = wlout
        return
      endif
      notfound = (line(2:5) /= '====')
    enddo

    do n=1,nwl
      read(lun,'(33x,4i5)',iostat=ios) iwl(n),jwl(n),igrid(n),ifile(n)
      if (ios /= 0) then
        write(iu06,*) 'Could not read wl station list file: wlstat.dat '
        wlout = .false.
        if (present(wls)) wls = wlout
        return
      endif
    enddo
    close(lun)

    write(iu06,*) 'station list file read: wlstat.dat'

    !---- assign and check grid indices ----------------------------------------
    indx(:)  = 0    ! initialise as land
    do n=1,nwl
      igrd = igrid(n)

      if (igrd < 1 .or. igrd > narea) then
        write(iu06,'(a,6i5)') 'out of bounds: wl point no', n, igrd
        cycle
      endif

      mx = mmx(igrd)
      nx = nmx(igrd)
      if (iwl(n) < 1 .or. iwl(n) > mx .or. jwl(n) < 1 .or. jwl(n) > nx) then
        write(iu06,'(a,6i5)') 'out of bounds: wl point no',                    &
                           n, iwl(n), mx, jwl(n), nx, igrd
      elseif (ifile(n) < 1 .or. ifile(n) > 2) then
        write(iu06,'(a,3i5)') 'out of bounds: wl point no', n, ifile(n), igrd
      else
        indx(n) = me1(igrd)%p(iwl(n),jwl(n))
      endif
    enddo
    deallocate(iwl,jwl)

    !---- count the number of stations for each file ---------------------------
    ii(:) = 0
    do n=1,nwl
      ii(ifile(n)) = ii(ifile(n)) + 1
    enddo
    if (ii(1) > 999 .or. ii(2) > 999) then
      write(iu06,*) 'Not designed for more than 999 wl stations in each file'
      wlout = .false.
      if (present(wls)) wls = wlout
      return
    endif

    !---- allocate two working arrays for wl output ----------------------------
    if (ii(1) > 0) then
      allocate(ieta(ii(1)))
      ieta(:) = 0

      !---- sea level files:
      inquire(file=trim(file1),exist=lfile)
      if (.not. lfile) then
        write(iu06,'(a)') 'Warning: Cannot find '//trim(file1)
        write(iu06,'(a)') '         ... creating a new '//trim(file1)//' file'
      endif
      lun1 = io_new_unit()
      open(lun1,file=trim(file1),iostat=ios)
      if (ios /= 0) then
        write(iu06,*) 'Could not open wl output file '//trim(file1)
        wlout = .false.
        if (present(wls)) wls = wlout
        return
      endif
      write(iu06,*) 'opened for output: ',trim(file1)

      lun2 = io_new_unit()
      open(lun2,file=trim(file1)//'.forecast',status='replace',iostat=ios)
      if (ios /= 0) then
        write(iu06,*) 'Could not open wl output file '//trim(file1)//'.forecast'
        wlout = .false.
        if (present(wls)) wls = wlout
        return
      endif
      write(iu06,*) 'opened for output: '//trim(file1)//'.forecast'

      !---- format statements for printout:
      write(fmt1(18:20),'(i3.3)') ii(1)

      !---- step forward in eternal wl file:
      if (lfile) call forward(lun1,tim,.false.)
    endif 

    if (ii(2) > 0) then
      allocate(ipeg(ii(2)))
      ipeg(:) = 0

      !---- sea level file:
      inquire(file=trim(file2),exist=lfile)
      if (.not. lfile) then
        write(iu06,'(a)') 'Warning: Cannot find '//trim(file2)
        write(iu06,'(a)') '         ... creating a new '//trim(file2)//' file'
      endif
      lun3 = io_new_unit()
      open(lun3,file=trim(file2),iostat=ios)
      if (ios /= 0) then
        write(iu06,*) 'Could not open wl output file '//trim(file2)
        wlout = .false.
        if (present(wls)) wls = wlout
        return
      endif
      write(iu06,*) 'opened for output: '//trim(file2)

      lun4 = io_new_unit()
      open(lun4,file=trim(file2)//'.forecast',status='replace',iostat=ios)
      if (ios /= 0) then
        write(iu06,*) 'Could not open wl output file '//trim(file2)//'.forecast'
        wlout = .false.
        if (present(wls)) wls = wlout
        return
      endif
      write(iu06,*) 'opened for output: '//trim(file2)//'.forecast'

      !---- format statements for printout:
      write(fmt2(18:20),'(i3.3)') ii(2)

      !---- step forward in eternal wl file:
      if (lfile) call forward(lun3,tim,.false.)
    endif

  else if (iopt == 1) then
    !---- write sea level data -------------------------------------------------
    ie = 0
    ip = 0
    do n=1,nwl
      igrd = igrid(n)
      if (ifile(n) == 1) then
        ie = ie + 1
        ieta(ie) = nint(iscale*z(igrd)%p(indx(n)))
      elseif (ifile(n) == 2) then
        ip = ip + 1
        ipeg(ip) = nint(iscale*z(igrd)%p(indx(n)))
      endif
    enddo

    if (ii(1) > 0) then
      write(lun1,fmt1) ctim, tim, itime, ieta
      write(lun2,fmt1) ctim, tim, itime, ieta
      call flush_unit(lun1)
      call flush_unit(lun2)
    endif
    if (ii(2) > 0) then
      write(lun3,fmt2) ctim, tim, itime, ipeg
      write(lun4,fmt2) ctim, tim, itime, ipeg
      call flush_unit(lun3)
      call flush_unit(lun4)
    endif

  else if (iopt == 2) then
    !---- close output files ---------------------------------------------------
    if (ii(1) > 0) then
      close(lun1)
      close(lun2)
    endif
    if (ii(2) > 0) then
      close(lun3)
      close(lun4)
    endif
  endif

  !- That's all folks! ---------------------------------------------------------

end subroutine wlstat


end module wlstat_wrap 
